// ------------------------------------------------------------
function areWeightsIdentical(weightsa, weightsb, tol) {
	//
	if (weightsa.length != weightsb.length) {
		return false;
	}
	
	//
	for (var i = 0; i < weightsa.length; ++i) {
		if (Math.abs(weightsa[i] - weightsb[i]) > tol) {
			return false;
		}
	}
	
	//
	return true;
}

// ------------------------------------------------------------
QUnit.module('Assets allocation module', {
before: function() {
		// Generate a random 2*2 covariance matrix 
		this.randomCovarianceMatrix = function() {
			// Generate a random 2*2 covariance matrix
			var var1 = 1 - Math.random(); // belongs to ]0,1]
			var var2 = 1 - Math.random(); // belongs to ]0,1]
			var cov = 2*Math.random() - 1; // belongs to ]-1,1[
			while (var1*var2 - cov*cov <= 0) { // The covariance matrix must be DP, so that the covariance cannot be any value: determinant must be strictly positive
				cov = 2*Math.random() - 1;
			}
			
			return [[var1, cov], [cov, var2]];
		};
	}
});


QUnit.test('Equal weights portfolio', function(assert) {    
	// Random data, n assets case
	{
		// Generate a random number of assets
		var nbAssets = Math.floor(Math.random()*(50-1+1) + 1); // max 50 min 1
		
		// Compute EW weights
		var weights = PortfolioAllocation.equalWeights(nbAssets);

		// Check the number of output weights
		assert.equal(weights.length, nbAssets, 'EW - Number of weights');
		
		// Compare EW weights to exact closed-form formula
		var expectedWeights = [];
		for (var i = 0; i < nbAssets; ++i) {
			expectedWeights[i] = 1/nbAssets;
		}
		for (var i = 0; i < weights.length; ++i) {
			assert.equal(weights[i], expectedWeights[i], 'EW - Values ' + i);
		}
	}  
});


QUnit.test('Post optimization weights utilities', function(assert) {    
	// Test roundlotting with static data from https://pyportfolioopt.readthedocs.io/en/latest/Postprocessing.html
	{
		// The symbols
		var symbols = { 'GOOG': {weight: 0.01269, price: 1019.969971},
						'AAPL': {weight: 0.09202, price: 172.440002},
						'FB'  : {weight: 0.19856, price: 166.320007},
						'BABA': {weight: 0.09642, price: 175.360001},
						'AMZN': {weight: 0.07158, price: 1427.050049},
						'GE'  : {weight: 0.0, price: 12.97},
						'AMD' : {weight: 0.0, price: 9.82},
						'WMT' : {weight: 0.0, price: 85.910004},
						'BAC' : {weight: 0.0, price: 29.9},
						'GM'  : {weight: 0.0, price: 39.0},
						'T'   : {weight: 0.0, price: 35.25},
						'UAA' : {weight: 0.0, price: 16.74},
						'SHLD': {weight: 0.0, price: 3.3},
						'XOM' : {weight: 0.0, price: 77.43},
						'RRC' : {weight: 0.0, price: 14.99},
						'BBY' : {weight: 0.06129, price: 70.910004},
						'MA'  : {weight: 0.24562, price: 172.360001},
						'PFE' : {weight: 0.18413, price: 35.790001},
						'JPM' : {weight: 0.0, price: 110.620003},
						'SBUX': {weight: 0.03769, price: 59.419998},
					  }
		
		// The parameters to use the roundlotting method
		var portfolioValue = 10000;
		var assetsPrices = [];
		var originalWeights = [];
		for (var key in symbols) {
			if (symbols.hasOwnProperty(key)) {           
				var symbol = symbols[key];
				
				originalWeights.push(symbol.weight);
				assetsPrices.push(symbol.price);
			}
		}
		
		// Test missing parameters
		assert.throws(function() { 
			PortfolioAllocation.postOptimizationWeights(originalWeights) },
			new Error('missing portfolio value'),
			"Post optimization weights, roundlotting missing portfolio value");

		assert.throws(function() { 
			PortfolioAllocation.postOptimizationWeights(originalWeights, {portfolioValue: portfolioValue}) },
			new Error('missing assets prices'),
			"Post optimization weights, roundlotting missing assets prices");

		// Test the roundlotting method:
		// - Output must be fractional weights, corresponding number of lots, cash position
		var weights = PortfolioAllocation.postOptimizationWeights(originalWeights, {//roundingOptimizationMethodParams: {nRounds: 10},
																					portfolioValue: portfolioValue, 
																					assetsPrices: assetsPrices});
				
        var weightsMatchLots = true;
		var weightsSumToOne = true;
		var valAssets = 0;
		var sumWeights = 0;
		for (var i = 0; i < assetsPrices.length; ++i) {
			var weight = weights[1][i];
			var lot = weights[0][i];
			
			// Computation of the sum of weights
			sumWeights += weight;
			
			// The returned fractional weights must match with the returned number of lots
			var assetValue = lot * assetsPrices[i];
			if (Math.abs(assetValue/portfolioValue - weight) > 1e-8) {
				weightsMatchLots = false;
				break;
			}
			
			// Computation of the sum of the assets value
			valAssets += assetValue;
		}
		sumWeights += weights[2]/portfolioValue; // add cash position to the sum of weights
		
        assert.equal(weightsMatchLots, true, 'Post optimization weights, roundlotting - Fractional weights matching quantity of lots');
		assert.equal(Math.abs(valAssets + weights[2] - portfolioValue) <= 1e-8, true, 'Post optimization weights, roundlotting - Value of assets plus cash equal portfolio value');
		assert.equal(Math.abs(sumWeights - 1) <= 1e-12, true, 'Post optimization weights, roundlotting - Sum of portfolio weights plus cash position equal one');
		assert.equal(weights[2] <= 85, true, 'Post optimization weights, roundlotting - Value of remaining cash'); // Note: might fail due to the stochastic nature of the TA algorithm used internally
				
		var sumSupposedNullWeights = 0;
		var i = 0;
		for (var key in symbols) {
			if (symbols.hasOwnProperty(key)) {           
				var symbol = symbols[key];
				var originalWeight = symbol.weight;
				
				if (originalWeight == 0 && weights[1][i] != 0) {
					assert.equal(false, true, 'Post optimization weights, roundlotting - Null original weights must stay null');
					break;
				}
			}
			++i;
		}
		
	}
	
	// Test roundlotting with static data from Research Insight, Managing Odd Lot Trades with the Barra Optimizer, September 2013
	{
		// Due to the stochastic nature of the TA algorithm used internally, 
		// three different solutions have been listed, with two of them included in the reference.
		var expectedPossibleWeights = [ [[15, 239, 146], [0.0375, 0.5975, 0.365], 0],
		                                [[15, 240, 145], [0.0375, 0.6, 0.3625], 0],
                                        [[14, 240, 146], [0.035, 0.6, 0.365], 0]
									   ];

		var weights = PortfolioAllocation.postOptimizationWeights([0.0369, 0.5990, 0.3641], {portfolioValue: 4000000, assetsPrices: [100, 100, 100], sizeLots: [100, 100, 100]});
		
		var weightsOK = true;
		for (var i = 0; i < 3; ++i) {
			if (weights[0][i] != expectedPossibleWeights[0][0][i] && 
			    weights[0][i] != expectedPossibleWeights[1][0][i] && 
				weights[0][i] != expectedPossibleWeights[2][0][i]) {
				weightsOK = false;
				break;
			}
		}
		assert.equal(weightsOK, true, 'Post optimization weights, roundlotting - Weights matching expected weights');
		assert.equal(weights[2] == 0, true, 'Post optimization weights, roundlotting - No cash remaining');
	}

	// Test roundlotting with static data from Accuracy and Rounding In Portfolio Construction, Andreas Steiner Consulting GmbH, April 2013
	{
		var weights = PortfolioAllocation.postOptimizationWeights([1/6, 1/6, 1/6, 1/6, 1/6, 1/6], {portfolioValue: 1940, assetsPrices: [100, 20, 12, 250, 50, 1]});
		// Nothing can really be tested due to the stochastic nature of the TA algorithm used internally,
		// apart from the cash position which must be very limited.
		assert.equal(weights[2] <= 22, true, 'Post optimization weights, roundlotting - Minimal cash remaining');
		
		// Add a test with fractional roundlotting
		var weights = PortfolioAllocation.postOptimizationWeights([1/6, 1/6, 1/6, 1/6, 1/6, 1/6], {portfolioValue: 1940, 
	                                                                                               assetsPrices: [100, 20, 12, 250, 50, 1],
																								   sizeLots: [0.001, 0.001, 0.001, 0.001, 0.001, 0.001]});
		
		// Nothing can really be tested due to the stochastic nature of the TA algorithm used internally,
		// apart from the cash position which must be very limited.
		assert.equal(weights[2] <= 0.005, true, 'Post optimization weights, roundlotting - Minimal cash remaining with fractional roundlotting');	
	}
	
	// Test roundlotting with static data
	{
		var weights = PortfolioAllocation.postOptimizationWeights([0.4,0.2,0.2,0.16,0.04,0], {portfolioValue: 12032.57, 
	                                                                                          assetsPrices: [25.28, 32.55, 46.42, 31.08,28.19,1]});
		var expectedWeights = [190, 74, 52, 62, 17, 0];
		assert.deepEqual(weights[0], expectedWeights, 'Post optimization weights, roundlotting - Weights exact matching expected weights');
	}
	
	// Test with weights summing to < 1
	//
	// This case checks that there is an automated process in place to deduce the cash weight, because with a default target
	// cash weight of zero, the expected quantities would be [3,1] + 200 cash.
	{
		var weights = PortfolioAllocation.postOptimizationWeights([0.2,0.5], {portfolioValue: 1000, assetsPrices: [100, 500]});
		var expectedWeights = [2, 1];
		var expectedCash = 300;
		
		assert.deepEqual(weights[0], expectedWeights, 'Post optimization weights, roundlotting - Sum weights < 1, 1');
		assert.deepEqual(weights[2], expectedCash, 'Post optimization weights, roundlotting - Sum weights < 1, 2');
	}
	
	// Test with weights summing to > 1
	//
	// This case checks that there is no error raised, because if the original weights were left untouched, an error 
	// with "initial cash is negative" would be raised.
	{
		var weights = PortfolioAllocation.postOptimizationWeights([0.3,0.8], {portfolioValue: 1000, assetsPrices: [10, 50]});
		var expectedWeights = [25, 15];
		var expectedCash = 0;
		assert.deepEqual(weights[0], expectedWeights, 'Post optimization weights, roundlotting - Sum weights > 1, 1');
		assert.deepEqual(weights[2], expectedCash, 'Post optimization weights, roundlotting - Sum weights > 1, 2');
	}
	
	// Test with weights summing to 0
	//
	// This case checks that no fancy business is done to forcefully have the portfolio invested
		var weights = PortfolioAllocation.postOptimizationWeights([0,0], {portfolioValue: 1000, assetsPrices: [10, 50]});
		var expectedWeights = [0, 0];
		var expectedCash = 1000;
		assert.deepEqual(weights[0], expectedWeights, 'Post optimization weights, roundlotting - Sum weights = 0, 1');
		assert.deepEqual(weights[2], expectedCash, 'Post optimization weights, roundlotting - Sum weights = 0, 2');	
});


QUnit.test('Inverse volatility portfolio', function(assert) {     
	// Reference: Carvalho, Raul Leote de and Xiao, Lu and Moulin, Pierre, Demystifying Equity Risk-Based Strategies: A Simple Alpha Plus Beta Description

	// Random data, n assets case
	{
		// Generate a random number of assets
		var nbAssets = Math.floor(Math.random()*(50-1+1) + 1); // max 50 min 1
		
		// Generate n random variances
		var sigma = new Array(nbAssets);
		for (var i = 0; i < nbAssets; ++i) {
			sigma[i] = 1 - Math.random(); // belongs to ]0,1];
		}
		
		// Compute ERB weights
		var weights = PortfolioAllocation.inverseVolatilityWeights(sigma);
		
		// Check the number of output weights
		assert.equal(weights.length, nbAssets, 'ERB - Number of weights');
		
		// Compare ERB weights to exact closed-form formula, c.f. formula 3 of the reference
		var denom = 0;
		for (var i = 0; i < nbAssets; ++i) { 
			denom += 1/Math.sqrt(sigma[i]); 
		}
		var expectedWeights = [];
		for (var i = 0; i < nbAssets; ++i) {
			expectedWeights[i] = 1/Math.sqrt(sigma[i]) / denom;
		}
		for (var i = 0; i < nbAssets; ++i) {
			assert.equal(Math.abs(weights[i] - expectedWeights[i]) <= 1e-8, true, 'ERB - Values ' + i);  
		}
	}
});


QUnit.test('Equal risk contributions portfolio', function(assert) {    
	// Reference: Maillard, S., Roncalli, T., Teiletche, J.: The properties of equally weighted risk contribution portfolios. J. Portf. Manag. 36, 60–70 (2010)

	// Example 1 p. 65 of reference with constant correlation equals to 0
	{
		// Test default parameters
		var weights = PortfolioAllocation.equalRiskContributionWeights([[0.01, 0, 0, 0], [0, 0.04, 0, 0], [0, 0, 0.09, 0], [0, 0, 0, 0.16]]); 
		var expectedWeights = [48, 24, 16, 12];
		for (var i = 0; i < expectedWeights.length; ++i) {
			assert.equal(Math.round(100 * weights[i]), expectedWeights[i], 'ERC - Values #1 ' + i);
		}
		
		// Test a number of cycles equal to 0, which must output
		// the initial internal weights, corresponding to an equally weighted portfolio
		var weights = PortfolioAllocation.equalRiskContributionWeights([[0.01, 0, 0, 0], [0, 0.04, 0, 0], [0, 0, 0.09, 0], [0, 0, 0, 0.16]], {nbCycles: 0}); 
		var expectedWeights = [25, 25, 25, 25];
		for (var i = 0; i < expectedWeights.length; ++i) {
			assert.equal(Math.round(100 * weights[i]), expectedWeights[i], 'ERC - Values #1, no cycle ' + i);
		}
		
		// Test a number of cycles equal to 1
		//
		// Note: weights taken from running the algorithm once
		var weights = PortfolioAllocation.equalRiskContributionWeights([[0.01, 0, 0, 0], [0, 0.04, 0, 0], [0, 0, 0.09, 0], [0, 0, 0, 0.16]], {nbCycles: 1}); 
		var expectedWeights = [0.38407763703774006, 0.24824686014490746, 0.1979048506362147, 0.16977065218113782];
		for (var i = 0; i < expectedWeights.length; ++i) {
			assert.equal(Math.round(100 * weights[i]), Math.round(100 * expectedWeights[i]), 'ERC - Values #1, one cycle ' + i);
		}
		
		// Test a number of cycles equal to 10, which must converge to the
		// same solution as the default parameters
		var weights = PortfolioAllocation.equalRiskContributionWeights([[0.01, 0, 0, 0], [0, 0.04, 0, 0], [0, 0, 0.09, 0], [0, 0, 0, 0.16]], {nbCycles: 10}); 
		var expectedWeights = [48, 24, 16, 12];
		for (var i = 0; i < expectedWeights.length; ++i) {
			assert.equal(Math.round(100 * weights[i]), expectedWeights[i], 'ERC - Values #1, ten cycles ' + i);
		}
		
		// Test a number of cycles equal to 100, which must converge to the
		// same solution as the default parameters
		var weights = PortfolioAllocation.equalRiskContributionWeights([[0.01, 0, 0, 0], [0, 0.04, 0, 0], [0, 0, 0.09, 0], [0, 0, 0, 0.16]], {nbCycles: 100}); 
		var expectedWeights = [48, 24, 16, 12];
		for (var i = 0; i < expectedWeights.length; ++i) {
			assert.equal(Math.round(100 * weights[i]), expectedWeights[i], 'ERC - Values #1, hundred cycles ' + i);
		}
	}


	// Example 2 p. 65 of reference (note: covariance matrix is provided here v.s. correlation matrix in the reference => matrix operations were needed)
	{
		var weights = PortfolioAllocation.equalRiskContributionWeights([[0.01, 0.016, 0, 0], [0.016, 0.04, 0, 0], [0, 0, 0.09, -0.06], [0, 0, -0.06, 0.16]]); 
		var expectedWeights =  [38.4, 19.2, 24.3, 18.2];
		for (var i = 0; i < expectedWeights.length; ++i) {
			assert.equal(Math.round(1000 * weights[i])/10, expectedWeights[i], 'ERC - Values #2 ' + i);
		}
	}

	// Random data, 2 assets case, c.f. p 62 of reference
	{
		// Generate a random 2*2 covariance matrix
		var covMat = this.randomCovarianceMatrix();
		
		// Compute ERC weights
		var weights = PortfolioAllocation.equalRiskContributionWeights(covMat);
		
		// Compare ERC weights to exact closed-form formula, c.f. p 62 of reference
		var var1 = covMat[0][0];
		var var2 = covMat[1][1];
		var std1 = Math.sqrt(var1);
		var std2 = Math.sqrt(var2);
		var expectedWeights =  [(1/std1) / (1/std1 + 1/std2), (1/std2) / (1/std1 + 1/std2)];
		for (var i = 0; i < expectedWeights.length; ++i) {
			assert.equal(Math.abs(weights[i] - expectedWeights[i]) <= 1e-8, true, 'ERC - Values #3 ' + i);  
		}
	}
	
	// Reference: Xi Bai, Katya Scheinberg, Reha Tutuncu, Least-squares approach to risk parity in portfolio selection
	//
	// To be noted that this test also test the option of providing the portfolio volatility in output.
	{
		var covMat = [[94.868,33.750,12.325,-1.178,8.778],
					[33.750,445.642,98.955,-7.901,84.954],
					[12.325,98.955,117.265,0.503,45.184],
					[-1.178,-7.901,0.503,5.460,1.057],
					[8.778,84.954,45.184,1.057,34.126]];
		
		var weights = PortfolioAllocation.equalRiskContributionWeights(covMat, {outputPortfolioVolatility: true}); 
		var expectedWeights = [0.125, 0.047, 0.083, 0.613, 0.132];
		var expectedVolatility = 3.0406150258182514;
		for (var i = 0; i < expectedWeights.length; ++i) {
			assert.equal(Math.abs(weights[0][i] - expectedWeights[i]) <= 1e-3, true, 'ERC - Values #4 ' + i);
		}
		assert.equal(Math.abs(weights[1] - expectedVolatility) <= 1e-3, true, 'ERC - Values #4, volatility');
	}
	
	// Reference: https://github.com/lequant40/portfolio_allocation_js/issues/3
	//
	// This is a test of the ERC method wit a semi-positive definite covariance matrix
	//
	// The ERC method converges in this case to a MV portfolio, and risk contributions
	// cannot be equal (there is no risk anymore).
	{
		var cov = PortfolioAllocation.covarianceMatrix([[-0.001697998787, 0.001427530069, -0.0005054947277, 0.0001213800916, -0.0001618204804],
                                                   [-0.002961208738, 0.001640864801, -0.0001441519189, 0.0004640260466, 0.000206138145],
                                                   [-0.001196934304, 0.0002291811055, -0.0004775812854, 0.001433428435, 0.0009680408618],
                                                   [0.0007749231464, -0.0002552713534, 0.0003744935825, 0.00247583719, 0.002435774483],
                                                   [-0.01361067487, -0.009173715734, -0.01167442776, -0.002090384233, -0.01495151011],
                                                   [-0.02573117139, 0.002675293971, 0.003048918071, -0.01685867172, -0.008688999321],
                                                   [-0.007799506409, 0.001570415597, -0.008377497996, -0.002298475244, -0.02687699473],
                                                   [-0.009589945124, -0.02074629016, 0.001121818449, -0.003655601888, 0.01279545706],
                                                   [-0.0001174153557, -0.0252073843, 0.008193945345, -0.006319268471, 0.002470726766],
                                                   [0.01103327496, 0.003395115191, -0.003901529538, 0.002079722704, -0.005188516084]], {method: "sample-covariance"});
		
		var weights = PortfolioAllocation.equalRiskContributionWeights(cov, {outputPortfolioVolatility: true}); 
		
		// Check the expected weights and the expected volatility
		var expectedWeights = [0.27653353547900505, 0.08992490010322292, 0.07660067915254527, 0.10281081303356884, 0.00627339817004538, 0.07548939567866735, 0.004543021673884871, 0.0858506918300374, 0.02204534860588049, 0.25992821627314233];
		var expectedVolatility = 0;
		for (var i = 0; i < expectedWeights.length; ++i) {
			assert.equal(Math.abs(weights[0][i] - expectedWeights[i]) <= 1e-4, true, 'ERC, semi-positive definite covariance matrix - Weight #' + i);
		}
		assert.equal(Math.abs(weights[1] - expectedVolatility) <= 1e-8, true, 'ERC, semi-positive definite covariance matrix - Volatility');
	}
	
	// Reference: Table 9 from the paper Constrained Risk Budgeting Portfolios https://arxiv.org/abs/1902.05710
	//
	// Test different coordinates sampling methods
	{
		/* The covariance matrix below has been obtained through Diag(STD)*CORR*Diag(STD)
		var stdDev = [0.05,0.05,0.07,0.1,0.15,0.15,0.15,0.18];
		var corrMat = [[100,  80,  60, -20, -10, -20, -20, -20],
						[ 80, 100,  40, -20, -20, -10, -20, -20],
						[ 60,  40, 100,  50,  30,  20,  20,  30],
						[-20, -20,  50, 100,  60,  60,  50,  60],
						[-10, -20,  30,  60, 100,  90,  70,  70],
						[-20, -10,  20,  60,  90, 100,  60,  70],
						[-20, -20,  20,  50,  70,  60, 100,  70],
						[-20, -20,  30,  60,  70,  70,  70, 100]]/100;
		*/
		var covMat = [[ 0.0025 ,  0.002  ,  0.0021 , -0.001  , -0.00075, -0.0015 , -0.0015 , -0.0018 ],
                      [ 0.002  ,  0.0025 ,  0.0014 , -0.001  , -0.0015 , -0.00075, -0.0015 , -0.0018 ],
                      [ 0.0021 ,  0.0014 ,  0.0049 ,  0.0035 ,  0.00315,  0.0021 , 0.0021 ,  0.00378],
                      [-0.001  , -0.001  ,  0.0035 ,  0.01   ,  0.009  ,  0.009  , 0.0075 ,  0.0108 ],
                      [-0.00075, -0.0015 ,  0.00315,  0.009  ,  0.0225 ,  0.02025, 0.01575,  0.0189 ],
                      [-0.0015 , -0.00075,  0.0021 ,  0.009  ,  0.02025,  0.0225 , 0.0135 ,  0.0189 ],
                      [-0.0015 , -0.0015 ,  0.0021 ,  0.0075 ,  0.01575,  0.0135 , 0.0225 ,  0.0189 ],
                      [-0.0018 , -0.0018 ,  0.00378,  0.0108 ,  0.0189 ,  0.0189 , 0.0189 ,  0.0324 ]];
					  
       var expectedWeights = [26.8306, 28.6769, 11.4095,  9.7985,  5.6135,  5.9029,  6.656,   5.1121];
	   
	   
	   // Check that the default coordinates sampler is 'cyclic', and that it converges
	   var defaultSamplingWeights = PortfolioAllocation.equalRiskContributionWeights(covMat);
	   var cyclicWeights = PortfolioAllocation.equalRiskContributionWeights(covMat, {coordinatesSampler: 'cyclic'});
	   assert.deepEqual(defaultSamplingWeights, cyclicWeights, 'ERC - Default coordinates sampler');
	   
	   for (var i = 0; i < expectedWeights.length; ++i) {
	      assert.equal(Math.abs(defaultSamplingWeights[i] - expectedWeights[i]/100) <= 1e-5, true, 'ERC - Default coordinates sampler weights ' + i);
	   }
	   
	   
	   // Check that the shuffled cyclic coordinates sampler converges
	   var shuffledCyclicWeights = PortfolioAllocation.equalRiskContributionWeights(covMat, {coordinatesSampler: 'shuffledCyclic'});
	   for (var i = 0; i < expectedWeights.length; ++i) {
	      assert.equal(Math.abs(defaultSamplingWeights[i] - expectedWeights[i]/100) <= 1e-5, true, 'ERC - Shuffled cyclic coordinates sampler weights ' + i);
	   }
	   
	   
	   // Check that the randomized coordinates sampler converges
	   //
	   // Note: There is no theoretical guarantees that the algorithm converges to a solution of the problem,
	   // but in practice with these data, it seems to converge.
	   var randomizedWeights = PortfolioAllocation.equalRiskContributionWeights(covMat, {coordinatesSampler: 'randomized'});
	   for (var i = 0; i < expectedWeights.length; ++i) {
	      assert.equal(Math.abs(randomizedWeights[i] - expectedWeights[i]/100) <= 1e-5, true, 'ERC - Randomized coordinates sampler weights ' + i);
	   }
	   
	   
	   // Check that the ACF coordinates samples converges
	   var acfWeights = PortfolioAllocation.equalRiskContributionWeights(covMat, {coordinatesSampler: 'acf'});
	   for (var i = 0; i < expectedWeights.length; ++i) {
	      assert.equal(Math.abs(acfWeights[i] - expectedWeights[i]/100) <= 1e-5, true, 'ERC - ACF coordinates sampler weights ' + i);
	   }
	}

	// Reference: Table 6 from the paper Constrained Risk Budgeting Portfolios https://arxiv.org/abs/1902.05710
	//
	// Test weights constraints on ERC portfolios
	{
		var covMat = [[0.0225, 0.003, 0.015, 0.0225, 0.0075],
		              [0.003, 0.04, 0.035, 0.024, 0.008],
					  [0.015, 0.035, 0.0625, 0.06, 0.00125],
					  [0.0225, 0.024, 0.06, 0.09, 0.003],
					  [0.0075, 0.008, 0.00125, 0.003, 0.01]];
		
		// Compute the min/max weights constraints
		var initWeights = [25.00, 25.00, 10.00, 10.00, 30.00];
		var minWeights = new Array(5);
		var maxWeights = new Array(5);
		for (var i = 0; i < minWeights.length; ++i) {
			minWeights[i] = initWeights[i]/100 - 0.05;
			maxWeights[i] = initWeights[i]/100 + 0.05;
		}
		
		// Compute the min/max weights-constrained ERC portfolio
		var weights = PortfolioAllocation.equalRiskContributionWeights(covMat, {maxCycles: -1, constraints: { minWeights: minWeights, maxWeights: maxWeights } } );

		//
		var expectedWeights = [22.89, 20.00, 11.69, 10.42, 35.00];
	    for (var i = 0; i < expectedWeights.length; ++i) {
	       assert.equal(Math.abs(weights[i] - expectedWeights[i]/100) <= 1e-4, true, 'ERC - Bounds constraints weights ' + i);
	    }
	}
	
	// This example shows that the default bracketing interval proposed in the paper Constrained Risk Budgeting Portfolios https://arxiv.org/abs/1902.05710
	// is not sufficient, because the corresponding interval for lambda is [0.27613738533391785,1.1045495413356714], and is not a bracketing interval (upper bound issue).
	{
		var minWeights = [0, 0.8213735428258124];
		var maxWeights = [0.17862645717418746, 1];
		var covMat = [[1, -0.38998515536757905 ],
                      [-0.38998515536757905, 1]];
	
		// Compute the min/max weights-constrained ERC portfolio
		var expectedWeights = [0.17862645717418746, 0.8213735428258124];
		var weights = PortfolioAllocation.equalRiskContributionWeights(covMat, {constraints: { minWeights: minWeights, maxWeights: maxWeights } } );

		//
		assert.equal(areWeightsIdentical(weights, expectedWeights, 0), true, 'Bounds constraints, bisection bracketing interval auto-expanding');
	}
	
	// This example shows that the default bracketing interval proposed in the paper Constrained Risk Budgeting Portfolios https://arxiv.org/abs/1902.05710
	// is not sufficient, because the corresponding interval for lambda is [0.020454667788860955,0.08181867115544382], and is not a bracketing interval (lower bound issue).
	{
		var minWeights = [0, 0, 0.5677906167311624, 0];
		var maxWeights = [0.15736356718154443, 0.19621014565577566, 0.8677906167311624, 0.17863567043151757];
		var covMat = [[1,  0.08799014129596319,  0.02081603966787144, -0.38844753961406986 ],
						[  0.08799014129596319, 1, -0.16869872718539686 , -0.6955169149559889 ],
						[  0.02081603966787144, -0.16869872718539686 , 1,  -0.5142013855358811 ],
						[ -0.38844753961406986,  -0.6955169149559889,  -0.5142013855358811, 1 ]];
	
		// Compute the min/max weights-constrained ERC portfolio
		var expectedWeights = [0.057362792952609716, 0.19621014565577566, 0.5677906167311624, 0.17863567043151757];
		var weights = PortfolioAllocation.equalRiskContributionWeights(covMat, {constraints: { minWeights: minWeights, maxWeights: maxWeights } } );
		
		//
		assert.equal(areWeightsIdentical(weights, expectedWeights, 0), true, 'Bounds constraints, bisection bracketing interval auto-expanding');
	}
	
	// This example shows that the the bracketing interval lower bound might never be associated to a value of lambda such 
	// that sum x_i(lambda) - 1 < 0 (numerically), but sum x_i -1 ~= 0, so that the optimal lambda cannot be found by bisection,
	// but still exists (and must be detected with an early stopping criterion).
	{
		var minWeights = [0, 0.835158066327829];
		var maxWeights = [0.16484193367217118, 1];
		var covMat = [[1, -0.8695857079139759],
					  [-0.8695857079139759, 1]];
	
		// Compute the min/max weights-constrained ERC portfolio
		var expectedWeights = [0.16484193367217118, 0.835158066327829];
		var weights = PortfolioAllocation.equalRiskContributionWeights(covMat, {constraints: { minWeights: minWeights, maxWeights: maxWeights } } );

		//
		assert.equal(areWeightsIdentical(weights, expectedWeights, 0), true, 'Bounds constraints, bisection algorithm not working due to numerically sum x_i constraint satisfied with wrong sign');
	}
	
	// Test for infeasible problem: minimum reachable sum x_i is 1.22302686548491195, at lambda = 1e-18
	{
		var minWeights = [0, 0.40, 0];
		var maxWeights = [1, 1, 1];
		var covMat = [[1, 0.074, -0.70],
					  [0.074, 1,  -0.69 ],
					  [-0.70, -0.69, 1 ]];
	
		// Compute the min/max weights-constrained ERC portfolio
		assert.throws(function() { 
			var weights = PortfolioAllocation.equalRiskContributionWeights(covMat, {constraints: { minWeights: minWeights, maxWeights: maxWeights } } );
		},
		new Error('internal error: maximum number of iterations reached when searching for a bracketing interval, the problem might be infeasible'),
		"Bounds constraints, infeasible sum x_i constraint");
	}
	
	
});


QUnit.test('Risk budgeting portfolio', function(assert) {	
	// Reference: Bruder, Benjamin and Roncalli, Thierry, Managing Risk Exposures Using the Risk Budgeting Approach
	
	// Random data, 2 assets case, c.f. p 6 of reference
	{
		// Generate a random 2*2 covariance matrix
		var covMat = this.randomCovarianceMatrix();
		
		// Generate random risk budgets
		var b = 1 - Math.random(); // belongs to ]0,1]

		// Compute RB weights
		var weights = PortfolioAllocation.riskBudgetingWeights(covMat, [b, 1 - b]);
		
		// Compare RB weights to exact closed-form formula, c.f. p 6 of reference
		var var1 = covMat[0][0];
		var var2 = covMat[1][1];
		var cov = covMat[0][1];
		var std1 = Math.sqrt(var1);
		var std2 = Math.sqrt(var2);
		var corr = cov/(std1*std2);
		var w_star = (b - 0.5)*corr*std1*std2 - b*var2 + std1*std2*Math.sqrt( (b - 0.5)*(b - 0.5)*corr*corr + b*(1 - b) );
		w_star /= (1 - b)*var1 - b*var2 + 2*(b - 0.5)*corr*std1*std2;	
		var expectedWeights = [w_star, 1 - w_star];
		for (var i = 0; i < expectedWeights.length; ++i) {
			assert.equal(Math.abs(weights[i] - expectedWeights[i]) <= 1e-8, true, 'RB - Values #1 ' + i);  
		}
	}
	
	
	// Random data, n assets case, 0 correlation, c.f. p 7 of reference
	// The associated matrix is always definite positive as long as all variances are strictly positive
	{
		// Generate a random number of assets
		var nbAssets = Math.floor(Math.random()*(20-2+1) + 2); // max 20 min 2
		
		// Generate a random n*n covariance matrix with 0 correlation
		var sigma = new Array(nbAssets);
		for (var i = 0; i < nbAssets; ++i) {
			sigma[i] = new Array(nbAssets);
			for (var j = 0; j < nbAssets; ++j) { 
				sigma[i][j] = 0; // 0 correlation
			}
			sigma[i][i] = 1 - Math.random(); // belongs to ]0,1];
		}
		
		// Generate n random risk budgets
		var rb = new Array(nbAssets);
		var sum_rb = 0;
		for (var i = 0; i < nbAssets; ++i) { 
			rb[i] = 1 - Math.random(); // belongs to ]0,1]; so that rb[i] > 0
			sum_rb += rb[i]; 
		}
		for (var i = 0; i < nbAssets; ++i) {
			rb[i] /= sum_rb; // normalization, so that sum(rb[i]) == 1
		}

		// Compute RB weights
		var weights = PortfolioAllocation.riskBudgetingWeights(sigma, rb, {eps: 1e-10, maxIter: 10000});
		
		// Compare RB weights to exact closed-form formula, c.f. p 7 of reference
		var denom = 0;
		for (var i = 0; i < nbAssets; ++i) { 
			denom += Math.sqrt(rb[i]) * 1/Math.sqrt(sigma[i][i]); 
		}
		for (var i = 0; i < nbAssets; ++i) {
			assert.equal(Math.abs(weights[i] - Math.sqrt(rb[i]) * 1/Math.sqrt(sigma[i][i]) / denom) <= 1e-8, true, 'RB - Values #2 ' + i);  
		}
	}
	
	// Reference: Table 1 from the paper Constrained Risk Budgeting Portfolios https://arxiv.org/abs/1902.05710
	//
	// Test weights constraints on ERC and RB portfolios
	{
		var covMat = [[0.01,0.0075,0.01,0.015],
		              [0.0075,0.0225,0.015,0.0225],
					  [ 0.01,0.015,0.04,0.045],
					  [0.015,0.0225,0.045,0.09]];
					  
		// Compute ERC weights with no bounds constraints
		{
			var expectedWeights = [41.01, 27.34, 18.99, 12.66];
			
			var rb = [0.25, 0.25, 0.25, 0.25];
			var weights = PortfolioAllocation.riskBudgetingWeights(covMat, rb);

			for (var i = 0; i < expectedWeights.length; ++i) {
				assert.equal(Math.abs(weights[i] - expectedWeights[i]/100) <= 1e-3, true, 'RB - Test #3 ERC, ' + i);
			}
		}

		// Compute ERC weights with bounds constraints x_i <= 30%, i=1..4
		{
			var expectedWeights = [30, 30, 24.00, 16.00]; // the weights in the reference paper are [30, 30, 24.57, 15.43]
			
			var rb = [0.25, 0.25, 0.25, 0.25];
			var weights = PortfolioAllocation.riskBudgetingWeights(covMat, rb, {constraints: { maxWeights: [0.3, 0.3, 0.3, 0.3] } } );
			
			for (var i = 0; i < expectedWeights.length; ++i) {
				assert.equal(Math.abs(weights[i] - expectedWeights[i]/100) <= 1e-3, true, 'RB - Test #3 ERC with bounds constraints, ' + i);
			}
		}
		
		// Compute RB weights with no bounds constraints
		{
			var expectedWeights = [45.05, 30.04, 14.67, 10.24];
			
			var rb = [0.3, 0.3, 0.195, 0.205];
			var weights = PortfolioAllocation.riskBudgetingWeights(covMat, rb);
			
			for (var i = 0; i < expectedWeights.length; ++i) {
				assert.equal(Math.abs(weights[i] - expectedWeights[i]/100) <= 1e-3, true, 'RB - Test #3 RB, ' + i);
			}
		}
		
		// Compute RB weights with bounds constraints x_i <= 30%, i=1..4
		{
			var expectedWeights = [30, 30, 23.56, 16.43]; // the weights in the reference are [30, 30, 24.43, 15.57]
			
			var rb = [0.3, 0.3, 0.195, 0.205];
			var weights = PortfolioAllocation.riskBudgetingWeights(covMat, rb, {constraints: { maxWeights: [0.3, 0.3, 0.3, 0.3] } } );

			for (var i = 0; i < expectedWeights.length; ++i) {
				assert.equal(Math.abs(weights[i] - expectedWeights[i]/100) <= 1e-3, true, 'RB - Test #3 RB with bounds constraints, ' + i);
			}
		}
	}
});


QUnit.test('Most diversified portfolio', function(assert) {    
	// Reference: Choueifaty, Yves and Froidure, Tristan and Reynier, Julien, Properties of the Most Diversified Portfolio (July 6, 2011). Journal of Investment Strategies, Vol.2(2), Spring 2013, pp.49-70.

	// Simple 2 assets case, p. 9 of the reference
	// Note: covariance matrix is provided here v.s. correlation matrix in the reference => matrix operations were needed
	{
		// Covariance matrix
		var sigma = [[0.0400, 0.0100], [0.0100, 0.0100]];
		
		// Compute MDP weights
		var weights = PortfolioAllocation.mostDiversifiedWeights(sigma);

		// Compare MDP weights to values provided p. 9 of the reference
		var expectedWeights = [0.33, 0.67];
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-2), true, 'MDP - Test #1');
	}
	
	// Duplicate 3 assets is skipped, as not a real example (asset A duplication is a polico, and thus would never be selected)
	
	// Leverage invariance 2 assets case, p. 10 of the reference
	// Note: covariance matrix is provided here v.s. correlation matrix in the reference => matrix operations were needed
	{
		// Covariance matrix
		var sigma = [[0.0025, 0.0025], [0.0025, 0.0100]];
		
		// Compute MDP weights
		var weights = PortfolioAllocation.mostDiversifiedWeights(sigma);
		
		// Compare MDP weights to values provided p. 11 of the reference
		var expectedWeights = [0.67, 0.33];
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-2), true, 'MDP - Test #2');
	}
	
	// Polico invariance 3 assets case, p. 10 of the reference
	// Note: covariance matrix is provided here v.s. correlation matrix in the reference => matrix operations were needed
	{
		// Covariance matrix
		var sigma = [[0.0400, 0.0100, 0.0225], [0.0100, 0.0100, 0.0075], [0.0225, 0.0075, 0.0131]];
		
		// Compute MDP weights
		var weights = PortfolioAllocation.mostDiversifiedWeights(sigma);
		
		// Compare MDP weights to values provided p. 11 of the reference
		var expectedWeights = [0.33, 0.67, 0];		
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-2), true, 'MDP - Test #3');
	}
	
	
	// Reference: Toward Maximum Diversification by Y. Choueifaty, Y. Coignard, The Journal of Portfolio Management, Fall 2008, Vol. 35, No. 1: pp. 40-51
	
	// Example 2 p.41 of the reference
	// Note: covariance matrix is provided here v.s. correlation matrix in the reference => matrix operations were needed
	{
		// Covariance matrix
		var sigma = [[0.0100, 0.0090, 0.0010], [0.0090, 0.0100, 0.0010], [0.0010, 0.0010, 0.0100]];
		
		// Compute MDP weights
		var weights = PortfolioAllocation.mostDiversifiedWeights(sigma);
		var weightsGsmo = PortfolioAllocation.mostDiversifiedWeights(sigma, {optimizationMethod: "gsmo"});

		// Compare MDP weights to values provided p. 11 of the reference
		var expectedWeights = [25.7, 25.7, 48.6];
		var roundedWeights = weights.map(x => Math.round(x *1000)/10);
		var roundedWeightsGsmo = weightsGsmo.map(x => Math.round(x *1000)/10);
		assert.equal(areWeightsIdentical(roundedWeights, expectedWeights, 1e-2), true, 'MDP - Test #4');
		assert.equal(areWeightsIdentical(roundedWeightsGsmo, expectedWeights, 1e-2), true, 'MDP - Test #4, GSMO');
	}
	
	
	// Static data: correlation matrix built with randcorr Matlab gallery function
	{
		var corr = [[1.0000, 0.1966, -0.2071, -0.2084, 0.2021], [0.1966, 1.0000, 0.1993, 0.2008, -0.1963], [-0.2071, 0.1993, 1.0000, -0.2037, 0.1960], [-0.2084, 0.2008, -0.2037, 1.0000, 0.1978], [0.2021, -0.1963, 0.1960, 0.1978, 1.0000]];
		var nbAssets = corr.length;
		
		{	
			// In case of a correlation matrix, maximizing the diversification ratio
			// is equivalent to minimizing the variance, c.f. formula 2 of the reference
			//
			// Variance minimization with the correlation matrix corr has been implemented
			// in Matlab, with CVX library, and output weights copy/pasted here.

			// Compute MDP weights
			var weights = PortfolioAllocation.mostDiversifiedWeights(corr, {eps: 1e-10});
			
			// Compare MDP weights to values obtained through Matlab CVX library usage
			var expectedWeights = [0.334077696834404, 0.000000171869935, 0.332780568181053, 0.333141372759862, 0.000000190356559];
			assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-6), true, 'MDP - Test #5');
		}
		
		{
			// In case of a covariance matrix, maximizing the diversification ratio
			// is equivalent to minimizing the variance using the correlation matrix,
			// plus rescaling the obtained weights by the assets volatilities and normalizing
			// them if no cash, c.f. discussion above formula 4 of the reference.
			
			// Generate random standard deviations
			var stddevs = new Array(nbAssets);
			for (var i = 0; i < nbAssets; ++i) { 
				stddevs[i] = 1 - Math.random(); // belongs to ]0,1]
			}
			
			// Generate the associated covariance matrix from the correlation matrix above
			var sigma = corr.slice();
			for (var i = 0; i < nbAssets; ++i) { 
				for (var j = 0; j < nbAssets; ++j) { 
					sigma[i][j] *= stddevs[i]*stddevs[j];
				}
			}
			
			// Compute MDP weights
			var weights = PortfolioAllocation.mostDiversifiedWeights(sigma);
			
			// Compute expected weights by rescaling + normalizing the weights obtained from the Matlab CVX library usage
			// on the correlation matrix above
			var expectedWeights = [0.334077696834404, 0.000000171869935, 0.332780568181053, 0.333141372759862, 0.000000190356559]; // correlation weights
			var sumWeights = 0.0;
			for (var i = 0; i < nbAssets; ++i) { 
				expectedWeights[i] /= stddevs[i]; // correlation rescaled weights
				sumWeights += expectedWeights[i];
			}
			for (var i = 0; i < nbAssets; ++i) { 
				expectedWeights[i] /= sumWeights; // correlation rescaled + normalized weights = final expected weights
			}
			
			// Compare MDP weights to final expected weights
			assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-4), true, 'MDP - Test #6');
		}
	}
	
	// Reference: Understanding the Impact of Weights Constraints in Portfolio Theory, Thierry Roncalli
	{
		var covMat = [[0.0225,0.0030,0.0150,0.0225], [0.0030,0.0400,0.0350,0.0240], [0.0150,0.0350,0.0625,0.0600], [0.0225,0.0240,0.0600,0.0900]];
		
		var expectedWeights =  [0.4, 0.4, 0, 0.2];
		var weights = PortfolioAllocation.mostDiversifiedWeights(covMat, { constraints: {maxWeights: [0.4, 0.4, 0.4, 0.4]}}); 
		var weightsGsmo = PortfolioAllocation.mostDiversifiedWeights(covMat, {optimizationMethod: "gsmo",  constraints: {maxWeights: [0.4, 0.4, 0.4, 0.4]}}); 
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-8), true, 'Values #7');
		assert.equal(areWeightsIdentical(weightsGsmo, expectedWeights, 1e-8), true, 'Values #7, GSMO');
	}
});



QUnit.test('Minimum correlation heuristic portfolio', function(assert) {    
	// Reference: David Varadi, Michael Kapler, Henry Bee, Corey Rittenhouse, Sep. 2012, The Minimum Correlation Algorithm: A Practical Diversification Tool

	// Example of the appendix, with MinCorr 2 example being inversed with MinCorr example...
	{
		// Data
		var sigma = [[0.019600000000000, 0.022680000000000, 0.026180000000000], [0.022680000000000, 0.032400000000000, 0.027720000000000], [0.026180000000000, 0.027720000000000, 0.048400000000000]];
		var nbAssets = sigma.length;

		// Compute MinCorr weights
		var weights = PortfolioAllocation.minimumCorrelationWeights(sigma);
		
		// Compare MinCorr weights to expected weights
		var expectedWeights = [0.21, 0.31, 0.48];
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-2), true, 'MinCorr - Values #1');
	}

	// Limit case: 2 assets, must be equal to IV
	{
		// Generate a random 2*2 covariance matrix
		var covMat = this.randomCovarianceMatrix();
		
		// Compute MinCorr weights
		var weights = PortfolioAllocation.minimumCorrelationWeights(covMat);
		
		// Compute IV weights
		var var1 = covMat[0][0];
		var var2 = covMat[1][1];
		var expectedWeights = PortfolioAllocation.inverseVolatilityWeights([var1, var2]);
		
		// Compare the weights
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-8), true, 'MinCorr - 2 assets');
	}  
});


QUnit.test('Proportional minimum variance heuristic portfolio', function(assert) {    
	// Reference: Minimum Variance Algorithm (MVA) Excel Sheet

	// Example of the Excel sheet
	{
		// Data
		var cov = [[0.000090, 0.000044, 0.000028, 0.000034],[0.000044, 0.000084, 0.000068, 0.000039],[0.000028, 0.000068, 0.000101, 0.000036],[0.000034, 0.000039, 0.000036, 0.000039]];
		var nbAssets = cov.length;

		// Compute MVA weights
		var weights = PortfolioAllocation.proportionalMinimumVarianceWeights(cov);
		
		// Compare MVA weights to expected weights
		var expectedWeights = [0.18, 0.07, 0.07, 0.68];
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-2), true, 'MinVar - Values #1');
	}
});



QUnit.test('Cluster risk parity portfolio', function(assert) {    
	// Error cases
	{
		// Dummy covariance matrix
		var sigma = [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]];

		// Unsupported clustering method
		assert.throws(function() { 
			PortfolioAllocation.clusterRiskParityWeights(sigma, { clusteringMode: 'none' }) },
			new Error('unsupported clustering method'),
			"CRP - Manual clustering checks unsupported clustering method");

		// In case manual clusters are provided for n assets, they must form a partition of the set [1..n]
		// Test no clusters
		assert.throws(function() { 
			PortfolioAllocation.clusterRiskParityWeights(sigma, { clusteringMode: 'manual' }) }, 
			new Error('missing asset index: 1'),
			"CRP - Manual clustering checks no clusters");
		
		// Test the presence of an empty cluster
		assert.throws(function() { 
			PortfolioAllocation.clusterRiskParityWeights(sigma, { clusteringMode: 'manual', clusters: [[1, 2, 3, 4], []] }) }, 
			new Error('empty cluster at index: 1'),
			"CRP - Manual clustering checks empty cluster");
		
		// Test asset index out of bounds #1
		assert.throws(function() { 
			PortfolioAllocation.clusterRiskParityWeights(sigma, { clusteringMode: 'manual', clusters: [[0, 1, 2, 3]] }) }, 
			new Error('asset index out of bounds: 0'),
			"CRP - Manual clustering checks out of bounds #1");
		
		// Test asset index out of bounds #2
		assert.throws(function() { 
			PortfolioAllocation.clusterRiskParityWeights(sigma, { clusteringMode: 'manual', clusters: [[1, 2, 3, 5]] }) }, 
			new Error('asset index out of bounds: 5'),
			"CRP - Manual clustering checks out of bounds #2")

		// Test missing asset index
		assert.throws(function() { 
			PortfolioAllocation.clusterRiskParityWeights(sigma, { clusteringMode: 'manual', clusters: [[1, 2, 3]] }) }, 
			new Error('missing asset index: 4'),
			"CRP - Manual clustering checks missing asset index")
		
		// Test duplicate asset index
		assert.throws(function() { 
			PortfolioAllocation.clusterRiskParityWeights(sigma, { clusteringMode: 'manual', clusters: [[1, 1, 2, 3]] }) }, 
			new Error('duplicate asset index: 1'),
			"CRP - Manual clustering checks duplicate asset index")
	} 


	// Limit cases
	{
		// Example taken from the ERC
		var sigma = [[0.01, 0.016, 0, 0], [0.016, 0.04, 0, 0], [0, 0, 0.09, -0.06], [0, 0, -0.06, 0.16]];
		var ercWeights = [38.4, 19.2, 24.3, 18.2];

		// Limit case #1: in case all assets belong to a unique cluster, the CRP weights must be equal to the ERC weights
		{
			var opt = { clusteringMode: 'manual', clusters: [[1, 2, 3, 4]] };
			var weights = PortfolioAllocation.clusterRiskParityWeights(sigma, opt); 
			var expectedWeights = ercWeights;
			for (var i = 0; i < expectedWeights.length; ++i) {
				assert.equal(Math.round(1000 * weights[i])/10, expectedWeights[i], 'CRP - Values #1 ' + i);
			}
		}

		// Limit case #2: in case each assets belong to one cluster, the CRP weights must be equal to the ERC weights
		{
			var opt = { clusteringMode: 'manual', clusters: [[1], [2], [3], [4]] };
			var weights = PortfolioAllocation.clusterRiskParityWeights(sigma, opt); 
			var expectedWeights = ercWeights;
			for (var i = 0; i < expectedWeights.length; ++i) {
				assert.equal(Math.round(1000 * weights[i])/10, expectedWeights[i], 'CRP - Values #2 ' + i);
			}
		}
	}


	// In case of 2 clusters of 2 assets each, the final weights must verify w_1f = p * w_1erc, w_2f = p * w_2erc,
	// w_3f = (1-p) * w_3erc, w_4f = (1-p) * w_4erc with w_ierc the weight for asset i obtained though the ERC applied
	// intra clusters and p a real number belonging to ]0,1[ dependant on the clusters variances/correlation.
	{
		// Generate two random 2*2 covariance matrices, one for each of the two clusters
		var sigma_c1 = this.randomCovarianceMatrix();
		var sigma_c2 = this.randomCovarianceMatrix();

		// Create a 4x4 bloc matrix using the two 2x2 matrices above
		var sigma = [[sigma_c1[0][0], sigma_c1[0][1], 0, 0], [sigma_c1[1][0], sigma_c1[1][1], 0, 0], [0, 0, sigma_c2[0][0], sigma_c2[0][1]], [0, 0, sigma_c2[1][0], sigma_c2[1][1]]];
		
		// Compute CRP weights for the whole portfolio
		var opt = { clusteringMode: 'manual', clusters: [[1,2], [3,4]] };
		var weights = PortfolioAllocation.clusterRiskParityWeights(sigma, opt); 
		
		// Compute ERC weights for each cluster
		var weights_c1 = PortfolioAllocation.equalRiskContributionWeights(sigma_c1); 
		var weights_c2 = PortfolioAllocation.equalRiskContributionWeights(sigma_c2); 

		// Ensure the relationships described above are satisfied
		var p = weights[0]/weights_c1[0];
		assert.equal(Math.abs(p - weights[1]/weights_c1[1]) <= 1e-8, true, 'CRP - Values #3 0');
		assert.equal(Math.abs(1-p - weights[2]/weights_c2[0]) <= 1e-8, true, 'CRP - Values #3 1');
		assert.equal(Math.abs(1-p - weights[3]/weights_c2[1]) <= 1e-8, true, 'CRP - Values #3 2');
	}


	// CRP using FTCA, using static data
	{
		// Example taken from the ERC
		var sigma = [[0.01, 0.016, 0, 0], [0.016, 0.04, 0, 0], [0, 0, 0.09, -0.06], [0, 0, -0.06, 0.16]];
		
		// Call CRP with no options, which should defaults to FTCA
		var expectedWeights = [ 0.3262379197337176, 0.16311896054014643, 0.29179607083498427, 0.21884704889115175 ];
		var weights = PortfolioAllocation.clusterRiskParityWeights(sigma);
		for (var i = 0; i < sigma[0].length; ++i) { 
			assert.equal(Math.abs(weights[i] - expectedWeights[i]) <= 1e-8, true, 'CRP - Values #4 ' + i);
		}
		
		// Call CRP with FTCA mode, which should output the same result as above
		var opt = { clusteringMode: 'ftca' };
		weights = PortfolioAllocation.clusterRiskParityWeights(sigma, opt); 		
		for (var i = 0; i < sigma[0].length; ++i) { 
			assert.equal(Math.abs(weights[i] - expectedWeights[i]) <= 1e-8, true, 'CRP - Values #5 ' + i);
		}

	}
});


QUnit.test('Global minimum variance portfolio', function(assert) {    
	// Reference: Portfolio Optimization versus Risk-Budgeting Allocation, Thierry Roncalli, WG RISK ESSEC, January 18, 2012
	{
		var weights = PortfolioAllocation.globalMinimumVarianceWeights([[0.0396, 0.0398], [0.0398, 0.0400]]); 
		var expectedWeights =  [1, 0];
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-8), true, 'GMV - Values #1');
	}
	{
		var weights = PortfolioAllocation.globalMinimumVarianceWeights([[0.0400, 0.0396, 0.0414], [0.0396, 0.0484, 0.0455], [0.0414, 0.0455, 0.0529]]); 
		var expectedWeights =  [0.9565217391304353, 0.04347826086956469, 0];
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-8), true, 'GMV - Values #2');
	}
	// Note: The volatility obtained by the computed portfolio is 0.019751470588235294, which is lower than the volatility of 0.2017 in the reference,
	// associated with the commented weights !
	{
		var weights = PortfolioAllocation.globalMinimumVarianceWeights([[0.0400, 0.0374, 0.0391], [0.0374, 0.0484, 0.0430], [0.0391, 0.0430, 0.0529]]); 
		//var expectedWeights =  [0.7009, 0.2378, 0.0613];
		var expectedWeights =  [0.8088235294117648, 0.19117647058823511, 0 ];
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-8), true, 'GMV - Values #3');
	}
	{
		var weights = PortfolioAllocation.globalMinimumVarianceWeights([[0.0400, 0.0418, 0.0437], [0.0418, 0.0484, 0.0481], [0.0437, 0.0481, 0.0529]]); 
		var expectedWeights =  [1, 0, 0];
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-8), true, 'GMV - Values #4');
	}

	// Reference: Understanding the Impact of Weights Constraints in Portfolio Theory, Thierry Roncalli
	{
		var covMat = [[0.0225,0.0030,0.0150,0.0225], [0.0030,0.0400,0.0350,0.0240], [0.0150,0.0350,0.0625,0.0600], [0.0225,0.0240,0.0600,0.0900]];
		
		var expectedWeights =  [0.65487, 0.34513, 0, 0];
		var weights = PortfolioAllocation.globalMinimumVarianceWeights(covMat); 
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-5), true, 'GMV - Values #5');

		var expectedWeights =  [0.56195, 0.23805, 0.1, 0.1];
		var weights = PortfolioAllocation.globalMinimumVarianceWeights(covMat, {constraints: {minWeights: [0.10, 0.10, 0.10, 0.10]}}); 
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-5), true, 'GMV - Values #5, 2');
		
		var expectedWeights =  [0.40, 0.20, 0.2, 0.2];
		var weights = PortfolioAllocation.globalMinimumVarianceWeights(covMat, {constraints: {minWeights: [0.20, 0.20, 0.20, 0.20]}}); 
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-6), true, 'GMV - Values #5, 3');
		
		var expectedWeights =  [0.50, 0.50, 0, 0];
		var weights = PortfolioAllocation.globalMinimumVarianceWeights(covMat, {constraints: {maxWeights: [0.50, 0.50, 0.50, 0.50]}}); 
		assert.equal(areWeightsIdentical(weights, expectedWeights, 0), true, 'GMV - Values #5, 4');
	}
	
	// Reference: Private communication with TrendXplorer
	{
		var covMat = [[0.03401428,0.0333167,-0.00614739,0.00926415,-0.0064081],
					  [0.0333167,0.06323421,-0.00855552,0.02245369,-0.00480642],
					  [-0.00614739,-0.00855552,0.01444902,-0.00432445,0.00690744],
					  [0.00926415,0.02245369,-0.00432445,0.02622712,0.0016983],
					  [-0.0064081,-0.00480642,0.00690744,0.00169834,0.0116492]];
		
		var weights = PortfolioAllocation.globalMinimumVarianceWeights(covMat);
		var weightsGsmo = PortfolioAllocation.globalMinimumVarianceWeights(covMat, {optimizationMethod: 'gsmo', optimizationMethodParams: {epsGsmo: 1e-12}});
		var weightsCla = PortfolioAllocation.globalMinimumVarianceWeights(covMat, {optimizationMethod: 'critical-line'});
		
		var expectedWeights = [0.22372357878117932, 0, 0.30887500093660675, 0.13381977357496827, 0.3335816467072458];
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-8), true, 'Test #6, AUTO');
		assert.equal(areWeightsIdentical(weightsGsmo, expectedWeights, 1e-8), true, 'Test #6, GSMO');
		assert.equal(areWeightsIdentical(weightsCla, expectedWeights, 1e-8), true, 'Test #6, CLA');		
	}
	
	// Reference: Private communication
	{
		var covMat = [[0.00008167944062,	0.00002381649417,	0.0000270928852,	0.00004867187199,	0.00003320921106],
					[0.00002381649417,	0.00002359207427,	0.000005565882395,	0.0000188629976,	0.00001333414531],
					[0.0000270928852,	0.000005565882395,	0.00001048947312,	0.0000168927258,	0.00001058070153],
					[0.00004867187199,	0.0000188629976,	0.0000168927258,	0.00003645410697,	0.00002279528797],
					[0.00003320921106,	0.00001333414531,	0.00001058070153,	0.00002279528797,	0.00001607412275]];
					
		// True expected values
		var weights = PortfolioAllocation.globalMinimumVarianceWeights(covMat);			
		var expectedWeights =  [0, 0.21453757583742858, 0.7854624241625714, 0, 0]
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-8), true, 'Test #7');
		
		// Wrong values, obtained by setting eps with a too big value for GSMO
		var weightsGsmo = PortfolioAllocation.globalMinimumVarianceWeights(covMat, {optimizationMethod: 'gsmo', optimizationMethodParams: {epsGsmo: 1e-4}}); 
		var expectedWeights =  [0.2, 0.2, 0.2, 0.2, 0.2]
		assert.equal(areWeightsIdentical(weightsGsmo, expectedWeights, 1e-8), true, 'Test #7, GSMO');
	}
	
	// Reference: Xi Bai, Katya Scheinberg, Reha Tutuncu, Least-squares approach to risk parity in portfolio selection
	{
		var covMat = [[94.868,33.750,12.325,-1.178,8.778],
					  [33.750,445.642,98.955,-7.901,84.954],
					  [12.325,98.955,117.265,0.503,45.184],
					  [-1.178,-7.901,0.503,5.460,1.057],
					  [8.778,84.954,45.184,1.057,34.126]];
			
		var weights = PortfolioAllocation.globalMinimumVarianceWeights(covMat); 
		var expectedWeights =  [0.050, 0.006, 0, 0.862, 0.082];
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-3), true, 'Test #8');
	}
	
	// Reference: Xi Bai, Katya Scheinberg, Reha Tutuncu, Least-squares approach to risk parity in portfolio selection
	// Constraints on assets weights
	{
		var covMat = [[94.868,33.750,12.325,-1.178,8.778],
					  [33.750,445.642,98.955,-7.901,84.954],
					  [12.325,98.955,117.265,0.503,45.184],
					  [-1.178,-7.901,0.503,5.460,1.057],
					  [8.778,84.954,45.184,1.057,34.126]];
			
		var minWeights = [0.05,0.05,0.05,0.05,0.05];
		var maxWeights = [0.35,0.35,0.35,0.35,0.35];
		var weights = PortfolioAllocation.globalMinimumVarianceWeights(covMat, { constraints: {minWeights: minWeights, maxWeights: maxWeights} }); 
		var expectedWeights =  [0.200,0.050,0.050,0.350,0.350];
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-8), true, 'Test #9, constrained');
		
		var minWeights = [0.35,0.05,0.05,0.05,0];
		var maxWeights = [0.35,0.35,0.35,0.35,0];
		var weights = PortfolioAllocation.globalMinimumVarianceWeights(covMat, { constraints: {minWeights: minWeights, maxWeights: maxWeights} }); 
		var expectedWeights =  [0.350,0.050,0.25,0.350,0];
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-8), true, 'Test #9, constrained with exact tight constraints on weights');
	}
	
	// Reference: private communication
	//
	// This method test the fact that for a nearly singular matrix [determinant ~5e-24), the GMV
	// is the same if the returns are considered as identical (e.g. null),
	// v.s. provided in input (because nearly singular <> singular).
	{
		var covMat = [[0.05434700626, 0.05450611079, 0.0545732999, 0.04429473013, 0.03183574594, 0.02601472703, -0.007675835176, -0.0003383554852],
					[0.05186508706, 0.05711441188, 0.05233270776, 0.03993880464, 0.02927676965, 0.02103022743, -0.004987619848, -0.0002253313801],
					[0.0545732999, 0.05499754323, 0.06984354197, 0.04885108716, 0.02943209273, 0.02965970237, -0.009724897251, -0.0003809302526],
					[0.04429473013, 0.04197252977, 0.04885108716, 0.04175438053, 0.02938755621, 0.0233266123, -0.009103127988, -0.00038995015],
					[0.03183574594, 0.03076757296, 0.02943209273, 0.02938755621, 0.02569034099, 0.01355929033, -0.006665862624, -0.0002719638161],
					[0.02601472703, 0.02210110831, 0.02965970237, 0.0233266123, 0.01355929033, 0.01851558467, -0.004491637207, -0.0002525477259],
					[-0.007675835176, -0.005241594596, -0.009724897251, -0.009103127988, -0.006665862624, -0.004491637207, 0.003760367113, 0.0001072357482],
					[-0.0003383554852, -0.0002368054864, -0.0003809302526, -0.00038995015, -0.0002719638161, -0.0002525477259, 0.0001072357482, 0.000007905033839]];
		var maxWeights = [0.25,0.25,0.25,0.25,0.25,0.25,1,1];

		//
		var weightsNullMu = PortfolioAllocation.globalMinimumVarianceWeights(covMat, { constraints: {maxWeights: maxWeights} }); 

		//
		var mu = [0.33/100, 1.69/100, -0.30/100, -0.08/100, 0.49/100, -0.01/100, 0.48/100, 0.05/100];
		var weightsMu = PortfolioAllocation.globalMinimumVarianceWeights(covMat, { mu: mu, constraints: {maxWeights: maxWeights} }); 
		assert.equal(areWeightsIdentical(weightsNullMu, weightsMu, 0), true, 'Test #9, nearly singular covariance matrix');
	}
	
	// Test that the GMV portfolio is efficient and supports by default equal returns, 
	// using a semi-definite positive covariance matrix.
	{
		var weights = PortfolioAllocation.globalMinimumVarianceWeights([[0.0400, 0.0400], [0.0400, 0.0400]], {mu: [0.01, 0.02]});
		var weightsCla = PortfolioAllocation.globalMinimumVarianceWeights([[0.0400, 0.0400], [0.0400, 0.0400]], {optimizationMethod:'critical-line', mu: [0.01, 0.02]});
		var expectedWeights = [0, 1];
		assert.equal(areWeightsIdentical(weights, weightsCla, 0), true, 'Test #10, semi-definite positive matrix, CLA');

		var weights = PortfolioAllocation.globalMinimumVarianceWeights([[0.0400, 0.0400], [0.0400, 0.0400]], {mu: [0.02, 0.01]});
		var weightsCla = PortfolioAllocation.globalMinimumVarianceWeights([[0.0400, 0.0400], [0.0400, 0.0400]], {optimizationMethod:'critical-line', mu: [0.02, 0.01]});
		var expectedWeights =  [1, 0];
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-12), true, 'Test #10, equal returns, AUTO');
		assert.equal(areWeightsIdentical(weightsCla, expectedWeights, 1e-12), true, 'Test #10, equal returns, CLA');
		
		var weightsNoMu = PortfolioAllocation.globalMinimumVarianceWeights([[0.0400, 0.0400], [0.0400, 0.0400]]);
		var weights = PortfolioAllocation.globalMinimumVarianceWeights([[0.0400, 0.0400], [0.0400, 0.0400]], {mu: [0.01, 0.01]});
		var expectedWeights =  [0.5, 0.5];
		assert.equal(areWeightsIdentical(weightsNoMu, expectedWeights, 1e-8), true, 'Test #10, semi-definite positive matrix and equal returns');
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-8), true, 'Test #10, semi-definite positive matrix and equal returns');
	}
	
	// Test with a pseudo cash position that 1) the execution time is not varying and 2) that everything
	// is invested in the pseudo cash position.
	{
		var covMat = [[ 94.868, 33.75, 12.325, -1.178, 8.778, 0 ],
						[ 33.75, 445.642, 98.955, -7.901, 84.954, 0 ],
						[ 12.325, 98.955, 117.265, 0.503, 45.184, 0 ],
						[ -1.178, -7.901, 0.503, 5.46, 1.057, 0 ],
						[ 8.778, 84.954, 45.184, 1.057, 34.126, 0 ],
						[ 0, 0, 0, 0, 0, 0 ]];
		var returns = [ 0, 0, 0, 0, 0, -0.01 ];
		var minWeights = [ 0, 0, 0, 0, 0, 0 ];
		var maxWeights = [ 1, 1, 1, 1, 1, 1 ];
		
		var weights = PortfolioAllocation.globalMinimumVarianceWeights(covMat, { mu: returns, constraints: {minWeights:minWeights, maxWeights: maxWeights} }); 
		var expectedWeights =  [0, 0, 0, 0, 0, 1];
		assert.equal(areWeightsIdentical(weights, expectedWeights), true, 'Test #11, pseudo cash position');
	}
});



QUnit.test('Numerical optimization portfolio', function(assert) {    
	// Test unsupported optimisation method
	{
		assert.throws(function() { 
			PortfolioAllocation.numericalOptimizationWeights(3,  function (arr) { return 1; }, {optimizationMethod: 'none'}) },
			new Error('unsupported optimisation method'),
			"Grid search portfolio - Unsupported optimisation method");
	}
	
	// Test grid search using static data
	{
		// Objective function: portfolio variance, for three assets
		// Covariance matrix taken from GMV test case #2
		// From GMV test case #2, expected (exact) weights are [0.9565927119697761, 0.043407288030223846, 0];
		function portfolio_variance_three_assets(arr) { 
			var covMat = [[0.0400, 0.0396, 0.0414], [0.0396, 0.0484, 0.0455], [0.0414, 0.0455, 0.0529]];
			
			return arr[0]*arr[0]*covMat[0][0] + arr[1]*arr[1]*covMat[1][1] + arr[2]*arr[2]*covMat[2][2] +
			2*arr[0]*arr[1]*covMat[0][1] + 2*arr[0]*arr[2]*covMat[0][2] + 2*arr[1]*arr[2]*covMat[1][2];
		}
		
		// Test default values for the grid search
		assert.deepEqual(PortfolioAllocation.numericalOptimizationWeights(3, portfolio_variance_three_assets), 
		PortfolioAllocation.numericalOptimizationWeights(3, portfolio_variance_three_assets, {optimizationMethod: 'grid-search', optimizationMethodParams: {k: 3}}), 
		'Numerical optimization portfolio - Default values');

		// Test finer and finer rational grids
		var expectedWeights = [[1,0,0]];
		var weights = PortfolioAllocation.numericalOptimizationWeights(3, portfolio_variance_three_assets, {optimizationMethodParams: {k: 3}});
		for (var i = 0; i < weights.length; ++i) {
			assert.equal(areWeightsIdentical(weights[i], expectedWeights[i], 0), true, 'Numerical optimization portfolio - Values #1' + i);
		}
		
		var expectedWeights = [[1,0,0]];
		var weights = PortfolioAllocation.numericalOptimizationWeights(3, portfolio_variance_three_assets, {optimizationMethodParams: {k: 10}});
		for (var i = 0; i < weights.length; ++i) { 
			assert.equal(areWeightsIdentical(weights[i], expectedWeights[i], 0), true, 'Numerical optimization portfolio - Values #2' + i);
		}
		
		var expectedWeights = [[0.96,0.04,0]];
		var weights = PortfolioAllocation.numericalOptimizationWeights(3, portfolio_variance_three_assets, {optimizationMethodParams: {k: 100}});
		for (var i = 0; i < weights.length; ++i) {
			assert.equal(areWeightsIdentical(weights[i], expectedWeights[i], 0), true, 'Numerical optimization portfolio - Values #3' + i);
		}
		
		var expectedWeights = [[0.957,0.043,0]];
		var weights = PortfolioAllocation.numericalOptimizationWeights(3, portfolio_variance_three_assets, {optimizationMethodParams: {k: 1000}});
		for (var i = 0; i < weights.length; ++i) { 
			assert.equal(areWeightsIdentical(weights[i], expectedWeights[i], 0), true, 'Numerical optimization portfolio - Values #4' + i);
		}
	}

	// Test grid search with bounds contraints using static data
	{
		// Objective function: portfolio return, for three assets
		function portfolio_return_three_assets(arr) { 
			var ret = [-0.1, 0.3, .1];
			
			return -(arr[0]*ret[0] + arr[1]*ret[1] + arr[2]*ret[2]);
		}
		
		var expectedWeights = [[0.5,0.5,0]];
		var weights = PortfolioAllocation.numericalOptimizationWeights(3, portfolio_return_three_assets, {optimizationMethodParams: {k: 10}, constraints: {minWeights: [0.5, 0, 0]}});
		for (var i = 0; i < weights.length; ++i) { 
			assert.equal(areWeightsIdentical(weights[i], expectedWeights[i], 0), true, 'Numerical optimization portfolio, bounds contraints - Values #5' + i);
		}
		
		var expectedWeights = [[0,0.5,0.5]];
		var weights = PortfolioAllocation.numericalOptimizationWeights(3, portfolio_return_three_assets, {optimizationMethodParams: {k: 10}, constraints: {maxWeights: [1, 0.5, 1]}});
		for (var i = 0; i < weights.length; ++i) {
			assert.equal(areWeightsIdentical(weights[i], expectedWeights[i], 0), true, 'Numerical optimization portfolio, bounds contraints - Values #6' + i);
		}
	}
	
	// TODO: Use random data with return as function + random contraints v.s. solving the same problem via linear programming
});



QUnit.test('Equal risk bounding portfolio', function(assert) {    
	// Reference: Cesarone, F. & Tardella F., Equal Risk Bounding is better than Risk Parity for portfolio selection, J Glob Optim (2017) 68: 439

	// Example of Table 1
	{
		// Data
		var sigma = [[1,-9/10, 3/5], [-9/10, 1,-1/5],[ 3/5, -1/5, 4]];
		var nbAssets = sigma.length;

		// Compute ERB weights
		var weights = PortfolioAllocation.equalRiskBoundingWeights(sigma);
		
		// Compare ERB weights to expected weights
		var expectedWeights = [0.5, 0.5, 0];
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-8), true, 'ERB - Values #1');
	}

	// Example of Table 2
	{
		// Data
		var sigma = [[1/100, 21/5000, -147/10000, 4/625, -31/2000], [21/5000, 1/25, -57/5000, -21/1250, 21/500], [-147/10000, -57/5000, 9/100, 36/625, 3/125], [4/625, -21/1250, 36/625, 4/25, 1/250], [-31/2000, 21/500, 3/125, 1/250,  1/4]];
		var nbAssets = sigma.length;

		// Compute ERB weights
		var weights = PortfolioAllocation.equalRiskBoundingWeights(sigma);
		
		// Compare ERB weights to expected weights
		var expectedWeights = [0.583, 0.157, 0.186, 0, 0.074];
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-3), true, 'ERB - Values #2');
	}
	
	// Examples with bounds constraints
	{
		// Data
		var sigma = [[1,-9/10, 3/5], [-9/10, 1,-1/5],[ 3/5, -1/5, 4]];
		
		// Completely Infeasible problem due to upper bound constraints
		assert.throws(function() { 
			var upperBounds = [0.2, 0.2, 0.2];
			var weights = PortfolioAllocation.equalRiskBoundingWeights(sigma, {constraints: {maxWeights: upperBounds}});
		},
		new Error('no feasible portfolio generated'),
		"Infeasible problem");
		
		// Partially feasible problem (only the 3 assets case)
		var lowerBounds = [0.2, 0.1, 0.3];
		var upperBounds = [0.4, 0.4, 0.4];
		var expectedWeights = [0.3, 0.4, 0.3];
		var weights = PortfolioAllocation.equalRiskBoundingWeights(sigma, {constraints: {minWeights: lowerBounds, maxWeights: upperBounds}});
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-6), true, 'ERB - Values #3');
		
		// Fully feasible problem
		var lowerBounds = [0.2, 0.1, 0.3];
		var upperBounds = [0.6, 0.6, 0.6];
		var expectedWeights = [0.5, 0.5, 0];
		var weights = PortfolioAllocation.equalRiskBoundingWeights(sigma, {constraints: {minWeights: lowerBounds, maxWeights: upperBounds}});
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-6), true, 'ERB - Values #4');
	}

/*
	// Static example with 20 assets, to show the tractability of the ERB algorithm with this number of assets
	// This takes around 50 seconds on a 2012 computer
	{
	// Data
	var sigma = [[1.0000,0.3010,-0.2531,0.0497,-0.1286,0.0689,-0.0366,-0.0950,0.0502,-0.0342,0.0074,0.0107,-0.0971,0.2335,0.0807,-0.0024,-0.1245,0.0835,0.1783,-0.0989],
				[0.3010,1.0000,0.1298,0.0349,-0.0470,-0.1580,0.0884,-0.0551,0.0455,-0.0042,-0.1111,-0.0481,0.2739,-0.1916,-0.0774,0.0735,0.0239,0.0936,0.0030,-0.2237],
				[-0.2531,0.1298,1.0000,-0.0888,0.0696,-0.0881,0.1338,-0.1019,-0.2108,0.0422,-0.1493,-0.2719,0.1637,-0.2037,-0.2762,-0.0429,-0.1202,0.1024,0.1826,-0.2397],
				[0.0497,0.0349,-0.0888,1.0000,-0.0711,0.0303,-0.2361,-0.2027,0.2024,0.3856,-0.0313,0.1930,0.1567,-0.2357,0.0109,-0.1681,-0.0752,0.2626,0.0212,0.0834],
				[-0.1286,-0.0470,0.0696,-0.0711,1.0000,-0.2109,0.0067,0.0439,-0.2620,-0.2619,0.0822,0.1585,-0.1585,0.0025,-0.3311,-0.1409,-0.0738,-0.0535,-0.0684,0.2921],
				[0.0689,-0.1580,-0.0881,0.0303,-0.2109,1.0000,-0.1430,-0.2636,-0.0971,0.1451,-0.1107,-0.0961,-0.1054,-0.0189,0.0006,0.0678,0.1373,0.3409,-0.0739,0.1559],
				[-0.0366,0.0884,0.1338,-0.2361,0.0067,-0.1430,1.0000,-0.1179,0.1548,-0.0829,0.1414,-0.3829,-0.0553,0.0796,0.0092,0.0328,-0.1051,0.0139,-0.0700,0.1420],
				[-0.0950,-0.0551,-0.1019,-0.2027,0.0439,-0.2636,-0.1179,1.0000,-0.1681,0.1354,-0.0915,-0.1071,-0.1485,-0.0956,0.0881,-0.0646,0.1750,-0.2343,0.0476,-0.0378],
				[0.0502,0.0455,-0.2108,0.2024,-0.2620,-0.0971,0.1548,-0.1681,1.0000,0.0967,0.1938,-0.0046,0.3219,0.1068,0.0268,0.0781,0.0529,0.0346,0.1081,0.1386],
				[-0.0342,-0.0042,0.0422,0.3856,-0.2619,0.1451,-0.0829,0.1354,0.0967,1.0000,0.0773,-0.1698,-0.1185,-0.1201,0.1164,0.0458,-0.0040,-0.0958,-0.2451,0.0366],
				[0.0074,-0.1111,-0.1493,-0.0313,0.0822,-0.1107,0.1414,-0.0915,0.1938,0.0773,1.0000,-0.1579,-0.2839,0.0309,-0.1498,0.3240,-0.0849,-0.1001,0.0279,0.1015],
				[0.0107,-0.0481,-0.2719,0.1930,0.1585,-0.0961,-0.3829,-0.1071,-0.0046,-0.1698,-0.1579,1.0000,0.2488,0.1018,-0.0556,-0.0657,0.0473,-0.1634,-0.1715,0.0021],
				[-0.0971,0.2739,0.1637,0.1567,-0.1585,-0.1054,-0.0553,-0.1485,0.3219,-0.1185,-0.2839,0.2488,1.0000,0.0940,-0.1194,0.1852,-0.1272,-0.1893,0.0399,-0.0084],
				[0.2335,-0.1916,-0.2037,-0.2357,0.0025,-0.0189,0.0796,-0.0956,0.1068,-0.1201,0.0309,0.1018,0.0940,1.0000,0.1020,0.0104,-0.1762,-0.2871,0.1063,0.0213],
				[0.0807,-0.0774,-0.2762,0.0109,-0.3311,0.0006,0.0092,0.0881,0.0268,0.1164,-0.1498,-0.0556,-0.1194,0.1020,1.0000,0.0257,-0.1306,0.1000,-0.0951,-0.2423],
				[-0.0024,0.0735,-0.0429,-0.1681,-0.1409,0.0678,0.0328,-0.0646,0.0781,0.0458,0.3240,-0.0657,0.1852,0.0104,0.0257,1.0000,-0.2438,-0.2582,-0.2274,0.0310],
				[-0.1245,0.0239,-0.1202,-0.0752,-0.0738,0.1373,-0.1051,0.1750,0.0529,-0.0040,-0.0849,0.0473,-0.1272,-0.1762,-0.1306,-0.2438,1.0000,-0.0231,0.1164,0.0304],
				[0.0835,0.0936,0.1024,0.2626,-0.0535,0.3409,0.0139,-0.2343,0.0346,-0.0958,-0.1001,-0.1634,-0.1893,-0.2871,0.1000,-0.2582,-0.0231,1.0000,0.0632,0.0336],
				[0.1783,0.0030,0.1826,0.0212,-0.0684,-0.0739,-0.0700,0.0476,0.1081,-0.2451,0.0279,-0.1715,0.0399,0.1063,-0.0951,-0.2274,0.1164,0.0632,1.0000,-0.3556],
				[-0.0989,-0.2237,-0.2397,0.0834,0.2921,0.1559,0.1420,-0.0378,0.1386,0.0366,0.1015,0.0021,-0.0084,0.0213,-0.2423,0.0310,0.0304,0.0336,-0.3556,1.0000]]
	var nbAssets = sigma.length;
				
	// Compute ERB weights
	var weights = PortfolioAllocation.equalRiskBoundingWeights(sigma);				
	}
*/
});


QUnit.test('Random weights portfolio', function(assert) {    
	// Test with random data, no constraints
	{
		// Setup static parameters of the random test
		var nbTests = 50;
		var nbAssetsMin = 1;
		var nbAssetsMax = 50;

		// Aim of these tests is to check that for a portfolio of n assets:
		// - The number of weights computed is n
		// - The weights belong to the interval [0, 1]
		// - The weights sum to 1	  
		for (var i = 0; i < nbTests; ++i) {
			// Generate a random number of assets
			var nbAssets = Math.floor(Math.random()*(nbAssetsMax - nbAssetsMin + 1) + nbAssetsMin);
			
			// Generate a random portfolio for this number of assets
			var randomWeights = PortfolioAllocation.randomWeights(nbAssets);
			
			// Check that the number of weights corresponds to the number of assets
			var nbWeights = randomWeights.length;
			assert.equal(nbWeights, nbAssets, "Random weights portfolio, number of weights - Test " + i);
			
			// Check that the weights belong to the unit interval
			var weightsBelongToUnitInterval = true;
			for (var k = 0; k < randomWeights.length; ++k) {
				if (randomWeights[k] > 1 || randomWeights[k] < 0) {
					weightsBelongToUnitInterval = false;
					break;
				}
			}
			assert.equal(weightsBelongToUnitInterval, true, "Random weights portfolio, weights in unit interval - Test " + i);

			// Check that the sum of the weights is 1, near to machine precision
			var weightsSum = 0;
			for (var k = 0; k < randomWeights.length; ++k) {
				weightsSum += randomWeights[k];
			}
			assert.equal(Math.abs(weightsSum - 1) <= 1e-15, true, "Random weights portfolio, weights sum to one - Test " + i);
		}
	}

	// Test with random data, cardinality constraints
	{
		// Setup static parameters of the random test
		var nbTests = 50;
		var nbAssetsMin = 1;
		var nbAssetsMax = 50;

		// Aim of these tests is to check that for a portfolio of n assets constrained to hold between i and j assets:
		// - The number of non-zero weights is between i and j
		for (var i = 0; i < nbTests; ++i) {
			// Generate a random number of assets
			var nbAssets = Math.floor(Math.random()*(nbAssetsMax - nbAssetsMin + 1) + nbAssetsMin);
			
			// Generate random cardinality constraints
			var maxAssets = Math.floor(Math.random()*(nbAssets - 1 + 1) + 1);
			var minAssets = Math.floor(Math.random()*(maxAssets - 1 + 1) + 1);
			
			// Generate a random portfolio for this number of assets and these cardinality constraints
			var randomWeights = PortfolioAllocation.randomWeights(nbAssets, { constraints: { minNbAssets: minAssets, maxNbAssets: maxAssets } });
			
			// Check that the number of non-zero weights is between minAssets and maxAssets
			var nbNonZeroAssets = 0;
			for (var k = 0; k < randomWeights.length; ++k) {
				if (randomWeights[k] != 0) {
					nbNonZeroAssets = nbNonZeroAssets + 1;
				}
			}
			assert.equal(maxAssets >= nbNonZeroAssets, true, "Random weights portfolio with cardinality constraints, number of non-zero weights - Test " + i + "/1");
			assert.equal(minAssets <= nbNonZeroAssets, true, "Random weights portfolio with cardinality constraints, number of non-zero weights - Test " + i + "/2");
		}
	}
	
	// Test with static data, min/max weights and cardinality constraints
	{
		//
		var nbAssets = 3;
		
		//
		var maxAssets = 2;
		var minAssets = 1;
		
		// Set non null minimum weights constraints, and feasible lower/upper bounds
		// when only 2 assets are selected.
		var lowerBounds = [0.25, 0.15, 0.10];
		var upperBounds = [0.85, 0.6, 0.75];
		
		// Generate a random portfolio
		var randomWeights = PortfolioAllocation.randomWeights(nbAssets, { constraints: { minWeights: lowerBounds, maxWeights: upperBounds,
																						 minNbAssets: minAssets, maxNbAssets: maxAssets } });
		
		// Check that there are between 1 and 2 non zero weights (here, must be 2 given the constraints)
		//
		// This in particular implies that the minimum weight constraint for the non-selected asset
		// has properly been discarded.
		var nbNonZeroAssets = 0;
		for (var k = 0; k < randomWeights.length; ++k) {
			if (randomWeights[k] != 0) {
				nbNonZeroAssets = nbNonZeroAssets + 1;
			}
		}
		assert.equal(maxAssets >= nbNonZeroAssets, true, "Random weights portfolio with min/max weights and cardinality constraints, number of non-zero weights 1");
		assert.equal(minAssets <= nbNonZeroAssets, true, "Random weights portfolio with min/max weights and  cardinality constraints, number of non-zero weights 2");
		
		// Check that the min/max weights constraints for the selected
		// assets have been taken into account.
		var weightsConstraintsOK = true;
		for (var k = 0; k < randomWeights.length; ++k) {
			if (randomWeights[k] != 0) {
				if (randomWeights[k] > upperBounds[k] || randomWeights[k] < lowerBounds[k]) {
					weightsConstraintsOK = false;
					break;
				}
			}
		}
		assert.equal(weightsConstraintsOK, true, "Random weights portfolio with min/max weights and cardinality constraints, weights constraints properly taken into account");
	}

	// Test with random data, exposure constraints
	{
		// Setup static parameters of the random test
		var nbTests = 10;
		var nbAssetsMin = 1;
		var nbAssetsMax = 50;

		// Aim of these tests is to check that for a portfolio of n assets with a constrained exposure,
		// the exposure is properly constrained.
		for (var i = 0; i < nbTests; ++i) {
			// Generate a random number of assets
			var nbAssets = Math.floor(Math.random()*(nbAssetsMax - nbAssetsMin + 1) + nbAssetsMin);
			
			// Generate a random exposure constraint
			var maxExposure = Math.random();
			var minExposure = Math.min(Math.random(), maxExposure);

			// Generate a random portfolio for this number of assets/exposure
			var randomWeights = PortfolioAllocation.randomWeights(nbAssets, { constraints: { minExposure: minExposure, maxExposure: maxExposure } });

			// Check that the sum of the weights is belongs to the exposure interval, near to machine precision
			var weightsSum = 0;
			for (var k = 0; k < randomWeights.length; ++k) {
				weightsSum += randomWeights[k];
			}
			assert.equal(weightsSum <= maxExposure + 1e-14, true, "Random weights portfolio with exposure constraints, weights lower than max exposure - Test " + i);
			assert.equal(weightsSum >= minExposure - 1e-14, true, "Random weights portfolio with exposure constraints, weights greater than min exposure - Test " + i);
		}
	}
});


QUnit.test('Minimax portfolio', function(assert) {    
	// Static data
	// Reference: Christos Papahristodoulou, Optimal portfolios using Linear Programming models (preliminary version on Internet)
	// (b) Maximin formulation example
	{
		// Define the assets returns
		var assetsReturns = [[0.054,  0.045,  -0.03, -0.018,  0.043,  0.047,  0.055,  0.036, -0.039, -0.043,  0.046, 0.052], // Asset 1 returns
		[0.032,  0.055, -0.036,  0.052,  0.047,  0.034,  0.063,  0.048,  0.025,   0.04,  0.036, -0.017], // ...
		[0.064,  0.056,  0.048,  0.007,  0.053,  0.036,  0.017,  0.047, -0.059,  0.047,   0.04, 0.032],
		[0.038,  0.062, -0.037,   0.05,  0.065, -0.043,  0.062,  0.034,  0.035,  0.056,  0.057, 0.025],
		[0.049,  0.067, -0.039,  0.051,  0.049,  0.037,  0.055,  0.025,  0.052,  0.02,   0.045,  0.04]];
		
		// Compute the associated minimax portfolio weights
		var minimaxWeights = PortfolioAllocation.minimaxWeights(assetsReturns);
		
		// Compare minimax weights to expected weights
		var expectedWeights = [0, 0, 0.459596, 0, 0.540404];
		for (var i = 0; i < assetsReturns.length; ++i) { 
			assert.equal(Math.abs(minimaxWeights[i] - expectedWeights[i]) <= 1e-3, true, 'Minimax - Values ' + i);
		}
		
		// Compute the minimum portfolio return
		var minimaxWeights = PortfolioAllocation.minimaxWeights(assetsReturns, {outputMinimumPortfolioReturn: true});
		var expectedMinPortfolioReturn = 0.000985;
		assert.equal(Math.abs(minimaxWeights[1] - expectedMinPortfolioReturn) <= 1e-3, true, 'Minimax - Minimum Portfolio Return');
	} 
	
	// Static data, no full investment
	// Reference: Christos Papahristodoulou, Optimal portfolios using Linear Programming models (preliminary version on Internet)
	// (b) Maximin formulation example, with March return of security C changed from 0.048 to -0.03
	{
		// Define the assets returns
		var assetsReturns = [[0.054,  0.045,  -0.03, -0.018,  0.043,  0.047,  0.055,  0.036, -0.039, -0.043,  0.046, 0.052], // Asset 1 returns
		[0.032,  0.055, -0.036,  0.052,  0.047,  0.034,  0.063,  0.048,  0.025,   0.04,  0.036, -0.017], // ...
		[0.064,  0.056,  -0.03,  0.007,  0.053,  0.036,  0.017,  0.047, -0.059,  0.047,   0.04, 0.032], 
		[0.038,  0.062, -0.037,   0.05,  0.065, -0.043,  0.062,  0.034,  0.035,  0.056,  0.057, 0.025],
		[0.049,  0.067, -0.039,  0.051,  0.049,  0.037,  0.055,  0.025,  0.052,  0.02,   0.045,  0.04]];
		
		// Compute the associated minimax portfolio weights
		var minimaxWeights = PortfolioAllocation.minimaxWeights(assetsReturns, {constraints: {fullInvestment: false}});
		
		// Compare minimax weights to expected weights
		var expectedWeights = [0, 0, 0, 0, 0];
		for (var i = 0; i < assetsReturns.length; ++i) { 
			assert.equal(Math.abs(minimaxWeights[i] - expectedWeights[i]) <= 0, true, 'Minimax #2 - Values ' + i);
		}
		
		// Compute the minimum portfolio return
		var minimaxWeights = PortfolioAllocation.minimaxWeights(assetsReturns, {outputMinimumPortfolioReturn: true});
		var expectedMinPortfolioReturn =  -0.030794545037876233;
		assert.equal(Math.abs(minimaxWeights[1] - expectedMinPortfolioReturn) <= 1e-3, true, 'Minimax #2 - Minimum Portfolio Return');
	} 
});



QUnit.test('Mean variance portfolio - error cases', function(assert) {    
	// Test using static data
	// Test the unsupported cases:
	// - Missing return, volatility or risk tolerance constraint
	// - Simultaneous exclusive constraints
	// - Inconsistent number of returns
	// - Inconsistent number of weight constraints
	// - Incorrect covariance matrix
	{
		var covMat = [[0.0146, 0.0187, 0.0145],
					 [0.0187, 0.0854, 0.0104],
					  [0.0145, 0.0104, 0.0289]];
		var returns = [0.062, 0.146, 0.128];
		
		assert.throws(function() { 
		PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat) },
			new Error('missing return, volatility or risk tolerance constraints'),
			"Mean variance portfolio - Missing mean-variance optimization method");

		assert.throws(function() { 
		PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { constraints: {return: 0.10, volatility: 0.1}}) },
			new Error('simultaneous return and volatility constraints'),
			"Mean variance portfolio - Simultaneous exclusive constraints");
			
		assert.throws(function() { 
		PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { constraints: {riskTolerance: 4.18, volatility: 0.1}}) },
			new Error('simultaneous risk tolerance and return or volatility constraints'),
			"Mean variance portfolio - Simultaneous exclusive constraints #2");
			
		assert.throws(function() { 
		PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { constraints: {riskTolerance: 4.18, return: 0.1}}) },
			new Error('simultaneous risk tolerance and return or volatility constraints'),
			"Mean variance portfolio - Simultaneous exclusive constraints #3");
					
		assert.throws(function() { 
			var minWeights = [1, 1, 1];
			var weights = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, {constraints: { return: 0.10, minWeights: minWeights } } );
		},
		new Error('infeasible problem detected: the restricted simplex is empty'),
		"Bounds constraints, infeasible weights constraints");
	}
	
});	

QUnit.test('Mean variance portfolio - return constraint', function(assert) {    
	// Test using random data
	// Test unreachable cases
	{
		var covMat = [[0.0146, 0.0187, 0.0145],
					 [0.0187, 0.0854, 0.0104],
					  [0.0145, 0.0104, 0.0289]];
		var returns = [0.062, 0.146, 0.128];
		
		var maxReturn = returns[1];
		var unreachableMaxReturn = maxReturn + (1 - Math.random()); // > maxReturn
		assert.throws(function() { 
		PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { constraints: {return: unreachableMaxReturn}}) },
			new Error('no matching efficient portfolio with a return equal to ' + unreachableMaxReturn),
			"Return constraint, unreachable return #1/1");

		assert.throws(function() { 
		PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { optimizationMethod: "critical-line", 
		                                                                       constraints: {return: unreachableMaxReturn}}) },
			new Error('no matching efficient portfolio with a return equal to ' + unreachableMaxReturn),
			"Return constraint, unreachable return #1/2");

			
		var minReturn = returns[0];
		var unreachableMinReturn = minReturn - (1 - Math.random()); // < minReturn
		assert.throws(function() { 
		PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { optimizationMethod: "critical-line",  
		                                                                       constraints: {return: unreachableMinReturn}}) },
			new Error('no matching efficient portfolio with a return equal to ' + unreachableMinReturn),
			"Return constraint, unreachable return #2/1");

		assert.throws(function() { 
		PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { constraints: {return: unreachableMinReturn}}) },
			new Error('no matching efficient portfolio with a return equal to ' + unreachableMinReturn),
			"Return constraint, unreachable return #2/2");
	}
	
	// Test using static data
	{
		var covMat = [[1, 0.3], [0.3, 1]];
		var returns = [0.1, 0.2];
		
		//
		var targetReturn = returns[1]; // the second asset
		var expectedWeights = [0, 1]; 		
		var weights = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { constraints: {return: targetReturn}});
		assert.equal(areWeightsIdentical(weights, expectedWeights, 0), true, 'Return constraint #0');
		
		//
		var targetReturn = 0.15; // exact mix of the two assets
		var expectedWeights = [0.5, 0.5];
		var weights = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { constraints: {return: targetReturn}});
		assert.equal(areWeightsIdentical(weights, expectedWeights, 0), true, 'Return constraint #0, bis');
	}
		
	// Test using static data (positive semi-definite covariance matrix)
	// Reference: Portfolio Selection, H. Markowitz example, chapter II "Illustrative portfolio analyses"
	{	
		var covMat = [[0.05338816358,0.02149069753,0.02865533642,0.04896485802,0.01624895062,0.03223945062,0.02425553395,0.03999812963,0.0361509784, 0],
					[0.02149069753,0.01468446914,0.01878391358,0.02441658642,0.008041938272,0.01002193827,0.01448993827,0.02536259259,0.02083593827, 0],
					[0.02865533642,0.01878391358,0.08550016358,0.06260714198,0.04439938272,0.01328671605,0.01043991049,0.06864603704,0.0420215216, 0],
					[0.04896485802,0.02441658642,0.06260714198,0.09546446914,0.05153806173,0.02902461728,0.02077028395,0.09002012963,0.03664589506, 0],
					[0.01624895062,0.008041938272,0.04439938272,0.05153806173,0.1278900988,0.0128384321,0.02091715432,0.1015344074,0.04497232099, 0],
					[0.03223945062,0.01002193827,0.01328671605,0.02902461728,0.0128384321,0.04125832099,0.01127854321,0.02960762963,0.02165332099, 0],
					[0.02425553395,0.01448993827,0.01043991049,0.02077028395,0.02091715432,0.01127854321,0.02883379321,0.02913762963,0.01739445988, 0],
					[0.03999812963,0.02536259259,0.06864603704,0.09002012963,0.1015344074,0.02960762963,0.02913762963,0.1467278889,0.05284057407, 0],
					[0.0361509784,0.02083593827,0.0420215216,0.03664589506,0.04497232099,0.02165332099,0.01739445988,0.05284057407,0.07926979321, 0],
					[0, 0, 0, 0, 0, 0, 0, 0, 0, 0]];
		var returns = [0.06594444444,0.06155555556,0.1460555556,0.1734444444,0.1981111111,0.05511111111,0.1276111111,0.1903333333,0.1156111111, 0];
		
		//
		var expectedRoundedWeights = [0, 0, 0.08, 0.015, 0.06, 0, 0.345, 0, 0, 0.5];	
		var weights = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { constraints: {return: 0.07}});
		var weightsCla = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { optimizationMethod: "critical-line", constraints: {return: 0.07}});
		var weightsGsmo = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, {optimizationMethod: "gsmo", constraints: {return: 0.07}});
		var roundedWeights = weights.map(x => Math.round(x *200)/200);
		var roundedWeightsCla = weightsCla.map(x => Math.round(x *200)/200);
		var roundedWeightsGsmo = weightsGsmo.map(x => Math.round(x *200)/200);
		assert.equal(areWeightsIdentical(roundedWeights, expectedRoundedWeights, 0), true, 'Return constraint #1, AUTO');
		assert.equal(areWeightsIdentical(roundedWeightsCla, expectedRoundedWeights, 0), true, 'Return constraint #1, CLA');
		assert.equal(areWeightsIdentical(roundedWeightsGsmo, expectedRoundedWeights, 0), true, 'Return constraint #1, GSMO');

		//
		var expectedRoundedWeights = [0, 0, 0, 0.335, 0.455, 0, 0.21, 0, 0, 0]; // In Markowitz book, this is actually [0, 0, 0, 0.33, 0.455, 0, 0.215, 0, 0, 0], but this seems a wrong rounding 
		var weights = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { constraints: {return: 0.175}});
		assert.equal(areWeightsIdentical(weights, expectedRoundedWeights, 1e-3), true, 'Return constraint #2');

		//
		var expectedRoundedWeights = [0, 0, 0, 0.3, 0.39, 0, 0.31, 0, 0, 0]; // In Markowitz book, this is actually [0, 0, 0, 0.29, 0.39, 0, 0.32, 0, 0, 0], but this seems a wrong rounding 		
		var weights = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { constraints: {return: 0.169}});
		assert.equal(areWeightsIdentical(weights, expectedRoundedWeights, 1e-2), true, 'Return constraint #3');
		
		//
		var expectedRoundedWeights = [0, 0, 0.09, 0.13, 0.21, 0, 0.57, 0, 0, 0];		
		var weights = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { constraints: {return: 0.150}});
		assert.equal(areWeightsIdentical(weights, expectedRoundedWeights, 1e-2), true, 'Return constraint #4');
	}
	
	// Reference: Understanding the Impact of Weights Constraints in Portfolio Theory, Thierry Roncalli
	{
		var covMat = [[0.0225,0.0030,0.0150,0.0225], [0.0030,0.0400,0.0350,0.0240], [0.0150,0.0350,0.0625,0.0600], [0.0225,0.0240,0.0600,0.0900]];
		var returns = [5/100, 3/100, 7/100, 7/100];
		
		var expectedWeights =  [0.4, 0.05, 0.4, 0.15];
		var weights = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { constraints: {return: 6/100, maxWeights: [0.4, 0.4, 0.4, 0.4]}}); 
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-5), true, 'Return constraint #5');
	}
});	


QUnit.test('Mean variance portfolio - volatility constraint', function(assert) {    
	function generateRandomValue(minVal, maxVal) {	
		return Math.random() * (maxVal - minVal) + minVal;
	}
	
	// Test using static data
	// Test unreachable cases
	{
		var covMat = [[0.0146, 0.0187, 0.0145],
					 [0.0187, 0.0854, 0.0104],
					  [0.0145, 0.0104, 0.0289]];
		var returns = [0.062, 0.146, 0.128];
		
		
		var nbSamples = 5;
		var minVolatility = 0.12082760588883482; // computed thanks to the efficient frontier
		var maxVolatility = Math.sqrt(covMat[1][1]);
		for (var k = 0; k < nbSamples; ++k) {
			var unreachableMaxVolatility = generateRandomValue(maxVolatility - 1e-6, 2 * maxVolatility); // > maxVolatility
			assert.throws(function() { 
			PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { optimizationMethod: "critical-line", constraints: {volatility: unreachableMaxVolatility}}) },
				new Error('no matching efficient portfolio with a volatility equal to ' + unreachableMaxVolatility),
				"Volatility constraint weights portfolio, unreachable volatility #1/1 - " + k + "/" + nbSamples);
				
			assert.throws(function() { 
			PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { constraints: {volatility: unreachableMaxVolatility}}) },
				new Error('no matching efficient portfolio with a volatility equal to ' + unreachableMaxVolatility),
				"Volatility constraint weights portfolio, unreachable volatility #1/2 - " + k + "/" + nbSamples);

			var unreachableMinVolatility = generateRandomValue(0, minVolatility - 1e-6); // < minVolatility
			assert.throws(function() { 
			PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { optimizationMethod: "critical-line", constraints: {volatility: unreachableMinVolatility}}) },
				new Error('no matching efficient portfolio with a volatility equal to ' + unreachableMinVolatility),
				"Volatility constraint weights portfolio, unreachable volatility #2/1 - " + k + "/" + nbSamples);
			assert.throws(function() { 
			PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { constraints: {volatility: unreachableMinVolatility}}) },
				new Error('no matching efficient portfolio with a volatility equal to ' + unreachableMinVolatility),
				"Volatility constraint weights portfolio, unreachable volatility #2/2 - " + k + "/" + nbSamples);
		}
	}
	
	// Test using static data (positive semi-definite covariance matrix)
	// Reference: Portfolio Selection, H. Markowitz example, chapter II "Illustrative portfolio analyses"
	{	
		var covMat = [[0.05338816358,0.02149069753,0.02865533642,0.04896485802,0.01624895062,0.03223945062,0.02425553395,0.03999812963,0.0361509784, 0],
					[0.02149069753,0.01468446914,0.01878391358,0.02441658642,0.008041938272,0.01002193827,0.01448993827,0.02536259259,0.02083593827, 0],
					[0.02865533642,0.01878391358,0.08550016358,0.06260714198,0.04439938272,0.01328671605,0.01043991049,0.06864603704,0.0420215216, 0],
					[0.04896485802,0.02441658642,0.06260714198,0.09546446914,0.05153806173,0.02902461728,0.02077028395,0.09002012963,0.03664589506, 0],
					[0.01624895062,0.008041938272,0.04439938272,0.05153806173,0.1278900988,0.0128384321,0.02091715432,0.1015344074,0.04497232099, 0],
					[0.03223945062,0.01002193827,0.01328671605,0.02902461728,0.0128384321,0.04125832099,0.01127854321,0.02960762963,0.02165332099, 0],
					[0.02425553395,0.01448993827,0.01043991049,0.02077028395,0.02091715432,0.01127854321,0.02883379321,0.02913762963,0.01739445988, 0],
					[0.03999812963,0.02536259259,0.06864603704,0.09002012963,0.1015344074,0.02960762963,0.02913762963,0.1467278889,0.05284057407, 0],
					[0.0361509784,0.02083593827,0.0420215216,0.03664589506,0.04497232099,0.02165332099,0.01739445988,0.05284057407,0.07926979321, 0],
					[0, 0, 0, 0, 0, 0, 0, 0, 0, 0]];
		var returns = [0.06594444444,0.06155555556,0.1460555556,0.1734444444,0.1981111111,0.05511111111,0.1276111111,0.1903333333,0.1156111111, 0];
		
		//
		var expectedRoundedWeights = [0, 0, 0.08, 0.015, 0.055, 0, 0.335, 0, 0, 0.515]; // In Markowitz book, this is actually [0, 0, 0.08, 0.015, 0.06, 0, 0.345, 0, 0, 0.5], but this seems a wrong rounding 		
		var weights = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { constraints: {volatility: 0.08}});
		var roundedWeights = weights.map(x => Math.round(x *200)/200);
		assert.equal(areWeightsIdentical(roundedWeights, expectedRoundedWeights, 0), true, 'Volatility constraint #1');
	}
	
	// Test using data for equal returns
	// https://github.com/lequant40/portfolio_allocation_js/issues/8
	{
		var covMat = [[2.3/100, 2.5/100, 3.3/100, 1.7/100, 2.3/100, -0.4/100, -0.3/100, -0.3/100, 0.1/100, 1.7/100],
					[2.5/100, 2.9/100, 3.8/100, 1.9/100, 2.5/100, -0.4/100, -0.3/100, -0.3/100, 0.2/100, 1.9/100],
					[3.3/100, 3.8/100, 5.3/100, 2.3/100, 3.1/100, -0.5/100, -0.3/100, -0.4/100, 0.3/100, 2.9/100],
					[1.7/100, 1.9/100, 2.3/100, 1.4/100, 1.8/100, -0.3/100, -0.3/100, -0.3/100, 0.0/100, 1.1/100],
					[2.3/100, 2.5/100, 3.1/100, 1.8/100, 4.0/100, -0.1/100, 0.0/100, 0.2/100, 0.7/100, 1.8/100],
					[-0.4/100, -0.4/100, -0.5/100, -0.3/100, -0.1/100, 0.2/100, 0.2/100, 0.2/100, 0.2/100, -0.1/100],
					[-0.3/100, -0.3/100, -0.3/100, -0.3/100, 0.0/100, 0.2/100, 0.4/100, 0.2/100, 0.3/100, 0.0/100],
					[-0.3/100, -0.3/100, -0.4/100, -0.3/100, 0.2/100, 0.2/100, 0.2/100, 0.3/100, 0.2/100, 0.0/100],
					[0.1/100, 0.2/100, 0.3/100, 0.0/100, 0.7/100, 0.2/100, 0.3/100, 0.2/100, 0.5/100, 0.4/100],
					[1.7/100, 1.9/100, 2.9/100, 1.1/100, 1.8/100, -0.1/100, 0.0/100, 0.0/100, 0.4/100, 2.0/100]];
		var returns = [4.7/100, 5.0/100, 6.1/100, 4.0/100, 5.4/100, 2.0/100, 2.2/100, 2.4/100, 2.4/100, 3.6/100];
		var minWeights = [3/100, 3/100, 3/100, 0/100, 0/100, 0/100, 0/100, 3/100, 0/100, 0/100];
		var maxWeights = [35/100, 35/100, 35/100, 35/100, 35/100, 35/100, 35/100, 35/100, 35/100, 35/100];
		
		//
		var expectedWeights = [0.030000000000000002,0.030000000000000002,0.10234790898307598,0.255841446754833,0,0.0828589119927753,0.1489517322693158,0.35,0, 0];
		var weights = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, {constraints: {minWeights: minWeights, maxWeights: maxWeights, volatility: 5.5/100}});
		var weightsGsmo = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, {optimizationMethod: "gsmo", constraints: {minWeights: minWeights, maxWeights: maxWeights, volatility: 5.5/100}});		
		var weightsCla = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, {optimizationMethod: "critical-line", constraints: {minWeights: minWeights, maxWeights: maxWeights, volatility: 5.5/100}});
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-8), true, 'Volatility constraint #2, AUTO');
		assert.equal(areWeightsIdentical(weightsCla, expectedWeights, 1e-8), true, 'Volatility constraint #2, CLA');
		assert.equal(areWeightsIdentical(weightsGsmo, expectedWeights, 1e-3), true, 'Volatility constraint #2, GSMO');
	}
});	


QUnit.test('Mean variance portfolio - risk tolerance constraint', function(assert) {    
	// Test using static data
	// Reference: https://web.stanford.edu/~wfsharpe/mia/opt/mia_opt3.htm
	// Note: this test also allows checking that numerically equal corner portfolios are NOT filtered out by the internal algorithm,
	// because otherwise portfolios with a risk tolerance between 11.1475 and 11.47 would not be constant.
	{	
		var covMat = [[1, 2.96, 2.31],
					[2.96, 54.76, 39.886],
					[2.31,	39.886,	237.16]];
		var returns = [2.8000, 6.3000, 10.8000];
		
		// Test a risk tolerance greater than the highest attainable risk tolerance, which is 20.898844444444443,
		// which must output the rightmost corner portfolio.
		var expectedWeights = [0.2, 0.3, 0.5];
		var weights = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { constraints: {riskTolerance: 40, 
		                                                                                                   minWeights: [0.2, 0.2, 0.2], maxWeights: [0.5, 0.5, 0.5]} });
		var weightsCla = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { optimizationMethod: "critical-line",
		                                                                                        constraints: {riskTolerance: 40, 
		                                                                                                      minWeights: [0.2, 0.2, 0.2], maxWeights: [0.5, 0.5, 0.5]} });
		var weightsGsmo = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { optimizationMethod: "gsmo",
		                                                                                         constraints: {riskTolerance: 40, 
		                                                                                                       minWeights: [0.2, 0.2, 0.2], maxWeights: [0.5, 0.5, 0.5]} });
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-8), true, 'Mean variance portfolio - risk tolerance constraint, rightmost corner portfolio, AUTO');
		assert.equal(areWeightsIdentical(weightsCla, expectedWeights, 1e-8), true, 'Mean variance portfolio - risk tolerance constraint, rightmost corner portfolio, CLA');
		assert.equal(areWeightsIdentical(weightsGsmo, expectedWeights, 1e-8), true, 'Mean variance portfolio - risk tolerance constraint, rightmost corner portfolio, GSMO');
		
		// Test a risk tolerance equal to 0, which must output the leftmost corner portfolio.
		var expectedWeights = [0.5, 0.3, 0.2];
		var weights = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { optimizationMethod: "critical-line",
		                                                                                     constraints: {riskTolerance: 0, 
		                                                                                                   minWeights: [0.2, 0.2, 0.2], maxWeights: [0.5, 0.5, 0.5]} });
		var weightsCla = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { optimizationMethod: "critical-line",
		                                                                                        constraints: {riskTolerance: 0, 
		                                                                                                      minWeights: [0.2, 0.2, 0.2], maxWeights: [0.5, 0.5, 0.5]} });
		var weightsGsmo = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { optimizationMethod: "gsmo",
		                                                                                         constraints: {riskTolerance: 0, 
		                                                                                                       minWeights: [0.2, 0.2, 0.2], maxWeights: [0.5, 0.5, 0.5]} });
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-8), true, 'Mean variance portfolio - risk tolerance constraint, leftmost corner portfolio, AUTO');
		assert.equal(areWeightsIdentical(weightsCla, expectedWeights, 1e-8), true, 'Mean variance portfolio - risk tolerance constraint, leftmost corner portfolio, CLA');
		assert.equal(areWeightsIdentical(weightsGsmo, expectedWeights, 1e-8), true, 'Mean variance portfolio - risk tolerance constraint, leftmost corner portfolio, GSMO');

		
		// Test a risk tolerance between 11.1475 and 11.47, which is a segment inside which all efficient portfolios
		// must have the same weights due to a kink.
		var expectedWeights = [0.2, 0.5, 0.3];
		var weights = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { constraints: {riskTolerance: 11.3, 
		                                                                                                   minWeights: [0.2, 0.2, 0.2], maxWeights: [0.5, 0.5, 0.5]} });
		var weightsCla = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { optimizationMethod: "critical-line", 
		                                                                                        constraints: {riskTolerance: 11.3, 
		                                                                                                      minWeights: [0.2, 0.2, 0.2], maxWeights: [0.5, 0.5, 0.5]} });
		var weightsGsmo = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { optimizationMethod: "gsmo",
		                                                                                         constraints: {riskTolerance: 11.3, 
		                                                                                                       minWeights: [0.2, 0.2, 0.2], maxWeights: [0.5, 0.5, 0.5]} });
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-8), true, 'Mean variance portfolio - risk tolerance constraint, kink, AUTO');
		assert.equal(areWeightsIdentical(weightsCla, expectedWeights, 1e-8), true, 'Mean variance portfolio - risk tolerance constraint, kink, CLA');
		assert.equal(areWeightsIdentical(weightsGsmo, expectedWeights, 1e-8), true, 'Mean variance portfolio - risk tolerance constraint, kink, GSMO');
	}
	
	// Test using partially random data
	// Reference: Portfolio Selection, H. Markowitz example, chapter VIII, section "The computing procedure"
	{
		var covMat = [[0.0146, 0.0187, 0.0145],
					  [0.0187, 0.0854, 0.0104],
					  [0.0145, 0.0104, 0.0289]];
		var returns = [0.062, 0.146, 0.128];

		var riskTolerance = Math.random() * 4.16666666666667;
		
		// From rt = 4.16666666666667 to rt = 0.1408064320019454,
		// the equation of the critical line is (x1,x2,x3) = (0, 0.198 + rt*0.193, 0.802 - rt*0.193)
		//
		// From rt = 0.1408064320019454 to rt = 0.03332764893133244,
		// the equation of the critical line is (x1,x2,x3) = (1.10 - rt*7.83, -0.07 + rt*2.09, -0.03 + rt*5.73)
		//
		// From rt = 0.03332764893133244 to rt = 0,
		// the equation of the critical line is (x1,x2,x3) = (0.9931 -rt*4.551, 0, 0.0069 + rt*4.552)
		var expectedWeights;
		if (0.1408064320019454 <= riskTolerance && riskTolerance <= 4.16666666666667) {
			expectedWeights = [0, 0.198 + riskTolerance*0.193, 0.802 - riskTolerance*0.193];
		}
		else if (0.03332764893133244 <= riskTolerance && riskTolerance <= 0.1408064320019454) {
			expectedWeights = [1.10 - riskTolerance*7.83, -0.07 + riskTolerance*2.09, -0.03 + riskTolerance*5.73];
		}
		else if (0 <= riskTolerance && riskTolerance <= 0.03332764893133244) {
			expectedWeights = [0.9931 -riskTolerance*4.551, 0, 0.0069 + riskTolerance*4.552];
		}
		else {
			throw new Error("internal error: incorrect risk tolerance generated")
		}
		var weights = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { constraints: {riskTolerance: riskTolerance} });
		var weightsCla = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { optimizationMethod: "critical-line", constraints: {riskTolerance: riskTolerance} });
		var weightsGsmo = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { optimizationMethod: "gsmo", constraints: {riskTolerance: riskTolerance} });
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-2), true, 'Mean variance portfolio - risk tolerance constraint, exact value, AUTO');
		assert.equal(areWeightsIdentical(weightsCla, expectedWeights, 1e-2), true, 'Mean variance portfolio - risk tolerance constraint, exact value, CLA');
		assert.equal(areWeightsIdentical(weightsGsmo, expectedWeights, 1e-2), true, 'Mean variance portfolio - risk tolerance constraint, exact value, GSMO');
	}
	
	// Test using static data
	// Reference: Chopra, Vijay K; Ziemba, William T, The effect of errors in means, variances, and covariances on optimal portfolio choice
	{
		// Note: the covariance matrix below is the result of Diag(STD)*CORR*Diag(STD), because
		// the covariance matrix is not directly provided in the reference paper.
		var ret = [1.5617, 1.9477, 1.907, 1.5801, 2.1643, 1.6010, 1.4892, 1.6248, 1.4075, 1.4537];
		/*
		var std = [8.8308, 8.4585, 10.040, 8.6215, 5.9886, 6.8767, 5.8162, 5.6385, 8.0047, 8.2125];
		var corr = [[1,      0.3660, 0.3457, 0.1606, 0.2279, 0.5133, 0.5203, 0.2176, 0.3267, 0.5101],
				   [0.3660, 1,      0.5379, 0.2165, 0.4986, 0.5823, 0.5569, 0.4760, 0.6517, 0.5853],
				   [0.3457, 0.5379, 1,      0.2218, 0.4283, 0.4051, 0.4492, 0.3867, 0.4883, 0.6569],
				   [0.1606, 0.2165, 0.2218, 1,      0.0569, 0.3609, 0.2325, 0.2289, 0.1726, 0.3814],
				   [0.2279, 0.4986, 0.4283, 0.0569, 1,      0.3619, 0.4811, 0.5952, 0.4378, 0.4368],
				   [0.5133, 0.5823, 0.4051, 0.3609, 0.3619, 1,      0.6167, 0.4996, 0.5811, 0.5644],
				   [0.5203, 0.5569, 0.4492, 0.2325, 0.4811, 0.6167, 1,      0.6037, 0.5671, 0.6032],
				   [0.2176, 0.4760, 0.3867, 0.2289, 0.5952, 0.4996, 0.6037, 1,      0.5012, 0.4772],
				   [0.3267, 0.6517, 0.4883, 0.1726, 0.4378, 0.5811, 0.5671, 0.5012, 1,      0.6039],
				   [0.5101, 0.5853, 0.6569, 0.3814, 0.4368, 0.5644, 0.6032, 0.4772, 0.6039, 1]];
	     */
		var cov = [[77.98302864,      27.3384877788 , 30.650187902399995 ,12.227239597319999  ,  12.052292971752, 31.171047119387996  ,  26.723491968888 ,10.834840558079998, 23.093738485091997  ,    36.9939542445],
				   [27.338487778800005,  71.54622225000001,       45.680264586, 15.788253352875001  ,   25.25637014766 , 33.87039193498501  ,   27.39743489613    ,   22.701988071  ,  44.125143900915, 40.658116910625004],
				   [30.6501879024,       45.680264586, 100.80159999999998    ,   19.198976948   ,   25.7517704952, 27.968941746799995  ,    26.2308758816, 21.891295817999996, 39.243297900399995    ,    54.16370415],
				   [12.227239597319999 ,   15.788253352875, 19.198976947999995 , 74.33026224999999 ,     2.93778767781, 21.396847580144996, 11.658565629749999, 11.127361821974999, 11.911561133229998 ,    27.00467182125],
				   [12.052292971752001 ,    25.25637014766,      25.7517704952 ,     2.93778767781  ,      35.86332996 ,   14.903695453878 ,   16.757143738452, 20.097952398719997 ,20.986795142675998, 21.482425692000003],
				   [31.171047119388  ,  33.870391934985, 27.968941746799995, 21.396847580144996, 14.903695453877999 , 47.28900288999999  ,  24.665695108418 ,19.371626765819997, 31.987184396738993, 31.874432854499997],
				   [26.723491968888  ,   27.39743489613  ,    26.2308758816, 11.658565629749999  ,  16.757143738452  ,  24.665695108418, 33.828182440000006 ,    19.79812640169 ,26.402438484994004   ,    28.812175236],
				   [10.834840558079998 ,22.701988071000002, 21.891295817999996 ,11.127361821974999, 20.097952398719997, 19.371626765819997 ,    19.79812640169, 31.792682249999995, 22.621411876139998 ,     22.0973096925],
				   [23.093738485091997 ,   44.125143900915 ,39.243297900399995 ,11.911561133229998 ,   20.986795142676, 31.987184396738993, 26.402438484994004, 22.621411876139998  ,      64.07522209, 39.699539785125005],
				   [36.9939542445 ,40.658116910625004  ,      54.16370415  ,   27.00467182125    ,   21.482425692   ,   31.8744328545   ,    28.812175236  ,    22.0973096925  ,  39.699539785125 , 67.44515625000001]];

		// In the reference paper, the risk tolerance is 50, which corresponds to 25 for this implementation of the MVO algorithm due to the /2 coefficient on the variance term.
		var expectedRoundedWeights = [0.0350, 0.0082, 0, 0.1626, 0.7940, 0, 0, 0, 0, 0]
		var weights = PortfolioAllocation.meanVarianceOptimizationWeights(ret, cov, { constraints: {riskTolerance: 25} });
		var roundedWeights = weights.map(x => Math.round(x *10000)/10000);
		assert.equal(areWeightsIdentical(roundedWeights, expectedRoundedWeights, 1e-4), true, 'Mean variance portfolio - risk tolerance constraint, Chopra-Ziemba example');
	}
});	

QUnit.test('Mean variance portfolio - maximum volatility constraint', function(assert) {    
	
	
	function generateRandomValue(minVal, maxVal) {	
		return Math.random() * (maxVal - minVal) + minVal;
	}
	
	// Test using semi random data
	{
		var covMat = [[0.0146, 0.0187, 0.0145],
					 [0.0187, 0.0854, 0.0104],
					  [0.0145, 0.0104, 0.0289]];
		var returns = [0.062, 0.146, 0.128];
		
		var covMatPartial = [[0.0146, 0.0187, 0.0145, 0],
							[0.0187, 0.0854, 0.0104, 0],
							[0.0145, 0.0104, 0.0289, 0],
							[0, 0, 0, 0]];
		var returnsPartial = [0.062, 0.146, 0.128, 0];
		
		var nbSamples = 5;
		var minAttainableVolatility = 0.12082760588883482; // computed thanks to the efficient frontier
		var maxAttainableVolatility = Math.sqrt(covMat[1][1]);
		for (var k = 0; k < nbSamples; ++k) {	
			// In case the maximum target volatility is lower than the minimum attainable volatility,
			// the portfolio cannot be fully invested and so must correspond to the efficient portfolio with cash
			// and with a target volatility constraint 
			var maxTargetVolatility = generateRandomValue(0, minAttainableVolatility - 1e-6); // < minAttainableVolatility
			assert.throws(function() { 
				PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat,
			                                                        { constraints: {maxVolatility: maxTargetVolatility}}); 
			 },
			new Error('no matching efficient portfolio with a volatility lower than or equal to ' + maxTargetVolatility),
			"Error case - Too low volatility #1" + k);

			var weights_mtv = PortfolioAllocation.meanVarianceOptimizationWeights(returnsPartial, covMatPartial,
			                                                                      { constraints: {maxVolatility: maxTargetVolatility}});
			var investment = weights_mtv[0] + weights_mtv[1] + weights_mtv[2];
			var cash = weights_mtv[3];
			assert.equal(investment < 1 && cash > 0, true, 'Maximum target volatility weights portfolio #1' + k);
			
			
			// In case the maximum target volatility is greater than the maximum attainable volatility,
			// the portfolio must be the efficient portfolio with the highest return
			var maxTargetVolatility = generateRandomValue(maxAttainableVolatility - 1e-6, 2 * maxAttainableVolatility); // > maxAttainableVolatility
			var weights = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, 
			                                                                  { constraints: {maxVolatility: maxTargetVolatility}});
			assert.equal(areWeightsIdentical(weights, [0, 1, 0], 0), true, 'Maximum target volatility weights portfolio #3' + k);
			
			// Otherwise, the portfolio must correspond to the efficient portfolio with a target volatility constraint
			var maxTargetVolatility = generateRandomValue(minAttainableVolatility + 1e-6, maxAttainableVolatility - 1e-6); // > minAttainableVolatility && < maxAttainableVolatility
			var weights_mtv = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, 
			                                                                      { constraints: {maxVolatility: maxTargetVolatility}});
			var weights_tv = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, 
			                                                                     { constraints: {volatility: maxTargetVolatility}});
			assert.equal(areWeightsIdentical(weights_mtv, weights_tv, 0), true, 'Maximum target volatility weights portfolio #4' + k);
		}
	}
	
	// Test using static data
	// Test that a target maximum volatility of zero produces a zero-weights portfolio
	{
		// Problem data
		var covMat =[[0.0146, 0.0187, 0.0145, 0],
					[0.0187, 0.0854, 0.0104, 0],
					[0.0145, 0.0104, 0.0289, 0],
					[0, 0, 0, 0]];
		var returns = [0.062, 0.146, 0.128, 0];
		
		// Compute the desired maximum volatility
		var maxVolatility = 0;
		
		// Test that a 0 maximum volatility target is properly supported when partial investment is allowed
		var expectedWeights = [0, 0, 0, 1];
		var weights = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { constraints: {maxVolatility: maxVolatility}});
		var weightsCla = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { optimizationMethod: "critical-line", constraints: {maxVolatility: maxVolatility}});
		var weightsGsmo = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { optimizationMethod: "gsmo", constraints: {maxVolatility: maxVolatility}});

		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-6), true, 'Maximum target volatility weights portfolio #4, AUTO');
		assert.equal(areWeightsIdentical(weightsCla, expectedWeights, 1e-6), true, 'Maximum target volatility weights portfolio #4, CLA');
		assert.equal(areWeightsIdentical(weightsGsmo, expectedWeights, 1e-6), true, 'Maximum target volatility weights portfolio #4, GSMO');
	}
	
	
	// Test using static data
	// Test the algorithm when all returns are negative; the associated portfolio must be null.
	{
		// Problem data
		var covMat = [[0.002533106361, 0.002161135405, 0.002605003686, 0],
                     [0.002161135405, 0.002252008852, 0.002439575054, 0],
                     [0.002605003686, 0.002439575054, 0.003315580611, 0],
					 [0, 0, 0, 0]];
		var returns = [-0.0000244756209, -0.003612424187, -0.004045817996, 0];
		
		// Compute the desired maximum volatility
		var maxVolatility = 0.10/Math.sqrt(12);
		
		// Test that the algorithm is behaving properly
		var expectedWeights = [0, 0, 0, 1];
		var weights = PortfolioAllocation.meanVarianceOptimizationWeights(returns, covMat, { constraints: {maxVolatility: maxVolatility}});
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-6), true, 'Test negative return');
	}	
});	


QUnit.test('Mean variance portfolio - efficient portfolios computation', function(assert) {    
	function areEfficientPortfoliosIdentical(efficientPortfoliosa, efficientPortfoliob, tol) {
		//
		if (efficientPortfoliosa.length != efficientPortfoliob.length) {
			return false;
		}
		
		//
		for (var i = 0; i < efficientPortfoliosa.length; ++i) {
			var weightsa = efficientPortfoliosa[i][0];
			var reta = efficientPortfoliosa[i][1];
			var vola = efficientPortfoliosa[i][2];
			
			var weightsb = efficientPortfoliob[i][0];
			var retb = efficientPortfoliob[i][1];
			var volb = efficientPortfoliob[i][2];
			
			if (!areWeightsIdentical(weightsa, weightsb, tol)) {
				return false;
			}
			
			if (Math.abs(reta - retb) > tol) {
				return false;
			}
			
			if (Math.abs(vola - volb) > tol) {
				return false;
			}
		}
		
		//
		return true;
	}
	
	// Test using static data
	// Test the limit case of only one corner portfolio
	{
		// Lower bounds binding (sum lb_i == 1)
		var expectedEfficientPortfolios = [[[0.4, 0.6], 0.16, 0.7211102550927979], [[0.4, 0.6], 0.16, 0.7211102550927979]]; // exactly the same portfolios !

		var efficientPortfolios = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios([0.1, 0.2], [[1,0],[0,1]], { nbPortfolios: 2, constraints: {minWeights: [0.4, 0.6]} });
		var efficientPortfoliosCla = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios([0.1, 0.2], [[1,0],[0,1]], { optimizationMethod: "critical-line", nbPortfolios: 2, constraints: {minWeights: [0.4, 0.6]} });
		var efficientPortfoliosGsmo = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios([0.1, 0.2], [[1,0],[0,1]], { optimizationMethod: "gsmo", nbPortfolios: 2, constraints: {minWeights: [0.4, 0.6]} });
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfolios, expectedEfficientPortfolios, 0), true, 'Efficient portfolios, lower bounds binding OK #1, AUTO');
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfoliosCla, expectedEfficientPortfolios, 0), true, 'Efficient portfolios, lower bounds binding OK #1, CLA');
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfoliosGsmo, expectedEfficientPortfolios, 0), true, 'Efficient portfolios, lower bounds binding OK #1, GSMO');
		
		var efficientPortfolios = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios([0.1, 0.2], [[1,0],[0,1]], { nbPortfolios: 1, constraints: {minWeights: [0.4, 0.6]} });
		var efficientPortfoliosCla = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios([0.1, 0.2], [[1,0],[0,1]], { optimizationMethod: "critical-line", nbPortfolios: 1, constraints: {minWeights: [0.4, 0.6]} });
		var efficientPortfoliosGsmo = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios([0.1, 0.2], [[1,0],[0,1]], { optimizationMethod: "gsmo", nbPortfolios: 1, constraints: {minWeights: [0.4, 0.6]} });
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfolios, [expectedEfficientPortfolios[0]], 0), true, 'Efficient portfolios, lower bounds binding OK #2, AUTO');
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfoliosCla, [expectedEfficientPortfolios[0]], 0), true, 'Efficient portfolios, lower bounds binding OK #2, CLA');
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfoliosGsmo, [expectedEfficientPortfolios[0]], 0), true, 'Efficient portfolios, lower bounds binding OK #2, GSMO');
	}

	// Test using static data
	// Test the limit case of only one efficient portfolio to be computed
	{
		var covMat = [[0.0146, 0.0187, 0.0145],
					 [0.0187, 0.0854, 0.0104],
					  [0.0145, 0.0104, 0.0289]];
		var returns = [0.062, 0.146, 0.128];
		
		var expectedEfficientPortfolios = [[[0.39289338965447984, 0.11991944022734498, 0.48718717011817525], 0.10422758620689655, 0.13736661084455756]];
		
		var efficientPortfolio = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios(returns, covMat, { nbPortfolios: 1 });
		var efficientPortfolioCla = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios(returns, covMat, { optimizationMethod: "critical-line", nbPortfolios: 1 });
		var efficientPortfolioGsmo = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios(returns, covMat, { optimizationMethod: "gsmo", nbPortfolios: 1 });
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfolio, expectedEfficientPortfolios, 1e-8), true, 'Efficient portfolios, one efficient portfolio, AUTO');
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfolioCla, expectedEfficientPortfolios, 1e-8), true, 'Efficient portfolios, one efficient portfolio, CLA');
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfolioGsmo, expectedEfficientPortfolios, 1e-4), true, 'Efficient portfolios, one efficient portfolio, GSMO');
	}
	
	// Test using static data that the different types of discretization are properly managed
	{		
		var covMat = [[0.0146, 0.0187, 0.0145],
					 [0.0187, 0.0854, 0.0104],
					  [0.0145, 0.0104, 0.0289]];
		var returns = [0.062, 0.146, 0.128];
		
		// Test that by default, 100 efficient portfolios are computed
		var efficientPortfolios = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios(returns, covMat);
		assert.equal(efficientPortfolios.length, 100, 'Efficient portfolios, default number of efficient portfolios computed');
		
		// Test a small number of efficient portfolios
		//
		var expectedEfficientPortfoliosRiskTolerance = [[[0.9931034482758623, 0, 0.006896551724137813], 0.0624551724137931, 0.12082760588883482], 
														[[0, 0.3983957219251336, 0.6016042780748664], 0.1351711229946524, 0.1702926860745384],
														[[0, 0.5989304812834224, 0.40106951871657753], 0.13878074866310158, 0.20069797992103713], 
														[[0,0.7994652406417112,0.20053475935828877],0.1423903743315508,0.2430633926246908], 
														[[0, 1, 0], 0.146, 0.29223278392404917]];
														
		var expectedEfficientPortfoliosRiskToleranceNonCla = [[[0.9931012778446591,0,0.006898722155340931],0.0624553156622525,0.12082760588911746],
		                                                      [[0,1,0],0.146,0.29223278392404917],
						                                      [[0,1,0],0.146,0.29223278392404917],
															  [[0,1,0],0.146,0.29223278392404917],
															  [[0,1,0],0.146,0.29223278392404917]]; // The duplication is not a mistake
		var efficientPortfoliosRiskTolerance = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios(returns, covMat, { discretizationType: "riskTolerance", nbPortfolios: 5 });
		var efficientPortfoliosRiskToleranceCla = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios(returns, covMat, { optimizationMethod: "critical-line", discretizationType: "riskTolerance", nbPortfolios: 5 });
		var efficientPortfoliosRiskToleranceGsmo = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios(returns, covMat, { optimizationMethod: "gsmo", discretizationType: "riskTolerance", nbPortfolios: 5 });
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfoliosRiskTolerance, expectedEfficientPortfoliosRiskTolerance, 1e-12), true, 'Risk tolerance discretization parameter, AUTO');
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfoliosRiskToleranceCla, expectedEfficientPortfoliosRiskTolerance, 1e-12), true, 'Risk tolerance discretization parameter, CLA');
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfoliosRiskToleranceGsmo, expectedEfficientPortfoliosRiskToleranceNonCla, 1e-8), true, 'Risk tolerance discretization parameter, GSMO');
		
		//
		var expectedEfficientPortfoliosReturn = [[[0.993103448275862, 0, 0.006896551724137945], 0.06245517241379311, 0.1208276058888348], 
													[[0.6878434301262003, 0.0410580943707805, 0.27109847550301913], 0.08334137931034481, 0.12598059118870247],
													[[0.39289338965447984, 0.11991944022734487, 0.48718717011817525], 0.10422758620689654, 0.13736661084455756], 
													[[0.09794334918275893, 0.19878078608390934, 0.7032758647333318],0.12511379310344828, 0.15310742669374616], 
													[[0, 1, 0], 0.146, 0.29223278392404917]];												
		var efficientPortfoliosReturn = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios(returns, covMat, { discretizationType: "return", nbPortfolios: 5 });
		var efficientPortfoliosReturnCla = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios(returns, covMat, { optimizationMethod: "critical-line", discretizationType: "return", nbPortfolios: 5 });
		var efficientPortfoliosReturnGsmo = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios(returns, covMat, { optimizationMethod: "gsmo", discretizationType: "return", nbPortfolios: 5 });											
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfoliosReturn, expectedEfficientPortfoliosReturn, 1e-12), true, 'Return discretization parameter, AUTO');
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfoliosReturnCla, expectedEfficientPortfoliosReturn, 1e-12), true, 'Return discretization parameter, CLA');
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfoliosReturnGsmo, expectedEfficientPortfoliosReturn, 1e-4), true, 'Return discretization parameter, GSMO');
		

		//
		var efficientPortfoliosDefault = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios(returns, covMat, { nbPortfolios: 5 });	
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfoliosDefault, expectedEfficientPortfoliosReturn, 1e-12), true, 'Default discretization parameter, AUTO');

		
		//
		var expectedEfficientPortfoliosVolatility = [[[0.993103448275862,0,0.006896551724137945],0.06245517241379311,0.1208276058888348],
		                                                [[0,0.3266650148358897,0.6733349851641103],0.133879970267046,0.16367890039763838],
														[[0,0.6294374683923241,0.3705625316076758],0.13932987443106182,0.20653019490644195],
														[[0,0.8265133287796433,0.17348667122035663],0.14287723991803356,0.24938148941524554],
														[[0,1,0],0.146,0.29223278392404917]];												
		var efficientPortfoliosVolatility = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios(returns, covMat, { discretizationType: "volatility", nbPortfolios: 5 });
		var efficientPortfoliosVolatilityCla = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios(returns, covMat, { optimizationMethod: "critical-line", discretizationType: "volatility", nbPortfolios: 5 });
		var efficientPortfoliosVolatilityGsmo = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios(returns, covMat, { optimizationMethod: "gsmo", discretizationType: "volatility", nbPortfolios: 5 });
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfoliosVolatility, expectedEfficientPortfoliosVolatility, 1e-12), true, 'Volatility discretization parameter, AUTO');
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfoliosVolatilityCla, expectedEfficientPortfoliosVolatility, 1e-12), true, 'Volatility discretization parameter, CLA');
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfoliosVolatilityGsmo, expectedEfficientPortfoliosVolatility, 1e-5), true, 'Volatility discretization parameter, GSMO');
	}
	
	// Test using static data that the partial investment constraint and constraints on weights are properly handled
	{		
		var covMat = [[0.0146, 0.0187, 0.0145, 0],
					 [0.0187, 0.0854, 0.0104, 0],
					  [0.0145, 0.0104, 0.0289, 0],
					  [0 , 0, 0 ,0]];
		var returns = [0.062, 0.146, 0.128, 0];
		
		var expectedEfficientPortfolios = [[[0.1,0.2,0, 0.7],0.0354,0.06565059024867942],
										  [[0.1,0.2,0.19609375, 0.50390625],0.0605,0.08249670662519751],
										  [[0.1,0.2,0.3921875000000001, 0.3078124999999999],0.08560000000000001,0.10734981446661017],
										  [[0.1,0.2,0.58828125, 0.1117187499999999],0.11069999999999999,0.13588534941646635],
										  [[0.1,0.8,0.09999999999999998, 0],0.1358,0.24502448857206088]];
		var efficientPortfolio = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios(returns, covMat, { constraints: {minWeights: [0.1, 0.2, 0, 0], maxWeights: [0.8, 0.8, 0.8, 1]}, nbPortfolios: 5 });
		var efficientPortfoliosCla = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios(returns, covMat, { optimizationMethod: "critical-line", constraints: {minWeights: [0.1, 0.2, 0, 0], maxWeights: [0.8, 0.8, 0.8, 1]}, nbPortfolios: 5 });
		var efficientPortfoliosGsmo = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios(returns, covMat, { optimizationMethod: "gsmo", constraints: {minWeights: [0.1, 0.2, 0, 0], maxWeights: [0.8, 0.8, 0.8, 1]}, nbPortfolios: 5 });
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfolio, expectedEfficientPortfolios, 1e-12), true, 'Partial investment constraint, AUTO');
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfoliosCla, expectedEfficientPortfolios, 1e-12), true, 'Partial investment constraint, CLA');
		assert.equal(areEfficientPortfoliosIdentical(efficientPortfoliosGsmo, expectedEfficientPortfolios, 1e-4), true, 'Partial investment constraint, GSMO');		
	}
	
	// This test is just here to stress the GSMO algorithm, and check the execution time does not diminish
	{
		var covMat = [[0.008626335034, 0.006715399917, 0.007042419992, 0.007048459, 0.0066366781, 0.006594382192, 0.006188482528, 0.006144514577, 0.005581524855, 0.005791371247, 0.006839503165, 0.006026119572, 0.00556042548, 0.004834846375, 0.004511975974, 0.00502810301, 0.004273827084, 0.004391065668, 0.004333287312, 0.005263026397, 0.006093797699, 0.005328759167, 0.004885570623, 0.005006898428, 0.004375058515, 0.004536227401, 0.00382383205, 0.004188885336, 0.004227284984, 0.004480470259, 0.005652719041, 0.004684925202, 0.004824278, 0.004694296476, 0.0038311798, 0.004131720831, 0.003918816918, 0.00363012853, 0.004303746538, 0.004208510197, 0.00500884843, 0.004497613403, 0.003925394458, 0.003826366745, 0.003479102582, 0.003217506668, 0.003092539967, 0.003387487433, 0.003395715859, 0.004000762357, 0.004921900893, 0.004177801985, 0.003654297707, 0.003598637829, 0.00347358152, 0.003243638023, 0.003030619633, 0.003440298944, 0.003931812779, 0.004004083997, 0.00379606579, 0.00338809299, 0.003208075825, 0.002935523324, 0.002579952426, 0.002747787374, 0.002993945136, 0.003157948431, 0.003715079122, 0.003337110603, 0.002848833667, 0.002949900901, 0.002982481082, 0.002356035502, 0.002656567017, 0.002544831954, 0.002931137112, 0.003179676958, 0.00410448382, 0.003094311644, 0.002412252584, 0.002356759221, 0.001942123379, 0.002101900141, 0.002560515946, 0.002299870077, 0.002083781247, 0.002701932892, 0.002757407576, 0.002009863366, 0.001980655134, 0.001876568152, 0.001624975344, 0.001797528616, 0.001651044294, 0.001615803773],
					[0.006715399917, 0.006683590368, 0.006014896374, 0.005908565288, 0.005722001587, 0.005670592412, 0.005404122902, 0.005228753424, 0.004781393518, 0.004867670897, 0.006057294555, 0.005460206075, 0.004834687069, 0.00427960842, 0.004094676601, 0.004430417541, 0.003713099085, 0.003918644507, 0.003805239886, 0.004724829646, 0.005547769669, 0.004641512869, 0.004275172741, 0.004535712695, 0.004002068481, 0.003998855101, 0.003461115736, 0.003868338392, 0.003743206232, 0.00397440804, 0.005132938989, 0.004315534403, 0.004334464152, 0.004243317135, 0.003586062754, 0.003673558324, 0.003599684906, 0.003328464186, 0.003895523432, 0.003933823625, 0.00454823586, 0.004173927655, 0.00350670524, 0.003414971713, 0.003313012908, 0.003017905708, 0.002835420141, 0.00304936103, 0.003091043629, 0.00371479407, 0.004518149563, 0.003944750725, 0.003314415963, 0.00343389213, 0.003213881416, 0.003034389219, 0.002929474526, 0.00334420473, 0.003452367593, 0.003555839981, 0.003597330441, 0.003228763006, 0.002824046695, 0.002719042358, 0.002595143005, 0.00264760704, 0.002929607457, 0.003064194774, 0.003378001969, 0.003206598288, 0.002600477361, 0.002858297419, 0.002724947082, 0.002440528653, 0.002583794193, 0.002526042736, 0.002966731622, 0.003089654503, 0.003539376119, 0.002988376321, 0.002327715382, 0.0023059398, 0.0020451527, 0.001968497738, 0.002535640801, 0.002346819807, 0.002105827515, 0.00248426969, 0.002719471677, 0.002091166113, 0.00204061465, 0.001945355621, 0.00177718764, 0.001833740965, 0.001833126743, 0.001802919713],
					[0.007042419992, 0.006014896374, 0.006439832216, 0.006144856481, 0.005759584222, 0.005845477915, 0.005529484399, 0.005390914697, 0.004908853124, 0.005103830919, 0.006073314853, 0.005499701853, 0.004969641429, 0.004398641598, 0.004134976123, 0.004525198382, 0.003883988522, 0.003985939094, 0.003937507735, 0.004796383717, 0.00554862731, 0.00472970036, 0.004383809539, 0.004507738633, 0.004099612817, 0.004120392418, 0.003465223346, 0.003916806442, 0.003893286734, 0.00427363146, 0.005206899053, 0.004323028253, 0.004343625848, 0.00432711529, 0.003575259214, 0.003722534095, 0.003569795438, 0.003341274835, 0.00395494306, 0.004053391677, 0.0045785797, 0.004162307351, 0.003616132532, 0.003500002623, 0.00324854751, 0.003014847819, 0.002964497291, 0.003052088947, 0.003176130571, 0.003789134736, 0.004571681602, 0.003864248166, 0.003371414581, 0.003366216157, 0.003231814936, 0.003016512653, 0.002922320365, 0.003274787097, 0.003685314298, 0.003723157506, 0.003694396668, 0.003228902136, 0.002950115286, 0.002751010898, 0.002432745554, 0.002705575637, 0.002880348639, 0.002975663554, 0.003502498329, 0.00322837047, 0.002724240891, 0.002751144754, 0.002720678945, 0.00232456853, 0.00253747154, 0.002549888275, 0.002757467982, 0.003098168353, 0.003999211129, 0.002958428789, 0.002268515803, 0.002255255898, 0.00193368615, 0.00201404631, 0.00253937852, 0.002258501074, 0.002197152045, 0.002615141548, 0.002802102426, 0.001993131787, 0.001941433241, 0.001881035982, 0.001629664591, 0.001757523692, 0.001675342356, 0.001705850964],
					[0.007048459, 0.005908565288, 0.006144856481, 0.006721824469, 0.005852814111, 0.005856465086, 0.005599641385, 0.005466069036, 0.005048488162, 0.005242939938, 0.006111949473, 0.005465037035, 0.005129873692, 0.004416730131, 0.004205399686, 0.004458656717, 0.003886593217, 0.003969078367, 0.003980490598, 0.004776577303, 0.005403948677, 0.004734314166, 0.004410779587, 0.004359942384, 0.004043738715, 0.00403985557, 0.003443041304, 0.003961335731, 0.003888049826, 0.004038458787, 0.004970339251, 0.004186782856, 0.004351604312, 0.004236631295, 0.003515810312, 0.003803261423, 0.003629734042, 0.003345952371, 0.003877361097, 0.003927757425, 0.004493401784, 0.004028057711, 0.003580285672, 0.00341839726, 0.003306099986, 0.003046845036, 0.002896100679, 0.00306826418, 0.003125160756, 0.003719892798, 0.004420189562, 0.003735997127, 0.003210599149, 0.003358889658, 0.003149259674, 0.002974762136, 0.002953982871, 0.003297600332, 0.003627668097, 0.003716680489, 0.003460502947, 0.003201833917, 0.002927183154, 0.002807933105, 0.002393663894, 0.002612167028, 0.002891811709, 0.002789078706, 0.00345071268, 0.003015226454, 0.002634119186, 0.002712130997, 0.002692079815, 0.00220023004, 0.002432468817, 0.002479192684, 0.002699152581, 0.002967319487, 0.003811419029, 0.002813450776, 0.002223566, 0.002173940325, 0.001910269587, 0.001976848994, 0.002485448666, 0.002206599569, 0.002148219717, 0.002382785761, 0.002707287202, 0.001804422913, 0.001881753917, 0.001755480548, 0.001556326854, 0.001713310797, 0.001644961803, 0.001681966033],
					[0.0066366781, 0.005722001587, 0.005759584222, 0.005852814111, 0.005804198968, 0.005518711262, 0.005212667932, 0.00520579585, 0.004695260663, 0.004884578153, 0.00563581538, 0.0051420724, 0.004752966669, 0.004155890204, 0.003897907344, 0.004250814324, 0.00363196973, 0.003733774514, 0.003756053426, 0.004551709944, 0.005102927184, 0.004374288211, 0.004156728923, 0.004290280092, 0.003757206817, 0.00387302364, 0.003394904912, 0.003736503159, 0.00370489262, 0.003989563848, 0.004710173158, 0.003965921161, 0.004097876261, 0.003968104092, 0.003341734566, 0.003518722423, 0.003444342914, 0.00311780955, 0.003726171349, 0.003807930121, 0.004239428375, 0.003945152337, 0.003383915216, 0.003340113351, 0.003251743842, 0.002902764019, 0.002782632546, 0.002923215582, 0.003007653059, 0.003567505347, 0.004212690159, 0.003554662776, 0.003178857495, 0.003158288587, 0.003031807835, 0.002869155025, 0.002756593414, 0.003070416077, 0.003440003388, 0.003495357866, 0.003285121985, 0.003009518249, 0.002794968285, 0.002665138781, 0.002399572175, 0.002482747036, 0.002766082028, 0.002807242675, 0.003305060105, 0.002907239952, 0.002521817319, 0.002658128393, 0.002581899568, 0.002227657241, 0.002470283393, 0.002403485387, 0.002673820064, 0.002835663113, 0.003627688117, 0.002699749688, 0.002162000371, 0.002126479925, 0.001833840701, 0.001935762069, 0.002350195792, 0.0021989561, 0.002055965995, 0.002359758516, 0.002504526251, 0.001829998069, 0.001840005844, 0.001776080995, 0.001555161991, 0.001703873611, 0.001733502399, 0.001681532412],
					[0.006594382192, 0.005670592412, 0.005845477915, 0.005856465086, 0.005518711262, 0.005885907481, 0.005249993788, 0.005186831761, 0.004725555913, 0.00494187006, 0.005657857003, 0.005163692584, 0.004706632845, 0.004173076723, 0.003902254463, 0.004284994782, 0.00371978044, 0.003800660873, 0.003827532969, 0.004633361599, 0.005034250859, 0.004381039134, 0.004140177808, 0.004271141624, 0.003815087583, 0.003888776041, 0.003412560729, 0.003715946693, 0.003694375629, 0.003998670233, 0.00466125509, 0.003956994273, 0.004064539484, 0.003977163851, 0.003358623301, 0.003518226553, 0.003373116109, 0.003176878082, 0.003775985543, 0.003852669217, 0.004223583686, 0.003867575054, 0.003372837448, 0.003276297438, 0.003134634966, 0.002941816614, 0.002837975836, 0.002966669732, 0.00306358095, 0.003607251209, 0.004047749529, 0.003518512219, 0.003220634491, 0.003225088129, 0.003074806453, 0.002869951923, 0.002794455413, 0.00314857206, 0.003445206259, 0.003484276388, 0.003242065918, 0.00296896478, 0.002867250372, 0.002673288533, 0.002378738807, 0.002476341968, 0.00278931335, 0.002744695175, 0.003419811664, 0.002865320479, 0.00253830332, 0.002672007559, 0.002546935189, 0.002227989655, 0.00248874816, 0.002472291663, 0.002701562387, 0.002937912347, 0.003746374369, 0.002725652958, 0.002111424879, 0.002183330207, 0.001826697256, 0.001950773694, 0.00231638808, 0.002196664573, 0.002116575774, 0.002469559865, 0.002616001643, 0.001804342367, 0.00183879967, 0.001747813931, 0.001571649058, 0.001717569968, 0.001702643639, 0.001731275443],
					[0.006188482528, 0.005404122902, 0.005529484399, 0.005599641385, 0.005212667932, 0.005249993788, 0.00529751027, 0.00488858012, 0.004479292431, 0.004661042738, 0.005244524452, 0.004844508527, 0.004519702335, 0.003901269768, 0.003814645021, 0.004000603532, 0.003560416181, 0.003645410684, 0.003622276933, 0.004387740333, 0.004834011926, 0.004163941856, 0.003798179843, 0.003995557898, 0.00373019354, 0.003728689545, 0.00322625399, 0.003643686193, 0.003498595594, 0.00389183741, 0.00449380729, 0.003763685986, 0.003833958135, 0.003889181773, 0.003237279705, 0.003384040493, 0.003371544744, 0.003103820812, 0.003638987516, 0.003674315677, 0.003997987643, 0.00374307051, 0.003281052417, 0.003160486446, 0.003002804477, 0.002773694236, 0.002668950943, 0.002792739793, 0.0028610154, 0.003470426784, 0.003914053963, 0.003418586327, 0.003024243577, 0.003035802487, 0.002865340173, 0.002746671583, 0.002713610852, 0.002952582409, 0.003384751458, 0.003398783072, 0.003164531892, 0.002893749342, 0.002576146487, 0.002534087762, 0.002298139155, 0.002369198989, 0.002650016671, 0.002596581078, 0.003333011815, 0.002746955289, 0.0024088781, 0.002490168387, 0.002476504628, 0.002163291487, 0.002345133043, 0.002378529183, 0.002524251707, 0.002792157251, 0.003547327024, 0.002582676068, 0.001961667199, 0.00198535973, 0.001802418959, 0.00179555948, 0.002302544639, 0.002102420911, 0.001963682724, 0.002249388048, 0.002506105964, 0.001728073759, 0.001709216314, 0.001684785524, 0.0015206781, 0.00160284881, 0.001599105457, 0.001639530125],
					[0.006144514577, 0.005228753424, 0.005390914697, 0.005466069036, 0.00520579585, 0.005186831761, 0.00488858012, 0.00512539599, 0.004447562299, 0.004656774587, 0.005183246677, 0.00477351631, 0.004412606969, 0.003868620735, 0.003722629721, 0.003997654346, 0.00351483726, 0.003593304073, 0.00361034607, 0.004428704724, 0.004702079826, 0.004092405047, 0.003907918293, 0.004086110471, 0.003608937357, 0.003750278469, 0.003359302132, 0.003640954706, 0.003612722983, 0.003832686285, 0.004280689398, 0.003780650207, 0.003853131458, 0.003750471713, 0.003221510183, 0.003375329421, 0.00337204551, 0.003078967411, 0.003631111805, 0.003705470881, 0.004051691901, 0.003758263597, 0.003219548501, 0.00322410537, 0.003053520724, 0.00282870831, 0.002747014377, 0.002873165543, 0.002847918425, 0.003448010781, 0.003931554582, 0.003313178608, 0.003028926938, 0.003032511577, 0.002911464346, 0.002745567374, 0.002658174324, 0.003029978077, 0.003456979638, 0.003418816348, 0.003089516401, 0.002823881199, 0.002694148822, 0.002581815013, 0.002274152315, 0.002452890052, 0.002639344627, 0.002777070666, 0.003271405217, 0.002731006516, 0.002417733311, 0.002545041644, 0.002445071567, 0.002144393999, 0.002389854799, 0.002344599062, 0.002619809282, 0.002852917885, 0.003539833913, 0.002638401879, 0.002021217408, 0.002049355105, 0.001773477959, 0.001877138149, 0.002263133226, 0.00219391231, 0.002079094009, 0.002384093743, 0.002466890693, 0.001729753224, 0.001775353392, 0.001711371032, 0.001494355257, 0.001666919889, 0.001640984047, 0.001597138892],
					[0.005581524855, 0.004781393518, 0.004908853124, 0.005048488162, 0.004695260663, 0.004725555913, 0.004479292431, 0.004447562299, 0.004275795316, 0.004278451337, 0.004720826626, 0.004336785326, 0.004055297277, 0.003554139547, 0.003381220835, 0.003618877409, 0.003181156489, 0.003217439503, 0.003259972434, 0.004035653591, 0.004306432777, 0.003736776838, 0.003600043099, 0.003660395382, 0.003208342404, 0.003419811102, 0.002923043046, 0.003287871386, 0.003183279271, 0.003487995237, 0.003933100761, 0.003328081687, 0.003459011173, 0.003390186576, 0.002954394167, 0.002954282232, 0.003005882414, 0.002767666317, 0.003242079448, 0.003357423284, 0.003561854758, 0.003334818137, 0.002942575524, 0.002935013377, 0.002794287977, 0.002568939824, 0.002487098668, 0.002538668543, 0.002648547527, 0.003152359262, 0.003521043502, 0.003050485455, 0.00263296343, 0.002732898344, 0.002639358885, 0.00245577162, 0.00240993133, 0.002745503964, 0.003105609008, 0.003080811196, 0.002765467253, 0.002570024624, 0.002403539014, 0.002365928298, 0.002060583994, 0.002128802882, 0.002378697448, 0.002375537224, 0.002979944128, 0.002466248967, 0.002183305605, 0.00233221606, 0.002247460908, 0.00188411888, 0.002077564129, 0.002123604214, 0.002368484015, 0.002484697379, 0.003263079647, 0.002315163065, 0.001846701912, 0.00181648464, 0.001622591534, 0.001671580843, 0.002067975693, 0.001944108977, 0.001812298457, 0.002112557084, 0.002276282331, 0.001537089641, 0.001600344549, 0.001539839272, 0.001380351461, 0.001445855309, 0.001467154418, 0.001461220643],
					[0.005791371247, 0.004867670897, 0.005103830919, 0.005242939938, 0.004884578153, 0.00494187006, 0.004661042738, 0.004656774587, 0.004278451337, 0.004640905014, 0.00482602218, 0.004458280353, 0.004183449769, 0.003645954911, 0.003425163222, 0.003676113678, 0.003263922036, 0.003322619244, 0.00337732856, 0.004206406044, 0.004412656627, 0.003819775592, 0.003648447356, 0.003715759146, 0.003384254134, 0.003439829175, 0.003038721767, 0.003383124035, 0.003306777089, 0.003614084622, 0.003957043704, 0.003394430058, 0.003520707926, 0.00340538183, 0.002993569578, 0.003161183487, 0.002984117306, 0.002825546864, 0.003317706912, 0.003465333124, 0.003620403799, 0.00336627656, 0.002999951399, 0.002888635953, 0.002803079025, 0.002559037881, 0.002522217349, 0.002613303501, 0.002654684242, 0.003261252232, 0.003583525581, 0.003009493202, 0.002728739572, 0.00276611704, 0.00257947995, 0.002479049424, 0.002497650293, 0.00278677727, 0.003201191769, 0.003171901734, 0.002852639794, 0.002556324433, 0.002395752497, 0.002348819196, 0.002024066215, 0.002145725437, 0.002443764267, 0.002445315495, 0.002970631164, 0.002471499072, 0.00217780669, 0.002264720672, 0.002206492379, 0.001912792792, 0.002093213831, 0.00213031204, 0.00237844092, 0.002584543946, 0.003406758634, 0.002316135476, 0.001841431471, 0.001807023938, 0.001595657068, 0.001697851274, 0.002067148721, 0.001983828019, 0.001884783965, 0.002081992134, 0.00232940124, 0.001503426019, 0.001564091681, 0.001499557562, 0.001293417718, 0.001433147338, 0.001459815065, 0.001456114868],
					[0.006839503165, 0.006057294555, 0.006073314853, 0.006111949473, 0.00563581538, 0.005657857003, 0.005244524452, 0.005183246677, 0.004720826626, 0.00482602218, 0.00715380894, 0.00568062383, 0.005189433557, 0.004630976216, 0.004338145108, 0.004523508387, 0.003894140167, 0.003935601645, 0.003681198312, 0.004681323607, 0.005995902923, 0.005062114042, 0.004514844807, 0.004573083939, 0.004188643902, 0.003980118185, 0.003462811048, 0.003809565319, 0.003914667109, 0.004022603493, 0.005713294222, 0.004837360272, 0.00468190493, 0.004517144876, 0.003730300128, 0.003878401571, 0.003657651758, 0.003235475906, 0.003778763194, 0.00390359752, 0.005197194723, 0.00435597295, 0.003811533041, 0.003647806326, 0.003519599805, 0.003250822993, 0.002851594993, 0.003133866747, 0.003116543132, 0.003691230013, 0.004951209144, 0.004333963728, 0.003606982422, 0.003729182959, 0.003338004342, 0.003044495193, 0.002987117227, 0.003204243619, 0.003555873242, 0.003764693424, 0.004334541133, 0.003553878345, 0.003231203484, 0.002943099406, 0.002696793804, 0.002640047186, 0.003154011213, 0.0030178547, 0.003233047781, 0.003705899965, 0.003057049188, 0.003098922292, 0.002903858749, 0.00246176927, 0.002599881832, 0.002592969402, 0.002729180908, 0.003147871729, 0.003568750149, 0.003247957188, 0.002593234347, 0.002508700781, 0.002074163928, 0.00205201753, 0.002574940276, 0.002281943433, 0.002136746694, 0.002578710274, 0.002542355418, 0.002360605143, 0.002220554732, 0.002113770986, 0.001821547472, 0.001864275149, 0.001779315548, 0.00185400446],
					[0.006026119572, 0.005460206075, 0.005499701853, 0.005465037035, 0.0051420724, 0.005163692584, 0.004844508527, 0.00477351631, 0.004336785326, 0.004458280353, 0.00568062383, 0.005716769838, 0.0045563704, 0.004052643741, 0.003908646935, 0.004102625285, 0.003502287588, 0.003641064362, 0.003561828604, 0.004279809736, 0.005264020749, 0.004513935274, 0.004092159872, 0.004119086638, 0.003783692589, 0.003874680981, 0.003291156127, 0.003635289158, 0.003635098529, 0.003767603955, 0.005032640564, 0.00427678359, 0.004236813741, 0.004152959872, 0.003422740202, 0.003607572409, 0.003390717311, 0.003054099004, 0.003663946352, 0.003692743513, 0.004504288936, 0.004016010617, 0.003455978013, 0.003492681159, 0.003262636438, 0.003004556681, 0.002798426204, 0.002855994378, 0.002949423992, 0.003455392502, 0.004442682175, 0.003868226071, 0.003236383866, 0.003332541527, 0.003093094158, 0.002851603943, 0.002816757742, 0.003150937277, 0.003371518839, 0.003421802213, 0.003810412801, 0.003295890373, 0.002979283605, 0.002720910432, 0.002541390204, 0.002527176121, 0.002777588485, 0.00282129975, 0.003289166972, 0.003287562718, 0.00280330556, 0.002900054774, 0.00263994758, 0.002390474875, 0.002500464799, 0.002486099168, 0.002786945071, 0.002980381819, 0.003553381874, 0.002975489791, 0.002443998787, 0.002360782909, 0.002067739944, 0.001998907565, 0.002534972683, 0.002236704862, 0.00217467103, 0.002359074018, 0.002538706295, 0.002178968325, 0.002083663409, 0.001987904628, 0.001794990531, 0.001780425059, 0.001803667429, 0.001750839599],
					[0.00556042548, 0.004834687069, 0.004969641429, 0.005129873692, 0.004752966669, 0.004706632845, 0.004519702335, 0.004412606969, 0.004055297277, 0.004183449769, 0.005189433557, 0.0045563704, 0.005036636106, 0.0038638914, 0.003697488713, 0.003872703062, 0.003331785478, 0.003494649964, 0.00343602649, 0.004112031878, 0.004721710351, 0.004106976616, 0.003784774051, 0.003711100379, 0.00356760794, 0.003569918691, 0.003247660017, 0.003570598902, 0.00352122161, 0.003795115997, 0.004639492099, 0.003769156389, 0.00392091161, 0.003720303187, 0.003305639622, 0.003329627441, 0.00330156276, 0.002915103266, 0.00347520935, 0.003564945967, 0.004115633012, 0.003837504568, 0.003139800334, 0.003134726059, 0.003276090681, 0.002948850289, 0.002638145651, 0.002842945493, 0.003000883638, 0.003455188678, 0.00405233259, 0.003689967188, 0.002976667966, 0.00317508292, 0.003002616169, 0.002716058382, 0.002865345579, 0.002911171213, 0.0033207877, 0.003383236234, 0.003420739267, 0.003050888097, 0.002846153878, 0.002658557789, 0.00248201665, 0.002494794761, 0.002796370631, 0.002825190452, 0.003221488961, 0.002953910851, 0.002647777308, 0.00267266428, 0.002675630159, 0.002214654882, 0.002358356459, 0.002414975934, 0.002506934412, 0.002874428286, 0.003548025327, 0.002676243201, 0.00227159661, 0.002203276806, 0.002040575238, 0.001982869769, 0.002484016339, 0.002223537119, 0.002091956136, 0.002527216656, 0.00245712161, 0.00192903563, 0.002014819832, 0.00197991251, 0.001669107432, 0.001871868301, 0.001839191813, 0.001779332353],
					[0.004834846375, 0.00427960842, 0.004398641598, 0.004416730131, 0.004155890204, 0.004173076723, 0.003901269768, 0.003868620735, 0.003554139547, 0.003645954911, 0.004630976216, 0.004052643741, 0.0038638914, 0.003873833194, 0.003250860823, 0.003363765299, 0.002882109091, 0.003108371262, 0.002917879079, 0.003545322717, 0.004265575331, 0.00365789709, 0.003297331834, 0.003420793192, 0.003095560522, 0.003088608654, 0.002741008061, 0.00297496778, 0.002989693088, 0.003099461678, 0.003987789659, 0.003450262327, 0.003397950778, 0.00336390694, 0.002851727225, 0.002891509442, 0.002818125281, 0.002423383822, 0.002985284814, 0.003095657796, 0.003702221795, 0.003339200871, 0.002930283919, 0.002769878876, 0.002819545433, 0.002524422025, 0.002300083628, 0.002393712985, 0.002549847288, 0.002984198221, 0.003592822144, 0.00315543466, 0.002677606164, 0.002848600657, 0.002637211277, 0.002388155994, 0.002352838588, 0.002637722498, 0.00285052315, 0.002949664632, 0.003039306391, 0.002737754808, 0.002598694665, 0.002320043343, 0.002149391811, 0.002141692279, 0.002455331697, 0.002372667954, 0.002748106064, 0.002744602577, 0.002357861268, 0.002469088971, 0.002266383371, 0.001915814607, 0.002137951166, 0.002032070219, 0.002162047775, 0.002401755863, 0.003083914722, 0.002498508202, 0.002029350921, 0.002015332046, 0.001718203143, 0.001682071098, 0.002112971385, 0.001897921416, 0.001773218193, 0.002041876751, 0.002086498421, 0.001765418608, 0.001761942727, 0.001704274891, 0.00152144715, 0.001538600425, 0.001536824264, 0.00156334151],
					[0.004511975974, 0.004094676601, 0.004134976123, 0.004205399686, 0.003897907344, 0.003902254463, 0.003814645021, 0.003722629721, 0.003381220835, 0.003425163222, 0.004338145108, 0.003908646935, 0.003697488713, 0.003250860823, 0.00382247693, 0.003213676416, 0.002929488868, 0.002993233232, 0.002912453456, 0.003561962143, 0.004050883743, 0.003683104325, 0.003225346346, 0.003335041919, 0.003178531074, 0.003193442983, 0.002712119593, 0.003110599618, 0.003033533277, 0.003223937921, 0.003787378884, 0.003347239651, 0.003280235516, 0.003345649319, 0.002756067678, 0.002767110209, 0.002914685529, 0.002613713399, 0.002927980931, 0.003180273915, 0.003624700079, 0.003335812529, 0.002856828165, 0.002905681077, 0.002759159271, 0.002564853416, 0.002352480996, 0.002384606717, 0.002531822685, 0.002957736965, 0.003428895014, 0.003150561594, 0.00263577638, 0.002772655784, 0.00257720391, 0.002387000641, 0.002357687016, 0.002638124, 0.002886483586, 0.002862711554, 0.002969202312, 0.002639004767, 0.002423556742, 0.002302659107, 0.002211500634, 0.002164797252, 0.002527220919, 0.002339467906, 0.002896492621, 0.002636127797, 0.002371647666, 0.002451257932, 0.002351451581, 0.001999883551, 0.002172617411, 0.002162700965, 0.002268925825, 0.002563441418, 0.003174658432, 0.002401444852, 0.001947953835, 0.001925131204, 0.001891678493, 0.001711701281, 0.002146911221, 0.002034977418, 0.001895268085, 0.002260311334, 0.002183371878, 0.001703799545, 0.001711400505, 0.001713412, 0.001534514796, 0.00171764512, 0.001629135281, 0.00166261154],
					[0.00502810301, 0.004430417541, 0.004525198382, 0.004458656717, 0.004250814324, 0.004284994782, 0.004000603532, 0.003997654346, 0.003618877409, 0.003676113678, 0.004523508387, 0.004102625285, 0.003872703062, 0.003363765299, 0.003213676416, 0.003883552949, 0.002967310677, 0.003138069179, 0.002950611747, 0.0036818911, 0.004142008143, 0.003493402694, 0.003368136036, 0.003444248522, 0.003129152572, 0.003263735276, 0.002896374512, 0.003062725198, 0.003016611448, 0.003231478037, 0.003999979567, 0.003306005668, 0.00343324518, 0.003199677237, 0.002782050118, 0.002870923467, 0.002859287353, 0.00266264615, 0.003115850747, 0.003073912489, 0.003631119945, 0.003343044549, 0.00289164188, 0.002883360707, 0.002722695521, 0.002568207708, 0.002351255139, 0.002444673168, 0.002568367833, 0.003064479536, 0.003569656968, 0.003122008957, 0.00274180904, 0.002790751101, 0.00260473893, 0.002401464726, 0.002339342052, 0.002678514626, 0.002819510233, 0.002970506193, 0.002821127856, 0.002624749646, 0.002501434852, 0.002245179492, 0.002192825626, 0.002164558022, 0.002394731624, 0.002495798699, 0.002804875205, 0.002463355151, 0.002245615832, 0.002447919046, 0.002278303381, 0.001985647385, 0.002228671992, 0.002114276384, 0.002285185976, 0.002486016887, 0.003060032788, 0.002355436976, 0.001903773391, 0.001912987547, 0.001729006647, 0.001644036295, 0.002084742402, 0.00191297072, 0.001819593425, 0.002199419717, 0.002243951131, 0.001705130056, 0.001712051313, 0.001694201176, 0.001474786573, 0.001594540806, 0.001590322296, 0.001530771977],
					[0.004273827084, 0.003713099085, 0.003883988522, 0.003886593217, 0.00363196973, 0.00371978044, 0.003560416181, 0.00351483726, 0.003181156489, 0.003263922036, 0.003894140167, 0.003502287588, 0.003331785478, 0.002882109091, 0.002929488868, 0.002967310677, 0.003044465794, 0.002733450929, 0.002635010927, 0.003274962376, 0.003681445239, 0.003179269931, 0.002838947865, 0.003071383135, 0.002865501922, 0.002941029134, 0.002605314467, 0.002783878304, 0.002705720095, 0.00292749477, 0.003485988704, 0.00309340092, 0.003076533575, 0.002884361566, 0.002539860536, 0.002578019495, 0.002610227175, 0.002350060347, 0.002642787593, 0.002800009414, 0.003194270936, 0.002960725557, 0.002526066962, 0.002600968519, 0.002373548708, 0.00233798433, 0.002110325803, 0.002179840156, 0.002177178217, 0.002647144259, 0.002943346366, 0.002783759962, 0.00247288644, 0.002423331136, 0.002272073889, 0.002157631265, 0.002150761998, 0.002429601622, 0.00258608262, 0.002629567419, 0.002489536208, 0.002321161174, 0.00216214058, 0.002093913888, 0.001967676888, 0.001976377986, 0.00212531014, 0.002052453022, 0.002483985727, 0.00221991072, 0.002045275133, 0.002103865601, 0.00200061205, 0.001752657322, 0.001914621865, 0.001904786973, 0.00196953169, 0.002308013374, 0.002725449108, 0.002072097471, 0.00164384809, 0.001742680996, 0.001600301608, 0.001561717078, 0.001859878712, 0.001712852285, 0.001679136737, 0.001911216259, 0.001958052651, 0.001415231144, 0.00147673357, 0.001456591754, 0.001216366287, 0.001374620282, 0.001367862053, 0.001422734186],
					[0.004391065668, 0.003918644507, 0.003985939094, 0.003969078367, 0.003733774514, 0.003800660873, 0.003645410684, 0.003593304073, 0.003217439503, 0.003322619244, 0.003935601645, 0.003641064362, 0.003494649964, 0.003108371262, 0.002993233232, 0.003138069179, 0.002733450929, 0.00321693965, 0.0027661938, 0.003352791662, 0.003773810179, 0.003292673967, 0.002995705642, 0.003130947967, 0.002948265548, 0.002959804007, 0.002697129301, 0.002903473022, 0.002869045667, 0.003031698175, 0.003634588359, 0.003086208451, 0.003203120949, 0.003092637923, 0.002672967873, 0.002738575719, 0.002763725286, 0.002485791671, 0.00294973082, 0.002929667055, 0.003251648963, 0.003135708876, 0.002704731628, 0.002625437784, 0.002637326282, 0.002359066446, 0.002255987615, 0.002265448903, 0.002420139337, 0.002924262235, 0.003350293021, 0.002951902598, 0.002461926978, 0.002655913987, 0.002527600965, 0.002327256475, 0.002300116212, 0.002474336877, 0.002739514386, 0.002810675895, 0.002664617975, 0.002577909445, 0.002402551657, 0.002241335738, 0.002078977645, 0.002098506638, 0.002261782071, 0.002340604819, 0.002765602047, 0.002390797271, 0.002148585308, 0.002367405345, 0.002224109391, 0.001910957354, 0.002059282986, 0.002030608105, 0.002143121664, 0.002335886857, 0.002960454181, 0.002227701645, 0.001910514222, 0.00190527573, 0.001734247778, 0.001685271276, 0.002121516629, 0.001922968469, 0.001818487049, 0.002148142537, 0.0021263257, 0.001657059318, 0.001654215713, 0.001605786675, 0.001515116099, 0.001592331939, 0.001566729393, 0.001568749327],
					[0.004333287312, 0.003805239886, 0.003937507735, 0.003980490598, 0.003756053426, 0.003827532969, 0.003622276933, 0.00361034607, 0.003259972434, 0.00337732856, 0.003681198312, 0.003561828604, 0.00343602649, 0.002917879079, 0.002912453456, 0.002950611747, 0.002635010927, 0.0027661938, 0.003139940126, 0.003427624549, 0.003578906624, 0.00307687328, 0.002969666323, 0.00302098847, 0.002743564708, 0.002933038244, 0.002606332896, 0.002895006413, 0.002805153446, 0.003141477769, 0.003278536375, 0.002795899555, 0.002969012724, 0.002875247054, 0.002543491766, 0.00256979023, 0.002599392839, 0.002394934838, 0.002872689712, 0.002960508812, 0.002908189163, 0.002926999021, 0.002450992361, 0.002432798607, 0.00239761976, 0.002189184483, 0.002227756109, 0.002331667699, 0.002421487646, 0.002875366694, 0.00298844369, 0.002606981108, 0.002317270052, 0.002405332647, 0.002385423734, 0.002234018339, 0.002249965789, 0.002484491958, 0.002751166494, 0.002687481463, 0.002383835896, 0.002218656927, 0.002133343147, 0.002062347094, 0.001937510594, 0.001986508049, 0.002164657603, 0.002328611956, 0.002693200054, 0.002124462034, 0.001949411141, 0.002055621468, 0.002065006753, 0.001803744266, 0.002000441696, 0.001965647569, 0.002239790169, 0.002399122799, 0.002997783046, 0.002008603922, 0.00166251154, 0.001645494544, 0.001594663869, 0.00156716494, 0.001981218007, 0.001904040506, 0.001840492434, 0.002101416757, 0.002207618824, 0.001389582413, 0.001484180669, 0.001422802845, 0.001374981132, 0.001391746185, 0.00151998509, 0.001501016057],
					[0.005263026397, 0.004724829646, 0.004796383717, 0.004776577303, 0.004551709944, 0.004633361599, 0.004387740333, 0.004428704724, 0.004035653591, 0.004206406044, 0.004681323607, 0.004279809736, 0.004112031878, 0.003545322717, 0.003561962143, 0.0036818911, 0.003274962376, 0.003352791662, 0.003427624549, 0.004713592054, 0.004412996599, 0.003691369524, 0.003495881485, 0.003740263973, 0.003380891517, 0.003446890802, 0.003165427335, 0.003558889199, 0.003415026759, 0.003849844066, 0.004096787328, 0.003443607935, 0.003573020979, 0.003413438583, 0.003179537078, 0.003132294152, 0.003151896462, 0.002894221711, 0.003341524005, 0.003697354707, 0.003778983674, 0.003473167568, 0.003193445134, 0.003074032778, 0.003001363495, 0.002813582463, 0.002728806664, 0.002756412871, 0.002729660122, 0.003574148845, 0.003742135969, 0.003261816697, 0.002881400693, 0.00295569575, 0.002771345251, 0.00263992019, 0.002789369528, 0.003027644353, 0.003339619025, 0.003391684513, 0.003009037919, 0.0027456759, 0.002538829406, 0.00246199063, 0.002241391942, 0.002453759042, 0.002658168093, 0.002796176512, 0.003167364769, 0.002702996108, 0.002418751103, 0.002507110088, 0.002455918705, 0.002200525629, 0.002357724883, 0.002368100476, 0.002607530716, 0.002848028378, 0.003700040109, 0.002599035893, 0.002003765691, 0.002031277, 0.001906474095, 0.001893518972, 0.002302300689, 0.002221521965, 0.002142081755, 0.002591435225, 0.002621234068, 0.001826813062, 0.001795320974, 0.001810262683, 0.001647121456, 0.001765786251, 0.001753677596, 0.00178581935],
					[0.006093797699, 0.005547769669, 0.00554862731, 0.005403948677, 0.005102927184, 0.005034250859, 0.004834011926, 0.004702079826, 0.004306432777, 0.004412656627, 0.005995902923, 0.005264020749, 0.004721710351, 0.004265575331, 0.004050883743, 0.004142008143, 0.003681445239, 0.003773810179, 0.003578906624, 0.004412996599, 0.006400274475, 0.004829108861, 0.004184353511, 0.004320974139, 0.003946196099, 0.003911279088, 0.003315031289, 0.003730880704, 0.003706376531, 0.003906636903, 0.005361479462, 0.004625380525, 0.004510725187, 0.004359761436, 0.003687701075, 0.003642718224, 0.003533366195, 0.003147419208, 0.003727981077, 0.003698115635, 0.0048289164, 0.004370629445, 0.003732576263, 0.003622061423, 0.003367869088, 0.003209881211, 0.002915577328, 0.003030936925, 0.003163595894, 0.003710364719, 0.004846839608, 0.004181164132, 0.003579984805, 0.003684863634, 0.003303553928, 0.002993021396, 0.003079061485, 0.003216678089, 0.003516415998, 0.003563033468, 0.004092782743, 0.003510117601, 0.003131965539, 0.002865249967, 0.002725611012, 0.002733212066, 0.002940756862, 0.002965537976, 0.003303248719, 0.003620976929, 0.003034990502, 0.003093187533, 0.002901151042, 0.002514304441, 0.002581041583, 0.002572983051, 0.002693267864, 0.003012297037, 0.003726597402, 0.00317777977, 0.002611472402, 0.00252861915, 0.002266167678, 0.002117982045, 0.002668235806, 0.002302778545, 0.00226003792, 0.002571792485, 0.002718526671, 0.002277131577, 0.002209301972, 0.002131248267, 0.001875955335, 0.00190288067, 0.001968385308, 0.00192768926],
					[0.005328759167, 0.004641512869, 0.00472970036, 0.004734314166, 0.004374288211, 0.004381039134, 0.004163941856, 0.004092405047, 0.003736776838, 0.003819775592, 0.005062114042, 0.004513935274, 0.004106976616, 0.00365789709, 0.003683104325, 0.003493402694, 0.003179269931, 0.003292673967, 0.00307687328, 0.003691369524, 0.004829108861, 0.004846353964, 0.003666264902, 0.003736210123, 0.003435652026, 0.003327434877, 0.002786963435, 0.003327749918, 0.003165502457, 0.003293778884, 0.00458887544, 0.003868415051, 0.003810806758, 0.003812135294, 0.003031235293, 0.00315819996, 0.003127765307, 0.002679971527, 0.003145197477, 0.003229609154, 0.004040609008, 0.003625382924, 0.003093686213, 0.003046358195, 0.002982998964, 0.002713429037, 0.00249017059, 0.002583465192, 0.002667415222, 0.002995624971, 0.004059070863, 0.003637699749, 0.002998394955, 0.003056517742, 0.002810718864, 0.002538625765, 0.0024774441, 0.00270785449, 0.003038493889, 0.003132510879, 0.003491058531, 0.00292253278, 0.002812505344, 0.002501828039, 0.002316819949, 0.002253495925, 0.002516935413, 0.002451923454, 0.002803321185, 0.003086465809, 0.002570900195, 0.002649884076, 0.002471890206, 0.002036089535, 0.002110086306, 0.002118461054, 0.002273008844, 0.002606396206, 0.003256030568, 0.002670710051, 0.002162432991, 0.002133878537, 0.00189527577, 0.001799982984, 0.00224771988, 0.001895963612, 0.001801342803, 0.002218445191, 0.00220851091, 0.001925310451, 0.001872333325, 0.001733466947, 0.001460972446, 0.001628559069, 0.001487323111, 0.001518940644],
					[0.004885570623, 0.004275172741, 0.004383809539, 0.004410779587, 0.004156728923, 0.004140177808, 0.003798179843, 0.003907918293, 0.003600043099, 0.003648447356, 0.004514844807, 0.004092159872, 0.003784774051, 0.003297331834, 0.003225346346, 0.003368136036, 0.002838947865, 0.002995705642, 0.002969666323, 0.003495881485, 0.004184353511, 0.003666264902, 0.004179783337, 0.003471414173, 0.003023854828, 0.003356610161, 0.002698519832, 0.003038886897, 0.003003883224, 0.003341485813, 0.003794583205, 0.003293611224, 0.003492744597, 0.003326876146, 0.002819414503, 0.002903537259, 0.002864338184, 0.002573591133, 0.002974747996, 0.003097156618, 0.003585739896, 0.003359651379, 0.002702284528, 0.00280620731, 0.002800027585, 0.002587252992, 0.00237875066, 0.002318001535, 0.002577506406, 0.003046000888, 0.003738788513, 0.003156040757, 0.002682901711, 0.002738199901, 0.002771310142, 0.002361977066, 0.00228325972, 0.00259919058, 0.00286215559, 0.002856169711, 0.002929151204, 0.002730446441, 0.002602943627, 0.002393241764, 0.002125213045, 0.002038210391, 0.002278688507, 0.002451203521, 0.002800159684, 0.002582754777, 0.002317755746, 0.002533026189, 0.002351341128, 0.001900481007, 0.002128911057, 0.002028752638, 0.002231438776, 0.002492451223, 0.003146094171, 0.002458733781, 0.001977984212, 0.002032270797, 0.001658662239, 0.001736552634, 0.002087295719, 0.001880784241, 0.001813132445, 0.00215663356, 0.002089571752, 0.00168525719, 0.001848983201, 0.001633653508, 0.00153148607, 0.001521509988, 0.001577008256, 0.001443522175],
					[0.005006898428, 0.004535712695, 0.004507738633, 0.004359942384, 0.004290280092, 0.004271141624, 0.003995557898, 0.004086110471, 0.003660395382, 0.003715759146, 0.004573083939, 0.004119086638, 0.003711100379, 0.003420793192, 0.003335041919, 0.003444248522, 0.003071383135, 0.003130947967, 0.00302098847, 0.003740263973, 0.004320974139, 0.003736210123, 0.003471414173, 0.004219814752, 0.003074347462, 0.00335558002, 0.002853020663, 0.003163075181, 0.003162460188, 0.003383226237, 0.003946854786, 0.003358245728, 0.003492581255, 0.00347055143, 0.002864913462, 0.002947059973, 0.002977872901, 0.002697692301, 0.00306046926, 0.003209536799, 0.003709842156, 0.003330488186, 0.002923999009, 0.003021807234, 0.002858419046, 0.002557295966, 0.002518513222, 0.00260775945, 0.002569833295, 0.003149431155, 0.003752861444, 0.003185902301, 0.002801078981, 0.002773726189, 0.002783920954, 0.002526022159, 0.002509305232, 0.00270304282, 0.002986993227, 0.003066809247, 0.003032034749, 0.002679082237, 0.002584795124, 0.00242750741, 0.002219893885, 0.002286300568, 0.002416626155, 0.002570982108, 0.002968969687, 0.002657204096, 0.002334412356, 0.002524391039, 0.00233510927, 0.002032861451, 0.002332175229, 0.002109761204, 0.002414349499, 0.002598142746, 0.003200476161, 0.002519373609, 0.00194844727, 0.002044466161, 0.001777307171, 0.001766203399, 0.002226324109, 0.002049160193, 0.00191435273, 0.002332399788, 0.002324302701, 0.001714808166, 0.001772016561, 0.001672395879, 0.001495796545, 0.001619156643, 0.001669654434, 0.001534071611],
					[0.004375058515, 0.004002068481, 0.004099612817, 0.004043738715, 0.003757206817, 0.003815087583, 0.00373019354, 0.003608937357, 0.003208342404, 0.003384254134, 0.004188643902, 0.003783692589, 0.00356760794, 0.003095560522, 0.003178531074, 0.003129152572, 0.002865501922, 0.002948265548, 0.002743564708, 0.003380891517, 0.003946196099, 0.003435652026, 0.003023854828, 0.003074347462, 0.003587875816, 0.002988766884, 0.002676701988, 0.002935990214, 0.002847540267, 0.003027385698, 0.003833575022, 0.003323279761, 0.003323942424, 0.003132949639, 0.002732966861, 0.002828162298, 0.002733380803, 0.002420330457, 0.002918191526, 0.002943742574, 0.003517684979, 0.003152410878, 0.002764708831, 0.002673637997, 0.002574437589, 0.002454666916, 0.002211056038, 0.002238484936, 0.002225958147, 0.002883740895, 0.00337803957, 0.003022747245, 0.002552899392, 0.002698310427, 0.002457512633, 0.0023063138, 0.002281525885, 0.002496782109, 0.002708649286, 0.002886268458, 0.002825941231, 0.002597498699, 0.002309029674, 0.002188957955, 0.002129636242, 0.002074964307, 0.002305354962, 0.002290907059, 0.002604838112, 0.002458205216, 0.002175005627, 0.002335369335, 0.002151563889, 0.001885691664, 0.002024016797, 0.001981590044, 0.002175490999, 0.002433038059, 0.002925505835, 0.002296204885, 0.001843445266, 0.001861205512, 0.001682774242, 0.001597012896, 0.002031397406, 0.001888393315, 0.001786999549, 0.002043348134, 0.002036454966, 0.001631303246, 0.001626562968, 0.001603387422, 0.00138335841, 0.001554231646, 0.00142310159, 0.001518029966],
					[0.004536227401, 0.003998855101, 0.004120392418, 0.00403985557, 0.00387302364, 0.003888776041, 0.003728689545, 0.003750278469, 0.003419811102, 0.003439829175, 0.003980118185, 0.003874680981, 0.003569918691, 0.003088608654, 0.003193442983, 0.003263735276, 0.002941029134, 0.002959804007, 0.002933038244, 0.003446890802, 0.003911279088, 0.003327434877, 0.003356610161, 0.00335558002, 0.002988766884, 0.003799319481, 0.002803502387, 0.003093195063, 0.003051601905, 0.003287937879, 0.003713693103, 0.003127171505, 0.003227872339, 0.003230790424, 0.002656555739, 0.002746432141, 0.002948685667, 0.002492862527, 0.003131702677, 0.003070495267, 0.003482263014, 0.003260850024, 0.002760310073, 0.00289823961, 0.002656191464, 0.002533025109, 0.002453741056, 0.002358155389, 0.002498554716, 0.002957772661, 0.003434297459, 0.003012799755, 0.002586533688, 0.002659368438, 0.002638276807, 0.002365156932, 0.002266873072, 0.002727073654, 0.00298885925, 0.002962162727, 0.002724667161, 0.002603957511, 0.00239513157, 0.002366851537, 0.002204672391, 0.002124099997, 0.002254771995, 0.002485193819, 0.002957245289, 0.002441650783, 0.002273750322, 0.002472290253, 0.002295598751, 0.001996875499, 0.002265404134, 0.002150671015, 0.002292409218, 0.002491206316, 0.003125602712, 0.00234600001, 0.001961811834, 0.001976809919, 0.001834543289, 0.001731717203, 0.002172197335, 0.0020091459, 0.001939627161, 0.002327275906, 0.002234689403, 0.001603869682, 0.001739008661, 0.001730536194, 0.00160289107, 0.00161558132, 0.001635188623, 0.001538851026],
					[0.00382383205, 0.003461115736, 0.003465223346, 0.003443041304, 0.003394904912, 0.003412560729, 0.00322625399, 0.003359302132, 0.002923043046, 0.003038721767, 0.003462811048, 0.003291156127, 0.003247660017, 0.002741008061, 0.002712119593, 0.002896374512, 0.002605314467, 0.002697129301, 0.002606332896, 0.003165427335, 0.003315031289, 0.002786963435, 0.002698519832, 0.002853020663, 0.002676701988, 0.002803502387, 0.00306700713, 0.002735058614, 0.002695636045, 0.002991305127, 0.00319868205, 0.002879138605, 0.002890740144, 0.00277773913, 0.002525763465, 0.002601870775, 0.002549273479, 0.002315693113, 0.002736115921, 0.00279314337, 0.002947256641, 0.002949088362, 0.002438530769, 0.002494613245, 0.002427808488, 0.0023698177, 0.002180214972, 0.002209502928, 0.002311303426, 0.002783458454, 0.002905864464, 0.002661201218, 0.002421328055, 0.00247621901, 0.002356458479, 0.002146446492, 0.002187583781, 0.002492489763, 0.002553266905, 0.002613346258, 0.002381112847, 0.002299599325, 0.002138305566, 0.002185859516, 0.002131839155, 0.002043831445, 0.002271061995, 0.00229176301, 0.00263039294, 0.00211106795, 0.002025041307, 0.002161650698, 0.002002517374, 0.001885994037, 0.002042060296, 0.001988821433, 0.002202094963, 0.002329232732, 0.002684325948, 0.002051566, 0.001725660176, 0.001823756082, 0.001687102396, 0.001621796456, 0.001933316665, 0.001898223352, 0.001877129076, 0.002004063264, 0.002049152968, 0.001490542702, 0.001573055978, 0.001556844155, 0.001397123182, 0.001458963187, 0.001493556021, 0.001504941644],
					[0.004188885336, 0.003868338392, 0.003916806442, 0.003961335731, 0.003736503159, 0.003715946693, 0.003643686193, 0.003640954706, 0.003287871386, 0.003383124035, 0.003809565319, 0.003635289158, 0.003570598902, 0.00297496778, 0.003110599618, 0.003062725198, 0.002783878304, 0.002903473022, 0.002895006413, 0.003558889199, 0.003730880704, 0.003327749918, 0.003038886897, 0.003163075181, 0.002935990214, 0.003093195063, 0.002735058614, 0.003367045221, 0.00296287914, 0.003316453212, 0.003559494386, 0.003022089541, 0.00315814943, 0.003016359076, 0.002667549281, 0.002677872445, 0.002914688297, 0.002509079189, 0.003004795826, 0.003058055976, 0.003248632596, 0.003158675801, 0.00260980252, 0.002709653171, 0.002696348858, 0.002456309315, 0.002423594261, 0.002423369531, 0.002442380546, 0.003036829971, 0.003284227606, 0.0029268648, 0.002456698341, 0.002705616909, 0.002486510885, 0.002321681675, 0.002466251737, 0.002640528653, 0.002961271301, 0.002871738576, 0.002652598675, 0.002495345023, 0.002318905424, 0.0022650826, 0.002093950813, 0.002171420609, 0.002351905388, 0.002521800233, 0.002877079814, 0.002384031672, 0.002151030603, 0.002317802626, 0.00226400016, 0.002005291199, 0.002147263607, 0.00215961835, 0.002237437268, 0.002500962852, 0.0032285676, 0.00226710001, 0.001868700425, 0.001834793041, 0.001821025008, 0.001726301118, 0.002169956628, 0.002032204841, 0.001974984732, 0.002326966181, 0.002348328682, 0.001651025562, 0.001691110869, 0.001726021726, 0.001591942106, 0.001656064673, 0.001621731298, 0.001688645496],
					[0.004227284984, 0.003743206232, 0.003893286734, 0.003888049826, 0.00370489262, 0.003694375629, 0.003498595594, 0.003612722983, 0.003183279271, 0.003306777089, 0.003914667109, 0.003635098529, 0.00352122161, 0.002989693088, 0.003033533277, 0.003016611448, 0.002705720095, 0.002869045667, 0.002805153446, 0.003415026759, 0.003706376531, 0.003165502457, 0.003003883224, 0.003162460188, 0.002847540267, 0.003051601905, 0.002695636045, 0.00296287914, 0.003230447816, 0.003234096768, 0.003466110855, 0.003037932947, 0.003089193132, 0.00300929479, 0.00257306987, 0.00266034213, 0.00273167138, 0.002420467153, 0.002915252933, 0.002969772584, 0.003252968055, 0.003136741139, 0.002526157358, 0.002702039418, 0.002609354892, 0.002335447056, 0.002329560571, 0.002357951196, 0.00246716127, 0.002996461714, 0.003250542802, 0.002808323347, 0.002478598141, 0.002593966541, 0.002468548448, 0.00228161994, 0.002341879803, 0.002538539563, 0.002828931675, 0.002913652699, 0.002624850473, 0.002433160548, 0.002332568494, 0.002191589414, 0.00211624525, 0.00214557857, 0.002334825057, 0.002464649154, 0.002830893306, 0.002365428219, 0.002215903769, 0.002292992308, 0.002186510036, 0.001928013396, 0.002127059326, 0.002026009422, 0.002247039578, 0.002468833821, 0.003034514466, 0.002220333684, 0.001885137482, 0.001871014565, 0.001755383966, 0.001768774387, 0.002103557194, 0.001956164834, 0.001946371307, 0.002220914779, 0.002292293272, 0.00153782559, 0.001667038649, 0.001664785053, 0.001486070658, 0.001637206048, 0.001584095463, 0.001656696165],
					[0.004480470259, 0.00397440804, 0.00427363146, 0.004038458787, 0.003989563848, 0.003998670233, 0.00389183741, 0.003832686285, 0.003487995237, 0.003614084622, 0.004022603493, 0.003767603955, 0.003795115997, 0.003099461678, 0.003223937921, 0.003231478037, 0.00292749477, 0.003031698175, 0.003141477769, 0.003849844066, 0.003906636903, 0.003293778884, 0.003341485813, 0.003383226237, 0.003027385698, 0.003287937879, 0.002991305127, 0.003316453212, 0.003234096768, 0.004257539481, 0.00360892735, 0.003089312147, 0.00314914342, 0.003293199865, 0.002775515786, 0.002848306622, 0.002993796676, 0.002688962135, 0.003248114478, 0.003503084115, 0.003305595067, 0.003361283934, 0.002744912459, 0.002780883803, 0.002735180879, 0.002597800854, 0.002580293557, 0.002562534438, 0.002716587516, 0.003329739, 0.003374641407, 0.002900609376, 0.002693202523, 0.002619729093, 0.002685051147, 0.00240451075, 0.002498561864, 0.002693658519, 0.003142076927, 0.003097697868, 0.00259576905, 0.002535927788, 0.002366890196, 0.002410639485, 0.002165431495, 0.002328296499, 0.00245450379, 0.002711650031, 0.003238324681, 0.002348861861, 0.002259480343, 0.002309906809, 0.002355579841, 0.002038839365, 0.002239321229, 0.002284708105, 0.002384424834, 0.002667066797, 0.003507441082, 0.002274031605, 0.001883094939, 0.001964218654, 0.00181361643, 0.001834717014, 0.002292201244, 0.002216786869, 0.002177387041, 0.002613953357, 0.002567909878, 0.001524680113, 0.001779855608, 0.001690825931, 0.001564463116, 0.001679989757, 0.001671031498, 0.001728721221],
					[0.005652719041, 0.005132938989, 0.005206899053, 0.004970339251, 0.004710173158, 0.00466125509, 0.00449380729, 0.004280689398, 0.003933100761, 0.003957043704, 0.005713294222, 0.005032640564, 0.004639492099, 0.003987789659, 0.003787378884, 0.003999979567, 0.003485988704, 0.003634588359, 0.003278536375, 0.004096787328, 0.005361479462, 0.00458887544, 0.003794583205, 0.003946854786, 0.003833575022, 0.003713693103, 0.00319868205, 0.003559494386, 0.003466110855, 0.00360892735, 0.006389629496, 0.004543286095, 0.00435339964, 0.00421183499, 0.003643843907, 0.003552145708, 0.003377381468, 0.002940465183, 0.003674445074, 0.003546939204, 0.004832200731, 0.004167746276, 0.003702399965, 0.003501187289, 0.003368893116, 0.003111353431, 0.002770718558, 0.002793733171, 0.002945230267, 0.003532627186, 0.004823624868, 0.004206178584, 0.003434773421, 0.003554204283, 0.00317215526, 0.002848030125, 0.002739330765, 0.003111718767, 0.003134480965, 0.003584368242, 0.004337202597, 0.003567835962, 0.003097652087, 0.002903839306, 0.002862427116, 0.002742707281, 0.00303943435, 0.002918479447, 0.002895614018, 0.003558485885, 0.003049905475, 0.003085204278, 0.0028736148, 0.002450573985, 0.002543276964, 0.002552567112, 0.002644966347, 0.002905540983, 0.003571788009, 0.003231873214, 0.002752491792, 0.00253094309, 0.002247082258, 0.002097977091, 0.002585262202, 0.002213443299, 0.002178675095, 0.002510933318, 0.002546967054, 0.002496554737, 0.002295747956, 0.002140649214, 0.00184150236, 0.001894217899, 0.001779130572, 0.00175253825],
					[0.004684925202, 0.004315534403, 0.004323028253, 0.004186782856, 0.003965921161, 0.003956994273, 0.003763685986, 0.003780650207, 0.003328081687, 0.003394430058, 0.004837360272, 0.00427678359, 0.003769156389, 0.003450262327, 0.003347239651, 0.003306005668, 0.00309340092, 0.003086208451, 0.002795899555, 0.003443607935, 0.004625380525, 0.003868415051, 0.003293611224, 0.003358245728, 0.003323279761, 0.003127171505, 0.002879138605, 0.003022089541, 0.003037932947, 0.003089312147, 0.004543286095, 0.004619143534, 0.003587010759, 0.003561213594, 0.003084867702, 0.003114246027, 0.002929868051, 0.002477946386, 0.003101819533, 0.003145192983, 0.004121189692, 0.003653686039, 0.003180433981, 0.003098678454, 0.002956622019, 0.002822098896, 0.002300704258, 0.002436448376, 0.002488366376, 0.002998644227, 0.003872081124, 0.003577494119, 0.003077587403, 0.003050500899, 0.002748571156, 0.002444597683, 0.002435739537, 0.002681119544, 0.002759195476, 0.003072875462, 0.003471787293, 0.00303231373, 0.002760249885, 0.002524300375, 0.002394184833, 0.002314679862, 0.002563165524, 0.002435194682, 0.002696053478, 0.003113548988, 0.002701328888, 0.00269269277, 0.002425354578, 0.002117866581, 0.002233633219, 0.002230699024, 0.002236322138, 0.002683642961, 0.002926002114, 0.002786528801, 0.002314710289, 0.002238214903, 0.001959224997, 0.001948823312, 0.002238119388, 0.001904738534, 0.001925369836, 0.002235240477, 0.002217031467, 0.002028930962, 0.002013238493, 0.001896815714, 0.001624917218, 0.001691962259, 0.001543014166, 0.001629675397],
					[0.004824278, 0.004334464152, 0.004343625848, 0.004351604312, 0.004097876261, 0.004064539484, 0.003833958135, 0.003853131458, 0.003459011173, 0.003520707926, 0.00468190493, 0.004236813741, 0.00392091161, 0.003397950778, 0.003280235516, 0.00343324518, 0.003076533575, 0.003203120949, 0.002969012724, 0.003573020979, 0.004510725187, 0.003810806758, 0.003492744597, 0.003492581255, 0.003323942424, 0.003227872339, 0.002890740144, 0.00315814943, 0.003089193132, 0.00314914342, 0.00435339964, 0.003587010759, 0.00433071444, 0.003369536117, 0.003050768156, 0.002964649715, 0.003030567194, 0.002635540325, 0.00313070882, 0.003219439668, 0.003833987856, 0.003641186078, 0.002890780361, 0.00295686402, 0.003012434425, 0.002712519187, 0.002492592931, 0.002536415273, 0.002606521851, 0.003156837627, 0.003945092908, 0.003427932124, 0.002807858163, 0.003140722747, 0.002845330975, 0.002578218747, 0.002494993998, 0.002687851375, 0.002980382082, 0.002966540382, 0.00324307985, 0.003040830314, 0.002747754518, 0.002462205185, 0.00236962802, 0.002326459033, 0.002444427051, 0.002492011945, 0.002907310129, 0.002862370156, 0.002476277853, 0.002678459174, 0.002505757031, 0.00214030932, 0.002277384155, 0.002153952799, 0.00231883584, 0.002832604928, 0.003215172404, 0.00257810375, 0.002168545284, 0.002214754088, 0.001950404018, 0.00188735876, 0.002251596033, 0.002023847758, 0.001959518691, 0.002326900403, 0.002387760302, 0.001960869109, 0.00197660144, 0.001872576902, 0.00167473677, 0.001639575669, 0.001682717651, 0.001722636547],
					[0.004694296476, 0.004243317135, 0.00432711529, 0.004236631295, 0.003968104092, 0.003977163851, 0.003889181773, 0.003750471713, 0.003390186576, 0.00340538183, 0.004517144876, 0.004152959872, 0.003720303187, 0.00336390694, 0.003345649319, 0.003199677237, 0.002884361566, 0.003092637923, 0.002875247054, 0.003413438583, 0.004359761436, 0.003812135294, 0.003326876146, 0.00347055143, 0.003132949639, 0.003230790424, 0.00277773913, 0.003016359076, 0.00300929479, 0.003293199865, 0.00421183499, 0.003561213594, 0.003369536117, 0.00399116046, 0.002839814268, 0.002955843653, 0.002934780701, 0.002577714692, 0.003126073595, 0.00308664757, 0.003817000885, 0.003431157463, 0.002981412052, 0.002907614761, 0.002769647323, 0.002650533834, 0.002431642779, 0.002420464412, 0.002619850514, 0.002904114795, 0.003772310698, 0.003295540396, 0.002766997167, 0.002774585934, 0.002778439634, 0.002491512264, 0.002358714498, 0.00263069605, 0.002811431806, 0.002873848244, 0.003285817904, 0.002866599622, 0.002514428298, 0.002443524259, 0.002274656068, 0.00221991396, 0.002444334685, 0.002513984598, 0.002864591507, 0.002849514422, 0.002484707529, 0.002510756872, 0.002307803728, 0.002045824989, 0.002175205296, 0.002223016351, 0.002255465655, 0.002509502702, 0.003086664146, 0.002559313469, 0.002125851234, 0.002042135601, 0.001865060743, 0.001717740157, 0.002234850987, 0.001997505152, 0.001918653967, 0.002221537306, 0.002176170363, 0.001858518206, 0.001821790762, 0.001711593746, 0.001589475474, 0.001632520931, 0.001637769712, 0.001580531026],
					[0.0038311798, 0.003586062754, 0.003575259214, 0.003515810312, 0.003341734566, 0.003358623301, 0.003237279705, 0.003221510183, 0.002954394167, 0.002993569578, 0.003730300128, 0.003422740202, 0.003305639622, 0.002851727225, 0.002756067678, 0.002782050118, 0.002539860536, 0.002672967873, 0.002543491766, 0.003179537078, 0.003687701075, 0.003031235293, 0.002819414503, 0.002864913462, 0.002732966861, 0.002656555739, 0.002525763465, 0.002667549281, 0.00257306987, 0.002775515786, 0.003643843907, 0.003084867702, 0.003050768156, 0.002839814268, 0.003113261711, 0.002605304699, 0.002523949341, 0.002256129179, 0.002623905379, 0.002786969656, 0.003148736152, 0.002983505359, 0.002586300099, 0.002513571863, 0.002510273881, 0.002352049805, 0.002093244918, 0.002053063802, 0.002270429913, 0.002740250691, 0.00312637474, 0.002822365865, 0.002439855205, 0.002520047139, 0.002422558995, 0.002079851963, 0.002235407413, 0.002373286904, 0.002521135822, 0.002633358523, 0.002729913204, 0.00243657735, 0.002218249685, 0.002148550824, 0.00206209881, 0.002026451581, 0.0021809998, 0.002184221162, 0.002537095063, 0.002429949287, 0.002107394936, 0.002216443786, 0.002131532637, 0.001853938575, 0.001872324134, 0.001909036257, 0.002066530273, 0.00233571516, 0.002790613686, 0.002256992943, 0.001784834378, 0.001836535011, 0.001683397023, 0.001668607442, 0.001948746689, 0.001817722133, 0.001742905187, 0.001849463061, 0.001900591909, 0.001683628808, 0.001661515879, 0.00162246157, 0.001411743762, 0.001469458031, 0.001513598958, 0.001448566505],
					[0.004131720831, 0.003673558324, 0.003722534095, 0.003803261423, 0.003518722423, 0.003518226553, 0.003384040493, 0.003375329421, 0.002954282232, 0.003161183487, 0.003878401571, 0.003607572409, 0.003329627441, 0.002891509442, 0.002767110209, 0.002870923467, 0.002578019495, 0.002738575719, 0.00256979023, 0.003132294152, 0.003642718224, 0.00315819996, 0.002903537259, 0.002947059973, 0.002828162298, 0.002746432141, 0.002601870775, 0.002677872445, 0.00266034213, 0.002848306622, 0.003552145708, 0.003114246027, 0.002964649715, 0.002955843653, 0.002605304699, 0.003246127894, 0.002597081723, 0.002289770099, 0.002681355027, 0.002807109181, 0.003251010168, 0.002906715171, 0.002690006606, 0.002534592074, 0.002472799134, 0.002380507514, 0.002121420707, 0.002153921171, 0.002199153858, 0.00274069719, 0.003199500798, 0.002891564344, 0.002373193014, 0.002513681059, 0.002319327304, 0.002084102308, 0.00224190194, 0.002394257036, 0.00256134582, 0.00252863471, 0.002803873853, 0.002419776366, 0.002180843066, 0.002103332855, 0.002042384937, 0.001968265032, 0.002240327534, 0.002214681019, 0.002462364389, 0.002424836448, 0.002049107198, 0.002083366443, 0.00197987481, 0.001875720046, 0.001887109246, 0.001949640749, 0.002115343694, 0.002307184219, 0.002697538545, 0.002179623843, 0.001839841472, 0.001812945473, 0.001639023288, 0.001532550038, 0.001956169542, 0.00177166698, 0.001680075186, 0.001911598533, 0.001958579231, 0.001715632971, 0.001600549369, 0.001576340268, 0.00143020767, 0.001363441344, 0.001368196308, 0.001363541314],
					[0.003918816918, 0.003599684906, 0.003569795438, 0.003629734042, 0.003444342914, 0.003373116109, 0.003371544744, 0.00337204551, 0.003005882414, 0.002984117306, 0.003657651758, 0.003390717311, 0.00330156276, 0.002818125281, 0.002914685529, 0.002859287353, 0.002610227175, 0.002763725286, 0.002599392839, 0.003151896462, 0.003533366195, 0.003127765307, 0.002864338184, 0.002977872901, 0.002733380803, 0.002948685667, 0.002549273479, 0.002914688297, 0.00273167138, 0.002993796676, 0.003377381468, 0.002929868051, 0.003030567194, 0.002934780701, 0.002523949341, 0.002597081723, 0.003241193577, 0.002403651445, 0.00292431403, 0.002867665129, 0.003037304181, 0.003101957178, 0.002501104067, 0.002572921285, 0.002675836699, 0.002472353739, 0.002332287125, 0.002256236355, 0.002316121493, 0.002808901081, 0.003053367321, 0.002762400551, 0.002330442759, 0.002681173546, 0.002527505865, 0.002247129268, 0.002221462671, 0.002480365542, 0.002739803121, 0.002757168903, 0.00244546032, 0.002450922864, 0.002309054272, 0.002160272368, 0.002095941077, 0.002156213221, 0.002307032847, 0.002425733908, 0.00276051117, 0.002288477239, 0.002059309865, 0.002395086863, 0.002338798784, 0.001965452817, 0.002133015049, 0.002127609696, 0.002180562054, 0.002459078114, 0.003023566438, 0.002211358705, 0.001910315612, 0.001850515018, 0.001865905876, 0.001676086598, 0.002156986478, 0.001938585802, 0.001892057444, 0.002211490898, 0.002181747593, 0.00161786628, 0.001671386618, 0.001651221385, 0.001634855501, 0.001659412043, 0.001620250311, 0.001693878156],
					[0.00363012853, 0.003328464186, 0.003341274835, 0.003345952371, 0.00311780955, 0.003176878082, 0.003103820812, 0.003078967411, 0.002767666317, 0.002825546864, 0.003235475906, 0.003054099004, 0.002915103266, 0.002423383822, 0.002613713399, 0.00266264615, 0.002350060347, 0.002485791671, 0.002394934838, 0.002894221711, 0.003147419208, 0.002679971527, 0.002573591133, 0.002697692301, 0.002420330457, 0.002492862527, 0.002315693113, 0.002509079189, 0.002420467153, 0.002688962135, 0.002940465183, 0.002477946386, 0.002635540325, 0.002577714692, 0.002256129179, 0.002289770099, 0.002403651445, 0.002585858455, 0.00253544555, 0.002491761963, 0.002702254236, 0.002568549821, 0.002301293365, 0.002340550062, 0.002103647344, 0.002006869365, 0.001983430368, 0.002017001486, 0.002145451789, 0.002531211976, 0.002653443198, 0.002399420782, 0.002150288726, 0.002176605807, 0.002159345857, 0.001969783828, 0.001983946335, 0.002136173418, 0.002299743107, 0.002245431262, 0.002152025938, 0.00202899222, 0.001868377558, 0.001880645303, 0.001805562653, 0.001860821108, 0.001986037911, 0.002040183949, 0.002325492097, 0.001988182248, 0.001757660917, 0.001900188593, 0.001913647432, 0.001699751327, 0.001726585286, 0.001847907701, 0.001886751418, 0.002194595047, 0.002489518257, 0.001822208334, 0.001510692335, 0.001534000403, 0.001554887457, 0.001469200418, 0.001768324436, 0.001737591391, 0.001611170221, 0.001913410436, 0.001930658194, 0.001297908725, 0.001312630625, 0.001291381265, 0.001203881102, 0.001290540791, 0.001357172477, 0.001286333176],
					[0.004303746538, 0.003895523432, 0.00395494306, 0.003877361097, 0.003726171349, 0.003775985543, 0.003638987516, 0.003631111805, 0.003242079448, 0.003317706912, 0.003778763194, 0.003663946352, 0.00347520935, 0.002985284814, 0.002927980931, 0.003115850747, 0.002642787593, 0.00294973082, 0.002872689712, 0.003341524005, 0.003727981077, 0.003145197477, 0.002974747996, 0.00306046926, 0.002918191526, 0.003131702677, 0.002736115921, 0.003004795826, 0.002915252933, 0.003248114478, 0.003674445074, 0.003101819533, 0.00313070882, 0.003126073595, 0.002623905379, 0.002681355027, 0.00292431403, 0.00253544555, 0.003536369603, 0.003104430334, 0.0032369198, 0.003232893531, 0.002606564187, 0.002637783682, 0.002598238767, 0.002373926728, 0.002323770748, 0.002325965117, 0.002503271213, 0.002921510826, 0.003188651794, 0.002843995701, 0.002491374206, 0.002673806675, 0.002571918771, 0.002379817153, 0.002362586563, 0.002660356607, 0.002844804395, 0.002827361159, 0.002658962851, 0.002517867529, 0.0022987776, 0.002230930061, 0.002179070837, 0.002236853026, 0.002385015081, 0.002537127192, 0.002868935179, 0.002362173787, 0.002143727368, 0.002323640337, 0.002270978414, 0.001988655607, 0.002219449236, 0.002179725325, 0.002221286989, 0.002421063393, 0.003063830896, 0.002320725035, 0.001971615854, 0.001932172136, 0.001822227792, 0.001709856006, 0.00215968941, 0.002086821575, 0.001997052557, 0.002346739526, 0.002276102439, 0.001649338559, 0.001713164575, 0.001690114187, 0.00157166069, 0.001657481338, 0.001697401999, 0.001620050265],
					[0.004208510197, 0.003933823625, 0.004053391677, 0.003927757425, 0.003807930121, 0.003852669217, 0.003674315677, 0.003705470881, 0.003357423284, 0.003465333124, 0.00390359752, 0.003692743513, 0.003564945967, 0.003095657796, 0.003180273915, 0.003073912489, 0.002800009414, 0.002929667055, 0.002960508812, 0.003697354707, 0.003698115635, 0.003229609154, 0.003097156618, 0.003209536799, 0.002943742574, 0.003070495267, 0.00279314337, 0.003058055976, 0.002969772584, 0.003503084115, 0.003546939204, 0.003145192983, 0.003219439668, 0.00308664757, 0.002786969656, 0.002807109181, 0.002867665129, 0.002491761963, 0.003104430334, 0.003862731071, 0.003180756389, 0.003252099621, 0.00273071594, 0.002610622681, 0.002728077169, 0.002511774866, 0.002401525174, 0.002438591233, 0.002548323656, 0.003072985728, 0.003316450463, 0.002800158523, 0.002509262718, 0.002616284985, 0.002480641143, 0.002301242311, 0.002420236356, 0.002662738394, 0.003062524055, 0.003058764787, 0.002719144755, 0.002530666059, 0.002268505579, 0.00224192054, 0.002096645124, 0.002253077345, 0.00241413202, 0.002519864315, 0.002964938217, 0.002489035842, 0.002126379973, 0.002259785774, 0.002215060521, 0.001987783614, 0.002186204378, 0.002162518963, 0.002429174419, 0.002663709273, 0.003367061096, 0.002292200105, 0.001857039652, 0.001928067481, 0.001756471334, 0.001761635758, 0.002084969638, 0.002156477303, 0.002065400695, 0.002442677105, 0.00243369268, 0.001676106267, 0.001761323688, 0.001715023211, 0.001546563915, 0.001547746646, 0.0016355327, 0.001723242545],
					[0.00500884843, 0.00454823586, 0.0045785797, 0.004493401784, 0.004239428375, 0.004223583686, 0.003997987643, 0.004051691901, 0.003561854758, 0.003620403799, 0.005197194723, 0.004504288936, 0.004115633012, 0.003702221795, 0.003624700079, 0.003631119945, 0.003194270936, 0.003251648963, 0.002908189163, 0.003778983674, 0.0048289164, 0.004040609008, 0.003585739896, 0.003709842156, 0.003517684979, 0.003482263014, 0.002947256641, 0.003248632596, 0.003252968055, 0.003305595067, 0.004832200731, 0.004121189692, 0.003833987856, 0.003817000885, 0.003148736152, 0.003251010168, 0.003037304181, 0.002702254236, 0.0032369198, 0.003180756389, 0.004867089328, 0.003781505789, 0.003375505154, 0.00328495968, 0.003029895698, 0.00294335548, 0.002422654082, 0.002583255508, 0.002618984466, 0.003097829045, 0.004262926982, 0.003772977814, 0.003133892725, 0.003135532692, 0.002847280697, 0.002614391589, 0.002467966524, 0.00279316539, 0.002985339315, 0.003151541273, 0.003871364107, 0.003256375384, 0.002828057206, 0.002597729241, 0.002459102341, 0.002355242093, 0.002675446452, 0.002647022174, 0.00282064572, 0.003334871745, 0.002860552143, 0.002801038174, 0.002492115849, 0.002183844354, 0.002363437762, 0.002239527586, 0.002268189282, 0.00267658867, 0.00312213084, 0.00295891475, 0.00243252121, 0.002317128886, 0.001976554273, 0.001845401642, 0.00234040907, 0.002079558915, 0.001874360924, 0.002396075867, 0.002140531648, 0.002277918698, 0.002093178431, 0.002042077396, 0.001717972348, 0.001779658493, 0.001629050738, 0.001584058446],
					[0.004497613403, 0.004173927655, 0.004162307351, 0.004028057711, 0.003945152337, 0.003867575054, 0.00374307051, 0.003758263597, 0.003334818137, 0.00336627656, 0.00435597295, 0.004016010617, 0.003837504568, 0.003339200871, 0.003335812529, 0.003343044549, 0.002960725557, 0.003135708876, 0.002926999021, 0.003473167568, 0.004370629445, 0.003625382924, 0.003359651379, 0.003330488186, 0.003152410878, 0.003260850024, 0.002949088362, 0.003158675801, 0.003136741139, 0.003361283934, 0.004167746276, 0.003653686039, 0.003641186078, 0.003431157463, 0.002983505359, 0.002906715171, 0.003101957178, 0.002568549821, 0.003232893531, 0.003252099621, 0.003781505789, 0.004323497568, 0.00299665323, 0.003046330957, 0.00307135378, 0.002827228213, 0.002427048034, 0.002554118485, 0.002790492181, 0.003161271791, 0.003790943132, 0.003395870217, 0.002897485451, 0.003176837063, 0.002927014815, 0.002613668375, 0.002481251664, 0.002773992033, 0.002958353609, 0.002879259816, 0.003210875327, 0.003026247196, 0.002743137921, 0.002572044672, 0.002498455262, 0.002438483166, 0.002639653815, 0.002611863186, 0.00311687834, 0.002884994166, 0.00258528776, 0.002802613883, 0.002626465312, 0.002368972886, 0.002417322856, 0.002343205445, 0.002340431675, 0.002674308167, 0.003246847673, 0.002728429499, 0.002414224911, 0.002314975858, 0.002088830654, 0.002000047114, 0.002459099058, 0.00217047674, 0.00203353407, 0.002300070477, 0.002407624321, 0.002036565539, 0.002083728244, 0.002002175109, 0.001875702655, 0.001848791536, 0.001887412237, 0.001908320189],
					[0.003925394458, 0.00350670524, 0.003616132532, 0.003580285672, 0.003383915216, 0.003372837448, 0.003281052417, 0.003219548501, 0.002942575524, 0.002999951399, 0.003811533041, 0.003455978013, 0.003139800334, 0.002930283919, 0.002856828165, 0.00289164188, 0.002526066962, 0.002704731628, 0.002450992361, 0.003193445134, 0.003732576263, 0.003093686213, 0.002702284528, 0.002923999009, 0.002764708831, 0.002760310073, 0.002438530769, 0.00260980252, 0.002526157358, 0.002744912459, 0.003702399965, 0.003180433981, 0.002890780361, 0.002981412052, 0.002586300099, 0.002690006606, 0.002501104067, 0.002301293365, 0.002606564187, 0.00273071594, 0.003375505154, 0.00299665323, 0.003236821869, 0.002687041388, 0.002555955639, 0.002407418934, 0.002158106374, 0.002110355545, 0.00219950667, 0.00265617112, 0.003335814834, 0.002947628681, 0.002472716885, 0.002548346256, 0.002319075618, 0.00211918581, 0.002132957306, 0.002253955311, 0.002451245485, 0.002538588326, 0.002866070977, 0.002630237528, 0.002252036354, 0.002199742604, 0.002015571044, 0.001971782777, 0.002144784117, 0.002030323869, 0.002408010983, 0.002520117346, 0.002221175975, 0.002228063953, 0.002056039595, 0.001878906302, 0.00193734829, 0.001977574331, 0.001972020328, 0.002174230852, 0.002690242371, 0.002311024277, 0.001885365619, 0.00180390305, 0.001602129373, 0.001579517001, 0.001933646242, 0.001659412079, 0.001606100416, 0.001862895769, 0.001994029146, 0.001724103763, 0.001573455973, 0.001526185446, 0.001367473141, 0.001388784454, 0.001310831026, 0.001294276419],
					[0.003826366745, 0.003414971713, 0.003500002623, 0.00341839726, 0.003340113351, 0.003276297438, 0.003160486446, 0.00322410537, 0.002935013377, 0.002888635953, 0.003647806326, 0.003492681159, 0.003134726059, 0.002769878876, 0.002905681077, 0.002883360707, 0.002600968519, 0.002625437784, 0.002432798607, 0.003074032778, 0.003622061423, 0.003046358195, 0.00280620731, 0.003021807234, 0.002673637997, 0.00289823961, 0.002494613245, 0.002709653171, 0.002702039418, 0.002780883803, 0.003501187289, 0.003098678454, 0.00295686402, 0.002907614761, 0.002513571863, 0.002534592074, 0.002572921285, 0.002340550062, 0.002637783682, 0.002610622681, 0.00328495968, 0.003046330957, 0.002687041388, 0.003172722247, 0.002703964364, 0.002380647836, 0.002275645923, 0.002110789127, 0.002225447566, 0.002748700439, 0.003147833761, 0.002897016079, 0.002525782909, 0.002544187618, 0.002423581281, 0.002150737582, 0.002264520511, 0.002413970608, 0.002546536325, 0.002455990775, 0.00277449007, 0.002477180466, 0.002369015847, 0.002247473907, 0.00215330362, 0.002032713052, 0.002175111753, 0.002230589291, 0.002510170656, 0.002475605517, 0.002287571624, 0.002359078671, 0.002171076485, 0.00200555453, 0.00202633518, 0.00214252625, 0.001987810504, 0.002343696514, 0.002688894305, 0.00223512425, 0.001907502635, 0.001900859149, 0.001811115973, 0.001725504202, 0.002020156092, 0.001792794597, 0.001736940424, 0.002063233178, 0.002008129068, 0.001679055023, 0.001674292826, 0.001687232765, 0.001571458355, 0.001584612812, 0.001583883056, 0.001552290241],
					[0.003479102582, 0.003313012908, 0.00324854751, 0.003306099986, 0.003251743842, 0.003134634966, 0.003002804477, 0.003053520724, 0.002794287977, 0.002803079025, 0.003519599805, 0.003262636438, 0.003276090681, 0.002819545433, 0.002759159271, 0.002722695521, 0.002373548708, 0.002637326282, 0.00239761976, 0.003001363495, 0.003367869088, 0.002982998964, 0.002800027585, 0.002858419046, 0.002574437589, 0.002656191464, 0.002427808488, 0.002696348858, 0.002609354892, 0.002735180879, 0.003368893116, 0.002956622019, 0.003012434425, 0.002769647323, 0.002510273881, 0.002472799134, 0.002675836699, 0.002103647344, 0.002598238767, 0.002728077169, 0.003029895698, 0.00307135378, 0.002555955639, 0.002703964364, 0.003326836059, 0.002469461262, 0.002218134497, 0.002089632921, 0.002262581838, 0.002719719359, 0.00317050846, 0.002873997188, 0.002347548213, 0.002657807666, 0.002457468729, 0.002198946237, 0.002201418296, 0.002347567932, 0.002524656713, 0.00247925764, 0.002722687972, 0.002562334361, 0.002485296261, 0.002264635978, 0.002122171843, 0.002073310735, 0.002203683292, 0.00230487603, 0.00248902096, 0.002451727614, 0.002269753022, 0.002469453739, 0.002207256383, 0.002010964229, 0.002105815658, 0.00210722721, 0.002005792894, 0.002344255713, 0.002866565122, 0.002265267561, 0.002038896069, 0.001958905783, 0.001872698697, 0.001695447636, 0.00206957304, 0.00182965049, 0.001730261696, 0.002064534697, 0.001966684595, 0.001747999998, 0.001856325133, 0.001757823844, 0.001656088474, 0.001698686823, 0.001684434505, 0.00171185129],
					[0.003217506668, 0.003017905708, 0.003014847819, 0.003046845036, 0.002902764019, 0.002941816614, 0.002773694236, 0.00282870831, 0.002568939824, 0.002559037881, 0.003250822993, 0.003004556681, 0.002948850289, 0.002524422025, 0.002564853416, 0.002568207708, 0.00233798433, 0.002359066446, 0.002189184483, 0.002813582463, 0.003209881211, 0.002713429037, 0.002587252992, 0.002557295966, 0.002454666916, 0.002533025109, 0.0023698177, 0.002456309315, 0.002335447056, 0.002597800854, 0.003111353431, 0.002822098896, 0.002712519187, 0.002650533834, 0.002352049805, 0.002380507514, 0.002472353739, 0.002006869365, 0.002373926728, 0.002511774866, 0.00294335548, 0.002827228213, 0.002407418934, 0.002380647836, 0.002469461262, 0.00273628785, 0.002050882213, 0.002019935754, 0.002077264058, 0.002482654811, 0.002798265508, 0.002574299859, 0.002235745093, 0.002374428454, 0.002322647411, 0.002029250779, 0.00195440405, 0.002153424168, 0.002335183798, 0.002326288792, 0.002413520365, 0.002282572339, 0.00210209921, 0.002039529681, 0.002037678444, 0.001915372075, 0.002023075094, 0.002057182916, 0.00232234439, 0.002193428133, 0.002100723821, 0.002191852187, 0.002013411673, 0.001817086302, 0.001956677058, 0.001922936359, 0.001926255715, 0.002134125577, 0.002562093626, 0.002117743289, 0.001801102602, 0.001774330601, 0.001647109971, 0.001577969709, 0.001900135167, 0.001711745192, 0.001660298011, 0.001848218608, 0.001831724253, 0.001566536254, 0.001659315375, 0.001530303126, 0.001421089374, 0.001542052019, 0.001504801496, 0.00145597054],
					[0.003092539967, 0.002835420141, 0.002964497291, 0.002896100679, 0.002782632546, 0.002837975836, 0.002668950943, 0.002747014377, 0.002487098668, 0.002522217349, 0.002851594993, 0.002798426204, 0.002638145651, 0.002300083628, 0.002352480996, 0.002351255139, 0.002110325803, 0.002255987615, 0.002227756109, 0.002728806664, 0.002915577328, 0.00249017059, 0.00237875066, 0.002518513222, 0.002211056038, 0.002453741056, 0.002180214972, 0.002423594261, 0.002329560571, 0.002580293557, 0.002770718558, 0.002300704258, 0.002492592931, 0.002431642779, 0.002093244918, 0.002121420707, 0.002332287125, 0.001983430368, 0.002323770748, 0.002401525174, 0.002422654082, 0.002427048034, 0.002158106374, 0.002275645923, 0.002218134497, 0.002050882213, 0.002292387056, 0.001877801099, 0.001959666183, 0.002394099937, 0.002641668335, 0.002242256865, 0.001981723824, 0.002117546945, 0.002090233351, 0.001864608629, 0.001948433257, 0.002073591892, 0.00233437301, 0.002300474881, 0.002038243146, 0.002012237245, 0.001985023885, 0.001852543988, 0.001776673135, 0.001841625137, 0.00188748672, 0.001934718146, 0.002294160273, 0.001867289702, 0.001752422095, 0.001966253632, 0.001836744414, 0.00163411733, 0.001779050864, 0.001877644921, 0.001820894591, 0.002145394252, 0.002643949387, 0.001795900975, 0.001530803387, 0.001533680315, 0.001455476451, 0.001490387539, 0.001746800454, 0.001655796305, 0.001665854964, 0.001962286642, 0.001977779456, 0.001304139577, 0.001386751283, 0.001319070256, 0.001340951321, 0.00135730855, 0.001383488268, 0.001367615045],
					[0.003387487433, 0.00304936103, 0.003052088947, 0.00306826418, 0.002923215582, 0.002966669732, 0.002792739793, 0.002873165543, 0.002538668543, 0.002613303501, 0.003133866747, 0.002855994378, 0.002842945493, 0.002393712985, 0.002384606717, 0.002444673168, 0.002179840156, 0.002265448903, 0.002331667699, 0.002756412871, 0.003030936925, 0.002583465192, 0.002318001535, 0.00260775945, 0.002238484936, 0.002358155389, 0.002209502928, 0.002423369531, 0.002357951196, 0.002562534438, 0.002793733171, 0.002436448376, 0.002536415273, 0.002420464412, 0.002053063802, 0.002153921171, 0.002256236355, 0.002017001486, 0.002325965117, 0.002438591233, 0.002583255508, 0.002554118485, 0.002110355545, 0.002110789127, 0.002089632921, 0.002019935754, 0.001877801099, 0.002255582718, 0.002068414027, 0.002404775849, 0.002593694069, 0.002293415019, 0.001993140333, 0.002173903779, 0.002054921614, 0.001893534002, 0.001927413423, 0.002076444179, 0.002311205946, 0.002274966194, 0.002120337891, 0.001995792556, 0.001938888603, 0.001855814132, 0.001722785106, 0.001832567826, 0.001856676975, 0.002010200721, 0.002260962916, 0.001973444816, 0.001803630977, 0.001894799551, 0.001817529201, 0.001639164356, 0.00178368941, 0.001739598415, 0.001934769853, 0.002056383569, 0.002448597599, 0.001835817712, 0.001554517899, 0.001528020481, 0.001494005094, 0.001475744626, 0.001733468366, 0.001701345065, 0.001608682258, 0.001935430794, 0.001909424209, 0.001308641757, 0.001388652916, 0.001359905258, 0.001246940377, 0.001404625531, 0.001371462861, 0.001372454729],
					[0.003395715859, 0.003091043629, 0.003176130571, 0.003125160756, 0.003007653059, 0.00306358095, 0.0028610154, 0.002847918425, 0.002648547527, 0.002654684242, 0.003116543132, 0.002949423992, 0.003000883638, 0.002549847288, 0.002531822685, 0.002568367833, 0.002177178217, 0.002420139337, 0.002421487646, 0.002729660122, 0.003163595894, 0.002667415222, 0.002577506406, 0.002569833295, 0.002225958147, 0.002498554716, 0.002311303426, 0.002442380546, 0.00246716127, 0.002716587516, 0.002945230267, 0.002488366376, 0.002606521851, 0.002619850514, 0.002270429913, 0.002199153858, 0.002316121493, 0.002145451789, 0.002503271213, 0.002548323656, 0.002618984466, 0.002790492181, 0.00219950667, 0.002225447566, 0.002262581838, 0.002077264058, 0.001959666183, 0.002068414027, 0.002620710447, 0.002603254346, 0.002640121329, 0.002404387829, 0.002117499327, 0.00231447102, 0.002240947952, 0.002012175463, 0.001954114917, 0.002216503475, 0.002286007278, 0.002303013543, 0.002179407599, 0.002098648833, 0.002033042901, 0.001928023816, 0.001856704911, 0.001840407715, 0.002022820615, 0.002108084727, 0.002461211042, 0.002013526607, 0.001899826248, 0.00206525987, 0.00192052398, 0.001766285481, 0.001905770945, 0.00183325678, 0.001966713259, 0.002198692526, 0.002631403393, 0.001867927429, 0.001659350044, 0.001623427959, 0.001643625112, 0.001512696411, 0.001903089234, 0.001771103512, 0.001686312902, 0.001977935665, 0.001958590849, 0.001336237643, 0.001430388361, 0.001420022489, 0.001365864643, 0.00148274205, 0.001519772785, 0.001519431052],
					[0.004000762357, 0.00371479407, 0.003789134736, 0.003719892798, 0.003567505347, 0.003607251209, 0.003470426784, 0.003448010781, 0.003152359262, 0.003261252232, 0.003691230013, 0.003455392502, 0.003455188678, 0.002984198221, 0.002957736965, 0.003064479536, 0.002647144259, 0.002924262235, 0.002875366694, 0.003574148845, 0.003710364719, 0.002995624971, 0.003046000888, 0.003149431155, 0.002883740895, 0.002957772661, 0.002783458454, 0.003036829971, 0.002996461714, 0.003329739, 0.003532627186, 0.002998644227, 0.003156837627, 0.002904114795, 0.002740250691, 0.00274069719, 0.002808901081, 0.002531211976, 0.002921510826, 0.003072985728, 0.003097829045, 0.003161271791, 0.00265617112, 0.002748700439, 0.002719719359, 0.002482654811, 0.002394099937, 0.002404775849, 0.002603254346, 0.003507874446, 0.003261877243, 0.002892954422, 0.002485275888, 0.002805093691, 0.002547038788, 0.002374874245, 0.00248914718, 0.002706362721, 0.00290711498, 0.002931542597, 0.002570636166, 0.002504877642, 0.00234094259, 0.002328788354, 0.002196628596, 0.002275081176, 0.002412950967, 0.002613051957, 0.00293017241, 0.002370038322, 0.002156137142, 0.002405012624, 0.002390016422, 0.002112982315, 0.002240209112, 0.002241507991, 0.002418662726, 0.002577853902, 0.003224991981, 0.002298418854, 0.001922046187, 0.001935690547, 0.001934885145, 0.001863184141, 0.002231743966, 0.002133414729, 0.002105693351, 0.002335114657, 0.002455373533, 0.00165968462, 0.001753961274, 0.001731624546, 0.0017056357, 0.001694330807, 0.001752820866, 0.001814234354],
					[0.004921900893, 0.004518149563, 0.004571681602, 0.004420189562, 0.004212690159, 0.004047749529, 0.003914053963, 0.003931554582, 0.003521043502, 0.003583525581, 0.004951209144, 0.004442682175, 0.00405233259, 0.003592822144, 0.003428895014, 0.003569656968, 0.002943346366, 0.003350293021, 0.00298844369, 0.003742135969, 0.004846839608, 0.004059070863, 0.003738788513, 0.003752861444, 0.00337803957, 0.003434297459, 0.002905864464, 0.003284227606, 0.003250542802, 0.003374641407, 0.004823624868, 0.003872081124, 0.003945092908, 0.003772310698, 0.00312637474, 0.003199500798, 0.003053367321, 0.002653443198, 0.003188651794, 0.003316450463, 0.004262926982, 0.003790943132, 0.003335814834, 0.003147833761, 0.00317050846, 0.002798265508, 0.002641668335, 0.002593694069, 0.002640121329, 0.003261877243, 0.005037470377, 0.003785148817, 0.003021184311, 0.003144722717, 0.002826997808, 0.002607740669, 0.002574420361, 0.002691890067, 0.003095029334, 0.003187072618, 0.003814743727, 0.003287373498, 0.002838411383, 0.002605977434, 0.002375888774, 0.002473344017, 0.002512097295, 0.002746438852, 0.002892256126, 0.003257937869, 0.002711491509, 0.002887254079, 0.002549283435, 0.002204302719, 0.00230289056, 0.002175761723, 0.002439105183, 0.002668846883, 0.003424183921, 0.002966632996, 0.002436055038, 0.002355746166, 0.002025684896, 0.001889482664, 0.002335763509, 0.002037849418, 0.002005744056, 0.002458236182, 0.002416220994, 0.002355888284, 0.002182515245, 0.002002751956, 0.001770018254, 0.001689529892, 0.001693807964, 0.001633407093],
					[0.004177801985, 0.003944750725, 0.003864248166, 0.003735997127, 0.003554662776, 0.003518512219, 0.003418586327, 0.003313178608, 0.003050485455, 0.003009493202, 0.004333963728, 0.003868226071, 0.003689967188, 0.00315543466, 0.003150561594, 0.003122008957, 0.002783759962, 0.002951902598, 0.002606981108, 0.003261816697, 0.004181164132, 0.003637699749, 0.003156040757, 0.003185902301, 0.003022747245, 0.003012799755, 0.002661201218, 0.0029268648, 0.002808323347, 0.002900609376, 0.004206178584, 0.003577494119, 0.003427932124, 0.003295540396, 0.002822365865, 0.002891564344, 0.002762400551, 0.002399420782, 0.002843995701, 0.002800158523, 0.003772977814, 0.003395870217, 0.002947628681, 0.002897016079, 0.002873997188, 0.002574299859, 0.002242256865, 0.002293415019, 0.002404387829, 0.002892954422, 0.003785148817, 0.003795038486, 0.00274344516, 0.002846294208, 0.002643327318, 0.002373651119, 0.002450064418, 0.002525335517, 0.002638444729, 0.002732705094, 0.003402279329, 0.002927043135, 0.002580653227, 0.002380309535, 0.002335352466, 0.002231337129, 0.0024369971, 0.002456676516, 0.002660254894, 0.002915525926, 0.002531438179, 0.002613211472, 0.002426121036, 0.002052740998, 0.002158481331, 0.0021140707, 0.002201615384, 0.00245250601, 0.002904037504, 0.002652139789, 0.002216415693, 0.002137932195, 0.001988579044, 0.001773262991, 0.002244430763, 0.001871709159, 0.001750546507, 0.002199654503, 0.002098263899, 0.002135683637, 0.002002634763, 0.001942345413, 0.001702335031, 0.001709835884, 0.001653003911, 0.001587142632],
					[0.003654297707, 0.003314415963, 0.003371414581, 0.003210599149, 0.003178857495, 0.003220634491, 0.003024243577, 0.003028926938, 0.00263296343, 0.002728739572, 0.003606982422, 0.003236383866, 0.002976667966, 0.002677606164, 0.00263577638, 0.00274180904, 0.00247288644, 0.002461926978, 0.002317270052, 0.002881400693, 0.003579984805, 0.002998394955, 0.002682901711, 0.002801078981, 0.002552899392, 0.002586533688, 0.002421328055, 0.002456698341, 0.002478598141, 0.002693202523, 0.003434773421, 0.003077587403, 0.002807858163, 0.002766997167, 0.002439855205, 0.002373193014, 0.002330442759, 0.002150288726, 0.002491374206, 0.002509262718, 0.003133892725, 0.002897485451, 0.002472716885, 0.002525782909, 0.002347548213, 0.002235745093, 0.001981723824, 0.001993140333, 0.002117499327, 0.002485275888, 0.003021184311, 0.00274344516, 0.002907841859, 0.002360326866, 0.002201432646, 0.00197900765, 0.002028213155, 0.002182706382, 0.002323935946, 0.002457778017, 0.002650656099, 0.00229871417, 0.002193479256, 0.002106481104, 0.001988621986, 0.001907748692, 0.002067338217, 0.002054173723, 0.002371889688, 0.002318416621, 0.002158656476, 0.002257302626, 0.001995026144, 0.001761774637, 0.001884896651, 0.001870519581, 0.001795517347, 0.002097052813, 0.002458630052, 0.00220955019, 0.001801943668, 0.001859650234, 0.00158182962, 0.00155812396, 0.001802007614, 0.00162487014, 0.001567712893, 0.001873815258, 0.001786576186, 0.001529337994, 0.001611606875, 0.001470394496, 0.001221941677, 0.001405332314, 0.001389306134, 0.001394237652],
					[0.003598637829, 0.00343389213, 0.003366216157, 0.003358889658, 0.003158288587, 0.003225088129, 0.003035802487, 0.003032511577, 0.002732898344, 0.00276611704, 0.003729182959, 0.003332541527, 0.00317508292, 0.002848600657, 0.002772655784, 0.002790751101, 0.002423331136, 0.002655913987, 0.002405332647, 0.00295569575, 0.003684863634, 0.003056517742, 0.002738199901, 0.002773726189, 0.002698310427, 0.002659368438, 0.00247621901, 0.002705616909, 0.002593966541, 0.002619729093, 0.003554204283, 0.003050500899, 0.003140722747, 0.002774585934, 0.002520047139, 0.002513681059, 0.002681173546, 0.002176605807, 0.002673806675, 0.002616284985, 0.003135532692, 0.003176837063, 0.002548346256, 0.002544187618, 0.002657807666, 0.002374428454, 0.002117546945, 0.002173903779, 0.00231447102, 0.002805093691, 0.003144722717, 0.002846294208, 0.002360326866, 0.003149523875, 0.002415837636, 0.002200086391, 0.002162371878, 0.002324971658, 0.002461091217, 0.002459801104, 0.00268770369, 0.00263931478, 0.002375925292, 0.002159224955, 0.002075736788, 0.002131126522, 0.002320906714, 0.002170903329, 0.002468893541, 0.002472894677, 0.002205323436, 0.002454136129, 0.002299884068, 0.00205816416, 0.002082048619, 0.002101663354, 0.00213781531, 0.002254171392, 0.002674545428, 0.002328507026, 0.002047874052, 0.001979971247, 0.001877199216, 0.001696418522, 0.002103916017, 0.001831609016, 0.001815691681, 0.002105651673, 0.002104577206, 0.001764384174, 0.001754955104, 0.001726516932, 0.00164620921, 0.00164119908, 0.001632717802, 0.00170226126],
					[0.00347358152, 0.003213881416, 0.003231814936, 0.003149259674, 0.003031807835, 0.003074806453, 0.002865340173, 0.002911464346, 0.002639358885, 0.00257947995, 0.003338004342, 0.003093094158, 0.003002616169, 0.002637211277, 0.00257720391, 0.00260473893, 0.002272073889, 0.002527600965, 0.002385423734, 0.002771345251, 0.003303553928, 0.002810718864, 0.002771310142, 0.002783920954, 0.002457512633, 0.002638276807, 0.002356458479, 0.002486510885, 0.002468548448, 0.002685051147, 0.00317215526, 0.002748571156, 0.002845330975, 0.002778439634, 0.002422558995, 0.002319327304, 0.002527505865, 0.002159345857, 0.002571918771, 0.002480641143, 0.002847280697, 0.002927014815, 0.002319075618, 0.002423581281, 0.002457468729, 0.002322647411, 0.002090233351, 0.002054921614, 0.002240947952, 0.002547038788, 0.002826997808, 0.002643327318, 0.002201432646, 0.002415837636, 0.002767315247, 0.002043651733, 0.002034993437, 0.002204876043, 0.00234345839, 0.002419818446, 0.002411671324, 0.002288762897, 0.002288017767, 0.002072733425, 0.002008139257, 0.001982204687, 0.002128501004, 0.002193425371, 0.002539332658, 0.002208228764, 0.002025280393, 0.002246620993, 0.002133769892, 0.001897569735, 0.002027384031, 0.001938074558, 0.001924333968, 0.002256134662, 0.002770495738, 0.00210142722, 0.001832840001, 0.001837111486, 0.001635391534, 0.001627968138, 0.002071366733, 0.00171055106, 0.001684877891, 0.001950284629, 0.002041913208, 0.0014921953, 0.001613936507, 0.001541106991, 0.001471749075, 0.001556784332, 0.001587092448, 0.001462726316],
					[0.003243638023, 0.003034389219, 0.003016512653, 0.002974762136, 0.002869155025, 0.002869951923, 0.002746671583, 0.002745567374, 0.00245577162, 0.002479049424, 0.003044495193, 0.002851603943, 0.002716058382, 0.002388155994, 0.002387000641, 0.002401464726, 0.002157631265, 0.002327256475, 0.002234018339, 0.00263992019, 0.002993021396, 0.002538625765, 0.002361977066, 0.002526022159, 0.0023063138, 0.002365156932, 0.002146446492, 0.002321681675, 0.00228161994, 0.00240451075, 0.002848030125, 0.002444597683, 0.002578218747, 0.002491512264, 0.002079851963, 0.002084102308, 0.002247129268, 0.001969783828, 0.002379817153, 0.002301242311, 0.002614391589, 0.002613668375, 0.00211918581, 0.002150737582, 0.002198946237, 0.002029250779, 0.001864608629, 0.001893534002, 0.002012175463, 0.002374874245, 0.002607740669, 0.002373651119, 0.00197900765, 0.002200086391, 0.002043651733, 0.002258678707, 0.001848273189, 0.002082300372, 0.002138891732, 0.002093305463, 0.002110123955, 0.002090690778, 0.001923976682, 0.001807177388, 0.001846959803, 0.001777795269, 0.001862867801, 0.002014353121, 0.002167454362, 0.00190608691, 0.001789749635, 0.001991950908, 0.001904169139, 0.001637374176, 0.00181615008, 0.001761109708, 0.001886317215, 0.001973870058, 0.002307852321, 0.001854531318, 0.001563540612, 0.001555407534, 0.001528674383, 0.001389151699, 0.001813067579, 0.00165444263, 0.001623472287, 0.00179482723, 0.001723097603, 0.001368238167, 0.001434811152, 0.001338967163, 0.001364436889, 0.001395907181, 0.001442945758, 0.001410140658],
					[0.003030619633, 0.002929474526, 0.002922320365, 0.002953982871, 0.002756593414, 0.002794455413, 0.002713610852, 0.002658174324, 0.00240993133, 0.002497650293, 0.002987117227, 0.002816757742, 0.002865345579, 0.002352838588, 0.002357687016, 0.002339342052, 0.002150761998, 0.002300116212, 0.002249965789, 0.002789369528, 0.003079061485, 0.0024774441, 0.00228325972, 0.002509305232, 0.002281525885, 0.002266873072, 0.002187583781, 0.002466251737, 0.002341879803, 0.002498561864, 0.002739330765, 0.002435739537, 0.002494993998, 0.002358714498, 0.002235407413, 0.00224190194, 0.002221462671, 0.001983946335, 0.002362586563, 0.002420236356, 0.002467966524, 0.002481251664, 0.002132957306, 0.002264520511, 0.002201418296, 0.00195440405, 0.001948433257, 0.001927413423, 0.001954114917, 0.00248914718, 0.002574420361, 0.002450064418, 0.002028213155, 0.002162371878, 0.002034993437, 0.001848273189, 0.002538786963, 0.00222124262, 0.002253509776, 0.002253188285, 0.002299911894, 0.0020225735, 0.001944254201, 0.001863789412, 0.001819545462, 0.001845145489, 0.001922706237, 0.002048916468, 0.00233515705, 0.002032495395, 0.001825373988, 0.001888469975, 0.001866935206, 0.001704091112, 0.001790801096, 0.001876568938, 0.001794484491, 0.002072880281, 0.00253783427, 0.001899653245, 0.001507957861, 0.001604302022, 0.001605942258, 0.001543079407, 0.001825533626, 0.001679398022, 0.001668945698, 0.001919123145, 0.001938805242, 0.001453504753, 0.001421291285, 0.001527167595, 0.001413216884, 0.001448010909, 0.001478431473, 0.001485237838],
					[0.003440298944, 0.00334420473, 0.003274787097, 0.003297600332, 0.003070416077, 0.00314857206, 0.002952582409, 0.003029978077, 0.002745503964, 0.00278677727, 0.003204243619, 0.003150937277, 0.002911171213, 0.002637722498, 0.002638124, 0.002678514626, 0.002429601622, 0.002474336877, 0.002484491958, 0.003027644353, 0.003216678089, 0.00270785449, 0.00259919058, 0.00270304282, 0.002496782109, 0.002727073654, 0.002492489763, 0.002640528653, 0.002538539563, 0.002693658519, 0.003111718767, 0.002681119544, 0.002687851375, 0.00263069605, 0.002373286904, 0.002394257036, 0.002480365542, 0.002136173418, 0.002660356607, 0.002662738394, 0.00279316539, 0.002773992033, 0.002253955311, 0.002413970608, 0.002347567932, 0.002153424168, 0.002073591892, 0.002076444179, 0.002216503475, 0.002706362721, 0.002691890067, 0.002525335517, 0.002182706382, 0.002324971658, 0.002204876043, 0.002082300372, 0.00222124262, 0.002888158837, 0.00246768224, 0.002545452725, 0.002354012316, 0.002171205185, 0.002126558897, 0.002073529808, 0.002061390344, 0.002037461461, 0.00216609843, 0.00225437478, 0.002504191509, 0.002126676968, 0.001914553843, 0.002159707946, 0.002024030435, 0.001879695348, 0.00199321726, 0.002005570746, 0.002129976208, 0.002232949935, 0.002727763283, 0.002024908208, 0.001716670092, 0.001777444928, 0.001822485902, 0.001672337281, 0.00202609524, 0.001939476904, 0.001907641035, 0.002103609887, 0.002081524935, 0.001497014391, 0.001545671162, 0.001640709803, 0.001567264754, 0.001563318738, 0.001658344814, 0.001638997416],
					[0.003931812779, 0.003452367593, 0.003685314298, 0.003627668097, 0.003440003388, 0.003445206259, 0.003384751458, 0.003456979638, 0.003105609008, 0.003201191769, 0.003555873242, 0.003371518839, 0.0033207877, 0.00285052315, 0.002886483586, 0.002819510233, 0.00258608262, 0.002739514386, 0.002751166494, 0.003339619025, 0.003516415998, 0.003038493889, 0.00286215559, 0.002986993227, 0.002708649286, 0.00298885925, 0.002553266905, 0.002961271301, 0.002828931675, 0.003142076927, 0.003134480965, 0.002759195476, 0.002980382082, 0.002811431806, 0.002521135822, 0.00256134582, 0.002739803121, 0.002299743107, 0.002844804395, 0.003062524055, 0.002985339315, 0.002958353609, 0.002451245485, 0.002546536325, 0.002524656713, 0.002335183798, 0.00233437301, 0.002311205946, 0.002286007278, 0.00290711498, 0.003095029334, 0.002638444729, 0.002323935946, 0.002461091217, 0.00234345839, 0.002138891732, 0.002253509776, 0.00246768224, 0.003350583833, 0.002772584344, 0.002545890912, 0.00228588627, 0.002134542835, 0.002105589144, 0.001972762346, 0.002065197164, 0.002134455168, 0.002349684212, 0.002885162927, 0.002390285755, 0.001977518945, 0.002179275575, 0.002159351368, 0.001854111527, 0.002045975845, 0.002050252042, 0.002150942185, 0.002416861729, 0.00308709855, 0.002117514202, 0.001769438899, 0.001741036688, 0.001707515433, 0.001644695146, 0.00200538994, 0.001967585467, 0.001912712957, 0.002272632476, 0.002188558276, 0.001539832924, 0.001644256099, 0.001611616989, 0.001492172521, 0.001583504098, 0.001624889665, 0.001665134061],
					[0.004004083997, 0.003555839981, 0.003723157506, 0.003716680489, 0.003495357866, 0.003484276388, 0.003398783072, 0.003418816348, 0.003080811196, 0.003171901734, 0.003764693424, 0.003421802213, 0.003383236234, 0.002949664632, 0.002862711554, 0.002970506193, 0.002629567419, 0.002810675895, 0.002687481463, 0.003391684513, 0.003563033468, 0.003132510879, 0.002856169711, 0.003066809247, 0.002886268458, 0.002962162727, 0.002613346258, 0.002871738576, 0.002913652699, 0.003097697868, 0.003584368242, 0.003072875462, 0.002966540382, 0.002873848244, 0.002633358523, 0.00252863471, 0.002757168903, 0.002245431262, 0.002827361159, 0.003058764787, 0.003151541273, 0.002879259816, 0.002538588326, 0.002455990775, 0.00247925764, 0.002326288792, 0.002300474881, 0.002274966194, 0.002303013543, 0.002931542597, 0.003187072618, 0.002732705094, 0.002457778017, 0.002459801104, 0.002419818446, 0.002093305463, 0.002253188285, 0.002545452725, 0.002772584344, 0.004079755781, 0.002537002164, 0.002413232926, 0.002321551262, 0.002189272422, 0.002113071136, 0.002099741406, 0.002304073073, 0.002393506739, 0.002693485986, 0.002321912481, 0.002075753427, 0.002323698689, 0.00217477644, 0.00173595849, 0.002037136712, 0.002029942012, 0.002329128467, 0.002497822871, 0.002997682812, 0.002208828426, 0.0018639162, 0.001807356548, 0.001639376038, 0.001659224599, 0.002125126885, 0.001869063588, 0.00195522756, 0.002256425015, 0.002297224771, 0.001523107793, 0.001675158753, 0.001569474671, 0.001440842522, 0.001625692915, 0.001511484483, 0.00162087977],
					[0.00379606579, 0.003597330441, 0.003694396668, 0.003460502947, 0.003285121985, 0.003242065918, 0.003164531892, 0.003089516401, 0.002765467253, 0.002852639794, 0.004334541133, 0.003810412801, 0.003420739267, 0.003039306391, 0.002969202312, 0.002821127856, 0.002489536208, 0.002664617975, 0.002383835896, 0.003009037919, 0.004092782743, 0.003491058531, 0.002929151204, 0.003032034749, 0.002825941231, 0.002724667161, 0.002381112847, 0.002652598675, 0.002624850473, 0.00259576905, 0.004337202597, 0.003471787293, 0.00324307985, 0.003285817904, 0.002729913204, 0.002803873853, 0.00244546032, 0.002152025938, 0.002658962851, 0.002719144755, 0.003871364107, 0.003210875327, 0.002866070977, 0.00277449007, 0.002722687972, 0.002413520365, 0.002038243146, 0.002120337891, 0.002179407599, 0.002570636166, 0.003814743727, 0.003402279329, 0.002650656099, 0.00268770369, 0.002411671324, 0.002110123955, 0.002299911894, 0.002354012316, 0.002545890912, 0.002537002164, 0.004115982628, 0.00285063903, 0.002468721806, 0.002265342652, 0.002177720248, 0.002024486222, 0.002346885696, 0.002171199726, 0.002353119122, 0.003107641537, 0.002544776951, 0.002392017713, 0.0021142444, 0.002042435891, 0.001953488988, 0.00201758249, 0.002020718018, 0.002274620913, 0.00277005577, 0.002698335303, 0.002290962724, 0.002012465449, 0.00185434367, 0.001648297176, 0.002036498587, 0.001800994314, 0.001637370346, 0.001989093522, 0.001834252413, 0.00224975273, 0.001895068104, 0.001898527721, 0.001539200166, 0.001481003002, 0.001471806275, 0.001417212846],
					[0.00338809299, 0.003228763006, 0.003228902136, 0.003201833917, 0.003009518249, 0.00296896478, 0.002893749342, 0.002823881199, 0.002570024624, 0.002556324433, 0.003553878345, 0.003295890373, 0.003050888097, 0.002737754808, 0.002639004767, 0.002624749646, 0.002321161174, 0.002577909445, 0.002218656927, 0.0027456759, 0.003510117601, 0.00292253278, 0.002730446441, 0.002679082237, 0.002597498699, 0.002603957511, 0.002299599325, 0.002495345023, 0.002433160548, 0.002535927788, 0.003567835962, 0.00303231373, 0.003040830314, 0.002866599622, 0.00243657735, 0.002419776366, 0.002450922864, 0.00202899222, 0.002517867529, 0.002530666059, 0.003256375384, 0.003026247196, 0.002630237528, 0.002477180466, 0.002562334361, 0.002282572339, 0.002012237245, 0.001995792556, 0.002098648833, 0.002504877642, 0.003287373498, 0.002927043135, 0.00229871417, 0.00263931478, 0.002288762897, 0.002090690778, 0.0020225735, 0.002171205185, 0.00228588627, 0.002413232926, 0.00285063903, 0.002928884545, 0.002343570116, 0.002190788495, 0.002009566027, 0.001967943614, 0.002072171205, 0.002006980689, 0.002390876113, 0.002517607928, 0.002260702104, 0.002359948953, 0.002099432758, 0.001857530485, 0.001919568355, 0.001893251837, 0.001914514415, 0.002093109033, 0.002478238425, 0.002328326893, 0.002022264844, 0.001956531438, 0.00171365323, 0.001608642565, 0.001983211094, 0.001738711241, 0.001653203216, 0.001914985238, 0.001870729067, 0.00181912707, 0.001775680814, 0.001709574839, 0.001544124909, 0.001523394415, 0.001502743129, 0.001442843411],
					[0.003208075825, 0.002824046695, 0.002950115286, 0.002927183154, 0.002794968285, 0.002867250372, 0.002576146487, 0.002694148822, 0.002403539014, 0.002395752497, 0.003231203484, 0.002979283605, 0.002846153878, 0.002598694665, 0.002423556742, 0.002501434852, 0.00216214058, 0.002402551657, 0.002133343147, 0.002538829406, 0.003131965539, 0.002812505344, 0.002602943627, 0.002584795124, 0.002309029674, 0.00239513157, 0.002138305566, 0.002318905424, 0.002332568494, 0.002366890196, 0.003097652087, 0.002760249885, 0.002747754518, 0.002514428298, 0.002218249685, 0.002180843066, 0.002309054272, 0.001868377558, 0.0022987776, 0.002268505579, 0.002828057206, 0.002743137921, 0.002252036354, 0.002369015847, 0.002485296261, 0.00210209921, 0.001985023885, 0.001938888603, 0.002033042901, 0.00234094259, 0.002838411383, 0.002580653227, 0.002193479256, 0.002375925292, 0.002288017767, 0.001923976682, 0.001944254201, 0.002126558897, 0.002134542835, 0.002321551262, 0.002468721806, 0.002343570116, 0.002667728836, 0.002059014354, 0.001875570377, 0.001815770394, 0.001926379175, 0.001910367164, 0.002259947412, 0.002210713674, 0.002105304897, 0.00227225667, 0.002006248791, 0.001678812607, 0.00184701393, 0.00179998138, 0.001704258851, 0.002055336166, 0.002590616688, 0.002073480187, 0.001874480495, 0.001817398768, 0.00158261204, 0.00160578189, 0.001834642637, 0.001570663998, 0.001592933608, 0.002057573856, 0.001761267425, 0.001508379831, 0.001597609687, 0.001539238963, 0.001439716463, 0.001489612646, 0.001449216874, 0.001435943811],
					[0.002935523324, 0.002719042358, 0.002751010898, 0.002807933105, 0.002665138781, 0.002673288533, 0.002534087762, 0.002581815013, 0.002365928298, 0.002348819196, 0.002943099406, 0.002720910432, 0.002658557789, 0.002320043343, 0.002302659107, 0.002245179492, 0.002093913888, 0.002241335738, 0.002062347094, 0.00246199063, 0.002865249967, 0.002501828039, 0.002393241764, 0.00242750741, 0.002188957955, 0.002366851537, 0.002185859516, 0.0022650826, 0.002191589414, 0.002410639485, 0.002903839306, 0.002524300375, 0.002462205185, 0.002443524259, 0.002148550824, 0.002103332855, 0.002160272368, 0.001880645303, 0.002230930061, 0.00224192054, 0.002597729241, 0.002572044672, 0.002199742604, 0.002247473907, 0.002264635978, 0.002039529681, 0.001852543988, 0.001855814132, 0.001928023816, 0.002328788354, 0.002605977434, 0.002380309535, 0.002106481104, 0.002159224955, 0.002072733425, 0.001807177388, 0.001863789412, 0.002073529808, 0.002105589144, 0.002189272422, 0.002265342652, 0.002190788495, 0.002059014354, 0.002239974825, 0.001860978705, 0.001763013832, 0.001851172919, 0.001813111348, 0.002145930135, 0.002033279861, 0.001984916688, 0.002110346607, 0.001896770696, 0.001675683985, 0.001751742541, 0.001808930336, 0.001737045766, 0.001859046805, 0.002208657034, 0.001916434512, 0.001698026684, 0.001711879501, 0.001552914235, 0.001560586611, 0.001794451241, 0.001666964874, 0.001556821933, 0.001761721129, 0.001674310159, 0.001338224906, 0.001469942223, 0.001372897973, 0.001312547653, 0.001355054415, 0.001392112944, 0.001311553054],
					[0.002579952426, 0.002595143005, 0.002432745554, 0.002393663894, 0.002399572175, 0.002378738807, 0.002298139155, 0.002274152315, 0.002060583994, 0.002024066215, 0.002696793804, 0.002541390204, 0.00248201665, 0.002149391811, 0.002211500634, 0.002192825626, 0.001967676888, 0.002078977645, 0.001937510594, 0.002241391942, 0.002725611012, 0.002316819949, 0.002125213045, 0.002219893885, 0.002129636242, 0.002204672391, 0.002131839155, 0.002093950813, 0.00211624525, 0.002165431495, 0.002862427116, 0.002394184833, 0.00236962802, 0.002274656068, 0.00206209881, 0.002042384937, 0.002095941077, 0.001805562653, 0.002179070837, 0.002096645124, 0.002459102341, 0.002498455262, 0.002015571044, 0.00215330362, 0.002122171843, 0.002037678444, 0.001776673135, 0.001722785106, 0.001856704911, 0.002196628596, 0.002375888774, 0.002335352466, 0.001988621986, 0.002075736788, 0.002008139257, 0.001846959803, 0.001819545462, 0.002061390344, 0.001972762346, 0.002113071136, 0.002177720248, 0.002009566027, 0.001875570377, 0.001860978705, 0.00217574472, 0.001695211453, 0.001905841881, 0.001938044055, 0.001955343717, 0.001947203413, 0.001835683003, 0.00199579587, 0.001878570542, 0.001637405318, 0.001789838079, 0.001715748504, 0.001829044759, 0.001879697817, 0.00210540579, 0.001800355523, 0.0016475462, 0.001614763026, 0.001592853435, 0.001466974956, 0.001730962735, 0.001611239802, 0.001517674255, 0.001656282469, 0.00164398893, 0.001406303153, 0.001505474531, 0.001409622401, 0.001346738712, 0.001380300963, 0.001465920283, 0.001417331467],
					[0.002747787374, 0.00264760704, 0.002705575637, 0.002612167028, 0.002482747036, 0.002476341968, 0.002369198989, 0.002452890052, 0.002128802882, 0.002145725437, 0.002640047186, 0.002527176121, 0.002494794761, 0.002141692279, 0.002164797252, 0.002164558022, 0.001976377986, 0.002098506638, 0.001986508049, 0.002453759042, 0.002733212066, 0.002253495925, 0.002038210391, 0.002286300568, 0.002074964307, 0.002124099997, 0.002043831445, 0.002171420609, 0.00214557857, 0.002328296499, 0.002742707281, 0.002314679862, 0.002326459033, 0.00221991396, 0.002026451581, 0.001968265032, 0.002156213221, 0.001860821108, 0.002236853026, 0.002253077345, 0.002355242093, 0.002438483166, 0.001971782777, 0.002032713052, 0.002073310735, 0.001915372075, 0.001841625137, 0.001832567826, 0.001840407715, 0.002275081176, 0.002473344017, 0.002231337129, 0.001907748692, 0.002131126522, 0.001982204687, 0.001777795269, 0.001845145489, 0.002037461461, 0.002065197164, 0.002099741406, 0.002024486222, 0.001967943614, 0.001815770394, 0.001763013832, 0.001695211453, 0.002226843776, 0.001831680806, 0.001941214445, 0.002092000333, 0.001879618498, 0.001749958111, 0.001914182171, 0.001800106007, 0.001640827494, 0.001703409595, 0.00178709596, 0.001800410688, 0.001956771025, 0.002342896959, 0.001864830712, 0.001542859377, 0.00157997703, 0.001570629667, 0.001497027997, 0.001760054456, 0.001638850111, 0.001596252795, 0.001828888462, 0.001783426135, 0.001414236235, 0.001439126627, 0.001421393745, 0.001257311173, 0.001417501875, 0.001434609122, 0.001413057928],
					[0.002993945136, 0.002929607457, 0.002880348639, 0.002891811709, 0.002766082028, 0.00278931335, 0.002650016671, 0.002639344627, 0.002378697448, 0.002443764267, 0.003154011213, 0.002777588485, 0.002796370631, 0.002455331697, 0.002527220919, 0.002394731624, 0.00212531014, 0.002261782071, 0.002164657603, 0.002658168093, 0.002940756862, 0.002516935413, 0.002278688507, 0.002416626155, 0.002305354962, 0.002254771995, 0.002271061995, 0.002351905388, 0.002334825057, 0.00245450379, 0.00303943435, 0.002563165524, 0.002444427051, 0.002444334685, 0.0021809998, 0.002240327534, 0.002307032847, 0.001986037911, 0.002385015081, 0.00241413202, 0.002675446452, 0.002639653815, 0.002144784117, 0.002175111753, 0.002203683292, 0.002023075094, 0.00188748672, 0.001856676975, 0.002022820615, 0.002412950967, 0.002512097295, 0.0024369971, 0.002067338217, 0.002320906714, 0.002128501004, 0.001862867801, 0.001922706237, 0.00216609843, 0.002134455168, 0.002304073073, 0.002346885696, 0.002072171205, 0.001926379175, 0.001851172919, 0.001905841881, 0.001831680806, 0.002362971986, 0.002013099909, 0.002211800006, 0.002037217978, 0.001839230725, 0.001984006171, 0.001901105125, 0.001751261424, 0.001834798411, 0.001852889709, 0.001891929825, 0.002045778129, 0.002521603449, 0.001912894773, 0.001712256401, 0.001621937332, 0.001611890165, 0.001470146558, 0.001863226423, 0.001699113779, 0.001664173048, 0.001824116362, 0.001925407783, 0.001432696674, 0.001447393032, 0.001488967379, 0.001353134112, 0.001441163181, 0.001423381553, 0.001498832196],
					[0.003157948431, 0.003064194774, 0.002975663554, 0.002789078706, 0.002807242675, 0.002744695175, 0.002596581078, 0.002777070666, 0.002375537224, 0.002445315495, 0.0030178547, 0.00282129975, 0.002825190452, 0.002372667954, 0.002339467906, 0.002495798699, 0.002052453022, 0.002340604819, 0.002328611956, 0.002796176512, 0.002965537976, 0.002451923454, 0.002451203521, 0.002570982108, 0.002290907059, 0.002485193819, 0.00229176301, 0.002521800233, 0.002464649154, 0.002711650031, 0.002918479447, 0.002435194682, 0.002492011945, 0.002513984598, 0.002184221162, 0.002214681019, 0.002425733908, 0.002040183949, 0.002537127192, 0.002519864315, 0.002647022174, 0.002611863186, 0.002030323869, 0.002230589291, 0.00230487603, 0.002057182916, 0.001934718146, 0.002010200721, 0.002108084727, 0.002613051957, 0.002746438852, 0.002456676516, 0.002054173723, 0.002170903329, 0.002193425371, 0.002014353121, 0.002048916468, 0.00225437478, 0.002349684212, 0.002393506739, 0.002171199726, 0.002006980689, 0.001910367164, 0.001813111348, 0.001938044055, 0.001941214445, 0.002013099909, 0.002913229606, 0.002342612304, 0.002031034186, 0.001806357272, 0.001986177557, 0.001948085404, 0.001738131697, 0.001961482042, 0.001842175714, 0.002085970703, 0.002257239072, 0.002693510647, 0.001937579107, 0.001624053584, 0.001595164909, 0.001618440704, 0.001438808775, 0.001889528393, 0.001777879072, 0.001792062365, 0.00216390251, 0.001970141906, 0.001485800171, 0.001609694131, 0.001519896088, 0.001452871784, 0.001512179815, 0.001595801621, 0.00157436436],
					[0.003715079122, 0.003378001969, 0.003502498329, 0.00345071268, 0.003305060105, 0.003419811664, 0.003333011815, 0.003271405217, 0.002979944128, 0.002970631164, 0.003233047781, 0.003289166972, 0.003221488961, 0.002748106064, 0.002896492621, 0.002804875205, 0.002483985727, 0.002765602047, 0.002693200054, 0.003167364769, 0.003303248719, 0.002803321185, 0.002800159684, 0.002968969687, 0.002604838112, 0.002957245289, 0.00263039294, 0.002877079814, 0.002830893306, 0.003238324681, 0.002895614018, 0.002696053478, 0.002907310129, 0.002864591507, 0.002537095063, 0.002462364389, 0.00276051117, 0.002325492097, 0.002868935179, 0.002964938217, 0.00282064572, 0.00311687834, 0.002408010983, 0.002510170656, 0.00248902096, 0.00232234439, 0.002294160273, 0.002260962916, 0.002461211042, 0.00293017241, 0.002892256126, 0.002660254894, 0.002371889688, 0.002468893541, 0.002539332658, 0.002167454362, 0.00233515705, 0.002504191509, 0.002885162927, 0.002693485986, 0.002353119122, 0.002390876113, 0.002259947412, 0.002145930135, 0.001955343717, 0.002092000333, 0.002211800006, 0.002342612304, 0.003658817031, 0.00212495448, 0.001982733737, 0.002274283814, 0.002126321301, 0.001906339992, 0.002098460066, 0.002175184895, 0.002123685539, 0.002434289261, 0.003287733483, 0.002143422865, 0.001787322454, 0.001853956393, 0.001761340961, 0.001761927511, 0.002223767524, 0.00202875252, 0.002015049614, 0.002359241942, 0.002349290251, 0.001502287893, 0.001690429149, 0.001701696033, 0.001637735267, 0.001694576331, 0.001727829921, 0.001714672068],
					[0.003337110603, 0.003206598288, 0.00322837047, 0.003015226454, 0.002907239952, 0.002865320479, 0.002746955289, 0.002731006516, 0.002466248967, 0.002471499072, 0.003705899965, 0.003287562718, 0.002953910851, 0.002744602577, 0.002636127797, 0.002463355151, 0.00221991072, 0.002390797271, 0.002124462034, 0.002702996108, 0.003620976929, 0.003086465809, 0.002582754777, 0.002657204096, 0.002458205216, 0.002441650783, 0.00211106795, 0.002384031672, 0.002365428219, 0.002348861861, 0.003558485885, 0.003113548988, 0.002862370156, 0.002849514422, 0.002429949287, 0.002424836448, 0.002288477239, 0.001988182248, 0.002362173787, 0.002489035842, 0.003334871745, 0.002884994166, 0.002520117346, 0.002475605517, 0.002451727614, 0.002193428133, 0.001867289702, 0.001973444816, 0.002013526607, 0.002370038322, 0.003257937869, 0.002915525926, 0.002318416621, 0.002472894677, 0.002208228764, 0.00190608691, 0.002032495395, 0.002126676968, 0.002390285755, 0.002321912481, 0.003107641537, 0.002517607928, 0.002210713674, 0.002033279861, 0.001947203413, 0.001879618498, 0.002037217978, 0.002031034186, 0.00212495448, 0.002986835437, 0.002274080231, 0.002230011678, 0.002045158491, 0.001799474408, 0.001782427378, 0.00181265829, 0.001825797102, 0.002054788709, 0.002457843811, 0.002349393317, 0.00199023081, 0.001857663227, 0.001724921252, 0.001551595006, 0.001880673805, 0.001614511409, 0.001502270361, 0.001976858326, 0.001714422288, 0.00193244399, 0.00173933189, 0.001693475392, 0.0014430677, 0.001442626283, 0.001374532828, 0.001388198613],
					[0.002848833667, 0.002600477361, 0.002724240891, 0.002634119186, 0.002521817319, 0.00253830332, 0.0024088781, 0.002417733311, 0.002183305605, 0.00217780669, 0.003057049188, 0.00280330556, 0.002647777308, 0.002357861268, 0.002371647666, 0.002245615832, 0.002045275133, 0.002148585308, 0.001949411141, 0.002418751103, 0.003034990502, 0.002570900195, 0.002317755746, 0.002334412356, 0.002175005627, 0.002273750322, 0.002025041307, 0.002151030603, 0.002215903769, 0.002259480343, 0.003049905475, 0.002701328888, 0.002476277853, 0.002484707529, 0.002107394936, 0.002049107198, 0.002059309865, 0.001757660917, 0.002143727368, 0.002126379973, 0.002860552143, 0.00258528776, 0.002221175975, 0.002287571624, 0.002269753022, 0.002100723821, 0.001752422095, 0.001803630977, 0.001899826248, 0.002156137142, 0.002711491509, 0.002531438179, 0.002158656476, 0.002205323436, 0.002025280393, 0.001789749635, 0.001825373988, 0.001914553843, 0.001977518945, 0.002075753427, 0.002544776951, 0.002260702104, 0.002105304897, 0.001984916688, 0.001835683003, 0.001749958111, 0.001839230725, 0.001806357272, 0.001982733737, 0.002274080231, 0.00229457775, 0.002117357162, 0.001845375086, 0.001633187581, 0.001749012712, 0.001681082922, 0.001634979053, 0.00184047917, 0.002192893711, 0.002051811528, 0.001802743262, 0.001717172202, 0.001572679132, 0.001451319933, 0.001753953548, 0.001512238654, 0.001449994849, 0.001779310054, 0.001611715347, 0.001544583527, 0.00155642601, 0.001497240486, 0.001291858116, 0.001423933613, 0.001357856058, 0.001340842036],
					[0.002949900901, 0.002858297419, 0.002751144754, 0.002712130997, 0.002658128393, 0.002672007559, 0.002490168387, 0.002545041644, 0.00233221606, 0.002264720672, 0.003098922292, 0.002900054774, 0.00267266428, 0.002469088971, 0.002451257932, 0.002447919046, 0.002103865601, 0.002367405345, 0.002055621468, 0.002507110088, 0.003093187533, 0.002649884076, 0.002533026189, 0.002524391039, 0.002335369335, 0.002472290253, 0.002161650698, 0.002317802626, 0.002292992308, 0.002309906809, 0.003085204278, 0.00269269277, 0.002678459174, 0.002510756872, 0.002216443786, 0.002083366443, 0.002395086863, 0.001900188593, 0.002323640337, 0.002259785774, 0.002801038174, 0.002802613883, 0.002228063953, 0.002359078671, 0.002469453739, 0.002191852187, 0.001966253632, 0.001894799551, 0.00206525987, 0.002405012624, 0.002887254079, 0.002613211472, 0.002257302626, 0.002454136129, 0.002246620993, 0.001991950908, 0.001888469975, 0.002159707946, 0.002179275575, 0.002323698689, 0.002392017713, 0.002359948953, 0.00227225667, 0.002110346607, 0.00199579587, 0.001914182171, 0.001984006171, 0.001986177557, 0.002274283814, 0.002230011678, 0.002117357162, 0.002621401631, 0.002126901078, 0.001825069475, 0.002014683763, 0.001882770839, 0.001953755803, 0.002056603229, 0.00240569299, 0.002109080628, 0.001936116943, 0.001858294563, 0.001781608032, 0.001627779459, 0.001970454534, 0.001773160458, 0.001636927628, 0.001953381381, 0.00179774555, 0.001620992162, 0.001691735074, 0.001597226743, 0.0015912288, 0.001611998182, 0.001645564968, 0.00160054969],
					[0.002982481082, 0.002724947082, 0.002720678945, 0.002692079815, 0.002581899568, 0.002546935189, 0.002476504628, 0.002445071567, 0.002247460908, 0.002206492379, 0.002903858749, 0.00263994758, 0.002675630159, 0.002266383371, 0.002351451581, 0.002278303381, 0.00200061205, 0.002224109391, 0.002065006753, 0.002455918705, 0.002901151042, 0.002471890206, 0.002351341128, 0.00233510927, 0.002151563889, 0.002295598751, 0.002002517374, 0.00226400016, 0.002186510036, 0.002355579841, 0.0028736148, 0.002425354578, 0.002505757031, 0.002307803728, 0.002131532637, 0.00197987481, 0.002338798784, 0.001913647432, 0.002270978414, 0.002215060521, 0.002492115849, 0.002626465312, 0.002056039595, 0.002171076485, 0.002207256383, 0.002013411673, 0.001836744414, 0.001817529201, 0.00192052398, 0.002390016422, 0.002549283435, 0.002426121036, 0.001995026144, 0.002299884068, 0.002133769892, 0.001904169139, 0.001866935206, 0.002024030435, 0.002159351368, 0.00217477644, 0.0021142444, 0.002099432758, 0.002006248791, 0.001896770696, 0.001878570542, 0.001800106007, 0.001901105125, 0.001948085404, 0.002126321301, 0.002045158491, 0.001845375086, 0.002126901078, 0.002276511839, 0.001678201675, 0.001833440004, 0.001792319511, 0.001811427404, 0.001902093582, 0.002396847716, 0.001904328088, 0.001718775095, 0.001651573889, 0.001642356721, 0.001562520092, 0.00190059034, 0.001684161069, 0.001583373631, 0.001914626162, 0.001781562354, 0.001410880098, 0.001494793329, 0.001431770042, 0.001430295881, 0.001462799528, 0.001531357426, 0.001467325825],
					[0.002356035502, 0.002440528653, 0.00232456853, 0.00220023004, 0.002227657241, 0.002227989655, 0.002163291487, 0.002144393999, 0.00188411888, 0.001912792792, 0.00246176927, 0.002390474875, 0.002214654882, 0.001915814607, 0.001999883551, 0.001985647385, 0.001752657322, 0.001910957354, 0.001803744266, 0.002200525629, 0.002514304441, 0.002036089535, 0.001900481007, 0.002032861451, 0.001885691664, 0.001996875499, 0.001885994037, 0.002005291199, 0.001928013396, 0.002038839365, 0.002450573985, 0.002117866581, 0.00214030932, 0.002045824989, 0.001853938575, 0.001875720046, 0.001965452817, 0.001699751327, 0.001988655607, 0.001987783614, 0.002183844354, 0.002368972886, 0.001878906302, 0.00200555453, 0.002010964229, 0.001817086302, 0.00163411733, 0.001639164356, 0.001766285481, 0.002112982315, 0.002204302719, 0.002052740998, 0.001761774637, 0.00205816416, 0.001897569735, 0.001637374176, 0.001704091112, 0.001879695348, 0.001854111527, 0.00173595849, 0.002042435891, 0.001857530485, 0.001678812607, 0.001675683985, 0.001637405318, 0.001640827494, 0.001751261424, 0.001738131697, 0.001906339992, 0.001799474408, 0.001633187581, 0.001825069475, 0.001678201675, 0.001901617019, 0.001689389323, 0.001681978974, 0.001659577292, 0.001815411915, 0.002077210542, 0.001680707469, 0.001479750176, 0.001444464685, 0.001531887011, 0.001363783562, 0.001697949351, 0.001514333937, 0.001467214757, 0.001568259676, 0.001741996881, 0.001342113321, 0.001277767174, 0.001379396749, 0.001353480725, 0.001308016213, 0.001373687721, 0.001352817337],
					[0.002656567017, 0.002583794193, 0.00253747154, 0.002432468817, 0.002470283393, 0.00248874816, 0.002345133043, 0.002389854799, 0.002077564129, 0.002093213831, 0.002599881832, 0.002500464799, 0.002358356459, 0.002137951166, 0.002172617411, 0.002228671992, 0.001914621865, 0.002059282986, 0.002000441696, 0.002357724883, 0.002581041583, 0.002110086306, 0.002128911057, 0.002332175229, 0.002024016797, 0.002265404134, 0.002042060296, 0.002147263607, 0.002127059326, 0.002239321229, 0.002543276964, 0.002233633219, 0.002277384155, 0.002175205296, 0.001872324134, 0.001887109246, 0.002133015049, 0.001726585286, 0.002219449236, 0.002186204378, 0.002363437762, 0.002417322856, 0.00193734829, 0.00202633518, 0.002105815658, 0.001956677058, 0.001779050864, 0.00178368941, 0.001905770945, 0.002240209112, 0.00230289056, 0.002158481331, 0.001884896651, 0.002082048619, 0.002027384031, 0.00181615008, 0.001790801096, 0.00199321726, 0.002045975845, 0.002037136712, 0.001953488988, 0.001919568355, 0.00184701393, 0.001751742541, 0.001789838079, 0.001703409595, 0.001834798411, 0.001961482042, 0.002098460066, 0.001782427378, 0.001749012712, 0.002014683763, 0.001833440004, 0.001689389323, 0.00208699488, 0.001712450073, 0.001813296546, 0.001918209543, 0.002270302748, 0.001759219865, 0.001542083706, 0.001524916872, 0.001519306473, 0.001411680124, 0.00173158236, 0.001670207893, 0.001543731544, 0.001747236502, 0.001804869635, 0.001302159289, 0.001408221757, 0.001359346906, 0.001375020272, 0.001435210999, 0.001468558918, 0.001440521518],
					[0.002544831954, 0.002526042736, 0.002549888275, 0.002479192684, 0.002403485387, 0.002472291663, 0.002378529183, 0.002344599062, 0.002123604214, 0.00213031204, 0.002592969402, 0.002486099168, 0.002414975934, 0.002032070219, 0.002162700965, 0.002114276384, 0.001904786973, 0.002030608105, 0.001965647569, 0.002368100476, 0.002572983051, 0.002118461054, 0.002028752638, 0.002109761204, 0.001981590044, 0.002150671015, 0.001988821433, 0.00215961835, 0.002026009422, 0.002284708105, 0.002552567112, 0.002230699024, 0.002153952799, 0.002223016351, 0.001909036257, 0.001949640749, 0.002127609696, 0.001847907701, 0.002179725325, 0.002162518963, 0.002239527586, 0.002343205445, 0.001977574331, 0.00214252625, 0.00210722721, 0.001922936359, 0.001877644921, 0.001739598415, 0.00183325678, 0.002241507991, 0.002175761723, 0.0021140707, 0.001870519581, 0.002101663354, 0.001938074558, 0.001761109708, 0.001876568938, 0.002005570746, 0.002050252042, 0.002029942012, 0.00201758249, 0.001893251837, 0.00179998138, 0.001808930336, 0.001715748504, 0.00178709596, 0.001852889709, 0.001842175714, 0.002175184895, 0.00181265829, 0.001681082922, 0.001882770839, 0.001792319511, 0.001681978974, 0.001712450073, 0.002175490603, 0.001743625395, 0.00199189188, 0.00226151043, 0.001711165998, 0.001540525879, 0.001519947769, 0.001537543115, 0.00147923634, 0.001773098228, 0.00170197913, 0.001674688296, 0.001831954788, 0.001804142466, 0.001318801308, 0.001365465744, 0.001388263461, 0.00141299199, 0.001423903284, 0.001504110078, 0.001507816599],
					[0.002931137112, 0.002966731622, 0.002757467982, 0.002699152581, 0.002673820064, 0.002701562387, 0.002524251707, 0.002619809282, 0.002368484015, 0.00237844092, 0.002729180908, 0.002786945071, 0.002506934412, 0.002162047775, 0.002268925825, 0.002285185976, 0.00196953169, 0.002143121664, 0.002239790169, 0.002607530716, 0.002693267864, 0.002273008844, 0.002231438776, 0.002414349499, 0.002175490999, 0.002292409218, 0.002202094963, 0.002237437268, 0.002247039578, 0.002384424834, 0.002644966347, 0.002236322138, 0.00231883584, 0.002255465655, 0.002066530273, 0.002115343694, 0.002180562054, 0.001886751418, 0.002221286989, 0.002429174419, 0.002268189282, 0.002340431675, 0.001972020328, 0.001987810504, 0.002005792894, 0.001926255715, 0.001820894591, 0.001934769853, 0.001966713259, 0.002418662726, 0.002439105183, 0.002201615384, 0.001795517347, 0.00213781531, 0.001924333968, 0.001886317215, 0.001794484491, 0.002129976208, 0.002150942185, 0.002329128467, 0.002020718018, 0.001914514415, 0.001704258851, 0.001737045766, 0.001829044759, 0.001800410688, 0.001891929825, 0.002085970703, 0.002123685539, 0.001825797102, 0.001634979053, 0.001953755803, 0.001811427404, 0.001659577292, 0.001813296546, 0.001743625395, 0.002560726752, 0.002083073614, 0.002325012524, 0.001761420977, 0.001542287283, 0.00150474773, 0.001612232676, 0.001417474375, 0.001773750525, 0.001785032259, 0.001751504465, 0.001892191381, 0.001904908508, 0.001385364582, 0.001454037065, 0.001371315787, 0.001356353647, 0.001369128949, 0.001459390899, 0.001517580729],
					[0.003179676958, 0.003089654503, 0.003098168353, 0.002967319487, 0.002835663113, 0.002937912347, 0.002792157251, 0.002852917885, 0.002484697379, 0.002584543946, 0.003147871729, 0.002980381819, 0.002874428286, 0.002401755863, 0.002563441418, 0.002486016887, 0.002308013374, 0.002335886857, 0.002399122799, 0.002848028378, 0.003012297037, 0.002606396206, 0.002492451223, 0.002598142746, 0.002433038059, 0.002491206316, 0.002329232732, 0.002500962852, 0.002468833821, 0.002667066797, 0.002905540983, 0.002683642961, 0.002832604928, 0.002509502702, 0.00233571516, 0.002307184219, 0.002459078114, 0.002194595047, 0.002421063393, 0.002663709273, 0.00267658867, 0.002674308167, 0.002174230852, 0.002343696514, 0.002344255713, 0.002134125577, 0.002145394252, 0.002056383569, 0.002198692526, 0.002577853902, 0.002668846883, 0.00245250601, 0.002097052813, 0.002254171392, 0.002256134662, 0.001973870058, 0.002072880281, 0.002232949935, 0.002416861729, 0.002497822871, 0.002274620913, 0.002093109033, 0.002055336166, 0.001859046805, 0.001879697817, 0.001956771025, 0.002045778129, 0.002257239072, 0.002434289261, 0.002054788709, 0.00184047917, 0.002056603229, 0.001902093582, 0.001815411915, 0.001918209543, 0.00199189188, 0.002083073614, 0.002980570262, 0.002776905698, 0.001883144273, 0.001638469279, 0.001646822523, 0.001668124132, 0.001633452578, 0.001904942811, 0.00181302171, 0.001790193869, 0.002071404173, 0.002108539649, 0.001503615307, 0.001489054779, 0.001519445517, 0.001495595492, 0.001462829871, 0.001541539547, 0.001597668889],
					[0.00410448382, 0.003539376119, 0.003999211129, 0.003811419029, 0.003627688117, 0.003746374369, 0.003547327024, 0.003539833913, 0.003263079647, 0.003406758634, 0.003568750149, 0.003553381874, 0.003548025327, 0.003083914722, 0.003174658432, 0.003060032788, 0.002725449108, 0.002960454181, 0.002997783046, 0.003700040109, 0.003726597402, 0.003256030568, 0.003146094171, 0.003200476161, 0.002925505835, 0.003125602712, 0.002684325948, 0.0032285676, 0.003034514466, 0.003507441082, 0.003571788009, 0.002926002114, 0.003215172404, 0.003086664146, 0.002790613686, 0.002697538545, 0.003023566438, 0.002489518257, 0.003063830896, 0.003367061096, 0.00312213084, 0.003246847673, 0.002690242371, 0.002688894305, 0.002866565122, 0.002562093626, 0.002643949387, 0.002448597599, 0.002631403393, 0.003224991981, 0.003424183921, 0.002904037504, 0.002458630052, 0.002674545428, 0.002770495738, 0.002307852321, 0.00253783427, 0.002727763283, 0.00308709855, 0.002997682812, 0.00277005577, 0.002478238425, 0.002590616688, 0.002208657034, 0.00210540579, 0.002342896959, 0.002521603449, 0.002693510647, 0.003287733483, 0.002457843811, 0.002192893711, 0.00240569299, 0.002396847716, 0.002077210542, 0.002270302748, 0.00226151043, 0.002325012524, 0.002776905698, 0.004495284329, 0.002342090592, 0.00196423921, 0.001968095145, 0.001963940144, 0.001932027363, 0.002236250352, 0.002092354005, 0.002225072321, 0.002734847721, 0.002739584572, 0.001727048959, 0.001749568877, 0.001809098989, 0.001689155895, 0.001757563505, 0.001734222474, 0.001758184786],
					[0.003094311644, 0.002988376321, 0.002958428789, 0.002813450776, 0.002699749688, 0.002725652958, 0.002582676068, 0.002638401879, 0.002315163065, 0.002316135476, 0.003247957188, 0.002975489791, 0.002676243201, 0.002498508202, 0.002401444852, 0.002355436976, 0.002072097471, 0.002227701645, 0.002008603922, 0.002599035893, 0.00317777977, 0.002670710051, 0.002458733781, 0.002519373609, 0.002296204885, 0.00234600001, 0.002051566, 0.00226710001, 0.002220333684, 0.002274031605, 0.003231873214, 0.002786528801, 0.00257810375, 0.002559313469, 0.002256992943, 0.002179623843, 0.002211358705, 0.001822208334, 0.002320725035, 0.002292200105, 0.00295891475, 0.002728429499, 0.002311024277, 0.00223512425, 0.002265267561, 0.002117743289, 0.001795900975, 0.001835817712, 0.001867927429, 0.002298418854, 0.002966632996, 0.002652139789, 0.00220955019, 0.002328507026, 0.00210142722, 0.001854531318, 0.001899653245, 0.002024908208, 0.002117514202, 0.002208828426, 0.002698335303, 0.002328326893, 0.002073480187, 0.001916434512, 0.001800355523, 0.001864830712, 0.001912894773, 0.001937579107, 0.002143422865, 0.002349393317, 0.002051811528, 0.002109080628, 0.001904328088, 0.001680707469, 0.001759219865, 0.001711165998, 0.001761420977, 0.001883144273, 0.002342090592, 0.00238108428, 0.00184448747, 0.001772725629, 0.001556963915, 0.001482927742, 0.001793079798, 0.001564612875, 0.001518886766, 0.001774290456, 0.001573060605, 0.001739458821, 0.001649662703, 0.001562606651, 0.001332975416, 0.00145065916, 0.001354462311, 0.00131376431],
					[0.002412252584, 0.002327715382, 0.002268515803, 0.002223566, 0.002162000371, 0.002111424879, 0.001961667199, 0.002021217408, 0.001846701912, 0.001841431471, 0.002593234347, 0.002443998787, 0.00227159661, 0.002029350921, 0.001947953835, 0.001903773391, 0.00164384809, 0.001910514222, 0.00166251154, 0.002003765691, 0.002611472402, 0.002162432991, 0.001977984212, 0.00194844727, 0.001843445266, 0.001961811834, 0.001725660176, 0.001868700425, 0.001885137482, 0.001883094939, 0.002752491792, 0.002314710289, 0.002168545284, 0.002125851234, 0.001784834378, 0.001839841472, 0.001910315612, 0.001510692335, 0.001971615854, 0.001857039652, 0.00243252121, 0.002414224911, 0.001885365619, 0.001907502635, 0.002038896069, 0.001801102602, 0.001530803387, 0.001554517899, 0.001659350044, 0.001922046187, 0.002436055038, 0.002216415693, 0.001801943668, 0.002047874052, 0.001832840001, 0.001563540612, 0.001507957861, 0.001716670092, 0.001769438899, 0.0018639162, 0.002290962724, 0.002022264844, 0.001874480495, 0.001698026684, 0.0016475462, 0.001542859377, 0.001712256401, 0.001624053584, 0.001787322454, 0.00199023081, 0.001802743262, 0.001936116943, 0.001718775095, 0.001479750176, 0.001542083706, 0.001540525879, 0.001542287283, 0.001638469279, 0.00196423921, 0.00184448747, 0.00192243273, 0.001573558801, 0.001467854587, 0.001348935717, 0.001656951629, 0.001443941498, 0.001344356872, 0.001572495244, 0.001432723256, 0.001472780098, 0.001464959819, 0.00137748601, 0.001316196758, 0.001285898014, 0.001326462412, 0.001307264791],
					[0.002356759221, 0.0023059398, 0.002255255898, 0.002173940325, 0.002126479925, 0.002183330207, 0.00198535973, 0.002049355105, 0.00181648464, 0.001807023938, 0.002508700781, 0.002360782909, 0.002203276806, 0.002015332046, 0.001925131204, 0.001912987547, 0.001742680996, 0.00190527573, 0.001645494544, 0.002031277, 0.00252861915, 0.002133878537, 0.002032270797, 0.002044466161, 0.001861205512, 0.001976809919, 0.001823756082, 0.001834793041, 0.001871014565, 0.001964218654, 0.00253094309, 0.002238214903, 0.002214754088, 0.002042135601, 0.001836535011, 0.001812945473, 0.001850515018, 0.001534000403, 0.001932172136, 0.001928067481, 0.002317128886, 0.002314975858, 0.00180390305, 0.001900859149, 0.001958905783, 0.001774330601, 0.001533680315, 0.001528020481, 0.001623427959, 0.001935690547, 0.002355746166, 0.002137932195, 0.001859650234, 0.001979971247, 0.001837111486, 0.001555407534, 0.001604302022, 0.001777444928, 0.001741036688, 0.001807356548, 0.002012465449, 0.001956531438, 0.001817398768, 0.001711879501, 0.001614763026, 0.00157997703, 0.001621937332, 0.001595164909, 0.001853956393, 0.001857663227, 0.001717172202, 0.001858294563, 0.001651573889, 0.001444464685, 0.001524916872, 0.001519947769, 0.00150474773, 0.001646822523, 0.001968095145, 0.001772725629, 0.001573558801, 0.001742285217, 0.001398927468, 0.00134751789, 0.001524758113, 0.001405406504, 0.001348093794, 0.001599095327, 0.001447074662, 0.001360521947, 0.001412518689, 0.001338489337, 0.001231215769, 0.001236032472, 0.001274523524, 0.001226998722],
					[0.001942123379, 0.0020451527, 0.00193368615, 0.001910269587, 0.001833840701, 0.001826697256, 0.001802418959, 0.001773477959, 0.001622591534, 0.001595657068, 0.002074163928, 0.002067739944, 0.002040575238, 0.001718203143, 0.001891678493, 0.001729006647, 0.001600301608, 0.001734247778, 0.001594663869, 0.001906474095, 0.002266167678, 0.00189527577, 0.001658662239, 0.001777307171, 0.001682774242, 0.001834543289, 0.001687102396, 0.001821025008, 0.001755383966, 0.00181361643, 0.002247082258, 0.001959224997, 0.001950404018, 0.001865060743, 0.001683397023, 0.001639023288, 0.001865905876, 0.001554887457, 0.001822227792, 0.001756471334, 0.001976554273, 0.002088830654, 0.001602129373, 0.001811115973, 0.001872698697, 0.001647109971, 0.001455476451, 0.001494005094, 0.001643625112, 0.001934885145, 0.002025684896, 0.001988579044, 0.00158182962, 0.001877199216, 0.001635391534, 0.001528674383, 0.001605942258, 0.001822485902, 0.001707515433, 0.001639376038, 0.00185434367, 0.00171365323, 0.00158261204, 0.001552914235, 0.001592853435, 0.001570629667, 0.001611890165, 0.001618440704, 0.001761340961, 0.001724921252, 0.001572679132, 0.001781608032, 0.001642356721, 0.001531887011, 0.001519306473, 0.001537543115, 0.001612232676, 0.001668124132, 0.001963940144, 0.001556963915, 0.001467854587, 0.001398927468, 0.001698701726, 0.001327940391, 0.001616634301, 0.001494835972, 0.001437515905, 0.001637855863, 0.001541159024, 0.001283218879, 0.001300200301, 0.001356399067, 0.001310637426, 0.001323083041, 0.001367549979, 0.001368370588],
					[0.002101900141, 0.001968497738, 0.00201404631, 0.001976848994, 0.001935762069, 0.001950773694, 0.00179555948, 0.001877138149, 0.001671580843, 0.001697851274, 0.00205201753, 0.001998907565, 0.001982869769, 0.001682071098, 0.001711701281, 0.001644036295, 0.001561717078, 0.001685271276, 0.00156716494, 0.001893518972, 0.002117982045, 0.001799982984, 0.001736552634, 0.001766203399, 0.001597012896, 0.001731717203, 0.001621796456, 0.001726301118, 0.001768774387, 0.001834717014, 0.002097977091, 0.001948823312, 0.00188735876, 0.001717740157, 0.001668607442, 0.001532550038, 0.001676086598, 0.001469200418, 0.001709856006, 0.001761635758, 0.001845401642, 0.002000047114, 0.001579517001, 0.001725504202, 0.001695447636, 0.001577969709, 0.001490387539, 0.001475744626, 0.001512696411, 0.001863184141, 0.001889482664, 0.001773262991, 0.00155812396, 0.001696418522, 0.001627968138, 0.001389151699, 0.001543079407, 0.001672337281, 0.001644695146, 0.001659224599, 0.001648297176, 0.001608642565, 0.00160578189, 0.001560586611, 0.001466974956, 0.001497027997, 0.001470146558, 0.001438808775, 0.001761927511, 0.001551595006, 0.001451319933, 0.001627779459, 0.001562520092, 0.001363783562, 0.001411680124, 0.00147923634, 0.001417474375, 0.001633452578, 0.001932027363, 0.001482927742, 0.001348935717, 0.00134751789, 0.001327940391, 0.001499097178, 0.001464277672, 0.001374150364, 0.001351491444, 0.001519239066, 0.001488590535, 0.001073216745, 0.001183648946, 0.001176539898, 0.001128847482, 0.00120403918, 0.001211319905, 0.001150033874],
					[0.002560515946, 0.002535640801, 0.00253937852, 0.002485448666, 0.002350195792, 0.00231638808, 0.002302544639, 0.002263133226, 0.002067975693, 0.002067148721, 0.002574940276, 0.002534972683, 0.002484016339, 0.002112971385, 0.002146911221, 0.002084742402, 0.001859878712, 0.002121516629, 0.001981218007, 0.002302300689, 0.002668235806, 0.00224771988, 0.002087295719, 0.002226324109, 0.002031397406, 0.002172197335, 0.001933316665, 0.002169956628, 0.002103557194, 0.002292201244, 0.002585262202, 0.002238119388, 0.002251596033, 0.002234850987, 0.001948746689, 0.001956169542, 0.002156986478, 0.001768324436, 0.00215968941, 0.002084969638, 0.00234040907, 0.002459099058, 0.001933646242, 0.002020156092, 0.00206957304, 0.001900135167, 0.001746800454, 0.001733468366, 0.001903089234, 0.002231743966, 0.002335763509, 0.002244430763, 0.001802007614, 0.002103916017, 0.002071366733, 0.001813067579, 0.001825533626, 0.00202609524, 0.00200538994, 0.002125126885, 0.002036498587, 0.001983211094, 0.001834642637, 0.001794451241, 0.001730962735, 0.001760054456, 0.001863226423, 0.001889528393, 0.002223767524, 0.001880673805, 0.001753953548, 0.001970454534, 0.00190059034, 0.001697949351, 0.00173158236, 0.001773098228, 0.001773750525, 0.001904942811, 0.002236250352, 0.001793079798, 0.001656951629, 0.001524758113, 0.001616634301, 0.001464277672, 0.002158109585, 0.00167480383, 0.001628784164, 0.001809829016, 0.001810650856, 0.001328240001, 0.001424361999, 0.001421208669, 0.00141370992, 0.001458385528, 0.001509848185, 0.001468393513],
					[0.002299870077, 0.002346819807, 0.002258501074, 0.002206599569, 0.0021989561, 0.002196664573, 0.002102420911, 0.00219391231, 0.001944108977, 0.001983828019, 0.002281943433, 0.002236704862, 0.002223537119, 0.001897921416, 0.002034977418, 0.00191297072, 0.001712852285, 0.001922968469, 0.001904040506, 0.002221521965, 0.002302778545, 0.001895963612, 0.001880784241, 0.002049160193, 0.001888393315, 0.0020091459, 0.001898223352, 0.002032204841, 0.001956164834, 0.002216786869, 0.002213443299, 0.001904738534, 0.002023847758, 0.001997505152, 0.001817722133, 0.00177166698, 0.001938585802, 0.001737591391, 0.002086821575, 0.002156477303, 0.002079558915, 0.00217047674, 0.001659412079, 0.001792794597, 0.00182965049, 0.001711745192, 0.001655796305, 0.001701345065, 0.001771103512, 0.002133414729, 0.002037849418, 0.001871709159, 0.00162487014, 0.001831609016, 0.00171055106, 0.00165444263, 0.001679398022, 0.001939476904, 0.001967585467, 0.001869063588, 0.001800994314, 0.001738711241, 0.001570663998, 0.001666964874, 0.001611239802, 0.001638850111, 0.001699113779, 0.001777879072, 0.00202875252, 0.001614511409, 0.001512238654, 0.001773160458, 0.001684161069, 0.001514333937, 0.001670207893, 0.00170197913, 0.001785032259, 0.00181302171, 0.002092354005, 0.001564612875, 0.001443941498, 0.001405406504, 0.001494835972, 0.001374150364, 0.00167480383, 0.001943278547, 0.001557263926, 0.001799886207, 0.001633126472, 0.001208674032, 0.001284975258, 0.001287052074, 0.001306314015, 0.001310153149, 0.001438865939, 0.001397350548],
					[0.002083781247, 0.002105827515, 0.002197152045, 0.002148219717, 0.002055965995, 0.002116575774, 0.001963682724, 0.002079094009, 0.001812298457, 0.001884783965, 0.002136746694, 0.00217467103, 0.002091956136, 0.001773218193, 0.001895268085, 0.001819593425, 0.001679136737, 0.001818487049, 0.001840492434, 0.002142081755, 0.00226003792, 0.001801342803, 0.001813132445, 0.00191435273, 0.001786999549, 0.001939627161, 0.001877129076, 0.001974984732, 0.001946371307, 0.002177387041, 0.002178675095, 0.001925369836, 0.001959518691, 0.001918653967, 0.001742905187, 0.001680075186, 0.001892057444, 0.001611170221, 0.001997052557, 0.002065400695, 0.001874360924, 0.00203353407, 0.001606100416, 0.001736940424, 0.001730261696, 0.001660298011, 0.001665854964, 0.001608682258, 0.001686312902, 0.002105693351, 0.002005744056, 0.001750546507, 0.001567712893, 0.001815691681, 0.001684877891, 0.001623472287, 0.001668945698, 0.001907641035, 0.001912712957, 0.00195522756, 0.001637370346, 0.001653203216, 0.001592933608, 0.001556821933, 0.001517674255, 0.001596252795, 0.001664173048, 0.001792062365, 0.002015049614, 0.001502270361, 0.001449994849, 0.001636927628, 0.001583373631, 0.001467214757, 0.001543731544, 0.001674688296, 0.001751504465, 0.001790193869, 0.002225072321, 0.001518886766, 0.001344356872, 0.001348093794, 0.001437515905, 0.001351491444, 0.001628784164, 0.001557263926, 0.001869148927, 0.001821622391, 0.001754296788, 0.001100220997, 0.001245637172, 0.001224408789, 0.001235302959, 0.001315514696, 0.001423332905, 0.001436971018],
					[0.002701932892, 0.00248426969, 0.002615141548, 0.002382785761, 0.002359758516, 0.002469559865, 0.002249388048, 0.002384093743, 0.002112557084, 0.002081992134, 0.002578710274, 0.002359074018, 0.002527216656, 0.002041876751, 0.002260311334, 0.002199419717, 0.001911216259, 0.002148142537, 0.002101416757, 0.002591435225, 0.002571792485, 0.002218445191, 0.00215663356, 0.002332399788, 0.002043348134, 0.002327275906, 0.002004063264, 0.002326966181, 0.002220914779, 0.002613953357, 0.002510933318, 0.002235240477, 0.002326900403, 0.002221537306, 0.001849463061, 0.001911598533, 0.002211490898, 0.001913410436, 0.002346739526, 0.002442677105, 0.002396075867, 0.002300070477, 0.001862895769, 0.002063233178, 0.002064534697, 0.001848218608, 0.001962286642, 0.001935430794, 0.001977935665, 0.002335114657, 0.002458236182, 0.002199654503, 0.001873815258, 0.002105651673, 0.001950284629, 0.00179482723, 0.001919123145, 0.002103609887, 0.002272632476, 0.002256425015, 0.001989093522, 0.001914985238, 0.002057573856, 0.001761721129, 0.001656282469, 0.001828888462, 0.001824116362, 0.00216390251, 0.002359241942, 0.001976858326, 0.001779310054, 0.001953381381, 0.001914626162, 0.001568259676, 0.001747236502, 0.001831954788, 0.001892191381, 0.002071404173, 0.002734847721, 0.001774290456, 0.001572495244, 0.001599095327, 0.001637855863, 0.001519239066, 0.001809829016, 0.001799886207, 0.001821622391, 0.002979244709, 0.002074992858, 0.001353083366, 0.001470737906, 0.001491283712, 0.001517181308, 0.001570266791, 0.001544666878, 0.001551373396],
					[0.002757407576, 0.002719471677, 0.002802102426, 0.002707287202, 0.002504526251, 0.002616001643, 0.002506105964, 0.002466890693, 0.002276282331, 0.00232940124, 0.002542355418, 0.002538706295, 0.00245712161, 0.002086498421, 0.002183371878, 0.002243951131, 0.001958052651, 0.0021263257, 0.002207618824, 0.002621234068, 0.002718526671, 0.00220851091, 0.002089571752, 0.002324302701, 0.002036454966, 0.002234689403, 0.002049152968, 0.002348328682, 0.002292293272, 0.002567909878, 0.002546967054, 0.002217031467, 0.002387760302, 0.002176170363, 0.001900591909, 0.001958579231, 0.002181747593, 0.001930658194, 0.002276102439, 0.00243369268, 0.002140531648, 0.002407624321, 0.001994029146, 0.002008129068, 0.001966684595, 0.001831724253, 0.001977779456, 0.001909424209, 0.001958590849, 0.002455373533, 0.002416220994, 0.002098263899, 0.001786576186, 0.002104577206, 0.002041913208, 0.001723097603, 0.001938805242, 0.002081524935, 0.002188558276, 0.002297224771, 0.001834252413, 0.001870729067, 0.001761267425, 0.001674310159, 0.00164398893, 0.001783426135, 0.001925407783, 0.001970141906, 0.002349290251, 0.001714422288, 0.001611715347, 0.00179774555, 0.001781562354, 0.001741996881, 0.001804869635, 0.001804142466, 0.001904908508, 0.002108539649, 0.002739584572, 0.001573060605, 0.001432723256, 0.001447074662, 0.001541159024, 0.001488590535, 0.001810650856, 0.001633126472, 0.001754296788, 0.002074992858, 0.002635131973, 0.001167564371, 0.00130376408, 0.001329507691, 0.001380001238, 0.001321609962, 0.001336970271, 0.001462465311],
					[0.002009863366, 0.002091166113, 0.001993131787, 0.001804422913, 0.001829998069, 0.001804342367, 0.001728073759, 0.001729753224, 0.001537089641, 0.001503426019, 0.002360605143, 0.002178968325, 0.00192903563, 0.001765418608, 0.001703799545, 0.001705130056, 0.001415231144, 0.001657059318, 0.001389582413, 0.001826813062, 0.002277131577, 0.001925310451, 0.00168525719, 0.001714808166, 0.001631303246, 0.001603869682, 0.001490542702, 0.001651025562, 0.00153782559, 0.001524680113, 0.002496554737, 0.002028930962, 0.001960869109, 0.001858518206, 0.001683628808, 0.001715632971, 0.00161786628, 0.001297908725, 0.001649338559, 0.001676106267, 0.002277918698, 0.002036565539, 0.001724103763, 0.001679055023, 0.001747999998, 0.001566536254, 0.001304139577, 0.001308641757, 0.001336237643, 0.00165968462, 0.002355888284, 0.002135683637, 0.001529337994, 0.001764384174, 0.0014921953, 0.001368238167, 0.001453504753, 0.001497014391, 0.001539832924, 0.001523107793, 0.00224975273, 0.00181912707, 0.001508379831, 0.001338224906, 0.001406303153, 0.001414236235, 0.001432696674, 0.001485800171, 0.001502287893, 0.00193244399, 0.001544583527, 0.001620992162, 0.001410880098, 0.001342113321, 0.001302159289, 0.001318801308, 0.001385364582, 0.001503615307, 0.001727048959, 0.001739458821, 0.001472780098, 0.001360521947, 0.001283218879, 0.001073216745, 0.001328240001, 0.001208674032, 0.001100220997, 0.001353083366, 0.001167564371, 0.001732174103, 0.001355419126, 0.001356884386, 0.001206983927, 0.001050330198, 0.001120576959, 0.001069502127],
					[0.001980655134, 0.00204061465, 0.001941433241, 0.001881753917, 0.001840005844, 0.00183879967, 0.001709216314, 0.001775353392, 0.001600344549, 0.001564091681, 0.002220554732, 0.002083663409, 0.002014819832, 0.001761942727, 0.001711400505, 0.001712051313, 0.00147673357, 0.001654215713, 0.001484180669, 0.001795320974, 0.002209301972, 0.001872333325, 0.001848983201, 0.001772016561, 0.001626562968, 0.001739008661, 0.001573055978, 0.001691110869, 0.001667038649, 0.001779855608, 0.002295747956, 0.002013238493, 0.00197660144, 0.001821790762, 0.001661515879, 0.001600549369, 0.001671386618, 0.001312630625, 0.001713164575, 0.001761323688, 0.002093178431, 0.002083728244, 0.001573455973, 0.001674292826, 0.001856325133, 0.001659315375, 0.001386751283, 0.001388652916, 0.001430388361, 0.001753961274, 0.002182515245, 0.002002634763, 0.001611606875, 0.001754955104, 0.001613936507, 0.001434811152, 0.001421291285, 0.001545671162, 0.001644256099, 0.001675158753, 0.001895068104, 0.001775680814, 0.001597609687, 0.001469942223, 0.001505474531, 0.001439126627, 0.001447393032, 0.001609694131, 0.001690429149, 0.00173933189, 0.00155642601, 0.001691735074, 0.001494793329, 0.001277767174, 0.001408221757, 0.001365465744, 0.001454037065, 0.001489054779, 0.001749568877, 0.001649662703, 0.001464959819, 0.001412518689, 0.001300200301, 0.001183648946, 0.001424361999, 0.001284975258, 0.001245637172, 0.001470737906, 0.00130376408, 0.001355419126, 0.001475355593, 0.001258830108, 0.001160337151, 0.001174410199, 0.001188517155, 0.001154916643],
					[0.001876568152, 0.001945355621, 0.001881035982, 0.001755480548, 0.001776080995, 0.001747813931, 0.001684785524, 0.001711371032, 0.001539839272, 0.001499557562, 0.002113770986, 0.001987904628, 0.00197991251, 0.001704274891, 0.001713412, 0.001694201176, 0.001456591754, 0.001605786675, 0.001422802845, 0.001810262683, 0.002131248267, 0.001733466947, 0.001633653508, 0.001672395879, 0.001603387422, 0.001730536194, 0.001556844155, 0.001726021726, 0.001664785053, 0.001690825931, 0.002140649214, 0.001896815714, 0.001872576902, 0.001711593746, 0.00162246157, 0.001576340268, 0.001651221385, 0.001291381265, 0.001690114187, 0.001715023211, 0.002042077396, 0.002002175109, 0.001526185446, 0.001687232765, 0.001757823844, 0.001530303126, 0.001319070256, 0.001359905258, 0.001420022489, 0.001731624546, 0.002002751956, 0.001942345413, 0.001470394496, 0.001726516932, 0.001541106991, 0.001338967163, 0.001527167595, 0.001640709803, 0.001611616989, 0.001569474671, 0.001898527721, 0.001709574839, 0.001539238963, 0.001372897973, 0.001409622401, 0.001421393745, 0.001488967379, 0.001519896088, 0.001701696033, 0.001693475392, 0.001497240486, 0.001597226743, 0.001431770042, 0.001379396749, 0.001359346906, 0.001388263461, 0.001371315787, 0.001519445517, 0.001809098989, 0.001562606651, 0.00137748601, 0.001338489337, 0.001356399067, 0.001176539898, 0.001421208669, 0.001287052074, 0.001224408789, 0.001491283712, 0.001329507691, 0.001356884386, 0.001258830108, 0.001477072021, 0.001255149956, 0.001200739301, 0.001226329989, 0.001201041421],
					[0.001624975344, 0.00177718764, 0.001629664591, 0.001556326854, 0.001555161991, 0.001571649058, 0.0015206781, 0.001494355257, 0.001380351461, 0.001293417718, 0.001821547472, 0.001794990531, 0.001669107432, 0.00152144715, 0.001534514796, 0.001474786573, 0.001216366287, 0.001515116099, 0.001374981132, 0.001647121456, 0.001875955335, 0.001460972446, 0.00153148607, 0.001495796545, 0.00138335841, 0.00160289107, 0.001397123182, 0.001591942106, 0.001486070658, 0.001564463116, 0.00184150236, 0.001624917218, 0.00167473677, 0.001589475474, 0.001411743762, 0.00143020767, 0.001634855501, 0.001203881102, 0.00157166069, 0.001546563915, 0.001717972348, 0.001875702655, 0.001367473141, 0.001571458355, 0.001656088474, 0.001421089374, 0.001340951321, 0.001246940377, 0.001365864643, 0.0017056357, 0.001770018254, 0.001702335031, 0.001221941677, 0.00164620921, 0.001471749075, 0.001364436889, 0.001413216884, 0.001567264754, 0.001492172521, 0.001440842522, 0.001539200166, 0.001544124909, 0.001439716463, 0.001312547653, 0.001346738712, 0.001257311173, 0.001353134112, 0.001452871784, 0.001637735267, 0.0014430677, 0.001291858116, 0.0015912288, 0.001430295881, 0.001353480725, 0.001375020272, 0.00141299199, 0.001356353647, 0.001495595492, 0.001689155895, 0.001332975416, 0.001316196758, 0.001231215769, 0.001310637426, 0.001128847482, 0.00141370992, 0.001306314015, 0.001235302959, 0.001517181308, 0.001380001238, 0.001206983927, 0.001160337151, 0.001255149956, 0.001555627002, 0.001175399778, 0.001255432815, 0.001297431844],
					[0.001797528616, 0.001833740965, 0.001757523692, 0.001713310797, 0.001703873611, 0.001717569968, 0.00160284881, 0.001666919889, 0.001445855309, 0.001433147338, 0.001864275149, 0.001780425059, 0.001871868301, 0.001538600425, 0.00171764512, 0.001594540806, 0.001374620282, 0.001592331939, 0.001391746185, 0.001765786251, 0.00190288067, 0.001628559069, 0.001521509988, 0.001619156643, 0.001554231646, 0.00161558132, 0.001458963187, 0.001656064673, 0.001637206048, 0.001679989757, 0.001894217899, 0.001691962259, 0.001639575669, 0.001632520931, 0.001469458031, 0.001363441344, 0.001659412043, 0.001290540791, 0.001657481338, 0.001547746646, 0.001779658493, 0.001848791536, 0.001388784454, 0.001584612812, 0.001698686823, 0.001542052019, 0.00135730855, 0.001404625531, 0.00148274205, 0.001694330807, 0.001689529892, 0.001709835884, 0.001405332314, 0.00164119908, 0.001556784332, 0.001395907181, 0.001448010909, 0.001563318738, 0.001583504098, 0.001625692915, 0.001481003002, 0.001523394415, 0.001489612646, 0.001355054415, 0.001380300963, 0.001417501875, 0.001441163181, 0.001512179815, 0.001694576331, 0.001442626283, 0.001423933613, 0.001611998182, 0.001462799528, 0.001308016213, 0.001435210999, 0.001423903284, 0.001369128949, 0.001462829871, 0.001757563505, 0.00145065916, 0.001285898014, 0.001236032472, 0.001323083041, 0.00120403918, 0.001458385528, 0.001310153149, 0.001315514696, 0.001570266791, 0.001321609962, 0.001050330198, 0.001174410199, 0.001200739301, 0.001175399778, 0.00148087213, 0.001291575366, 0.001280782044],
					[0.001651044294, 0.001833126743, 0.001675342356, 0.001644961803, 0.001733502399, 0.001702643639, 0.001599105457, 0.001640984047, 0.001467154418, 0.001459815065, 0.001779315548, 0.001803667429, 0.001839191813, 0.001536824264, 0.001629135281, 0.001590322296, 0.001367862053, 0.001566729393, 0.00151998509, 0.001753677596, 0.001968385308, 0.001487323111, 0.001577008256, 0.001669654434, 0.00142310159, 0.001635188623, 0.001493556021, 0.001621731298, 0.001584095463, 0.001671031498, 0.001779130572, 0.001543014166, 0.001682717651, 0.001637769712, 0.001513598958, 0.001368196308, 0.001620250311, 0.001357172477, 0.001697401999, 0.0016355327, 0.001629050738, 0.001887412237, 0.001310831026, 0.001583883056, 0.001684434505, 0.001504801496, 0.001383488268, 0.001371462861, 0.001519772785, 0.001752820866, 0.001693807964, 0.001653003911, 0.001389306134, 0.001632717802, 0.001587092448, 0.001442945758, 0.001478431473, 0.001658344814, 0.001624889665, 0.001511484483, 0.001471806275, 0.001502743129, 0.001449216874, 0.001392112944, 0.001465920283, 0.001434609122, 0.001423381553, 0.001595801621, 0.001727829921, 0.001374532828, 0.001357856058, 0.001645564968, 0.001531357426, 0.001373687721, 0.001468558918, 0.001504110078, 0.001459390899, 0.001541539547, 0.001734222474, 0.001354462311, 0.001326462412, 0.001274523524, 0.001367549979, 0.001211319905, 0.001509848185, 0.001438865939, 0.001423332905, 0.001544666878, 0.001336970271, 0.001120576959, 0.001188517155, 0.001226329989, 0.001255432815, 0.001291575366, 0.001652398889, 0.001400097182],
					[0.001615803773, 0.001802919713, 0.001705850964, 0.001681966033, 0.001681532412, 0.001731275443, 0.001639530125, 0.001597138892, 0.001461220643, 0.001456114868, 0.00185400446, 0.001750839599, 0.001779332353, 0.00156334151, 0.00166261154, 0.001530771977, 0.001422734186, 0.001568749327, 0.001501016057, 0.00178581935, 0.00192768926, 0.001518940644, 0.001443522175, 0.001534071611, 0.001518029966, 0.001538851026, 0.001504941644, 0.001688645496, 0.001656696165, 0.001728721221, 0.00175253825, 0.001629675397, 0.001722636547, 0.001580531026, 0.001448566505, 0.001363541314, 0.001693878156, 0.001286333176, 0.001620050265, 0.001723242545, 0.001584058446, 0.001908320189, 0.001294276419, 0.001552290241, 0.00171185129, 0.00145597054, 0.001367615045, 0.001372454729, 0.001519431052, 0.001814234354, 0.001633407093, 0.001587142632, 0.001394237652, 0.00170226126, 0.001462726316, 0.001410140658, 0.001485237838, 0.001638997416, 0.001665134061, 0.00162087977, 0.001417212846, 0.001442843411, 0.001435943811, 0.001311553054, 0.001417331467, 0.001413057928, 0.001498832196, 0.00157436436, 0.001714672068, 0.001388198613, 0.001340842036, 0.00160054969, 0.001467325825, 0.001352817337, 0.001440521518, 0.001507816599, 0.001517580729, 0.001597668889, 0.001758184786, 0.00131376431, 0.001307264791, 0.001226998722, 0.001368370588, 0.001150033874, 0.001468393513, 0.001397350548, 0.001436971018, 0.001551373396, 0.001462465311, 0.001069502127, 0.001154916643, 0.001201041421, 0.001297431844, 0.001280782044, 0.001400097182, 0.001685882495]];

		var returns = [0.005873325, 0.007285383333, 0.0088992, 0.01417630833, 0.009235308333, 0.01318765, 0.01406648333, 0.01360380833, 0.01448148333, 0.01661586667, 0.004312425, 0.008158858333, 0.006681016667, 0.0005192, 0.006546558333, 0.005799741667, 0.009362716667, 0.005585858333, 0.01105469167, 0.01514394167, 0.0008603833333, 0.005390983333, 0.004285958333, 0.002965683333, 0.00882215, 0.01205829167, 0.007865941667, 0.009536483333, 0.01536361667, 0.01206999167, 0.001890775, 0.001219958333, 0.005129775, 0.00944175, 0.008676658333, 0.01137755, 0.008224466667, 0.01020840833, 0.01157376667, 0.007151091667, 0.005009416667, 0.005049083333, 0.006737966667, 0.01528215, 0.004935233333, 0.01058240833, 0.008808275, 0.0079628, 0.01132546667, 0.009103333333, 0.002207241667, 0.001660125, 0.00703745, 0.003528141667, 0.0058985, 0.00719915, 0.01374685833, 0.01278683333, 0.0099985, 0.008261508333, 0.0080636, 0.002264716667, 0.004111808333, 0.0021296, 0.0056057, 0.005727558333, 0.009533016667, 0.01281698333, 0.010599625, 0.006604125, 0.0015463, 0.001751933333, 0.003734616667, 0.008640916667, 0.007365616667, 0.0103935, 0.01234144167, 0.009268191667, 0.0082898, 0.004892658333, 0.004707458333, 0.005013725, 0.004047533333, 0.0034394, 0.006425441667, 0.007648291667, 0.005843025, 0.00818545, 0.0080509, 0.007949991667, 0.0058026, 0.003725691667, 0.001433233333, 0.002909541667, 0.006006575, 0.007056141667];
		
		var efficientPortfoliosVolatilityGsmo = PortfolioAllocation.meanVarianceEfficientFrontierPortfolios(returns, covMat, { optimizationMethod: "gsmo", discretizationType: "return", nbPortfolios: 100 });
}

});




QUnit.test('Projection on mean-variance efficient frontier', function(assert) {    	
	// Test with static data
	// As the corner portfolios are known for the data below, c.f. other tests,
	// it is easy to determine different test cases
	{
		var covMat = [[1, 2.96, 2.31],
					  [2.96, 54.76, 39.886],
					  [2.31,	39.886,	237.16]];
		var returns = [2.8000, 6.3000, 10.8000];

		// Tests with projection equal to one corner portfolio
		//
		var expectedWeights = [0.2, 0.30000000000000004, 0.5];
		var weights = PortfolioAllocation.meanVarianceEfficientFrontierNearestWeights([0.1, 0.1, 0.8], returns, covMat, { constraints: {minWeights: [0.2, 0.2, 0.2], maxWeights: [0.5, 0.5, 0.5]} });
		assert.equal(areWeightsIdentical(weights,expectedWeights, 1e-12), true, 'Projection equal to a corner portfolio, CLA');
		var weightsGsmo = PortfolioAllocation.meanVarianceEfficientFrontierNearestWeights([0.1, 0.1, 0.8], returns, covMat, { optimizationMethod: "gsmo", constraints: {minWeights: [0.2, 0.2, 0.2], maxWeights: [0.5, 0.5, 0.5]} });
		assert.equal(areWeightsIdentical(weightsGsmo,expectedWeights, 1e-12), true, 'Projection equal to a corner portfolio, GSMO');

		//
		var expectedWeights = [0.22180737780348653, 0.5, 0.27819262219651353];
		var weights = PortfolioAllocation.meanVarianceEfficientFrontierNearestWeights([0.1, 0.8, 0.1], returns, covMat, { constraints: {minWeights: [0.2, 0.2, 0.2], maxWeights: [0.5, 0.5, 0.5]} });
		assert.equal(areWeightsIdentical(weights,expectedWeights, 1e-12), true, 'Projection equal to a corner portfolio, CLA');
		var weightsGsmo = PortfolioAllocation.meanVarianceEfficientFrontierNearestWeights([0.1, 0.8, 0.1], returns, covMat, { optimizationMethod: "gsmo", constraints: {minWeights: [0.2, 0.2, 0.2], maxWeights: [0.5, 0.5, 0.5]} });
		assert.equal(areWeightsIdentical(weightsGsmo,expectedWeights, 1e-03), true, 'Projection equal to a corner portfolio, GSMO OK');
		
		//
		var expectedWeights = [0.5, 0.2999999999999999, 0.2];
		var weights = PortfolioAllocation.meanVarianceEfficientFrontierNearestWeights([0.8, 0.1, 0.1], returns, covMat, { constraints: {minWeights: [0.2, 0.2, 0.2], maxWeights: [0.5, 0.5, 0.5]} });
		assert.equal(areWeightsIdentical(weights,expectedWeights, 1e-12), true, 'Projection equal to a corner portfolio, CLA');
		var weightsGsmo = PortfolioAllocation.meanVarianceEfficientFrontierNearestWeights([0.8, 0.1, 0.1], returns, covMat, { optimizationMethod: "gsmo", constraints: {minWeights: [0.2, 0.2, 0.2], maxWeights: [0.5, 0.5, 0.5]} });
		assert.equal(areWeightsIdentical(weightsGsmo,expectedWeights, 1e-12), true, 'Projection equal to a corner portfolio, GSMO');

		
		// Test with projection equal to the initial portfolio (i.e., the initial portfolio is efficient)
		var expectedWeights = [0.2, 0.4, 0.4];
		var weights = PortfolioAllocation.meanVarianceEfficientFrontierNearestWeights([0.2, 0.4, 0.4], returns, covMat, { constraints: {minWeights: [0.2, 0.2, 0.2], maxWeights: [0.5, 0.5, 0.5]} });
		assert.equal(areWeightsIdentical(weights,expectedWeights, 1e-12), true, 'Projection equal to the initial portfolio, CLA');
		var weightsGsmo = PortfolioAllocation.meanVarianceEfficientFrontierNearestWeights([0.2, 0.4, 0.4], returns, covMat, { optimizationMethod: "gsmo", constraints: {minWeights: [0.2, 0.2, 0.2], maxWeights: [0.5, 0.5, 0.5]} });
		assert.equal(areWeightsIdentical(weightsGsmo,expectedWeights, 1e-12), true, 'Projection equal to the initial portfolio, GSMO');		
		
		
		// Test with projection equal to a portfolio strictly on an efficient segment
		var expectedWeights = [0.2, 0.405, 0.395];
		var weights = PortfolioAllocation.meanVarianceEfficientFrontierNearestWeights([0.19, 0.41, 0.4], returns, covMat, { constraints: {minWeights: [0.2, 0.2, 0.2], maxWeights: [0.5, 0.5, 0.5]} });
		assert.equal(areWeightsIdentical(weights,expectedWeights, 1e-12), true, 'Projection equal to a strict efficient portfolio, CLA');
		var weightsGsmo = PortfolioAllocation.meanVarianceEfficientFrontierNearestWeights([0.19, 0.41, 0.4], returns, covMat, { constraints: {minWeights: [0.2, 0.2, 0.2], maxWeights: [0.5, 0.5, 0.5]} });
		assert.equal(areWeightsIdentical(weightsGsmo,expectedWeights, 1e-12), true, 'Projection equal to a strict efficient portfolio, GSMO');
	} 
	
	// Test with static data
	// Test that the partial investment constraint is taken into account
	{
		var covMat = [[1, 2.96, 2.31],
					  [2.96, 54.76, 39.886],
					  [2.31,	39.886,	237.16]];
		var returns = [2.8000, 6.3000, 10.8000];

		var expectedWeights = [0.5, 0.2999999999999999, 0.2];
		var weights = PortfolioAllocation.meanVarianceEfficientFrontierNearestWeights([0.6, 0.1, 0.3], returns, covMat, { constraints: {minWeights: [0.2, 0.2, 0.2], maxWeights: [0.5, 0.5, 0.5]} });
		assert.equal(areWeightsIdentical(weights,expectedWeights, 1e-12), true, 'Partial investment constraint #1/1, CLA');
		var weightsGsmo = PortfolioAllocation.meanVarianceEfficientFrontierNearestWeights([0.6, 0.1, 0.3], returns, covMat, { optimizationMethod: "gsmo", constraints: {minWeights: [0.2, 0.2, 0.2], maxWeights: [0.5, 0.5, 0.5]} });
		assert.equal(areWeightsIdentical(weightsGsmo,expectedWeights, 1e-12), true, 'Partial investment constraint #1/1, GSMO');


		var covMat = [[1, 2.96, 2.31, 0],
					  [2.96, 54.76, 39.886, 0],
					  [2.31,	39.886,	237.16, 0],
					  [0, 0, 0, 0]];

		var returns = [2.8000, 6.3000, 10.8000, 0];

		var expectedWeights = [0.5, 0.2, 0.2, 0.1];
		var weights = PortfolioAllocation.meanVarianceEfficientFrontierNearestWeights([0.6, 0.1, 0.3, 0], returns, covMat, { constraints: {minWeights: [0.2, 0.2, 0.2, 0], maxWeights: [0.5, 0.5, 0.5, 1]} });
		assert.equal(areWeightsIdentical(weights,expectedWeights, 1e-12), true, 'Partial investment constraint #1/2, CLA');
		var weightsGsmo = PortfolioAllocation.meanVarianceEfficientFrontierNearestWeights([0.6, 0.1, 0.3, 0], returns, covMat, { optimizationMethod: "gsmo", constraints: {minWeights: [0.2, 0.2, 0.2, 0], maxWeights: [0.5, 0.5, 0.5, 1]} });
		assert.equal(areWeightsIdentical(weightsGsmo,expectedWeights, 1e-2), true, 'Partial investment constraint #1/2, GSMO');
	}
});



QUnit.test('Mean variance portfolio - maximum Sharpe ratio computation', function(assert) {    
	// Test using static data
	// Reference: An Open-Source Implementation of the Critical-Line Algorithm for Portfolio Optimization, David H. Bailey and Marcos Lopez de Prado
	{
		// Note: this portfolio has a Sharpe ratio of ~4.4535(3...), with a null risk free rate, which
		// is exactly the same value as in the reference.
		var expectedWeights = [0.08397318948217423, 0.04890598613711377, 0, 0.21830925954049477, 0.0016773041709357821,
							   0.18120064671441183, 0, 0.031183038765169507, 0.007859012532850177, 0.42689156265685];
										
		var covMat = [[0.40755159,0.03175842,0.05183923,0.05663904,0.0330226,0.00827775,0.02165938,0.01332419,0.0343476,0.02249903],
					[0.03175842,0.9063047,0.03136385,0.02687256,0.01917172,0.00934384,0.02495043,0.00761036,0.02874874,0.01336866],
					[0.05183923,0.03136385,0.19490901,0.04408485,0.03006772,0.01322738,0.03525971,0.0115493,0.0427563,0.02057303],
					[0.05663904,0.02687256,0.04408485,0.19528471,0.02777345,0.00526665,0.01375808,0.00780878,0.02914176,0.01640377],
					[0.0330226,0.01917172,0.03006772,0.02777345,0.34059105,0.00777055,0.02067844,0.00736409,0.02542657,0.01284075],
					[0.00827775,0.00934384,0.01322738,0.00526665,0.00777055,0.15983874,0.02105575,0.00518686,0.01723737,0.00723779],
					[0.02165938,0.02495043,0.03525971,0.01375808,0.02067844,0.02105575,0.68056711,0.01377882,0.04627027,0.01926088],
					[0.01332419,0.00761036,0.0115493,0.00780878,0.00736409,0.00518686,0.01377882,0.95526918,0.0106553,0.00760955],
					[0.0343476,0.02874874,0.0427563,0.02914176,0.02542657,0.01723737,0.04627027,0.0106553,0.31681584,0.01854318],
					[0.02249903,0.01336866,0.02057303,0.01640377,0.01284075,0.00723779,0.01926088,0.00760955,0.01854318,0.11079287]];
		var returns = [1.175,1.19,0.396,1.12,0.346,0.679,0.089,0.73,0.481,1.08];
		var rf = 0;
		
		var maxSharpeRatioPortfolioWeights = PortfolioAllocation.maximumSharpeRatioWeights(returns, covMat, rf);
		var maxSharpeRatioPortfolioWeightsCla = PortfolioAllocation.maximumSharpeRatioWeights(returns, covMat, rf, {optimizationMethod: 'critical-line'});
		var maxSharpeRatioPortfolioWeightsGsmo = PortfolioAllocation.maximumSharpeRatioWeights(returns, covMat, rf, {optimizationMethod: 'gsmo'});

		assert.equal(areWeightsIdentical(maxSharpeRatioPortfolioWeights, expectedWeights, 1e-8), true, "Test #1, AUTO");
		assert.equal(areWeightsIdentical(maxSharpeRatioPortfolioWeightsCla, expectedWeights, 1e-8), true, "Test #1, CLA");
		assert.equal(areWeightsIdentical(maxSharpeRatioPortfolioWeightsGsmo, expectedWeights, 1e-5), true, "Test #1, GSMO");
	}
	
	// Test using static data
	// Reference: Optimal Portfolio Selection Without Short Sales Under the Full-Information Covariance Structure: A Pedagogic Consideration, Clarence C. Y. Kwan and Yufei Yuan
	{
		var expectedRoundedWeights = [0.3318, 0, 0.2443, 0, 0.4239, 0];

		// Note: the covariance matrix below is the result of Diag(STD)*CORR*Diag(STD), because
		// the covariance matrix as displayed in the reference has some digits truncated.
		var covMat = [[0.010, 0.010, -0.0018, 0.0024, 0.0016, 0.0048],
		              [0.010, 0.0625, 0.0135, 0.009, 0.002, 0.008],
					  [-0.0018, 0.0135, 0.0324, 0.00432, -0.00288, 0.00864],
					  [0.0024, 0.009, 0.00432, 0.0144, 0.0096, 0.00192],
					  [0.0016, 0.002, -0.00288, 0.0096, 0.0064, 0.00256],
					  [0.0048, 0.008, 0.00864, 0.00192, 0.00256, 0.0256]];
		var returns = [0.15, 0.18, 0.20, 0.11, 0.13, 0.12];
		var rf = 0.08;
		
		var maxSharpeRatioPortfolioWeights = PortfolioAllocation.maximumSharpeRatioWeights(returns, covMat, rf);
		assert.equal(areWeightsIdentical(maxSharpeRatioPortfolioWeights, expectedRoundedWeights, 1e-4), true, "Test #2");
	}
	
	// Test using static data
	// Reference: Maximizing the Sharpe Ratio and Information Ratio in the Barra Optimizer, Leonid Kopman, Scott Liu
	{
		var expectedRoundedWeights = [0.2727, 0.7273];
		
		var covMat = [[0.05, 0.02],
		              [0.02, 0.07]];
		var returns = [0.05, 0.1];
		var rf = 0;
		
		var maxSharpeRatioPortfolioWeights = PortfolioAllocation.maximumSharpeRatioWeights(returns, covMat, rf);
		assert.equal(areWeightsIdentical(maxSharpeRatioPortfolioWeights, expectedRoundedWeights, 1e-4), true, "Test #3");
	}

	// Test using static data
	// Reference: Optimal mean-variance portfolio selection using Cauchy–Schwarz maximization, Hsin-Hung Chen, Hsien-Tang Tsai, Dennis K. J. Lin
	{
		// Note: the weights in the reference do not sum to 1 (they sum to 0.999)
		//
		// Note 2: the weights in the reference do not reach the maximum Sharpe ratio, 
		// because they reach ~0.743369... compared to ~0.743391... for the computed weights
		// below, plus there is a mismatch between the computed Sharpe ratio and the 
		// Sharpe ratio provided in the reference.
		//
		// Most probably, this is due to truncated numerical data displayed in the reference
		// v.s. non truncated data used by the researchers in their algorithm.
		
		//var referenceRoundedWeights = [0, 0.1355, 0, 0.5153, 0.2292, 0.119];
		var expectedRoundedWeights = [0, 0.1355, 0, 0.5153, 0.2290, 0.1202];
		
		/* The covariance matrix below has been obtained through Diag(STD)*CORR*Diag(STD)
		var stdDev = [0.2929, 0.3252, 0.3278, 0.2555, 0.3943, 0.5698];
		var corrMat = [[1.0000, 0.1484, 0.3076, 0.3685, 0.4482, 0.3802],
		              [0.1484, 1.0000, 0.0382, -0.1451, 0.2714, 0.1025],
					  [0.3076, 0.0382, 1.0000, 0.2990, 0.3634, 0.4388],
					  [0.3685, -0.1451, 0.2990, 1.0000, 0.2756, 0.2796],
					  [0.4482, 0.2714, 0.3634, 0.2756, 1.0000, 0.7542],
					  [0.3802, 0.1025, 0.4388, 0.2796, 0.7542, 1.0000]];
		*/
		var covMat = [[0.08579041, 0.014135260272, 0.029533481911999993, 0.027577047575, 0.05176282865399999, 0.06345325848399999 ],
						[0.014135260271999999, 0.10575504, 0.004072141391999999, -0.01205615586, 0.034800634104, 0.018993143399999995 ],
						[0.029533481911999997, 0.004072141392, 0.10745284, 0.0250421171, 0.046970009635999996, 0.081959257072 ],
						[0.027577047575000003, -0.01205615586, 0.0250421171, 0.06528025, 0.02776494994, 0.04070525844 ],
						[0.051762828654, 0.03480063410399999, 0.046970009635999996, 0.027764949940000002, 0.15547249, 0.16944772798799998 ],
						[0.06345325848399999, 0.018993143399999995, 0.081959257072, 0.04070525844, 0.16944772798799998, 0.32467204 ]];
		var returns = [0.0912, 0.1087, 0.1283, 0.1882, 0.2896, 0.3739];
		var rf = 0.05;
		
		var maxSharpeRatioPortfolioWeights = PortfolioAllocation.maximumSharpeRatioWeights(returns, covMat, rf);
		assert.equal(areWeightsIdentical(maxSharpeRatioPortfolioWeights, expectedRoundedWeights, 1e-4), true, "Test #4");
	}
	
	// Test using static data
	// Test the case of only one portfolio on the efficient frontier
	{
		var expectedWeights = [0.4, 0.6];
		
		var covMat = [[1,0],[0,1]];
		var returns = [0.1, 0.2];
		var rf = 0;
		
		var maxSharpeRatioPortfolioWeights = PortfolioAllocation.maximumSharpeRatioWeights(returns, covMat, rf, { constraints: {minWeights: [0.4, 0.6]} });
		assert.equal(areWeightsIdentical(maxSharpeRatioPortfolioWeights, expectedWeights, 1e-8), true, "Test #6");
	}
	
	// Test using random data
	// Test the case of a too high risk free rate
	// Reference: Maximizing the Sharpe Ratio and Information Ratio in the Barra Optimizer, Leonid Kopman, Scott Liu
	{
		var covMat = [[0.05, 0.02],
		              [0.02, 0.07]];
		var returns = [0.05, 0.1];
		var rf = 0.1 + (1 - Math.random()); // rf will be > 0.1, which is the maximum return attainable
		
		assert.throws(function() { 
			PortfolioAllocation.maximumSharpeRatioWeights(returns, covMat, rf) },
			new Error('impossible to restrict the efficient frontier: the minimum return constraint is not feasible'),
			"Maximum Sharpe Ratio portfolio #7");
	}
	
	// Test with static data for numerical precision on the right boundary of the efficient frontier,
	// for the different possible algorithms.
	{
		var covMat = [[1, 0], [0, 2]];
		var returns = [1,2];
		
		// A risk-free rate of 2 makes the excess return exactly zero
		var rf = 2;
		var expectedWeights = [0, 1];
		var weights = PortfolioAllocation.maximumSharpeRatioWeights(returns, covMat, rf);
		var weightsCla = PortfolioAllocation.maximumSharpeRatioWeights(returns, covMat, rf, {optimizationMethod:'critical-line'});
		var weightsGsmo = PortfolioAllocation.maximumSharpeRatioWeights(returns, covMat, rf, {optimizationMethod:'gsmo'});

		assert.equal(areWeightsIdentical(weights, expectedWeights, 0), true, "Test #7, AUTO");
		assert.equal(areWeightsIdentical(weightsCla, expectedWeights, 0), true, "Test #7, CLA");
		assert.equal(areWeightsIdentical(weightsGsmo, expectedWeights, 0), true, "Test #7, GSMO");
	}
});


QUnit.test('Random subspace optimization method - error and limit cases', function(assert) {    
	// Test using static data
	// Test the unsupported cases:
	// - Unsupported subsets generation method
	// - Unsupported subsets aggregation method
	// - Wrong size of subsets of assets to generate
	// - No feasible portfolio generated
	// - Number of infeasible portfolios generated greater than the admissible ratio
	// - Wrong number of weights returned by the portfolio optimization method
	{
		// Problem data
		var nbAssets = 3;

		// Test an unsupported subsets generation method
		assert.throws(function() { 
			PortfolioAllocation.randomSubspaceOptimizationWeights(nbAssets, null, { subsetsGenerationMethod: 'unknown' })},
			new Error('unsupported subsets of assets generation method'),
			"Random subspace optimization method - Unsupported subsets generation method");

		// Test an unsupported subsets aggregation method
		assert.throws(function() { 
			PortfolioAllocation.randomSubspaceOptimizationWeights(nbAssets, null, { subsetsAggregationMethod: 'unknown' }) },
			new Error('unsupported aggregation method'),
			"Random subspace optimization method - Unsupported aggregation method");

		// Test wrong size of subsets of assets to generate: < 2
		assert.throws(function() { 
			PortfolioAllocation.randomSubspaceOptimizationWeights(nbAssets, null, { sizeSubsets: 1 }) },
			new Error('the size of the subsets of assets must be between 2 and 3'),
			"Random subspace optimization method - Size of subsets of assets to generate < 2");

		// Test wrong size of subsets of assets to generate: > n
		assert.throws(function() { 
			PortfolioAllocation.randomSubspaceOptimizationWeights(nbAssets, null, { sizeSubsets: 4 }) },
			new Error('the size of the subsets of assets must be between 2 and 3'),
			"Random subspace optimization method - Size of subsets of assets to generate > n");
		
		// Test no feasible portfolio generated
		assert.throws(function() { 
			PortfolioAllocation.randomSubspaceOptimizationWeights(nbAssets, 
			                                                      function(idx, opt) { throw new Error('infeasible portfolio optimization problem'); }) },
			new Error('no feasible portfolio generated'),
			"Random subspace optimization method - No feasible portfolio generated");	
		
		// Test number of feasible portfolios generated less than the admissible ratio
		assert.throws(function() { 
			PortfolioAllocation.randomSubspaceOptimizationWeights(nbAssets, 
			                                                      function(idx, opt) { if (idx[0] == 1) { throw new Error('infeasible portfolio optimization problem'); } else { return [0, 1]; } }, 
																  { subsetsGenerationMethod: 'deterministic',
                                                                    maxInfeasibleSubsetsRatio: 2/3}) },
			new Error('too many infeasible portfolios generated'),
			"Random subspace optimization method - Too many infeasible portfolios generated");
		
		// Test wrong number of weights returned by the portfolio optimization method
		assert.throws(function() { 
			PortfolioAllocation.randomSubspaceOptimizationWeights(nbAssets, 
			                                                      function(idx, opt) { return [0, 0, 0, 1]; }) },
			new Error('internal error: the portfolio optimization method did not return the expected number of weights'),
			"Random subspace optimization method - Wrong number of weights returned by the portfolio optimization method");	
	}
	
	// Test using random data
	// Test the limit cases:
	// - 1 asset
	// - Number of assets equal to the size of the subsets to generate
	{
		// Test 1 asset
		{
			var expectedWeights = [1];
			var weights = PortfolioAllocation.randomSubspaceOptimizationWeights(1, null);
			assert.equal(areWeightsIdentical(weights, expectedWeights, 0), true, 'Random subspace optimization method - 1 asset');	
		}
		
		// Test number of assets equal to the size of the subsets to generate
		//
		// For performances reasons, it must be tested that only one call to the portfolio optimization method takes place
		{
			// Generate a random number of assets
			var nbAssets = Math.floor(Math.random()*(25-2+1) + 2); // max 25 min 2
			
			// Define the portfolio optimization method: an equal weigths portfolio optimization method, for simplicity
			function subsetEqualWeightsOptimization(subsetAssetsIdx, subsetOptimizationFctOpt) {
				// Count the number of function calls
				subsetEqualWeightsOptimization.called++;
				
				return PortfolioAllocation.equalWeights(subsetAssetsIdx.length);
			}
			subsetEqualWeightsOptimization.called = 0;
			
			// Compute the RSO equally weighted portfolio
			var weights = PortfolioAllocation.randomSubspaceOptimizationWeights(nbAssets, subsetEqualWeightsOptimization, {sizeSubsets: nbAssets});
			
			// Compare the weights as well as the number of calls of the portfolio optimization method
			var expectedWeights = PortfolioAllocation.equalWeights(nbAssets);
			assert.deepEqual(weights, expectedWeights, 'Random subspace optimization method -  Number of assets equal to the size of the subsets to generate, #1');	
			
			var expectedFunctionCalls = 1;
			assert.equal(subsetEqualWeightsOptimization.called, expectedFunctionCalls, 'Random subspace optimization method - Number of assets equal to the size of the subsets to generate, #2');	
		}
	}
});	

QUnit.test('Random subspace optimization method', function(assert) {    
	// Tests the properties of the number of subsets to generate in case of random subsets generation with 2 assets, using random data 
	{
		// Limit case to test
		var nbAssets = 2;
		
		// Define the portfolio optimization method: an equal weigths portfolio optimization method, for simplicity
		function subsetEqualWeightsOptimization(subsetAssetsIdx, subsetOptimizationFctOpt) {
			// Count the number of function calls
			subsetEqualWeightsOptimization.called++;
			
			return PortfolioAllocation.equalWeights(subsetAssetsIdx.length);
		}
		
		// Compute the RSO equally weighted portfolio
		// The default number of subsets in case of random subsets generation method to generate must be equal to 128
		var expectedFunctionCalls = 1;
		subsetEqualWeightsOptimization.called = 0;
		var weights = PortfolioAllocation.randomSubspaceOptimizationWeights(nbAssets, subsetEqualWeightsOptimization);
		assert.equal(subsetEqualWeightsOptimization.called, expectedFunctionCalls, 'Random subspace optimization method - Default number of random subsets to generate with 2 assets');
		
		// Compute the RSO equally weighted portfolio
		// The number of subsets to generate in case of random subsets generation method must be equal to the specified one
		var nbSubsetsToGenerate = Math.floor(Math.random()*(1000-1+1) + 1); // max 1000 min 1
		var expectedFunctionCalls = 1;
		subsetEqualWeightsOptimization.called = 0;
		var weights = PortfolioAllocation.randomSubspaceOptimizationWeights(nbAssets, subsetEqualWeightsOptimization, {nbRandomSubsets: nbSubsetsToGenerate});
		assert.equal(subsetEqualWeightsOptimization.called, expectedFunctionCalls, 'Random subspace optimization method - Specified number of random subsets to generate with 2 assets');							
	}
	
	// Tests the properties of the number of subsets to generate in case of random subsets generation, using random data 
	{
		// Generate a random number of assets
		var nbAssets = Math.floor(Math.random()*(25-3+1) + 3); // max 25 min 3
			
		// Define the portfolio optimization method: an equal weigths portfolio optimization method, for simplicity
		function subsetEqualWeightsOptimization(subsetAssetsIdx, subsetOptimizationFctOpt) {
			// Count the number of function calls
			subsetEqualWeightsOptimization.called++;
			
			return PortfolioAllocation.equalWeights(subsetAssetsIdx.length);
		}
		
		// Compute the RSO equally weighted portfolio
		// The default number of subsets in case of random subsets generation method to generate must be equal to 128
		var expectedFunctionCalls = 128;
		subsetEqualWeightsOptimization.called = 0;
		var weights = PortfolioAllocation.randomSubspaceOptimizationWeights(nbAssets, subsetEqualWeightsOptimization);
		assert.equal(subsetEqualWeightsOptimization.called, expectedFunctionCalls, 'Random subspace optimization method - Default number of random subsets to generate');
		
		// Compute the RSO equally weighted portfolio
		// The number of subsets to generate in case of random subsets generation method must be equal to the specified one
		var nbSubsetsToGenerate = Math.floor(Math.random()*(1000-1+1) + 1); // max 1000 min 1
		var expectedFunctionCalls = nbSubsetsToGenerate;
		subsetEqualWeightsOptimization.called = 0;
		var weights = PortfolioAllocation.randomSubspaceOptimizationWeights(nbAssets, subsetEqualWeightsOptimization, {nbRandomSubsets: nbSubsetsToGenerate});
		assert.equal(subsetEqualWeightsOptimization.called, expectedFunctionCalls, 'Random subspace optimization method - Specified number of random subsets to generate');					
	}

	// Test the properties of the size of the generated subsets in case of random subsets generation, using random data  
	{
		// Generate a random number of assets
		var nbAssets = Math.floor(Math.random()*(50-2+1) + 2); // max 50 min 2
			
		// Define the portfolio optimization method: an equal weigths portfolio optimization method, for simplicity
		function subsetEqualWeightsOptimization(subsetAssetsIdx, subsetOptimizationFctOpt) {
			// In case the size of the generated subsets is not correct, throw
			assert.equal(subsetAssetsIdx.length, expectedSizeSubsets, 'Random subspace optimization method - Size of random subsets to generate');
			
			return PortfolioAllocation.equalWeights(subsetAssetsIdx.length);
		}
		
		// Compute the RSO equally weighted portfolio
		// The default size of the subsets to generate must be FLOOR(SQRT(nbAssets))
		var expectedSizeSubsets = Math.max(2, Math.floor(Math.sqrt(nbAssets)));
		var weights = PortfolioAllocation.randomSubspaceOptimizationWeights(nbAssets, subsetEqualWeightsOptimization);	
		
		// Compute the RSO equally weighted portfolio
		// The size of the generated subsets must be equal to the specified one
		var sizeSubsetsToGenerate = Math.floor(Math.random()*(nbAssets-2+1) + 2); // max nbAssets min 2
		var expectedSizeSubsets = sizeSubsetsToGenerate;
		var weights = PortfolioAllocation.randomSubspaceOptimizationWeights(nbAssets, subsetEqualWeightsOptimization, {sizeSubsets: sizeSubsetsToGenerate});			
	}
	
	// Tests the properties of the number of the subsets to generate in case of deterministic subsets generation, using random data 
	{
		function binomial(n, k) { if (k === 0) return 1; return (n * binomial(n-1, k-1)) / k; }
		
		// Generate a random number of assets
		var nbAssets = Math.floor(Math.random()*(15-2+1) + 2); // max 15 min 2
			
		// Define the portfolio optimization method: an equal weigths portfolio optimization method, for simplicity
		function subsetEqualWeightsOptimization(subsetAssetsIdx, subsetOptimizationFctOpt) {
			// Count the number of function calls
			subsetEqualWeightsOptimization.called++;
			
			return PortfolioAllocation.equalWeights(subsetAssetsIdx.length);
		}

		// Compute the RSO equally weighted portfolio
		// The number of subsets to generate in case of deterministic subsets generation method must be equal to Binomial(nbAssets, sizeSubsets)
		var sizeSubsetsToGenerate = Math.floor(Math.random()*(nbAssets-2+1) + 2); // max nbAssets min 2		
		var expectedFunctionCalls = binomial(nbAssets, sizeSubsetsToGenerate);
		subsetEqualWeightsOptimization.called = 0;
		var weights = PortfolioAllocation.randomSubspaceOptimizationWeights(nbAssets, subsetEqualWeightsOptimization, {subsetsGenerationMethod: 'deterministic', sizeSubsets: sizeSubsetsToGenerate});
		assert.equal(subsetEqualWeightsOptimization.called, expectedFunctionCalls, 'Random subspace optimization method - Number of deterministic subsets to generate');
	}
	
	// Test the behaviour of the deterministic subsets generation method, using random data 
	{
		// Generate a random number of assets
		var nbAssets = Math.floor(Math.random()*(15-2+1) + 2); // max 15 min 2
			
		// Define the portfolio optimization method: an equal weigths portfolio optimization method, for simplicity
		function subsetEqualWeightsOptimization(subsetAssetsIdx, subsetOptimizationFctOpt) {
			return PortfolioAllocation.equalWeights(subsetAssetsIdx.length);
		}

		// Compute the RSO equally weighted portfolio
		// In the specific case of the deterministic subsets generation method, an RSO equally weighted portfolio
		// must be equal to an equally weighted portfolio over all the original assets
		var expectedWeights = PortfolioAllocation.equalWeights(nbAssets);
		var weights = PortfolioAllocation.randomSubspaceOptimizationWeights(nbAssets, subsetEqualWeightsOptimization, {subsetsGenerationMethod: 'deterministic'});
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-8), true, 'Random subspace optimization method - Deterministic subsets generation method');
	}

	// Test the behaviour of the automated subsetting in case minWeights/maxWeights constraints are provided, using random data 
	{
		// Generate a random number of assets
		var nbAssets = Math.floor(Math.random()*(50-2+1) + 2); // max 50 min 2
				
		// Generate random lower and upper bounds on these assets 
		// (feasibility of the associated portfolio optimization problem is not relevant).
		var lowerBounds = new Array(nbAssets);
		var upperBounds = new Array(nbAssets);
		for (var i = 0; i < nbAssets; ++i) {
			lowerBounds[i] = Math.random();
			upperBounds[i] = Math.random();
		}
		
		// Define the portfolio optimization method: an equal weigths portfolio optimization method, for simplicity
		function subsetEqualWeightsOptimization(subsetAssetsIdx, subsetOptimizationFctOpt) {
			// In case the automated subsetting of minWeights/maxWeights is not correct, throw
			var boundsOK = true;

			for (var i = 0; i < subsetAssetsIdx.length; ++i) {
				if (subsetOptimizationFctOpt.constraints.minWeights[i] != lowerBounds[subsetAssetsIdx[i]-1]) {
					boundsOK = false;
					break;
				}
				if (subsetOptimizationFctOpt.constraints.maxWeights[i] != upperBounds[subsetAssetsIdx[i]-1]) {
					boundsOK = false;
					break;
				}
			}
			assert.equal(boundsOK, true, 'Random subspace optimization method -  Automated subsetting of minWeights/maxWeights constraints');
			
			return PortfolioAllocation.equalWeights(subsetAssetsIdx.length);
		}
		
		// Compute the RSO equally weighted portfolio
		var weights = PortfolioAllocation.randomSubspaceOptimizationWeights(nbAssets, subsetEqualWeightsOptimization, { constraints: { minWeights: lowerBounds, maxWeights: upperBounds}});	
	}
	
	// Test the subsets aggregation method properties, using static data
	{
		var nbAssets = 3;
		
		// Define the portfolio optimization method to test the subsets aggregation method
		function subsetOptimization(subsetAssetsIdx, subsetOptimizationFctOpt) {
			// Use the number of function calls to compute the subset portfolio weights
			subsetOptimization.called++;

			if (subsetOptimization.called === 1) {
				return [1, 0]; // subsetAssetsIdx is equal to [1, 2] -> [1, 0, 0] for original portfolio weights
			}
			else if (subsetOptimization.called === 2) {
				return [0, 1]; // subsetAssetsIdx is equal to [1, 3] -> [0, 0, 1] for original portfolio weights
			}
			else {
				return [0, 0]; // subsetAssetsIdx is equal to [2, 3] -> [0, 0, 0] for original portfolio weights
			}
		}
		
		// Compute the RSO portfolio
		// Test the 'average' subsets aggregation method
		var expectedWeights = [0.3333333333333333, 0, 0.3333333333333333];
		subsetOptimization.called = 0;	
		var averageWeights = PortfolioAllocation.randomSubspaceOptimizationWeights(nbAssets, subsetOptimization, {subsetsGenerationMethod: 'deterministic', subsetsAggregationMethod: 'average'});
		assert.equal(areWeightsIdentical(averageWeights, expectedWeights, 0), true, 'Random subspace optimization method - Average subsets aggregation method');

		// Compute the RSO portfolio
		// The default subsets aggregation method must be 'average'		
		subsetOptimization.called = 0;
		var weights = PortfolioAllocation.randomSubspaceOptimizationWeights(nbAssets, subsetOptimization, {subsetsGenerationMethod: 'deterministic'});
		assert.equal(areWeightsIdentical(averageWeights, expectedWeights, 0), true, 'Random subspace optimization method - Default subsets aggregation method');

		// Compute the RSO portfolio
		// Test the 'median' subsets aggregation method
		var expectedWeights = [0.21134794394954964, 0, 0.21134794394954964];
		subsetOptimization.called = 0;	
		var medianWeights = PortfolioAllocation.randomSubspaceOptimizationWeights(nbAssets, subsetOptimization, {subsetsGenerationMethod: 'deterministic', subsetsAggregationMethod: 'median'});
		assert.equal(areWeightsIdentical(medianWeights, expectedWeights, 0), true, 'Random subspace optimization method - Median subsets aggregation method');		
	}
});	



QUnit.test('Random subspace mean variance portfolio - error cases', function(assert) {    	
	// Test using static data
	// Test the unsupported cases:
	// - Missing return/volatility constraint
	{
		// Problem data
		var covMat =[[0.0146, 0.0187, 0.0145],
					[0.0187, 0.0854, 0.0104],
					[0.0145, 0.0104, 0.0289]];
		var returns = [0.062, 0.146, 0.128];
		
		// Compute the target maximum volatility
		var maxVolatility = covMat[0][0];
		
		// Test missing mean-variance optimization method
		assert.throws(function() { 
			PortfolioAllocation.randomSubspaceMeanVarianceOptimizationWeights(returns, covMat)},
			new Error('missing return, volatility or risk tolerance constraints'),
			"Random subspace mean variance portfolio - Missing mean-variance optimization method");
	}
});	


QUnit.test('Random subspace mean variance portfolio', function(assert) {    
	// Test using static data
	{
		var covMat = [[1, 0.3, -0.2], 
		              [0.3, 1, 0.6], 
					  [-0.2, 0.6, 1]];
		var returns = [0.1, 0.2, 0.15];
		
		var expectedWeights = [0.09395200108235617, 0.1853986157617026, 0.11749197001156553];
		var weights = PortfolioAllocation.randomSubspaceMeanVarianceOptimizationWeights(returns, covMat, 
		                                                                                { constraints: { minExposure: 0},
																						  subsetsGenerationMethod: 'deterministic', 
																			              subsetsOpt: {
																							  constraints: {maxVolatility: Math.sqrt(0.10)}
																						  }
																						});
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-6), true, 'Random subspace mean variance portfolio, #1');
	}
	
	// Test using static data infeasible cases
	{
		var covMat = [[1, 0.3, -0.2], 
		              [0.3, 1, 0.6], 
					  [-0.2, 0.6, 1]];
		var returns = [0.1, 0.2, 0.15];
		
		// Test a non-reachable target return
		assert.throws(function() { 
			PortfolioAllocation.randomSubspaceMeanVarianceOptimizationWeights(returns, covMat, 
		                                                                                { subsetsGenerationMethod: 'deterministic', 
																			              subsetsOpt: {
																							  constraints: {return: 0.3}
																						  }
																						}) },
			new Error('no feasible portfolio generated'),
			"Random subspace mean variance portfolio - No feasible portfolio generated #1");
			
		// Test a non-reachable target volatility
		assert.throws(function() { 
			PortfolioAllocation.randomSubspaceMeanVarianceOptimizationWeights(returns, covMat, 
		                                                                                { subsetsGenerationMethod: 'deterministic', 
																			              subsetsOpt: {
																							  constraints: {volatility: 0.1}
																						  }
																						}) },
			new Error('no feasible portfolio generated'),
			"Random subspace mean variance portfolio - No feasible portfolio generated #2");
	}

	// Test using static data
	// Reference: http://mba.tuck.dartmouth.edu/pages/faculty/ken.french/data_library.html, 100 Portfolios Formed on Size and Book-to-Market (10 x 10)
	// Returns data from 1963/07 to 1973/06 used to computed the expected returns/covariance matrix
	// 4 portfolios have been removed due to missing data
	//
	// This example is used to stress-test the algorithm, but it is not possible to compare the results to some expected weights 
	{
		var covMat = [[0.008626335034, 0.006715399917, 0.007042419992, 0.007048459, 0.0066366781, 0.006594382192, 0.006188482528, 0.006144514577, 0.005581524855, 0.005791371247, 0.006839503165, 0.006026119572, 0.00556042548, 0.004834846375, 0.004511975974, 0.00502810301, 0.004273827084, 0.004391065668, 0.004333287312, 0.005263026397, 0.006093797699, 0.005328759167, 0.004885570623, 0.005006898428, 0.004375058515, 0.004536227401, 0.00382383205, 0.004188885336, 0.004227284984, 0.004480470259, 0.005652719041, 0.004684925202, 0.004824278, 0.004694296476, 0.0038311798, 0.004131720831, 0.003918816918, 0.00363012853, 0.004303746538, 0.004208510197, 0.00500884843, 0.004497613403, 0.003925394458, 0.003826366745, 0.003479102582, 0.003217506668, 0.003092539967, 0.003387487433, 0.003395715859, 0.004000762357, 0.004921900893, 0.004177801985, 0.003654297707, 0.003598637829, 0.00347358152, 0.003243638023, 0.003030619633, 0.003440298944, 0.003931812779, 0.004004083997, 0.00379606579, 0.00338809299, 0.003208075825, 0.002935523324, 0.002579952426, 0.002747787374, 0.002993945136, 0.003157948431, 0.003715079122, 0.003337110603, 0.002848833667, 0.002949900901, 0.002982481082, 0.002356035502, 0.002656567017, 0.002544831954, 0.002931137112, 0.003179676958, 0.00410448382, 0.003094311644, 0.002412252584, 0.002356759221, 0.001942123379, 0.002101900141, 0.002560515946, 0.002299870077, 0.002083781247, 0.002701932892, 0.002757407576, 0.002009863366, 0.001980655134, 0.001876568152, 0.001624975344, 0.001797528616, 0.001651044294, 0.001615803773],
						[0.006715399917, 0.006683590368, 0.006014896374, 0.005908565288, 0.005722001587, 0.005670592412, 0.005404122902, 0.005228753424, 0.004781393518, 0.004867670897, 0.006057294555, 0.005460206075, 0.004834687069, 0.00427960842, 0.004094676601, 0.004430417541, 0.003713099085, 0.003918644507, 0.003805239886, 0.004724829646, 0.005547769669, 0.004641512869, 0.004275172741, 0.004535712695, 0.004002068481, 0.003998855101, 0.003461115736, 0.003868338392, 0.003743206232, 0.00397440804, 0.005132938989, 0.004315534403, 0.004334464152, 0.004243317135, 0.003586062754, 0.003673558324, 0.003599684906, 0.003328464186, 0.003895523432, 0.003933823625, 0.00454823586, 0.004173927655, 0.00350670524, 0.003414971713, 0.003313012908, 0.003017905708, 0.002835420141, 0.00304936103, 0.003091043629, 0.00371479407, 0.004518149563, 0.003944750725, 0.003314415963, 0.00343389213, 0.003213881416, 0.003034389219, 0.002929474526, 0.00334420473, 0.003452367593, 0.003555839981, 0.003597330441, 0.003228763006, 0.002824046695, 0.002719042358, 0.002595143005, 0.00264760704, 0.002929607457, 0.003064194774, 0.003378001969, 0.003206598288, 0.002600477361, 0.002858297419, 0.002724947082, 0.002440528653, 0.002583794193, 0.002526042736, 0.002966731622, 0.003089654503, 0.003539376119, 0.002988376321, 0.002327715382, 0.0023059398, 0.0020451527, 0.001968497738, 0.002535640801, 0.002346819807, 0.002105827515, 0.00248426969, 0.002719471677, 0.002091166113, 0.00204061465, 0.001945355621, 0.00177718764, 0.001833740965, 0.001833126743, 0.001802919713],
						[0.007042419992, 0.006014896374, 0.006439832216, 0.006144856481, 0.005759584222, 0.005845477915, 0.005529484399, 0.005390914697, 0.004908853124, 0.005103830919, 0.006073314853, 0.005499701853, 0.004969641429, 0.004398641598, 0.004134976123, 0.004525198382, 0.003883988522, 0.003985939094, 0.003937507735, 0.004796383717, 0.00554862731, 0.00472970036, 0.004383809539, 0.004507738633, 0.004099612817, 0.004120392418, 0.003465223346, 0.003916806442, 0.003893286734, 0.00427363146, 0.005206899053, 0.004323028253, 0.004343625848, 0.00432711529, 0.003575259214, 0.003722534095, 0.003569795438, 0.003341274835, 0.00395494306, 0.004053391677, 0.0045785797, 0.004162307351, 0.003616132532, 0.003500002623, 0.00324854751, 0.003014847819, 0.002964497291, 0.003052088947, 0.003176130571, 0.003789134736, 0.004571681602, 0.003864248166, 0.003371414581, 0.003366216157, 0.003231814936, 0.003016512653, 0.002922320365, 0.003274787097, 0.003685314298, 0.003723157506, 0.003694396668, 0.003228902136, 0.002950115286, 0.002751010898, 0.002432745554, 0.002705575637, 0.002880348639, 0.002975663554, 0.003502498329, 0.00322837047, 0.002724240891, 0.002751144754, 0.002720678945, 0.00232456853, 0.00253747154, 0.002549888275, 0.002757467982, 0.003098168353, 0.003999211129, 0.002958428789, 0.002268515803, 0.002255255898, 0.00193368615, 0.00201404631, 0.00253937852, 0.002258501074, 0.002197152045, 0.002615141548, 0.002802102426, 0.001993131787, 0.001941433241, 0.001881035982, 0.001629664591, 0.001757523692, 0.001675342356, 0.001705850964],
						[0.007048459, 0.005908565288, 0.006144856481, 0.006721824469, 0.005852814111, 0.005856465086, 0.005599641385, 0.005466069036, 0.005048488162, 0.005242939938, 0.006111949473, 0.005465037035, 0.005129873692, 0.004416730131, 0.004205399686, 0.004458656717, 0.003886593217, 0.003969078367, 0.003980490598, 0.004776577303, 0.005403948677, 0.004734314166, 0.004410779587, 0.004359942384, 0.004043738715, 0.00403985557, 0.003443041304, 0.003961335731, 0.003888049826, 0.004038458787, 0.004970339251, 0.004186782856, 0.004351604312, 0.004236631295, 0.003515810312, 0.003803261423, 0.003629734042, 0.003345952371, 0.003877361097, 0.003927757425, 0.004493401784, 0.004028057711, 0.003580285672, 0.00341839726, 0.003306099986, 0.003046845036, 0.002896100679, 0.00306826418, 0.003125160756, 0.003719892798, 0.004420189562, 0.003735997127, 0.003210599149, 0.003358889658, 0.003149259674, 0.002974762136, 0.002953982871, 0.003297600332, 0.003627668097, 0.003716680489, 0.003460502947, 0.003201833917, 0.002927183154, 0.002807933105, 0.002393663894, 0.002612167028, 0.002891811709, 0.002789078706, 0.00345071268, 0.003015226454, 0.002634119186, 0.002712130997, 0.002692079815, 0.00220023004, 0.002432468817, 0.002479192684, 0.002699152581, 0.002967319487, 0.003811419029, 0.002813450776, 0.002223566, 0.002173940325, 0.001910269587, 0.001976848994, 0.002485448666, 0.002206599569, 0.002148219717, 0.002382785761, 0.002707287202, 0.001804422913, 0.001881753917, 0.001755480548, 0.001556326854, 0.001713310797, 0.001644961803, 0.001681966033],
						[0.0066366781, 0.005722001587, 0.005759584222, 0.005852814111, 0.005804198968, 0.005518711262, 0.005212667932, 0.00520579585, 0.004695260663, 0.004884578153, 0.00563581538, 0.0051420724, 0.004752966669, 0.004155890204, 0.003897907344, 0.004250814324, 0.00363196973, 0.003733774514, 0.003756053426, 0.004551709944, 0.005102927184, 0.004374288211, 0.004156728923, 0.004290280092, 0.003757206817, 0.00387302364, 0.003394904912, 0.003736503159, 0.00370489262, 0.003989563848, 0.004710173158, 0.003965921161, 0.004097876261, 0.003968104092, 0.003341734566, 0.003518722423, 0.003444342914, 0.00311780955, 0.003726171349, 0.003807930121, 0.004239428375, 0.003945152337, 0.003383915216, 0.003340113351, 0.003251743842, 0.002902764019, 0.002782632546, 0.002923215582, 0.003007653059, 0.003567505347, 0.004212690159, 0.003554662776, 0.003178857495, 0.003158288587, 0.003031807835, 0.002869155025, 0.002756593414, 0.003070416077, 0.003440003388, 0.003495357866, 0.003285121985, 0.003009518249, 0.002794968285, 0.002665138781, 0.002399572175, 0.002482747036, 0.002766082028, 0.002807242675, 0.003305060105, 0.002907239952, 0.002521817319, 0.002658128393, 0.002581899568, 0.002227657241, 0.002470283393, 0.002403485387, 0.002673820064, 0.002835663113, 0.003627688117, 0.002699749688, 0.002162000371, 0.002126479925, 0.001833840701, 0.001935762069, 0.002350195792, 0.0021989561, 0.002055965995, 0.002359758516, 0.002504526251, 0.001829998069, 0.001840005844, 0.001776080995, 0.001555161991, 0.001703873611, 0.001733502399, 0.001681532412],
						[0.006594382192, 0.005670592412, 0.005845477915, 0.005856465086, 0.005518711262, 0.005885907481, 0.005249993788, 0.005186831761, 0.004725555913, 0.00494187006, 0.005657857003, 0.005163692584, 0.004706632845, 0.004173076723, 0.003902254463, 0.004284994782, 0.00371978044, 0.003800660873, 0.003827532969, 0.004633361599, 0.005034250859, 0.004381039134, 0.004140177808, 0.004271141624, 0.003815087583, 0.003888776041, 0.003412560729, 0.003715946693, 0.003694375629, 0.003998670233, 0.00466125509, 0.003956994273, 0.004064539484, 0.003977163851, 0.003358623301, 0.003518226553, 0.003373116109, 0.003176878082, 0.003775985543, 0.003852669217, 0.004223583686, 0.003867575054, 0.003372837448, 0.003276297438, 0.003134634966, 0.002941816614, 0.002837975836, 0.002966669732, 0.00306358095, 0.003607251209, 0.004047749529, 0.003518512219, 0.003220634491, 0.003225088129, 0.003074806453, 0.002869951923, 0.002794455413, 0.00314857206, 0.003445206259, 0.003484276388, 0.003242065918, 0.00296896478, 0.002867250372, 0.002673288533, 0.002378738807, 0.002476341968, 0.00278931335, 0.002744695175, 0.003419811664, 0.002865320479, 0.00253830332, 0.002672007559, 0.002546935189, 0.002227989655, 0.00248874816, 0.002472291663, 0.002701562387, 0.002937912347, 0.003746374369, 0.002725652958, 0.002111424879, 0.002183330207, 0.001826697256, 0.001950773694, 0.00231638808, 0.002196664573, 0.002116575774, 0.002469559865, 0.002616001643, 0.001804342367, 0.00183879967, 0.001747813931, 0.001571649058, 0.001717569968, 0.001702643639, 0.001731275443],
						[0.006188482528, 0.005404122902, 0.005529484399, 0.005599641385, 0.005212667932, 0.005249993788, 0.00529751027, 0.00488858012, 0.004479292431, 0.004661042738, 0.005244524452, 0.004844508527, 0.004519702335, 0.003901269768, 0.003814645021, 0.004000603532, 0.003560416181, 0.003645410684, 0.003622276933, 0.004387740333, 0.004834011926, 0.004163941856, 0.003798179843, 0.003995557898, 0.00373019354, 0.003728689545, 0.00322625399, 0.003643686193, 0.003498595594, 0.00389183741, 0.00449380729, 0.003763685986, 0.003833958135, 0.003889181773, 0.003237279705, 0.003384040493, 0.003371544744, 0.003103820812, 0.003638987516, 0.003674315677, 0.003997987643, 0.00374307051, 0.003281052417, 0.003160486446, 0.003002804477, 0.002773694236, 0.002668950943, 0.002792739793, 0.0028610154, 0.003470426784, 0.003914053963, 0.003418586327, 0.003024243577, 0.003035802487, 0.002865340173, 0.002746671583, 0.002713610852, 0.002952582409, 0.003384751458, 0.003398783072, 0.003164531892, 0.002893749342, 0.002576146487, 0.002534087762, 0.002298139155, 0.002369198989, 0.002650016671, 0.002596581078, 0.003333011815, 0.002746955289, 0.0024088781, 0.002490168387, 0.002476504628, 0.002163291487, 0.002345133043, 0.002378529183, 0.002524251707, 0.002792157251, 0.003547327024, 0.002582676068, 0.001961667199, 0.00198535973, 0.001802418959, 0.00179555948, 0.002302544639, 0.002102420911, 0.001963682724, 0.002249388048, 0.002506105964, 0.001728073759, 0.001709216314, 0.001684785524, 0.0015206781, 0.00160284881, 0.001599105457, 0.001639530125],
						[0.006144514577, 0.005228753424, 0.005390914697, 0.005466069036, 0.00520579585, 0.005186831761, 0.00488858012, 0.00512539599, 0.004447562299, 0.004656774587, 0.005183246677, 0.00477351631, 0.004412606969, 0.003868620735, 0.003722629721, 0.003997654346, 0.00351483726, 0.003593304073, 0.00361034607, 0.004428704724, 0.004702079826, 0.004092405047, 0.003907918293, 0.004086110471, 0.003608937357, 0.003750278469, 0.003359302132, 0.003640954706, 0.003612722983, 0.003832686285, 0.004280689398, 0.003780650207, 0.003853131458, 0.003750471713, 0.003221510183, 0.003375329421, 0.00337204551, 0.003078967411, 0.003631111805, 0.003705470881, 0.004051691901, 0.003758263597, 0.003219548501, 0.00322410537, 0.003053520724, 0.00282870831, 0.002747014377, 0.002873165543, 0.002847918425, 0.003448010781, 0.003931554582, 0.003313178608, 0.003028926938, 0.003032511577, 0.002911464346, 0.002745567374, 0.002658174324, 0.003029978077, 0.003456979638, 0.003418816348, 0.003089516401, 0.002823881199, 0.002694148822, 0.002581815013, 0.002274152315, 0.002452890052, 0.002639344627, 0.002777070666, 0.003271405217, 0.002731006516, 0.002417733311, 0.002545041644, 0.002445071567, 0.002144393999, 0.002389854799, 0.002344599062, 0.002619809282, 0.002852917885, 0.003539833913, 0.002638401879, 0.002021217408, 0.002049355105, 0.001773477959, 0.001877138149, 0.002263133226, 0.00219391231, 0.002079094009, 0.002384093743, 0.002466890693, 0.001729753224, 0.001775353392, 0.001711371032, 0.001494355257, 0.001666919889, 0.001640984047, 0.001597138892],
						[0.005581524855, 0.004781393518, 0.004908853124, 0.005048488162, 0.004695260663, 0.004725555913, 0.004479292431, 0.004447562299, 0.004275795316, 0.004278451337, 0.004720826626, 0.004336785326, 0.004055297277, 0.003554139547, 0.003381220835, 0.003618877409, 0.003181156489, 0.003217439503, 0.003259972434, 0.004035653591, 0.004306432777, 0.003736776838, 0.003600043099, 0.003660395382, 0.003208342404, 0.003419811102, 0.002923043046, 0.003287871386, 0.003183279271, 0.003487995237, 0.003933100761, 0.003328081687, 0.003459011173, 0.003390186576, 0.002954394167, 0.002954282232, 0.003005882414, 0.002767666317, 0.003242079448, 0.003357423284, 0.003561854758, 0.003334818137, 0.002942575524, 0.002935013377, 0.002794287977, 0.002568939824, 0.002487098668, 0.002538668543, 0.002648547527, 0.003152359262, 0.003521043502, 0.003050485455, 0.00263296343, 0.002732898344, 0.002639358885, 0.00245577162, 0.00240993133, 0.002745503964, 0.003105609008, 0.003080811196, 0.002765467253, 0.002570024624, 0.002403539014, 0.002365928298, 0.002060583994, 0.002128802882, 0.002378697448, 0.002375537224, 0.002979944128, 0.002466248967, 0.002183305605, 0.00233221606, 0.002247460908, 0.00188411888, 0.002077564129, 0.002123604214, 0.002368484015, 0.002484697379, 0.003263079647, 0.002315163065, 0.001846701912, 0.00181648464, 0.001622591534, 0.001671580843, 0.002067975693, 0.001944108977, 0.001812298457, 0.002112557084, 0.002276282331, 0.001537089641, 0.001600344549, 0.001539839272, 0.001380351461, 0.001445855309, 0.001467154418, 0.001461220643],
						[0.005791371247, 0.004867670897, 0.005103830919, 0.005242939938, 0.004884578153, 0.00494187006, 0.004661042738, 0.004656774587, 0.004278451337, 0.004640905014, 0.00482602218, 0.004458280353, 0.004183449769, 0.003645954911, 0.003425163222, 0.003676113678, 0.003263922036, 0.003322619244, 0.00337732856, 0.004206406044, 0.004412656627, 0.003819775592, 0.003648447356, 0.003715759146, 0.003384254134, 0.003439829175, 0.003038721767, 0.003383124035, 0.003306777089, 0.003614084622, 0.003957043704, 0.003394430058, 0.003520707926, 0.00340538183, 0.002993569578, 0.003161183487, 0.002984117306, 0.002825546864, 0.003317706912, 0.003465333124, 0.003620403799, 0.00336627656, 0.002999951399, 0.002888635953, 0.002803079025, 0.002559037881, 0.002522217349, 0.002613303501, 0.002654684242, 0.003261252232, 0.003583525581, 0.003009493202, 0.002728739572, 0.00276611704, 0.00257947995, 0.002479049424, 0.002497650293, 0.00278677727, 0.003201191769, 0.003171901734, 0.002852639794, 0.002556324433, 0.002395752497, 0.002348819196, 0.002024066215, 0.002145725437, 0.002443764267, 0.002445315495, 0.002970631164, 0.002471499072, 0.00217780669, 0.002264720672, 0.002206492379, 0.001912792792, 0.002093213831, 0.00213031204, 0.00237844092, 0.002584543946, 0.003406758634, 0.002316135476, 0.001841431471, 0.001807023938, 0.001595657068, 0.001697851274, 0.002067148721, 0.001983828019, 0.001884783965, 0.002081992134, 0.00232940124, 0.001503426019, 0.001564091681, 0.001499557562, 0.001293417718, 0.001433147338, 0.001459815065, 0.001456114868],
						[0.006839503165, 0.006057294555, 0.006073314853, 0.006111949473, 0.00563581538, 0.005657857003, 0.005244524452, 0.005183246677, 0.004720826626, 0.00482602218, 0.00715380894, 0.00568062383, 0.005189433557, 0.004630976216, 0.004338145108, 0.004523508387, 0.003894140167, 0.003935601645, 0.003681198312, 0.004681323607, 0.005995902923, 0.005062114042, 0.004514844807, 0.004573083939, 0.004188643902, 0.003980118185, 0.003462811048, 0.003809565319, 0.003914667109, 0.004022603493, 0.005713294222, 0.004837360272, 0.00468190493, 0.004517144876, 0.003730300128, 0.003878401571, 0.003657651758, 0.003235475906, 0.003778763194, 0.00390359752, 0.005197194723, 0.00435597295, 0.003811533041, 0.003647806326, 0.003519599805, 0.003250822993, 0.002851594993, 0.003133866747, 0.003116543132, 0.003691230013, 0.004951209144, 0.004333963728, 0.003606982422, 0.003729182959, 0.003338004342, 0.003044495193, 0.002987117227, 0.003204243619, 0.003555873242, 0.003764693424, 0.004334541133, 0.003553878345, 0.003231203484, 0.002943099406, 0.002696793804, 0.002640047186, 0.003154011213, 0.0030178547, 0.003233047781, 0.003705899965, 0.003057049188, 0.003098922292, 0.002903858749, 0.00246176927, 0.002599881832, 0.002592969402, 0.002729180908, 0.003147871729, 0.003568750149, 0.003247957188, 0.002593234347, 0.002508700781, 0.002074163928, 0.00205201753, 0.002574940276, 0.002281943433, 0.002136746694, 0.002578710274, 0.002542355418, 0.002360605143, 0.002220554732, 0.002113770986, 0.001821547472, 0.001864275149, 0.001779315548, 0.00185400446],
						[0.006026119572, 0.005460206075, 0.005499701853, 0.005465037035, 0.0051420724, 0.005163692584, 0.004844508527, 0.00477351631, 0.004336785326, 0.004458280353, 0.00568062383, 0.005716769838, 0.0045563704, 0.004052643741, 0.003908646935, 0.004102625285, 0.003502287588, 0.003641064362, 0.003561828604, 0.004279809736, 0.005264020749, 0.004513935274, 0.004092159872, 0.004119086638, 0.003783692589, 0.003874680981, 0.003291156127, 0.003635289158, 0.003635098529, 0.003767603955, 0.005032640564, 0.00427678359, 0.004236813741, 0.004152959872, 0.003422740202, 0.003607572409, 0.003390717311, 0.003054099004, 0.003663946352, 0.003692743513, 0.004504288936, 0.004016010617, 0.003455978013, 0.003492681159, 0.003262636438, 0.003004556681, 0.002798426204, 0.002855994378, 0.002949423992, 0.003455392502, 0.004442682175, 0.003868226071, 0.003236383866, 0.003332541527, 0.003093094158, 0.002851603943, 0.002816757742, 0.003150937277, 0.003371518839, 0.003421802213, 0.003810412801, 0.003295890373, 0.002979283605, 0.002720910432, 0.002541390204, 0.002527176121, 0.002777588485, 0.00282129975, 0.003289166972, 0.003287562718, 0.00280330556, 0.002900054774, 0.00263994758, 0.002390474875, 0.002500464799, 0.002486099168, 0.002786945071, 0.002980381819, 0.003553381874, 0.002975489791, 0.002443998787, 0.002360782909, 0.002067739944, 0.001998907565, 0.002534972683, 0.002236704862, 0.00217467103, 0.002359074018, 0.002538706295, 0.002178968325, 0.002083663409, 0.001987904628, 0.001794990531, 0.001780425059, 0.001803667429, 0.001750839599],
						[0.00556042548, 0.004834687069, 0.004969641429, 0.005129873692, 0.004752966669, 0.004706632845, 0.004519702335, 0.004412606969, 0.004055297277, 0.004183449769, 0.005189433557, 0.0045563704, 0.005036636106, 0.0038638914, 0.003697488713, 0.003872703062, 0.003331785478, 0.003494649964, 0.00343602649, 0.004112031878, 0.004721710351, 0.004106976616, 0.003784774051, 0.003711100379, 0.00356760794, 0.003569918691, 0.003247660017, 0.003570598902, 0.00352122161, 0.003795115997, 0.004639492099, 0.003769156389, 0.00392091161, 0.003720303187, 0.003305639622, 0.003329627441, 0.00330156276, 0.002915103266, 0.00347520935, 0.003564945967, 0.004115633012, 0.003837504568, 0.003139800334, 0.003134726059, 0.003276090681, 0.002948850289, 0.002638145651, 0.002842945493, 0.003000883638, 0.003455188678, 0.00405233259, 0.003689967188, 0.002976667966, 0.00317508292, 0.003002616169, 0.002716058382, 0.002865345579, 0.002911171213, 0.0033207877, 0.003383236234, 0.003420739267, 0.003050888097, 0.002846153878, 0.002658557789, 0.00248201665, 0.002494794761, 0.002796370631, 0.002825190452, 0.003221488961, 0.002953910851, 0.002647777308, 0.00267266428, 0.002675630159, 0.002214654882, 0.002358356459, 0.002414975934, 0.002506934412, 0.002874428286, 0.003548025327, 0.002676243201, 0.00227159661, 0.002203276806, 0.002040575238, 0.001982869769, 0.002484016339, 0.002223537119, 0.002091956136, 0.002527216656, 0.00245712161, 0.00192903563, 0.002014819832, 0.00197991251, 0.001669107432, 0.001871868301, 0.001839191813, 0.001779332353],
						[0.004834846375, 0.00427960842, 0.004398641598, 0.004416730131, 0.004155890204, 0.004173076723, 0.003901269768, 0.003868620735, 0.003554139547, 0.003645954911, 0.004630976216, 0.004052643741, 0.0038638914, 0.003873833194, 0.003250860823, 0.003363765299, 0.002882109091, 0.003108371262, 0.002917879079, 0.003545322717, 0.004265575331, 0.00365789709, 0.003297331834, 0.003420793192, 0.003095560522, 0.003088608654, 0.002741008061, 0.00297496778, 0.002989693088, 0.003099461678, 0.003987789659, 0.003450262327, 0.003397950778, 0.00336390694, 0.002851727225, 0.002891509442, 0.002818125281, 0.002423383822, 0.002985284814, 0.003095657796, 0.003702221795, 0.003339200871, 0.002930283919, 0.002769878876, 0.002819545433, 0.002524422025, 0.002300083628, 0.002393712985, 0.002549847288, 0.002984198221, 0.003592822144, 0.00315543466, 0.002677606164, 0.002848600657, 0.002637211277, 0.002388155994, 0.002352838588, 0.002637722498, 0.00285052315, 0.002949664632, 0.003039306391, 0.002737754808, 0.002598694665, 0.002320043343, 0.002149391811, 0.002141692279, 0.002455331697, 0.002372667954, 0.002748106064, 0.002744602577, 0.002357861268, 0.002469088971, 0.002266383371, 0.001915814607, 0.002137951166, 0.002032070219, 0.002162047775, 0.002401755863, 0.003083914722, 0.002498508202, 0.002029350921, 0.002015332046, 0.001718203143, 0.001682071098, 0.002112971385, 0.001897921416, 0.001773218193, 0.002041876751, 0.002086498421, 0.001765418608, 0.001761942727, 0.001704274891, 0.00152144715, 0.001538600425, 0.001536824264, 0.00156334151],
						[0.004511975974, 0.004094676601, 0.004134976123, 0.004205399686, 0.003897907344, 0.003902254463, 0.003814645021, 0.003722629721, 0.003381220835, 0.003425163222, 0.004338145108, 0.003908646935, 0.003697488713, 0.003250860823, 0.00382247693, 0.003213676416, 0.002929488868, 0.002993233232, 0.002912453456, 0.003561962143, 0.004050883743, 0.003683104325, 0.003225346346, 0.003335041919, 0.003178531074, 0.003193442983, 0.002712119593, 0.003110599618, 0.003033533277, 0.003223937921, 0.003787378884, 0.003347239651, 0.003280235516, 0.003345649319, 0.002756067678, 0.002767110209, 0.002914685529, 0.002613713399, 0.002927980931, 0.003180273915, 0.003624700079, 0.003335812529, 0.002856828165, 0.002905681077, 0.002759159271, 0.002564853416, 0.002352480996, 0.002384606717, 0.002531822685, 0.002957736965, 0.003428895014, 0.003150561594, 0.00263577638, 0.002772655784, 0.00257720391, 0.002387000641, 0.002357687016, 0.002638124, 0.002886483586, 0.002862711554, 0.002969202312, 0.002639004767, 0.002423556742, 0.002302659107, 0.002211500634, 0.002164797252, 0.002527220919, 0.002339467906, 0.002896492621, 0.002636127797, 0.002371647666, 0.002451257932, 0.002351451581, 0.001999883551, 0.002172617411, 0.002162700965, 0.002268925825, 0.002563441418, 0.003174658432, 0.002401444852, 0.001947953835, 0.001925131204, 0.001891678493, 0.001711701281, 0.002146911221, 0.002034977418, 0.001895268085, 0.002260311334, 0.002183371878, 0.001703799545, 0.001711400505, 0.001713412, 0.001534514796, 0.00171764512, 0.001629135281, 0.00166261154],
						[0.00502810301, 0.004430417541, 0.004525198382, 0.004458656717, 0.004250814324, 0.004284994782, 0.004000603532, 0.003997654346, 0.003618877409, 0.003676113678, 0.004523508387, 0.004102625285, 0.003872703062, 0.003363765299, 0.003213676416, 0.003883552949, 0.002967310677, 0.003138069179, 0.002950611747, 0.0036818911, 0.004142008143, 0.003493402694, 0.003368136036, 0.003444248522, 0.003129152572, 0.003263735276, 0.002896374512, 0.003062725198, 0.003016611448, 0.003231478037, 0.003999979567, 0.003306005668, 0.00343324518, 0.003199677237, 0.002782050118, 0.002870923467, 0.002859287353, 0.00266264615, 0.003115850747, 0.003073912489, 0.003631119945, 0.003343044549, 0.00289164188, 0.002883360707, 0.002722695521, 0.002568207708, 0.002351255139, 0.002444673168, 0.002568367833, 0.003064479536, 0.003569656968, 0.003122008957, 0.00274180904, 0.002790751101, 0.00260473893, 0.002401464726, 0.002339342052, 0.002678514626, 0.002819510233, 0.002970506193, 0.002821127856, 0.002624749646, 0.002501434852, 0.002245179492, 0.002192825626, 0.002164558022, 0.002394731624, 0.002495798699, 0.002804875205, 0.002463355151, 0.002245615832, 0.002447919046, 0.002278303381, 0.001985647385, 0.002228671992, 0.002114276384, 0.002285185976, 0.002486016887, 0.003060032788, 0.002355436976, 0.001903773391, 0.001912987547, 0.001729006647, 0.001644036295, 0.002084742402, 0.00191297072, 0.001819593425, 0.002199419717, 0.002243951131, 0.001705130056, 0.001712051313, 0.001694201176, 0.001474786573, 0.001594540806, 0.001590322296, 0.001530771977],
						[0.004273827084, 0.003713099085, 0.003883988522, 0.003886593217, 0.00363196973, 0.00371978044, 0.003560416181, 0.00351483726, 0.003181156489, 0.003263922036, 0.003894140167, 0.003502287588, 0.003331785478, 0.002882109091, 0.002929488868, 0.002967310677, 0.003044465794, 0.002733450929, 0.002635010927, 0.003274962376, 0.003681445239, 0.003179269931, 0.002838947865, 0.003071383135, 0.002865501922, 0.002941029134, 0.002605314467, 0.002783878304, 0.002705720095, 0.00292749477, 0.003485988704, 0.00309340092, 0.003076533575, 0.002884361566, 0.002539860536, 0.002578019495, 0.002610227175, 0.002350060347, 0.002642787593, 0.002800009414, 0.003194270936, 0.002960725557, 0.002526066962, 0.002600968519, 0.002373548708, 0.00233798433, 0.002110325803, 0.002179840156, 0.002177178217, 0.002647144259, 0.002943346366, 0.002783759962, 0.00247288644, 0.002423331136, 0.002272073889, 0.002157631265, 0.002150761998, 0.002429601622, 0.00258608262, 0.002629567419, 0.002489536208, 0.002321161174, 0.00216214058, 0.002093913888, 0.001967676888, 0.001976377986, 0.00212531014, 0.002052453022, 0.002483985727, 0.00221991072, 0.002045275133, 0.002103865601, 0.00200061205, 0.001752657322, 0.001914621865, 0.001904786973, 0.00196953169, 0.002308013374, 0.002725449108, 0.002072097471, 0.00164384809, 0.001742680996, 0.001600301608, 0.001561717078, 0.001859878712, 0.001712852285, 0.001679136737, 0.001911216259, 0.001958052651, 0.001415231144, 0.00147673357, 0.001456591754, 0.001216366287, 0.001374620282, 0.001367862053, 0.001422734186],
						[0.004391065668, 0.003918644507, 0.003985939094, 0.003969078367, 0.003733774514, 0.003800660873, 0.003645410684, 0.003593304073, 0.003217439503, 0.003322619244, 0.003935601645, 0.003641064362, 0.003494649964, 0.003108371262, 0.002993233232, 0.003138069179, 0.002733450929, 0.00321693965, 0.0027661938, 0.003352791662, 0.003773810179, 0.003292673967, 0.002995705642, 0.003130947967, 0.002948265548, 0.002959804007, 0.002697129301, 0.002903473022, 0.002869045667, 0.003031698175, 0.003634588359, 0.003086208451, 0.003203120949, 0.003092637923, 0.002672967873, 0.002738575719, 0.002763725286, 0.002485791671, 0.00294973082, 0.002929667055, 0.003251648963, 0.003135708876, 0.002704731628, 0.002625437784, 0.002637326282, 0.002359066446, 0.002255987615, 0.002265448903, 0.002420139337, 0.002924262235, 0.003350293021, 0.002951902598, 0.002461926978, 0.002655913987, 0.002527600965, 0.002327256475, 0.002300116212, 0.002474336877, 0.002739514386, 0.002810675895, 0.002664617975, 0.002577909445, 0.002402551657, 0.002241335738, 0.002078977645, 0.002098506638, 0.002261782071, 0.002340604819, 0.002765602047, 0.002390797271, 0.002148585308, 0.002367405345, 0.002224109391, 0.001910957354, 0.002059282986, 0.002030608105, 0.002143121664, 0.002335886857, 0.002960454181, 0.002227701645, 0.001910514222, 0.00190527573, 0.001734247778, 0.001685271276, 0.002121516629, 0.001922968469, 0.001818487049, 0.002148142537, 0.0021263257, 0.001657059318, 0.001654215713, 0.001605786675, 0.001515116099, 0.001592331939, 0.001566729393, 0.001568749327],
						[0.004333287312, 0.003805239886, 0.003937507735, 0.003980490598, 0.003756053426, 0.003827532969, 0.003622276933, 0.00361034607, 0.003259972434, 0.00337732856, 0.003681198312, 0.003561828604, 0.00343602649, 0.002917879079, 0.002912453456, 0.002950611747, 0.002635010927, 0.0027661938, 0.003139940126, 0.003427624549, 0.003578906624, 0.00307687328, 0.002969666323, 0.00302098847, 0.002743564708, 0.002933038244, 0.002606332896, 0.002895006413, 0.002805153446, 0.003141477769, 0.003278536375, 0.002795899555, 0.002969012724, 0.002875247054, 0.002543491766, 0.00256979023, 0.002599392839, 0.002394934838, 0.002872689712, 0.002960508812, 0.002908189163, 0.002926999021, 0.002450992361, 0.002432798607, 0.00239761976, 0.002189184483, 0.002227756109, 0.002331667699, 0.002421487646, 0.002875366694, 0.00298844369, 0.002606981108, 0.002317270052, 0.002405332647, 0.002385423734, 0.002234018339, 0.002249965789, 0.002484491958, 0.002751166494, 0.002687481463, 0.002383835896, 0.002218656927, 0.002133343147, 0.002062347094, 0.001937510594, 0.001986508049, 0.002164657603, 0.002328611956, 0.002693200054, 0.002124462034, 0.001949411141, 0.002055621468, 0.002065006753, 0.001803744266, 0.002000441696, 0.001965647569, 0.002239790169, 0.002399122799, 0.002997783046, 0.002008603922, 0.00166251154, 0.001645494544, 0.001594663869, 0.00156716494, 0.001981218007, 0.001904040506, 0.001840492434, 0.002101416757, 0.002207618824, 0.001389582413, 0.001484180669, 0.001422802845, 0.001374981132, 0.001391746185, 0.00151998509, 0.001501016057],
						[0.005263026397, 0.004724829646, 0.004796383717, 0.004776577303, 0.004551709944, 0.004633361599, 0.004387740333, 0.004428704724, 0.004035653591, 0.004206406044, 0.004681323607, 0.004279809736, 0.004112031878, 0.003545322717, 0.003561962143, 0.0036818911, 0.003274962376, 0.003352791662, 0.003427624549, 0.004713592054, 0.004412996599, 0.003691369524, 0.003495881485, 0.003740263973, 0.003380891517, 0.003446890802, 0.003165427335, 0.003558889199, 0.003415026759, 0.003849844066, 0.004096787328, 0.003443607935, 0.003573020979, 0.003413438583, 0.003179537078, 0.003132294152, 0.003151896462, 0.002894221711, 0.003341524005, 0.003697354707, 0.003778983674, 0.003473167568, 0.003193445134, 0.003074032778, 0.003001363495, 0.002813582463, 0.002728806664, 0.002756412871, 0.002729660122, 0.003574148845, 0.003742135969, 0.003261816697, 0.002881400693, 0.00295569575, 0.002771345251, 0.00263992019, 0.002789369528, 0.003027644353, 0.003339619025, 0.003391684513, 0.003009037919, 0.0027456759, 0.002538829406, 0.00246199063, 0.002241391942, 0.002453759042, 0.002658168093, 0.002796176512, 0.003167364769, 0.002702996108, 0.002418751103, 0.002507110088, 0.002455918705, 0.002200525629, 0.002357724883, 0.002368100476, 0.002607530716, 0.002848028378, 0.003700040109, 0.002599035893, 0.002003765691, 0.002031277, 0.001906474095, 0.001893518972, 0.002302300689, 0.002221521965, 0.002142081755, 0.002591435225, 0.002621234068, 0.001826813062, 0.001795320974, 0.001810262683, 0.001647121456, 0.001765786251, 0.001753677596, 0.00178581935],
						[0.006093797699, 0.005547769669, 0.00554862731, 0.005403948677, 0.005102927184, 0.005034250859, 0.004834011926, 0.004702079826, 0.004306432777, 0.004412656627, 0.005995902923, 0.005264020749, 0.004721710351, 0.004265575331, 0.004050883743, 0.004142008143, 0.003681445239, 0.003773810179, 0.003578906624, 0.004412996599, 0.006400274475, 0.004829108861, 0.004184353511, 0.004320974139, 0.003946196099, 0.003911279088, 0.003315031289, 0.003730880704, 0.003706376531, 0.003906636903, 0.005361479462, 0.004625380525, 0.004510725187, 0.004359761436, 0.003687701075, 0.003642718224, 0.003533366195, 0.003147419208, 0.003727981077, 0.003698115635, 0.0048289164, 0.004370629445, 0.003732576263, 0.003622061423, 0.003367869088, 0.003209881211, 0.002915577328, 0.003030936925, 0.003163595894, 0.003710364719, 0.004846839608, 0.004181164132, 0.003579984805, 0.003684863634, 0.003303553928, 0.002993021396, 0.003079061485, 0.003216678089, 0.003516415998, 0.003563033468, 0.004092782743, 0.003510117601, 0.003131965539, 0.002865249967, 0.002725611012, 0.002733212066, 0.002940756862, 0.002965537976, 0.003303248719, 0.003620976929, 0.003034990502, 0.003093187533, 0.002901151042, 0.002514304441, 0.002581041583, 0.002572983051, 0.002693267864, 0.003012297037, 0.003726597402, 0.00317777977, 0.002611472402, 0.00252861915, 0.002266167678, 0.002117982045, 0.002668235806, 0.002302778545, 0.00226003792, 0.002571792485, 0.002718526671, 0.002277131577, 0.002209301972, 0.002131248267, 0.001875955335, 0.00190288067, 0.001968385308, 0.00192768926],
						[0.005328759167, 0.004641512869, 0.00472970036, 0.004734314166, 0.004374288211, 0.004381039134, 0.004163941856, 0.004092405047, 0.003736776838, 0.003819775592, 0.005062114042, 0.004513935274, 0.004106976616, 0.00365789709, 0.003683104325, 0.003493402694, 0.003179269931, 0.003292673967, 0.00307687328, 0.003691369524, 0.004829108861, 0.004846353964, 0.003666264902, 0.003736210123, 0.003435652026, 0.003327434877, 0.002786963435, 0.003327749918, 0.003165502457, 0.003293778884, 0.00458887544, 0.003868415051, 0.003810806758, 0.003812135294, 0.003031235293, 0.00315819996, 0.003127765307, 0.002679971527, 0.003145197477, 0.003229609154, 0.004040609008, 0.003625382924, 0.003093686213, 0.003046358195, 0.002982998964, 0.002713429037, 0.00249017059, 0.002583465192, 0.002667415222, 0.002995624971, 0.004059070863, 0.003637699749, 0.002998394955, 0.003056517742, 0.002810718864, 0.002538625765, 0.0024774441, 0.00270785449, 0.003038493889, 0.003132510879, 0.003491058531, 0.00292253278, 0.002812505344, 0.002501828039, 0.002316819949, 0.002253495925, 0.002516935413, 0.002451923454, 0.002803321185, 0.003086465809, 0.002570900195, 0.002649884076, 0.002471890206, 0.002036089535, 0.002110086306, 0.002118461054, 0.002273008844, 0.002606396206, 0.003256030568, 0.002670710051, 0.002162432991, 0.002133878537, 0.00189527577, 0.001799982984, 0.00224771988, 0.001895963612, 0.001801342803, 0.002218445191, 0.00220851091, 0.001925310451, 0.001872333325, 0.001733466947, 0.001460972446, 0.001628559069, 0.001487323111, 0.001518940644],
						[0.004885570623, 0.004275172741, 0.004383809539, 0.004410779587, 0.004156728923, 0.004140177808, 0.003798179843, 0.003907918293, 0.003600043099, 0.003648447356, 0.004514844807, 0.004092159872, 0.003784774051, 0.003297331834, 0.003225346346, 0.003368136036, 0.002838947865, 0.002995705642, 0.002969666323, 0.003495881485, 0.004184353511, 0.003666264902, 0.004179783337, 0.003471414173, 0.003023854828, 0.003356610161, 0.002698519832, 0.003038886897, 0.003003883224, 0.003341485813, 0.003794583205, 0.003293611224, 0.003492744597, 0.003326876146, 0.002819414503, 0.002903537259, 0.002864338184, 0.002573591133, 0.002974747996, 0.003097156618, 0.003585739896, 0.003359651379, 0.002702284528, 0.00280620731, 0.002800027585, 0.002587252992, 0.00237875066, 0.002318001535, 0.002577506406, 0.003046000888, 0.003738788513, 0.003156040757, 0.002682901711, 0.002738199901, 0.002771310142, 0.002361977066, 0.00228325972, 0.00259919058, 0.00286215559, 0.002856169711, 0.002929151204, 0.002730446441, 0.002602943627, 0.002393241764, 0.002125213045, 0.002038210391, 0.002278688507, 0.002451203521, 0.002800159684, 0.002582754777, 0.002317755746, 0.002533026189, 0.002351341128, 0.001900481007, 0.002128911057, 0.002028752638, 0.002231438776, 0.002492451223, 0.003146094171, 0.002458733781, 0.001977984212, 0.002032270797, 0.001658662239, 0.001736552634, 0.002087295719, 0.001880784241, 0.001813132445, 0.00215663356, 0.002089571752, 0.00168525719, 0.001848983201, 0.001633653508, 0.00153148607, 0.001521509988, 0.001577008256, 0.001443522175],
						[0.005006898428, 0.004535712695, 0.004507738633, 0.004359942384, 0.004290280092, 0.004271141624, 0.003995557898, 0.004086110471, 0.003660395382, 0.003715759146, 0.004573083939, 0.004119086638, 0.003711100379, 0.003420793192, 0.003335041919, 0.003444248522, 0.003071383135, 0.003130947967, 0.00302098847, 0.003740263973, 0.004320974139, 0.003736210123, 0.003471414173, 0.004219814752, 0.003074347462, 0.00335558002, 0.002853020663, 0.003163075181, 0.003162460188, 0.003383226237, 0.003946854786, 0.003358245728, 0.003492581255, 0.00347055143, 0.002864913462, 0.002947059973, 0.002977872901, 0.002697692301, 0.00306046926, 0.003209536799, 0.003709842156, 0.003330488186, 0.002923999009, 0.003021807234, 0.002858419046, 0.002557295966, 0.002518513222, 0.00260775945, 0.002569833295, 0.003149431155, 0.003752861444, 0.003185902301, 0.002801078981, 0.002773726189, 0.002783920954, 0.002526022159, 0.002509305232, 0.00270304282, 0.002986993227, 0.003066809247, 0.003032034749, 0.002679082237, 0.002584795124, 0.00242750741, 0.002219893885, 0.002286300568, 0.002416626155, 0.002570982108, 0.002968969687, 0.002657204096, 0.002334412356, 0.002524391039, 0.00233510927, 0.002032861451, 0.002332175229, 0.002109761204, 0.002414349499, 0.002598142746, 0.003200476161, 0.002519373609, 0.00194844727, 0.002044466161, 0.001777307171, 0.001766203399, 0.002226324109, 0.002049160193, 0.00191435273, 0.002332399788, 0.002324302701, 0.001714808166, 0.001772016561, 0.001672395879, 0.001495796545, 0.001619156643, 0.001669654434, 0.001534071611],
						[0.004375058515, 0.004002068481, 0.004099612817, 0.004043738715, 0.003757206817, 0.003815087583, 0.00373019354, 0.003608937357, 0.003208342404, 0.003384254134, 0.004188643902, 0.003783692589, 0.00356760794, 0.003095560522, 0.003178531074, 0.003129152572, 0.002865501922, 0.002948265548, 0.002743564708, 0.003380891517, 0.003946196099, 0.003435652026, 0.003023854828, 0.003074347462, 0.003587875816, 0.002988766884, 0.002676701988, 0.002935990214, 0.002847540267, 0.003027385698, 0.003833575022, 0.003323279761, 0.003323942424, 0.003132949639, 0.002732966861, 0.002828162298, 0.002733380803, 0.002420330457, 0.002918191526, 0.002943742574, 0.003517684979, 0.003152410878, 0.002764708831, 0.002673637997, 0.002574437589, 0.002454666916, 0.002211056038, 0.002238484936, 0.002225958147, 0.002883740895, 0.00337803957, 0.003022747245, 0.002552899392, 0.002698310427, 0.002457512633, 0.0023063138, 0.002281525885, 0.002496782109, 0.002708649286, 0.002886268458, 0.002825941231, 0.002597498699, 0.002309029674, 0.002188957955, 0.002129636242, 0.002074964307, 0.002305354962, 0.002290907059, 0.002604838112, 0.002458205216, 0.002175005627, 0.002335369335, 0.002151563889, 0.001885691664, 0.002024016797, 0.001981590044, 0.002175490999, 0.002433038059, 0.002925505835, 0.002296204885, 0.001843445266, 0.001861205512, 0.001682774242, 0.001597012896, 0.002031397406, 0.001888393315, 0.001786999549, 0.002043348134, 0.002036454966, 0.001631303246, 0.001626562968, 0.001603387422, 0.00138335841, 0.001554231646, 0.00142310159, 0.001518029966],
						[0.004536227401, 0.003998855101, 0.004120392418, 0.00403985557, 0.00387302364, 0.003888776041, 0.003728689545, 0.003750278469, 0.003419811102, 0.003439829175, 0.003980118185, 0.003874680981, 0.003569918691, 0.003088608654, 0.003193442983, 0.003263735276, 0.002941029134, 0.002959804007, 0.002933038244, 0.003446890802, 0.003911279088, 0.003327434877, 0.003356610161, 0.00335558002, 0.002988766884, 0.003799319481, 0.002803502387, 0.003093195063, 0.003051601905, 0.003287937879, 0.003713693103, 0.003127171505, 0.003227872339, 0.003230790424, 0.002656555739, 0.002746432141, 0.002948685667, 0.002492862527, 0.003131702677, 0.003070495267, 0.003482263014, 0.003260850024, 0.002760310073, 0.00289823961, 0.002656191464, 0.002533025109, 0.002453741056, 0.002358155389, 0.002498554716, 0.002957772661, 0.003434297459, 0.003012799755, 0.002586533688, 0.002659368438, 0.002638276807, 0.002365156932, 0.002266873072, 0.002727073654, 0.00298885925, 0.002962162727, 0.002724667161, 0.002603957511, 0.00239513157, 0.002366851537, 0.002204672391, 0.002124099997, 0.002254771995, 0.002485193819, 0.002957245289, 0.002441650783, 0.002273750322, 0.002472290253, 0.002295598751, 0.001996875499, 0.002265404134, 0.002150671015, 0.002292409218, 0.002491206316, 0.003125602712, 0.00234600001, 0.001961811834, 0.001976809919, 0.001834543289, 0.001731717203, 0.002172197335, 0.0020091459, 0.001939627161, 0.002327275906, 0.002234689403, 0.001603869682, 0.001739008661, 0.001730536194, 0.00160289107, 0.00161558132, 0.001635188623, 0.001538851026],
						[0.00382383205, 0.003461115736, 0.003465223346, 0.003443041304, 0.003394904912, 0.003412560729, 0.00322625399, 0.003359302132, 0.002923043046, 0.003038721767, 0.003462811048, 0.003291156127, 0.003247660017, 0.002741008061, 0.002712119593, 0.002896374512, 0.002605314467, 0.002697129301, 0.002606332896, 0.003165427335, 0.003315031289, 0.002786963435, 0.002698519832, 0.002853020663, 0.002676701988, 0.002803502387, 0.00306700713, 0.002735058614, 0.002695636045, 0.002991305127, 0.00319868205, 0.002879138605, 0.002890740144, 0.00277773913, 0.002525763465, 0.002601870775, 0.002549273479, 0.002315693113, 0.002736115921, 0.00279314337, 0.002947256641, 0.002949088362, 0.002438530769, 0.002494613245, 0.002427808488, 0.0023698177, 0.002180214972, 0.002209502928, 0.002311303426, 0.002783458454, 0.002905864464, 0.002661201218, 0.002421328055, 0.00247621901, 0.002356458479, 0.002146446492, 0.002187583781, 0.002492489763, 0.002553266905, 0.002613346258, 0.002381112847, 0.002299599325, 0.002138305566, 0.002185859516, 0.002131839155, 0.002043831445, 0.002271061995, 0.00229176301, 0.00263039294, 0.00211106795, 0.002025041307, 0.002161650698, 0.002002517374, 0.001885994037, 0.002042060296, 0.001988821433, 0.002202094963, 0.002329232732, 0.002684325948, 0.002051566, 0.001725660176, 0.001823756082, 0.001687102396, 0.001621796456, 0.001933316665, 0.001898223352, 0.001877129076, 0.002004063264, 0.002049152968, 0.001490542702, 0.001573055978, 0.001556844155, 0.001397123182, 0.001458963187, 0.001493556021, 0.001504941644],
						[0.004188885336, 0.003868338392, 0.003916806442, 0.003961335731, 0.003736503159, 0.003715946693, 0.003643686193, 0.003640954706, 0.003287871386, 0.003383124035, 0.003809565319, 0.003635289158, 0.003570598902, 0.00297496778, 0.003110599618, 0.003062725198, 0.002783878304, 0.002903473022, 0.002895006413, 0.003558889199, 0.003730880704, 0.003327749918, 0.003038886897, 0.003163075181, 0.002935990214, 0.003093195063, 0.002735058614, 0.003367045221, 0.00296287914, 0.003316453212, 0.003559494386, 0.003022089541, 0.00315814943, 0.003016359076, 0.002667549281, 0.002677872445, 0.002914688297, 0.002509079189, 0.003004795826, 0.003058055976, 0.003248632596, 0.003158675801, 0.00260980252, 0.002709653171, 0.002696348858, 0.002456309315, 0.002423594261, 0.002423369531, 0.002442380546, 0.003036829971, 0.003284227606, 0.0029268648, 0.002456698341, 0.002705616909, 0.002486510885, 0.002321681675, 0.002466251737, 0.002640528653, 0.002961271301, 0.002871738576, 0.002652598675, 0.002495345023, 0.002318905424, 0.0022650826, 0.002093950813, 0.002171420609, 0.002351905388, 0.002521800233, 0.002877079814, 0.002384031672, 0.002151030603, 0.002317802626, 0.00226400016, 0.002005291199, 0.002147263607, 0.00215961835, 0.002237437268, 0.002500962852, 0.0032285676, 0.00226710001, 0.001868700425, 0.001834793041, 0.001821025008, 0.001726301118, 0.002169956628, 0.002032204841, 0.001974984732, 0.002326966181, 0.002348328682, 0.001651025562, 0.001691110869, 0.001726021726, 0.001591942106, 0.001656064673, 0.001621731298, 0.001688645496],
						[0.004227284984, 0.003743206232, 0.003893286734, 0.003888049826, 0.00370489262, 0.003694375629, 0.003498595594, 0.003612722983, 0.003183279271, 0.003306777089, 0.003914667109, 0.003635098529, 0.00352122161, 0.002989693088, 0.003033533277, 0.003016611448, 0.002705720095, 0.002869045667, 0.002805153446, 0.003415026759, 0.003706376531, 0.003165502457, 0.003003883224, 0.003162460188, 0.002847540267, 0.003051601905, 0.002695636045, 0.00296287914, 0.003230447816, 0.003234096768, 0.003466110855, 0.003037932947, 0.003089193132, 0.00300929479, 0.00257306987, 0.00266034213, 0.00273167138, 0.002420467153, 0.002915252933, 0.002969772584, 0.003252968055, 0.003136741139, 0.002526157358, 0.002702039418, 0.002609354892, 0.002335447056, 0.002329560571, 0.002357951196, 0.00246716127, 0.002996461714, 0.003250542802, 0.002808323347, 0.002478598141, 0.002593966541, 0.002468548448, 0.00228161994, 0.002341879803, 0.002538539563, 0.002828931675, 0.002913652699, 0.002624850473, 0.002433160548, 0.002332568494, 0.002191589414, 0.00211624525, 0.00214557857, 0.002334825057, 0.002464649154, 0.002830893306, 0.002365428219, 0.002215903769, 0.002292992308, 0.002186510036, 0.001928013396, 0.002127059326, 0.002026009422, 0.002247039578, 0.002468833821, 0.003034514466, 0.002220333684, 0.001885137482, 0.001871014565, 0.001755383966, 0.001768774387, 0.002103557194, 0.001956164834, 0.001946371307, 0.002220914779, 0.002292293272, 0.00153782559, 0.001667038649, 0.001664785053, 0.001486070658, 0.001637206048, 0.001584095463, 0.001656696165],
						[0.004480470259, 0.00397440804, 0.00427363146, 0.004038458787, 0.003989563848, 0.003998670233, 0.00389183741, 0.003832686285, 0.003487995237, 0.003614084622, 0.004022603493, 0.003767603955, 0.003795115997, 0.003099461678, 0.003223937921, 0.003231478037, 0.00292749477, 0.003031698175, 0.003141477769, 0.003849844066, 0.003906636903, 0.003293778884, 0.003341485813, 0.003383226237, 0.003027385698, 0.003287937879, 0.002991305127, 0.003316453212, 0.003234096768, 0.004257539481, 0.00360892735, 0.003089312147, 0.00314914342, 0.003293199865, 0.002775515786, 0.002848306622, 0.002993796676, 0.002688962135, 0.003248114478, 0.003503084115, 0.003305595067, 0.003361283934, 0.002744912459, 0.002780883803, 0.002735180879, 0.002597800854, 0.002580293557, 0.002562534438, 0.002716587516, 0.003329739, 0.003374641407, 0.002900609376, 0.002693202523, 0.002619729093, 0.002685051147, 0.00240451075, 0.002498561864, 0.002693658519, 0.003142076927, 0.003097697868, 0.00259576905, 0.002535927788, 0.002366890196, 0.002410639485, 0.002165431495, 0.002328296499, 0.00245450379, 0.002711650031, 0.003238324681, 0.002348861861, 0.002259480343, 0.002309906809, 0.002355579841, 0.002038839365, 0.002239321229, 0.002284708105, 0.002384424834, 0.002667066797, 0.003507441082, 0.002274031605, 0.001883094939, 0.001964218654, 0.00181361643, 0.001834717014, 0.002292201244, 0.002216786869, 0.002177387041, 0.002613953357, 0.002567909878, 0.001524680113, 0.001779855608, 0.001690825931, 0.001564463116, 0.001679989757, 0.001671031498, 0.001728721221],
						[0.005652719041, 0.005132938989, 0.005206899053, 0.004970339251, 0.004710173158, 0.00466125509, 0.00449380729, 0.004280689398, 0.003933100761, 0.003957043704, 0.005713294222, 0.005032640564, 0.004639492099, 0.003987789659, 0.003787378884, 0.003999979567, 0.003485988704, 0.003634588359, 0.003278536375, 0.004096787328, 0.005361479462, 0.00458887544, 0.003794583205, 0.003946854786, 0.003833575022, 0.003713693103, 0.00319868205, 0.003559494386, 0.003466110855, 0.00360892735, 0.006389629496, 0.004543286095, 0.00435339964, 0.00421183499, 0.003643843907, 0.003552145708, 0.003377381468, 0.002940465183, 0.003674445074, 0.003546939204, 0.004832200731, 0.004167746276, 0.003702399965, 0.003501187289, 0.003368893116, 0.003111353431, 0.002770718558, 0.002793733171, 0.002945230267, 0.003532627186, 0.004823624868, 0.004206178584, 0.003434773421, 0.003554204283, 0.00317215526, 0.002848030125, 0.002739330765, 0.003111718767, 0.003134480965, 0.003584368242, 0.004337202597, 0.003567835962, 0.003097652087, 0.002903839306, 0.002862427116, 0.002742707281, 0.00303943435, 0.002918479447, 0.002895614018, 0.003558485885, 0.003049905475, 0.003085204278, 0.0028736148, 0.002450573985, 0.002543276964, 0.002552567112, 0.002644966347, 0.002905540983, 0.003571788009, 0.003231873214, 0.002752491792, 0.00253094309, 0.002247082258, 0.002097977091, 0.002585262202, 0.002213443299, 0.002178675095, 0.002510933318, 0.002546967054, 0.002496554737, 0.002295747956, 0.002140649214, 0.00184150236, 0.001894217899, 0.001779130572, 0.00175253825],
						[0.004684925202, 0.004315534403, 0.004323028253, 0.004186782856, 0.003965921161, 0.003956994273, 0.003763685986, 0.003780650207, 0.003328081687, 0.003394430058, 0.004837360272, 0.00427678359, 0.003769156389, 0.003450262327, 0.003347239651, 0.003306005668, 0.00309340092, 0.003086208451, 0.002795899555, 0.003443607935, 0.004625380525, 0.003868415051, 0.003293611224, 0.003358245728, 0.003323279761, 0.003127171505, 0.002879138605, 0.003022089541, 0.003037932947, 0.003089312147, 0.004543286095, 0.004619143534, 0.003587010759, 0.003561213594, 0.003084867702, 0.003114246027, 0.002929868051, 0.002477946386, 0.003101819533, 0.003145192983, 0.004121189692, 0.003653686039, 0.003180433981, 0.003098678454, 0.002956622019, 0.002822098896, 0.002300704258, 0.002436448376, 0.002488366376, 0.002998644227, 0.003872081124, 0.003577494119, 0.003077587403, 0.003050500899, 0.002748571156, 0.002444597683, 0.002435739537, 0.002681119544, 0.002759195476, 0.003072875462, 0.003471787293, 0.00303231373, 0.002760249885, 0.002524300375, 0.002394184833, 0.002314679862, 0.002563165524, 0.002435194682, 0.002696053478, 0.003113548988, 0.002701328888, 0.00269269277, 0.002425354578, 0.002117866581, 0.002233633219, 0.002230699024, 0.002236322138, 0.002683642961, 0.002926002114, 0.002786528801, 0.002314710289, 0.002238214903, 0.001959224997, 0.001948823312, 0.002238119388, 0.001904738534, 0.001925369836, 0.002235240477, 0.002217031467, 0.002028930962, 0.002013238493, 0.001896815714, 0.001624917218, 0.001691962259, 0.001543014166, 0.001629675397],
						[0.004824278, 0.004334464152, 0.004343625848, 0.004351604312, 0.004097876261, 0.004064539484, 0.003833958135, 0.003853131458, 0.003459011173, 0.003520707926, 0.00468190493, 0.004236813741, 0.00392091161, 0.003397950778, 0.003280235516, 0.00343324518, 0.003076533575, 0.003203120949, 0.002969012724, 0.003573020979, 0.004510725187, 0.003810806758, 0.003492744597, 0.003492581255, 0.003323942424, 0.003227872339, 0.002890740144, 0.00315814943, 0.003089193132, 0.00314914342, 0.00435339964, 0.003587010759, 0.00433071444, 0.003369536117, 0.003050768156, 0.002964649715, 0.003030567194, 0.002635540325, 0.00313070882, 0.003219439668, 0.003833987856, 0.003641186078, 0.002890780361, 0.00295686402, 0.003012434425, 0.002712519187, 0.002492592931, 0.002536415273, 0.002606521851, 0.003156837627, 0.003945092908, 0.003427932124, 0.002807858163, 0.003140722747, 0.002845330975, 0.002578218747, 0.002494993998, 0.002687851375, 0.002980382082, 0.002966540382, 0.00324307985, 0.003040830314, 0.002747754518, 0.002462205185, 0.00236962802, 0.002326459033, 0.002444427051, 0.002492011945, 0.002907310129, 0.002862370156, 0.002476277853, 0.002678459174, 0.002505757031, 0.00214030932, 0.002277384155, 0.002153952799, 0.00231883584, 0.002832604928, 0.003215172404, 0.00257810375, 0.002168545284, 0.002214754088, 0.001950404018, 0.00188735876, 0.002251596033, 0.002023847758, 0.001959518691, 0.002326900403, 0.002387760302, 0.001960869109, 0.00197660144, 0.001872576902, 0.00167473677, 0.001639575669, 0.001682717651, 0.001722636547],
						[0.004694296476, 0.004243317135, 0.00432711529, 0.004236631295, 0.003968104092, 0.003977163851, 0.003889181773, 0.003750471713, 0.003390186576, 0.00340538183, 0.004517144876, 0.004152959872, 0.003720303187, 0.00336390694, 0.003345649319, 0.003199677237, 0.002884361566, 0.003092637923, 0.002875247054, 0.003413438583, 0.004359761436, 0.003812135294, 0.003326876146, 0.00347055143, 0.003132949639, 0.003230790424, 0.00277773913, 0.003016359076, 0.00300929479, 0.003293199865, 0.00421183499, 0.003561213594, 0.003369536117, 0.00399116046, 0.002839814268, 0.002955843653, 0.002934780701, 0.002577714692, 0.003126073595, 0.00308664757, 0.003817000885, 0.003431157463, 0.002981412052, 0.002907614761, 0.002769647323, 0.002650533834, 0.002431642779, 0.002420464412, 0.002619850514, 0.002904114795, 0.003772310698, 0.003295540396, 0.002766997167, 0.002774585934, 0.002778439634, 0.002491512264, 0.002358714498, 0.00263069605, 0.002811431806, 0.002873848244, 0.003285817904, 0.002866599622, 0.002514428298, 0.002443524259, 0.002274656068, 0.00221991396, 0.002444334685, 0.002513984598, 0.002864591507, 0.002849514422, 0.002484707529, 0.002510756872, 0.002307803728, 0.002045824989, 0.002175205296, 0.002223016351, 0.002255465655, 0.002509502702, 0.003086664146, 0.002559313469, 0.002125851234, 0.002042135601, 0.001865060743, 0.001717740157, 0.002234850987, 0.001997505152, 0.001918653967, 0.002221537306, 0.002176170363, 0.001858518206, 0.001821790762, 0.001711593746, 0.001589475474, 0.001632520931, 0.001637769712, 0.001580531026],
						[0.0038311798, 0.003586062754, 0.003575259214, 0.003515810312, 0.003341734566, 0.003358623301, 0.003237279705, 0.003221510183, 0.002954394167, 0.002993569578, 0.003730300128, 0.003422740202, 0.003305639622, 0.002851727225, 0.002756067678, 0.002782050118, 0.002539860536, 0.002672967873, 0.002543491766, 0.003179537078, 0.003687701075, 0.003031235293, 0.002819414503, 0.002864913462, 0.002732966861, 0.002656555739, 0.002525763465, 0.002667549281, 0.00257306987, 0.002775515786, 0.003643843907, 0.003084867702, 0.003050768156, 0.002839814268, 0.003113261711, 0.002605304699, 0.002523949341, 0.002256129179, 0.002623905379, 0.002786969656, 0.003148736152, 0.002983505359, 0.002586300099, 0.002513571863, 0.002510273881, 0.002352049805, 0.002093244918, 0.002053063802, 0.002270429913, 0.002740250691, 0.00312637474, 0.002822365865, 0.002439855205, 0.002520047139, 0.002422558995, 0.002079851963, 0.002235407413, 0.002373286904, 0.002521135822, 0.002633358523, 0.002729913204, 0.00243657735, 0.002218249685, 0.002148550824, 0.00206209881, 0.002026451581, 0.0021809998, 0.002184221162, 0.002537095063, 0.002429949287, 0.002107394936, 0.002216443786, 0.002131532637, 0.001853938575, 0.001872324134, 0.001909036257, 0.002066530273, 0.00233571516, 0.002790613686, 0.002256992943, 0.001784834378, 0.001836535011, 0.001683397023, 0.001668607442, 0.001948746689, 0.001817722133, 0.001742905187, 0.001849463061, 0.001900591909, 0.001683628808, 0.001661515879, 0.00162246157, 0.001411743762, 0.001469458031, 0.001513598958, 0.001448566505],
						[0.004131720831, 0.003673558324, 0.003722534095, 0.003803261423, 0.003518722423, 0.003518226553, 0.003384040493, 0.003375329421, 0.002954282232, 0.003161183487, 0.003878401571, 0.003607572409, 0.003329627441, 0.002891509442, 0.002767110209, 0.002870923467, 0.002578019495, 0.002738575719, 0.00256979023, 0.003132294152, 0.003642718224, 0.00315819996, 0.002903537259, 0.002947059973, 0.002828162298, 0.002746432141, 0.002601870775, 0.002677872445, 0.00266034213, 0.002848306622, 0.003552145708, 0.003114246027, 0.002964649715, 0.002955843653, 0.002605304699, 0.003246127894, 0.002597081723, 0.002289770099, 0.002681355027, 0.002807109181, 0.003251010168, 0.002906715171, 0.002690006606, 0.002534592074, 0.002472799134, 0.002380507514, 0.002121420707, 0.002153921171, 0.002199153858, 0.00274069719, 0.003199500798, 0.002891564344, 0.002373193014, 0.002513681059, 0.002319327304, 0.002084102308, 0.00224190194, 0.002394257036, 0.00256134582, 0.00252863471, 0.002803873853, 0.002419776366, 0.002180843066, 0.002103332855, 0.002042384937, 0.001968265032, 0.002240327534, 0.002214681019, 0.002462364389, 0.002424836448, 0.002049107198, 0.002083366443, 0.00197987481, 0.001875720046, 0.001887109246, 0.001949640749, 0.002115343694, 0.002307184219, 0.002697538545, 0.002179623843, 0.001839841472, 0.001812945473, 0.001639023288, 0.001532550038, 0.001956169542, 0.00177166698, 0.001680075186, 0.001911598533, 0.001958579231, 0.001715632971, 0.001600549369, 0.001576340268, 0.00143020767, 0.001363441344, 0.001368196308, 0.001363541314],
						[0.003918816918, 0.003599684906, 0.003569795438, 0.003629734042, 0.003444342914, 0.003373116109, 0.003371544744, 0.00337204551, 0.003005882414, 0.002984117306, 0.003657651758, 0.003390717311, 0.00330156276, 0.002818125281, 0.002914685529, 0.002859287353, 0.002610227175, 0.002763725286, 0.002599392839, 0.003151896462, 0.003533366195, 0.003127765307, 0.002864338184, 0.002977872901, 0.002733380803, 0.002948685667, 0.002549273479, 0.002914688297, 0.00273167138, 0.002993796676, 0.003377381468, 0.002929868051, 0.003030567194, 0.002934780701, 0.002523949341, 0.002597081723, 0.003241193577, 0.002403651445, 0.00292431403, 0.002867665129, 0.003037304181, 0.003101957178, 0.002501104067, 0.002572921285, 0.002675836699, 0.002472353739, 0.002332287125, 0.002256236355, 0.002316121493, 0.002808901081, 0.003053367321, 0.002762400551, 0.002330442759, 0.002681173546, 0.002527505865, 0.002247129268, 0.002221462671, 0.002480365542, 0.002739803121, 0.002757168903, 0.00244546032, 0.002450922864, 0.002309054272, 0.002160272368, 0.002095941077, 0.002156213221, 0.002307032847, 0.002425733908, 0.00276051117, 0.002288477239, 0.002059309865, 0.002395086863, 0.002338798784, 0.001965452817, 0.002133015049, 0.002127609696, 0.002180562054, 0.002459078114, 0.003023566438, 0.002211358705, 0.001910315612, 0.001850515018, 0.001865905876, 0.001676086598, 0.002156986478, 0.001938585802, 0.001892057444, 0.002211490898, 0.002181747593, 0.00161786628, 0.001671386618, 0.001651221385, 0.001634855501, 0.001659412043, 0.001620250311, 0.001693878156],
						[0.00363012853, 0.003328464186, 0.003341274835, 0.003345952371, 0.00311780955, 0.003176878082, 0.003103820812, 0.003078967411, 0.002767666317, 0.002825546864, 0.003235475906, 0.003054099004, 0.002915103266, 0.002423383822, 0.002613713399, 0.00266264615, 0.002350060347, 0.002485791671, 0.002394934838, 0.002894221711, 0.003147419208, 0.002679971527, 0.002573591133, 0.002697692301, 0.002420330457, 0.002492862527, 0.002315693113, 0.002509079189, 0.002420467153, 0.002688962135, 0.002940465183, 0.002477946386, 0.002635540325, 0.002577714692, 0.002256129179, 0.002289770099, 0.002403651445, 0.002585858455, 0.00253544555, 0.002491761963, 0.002702254236, 0.002568549821, 0.002301293365, 0.002340550062, 0.002103647344, 0.002006869365, 0.001983430368, 0.002017001486, 0.002145451789, 0.002531211976, 0.002653443198, 0.002399420782, 0.002150288726, 0.002176605807, 0.002159345857, 0.001969783828, 0.001983946335, 0.002136173418, 0.002299743107, 0.002245431262, 0.002152025938, 0.00202899222, 0.001868377558, 0.001880645303, 0.001805562653, 0.001860821108, 0.001986037911, 0.002040183949, 0.002325492097, 0.001988182248, 0.001757660917, 0.001900188593, 0.001913647432, 0.001699751327, 0.001726585286, 0.001847907701, 0.001886751418, 0.002194595047, 0.002489518257, 0.001822208334, 0.001510692335, 0.001534000403, 0.001554887457, 0.001469200418, 0.001768324436, 0.001737591391, 0.001611170221, 0.001913410436, 0.001930658194, 0.001297908725, 0.001312630625, 0.001291381265, 0.001203881102, 0.001290540791, 0.001357172477, 0.001286333176],
						[0.004303746538, 0.003895523432, 0.00395494306, 0.003877361097, 0.003726171349, 0.003775985543, 0.003638987516, 0.003631111805, 0.003242079448, 0.003317706912, 0.003778763194, 0.003663946352, 0.00347520935, 0.002985284814, 0.002927980931, 0.003115850747, 0.002642787593, 0.00294973082, 0.002872689712, 0.003341524005, 0.003727981077, 0.003145197477, 0.002974747996, 0.00306046926, 0.002918191526, 0.003131702677, 0.002736115921, 0.003004795826, 0.002915252933, 0.003248114478, 0.003674445074, 0.003101819533, 0.00313070882, 0.003126073595, 0.002623905379, 0.002681355027, 0.00292431403, 0.00253544555, 0.003536369603, 0.003104430334, 0.0032369198, 0.003232893531, 0.002606564187, 0.002637783682, 0.002598238767, 0.002373926728, 0.002323770748, 0.002325965117, 0.002503271213, 0.002921510826, 0.003188651794, 0.002843995701, 0.002491374206, 0.002673806675, 0.002571918771, 0.002379817153, 0.002362586563, 0.002660356607, 0.002844804395, 0.002827361159, 0.002658962851, 0.002517867529, 0.0022987776, 0.002230930061, 0.002179070837, 0.002236853026, 0.002385015081, 0.002537127192, 0.002868935179, 0.002362173787, 0.002143727368, 0.002323640337, 0.002270978414, 0.001988655607, 0.002219449236, 0.002179725325, 0.002221286989, 0.002421063393, 0.003063830896, 0.002320725035, 0.001971615854, 0.001932172136, 0.001822227792, 0.001709856006, 0.00215968941, 0.002086821575, 0.001997052557, 0.002346739526, 0.002276102439, 0.001649338559, 0.001713164575, 0.001690114187, 0.00157166069, 0.001657481338, 0.001697401999, 0.001620050265],
						[0.004208510197, 0.003933823625, 0.004053391677, 0.003927757425, 0.003807930121, 0.003852669217, 0.003674315677, 0.003705470881, 0.003357423284, 0.003465333124, 0.00390359752, 0.003692743513, 0.003564945967, 0.003095657796, 0.003180273915, 0.003073912489, 0.002800009414, 0.002929667055, 0.002960508812, 0.003697354707, 0.003698115635, 0.003229609154, 0.003097156618, 0.003209536799, 0.002943742574, 0.003070495267, 0.00279314337, 0.003058055976, 0.002969772584, 0.003503084115, 0.003546939204, 0.003145192983, 0.003219439668, 0.00308664757, 0.002786969656, 0.002807109181, 0.002867665129, 0.002491761963, 0.003104430334, 0.003862731071, 0.003180756389, 0.003252099621, 0.00273071594, 0.002610622681, 0.002728077169, 0.002511774866, 0.002401525174, 0.002438591233, 0.002548323656, 0.003072985728, 0.003316450463, 0.002800158523, 0.002509262718, 0.002616284985, 0.002480641143, 0.002301242311, 0.002420236356, 0.002662738394, 0.003062524055, 0.003058764787, 0.002719144755, 0.002530666059, 0.002268505579, 0.00224192054, 0.002096645124, 0.002253077345, 0.00241413202, 0.002519864315, 0.002964938217, 0.002489035842, 0.002126379973, 0.002259785774, 0.002215060521, 0.001987783614, 0.002186204378, 0.002162518963, 0.002429174419, 0.002663709273, 0.003367061096, 0.002292200105, 0.001857039652, 0.001928067481, 0.001756471334, 0.001761635758, 0.002084969638, 0.002156477303, 0.002065400695, 0.002442677105, 0.00243369268, 0.001676106267, 0.001761323688, 0.001715023211, 0.001546563915, 0.001547746646, 0.0016355327, 0.001723242545],
						[0.00500884843, 0.00454823586, 0.0045785797, 0.004493401784, 0.004239428375, 0.004223583686, 0.003997987643, 0.004051691901, 0.003561854758, 0.003620403799, 0.005197194723, 0.004504288936, 0.004115633012, 0.003702221795, 0.003624700079, 0.003631119945, 0.003194270936, 0.003251648963, 0.002908189163, 0.003778983674, 0.0048289164, 0.004040609008, 0.003585739896, 0.003709842156, 0.003517684979, 0.003482263014, 0.002947256641, 0.003248632596, 0.003252968055, 0.003305595067, 0.004832200731, 0.004121189692, 0.003833987856, 0.003817000885, 0.003148736152, 0.003251010168, 0.003037304181, 0.002702254236, 0.0032369198, 0.003180756389, 0.004867089328, 0.003781505789, 0.003375505154, 0.00328495968, 0.003029895698, 0.00294335548, 0.002422654082, 0.002583255508, 0.002618984466, 0.003097829045, 0.004262926982, 0.003772977814, 0.003133892725, 0.003135532692, 0.002847280697, 0.002614391589, 0.002467966524, 0.00279316539, 0.002985339315, 0.003151541273, 0.003871364107, 0.003256375384, 0.002828057206, 0.002597729241, 0.002459102341, 0.002355242093, 0.002675446452, 0.002647022174, 0.00282064572, 0.003334871745, 0.002860552143, 0.002801038174, 0.002492115849, 0.002183844354, 0.002363437762, 0.002239527586, 0.002268189282, 0.00267658867, 0.00312213084, 0.00295891475, 0.00243252121, 0.002317128886, 0.001976554273, 0.001845401642, 0.00234040907, 0.002079558915, 0.001874360924, 0.002396075867, 0.002140531648, 0.002277918698, 0.002093178431, 0.002042077396, 0.001717972348, 0.001779658493, 0.001629050738, 0.001584058446],
						[0.004497613403, 0.004173927655, 0.004162307351, 0.004028057711, 0.003945152337, 0.003867575054, 0.00374307051, 0.003758263597, 0.003334818137, 0.00336627656, 0.00435597295, 0.004016010617, 0.003837504568, 0.003339200871, 0.003335812529, 0.003343044549, 0.002960725557, 0.003135708876, 0.002926999021, 0.003473167568, 0.004370629445, 0.003625382924, 0.003359651379, 0.003330488186, 0.003152410878, 0.003260850024, 0.002949088362, 0.003158675801, 0.003136741139, 0.003361283934, 0.004167746276, 0.003653686039, 0.003641186078, 0.003431157463, 0.002983505359, 0.002906715171, 0.003101957178, 0.002568549821, 0.003232893531, 0.003252099621, 0.003781505789, 0.004323497568, 0.00299665323, 0.003046330957, 0.00307135378, 0.002827228213, 0.002427048034, 0.002554118485, 0.002790492181, 0.003161271791, 0.003790943132, 0.003395870217, 0.002897485451, 0.003176837063, 0.002927014815, 0.002613668375, 0.002481251664, 0.002773992033, 0.002958353609, 0.002879259816, 0.003210875327, 0.003026247196, 0.002743137921, 0.002572044672, 0.002498455262, 0.002438483166, 0.002639653815, 0.002611863186, 0.00311687834, 0.002884994166, 0.00258528776, 0.002802613883, 0.002626465312, 0.002368972886, 0.002417322856, 0.002343205445, 0.002340431675, 0.002674308167, 0.003246847673, 0.002728429499, 0.002414224911, 0.002314975858, 0.002088830654, 0.002000047114, 0.002459099058, 0.00217047674, 0.00203353407, 0.002300070477, 0.002407624321, 0.002036565539, 0.002083728244, 0.002002175109, 0.001875702655, 0.001848791536, 0.001887412237, 0.001908320189],
						[0.003925394458, 0.00350670524, 0.003616132532, 0.003580285672, 0.003383915216, 0.003372837448, 0.003281052417, 0.003219548501, 0.002942575524, 0.002999951399, 0.003811533041, 0.003455978013, 0.003139800334, 0.002930283919, 0.002856828165, 0.00289164188, 0.002526066962, 0.002704731628, 0.002450992361, 0.003193445134, 0.003732576263, 0.003093686213, 0.002702284528, 0.002923999009, 0.002764708831, 0.002760310073, 0.002438530769, 0.00260980252, 0.002526157358, 0.002744912459, 0.003702399965, 0.003180433981, 0.002890780361, 0.002981412052, 0.002586300099, 0.002690006606, 0.002501104067, 0.002301293365, 0.002606564187, 0.00273071594, 0.003375505154, 0.00299665323, 0.003236821869, 0.002687041388, 0.002555955639, 0.002407418934, 0.002158106374, 0.002110355545, 0.00219950667, 0.00265617112, 0.003335814834, 0.002947628681, 0.002472716885, 0.002548346256, 0.002319075618, 0.00211918581, 0.002132957306, 0.002253955311, 0.002451245485, 0.002538588326, 0.002866070977, 0.002630237528, 0.002252036354, 0.002199742604, 0.002015571044, 0.001971782777, 0.002144784117, 0.002030323869, 0.002408010983, 0.002520117346, 0.002221175975, 0.002228063953, 0.002056039595, 0.001878906302, 0.00193734829, 0.001977574331, 0.001972020328, 0.002174230852, 0.002690242371, 0.002311024277, 0.001885365619, 0.00180390305, 0.001602129373, 0.001579517001, 0.001933646242, 0.001659412079, 0.001606100416, 0.001862895769, 0.001994029146, 0.001724103763, 0.001573455973, 0.001526185446, 0.001367473141, 0.001388784454, 0.001310831026, 0.001294276419],
						[0.003826366745, 0.003414971713, 0.003500002623, 0.00341839726, 0.003340113351, 0.003276297438, 0.003160486446, 0.00322410537, 0.002935013377, 0.002888635953, 0.003647806326, 0.003492681159, 0.003134726059, 0.002769878876, 0.002905681077, 0.002883360707, 0.002600968519, 0.002625437784, 0.002432798607, 0.003074032778, 0.003622061423, 0.003046358195, 0.00280620731, 0.003021807234, 0.002673637997, 0.00289823961, 0.002494613245, 0.002709653171, 0.002702039418, 0.002780883803, 0.003501187289, 0.003098678454, 0.00295686402, 0.002907614761, 0.002513571863, 0.002534592074, 0.002572921285, 0.002340550062, 0.002637783682, 0.002610622681, 0.00328495968, 0.003046330957, 0.002687041388, 0.003172722247, 0.002703964364, 0.002380647836, 0.002275645923, 0.002110789127, 0.002225447566, 0.002748700439, 0.003147833761, 0.002897016079, 0.002525782909, 0.002544187618, 0.002423581281, 0.002150737582, 0.002264520511, 0.002413970608, 0.002546536325, 0.002455990775, 0.00277449007, 0.002477180466, 0.002369015847, 0.002247473907, 0.00215330362, 0.002032713052, 0.002175111753, 0.002230589291, 0.002510170656, 0.002475605517, 0.002287571624, 0.002359078671, 0.002171076485, 0.00200555453, 0.00202633518, 0.00214252625, 0.001987810504, 0.002343696514, 0.002688894305, 0.00223512425, 0.001907502635, 0.001900859149, 0.001811115973, 0.001725504202, 0.002020156092, 0.001792794597, 0.001736940424, 0.002063233178, 0.002008129068, 0.001679055023, 0.001674292826, 0.001687232765, 0.001571458355, 0.001584612812, 0.001583883056, 0.001552290241],
						[0.003479102582, 0.003313012908, 0.00324854751, 0.003306099986, 0.003251743842, 0.003134634966, 0.003002804477, 0.003053520724, 0.002794287977, 0.002803079025, 0.003519599805, 0.003262636438, 0.003276090681, 0.002819545433, 0.002759159271, 0.002722695521, 0.002373548708, 0.002637326282, 0.00239761976, 0.003001363495, 0.003367869088, 0.002982998964, 0.002800027585, 0.002858419046, 0.002574437589, 0.002656191464, 0.002427808488, 0.002696348858, 0.002609354892, 0.002735180879, 0.003368893116, 0.002956622019, 0.003012434425, 0.002769647323, 0.002510273881, 0.002472799134, 0.002675836699, 0.002103647344, 0.002598238767, 0.002728077169, 0.003029895698, 0.00307135378, 0.002555955639, 0.002703964364, 0.003326836059, 0.002469461262, 0.002218134497, 0.002089632921, 0.002262581838, 0.002719719359, 0.00317050846, 0.002873997188, 0.002347548213, 0.002657807666, 0.002457468729, 0.002198946237, 0.002201418296, 0.002347567932, 0.002524656713, 0.00247925764, 0.002722687972, 0.002562334361, 0.002485296261, 0.002264635978, 0.002122171843, 0.002073310735, 0.002203683292, 0.00230487603, 0.00248902096, 0.002451727614, 0.002269753022, 0.002469453739, 0.002207256383, 0.002010964229, 0.002105815658, 0.00210722721, 0.002005792894, 0.002344255713, 0.002866565122, 0.002265267561, 0.002038896069, 0.001958905783, 0.001872698697, 0.001695447636, 0.00206957304, 0.00182965049, 0.001730261696, 0.002064534697, 0.001966684595, 0.001747999998, 0.001856325133, 0.001757823844, 0.001656088474, 0.001698686823, 0.001684434505, 0.00171185129],
						[0.003217506668, 0.003017905708, 0.003014847819, 0.003046845036, 0.002902764019, 0.002941816614, 0.002773694236, 0.00282870831, 0.002568939824, 0.002559037881, 0.003250822993, 0.003004556681, 0.002948850289, 0.002524422025, 0.002564853416, 0.002568207708, 0.00233798433, 0.002359066446, 0.002189184483, 0.002813582463, 0.003209881211, 0.002713429037, 0.002587252992, 0.002557295966, 0.002454666916, 0.002533025109, 0.0023698177, 0.002456309315, 0.002335447056, 0.002597800854, 0.003111353431, 0.002822098896, 0.002712519187, 0.002650533834, 0.002352049805, 0.002380507514, 0.002472353739, 0.002006869365, 0.002373926728, 0.002511774866, 0.00294335548, 0.002827228213, 0.002407418934, 0.002380647836, 0.002469461262, 0.00273628785, 0.002050882213, 0.002019935754, 0.002077264058, 0.002482654811, 0.002798265508, 0.002574299859, 0.002235745093, 0.002374428454, 0.002322647411, 0.002029250779, 0.00195440405, 0.002153424168, 0.002335183798, 0.002326288792, 0.002413520365, 0.002282572339, 0.00210209921, 0.002039529681, 0.002037678444, 0.001915372075, 0.002023075094, 0.002057182916, 0.00232234439, 0.002193428133, 0.002100723821, 0.002191852187, 0.002013411673, 0.001817086302, 0.001956677058, 0.001922936359, 0.001926255715, 0.002134125577, 0.002562093626, 0.002117743289, 0.001801102602, 0.001774330601, 0.001647109971, 0.001577969709, 0.001900135167, 0.001711745192, 0.001660298011, 0.001848218608, 0.001831724253, 0.001566536254, 0.001659315375, 0.001530303126, 0.001421089374, 0.001542052019, 0.001504801496, 0.00145597054],
						[0.003092539967, 0.002835420141, 0.002964497291, 0.002896100679, 0.002782632546, 0.002837975836, 0.002668950943, 0.002747014377, 0.002487098668, 0.002522217349, 0.002851594993, 0.002798426204, 0.002638145651, 0.002300083628, 0.002352480996, 0.002351255139, 0.002110325803, 0.002255987615, 0.002227756109, 0.002728806664, 0.002915577328, 0.00249017059, 0.00237875066, 0.002518513222, 0.002211056038, 0.002453741056, 0.002180214972, 0.002423594261, 0.002329560571, 0.002580293557, 0.002770718558, 0.002300704258, 0.002492592931, 0.002431642779, 0.002093244918, 0.002121420707, 0.002332287125, 0.001983430368, 0.002323770748, 0.002401525174, 0.002422654082, 0.002427048034, 0.002158106374, 0.002275645923, 0.002218134497, 0.002050882213, 0.002292387056, 0.001877801099, 0.001959666183, 0.002394099937, 0.002641668335, 0.002242256865, 0.001981723824, 0.002117546945, 0.002090233351, 0.001864608629, 0.001948433257, 0.002073591892, 0.00233437301, 0.002300474881, 0.002038243146, 0.002012237245, 0.001985023885, 0.001852543988, 0.001776673135, 0.001841625137, 0.00188748672, 0.001934718146, 0.002294160273, 0.001867289702, 0.001752422095, 0.001966253632, 0.001836744414, 0.00163411733, 0.001779050864, 0.001877644921, 0.001820894591, 0.002145394252, 0.002643949387, 0.001795900975, 0.001530803387, 0.001533680315, 0.001455476451, 0.001490387539, 0.001746800454, 0.001655796305, 0.001665854964, 0.001962286642, 0.001977779456, 0.001304139577, 0.001386751283, 0.001319070256, 0.001340951321, 0.00135730855, 0.001383488268, 0.001367615045],
						[0.003387487433, 0.00304936103, 0.003052088947, 0.00306826418, 0.002923215582, 0.002966669732, 0.002792739793, 0.002873165543, 0.002538668543, 0.002613303501, 0.003133866747, 0.002855994378, 0.002842945493, 0.002393712985, 0.002384606717, 0.002444673168, 0.002179840156, 0.002265448903, 0.002331667699, 0.002756412871, 0.003030936925, 0.002583465192, 0.002318001535, 0.00260775945, 0.002238484936, 0.002358155389, 0.002209502928, 0.002423369531, 0.002357951196, 0.002562534438, 0.002793733171, 0.002436448376, 0.002536415273, 0.002420464412, 0.002053063802, 0.002153921171, 0.002256236355, 0.002017001486, 0.002325965117, 0.002438591233, 0.002583255508, 0.002554118485, 0.002110355545, 0.002110789127, 0.002089632921, 0.002019935754, 0.001877801099, 0.002255582718, 0.002068414027, 0.002404775849, 0.002593694069, 0.002293415019, 0.001993140333, 0.002173903779, 0.002054921614, 0.001893534002, 0.001927413423, 0.002076444179, 0.002311205946, 0.002274966194, 0.002120337891, 0.001995792556, 0.001938888603, 0.001855814132, 0.001722785106, 0.001832567826, 0.001856676975, 0.002010200721, 0.002260962916, 0.001973444816, 0.001803630977, 0.001894799551, 0.001817529201, 0.001639164356, 0.00178368941, 0.001739598415, 0.001934769853, 0.002056383569, 0.002448597599, 0.001835817712, 0.001554517899, 0.001528020481, 0.001494005094, 0.001475744626, 0.001733468366, 0.001701345065, 0.001608682258, 0.001935430794, 0.001909424209, 0.001308641757, 0.001388652916, 0.001359905258, 0.001246940377, 0.001404625531, 0.001371462861, 0.001372454729],
						[0.003395715859, 0.003091043629, 0.003176130571, 0.003125160756, 0.003007653059, 0.00306358095, 0.0028610154, 0.002847918425, 0.002648547527, 0.002654684242, 0.003116543132, 0.002949423992, 0.003000883638, 0.002549847288, 0.002531822685, 0.002568367833, 0.002177178217, 0.002420139337, 0.002421487646, 0.002729660122, 0.003163595894, 0.002667415222, 0.002577506406, 0.002569833295, 0.002225958147, 0.002498554716, 0.002311303426, 0.002442380546, 0.00246716127, 0.002716587516, 0.002945230267, 0.002488366376, 0.002606521851, 0.002619850514, 0.002270429913, 0.002199153858, 0.002316121493, 0.002145451789, 0.002503271213, 0.002548323656, 0.002618984466, 0.002790492181, 0.00219950667, 0.002225447566, 0.002262581838, 0.002077264058, 0.001959666183, 0.002068414027, 0.002620710447, 0.002603254346, 0.002640121329, 0.002404387829, 0.002117499327, 0.00231447102, 0.002240947952, 0.002012175463, 0.001954114917, 0.002216503475, 0.002286007278, 0.002303013543, 0.002179407599, 0.002098648833, 0.002033042901, 0.001928023816, 0.001856704911, 0.001840407715, 0.002022820615, 0.002108084727, 0.002461211042, 0.002013526607, 0.001899826248, 0.00206525987, 0.00192052398, 0.001766285481, 0.001905770945, 0.00183325678, 0.001966713259, 0.002198692526, 0.002631403393, 0.001867927429, 0.001659350044, 0.001623427959, 0.001643625112, 0.001512696411, 0.001903089234, 0.001771103512, 0.001686312902, 0.001977935665, 0.001958590849, 0.001336237643, 0.001430388361, 0.001420022489, 0.001365864643, 0.00148274205, 0.001519772785, 0.001519431052],
						[0.004000762357, 0.00371479407, 0.003789134736, 0.003719892798, 0.003567505347, 0.003607251209, 0.003470426784, 0.003448010781, 0.003152359262, 0.003261252232, 0.003691230013, 0.003455392502, 0.003455188678, 0.002984198221, 0.002957736965, 0.003064479536, 0.002647144259, 0.002924262235, 0.002875366694, 0.003574148845, 0.003710364719, 0.002995624971, 0.003046000888, 0.003149431155, 0.002883740895, 0.002957772661, 0.002783458454, 0.003036829971, 0.002996461714, 0.003329739, 0.003532627186, 0.002998644227, 0.003156837627, 0.002904114795, 0.002740250691, 0.00274069719, 0.002808901081, 0.002531211976, 0.002921510826, 0.003072985728, 0.003097829045, 0.003161271791, 0.00265617112, 0.002748700439, 0.002719719359, 0.002482654811, 0.002394099937, 0.002404775849, 0.002603254346, 0.003507874446, 0.003261877243, 0.002892954422, 0.002485275888, 0.002805093691, 0.002547038788, 0.002374874245, 0.00248914718, 0.002706362721, 0.00290711498, 0.002931542597, 0.002570636166, 0.002504877642, 0.00234094259, 0.002328788354, 0.002196628596, 0.002275081176, 0.002412950967, 0.002613051957, 0.00293017241, 0.002370038322, 0.002156137142, 0.002405012624, 0.002390016422, 0.002112982315, 0.002240209112, 0.002241507991, 0.002418662726, 0.002577853902, 0.003224991981, 0.002298418854, 0.001922046187, 0.001935690547, 0.001934885145, 0.001863184141, 0.002231743966, 0.002133414729, 0.002105693351, 0.002335114657, 0.002455373533, 0.00165968462, 0.001753961274, 0.001731624546, 0.0017056357, 0.001694330807, 0.001752820866, 0.001814234354],
						[0.004921900893, 0.004518149563, 0.004571681602, 0.004420189562, 0.004212690159, 0.004047749529, 0.003914053963, 0.003931554582, 0.003521043502, 0.003583525581, 0.004951209144, 0.004442682175, 0.00405233259, 0.003592822144, 0.003428895014, 0.003569656968, 0.002943346366, 0.003350293021, 0.00298844369, 0.003742135969, 0.004846839608, 0.004059070863, 0.003738788513, 0.003752861444, 0.00337803957, 0.003434297459, 0.002905864464, 0.003284227606, 0.003250542802, 0.003374641407, 0.004823624868, 0.003872081124, 0.003945092908, 0.003772310698, 0.00312637474, 0.003199500798, 0.003053367321, 0.002653443198, 0.003188651794, 0.003316450463, 0.004262926982, 0.003790943132, 0.003335814834, 0.003147833761, 0.00317050846, 0.002798265508, 0.002641668335, 0.002593694069, 0.002640121329, 0.003261877243, 0.005037470377, 0.003785148817, 0.003021184311, 0.003144722717, 0.002826997808, 0.002607740669, 0.002574420361, 0.002691890067, 0.003095029334, 0.003187072618, 0.003814743727, 0.003287373498, 0.002838411383, 0.002605977434, 0.002375888774, 0.002473344017, 0.002512097295, 0.002746438852, 0.002892256126, 0.003257937869, 0.002711491509, 0.002887254079, 0.002549283435, 0.002204302719, 0.00230289056, 0.002175761723, 0.002439105183, 0.002668846883, 0.003424183921, 0.002966632996, 0.002436055038, 0.002355746166, 0.002025684896, 0.001889482664, 0.002335763509, 0.002037849418, 0.002005744056, 0.002458236182, 0.002416220994, 0.002355888284, 0.002182515245, 0.002002751956, 0.001770018254, 0.001689529892, 0.001693807964, 0.001633407093],
						[0.004177801985, 0.003944750725, 0.003864248166, 0.003735997127, 0.003554662776, 0.003518512219, 0.003418586327, 0.003313178608, 0.003050485455, 0.003009493202, 0.004333963728, 0.003868226071, 0.003689967188, 0.00315543466, 0.003150561594, 0.003122008957, 0.002783759962, 0.002951902598, 0.002606981108, 0.003261816697, 0.004181164132, 0.003637699749, 0.003156040757, 0.003185902301, 0.003022747245, 0.003012799755, 0.002661201218, 0.0029268648, 0.002808323347, 0.002900609376, 0.004206178584, 0.003577494119, 0.003427932124, 0.003295540396, 0.002822365865, 0.002891564344, 0.002762400551, 0.002399420782, 0.002843995701, 0.002800158523, 0.003772977814, 0.003395870217, 0.002947628681, 0.002897016079, 0.002873997188, 0.002574299859, 0.002242256865, 0.002293415019, 0.002404387829, 0.002892954422, 0.003785148817, 0.003795038486, 0.00274344516, 0.002846294208, 0.002643327318, 0.002373651119, 0.002450064418, 0.002525335517, 0.002638444729, 0.002732705094, 0.003402279329, 0.002927043135, 0.002580653227, 0.002380309535, 0.002335352466, 0.002231337129, 0.0024369971, 0.002456676516, 0.002660254894, 0.002915525926, 0.002531438179, 0.002613211472, 0.002426121036, 0.002052740998, 0.002158481331, 0.0021140707, 0.002201615384, 0.00245250601, 0.002904037504, 0.002652139789, 0.002216415693, 0.002137932195, 0.001988579044, 0.001773262991, 0.002244430763, 0.001871709159, 0.001750546507, 0.002199654503, 0.002098263899, 0.002135683637, 0.002002634763, 0.001942345413, 0.001702335031, 0.001709835884, 0.001653003911, 0.001587142632],
						[0.003654297707, 0.003314415963, 0.003371414581, 0.003210599149, 0.003178857495, 0.003220634491, 0.003024243577, 0.003028926938, 0.00263296343, 0.002728739572, 0.003606982422, 0.003236383866, 0.002976667966, 0.002677606164, 0.00263577638, 0.00274180904, 0.00247288644, 0.002461926978, 0.002317270052, 0.002881400693, 0.003579984805, 0.002998394955, 0.002682901711, 0.002801078981, 0.002552899392, 0.002586533688, 0.002421328055, 0.002456698341, 0.002478598141, 0.002693202523, 0.003434773421, 0.003077587403, 0.002807858163, 0.002766997167, 0.002439855205, 0.002373193014, 0.002330442759, 0.002150288726, 0.002491374206, 0.002509262718, 0.003133892725, 0.002897485451, 0.002472716885, 0.002525782909, 0.002347548213, 0.002235745093, 0.001981723824, 0.001993140333, 0.002117499327, 0.002485275888, 0.003021184311, 0.00274344516, 0.002907841859, 0.002360326866, 0.002201432646, 0.00197900765, 0.002028213155, 0.002182706382, 0.002323935946, 0.002457778017, 0.002650656099, 0.00229871417, 0.002193479256, 0.002106481104, 0.001988621986, 0.001907748692, 0.002067338217, 0.002054173723, 0.002371889688, 0.002318416621, 0.002158656476, 0.002257302626, 0.001995026144, 0.001761774637, 0.001884896651, 0.001870519581, 0.001795517347, 0.002097052813, 0.002458630052, 0.00220955019, 0.001801943668, 0.001859650234, 0.00158182962, 0.00155812396, 0.001802007614, 0.00162487014, 0.001567712893, 0.001873815258, 0.001786576186, 0.001529337994, 0.001611606875, 0.001470394496, 0.001221941677, 0.001405332314, 0.001389306134, 0.001394237652],
						[0.003598637829, 0.00343389213, 0.003366216157, 0.003358889658, 0.003158288587, 0.003225088129, 0.003035802487, 0.003032511577, 0.002732898344, 0.00276611704, 0.003729182959, 0.003332541527, 0.00317508292, 0.002848600657, 0.002772655784, 0.002790751101, 0.002423331136, 0.002655913987, 0.002405332647, 0.00295569575, 0.003684863634, 0.003056517742, 0.002738199901, 0.002773726189, 0.002698310427, 0.002659368438, 0.00247621901, 0.002705616909, 0.002593966541, 0.002619729093, 0.003554204283, 0.003050500899, 0.003140722747, 0.002774585934, 0.002520047139, 0.002513681059, 0.002681173546, 0.002176605807, 0.002673806675, 0.002616284985, 0.003135532692, 0.003176837063, 0.002548346256, 0.002544187618, 0.002657807666, 0.002374428454, 0.002117546945, 0.002173903779, 0.00231447102, 0.002805093691, 0.003144722717, 0.002846294208, 0.002360326866, 0.003149523875, 0.002415837636, 0.002200086391, 0.002162371878, 0.002324971658, 0.002461091217, 0.002459801104, 0.00268770369, 0.00263931478, 0.002375925292, 0.002159224955, 0.002075736788, 0.002131126522, 0.002320906714, 0.002170903329, 0.002468893541, 0.002472894677, 0.002205323436, 0.002454136129, 0.002299884068, 0.00205816416, 0.002082048619, 0.002101663354, 0.00213781531, 0.002254171392, 0.002674545428, 0.002328507026, 0.002047874052, 0.001979971247, 0.001877199216, 0.001696418522, 0.002103916017, 0.001831609016, 0.001815691681, 0.002105651673, 0.002104577206, 0.001764384174, 0.001754955104, 0.001726516932, 0.00164620921, 0.00164119908, 0.001632717802, 0.00170226126],
						[0.00347358152, 0.003213881416, 0.003231814936, 0.003149259674, 0.003031807835, 0.003074806453, 0.002865340173, 0.002911464346, 0.002639358885, 0.00257947995, 0.003338004342, 0.003093094158, 0.003002616169, 0.002637211277, 0.00257720391, 0.00260473893, 0.002272073889, 0.002527600965, 0.002385423734, 0.002771345251, 0.003303553928, 0.002810718864, 0.002771310142, 0.002783920954, 0.002457512633, 0.002638276807, 0.002356458479, 0.002486510885, 0.002468548448, 0.002685051147, 0.00317215526, 0.002748571156, 0.002845330975, 0.002778439634, 0.002422558995, 0.002319327304, 0.002527505865, 0.002159345857, 0.002571918771, 0.002480641143, 0.002847280697, 0.002927014815, 0.002319075618, 0.002423581281, 0.002457468729, 0.002322647411, 0.002090233351, 0.002054921614, 0.002240947952, 0.002547038788, 0.002826997808, 0.002643327318, 0.002201432646, 0.002415837636, 0.002767315247, 0.002043651733, 0.002034993437, 0.002204876043, 0.00234345839, 0.002419818446, 0.002411671324, 0.002288762897, 0.002288017767, 0.002072733425, 0.002008139257, 0.001982204687, 0.002128501004, 0.002193425371, 0.002539332658, 0.002208228764, 0.002025280393, 0.002246620993, 0.002133769892, 0.001897569735, 0.002027384031, 0.001938074558, 0.001924333968, 0.002256134662, 0.002770495738, 0.00210142722, 0.001832840001, 0.001837111486, 0.001635391534, 0.001627968138, 0.002071366733, 0.00171055106, 0.001684877891, 0.001950284629, 0.002041913208, 0.0014921953, 0.001613936507, 0.001541106991, 0.001471749075, 0.001556784332, 0.001587092448, 0.001462726316],
						[0.003243638023, 0.003034389219, 0.003016512653, 0.002974762136, 0.002869155025, 0.002869951923, 0.002746671583, 0.002745567374, 0.00245577162, 0.002479049424, 0.003044495193, 0.002851603943, 0.002716058382, 0.002388155994, 0.002387000641, 0.002401464726, 0.002157631265, 0.002327256475, 0.002234018339, 0.00263992019, 0.002993021396, 0.002538625765, 0.002361977066, 0.002526022159, 0.0023063138, 0.002365156932, 0.002146446492, 0.002321681675, 0.00228161994, 0.00240451075, 0.002848030125, 0.002444597683, 0.002578218747, 0.002491512264, 0.002079851963, 0.002084102308, 0.002247129268, 0.001969783828, 0.002379817153, 0.002301242311, 0.002614391589, 0.002613668375, 0.00211918581, 0.002150737582, 0.002198946237, 0.002029250779, 0.001864608629, 0.001893534002, 0.002012175463, 0.002374874245, 0.002607740669, 0.002373651119, 0.00197900765, 0.002200086391, 0.002043651733, 0.002258678707, 0.001848273189, 0.002082300372, 0.002138891732, 0.002093305463, 0.002110123955, 0.002090690778, 0.001923976682, 0.001807177388, 0.001846959803, 0.001777795269, 0.001862867801, 0.002014353121, 0.002167454362, 0.00190608691, 0.001789749635, 0.001991950908, 0.001904169139, 0.001637374176, 0.00181615008, 0.001761109708, 0.001886317215, 0.001973870058, 0.002307852321, 0.001854531318, 0.001563540612, 0.001555407534, 0.001528674383, 0.001389151699, 0.001813067579, 0.00165444263, 0.001623472287, 0.00179482723, 0.001723097603, 0.001368238167, 0.001434811152, 0.001338967163, 0.001364436889, 0.001395907181, 0.001442945758, 0.001410140658],
						[0.003030619633, 0.002929474526, 0.002922320365, 0.002953982871, 0.002756593414, 0.002794455413, 0.002713610852, 0.002658174324, 0.00240993133, 0.002497650293, 0.002987117227, 0.002816757742, 0.002865345579, 0.002352838588, 0.002357687016, 0.002339342052, 0.002150761998, 0.002300116212, 0.002249965789, 0.002789369528, 0.003079061485, 0.0024774441, 0.00228325972, 0.002509305232, 0.002281525885, 0.002266873072, 0.002187583781, 0.002466251737, 0.002341879803, 0.002498561864, 0.002739330765, 0.002435739537, 0.002494993998, 0.002358714498, 0.002235407413, 0.00224190194, 0.002221462671, 0.001983946335, 0.002362586563, 0.002420236356, 0.002467966524, 0.002481251664, 0.002132957306, 0.002264520511, 0.002201418296, 0.00195440405, 0.001948433257, 0.001927413423, 0.001954114917, 0.00248914718, 0.002574420361, 0.002450064418, 0.002028213155, 0.002162371878, 0.002034993437, 0.001848273189, 0.002538786963, 0.00222124262, 0.002253509776, 0.002253188285, 0.002299911894, 0.0020225735, 0.001944254201, 0.001863789412, 0.001819545462, 0.001845145489, 0.001922706237, 0.002048916468, 0.00233515705, 0.002032495395, 0.001825373988, 0.001888469975, 0.001866935206, 0.001704091112, 0.001790801096, 0.001876568938, 0.001794484491, 0.002072880281, 0.00253783427, 0.001899653245, 0.001507957861, 0.001604302022, 0.001605942258, 0.001543079407, 0.001825533626, 0.001679398022, 0.001668945698, 0.001919123145, 0.001938805242, 0.001453504753, 0.001421291285, 0.001527167595, 0.001413216884, 0.001448010909, 0.001478431473, 0.001485237838],
						[0.003440298944, 0.00334420473, 0.003274787097, 0.003297600332, 0.003070416077, 0.00314857206, 0.002952582409, 0.003029978077, 0.002745503964, 0.00278677727, 0.003204243619, 0.003150937277, 0.002911171213, 0.002637722498, 0.002638124, 0.002678514626, 0.002429601622, 0.002474336877, 0.002484491958, 0.003027644353, 0.003216678089, 0.00270785449, 0.00259919058, 0.00270304282, 0.002496782109, 0.002727073654, 0.002492489763, 0.002640528653, 0.002538539563, 0.002693658519, 0.003111718767, 0.002681119544, 0.002687851375, 0.00263069605, 0.002373286904, 0.002394257036, 0.002480365542, 0.002136173418, 0.002660356607, 0.002662738394, 0.00279316539, 0.002773992033, 0.002253955311, 0.002413970608, 0.002347567932, 0.002153424168, 0.002073591892, 0.002076444179, 0.002216503475, 0.002706362721, 0.002691890067, 0.002525335517, 0.002182706382, 0.002324971658, 0.002204876043, 0.002082300372, 0.00222124262, 0.002888158837, 0.00246768224, 0.002545452725, 0.002354012316, 0.002171205185, 0.002126558897, 0.002073529808, 0.002061390344, 0.002037461461, 0.00216609843, 0.00225437478, 0.002504191509, 0.002126676968, 0.001914553843, 0.002159707946, 0.002024030435, 0.001879695348, 0.00199321726, 0.002005570746, 0.002129976208, 0.002232949935, 0.002727763283, 0.002024908208, 0.001716670092, 0.001777444928, 0.001822485902, 0.001672337281, 0.00202609524, 0.001939476904, 0.001907641035, 0.002103609887, 0.002081524935, 0.001497014391, 0.001545671162, 0.001640709803, 0.001567264754, 0.001563318738, 0.001658344814, 0.001638997416],
						[0.003931812779, 0.003452367593, 0.003685314298, 0.003627668097, 0.003440003388, 0.003445206259, 0.003384751458, 0.003456979638, 0.003105609008, 0.003201191769, 0.003555873242, 0.003371518839, 0.0033207877, 0.00285052315, 0.002886483586, 0.002819510233, 0.00258608262, 0.002739514386, 0.002751166494, 0.003339619025, 0.003516415998, 0.003038493889, 0.00286215559, 0.002986993227, 0.002708649286, 0.00298885925, 0.002553266905, 0.002961271301, 0.002828931675, 0.003142076927, 0.003134480965, 0.002759195476, 0.002980382082, 0.002811431806, 0.002521135822, 0.00256134582, 0.002739803121, 0.002299743107, 0.002844804395, 0.003062524055, 0.002985339315, 0.002958353609, 0.002451245485, 0.002546536325, 0.002524656713, 0.002335183798, 0.00233437301, 0.002311205946, 0.002286007278, 0.00290711498, 0.003095029334, 0.002638444729, 0.002323935946, 0.002461091217, 0.00234345839, 0.002138891732, 0.002253509776, 0.00246768224, 0.003350583833, 0.002772584344, 0.002545890912, 0.00228588627, 0.002134542835, 0.002105589144, 0.001972762346, 0.002065197164, 0.002134455168, 0.002349684212, 0.002885162927, 0.002390285755, 0.001977518945, 0.002179275575, 0.002159351368, 0.001854111527, 0.002045975845, 0.002050252042, 0.002150942185, 0.002416861729, 0.00308709855, 0.002117514202, 0.001769438899, 0.001741036688, 0.001707515433, 0.001644695146, 0.00200538994, 0.001967585467, 0.001912712957, 0.002272632476, 0.002188558276, 0.001539832924, 0.001644256099, 0.001611616989, 0.001492172521, 0.001583504098, 0.001624889665, 0.001665134061],
						[0.004004083997, 0.003555839981, 0.003723157506, 0.003716680489, 0.003495357866, 0.003484276388, 0.003398783072, 0.003418816348, 0.003080811196, 0.003171901734, 0.003764693424, 0.003421802213, 0.003383236234, 0.002949664632, 0.002862711554, 0.002970506193, 0.002629567419, 0.002810675895, 0.002687481463, 0.003391684513, 0.003563033468, 0.003132510879, 0.002856169711, 0.003066809247, 0.002886268458, 0.002962162727, 0.002613346258, 0.002871738576, 0.002913652699, 0.003097697868, 0.003584368242, 0.003072875462, 0.002966540382, 0.002873848244, 0.002633358523, 0.00252863471, 0.002757168903, 0.002245431262, 0.002827361159, 0.003058764787, 0.003151541273, 0.002879259816, 0.002538588326, 0.002455990775, 0.00247925764, 0.002326288792, 0.002300474881, 0.002274966194, 0.002303013543, 0.002931542597, 0.003187072618, 0.002732705094, 0.002457778017, 0.002459801104, 0.002419818446, 0.002093305463, 0.002253188285, 0.002545452725, 0.002772584344, 0.004079755781, 0.002537002164, 0.002413232926, 0.002321551262, 0.002189272422, 0.002113071136, 0.002099741406, 0.002304073073, 0.002393506739, 0.002693485986, 0.002321912481, 0.002075753427, 0.002323698689, 0.00217477644, 0.00173595849, 0.002037136712, 0.002029942012, 0.002329128467, 0.002497822871, 0.002997682812, 0.002208828426, 0.0018639162, 0.001807356548, 0.001639376038, 0.001659224599, 0.002125126885, 0.001869063588, 0.00195522756, 0.002256425015, 0.002297224771, 0.001523107793, 0.001675158753, 0.001569474671, 0.001440842522, 0.001625692915, 0.001511484483, 0.00162087977],
						[0.00379606579, 0.003597330441, 0.003694396668, 0.003460502947, 0.003285121985, 0.003242065918, 0.003164531892, 0.003089516401, 0.002765467253, 0.002852639794, 0.004334541133, 0.003810412801, 0.003420739267, 0.003039306391, 0.002969202312, 0.002821127856, 0.002489536208, 0.002664617975, 0.002383835896, 0.003009037919, 0.004092782743, 0.003491058531, 0.002929151204, 0.003032034749, 0.002825941231, 0.002724667161, 0.002381112847, 0.002652598675, 0.002624850473, 0.00259576905, 0.004337202597, 0.003471787293, 0.00324307985, 0.003285817904, 0.002729913204, 0.002803873853, 0.00244546032, 0.002152025938, 0.002658962851, 0.002719144755, 0.003871364107, 0.003210875327, 0.002866070977, 0.00277449007, 0.002722687972, 0.002413520365, 0.002038243146, 0.002120337891, 0.002179407599, 0.002570636166, 0.003814743727, 0.003402279329, 0.002650656099, 0.00268770369, 0.002411671324, 0.002110123955, 0.002299911894, 0.002354012316, 0.002545890912, 0.002537002164, 0.004115982628, 0.00285063903, 0.002468721806, 0.002265342652, 0.002177720248, 0.002024486222, 0.002346885696, 0.002171199726, 0.002353119122, 0.003107641537, 0.002544776951, 0.002392017713, 0.0021142444, 0.002042435891, 0.001953488988, 0.00201758249, 0.002020718018, 0.002274620913, 0.00277005577, 0.002698335303, 0.002290962724, 0.002012465449, 0.00185434367, 0.001648297176, 0.002036498587, 0.001800994314, 0.001637370346, 0.001989093522, 0.001834252413, 0.00224975273, 0.001895068104, 0.001898527721, 0.001539200166, 0.001481003002, 0.001471806275, 0.001417212846],
						[0.00338809299, 0.003228763006, 0.003228902136, 0.003201833917, 0.003009518249, 0.00296896478, 0.002893749342, 0.002823881199, 0.002570024624, 0.002556324433, 0.003553878345, 0.003295890373, 0.003050888097, 0.002737754808, 0.002639004767, 0.002624749646, 0.002321161174, 0.002577909445, 0.002218656927, 0.0027456759, 0.003510117601, 0.00292253278, 0.002730446441, 0.002679082237, 0.002597498699, 0.002603957511, 0.002299599325, 0.002495345023, 0.002433160548, 0.002535927788, 0.003567835962, 0.00303231373, 0.003040830314, 0.002866599622, 0.00243657735, 0.002419776366, 0.002450922864, 0.00202899222, 0.002517867529, 0.002530666059, 0.003256375384, 0.003026247196, 0.002630237528, 0.002477180466, 0.002562334361, 0.002282572339, 0.002012237245, 0.001995792556, 0.002098648833, 0.002504877642, 0.003287373498, 0.002927043135, 0.00229871417, 0.00263931478, 0.002288762897, 0.002090690778, 0.0020225735, 0.002171205185, 0.00228588627, 0.002413232926, 0.00285063903, 0.002928884545, 0.002343570116, 0.002190788495, 0.002009566027, 0.001967943614, 0.002072171205, 0.002006980689, 0.002390876113, 0.002517607928, 0.002260702104, 0.002359948953, 0.002099432758, 0.001857530485, 0.001919568355, 0.001893251837, 0.001914514415, 0.002093109033, 0.002478238425, 0.002328326893, 0.002022264844, 0.001956531438, 0.00171365323, 0.001608642565, 0.001983211094, 0.001738711241, 0.001653203216, 0.001914985238, 0.001870729067, 0.00181912707, 0.001775680814, 0.001709574839, 0.001544124909, 0.001523394415, 0.001502743129, 0.001442843411],
						[0.003208075825, 0.002824046695, 0.002950115286, 0.002927183154, 0.002794968285, 0.002867250372, 0.002576146487, 0.002694148822, 0.002403539014, 0.002395752497, 0.003231203484, 0.002979283605, 0.002846153878, 0.002598694665, 0.002423556742, 0.002501434852, 0.00216214058, 0.002402551657, 0.002133343147, 0.002538829406, 0.003131965539, 0.002812505344, 0.002602943627, 0.002584795124, 0.002309029674, 0.00239513157, 0.002138305566, 0.002318905424, 0.002332568494, 0.002366890196, 0.003097652087, 0.002760249885, 0.002747754518, 0.002514428298, 0.002218249685, 0.002180843066, 0.002309054272, 0.001868377558, 0.0022987776, 0.002268505579, 0.002828057206, 0.002743137921, 0.002252036354, 0.002369015847, 0.002485296261, 0.00210209921, 0.001985023885, 0.001938888603, 0.002033042901, 0.00234094259, 0.002838411383, 0.002580653227, 0.002193479256, 0.002375925292, 0.002288017767, 0.001923976682, 0.001944254201, 0.002126558897, 0.002134542835, 0.002321551262, 0.002468721806, 0.002343570116, 0.002667728836, 0.002059014354, 0.001875570377, 0.001815770394, 0.001926379175, 0.001910367164, 0.002259947412, 0.002210713674, 0.002105304897, 0.00227225667, 0.002006248791, 0.001678812607, 0.00184701393, 0.00179998138, 0.001704258851, 0.002055336166, 0.002590616688, 0.002073480187, 0.001874480495, 0.001817398768, 0.00158261204, 0.00160578189, 0.001834642637, 0.001570663998, 0.001592933608, 0.002057573856, 0.001761267425, 0.001508379831, 0.001597609687, 0.001539238963, 0.001439716463, 0.001489612646, 0.001449216874, 0.001435943811],
						[0.002935523324, 0.002719042358, 0.002751010898, 0.002807933105, 0.002665138781, 0.002673288533, 0.002534087762, 0.002581815013, 0.002365928298, 0.002348819196, 0.002943099406, 0.002720910432, 0.002658557789, 0.002320043343, 0.002302659107, 0.002245179492, 0.002093913888, 0.002241335738, 0.002062347094, 0.00246199063, 0.002865249967, 0.002501828039, 0.002393241764, 0.00242750741, 0.002188957955, 0.002366851537, 0.002185859516, 0.0022650826, 0.002191589414, 0.002410639485, 0.002903839306, 0.002524300375, 0.002462205185, 0.002443524259, 0.002148550824, 0.002103332855, 0.002160272368, 0.001880645303, 0.002230930061, 0.00224192054, 0.002597729241, 0.002572044672, 0.002199742604, 0.002247473907, 0.002264635978, 0.002039529681, 0.001852543988, 0.001855814132, 0.001928023816, 0.002328788354, 0.002605977434, 0.002380309535, 0.002106481104, 0.002159224955, 0.002072733425, 0.001807177388, 0.001863789412, 0.002073529808, 0.002105589144, 0.002189272422, 0.002265342652, 0.002190788495, 0.002059014354, 0.002239974825, 0.001860978705, 0.001763013832, 0.001851172919, 0.001813111348, 0.002145930135, 0.002033279861, 0.001984916688, 0.002110346607, 0.001896770696, 0.001675683985, 0.001751742541, 0.001808930336, 0.001737045766, 0.001859046805, 0.002208657034, 0.001916434512, 0.001698026684, 0.001711879501, 0.001552914235, 0.001560586611, 0.001794451241, 0.001666964874, 0.001556821933, 0.001761721129, 0.001674310159, 0.001338224906, 0.001469942223, 0.001372897973, 0.001312547653, 0.001355054415, 0.001392112944, 0.001311553054],
						[0.002579952426, 0.002595143005, 0.002432745554, 0.002393663894, 0.002399572175, 0.002378738807, 0.002298139155, 0.002274152315, 0.002060583994, 0.002024066215, 0.002696793804, 0.002541390204, 0.00248201665, 0.002149391811, 0.002211500634, 0.002192825626, 0.001967676888, 0.002078977645, 0.001937510594, 0.002241391942, 0.002725611012, 0.002316819949, 0.002125213045, 0.002219893885, 0.002129636242, 0.002204672391, 0.002131839155, 0.002093950813, 0.00211624525, 0.002165431495, 0.002862427116, 0.002394184833, 0.00236962802, 0.002274656068, 0.00206209881, 0.002042384937, 0.002095941077, 0.001805562653, 0.002179070837, 0.002096645124, 0.002459102341, 0.002498455262, 0.002015571044, 0.00215330362, 0.002122171843, 0.002037678444, 0.001776673135, 0.001722785106, 0.001856704911, 0.002196628596, 0.002375888774, 0.002335352466, 0.001988621986, 0.002075736788, 0.002008139257, 0.001846959803, 0.001819545462, 0.002061390344, 0.001972762346, 0.002113071136, 0.002177720248, 0.002009566027, 0.001875570377, 0.001860978705, 0.00217574472, 0.001695211453, 0.001905841881, 0.001938044055, 0.001955343717, 0.001947203413, 0.001835683003, 0.00199579587, 0.001878570542, 0.001637405318, 0.001789838079, 0.001715748504, 0.001829044759, 0.001879697817, 0.00210540579, 0.001800355523, 0.0016475462, 0.001614763026, 0.001592853435, 0.001466974956, 0.001730962735, 0.001611239802, 0.001517674255, 0.001656282469, 0.00164398893, 0.001406303153, 0.001505474531, 0.001409622401, 0.001346738712, 0.001380300963, 0.001465920283, 0.001417331467],
						[0.002747787374, 0.00264760704, 0.002705575637, 0.002612167028, 0.002482747036, 0.002476341968, 0.002369198989, 0.002452890052, 0.002128802882, 0.002145725437, 0.002640047186, 0.002527176121, 0.002494794761, 0.002141692279, 0.002164797252, 0.002164558022, 0.001976377986, 0.002098506638, 0.001986508049, 0.002453759042, 0.002733212066, 0.002253495925, 0.002038210391, 0.002286300568, 0.002074964307, 0.002124099997, 0.002043831445, 0.002171420609, 0.00214557857, 0.002328296499, 0.002742707281, 0.002314679862, 0.002326459033, 0.00221991396, 0.002026451581, 0.001968265032, 0.002156213221, 0.001860821108, 0.002236853026, 0.002253077345, 0.002355242093, 0.002438483166, 0.001971782777, 0.002032713052, 0.002073310735, 0.001915372075, 0.001841625137, 0.001832567826, 0.001840407715, 0.002275081176, 0.002473344017, 0.002231337129, 0.001907748692, 0.002131126522, 0.001982204687, 0.001777795269, 0.001845145489, 0.002037461461, 0.002065197164, 0.002099741406, 0.002024486222, 0.001967943614, 0.001815770394, 0.001763013832, 0.001695211453, 0.002226843776, 0.001831680806, 0.001941214445, 0.002092000333, 0.001879618498, 0.001749958111, 0.001914182171, 0.001800106007, 0.001640827494, 0.001703409595, 0.00178709596, 0.001800410688, 0.001956771025, 0.002342896959, 0.001864830712, 0.001542859377, 0.00157997703, 0.001570629667, 0.001497027997, 0.001760054456, 0.001638850111, 0.001596252795, 0.001828888462, 0.001783426135, 0.001414236235, 0.001439126627, 0.001421393745, 0.001257311173, 0.001417501875, 0.001434609122, 0.001413057928],
						[0.002993945136, 0.002929607457, 0.002880348639, 0.002891811709, 0.002766082028, 0.00278931335, 0.002650016671, 0.002639344627, 0.002378697448, 0.002443764267, 0.003154011213, 0.002777588485, 0.002796370631, 0.002455331697, 0.002527220919, 0.002394731624, 0.00212531014, 0.002261782071, 0.002164657603, 0.002658168093, 0.002940756862, 0.002516935413, 0.002278688507, 0.002416626155, 0.002305354962, 0.002254771995, 0.002271061995, 0.002351905388, 0.002334825057, 0.00245450379, 0.00303943435, 0.002563165524, 0.002444427051, 0.002444334685, 0.0021809998, 0.002240327534, 0.002307032847, 0.001986037911, 0.002385015081, 0.00241413202, 0.002675446452, 0.002639653815, 0.002144784117, 0.002175111753, 0.002203683292, 0.002023075094, 0.00188748672, 0.001856676975, 0.002022820615, 0.002412950967, 0.002512097295, 0.0024369971, 0.002067338217, 0.002320906714, 0.002128501004, 0.001862867801, 0.001922706237, 0.00216609843, 0.002134455168, 0.002304073073, 0.002346885696, 0.002072171205, 0.001926379175, 0.001851172919, 0.001905841881, 0.001831680806, 0.002362971986, 0.002013099909, 0.002211800006, 0.002037217978, 0.001839230725, 0.001984006171, 0.001901105125, 0.001751261424, 0.001834798411, 0.001852889709, 0.001891929825, 0.002045778129, 0.002521603449, 0.001912894773, 0.001712256401, 0.001621937332, 0.001611890165, 0.001470146558, 0.001863226423, 0.001699113779, 0.001664173048, 0.001824116362, 0.001925407783, 0.001432696674, 0.001447393032, 0.001488967379, 0.001353134112, 0.001441163181, 0.001423381553, 0.001498832196],
						[0.003157948431, 0.003064194774, 0.002975663554, 0.002789078706, 0.002807242675, 0.002744695175, 0.002596581078, 0.002777070666, 0.002375537224, 0.002445315495, 0.0030178547, 0.00282129975, 0.002825190452, 0.002372667954, 0.002339467906, 0.002495798699, 0.002052453022, 0.002340604819, 0.002328611956, 0.002796176512, 0.002965537976, 0.002451923454, 0.002451203521, 0.002570982108, 0.002290907059, 0.002485193819, 0.00229176301, 0.002521800233, 0.002464649154, 0.002711650031, 0.002918479447, 0.002435194682, 0.002492011945, 0.002513984598, 0.002184221162, 0.002214681019, 0.002425733908, 0.002040183949, 0.002537127192, 0.002519864315, 0.002647022174, 0.002611863186, 0.002030323869, 0.002230589291, 0.00230487603, 0.002057182916, 0.001934718146, 0.002010200721, 0.002108084727, 0.002613051957, 0.002746438852, 0.002456676516, 0.002054173723, 0.002170903329, 0.002193425371, 0.002014353121, 0.002048916468, 0.00225437478, 0.002349684212, 0.002393506739, 0.002171199726, 0.002006980689, 0.001910367164, 0.001813111348, 0.001938044055, 0.001941214445, 0.002013099909, 0.002913229606, 0.002342612304, 0.002031034186, 0.001806357272, 0.001986177557, 0.001948085404, 0.001738131697, 0.001961482042, 0.001842175714, 0.002085970703, 0.002257239072, 0.002693510647, 0.001937579107, 0.001624053584, 0.001595164909, 0.001618440704, 0.001438808775, 0.001889528393, 0.001777879072, 0.001792062365, 0.00216390251, 0.001970141906, 0.001485800171, 0.001609694131, 0.001519896088, 0.001452871784, 0.001512179815, 0.001595801621, 0.00157436436],
						[0.003715079122, 0.003378001969, 0.003502498329, 0.00345071268, 0.003305060105, 0.003419811664, 0.003333011815, 0.003271405217, 0.002979944128, 0.002970631164, 0.003233047781, 0.003289166972, 0.003221488961, 0.002748106064, 0.002896492621, 0.002804875205, 0.002483985727, 0.002765602047, 0.002693200054, 0.003167364769, 0.003303248719, 0.002803321185, 0.002800159684, 0.002968969687, 0.002604838112, 0.002957245289, 0.00263039294, 0.002877079814, 0.002830893306, 0.003238324681, 0.002895614018, 0.002696053478, 0.002907310129, 0.002864591507, 0.002537095063, 0.002462364389, 0.00276051117, 0.002325492097, 0.002868935179, 0.002964938217, 0.00282064572, 0.00311687834, 0.002408010983, 0.002510170656, 0.00248902096, 0.00232234439, 0.002294160273, 0.002260962916, 0.002461211042, 0.00293017241, 0.002892256126, 0.002660254894, 0.002371889688, 0.002468893541, 0.002539332658, 0.002167454362, 0.00233515705, 0.002504191509, 0.002885162927, 0.002693485986, 0.002353119122, 0.002390876113, 0.002259947412, 0.002145930135, 0.001955343717, 0.002092000333, 0.002211800006, 0.002342612304, 0.003658817031, 0.00212495448, 0.001982733737, 0.002274283814, 0.002126321301, 0.001906339992, 0.002098460066, 0.002175184895, 0.002123685539, 0.002434289261, 0.003287733483, 0.002143422865, 0.001787322454, 0.001853956393, 0.001761340961, 0.001761927511, 0.002223767524, 0.00202875252, 0.002015049614, 0.002359241942, 0.002349290251, 0.001502287893, 0.001690429149, 0.001701696033, 0.001637735267, 0.001694576331, 0.001727829921, 0.001714672068],
						[0.003337110603, 0.003206598288, 0.00322837047, 0.003015226454, 0.002907239952, 0.002865320479, 0.002746955289, 0.002731006516, 0.002466248967, 0.002471499072, 0.003705899965, 0.003287562718, 0.002953910851, 0.002744602577, 0.002636127797, 0.002463355151, 0.00221991072, 0.002390797271, 0.002124462034, 0.002702996108, 0.003620976929, 0.003086465809, 0.002582754777, 0.002657204096, 0.002458205216, 0.002441650783, 0.00211106795, 0.002384031672, 0.002365428219, 0.002348861861, 0.003558485885, 0.003113548988, 0.002862370156, 0.002849514422, 0.002429949287, 0.002424836448, 0.002288477239, 0.001988182248, 0.002362173787, 0.002489035842, 0.003334871745, 0.002884994166, 0.002520117346, 0.002475605517, 0.002451727614, 0.002193428133, 0.001867289702, 0.001973444816, 0.002013526607, 0.002370038322, 0.003257937869, 0.002915525926, 0.002318416621, 0.002472894677, 0.002208228764, 0.00190608691, 0.002032495395, 0.002126676968, 0.002390285755, 0.002321912481, 0.003107641537, 0.002517607928, 0.002210713674, 0.002033279861, 0.001947203413, 0.001879618498, 0.002037217978, 0.002031034186, 0.00212495448, 0.002986835437, 0.002274080231, 0.002230011678, 0.002045158491, 0.001799474408, 0.001782427378, 0.00181265829, 0.001825797102, 0.002054788709, 0.002457843811, 0.002349393317, 0.00199023081, 0.001857663227, 0.001724921252, 0.001551595006, 0.001880673805, 0.001614511409, 0.001502270361, 0.001976858326, 0.001714422288, 0.00193244399, 0.00173933189, 0.001693475392, 0.0014430677, 0.001442626283, 0.001374532828, 0.001388198613],
						[0.002848833667, 0.002600477361, 0.002724240891, 0.002634119186, 0.002521817319, 0.00253830332, 0.0024088781, 0.002417733311, 0.002183305605, 0.00217780669, 0.003057049188, 0.00280330556, 0.002647777308, 0.002357861268, 0.002371647666, 0.002245615832, 0.002045275133, 0.002148585308, 0.001949411141, 0.002418751103, 0.003034990502, 0.002570900195, 0.002317755746, 0.002334412356, 0.002175005627, 0.002273750322, 0.002025041307, 0.002151030603, 0.002215903769, 0.002259480343, 0.003049905475, 0.002701328888, 0.002476277853, 0.002484707529, 0.002107394936, 0.002049107198, 0.002059309865, 0.001757660917, 0.002143727368, 0.002126379973, 0.002860552143, 0.00258528776, 0.002221175975, 0.002287571624, 0.002269753022, 0.002100723821, 0.001752422095, 0.001803630977, 0.001899826248, 0.002156137142, 0.002711491509, 0.002531438179, 0.002158656476, 0.002205323436, 0.002025280393, 0.001789749635, 0.001825373988, 0.001914553843, 0.001977518945, 0.002075753427, 0.002544776951, 0.002260702104, 0.002105304897, 0.001984916688, 0.001835683003, 0.001749958111, 0.001839230725, 0.001806357272, 0.001982733737, 0.002274080231, 0.00229457775, 0.002117357162, 0.001845375086, 0.001633187581, 0.001749012712, 0.001681082922, 0.001634979053, 0.00184047917, 0.002192893711, 0.002051811528, 0.001802743262, 0.001717172202, 0.001572679132, 0.001451319933, 0.001753953548, 0.001512238654, 0.001449994849, 0.001779310054, 0.001611715347, 0.001544583527, 0.00155642601, 0.001497240486, 0.001291858116, 0.001423933613, 0.001357856058, 0.001340842036],
						[0.002949900901, 0.002858297419, 0.002751144754, 0.002712130997, 0.002658128393, 0.002672007559, 0.002490168387, 0.002545041644, 0.00233221606, 0.002264720672, 0.003098922292, 0.002900054774, 0.00267266428, 0.002469088971, 0.002451257932, 0.002447919046, 0.002103865601, 0.002367405345, 0.002055621468, 0.002507110088, 0.003093187533, 0.002649884076, 0.002533026189, 0.002524391039, 0.002335369335, 0.002472290253, 0.002161650698, 0.002317802626, 0.002292992308, 0.002309906809, 0.003085204278, 0.00269269277, 0.002678459174, 0.002510756872, 0.002216443786, 0.002083366443, 0.002395086863, 0.001900188593, 0.002323640337, 0.002259785774, 0.002801038174, 0.002802613883, 0.002228063953, 0.002359078671, 0.002469453739, 0.002191852187, 0.001966253632, 0.001894799551, 0.00206525987, 0.002405012624, 0.002887254079, 0.002613211472, 0.002257302626, 0.002454136129, 0.002246620993, 0.001991950908, 0.001888469975, 0.002159707946, 0.002179275575, 0.002323698689, 0.002392017713, 0.002359948953, 0.00227225667, 0.002110346607, 0.00199579587, 0.001914182171, 0.001984006171, 0.001986177557, 0.002274283814, 0.002230011678, 0.002117357162, 0.002621401631, 0.002126901078, 0.001825069475, 0.002014683763, 0.001882770839, 0.001953755803, 0.002056603229, 0.00240569299, 0.002109080628, 0.001936116943, 0.001858294563, 0.001781608032, 0.001627779459, 0.001970454534, 0.001773160458, 0.001636927628, 0.001953381381, 0.00179774555, 0.001620992162, 0.001691735074, 0.001597226743, 0.0015912288, 0.001611998182, 0.001645564968, 0.00160054969],
						[0.002982481082, 0.002724947082, 0.002720678945, 0.002692079815, 0.002581899568, 0.002546935189, 0.002476504628, 0.002445071567, 0.002247460908, 0.002206492379, 0.002903858749, 0.00263994758, 0.002675630159, 0.002266383371, 0.002351451581, 0.002278303381, 0.00200061205, 0.002224109391, 0.002065006753, 0.002455918705, 0.002901151042, 0.002471890206, 0.002351341128, 0.00233510927, 0.002151563889, 0.002295598751, 0.002002517374, 0.00226400016, 0.002186510036, 0.002355579841, 0.0028736148, 0.002425354578, 0.002505757031, 0.002307803728, 0.002131532637, 0.00197987481, 0.002338798784, 0.001913647432, 0.002270978414, 0.002215060521, 0.002492115849, 0.002626465312, 0.002056039595, 0.002171076485, 0.002207256383, 0.002013411673, 0.001836744414, 0.001817529201, 0.00192052398, 0.002390016422, 0.002549283435, 0.002426121036, 0.001995026144, 0.002299884068, 0.002133769892, 0.001904169139, 0.001866935206, 0.002024030435, 0.002159351368, 0.00217477644, 0.0021142444, 0.002099432758, 0.002006248791, 0.001896770696, 0.001878570542, 0.001800106007, 0.001901105125, 0.001948085404, 0.002126321301, 0.002045158491, 0.001845375086, 0.002126901078, 0.002276511839, 0.001678201675, 0.001833440004, 0.001792319511, 0.001811427404, 0.001902093582, 0.002396847716, 0.001904328088, 0.001718775095, 0.001651573889, 0.001642356721, 0.001562520092, 0.00190059034, 0.001684161069, 0.001583373631, 0.001914626162, 0.001781562354, 0.001410880098, 0.001494793329, 0.001431770042, 0.001430295881, 0.001462799528, 0.001531357426, 0.001467325825],
						[0.002356035502, 0.002440528653, 0.00232456853, 0.00220023004, 0.002227657241, 0.002227989655, 0.002163291487, 0.002144393999, 0.00188411888, 0.001912792792, 0.00246176927, 0.002390474875, 0.002214654882, 0.001915814607, 0.001999883551, 0.001985647385, 0.001752657322, 0.001910957354, 0.001803744266, 0.002200525629, 0.002514304441, 0.002036089535, 0.001900481007, 0.002032861451, 0.001885691664, 0.001996875499, 0.001885994037, 0.002005291199, 0.001928013396, 0.002038839365, 0.002450573985, 0.002117866581, 0.00214030932, 0.002045824989, 0.001853938575, 0.001875720046, 0.001965452817, 0.001699751327, 0.001988655607, 0.001987783614, 0.002183844354, 0.002368972886, 0.001878906302, 0.00200555453, 0.002010964229, 0.001817086302, 0.00163411733, 0.001639164356, 0.001766285481, 0.002112982315, 0.002204302719, 0.002052740998, 0.001761774637, 0.00205816416, 0.001897569735, 0.001637374176, 0.001704091112, 0.001879695348, 0.001854111527, 0.00173595849, 0.002042435891, 0.001857530485, 0.001678812607, 0.001675683985, 0.001637405318, 0.001640827494, 0.001751261424, 0.001738131697, 0.001906339992, 0.001799474408, 0.001633187581, 0.001825069475, 0.001678201675, 0.001901617019, 0.001689389323, 0.001681978974, 0.001659577292, 0.001815411915, 0.002077210542, 0.001680707469, 0.001479750176, 0.001444464685, 0.001531887011, 0.001363783562, 0.001697949351, 0.001514333937, 0.001467214757, 0.001568259676, 0.001741996881, 0.001342113321, 0.001277767174, 0.001379396749, 0.001353480725, 0.001308016213, 0.001373687721, 0.001352817337],
						[0.002656567017, 0.002583794193, 0.00253747154, 0.002432468817, 0.002470283393, 0.00248874816, 0.002345133043, 0.002389854799, 0.002077564129, 0.002093213831, 0.002599881832, 0.002500464799, 0.002358356459, 0.002137951166, 0.002172617411, 0.002228671992, 0.001914621865, 0.002059282986, 0.002000441696, 0.002357724883, 0.002581041583, 0.002110086306, 0.002128911057, 0.002332175229, 0.002024016797, 0.002265404134, 0.002042060296, 0.002147263607, 0.002127059326, 0.002239321229, 0.002543276964, 0.002233633219, 0.002277384155, 0.002175205296, 0.001872324134, 0.001887109246, 0.002133015049, 0.001726585286, 0.002219449236, 0.002186204378, 0.002363437762, 0.002417322856, 0.00193734829, 0.00202633518, 0.002105815658, 0.001956677058, 0.001779050864, 0.00178368941, 0.001905770945, 0.002240209112, 0.00230289056, 0.002158481331, 0.001884896651, 0.002082048619, 0.002027384031, 0.00181615008, 0.001790801096, 0.00199321726, 0.002045975845, 0.002037136712, 0.001953488988, 0.001919568355, 0.00184701393, 0.001751742541, 0.001789838079, 0.001703409595, 0.001834798411, 0.001961482042, 0.002098460066, 0.001782427378, 0.001749012712, 0.002014683763, 0.001833440004, 0.001689389323, 0.00208699488, 0.001712450073, 0.001813296546, 0.001918209543, 0.002270302748, 0.001759219865, 0.001542083706, 0.001524916872, 0.001519306473, 0.001411680124, 0.00173158236, 0.001670207893, 0.001543731544, 0.001747236502, 0.001804869635, 0.001302159289, 0.001408221757, 0.001359346906, 0.001375020272, 0.001435210999, 0.001468558918, 0.001440521518],
						[0.002544831954, 0.002526042736, 0.002549888275, 0.002479192684, 0.002403485387, 0.002472291663, 0.002378529183, 0.002344599062, 0.002123604214, 0.00213031204, 0.002592969402, 0.002486099168, 0.002414975934, 0.002032070219, 0.002162700965, 0.002114276384, 0.001904786973, 0.002030608105, 0.001965647569, 0.002368100476, 0.002572983051, 0.002118461054, 0.002028752638, 0.002109761204, 0.001981590044, 0.002150671015, 0.001988821433, 0.00215961835, 0.002026009422, 0.002284708105, 0.002552567112, 0.002230699024, 0.002153952799, 0.002223016351, 0.001909036257, 0.001949640749, 0.002127609696, 0.001847907701, 0.002179725325, 0.002162518963, 0.002239527586, 0.002343205445, 0.001977574331, 0.00214252625, 0.00210722721, 0.001922936359, 0.001877644921, 0.001739598415, 0.00183325678, 0.002241507991, 0.002175761723, 0.0021140707, 0.001870519581, 0.002101663354, 0.001938074558, 0.001761109708, 0.001876568938, 0.002005570746, 0.002050252042, 0.002029942012, 0.00201758249, 0.001893251837, 0.00179998138, 0.001808930336, 0.001715748504, 0.00178709596, 0.001852889709, 0.001842175714, 0.002175184895, 0.00181265829, 0.001681082922, 0.001882770839, 0.001792319511, 0.001681978974, 0.001712450073, 0.002175490603, 0.001743625395, 0.00199189188, 0.00226151043, 0.001711165998, 0.001540525879, 0.001519947769, 0.001537543115, 0.00147923634, 0.001773098228, 0.00170197913, 0.001674688296, 0.001831954788, 0.001804142466, 0.001318801308, 0.001365465744, 0.001388263461, 0.00141299199, 0.001423903284, 0.001504110078, 0.001507816599],
						[0.002931137112, 0.002966731622, 0.002757467982, 0.002699152581, 0.002673820064, 0.002701562387, 0.002524251707, 0.002619809282, 0.002368484015, 0.00237844092, 0.002729180908, 0.002786945071, 0.002506934412, 0.002162047775, 0.002268925825, 0.002285185976, 0.00196953169, 0.002143121664, 0.002239790169, 0.002607530716, 0.002693267864, 0.002273008844, 0.002231438776, 0.002414349499, 0.002175490999, 0.002292409218, 0.002202094963, 0.002237437268, 0.002247039578, 0.002384424834, 0.002644966347, 0.002236322138, 0.00231883584, 0.002255465655, 0.002066530273, 0.002115343694, 0.002180562054, 0.001886751418, 0.002221286989, 0.002429174419, 0.002268189282, 0.002340431675, 0.001972020328, 0.001987810504, 0.002005792894, 0.001926255715, 0.001820894591, 0.001934769853, 0.001966713259, 0.002418662726, 0.002439105183, 0.002201615384, 0.001795517347, 0.00213781531, 0.001924333968, 0.001886317215, 0.001794484491, 0.002129976208, 0.002150942185, 0.002329128467, 0.002020718018, 0.001914514415, 0.001704258851, 0.001737045766, 0.001829044759, 0.001800410688, 0.001891929825, 0.002085970703, 0.002123685539, 0.001825797102, 0.001634979053, 0.001953755803, 0.001811427404, 0.001659577292, 0.001813296546, 0.001743625395, 0.002560726752, 0.002083073614, 0.002325012524, 0.001761420977, 0.001542287283, 0.00150474773, 0.001612232676, 0.001417474375, 0.001773750525, 0.001785032259, 0.001751504465, 0.001892191381, 0.001904908508, 0.001385364582, 0.001454037065, 0.001371315787, 0.001356353647, 0.001369128949, 0.001459390899, 0.001517580729],
						[0.003179676958, 0.003089654503, 0.003098168353, 0.002967319487, 0.002835663113, 0.002937912347, 0.002792157251, 0.002852917885, 0.002484697379, 0.002584543946, 0.003147871729, 0.002980381819, 0.002874428286, 0.002401755863, 0.002563441418, 0.002486016887, 0.002308013374, 0.002335886857, 0.002399122799, 0.002848028378, 0.003012297037, 0.002606396206, 0.002492451223, 0.002598142746, 0.002433038059, 0.002491206316, 0.002329232732, 0.002500962852, 0.002468833821, 0.002667066797, 0.002905540983, 0.002683642961, 0.002832604928, 0.002509502702, 0.00233571516, 0.002307184219, 0.002459078114, 0.002194595047, 0.002421063393, 0.002663709273, 0.00267658867, 0.002674308167, 0.002174230852, 0.002343696514, 0.002344255713, 0.002134125577, 0.002145394252, 0.002056383569, 0.002198692526, 0.002577853902, 0.002668846883, 0.00245250601, 0.002097052813, 0.002254171392, 0.002256134662, 0.001973870058, 0.002072880281, 0.002232949935, 0.002416861729, 0.002497822871, 0.002274620913, 0.002093109033, 0.002055336166, 0.001859046805, 0.001879697817, 0.001956771025, 0.002045778129, 0.002257239072, 0.002434289261, 0.002054788709, 0.00184047917, 0.002056603229, 0.001902093582, 0.001815411915, 0.001918209543, 0.00199189188, 0.002083073614, 0.002980570262, 0.002776905698, 0.001883144273, 0.001638469279, 0.001646822523, 0.001668124132, 0.001633452578, 0.001904942811, 0.00181302171, 0.001790193869, 0.002071404173, 0.002108539649, 0.001503615307, 0.001489054779, 0.001519445517, 0.001495595492, 0.001462829871, 0.001541539547, 0.001597668889],
						[0.00410448382, 0.003539376119, 0.003999211129, 0.003811419029, 0.003627688117, 0.003746374369, 0.003547327024, 0.003539833913, 0.003263079647, 0.003406758634, 0.003568750149, 0.003553381874, 0.003548025327, 0.003083914722, 0.003174658432, 0.003060032788, 0.002725449108, 0.002960454181, 0.002997783046, 0.003700040109, 0.003726597402, 0.003256030568, 0.003146094171, 0.003200476161, 0.002925505835, 0.003125602712, 0.002684325948, 0.0032285676, 0.003034514466, 0.003507441082, 0.003571788009, 0.002926002114, 0.003215172404, 0.003086664146, 0.002790613686, 0.002697538545, 0.003023566438, 0.002489518257, 0.003063830896, 0.003367061096, 0.00312213084, 0.003246847673, 0.002690242371, 0.002688894305, 0.002866565122, 0.002562093626, 0.002643949387, 0.002448597599, 0.002631403393, 0.003224991981, 0.003424183921, 0.002904037504, 0.002458630052, 0.002674545428, 0.002770495738, 0.002307852321, 0.00253783427, 0.002727763283, 0.00308709855, 0.002997682812, 0.00277005577, 0.002478238425, 0.002590616688, 0.002208657034, 0.00210540579, 0.002342896959, 0.002521603449, 0.002693510647, 0.003287733483, 0.002457843811, 0.002192893711, 0.00240569299, 0.002396847716, 0.002077210542, 0.002270302748, 0.00226151043, 0.002325012524, 0.002776905698, 0.004495284329, 0.002342090592, 0.00196423921, 0.001968095145, 0.001963940144, 0.001932027363, 0.002236250352, 0.002092354005, 0.002225072321, 0.002734847721, 0.002739584572, 0.001727048959, 0.001749568877, 0.001809098989, 0.001689155895, 0.001757563505, 0.001734222474, 0.001758184786],
						[0.003094311644, 0.002988376321, 0.002958428789, 0.002813450776, 0.002699749688, 0.002725652958, 0.002582676068, 0.002638401879, 0.002315163065, 0.002316135476, 0.003247957188, 0.002975489791, 0.002676243201, 0.002498508202, 0.002401444852, 0.002355436976, 0.002072097471, 0.002227701645, 0.002008603922, 0.002599035893, 0.00317777977, 0.002670710051, 0.002458733781, 0.002519373609, 0.002296204885, 0.00234600001, 0.002051566, 0.00226710001, 0.002220333684, 0.002274031605, 0.003231873214, 0.002786528801, 0.00257810375, 0.002559313469, 0.002256992943, 0.002179623843, 0.002211358705, 0.001822208334, 0.002320725035, 0.002292200105, 0.00295891475, 0.002728429499, 0.002311024277, 0.00223512425, 0.002265267561, 0.002117743289, 0.001795900975, 0.001835817712, 0.001867927429, 0.002298418854, 0.002966632996, 0.002652139789, 0.00220955019, 0.002328507026, 0.00210142722, 0.001854531318, 0.001899653245, 0.002024908208, 0.002117514202, 0.002208828426, 0.002698335303, 0.002328326893, 0.002073480187, 0.001916434512, 0.001800355523, 0.001864830712, 0.001912894773, 0.001937579107, 0.002143422865, 0.002349393317, 0.002051811528, 0.002109080628, 0.001904328088, 0.001680707469, 0.001759219865, 0.001711165998, 0.001761420977, 0.001883144273, 0.002342090592, 0.00238108428, 0.00184448747, 0.001772725629, 0.001556963915, 0.001482927742, 0.001793079798, 0.001564612875, 0.001518886766, 0.001774290456, 0.001573060605, 0.001739458821, 0.001649662703, 0.001562606651, 0.001332975416, 0.00145065916, 0.001354462311, 0.00131376431],
						[0.002412252584, 0.002327715382, 0.002268515803, 0.002223566, 0.002162000371, 0.002111424879, 0.001961667199, 0.002021217408, 0.001846701912, 0.001841431471, 0.002593234347, 0.002443998787, 0.00227159661, 0.002029350921, 0.001947953835, 0.001903773391, 0.00164384809, 0.001910514222, 0.00166251154, 0.002003765691, 0.002611472402, 0.002162432991, 0.001977984212, 0.00194844727, 0.001843445266, 0.001961811834, 0.001725660176, 0.001868700425, 0.001885137482, 0.001883094939, 0.002752491792, 0.002314710289, 0.002168545284, 0.002125851234, 0.001784834378, 0.001839841472, 0.001910315612, 0.001510692335, 0.001971615854, 0.001857039652, 0.00243252121, 0.002414224911, 0.001885365619, 0.001907502635, 0.002038896069, 0.001801102602, 0.001530803387, 0.001554517899, 0.001659350044, 0.001922046187, 0.002436055038, 0.002216415693, 0.001801943668, 0.002047874052, 0.001832840001, 0.001563540612, 0.001507957861, 0.001716670092, 0.001769438899, 0.0018639162, 0.002290962724, 0.002022264844, 0.001874480495, 0.001698026684, 0.0016475462, 0.001542859377, 0.001712256401, 0.001624053584, 0.001787322454, 0.00199023081, 0.001802743262, 0.001936116943, 0.001718775095, 0.001479750176, 0.001542083706, 0.001540525879, 0.001542287283, 0.001638469279, 0.00196423921, 0.00184448747, 0.00192243273, 0.001573558801, 0.001467854587, 0.001348935717, 0.001656951629, 0.001443941498, 0.001344356872, 0.001572495244, 0.001432723256, 0.001472780098, 0.001464959819, 0.00137748601, 0.001316196758, 0.001285898014, 0.001326462412, 0.001307264791],
						[0.002356759221, 0.0023059398, 0.002255255898, 0.002173940325, 0.002126479925, 0.002183330207, 0.00198535973, 0.002049355105, 0.00181648464, 0.001807023938, 0.002508700781, 0.002360782909, 0.002203276806, 0.002015332046, 0.001925131204, 0.001912987547, 0.001742680996, 0.00190527573, 0.001645494544, 0.002031277, 0.00252861915, 0.002133878537, 0.002032270797, 0.002044466161, 0.001861205512, 0.001976809919, 0.001823756082, 0.001834793041, 0.001871014565, 0.001964218654, 0.00253094309, 0.002238214903, 0.002214754088, 0.002042135601, 0.001836535011, 0.001812945473, 0.001850515018, 0.001534000403, 0.001932172136, 0.001928067481, 0.002317128886, 0.002314975858, 0.00180390305, 0.001900859149, 0.001958905783, 0.001774330601, 0.001533680315, 0.001528020481, 0.001623427959, 0.001935690547, 0.002355746166, 0.002137932195, 0.001859650234, 0.001979971247, 0.001837111486, 0.001555407534, 0.001604302022, 0.001777444928, 0.001741036688, 0.001807356548, 0.002012465449, 0.001956531438, 0.001817398768, 0.001711879501, 0.001614763026, 0.00157997703, 0.001621937332, 0.001595164909, 0.001853956393, 0.001857663227, 0.001717172202, 0.001858294563, 0.001651573889, 0.001444464685, 0.001524916872, 0.001519947769, 0.00150474773, 0.001646822523, 0.001968095145, 0.001772725629, 0.001573558801, 0.001742285217, 0.001398927468, 0.00134751789, 0.001524758113, 0.001405406504, 0.001348093794, 0.001599095327, 0.001447074662, 0.001360521947, 0.001412518689, 0.001338489337, 0.001231215769, 0.001236032472, 0.001274523524, 0.001226998722],
						[0.001942123379, 0.0020451527, 0.00193368615, 0.001910269587, 0.001833840701, 0.001826697256, 0.001802418959, 0.001773477959, 0.001622591534, 0.001595657068, 0.002074163928, 0.002067739944, 0.002040575238, 0.001718203143, 0.001891678493, 0.001729006647, 0.001600301608, 0.001734247778, 0.001594663869, 0.001906474095, 0.002266167678, 0.00189527577, 0.001658662239, 0.001777307171, 0.001682774242, 0.001834543289, 0.001687102396, 0.001821025008, 0.001755383966, 0.00181361643, 0.002247082258, 0.001959224997, 0.001950404018, 0.001865060743, 0.001683397023, 0.001639023288, 0.001865905876, 0.001554887457, 0.001822227792, 0.001756471334, 0.001976554273, 0.002088830654, 0.001602129373, 0.001811115973, 0.001872698697, 0.001647109971, 0.001455476451, 0.001494005094, 0.001643625112, 0.001934885145, 0.002025684896, 0.001988579044, 0.00158182962, 0.001877199216, 0.001635391534, 0.001528674383, 0.001605942258, 0.001822485902, 0.001707515433, 0.001639376038, 0.00185434367, 0.00171365323, 0.00158261204, 0.001552914235, 0.001592853435, 0.001570629667, 0.001611890165, 0.001618440704, 0.001761340961, 0.001724921252, 0.001572679132, 0.001781608032, 0.001642356721, 0.001531887011, 0.001519306473, 0.001537543115, 0.001612232676, 0.001668124132, 0.001963940144, 0.001556963915, 0.001467854587, 0.001398927468, 0.001698701726, 0.001327940391, 0.001616634301, 0.001494835972, 0.001437515905, 0.001637855863, 0.001541159024, 0.001283218879, 0.001300200301, 0.001356399067, 0.001310637426, 0.001323083041, 0.001367549979, 0.001368370588],
						[0.002101900141, 0.001968497738, 0.00201404631, 0.001976848994, 0.001935762069, 0.001950773694, 0.00179555948, 0.001877138149, 0.001671580843, 0.001697851274, 0.00205201753, 0.001998907565, 0.001982869769, 0.001682071098, 0.001711701281, 0.001644036295, 0.001561717078, 0.001685271276, 0.00156716494, 0.001893518972, 0.002117982045, 0.001799982984, 0.001736552634, 0.001766203399, 0.001597012896, 0.001731717203, 0.001621796456, 0.001726301118, 0.001768774387, 0.001834717014, 0.002097977091, 0.001948823312, 0.00188735876, 0.001717740157, 0.001668607442, 0.001532550038, 0.001676086598, 0.001469200418, 0.001709856006, 0.001761635758, 0.001845401642, 0.002000047114, 0.001579517001, 0.001725504202, 0.001695447636, 0.001577969709, 0.001490387539, 0.001475744626, 0.001512696411, 0.001863184141, 0.001889482664, 0.001773262991, 0.00155812396, 0.001696418522, 0.001627968138, 0.001389151699, 0.001543079407, 0.001672337281, 0.001644695146, 0.001659224599, 0.001648297176, 0.001608642565, 0.00160578189, 0.001560586611, 0.001466974956, 0.001497027997, 0.001470146558, 0.001438808775, 0.001761927511, 0.001551595006, 0.001451319933, 0.001627779459, 0.001562520092, 0.001363783562, 0.001411680124, 0.00147923634, 0.001417474375, 0.001633452578, 0.001932027363, 0.001482927742, 0.001348935717, 0.00134751789, 0.001327940391, 0.001499097178, 0.001464277672, 0.001374150364, 0.001351491444, 0.001519239066, 0.001488590535, 0.001073216745, 0.001183648946, 0.001176539898, 0.001128847482, 0.00120403918, 0.001211319905, 0.001150033874],
						[0.002560515946, 0.002535640801, 0.00253937852, 0.002485448666, 0.002350195792, 0.00231638808, 0.002302544639, 0.002263133226, 0.002067975693, 0.002067148721, 0.002574940276, 0.002534972683, 0.002484016339, 0.002112971385, 0.002146911221, 0.002084742402, 0.001859878712, 0.002121516629, 0.001981218007, 0.002302300689, 0.002668235806, 0.00224771988, 0.002087295719, 0.002226324109, 0.002031397406, 0.002172197335, 0.001933316665, 0.002169956628, 0.002103557194, 0.002292201244, 0.002585262202, 0.002238119388, 0.002251596033, 0.002234850987, 0.001948746689, 0.001956169542, 0.002156986478, 0.001768324436, 0.00215968941, 0.002084969638, 0.00234040907, 0.002459099058, 0.001933646242, 0.002020156092, 0.00206957304, 0.001900135167, 0.001746800454, 0.001733468366, 0.001903089234, 0.002231743966, 0.002335763509, 0.002244430763, 0.001802007614, 0.002103916017, 0.002071366733, 0.001813067579, 0.001825533626, 0.00202609524, 0.00200538994, 0.002125126885, 0.002036498587, 0.001983211094, 0.001834642637, 0.001794451241, 0.001730962735, 0.001760054456, 0.001863226423, 0.001889528393, 0.002223767524, 0.001880673805, 0.001753953548, 0.001970454534, 0.00190059034, 0.001697949351, 0.00173158236, 0.001773098228, 0.001773750525, 0.001904942811, 0.002236250352, 0.001793079798, 0.001656951629, 0.001524758113, 0.001616634301, 0.001464277672, 0.002158109585, 0.00167480383, 0.001628784164, 0.001809829016, 0.001810650856, 0.001328240001, 0.001424361999, 0.001421208669, 0.00141370992, 0.001458385528, 0.001509848185, 0.001468393513],
						[0.002299870077, 0.002346819807, 0.002258501074, 0.002206599569, 0.0021989561, 0.002196664573, 0.002102420911, 0.00219391231, 0.001944108977, 0.001983828019, 0.002281943433, 0.002236704862, 0.002223537119, 0.001897921416, 0.002034977418, 0.00191297072, 0.001712852285, 0.001922968469, 0.001904040506, 0.002221521965, 0.002302778545, 0.001895963612, 0.001880784241, 0.002049160193, 0.001888393315, 0.0020091459, 0.001898223352, 0.002032204841, 0.001956164834, 0.002216786869, 0.002213443299, 0.001904738534, 0.002023847758, 0.001997505152, 0.001817722133, 0.00177166698, 0.001938585802, 0.001737591391, 0.002086821575, 0.002156477303, 0.002079558915, 0.00217047674, 0.001659412079, 0.001792794597, 0.00182965049, 0.001711745192, 0.001655796305, 0.001701345065, 0.001771103512, 0.002133414729, 0.002037849418, 0.001871709159, 0.00162487014, 0.001831609016, 0.00171055106, 0.00165444263, 0.001679398022, 0.001939476904, 0.001967585467, 0.001869063588, 0.001800994314, 0.001738711241, 0.001570663998, 0.001666964874, 0.001611239802, 0.001638850111, 0.001699113779, 0.001777879072, 0.00202875252, 0.001614511409, 0.001512238654, 0.001773160458, 0.001684161069, 0.001514333937, 0.001670207893, 0.00170197913, 0.001785032259, 0.00181302171, 0.002092354005, 0.001564612875, 0.001443941498, 0.001405406504, 0.001494835972, 0.001374150364, 0.00167480383, 0.001943278547, 0.001557263926, 0.001799886207, 0.001633126472, 0.001208674032, 0.001284975258, 0.001287052074, 0.001306314015, 0.001310153149, 0.001438865939, 0.001397350548],
						[0.002083781247, 0.002105827515, 0.002197152045, 0.002148219717, 0.002055965995, 0.002116575774, 0.001963682724, 0.002079094009, 0.001812298457, 0.001884783965, 0.002136746694, 0.00217467103, 0.002091956136, 0.001773218193, 0.001895268085, 0.001819593425, 0.001679136737, 0.001818487049, 0.001840492434, 0.002142081755, 0.00226003792, 0.001801342803, 0.001813132445, 0.00191435273, 0.001786999549, 0.001939627161, 0.001877129076, 0.001974984732, 0.001946371307, 0.002177387041, 0.002178675095, 0.001925369836, 0.001959518691, 0.001918653967, 0.001742905187, 0.001680075186, 0.001892057444, 0.001611170221, 0.001997052557, 0.002065400695, 0.001874360924, 0.00203353407, 0.001606100416, 0.001736940424, 0.001730261696, 0.001660298011, 0.001665854964, 0.001608682258, 0.001686312902, 0.002105693351, 0.002005744056, 0.001750546507, 0.001567712893, 0.001815691681, 0.001684877891, 0.001623472287, 0.001668945698, 0.001907641035, 0.001912712957, 0.00195522756, 0.001637370346, 0.001653203216, 0.001592933608, 0.001556821933, 0.001517674255, 0.001596252795, 0.001664173048, 0.001792062365, 0.002015049614, 0.001502270361, 0.001449994849, 0.001636927628, 0.001583373631, 0.001467214757, 0.001543731544, 0.001674688296, 0.001751504465, 0.001790193869, 0.002225072321, 0.001518886766, 0.001344356872, 0.001348093794, 0.001437515905, 0.001351491444, 0.001628784164, 0.001557263926, 0.001869148927, 0.001821622391, 0.001754296788, 0.001100220997, 0.001245637172, 0.001224408789, 0.001235302959, 0.001315514696, 0.001423332905, 0.001436971018],
						[0.002701932892, 0.00248426969, 0.002615141548, 0.002382785761, 0.002359758516, 0.002469559865, 0.002249388048, 0.002384093743, 0.002112557084, 0.002081992134, 0.002578710274, 0.002359074018, 0.002527216656, 0.002041876751, 0.002260311334, 0.002199419717, 0.001911216259, 0.002148142537, 0.002101416757, 0.002591435225, 0.002571792485, 0.002218445191, 0.00215663356, 0.002332399788, 0.002043348134, 0.002327275906, 0.002004063264, 0.002326966181, 0.002220914779, 0.002613953357, 0.002510933318, 0.002235240477, 0.002326900403, 0.002221537306, 0.001849463061, 0.001911598533, 0.002211490898, 0.001913410436, 0.002346739526, 0.002442677105, 0.002396075867, 0.002300070477, 0.001862895769, 0.002063233178, 0.002064534697, 0.001848218608, 0.001962286642, 0.001935430794, 0.001977935665, 0.002335114657, 0.002458236182, 0.002199654503, 0.001873815258, 0.002105651673, 0.001950284629, 0.00179482723, 0.001919123145, 0.002103609887, 0.002272632476, 0.002256425015, 0.001989093522, 0.001914985238, 0.002057573856, 0.001761721129, 0.001656282469, 0.001828888462, 0.001824116362, 0.00216390251, 0.002359241942, 0.001976858326, 0.001779310054, 0.001953381381, 0.001914626162, 0.001568259676, 0.001747236502, 0.001831954788, 0.001892191381, 0.002071404173, 0.002734847721, 0.001774290456, 0.001572495244, 0.001599095327, 0.001637855863, 0.001519239066, 0.001809829016, 0.001799886207, 0.001821622391, 0.002979244709, 0.002074992858, 0.001353083366, 0.001470737906, 0.001491283712, 0.001517181308, 0.001570266791, 0.001544666878, 0.001551373396],
						[0.002757407576, 0.002719471677, 0.002802102426, 0.002707287202, 0.002504526251, 0.002616001643, 0.002506105964, 0.002466890693, 0.002276282331, 0.00232940124, 0.002542355418, 0.002538706295, 0.00245712161, 0.002086498421, 0.002183371878, 0.002243951131, 0.001958052651, 0.0021263257, 0.002207618824, 0.002621234068, 0.002718526671, 0.00220851091, 0.002089571752, 0.002324302701, 0.002036454966, 0.002234689403, 0.002049152968, 0.002348328682, 0.002292293272, 0.002567909878, 0.002546967054, 0.002217031467, 0.002387760302, 0.002176170363, 0.001900591909, 0.001958579231, 0.002181747593, 0.001930658194, 0.002276102439, 0.00243369268, 0.002140531648, 0.002407624321, 0.001994029146, 0.002008129068, 0.001966684595, 0.001831724253, 0.001977779456, 0.001909424209, 0.001958590849, 0.002455373533, 0.002416220994, 0.002098263899, 0.001786576186, 0.002104577206, 0.002041913208, 0.001723097603, 0.001938805242, 0.002081524935, 0.002188558276, 0.002297224771, 0.001834252413, 0.001870729067, 0.001761267425, 0.001674310159, 0.00164398893, 0.001783426135, 0.001925407783, 0.001970141906, 0.002349290251, 0.001714422288, 0.001611715347, 0.00179774555, 0.001781562354, 0.001741996881, 0.001804869635, 0.001804142466, 0.001904908508, 0.002108539649, 0.002739584572, 0.001573060605, 0.001432723256, 0.001447074662, 0.001541159024, 0.001488590535, 0.001810650856, 0.001633126472, 0.001754296788, 0.002074992858, 0.002635131973, 0.001167564371, 0.00130376408, 0.001329507691, 0.001380001238, 0.001321609962, 0.001336970271, 0.001462465311],
						[0.002009863366, 0.002091166113, 0.001993131787, 0.001804422913, 0.001829998069, 0.001804342367, 0.001728073759, 0.001729753224, 0.001537089641, 0.001503426019, 0.002360605143, 0.002178968325, 0.00192903563, 0.001765418608, 0.001703799545, 0.001705130056, 0.001415231144, 0.001657059318, 0.001389582413, 0.001826813062, 0.002277131577, 0.001925310451, 0.00168525719, 0.001714808166, 0.001631303246, 0.001603869682, 0.001490542702, 0.001651025562, 0.00153782559, 0.001524680113, 0.002496554737, 0.002028930962, 0.001960869109, 0.001858518206, 0.001683628808, 0.001715632971, 0.00161786628, 0.001297908725, 0.001649338559, 0.001676106267, 0.002277918698, 0.002036565539, 0.001724103763, 0.001679055023, 0.001747999998, 0.001566536254, 0.001304139577, 0.001308641757, 0.001336237643, 0.00165968462, 0.002355888284, 0.002135683637, 0.001529337994, 0.001764384174, 0.0014921953, 0.001368238167, 0.001453504753, 0.001497014391, 0.001539832924, 0.001523107793, 0.00224975273, 0.00181912707, 0.001508379831, 0.001338224906, 0.001406303153, 0.001414236235, 0.001432696674, 0.001485800171, 0.001502287893, 0.00193244399, 0.001544583527, 0.001620992162, 0.001410880098, 0.001342113321, 0.001302159289, 0.001318801308, 0.001385364582, 0.001503615307, 0.001727048959, 0.001739458821, 0.001472780098, 0.001360521947, 0.001283218879, 0.001073216745, 0.001328240001, 0.001208674032, 0.001100220997, 0.001353083366, 0.001167564371, 0.001732174103, 0.001355419126, 0.001356884386, 0.001206983927, 0.001050330198, 0.001120576959, 0.001069502127],
						[0.001980655134, 0.00204061465, 0.001941433241, 0.001881753917, 0.001840005844, 0.00183879967, 0.001709216314, 0.001775353392, 0.001600344549, 0.001564091681, 0.002220554732, 0.002083663409, 0.002014819832, 0.001761942727, 0.001711400505, 0.001712051313, 0.00147673357, 0.001654215713, 0.001484180669, 0.001795320974, 0.002209301972, 0.001872333325, 0.001848983201, 0.001772016561, 0.001626562968, 0.001739008661, 0.001573055978, 0.001691110869, 0.001667038649, 0.001779855608, 0.002295747956, 0.002013238493, 0.00197660144, 0.001821790762, 0.001661515879, 0.001600549369, 0.001671386618, 0.001312630625, 0.001713164575, 0.001761323688, 0.002093178431, 0.002083728244, 0.001573455973, 0.001674292826, 0.001856325133, 0.001659315375, 0.001386751283, 0.001388652916, 0.001430388361, 0.001753961274, 0.002182515245, 0.002002634763, 0.001611606875, 0.001754955104, 0.001613936507, 0.001434811152, 0.001421291285, 0.001545671162, 0.001644256099, 0.001675158753, 0.001895068104, 0.001775680814, 0.001597609687, 0.001469942223, 0.001505474531, 0.001439126627, 0.001447393032, 0.001609694131, 0.001690429149, 0.00173933189, 0.00155642601, 0.001691735074, 0.001494793329, 0.001277767174, 0.001408221757, 0.001365465744, 0.001454037065, 0.001489054779, 0.001749568877, 0.001649662703, 0.001464959819, 0.001412518689, 0.001300200301, 0.001183648946, 0.001424361999, 0.001284975258, 0.001245637172, 0.001470737906, 0.00130376408, 0.001355419126, 0.001475355593, 0.001258830108, 0.001160337151, 0.001174410199, 0.001188517155, 0.001154916643],
						[0.001876568152, 0.001945355621, 0.001881035982, 0.001755480548, 0.001776080995, 0.001747813931, 0.001684785524, 0.001711371032, 0.001539839272, 0.001499557562, 0.002113770986, 0.001987904628, 0.00197991251, 0.001704274891, 0.001713412, 0.001694201176, 0.001456591754, 0.001605786675, 0.001422802845, 0.001810262683, 0.002131248267, 0.001733466947, 0.001633653508, 0.001672395879, 0.001603387422, 0.001730536194, 0.001556844155, 0.001726021726, 0.001664785053, 0.001690825931, 0.002140649214, 0.001896815714, 0.001872576902, 0.001711593746, 0.00162246157, 0.001576340268, 0.001651221385, 0.001291381265, 0.001690114187, 0.001715023211, 0.002042077396, 0.002002175109, 0.001526185446, 0.001687232765, 0.001757823844, 0.001530303126, 0.001319070256, 0.001359905258, 0.001420022489, 0.001731624546, 0.002002751956, 0.001942345413, 0.001470394496, 0.001726516932, 0.001541106991, 0.001338967163, 0.001527167595, 0.001640709803, 0.001611616989, 0.001569474671, 0.001898527721, 0.001709574839, 0.001539238963, 0.001372897973, 0.001409622401, 0.001421393745, 0.001488967379, 0.001519896088, 0.001701696033, 0.001693475392, 0.001497240486, 0.001597226743, 0.001431770042, 0.001379396749, 0.001359346906, 0.001388263461, 0.001371315787, 0.001519445517, 0.001809098989, 0.001562606651, 0.00137748601, 0.001338489337, 0.001356399067, 0.001176539898, 0.001421208669, 0.001287052074, 0.001224408789, 0.001491283712, 0.001329507691, 0.001356884386, 0.001258830108, 0.001477072021, 0.001255149956, 0.001200739301, 0.001226329989, 0.001201041421],
						[0.001624975344, 0.00177718764, 0.001629664591, 0.001556326854, 0.001555161991, 0.001571649058, 0.0015206781, 0.001494355257, 0.001380351461, 0.001293417718, 0.001821547472, 0.001794990531, 0.001669107432, 0.00152144715, 0.001534514796, 0.001474786573, 0.001216366287, 0.001515116099, 0.001374981132, 0.001647121456, 0.001875955335, 0.001460972446, 0.00153148607, 0.001495796545, 0.00138335841, 0.00160289107, 0.001397123182, 0.001591942106, 0.001486070658, 0.001564463116, 0.00184150236, 0.001624917218, 0.00167473677, 0.001589475474, 0.001411743762, 0.00143020767, 0.001634855501, 0.001203881102, 0.00157166069, 0.001546563915, 0.001717972348, 0.001875702655, 0.001367473141, 0.001571458355, 0.001656088474, 0.001421089374, 0.001340951321, 0.001246940377, 0.001365864643, 0.0017056357, 0.001770018254, 0.001702335031, 0.001221941677, 0.00164620921, 0.001471749075, 0.001364436889, 0.001413216884, 0.001567264754, 0.001492172521, 0.001440842522, 0.001539200166, 0.001544124909, 0.001439716463, 0.001312547653, 0.001346738712, 0.001257311173, 0.001353134112, 0.001452871784, 0.001637735267, 0.0014430677, 0.001291858116, 0.0015912288, 0.001430295881, 0.001353480725, 0.001375020272, 0.00141299199, 0.001356353647, 0.001495595492, 0.001689155895, 0.001332975416, 0.001316196758, 0.001231215769, 0.001310637426, 0.001128847482, 0.00141370992, 0.001306314015, 0.001235302959, 0.001517181308, 0.001380001238, 0.001206983927, 0.001160337151, 0.001255149956, 0.001555627002, 0.001175399778, 0.001255432815, 0.001297431844],
						[0.001797528616, 0.001833740965, 0.001757523692, 0.001713310797, 0.001703873611, 0.001717569968, 0.00160284881, 0.001666919889, 0.001445855309, 0.001433147338, 0.001864275149, 0.001780425059, 0.001871868301, 0.001538600425, 0.00171764512, 0.001594540806, 0.001374620282, 0.001592331939, 0.001391746185, 0.001765786251, 0.00190288067, 0.001628559069, 0.001521509988, 0.001619156643, 0.001554231646, 0.00161558132, 0.001458963187, 0.001656064673, 0.001637206048, 0.001679989757, 0.001894217899, 0.001691962259, 0.001639575669, 0.001632520931, 0.001469458031, 0.001363441344, 0.001659412043, 0.001290540791, 0.001657481338, 0.001547746646, 0.001779658493, 0.001848791536, 0.001388784454, 0.001584612812, 0.001698686823, 0.001542052019, 0.00135730855, 0.001404625531, 0.00148274205, 0.001694330807, 0.001689529892, 0.001709835884, 0.001405332314, 0.00164119908, 0.001556784332, 0.001395907181, 0.001448010909, 0.001563318738, 0.001583504098, 0.001625692915, 0.001481003002, 0.001523394415, 0.001489612646, 0.001355054415, 0.001380300963, 0.001417501875, 0.001441163181, 0.001512179815, 0.001694576331, 0.001442626283, 0.001423933613, 0.001611998182, 0.001462799528, 0.001308016213, 0.001435210999, 0.001423903284, 0.001369128949, 0.001462829871, 0.001757563505, 0.00145065916, 0.001285898014, 0.001236032472, 0.001323083041, 0.00120403918, 0.001458385528, 0.001310153149, 0.001315514696, 0.001570266791, 0.001321609962, 0.001050330198, 0.001174410199, 0.001200739301, 0.001175399778, 0.00148087213, 0.001291575366, 0.001280782044],
						[0.001651044294, 0.001833126743, 0.001675342356, 0.001644961803, 0.001733502399, 0.001702643639, 0.001599105457, 0.001640984047, 0.001467154418, 0.001459815065, 0.001779315548, 0.001803667429, 0.001839191813, 0.001536824264, 0.001629135281, 0.001590322296, 0.001367862053, 0.001566729393, 0.00151998509, 0.001753677596, 0.001968385308, 0.001487323111, 0.001577008256, 0.001669654434, 0.00142310159, 0.001635188623, 0.001493556021, 0.001621731298, 0.001584095463, 0.001671031498, 0.001779130572, 0.001543014166, 0.001682717651, 0.001637769712, 0.001513598958, 0.001368196308, 0.001620250311, 0.001357172477, 0.001697401999, 0.0016355327, 0.001629050738, 0.001887412237, 0.001310831026, 0.001583883056, 0.001684434505, 0.001504801496, 0.001383488268, 0.001371462861, 0.001519772785, 0.001752820866, 0.001693807964, 0.001653003911, 0.001389306134, 0.001632717802, 0.001587092448, 0.001442945758, 0.001478431473, 0.001658344814, 0.001624889665, 0.001511484483, 0.001471806275, 0.001502743129, 0.001449216874, 0.001392112944, 0.001465920283, 0.001434609122, 0.001423381553, 0.001595801621, 0.001727829921, 0.001374532828, 0.001357856058, 0.001645564968, 0.001531357426, 0.001373687721, 0.001468558918, 0.001504110078, 0.001459390899, 0.001541539547, 0.001734222474, 0.001354462311, 0.001326462412, 0.001274523524, 0.001367549979, 0.001211319905, 0.001509848185, 0.001438865939, 0.001423332905, 0.001544666878, 0.001336970271, 0.001120576959, 0.001188517155, 0.001226329989, 0.001255432815, 0.001291575366, 0.001652398889, 0.001400097182],
						[0.001615803773, 0.001802919713, 0.001705850964, 0.001681966033, 0.001681532412, 0.001731275443, 0.001639530125, 0.001597138892, 0.001461220643, 0.001456114868, 0.00185400446, 0.001750839599, 0.001779332353, 0.00156334151, 0.00166261154, 0.001530771977, 0.001422734186, 0.001568749327, 0.001501016057, 0.00178581935, 0.00192768926, 0.001518940644, 0.001443522175, 0.001534071611, 0.001518029966, 0.001538851026, 0.001504941644, 0.001688645496, 0.001656696165, 0.001728721221, 0.00175253825, 0.001629675397, 0.001722636547, 0.001580531026, 0.001448566505, 0.001363541314, 0.001693878156, 0.001286333176, 0.001620050265, 0.001723242545, 0.001584058446, 0.001908320189, 0.001294276419, 0.001552290241, 0.00171185129, 0.00145597054, 0.001367615045, 0.001372454729, 0.001519431052, 0.001814234354, 0.001633407093, 0.001587142632, 0.001394237652, 0.00170226126, 0.001462726316, 0.001410140658, 0.001485237838, 0.001638997416, 0.001665134061, 0.00162087977, 0.001417212846, 0.001442843411, 0.001435943811, 0.001311553054, 0.001417331467, 0.001413057928, 0.001498832196, 0.00157436436, 0.001714672068, 0.001388198613, 0.001340842036, 0.00160054969, 0.001467325825, 0.001352817337, 0.001440521518, 0.001507816599, 0.001517580729, 0.001597668889, 0.001758184786, 0.00131376431, 0.001307264791, 0.001226998722, 0.001368370588, 0.001150033874, 0.001468393513, 0.001397350548, 0.001436971018, 0.001551373396, 0.001462465311, 0.001069502127, 0.001154916643, 0.001201041421, 0.001297431844, 0.001280782044, 0.001400097182, 0.001685882495]];

		var returns = [0.005873325, 0.007285383333, 0.0088992, 0.01417630833, 0.009235308333, 0.01318765, 0.01406648333, 0.01360380833, 0.01448148333, 0.01661586667, 0.004312425, 0.008158858333, 0.006681016667, 0.0005192, 0.006546558333, 0.005799741667, 0.009362716667, 0.005585858333, 0.01105469167, 0.01514394167, 0.0008603833333, 0.005390983333, 0.004285958333, 0.002965683333, 0.00882215, 0.01205829167, 0.007865941667, 0.009536483333, 0.01536361667, 0.01206999167, 0.001890775, 0.001219958333, 0.005129775, 0.00944175, 0.008676658333, 0.01137755, 0.008224466667, 0.01020840833, 0.01157376667, 0.007151091667, 0.005009416667, 0.005049083333, 0.006737966667, 0.01528215, 0.004935233333, 0.01058240833, 0.008808275, 0.0079628, 0.01132546667, 0.009103333333, 0.002207241667, 0.001660125, 0.00703745, 0.003528141667, 0.0058985, 0.00719915, 0.01374685833, 0.01278683333, 0.0099985, 0.008261508333, 0.0080636, 0.002264716667, 0.004111808333, 0.0021296, 0.0056057, 0.005727558333, 0.009533016667, 0.01281698333, 0.010599625, 0.006604125, 0.0015463, 0.001751933333, 0.003734616667, 0.008640916667, 0.007365616667, 0.0103935, 0.01234144167, 0.009268191667, 0.0082898, 0.004892658333, 0.004707458333, 0.005013725, 0.004047533333, 0.0034394, 0.006425441667, 0.007648291667, 0.005843025, 0.00818545, 0.0080509, 0.007949991667, 0.0058026, 0.003725691667, 0.001433233333, 0.002909541667, 0.006006575, 0.007056141667];
		
		// Compute the maximum desired volatility
		var maxVolatility = 0.10;
		
		// Compute the RSO-MVO weights
		var weights = PortfolioAllocation.randomSubspaceMeanVarianceOptimizationWeights(returns, covMat, { nbRandomSubsets: 1024*1, 
		                                                                                                   subsetsAggregationMethod: 'median',
																										   constraints: {minExposure: 0},
																										   subsetsOpt: { 
																											   constraints: {
																												   maxVolatility: maxVolatility 
																											    }
																										     }
																										  });
																										 
		var weightsGsmo = PortfolioAllocation.randomSubspaceMeanVarianceOptimizationWeights(returns, covMat, { nbRandomSubsets: 1024*1, 
		                                                                                                       constraints: {minExposure: 0},
																											   subsetsAggregationMethod: 'median',
																										       subsetsOpt: { 
																											       optimizationMethod: 'gsmo', 
																											       constraints: {
																												       maxVolatility: maxVolatility 
																											        }
																										         }
																										      }); 
																											  
	}
});	



QUnit.test('Random subspace global minimum variance portfolio', function(assert) {    
	// Test using static data, with a semi-positive definite covariance matrix, to test that the
	// optional asset returns are properly provided to the subsets.
	{
		var covMat = [[0.0400, 0.0400, 0.0400], [0.0400, 0.0400, 0.0400], [0.0400, 0.0400, 0.0400]];
		var returns = [0.01, 0.02, 0.03];
		
		// The GMV portfolios associated to the three sub-optimization problems with 2 assets and their returns are:
		// - Assets 1 and 2: [0, 1]
		// - Assets 1 and 3: [0, 1]
		// - Assets 2 and 3: [0, 1]
		// So, the final portfolio must be [0, 1/3, 2/3]
		var expectedWeights = [0, 1/3, 2/3];
		var weights = PortfolioAllocation.randomSubspaceGlobalMinimumVarianceOptimizationWeights(covMat, 
																								{ sizeSubsets: 2, mu: returns, subsetsGenerationMethod: 'deterministic'});
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-6), true, 'Test #1, semi-positive definite covariance matrix, optional returns provided');
		
		// The GMV portfolios associated to the three sub-optimization problems with 2 assets and no returns are:
		// - Assets 1 and 2: [1/2, 1/2]
		// - Assets 1 and 3: [1/2, 1/2]
		// - Assets 2 and 3: [1/2, 1/2]
		// So, the final portfolio must be [1/3, 1/3, 1/3]
		var expectedWeights = [1/3, 1/3, 1/3];
		var weights = PortfolioAllocation.randomSubspaceGlobalMinimumVarianceOptimizationWeights(covMat, 
																								{ sizeSubsets: 2, subsetsGenerationMethod: 'deterministic'});
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-6), true, 'Test #1, semi-positive definite covariance matrix, no returns provided');
	}
	
	// Test that with a non-limited cash asset, the generated portfolio is unique, equal to cash...
	{
		var covMat = [[1, 0, 0], [0, 1, 0], [0, 0, 1]];
		var expectedWeights = [0, 0, 0];
		var weights = PortfolioAllocation.randomSubspaceGlobalMinimumVarianceOptimizationWeights(covMat, {constraints: {maxCashWeight: 1}});
		
		assert.equal(areWeightsIdentical(weights, expectedWeights), true, 'Test #2, full cash position allowed');
	}
	
});	

QUnit.test('Random subspace maximum Sharpe ratio portfolio', function(assert) {    
	// Test using static data
	{
		var covMat = [[1, 0, 0], [0, 2, 0], [0, 0, 3]];
		var returns = [1, 2, 3];
		
		// The MSR portfolios associated to the three sub-optimization problems with 2 assets are:
		// - Assets 1 and 2: [0, 1]
		// - Assets 1 and 3: [0, 1]
		// - Assets 2 and 3: [0, 1]
		// So, the final portfolio must be [0, 1/3, 2/3]
		var rf = 2;
		var expectedWeights = [0, 1/3, 2/3];
		var weights = PortfolioAllocation.randomSubspaceMaximumSharpeRatioWeights(returns, covMat, rf,
																				  { sizeSubsets: 2, subsetsGenerationMethod: 'deterministic'});
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-8), true, 'Test #1');
		
		
		// When rf > 3, there is no portfolio with a positive excess return, so that the computation must fail
		var rf = 3 + 1e-6;
		assert.throws(function() { 
			var weights = PortfolioAllocation.randomSubspaceMaximumSharpeRatioWeights(returns, covMat, rf,
 																		              { sizeSubsets: 2, subsetsGenerationMethod: 'deterministic'});
		},
		new Error('no feasible portfolio generated'),
		"Test #2, no feasible portfolios");
	}
});




QUnit.test('Best constantly rebalanced portfolio', function(assert) {    
	// Test with static data
	{
		// Define the assets price relatives
		var priceRelatives = [[1 - 0.05, 1 - 0.05, 1 - 0.05], // Asset 1 returns, all negatives
		                      [1, 1, 1], // Asset 2 returns, constants
							  [1 + 0.05, 1 + 0.05, 1 + 0.05]]; // Asset 3 returns, all positives
		
		// The BRCP should be the portfolio fully invested in the asset 3, as any other
		// portfolio would lead to a lower terminal wealth.
		var expectedWeights = [0, 0, 1];
		
		// Compute the weights of the BCRP
		var weights = PortfolioAllocation.bestConstantlyRebalancedWeights(priceRelatives);

		//
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-8), true, 'Best constantly rebalanced portfolio, #1');
	}  
	

	// Test with random data, comparing the BCRP algorithm with an exhaustive search algorithm
	{
		// Define random assets price relatives for 5 assets and 24 periods
		var nbAssets = 5;
		var nbPeriods = 24;
		var priceRelativeMax = 1.20;
		var priceRelativeMin = 0.80;
		
		var priceRelatives = [];
		for (var i = 0; i < nbAssets; ++i) {
			// Initialization
			priceRelatives[i] = new Array(nbPeriods);
			
			//
			for (var j = 0; j < nbPeriods; ++j) {
				priceRelatives[i][j] = Math.random() * (priceRelativeMax - priceRelativeMin) + priceRelativeMin; // belongs to (priceRelativeMin, priceRelativeMax)
			}
		}
		
		// Compute the weights of the associated BCRP
		var weights = PortfolioAllocation.bestConstantlyRebalancedWeights(priceRelatives);
		
		// Compute the weights of the BRCP using a grid search algorithm:
		//
		// - Objective function to minimize: opposite of the portfolio cumulative wealth
		function portfolio_cumulative_wealth(arr) { 
			// Initialization
			var cumWealthPortfolio = 1;
			
			// Compute the evolution of the portfolio wealth
			for (var j = 0; j < nbPeriods; ++j) {
				// Compute the wealth relative for the current period
				var cumWealthPeriod = 0;
				for (var i = 0; i < nbAssets; ++i) {
					cumWealthPeriod += arr[i] * priceRelatives[i][j];
				}
				
				// Update the portfolio cumulative wealth
				cumWealthPortfolio *= cumWealthPeriod;
			}
			
			// Return it
			return -cumWealthPortfolio;
		}
		
		// - Numerical weights computation
		var numericalWeights = PortfolioAllocation.numericalOptimizationWeights(nbAssets, portfolio_cumulative_wealth, {optimizationMethod: 'grid-search', optimizationMethodParams: {k: 10}})
		
		// Compare the resulting cumulative portfolio wealths, which must be numerically equal
		assert.equal(Math.abs(portfolio_cumulative_wealth(weights) - portfolio_cumulative_wealth(numericalWeights[0])) <= 1e-2, true, 'Best constantly rebalanced portfolio, #2');
	}

});


QUnit.test('Minimum tracking error portfolio', function(assert) {    
	// Test with static data: 1 EEM ETF, 1 WLD ETF, 1 ACWI ETF as benchmark
	{
		// Define the assets returns
		var assetsReturns = [[0.078129363,0.03663478,0.028787879,0.031491869,-0.019533386,-0.042308288,-0.062133001,-0.103910104,-0.027053038,0.082179315,0.004579357,-0.049970429], // Asset 1 returns in 2015 (EEM)
		                     [0.05289784,0.065017266,0.028012939,-0.019089773,0.025592817,-0.03893898,0.026564726,-0.07927768,-0.033253707,0.090633334,0.04075,-0.044859476]]; // Asset 2 returns in 2015 (WLD)
		
		// Define the benchmark returns in 2015 (ACWI)
		var benchmarkReturns = [0.055518502,0.062009871,0.028003081,-0.013945023,0.020509392,-0.03935567,0.017029792,-0.081809115,-0.032750097,0.089738414,0.037148934,-0.045498531];

		// Prior to any numerical computation, since ACWI index in 2015 ~= 90% World index + 10% Emerging index,
		// the minimum tracking error portfolio should have weights close to [0.1, 0.9], which has been
		// validated by the numerical computation below.
		var expectedWeights = [0.10614062381555761, 0.8938593761844424];
		
		// Compute the weights of the minimum tracking error portfolio
		var weights = PortfolioAllocation.minimumTrackingErrorWeights(assetsReturns, benchmarkReturns);
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-4), true, 'Minimum tracking error portfolio, #1');
	}  
	
	// Test with static data: 1 EEM ETF, 1 WLD ETF, and 1 ACWI ETF with returns halved as benchmark
	//
	// The only way to be able to minimize the tracking error is to invest partially, about half position of the test above.
	{
		// Define the assets returns	
		var assetsReturns = [[0.078129363,0.03663478,0.028787879,0.031491869,-0.019533386,-0.042308288,-0.062133001,-0.103910104,-0.027053038,0.082179315,0.004579357,-0.049970429], // Asset 1 returns in 2015 (EEM)
							[0.05289784,0.065017266,0.028012939,-0.019089773,0.025592817,-0.03893898,0.026564726,-0.07927768,-0.033253707,0.090633334,0.04075,-0.044859476], // Asset 2 returns in 2015 (WLD)
							[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]]; // Asset 3 return = 0 (cash)

		// Define the benchmark returns in 2015 (ACWI), and halve them
		var benchmarkReturns = [0.055518502,0.062009871,0.028003081,-0.013945023,0.020509392,-0.03935567,0.017029792,-0.081809115,-0.032750097,0.089738414,0.037148934,-0.045498531];
		for (var i = 0; i < benchmarkReturns.length; ++i) {
			benchmarkReturns[i] /= 2;
		}

		// Prior to any numerical computation, since ACWI index in 2015 ~= 90% World index + 10% Emerging index,
		// the minimum tracking error portfolio partially invested should have weights close to [0.1, 0.9]/2, which has been
		// validated by the numerical computation below.
		var expectedWeights = [0.05641314677088126, 0.44330002122510276, 0.5002868320040159];
		
		// Compute the weights of the minimum tracking error portfolio
		var weights = PortfolioAllocation.minimumTrackingErrorWeights(assetsReturns, benchmarkReturns);
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-4), true, 'Minimum tracking error portfolio, partial investment constraint #1');
	}

	
	// Test with static data for redundancy: 2 EEM ETFs, 2 WLD ETFs, 1 ACWI ETF as benchmark
	{
		// Define the assets returns
		var assetsReturns = [[0.078129363,0.03663478,0.028787879,0.031491869,-0.019533386,-0.042308288,-0.062133001,-0.103910104,-0.027053038,0.082179315,0.004579357,-0.049970429], // Asset 1 returns in 2015 (EEM)
		                     [0.05289784,0.065017266,0.028012939,-0.019089773,0.025592817,-0.03893898,0.026564726,-0.07927768,-0.033253707,0.090633334,0.04075,-0.044859476], // Asset 2 returns in 2015 (WLD)
							 [0.055323183,0.065000264,0.027985908,-0.019018198,0.025635979,-0.038860104,0.026504942,-0.079552638,-0.033599239,0.090526431,0.041004562,-0.044589974], // Asset 3 returns in 2015 (WLD #2)
							 [0.078085397,0.036893028,0.02912706,0.031861707,-0.019134434,-0.041778299,-0.0617737,-0.103650587,-0.026909091,0.082105488,0.005130735,-0.049278492]]; // Asset 4 returns in 2015 (EEM #2)
		
		// Define the benchmark returns in 2015 (ACWI)
		var benchmarkReturns = [0.055518502,0.062009871,0.028003081,-0.013945023,0.020509392,-0.03935567,0.017029792,-0.081809115,-0.032750097,0.089738414,0.037148934,-0.045498531];

		
		// The default numerical computation allocates evenly between the 2 WLD ETFs and the 2 EEM ETFs.
		var expectedWeightsDefault = [0.05152357319477277, 0.4486638440288671, 0.44666368775168186, 0.053148895024678164];
		
		// Compute the weights of the minimum tracking error portfolio
		var weights = PortfolioAllocation.minimumTrackingErrorWeights(assetsReturns, benchmarkReturns);
		assert.equal(areWeightsIdentical(weights, expectedWeightsDefault, 1e-4), true, 'Minimum tracking error portfolio, no cardinality constraints, duplicate assets');
		
		
		// The weights constrained numerical computation allocates between the 2 WLD ETFs and the 2 EEM ETFs according to the weights constraints.
		var expectedWeightsMinMaxWeights = [0.25, 0.3268282983256891, 0.32317170167431086, 0.1];
		
		// Compute the weights of the minimum tracking error portfolio
		var weights = PortfolioAllocation.minimumTrackingErrorWeights(assetsReturns, benchmarkReturns, {constraints: { minWeights: [0.25, 0, 0, 0.1], maxWeights: [1, 0.5, 1, 1] }});
		assert.equal(areWeightsIdentical(weights, expectedWeightsMinMaxWeights, 1e-4), true, 'Minimum tracking error portfolio, weights constraints, duplicate assets #1');
		
		
		// Note: all the expected cardinality constrained weights below have been validated with a combinatorial optimization
		
		// The 2-2 cardinality constrained numerical computation must allocate 1 WLD ETF and 1 EEM ETF,
		// and whatever the optimization method, the result should be approximately identical.
		var expectedWeightsTwoTwo = [0.10614062381555761, 0.8938593761844424, 0, 0];
																											
		// Compute the weights of the minimum tracking error portfolio
		var weights = PortfolioAllocation.minimumTrackingErrorWeights(assetsReturns, benchmarkReturns, { constraints: { minNbAssets: 2, maxNbAssets: 2 }});
		assert.equal(areWeightsIdentical(weights, expectedWeightsTwoTwo, 5e-3), true, 'Minimum tracking error portfolio, cardinality constraints, duplicate assets #1');

		
		// The 2-2 cardinality constrained numerical computation allocates to the same EEM ETF as above, 
		// but to the other WLD ETF in case maximum weights constraints are additionally imposed on the first WLD ETF.
		var expectedWeightsMaxTwoTwo = [0.10292768040365563, 0, 0.8970723195963444, 0];
				
		// Compute the weights of the minimum tracking error portfolio
		var weights = PortfolioAllocation.minimumTrackingErrorWeights(assetsReturns, benchmarkReturns, {constraints: { maxWeights: [1,0.25,1,1], minNbAssets: 2, maxNbAssets: 2 }});
		assert.equal(areWeightsIdentical(weights, expectedWeightsMaxTwoTwo, 1e-3), true, 'Minimum tracking error portfolio, cardinality and max weights constraints, duplicate assets #1');
				
		
		// The 2-3 cardinality constrained numerical computation allocates to the same EEM ETF as above, 
		// but dispatch to the two WLD ETFs in case maximum weights constraints are additionally imposed on the first WLD ETF.
		var expectedWeightsMaxTwoThree = [0.09775016142607804, 0.2441086935007033, 0.6581411450732186, 0];
		
		// Compute the weights of the minimum tracking error portfolio
		var weights = PortfolioAllocation.minimumTrackingErrorWeights(assetsReturns, benchmarkReturns, {constraints: { maxWeights: [1,0.25,1,1], minNbAssets: 2, maxNbAssets: 3 }});
		assert.equal(areWeightsIdentical(weights, expectedWeightsMaxTwoThree, 5e-2), true, 'Minimum tracking error portfolio, relaxed cardinality and max weights constraints, duplicate assets #1');
		
																										
		// The exact 2-2 cardinality constrained with min weights numerical computation allocates 
		// to a different WLD ETF than above in case the min weight constraint would force
		// the ideal WLD ETF to have too much weight (here, 95% v.s. its ideal allocation of ~90%)
		var expectedWeightsMinTwoTwo = [0, 0, 0.8, 0.2];
			
		// Compute the weights of the minimum tracking error portfolio
		var weights = PortfolioAllocation.minimumTrackingErrorWeights(assetsReturns, benchmarkReturns, {constraints: { minWeights: [0.25, 0.95, 0.1, 0.2], minNbAssets: 2, maxNbAssets: 2 }});
		assert.equal(areWeightsIdentical(weights, expectedWeightsMinTwoTwo, 1e-4), true, 'Minimum tracking error portfolio, cardinality and min weights constraints, duplicate assets #1');

		
		// Infeasible 2-2 cardinality constrained computation
		assert.throws(function() { 
			PortfolioAllocation.minimumTrackingErrorWeights(assetsReturns, benchmarkReturns, {constraints: { minWeights: [0.6, 0.6, 0.6, 0.6], minNbAssets: 2, maxNbAssets: 2 }}); },
			new Error('infeasible problem detected'),
			"Minimum tracking error portfolio, infeasible cardinality constraints, duplicate assets #1");
		
		
		// The 1-4 cardinality constrained numerical computation allocates the same as the 2-2.
		var expectedWeightsOneFour = expectedWeightsTwoTwo;
	
		// Compute the weights of the minimum tracking error portfolio
		var weights = PortfolioAllocation.minimumTrackingErrorWeights(assetsReturns, benchmarkReturns, {constraints: { minNbAssets: 1, maxNbAssets: 4 }});
		assert.equal(areWeightsIdentical(weights, expectedWeightsOneFour, 1e-3), true, 'Minimum tracking error portfolio, cardinality constraints, duplicate assets #5');
	} 

	// Test cardinality constrained, with partial investment constraint
	{
		// Define the assets returns
		var assetsReturns = [[0.078129363,0.03663478,0.028787879,0.031491869,-0.019533386,-0.042308288,-0.062133001,-0.103910104,-0.027053038,0.082179315,0.004579357,-0.049970429], // Asset 1 returns in 2015 (EEM)
		                     [0.05289784,0.065017266,0.028012939,-0.019089773,0.025592817,-0.03893898,0.026564726,-0.07927768,-0.033253707,0.090633334,0.04075,-0.044859476], // Asset 2 returns in 2015 (WLD)
							 [0.055323183,0.065000264,0.027985908,-0.019018198,0.025635979,-0.038860104,0.026504942,-0.079552638,-0.033599239,0.090526431,0.041004562,-0.044589974], // Asset 3 returns in 2015 (WLD #2)
							 [0.078085397,0.036893028,0.02912706,0.031861707,-0.019134434,-0.041778299,-0.0617737,-0.103650587,-0.026909091,0.082105488,0.005130735,-0.049278492]]; // Asset 4 returns in 2015 (EEM #2)
		
		// Define the benchmark returns in 2015 (ACWI), and halve them
		var benchmarkReturns = [0.055518502,0.062009871,0.028003081,-0.013945023,0.020509392,-0.03935567,0.017029792,-0.081809115,-0.032750097,0.089738414,0.037148934,-0.045498531];
		for (var i = 0; i < benchmarkReturns.length; ++i) {
			benchmarkReturns[i] /= 2;
		}
		
		// The 1-4 cardinality constrained numerical computation with partial investment constraint should allocates half the quantity of the 1-4 with 
		// full investment constraint, but possibly different assets, which is the case.
		var expectedWeights = [0.024819829894880664, 0.4490162021223922, 0, 0.026641220660455277];
				
		// Compute the weights of the minimum tracking error portfolio using Threshold Accepting
		var weights = PortfolioAllocation.minimumTrackingErrorWeights(assetsReturns, benchmarkReturns, { constraints: { fullInvestment: false, minNbAssets: 1, maxNbAssets: 4 }});
																										 
	    // The only test which can be done due to the stochasticity is that the portfolio size should be less than ~50% invested
		var weightsOK = true;
		if (expectedWeights.length != weights.length) {
			weightsOK = false;
		}
		else {
			var sumWeights = 0;
			for (var i = 0; i < expectedWeights.length; ++i) {
				sumWeights += weights[i];
			}
		}
		assert.equal(sumWeights <= 0.51, true, 'Minimum tracking error portfolio, cardinality constraints and partial investment constraint, duplicate assets #1');
	}
	


	// INDTRACK1 (Hang Seng Index and Constituents)
	// Reference: OR-library http://people.brunel.ac.uk/~mastjjb/jeb/orlib/indtrackinfo.html
	{
		// Define the index prices
		var indexPrices = [8749.31759356, 8713.53262793, 9001.60515134, 8903.30299873, 8888.69435453, 8727.37800153, 8735.20596277, 9022.01820217, 9461.30705604, 9734.73543429, 9954.80587273, 10125.76073581, 10401.21266867, 10792.5574788, 10713.83410451, 10329.12797352, 10273.26721616, 10851.04530705, 10760.07410226, 10617.20049693, 10877.60002361, 10246.92550535, 10439.28744609, 10385.68099956, 10335.37614212, 9779.43114037, 9991.03460033, 10131.10362998, 9828.95497674, 10098.81550871, 10093.20635736, 9932.17401202, 9934.42832285, 10624.24743709, 11116.30846497, 10988.77127348, 11209.4452282, 11300.94894735, 10434.05105479, 10626.62600127, 9351.12983299, 9324.64611835, 9217.22021955, 9659.82841299, 9784.75628407, 9815.23385648, 10422.72624919, 10498.32553918, 10209.01048224, 10201.57303155, 10398.49684538, 10952.20528678, 11275.08649948, 11542.7992235, 10952.75555162, 10807.09512109, 11256.02248505, 11368.18776174, 11157.36532286, 11887.76203205, 11983.2418585, 12124.48241977, 12067.37912888, 12434.72529143, 12727.02242884, 13131.64459726, 12704.88758155, 12898.81156444, 12781.67615363, 12450.32796246, 12812.17147652, 12550.67142015, 12292.59720621, 11982.15907928, 12405.73875925, 13129.62104265, 13121.59782615, 13393.37540986, 13176.14504807, 13334.3905682, 13478.73746329, 13167.5005648, 13357.32418705, 13625.65817783, 14210.25245265, 15556.48427968, 15477.21064054, 16559.58159748, 15969.98168759, 17220.18342331, 16443.93722691, 16462.53972888, 16497.9164335, 18155.385159, 18760.23273035, 19596.17378757, 21102.46168358, 19528.01194828, 19124.72106578, 20340.93063426, 20196.19322863, 21580.26907684, 20420.20427339, 19216.45554117, 17928.42753473, 17605.26231437, 17583.0209642, 16210.28767206, 16391.16505289, 16028.52276727, 16504.78586886, 16926.96326099, 16253.84734748, 15915.20370944, 15301.39214621, 16214.56553749, 17096.60458863, 16809.93435306, 16391.27155577, 16172.74540901, 16177.71554312, 15764.20038474, 15326.42032158, 14968.90792439, 16183.14718969, 16246.99566259, 16832.44196041, 17044.38267946, 16800.04733626, 16693.33145662, 16683.81719988, 17575.74326782, 18007.89642914, 17694.60047509, 17098.09562887, 16900.65705114, 16481.1067299, 16953.35822324, 16576.44455252, 16649.0085106, 16916.91648989, 16628.38245402, 16734.15755828, 15369.83799317, 14593.68054917, 13825.97233315, 14495.73340614, 14753.98512486, 14539.48833683, 13638.13676456, 12873.2508747, 12918.97610856, 12952.73751959, 13275.4412275, 14223.1393004, 14276.72799645, 14589.02992367, 14529.03330471, 14110.54801221, 15149.51904804, 15060.67790073, 15243.61433712, 15035.01070812, 15375.44714453, 15345.98134942, 14841.22872906, 14790.00084672, 16362.08976832, 15999.07472264, 16559.74135179, 16968.99639522, 16450.46940317, 16532.7073723, 16373.68083109, 16342.04947755, 17097.70511833, 17267.59495267, 16703.60898395, 16777.20246981, 16619.47171357, 15987.73216657, 15790.50659459, 16118.12718518, 16324.17474521, 16673.39766872, 17390.85427873, 16937.27628928, 17122.71554322, 17526.64544295, 17544.18291619, 17564.5249651, 17183.7949414, 17494.51707599, 16706.48456154, 16486.46737455, 16842.98574492, 17506.49864931, 17508.38020008, 17499.5759625, 17630.13073542, 17880.74974818, 18691.07686423, 18709.02259848, 19106.77533153, 19724.1014896, 20358.73436468, 20076.2887431, 20581.66263022, 20218.52333119, 19871.55471851, 19912.11456299, 18740.21019006, 19572.97391154, 19449.55483117, 19773.82058123, 19258.91468689, 19203.3379372, 19051.21633232, 19053.79015177, 18811.47836317, 19200.42685865, 19559.53679895, 19995.43531133, 19874.41254563, 19285.87766447, 19268.65969986, 19562.62538229, 19839.94111544, 19175.27442993, 19250.92697135, 19002.89952852, 19458.02180965, 19710.18511408, 19837.89981036, 20279.28322075, 19807.794998, 19570.95035694, 20180.5905576, 20576.99425425, 20873.4805047, 21132.8505036, 21688.2452405, 22205.93796003, 21989.96788225, 22240.05438064, 22633.91975879, 22879.24912881, 23282.93052184, 23774.86729638, 23257.97334839, 22692.33658512, 23308.88172212, 23792.99053542, 23471.08559906, 23415.54435033, 24595.7736979, 23749.3421076, 23646.81534099, 24248.04181463, 23276.66460276, 23865.25273536, 23783.36977581, 23674.43508629, 22607.95080804, 22169.10571615, 22249.01837253, 21663.73182902, 22217.56452377, 22261.19520111, 22446.82971031, 23220.64409109, 24727.83726153, 24961.38031351, 25439.418463, 26195.8196239, 26013.5577057, 25050.45221705, 26899.71486797, 26975.03015029, 26311.48174495, 27025.61901539, 27638.20579557, 27793.91299721, 29073.90003667, 29550.18088874, 28572.73301307, 27388.54530868, 25090.74580434, 25850.99881918, 25685.75961033, 25532.51972527];
		var indexReturns = new Array(indexPrices.length - 1);
		for (var i = 1; i < indexPrices.length; ++i) {
			indexReturns[i-1] = (indexPrices[i] - indexPrices[i-1])/indexPrices[i-1];
		}
		
		// Define the 31 assets prices
		var assetsPrices = [[9.33675195, 9.86926631, 9.95801871, 9.15924716, 9.24799955, 8.98174237, 9.07049476, 8.73323566, 9.24799955, 9.33675195, 10.93429505, 10.93429505, 10.93429505, 11.37805703, 12.07032571, 11.44905894, 11.37805703, 11.53781134, 11.11179984, 10.93429505, 11.02304745, 10.31302829, 10.0467711, 10.0467711, 10.31302829, 9.51425673, 9.69176152, 9.78051392, 9.60300913, 10.75679026, 10.93429505, 10.49053308, 10.57928547, 11.28930463, 11.37805703, 10.93429505, 11.37805703, 12.1590781, 10.93429505, 11.28930463, 9.33675195, 8.82198805, 8.89298997, 9.51425673, 9.69176152, 10.0467711, 10.66803787, 11.02304745, 10.1355235, 10.40178068, 10.40178068, 10.57928547, 11.20055224, 11.44905894, 11.02304745, 11.02304745, 11.71531613, 11.89282092, 11.37805703, 11.89282092, 11.89282092, 12.51408768, 13.13535445, 12.78034487, 12.60284008, 13.04660205, 12.60284008, 13.13535445, 12.95784966, 12.95784966, 12.95784966, 12.78034487, 12.1590781, 11.44905894, 11.89282092, 12.42533529, 11.89282092, 11.62656373, 12.69159247, 12.69159247, 12.78034487, 12.86909726, 13.75662121, 13.66786882, 14.00512792, 15.24766145, 16.1351854, 18.33624479, 16.56119689, 18.15874, 18.15874, 17.62622563, 17.53747324, 19.38352305, 19.04626395, 21.67333484, 26.27070889, 21.15857095, 20.62605658, 22.91586837, 23.62588753, 25.38318495, 24.495661, 22.73836358, 22.02834442, 21.67333484, 21.15857095, 20.44855179, 19.91603742, 20.09354221, 20.44855179, 21.85083963, 20.98106616, 20.98106616, 20.09354221, 21.85083963, 21.85083963, 21.67333484, 21.67333484, 21.33607574, 20.98106616, 20.62605658, 20.89231376, 20.44855179, 20.71480897, 20.62605658, 20.80356137, 20.80356137, 20.80356137, 21.33607574, 20.1822946, 20.98106616, 21.67333484, 21.06981855, 20.44855179, 20.1822946, 19.56102784, 19.38352305, 18.60250197, 18.42499718, 19.11726586, 18.15874, 17.89248281, 15.86892821, 14.53764229, 13.57911642, 15.69142342, 15.42516624, 15.78017582, 14.62639468, 13.31285924, 13.13535445, 12.69159247, 13.31285924, 14.44888989, 14.53764229, 15.15890905, 15.24766145, 15.42516624, 15.60267103, 15.86892821, 16.046433, 16.64994929, 16.64994929, 16.56119689, 16.56119689, 16.1351854, 17.71497802, 17.27121605, 17.80373042, 18.24749239, 16.91620647, 17.18246366, 17.00495887, 16.82745408, 17.53747324, 17.44872084, 16.91620647, 17.18246366, 17.35996845, 16.49019498, 16.1351854, 16.1351854, 16.64994929, 17.09371126, 17.09371126, 16.91620647, 17.00495887, 17.18246366, 17.00495887, 17.00495887, 16.56119689, 17.35996845, 16.31269019, 16.046433, 15.95768061, 17.44872084, 18.15874, 18.95751155, 19.47227544, 19.73853263, 21.06981855, 20.44855179, 21.85083963, 22.02834442, 21.51358053, 21.06981855, 22.383354, 22.02834442, 21.67333484, 22.02834442, 21.06981855, 22.47210639, 22.2946016, 22.11709681, 22.20584921, 23.27087795, 23.27087795, 23.00462076, 22.82711597, 22.82711597, 23.44838274, 24.0696495, 23.62588753, 23.62588753, 23.35963034, 23.27087795, 23.53713513, 23.27087795, 23.27087795, 23.35963034, 23.27087795, 23.35963034, 23.18212555, 23.98089711, 23.53713513, 23.35963034, 24.0696495, 23.98089711, 24.31815621, 24.22940381, 24.67316579, 25.11692776, 24.4069086, 25.47193734, 24.58441339, 24.76191818, 25.47193734, 27.33573763, 26.53696608, 24.93942297, 25.73819453, 26.0932041, 25.02817537, 25.38318495, 26.27070889, 25.29443255, 25.29443255, 25.29443255, 24.22940381, 25.20568016, 24.22940381, 24.495661, 22.36560352, 21.12306999, 21.74433675, 20.9455652, 21.47807957, 20.50180323, 19.88053646, 20.50180323, 21.03431759, 21.30057478, 21.21182238, 21.30057478, 20.8568128, 21.38932717, 22.18809873, 21.92184154, 20.76806041, 20.9455652, 20.50180323, 19.34802209, 18.46049814, 19.1705173, 18.63800293, 17.57297419, 16.86295503, 17.04045982, 17.04045982, 17.12921222],
							[14.37788798, 14.18263271, 14.59089372, 15.74467486, 16.29493971, 15.6204215, 15.74467486, 16.98720839, 18.97526203, 18.76225628, 18.90426012, 18.97526203, 20.12904317, 22.08159585, 22.25910064, 21.92184154, 22.25910064, 24.31815621, 26.02220219, 25.68494309, 26.35946129, 24.14065142, 24.47791052, 23.96314663, 24.31815621, 23.28862843, 23.11112364, 23.80339232, 23.28862843, 23.44838274, 23.62588753, 23.44838274, 23.96314663, 25.1701792, 25.68494309, 24.65541531, 24.8329201, 24.8329201, 22.77386454, 24.14065142, 20.41305083, 20.71480897, 21.92184154, 22.59635975, 23.28862843, 23.96314663, 24.47791052, 23.96314663, 24.47791052, 24.65541531, 25.47193734, 25.27668207, 26.0932041, 27.31798715, 26.0932041, 25.47193734, 26.0932041, 26.0932041, 25.88019836, 27.93925392, 28.34751494, 29.37704272, 29.18178745, 29.18178745, 29.37704272, 31.43609828, 30.81483151, 31.22309253, 30.81483151, 29.78530373, 30.4065705, 29.78530373, 29.37704272, 28.75577595, 27.74399865, 28.15225967, 27.74399865, 28.15225967, 26.90972614, 26.71447087, 26.71447087, 26.28845937, 26.90972614, 26.71447087, 28.34751494, 29.78530373, 29.37704272, 31.22309253, 30.19356475, 32.25262031, 32.46562606, 34.09867013, 36.15772569, 39.03330328, 40.8793531, 41.50061986, 43.55967542, 45.60098051, 46.84351404, 45.19271949, 43.96793644, 46.026992, 44.78445847, 41.85562944, 35.18144934, 32.87388708, 34.15192156, 32.092866, 34.93294264, 33.63715767, 33.63715767, 34.66668545, 33.90341486, 33.38865097, 32.87388708, 34.15192156, 37.7552688, 38.51853939, 37.24050491, 36.21097713, 36.97424772, 36.21097713, 34.205173, 31.73785642, 34.31167587, 32.55437845, 32.98038995, 33.17564522, 32.7673842, 32.25262031, 31.54260115, 33.28214809, 36.15772569, 35.53645892, 34.205173, 33.90341486, 33.17564522, 34.205173, 33.90341486, 33.79691198, 34.82643976, 34.82643976, 35.12819791, 33.49515384, 31.8443593, 28.45401781, 29.99830948, 32.05736504, 31.73785642, 30.56632481, 28.86227883, 28.70252452, 24.14065142, 23.87439423, 26.19970698, 26.19970698, 25.82694692, 25.63169165, 24.70866674, 26.07545363, 27.05172997, 26.92747662, 26.12870506, 26.67896991, 26.57246704, 25.95120027, 25.40093542, 27.92150344, 27.79725009, 28.41851685, 28.66702356, 28.59602164, 28.84452835, 28.72027499, 28.72027499, 31.73785642, 31.98636313, 31.56035163, 32.28812127, 32.05736504, 30.93908487, 29.94505804, 30.56632481, 30.56632481, 31.06333822, 31.86210977, 30.19356475, 30.88583343, 32.11061648, 32.78513468, 33.65490815, 33.33539953, 33.90341486, 33.03364139, 33.28214809, 33.45965288, 34.6311845, 34.15192156, 33.83241294, 33.95666629, 34.205173, 35.44770653, 35.87371802, 36.97424772, 38.83804801, 40.68409783, 38.83804801, 40.93260453, 40.31133777, 40.68409783, 40.4533416, 38.0925279, 39.63681957, 39.19305759, 39.63681957, 38.74929562, 39.56581765, 38.02152598, 38.16352981, 37.4180097, 38.53628987, 39.49481574, 40.38233969, 42.08638567, 40.82610166, 40.82610166, 41.85562944, 42.08638567, 40.38233969, 40.96810549, 40.96810549, 41.57162178, 43.18691537, 41.19886172, 41.85562944, 41.48286938, 40.75509974, 41.78462753, 42.15738758, 41.71362561, 42.53014764, 43.41767159, 44.81995943, 44.6779556, 44.51820129, 46.4530035, 47.18077314, 49.39958301, 48.95582103, 50.00309929, 49.11557535, 49.55933732, 50.73086893, 50.28710696, 51.33438522, 55.02648485, 52.50591683, 51.61839288, 50.73086893, 49.55933732, 50.00309929, 50.14510313, 48.10379804, 46.23999775, 46.23999775, 46.68375973, 45.97374057, 46.59500733, 46.68375973, 46.77251212, 47.30502649, 50.05635073, 49.96759834, 50.14510313, 50.05635073, 51.56514145, 50.14510313, 53.25143695, 57.33404711, 56.44652317, 56.44652317, 56.80153275, 56.62402796, 56.97903754, 55.38149443, 55.55899922, 53.78395132, 49.08007439, 48.28130283, 48.8138172, 48.28130283],
							[7.36644878, 7.40194974, 7.77470979, 7.63270596, 7.49070213, 7.54395357, 7.66820692, 8.0232165, 8.66223374, 9.40775386, 9.23024907, 9.85151584, 9.76276344, 9.94026823, 9.67401105, 9.67401105, 9.31900147, 9.94026823, 9.85151584, 9.58525865, 9.76276344, 9.31900147, 9.31900147, 9.49650626, 9.49650626, 8.66223374, 9.05274428, 9.05274428, 9.05274428, 9.23024907, 8.82198805, 8.7864871, 8.66223374, 8.7864871, 9.40775386, 8.96399189, 8.73323566, 8.7864871, 7.98771554, 8.34272512, 7.27769638, 7.18894399, 7.27769638, 7.66820692, 7.66820692, 7.81021075, 8.34272512, 8.37822608, 7.93446411, 7.81021075, 7.98771554, 8.66223374, 8.96399189, 9.05274428, 8.25397273, 8.46697847, 8.6977347, 8.6977347, 8.16522033, 8.73323566, 8.87523949, 9.14149668, 8.82198805, 9.14149668, 9.14149668, 9.40775386, 8.82198805, 8.96399189, 8.60898231, 8.34272512, 8.28947369, 7.98771554, 7.72145836, 7.81021075, 8.16522033, 8.60898231, 8.34272512, 8.37822608, 8.66223374, 8.96399189, 9.40775386, 8.96399189, 9.58525865, 10.73903978, 11.53781134, 11.89282092, 11.62656373, 13.75662121, 13.49036403, 15.97543108, 14.82164995, 15.6204215, 15.35416432, 15.79792629, 15.35416432, 15.35416432, 17.92798377, 16.24168827, 15.26541193, 16.59669785, 15.7091739, 15.7091739, 14.82164995, 13.84537361, 13.04660205, 13.13535445, 12.78034487, 12.33658289, 13.31285924, 12.86909726, 12.86909726, 13.22410684, 12.69159247, 12.2478305, 11.62656373, 11.80406852, 12.07032571, 12.07032571, 11.89282092, 11.89282092, 12.07032571, 11.53781134, 11.18280176, 10.68578835, 12.01707427, 11.98157331, 12.78034487, 12.33658289, 12.21232954, 11.92832188, 11.80406852, 12.37208385, 12.90459822, 12.72709343, 12.86909726, 12.86909726, 13.04660205, 13.66786882, 13.31285924, 13.57911642, 13.934126, 13.75662121, 14.20038319, 12.60284008, 11.18280176, 10.73903978, 11.71531613, 11.62656373, 11.36030655, 10.56153499, 9.49650626, 9.76276344, 9.49650626, 10.02902063, 10.82779218, 11.09404936, 10.29527781, 10.29527781, 9.76276344, 9.67401105, 9.76276344, 9.67401105, 9.58525865, 9.58525865, 10.11777302, 9.76276344, 9.85151584, 11.36030655, 11.00529697, 11.36030655, 11.98157331, 11.89282092, 11.80406852, 11.53781134, 11.62656373, 11.80406852, 12.42533529, 11.89282092, 11.89282092, 11.44905894, 9.85151584, 10.11777302, 9.94026823, 9.76276344, 10.11777302, 11.00529697, 9.85151584, 9.31900147, 9.23024907, 9.40775386, 9.05274428, 8.73323566, 8.66223374, 8.46697847, 7.40194974, 7.66820692, 7.57945453, 7.40194974, 7.18894399, 7.13569255, 7.89896315, 7.81021075, 8.7864871, 8.87523949, 9.14149668, 9.23024907, 9.40775386, 10.4727826, 10.11777302, 10.02902063, 9.67401105, 8.6977347, 8.55573087, 7.98771554, 8.55573087, 8.16522033, 8.07646794, 8.21847177, 8.1119689, 7.98771554, 7.89896315, 7.98771554, 8.25397273, 7.98771554, 7.81021075, 7.72145836, 7.45520117, 7.57945453, 7.36644878, 7.63270596, 7.63270596, 7.63270596, 7.77470979, 7.63270596, 7.89896315, 7.33094782, 7.27769638, 7.40194974, 7.57945453, 7.57945453, 7.81021075, 8.7864871, 9.40775386, 9.49650626, 8.96399189, 8.82198805, 8.82198805, 9.23024907, 9.05274428, 8.87523949, 8.82198805, 8.82198805, 8.87523949, 8.73323566, 8.55573087, 8.52022991, 8.25397273, 8.28947369, 8.34272512, 7.72145836, 7.89896315, 8.21847177, 7.93446411, 7.63270596, 7.27769638, 7.27769638, 7.0114392, 7.22444495, 7.13569255, 7.18894399, 7.04694016, 7.18894399, 7.10019159, 7.13569255, 7.66820692, 7.36644878, 7.10019159, 7.63270596, 8.6977347, 7.77470979, 8.7864871, 9.14149668, 8.37822608, 8.60898231, 8.34272512, 8.82198805, 8.46697847, 8.21847177, 7.77470979, 8.28947369, 8.25397273],
							[29.46579511, 27.33573763, 28.75577595, 28.75577595, 27.868252, 26.62571847, 26.98072805, 27.51324242, 29.11078553, 30.88583343, 32.30587175, 31.24084301, 31.4183478, 35.32345318, 34.43592923, 32.66088133, 31.59585259, 32.83838612, 31.77335738, 30.88583343, 31.06333822, 28.40076637, 28.93328074, 28.22326158, 28.04575679, 27.15823284, 27.69074721, 27.868252, 26.80322326, 27.868252, 27.33573763, 26.44821368, 26.44821368, 27.51324242, 31.24084301, 29.46579511, 28.57827116, 29.82080469, 27.868252, 28.75577595, 25.38318495, 25.20568016, 24.85067058, 26.98072805, 26.80322326, 26.80322326, 29.28829032, 29.82080469, 28.93328074, 28.40076637, 28.93328074, 30.88583343, 31.4183478, 31.95086217, 28.93328074, 28.75577595, 30.17581427, 30.53082385, 29.82080469, 31.24084301, 32.83838612, 35.32345318, 36.21097713, 35.50095797, 36.92099628, 39.58356813, 37.09850107, 39.05105376, 37.63101544, 36.03347234, 38.16352981, 37.09850107, 35.85596755, 34.61343402, 37.27600586, 39.05105376, 39.76107292, 40.82610166, 40.29358729, 42.77865435, 44.02118788, 41.00360645, 41.53612082, 42.06863519, 44.02118788, 47.92629325, 48.10379804, 51.65389384, 48.8138172, 53.07393216, 55.02648485, 57.24529472, 55.47024682, 65.67677224, 67.45182014, 67.45182014, 68.33934408, 62.57043841, 61.68291447, 70.11439198, 71.44567791, 74.55201173, 68.78310606, 64.34548631, 59.46410459, 60.35162854, 58.13281867, 53.69519892, 56.80153275, 55.9140088, 57.24529472, 58.57658064, 58.13281867, 56.35777077, 51.12137947, 54.58272287, 59.02034262, 59.46410459, 56.80153275, 56.80153275, 56.35777077, 57.68905669, 55.38149443, 55.02648485, 59.46410459, 61.0616477, 63.01420039, 63.90172434, 65.14425787, 64.25673392, 63.54671476, 68.16183929, 67.80682971, 67.9843345, 66.20928661, 64.78924829, 63.36920997, 62.30418123, 60.35162854, 61.59416207, 62.48168602, 60.52913333, 61.77166686, 55.55899922, 52.363913, 51.65389384, 55.20398964, 56.80153275, 55.55899922, 51.38763666, 45.08621662, 46.86126451, 48.01504565, 49.70134115, 54.31646569, 55.73650401, 57.33404711, 56.09151359, 55.20398964, 59.64160938, 57.86656148, 60.52913333, 59.46410459, 61.41665728, 62.12667644, 59.64160938, 60.35162854, 68.16183929, 67.27431535, 70.29189677, 70.11439198, 64.6117435, 65.85427703, 65.85427703, 66.3867914, 70.46940156, 70.82441114, 69.04936324, 69.22686803, 68.16183929, 67.27431535, 66.03178182, 65.85427703, 66.74180098, 68.69435366, 73.13197341, 72.77696383, 74.37450694, 76.50456442, 75.61704047, 75.97205005, 74.55201173, 76.14955484, 73.66448778, 73.48698299, 74.55201173, 78.27961231, 78.81212668, 79.52214584, 82.0072129, 83.60475601, 89.19615689, 87.68736618, 86.44483265, 88.39738533, 88.75239491, 90.97120479, 89.63991886, 88.57489012, 88.04237576, 87.68736618, 80.40966979, 85.91231828, 86.62233744, 87.50986139, 86.26732786, 84.49227996, 84.49227996, 83.42725122, 82.53972727, 84.66978475, 85.73481349, 88.21988055, 87.68736618, 86.44483265, 85.73481349, 86.44483265, 86.44483265, 84.66978475, 84.66978475, 82.71723206, 83.60475601, 83.60475601, 83.24974643, 85.5573087, 83.07224164, 83.7822608, 86.62233744, 89.63991886, 90.97120479, 90.97120479, 94.07753861, 95.40882453, 94.52130058, 95.85258651, 101.1777302, 106.05911192, 107.39039785, 106.05911192, 104.28406402, 98.95892033, 106.94663587, 108.72168377, 103.84030205, 100.73396823, 109.16544574, 108.72168377, 103.84030205, 106.05911192, 99.84644428, 103.39654008, 104.727826, 103.84030205, 96.29634848, 97.62763441, 103.39654008, 98.07139638, 100.73396823, 101.62149218, 100.73396823, 104.28406402, 111.82801759, 109.16544574, 113.15930352, 114.49058944, 108.72168377, 107.83415982, 114.04682747, 118.92820919, 113.15930352, 118.48444721, 118.48444721, 123.80959091, 137.12245014, 143.33511779, 141.56006989, 135.79116422, 114.49058944, 126.02840078, 126.02840078, 123.80959091],
							[5.55589992, 5.44939705, 5.80440663, 5.50264848, 5.50264848, 5.44939705, 5.32514369, 5.32514369, 5.62690184, 5.71565423, 6.12391525, 6.336921, 6.78068297, 7.45520117, 7.33094782, 7.22444495, 7.27769638, 7.54395357, 7.36644878, 7.27769638, 7.18894399, 6.74518201, 6.69193058, 6.69193058, 6.74518201, 6.336921, 6.2481686, 6.21266764, 6.03516285, 6.12391525, 6.12391525, 5.80440663, 6.07066381, 6.69193058, 6.95818776, 6.95818776, 7.04694016, 6.95818776, 6.51442579, 6.78068297, 5.68015327, 5.94641046, 5.85765806, 5.89315902, 6.03516285, 5.9996619, 6.21266764, 6.44342387, 6.03516285, 6.2481686, 6.336921, 6.74518201, 7.22444495, 7.13569255, 6.9226868, 6.56767722, 6.74518201, 6.51442579, 6.83393441, 7.0114392, 6.88718585, 6.88718585, 6.88718585, 7.18894399, 8.1119689, 8.43147752, 8.25397273, 7.81021075, 7.72145836, 7.54395357, 7.49070213, 7.27769638, 6.88718585, 6.88718585, 7.0114392, 7.0114392, 7.10019159, 7.13569255, 7.18894399, 7.13569255, 7.10019159, 7.04694016, 7.04694016, 7.33094782, 7.36644878, 8.46697847, 9.40775386, 9.76276344, 9.23024907, 11.62656373, 11.44905894, 11.44905894, 11.80406852, 12.86909726, 12.78034487, 13.13535445, 15.6204215, 14.46664037, 13.40161163, 14.64414516, 14.64414516, 15.6204215, 14.82164995, 13.66786882, 12.2478305, 12.2478305, 12.1590781, 10.73903978, 11.80406852, 11.62656373, 11.71531613, 12.1590781, 10.91654457, 10.20652542, 9.49650626, 10.11777302, 11.09404936, 10.73903978, 10.65028739, 10.73903978, 10.73903978, 11.00529697, 10.33077877, 10.29527781, 11.04079793, 11.44905894, 11.92832188, 12.01707427, 11.98157331, 12.10582667, 11.62656373, 12.33658289, 13.17085541, 12.95784966, 13.04660205, 12.01707427, 11.62656373, 11.98157331, 11.71531613, 12.07032571, 12.2478305, 11.80406852, 11.53781134, 9.94026823, 8.82198805, 8.1119689, 9.31900147, 9.23024907, 9.05274428, 8.43147752, 7.84571171, 7.54395357, 7.40194974, 7.54395357, 8.82198805, 9.14149668, 9.40775386, 9.31900147, 8.73323566, 9.40775386, 9.58525865, 10.38403021, 9.85151584, 10.20652542, 10.38403021, 9.76276344, 9.76276344, 11.00529697, 10.91654457, 11.09404936, 11.44905894, 11.36030655, 11.09404936, 11.18280176, 11.27155415, 11.98157331, 11.98157331, 11.44905894, 11.53781134, 11.27155415, 10.73903978, 10.73903978, 10.82779218, 10.38403021, 11.00529697, 11.71531613, 11.09404936, 11.44905894, 11.98157331, 11.89282092, 12.07032571, 11.80406852, 11.18280176, 10.82779218, 10.82779218, 11.36030655, 11.62656373, 11.36030655, 11.36030655, 11.18280176, 11.27155415, 12.33658289, 12.33658289, 12.78034487, 12.78034487, 12.86909726, 12.69159247, 13.13535445, 12.78034487, 12.78034487, 13.04660205, 12.42533529, 12.78034487, 12.60284008, 12.60284008, 12.2478305, 12.33658289, 12.07032571, 11.62656373, 11.36030655, 11.44905894, 11.80406852, 12.1590781, 11.98157331, 11.89282092, 11.98157331, 12.33658289, 12.42533529, 12.95784966, 13.22410684, 12.78034487, 13.04660205, 13.22410684, 13.934126, 14.99915474, 14.73289756, 14.46664037, 14.37788798, 14.91040235, 15.53166911, 15.35416432, 15.6204215, 15.6204215, 15.53166911, 14.99915474, 14.99915474, 14.82164995, 16.68545024, 16.68545024, 16.33044066, 16.06418348, 15.97543108, 16.50794545, 15.88667869, 15.6204215, 16.59669785, 16.33044066, 15.88667869, 15.97543108, 15.6204215, 15.7091739, 15.44291672, 15.53166911, 14.91040235, 13.40161163, 14.20038319, 13.31285924, 13.84537361, 13.75662121, 14.0228784, 15.44291672, 15.6204215, 15.6204215, 15.35416432, 16.06418348, 15.53166911, 14.28913558, 15.7091739, 15.17665953, 14.73289756, 14.99915474, 15.35416432, 15.17665953, 15.97543108, 15.7091739, 16.06418348, 15.88667869, 14.55539277, 15.53166911, 14.99915474, 15.97543108],
							[8.55573087, 8.50247943, 8.73323566, 8.85748901, 9.21249859, 9.15924716, 9.05274428, 8.76873662, 9.56750817, 9.709512, 9.78051392, 9.47875578, 9.99351967, 9.78051392, 9.78051392, 9.40775386, 9.40775386, 9.709512, 9.56750817, 9.47875578, 9.78051392, 9.33675195, 9.85151584, 9.63851009, 9.40775386, 9.33675195, 9.40775386, 9.709512, 9.40775386, 9.63851009, 9.56750817, 9.33675195, 9.19474811, 10.34852925, 10.84554266, 11.20055224, 11.83956948, 11.76856757, 10.84554266, 11.20055224, 10.06452158, 10.27752733, 10.49053308, 10.91654457, 11.20055224, 11.27155415, 11.12955032, 11.55556182, 11.05854841, 10.77454074, 11.34255607, 11.55556182, 12.26558098, 12.12357715, 11.69756565, 11.4845599, 11.76856757, 11.76856757, 12.05257523, 12.62059056, 12.40758481, 12.40758481, 12.1590781, 12.07032571, 12.40758481, 12.40758481, 12.5850896, 12.40758481, 12.31883241, 12.07032571, 12.2478305, 12.07032571, 11.89282092, 11.55556182, 11.73306661, 12.40758481, 12.2478305, 12.2478305, 12.1590781, 12.67384199, 12.5850896, 12.40758481, 12.67384199, 12.67384199, 12.9223487, 14.12938127, 15.15890905, 16.09968444, 15.22991097, 17.37771892, 17.11146174, 16.61444833, 16.61444833, 17.46647132, 17.28896653, 17.62622563, 19.1705173, 18.8332582, 17.62622563, 18.65575341, 19.2592697, 19.43677449, 18.56700102, 17.46647132, 16.77420264, 16.34819114, 16.52569593, 15.49616815, 15.58492055, 16.52569593, 16.33044066, 17.35996845, 17.46647132, 17.0582103, 15.40741576, 16.63219881, 17.1469627, 16.54344641, 16.33044066, 15.92217965, 15.51391863, 15.19441001, 15.10565761, 13.66786882, 14.32463654, 14.64414516, 14.9459033, 14.6973966, 14.6973966, 14.64414516, 14.53764229, 14.99915474, 15.97543108, 15.56717007, 14.89265187, 14.89265187, 14.59089372, 14.83940043, 14.37788798, 14.43113941, 14.59089372, 14.48439085, 14.32463654, 13.66786882, 12.74484391, 11.30705511, 12.69159247, 13.29510876, 12.8335963, 12.17682858, 12.12357715, 11.76856757, 11.66206469, 12.28333146, 13.3483602, 13.2596078, 13.72112025, 14.21813367, 13.80987265, 14.21813367, 14.07612983, 14.18263271, 13.80987265, 13.98737744, 13.80987265, 13.86312409, 13.61461738, 15.15890905, 15.10565761, 15.53166911, 15.7091739, 15.22991097, 15.2831624, 15.35416432, 15.22991097, 15.35416432, 15.65592246, 15.2831624, 15.22991097, 15.0346557, 14.66189564, 14.66189564, 14.73289756, 15.10565761, 15.15890905, 15.65592246, 15.35416432, 15.35416432, 15.58492055, 15.58492055, 15.58492055, 15.35416432, 15.35416432, 15.35416432, 15.2831624, 15.2831624, 15.65592246, 15.53166911, 15.65592246, 15.53166911, 15.35416432, 15.78017582, 16.57894737, 16.95170743, 16.63219881, 17.69722755, 17.87473234, 19.09951539, 18.17649048, 18.79775724, 18.79775724, 18.05223713, 18.62025245, 18.79775724, 19.04626395, 18.85100868, 18.63800293, 18.26524287, 18.26524287, 17.75047898, 17.96348473, 17.75047898, 18.19424096, 18.19424096, 17.89248281, 18.12323904, 18.26524287, 18.19424096, 17.8214809, 18.19424096, 18.19424096, 18.33624479, 18.40724671, 18.19424096, 19.01076299, 18.4959991, 18.93976107, 19.22376874, 19.29477065, 19.22376874, 19.22376874, 19.38352305, 19.82728502, 19.66753071, 20.11129269, 21.06981855, 21.30057478, 22.40110448, 22.70286262, 21.88634059, 21.60233292, 22.11709681, 22.18809873, 21.88634059, 21.88634059, 22.11709681, 21.44257861, 21.60233292, 22.0460949, 21.30057478, 21.60233292, 21.74433675, 22.18809873, 21.81533867, 21.67333484, 21.51358053, 21.15857095, 21.67333484, 22.18809873, 22.01059394, 21.83308915, 22.45435591, 22.18809873, 22.09934633, 23.96314663, 24.31815621, 25.47193734, 27.42449003, 27.51324242, 27.15823284, 26.0932041, 28.66702356, 28.31201398, 29.28829032, 29.55454751, 29.6432999, 29.11078553, 25.56068974, 27.33573763, 27.42449003, 26.98072805],
							[16.41919306, 16.41919306, 16.59669785, 16.59669785, 16.77420264, 16.50794545, 15.79792629, 16.50794545, 17.3954694, 18.28299335, 20.05804125, 21.83308915, 23.96314663, 24.495661, 24.67316579, 23.43063226, 23.07562268, 25.56068974, 24.85067058, 24.85067058, 24.495661, 23.78564184, 22.89811789, 22.18809873, 22.36560352, 20.59055562, 20.76806041, 22.54310831, 22.01059394, 23.78564184, 23.78564184, 23.96314663, 23.78564184, 25.20568016, 26.27070889, 27.33573763, 26.80322326, 26.98072805, 25.38318495, 25.20568016, 22.01059394, 22.01059394, 21.12306999, 22.54310831, 22.01059394, 21.83308915, 23.43063226, 24.495661, 23.07562268, 23.78564184, 24.85067058, 26.44821368, 27.33573763, 27.33573763, 25.56068974, 24.14065142, 25.73819453, 25.73819453, 24.31815621, 25.91569932, 26.0932041, 25.73819453, 26.80322326, 29.28829032, 29.11078553, 29.6432999, 29.11078553, 29.6432999, 29.6432999, 28.75577595, 28.75577595, 29.11078553, 28.57827116, 27.51324242, 28.22326158, 29.99830948, 28.40076637, 29.46579511, 28.75577595, 29.46579511, 30.35331906, 29.11078553, 30.88583343, 32.12836696, 32.83838612, 35.67846276, 36.03347234, 39.76107292, 39.93857771, 42.24613998, 38.51853939, 36.5659867, 38.87354897, 45.26372141, 46.68375973, 49.70134115, 54.58272287, 47.39377888, 45.08621662, 48.99132199, 49.70134115, 55.47024682, 51.65389384, 48.8138172, 44.90871183, 44.19869267, 46.15124536, 44.19869267, 44.90871183, 43.84368309, 43.84368309, 43.84368309, 41.53612082, 40.1160825, 38.3410346, 42.60114956, 44.37619746, 41.53612082, 39.22855855, 37.63101544, 37.80852023, 37.80852023, 36.29972952, 35.05719599, 38.78479658, 38.3410346, 39.31731095, 41.35861603, 40.64859687, 40.47109208, 38.96230137, 41.44736843, 44.10994027, 42.24613998, 40.73734927, 38.3410346, 36.5659867, 36.5659867, 35.05719599, 35.50095797, 35.76721515, 35.05719599, 34.9684436, 31.50710019, 28.13450919, 25.47193734, 27.06948045, 27.24698524, 27.24698524, 25.29443255, 24.58441339, 24.58441339, 24.495661, 24.67316579, 26.89197566, 27.51324242, 27.60199482, 27.69074721, 26.80322326, 29.02203314, 31.24084301, 30.53082385, 29.90955709, 29.90955709, 30.08706188, 29.28829032, 28.75577595, 33.54840528, 32.39462414, 32.39462414, 32.66088133, 33.01589091, 32.21711935, 31.59585259, 31.4183478, 33.1933957, 33.54840528, 33.90341486, 33.45965288, 32.92713851, 29.6432999, 28.75577595, 29.28829032, 28.75577595, 30.26456667, 32.21711935, 31.77335738, 32.92713851, 33.90341486, 35.32345318, 36.29972952, 35.32345318, 35.14594839, 33.72591007, 33.45965288, 33.28214809, 35.50095797, 35.67846276, 34.79093881, 35.94471994, 36.29972952, 38.87354897, 38.07477742, 38.96230137, 41.35861603, 42.77865435, 41.802378, 43.39992111, 41.97988279, 40.55984448, 42.33489237, 39.31731095, 44.46494985, 44.28744506, 44.28744506, 42.15738758, 42.86740674, 42.60114956, 42.24613998, 40.55984448, 41.71362561, 43.04491153, 44.28744506, 43.39992111, 41.71362561, 42.42364477, 42.06863519, 41.8911304, 39.22855855, 39.22855855, 38.16352981, 40.29358729, 41.09235885, 41.71362561, 44.02118788, 42.33489237, 41.00360645, 41.35861603, 41.97988279, 43.04491153, 43.22241632, 43.5774259, 44.19869267, 42.95615914, 45.26372141, 47.39377888, 48.72506481, 49.96759834, 52.45266539, 50.85512229, 51.03262708, 51.21013187, 52.89642737, 52.54141779, 51.74264624, 51.83139863, 51.21013187, 50.05635073, 50.05635073, 48.01504565, 48.54756002, 47.39377888, 46.41750254, 42.77865435, 40.29358729, 41.18111124, 40.47109208, 40.73734927, 39.67232053, 37.89727263, 40.2048349, 41.8911304, 41.8911304, 44.73120704, 45.88498817, 43.48867351, 39.93857771, 42.06863519, 40.55984448, 39.49481574, 41.35861603, 41.71362561, 42.68990195, 44.64245464, 47.48253128, 45.08621662, 42.51239716, 39.93857771, 41.53612082, 41.53612082, 41.35861603],
							[9.95801871, 9.28350051, 8.85748901, 8.76873662, 8.87523949, 8.6977347, 8.43147752, 8.66223374, 9.14149668, 9.14149668, 9.67401105, 9.58525865, 10.56153499, 11.27155415, 10.91654457, 10.73903978, 10.4727826, 11.00529697, 10.65028739, 10.82779218, 10.65028739, 10.02902063, 10.38403021, 10.65028739, 10.38403021, 10.02902063, 10.11777302, 10.11777302, 10.20652542, 10.73903978, 11.09404936, 11.18280176, 11.89282092, 12.51408768, 13.13535445, 13.04660205, 12.86909726, 13.04660205, 11.62656373, 12.1590781, 11.09404936, 10.29527781, 10.20652542, 10.73903978, 11.00529697, 11.18280176, 12.33658289, 11.98157331, 11.36030655, 11.36030655, 11.89282092, 12.60284008, 13.934126, 13.22410684, 12.33658289, 12.78034487, 13.22410684, 13.13535445, 12.86909726, 13.66786882, 13.13535445, 13.84537361, 13.31285924, 13.84537361, 14.0228784, 14.82164995, 14.46664037, 14.37788798, 14.20038319, 13.40161163, 13.40161163, 13.04660205, 12.78034487, 13.04660205, 13.40161163, 14.20038319, 13.75662121, 13.57911642, 14.11163079, 14.82164995, 15.08790714, 14.37788798, 15.08790714, 15.44291672, 15.79792629, 16.33044066, 16.59669785, 17.75047898, 17.4842218, 17.92798377, 17.30671701, 17.66172659, 17.75047898, 18.81550772, 19.88053646, 21.12306999, 22.7206131, 21.12306999, 21.47807957, 22.54310831, 22.36560352, 23.78564184, 22.7206131, 21.83308915, 20.59055562, 19.52552688, 20.05804125, 18.81550772, 19.34802209, 18.81550772, 19.1705173, 18.99301251, 17.92798377, 18.10548856, 17.57297419, 19.52552688, 19.88053646, 18.99301251, 18.28299335, 17.75047898, 17.21796461, 16.50794545, 15.92217965, 15.38966528, 16.27718923, 17.16471318, 17.4842218, 17.53747324, 17.16471318, 16.86295503, 15.92217965, 16.54344641, 18.46049814, 18.10548856, 18.46049814, 17.60847515, 17.21796461, 17.66172659, 16.95170743, 16.77420264, 16.77420264, 16.68545024, 16.06418348, 14.11163079, 13.04660205, 11.80406852, 13.04660205, 12.69159247, 12.42533529, 11.89282092, 11.62656373, 11.62656373, 11.98157331, 11.98157331, 12.60284008, 12.60284008, 12.60284008, 12.86909726, 12.1590781, 12.78034487, 12.78034487, 13.13535445, 11.89282092, 12.33658289, 12.07032571, 11.44905894, 11.36030655, 12.86909726, 12.42533529, 12.42533529, 12.51408768, 12.60284008, 12.33658289, 12.1590781, 12.07032571, 13.04660205, 13.66786882, 12.78034487, 12.86909726, 12.86909726, 12.2478305, 11.98157331, 12.2478305, 12.2478305, 12.51408768, 13.66786882, 12.95784966, 12.95784966, 13.40161163, 13.66786882, 13.75662121, 13.40161163, 13.66786882, 12.78034487, 12.69159247, 12.60284008, 13.49036403, 13.49036403, 13.49036403, 13.84537361, 13.66786882, 14.55539277, 15.08790714, 15.08790714, 16.06418348, 17.3954694, 16.77420264, 16.77420264, 16.50794545, 16.15293587, 16.77420264, 15.6204215, 16.15293587, 16.06418348, 15.97543108, 15.26541193, 15.26541193, 15.35416432, 15.26541193, 14.73289756, 15.44291672, 16.59669785, 17.04045982, 16.68545024, 16.50794545, 16.33044066, 16.59669785, 16.77420264, 16.33044066, 15.79792629, 15.6204215, 16.06418348, 16.06418348, 16.15293587, 16.50794545, 15.97543108, 15.44291672, 16.06418348, 16.15293587, 15.7091739, 16.24168827, 16.77420264, 17.04045982, 16.77420264, 16.86295503, 17.04045982, 18.28299335, 19.70303167, 19.1705173, 19.08176491, 18.90426012, 19.1705173, 19.96928886, 19.52552688, 19.2592697, 19.08176491, 18.01673617, 17.4842218, 17.75047898, 17.57297419, 17.04045982, 16.41919306, 16.41919306, 15.44291672, 14.91040235, 14.55539277, 14.20038319, 13.84537361, 13.66786882, 13.40161163, 14.0228784, 14.73289756, 14.91040235, 15.79792629, 15.6204215, 14.82164995, 13.84537361, 14.82164995, 15.17665953, 14.28913558, 14.73289756, 14.28913558, 14.37788798, 15.79792629, 17.21796461, 15.88667869, 15.26541193, 13.84537361, 13.84537361, 13.66786882, 14.55539277],
							[5.85765806, 6.2481686, 6.51442579, 6.30142004, 6.30142004, 6.39017243, 6.21266764, 6.39017243, 6.39017243, 6.69193058, 6.74518201, 7.04694016, 7.13569255, 7.04694016, 7.22444495, 6.78068297, 6.83393441, 7.36644878, 7.63270596, 7.27769638, 7.54395357, 7.36644878, 7.27769638, 7.45520117, 7.57945453, 7.22444495, 7.10019159, 7.0114392, 6.9226868, 7.13569255, 7.36644878, 7.45520117, 7.10019159, 7.36644878, 7.36644878, 7.33094782, 7.33094782, 7.36644878, 7.40194974, 7.22444495, 7.10019159, 6.74518201, 6.83393441, 7.10019159, 6.88718585, 7.10019159, 7.13569255, 7.27769638, 7.13569255, 7.22444495, 7.33094782, 7.40194974, 7.04694016, 7.54395357, 7.27769638, 7.04694016, 7.04694016, 7.33094782, 7.04694016, 7.10019159, 7.27769638, 7.13569255, 7.10019159, 7.33094782, 7.81021075, 7.98771554, 7.89896315, 7.98771554, 8.28947369, 8.07646794, 8.34272512, 8.37822608, 8.52022991, 8.37822608, 8.46697847, 8.7864871, 8.96399189, 8.87523949, 8.52022991, 9.49650626, 9.49650626, 8.28947369, 8.16522033, 8.60898231, 8.28947369, 7.77470979, 8.34272512, 8.25397273, 8.25397273, 8.25397273, 7.84571171, 7.66820692, 7.84571171, 8.43147752, 7.77470979, 7.98771554, 8.34272512, 8.52022991, 8.25397273, 8.37822608, 8.16522033, 8.37822608, 7.93446411, 7.81021075, 7.27769638, 7.40194974, 7.89896315, 7.72145836, 7.66820692, 7.66820692, 7.72145836, 7.77470979, 8.0232165, 8.28947369, 7.98771554, 8.25397273, 8.21847177, 8.25397273, 8.16522033, 8.0232165, 8.07646794, 7.98771554, 7.98771554, 7.89896315, 8.07646794, 8.07646794, 8.34272512, 8.21847177, 8.3604756, 8.16522033, 8.34272512, 7.98771554, 8.52022991, 8.48472895, 8.20072129, 8.449228, 8.34272512, 8.52022991, 8.52022991, 8.60898231, 8.52022991, 8.34272512, 8.52022991, 8.43147752, 7.93446411, 7.72145836, 8.07646794, 7.98771554, 8.0232165, 8.16522033, 7.77470979, 7.81021075, 7.63270596, 7.72145836, 7.54395357, 7.33094782, 7.72145836, 7.45520117, 7.72145836, 7.93446411, 7.81021075, 7.66820692, 7.81021075, 7.98771554, 7.84571171, 8.16522033, 8.07646794, 8.6977347, 8.82198805, 8.6977347, 8.21847177, 8.34272512, 8.07646794, 8.1119689, 8.25397273, 8.25397273, 8.34272512, 8.1119689, 8.0232165, 7.89896315, 7.54395357, 7.49070213, 7.84571171, 7.63270596, 7.81021075, 8.16522033, 8.1119689, 8.1119689, 8.1119689, 7.98771554, 7.98771554, 7.89896315, 7.98771554, 7.98771554, 8.07646794, 8.0232165, 8.25397273, 8.28947369, 8.37822608, 8.34272512, 8.37822608, 8.55573087, 8.43147752, 8.82198805, 9.40775386, 9.76276344, 9.85151584, 9.58525865, 9.49650626, 9.58525865, 9.58525865, 9.14149668, 9.05274428, 8.96399189, 8.96399189, 8.82198805, 8.87523949, 9.05274428, 9.23024907, 8.96399189, 9.94026823, 10.11777302, 9.94026823, 9.85151584, 9.49650626, 9.94026823, 9.40775386, 9.14149668, 8.87523949, 8.96399189, 9.05274428, 9.67401105, 9.49650626, 9.76276344, 10.38403021, 9.67401105, 9.58525865, 9.67401105, 9.40775386, 9.94026823, 10.65028739, 11.09404936, 11.89282092, 11.09404936, 11.98157331, 13.13535445, 12.1590781, 12.07032571, 12.51408768, 11.71531613, 11.53781134, 11.71531613, 11.36030655, 11.18280176, 11.44905894, 11.44905894, 11.44905894, 11.00529697, 11.27155415, 11.27155415, 13.04660205, 12.42533529, 12.60284008, 12.07032571, 12.1590781, 11.71531613, 11.53781134, 11.71531613, 11.98157331, 11.80406852, 11.80406852, 12.42533529, 12.60284008, 12.51408768, 13.22410684, 12.78034487, 12.86909726, 13.04660205, 13.49036403, 12.78034487, 13.04660205, 13.934126, 13.57911642, 13.934126, 13.934126, 12.78034487, 12.42533529, 11.09404936, 11.44905894, 11.53781134, 11.53781134],
							[1.79279838, 1.81054886, 1.88155077, 2.07680604, 1.97030317, 1.89930125, 1.88155077, 1.84604981, 1.82829934, 2.27206131, 2.20105939, 2.23656035, 2.30756227, 2.69807281, 2.71582328, 2.48506706, 2.37856418, 2.50281754, 2.4495661, 2.39631466, 2.50281754, 2.30756227, 2.23656035, 2.20105939, 2.27206131, 2.09455652, 2.27206131, 2.23656035, 2.13005748, 2.18330891, 2.16555844, 2.112307, 2.09455652, 2.112307, 2.37856418, 2.20105939, 2.27206131, 2.27206131, 2.07680604, 2.16555844, 1.86380029, 1.86380029, 1.88155077, 1.89930125, 1.93480221, 2.04130508, 2.112307, 2.21880987, 2.112307, 2.20105939, 2.23656035, 2.28981179, 2.4495661, 2.94657951, 2.80457568, 2.7868252, 3.14183478, 3.2128367, 3.26608813, 3.53234532, 3.37259101, 3.5500958, 3.85185394, 4.02935873, 3.85185394, 3.81635298, 3.78085202, 3.72760059, 3.63884819, 3.4613434, 3.63884819, 3.63884819, 3.67434915, 3.63884819, 4.08261017, 4.26011496, 4.29561591, 4.38436831, 4.43761975, 4.61512454, 4.4731207, 4.29561591, 4.29561591, 4.5618731, 4.5618731, 5.18313986, 5.32514369, 5.62690184, 5.76890567, 7.04694016, 6.83393441, 7.04694016, 6.69193058, 7.27769638, 7.77470979, 7.63270596, 7.49070213, 7.84571171, 7.04694016, 7.66820692, 6.95818776, 7.04694016, 6.47892483, 6.15941621, 6.21266764, 6.21266764, 5.76890567, 5.85765806, 5.62690184, 5.76890567, 6.07066381, 6.07066381, 6.44342387, 6.15941621, 6.07066381, 7.0114392, 7.36644878, 7.22444495, 7.33094782, 6.95818776, 7.0114392, 7.10019159, 7.36644878, 7.63270596, 7.79246027, 7.65045644, 8.07646794, 8.91074045, 8.67998422, 8.83973853, 8.96399189, 9.40775386, 9.7982644, 10.15327398, 9.99351967, 10.02902063, 9.49650626, 9.76276344, 9.40775386, 9.14149668, 9.58525865, 10.02902063, 10.65028739, 9.94026823, 9.31900147, 8.60898231, 9.14149668, 9.67401105, 10.02902063, 9.58525865, 8.34272512, 8.43147752, 8.25397273, 8.60898231, 9.23024907, 9.05274428, 9.14149668, 9.67401105, 9.05274428, 9.76276344, 9.49650626, 10.02902063, 11.18280176, 11.36030655, 11.00529697, 11.27155415, 11.36030655, 12.86909726, 12.51408768, 12.60284008, 13.49036403, 12.69159247, 12.2478305, 12.51408768, 12.1590781, 12.86909726, 13.22410684, 13.13535445, 14.64414516, 14.20038319, 13.934126, 14.11163079, 14.37788798, 14.91040235, 14.20038319, 14.64414516, 14.28913558, 14.64414516, 16.06418348, 15.97543108, 16.06418348, 15.53166911, 15.79792629, 14.99915474, 14.55539277, 15.08790714, 15.08790714, 14.91040235, 15.26541193, 15.08790714, 15.26541193, 15.44291672, 16.50794545, 17.04045982, 16.95170743, 20.05804125, 18.46049814, 18.90426012, 18.72675533, 20.9455652, 19.79178407, 19.1705173, 18.99301251, 19.52552688, 19.34802209, 18.54925054, 18.28299335, 17.83923138, 18.46049814, 17.92798377, 18.46049814, 18.46049814, 18.90426012, 19.88053646, 19.70303167, 19.96928886, 21.12306999, 21.30057478, 21.38932717, 22.54310831, 20.76806041, 21.74433675, 21.65558436, 21.92184154, 22.27685112, 22.27685112, 20.76806041, 21.12306999, 20.8568128, 20.9455652, 20.50180323, 20.32429844, 20.41305083, 19.61427928, 18.46049814, 18.72675533, 18.10548856, 19.1705173, 19.08176491, 17.92798377, 16.68545024, 17.21796461, 17.3954694, 17.75047898, 16.95170743, 19.34802209, 18.81550772, 18.63800293, 18.99301251, 19.52552688, 20.05804125, 19.2592697, 20.05804125, 19.08176491, 17.21796461, 17.4842218, 17.12921222, 17.83923138, 17.3954694, 16.50794545, 16.68545024, 16.68545024, 16.95170743, 16.77420264, 17.57297419, 17.57297419, 16.59669785, 17.04045982, 17.57297419, 16.68545024, 16.15293587, 16.50794545, 16.24168827, 16.33044066, 17.04045982, 16.50794545, 16.06418348, 13.75662121, 13.40161163, 14.20038319, 14.11163079],
							[14.91040235, 14.99915474, 14.99915474, 14.99915474, 15.08790714, 14.91040235, 14.91040235, 15.17665953, 16.59669785, 16.50794545, 17.57297419, 17.66172659, 17.92798377, 17.30671701, 16.24168827, 15.7091739, 15.08790714, 15.79792629, 15.6204215, 16.06418348, 16.06418348, 15.17665953, 15.44291672, 15.7091739, 16.15293587, 15.44291672, 15.53166911, 15.44291672, 15.08790714, 15.26541193, 14.91040235, 14.64414516, 14.28913558, 15.7091739, 16.24168827, 16.06418348, 17.04045982, 17.3954694, 16.68545024, 17.30671701, 15.44291672, 15.97543108, 15.97543108, 16.68545024, 17.12921222, 16.95170743, 18.10548856, 18.28299335, 17.75047898, 17.30671701, 17.4842218, 18.10548856, 17.92798377, 17.57297419, 17.12921222, 17.12921222, 17.66172659, 17.75047898, 17.3954694, 18.10548856, 17.66172659, 18.28299335, 18.28299335, 19.1705173, 19.88053646, 20.76806041, 19.52552688, 19.1705173, 18.99301251, 18.81550772, 19.34802209, 18.63800293, 18.10548856, 18.63800293, 19.34802209, 20.41305083, 20.05804125, 21.12306999, 21.30057478, 22.54310831, 21.65558436, 22.18809873, 22.18809873, 22.54310831, 24.31815621, 29.6432999, 27.51324242, 29.6432999, 28.04575679, 29.99830948, 27.15823284, 26.44821368, 26.0932041, 28.40076637, 27.69074721, 28.40076637, 28.93328074, 26.98072805, 26.98072805, 28.75577595, 28.04575679, 28.75577595, 27.868252, 26.44821368, 25.20568016, 24.85067058, 24.85067058, 22.7206131, 22.89811789, 22.7206131, 23.96314663, 24.67316579, 26.0932041, 26.27070889, 24.67316579, 26.62571847, 27.15823284, 27.33573763, 26.80322326, 26.62571847, 26.98072805, 26.80322326, 25.47193734, 24.85067058, 26.44821368, 26.71447087, 27.33573763, 28.04575679, 27.69074721, 28.84452835, 29.11078553, 29.55454751, 28.75577595, 28.40076637, 27.51324242, 27.42449003, 26.80322326, 28.66702356, 27.77949961, 27.9570044, 28.31201398, 27.60199482, 27.9570044, 26.71447087, 26.44821368, 24.85067058, 25.73819453, 26.62571847, 26.1819565, 24.85067058, 23.60813705, 24.4069086, 24.495661, 24.58441339, 25.02817537, 24.67316579, 24.58441339, 24.76191818, 24.31815621, 26.0932041, 26.27070889, 26.71447087, 26.35946129, 26.98072805, 27.06948045, 26.89197566, 26.71447087, 28.40076637, 27.42449003, 28.75577595, 29.37704272, 27.77949961, 28.04575679, 27.51324242, 27.15823284, 27.9570044, 27.15823284, 26.00445171, 25.20568016, 24.58441339, 23.60813705, 23.69688944, 24.58441339, 24.76191818, 24.93942297, 25.47193734, 24.76191818, 24.93942297, 24.93942297, 24.85067058, 24.4069086, 23.60813705, 23.96314663, 22.54310831, 22.18809873, 22.7206131, 23.25312747, 24.05189902, 23.87439423, 24.31815621, 24.495661, 25.47193734, 26.1819565, 26.00445171, 25.73819453, 27.77949961, 28.13450919, 29.11078553, 28.13450919, 26.98072805, 27.15823284, 26.71447087, 26.53696608, 27.42449003, 27.06948045, 25.73819453, 25.82694692, 25.64944213, 26.1819565, 26.62571847, 26.71447087, 26.00445171, 25.47193734, 25.11692776, 23.69688944, 23.78564184, 24.67316579, 24.495661, 23.43063226, 22.98687028, 22.6318607, 23.34187986, 22.80936549, 22.80936549, 23.16437507, 22.98687028, 22.6318607, 22.80936549, 23.60813705, 24.31815621, 25.02817537, 24.58441339, 24.22940381, 23.87439423, 24.05189902, 22.89811789, 23.78564184, 23.34187986, 23.78564184, 23.25312747, 22.89811789, 22.7206131, 22.36560352, 21.83308915, 23.60813705, 23.87439423, 22.6318607, 23.78564184, 24.22940381, 23.16437507, 23.96314663, 23.69688944, 24.85067058, 24.85067058, 25.56068974, 23.51938465, 23.34187986, 24.22940381, 23.78564184, 23.07562268, 23.43063226, 25.73819453, 27.60199482, 28.40076637, 30.44207146, 33.72591007, 29.02203314, 32.21711935, 32.83838612, 31.59585259, 32.21711935, 35.76721515, 35.32345318, 35.14594839, 34.43592923, 32.74963372, 31.59585259, 28.93328074, 30.61957625, 30.35331906, 29.7320523],
							[22.54310831, 23.60813705, 24.495661, 24.495661, 23.96314663, 23.78564184, 23.60813705, 24.31815621, 25.02817537, 26.44821368, 26.27070889, 27.868252, 30.17581427, 31.95086217, 31.95086217, 31.06333822, 30.35331906, 31.59585259, 30.70832864, 30.70832864, 30.70832864, 29.11078553, 29.11078553, 27.868252, 27.51324242, 26.27070889, 26.27070889, 26.80322326, 26.80322326, 28.93328074, 28.57827116, 28.40076637, 28.57827116, 31.06333822, 31.24084301, 31.24084301, 31.77335738, 32.30587175, 30.35331906, 30.53082385, 25.73819453, 26.27070889, 26.27070889, 27.15823284, 27.15823284, 27.868252, 30.17581427, 31.24084301, 30.35331906, 30.35331906, 31.06333822, 31.95086217, 32.48337654, 31.95086217, 29.99830948, 30.88583343, 31.77335738, 32.66088133, 31.95086217, 34.79093881, 36.03347234, 34.61343402, 34.61343402, 35.67846276, 36.74349149, 37.45351065, 36.03347234, 37.09850107, 37.27600586, 35.67846276, 36.5659867, 35.50095797, 35.14594839, 33.1933957, 35.32345318, 37.09850107, 36.38848191, 36.92099628, 35.85596755, 35.50095797, 36.74349149, 35.85596755, 37.09850107, 37.27600586, 39.40606334, 44.90871183, 47.21627409, 50.5888651, 50.76636989, 52.89642737, 48.8138172, 48.45880762, 49.70134115, 53.69519892, 58.13281867, 60.79539052, 65.23301026, 59.90786657, 57.24529472, 62.12667644, 61.68291447, 67.45182014, 61.68291447, 55.9140088, 53.69519892, 54.58272287, 55.02648485, 52.18640821, 56.35777077, 52.89642737, 56.80153275, 58.13281867, 55.02648485, 52.89642737, 48.28130283, 52.71892258, 57.24529472, 57.24529472, 53.25143695, 53.07393216, 51.29888426, 50.5888651, 49.79009355, 48.28130283, 55.20398964, 55.02648485, 57.68905669, 55.9140088, 56.09151359, 56.09151359, 55.38149443, 58.39907585, 61.23915249, 59.10909501, 56.62402796, 55.20398964, 53.25143695, 54.1389609, 51.92015103, 51.65389384, 53.78395132, 52.80767497, 53.78395132, 48.63631241, 45.97374057, 40.29358729, 45.08621662, 46.32875015, 46.32875015, 41.26986364, 38.3410346, 39.84982532, 41.18111124, 42.33489237, 45.61873099, 46.32875015, 47.0387693, 44.99746422, 42.42364477, 46.95001691, 44.90871183, 44.81995943, 43.39992111, 44.46494985, 43.39992111, 41.18111124, 41.09235885, 46.41750254, 44.99746422, 44.02118788, 45.61873099, 45.52997859, 45.4412262, 44.64245464, 44.81995943, 46.41750254, 46.23999775, 43.84368309, 43.6661783, 42.42364477, 40.47109208, 40.02733011, 40.64859687, 39.22855855, 41.44736843, 44.64245464, 42.51239716, 42.77865435, 44.81995943, 45.4412262, 46.50625494, 46.15124536, 46.41750254, 43.6661783, 42.42364477, 43.13366393, 45.61873099, 45.52997859, 45.88498817, 46.15124536, 45.70748338, 51.92015103, 50.6776175, 52.71892258, 52.80767497, 55.9140088, 54.1389609, 55.9140088, 56.26901838, 53.96145611, 54.31646569, 50.32260792, 53.78395132, 51.83139863, 52.89642737, 50.76636989, 50.50011271, 50.41136031, 50.32260792, 49.79009355, 50.5888651, 51.92015103, 52.98517976, 50.76636989, 49.70134115, 49.52383636, 49.16882678, 50.14510313, 47.30502649, 47.92629325, 48.28130283, 50.14510313, 50.50011271, 51.47638905, 52.71892258, 51.83139863, 51.83139863, 53.78395132, 55.20398964, 55.38149443, 56.26901838, 57.33404711, 56.97903754, 54.67147527, 57.68905669, 62.12667644, 61.77166686, 66.91930577, 71.00191593, 66.3867914, 66.03178182, 68.51684887, 69.93688719, 66.03178182, 67.45182014, 69.04936324, 64.96675308, 62.12667644, 63.72421955, 58.75408543, 63.90172434, 61.0616477, 59.46410459, 55.73650401, 51.74264624, 52.63017018, 52.80767497, 54.84898006, 53.25143695, 52.18640821, 51.74264624, 54.84898006, 54.31646569, 59.46410459, 61.59416207, 59.81911417, 56.80153275, 61.77166686, 59.64160938, 56.62402796, 57.5115519, 56.26901838, 54.49397048, 58.75408543, 59.64160938, 55.73650401, 52.45266539, 49.79009355, 51.29888426, 48.99132199, 47.83754086],
							[26.44821368, 25.73819453, 26.80322326, 26.0932041, 26.44821368, 26.27070889, 26.0932041, 26.62571847, 29.46579511, 31.06333822, 31.59585259, 31.77335738, 32.83838612, 32.48337654, 32.66088133, 31.24084301, 31.06333822, 33.72591007, 33.54840528, 31.59585259, 31.4183478, 30.17581427, 30.35331906, 28.93328074, 28.93328074, 26.0932041, 27.69074721, 28.57827116, 26.80322326, 29.46579511, 29.82080469, 29.82080469, 29.6432999, 32.12836696, 32.48337654, 32.66088133, 34.08091965, 34.61343402, 31.06333822, 32.48337654, 28.93328074, 27.15823284, 26.98072805, 28.57827116, 29.11078553, 29.82080469, 31.06333822, 30.35331906, 29.11078553, 29.46579511, 29.82080469, 32.48337654, 34.43592923, 35.14594839, 33.1933957, 32.83838612, 34.25842444, 33.54840528, 32.83838612, 35.32345318, 36.38848191, 36.03347234, 36.5659867, 37.98602502, 38.51853939, 40.29358729, 37.45351065, 38.16352981, 37.45351065, 34.9684436, 35.67846276, 34.9684436, 33.72591007, 33.01589091, 33.90341486, 36.5659867, 36.03347234, 36.92099628, 34.9684436, 34.43592923, 34.9684436, 33.72591007, 35.67846276, 36.92099628, 38.16352981, 42.42364477, 42.06863519, 47.57128367, 47.57128367, 50.5888651, 48.99132199, 48.45880762, 48.8138172, 54.58272287, 60.79539052, 65.23301026, 72.77696383, 61.68291447, 59.02034262, 65.23301026, 64.78924829, 68.33934408, 63.45796236, 59.46410459, 55.47024682, 55.02648485, 55.9140088, 50.41136031, 50.41136031, 47.39377888, 48.10379804, 49.16882678, 42.06863519, 42.06863519, 40.1160825, 42.42364477, 45.61873099, 44.55370225, 42.60114956, 42.95615914, 41.71362561, 39.58356813, 37.54226305, 36.38848191, 40.91485406, 40.91485406, 44.28744506, 44.99746422, 45.4412262, 44.81995943, 44.10994027, 47.30502649, 50.14510313, 49.61258876, 47.39377888, 47.30502649, 45.26372141, 45.97374057, 43.75493069, 42.68990195, 43.5774259, 43.48867351, 44.55370225, 39.49481574, 37.18725347, 33.90341486, 36.03347234, 37.18725347, 36.6547391, 33.81466246, 31.95086217, 30.08706188, 29.6432999, 30.97458583, 33.90341486, 34.52468162, 36.6547391, 35.85596755, 34.9684436, 38.16352981, 38.3410346, 37.36475826, 36.6547391, 37.63101544, 36.29972952, 35.67846276, 33.99216725, 42.33489237, 40.64859687, 42.42364477, 43.84368309, 42.51239716, 44.10994027, 44.37619746, 45.70748338, 48.72506481, 49.34633157, 49.08007439, 49.96759834, 50.14510313, 47.39377888, 48.28130283, 50.23385552, 50.23385552, 51.47638905, 55.55899922, 53.25143695, 54.1389609, 55.9140088, 55.9140088, 56.80153275, 52.71892258, 52.98517976, 53.25143695, 51.65389384, 54.67147527, 58.04406627, 56.97903754, 57.33404711, 58.93159022, 59.81911417, 61.77166686, 60.88414291, 61.23915249, 66.03178182, 69.40437282, 64.96675308, 69.7593824, 68.33934408, 66.91930577, 67.09681056, 61.59416207, 65.85427703, 63.90172434, 64.78924829, 61.23915249, 61.41665728, 59.99661896, 59.10909501, 58.93159022, 60.52913333, 62.12667644, 65.14425787, 66.20928661, 62.65919081, 62.65919081, 63.72421955, 65.49926745, 62.30418123, 63.54671476, 62.48168602, 63.19170518, 63.36920997, 65.49926745, 68.51684887, 66.56429619, 66.20928661, 68.33934408, 69.93688719, 71.17942072, 71.00191593, 73.66448778, 76.68206921, 77.56959316, 81.65220332, 81.82970811, 82.53972727, 85.02479433, 92.74625269, 87.3323566, 82.89473685, 86.44483265, 91.85872874, 92.74625269, 89.19615689, 94.07753861, 87.3323566, 84.84728954, 89.19615689, 83.7822608, 86.44483265, 85.20229912, 82.71723206, 74.90702131, 73.13197341, 74.19700215, 73.3094782, 72.42195425, 70.64690635, 73.48698299, 81.11968895, 86.79984223, 86.44483265, 87.86487097, 87.3323566, 85.5573087, 79.87715542, 86.08982307, 82.0072129, 82.18471769, 85.37980391, 85.73481349, 87.86487097, 95.85258651, 104.28406402, 91.41496676, 92.30249071, 85.73481349, 84.84728954, 84.13727038, 80.58717458],
							[21.30057478, 20.9455652, 22.36560352, 21.12306999, 20.9455652, 20.76806041, 20.76806041, 20.59055562, 21.12306999, 21.12306999, 22.54310831, 22.54310831, 24.31815621, 22.89811789, 22.7206131, 21.12306999, 20.9455652, 22.01059394, 20.9455652, 21.12306999, 22.18809873, 21.12306999, 21.83308915, 21.30057478, 20.76806041, 19.1705173, 18.63800293, 19.52552688, 18.81550772, 19.34802209, 19.34802209, 18.81550772, 18.81550772, 19.34802209, 20.41305083, 19.70303167, 19.88053646, 19.70303167, 18.10548856, 17.92798377, 15.7091739, 16.50794545, 16.06418348, 16.41919306, 16.77420264, 16.41919306, 16.68545024, 16.59669785, 16.50794545, 16.50794545, 16.68545024, 17.92798377, 17.75047898, 17.75047898, 17.04045982, 16.06418348, 16.59669785, 16.77420264, 16.59669785, 16.86295503, 17.04045982, 18.81550772, 18.63800293, 18.99301251, 19.52552688, 19.52552688, 19.52552688, 19.1705173, 19.34802209, 17.75047898, 17.57297419, 18.28299335, 17.92798377, 17.30671701, 17.3954694, 17.92798377, 18.10548856, 19.1705173, 18.28299335, 18.28299335, 19.1705173, 18.63800293, 18.81550772, 18.81550772, 19.88053646, 20.76806041, 21.30057478, 22.36560352, 22.01059394, 23.43063226, 21.30057478, 21.12306999, 21.12306999, 22.01059394, 22.89811789, 23.07562268, 26.62571847, 23.60813705, 22.54310831, 23.78564184, 23.60813705, 24.85067058, 24.31815621, 23.78564184, 21.30057478, 21.65558436, 21.47807957, 20.41305083, 19.88053646, 19.34802209, 19.52552688, 20.41305083, 19.70303167, 19.88053646, 19.70303167, 19.88053646, 20.23554604, 19.52552688, 19.52552688, 19.34802209, 19.88053646, 19.52552688, 20.23554604, 21.47807957, 20.9455652, 21.47807957, 22.18809873, 22.54310831, 22.09934633, 22.18809873, 22.27685112, 22.7206131, 23.34187986, 22.98687028, 22.36560352, 21.74433675, 21.12306999, 20.9455652, 20.14679365, 20.05804125, 20.41305083, 19.52552688, 19.70303167, 19.52552688, 19.52552688, 18.46049814, 19.34802209, 19.43677449, 19.96928886, 19.61427928, 18.99301251, 18.72675533, 18.37174575, 18.90426012, 20.9455652, 21.83308915, 22.45435591, 21.92184154, 21.83308915, 21.92184154, 21.38932717, 21.21182238, 20.76806041, 20.8568128, 20.23554604, 19.2592697, 18.10548856, 19.70303167, 19.88053646, 20.05804125, 20.76806041, 20.32429844, 20.32429844, 19.61427928, 20.05804125, 20.50180323, 20.23554604, 19.79178407, 19.52552688, 19.52552688, 20.9455652, 20.41305083, 20.67930802, 20.59055562, 21.38932717, 21.12306999, 20.67930802, 20.8568128, 21.30057478, 21.30057478, 21.47807957, 19.79178407, 20.14679365, 19.2592697, 18.99301251, 19.70303167, 20.8568128, 21.65558436, 21.30057478, 20.9455652, 20.9455652, 22.18809873, 22.54310831, 22.54310831, 23.87439423, 24.93942297, 24.31815621, 24.93942297, 24.76191818, 25.02817537, 25.47193734, 23.16437507, 24.14065142, 24.14065142, 23.96314663, 22.54310831, 22.18809873, 22.89811789, 24.14065142, 24.22940381, 24.4069086, 25.20568016, 25.38318495, 26.00445171, 24.58441339, 24.85067058, 25.20568016, 24.85067058, 23.51938465, 24.31815621, 22.27685112, 22.98687028, 22.89811789, 22.6318607, 23.07562268, 22.80936549, 22.36560352, 22.54310831, 22.36560352, 22.45435591, 22.54310831, 22.6318607, 22.36560352, 21.38932717, 21.21182238, 20.76806041, 21.03431759, 20.8568128, 22.6318607, 21.38932717, 21.03431759, 21.83308915, 21.38932717, 21.38932717, 21.03431759, 21.38932717, 20.8568128, 21.30057478, 21.21182238, 20.9455652, 21.47807957, 21.38932717, 21.21182238, 21.12306999, 20.8568128, 20.59055562, 20.76806041, 21.21182238, 21.12306999, 20.8568128, 20.8568128, 21.12306999, 21.30057478, 20.8568128, 20.67930802, 21.03431759, 22.36560352, 25.73819453, 28.48951877, 24.85067058, 25.56068974, 26.27070889, 26.98072805, 25.73819453, 24.67316579, 24.85067058, 23.16437507, 22.45435591, 22.54310831, 22.54310831, 22.18809873],
							[75.88329765, 75.43953568, 80.3209174, 73.2207258, 71.44567791, 68.33934408, 69.67063001, 73.66448778, 72.77696383, 73.66448778, 77.21458358, 78.5458695, 78.5458695, 84.75853714, 85.64606109, 78.5458695, 80.76467937, 85.20229912, 88.75239491, 86.97734702, 94.96506256, 92.30249071, 95.85258651, 94.07753861, 94.96506256, 89.63991886, 96.74011046, 98.51515836, 94.96506256, 96.74011046, 98.51515836, 98.51515836, 96.74011046, 102.9527781, 110.05296969, 110.94049364, 116.26563734, 116.26563734, 108.2779218, 106.5028739, 97.62763441, 97.62763441, 92.30249071, 95.85258651, 98.51515836, 99.4026823, 106.5028739, 106.5028739, 106.5028739, 108.2779218, 109.16544574, 113.60306549, 115.37811339, 120.70325708, 117.15316129, 116.26563734, 119.81573313, 122.47830498, 121.59078103, 129.57849658, 129.57849658, 126.91592473, 122.47830498, 125.14087683, 127.80344868, 132.24106842, 127.80344868, 129.57849658, 130.46602052, 127.80344868, 132.24106842, 132.24106842, 129.57849658, 126.02840078, 131.35354447, 141.11630791, 145.55392766, 148.21649951, 147.32897556, 146.44145161, 147.32897556, 143.77887976, 142.00383186, 144.66640371, 149.10402346, 155.3166911, 152.65411925, 158.8667869, 153.5416432, 158.8667869, 152.65411925, 153.5416432, 155.3166911, 169.51707429, 177.50478983, 188.15507722, 204.1305083, 189.93012512, 189.93012512, 202.35546041, 200.58041251, 232.53127468, 220.10593939, 209.455652, 197.03031671, 179.27983773, 181.05488563, 164.19193059, 158.8667869, 154.42916715, 158.8667869, 163.30440664, 156.20421505, 151.7665953, 147.32897556, 151.7665953, 160.6418348, 157.091739, 152.65411925, 150.87907135, 154.42916715, 148.21649951, 149.54778543, 145.55392766, 155.3166911, 158.8667869, 165.07945454, 169.96083626, 162.41688269, 156.20421505, 156.20421505, 160.6418348, 165.52321652, 161.08559677, 155.76045308, 153.09788123, 152.21035728, 161.52935874, 160.19807282, 159.75431085, 161.52935874, 160.19807282, 161.97312072, 152.65411925, 148.66026148, 144.66640371, 146.44145161, 150.87907135, 148.21649951, 143.77887976, 138.00997409, 133.57235435, 132.6848304, 134.4598783, 142.89135581, 142.00383186, 145.11016569, 143.33511779, 138.45373607, 150.43530938, 154.87292913, 154.87292913, 156.64797702, 161.52935874, 161.97312072, 159.31054887, 161.08559677, 170.40459824, 168.18578836, 175.72974193, 182.82993352, 181.05488563, 178.39231378, 179.27983773, 176.17350391, 187.26755327, 189.04260117, 181.05488563, 185.49250537, 187.26755327, 180.16736168, 177.50478983, 182.82993352, 184.60498142, 189.04260117, 195.25526881, 190.81764907, 190.81764907, 197.03031671, 197.03031671, 202.35546041, 197.03031671, 202.35546041, 194.36774486, 196.14279276, 197.91784066, 203.24298435, 205.01803225, 205.9055562, 205.01803225, 207.6806041, 210.34317595, 208.56812805, 217.44336754, 222.76851124, 226.31860703, 224.54355913, 227.20613098, 225.43108308, 220.10593939, 220.10593939, 203.24298435, 211.2306999, 205.9055562, 209.455652, 204.1305083, 202.35546041, 202.35546041, 200.58041251, 197.03031671, 198.80536461, 203.24298435, 207.6806041, 209.455652, 205.01803225, 203.24298435, 207.6806041, 219.21841544, 213.89327174, 214.78079569, 219.21841544, 225.43108308, 236.08137047, 239.63146627, 241.40651417, 236.96889442, 233.41879863, 244.06908602, 251.16927761, 254.7193734, 254.7193734, 261.819565, 275.13242424, 277.79499608, 283.12013978, 287.55775952, 284.00766373, 276.01994818, 285.78271163, 280.45756793, 274.24490029, 284.89518768, 291.10785532, 292.88290322, 298.20804691, 324.83376539, 315.07100195, 318.62109774, 326.60881329, 325.72128934, 335.48405278, 335.48405278, 337.25910068, 328.38386118, 313.29595405, 319.50862169, 313.29595405, 326.60881329, 331.04643303, 336.37157673, 350.57195991, 385.18539393, 390.51053762, 395.83568132, 417.1362561, 404.71092081, 408.26101661, 413.5861603, 413.5861603, 427.78654349, 434.88673508, 470.38769305, 465.06254935, 475.71283674, 484.58807623, 482.81302834, 457.96235776, 418.911304, 424.23644769, 408.26101661, 410.03606451],
							[3.23058717, 3.30158909, 3.5500958, 4.11811112, 4.11811112, 4.43761975, 4.38436831, 5.14763891, 4.88138172, 5.00563507, 5.11213795, 5.44939705, 5.80440663, 5.71565423, 5.85765806, 5.71565423, 5.80440663, 6.21266764, 6.74518201, 6.74518201, 6.44342387, 5.76890567, 6.12391525, 6.03516285, 5.89315902, 5.50264848, 5.50264848, 5.62690184, 5.36064465, 5.14763891, 5.18313986, 4.34886735, 5.00563507, 5.2363913, 5.44939705, 4.88138172, 5.2363913, 5.00563507, 4.66837597, 4.66837597, 3.72760059, 3.67434915, 3.85185394, 4.08261017, 4.17136256, 4.02935873, 4.11811112, 4.17136256, 3.90510538, 3.90510538, 4.43761975, 4.91688268, 4.97013412, 4.97013412, 4.66837597, 4.38436831, 4.66837597, 5.00563507, 5.18313986, 5.62690184, 5.50264848, 5.59140088, 5.68015327, 5.71565423, 5.41389609, 5.44939705, 4.82813028, 4.97013412, 4.79262933, 4.52637214, 4.91688268, 4.61512454, 4.61512454, 4.43761975, 4.34886735, 4.79262933, 4.91688268, 4.97013412, 4.79262933, 4.88138172, 5.27189226, 5.14763891, 5.36064465, 5.50264848, 5.76890567, 6.21266764, 6.30142004, 7.63270596, 7.10019159, 8.6977347, 8.6977347, 8.60898231, 9.58525865, 10.20652542, 10.20652542, 9.85151584, 10.29527781, 9.23024907, 9.14149668, 8.87523949, 8.52022991, 8.66223374, 8.7864871, 8.43147752, 7.45520117, 7.22444495, 7.49070213, 6.9226868, 8.0232165, 7.89896315, 7.49070213, 8.21847177, 7.72145836, 7.72145836, 7.22444495, 7.89896315, 8.0232165, 7.93446411, 7.98771554, 8.07646794, 7.98771554, 7.81021075, 8.00546602, 7.65045644, 8.50247943, 8.67998422, 9.05274428, 9.17699763, 9.31900147, 9.17699763, 8.91074045, 9.14149668, 9.62075961, 9.709512, 9.26575003, 9.14149668, 8.43147752, 8.6977347, 8.16522033, 8.37822608, 8.55573087, 8.55573087, 8.6977347, 7.72145836, 7.10019159, 6.74518201, 7.10019159, 7.0114392, 6.78068297, 6.78068297, 5.9996619, 5.89315902, 5.71565423, 5.68015327, 6.44342387, 6.56767722, 6.39017243, 6.2481686, 5.9996619, 6.44342387, 6.60317818, 6.65642962, 6.47892483, 6.51442579, 6.47892483, 6.336921, 6.15941621, 6.83393441, 7.04694016, 7.36644878, 7.72145836, 7.40194974, 7.40194974, 7.36644878, 7.49070213, 7.98771554, 8.1119689, 8.16522033, 8.1119689, 8.16522033, 7.45520117, 7.54395357, 7.84571171, 7.84571171, 7.93446411, 8.37822608, 8.07646794, 8.21847177, 8.34272512, 8.55573087, 8.28947369, 8.07646794, 8.34272512, 8.0232165, 8.1119689, 7.72145836, 8.16522033, 8.25397273, 8.34272512, 8.28947369, 8.25397273, 9.40775386, 10.02902063, 9.85151584, 9.76276344, 9.58525865, 9.40775386, 9.40775386, 9.40775386, 8.96399189, 9.23024907, 8.37822608, 9.05274428, 8.6977347, 9.05274428, 8.73323566, 8.6977347, 8.34272512, 8.37822608, 8.34272512, 8.66223374, 8.66223374, 8.96399189, 8.82198805, 8.66223374, 8.66223374, 8.6977347, 8.52022991, 8.55573087, 8.7864871, 8.7864871, 8.96399189, 9.85151584, 9.94026823, 10.02902063, 9.85151584, 9.76276344, 9.76276344, 9.67401105, 9.67401105, 10.20652542, 9.94026823, 9.94026823, 10.02902063, 9.67401105, 9.49650626, 10.02902063, 10.91654457, 11.44905894, 10.4727826, 10.73903978, 11.09404936, 11.53781134, 13.13535445, 12.69159247, 12.78034487, 12.07032571, 12.33658289, 12.60284008, 11.89282092, 12.51408768, 13.40161163, 12.95784966, 12.51408768, 11.44905894, 11.80406852, 11.00529697, 11.80406852, 12.2478305, 13.13535445, 12.86909726, 12.78034487, 12.95784966, 15.26541193, 18.19424096, 17.30671701, 16.77420264, 19.61427928, 20.67930802, 20.41305083, 19.70303167, 19.61427928, 19.08176491, 19.70303167, 19.70303167, 18.99301251, 18.28299335, 16.59669785, 14.55539277, 16.15293587, 14.99915474],
							[8.12971937, 8.07646794, 8.25397273, 8.16522033, 8.12971937, 8.07646794, 8.25397273, 8.82198805, 8.73323566, 9.21249859, 8.96399189, 9.72726248, 10.40178068, 10.22427589, 9.97576919, 9.72726248, 10.1355235, 9.88701679, 9.97576919, 9.7982644, 9.7982644, 8.6977347, 8.96399189, 8.6977347, 9.21249859, 8.7864871, 9.1237462, 8.96399189, 8.53798039, 9.21249859, 9.4610053, 9.39000338, 9.30125099, 9.63851009, 10.31302829, 10.4727826, 10.82779218, 10.98754649, 9.7982644, 10.1355235, 9.4610053, 8.96399189, 9.21249859, 9.21249859, 9.30125099, 9.30125099, 9.4610053, 9.30125099, 8.96399189, 8.7864871, 9.39000338, 9.21249859, 10.06452158, 9.7982644, 9.4610053, 10.06452158, 9.88701679, 9.88701679, 9.97576919, 9.97576919, 10.40178068, 11.07629889, 10.65028739, 10.73903978, 10.73903978, 11.2360532, 10.98754649, 11.5733123, 12.08807619, 11.9105714, 12.85134678, 12.76259439, 12.60284008, 11.83956948, 12.51408768, 12.42533529, 12.51408768, 12.85134678, 12.60284008, 12.76259439, 12.76259439, 12.60284008, 12.76259439, 12.67384199, 12.85134678, 13.61461738, 14.62639468, 15.54941959, 15.05240618, 16.64994929, 15.38966528, 15.30091288, 15.30091288, 16.56119689, 15.79792629, 16.56119689, 17.07596078, 15.05240618, 14.20038319, 15.30091288, 15.38966528, 15.46066719, 14.11163079, 14.53764229, 14.28913558, 14.78614899, 15.63817198, 14.71514708, 15.21216049, 14.71514708, 14.53764229, 14.71514708, 14.44888989, 13.95187648, 13.52586498, 14.28913558, 14.0228784, 14.28913558, 13.95187648, 13.77437169, 13.6856193, 12.94009918, 13.52586498, 13.13535445, 13.77437169, 12.85134678, 12.85134678, 12.79809535, 12.54958864, 10.98754649, 11.87507044, 11.2360532, 11.62656373, 11.66206469, 11.07629889, 10.98754649, 10.82779218, 11.83956948, 11.41355799, 11.50231038, 11.32480559, 11.07629889, 10.82779218, 10.56153499, 9.05274428, 8.0232165, 9.1237462, 8.87523949, 9.30125099, 8.7864871, 8.07646794, 7.18894399, 7.43745069, 7.73920884, 7.52620309, 7.31319734, 7.40194974, 7.27769638, 6.79843345, 7.10019159, 7.43745069, 7.73920884, 7.89896315, 7.89896315, 8.28947369, 7.95221458, 7.52620309, 8.87523949, 8.6977347, 9.05274428, 9.05274428, 9.05274428, 9.88701679, 9.88701679, 10.40178068, 10.56153499, 11.07629889, 10.82779218, 10.98754649, 11.16505128, 10.56153499, 10.22427589, 10.40178068, 10.65028739, 10.40178068, 10.4727826, 10.4727826, 10.31302829, 10.1355235, 10.56153499, 10.1355235, 10.1355235, 10.22427589, 9.88701679, 9.05274428, 8.6977347, 8.96399189, 9.21249859, 9.21249859, 8.87523949, 9.21249859, 9.7982644, 9.7982644, 9.63851009, 10.06452158, 10.22427589, 9.7982644, 9.88701679, 9.63851009, 9.72726248, 9.7982644, 9.05274428, 9.30125099, 9.1237462, 9.39000338, 9.4610053, 9.4610053, 9.4610053, 9.30125099, 8.73323566, 8.6977347, 9.05274428, 9.14149668, 8.96399189, 8.52022991, 8.60898231, 8.43147752, 8.43147752, 8.25397273, 8.21847177, 8.07646794, 8.34272512, 8.46697847, 8.16522033, 8.1119689, 8.07646794, 7.84571171, 7.89896315, 7.89896315, 9.14149668, 8.52022991, 9.23024907, 9.31900147, 9.14149668, 9.14149668, 9.40775386, 9.23024907, 9.40775386, 9.58525865, 9.14149668, 9.14149668, 9.23024907, 9.23024907, 9.14149668, 9.14149668, 9.58525865, 9.49650626, 9.40775386, 10.02902063, 9.58525865, 9.49650626, 9.67401105, 9.76276344, 9.40775386, 9.14149668, 8.96399189, 9.05274428, 8.82198805, 8.73323566, 8.7864871, 8.87523949, 8.7864871, 8.60898231, 8.87523949, 9.05274428, 8.96399189, 8.46697847, 8.60898231, 8.43147752, 8.34272512, 8.46697847, 8.34272512, 8.0232165, 8.21847177, 8.60898231, 8.34272512, 8.16522033, 7.77470979, 7.89896315, 8.21847177, 8.82198805],
							[49.70134115, 49.34633157, 51.12137947, 49.34633157, 50.5888651, 50.76636989, 51.12137947, 52.18640821, 55.47024682, 56.80153275, 60.35162854, 59.02034262, 61.68291447, 64.78924829, 62.12667644, 59.02034262, 61.23915249, 65.23301026, 65.23301026, 63.01420039, 64.34548631, 57.24529472, 59.90786657, 62.57043841, 61.68291447, 56.35777077, 59.46410459, 62.12667644, 59.90786657, 64.34548631, 64.34548631, 64.78924829, 63.45796236, 67.45182014, 65.23301026, 66.56429619, 65.67677224, 64.34548631, 58.13281867, 57.24529472, 49.52383636, 49.87884594, 48.28130283, 51.65389384, 52.00890342, 52.18640821, 55.9140088, 55.02648485, 54.1389609, 54.1389609, 54.58272287, 58.57658064, 59.90786657, 61.68291447, 55.9140088, 55.9140088, 58.13281867, 60.35162854, 61.23915249, 65.23301026, 66.12053421, 67.89558211, 67.89558211, 68.33934408, 70.11439198, 73.2207258, 70.55815396, 73.66448778, 70.55815396, 68.78310606, 73.66448778, 71.44567791, 67.89558211, 64.34548631, 66.56429619, 70.11439198, 69.22686803, 71.44567791, 65.23301026, 67.00805816, 70.11439198, 68.33934408, 71.00191593, 71.88943988, 77.21458358, 82.53972727, 82.0959653, 91.41496676, 88.30863294, 98.51515836, 94.07753861, 96.74011046, 93.19001466, 109.16544574, 110.05296969, 112.71554154, 123.36582893, 105.61534995, 104.727826, 113.60306549, 111.82801759, 118.04068524, 113.60306549, 102.06525415, 94.96506256, 95.85258651, 94.07753861, 93.19001466, 100.29020625, 94.96506256, 99.4026823, 102.06525415, 100.29020625, 98.51515836, 94.07753861, 102.9527781, 103.84030205, 106.5028739, 103.84030205, 103.84030205, 102.06525415, 101.1777302, 98.07139638, 98.95892033, 104.727826, 106.05911192, 112.71554154, 116.70939931, 104.28406402, 106.5028739, 110.49673167, 112.27177957, 115.37811339, 112.71554154, 110.05296969, 107.39039785, 103.39654008, 102.50901613, 97.18387243, 101.1777302, 101.62149218, 101.62149218, 100.73396823, 92.30249071, 87.68736618, 80.05466021, 83.07224164, 84.84728954, 85.5573087, 77.03707879, 72.24444946, 71.00191593, 73.66448778, 77.39208837, 83.60475601, 87.68736618, 91.41496676, 92.74625269, 90.08368084, 95.85258651, 93.19001466, 93.63377663, 93.63377663, 95.40882453, 96.74011046, 91.85872874, 89.19615689, 100.73396823, 96.74011046, 102.9527781, 108.2779218, 103.84030205, 105.61534995, 102.9527781, 104.727826, 109.60920772, 110.49673167, 106.94663587, 107.39039785, 106.5028739, 100.73396823, 99.84644428, 102.06525415, 104.28406402, 106.5028739, 109.16544574, 107.39039785, 108.72168377, 112.27177957, 112.71554154, 108.72168377, 102.50901613, 104.28406402, 99.4026823, 95.40882453, 98.07139638, 105.17158797, 106.5028739, 105.17158797, 106.94663587, 106.5028739, 111.82801759, 110.94049364, 113.15930352, 118.48444721, 125.5846388, 122.92206696, 127.3596867, 125.5846388, 121.14701906, 120.70325708, 115.37811339, 120.25949511, 120.70325708, 120.25949511, 112.27177957, 111.82801759, 110.94049364, 113.15930352, 113.15930352, 114.49058944, 120.70325708, 122.03454301, 119.81573313, 113.60306549, 114.04682747, 117.59692326, 120.25949511, 118.04068524, 119.81573313, 117.59692326, 119.37197116, 118.92820919, 118.92820919, 121.59078103, 122.03454301, 118.48444721, 121.14701906, 120.70325708, 121.59078103, 120.70325708, 124.69711486, 127.3596867, 123.80959091, 120.25949511, 120.70325708, 122.03454301, 133.57235435, 130.02225855, 126.47216275, 125.5846388, 130.02225855, 130.9097825, 129.57849658, 125.5846388, 127.3596867, 130.9097825, 126.47216275, 126.91592473, 115.82187536, 119.37197116, 118.04068524, 117.59692326, 109.16544574, 103.84030205, 108.2779218, 107.83415982, 106.94663587, 104.28406402, 104.727826, 106.05911192, 111.38425562, 110.94049364, 115.37811339, 115.37811339, 121.59078103, 114.93435141, 122.47830498, 123.80959091, 118.04068524, 122.03454301, 120.70325708, 122.03454301, 130.46602052, 127.80344868, 121.14701906, 115.82187536, 105.17158797, 114.49058944, 111.82801759, 108.72168377],
							[5.85765806, 5.76890567, 6.01741238, 6.01741238, 5.89315902, 5.76890567, 5.68015327, 6.17716669, 7.06469064, 7.18894399, 8.23622225, 8.39597656, 9.28350051, 9.67401105, 9.51425673, 8.7864871, 8.62673279, 9.76276344, 9.35450242, 9.19474811, 9.85151584, 9.0349938, 9.35450242, 9.28350051, 9.0349938, 8.71548518, 8.7864871, 8.71548518, 8.55573087, 8.62673279, 8.87523949, 8.46697847, 8.7864871, 9.0349938, 9.60300913, 9.35450242, 9.51425673, 9.28350051, 8.23622225, 8.39597656, 6.53217627, 6.37242195, 6.78068297, 7.06469064, 7.54395357, 7.45520117, 7.89896315, 8.14746985, 7.57945453, 7.50845261, 7.79246027, 8.30722416, 8.71548518, 8.7864871, 8.14746985, 7.63270596, 7.79246027, 7.73920884, 7.63270596, 8.30722416, 8.30722416, 8.96399189, 8.87523949, 9.28350051, 9.35450242, 10.01127015, 9.67401105, 10.17102446, 10.17102446, 9.67401105, 9.92251775, 9.51425673, 9.19474811, 8.55573087, 8.96399189, 9.67401105, 9.28350051, 9.28350051, 9.0349938, 9.19474811, 9.28350051, 8.96399189, 9.0349938, 9.1237462, 9.35450242, 10.33077877, 10.33077877, 11.28930463, 11.28930463, 11.94607236, 11.37805703, 11.37805703, 11.21830272, 12.90459822, 15.17665953, 15.97543108, 19.52552688, 18.72675533, 18.38949623, 19.52552688, 18.54925054, 19.52552688, 17.91023329, 15.88667869, 13.79212217, 14.3601375, 14.76839851, 12.99335062, 13.95187648, 13.47261355, 13.56136594, 13.95187648, 12.74484391, 12.51408768, 11.69756565, 12.90459822, 13.31285924, 13.15310493, 12.74484391, 12.74484391, 12.42533529, 12.51408768, 11.85731996, 11.00529697, 12.42533529, 12.17682858, 12.74484391, 13.06435253, 12.8335963, 12.86909726, 12.46083625, 13.06435253, 13.50811451, 13.15310493, 13.40161163, 14.44888989, 14.11163079, 14.28913558, 14.28913558, 14.28913558, 14.44888989, 14.67964612, 14.20038319, 12.69159247, 10.91654457, 10.02902063, 10.91654457, 10.73903978, 10.65028739, 9.85151584, 8.82198805, 8.7864871, 9.49650626, 9.94026823, 10.56153499, 9.76276344, 10.56153499, 10.65028739, 10.11777302, 10.38403021, 10.11777302, 10.02902063, 9.76276344, 9.76276344, 9.58525865, 9.49650626, 9.40775386, 10.56153499, 10.02902063, 10.11777302, 10.20652542, 10.02902063, 10.02902063, 10.02902063, 10.02902063, 10.20652542, 10.56153499, 10.20652542, 10.20652542, 9.94026823, 9.67401105, 9.49650626, 9.58525865, 9.58525865, 9.76276344, 9.94026823, 9.67401105, 9.40775386, 10.29527781, 10.4727826, 10.91654457, 10.38403021, 10.73903978, 10.20652542, 10.02902063, 10.20652542, 10.91654457, 10.56153499, 10.29527781, 10.4727826, 10.65028739, 11.98157331, 12.42533529, 13.57911642, 14.73289756, 15.08790714, 14.73289756, 15.17665953, 14.99915474, 14.55539277, 14.55539277, 12.95784966, 13.66786882, 13.66786882, 14.0228784, 13.40161163, 13.66786882, 13.57911642, 13.13535445, 12.86909726, 13.22410684, 13.66786882, 13.934126, 13.84537361, 13.66786882, 13.49036403, 13.57911642, 14.0228784, 13.57911642, 13.49036403, 13.22410684, 13.22410684, 13.66786882, 13.934126, 14.82164995, 14.28913558, 13.84537361, 14.46664037, 14.46664037, 14.91040235, 15.26541193, 15.35416432, 15.53166911, 15.17665953, 15.44291672, 15.88667869, 16.33044066, 17.57297419, 17.4842218, 16.86295503, 17.04045982, 17.04045982, 16.86295503, 16.41919306, 15.6204215, 16.50794545, 16.15293587, 16.15293587, 16.15293587, 15.44291672, 15.88667869, 15.88667869, 15.35416432, 14.55539277, 14.55539277, 14.20038319, 13.22410684, 13.40161163, 12.78034487, 12.60284008, 14.11163079, 14.73289756, 14.46664037, 14.99915474, 16.06418348, 15.88667869, 14.91040235, 15.17665953, 14.91040235, 13.934126, 14.0228784, 13.31285924, 13.22410684, 14.64414516, 14.91040235, 14.11163079, 13.57911642, 13.04660205, 12.78034487, 12.78034487, 12.78034487],
							[16.95170743, 16.41919306, 16.41919306, 16.50794545, 16.50794545, 16.86295503, 16.50794545, 16.50794545, 17.3954694, 17.92798377, 18.99301251, 18.99301251, 19.88053646, 22.18809873, 22.01059394, 20.76806041, 19.88053646, 21.65558436, 20.9455652, 20.41305083, 20.41305083, 18.63800293, 18.46049814, 18.46049814, 18.46049814, 17.66172659, 17.3954694, 17.75047898, 17.4842218, 17.66172659, 18.28299335, 18.46049814, 19.34802209, 20.23554604, 21.47807957, 21.30057478, 20.76806041, 20.59055562, 18.28299335, 18.46049814, 15.53166911, 14.99915474, 15.26541193, 15.97543108, 16.41919306, 15.7091739, 16.77420264, 16.77420264, 16.59669785, 16.59669785, 17.21796461, 18.99301251, 20.05804125, 20.41305083, 18.99301251, 18.10548856, 19.1705173, 18.46049814, 18.10548856, 19.70303167, 19.70303167, 20.9455652, 20.59055562, 22.01059394, 23.07562268, 23.43063226, 22.18809873, 22.54310831, 22.01059394, 20.76806041, 21.30057478, 20.05804125, 20.23554604, 20.23554604, 20.9455652, 21.65558436, 20.59055562, 21.65558436, 20.9455652, 21.12306999, 21.30057478, 20.9455652, 20.9455652, 21.47807957, 22.36560352, 24.495661, 24.14065142, 26.44821368, 26.27070889, 27.868252, 25.38318495, 24.67316579, 24.85067058, 28.04575679, 29.6432999, 30.53082385, 33.54840528, 33.1933957, 32.12836696, 34.25842444, 33.01589091, 33.90341486, 32.12836696, 29.28829032, 27.868252, 27.69074721, 28.04575679, 26.0932041, 26.27070889, 26.27070889, 26.0932041, 26.27070889, 24.67316579, 23.60813705, 21.65558436, 23.78564184, 24.85067058, 25.02817537, 23.60813705, 23.25312747, 22.89811789, 21.83308915, 20.8568128, 21.12306999, 23.60813705, 23.78564184, 24.495661, 25.29443255, 24.85067058, 24.05189902, 23.69688944, 25.20568016, 26.62571847, 25.64944213, 25.20568016, 25.29443255, 24.4069086, 24.76191818, 24.14065142, 24.31815621, 24.4069086, 24.58441339, 23.51938465, 20.9455652, 20.14679365, 19.1705173, 20.05804125, 20.05804125, 19.52552688, 18.46049814, 18.46049814, 17.66172659, 17.21796461, 17.57297419, 19.1705173, 19.52552688, 19.88053646, 19.70303167, 18.99301251, 20.41305083, 20.23554604, 20.9455652, 19.88053646, 19.96928886, 20.41305083, 20.23554604, 20.14679365, 22.09934633, 21.92184154, 22.18809873, 22.36560352, 22.27685112, 22.45435591, 21.65558436, 21.83308915, 23.16437507, 23.51938465, 22.6318607, 22.6318607, 22.09934633, 20.67930802, 21.03431759, 21.03431759, 20.9455652, 21.74433675, 23.34187986, 22.27685112, 22.09934633, 22.89811789, 22.89811789, 22.7206131, 22.54310831, 22.89811789, 20.9455652, 20.9455652, 20.8568128, 22.01059394, 21.12306999, 21.47807957, 21.47807957, 21.83308915, 23.34187986, 23.60813705, 25.56068974, 26.53696608, 27.69074721, 26.62571847, 27.42449003, 26.62571847, 26.00445171, 26.62571847, 25.29443255, 26.1819565, 26.1819565, 26.27070889, 26.0932041, 25.11692776, 25.38318495, 25.29443255, 24.76191818, 25.73819453, 25.82694692, 26.71447087, 26.27070889, 25.82694692, 25.73819453, 25.82694692, 26.00445171, 24.93942297, 25.11692776, 25.38318495, 26.27070889, 26.1819565, 26.0932041, 26.53696608, 25.91569932, 25.47193734, 26.27070889, 26.35946129, 25.82694692, 26.27070889, 26.98072805, 27.15823284, 26.71447087, 27.24698524, 27.60199482, 28.48951877, 29.46579511, 30.61957625, 29.55454751, 28.75577595, 29.82080469, 30.17581427, 29.37704272, 28.04575679, 29.82080469, 28.31201398, 28.22326158, 28.31201398, 27.868252, 27.69074721, 27.69074721, 28.22326158, 26.53696608, 25.02817537, 24.93942297, 25.47193734, 25.29443255, 25.11692776, 25.20568016, 25.64944213, 26.44821368, 26.0932041, 26.62571847, 26.35946129, 26.00445171, 25.20568016, 25.47193734, 25.20568016, 24.85067058, 24.85067058, 24.85067058, 25.11692776, 27.60199482, 27.60199482, 26.0932041, 25.82694692, 24.85067058, 25.02817537, 24.85067058, 26.0932041],
							[38.87354897, 37.27600586, 39.05105376, 41.18111124, 39.40606334, 37.27600586, 37.45351065, 36.92099628, 39.58356813, 41.53612082, 42.42364477, 42.42364477, 42.60114956, 47.39377888, 47.0387693, 45.08621662, 44.19869267, 47.39377888, 45.79623578, 44.55370225, 45.26372141, 43.13366393, 43.48867351, 42.06863519, 42.06863519, 38.69604418, 39.22855855, 39.76107292, 39.05105376, 38.87354897, 37.63101544, 36.74349149, 36.5659867, 38.69604418, 42.60114956, 40.82610166, 40.47109208, 40.64859687, 36.92099628, 37.63101544, 33.01589091, 31.95086217, 31.4183478, 33.1933957, 33.37090049, 33.1933957, 34.25842444, 36.5659867, 34.43592923, 35.85596755, 36.38848191, 38.69604418, 39.76107292, 41.35861603, 39.22855855, 38.87354897, 40.1160825, 40.1160825, 39.22855855, 40.64859687, 43.48867351, 46.32875015, 47.74878846, 47.92629325, 49.16882678, 49.70134115, 48.8138172, 49.16882678, 48.99132199, 46.50625494, 48.8138172, 46.86126451, 44.90871183, 43.48867351, 44.73120704, 49.34633157, 49.52383636, 50.23385552, 49.52383636, 50.05635073, 49.52383636, 47.92629325, 48.28130283, 48.8138172, 50.94387468, 59.02034262, 59.02034262, 64.78924829, 62.12667644, 67.89558211, 62.57043841, 64.34548631, 63.45796236, 72.33320186, 75.43953568, 78.10210752, 83.87101319, 80.76467937, 78.5458695, 83.87101319, 84.31477517, 88.75239491, 84.75853714, 78.98963147, 75.43953568, 77.65834555, 75.43953568, 66.56429619, 72.33320186, 70.11439198, 71.44567791, 71.00191593, 64.78924829, 64.78924829, 62.57043841, 67.00805816, 70.55815396, 67.00805816, 67.00805816, 66.12053421, 64.34548631, 61.68291447, 58.39907585, 54.67147527, 62.12667644, 62.48168602, 67.27431535, 67.9843345, 67.09681056, 65.14425787, 64.78924829, 70.11439198, 72.24444946, 70.82441114, 68.51684887, 66.74180098, 64.6117435, 65.85427703, 63.90172434, 64.6117435, 65.32176266, 65.14425787, 65.32176266, 59.2865998, 55.20398964, 54.1389609, 56.62402796, 57.5115519, 55.9140088, 51.92015103, 46.68375973, 46.15124536, 47.92629325, 50.14510313, 55.20398964, 56.62402796, 59.10909501, 57.68905669, 55.73650401, 61.23915249, 57.15654233, 59.81911417, 60.35162854, 62.30418123, 62.12667644, 57.86656148, 58.39907585, 68.16183929, 65.85427703, 68.16183929, 68.87185845, 65.49926745, 68.69435366, 68.69435366, 67.9843345, 71.88943988, 73.13197341, 70.29189677, 69.22686803, 68.51684887, 66.74180098, 66.03178182, 67.09681056, 67.9843345, 71.00191593, 75.43953568, 72.95446862, 74.72951652, 77.74709795, 77.56959316, 76.859574, 76.50456442, 77.74709795, 74.90702131, 74.37450694, 75.61704047, 78.63462189, 78.27961231, 79.34464105, 80.76467937, 83.60475601, 90.52744281, 90.52744281, 91.85872874, 100.29020625, 99.84644428, 96.29634848, 101.1777302, 98.07139638, 95.40882453, 95.40882453, 91.41496676, 96.74011046, 96.74011046, 101.1777302, 97.62763441, 94.96506256, 94.52130058, 94.96506256, 91.41496676, 94.52130058, 97.62763441, 101.62149218, 99.84644428, 97.18387243, 97.18387243, 98.95892033, 101.62149218, 96.29634848, 96.29634848, 93.19001466, 97.18387243, 97.62763441, 99.4026823, 100.73396823, 96.29634848, 96.29634848, 98.95892033, 100.73396823, 102.06525415, 105.17158797, 107.39039785, 110.05296969, 108.72168377, 110.05296969, 115.82187536, 117.59692326, 121.14701906, 120.70325708, 118.92820919, 116.70939931, 117.15316129, 123.36582893, 120.25949511, 126.47216275, 134.90364027, 130.9097825, 128.24721065, 134.4598783, 127.3596867, 127.3596867, 131.35354447, 132.24106842, 122.47830498, 121.59078103, 121.14701906, 114.49058944, 116.26563734, 119.37197116, 120.25949511, 125.5846388, 132.24106842, 137.12245014, 136.67868817, 140.67254594, 135.34740224, 128.69097263, 138.45373607, 135.79116422, 127.3596867, 137.56621212, 132.24106842, 134.4598783, 154.87292913, 170.84836021, 163.30440664, 155.3166911, 145.55392766, 149.54778543, 149.99154741, 149.99154741],
							[29.28829032, 28.93328074, 28.57827116, 27.868252, 28.93328074, 28.22326158, 28.22326158, 28.75577595, 30.53082385, 29.99830948, 29.99830948, 31.95086217, 32.83838612, 32.66088133, 34.43592923, 33.72591007, 32.48337654, 34.43592923, 34.79093881, 34.08091965, 34.61343402, 32.48337654, 32.48337654, 32.12836696, 31.24084301, 29.82080469, 30.17581427, 29.99830948, 28.93328074, 28.93328074, 29.11078553, 27.33573763, 27.69074721, 29.99830948, 30.88583343, 30.70832864, 31.59585259, 31.06333822, 28.75577595, 29.99830948, 26.27070889, 26.98072805, 26.80322326, 27.69074721, 27.69074721, 27.15823284, 28.22326158, 29.6432999, 28.22326158, 27.69074721, 28.75577595, 29.82080469, 30.88583343, 31.4183478, 29.46579511, 28.57827116, 29.46579511, 29.82080469, 29.99830948, 31.24084301, 31.4183478, 31.4183478, 30.35331906, 31.4183478, 33.01589091, 33.01589091, 31.95086217, 31.95086217, 30.88583343, 31.4183478, 31.59585259, 31.24084301, 31.24084301, 30.35331906, 31.24084301, 33.01589091, 32.48337654, 34.61343402, 34.79093881, 36.5659867, 36.5659867, 35.67846276, 36.92099628, 37.63101544, 38.87354897, 42.06863519, 41.35861603, 44.37619746, 42.06863519, 48.28130283, 46.86126451, 45.97374057, 45.26372141, 47.39377888, 48.10379804, 50.94387468, 57.24529472, 50.05635073, 49.34633157, 50.5888651, 52.54141779, 56.35777077, 51.29888426, 48.99132199, 42.77865435, 42.24613998, 41.53612082, 37.27600586, 39.40606334, 37.80852023, 38.3410346, 41.00360645, 41.71362561, 40.47109208, 39.76107292, 41.35861603, 43.6661783, 42.95615914, 42.77865435, 43.13366393, 43.6661783, 42.60114956, 40.29358729, 39.93857771, 41.8911304, 41.62487321, 41.53612082, 41.71362561, 43.22241632, 43.48867351, 45.26372141, 47.21627409, 47.30502649, 45.52997859, 46.41750254, 44.99746422, 42.24613998, 43.31116872, 41.35861603, 41.44736843, 41.97988279, 41.802378, 41.62487321, 37.18725347, 34.9684436, 33.81466246, 37.80852023, 38.51853939, 37.54226305, 36.47723431, 37.54226305, 36.29972952, 36.5659867, 37.63101544, 40.55984448, 40.55984448, 40.64859687, 40.02733011, 40.47109208, 44.02118788, 44.02118788, 43.93243548, 43.6661783, 44.64245464, 45.3524738, 42.15738758, 41.00360645, 44.99746422, 46.15124536, 48.54756002, 49.52383636, 47.48253128, 48.8138172, 47.92629325, 46.68375973, 47.92629325, 49.61258876, 47.30502649, 49.34633157, 48.19255044, 46.15124536, 45.17496901, 45.88498817, 47.21627409, 47.74878846, 48.37005523, 46.50625494, 45.88498817, 48.10379804, 48.28130283, 47.57128367, 46.23999775, 47.57128367, 45.88498817, 45.3524738, 45.70748338, 46.32875015, 44.37619746, 45.08621662, 43.84368309, 44.99746422, 46.06249296, 45.97374057, 46.68375973, 45.88498817, 47.1275217, 48.01504565, 48.28130283, 47.48253128, 46.50625494, 47.21627409, 44.81995943, 45.17496901, 44.64245464, 45.3524738, 44.28744506, 44.19869267, 44.28744506, 43.48867351, 42.51239716, 43.48867351, 42.95615914, 44.02118788, 43.5774259, 42.42364477, 41.8911304, 41.8911304, 43.93243548, 42.51239716, 41.97988279, 40.55984448, 41.18111124, 40.91485406, 40.55984448, 42.06863519, 41.00360645, 41.26986364, 41.62487321, 42.60114956, 44.46494985, 43.6661783, 44.02118788, 44.37619746, 44.19869267, 44.28744506, 43.5774259, 43.75493069, 43.75493069, 44.02118788, 45.70748338, 45.4412262, 45.3524738, 45.79623578, 44.81995943, 48.01504565, 48.37005523, 47.57128367, 48.9025696, 49.52383636, 47.74878846, 47.74878846, 47.57128367, 47.39377888, 48.37005523, 48.63631241, 48.54756002, 48.10379804, 48.8138172, 48.10379804, 47.74878846, 46.59500733, 48.99132199, 49.08007439, 48.10379804, 49.52383636, 50.5888651, 52.2751606, 58.04406627, 55.38149443, 53.25143695, 53.42894174, 54.84898006, 56.62402796, 55.9140088, 55.73650401, 53.60644653, 50.05635073, 48.10379804, 50.41136031, 50.50011271, 50.50011271],
							[12.07032571, 12.07032571, 12.78034487, 12.60284008, 13.49036403, 13.13535445, 12.86909726, 13.84537361, 14.91040235, 14.46664037, 14.37788798, 14.37788798, 14.20038319, 14.82164995, 15.08790714, 16.33044066, 15.97543108, 16.15293587, 16.24168827, 17.3954694, 17.75047898, 16.68545024, 17.4842218, 16.33044066, 16.41919306, 16.06418348, 15.97543108, 16.24168827, 16.15293587, 15.26541193, 16.41919306, 16.59669785, 16.59669785, 17.04045982, 17.57297419, 17.66172659, 18.10548856, 21.30057478, 19.34802209, 20.41305083, 18.99301251, 19.88053646, 19.52552688, 20.23554604, 20.9455652, 20.76806041, 23.96314663, 21.65558436, 22.54310831, 22.36560352, 23.96314663, 25.20568016, 26.44821368, 26.80322326, 26.0932041, 29.11078553, 28.22326158, 28.75577595, 28.57827116, 29.99830948, 30.53082385, 30.70832864, 30.70832864, 32.48337654, 34.61343402, 36.21097713, 35.50095797, 37.80852023, 37.98602502, 37.63101544, 35.32345318, 35.85596755, 37.27600586, 36.38848191, 34.79093881, 39.22855855, 44.37619746, 44.19869267, 45.97374057, 45.79623578, 43.84368309, 44.90871183, 47.21627409, 45.97374057, 49.34633157, 49.70134115, 48.10379804, 50.05635073, 46.86126451, 47.0387693, 46.86126451, 45.79623578, 51.83139863, 52.00890342, 52.18640821, 53.25143695, 56.80153275, 59.02034262, 55.02648485, 58.13281867, 55.9140088, 59.02034262, 56.35777077, 50.76636989, 44.37619746, 50.94387468, 50.76636989, 50.23385552, 51.65389384, 48.45880762, 52.89642737, 49.52383636, 49.70134115, 51.12137947, 51.47638905, 51.29888426, 52.89642737, 54.1389609, 55.02648485, 55.47024682, 52.89642737, 50.23385552, 52.98517976, 55.02648485, 55.9140088, 54.84898006, 60.88414291, 61.77166686, 61.77166686, 61.23915249, 61.94917165, 64.43423871, 65.32176266, 61.41665728, 61.41665728, 63.72421955, 63.19170518, 63.72421955, 63.36920997, 60.35162854, 59.46410459, 57.68905669, 59.81911417, 56.44652317, 53.78395132, 53.42894174, 59.64160938, 57.33404711, 54.84898006, 48.8138172, 45.4412262, 50.05635073, 51.12137947, 52.71892258, 57.33404711, 49.70134115, 51.12137947, 51.56514145, 47.0387693, 48.72506481, 48.54756002, 46.86126451, 47.1275217, 49.96759834, 53.25143695, 51.12137947, 51.65389384, 52.54141779, 53.42894174, 52.54141779, 54.84898006, 53.16268455, 53.07393216, 47.83754086, 48.28130283, 50.41136031, 50.05635073, 49.70134115, 51.12137947, 51.12137947, 49.16882678, 49.34633157, 51.29888426, 50.94387468, 49.43508397, 53.25143695, 51.83139863, 55.20398964, 53.25143695, 55.38149443, 54.84898006, 54.84898006, 55.38149443, 52.363913, 51.12137947, 50.76636989, 52.363913, 50.76636989, 48.10379804, 49.08007439, 48.9025696, 53.78395132, 53.78395132, 55.73650401, 54.1389609, 53.60644653, 50.85512229, 50.23385552, 50.85512229, 52.45266539, 50.94387468, 48.45880762, 49.96759834, 50.50011271, 49.52383636, 51.56514145, 53.78395132, 54.31646569, 55.02648485, 53.25143695, 52.363913, 52.89642737, 54.31646569, 52.98517976, 52.18640821, 50.32260792, 51.56514145, 52.80767497, 50.94387468, 50.94387468, 50.50011271, 50.94387468, 52.54141779, 51.21013187, 52.00890342, 48.8138172, 49.70134115, 50.05635073, 51.12137947, 51.12137947, 49.16882678, 47.48253128, 48.63631241, 47.21627409, 48.54756002, 50.5888651, 48.63631241, 49.70134115, 52.71892258, 52.09765581, 51.29888426, 51.83139863, 54.1389609, 52.54141779, 55.73650401, 57.68905669, 58.57658064, 60.70663812, 61.23915249, 61.59416207, 60.35162854, 58.57658064, 58.57658064, 57.86656148, 56.44652317, 55.9140088, 55.20398964, 56.26901838, 57.5115519, 56.80153275, 56.62402796, 57.86656148, 57.68905669, 58.04406627, 59.99661896, 57.86656148, 59.2865998, 59.46410459, 61.77166686, 59.81911417, 61.23915249, 61.23915249, 58.75408543, 51.47638905, 50.94387468, 49.96759834, 49.43508397, 48.63631241, 48.9025696, 48.99132199, 48.45880762],
							[17.3954694, 17.57297419, 17.75047898, 18.99301251, 18.63800293, 17.92798377, 18.28299335, 18.63800293, 19.88053646, 20.76806041, 20.23554604, 20.41305083, 20.76806041, 22.89811789, 22.89811789, 22.7206131, 22.36560352, 23.60813705, 23.78564184, 23.60813705, 24.31815621, 22.54310831, 23.60813705, 23.43063226, 22.89811789, 22.18809873, 22.18809873, 22.01059394, 21.83308915, 22.01059394, 22.7206131, 22.18809873, 22.01059394, 23.78564184, 24.14065142, 23.60813705, 24.67316579, 24.85067058, 22.89811789, 24.31815621, 22.18809873, 21.83308915, 21.83308915, 23.96314663, 24.31815621, 23.78564184, 25.56068974, 25.02817537, 24.495661, 24.495661, 25.02817537, 26.44821368, 28.04575679, 29.28829032, 28.22326158, 29.11078553, 30.17581427, 31.06333822, 30.35331906, 33.72591007, 33.90341486, 33.37090049, 32.66088133, 32.83838612, 32.83838612, 32.66088133, 30.70832864, 31.06333822, 30.70832864, 29.99830948, 30.53082385, 27.868252, 28.04575679, 27.69074721, 27.51324242, 28.75577595, 29.82080469, 29.99830948, 28.93328074, 28.75577595, 30.35331906, 29.6432999, 30.53082385, 30.88583343, 32.48337654, 34.43592923, 34.43592923, 37.27600586, 35.14594839, 37.63101544, 36.74349149, 36.74349149, 35.14594839, 39.76107292, 39.76107292, 42.95615914, 45.26372141, 41.53612082, 43.13366393, 44.37619746, 44.37619746, 46.15124536, 43.6661783, 42.06863519, 39.05105376, 36.74349149, 39.05105376, 36.5659867, 39.58356813, 38.51853939, 41.18111124, 42.24613998, 38.51853939, 38.16352981, 36.92099628, 38.51853939, 39.93857771, 39.93857771, 39.22855855, 38.87354897, 39.58356813, 38.51853939, 36.74349149, 35.67846276, 39.76107292, 41.26986364, 42.60114956, 42.24613998, 42.68990195, 42.33489237, 42.33489237, 42.60114956, 43.22241632, 43.48867351, 43.13366393, 42.42364477, 41.53612082, 41.97988279, 40.02733011, 40.1160825, 42.24613998, 40.82610166, 40.29358729, 36.92099628, 31.95086217, 31.50710019, 33.99216725, 33.63715767, 33.1046433, 31.15209062, 26.27070889, 28.84452835, 30.08706188, 31.95086217, 34.08091965, 33.45965288, 33.63715767, 33.28214809, 31.68460498, 34.08091965, 32.48337654, 33.90341486, 34.34717683, 34.52468162, 34.8796912, 33.63715767, 33.72591007, 38.69604418, 34.9684436, 36.12222473, 35.94471994, 36.12222473, 34.61343402, 34.16967204, 34.52468162, 36.03347234, 38.16352981, 36.74349149, 36.47723431, 37.36475826, 36.21097713, 35.94471994, 37.71976784, 39.31731095, 40.64859687, 40.82610166, 39.40606334, 41.44736843, 41.8911304, 42.33489237, 42.15738758, 41.71362561, 43.48867351, 40.1160825, 41.09235885, 42.42364477, 43.39992111, 44.90871183, 44.73120704, 45.26372141, 46.95001691, 48.19255044, 49.25757918, 51.65389384, 52.45266539, 55.73650401, 53.25143695, 57.68905669, 55.55899922, 54.31646569, 54.67147527, 50.41136031, 52.00890342, 53.25143695, 54.49397048, 52.89642737, 53.96145611, 54.31646569, 52.09765581, 50.50011271, 53.16268455, 54.1389609, 56.44652317, 55.9140088, 53.96145611, 54.31646569, 55.55899922, 55.9140088, 55.20398964, 56.80153275, 55.9140088, 57.33404711, 59.99661896, 60.52913333, 61.77166686, 60.35162854, 59.10909501, 59.64160938, 61.41665728, 62.12667644, 63.01420039, 64.96675308, 66.91930577, 66.20928661, 66.56429619, 67.45182014, 68.51684887, 70.11439198, 71.5344303, 70.46940156, 71.00191593, 71.88943988, 77.39208837, 75.61704047, 70.11439198, 72.06694467, 68.16183929, 67.80682971, 70.82441114, 66.3867914, 68.51684887, 70.11439198, 71.35692551, 67.62932492, 67.45182014, 68.16183929, 67.09681056, 67.80682971, 68.69435366, 73.84199257, 75.0845261, 78.81212668, 75.97205005, 79.16713626, 78.63462189, 84.66978475, 80.58717458, 84.13727038, 85.91231828, 84.13727038, 84.66978475, 83.42725122, 84.31477517, 86.62233744, 90.97120479, 85.20229912, 80.05466021, 73.3094782, 76.14955484, 78.4571171, 78.63462189],
							[24.61991435, 24.85067058, 26.39496225, 26.62571847, 25.73819453, 25.06367632, 24.17615237, 24.61991435, 26.62571847, 27.2824862, 28.84452835, 31.06333822, 32.60762989, 35.05719599, 33.72591007, 31.72010594, 33.05139187, 35.50095797, 32.60762989, 32.83838612, 33.72591007, 31.50710019, 31.50710019, 33.28214809, 33.49515384, 29.96280852, 32.1816184, 32.39462414, 29.96280852, 31.06333822, 32.60762989, 30.83258199, 31.50710019, 33.28214809, 34.82643976, 33.72591007, 33.93891582, 35.27020174, 30.38882002, 32.1816184, 28.17001015, 26.83872422, 26.83872422, 27.2824862, 27.06948045, 27.51324242, 29.7320523, 29.7320523, 28.17001015, 27.51324242, 27.74399865, 30.17581427, 30.61957625, 31.72010594, 29.28829032, 27.51324242, 29.7320523, 29.0575341, 28.61377212, 30.17581427, 29.7320523, 31.50710019, 32.1816184, 34.82643976, 35.50095797, 35.71396371, 34.61343402, 35.71396371, 34.61343402, 33.49515384, 33.93891582, 32.83838612, 31.50710019, 29.96280852, 31.95086217, 31.95086217, 32.83838612, 33.05139187, 31.95086217, 32.39462414, 32.83838612, 31.50710019, 31.95086217, 31.72010594, 34.82643976, 36.38848191, 38.16352981, 44.14544123, 44.37619746, 48.37005523, 46.59500733, 46.15124536, 48.8138172, 55.9140088, 52.80767497, 55.47024682, 62.57043841, 55.9140088, 57.24529472, 60.79539052, 62.12667644, 66.12053421, 63.01420039, 59.46410459, 51.92015103, 50.5888651, 52.363913, 46.59500733, 46.15124536, 43.27566776, 43.93243548, 44.81995943, 41.26986364, 40.38233969, 39.93857771, 45.26372141, 46.59500733, 47.0387693, 44.81995943, 41.48286938, 40.82610166, 38.16352981, 37.71976784, 37.89727263, 41.09235885, 39.49481574, 41.8911304, 41.97988279, 42.06863519, 42.06863519, 41.18111124, 42.24613998, 43.75493069, 42.60114956, 41.00360645, 39.76107292, 37.48901161, 39.05105376, 37.48901161, 37.71976784, 37.93277359, 37.04524964, 37.04524964, 32.60762989, 28.17001015, 26.83872422, 28.40076637, 27.74399865, 27.51324242, 25.06367632, 24.85067058, 25.73819453, 28.84452835, 27.9570044, 26.83872422, 28.40076637, 27.9570044, 27.74399865, 25.95120027, 27.15823284, 27.77949961, 28.13450919, 27.51324242, 26.71447087, 26.00445171, 24.67316579, 25.11692776, 29.7320523, 31.24084301, 31.15209062, 30.26456667, 29.28829032, 29.46579511, 29.02203314, 29.28829032, 31.15209062, 33.45965288, 31.95086217, 32.83838612, 31.77335738, 28.93328074, 28.57827116, 28.75577595, 28.40076637, 28.84452835, 33.28214809, 33.01589091, 33.90341486, 37.45351065, 37.36475826, 36.6547391, 36.74349149, 37.63101544, 35.94471994, 34.61343402, 35.14594839, 36.6547391, 35.41220557, 36.83224389, 36.21097713, 35.50095797, 39.40606334, 38.87354897, 39.22855855, 41.97988279, 42.95615914, 38.3410346, 41.71362561, 41.35861603, 38.51853939, 40.29358729, 39.05105376, 42.42364477, 40.55984448, 40.47109208, 38.78479658, 39.76107292, 38.96230137, 38.78479658, 37.54226305, 37.54226305, 39.05105376, 42.06863519, 41.35861603, 38.78479658, 38.78479658, 40.2048349, 39.84982532, 39.49481574, 38.96230137, 38.429787, 39.93857771, 39.67232053, 40.38233969, 41.44736843, 39.76107292, 39.22855855, 39.40606334, 40.91485406, 40.29358729, 42.06863519, 43.13366393, 43.84368309, 43.75493069, 44.99746422, 51.29888426, 51.29888426, 55.02648485, 56.97903754, 55.55899922, 54.84898006, 55.73650401, 58.22157106, 56.62402796, 58.22157106, 63.01420039, 59.64160938, 55.73650401, 58.22157106, 55.73650401, 53.25143695, 51.56514145, 52.45266539, 50.32260792, 46.06249296, 45.4412262, 42.60114956, 41.97988279, 41.71362561, 41.00360645, 43.13366393, 47.48253128, 46.06249296, 47.92629325, 48.63631241, 47.48253128, 42.51239716, 45.88498817, 45.3524738, 43.31116872, 44.02118788, 42.06863519, 42.33489237, 48.45880762, 52.71892258, 47.30502649, 43.31116872, 40.47109208, 42.24613998, 41.71362561, 41.00360645],
							[43.70167926, 44.78445847, 47.78428942, 48.88481912, 49.96759834, 50.251606, 50.51786319, 57.68905669, 58.0263158, 59.05584358, 59.05584358, 59.05584358, 60.75988956, 62.12667644, 62.46393554, 60.75988956, 62.46393554, 68.26834217, 71.67643413, 68.26834217, 71.67643413, 71.00191593, 73.04322101, 75.77679478, 73.7354897, 70.32739773, 69.63512905, 71.00191593, 68.96061085, 69.63512905, 68.96061085, 68.26834217, 67.23881439, 72.36870281, 74.4100079, 73.04322101, 76.46906346, 79.20263722, 73.04322101, 75.10227658, 64.8602502, 65.1975093, 67.23881439, 71.00191593, 72.36870281, 73.04322101, 77.83585034, 79.20263722, 77.83585034, 77.83585034, 79.20263722, 81.2439423, 86.01882115, 91.48596868, 90.1191818, 88.07787671, 94.21954244, 99.4026823, 96.74011046, 103.84030205, 103.84030205, 101.1777302, 101.1777302, 102.06525415, 100.29020625, 104.727826, 103.84030205, 103.84030205, 104.727826, 101.1777302, 103.84030205, 102.06525415, 102.9527781, 100.29020625, 103.84030205, 109.16544574, 107.39039785, 102.06525415, 97.62763441, 95.85258651, 95.85258651, 93.19001466, 94.96506256, 94.96506256, 97.62763441, 102.9527781, 101.1777302, 104.727826, 101.1777302, 105.61534995, 103.84030205, 106.5028739, 106.5028739, 114.49058944, 129.57849658, 128.69097263, 134.01611632, 126.91592473, 126.91592473, 131.35354447, 131.35354447, 142.00383186, 137.56621212, 129.57849658, 119.81573313, 102.06525415, 101.1777302, 93.19001466, 93.19001466, 89.63991886, 91.41496676, 93.19001466, 92.30249071, 91.41496676, 88.30863294, 91.41496676, 100.29020625, 98.51515836, 95.85258651, 93.19001466, 96.74011046, 89.63991886, 91.85872874, 89.63991886, 96.74011046, 95.85258651, 97.18387243, 97.18387243, 96.74011046, 93.63377663, 93.19001466, 100.73396823, 99.4026823, 98.95892033, 96.74011046, 96.29634848, 94.07753861, 94.96506256, 98.51515836, 98.07139638, 101.62149218, 102.50901613, 105.17158797, 101.62149218, 100.29020625, 94.07753861, 98.07139638, 100.29020625, 98.51515836, 93.19001466, 84.84728954, 83.24974643, 75.97205005, 75.97205005, 81.65220332, 81.47469853, 82.36222248, 85.5573087, 83.60475601, 91.41496676, 94.07753861, 96.74011046, 90.52744281, 93.19001466, 94.07753861, 90.52744281, 90.52744281, 100.73396823, 101.1777302, 104.28406402, 107.83415982, 103.84030205, 103.39654008, 103.39654008, 104.727826, 110.05296969, 114.49058944, 111.82801759, 113.60306549, 114.49058944, 114.04682747, 108.2779218, 110.94049364, 111.82801759, 113.60306549, 115.37811339, 111.82801759, 113.15930352, 113.60306549, 113.15930352, 115.37811339, 113.60306549, 116.70939931, 113.60306549, 111.38425562, 114.93435141, 120.70325708, 122.03454301, 122.03454301, 122.03454301, 122.92206696, 124.25335288, 123.36582893, 126.91592473, 130.9097825, 130.9097825, 131.35354447, 133.12859237, 130.9097825, 137.56621212, 135.34740224, 133.12859237, 140.67254594, 140.22878397, 138.89749804, 140.67254594, 142.00383186, 135.34740224, 135.79116422, 134.4598783, 139.34126002, 140.22878397, 143.33511779, 141.11630791, 138.00997409, 138.45373607, 138.45373607, 138.45373607, 134.90364027, 136.23492619, 136.67868817, 139.34126002, 141.11630791, 140.67254594, 144.22264174, 140.67254594, 138.45373607, 143.33511779, 140.67254594, 143.33511779, 148.66026148, 155.76045308, 165.07945454, 161.97312072, 164.19193059, 168.18578836, 168.18578836, 161.52935874, 165.52321652, 160.19807282, 159.75431085, 163.30440664, 164.19193059, 164.19193059, 161.08559677, 169.51707429, 161.97312072, 161.97312072, 165.52321652, 157.091739, 161.97312072, 158.8667869, 149.54778543, 145.99768963, 145.11016569, 142.00383186, 140.67254594, 144.66640371, 147.32897556, 152.21035728, 156.64797702, 166.41074046, 164.19193059, 160.19807282, 165.07945454, 158.8667869, 157.091739, 194.36774486, 196.14279276, 196.14279276, 197.91784066, 197.03031671, 202.35546041, 199.69288856, 182.82993352, 185.49250537, 181.94240957, 165.96697849, 169.51707429, 166.85450244, 160.6418348],
							[44.21644315, 43.73718021, 44.69570608, 43.89693452, 43.89693452, 43.08041249, 43.2401668, 45.17496901, 46.7902626, 50.42911079, 48.8138172, 49.61258876, 51.22788234, 54.0502085, 53.65969797, 51.63614336, 53.25143695, 57.28079568, 54.86673054, 54.0502085, 54.0502085, 48.40555619, 50.02084977, 50.02084977, 49.22207822, 47.60678463, 47.76653894, 50.02084977, 48.08604756, 49.22207822, 49.22207822, 50.42911079, 52.04440438, 54.45846952, 57.68905669, 55.27499155, 56.07376311, 54.86673054, 50.02084977, 50.42911079, 44.69570608, 42.12188663, 42.60114956, 44.69570608, 45.0152147, 44.69570608, 46.31099967, 46.95001691, 44.53595177, 44.05668884, 46.63050829, 50.83737181, 53.25143695, 54.45846952, 49.61258876, 49.22207822, 53.25143695, 52.45266539, 51.63614336, 56.89028514, 56.89028514, 58.48782825, 60.51138285, 62.53493746, 62.92544799, 63.74197003, 61.71841542, 63.33370901, 62.53493746, 61.32790489, 63.74197003, 61.71841542, 59.7126113, 57.28079568, 60.10312184, 63.33370901, 62.12667644, 62.92544799, 61.32790489, 60.51138285, 62.92544799, 61.71841542, 62.53493746, 65.76552463, 68.58785079, 75.84779669, 77.85360082, 85.52180774, 85.52180774, 96.01234082, 90.3676885, 87.86487097, 88.75239491, 98.51515836, 103.84030205, 118.04068524, 126.02840078, 123.36582893, 115.37811339, 117.15316129, 115.37811339, 122.47830498, 112.71554154, 107.39039785, 101.1777302, 104.727826, 104.727826, 94.07753861, 98.51515836, 94.07753861, 98.51515836, 95.85258651, 82.98348925, 82.53972727, 82.0959653, 86.97734702, 94.96506256, 90.52744281, 89.63991886, 88.75239491, 86.08982307, 82.0959653, 76.50456442, 75.79454526, 85.37980391, 84.49227996, 89.19615689, 90.97120479, 90.97120479, 92.74625269, 94.52130058, 100.73396823, 106.05911192, 106.5028739, 102.50901613, 102.06525415, 98.51515836, 101.62149218, 99.84644428, 101.62149218, 102.50901613, 101.62149218, 100.29020625, 88.75239491, 85.5573087, 81.65220332, 83.07224164, 83.42725122, 82.0072129, 73.3094782, 68.51684887, 70.82441114, 74.37450694, 76.859574, 83.95976559, 84.13727038, 89.19615689, 91.41496676, 87.3323566, 96.29634848, 92.74625269, 93.63377663, 92.30249071, 94.07753861, 91.85872874, 87.68736618, 88.21988055, 99.4026823, 94.52130058, 98.07139638, 102.50901613, 100.29020625, 102.50901613, 101.62149218, 101.62149218, 105.17158797, 106.05911192, 102.9527781, 102.9527781, 101.1777302, 94.52130058, 93.63377663, 96.29634848, 102.06525415, 102.9527781, 110.94049364, 109.16544574, 111.38425562, 111.82801759, 111.38425562, 108.2779218, 107.39039785, 110.05296969, 103.84030205, 101.62149218, 106.5028739, 110.94049364, 108.2779218, 106.94663587, 109.16544574, 112.27177957, 118.48444721, 114.93435141, 114.93435141, 124.25335288, 131.35354447, 125.5846388, 129.57849658, 126.91592473, 124.25335288, 123.36582893, 117.15316129, 125.14087683, 122.92206696, 131.79730645, 128.69097263, 128.24721065, 126.02840078, 128.24721065, 128.69097263, 132.24106842, 138.00997409, 140.22878397, 140.67254594, 136.67868817, 138.00997409, 138.89749804, 138.89749804, 132.24106842, 133.57235435, 129.1347346, 131.79730645, 134.01611632, 134.90364027, 138.45373607, 134.01611632, 132.24106842, 137.12245014, 140.22878397, 143.33511779, 146.44145161, 153.98540518, 153.98540518, 152.65411925, 154.42916715, 157.091739, 160.6418348, 167.74202639, 170.40459824, 171.29212219, 161.08559677, 166.41074046, 169.96083626, 167.29826441, 161.97312072, 166.41074046, 159.75431085, 155.3166911, 165.07945454, 157.97926295, 160.6418348, 158.8667869, 157.53550097, 148.21649951, 145.99768963, 145.55392766, 138.45373607, 138.00997409, 138.89749804, 143.77887976, 155.3166911, 162.86064467, 166.41074046, 168.18578836, 169.07331231, 164.63569257, 152.21035728, 169.07331231, 165.52321652, 153.09788123, 164.19193059, 157.97926295, 161.97312072, 179.27983773, 185.49250537, 170.84836021, 166.41074046, 156.20421505, 156.64797702, 158.8667869, 163.74816862],
							[41.1278598, 41.41186747, 43.18691537, 45.84948721, 45.49447763, 45.84948721, 46.59500733, 48.44105714, 50.65986702, 49.55933732, 49.55933732, 50.28710696, 51.03262708, 51.03262708, 52.15090725, 51.03262708, 49.93209738, 51.03262708, 51.03262708, 49.93209738, 50.65986702, 46.96776739, 47.34052745, 48.06829709, 48.06829709, 46.96776739, 47.34052745, 46.59500733, 45.49447763, 46.59500733, 47.34052745, 47.34052745, 49.55933732, 53.25143695, 53.25143695, 53.99695707, 52.15090725, 53.25143695, 48.44105714, 49.93209738, 44.74895752, 44.74895752, 45.49447763, 47.71328751, 48.8138172, 48.8138172, 50.65986702, 50.65986702, 51.40538713, 50.28710696, 51.40538713, 51.40538713, 54.7247267, 54.7247267, 52.15090725, 52.15090725, 54.35196665, 53.25143695, 52.15090725, 53.62419701, 54.35196665, 53.99695707, 53.25143695, 55.84300688, 56.94353658, 58.78958639, 58.06181675, 58.43457681, 58.06181675, 58.43457681, 59.90786657, 58.43457681, 57.68905669, 55.47024682, 58.06181675, 61.75391638, 62.12667644, 63.97272625, 62.87219656, 63.97272625, 64.34548631, 62.12667644, 64.71824637, 66.56429619, 69.88363576, 74.69401556, 73.96624592, 76.91282543, 73.2207258, 80.62267554, 79.87715542, 82.84148541, 82.0959653, 92.30249071, 91.41496676, 93.19001466, 100.29020625, 89.63991886, 88.30863294, 87.42110899, 85.64606109, 86.53358504, 82.53972727, 78.98963147, 72.77696383, 76.32705963, 75.88329765, 70.55815396, 71.44567791, 70.55815396, 72.33320186, 78.10210752, 77.65834555, 71.44567791, 71.00191593, 73.2207258, 75.88329765, 74.55201173, 73.66448778, 73.2207258, 74.55201173, 72.77696383, 68.69435366, 66.56429619, 69.40437282, 70.46940156, 72.06694467, 70.29189677, 68.87185845, 69.04936324, 69.40437282, 72.59945904, 74.37450694, 72.06694467, 70.11439198, 69.7593824, 69.40437282, 71.5344303, 69.7593824, 70.11439198, 69.40437282, 67.9843345, 68.16183929, 63.19170518, 56.62402796, 55.02648485, 58.93159022, 60.88414291, 58.57658064, 53.96145611, 55.20398964, 55.73650401, 57.68905669, 59.81911417, 66.20928661, 67.09681056, 66.56429619, 64.07922913, 63.72421955, 66.56429619, 67.09681056, 66.56429619, 64.96675308, 66.20928661, 64.96675308, 64.78924829, 64.78924829, 69.22686803, 70.11439198, 73.13197341, 75.43953568, 72.95446862, 72.59945904, 71.17942072, 70.64690635, 74.37450694, 71.88943988, 69.93688719, 72.24444946, 70.46940156, 66.74180098, 67.27431535, 66.56429619, 67.62932492, 69.7593824, 72.42195425, 70.82441114, 71.00191593, 72.77696383, 72.77696383, 72.42195425, 72.42195425, 73.48698299, 65.14425787, 62.48168602, 64.25673392, 65.14425787, 63.54671476, 62.8366956, 61.94917165, 63.19170518, 64.25673392, 64.78924829, 67.45182014, 66.20928661, 67.45182014, 66.91930577, 67.09681056, 66.20928661, 64.43423871, 64.43423871, 60.17412375, 62.8366956, 61.94917165, 63.19170518, 61.23915249, 62.30418123, 63.54671476, 63.72421955, 65.14425787, 63.36920997, 63.90172434, 65.32176266, 64.96675308, 63.01420039, 62.65919081, 62.30418123, 61.41665728, 59.64160938, 59.46410459, 58.39907585, 57.5115519, 57.33404711, 56.97903754, 59.46410459, 58.39907585, 57.5115519, 60.17412375, 63.54671476, 63.19170518, 63.01420039, 65.49926745, 66.03178182, 63.54671476, 64.25673392, 61.0616477, 59.2865998, 60.35162854, 58.93159022, 59.46410459, 59.99661896, 59.46410459, 59.2865998, 60.35162854, 59.81911417, 60.35162854, 58.75408543, 64.43423871, 64.78924829, 62.12667644, 61.41665728, 63.54671476, 63.19170518, 62.48168602, 61.59416207, 60.52913333, 59.10909501, 62.65919081, 61.23915249, 61.94917165, 61.23915249, 63.90172434, 65.67677224, 67.80682971, 68.87185845, 79.34464105, 75.0845261, 79.34464105, 77.92460274, 75.43953568, 75.97205005, 78.81212668, 78.98963147, 78.4571171, 78.10210752, 74.55201173, 68.33934408, 63.72421955, 69.58187761, 74.55201173, 73.3094782],
							[2.43181562, 2.34306323, 2.30756227, 3.81635298, 3.67434915, 4.08261017, 4.11811112, 4.4731207, 4.4731207, 4.43761975, 4.61512454, 4.43761975, 4.29561591, 4.34886735, 4.38436831, 4.11811112, 4.91688268, 5.11213795, 5.85765806, 6.03516285, 6.07066381, 5.36064465, 5.62690184, 5.89315902, 6.39017243, 6.336921, 6.15941621, 6.07066381, 6.12391525, 5.89315902, 5.68015327, 5.50264848, 5.76890567, 5.9996619, 6.336921, 6.336921, 6.95818776, 6.65642962, 6.07066381, 6.21266764, 5.36064465, 5.11213795, 5.14763891, 5.36064465, 5.41389609, 5.71565423, 5.85765806, 5.94641046, 5.59140088, 5.55589992, 5.62690184, 5.80440663, 5.80440663, 6.44342387, 6.47892483, 6.07066381, 5.9996619, 6.51442579, 6.21266764, 6.39017243, 6.56767722, 6.56767722, 6.336921, 6.30142004, 6.39017243, 6.39017243, 6.15941621, 6.15941621, 6.21266764, 5.89315902, 5.89315902, 5.62690184, 5.50264848, 5.55589992, 5.44939705, 5.32514369, 5.32514369, 5.27189226, 5.18313986, 5.2363913, 5.55589992, 5.14763891, 5.36064465, 5.18313986, 5.32514369, 5.80440663, 6.56767722, 6.78068297, 6.44342387, 6.95818776, 6.78068297, 6.56767722, 6.74518201, 7.04694016, 8.37822608, 7.98771554, 7.89896315, 8.16522033, 7.33094782, 7.36644878, 7.40194974, 7.36644878, 7.18894399, 7.10019159, 6.30142004, 6.12391525, 6.39017243, 5.80440663, 6.21266764, 5.62690184, 5.62690184, 5.62690184, 5.2363913, 5.18313986, 5.00563507, 5.05888651, 5.62690184, 5.62690184, 5.85765806, 5.50264848, 5.32514369, 5.32514369, 5.27189226, 5.25414178, 5.60915136, 5.30739322, 5.46714753, 5.9109095, 5.83990759, 5.64465232, 5.71565423, 5.76890567, 6.15941621, 5.89315902, 5.80440663, 5.37839513, 5.27189226, 5.11213795, 4.97013412, 5.32514369, 5.2363913, 5.05888651, 5.00563507, 4.5618731, 3.81635298, 3.53234532, 3.72760059, 3.47909388, 3.5500958, 3.24833765, 2.59156993, 2.92882903, 2.82232616, 2.71582328, 3.24833765, 3.1240843, 3.1240843, 2.98208047, 2.89332807, 3.03533191, 2.96432999, 3.4613434, 3.30158909, 3.37259101, 3.44359292, 3.44359292, 3.58559675, 4.4731207, 4.43761975, 4.82813028, 4.88138172, 4.70387693, 4.52637214, 4.5618731, 4.52637214, 4.5618731, 5.00563507, 4.82813028, 4.91688268, 5.41389609, 5.41389609, 5.55589992, 5.50264848, 5.62690184, 6.15941621, 6.30142004, 6.2481686, 5.89315902, 5.89315902, 6.12391525, 6.21266764, 6.07066381, 6.07066381, 5.85765806, 5.62690184, 6.03516285, 6.07066381, 6.336921, 7.04694016, 6.9226868, 7.10019159, 8.1119689, 8.43147752, 7.98771554, 7.66820692, 8.1119689, 7.93446411, 8.1119689, 7.93446411, 7.84571171, 7.81021075, 7.36644878, 8.21847177, 7.89896315, 7.72145836, 7.93446411, 8.07646794, 8.52022991, 8.55573087, 8.60898231, 9.23024907, 9.76276344, 10.02902063, 10.4727826, 9.49650626, 9.31900147, 9.67401105, 10.73903978, 11.09404936, 11.36030655, 11.00529697, 11.27155415, 12.07032571, 12.1590781, 12.78034487, 12.33658289, 11.98157331, 12.51408768, 12.95784966, 13.40161163, 15.17665953, 15.08790714, 14.73289756, 14.99915474, 15.79792629, 16.95170743, 16.41919306, 17.57297419, 19.70303167, 23.34187986, 24.85067058, 24.05189902, 27.24698524, 31.50710019, 29.02203314, 30.35331906, 28.04575679, 29.37704272, 29.46579511, 28.22326158, 28.93328074, 30.88583343, 31.24084301, 30.70832864, 28.48951877, 29.6432999, 28.66702356, 31.86210977, 34.8796912, 41.18111124, 38.78479658, 39.84982532, 38.3410346, 46.50625494, 48.28130283, 47.57128367, 46.50625494, 54.1389609, 67.45182014, 63.72421955, 61.41665728, 63.54671476, 63.72421955, 67.62932492, 67.62932492, 66.56429619, 63.01420039, 57.68905669, 54.84898006, 59.99661896, 56.80153275],
							[29.11078553, 27.868252, 29.28829032, 28.40076637, 28.57827116, 28.93328074, 28.75577595, 29.82080469, 31.06333822, 33.1933957, 32.66088133, 33.1933957, 33.72591007, 34.08091965, 34.9684436, 33.72591007, 33.90341486, 34.9684436, 33.90341486, 33.54840528, 33.72591007, 31.77335738, 30.35331906, 29.82080469, 30.35331906, 28.75577595, 29.11078553, 28.93328074, 27.868252, 28.22326158, 28.04575679, 28.04575679, 27.868252, 30.53082385, 32.30587175, 32.30587175, 32.83838612, 32.66088133, 29.82080469, 29.99830948, 25.02817537, 23.96314663, 23.43063226, 24.85067058, 25.73819453, 25.56068974, 27.33573763, 28.04575679, 27.51324242, 28.04575679, 28.75577595, 31.06333822, 32.48337654, 32.83838612, 31.77335738, 30.17581427, 32.48337654, 32.66088133, 32.12836696, 33.72591007, 34.08091965, 34.61343402, 34.43592923, 38.69604418, 39.40606334, 41.8911304, 39.05105376, 38.16352981, 38.3410346, 38.16352981, 39.05105376, 37.98602502, 36.38848191, 34.79093881, 37.27600586, 39.22855855, 37.45351065, 38.16352981, 36.38848191, 35.85596755, 36.74349149, 36.5659867, 37.63101544, 42.06863519, 42.95615914, 51.47638905, 53.69519892, 56.35777077, 52.363913, 60.35162854, 61.23915249, 60.35162854, 63.01420039, 69.67063001, 72.33320186, 78.10210752, 101.1777302, 89.63991886, 88.30863294, 96.74011046, 94.07753861, 97.62763441, 94.96506256, 84.75853714, 79.87715542, 83.87101319, 84.31477517, 82.53972727, 79.43339345, 74.55201173, 78.10210752, 77.65834555, 67.89558211, 67.45182014, 65.67677224, 71.88943988, 75.88329765, 71.88943988, 71.00191593, 71.00191593, 68.33934408, 65.67677224, 62.30418123, 59.81911417, 69.58187761, 68.33934408, 72.42195425, 71.5344303, 70.46940156, 73.3094782, 73.48698299, 79.34464105, 86.62233744, 87.15485181, 85.02479433, 85.20229912, 83.95976559, 87.68736618, 86.26732786, 87.15485181, 90.52744281, 85.91231828, 88.57489012, 76.50456442, 71.00191593, 64.25673392, 63.72421955, 63.72421955, 65.49926745, 60.70663812, 59.64160938, 58.04406627, 59.64160938, 61.23915249, 66.74180098, 65.67677224, 73.3094782, 75.97205005, 73.66448778, 79.69965063, 76.68206921, 77.21458358, 77.03707879, 79.16713626, 77.56959316, 70.82441114, 69.93688719, 79.87715542, 75.97205005, 78.10210752, 78.63462189, 77.39208837, 77.56959316, 76.14955484, 75.26203089, 78.98963147, 81.82970811, 78.98963147, 77.92460274, 76.14955484, 73.3094782, 72.24444946, 73.84199257, 74.90702131, 76.32705963, 81.29719374, 80.58717458, 82.89473685, 86.44483265, 86.79984223, 85.73481349, 85.37980391, 82.89473685, 79.87715542, 78.10210752, 81.29719374, 84.84728954, 81.82970811, 81.82970811, 82.36222248, 82.71723206, 87.15485181, 90.08368084, 92.74625269, 98.95892033, 104.28406402, 102.06525415, 103.39654008, 102.06525415, 99.84644428, 100.29020625, 93.63377663, 98.51515836, 96.29634848, 98.07139638, 97.62763441, 97.18387243, 94.96506256, 95.85258651, 92.30249071, 96.29634848, 98.51515836, 103.84030205, 101.62149218, 98.07139638, 98.07139638, 102.9527781, 105.17158797, 100.73396823, 101.1777302, 97.62763441, 101.62149218, 103.39654008, 107.39039785, 108.72168377, 107.39039785, 106.05911192, 112.71554154, 114.49058944, 115.37811339, 118.92820919, 122.03454301, 124.69711486, 126.47216275, 124.25335288, 125.5846388, 126.47216275, 135.34740224, 138.00997409, 134.90364027, 127.80344868, 131.35354447, 135.34740224, 132.6848304, 128.24721065, 134.4598783, 127.3596867, 127.3596867, 130.46602052, 122.47830498, 126.91592473, 122.92206696, 122.03454301, 111.82801759, 113.60306549, 114.04682747, 109.60920772, 114.04682747, 114.93435141, 114.93435141, 119.81573313, 126.02840078, 127.80344868, 133.57235435, 134.01611632, 127.80344868, 116.26563734, 128.24721065, 122.03454301, 117.15316129, 123.36582893, 118.48444721, 120.25949511, 133.12859237, 133.12859237, 124.69711486, 120.70325708, 114.04682747, 118.48444721, 115.82187536, 117.15316129],
							[10.91654457, 11.27155415, 11.62656373, 11.36030655, 11.36030655, 11.18280176, 10.73903978, 11.18280176, 11.44905894, 11.98157331, 12.07032571, 13.31285924, 14.64414516, 15.08790714, 14.73289756, 14.55539277, 14.46664037, 15.53166911, 14.99915474, 14.91040235, 14.64414516, 13.57911642, 13.84537361, 14.37788798, 14.11163079, 13.49036403, 13.49036403, 13.40161163, 13.40161163, 13.75662121, 14.11163079, 13.22410684, 13.66786882, 14.55539277, 14.99915474, 15.08790714, 15.35416432, 15.53166911, 14.46664037, 14.55539277, 11.71531613, 11.53781134, 11.44905894, 12.07032571, 11.89282092, 11.98157331, 13.04660205, 13.40161163, 12.42533529, 12.60284008, 13.13535445, 13.75662121, 14.46664037, 14.91040235, 13.934126, 13.04660205, 13.31285924, 13.75662121, 13.13535445, 14.46664037, 14.99915474, 15.08790714, 14.82164995, 15.35416432, 16.06418348, 16.50794545, 15.97543108, 16.68545024, 16.95170743, 16.15293587, 16.86295503, 16.77420264, 16.24168827, 16.24168827, 16.68545024, 17.4842218, 17.75047898, 19.88053646, 20.23554604, 19.88053646, 19.88053646, 19.52552688, 20.41305083, 22.18809873, 23.07562268, 26.0932041, 26.62571847, 29.28829032, 29.6432999, 30.88583343, 28.75577595, 29.28829032, 31.24084301, 33.37090049, 36.21097713, 37.80852023, 38.69604418, 36.5659867, 35.85596755, 38.51853939, 39.76107292, 39.93857771, 36.74349149, 34.25842444, 32.83838612, 33.01589091, 32.83838612, 29.6432999, 31.77335738, 28.93328074, 31.24084301, 33.37090049, 32.66088133, 31.59585259, 30.53082385, 33.1933957, 34.08091965, 33.01589091, 32.12836696, 30.53082385, 29.82080469, 29.6432999, 28.75577595, 27.33573763, 29.55454751, 28.57827116, 30.61957625, 31.4183478, 31.3295954, 30.70832864, 29.90955709, 31.86210977, 32.21711935, 31.50710019, 30.17581427, 30.17581427, 29.11078553, 29.55454751, 29.11078553, 28.75577595, 29.02203314, 28.57827116, 28.31201398, 24.76191818, 23.34187986, 21.47807957, 22.98687028, 23.34187986, 22.80936549, 19.79178407, 18.99301251, 18.90426012, 18.72675533, 20.41305083, 22.7206131, 22.45435591, 23.07562268, 22.27685112, 21.56683196, 22.6318607, 22.45435591, 23.25312747, 22.54310831, 22.89811789, 22.7206131, 21.47807957, 21.47807957, 24.76191818, 23.96314663, 24.14065142, 24.31815621, 23.34187986, 23.07562268, 22.09934633, 22.80936549, 23.34187986, 23.69688944, 22.98687028, 22.80936549, 22.36560352, 21.38932717, 20.76806041, 20.67930802, 20.9455652, 20.9455652, 22.18809873, 20.41305083, 20.50180323, 21.47807957, 22.01059394, 22.6318607, 22.36560352, 22.98687028, 21.83308915, 21.03431759, 22.01059394, 23.16437507, 22.98687028, 23.07562268, 23.16437507, 23.51938465, 26.0932041, 27.06948045, 28.48951877, 29.11078553, 29.7320523, 29.11078553, 29.99830948, 29.55454751, 27.69074721, 29.19953793, 26.71447087, 27.51324242, 27.51324242, 27.60199482, 26.89197566, 27.42449003, 27.33573763, 26.71447087, 24.93942297, 25.64944213, 27.06948045, 27.77949961, 27.15823284, 26.35946129, 26.0932041, 27.15823284, 26.89197566, 26.44821368, 26.00445171, 25.64944213, 26.98072805, 26.00445171, 26.27070889, 27.77949961, 27.69074721, 27.24698524, 28.40076637, 29.82080469, 29.99830948, 29.90955709, 30.44207146, 31.86210977, 30.44207146, 31.50710019, 32.74963372, 35.23470078, 37.09850107, 40.29358729, 38.16352981, 37.27600586, 38.16352981, 39.31731095, 38.25228221, 37.54226305, 41.00360645, 38.96230137, 38.25228221, 38.3410346, 35.32345318, 37.89727263, 36.92099628, 35.23470078, 31.59585259, 30.17581427, 30.08706188, 28.93328074, 29.82080469, 29.02203314, 28.57827116, 28.66702356, 31.68460498, 32.57212893, 34.79093881, 33.63715767, 33.28214809, 31.4183478, 31.24084301, 33.81466246, 32.03961456, 31.77335738, 31.4183478, 31.24084301, 33.90341486, 34.25842444, 31.86210977, 31.24084301, 28.40076637, 29.28829032, 28.75577595, 28.31201398]];
		var assetsReturns = new Array(assetsPrices.length);
		for (var k = 0; k < assetsPrices.length; ++k) {
			assetsReturns[k] = new Array(assetsPrices[k].length - 1);
			for (var i = 1; i < assetsPrices[k].length; ++i) {
				assetsReturns[k][i-1] = (assetsPrices[k][i] - assetsPrices[k][i-1])/assetsPrices[k][i-1];
			}
		}
		
		// First compute the minimum tracking weights with no cardinality constraints
		// Expected weights have been computed with the method below, and value of the objective function verified to be of ~6.8e-4, which is good !
		var expectedWeights = [0.011715036820412077, 0.010982741983461726, 0.014329219601669463, 0.08138764092018093, 0.005174521042167091, 0.03731032609992767, 0.026151093715953894, 0.004251777653722618, 0.006812177865239405, 0.003484984901039915, 0.09715701892213753, 0.02866896503974561, 0.036382118690755604, 0.025350781767040753, 0.17571563315373878, 0.0017205161735978484, 0.0008476736457532089, 0.025294354986590422, 0, 0.022574142290264214, 0.04012875317955057, 0.029974000042811807, 0.012634234356196011, 0.031698301448724105, 0.01762436687114238, 0.0672892678807561, 0.08493973626173573, 0.03733945213481528, 0, 0.034618365316269016, 0.028442797234600012];
		
		// Compute the weights of the minimum tracking error portfolio
		var weights = PortfolioAllocation.minimumTrackingErrorWeights(assetsReturns, indexReturns);

		//
		assert.equal(areWeightsIdentical(weights, expectedWeights, 1e-4), true, 'Minimum tracking error portfolio, indtrack1 OR data set (Hang Seng Index)');
		
		
		// Now compute the weights of the minimum tracking error portfolio using the Threshold Accepting algorithm,
		// with no tests on the numerical value of the weights, as this is a heuristic method.
			
		// Construct minimum weights constraints
		var lowerBounds = new Array(assetsPrices.length);
		for (var k = 0; k < assetsPrices.length; ++k) {
			lowerBounds[k] = 0.01;
		}
		
		// Test an exact number of assets
		var weights = PortfolioAllocation.minimumTrackingErrorWeights(assetsReturns, indexReturns, {constraints: {minWeights: lowerBounds, minNbAssets: 10, maxNbAssets: 10 }});
		
		// Check that the number of non-zero weights is equal to 10, and that the minimum weights matches with the lower bounds
		var nbNonZeroAssets = 0;
		var lowerBoundsOk = true;
		for (var k = 0; k < weights.length; ++k) {
			if (weights[k] != 0) {
				nbNonZeroAssets = nbNonZeroAssets + 1;
				
				if (weights[k] < lowerBounds[k]) {
					lowerBoundsOk = false;
				}
			}
		}
		assert.equal(nbNonZeroAssets == 10, true, "Minimum tracking error portfolio, indtrack1 OR data set ((Hang Seng Index)), exact number of assets");
		assert.equal(lowerBoundsOk, true, "Minimum tracking error portfolio, indtrack1 OR data set ((Hang Seng Index)), minimum bounds");


		// Test an exact full number of assets
		var weights = PortfolioAllocation.minimumTrackingErrorWeights(assetsReturns, indexReturns, {constraints: {minWeights: lowerBounds, minNbAssets: 31, maxNbAssets: 31 }});
		
		// Check that the number of non-zero weights is equal to 31
		var nbNonZeroAssets = 0;
		var lowerBoundsOk = true;
		for (var k = 0; k < weights.length; ++k) {
			if (weights[k] != 0) {
				nbNonZeroAssets = nbNonZeroAssets + 1;
				
				if (weights[k] < lowerBounds[k]) {
					lowerBoundsOk = false;
				}
			}
		}
		assert.equal(nbNonZeroAssets == 31, true, "Minimum tracking error portfolio, indtrack1 OR data set ((Hang Seng Index)), full exact number of assets");
		assert.equal(lowerBoundsOk, true, "Minimum tracking error portfolio, indtrack1 OR data set ((Hang Seng Index)), full minimum bounds");
		
		// Test a non-exact number of assets
		var weights = PortfolioAllocation.minimumTrackingErrorWeights(assetsReturns, indexReturns, {constraints: { minNbAssets: 10, maxNbAssets: 20 }});

		// Check that the number of non-zero weights is between 10 and 20
		var nbNonZeroAssets = 0;
		var lowerBoundsOk = true;
		for (var k = 0; k < weights.length; ++k) {
			if (weights[k] != 0) {
				nbNonZeroAssets = nbNonZeroAssets + 1;
				
				if (weights[k] < lowerBounds[k]) {
					lowerBoundsOk = false;
				}
			}
		}
		assert.equal(nbNonZeroAssets >= 10 && nbNonZeroAssets <= 20, true, "Minimum tracking error portfolio, indtrack1 OR data set ((Hang Seng Index)), min/max number of assets");		
		assert.equal(lowerBoundsOk, true, "Minimum tracking error portfolio, indtrack1 OR data set ((Hang Seng Index)), min/max number of assets minimum bounds");
	}

	
	
	// INDTRACK2 (Nikkei Index and Constituents)
	// Reference: OR-library http://people.brunel.ac.uk/~mastjjb/jeb/orlib/indtrackinfo.html
	// Test to test the performances of the algorithm: ~500 ms seconds for the problem below on a 2018 machine
	{
		// Define the index prices
		var indexPrices = [25677.77068055, 25261.63282391, 24615.52531315, 24289.45719964, 23629.90841927, 22333.57798666, 21480.35218349, 21155.48740655, 21109.47181567, 20821.80217248, 22112.45285643, 21749.43027649, 21925.99585311, 22078.51876483, 21407.40591991, 20918.41806662, 19878.96389407, 19028.03646377, 20117.11623694, 20196.46425113, 19912.89798568, 18649.05764077, 19145.42194730, 18673.70197396, 17833.74897328, 19514.36494322, 21624.90900703, 22328.27127234, 21789.64578505, 21860.77501039, 22135.08761753, 20846.68717299, 20528.65734784, 20901.72778823, 20598.29443582, 20176.82579812, 20300.05949744, 19651.43701323, 20497.15399613, 21023.02411565, 20812.53648080, 20987.41738620, 21275.88123154, 21127.02849640, 20366.41147654, 20017.13100037, 19873.82564687, 19658.68109945, 20485.33723090, 20857.31263500, 20278.03843799, 20468.79135291, 20400.58623546, 20237.35362817, 21705.20765712, 22306.45478011, 22609.13003048, 23401.21430194, 23936.98788153, 24425.15746595, 23710.72450409, 25172.81455082, 25043.07080059, 24637.29367192, 24737.55567589, 25081.97467228, 25128.36329750, 24669.54309230, 23831.52746352, 23657.07975929, 23611.22060216, 23919.19053349, 24465.67380865, 23747.33000290, 24524.16799988, 24497.45392777, 24963.42195226, 24797.46980429, 25019.38913667, 25409.90795749, 25051.03688876, 24537.28436862, 24436.79373069, 24407.43231810, 24522.36299501, 24276.61759832, 24385.41125866, 23709.30456693, 22370.58058654, 22253.96523842, 21589.29024422, 20127.45289817, 21009.47454574, 20766.49682319, 21722.52367053, 20626.52471202, 20958.80204229, 21809.28423806, 22831.74729807, 23233.33681545, 22572.04319703, 24429.45337754, 24055.54060155, 22814.78025227, 23830.13159308, 24025.81818798, 24205.48837298, 24631.63798998, 23869.96203393, 23196.91181712, 23988.50272059, 24264.83693318, 24023.88081609, 23736.11490596, 23901.23675169, 24392.53501122, 24478.47730988, 25001.91668950, 25214.94336454, 25745.60276366, 25875.70751486, 24989.38995569, 24720.63676356, 24700.30037533, 24993.48130007, 24623.74410201, 24607.49905816, 24694.51232637, 24865.54255471, 24683.68229714, 24634.09279661, 24853.50918890, 23943.84690004, 23821.56383662, 23866.58065814, 23541.84824823, 23759.57996931, 24029.77716534, 23945.29090394, 23832.27353220, 23839.97488632, 23205.57584051, 23227.47656629, 22462.59973494, 22861.34937798, 22837.28264635, 23060.05634770, 23625.74487470, 23733.47959885, 23488.48027084, 23261.90402589, 22671.12593116, 21785.62664087, 22308.62078596, 22010.65058163, 21684.73890187, 21025.82788989, 20504.39808235, 19684.63706951, 19555.69955480, 18952.21422576, 19421.79226000, 18915.84939427, 19311.01309430, 20418.50391716, 20224.17709260, 20563.40970832, 19759.70120547, 19422.87526292, 18885.46514558, 19071.83791534, 18103.21213399, 17692.86232630, 18369.14951852, 17469.31848954, 19509.79226421, 19876.35265369, 19962.25885225, 20034.89224832, 20145.29837968, 20206.83701246, 21699.15487412, 21384.10932369, 21805.33729408, 21996.45120996, 22572.84943254, 21315.81997269, 21555.44038618, 22269.28371311, 21516.65684815, 21849.37941296, 20862.47494894, 21694.71456213, 21471.80849376, 21841.95482625, 21919.05260104, 22662.55817470, 23208.71654899, 23280.41134252, 23759.18286823, 23908.07170347, 24871.79990494, 24412.59463204, 24506.86401984, 24865.55458808, 25154.58400161, 25191.63473496, 25032.73413936, 24428.18987413, 24269.65027951, 24254.29570473, 24296.38841835, 24910.17430852, 25759.64570156, 26107.39794028, 26064.83592538, 26333.62521761, 26756.22499170, 26067.13429826, 25775.61397800, 26373.04652403, 26230.84824018, 26420.68661929, 26174.74868875, 26821.63836828, 27111.95535195, 27112.07568561, 26753.08428322, 26059.99851233, 25843.20539379, 25420.36495238, 25198.36138645, 24729.83025504, 25070.31434080, 25545.39162321, 24267.56850722, 24250.27656055, 25080.71116886, 25405.13071126, 25928.31739020, 25448.19812752, 25231.75397659, 26006.87120225, 24957.16460203, 24828.51588810, 25511.98699971, 25185.50975176, 25530.12128199, 25294.56814614, 24399.69486388, 24477.53870734, 23694.25082629, 23307.47438224, 23298.22072392, 20822.11503999, 21768.40689439, 21286.25399288, 22057.17157387, 21500.06283669, 22528.86748049, 22904.95829570, 22330.31694452, 21899.20958081, 21568.17168721, 22421.96305858, 21888.35548484, 21492.30131574, 21475.92390487, 22083.80141242, 22397.53532597, 23482.81255554, 23829.40959114, 24457.49111989, 24077.56166099, 24149.53322193, 24651.25237626, 24702.51451464, 24530.66601742, 24696.97916636, 24028.22486115, 23916.90419398, 24366.74750828, 24535.47936375, 23831.33492967, 23590.76388027, 23255.71887586, 22442.43181383, 21936.12794713, 22442.43181383, 21618.90435749, 21730.10469099];
		var indexReturns = new Array(indexPrices.length - 1);
		for (var i = 1; i < indexPrices.length; ++i) {
			indexReturns[i-1] = (indexPrices[i] - indexPrices[i-1])/indexPrices[i-1];
		}
		
		// Define the 225 assets prices
		var assetsPrices = [[938.60253364, 915.73913859, 890.46907037, 884.45238747, 859.18231925, 841.13227053, 794.20214385, 806.23550966, 826.69223155, 830.30224130, 844.74228027, 845.94561686, 872.41902165, 872.41902165, 842.33560711, 812.25219257, 794.20214385, 753.28870007, 806.23550966, 784.57545120, 764.11872931, 722.00194895, 758.10204640, 730.42530502, 675.07182227, 764.11872931, 839.92893395, 826.69223155, 812.25219257, 784.57545120, 831.50557788, 807.43884625, 788.18546094, 790.59213410, 759.30538298, 755.69537324, 734.03531477, 715.98526604, 726.81529528, 749.67869033, 715.98526604, 740.05199768, 735.23865135, 734.03531477, 711.17191972, 705.15523681, 697.93521732, 705.15523681, 742.45867084, 737.64532451, 731.62864161, 719.59527579, 728.01863186, 719.59527579, 743.66200742, 795.40548043, 854.36897293, 944.61921655, 936.19586048, 925.36583124, 934.99252389, 1027.64944068, 1058.93619180, 1024.03943093, 1021.63275777, 1012.00606512, 999.97269930, 1010.80272853, 943.41587996, 930.17917757, 944.61921655, 962.66926527, 974.70263109, 934.99252389, 972.29595792, 960.26259211, 1009.59939195, 966.27927501, 967.48261160, 972.29595792, 953.04257262, 961.46592869, 954.24590920, 939.80587022, 943.41587996, 945.82255313, 926.56916782, 913.33246543, 855.57230951, 861.58899242, 844.74228027, 819.47221206, 811.04885599, 805.03217308, 827.89556813, 795.40548043, 786.98212436, 796.60881701, 842.33560711, 860.38565584, 830.30224130, 890.46907037, 866.40233874, 857.97898267, 878.43570456, 891.67240696, 896.48575328, 872.41902165, 854.36897293, 844.74228027, 860.38565584, 886.85906063, 874.82569481, 866.40233874, 870.01234849, 903.70577277, 900.09576303, 913.33246543, 925.36583124, 914.53580201, 922.95915808, 902.50243619, 924.16249466, 890.46907037, 902.50243619, 906.11244594, 901.29909961, 902.50243619, 914.53580201, 904.90910935, 906.11244594, 894.07908012, 873.62235823, 872.41902165, 854.36897293, 827.89556813, 830.30224130, 824.28555839, 829.09890471, 835.11558762, 830.30224130, 818.26887548, 824.28555839, 803.82883650, 809.84551941, 802.62549992, 801.42216334, 788.18546094, 797.81215359, 794.20214385, 792.99880727, 782.16877803, 755.69537324, 780.96544145, 758.10204640, 740.05199768, 712.37525630, 709.96858314, 695.52854416, 682.29184176, 661.83511987, 660.63178329, 643.78507115, 677.47849543, 688.30852467, 685.90185151, 712.37525630, 663.03845645, 657.02177355, 672.66514911, 659.42844671, 643.78507115, 658.22511013, 667.85180278, 641.37839799, 667.85180278, 640.17506140, 644.98840773, 655.81843697, 661.83511987, 655.81843697, 655.81843697, 664.24179304, 672.66514911, 659.42844671, 661.83511987, 632.95504192, 642.58173457, 675.07182227, 655.81843697, 664.24179304, 647.39508089, 647.39508089, 661.83511987, 657.02177355, 666.64846620, 693.12187099, 688.30852467, 695.52854416, 722.00194895, 779.76210487, 766.52540247, 758.10204640, 717.18860263, 722.00194895, 782.16877803, 794.20214385, 778.55876829, 765.32206589, 766.52540247, 760.50871956, 771.33874880, 780.96544145, 821.87888522, 830.30224130, 855.57230951, 873.62235823, 877.23236798, 836.31892420, 824.28555839, 835.11558762, 855.57230951, 860.38565584, 848.35229002, 912.12912884, 900.09576303, 894.07908012, 862.79232900, 853.16563635, 836.31892420, 856.77564609, 833.91225104, 815.86220232, 818.26887548, 841.13227053, 800.21882676, 820.67554864, 815.86220232, 800.21882676, 818.26887548, 797.81215359, 779.76210487, 807.43884625, 784.57545120, 782.16877803, 807.43884625, 811.04885599, 811.04885599, 773.74542196, 713.57859288, 694.32520758, 703.95190023, 709.96858314, 715.98526604, 664.24179304, 703.95190023, 648.59841748, 649.80175406, 624.53168584, 602.87162738, 612.49832003, 612.49832003, 594.44827130, 557.14483728, 601.66829079, 623.32834926, 593.24493472, 619.71833952, 651.00509064, 659.42844671, 722.00194895, 758.10204640, 737.64532451, 740.05199768, 709.96858314, 737.64532451, 765.32206589, 760.50871956, 752.08536349, 738.84866109, 712.37525630, 682.29184176, 676.27515885, 625.73502243, 625.73502243, 605.27830054, 638.97172482, 648.59841748, 619.71833952, 601.66829079, 584.82157865], 
							[578.80489574, 561.95818360, 577.60155916, 541.50146171, 548.72148120, 504.19802769, 505.40136427, 496.97800820, 509.01137401, 512.62138376, 535.48477881, 537.89145197, 540.29812513, 564.36485676, 567.97486651, 567.97486651, 542.70479830, 518.63806666, 566.77152993, 559.55151044, 541.50146171, 517.43473008, 502.99469110, 475.31794973, 445.23453519, 492.16466187, 565.56819335, 601.66829079, 580.00823233, 599.26161763, 617.31166635, 611.29498345, 588.43158840, 571.58487625, 571.58487625, 595.65160789, 572.78821284, 529.46809590, 548.72148120, 548.72148120, 555.94150069, 554.73816411, 554.73816411, 541.50146171, 534.28144222, 531.87476906, 518.63806666, 487.35131554, 517.43473008, 506.60470085, 498.18134478, 496.97800820, 468.09793024, 436.81117912, 546.31480804, 541.50146171, 563.16152018, 655.81843697, 634.15837850, 641.37839799, 618.51500294, 673.86848569, 679.88516860, 659.42844671, 658.22511013, 661.83511987, 693.12187099, 679.88516860, 634.15837850, 642.58173457, 643.78507115, 644.98840773, 651.00509064, 643.78507115, 671.46181253, 657.02177355, 695.52854416, 669.05513936, 664.24179304, 664.24179304, 594.44827130, 601.66829079, 610.09164687, 589.63492498, 611.29498345, 610.09164687, 616.10832977, 600.46495421, 545.11147146, 524.65474957, 517.43473008, 481.33463264, 495.77467161, 498.18134478, 505.40136427, 477.72462289, 487.35131554, 487.35131554, 518.63806666, 517.43473008, 505.40136427, 531.87476906, 512.62138376, 500.58801794, 516.23139350, 515.02805692, 527.06142274, 548.72148120, 525.85808615, 519.84140325, 511.41804717, 500.58801794, 504.19802769, 494.57133503, 502.99469110, 504.19802769, 506.60470085, 528.26475932, 541.50146171, 537.89145197, 553.53482753, 542.70479830, 541.50146171, 575.19488600, 589.63492498, 566.77152993, 563.16152018, 563.16152018, 561.95818360, 553.53482753, 536.68811539, 530.67143248, 515.02805692, 500.58801794, 513.82472034, 510.21471059, 507.80803743, 507.80803743, 510.21471059, 506.60470085, 502.99469110, 484.94464238, 482.53796922, 456.06456442, 466.89459366, 458.47123758, 451.25121810, 463.28458391, 474.11461315, 460.87791075, 457.26790100, 441.62452544, 427.18448646, 440.42118886, 433.20116937, 438.01451570, 417.55779381, 413.94778407, 403.11775483, 397.10107192, 371.83100371, 392.28772560, 386.27104269, 400.71108167, 413.94778407, 401.91441825, 428.38782305, 405.52442800, 377.84768662, 375.44101346, 357.39096473, 318.88419412, 338.13757943, 339.34091601, 330.91755994, 379.05102320, 365.81432080, 373.03434029, 371.83100371, 354.98429157, 342.95092575, 339.34091601, 368.22099397, 367.01765738, 353.78095499, 354.98429157, 336.93424284, 348.96760866, 357.39096473, 348.96760866, 350.17094524, 346.56093550, 352.57761841, 341.74758917, 371.83100371, 377.84768662, 385.06770611, 403.11775483, 391.08438902, 401.91441825, 416.35445723, 425.98114988, 409.13443774, 388.67771585, 391.08438902, 445.23453519, 460.87791075, 438.01451570, 442.82786202, 438.01451570, 433.20116937, 433.20116937, 472.91127656, 471.70793998, 480.13129605, 510.21471059, 498.18134478, 517.43473008, 501.79135452, 502.99469110, 494.57133503, 486.14797896, 501.79135452, 487.35131554, 496.97800820, 493.36799845, 489.75798871, 481.33463264, 480.13129605, 472.91127656, 468.09793024, 459.67457417, 458.47123758, 463.28458391, 471.70793998, 450.04788151, 468.09793024, 472.91127656, 469.30126682, 469.30126682, 468.09793024, 450.04788151, 464.48792049, 456.06456442, 463.28458391, 474.11461315, 471.70793998, 472.91127656, 445.23453519, 407.93110116, 403.11775483, 376.64435004, 371.83100371, 379.05102320, 344.15426233, 359.79763789, 344.15426233, 354.98429157, 361.00097448, 347.76427208, 352.57761841, 341.74758917, 341.74758917, 340.54425259, 368.22099397, 357.39096473, 334.52756968, 329.71422336, 330.91755994, 350.17094524, 352.57761841, 359.79763789, 338.13757943, 341.74758917, 351.37428182, 358.59430131, 353.78095499, 357.39096473, 362.20431106, 338.13757943, 334.52756968, 314.07084779, 340.54425259, 299.63080882, 296.02079907, 270.75073086, 264.73404795, 256.31069188, 239.46397974, 231.04062366, 228.63395050], 
							[1516.20409280, 1419.93716627, 1468.07062954, 1407.90380046, 938.60253364, 830.30224130, 762.91539273, 746.06868058, 709.96858314, 643.78507115, 784.57545120, 794.20214385, 783.37211461, 785.77878778, 728.01863186, 730.42530502, 673.86848569, 613.70165661, 595.65160789, 619.71833952, 628.14169559, 565.56819335, 607.68497370, 529.46809590, 439.21785228, 466.89459366, 696.73188074, 671.46181253, 659.42844671, 648.59841748, 681.08850518, 715.98526604, 670.25847594, 722.00194895, 696.73188074, 651.00509064, 675.07182227, 642.58173457, 693.12187099, 709.96858314, 694.32520758, 673.86848569, 643.78507115, 624.53168584, 601.66829079, 583.61824207, 587.22825181, 565.56819335, 575.19488600, 652.20842722, 642.58173457, 655.81843697, 683.49517834, 703.95190023, 841.13227053, 983.12598716, 980.71931399, 854.36897293, 817.06553890, 886.85906063, 860.38565584, 914.53580201, 890.46907037, 865.19900216, 866.40233874, 862.79232900, 856.77564609, 820.67554864, 809.84551941, 776.15209512, 776.15209512, 782.16877803, 829.09890471, 774.94875854, 794.20214385, 812.25219257, 813.45552915, 824.28555839, 842.33560711, 851.96229976, 806.23550966, 795.40548043, 782.16877803, 803.82883650, 746.06868058, 722.00194895, 685.90185151, 613.70165661, 577.60155916, 545.11147146, 505.40136427, 422.37114014, 500.58801794, 481.33463264, 501.79135452, 457.26790100, 499.38468136, 511.41804717, 576.39822258, 567.97486651, 540.29812513, 613.70165661, 601.66829079, 533.07810564, 587.22825181, 601.66829079, 605.27830054, 693.12187099, 679.88516860, 637.76838824, 660.63178329, 661.83511987, 671.46181253, 637.76838824, 649.80175406, 661.83511987, 657.02177355, 746.06868058, 723.20528553, 762.91539273, 780.96544145, 709.96858314, 711.17191972, 684.69851492, 707.56190997, 697.93521732, 703.95190023, 732.83197819, 709.96858314, 685.90185151, 693.12187099, 683.49517834, 700.34189048, 663.03845645, 628.14169559, 601.66829079, 625.73502243, 664.24179304, 695.52854416, 708.76524656, 717.18860263, 717.18860263, 685.90185151, 643.78507115, 657.02177355, 694.32520758, 701.54522707, 730.42530502, 687.10518809, 719.59527579, 746.06868058, 740.05199768, 673.86848569, 706.35857339, 691.91853441, 661.83511987, 626.93835901, 584.82157865, 551.12815437, 555.94150069, 554.73816411, 529.46809590, 521.04473983, 534.28144222, 586.02491523, 586.02491523, 606.48163712, 584.82157865, 570.38153967, 533.07810564, 565.56819335, 518.63806666, 488.55465212, 533.07810564, 487.35131554, 558.34817386, 576.39822258, 570.38153967, 588.43158840, 569.17820309, 566.77152993, 643.78507115, 625.73502243, 649.80175406, 663.03845645, 673.86848569, 611.29498345, 613.70165661, 649.80175406, 614.90499319, 646.19174431, 607.68497370, 631.75170533, 613.70165661, 634.15837850, 634.15837850, 663.03845645, 685.90185151, 693.12187099, 703.95190023, 709.96858314, 709.96858314, 712.37525630, 715.98526604, 752.08536349, 777.35543171, 744.86534400, 722.00194895, 709.96858314, 709.96858314, 735.23865135, 748.47535375, 758.10204640, 794.20214385, 806.23550966, 772.54208538, 773.74542196, 779.76210487, 756.89870982, 764.11872931, 806.23550966, 824.28555839, 830.30224130, 813.45552915, 830.30224130, 865.19900216, 824.28555839, 842.33560711, 788.18546094, 844.74228027, 803.82883650, 791.79547068, 766.52540247, 767.72873905, 814.65886573, 746.06868058, 743.66200742, 800.21882676, 832.70891446, 854.36897293, 820.67554864, 825.48889497, 880.84237772, 830.30224130, 830.30224130, 878.43570456, 854.36897293, 901.29909961, 902.50243619, 806.23550966, 878.43570456, 812.25219257, 758.10204640, 741.25533426, 595.65160789, 631.75170533, 625.73502243, 652.20842722, 616.10832977, 634.15837850, 636.56505166, 620.92167610, 610.09164687, 589.63492498, 630.54836875, 571.58487625, 493.36799845, 509.01137401, 529.46809590, 487.35131554, 547.51814462, 583.61824207, 634.15837850, 624.53168584, 649.80175406, 665.44512962, 652.20842722, 651.00509064, 670.25847594, 604.07496396, 583.61824207, 563.16152018, 527.06142274, 471.70793998, 386.27104269, 380.25435978, 374.23767687, 363.40764764, 368.22099397, 357.39096473, 318.88419412], 
							[1383.83706883, 1371.80370301, 1347.73697138, 1287.57014230, 1251.47004485, 1239.43667904, 1203.33658159, 1239.43667904, 1275.53677648, 1251.47004485, 1299.60350811, 1239.43667904, 1239.43667904, 1227.40331322, 1196.11656210, 1180.47318654, 1073.37623078, 1058.93619180, 1060.13952838, 1083.00292343, 1107.06965506, 1054.12284547, 1074.57956736, 1048.10616256, 1036.07279675, 1072.17289419, 1141.96641593, 1180.47318654, 1155.20311832, 1114.28967455, 1090.22294292, 1030.05611384, 1034.86946017, 1033.66612358, 1031.25945042, 1031.25945042, 1032.46278700, 1016.81941144, 1008.39605537, 1025.24276751, 1027.64944068, 1021.63275777, 1032.46278700, 1057.73285522, 1027.64944068, 1019.22608460, 1008.39605537, 1007.19271879, 1025.24276751, 1032.46278700, 1010.80272853, 1015.61607486, 999.97269930, 991.54934323, 1072.17289419, 1152.79644516, 1161.21980123, 1239.43667904, 1227.40331322, 1275.53677648, 1239.43667904, 1335.70360556, 1323.67023975, 1299.60350811, 1275.53677648, 1299.60350811, 1299.60350811, 1311.63687393, 1275.53677648, 1275.53677648, 1263.50341067, 1263.50341067, 1287.57014230, 1275.53677648, 1323.67023975, 1323.67023975, 1395.87043464, 1371.80370301, 1347.73697138, 1359.77033719, 1359.77033719, 1359.77033719, 1383.83706883, 1407.90380046, 1431.97053209, 1395.87043464, 1419.93716627, 1431.97053209, 1395.87043464, 1335.70360556, 1311.63687393, 1287.57014230, 1251.47004485, 1251.47004485, 1287.57014230, 1251.47004485, 1287.57014230, 1263.50341067, 1287.57014230, 1299.60350811, 1275.53677648, 1359.77033719, 1311.63687393, 1287.57014230, 1311.63687393, 1299.60350811, 1287.57014230, 1347.73697138, 1383.83706883, 1323.67023975, 1407.90380046, 1419.93716627, 1359.77033719, 1431.97053209, 1407.90380046, 1444.00389791, 1395.87043464, 1431.97053209, 1504.17072698, 1468.07062954, 1468.07062954, 1468.07062954, 1468.07062954, 1395.87043464, 1444.00389791, 1456.03726372, 1444.00389791, 1456.03726372, 1480.10399535, 1456.03726372, 1456.03726372, 1528.23745862, 1456.03726372, 1480.10399535, 1492.13736117, 1480.10399535, 1492.13736117, 1516.20409280, 1540.27082443, 1552.30419025, 1588.40428770, 1552.30419025, 1528.23745862, 1528.23745862, 1552.30419025, 1540.27082443, 1540.27082443, 1552.30419025, 1576.37092188, 1552.30419025, 1492.13736117, 1419.93716627, 1383.83706883, 1371.80370301, 1311.63687393, 1263.50341067, 1215.36994740, 1227.40331322, 1227.40331322, 1239.43667904, 1176.86317679, 1263.50341067, 1145.57642567, 1203.33658159, 1287.57014230, 1275.53677648, 1287.57014230, 1239.43667904, 1182.87985970, 1190.09987919, 1179.26984996, 1107.06965506, 1098.64629899, 1176.86317679, 1161.21980123, 1203.33658159, 1239.43667904, 1251.47004485, 1239.43667904, 1251.47004485, 1215.36994740, 1287.57014230, 1287.57014230, 1299.60350811, 1299.60350811, 1383.83706883, 1311.63687393, 1335.70360556, 1383.83706883, 1335.70360556, 1395.87043464, 1311.63687393, 1371.80370301, 1347.73697138, 1383.83706883, 1419.93716627, 1468.07062954, 1504.17072698, 1492.13736117, 1492.13736117, 1528.23745862, 1600.43765351, 1528.23745862, 1540.27082443, 1552.30419025, 1624.50438514, 1588.40428770, 1588.40428770, 1576.37092188, 1540.27082443, 1504.17072698, 1504.17072698, 1624.50438514, 1684.67121422, 1696.70458004, 1732.80467749, 1756.87140912, 1768.90477493, 1768.90477493, 1744.83804330, 1720.77131167, 1744.83804330, 1780.93814075, 1744.83804330, 1756.87140912, 1720.77131167, 1732.80467749, 1768.90477493, 1720.77131167, 1720.77131167, 1636.53775096, 1624.50438514, 1600.43765351, 1600.43765351, 1576.37092188, 1456.03726372, 1492.13736117, 1588.40428770, 1636.53775096, 1684.67121422, 1600.43765351, 1612.47101933, 1636.53775096, 1564.33755606, 1540.27082443, 1612.47101933, 1588.40428770, 1648.57111678, 1588.40428770, 1528.23745862, 1492.13736117, 1431.97053209, 1419.93716627, 1444.00389791, 1215.36994740, 1311.63687393, 1227.40331322, 1347.73697138, 1323.67023975, 1347.73697138, 1395.87043464, 1347.73697138, 1287.57014230, 1275.53677648, 1359.77033719, 1347.73697138, 1311.63687393, 1287.57014230, 1371.80370301, 1383.83706883, 1492.13736117, 1648.57111678, 1672.63784841, 1636.53775096, 1648.57111678, 1660.60448259, 1672.63784841, 1684.67121422, 1696.70458004, 1588.40428770, 1552.30419025, 1540.27082443, 1552.30419025, 1468.07062954, 1516.20409280, 1564.33755606, 1492.13736117, 1431.97053209, 1347.73697138, 1287.57014230, 1311.63687393], 
							[1540.27082443, 1468.07062954, 1419.93716627, 1456.03726372, 1299.60350811, 1251.47004485, 1125.11970378, 1169.64315730, 1176.86317679, 1163.62647440, 1263.50341067, 1275.53677648, 1188.89654261, 1251.47004485, 1156.40645491, 1095.03628924, 1007.19271879, 934.99252389, 1020.42942119, 978.31264083, 950.63589945, 866.40233874, 1009.59939195, 980.71931399, 985.53266032, 1173.25316705, 1335.70360556, 1299.60350811, 1215.36994740, 1203.33658159, 1227.40331322, 1167.23648414, 1167.23648414, 1156.40645491, 1132.33972327, 1113.08633797, 1073.37623078, 1050.51283573, 1078.18957710, 1153.99978174, 1143.16975251, 1227.40331322, 1239.43667904, 1227.40331322, 1179.26984996, 1156.40645491, 1158.81312807, 1172.04983047, 1202.13324501, 1215.36994740, 1167.23648414, 1185.28653286, 1179.26984996, 1185.28653286, 1335.70360556, 1480.10399535, 1540.27082443, 1756.87140912, 1660.60448259, 1612.47101933, 1600.43765351, 1660.60448259, 1720.77131167, 1660.60448259, 1612.47101933, 1660.60448259, 1612.47101933, 1612.47101933, 1504.17072698, 1504.17072698, 1492.13736117, 1540.27082443, 1576.37092188, 1516.20409280, 1516.20409280, 1552.30419025, 1612.47101933, 1600.43765351, 1576.37092188, 1612.47101933, 1588.40428770, 1576.37092188, 1576.37092188, 1564.33755606, 1600.43765351, 1552.30419025, 1576.37092188, 1576.37092188, 1492.13736117, 1528.23745862, 1492.13736117, 1251.47004485, 1359.77033719, 1407.90380046, 1444.00389791, 1407.90380046, 1359.77033719, 1383.83706883, 1468.07062954, 1564.33755606, 1528.23745862, 1684.67121422, 1600.43765351, 1516.20409280, 1564.33755606, 1504.17072698, 1468.07062954, 1456.03726372, 1419.93716627, 1395.87043464, 1444.00389791, 1444.00389791, 1504.17072698, 1480.10399535, 1456.03726372, 1564.33755606, 1516.20409280, 1516.20409280, 1564.33755606, 1528.23745862, 1504.17072698, 1431.97053209, 1431.97053209, 1395.87043464, 1407.90380046, 1407.90380046, 1444.00389791, 1431.97053209, 1431.97053209, 1480.10399535, 1431.97053209, 1456.03726372, 1419.93716627, 1347.73697138, 1359.77033719, 1359.77033719, 1347.73697138, 1347.73697138, 1299.60350811, 1287.57014230, 1275.53677648, 1227.40331322, 1197.31989868, 1196.11656210, 1215.36994740, 1239.43667904, 1227.40331322, 1263.50341067, 1275.53677648, 1227.40331322, 1239.43667904, 1239.43667904, 1178.06651337, 1323.67023975, 1275.53677648, 1239.43667904, 1191.30321577, 1215.36994740, 1167.23648414, 1169.64315730, 1139.55974276, 1152.79644516, 1095.03628924, 1119.10302088, 1193.70988894, 1176.86317679, 1198.52323526, 1197.31989868, 1179.26984996, 1164.82981098, 1161.21980123, 1132.33972327, 1133.54305986, 1193.70988894, 1168.43982072, 1251.47004485, 1287.57014230, 1311.63687393, 1347.73697138, 1371.80370301, 1323.67023975, 1492.13736117, 1480.10399535, 1528.23745862, 1516.20409280, 1480.10399535, 1395.87043464, 1431.97053209, 1444.00389791, 1347.73697138, 1395.87043464, 1335.70360556, 1383.83706883, 1371.80370301, 1407.90380046, 1444.00389791, 1540.27082443, 1528.23745862, 1528.23745862, 1564.33755606, 1528.23745862, 1636.53775096, 1588.40428770, 1600.43765351, 1504.17072698, 1540.27082443, 1528.23745862, 1552.30419025, 1504.17072698, 1468.07062954, 1444.00389791, 1516.20409280, 1588.40428770, 1672.63784841, 1684.67121422, 1648.57111678, 1600.43765351, 1660.60448259, 1648.57111678, 1600.43765351, 1648.57111678, 1648.57111678, 1648.57111678, 1660.60448259, 1672.63784841, 1672.63784841, 1780.93814075, 1768.90477493, 1708.73794585, 1696.70458004, 1660.60448259, 1696.70458004, 1672.63784841, 1684.67121422, 1684.67121422, 1612.47101933, 1600.43765351, 1684.67121422, 1684.67121422, 1756.87140912, 1720.77131167, 1696.70458004, 1805.00487238, 1684.67121422, 1732.80467749, 1696.70458004, 1696.70458004, 1720.77131167, 1624.50438514, 1540.27082443, 1492.13736117, 1419.93716627, 1371.80370301, 1395.87043464, 1239.43667904, 1323.67023975, 1359.77033719, 1504.17072698, 1347.73697138, 1540.27082443, 1636.53775096, 1576.37092188, 1576.37092188, 1552.30419025, 1648.57111678, 1588.40428770, 1588.40428770, 1612.47101933, 1672.63784841, 1768.90477493, 1732.80467749, 1684.67121422, 1768.90477493, 1792.97150657, 1768.90477493, 1889.23843309, 1889.23843309, 1841.10496983, 1853.13833565, 1841.10496983, 1841.10496983, 1805.00487238, 1901.27179891, 1756.87140912, 1732.80467749, 1780.93814075, 1708.73794585, 1684.67121422, 1672.63784841, 1672.63784841, 1829.07160401], 
							[652.20842722, 622.12501268, 606.48163712, 586.02491523, 576.39822258, 531.87476906, 523.45141299, 489.75798871, 499.38468136, 487.35131554, 527.06142274, 529.46809590, 531.87476906, 547.51814462, 511.41804717, 488.55465212, 489.75798871, 466.89459366, 500.58801794, 506.60470085, 475.31794973, 465.69125707, 469.30126682, 439.21785228, 410.33777432, 438.01451570, 519.84140325, 547.51814462, 499.38468136, 535.48477881, 515.02805692, 475.31794973, 468.09793024, 490.96132529, 483.74130580, 458.47123758, 483.74130580, 492.16466187, 494.57133503, 504.19802769, 495.77467161, 489.75798871, 486.14797896, 482.53796922, 475.31794973, 451.25121810, 448.84454493, 436.81117912, 435.60784253, 453.65789126, 429.59115963, 450.04788151, 438.01451570, 425.98114988, 459.67457417, 512.62138376, 572.78821284, 581.21156891, 552.33149095, 554.73816411, 546.31480804, 589.63492498, 580.00823233, 569.17820309, 563.16152018, 589.63492498, 589.63492498, 576.39822258, 553.53482753, 536.68811539, 548.72148120, 564.36485676, 583.61824207, 575.19488600, 586.02491523, 587.22825181, 613.70165661, 605.27830054, 605.27830054, 628.14169559, 624.53168584, 595.65160789, 604.07496396, 613.70165661, 613.70165661, 608.88831028, 614.90499319, 589.63492498, 567.97486651, 549.92481779, 527.06142274, 487.35131554, 531.87476906, 499.38468136, 525.85808615, 489.75798871, 505.40136427, 495.77467161, 523.45141299, 546.31480804, 548.72148120, 576.39822258, 567.97486651, 535.48477881, 581.21156891, 571.58487625, 566.77152993, 577.60155916, 546.31480804, 536.68811539, 560.75484702, 577.60155916, 567.97486651, 553.53482753, 555.94150069, 571.58487625, 571.58487625, 623.32834926, 635.36171508, 666.64846620, 648.59841748, 652.20842722, 637.76838824, 620.92167610, 649.80175406, 622.12501268, 634.15837850, 642.58173457, 661.83511987, 638.97172482, 644.98840773, 643.78507115, 634.15837850, 631.75170533, 635.36171508, 644.98840773, 643.78507115, 648.59841748, 642.58173457, 638.97172482, 640.17506140, 624.53168584, 616.10832977, 601.66829079, 604.07496396, 596.85494447, 604.07496396, 611.29498345, 619.71833952, 610.09164687, 600.46495421, 606.48163712, 555.94150069, 580.00823233, 564.36485676, 559.55151044, 539.09478855, 539.09478855, 530.67143248, 534.28144222, 506.60470085, 540.29812513, 493.36799845, 510.21471059, 539.09478855, 517.43473008, 510.21471059, 512.62138376, 495.77467161, 469.30126682, 462.08124733, 399.50774509, 407.93110116, 428.38782305, 410.33777432, 499.38468136, 445.23453519, 460.87791075, 463.28458391, 462.08124733, 483.74130580, 494.57133503, 487.35131554, 465.69125707, 471.70793998, 480.13129605, 465.69125707, 481.33463264, 478.92795947, 460.87791075, 478.92795947, 448.84454493, 469.30126682, 522.24807641, 542.70479830, 565.56819335, 655.81843697, 659.42844671, 636.56505166, 632.95504192, 637.76838824, 651.00509064, 671.46181253, 655.81843697, 652.20842722, 654.61510038, 648.59841748, 616.10832977, 594.44827130, 598.05828105, 601.66829079, 589.63492498, 623.32834926, 657.02177355, 693.12187099, 724.40862212, 707.56190997, 732.83197819, 720.79861237, 746.06868058, 749.67869033, 730.42530502, 736.44198793, 709.96858314, 697.93521732, 743.66200742, 742.45867084, 737.64532451, 684.69851492, 696.73188074, 646.19174431, 625.73502243, 624.53168584, 623.32834926, 632.95504192, 616.10832977, 610.09164687, 608.88831028, 600.46495421, 620.92167610, 619.71833952, 617.31166635, 628.14169559, 623.32834926, 613.70165661, 616.10832977, 610.09164687, 606.48163712, 619.71833952, 580.00823233, 589.63492498, 576.39822258, 546.31480804, 543.90813488, 445.23453519, 494.57133503, 484.94464238, 531.87476906, 500.58801794, 517.43473008, 539.09478855, 547.51814462, 513.82472034, 504.19802769, 498.18134478, 492.16466187, 499.38468136, 460.87791075, 487.35131554, 505.40136427, 523.45141299, 512.62138376, 541.50146171, 586.02491523, 582.41490549, 595.65160789, 608.88831028, 611.29498345, 623.32834926, 624.53168584, 614.90499319, 600.46495421, 593.24493472, 573.99154942, 573.99154942, 537.89145197, 512.62138376, 545.11147146, 552.33149095, 528.26475932, 510.21471059], 
							[792.99880727, 812.25219257, 768.93207563, 752.08536349, 703.95190023, 630.54836875, 614.90499319, 619.71833952, 634.15837850, 658.22511013, 703.95190023, 693.12187099, 711.17191972, 707.56190997, 715.98526604, 691.91853441, 661.83511987, 608.88831028, 655.81843697, 629.34503217, 619.71833952, 560.75484702, 586.02491523, 559.55151044, 501.79135452, 541.50146171, 618.51500294, 638.97172482, 659.42844671, 649.80175406, 669.05513936, 660.63178329, 637.76838824, 654.61510038, 619.71833952, 608.88831028, 602.87162738, 581.21156891, 608.88831028, 607.68497370, 594.44827130, 586.02491523, 601.66829079, 584.82157865, 566.77152993, 535.48477881, 527.06142274, 506.60470085, 536.68811539, 551.12815437, 543.90813488, 565.56819335, 549.92481779, 555.94150069, 570.38153967, 586.02491523, 613.70165661, 604.07496396, 619.71833952, 664.24179304, 635.36171508, 670.25847594, 671.46181253, 688.30852467, 687.10518809, 676.27515885, 669.05513936, 654.61510038, 653.41176380, 626.93835901, 626.93835901, 637.76838824, 640.17506140, 620.92167610, 632.95504192, 624.53168584, 628.14169559, 630.54836875, 646.19174431, 658.22511013, 635.36171508, 610.09164687, 628.14169559, 606.48163712, 632.95504192, 599.26161763, 619.71833952, 590.83826156, 542.70479830, 529.46809590, 523.45141299, 481.33463264, 513.82472034, 477.72462289, 513.82472034, 481.33463264, 505.40136427, 531.87476906, 553.53482753, 536.68811539, 506.60470085, 570.38153967, 553.53482753, 512.62138376, 524.65474957, 563.16152018, 554.73816411, 577.60155916, 551.12815437, 542.70479830, 547.51814462, 557.14483728, 557.14483728, 549.92481779, 555.94150069, 558.34817386, 566.77152993, 578.80489574, 577.60155916, 608.88831028, 617.31166635, 580.00823233, 564.36485676, 559.55151044, 566.77152993, 566.77152993, 569.17820309, 572.78821284, 588.43158840, 577.60155916, 571.58487625, 575.19488600, 545.11147146, 554.73816411, 558.34817386, 546.31480804, 546.31480804, 543.90813488, 549.92481779, 529.46809590, 553.53482753, 551.12815437, 537.89145197, 481.33463264, 512.62138376, 511.41804717, 509.01137401, 529.46809590, 522.24807641, 519.84140325, 533.07810564, 539.09478855, 487.35131554, 502.99469110, 492.16466187, 478.92795947, 438.01451570, 427.18448646, 403.11775483, 375.44101346, 397.10107192, 397.10107192, 376.64435004, 392.28772560, 433.20116937, 422.37114014, 423.57447672, 398.30440851, 389.88105243, 364.61098422, 388.67771585, 371.83100371, 327.30755019, 335.73090626, 300.83414540, 373.03434029, 361.00097448, 361.00097448, 379.05102320, 373.03434029, 361.00097448, 391.08438902, 388.67771585, 397.10107192, 397.10107192, 409.13443774, 397.10107192, 388.67771585, 404.32109141, 382.66103294, 382.66103294, 365.81432080, 382.66103294, 361.00097448, 377.84768662, 367.01765738, 373.03434029, 380.25435978, 387.47437927, 392.28772560, 392.28772560, 413.94778407, 427.18448646, 431.99783279, 431.99783279, 452.45455468, 469.30126682, 456.06456442, 436.81117912, 451.25121810, 441.62452544, 451.25121810, 445.23453519, 465.69125707, 500.58801794, 512.62138376, 492.16466187, 498.18134478, 487.35131554, 483.74130580, 495.77467161, 480.13129605, 499.38468136, 483.74130580, 500.58801794, 488.55465212, 501.79135452, 493.36799845, 470.50460340, 481.33463264, 453.65789126, 465.69125707, 452.45455468, 469.30126682, 484.94464238, 456.06456442, 448.84454493, 470.50460340, 478.92795947, 480.13129605, 456.06456442, 448.84454493, 466.89459366, 440.42118886, 436.81117912, 452.45455468, 441.62452544, 434.40450595, 406.72776458, 376.64435004, 397.10107192, 371.83100371, 376.64435004, 389.88105243, 304.44415514, 338.13757943, 323.69754045, 326.10421361, 308.05416489, 302.03748198, 290.00411616, 283.98743325, 264.73404795, 238.26064315, 249.09067239, 228.63395050, 215.39724810, 227.43061392, 243.07398948, 228.63395050, 234.65063341, 238.26064315, 247.88733581, 247.88733581, 246.68399923, 252.70068213, 251.49734555, 251.49734555, 247.88733581, 241.87065290, 231.04062366, 228.63395050, 223.82060418, 204.56721887, 199.75387254, 184.11049698, 168.46712142, 157.63709219, 150.41707270, 131.16368739, 114.31697525], 
							[1407.90380046, 1371.80370301, 1335.70360556, 1287.57014230, 1347.73697138, 1347.73697138, 1311.63687393, 1263.50341067, 1263.50341067, 1287.57014230, 1299.60350811, 1299.60350811, 1335.70360556, 1311.63687393, 1275.53677648, 1239.43667904, 1215.36994740, 1184.08319628, 1239.43667904, 1227.40331322, 1239.43667904, 1147.98309883, 1070.96955761, 1086.61293317, 1078.18957710, 1131.13638669, 1191.30321577, 1188.89654261, 1163.62647440, 1203.33658159, 1192.50655235, 1203.33658159, 1180.47318654, 1174.45650363, 1194.91322552, 1180.47318654, 1167.23648414, 1147.98309883, 1191.30321577, 1227.40331322, 1170.84649388, 1182.87985970, 1215.36994740, 1215.36994740, 1180.47318654, 1184.08319628, 1172.04983047, 1172.04983047, 1200.92990842, 1200.92990842, 1198.52323526, 1193.70988894, 1215.36994740, 1215.36994740, 1287.57014230, 1239.43667904, 1239.43667904, 1263.50341067, 1444.00389791, 1444.00389791, 1431.97053209, 1468.07062954, 1419.93716627, 1395.87043464, 1359.77033719, 1371.80370301, 1383.83706883, 1359.77033719, 1263.50341067, 1347.73697138, 1299.60350811, 1323.67023975, 1359.77033719, 1311.63687393, 1383.83706883, 1371.80370301, 1407.90380046, 1395.87043464, 1419.93716627, 1456.03726372, 1444.00389791, 1419.93716627, 1419.93716627, 1407.90380046, 1419.93716627, 1419.93716627, 1444.00389791, 1444.00389791, 1431.97053209, 1407.90380046, 1335.70360556, 1299.60350811, 1299.60350811, 1287.57014230, 1359.77033719, 1287.57014230, 1275.53677648, 1347.73697138, 1359.77033719, 1383.83706883, 1347.73697138, 1431.97053209, 1407.90380046, 1347.73697138, 1359.77033719, 1395.87043464, 1407.90380046, 1419.93716627, 1371.80370301, 1359.77033719, 1444.00389791, 1431.97053209, 1419.93716627, 1419.93716627, 1419.93716627, 1468.07062954, 1456.03726372, 1456.03726372, 1528.23745862, 1540.27082443, 1552.30419025, 1480.10399535, 1456.03726372, 1456.03726372, 1468.07062954, 1444.00389791, 1419.93716627, 1444.00389791, 1468.07062954, 1516.20409280, 1492.13736117, 1528.23745862, 1456.03726372, 1456.03726372, 1468.07062954, 1468.07062954, 1492.13736117, 1504.17072698, 1492.13736117, 1504.17072698, 1504.17072698, 1468.07062954, 1480.10399535, 1456.03726372, 1444.00389791, 1431.97053209, 1444.00389791, 1456.03726372, 1480.10399535, 1468.07062954, 1456.03726372, 1444.00389791, 1407.90380046, 1407.90380046, 1323.67023975, 1335.70360556, 1311.63687393, 1299.60350811, 1275.53677648, 1263.50341067, 1215.36994740, 1227.40331322, 1239.43667904, 1299.60350811, 1335.70360556, 1323.67023975, 1335.70360556, 1239.43667904, 1275.53677648, 1251.47004485, 1239.43667904, 1196.11656210, 1145.57642567, 1173.25316705, 1126.32304037, 1239.43667904, 1227.40331322, 1227.40331322, 1239.43667904, 1251.47004485, 1251.47004485, 1359.77033719, 1347.73697138, 1347.73697138, 1311.63687393, 1371.80370301, 1275.53677648, 1287.57014230, 1359.77033719, 1299.60350811, 1299.60350811, 1185.28653286, 1239.43667904, 1251.47004485, 1239.43667904, 1263.50341067, 1323.67023975, 1347.73697138, 1371.80370301, 1359.77033719, 1383.83706883, 1395.87043464, 1407.90380046, 1431.97053209, 1456.03726372, 1492.13736117, 1480.10399535, 1444.00389791, 1456.03726372, 1468.07062954, 1407.90380046, 1407.90380046, 1468.07062954, 1492.13736117, 1468.07062954, 1480.10399535, 1540.27082443, 1528.23745862, 1492.13736117, 1492.13736117, 1528.23745862, 1552.30419025, 1564.33755606, 1552.30419025, 1588.40428770, 1600.43765351, 1576.37092188, 1564.33755606, 1516.20409280, 1516.20409280, 1480.10399535, 1492.13736117, 1456.03726372, 1480.10399535, 1480.10399535, 1456.03726372, 1431.97053209, 1456.03726372, 1444.00389791, 1480.10399535, 1444.00389791, 1444.00389791, 1492.13736117, 1431.97053209, 1444.00389791, 1444.00389791, 1431.97053209, 1419.93716627, 1395.87043464, 1359.77033719, 1359.77033719, 1311.63687393, 1287.57014230, 1311.63687393, 1215.36994740, 1287.57014230, 1251.47004485, 1299.60350811, 1251.47004485, 1323.67023975, 1335.70360556, 1287.57014230, 1287.57014230, 1311.63687393, 1311.63687393, 1323.67023975, 1287.57014230, 1311.63687393, 1335.70360556, 1335.70360556, 1407.90380046, 1395.87043464, 1407.90380046, 1395.87043464, 1359.77033719, 1383.83706883, 1359.77033719, 1347.73697138, 1335.70360556, 1335.70360556, 1287.57014230, 1263.50341067, 1287.57014230, 1263.50341067, 1287.57014230, 1197.31989868, 1132.33972327, 1103.45964532, 1098.64629899, 1101.05297215, 1104.66298190], 
							[1275.53677648, 1182.87985970, 1137.15306960, 1147.98309883, 1149.18643542, 1155.20311832, 1203.33658159, 1239.43667904, 1239.43667904, 1203.33658159, 1215.36994740, 1181.67652312, 1170.84649388, 1160.01646465, 1095.03628924, 1062.54620154, 993.95601639, 971.09262134, 1046.90282598, 1027.64944068, 975.90596767, 938.60253364, 914.53580201, 934.99252389, 926.56916782, 962.66926527, 1015.61607486, 987.93933348, 977.10930425, 925.36583124, 879.63904114, 830.30224130, 824.28555839, 849.55562660, 820.67554864, 808.64218283, 789.38879752, 754.49203666, 768.93207563, 807.43884625, 782.16877803, 831.50557788, 837.52226079, 832.70891446, 819.47221206, 802.62549992, 765.32206589, 759.30538298, 786.98212436, 779.76210487, 754.49203666, 760.50871956, 767.72873905, 795.40548043, 937.39919706, 930.17917757, 1075.78290394, 1161.21980123, 1135.94973302, 1155.20311832, 1108.27299164, 1263.50341067, 1239.43667904, 1251.47004485, 1227.40331322, 1227.40331322, 1251.47004485, 1239.43667904, 1167.23648414, 1131.13638669, 1134.74639644, 1151.59310858, 1227.40331322, 1186.48986945, 1227.40331322, 1184.08319628, 1251.47004485, 1199.72657184, 1203.33658159, 1239.43667904, 1193.70988894, 1149.18643542, 1132.33972327, 1166.03314756, 1188.89654261, 1197.31989868, 1202.13324501, 1144.37308909, 1095.03628924, 1109.47632822, 1076.98624052, 972.29595792, 1045.69948940, 1058.93619180, 1091.42627950, 1018.02274802, 1028.85277726, 1157.60979149, 1135.94973302, 1181.67652312, 1162.42313781, 1263.50341067, 1251.47004485, 1178.06651337, 1251.47004485, 1227.40331322, 1239.43667904, 1335.70360556, 1323.67023975, 1287.57014230, 1335.70360556, 1383.83706883, 1383.83706883, 1359.77033719, 1371.80370301, 1419.93716627, 1431.97053209, 1431.97053209, 1456.03726372, 1528.23745862, 1564.33755606, 1492.13736117, 1492.13736117, 1504.17072698, 1492.13736117, 1444.00389791, 1371.80370301, 1431.97053209, 1431.97053209, 1419.93716627, 1431.97053209, 1504.17072698, 1444.00389791, 1407.90380046, 1419.93716627, 1431.97053209, 1480.10399535, 1516.20409280, 1480.10399535, 1468.07062954, 1468.07062954, 1383.83706883, 1444.00389791, 1371.80370301, 1395.87043464, 1359.77033719, 1347.73697138, 1347.73697138, 1371.80370301, 1359.77033719, 1311.63687393, 1197.31989868, 1149.18643542, 1153.99978174, 1113.08633797, 1099.84963557, 1101.05297215, 1131.13638669, 1113.08633797, 1075.78290394, 1051.71617231, 1117.89968430, 1078.18957710, 1080.59625027, 1119.10302088, 1119.10302088, 1117.89968430, 1090.22294292, 1098.64629899, 1087.81626976, 1101.05297215, 1060.13952838, 1036.07279675, 1099.84963557, 1117.89968430, 1227.40331322, 1239.43667904, 1251.47004485, 1251.47004485, 1323.67023975, 1371.80370301, 1431.97053209, 1444.00389791, 1600.43765351, 1720.77131167, 1768.90477493, 1624.50438514, 1660.60448259, 1696.70458004, 1600.43765351, 1696.70458004, 1600.43765351, 1684.67121422, 1636.53775096, 1552.30419025, 1492.13736117, 1564.33755606, 1612.47101933, 1492.13736117, 1492.13736117, 1516.20409280, 1576.37092188, 1516.20409280, 1564.33755606, 1528.23745862, 1552.30419025, 1564.33755606, 1528.23745862, 1492.13736117, 1504.17072698, 1395.87043464, 1371.80370301, 1395.87043464, 1492.13736117, 1504.17072698, 1504.17072698, 1576.37092188, 1636.53775096, 1588.40428770, 1516.20409280, 1540.27082443, 1456.03726372, 1431.97053209, 1407.90380046, 1480.10399535, 1431.97053209, 1431.97053209, 1444.00389791, 1407.90380046, 1359.77033719, 1359.77033719, 1359.77033719, 1371.80370301, 1407.90380046, 1431.97053209, 1395.87043464, 1395.87043464, 1407.90380046, 1456.03726372, 1552.30419025, 1492.13736117, 1504.17072698, 1528.23745862, 1468.07062954, 1516.20409280, 1552.30419025, 1588.40428770, 1660.60448259, 1660.60448259, 1636.53775096, 1660.60448259, 1624.50438514, 1684.67121422, 1684.67121422, 1648.57111678, 1672.63784841, 1684.67121422, 1708.73794585, 1660.60448259, 1756.87140912, 1732.80467749, 1684.67121422, 1684.67121422, 1672.63784841, 1648.57111678, 1672.63784841, 1768.90477493, 1829.07160401, 1805.00487238, 1841.10496983, 1913.30516472, 1937.37189636, 1925.33853054, 1889.23843309, 1949.40526217, 1973.47199380, 1973.47199380, 1985.50535962, 1997.53872544, 1877.20506728, 1949.40526217, 1985.50535962, 1973.47199380, 2057.70555452, 1901.27179891, 1961.43862799, 1853.13833565, 1624.50438514, 1684.67121422, 1588.40428770, 1672.63784841], 
							[1780.93814075, 1780.93814075, 1756.87140912, 1756.87140912, 1805.00487238, 1744.83804330, 1672.63784841, 1768.90477493, 1756.87140912, 1720.77131167, 1768.90477493, 1780.93814075, 1684.67121422, 1732.80467749, 1660.60448259, 1636.53775096, 1576.37092188, 1600.43765351, 1636.53775096, 1672.63784841, 1648.57111678, 1528.23745862, 1612.47101933, 1612.47101933, 1504.17072698, 1600.43765351, 1648.57111678, 1708.73794585, 1660.60448259, 1636.53775096, 1648.57111678, 1636.53775096, 1636.53775096, 1636.53775096, 1588.40428770, 1648.57111678, 1612.47101933, 1516.20409280, 1528.23745862, 1624.50438514, 1612.47101933, 1600.43765351, 1576.37092188, 1576.37092188, 1516.20409280, 1552.30419025, 1552.30419025, 1528.23745862, 1624.50438514, 1624.50438514, 1588.40428770, 1540.27082443, 1552.30419025, 1504.17072698, 1588.40428770, 1636.53775096, 1600.43765351, 1792.97150657, 1901.27179891, 2057.70555452, 1961.43862799, 2105.83901778, 2033.63882288, 2021.60545707, 1925.33853054, 1961.43862799, 1997.53872544, 1949.40526217, 1913.30516472, 1949.40526217, 1985.50535962, 1937.37189636, 1985.50535962, 1961.43862799, 2021.60545707, 2021.60545707, 2009.57209125, 1973.47199380, 1961.43862799, 1985.50535962, 2045.67218870, 2045.67218870, 2033.63882288, 2117.87238359, 2117.87238359, 2105.83901778, 2105.83901778, 2057.70555452, 1985.50535962, 1961.43862799, 1997.53872544, 1913.30516472, 1973.47199380, 1937.37189636, 1997.53872544, 1865.17170146, 1913.30516472, 2033.63882288, 2045.67218870, 2045.67218870, 1997.53872544, 2238.20604175, 2190.07257849, 2178.03921267, 2298.37287083, 2214.13931012, 2214.13931012, 2262.27277339, 2226.17267594, 2117.87238359, 2226.17267594, 2322.43960246, 2153.97248104, 2105.83901778, 2166.00584686, 2214.13931012, 2214.13931012, 2262.27277339, 2262.27277339, 2346.50633410, 2322.43960246, 2262.27277339, 2298.37287083, 2238.20604175, 2250.23940757, 2238.20604175, 2322.43960246, 2274.30613920, 2310.40623665, 2262.27277339, 2286.33950502, 2286.33950502, 2202.10594431, 2190.07257849, 2214.13931012, 2153.97248104, 2226.17267594, 2178.03921267, 2190.07257849, 2141.93911523, 2141.93911523, 2045.67218870, 2081.77228615, 2009.57209125, 2033.63882288, 1997.53872544, 2033.63882288, 2057.70555452, 2045.67218870, 2021.60545707, 2009.57209125, 1985.50535962, 1889.23843309, 1901.27179891, 1865.17170146, 1865.17170146, 1768.90477493, 1732.80467749, 1624.50438514, 1636.53775096, 1696.70458004, 1624.50438514, 1648.57111678, 1684.67121422, 1720.77131167, 1696.70458004, 1708.73794585, 1636.53775096, 1612.47101933, 1636.53775096, 1576.37092188, 1540.27082443, 1552.30419025, 1660.60448259, 1624.50438514, 1720.77131167, 1696.70458004, 1684.67121422, 1696.70458004, 1684.67121422, 1720.77131167, 1889.23843309, 1877.20506728, 1913.30516472, 1901.27179891, 1997.53872544, 1817.03823820, 1889.23843309, 2021.60545707, 1961.43862799, 2009.57209125, 1877.20506728, 2009.57209125, 2069.73892033, 2045.67218870, 2093.80565196, 2105.83901778, 2081.77228615, 2093.80565196, 2202.10594431, 2105.83901778, 2214.13931012, 2178.03921267, 2178.03921267, 2153.97248104, 2226.17267594, 2274.30613920, 2274.30613920, 2214.13931012, 2178.03921267, 2202.10594431, 2238.20604175, 2250.23940757, 2346.50633410, 2370.57306573, 2358.53969991, 2334.47296828, 2406.67316318, 2322.43960246, 2334.47296828, 2358.53969991, 2322.43960246, 2406.67316318, 2382.60643154, 2430.73989481, 2490.90672389, 2551.07355297, 2527.00682133, 2466.83999225, 2442.77326062, 2430.73989481, 2430.73989481, 2406.67316318, 2442.77326062, 2442.77326062, 2298.37287083, 2286.33950502, 2370.57306573, 2442.77326062, 2466.83999225, 2418.70652899, 2430.73989481, 2514.97345552, 2358.53969991, 2346.50633410, 2430.73989481, 2418.70652899, 2466.83999225, 2502.94008970, 2430.73989481, 2454.80662644, 2382.60643154, 2418.70652899, 2442.77326062, 2238.20604175, 2262.27277339, 2178.03921267, 2370.57306573, 2346.50633410, 2394.63979736, 2466.83999225, 2418.70652899, 2346.50633410, 2370.57306573, 2418.70652899, 2478.87335807, 2527.00682133, 2623.27374786, 2659.37384531, 2755.64077184, 2839.87433255, 2900.04116163, 2972.24135652, 2972.24135652, 2803.77423510, 2888.00779581, 2912.07452744, 2996.30808815, 3080.54164886, 3032.40818560, 3044.44155142, 3140.70847794, 3164.77520958, 3285.10886773, 3285.10886773, 3333.24233100, 3285.10886773, 3032.40818560, 3176.80857539, 2948.17462489, 2912.07452744], 
							[2833.37631501, 2756.79597495, 2691.16599779, 2713.04265685, 2647.40064632, 2625.52398727, 2559.88197674, 2505.19032911, 2538.00531769, 2538.00531769, 2855.25297406, 2691.16599779, 2713.04265685, 2767.74633785, 2702.10432732, 2691.16599779, 2647.40064632, 2603.64732822, 2713.04265685, 2636.46231680, 2603.64732822, 2472.36330716, 2297.33800137, 2308.27633090, 2209.81933179, 2439.54831858, 2559.88197674, 2636.46231680, 2559.88197674, 2527.06698816, 2603.64732822, 2505.19032911, 2505.19032911, 2548.94364722, 2527.06698816, 2505.19032911, 2516.12865864, 2428.60998906, 2559.88197674, 2614.58565774, 2592.70899869, 2713.04265685, 2734.91931590, 2800.56132643, 2734.91931590, 2745.85764543, 2734.91931590, 2745.85764543, 2767.74633785, 2789.62299690, 2723.98098637, 2778.68466737, 2734.91931590, 2658.33897585, 2702.10432732, 2636.46231680, 2505.19032911, 2483.31367006, 2527.06698816, 2669.28933874, 2625.52398727, 2669.28933874, 2647.40064632, 2625.52398727, 2669.28933874, 2680.22766827, 2581.77066916, 2581.77066916, 2581.77066916, 2592.70899869, 2603.64732822, 2658.33897585, 2702.10432732, 2581.77066916, 2691.16599779, 2636.46231680, 2636.46231680, 2669.28933874, 2756.79597495, 3117.80898280, 3095.93232374, 3063.11733516, 3216.26598190, 3292.84632195, 3095.93232374, 2986.53699511, 3128.74731232, 3074.05566469, 3041.22864274, 2953.70997317, 2899.01832553, 2680.22766827, 2647.40064632, 2680.22766827, 2723.98098637, 2516.12865864, 2417.67165953, 2636.46231680, 2822.43798548, 2855.25297406, 2789.62299690, 2811.49965595, 2778.68466737, 2658.33897585, 2680.22766827, 2680.22766827, 2767.74633785, 2756.79597495, 2527.06698816, 2516.12865864, 2494.25199958, 2527.06698816, 2548.94364722, 2527.06698816, 2559.88197674, 2570.83233964, 2570.83233964, 2538.00531769, 2548.94364722, 2570.83233964, 2559.88197674, 2439.54831858, 2373.90630806, 2505.19032911, 2516.12865864, 2516.12865864, 2516.12865864, 2461.42497764, 2505.19032911, 2516.12865864, 2636.46231680, 2581.77066916, 2570.83233964, 2581.77066916, 2680.22766827, 2713.04265685, 2734.91931590, 2691.16599779, 2691.16599779, 2734.91931590, 2702.10432732, 2723.98098637, 2702.10432732, 2636.46231680, 2734.91931590, 2723.98098637, 2702.10432732, 2745.85764543, 2713.04265685, 2713.04265685, 2745.85764543, 2756.79597495, 2603.64732822, 2614.58565774, 2614.58565774, 2527.06698816, 2472.36330716, 2373.90630806, 2286.39967185, 2253.57264990, 2297.33800137, 2406.67316318, 2334.47296828, 2382.60643154, 2490.90672389, 2430.73989481, 2430.73989481, 2370.57306573, 2394.63979736, 2406.67316318, 2418.70652899, 2382.60643154, 2382.60643154, 2406.67316318, 2370.57306573, 2490.90672389, 2466.83999225, 2575.14028460, 2539.04018715, 2490.90672389, 2442.77326062, 2551.07355297, 2611.24038205, 2623.27374786, 2719.54067439, 2719.54067439, 2587.17365041, 2719.54067439, 2803.77423510, 2707.50730857, 2671.40721112, 2683.44057694, 2719.54067439, 2719.54067439, 2647.34047949, 2683.44057694, 2719.54067439, 2767.67413765, 2731.57404020, 2815.80760092, 2791.74086928, 2912.07452744, 2996.30808815, 3044.44155142, 3068.50828305, 3056.47491723, 3092.57501468, 2984.27472234, 2960.20799071, 2960.20799071, 2912.07452744, 2972.24135652, 3056.47491723, 2948.17462489, 2996.30808815, 2984.27472234, 2972.24135652, 3044.44155142, 3020.37481979, 3056.47491723, 3044.44155142, 3116.64174631, 3080.54164886, 3116.64174631, 3285.10886773, 3369.34242845, 3417.47589171, 3345.27569681, 3489.67608660, 3417.47589171, 3309.17559937, 3249.00877029, 3224.94203866, 3200.87530702, 3417.47589171, 3249.00877029, 3224.94203866, 3345.27569681, 3345.27569681, 3429.50925753, 3285.10886773, 3357.30906263, 3393.40916008, 3369.34242845, 3405.44252589, 3513.74281824, 3513.74281824, 3597.97637895, 3670.17657384, 3934.91062179, 3790.51023200, 3814.57696363, 3983.04408506, 3946.94398761, 3730.34340292, 3874.74379271, 3826.61032945, 3802.54359782, 3995.07745087, 4247.77813300, 4163.54457229, 4043.21091413, 3971.01071924, 4043.21091413, 4103.37774321, 4151.51120648, 4127.44447485, 3922.87725598, 3922.87725598, 4055.24427995, 4067.27764577, 4031.17754832, 4271.84486464, 4440.31198606, 4440.31198606, 4524.54554677, 4536.57891259, 4452.34535187, 4548.61227840, 4789.27959472, 4813.34632635, 5017.91354522, 5222.48076409, 5066.04700848, 5234.51412991, 4801.31296054, 4753.17949727, 4404.21188861, 4705.04603401, 4464.37871769, 4813.34632635], 
							[1889.23843309, 1684.67121422, 1588.40428770, 1564.33755606, 1552.30419025, 1516.20409280, 1648.57111678, 1660.60448259, 1504.17072698, 1468.07062954, 1564.33755606, 1588.40428770, 1588.40428770, 1600.43765351, 1504.17072698, 1468.07062954, 1395.87043464, 1395.87043464, 1431.97053209, 1359.77033719, 1323.67023975, 1203.33658159, 1227.40331322, 1197.31989868, 1174.45650363, 1287.57014230, 1468.07062954, 1419.93716627, 1359.77033719, 1407.90380046, 1431.97053209, 1335.70360556, 1227.40331322, 1275.53677648, 1203.33658159, 1215.36994740, 1203.33658159, 1203.33658159, 1197.31989868, 1287.57014230, 1215.36994740, 1239.43667904, 1251.47004485, 1275.53677648, 1239.43667904, 1323.67023975, 1323.67023975, 1371.80370301, 1395.87043464, 1444.00389791, 1492.13736117, 1504.17072698, 1468.07062954, 1552.30419025, 1612.47101933, 1636.53775096, 1696.70458004, 1865.17170146, 1768.90477493, 1901.27179891, 1901.27179891, 1985.50535962, 2045.67218870, 1889.23843309, 1949.40526217, 1997.53872544, 1961.43862799, 1949.40526217, 1805.00487238, 1853.13833565, 1768.90477493, 1829.07160401, 1949.40526217, 1817.03823820, 1985.50535962, 1913.30516472, 1853.13833565, 1901.27179891, 1961.43862799, 2069.73892033, 2117.87238359, 2105.83901778, 1997.53872544, 2093.80565196, 2057.70555452, 1961.43862799, 1889.23843309, 1901.27179891, 1889.23843309, 1949.40526217, 1901.27179891, 1805.00487238, 1817.03823820, 1768.90477493, 1853.13833565, 1756.87140912, 1720.77131167, 1744.83804330, 1973.47199380, 2057.70555452, 1985.50535962, 2166.00584686, 2069.73892033, 2057.70555452, 2117.87238359, 2178.03921267, 2105.83901778, 2093.80565196, 2045.67218870, 1913.30516472, 1973.47199380, 1997.53872544, 1973.47199380, 2021.60545707, 2021.60545707, 2081.77228615, 2081.77228615, 2129.90574941, 2117.87238359, 2081.77228615, 2141.93911523, 2153.97248104, 2262.27277339, 2358.53969991, 2358.53969991, 2298.37287083, 2262.27277339, 2262.27277339, 2202.10594431, 2250.23940757, 2202.10594431, 2274.30613920, 2190.07257849, 2069.73892033, 2045.67218870, 2069.73892033, 2045.67218870, 2045.67218870, 2093.80565196, 2093.80565196, 2081.77228615, 2057.70555452, 1997.53872544, 1985.50535962, 1997.53872544, 2009.57209125, 2105.83901778, 2141.93911523, 2190.07257849, 2141.93911523, 1997.53872544, 1949.40526217, 1744.83804330, 1841.10496983, 1817.03823820, 1768.90477493, 1696.70458004, 1756.87140912, 1636.53775096, 1540.27082443, 1528.23745862, 1648.57111678, 1576.37092188, 1552.30419025, 1624.50438514, 1552.30419025, 1552.30419025, 1552.30419025, 1600.43765351, 1576.37092188, 1540.27082443, 1528.23745862, 1600.43765351, 1576.37092188, 1624.50438514, 1648.57111678, 1708.73794585, 1684.67121422, 1720.77131167, 1805.00487238, 1805.00487238, 2009.57209125, 1997.53872544, 2045.67218870, 2069.73892033, 2117.87238359, 2033.63882288, 2226.17267594, 2238.20604175, 2153.97248104, 2178.03921267, 2093.80565196, 2117.87238359, 2141.93911523, 2190.07257849, 2214.13931012, 2262.27277339, 2322.43960246, 2418.70652899, 2490.90672389, 2587.17365041, 2719.54067439, 2587.17365041, 2539.04018715, 2514.97345552, 2430.73989481, 2478.87335807, 2587.17365041, 2454.80662644, 2418.70652899, 2514.97345552, 2454.80662644, 2635.30711368, 2803.77423510, 2695.47394276, 2623.27374786, 2695.47394276, 2827.84096673, 2695.47394276, 2623.27374786, 2671.40721112, 2647.34047949, 2671.40721112, 2659.37384531, 2779.70750347, 2827.84096673, 2924.10789326, 2863.94106418, 2827.84096673, 2791.74086928, 2683.44057694, 2719.54067439, 2659.37384531, 2695.47394276, 2767.67413765, 2563.10691878, 2539.04018715, 2527.00682133, 2551.07355297, 2647.34047949, 2611.24038205, 2623.27374786, 2659.37384531, 2575.14028460, 2563.10691878, 2647.34047949, 2635.30711368, 2623.27374786, 2599.20701623, 2575.14028460, 2527.00682133, 2551.07355297, 2490.90672389, 2514.97345552, 2190.07257849, 2141.93911523, 2045.67218870, 2069.73892033, 1961.43862799, 2105.83901778, 2298.37287083, 2081.77228615, 2033.63882288, 2033.63882288, 2093.80565196, 2166.00584686, 2202.10594431, 2093.80565196, 2346.50633410, 2454.80662644, 2539.04018715, 2575.14028460, 2635.30711368, 2611.24038205, 2611.24038205, 2719.54067439, 2731.57404020, 2731.57404020, 2551.07355297, 2454.80662644, 2563.10691878, 2454.80662644, 2514.97345552, 2298.37287083, 2346.50633410, 2190.07257849, 2358.53969991, 2298.37287083, 2334.47296828, 2454.80662644, 2418.70652899], 
							[676.27515885, 672.66514911, 643.78507115, 695.52854416, 655.81843697, 602.87162738, 571.58487625, 519.84140325, 511.41804717, 529.46809590, 589.63492498, 599.26161763, 569.17820309, 583.61824207, 541.50146171, 518.63806666, 475.31794973, 433.20116937, 493.36799845, 483.74130580, 480.13129605, 480.13129605, 499.38468136, 489.75798871, 451.25121810, 466.89459366, 510.21471059, 517.43473008, 506.60470085, 534.28144222, 553.53482753, 482.53796922, 483.74130580, 494.57133503, 489.75798871, 451.25121810, 469.30126682, 424.77781330, 476.52128631, 474.11461315, 481.33463264, 481.33463264, 490.96132529, 476.52128631, 446.43787177, 425.98114988, 418.76113039, 398.30440851, 418.76113039, 452.45455468, 433.20116937, 453.65789126, 454.86122784, 445.23453519, 478.92795947, 477.72462289, 480.13129605, 470.50460340, 511.41804717, 535.48477881, 545.11147146, 601.66829079, 586.02491523, 598.05828105, 623.32834926, 620.92167610, 620.92167610, 595.65160789, 578.80489574, 549.92481779, 529.46809590, 534.28144222, 537.89145197, 518.63806666, 553.53482753, 543.90813488, 551.12815437, 554.73816411, 572.78821284, 565.56819335, 545.11147146, 533.07810564, 527.06142274, 541.50146171, 528.26475932, 523.45141299, 523.45141299, 487.35131554, 429.59115963, 423.57447672, 412.74444748, 362.20431106, 407.93110116, 406.72776458, 394.69439876, 383.86436953, 391.08438902, 391.08438902, 419.96446697, 435.60784253, 410.33777432, 442.82786202, 451.25121810, 470.50460340, 487.35131554, 502.99469110, 515.02805692, 547.51814462, 534.28144222, 505.40136427, 483.74130580, 519.84140325, 505.40136427, 504.19802769, 510.21471059, 522.24807641, 517.43473008, 533.07810564, 547.51814462, 581.21156891, 577.60155916, 535.48477881, 539.09478855, 535.48477881, 523.45141299, 529.46809590, 523.45141299, 528.26475932, 536.68811539, 529.46809590, 523.45141299, 516.23139350, 507.80803743, 480.13129605, 466.89459366, 470.50460340, 495.77467161, 501.79135452, 528.26475932, 528.26475932, 528.26475932, 519.84140325, 521.04473983, 488.55465212, 481.33463264, 505.40136427, 505.40136427, 525.85808615, 519.84140325, 513.82472034, 511.41804717, 482.53796922, 466.89459366, 483.74130580, 487.35131554, 495.77467161, 453.65789126, 435.60784253, 400.71108167, 411.54111090, 375.44101346, 387.47437927, 399.50774509, 415.15112065, 450.04788151, 424.77781330, 441.62452544, 431.99783279, 421.16780356, 395.89773534, 397.10107192, 354.98429157, 308.05416489, 330.91755994, 316.47752096, 397.10107192, 404.32109141, 409.13443774, 405.52442800, 397.10107192, 392.28772560, 439.21785228, 439.21785228, 450.04788151, 431.99783279, 489.75798871, 451.25121810, 445.23453519, 469.30126682, 475.31794973, 483.74130580, 442.82786202, 481.33463264, 452.45455468, 464.48792049, 458.47123758, 468.09793024, 477.72462289, 465.69125707, 478.92795947, 496.97800820, 501.79135452, 530.67143248, 529.46809590, 535.48477881, 577.60155916, 566.77152993, 553.53482753, 541.50146171, 530.67143248, 569.17820309, 587.22825181, 595.65160789, 613.70165661, 616.10832977, 611.29498345, 618.51500294, 637.76838824, 635.36171508, 612.49832003, 629.34503217, 604.07496396, 626.93835901, 618.51500294, 642.58173457, 647.39508089, 643.78507115, 640.17506140, 631.75170533, 637.76838824, 619.71833952, 604.07496396, 567.97486651, 577.60155916, 598.05828105, 542.70479830, 559.55151044, 590.83826156, 611.29498345, 590.83826156, 595.65160789, 582.41490549, 607.68497370, 570.38153967, 559.55151044, 594.44827130, 573.99154942, 584.82157865, 563.16152018, 517.43473008, 541.50146171, 495.77467161, 475.31794973, 463.28458391, 409.13443774, 422.37114014, 405.52442800, 434.40450595, 416.35445723, 428.38782305, 434.40450595, 435.60784253, 429.59115963, 419.96446697, 446.43787177, 405.52442800, 380.25435978, 382.66103294, 412.74444748, 407.93110116, 438.01451570, 438.01451570, 469.30126682, 448.84454493, 472.91127656, 472.91127656, 460.87791075, 457.26790100, 452.45455468, 428.38782305, 409.13443774, 422.37114014, 421.16780356, 358.59430131, 367.01765738, 344.15426233, 336.93424284, 333.32423310, 329.71422336, 310.46083805, 310.46083805], 
							[1822.16445204, 1845.07598055, 1787.78512590, 1833.62021629, 1845.07598055, 1845.07598055, 1845.07598055, 1810.69665441, 1787.78512590, 1776.31732828, 1867.99954243, 1810.69665441, 1696.10291175, 1810.69665441, 1741.93800214, 1707.55867600, 1604.42069760, 1512.73848345, 1627.33222611, 1592.95289997, 1604.42069760, 1478.35915731, 1604.42069760, 1558.57357384, 1524.19424770, 1604.42069760, 1707.55867600, 1764.86156402, 1684.63511413, 1741.93800214, 1730.48223788, 1501.27068582, 1558.57357384, 1638.80002373, 1615.87646185, 1627.33222611, 1673.17934987, 1673.17934987, 1707.55867600, 1776.31732828, 1764.86156402, 1810.69665441, 1799.24089016, 1833.62021629, 1764.86156402, 1753.39376640, 1684.63511413, 1730.48223788, 1776.31732828, 1822.16445204, 1776.31732828, 1822.16445204, 1822.16445204, 1810.69665441, 1856.54377817, 1936.75819470, 1879.45530668, 1856.54377817, 2005.51684697, 2097.19906112, 2028.44040885, 2143.04618488, 2108.66685875, 2085.74329687, 2062.81973499, 2085.74329687, 2039.89617311, 2005.51684697, 1994.06108271, 1971.13752084, 1959.68175658, 1971.13752084, 2062.81973499, 1994.06108271, 2016.98464459, 2016.98464459, 2108.66685875, 2039.89617311, 2062.81973499, 2085.74329687, 2005.51684697, 1971.13752084, 1936.75819470, 1936.75819470, 2028.44040885, 2005.51684697, 2051.36397073, 2039.89617311, 1959.68175658, 1925.30243044, 1856.54377817, 1753.39376640, 1753.39376640, 1764.86156402, 1982.60531846, 1890.92310431, 1890.92310431, 2039.89617311, 2051.36397073, 2143.04618488, 2074.28753261, 2211.80483715, 2211.80483715, 2211.80483715, 2452.47215347, 2383.70146783, 2475.38368198, 2486.85147961, 2360.78993932, 2223.26060141, 2314.94281556, 2337.86637744, 2395.16926546, 2395.16926546, 2452.47215347, 2395.16926546, 2452.47215347, 2498.30724386, 2509.76300812, 2578.52166039, 2635.82454841, 2486.85147961, 2429.54859159, 2418.08079397, 2452.47215347, 2418.08079397, 2372.24570358, 2395.16926546, 2429.54859159, 2372.24570358, 2337.86637744, 2395.16926546, 2326.41061318, 2326.41061318, 2383.70146783, 2326.41061318, 2383.70146783, 2360.78993932, 2349.32214170, 2349.32214170, 2314.94281556, 2223.26060141, 2200.34907290, 2188.88127527, 2211.80483715, 2188.88127527, 2188.88127527, 2200.34907290, 2269.10772517, 2246.18416329, 2177.42551102, 2028.44040885, 1890.92310431, 1948.22599232, 1959.68175658, 1890.92310431, 1856.54377817, 1822.16445204, 1776.31732828, 1810.69665441, 1673.17934987, 1867.99954243, 1787.78512590, 1822.16445204, 1879.45530668, 1867.99954243, 1890.92310431, 1822.16445204, 1822.16445204, 1810.69665441, 1787.78512590, 1707.55867600, 1764.86156402, 1776.31732828, 1707.55867600, 1971.13752084, 2097.19906112, 2062.81973499, 2028.44040885, 2074.28753261, 2085.74329687, 2257.63992755, 2131.57838726, 2257.63992755, 2326.41061318, 2280.56348943, 2120.12262300, 2200.34907290, 2200.34907290, 2246.18416329, 2314.94281556, 2349.32214170, 2441.00435585, 2475.38368198, 2441.00435585, 2349.32214170, 2372.24570358, 2372.24570358, 2349.32214170, 2429.54859159, 2452.47215347, 2498.30724386, 2395.16926546, 2441.00435585, 2463.92791773, 2475.38368198, 2509.76300812, 2521.23080574, 2429.54859159, 2429.54859159, 2406.62502971, 2372.24570358, 2509.76300812, 2478.87335807, 2502.94008970, 2599.20701623, 2671.40721112, 2767.67413765, 2683.44057694, 2611.24038205, 2719.54067439, 2587.17365041, 2599.20701623, 2587.17365041, 2671.40721112, 2635.30711368, 2527.00682133, 2490.90672389, 2430.73989481, 2346.50633410, 2310.40623665, 2298.37287083, 2274.30613920, 2286.33950502, 2358.53969991, 2310.40623665, 2334.47296828, 2286.33950502, 2346.50633410, 2406.67316318, 2394.63979736, 2370.57306573, 2442.77326062, 2346.50633410, 2334.47296828, 2370.57306573, 2382.60643154, 2442.77326062, 2502.94008970, 2527.00682133, 2539.04018715, 2527.00682133, 2539.04018715, 2539.04018715, 2478.87335807, 2539.04018715, 2599.20701623, 2671.40721112, 2683.44057694, 2924.10789326, 2863.94106418, 2803.77423510, 2779.70750347, 2695.47394276, 2755.64077184, 2803.77423510, 2936.14125907, 3008.34145397, 2924.10789326, 3008.34145397, 3188.84194121, 3224.94203866, 3164.77520958, 3116.64174631, 3513.74281824, 3465.60935497, 3561.87628150, 3694.24330547, 3622.04311058, 3549.84291568, 3778.47686619, 4007.11081669, 4079.31101158, 3995.07745087, 3958.97735342, 4067.27764577, 3730.34340292, 3706.27667129, 3850.67706108, 3682.20993966, 3862.71042690], 
							[1124.18110125, 1146.00962684, 1113.26683846, 1053.23237640, 1124.18110125, 1058.69552448, 1015.03847330, 896.06458548, 904.80080906, 896.06458548, 981.20064863, 1024.85769981, 965.91827404, 1004.12421051, 933.17548566, 901.52773356, 829.49600579, 818.58174299, 894.98158256, 987.74679963, 962.64519854, 927.72437094, 982.29568492, 950.63589945, 969.19134954, 996.48302321, 1066.33671177, 1124.18110125, 1102.35257566, 1124.18110125, 1102.35257566, 1085.97516479, 1055.42244898, 1066.33671177, 1069.60978728, 1044.50818618, 1085.97516479, 1080.52405007, 1102.35257566, 1124.18110125, 1113.26683846, 1113.26683846, 1124.18110125, 1110.51119768, 1065.81927704, 1039.45417254, 1057.79302204, 1003.93167665, 1079.57341417, 1088.73080556, 1055.50668254, 1031.43995091, 1070.40398942, 1029.14157804, 1077.27504130, 1058.93619180, 1036.01262992, 1026.85523853, 1071.54715917, 1127.70687743, 1093.31551793, 1111.65436744, 1117.38224956, 1191.88081733, 1157.50149120, 1226.26014347, 1226.26014347, 1180.42505307, 1135.72109907, 1140.30581144, 1134.57792932, 1123.12216506, 1157.50149120, 1121.96696194, 1129.99321694, 1127.70687743, 1168.95725545, 1134.57792932, 1168.95725545, 1180.42505307, 1180.42505307, 1138.01947194, 1146.03369357, 1132.27955645, 1132.27955645, 1135.72109907, 1123.12216506, 1095.61389080, 1019.97215329, 1003.93167665, 971.83869002, 912.24946250, 985.59282715, 929.43310889, 979.86494502, 928.28993913, 968.39714740, 966.11080789, 1036.01262992, 1048.62359729, 1002.78850690, 1096.75706056, 1060.07936155, 997.04859141, 1005.07484641, 1006.21801616, 1023.41369591, 1049.76676705, 998.20379452, 1000.49013403, 1007.37321928, 1009.65955878, 992.46387903, 979.86494502, 1008.51638903, 1002.78850690, 981.00811477, 1018.82898353, 1008.51638903, 1019.97215329, 1019.97215329, 968.39714740, 974.13706289, 968.39714740, 1016.53061066, 1033.72629041, 1031.43995091, 1025.70003541, 1009.65955878, 1014.24427116, 1011.94589829, 1025.70003541, 975.28023265, 964.96763814, 960.38292577, 937.45936389, 940.90090651, 947.77195839, 968.39714740, 962.66926527, 960.38292577, 967.25397765, 950.07033126, 907.66475013, 907.66475013, 905.36637726, 937.45936389, 950.07033126, 958.08455289, 956.94138314, 909.95108963, 922.56205701, 869.84388137, 882.45484874, 892.76744325, 908.80791988, 849.21869236, 789.61743147, 744.92551083, 765.55069984, 767.84907271, 727.72983108, 712.83252420, 731.17137370, 802.22839885, 807.95628098, 829.73667310, 773.57695484, 786.17588885, 771.27858197, 752.95176583, 704.81830257, 672.72531594, 735.75608608, 675.01165544, 804.51473835, 802.22839885, 830.87984285, 833.16618236, 813.68416310, 790.76060122, 835.46455523, 818.26887548, 842.33560711, 870.98705112, 865.25916899, 821.71041810, 814.82733286, 865.25916899, 836.60772498, 859.53128686, 812.54099335, 858.38811711, 812.54099335, 849.21869236, 841.19243736, 854.94657449, 879.01330612, 882.45484874, 916.83417488, 914.53580201, 938.60253364, 961.46592869, 979.51597741, 1022.83609435, 1020.42942119, 993.95601639, 997.56602614, 969.88928476, 959.05925553, 941.00920680, 956.65258236, 950.63589945, 980.71931399, 1010.80272853, 974.70263109, 1008.39605537, 1008.39605537, 992.75267981, 974.70263109, 1034.86946017, 1068.56288445, 1039.68280649, 1033.66612358, 1066.15621129, 1074.57956736, 1076.98624052, 1067.35954787, 1055.32618205, 1050.51283573, 1014.41273828, 990.34600665, 980.71931399, 1002.37937246, 1022.83609435, 954.24590920, 974.70263109, 991.54934323, 1012.00606512, 1012.00606512, 991.54934323, 986.73599690, 1013.20940170, 1004.78604563, 998.76936272, 1045.69948940, 1028.85277726, 1057.73285522, 1028.85277726, 930.17917757, 948.22922629, 895.28241670, 818.26887548, 824.28555839, 724.40862212, 774.94875854, 780.96544145, 821.87888522, 762.91539273, 790.59213410, 811.04885599, 783.37211461, 754.49203666, 750.88202691, 815.86220232, 790.59213410, 724.40862212, 747.27201717, 768.93207563, 772.54208538, 837.52226079, 839.92893395, 934.99252389, 924.16249466, 926.56916782, 941.00920680, 944.61921655, 930.17917757, 903.70577277, 855.57230951, 835.11558762, 879.63904114, 878.43570456, 813.45552915, 800.21882676, 752.08536349, 777.35543171, 818.26887548, 829.09890471, 814.65886573, 854.36897293], 
							[3443.14306099, 3204.87038447, 3038.07590090, 3216.78341663, 3216.78341663, 3038.07590090, 2859.35635180, 2954.67264243, 2954.67264243, 2895.10748164, 2930.84657812, 2954.67264243, 3049.98893306, 3026.15083538, 2930.84657812, 3121.46712600, 2859.35635180, 2764.05209454, 2942.75961027, 2918.93354596, 2859.35635180, 2883.18241612, 2930.84657812, 2775.96512670, 2716.38793254, 3038.07590090, 3264.43554526, 3288.26160958, 3216.78341663, 3109.55409385, 3335.91373821, 2978.49870675, 2990.41173890, 2978.49870675, 2895.10748164, 2859.35635180, 2907.02051380, 2752.13906238, 2847.44331965, 2847.44331965, 2799.79119101, 2823.61725533, 2859.35635180, 2907.02051380, 2811.70422317, 2764.05209454, 2704.47490039, 2728.31299807, 2799.79119101, 2835.53028749, 2907.02051380, 2895.10748164, 2978.49870675, 3061.90196522, 3121.46712600, 3192.95735232, 3145.29319032, 4050.75583450, 3669.51473872, 3526.54631946, 3562.28541594, 3586.11148025, 3526.54631946, 3466.98115868, 3502.72025515, 3490.80722299, 3490.80722299, 3538.45935162, 3371.66486805, 3407.40396452, 3419.31699668, 3574.19844810, 3609.93754457, 3550.37238378, 3609.93754457, 3598.02451241, 3621.86261009, 3633.77564225, 3502.72025515, 3490.80722299, 3502.72025515, 3371.66486805, 3466.98115868, 3419.31699668, 3633.77564225, 3574.19844810, 3669.51473872, 3502.72025515, 3502.72025515, 3609.93754457, 3478.89419083, 3395.49093236, 3490.80722299, 3562.28541594, 3455.05609315, 3383.57790021, 3383.57790021, 3383.57790021, 3407.40396452, 3359.75183589, 3300.17464174, 3371.66486805, 3383.57790021, 3419.31699668, 3419.31699668, 3443.14306099, 3419.31699668, 3395.49093236, 3347.83880373, 3216.78341663, 3181.03228679, 3181.03228679, 3169.11925464, 3157.20622248, 3228.69644879, 3252.52251311, 3240.60948095, 3181.03228679, 3216.78341663, 3181.03228679, 3133.38015816, 3133.38015816, 3157.20622248, 3121.46712600, 3145.29319032, 3109.55409385, 3133.38015816, 3109.55409385, 3169.11925464, 3085.72802953, 3097.64106169, 3097.64106169, 3061.90196522, 3097.64106169, 3061.90196522, 3002.32477106, 2930.84657812, 2990.41173890, 2966.58567459, 2918.93354596, 2930.84657812, 2907.02051380, 2871.26938396, 2871.26938396, 2835.53028749, 2811.70422317, 2847.44331965, 2871.26938396, 2895.10748164, 2847.44331965, 2799.79119101, 2775.96512670, 2632.99670744, 2823.61725533, 2799.79119101, 2811.70422317, 2811.70422317, 2787.87815886, 2716.38793254, 2740.22603023, 2764.05209454, 2764.05209454, 2799.79119101, 2799.79119101, 2895.10748164, 2787.87815886, 2883.18241612, 2823.61725533, 2799.79119101, 2787.87815886, 2764.05209454, 2787.87815886, 2752.13906238, 2787.87815886, 2764.05209454, 2859.35635180, 2755.64077184, 2767.67413765, 2767.67413765, 2695.47394276, 2719.54067439, 2791.74086928, 2779.70750347, 2767.67413765, 2827.84096673, 2984.27472234, 2984.27472234, 2888.00779581, 2924.10789326, 2900.04116163, 2948.17462489, 2851.90769836, 2900.04116163, 2875.97442999, 2888.00779581, 2900.04116163, 2912.07452744, 2863.94106418, 2875.97442999, 2996.30808815, 2936.14125907, 2996.30808815, 2948.17462489, 2972.24135652, 2960.20799071, 2996.30808815, 2984.27472234, 2972.24135652, 2924.10789326, 2948.17462489, 2948.17462489, 2900.04116163, 2996.30808815, 2960.20799071, 3044.44155142, 3032.40818560, 3152.74184376, 3200.87530702, 3164.77520958, 3080.54164886, 3116.64174631, 3092.57501468, 3020.37481979, 3020.37481979, 2948.17462489, 2984.27472234, 2984.27472234, 2948.17462489, 2888.00779581, 2900.04116163, 2900.04116163, 2948.17462489, 2900.04116163, 2924.10789326, 2912.07452744, 2888.00779581, 2863.94106418, 2888.00779581, 2888.00779581, 2863.94106418, 2900.04116163, 2851.90769836, 2863.94106418, 2827.84096673, 2827.84096673, 2827.84096673, 2803.77423510, 2803.77423510, 2803.77423510, 2791.74086928, 2755.64077184, 2779.70750347, 2791.74086928, 2767.67413765, 2647.34047949, 2719.54067439, 2707.50730857, 2683.44057694, 2442.77326062, 2430.73989481, 2466.83999225, 2563.10691878, 2563.10691878, 2599.20701623, 2623.27374786, 2478.87335807, 2478.87335807, 2406.67316318, 2502.94008970, 2490.90672389, 2478.87335807, 2478.87335807, 2478.87335807, 2575.14028460, 2539.04018715, 2527.00682133, 2575.14028460, 2551.07355297, 2478.87335807, 2454.80662644, 2478.87335807, 2430.73989481, 2442.77326062, 2442.77326062, 2466.83999225, 2430.73989481, 2418.70652899, 2430.73989481, 2454.80662644, 2466.83999225, 2490.90672389], 
							[1612.47101933, 1576.37092188, 1564.33755606, 1516.20409280, 1407.90380046, 1323.67023975, 1113.08633797, 1025.24276751, 960.26259211, 926.56916782, 1039.68280649, 1046.90282598, 1064.95287471, 1086.61293317, 1061.34286496, 1025.24276751, 913.33246543, 866.40233874, 932.58585073, 928.97584099, 910.92579226, 862.79232900, 908.51911910, 863.99566558, 824.28555839, 896.48575328, 962.66926527, 1038.47946991, 1024.03943093, 1083.00292343, 1116.69634771, 1030.05611384, 967.48261160, 1046.90282598, 1040.88614307, 983.12598716, 1031.25945042, 957.85591894, 1036.07279675, 1064.95287471, 1076.98624052, 1037.27613333, 1046.90282598, 1012.00606512, 967.48261160, 965.07593843, 954.24590920, 968.68594818, 965.07593843, 1001.17603588, 969.88928476, 956.65258236, 984.32932374, 980.71931399, 1091.42627950, 1083.00292343, 1080.59625027, 1081.79958685, 1004.78604563, 1060.13952838, 1018.02274802, 1070.96955761, 1057.73285522, 1073.37623078, 1073.37623078, 1158.81312807, 1181.67652312, 1161.21980123, 1092.62961608, 1070.96955761, 1141.96641593, 1093.83295266, 1122.71303062, 1078.18957710, 1096.23962583, 1138.35640618, 1155.20311832, 1161.21980123, 1191.30321577, 1173.25316705, 1185.28653286, 1153.99978174, 1161.21980123, 1140.76307935, 1079.39291368, 1060.13952838, 1078.18957710, 965.07593843, 824.28555839, 849.55562660, 803.82883650, 728.01863186, 794.20214385, 762.91539273, 815.86220232, 753.28870007, 780.96544145, 817.06553890, 884.45238747, 897.68908986, 860.38565584, 960.26259211, 926.56916782, 843.53894369, 886.85906063, 888.06239721, 916.94247517, 961.46592869, 906.11244594, 883.24905089, 913.33246543, 926.56916782, 878.43570456, 884.45238747, 895.28241670, 972.29595792, 978.31264083, 986.73599690, 974.70263109, 1002.37937246, 998.76936272, 986.73599690, 959.05925553, 962.66926527, 986.73599690, 986.73599690, 972.29595792, 974.70263109, 1007.19271879, 1001.17603588, 1027.64944068, 1014.41273828, 963.87260185, 980.71931399, 986.73599690, 916.94247517, 908.51911910, 926.56916782, 944.61921655, 950.63589945, 969.88928476, 974.70263109, 920.55248491, 850.75896318, 868.80901191, 897.68908986, 920.55248491, 932.58585073, 926.56916782, 913.33246543, 937.39919706, 926.56916782, 882.04571430, 963.87260185, 944.61921655, 962.66926527, 920.55248491, 853.16563635, 770.13541222, 813.45552915, 818.26887548, 758.10204640, 773.74542196, 801.42216334, 871.21568507, 885.65572405, 890.46907037, 843.53894369, 803.82883650, 770.13541222, 799.01549017, 777.35543171, 715.98526604, 719.59527579, 672.66514911, 805.03217308, 824.28555839, 812.25219257, 842.33560711, 812.25219257, 805.03217308, 903.70577277, 866.40233874, 885.65572405, 926.56916782, 932.58585073, 900.09576303, 916.94247517, 974.70263109, 938.60253364, 962.66926527, 938.60253364, 938.60253364, 920.55248491, 972.29595792, 941.00920680, 987.93933348, 1034.86946017, 1032.46278700, 998.76936272, 1021.63275777, 1049.30949914, 1145.57642567, 1126.32304037, 1176.86317679, 1299.60350811, 1239.43667904, 1239.43667904, 1157.60979149, 1192.50655235, 1227.40331322, 1251.47004485, 1203.33658159, 1287.57014230, 1263.50341067, 1227.40331322, 1263.50341067, 1251.47004485, 1227.40331322, 1199.72657184, 1287.57014230, 1347.73697138, 1383.83706883, 1371.80370301, 1359.77033719, 1407.90380046, 1383.83706883, 1359.77033719, 1335.70360556, 1311.63687393, 1287.57014230, 1275.53677648, 1170.84649388, 1203.33658159, 1275.53677648, 1180.47318654, 1192.50655235, 1275.53677648, 1359.77033719, 1371.80370301, 1323.67023975, 1347.73697138, 1383.83706883, 1347.73697138, 1275.53677648, 1407.90380046, 1359.77033719, 1383.83706883, 1480.10399535, 1383.83706883, 1371.80370301, 1299.60350811, 1215.36994740, 1203.33658159, 961.46592869, 993.95601639, 962.66926527, 973.49929450, 944.61921655, 948.22922629, 977.10930425, 951.83923604, 947.02588971, 918.14581175, 992.75267981, 903.70577277, 807.43884625, 784.57545120, 845.94561686, 791.79547068, 878.43570456, 936.19586048, 1010.80272853, 977.10930425, 1003.58270904, 1028.85277726, 1024.03943093, 1016.81941144, 1004.78604563, 939.80587022, 896.48575328, 949.43256287, 913.33246543, 818.26887548, 797.81215359, 734.03531477, 699.13855390, 685.90185151, 690.71519783, 655.81843697, 643.78507115], 
							[4222.61636508, 4244.49302414, 4102.28270693, 3981.94904877, 4025.70236687, 3905.36870871, 3511.55274566, 3424.03407608, 3424.03407608, 3369.34242845, 3500.61441613, 3478.73775708, 3555.31809713, 3555.31809713, 3467.79942755, 3456.86109802, 3216.19378171, 3139.61344165, 3227.13211123, 3358.40409892, 3281.82375887, 2975.52646539, 3238.07044076, 3194.31712265, 3106.79845307, 3347.46576939, 3610.00974476, 3730.34340292, 3620.94807429, 3752.22006198, 3686.59008482, 3336.52743987, 3270.88542934, 3391.21908750, 3391.21908750, 3270.88542934, 3402.15741703, 3380.28075797, 3544.36773424, 3642.82473334, 3544.36773424, 3577.19475618, 3500.61441613, 3500.61441613, 3314.65078081, 3216.19378171, 3117.73678260, 3106.79845307, 3270.88542934, 3336.52743987, 3227.13211123, 3161.49010071, 3249.00877029, 3325.58911034, 3511.55274566, 3610.00974476, 3806.92374297, 3500.61441613, 3445.92276850, 3642.82473334, 3522.49107518, 3741.28173245, 3631.88640382, 3511.55274566, 3434.98443897, 3391.21908750, 3358.40409892, 3292.76208839, 3270.88542934, 3358.40409892, 3391.21908750, 3369.34242845, 3434.98443897, 3347.46576939, 3413.09574655, 3445.92276850, 3445.92276850, 3391.21908750, 3391.21908750, 3544.36773424, 3500.61441613, 3456.86109802, 3467.79942755, 3478.73775708, 3358.40409892, 3314.65078081, 3270.88542934, 3183.36675976, 2953.64980634, 2997.40312444, 2898.94612534, 3019.27978350, 3117.73678260, 3008.34145397, 3128.67511213, 2964.58813586, 3063.03310160, 3281.82375887, 3358.40409892, 3358.40409892, 3270.88542934, 3566.25642666, 3522.49107518, 3150.55177118, 3347.46576939, 3445.92276850, 3325.58911034, 3445.92276850, 3281.82375887, 3172.42843023, 3183.36675976, 3128.67511213, 3128.67511213, 3128.67511213, 3238.07044076, 3161.49010071, 3172.42843023, 3216.19378171, 3336.52743987, 3577.19475618, 3675.65175529, 3445.92276850, 3467.79942755, 3631.88640382, 3642.82473334, 3631.88640382, 3599.07141524, 3697.52841434, 3653.76306287, 3533.42940471, 3533.42940471, 3675.65175529, 3445.92276850, 3511.55274566, 3489.67608660, 3369.34242845, 3402.15741703, 3467.79942755, 3533.42940471, 3522.49107518, 3544.36773424, 3577.19475618, 3653.76306287, 3588.13308571, 3588.13308571, 3588.13308571, 3653.76306287, 3675.65175529, 3631.88640382, 3610.00974476, 3555.31809713, 3511.55274566, 3227.13211123, 3391.21908750, 3380.28075797, 3456.86109802, 3281.82375887, 3205.25545218, 2997.40312444, 3052.09477208, 3019.27978350, 2953.64980634, 2898.94612534, 2898.94612534, 3084.92179402, 3172.42843023, 3303.70041792, 3117.73678260, 3227.13211123, 3073.98346450, 3084.92179402, 2822.36578529, 2800.48912623, 2953.64980634, 2712.98249002, 3227.13211123, 3194.31712265, 3238.07044076, 3303.70041792, 3205.25545218, 3063.03310160, 3467.79942755, 3336.52743987, 3555.31809713, 3719.40507340, 3719.40507340, 3424.03407608, 3380.28075797, 3566.25642666, 3314.65078081, 3369.34242845, 3227.13211123, 3194.31712265, 3041.15644255, 3259.94709981, 3238.07044076, 3369.34242845, 3358.40409892, 3413.09574655, 3380.28075797, 3380.28075797, 3391.21908750, 3336.52743987, 3522.49107518, 3719.40507340, 3620.94807429, 3588.13308571, 3741.28173245, 3631.88640382, 3653.76306287, 3631.88640382, 3522.49107518, 3653.76306287, 4124.15936598, 4102.28270693, 3960.07238971, 4156.97435456, 4244.49302414, 3905.36870871, 4014.76403735, 4102.28270693, 4342.95002324, 4222.61636508, 4189.80137650, 4156.97435456, 4156.97435456, 4178.86304698, 4102.28270693, 4003.82570782, 4091.34437740, 4025.70236687, 3971.01071924, 3927.24536777, 3960.07238971, 4146.03602503, 3817.86207250, 3620.94807429, 3763.15839150, 4014.76403735, 3960.07238971, 4069.46771835, 4102.28270693, 4167.92471745, 4080.40604787, 3949.13406019, 4047.57902593, 3949.13406019, 4124.15936598, 3949.13406019, 3795.98541345, 3828.80040203, 3577.19475618, 3238.07044076, 3052.09477208, 2625.46382044, 3073.98346450, 3073.98346450, 3434.98443897, 3084.92179402, 3150.55177118, 3238.07044076, 3095.86012355, 2986.46479492, 2844.25447770, 3095.86012355, 2839.87433255, 2514.97345552, 2635.30711368, 2671.40721112, 2683.44057694, 2815.80760092, 3008.34145397, 3044.44155142, 2839.87433255, 2671.40721112, 2839.87433255, 2791.74086928, 2875.97442999, 2960.20799071, 2791.74086928, 2767.67413765, 2490.90672389, 2394.63979736, 2370.57306573, 2527.00682133, 2478.87335807, 2502.94008970, 2502.94008970, 2527.00682133, 2430.73989481, 2502.94008970], 
							[839975.98440862, 806943.31224111, 773910.51973994, 782168.65769841, 755034.62112018, 707845.05507127, 717282.94421432, 744416.98079254, 719642.44658350, 744416.98079254, 753854.99026925, 755034.62112018, 757394.12348936, 749135.98553090, 720822.19776809, 711384.18829138, 707845.05507127, 684250.15171315, 742057.47842337, 739697.97605419, 724361.45132185, 697227.29440997, 693688.04085620, 600488.53960931, 563916.49355437, 654756.61276575, 738518.22486960, 713743.69066055, 679531.14697480, 697227.29440997, 684250.15171315, 637060.58566424, 632341.46059223, 642959.22125352, 704305.80151750, 651217.35921199, 631161.82974130, 600488.53960931, 624083.32263376, 631161.82974130, 618184.56671082, 640599.71888435, 664194.62224246, 684250.15171315, 673632.51138551, 657116.11513493, 622903.57144918, 619364.31789541, 678351.39579021, 681890.64934398, 686609.65408233, 718462.69539891, 726720.95369103, 831717.84645016, 955590.75816270, 965028.76763941, 994522.30658681, 1179741.67822968, 1215134.09343368, 1191539.06974190, 1075924.41632148, 1160865.77960992, 1140810.25013922, 1162045.65112816, 1143169.63217474, 1152607.64165145, 1141989.88099016, 1143169.63217474, 1073564.91395230, 1044071.37500490, 1031094.23230809, 1042891.62382031, 1079463.66987525, 1046430.87737408, 1073564.91395230, 1080643.30072618, 1150248.13928228, 1112496.46237641, 1117215.34678111, 1126653.23592416, 1087721.80783371, 1045251.12618949, 1000420.94217610, 1038172.73941562, 1042891.62382031, 1068845.90921395, 1036992.86789737, 980365.29237175, 955590.75816270, 942613.61546588, 902502.31585719, 808123.06342570, 858851.88302838, 861211.50573121, 930816.22395366, 852953.12710544, 877727.90198180, 928456.72158448, 967388.27000859, 1038172.73941562, 1035813.23704644, 1164405.03316368, 1126653.23592416, 1061767.52244007, 1123114.10270405, 1130192.48947793, 1117215.34678111, 1138450.74777005, 1085362.30546453, 1067666.27836302, 1066486.40684477, 1084182.67461360, 1058228.26888631, 1040532.12145114, 1053509.26414795, 1049970.13092785, 994522.30658681, 1024015.72520055, 1045251.12618949, 1031094.23230809, 996881.68862233, 983904.54592551, 1039352.37026655, 1008679.20046822, 1013398.08487291, 993342.43506856, 987443.79947928, 1007499.32894997, 1014577.83605750, 1026375.22756973, 1038172.73941562, 1097159.69697676, 1044071.37500490, 1055868.76651713, 1044071.37500490, 1038172.73941562, 1020476.47164679, 1045251.12618949, 1055868.76651713, 1049970.13092785, 1079463.66987525, 1020476.47164679, 1028734.72993891, 978005.91033623, 979185.54118716, 979185.54118716, 992162.80421763, 1019296.84079586, 1039352.37026655, 1014577.83605750, 1019296.84079586, 979185.54118716, 904861.93856002, 914299.82770307, 883626.53757108, 856492.38065920, 836436.85118851, 884806.28875567, 845874.74033156, 844694.98914697, 826998.96204546, 882446.78638649, 881267.03520191, 888345.54230944, 913119.95618483, 876548.03046355, 885985.91960660, 867110.02098684, 847054.49151615, 848234.24270074, 848234.24270074, 828178.59289639, 814021.69901498, 854132.99862368, 837616.60237310, 880087.28401732, 928456.72158448, 893064.42671413, 896603.55993424, 917838.96092318, 920198.46329236, 1015757.58724209, 1007499.32894997, 1061767.52244007, 1081823.17224442, 1119574.84915029, 1028734.72993891, 1026446.10409439, 1066156.21128678, 1030056.11383915, 1054122.84547091, 1025242.76751280, 985532.66032040, 969889.28475976, 986735.99690199, 986735.99690199, 1019226.08460486, 1024039.43093121, 993956.01639152, 1002379.37246263, 1004786.04562581, 1033666.12358391, 971092.62134135, 984329.32373881, 977109.30424929, 981922.65057564, 986735.99690199, 985532.66032040, 968685.94817817, 933789.18731213, 913332.46542513, 901299.09960926, 947025.88970959, 941009.20680165, 996362.68955469, 963872.60185182, 981922.65057564, 995159.35297310, 965075.93843341, 957855.91894388, 971092.62134135, 936195.86047530, 945822.55312800, 948229.22629118, 980719.31399405, 975905.96766770, 977109.30424929, 966279.27501500, 962669.26527023, 951839.23603594, 938602.53363848, 937399.19705689, 914535.80200672, 984329.32373881, 975905.96766770, 927772.50440419, 966279.27501500, 971092.62134135, 984329.32373881, 984329.32373881, 962669.26527023, 956652.58236229, 977109.30424929, 950635.89945436, 947025.88970959, 962669.26527023, 961465.92868865, 986735.99690199, 977109.30424929, 1016819.41144168, 1090222.94291854, 1048106.16256297, 1054122.84547091, 1056529.51863408, 1024039.43093121, 1022836.09434962, 1022836.09434962, 1052919.50888932, 1016819.41144168, 1064952.87470520, 1061342.86496043, 1033666.12358391, 1027649.44067597, 1013209.40169692, 1040886.14307344, 1044496.15281820, 1026446.10409439, 1045699.48939979, 1067359.54786837, 1063749.53812361, 1167236.48414016, 1180473.18653762, 1287570.14229894, 1263503.41066718, 1335703.60556245, 1299603.50811481, 1323670.23974657, 1287570.14229894, 1311636.87393069, 1287570.14229894, 1311636.87393069, 1371803.70301008, 1383837.06882596, 1468070.62953711, 1468070.62953711, 1419937.16627359, 1347736.97137833, 1359770.33719420, 1359770.33719420, 1239436.67903543, 1287570.14229894], 
							[593.24493472, 566.77152993, 545.11147146, 547.51814462, 527.06142274, 516.23139350, 504.19802769, 493.36799845, 510.21471059, 511.41804717, 534.28144222, 506.60470085, 506.60470085, 512.62138376, 492.16466187, 480.13129605, 442.82786202, 416.35445723, 459.67457417, 469.30126682, 440.42118886, 439.21785228, 434.40450595, 421.16780356, 393.49106218, 450.04788151, 513.82472034, 531.87476906, 499.38468136, 507.80803743, 504.19802769, 469.30126682, 477.72462289, 469.30126682, 454.86122784, 450.04788151, 460.87791075, 441.62452544, 453.65789126, 477.72462289, 453.65789126, 463.28458391, 470.50460340, 465.69125707, 446.43787177, 436.81117912, 431.99783279, 418.76113039, 456.06456442, 459.67457417, 421.16780356, 416.35445723, 428.38782305, 409.13443774, 442.82786202, 552.33149095, 559.55151044, 587.22825181, 587.22825181, 578.80489574, 570.38153967, 622.12501268, 610.09164687, 581.21156891, 596.85494447, 595.65160789, 602.87162738, 586.02491523, 565.56819335, 571.58487625, 565.56819335, 573.99154942, 590.83826156, 567.97486651, 588.43158840, 587.22825181, 622.12501268, 602.87162738, 613.70165661, 646.19174431, 638.97172482, 613.70165661, 630.54836875, 628.14169559, 643.78507115, 631.75170533, 629.34503217, 618.51500294, 598.05828105, 601.66829079, 577.60155916, 554.73816411, 569.17820309, 554.73816411, 570.38153967, 518.63806666, 533.07810564, 543.90813488, 569.17820309, 571.58487625, 553.53482753, 596.85494447, 583.61824207, 592.04159814, 587.22825181, 602.87162738, 605.27830054, 606.48163712, 593.24493472, 589.63492498, 593.24493472, 616.10832977, 626.93835901, 624.53168584, 642.58173457, 649.80175406, 638.97172482, 664.24179304, 679.88516860, 689.51186125, 675.07182227, 672.66514911, 654.61510038, 649.80175406, 666.64846620, 640.17506140, 640.17506140, 665.44512962, 671.46181253, 665.44512962, 661.83511987, 655.81843697, 660.63178329, 646.19174431, 661.83511987, 660.63178329, 651.00509064, 666.64846620, 669.05513936, 667.85180278, 666.64846620, 642.58173457, 655.81843697, 618.51500294, 641.37839799, 632.95504192, 640.17506140, 636.56505166, 661.83511987, 657.02177355, 641.37839799, 602.87162738, 572.78821284, 599.26161763, 566.77152993, 561.95818360, 548.72148120, 571.58487625, 549.92481779, 542.70479830, 517.43473008, 530.67143248, 521.04473983, 537.89145197, 586.02491523, 557.14483728, 571.58487625, 553.53482753, 548.72148120, 554.73816411, 543.90813488, 523.45141299, 507.80803743, 541.50146171, 518.63806666, 559.55151044, 558.34817386, 547.51814462, 545.11147146, 519.84140325, 553.53482753, 595.65160789, 602.87162738, 595.65160789, 602.87162738, 631.75170533, 580.00823233, 595.65160789, 628.14169559, 607.68497370, 610.09164687, 578.80489574, 599.26161763, 595.65160789, 599.26161763, 618.51500294, 649.80175406, 666.64846620, 649.80175406, 670.25847594, 672.66514911, 693.12187099, 691.91853441, 679.88516860, 677.47849543, 685.90185151, 679.88516860, 683.49517834, 666.64846620, 675.07182227, 672.66514911, 672.66514911, 667.85180278, 720.79861237, 734.03531477, 766.52540247, 732.83197819, 756.89870982, 740.05199768, 732.83197819, 717.18860263, 703.95190023, 708.76524656, 701.54522707, 726.81529528, 717.18860263, 722.00194895, 718.39193921, 678.68183202, 681.08850518, 677.47849543, 685.90185151, 671.46181253, 664.24179304, 676.27515885, 635.36171508, 626.93835901, 670.25847594, 669.05513936, 695.52854416, 677.47849543, 665.44512962, 673.86848569, 643.78507115, 629.34503217, 647.39508089, 637.76838824, 625.73502243, 619.71833952, 601.66829079, 598.05828105, 590.83826156, 592.04159814, 599.26161763, 515.02805692, 557.14483728, 536.68811539, 577.60155916, 576.39822258, 572.78821284, 588.43158840, 567.97486651, 537.89145197, 547.51814462, 586.02491523, 582.41490549, 576.39822258, 549.92481779, 561.95818360, 584.82157865, 586.02491523, 600.46495421, 611.29498345, 624.53168584, 595.65160789, 613.70165661, 620.92167610, 620.92167610, 617.31166635, 612.49832003, 595.65160789, 607.68497370, 602.87162738, 577.60155916, 541.50146171, 537.89145197, 517.43473008, 525.85808615, 513.82472034, 495.77467161, 481.33463264], 
							[1263.50341067, 1275.53677648, 1203.33658159, 1215.36994740, 1215.36994740, 1162.42313781, 1167.23648414, 1191.30321577, 1215.36994740, 1181.67652312, 1263.50341067, 1215.36994740, 1263.50341067, 1227.40331322, 1203.33658159, 1197.31989868, 1155.20311832, 1191.30321577, 1192.50655235, 1239.43667904, 1194.91322552, 1199.72657184, 1251.47004485, 1239.43667904, 1169.64315730, 1419.93716627, 1504.17072698, 1552.30419025, 1395.87043464, 1347.73697138, 1419.93716627, 1335.70360556, 1323.67023975, 1323.67023975, 1347.73697138, 1323.67023975, 1335.70360556, 1299.60350811, 1371.80370301, 1407.90380046, 1383.83706883, 1444.00389791, 1444.00389791, 1480.10399535, 1456.03726372, 1395.87043464, 1407.90380046, 1383.83706883, 1431.97053209, 1456.03726372, 1359.77033719, 1383.83706883, 1383.83706883, 1299.60350811, 1359.77033719, 1492.13736117, 1468.07062954, 1648.57111678, 1660.60448259, 1612.47101933, 1576.37092188, 1684.67121422, 1660.60448259, 1576.37092188, 1564.33755606, 1648.57111678, 1660.60448259, 1600.43765351, 1552.30419025, 1540.27082443, 1516.20409280, 1564.33755606, 1600.43765351, 1540.27082443, 1588.40428770, 1624.50438514, 1624.50438514, 1612.47101933, 1600.43765351, 1636.53775096, 1552.30419025, 1552.30419025, 1516.20409280, 1528.23745862, 1552.30419025, 1528.23745862, 1576.37092188, 1600.43765351, 1552.30419025, 1576.37092188, 1516.20409280, 1407.90380046, 1395.87043464, 1456.03726372, 1492.13736117, 1371.80370301, 1468.07062954, 1468.07062954, 1504.17072698, 1528.23745862, 1480.10399535, 1564.33755606, 1588.40428770, 1528.23745862, 1612.47101933, 1576.37092188, 1540.27082443, 1600.43765351, 1540.27082443, 1468.07062954, 1504.17072698, 1624.50438514, 1576.37092188, 1564.33755606, 1564.33755606, 1624.50438514, 1600.43765351, 1624.50438514, 1600.43765351, 1588.40428770, 1612.47101933, 1528.23745862, 1492.13736117, 1504.17072698, 1492.13736117, 1528.23745862, 1540.27082443, 1528.23745862, 1480.10399535, 1492.13736117, 1468.07062954, 1480.10399535, 1431.97053209, 1383.83706883, 1468.07062954, 1419.93716627, 1371.80370301, 1383.83706883, 1359.77033719, 1383.83706883, 1371.80370301, 1347.73697138, 1383.83706883, 1311.63687393, 1407.90380046, 1371.80370301, 1383.83706883, 1444.00389791, 1468.07062954, 1444.00389791, 1371.80370301, 1323.67023975, 1251.47004485, 1287.57014230, 1287.57014230, 1239.43667904, 1215.36994740, 1251.47004485, 1215.36994740, 1215.36994740, 1137.15306960, 1185.28653286, 1193.70988894, 1170.84649388, 1239.43667904, 1203.33658159, 1203.33658159, 1184.08319628, 1202.13324501, 1178.06651337, 1191.30321577, 1149.18643542, 1127.52637695, 1167.23648414, 1169.64315730, 1215.36994740, 1251.47004485, 1239.43667904, 1275.53677648, 1287.57014230, 1263.50341067, 1383.83706883, 1371.80370301, 1383.83706883, 1347.73697138, 1347.73697138, 1275.53677648, 1287.57014230, 1323.67023975, 1287.57014230, 1311.63687393, 1251.47004485, 1299.60350811, 1323.67023975, 1323.67023975, 1335.70360556, 1431.97053209, 1444.00389791, 1492.13736117, 1624.50438514, 1624.50438514, 1720.77131167, 1576.37092188, 1588.40428770, 1600.43765351, 1624.50438514, 1648.57111678, 1648.57111678, 1576.37092188, 1540.27082443, 1492.13736117, 1528.23745862, 1624.50438514, 1672.63784841, 1660.60448259, 1684.67121422, 1684.67121422, 1720.77131167, 1684.67121422, 1600.43765351, 1636.53775096, 1672.63784841, 1684.67121422, 1648.57111678, 1708.73794585, 1684.67121422, 1756.87140912, 1720.77131167, 1612.47101933, 1648.57111678, 1600.43765351, 1588.40428770, 1504.17072698, 1492.13736117, 1564.33755606, 1492.13736117, 1504.17072698, 1528.23745862, 1516.20409280, 1564.33755606, 1528.23745862, 1528.23745862, 1564.33755606, 1504.17072698, 1504.17072698, 1540.27082443, 1516.20409280, 1540.27082443, 1516.20409280, 1383.83706883, 1371.80370301, 1299.60350811, 1299.60350811, 1311.63687393, 1227.40331322, 1227.40331322, 1227.40331322, 1323.67023975, 1275.53677648, 1299.60350811, 1480.10399535, 1383.83706883, 1552.30419025, 1407.90380046, 1516.20409280, 1528.23745862, 1480.10399535, 1431.97053209, 1468.07062954, 1468.07062954, 1528.23745862, 1636.53775096, 1696.70458004, 1708.73794585, 1648.57111678, 1660.60448259, 1660.60448259, 1696.70458004, 1780.93814075, 1732.80467749, 1696.70458004, 1744.83804330, 1780.93814075, 1780.93814075, 1792.97150657, 1829.07160401, 1696.70458004, 1660.60448259, 1684.67121422, 1624.50438514, 1660.60448259], 
							[4909.61325288, 4813.34632635, 4765.21286309, 4729.11276564, 4813.34632635, 4945.71335033, 5090.11374012, 5041.98027685, 5270.61422735, 5174.34730083, 5390.94788551, 5162.31393501, 5126.21383756, 5198.41403246, 5198.41403246, 4945.71335033, 5005.88017941, 4945.71335033, 5029.94691104, 5029.94691104, 4933.67998451, 4632.84583911, 4789.27959472, 4705.04603401, 4572.67901003, 4897.57988706, 5054.01364267, 5078.08037430, 4945.71335033, 5150.28056920, 4993.84681359, 4777.24622890, 4753.17949727, 4885.54652125, 4909.61325288, 4849.44642380, 4717.07939982, 4596.74574167, 4620.81247330, 4813.34632635, 4873.51315543, 4945.71335033, 5054.01364267, 5186.38066664, 5126.21383756, 5029.94691104, 4885.54652125, 4945.71335033, 4933.67998451, 5029.94691104, 4885.54652125, 4813.34632635, 4789.27959472, 4717.07939982, 4945.71335033, 4825.37969217, 5090.11374012, 5198.41403246, 5824.14905488, 5848.21578652, 5655.68193346, 6016.68290794, 5884.31588396, 5715.84876254, 5691.78203091, 5475.18144622, 5643.64856765, 5427.04798296, 5246.54749572, 5174.34730083, 5150.28056920, 5234.51412991, 5619.58183601, 5378.91451970, 5415.01461715, 5318.74769062, 5270.61422735, 5366.88115388, 5390.94788551, 5619.58183601, 5535.34827530, 5234.51412991, 5511.28154367, 5487.21481204, 5487.21481204, 5547.38164112, 5896.34924978, 5920.41598141, 5956.51607886, 5884.31588396, 5824.14905488, 5571.44837275, 5932.44934723, 6257.35022426, 6630.38456455, 6498.01754057, 6630.38456455, 6919.18534413, 7099.68583137, 7051.55236810, 7015.45227066, 7556.95373237, 7568.98709819, 7256.11958697, 7665.25402471, 7436.62007421, 7532.88700074, 7568.98709819, 7340.35314769, 7135.78592882, 7232.05285534, 7159.85266045, 7111.71919718, 6931.21870995, 6859.01851505, 7075.61909974, 7123.75256300, 7280.18631861, 7544.92036656, 7665.25402471, 7629.15392727, 7364.41987932, 7232.05285534, 7256.11958697, 7159.85266045, 7159.85266045, 7063.58573392, 7099.68583137, 7183.91939208, 7099.68583137, 7015.45227066, 7340.35314769, 6955.28544158, 7075.61909974, 7051.55236810, 6931.21870995, 7099.68583137, 7183.91939208, 6991.38553903, 7015.45227066, 7015.45227066, 6919.18534413, 6594.28446710, 6209.21676099, 6522.08427221, 6353.61715078, 6485.98417476, 6726.65149108, 6798.85168597, 6774.78495434, 6485.98417476, 6064.81637120, 5776.01559162, 5559.41500694, 5739.91549417, 5294.68095899, 5138.24720338, 5354.84778807, 5054.01364267, 4909.61325288, 4897.57988706, 5234.51412991, 4969.78008196, 4945.71335033, 5174.34730083, 5102.14710593, 5174.34730083, 4945.71335033, 5054.01364267, 4933.67998451, 5090.11374012, 4885.54652125, 4632.84583911, 4969.78008196, 4897.57988706, 5198.41403246, 5715.84876254, 5751.94885999, 5776.01559162, 6365.65051660, 6353.61715078, 6522.08427221, 6221.25012681, 6401.75061405, 6389.71724823, 6437.85071149, 6317.51705334, 6185.15002936, 6401.75061405, 6088.88310283, 6257.35022426, 5715.84876254, 5860.24915233, 5896.34924978, 6064.81637120, 6245.31685844, 6618.35119873, 6955.28544158, 6822.91841760, 7220.01948953, 7448.65344003, 8158.62202317, 7761.52095124, 7833.72114614, 7581.02046400, 7821.68778032, 7761.52095124, 7821.68778032, 7556.95373237, 7364.41987932, 7520.85363492, 7544.92036656, 7665.25402471, 7689.32075635, 7749.48758543, 7701.35412216, 8014.22163337, 8303.02241296, 8146.58865735, 8038.28836501, 8278.95568132, 8254.88894969, 8266.92231551, 8375.22260785, 8579.78982672, 8543.68972927, 8676.05675325, 8639.95665580, 8543.68972927, 8218.78885224, 8218.78885224, 8327.08914459, 8206.75548643, 8303.02241296, 8411.32270530, 8206.75548643, 8206.75548643, 8363.18924204, 8375.22260785, 8423.35607111, 8423.35607111, 8411.32270530, 8495.55626601, 8254.88894969, 8134.55529153, 8290.98904714, 8363.18924204, 8567.75646091, 8772.32367977, 8988.92426446, 9025.02436191, 8928.75743538, 9109.25792262, 9133.32465425, 8808.42377722, 9217.55821496, 9337.89187312, 9843.29323739, 10180.22748023, 10926.29616082, 10529.19508889, 10481.06162563, 10553.26182052, 10637.49538124, 10348.69460166, 10396.82806492, 10757.82903939, 10890.19606337, 10649.52874705, 10998.49635571, 11263.23040366, 11455.76425672, 11816.76523119, 11648.29810977, 11804.73186538, 11973.19898680, 11888.96542609, 11828.79859701, 12021.33245006, 11648.29810977, 11876.93206027, 12875.70142299, 12996.03508115, 14680.70629537, 14319.70532089, 14319.70532089, 13236.70239747, 12635.03410667, 13718.03703010, 12996.03508115, 13116.36873931], 
							[703.95190023, 669.05513936, 659.42844671, 624.53168584, 649.80175406, 590.83826156, 567.97486651, 565.56819335, 537.89145197, 531.87476906, 577.60155916, 580.00823233, 577.60155916, 577.60155916, 569.17820309, 535.48477881, 511.41804717, 493.36799845, 521.04473983, 525.85808615, 527.06142274, 505.40136427, 510.21471059, 495.77467161, 459.67457417, 472.91127656, 527.06142274, 570.38153967, 559.55151044, 576.39822258, 577.60155916, 537.89145197, 529.46809590, 553.53482753, 540.29812513, 541.50146171, 529.46809590, 523.45141299, 539.09478855, 541.50146171, 535.48477881, 551.12815437, 552.33149095, 540.29812513, 529.46809590, 505.40136427, 505.40136427, 510.21471059, 539.09478855, 537.89145197, 536.68811539, 543.90813488, 558.34817386, 536.68811539, 560.75484702, 548.72148120, 575.19488600, 649.80175406, 649.80175406, 661.83511987, 632.95504192, 673.86848569, 661.83511987, 695.52854416, 743.66200742, 760.50871956, 755.69537324, 735.23865135, 682.29184176, 688.30852467, 697.93521732, 717.18860263, 728.01863186, 703.95190023, 754.49203666, 747.27201717, 760.50871956, 747.27201717, 748.47535375, 761.71205615, 747.27201717, 731.62864161, 719.59527579, 697.93521732, 703.95190023, 697.93521732, 696.73188074, 652.20842722, 634.15837850, 596.85494447, 541.50146171, 512.62138376, 541.50146171, 505.40136427, 558.34817386, 525.85808615, 535.48477881, 541.50146171, 580.00823233, 611.29498345, 583.61824207, 618.51500294, 616.10832977, 580.00823233, 590.83826156, 601.66829079, 640.17506140, 691.91853441, 715.98526604, 679.88516860, 673.86848569, 697.93521732, 676.27515885, 657.02177355, 652.20842722, 651.00509064, 665.44512962, 682.29184176, 696.73188074, 701.54522707, 684.69851492, 672.66514911, 663.03845645, 666.64846620, 669.05513936, 652.20842722, 655.81843697, 658.22511013, 659.42844671, 701.54522707, 687.10518809, 681.08850518, 657.02177355, 643.78507115, 644.98840773, 618.51500294, 612.49832003, 640.17506140, 625.73502243, 635.36171508, 630.54836875, 631.75170533, 635.36171508, 617.31166635, 620.92167610, 625.73502243, 626.93835901, 641.37839799, 644.98840773, 642.58173457, 629.34503217, 667.85180278, 690.71519783, 666.64846620, 643.78507115, 635.36171508, 620.92167610, 573.99154942, 552.33149095, 554.73816411, 528.26475932, 561.95818360, 536.68811539, 557.14483728, 565.56819335, 549.92481779, 566.77152993, 518.63806666, 529.46809590, 513.82472034, 515.02805692, 452.45455468, 444.03119861, 448.84454493, 456.06456442, 512.62138376, 499.38468136, 476.52128631, 498.18134478, 496.97800820, 512.62138376, 543.90813488, 543.90813488, 539.09478855, 523.45141299, 540.29812513, 502.99469110, 511.41804717, 541.50146171, 511.41804717, 543.90813488, 499.38468136, 517.43473008, 493.36799845, 513.82472034, 523.45141299, 541.50146171, 534.28144222, 539.09478855, 535.48477881, 540.29812513, 577.60155916, 567.97486651, 571.58487625, 577.60155916, 620.92167610, 618.51500294, 600.46495421, 572.78821284, 575.19488600, 611.29498345, 611.29498345, 613.70165661, 638.97172482, 643.78507115, 664.24179304, 641.37839799, 641.37839799, 632.95504192, 651.00509064, 644.98840773, 631.75170533, 622.12501268, 614.90499319, 631.75170533, 658.22511013, 646.19174431, 642.58173457, 641.37839799, 631.75170533, 602.87162738, 581.21156891, 600.46495421, 589.63492498, 592.04159814, 561.95818360, 557.14483728, 593.24493472, 590.83826156, 598.05828105, 593.24493472, 582.41490549, 589.63492498, 578.80489574, 575.19488600, 583.61824207, 581.21156891, 565.56819335, 564.36485676, 547.51814462, 547.51814462, 500.58801794, 493.36799845, 494.57133503, 409.13443774, 439.21785228, 421.16780356, 487.35131554, 445.23453519, 448.84454493, 499.38468136, 463.28458391, 448.84454493, 456.06456442, 477.72462289, 463.28458391, 452.45455468, 444.03119861, 475.31794973, 468.09793024, 481.33463264, 487.35131554, 539.09478855, 543.90813488, 518.63806666, 543.90813488, 534.28144222, 530.67143248, 501.79135452, 529.46809590, 498.18134478, 472.91127656, 481.33463264, 433.20116937, 388.67771585, 370.62766713, 354.98429157, 371.83100371, 352.57761841, 340.54425259, 327.30755019], 
							[2334.47296828, 2250.23940757, 2202.10594431, 2081.77228615, 1323.67023975, 782.16877803, 647.39508089, 801.42216334, 709.96858314, 715.98526604, 753.28870007, 752.08536349, 812.25219257, 780.96544145, 748.47535375, 749.67869033, 687.10518809, 589.63492498, 636.56505166, 601.66829079, 584.82157865, 541.50146171, 481.33463264, 423.57447672, 379.05102320, 487.35131554, 589.63492498, 644.98840773, 640.17506140, 611.29498345, 598.05828105, 541.50146171, 547.51814462, 539.09478855, 549.92481779, 512.62138376, 494.57133503, 465.69125707, 484.94464238, 499.38468136, 605.27830054, 589.63492498, 565.56819335, 553.53482753, 529.46809590, 511.41804717, 495.77467161, 489.75798871, 518.63806666, 536.68811539, 525.85808615, 540.29812513, 504.19802769, 498.18134478, 523.45141299, 530.67143248, 528.26475932, 529.46809590, 606.48163712, 637.76838824, 602.87162738, 637.76838824, 642.58173457, 637.76838824, 628.14169559, 670.25847594, 717.18860263, 736.44198793, 673.86848569, 649.80175406, 649.80175406, 653.41176380, 652.20842722, 616.10832977, 636.56505166, 643.78507115, 649.80175406, 641.37839799, 640.17506140, 644.98840773, 636.56505166, 607.68497370, 589.63492498, 577.60155916, 571.58487625, 541.50146171, 531.87476906, 494.57133503, 435.60784253, 456.06456442, 423.57447672, 373.03434029, 425.98114988, 412.74444748, 469.30126682, 427.18448646, 444.03119861, 463.28458391, 505.40136427, 509.01137401, 469.30126682, 525.85808615, 517.43473008, 466.89459366, 488.55465212, 484.94464238, 496.97800820, 584.82157865, 565.56819335, 547.51814462, 552.33149095, 551.12815437, 543.90813488, 531.87476906, 535.48477881, 531.87476906, 534.28144222, 566.77152993, 571.58487625, 594.44827130, 612.49832003, 577.60155916, 575.19488600, 553.53482753, 578.80489574, 561.95818360, 555.94150069, 541.50146171, 552.33149095, 530.67143248, 535.48477881, 531.87476906, 505.40136427, 476.52128631, 486.14797896, 475.31794973, 484.94464238, 493.36799845, 487.35131554, 481.33463264, 482.53796922, 453.65789126, 456.06456442, 421.16780356, 433.20116937, 433.20116937, 425.98114988, 459.67457417, 470.50460340, 453.65789126, 456.06456442, 442.82786202, 425.98114988, 445.23453519, 478.92795947, 451.25121810, 433.20116937, 427.18448646, 405.52442800, 403.11775483, 379.05102320, 375.44101346, 381.45769636, 395.89773534, 417.55779381, 403.11775483, 411.54111090, 401.91441825, 385.06770611, 367.01765738, 382.66103294, 354.98429157, 326.10421361, 342.95092575, 312.86751121, 407.93110116, 422.37114014, 417.55779381, 409.13443774, 389.88105243, 398.30440851, 468.09793024, 439.21785228, 457.26790100, 511.41804717, 512.62138376, 480.13129605, 471.70793998, 469.30126682, 457.26790100, 488.55465212, 458.47123758, 482.53796922, 477.72462289, 518.63806666, 505.40136427, 507.80803743, 507.80803743, 616.10832977, 565.56819335, 587.22825181, 576.39822258, 566.77152993, 565.56819335, 577.60155916, 616.10832977, 596.85494447, 584.82157865, 560.75484702, 539.09478855, 567.97486651, 559.55151044, 557.14483728, 589.63492498, 610.09164687, 608.88831028, 614.90499319, 617.31166635, 622.12501268, 613.70165661, 604.07496396, 599.26161763, 587.22825181, 577.60155916, 610.09164687, 625.73502243, 606.48163712, 596.85494447, 588.43158840, 578.80489574, 571.58487625, 534.28144222, 509.01137401, 560.75484702, 554.73816411, 533.07810564, 525.85808615, 548.72148120, 572.78821284, 575.19488600, 558.34817386, 525.85808615, 560.75484702, 517.43473008, 516.23139350, 535.48477881, 535.48477881, 525.85808615, 515.02805692, 474.11461315, 493.36799845, 466.89459366, 436.81117912, 433.20116937, 375.44101346, 400.71108167, 398.30440851, 407.93110116, 387.47437927, 400.71108167, 407.93110116, 403.11775483, 399.50774509, 385.06770611, 403.11775483, 385.06770611, 354.98429157, 347.76427208, 381.45769636, 368.22099397, 385.06770611, 415.15112065, 454.86122784, 433.20116937, 469.30126682, 463.28458391, 458.47123758, 444.03119861, 442.82786202, 416.35445723, 403.11775483, 409.13443774, 421.16780356, 379.05102320, 357.39096473, 348.96760866, 339.34091601, 330.91755994, 347.76427208, 324.90087703, 308.05416489], 
							[1516.20409280, 1480.10399535, 1395.87043464, 1431.97053209, 1456.03726372, 1299.60350811, 1263.50341067, 1275.53677648, 1323.67023975, 1299.60350811, 1323.67023975, 1323.67023975, 1299.60350811, 1239.43667904, 1215.36994740, 1155.20311832, 1093.83295266, 978.31264083, 1058.93619180, 1050.51283573, 1038.47946991, 944.61921655, 932.58585073, 926.56916782, 915.73913859, 1058.93619180, 1215.36994740, 1173.25316705, 1144.37308909, 1125.11970378, 1113.08633797, 1016.81941144, 1043.29281624, 1038.47946991, 1028.85277726, 968.68594818, 1004.78604563, 980.71931399, 998.76936272, 1019.22608460, 1013.20940170, 977.10930425, 1010.80272853, 1001.17603588, 969.88928476, 934.99252389, 921.75582150, 866.40233874, 904.90910935, 902.50243619, 889.26573379, 918.14581175, 907.31578252, 863.99566558, 928.97584099, 967.48261160, 972.29595792, 1032.46278700, 1093.83295266, 1121.50969404, 1083.00292343, 1109.47632822, 1125.11970378, 1095.03628924, 1113.08633797, 1048.10616256, 1063.74953812, 1049.30949914, 991.54934323, 975.90596767, 968.68594818, 975.90596767, 991.54934323, 924.16249466, 950.63589945, 956.65258236, 985.53266032, 962.66926527, 981.92265058, 998.76936272, 997.56602614, 1013.20940170, 997.56602614, 998.76936272, 990.34600665, 1034.86946017, 1034.86946017, 1018.02274802, 1014.41273828, 1022.83609435, 1016.81941144, 1009.59939195, 1028.85277726, 998.76936272, 1010.80272853, 1002.37937246, 995.15935297, 1018.02274802, 1056.52951863, 1125.11970378, 1098.64629899, 1163.62647440, 1145.57642567, 1097.44296241, 1152.79644516, 1173.25316705, 1157.60979149, 1113.08633797, 1060.13952838, 1070.96955761, 1115.49301113, 1113.08633797, 1119.10302088, 1108.27299164, 1110.67966481, 1128.72971353, 1139.55974276, 1119.10302088, 1156.40645491, 1155.20311832, 1196.11656210, 1175.65984021, 1202.13324501, 1168.43982072, 1180.47318654, 1158.81312807, 1169.64315730, 1153.99978174, 1157.60979149, 1169.64315730, 1191.30321577, 1239.43667904, 1184.08319628, 1179.26984996, 1215.36994740, 1180.47318654, 1176.86317679, 1191.30321577, 1146.77976225, 1110.67966481, 1052.91950889, 1040.88614307, 1058.93619180, 1007.19271879, 1003.58270904, 998.76936272, 998.76936272, 1004.78604563, 1027.64944068, 1027.64944068, 1031.25945042, 1049.30949914, 1070.96955761, 1040.88614307, 1064.95287471, 1063.74953812, 1038.47946991, 1052.91950889, 1040.88614307, 1019.22608460, 974.70263109, 1012.00606512, 966.27927501, 1003.58270904, 1007.19271879, 1010.80272853, 1032.46278700, 1040.88614307, 1031.25945042, 990.34600665, 1015.61607486, 995.15935297, 986.73599690, 1026.44610409, 1013.20940170, 1089.01960634, 1079.39291368, 1075.78290394, 1075.78290394, 1086.61293317, 1105.86631848, 1179.26984996, 1146.77976225, 1188.89654261, 1227.40331322, 1251.47004485, 1179.26984996, 1174.45650363, 1190.09987919, 1176.86317679, 1181.67652312, 1120.30635746, 1153.99978174, 1151.59310858, 1160.01646465, 1202.13324501, 1176.86317679, 1203.33658159, 1215.36994740, 1227.40331322, 1227.40331322, 1311.63687393, 1251.47004485, 1299.60350811, 1323.67023975, 1383.83706883, 1395.87043464, 1335.70360556, 1287.57014230, 1275.53677648, 1239.43667904, 1239.43667904, 1299.60350811, 1383.83706883, 1431.97053209, 1395.87043464, 1419.93716627, 1395.87043464, 1407.90380046, 1371.80370301, 1383.83706883, 1311.63687393, 1359.77033719, 1335.70360556, 1299.60350811, 1347.73697138, 1359.77033719, 1299.60350811, 1263.50341067, 1263.50341067, 1251.47004485, 1227.40331322, 1227.40331322, 1215.36994740, 1215.36994740, 1172.04983047, 1167.23648414, 1191.30321577, 1181.67652312, 1227.40331322, 1199.72657184, 1203.33658159, 1275.53677648, 1203.33658159, 1162.42313781, 1194.91322552, 1145.57642567, 1150.38977200, 1095.03628924, 1068.56288445, 1020.42942119, 985.53266032, 962.66926527, 996.36268955, 811.04885599, 892.87574354, 866.40233874, 902.50243619, 890.46907037, 848.35229002, 851.96229976, 794.20214385, 758.10204640, 673.86848569, 724.40862212, 718.39193921, 644.98840773, 588.43158840, 637.76838824, 657.02177355, 722.00194895, 782.16877803, 847.14895344, 794.20214385, 783.37211461, 782.16877803, 754.49203666, 807.43884625, 794.20214385, 791.79547068, 736.44198793, 753.28870007, 755.69537324, 673.86848569, 673.86848569, 734.03531477, 770.13541222, 762.91539273, 742.45867084, 701.54522707, 676.27515885], 
							[1051.71617231, 997.56602614, 931.38251415, 839.92893395, 942.21254338, 786.98212436, 778.55876829, 740.05199768, 791.79547068, 754.49203666, 823.08222181, 878.43570456, 806.23550966, 870.01234849, 806.23550966, 709.96858314, 673.86848569, 595.65160789, 637.76838824, 664.24179304, 643.78507115, 595.65160789, 679.88516860, 658.22511013, 613.70165661, 679.88516860, 764.11872931, 831.50557788, 759.30538298, 776.15209512, 788.18546094, 759.30538298, 759.30538298, 776.15209512, 759.30538298, 699.13855390, 766.52540247, 742.45867084, 835.11558762, 807.43884625, 806.23550966, 765.32206589, 779.76210487, 779.76210487, 734.03531477, 770.13541222, 778.55876829, 740.05199768, 770.13541222, 817.06553890, 809.84551941, 818.26887548, 848.35229002, 839.92893395, 1005.98938221, 1107.06965506, 1086.61293317, 1131.13638669, 1030.05611384, 1060.13952838, 1070.96955761, 1128.72971353, 1157.60979149, 1153.99978174, 1163.62647440, 1178.06651337, 1157.60979149, 1132.33972327, 1108.27299164, 1107.06965506, 1120.30635746, 1098.64629899, 1192.50655235, 1141.96641593, 1251.47004485, 1299.60350811, 1299.60350811, 1311.63687393, 1311.63687393, 1335.70360556, 1275.53677648, 1203.33658159, 1239.43667904, 1299.60350811, 1287.57014230, 1193.70988894, 1275.53677648, 1227.40331322, 1157.60979149, 1164.82981098, 1162.42313781, 979.51597741, 1131.13638669, 1095.03628924, 1178.06651337, 1131.13638669, 1110.67966481, 1135.94973302, 1239.43667904, 1239.43667904, 1227.40331322, 1287.57014230, 1227.40331322, 1152.79644516, 1185.28653286, 1239.43667904, 1251.47004485, 1239.43667904, 1227.40331322, 1173.25316705, 1203.33658159, 1202.13324501, 1176.86317679, 1155.20311832, 1143.16975251, 1186.48986945, 1133.54305986, 1188.89654261, 1175.65984021, 1239.43667904, 1185.28653286, 1146.77976225, 1129.93305011, 1128.72971353, 1138.35640618, 1116.69634771, 1178.06651337, 1115.49301113, 1134.74639644, 1113.08633797, 1143.16975251, 1119.10302088, 1092.62961608, 1096.23962583, 1083.00292343, 1009.59939195, 968.68594818, 974.70263109, 930.17917757, 904.90910935, 938.60253364, 907.31578252, 934.99252389, 907.31578252, 956.65258236, 904.90910935, 955.44924578, 950.63589945, 926.56916782, 931.38251415, 950.63589945, 960.26259211, 901.29909961, 925.36583124, 914.53580201, 933.78918731, 878.43570456, 829.09890471, 748.47535375, 767.72873905, 756.89870982, 734.03531477, 746.06868058, 818.26887548, 839.92893395, 801.42216334, 830.30224130, 784.57545120, 797.81215359, 725.61195870, 742.45867084, 684.69851492, 685.90185151, 703.95190023, 654.61510038, 758.10204640, 755.69537324, 767.72873905, 722.00194895, 703.95190023, 735.23865135, 758.10204640, 744.86534400, 756.89870982, 744.86534400, 758.10204640, 756.89870982, 746.06868058, 768.93207563, 756.89870982, 788.18546094, 709.96858314, 758.10204640, 758.10204640, 764.11872931, 797.81215359, 829.09890471, 936.19586048, 932.58585073, 947.02588971, 938.60253364, 1040.88614307, 971.09262134, 975.90596767, 985.53266032, 957.85591894, 934.99252389, 914.53580201, 878.43570456, 873.62235823, 926.56916782, 868.80901191, 912.12912884, 992.75267981, 932.58585073, 950.63589945, 966.27927501, 944.61921655, 928.97584099, 897.68908986, 919.34914833, 944.61921655, 962.66926527, 919.34914833, 956.65258236, 947.02588971, 938.60253364, 966.27927501, 928.97584099, 914.53580201, 872.41902165, 847.14895344, 824.28555839, 856.77564609, 870.01234849, 835.11558762, 806.23550966, 821.87888522, 908.51911910, 924.16249466, 872.41902165, 850.75896318, 896.48575328, 860.38565584, 848.35229002, 853.16563635, 860.38565584, 884.45238747, 857.97898267, 794.20214385, 818.26887548, 819.47221206, 774.94875854, 782.16877803, 728.01863186, 693.12187099, 610.09164687, 646.19174431, 590.83826156, 602.87162738, 638.97172482, 613.70165661, 582.41490549, 549.92481779, 560.75484702, 528.26475932, 494.57133503, 406.72776458, 453.65789126, 424.77781330, 453.65789126, 468.09793024, 501.79135452, 488.55465212, 483.74130580, 480.13129605, 481.33463264, 493.36799845, 501.79135452, 476.52128631, 469.30126682, 434.40450595, 445.23453519, 376.64435004, 389.88105243, 365.81432080, 362.20431106, 351.37428182, 363.40764764, 338.13757943, 347.76427208], 
							[1107.06965506, 1083.00292343, 1096.23962583, 1042.08947966, 1004.78604563, 1031.25945042, 902.50243619, 836.31892420, 851.96229976, 811.04885599, 896.48575328, 845.94561686, 862.79232900, 895.28241670, 837.52226079, 827.89556813, 758.10204640, 707.56190997, 755.69537324, 761.71205615, 735.23865135, 691.91853441, 700.34189048, 715.98526604, 678.68183202, 709.96858314, 801.42216334, 825.48889497, 788.18546094, 819.47221206, 841.13227053, 760.50871956, 720.79861237, 746.06868058, 747.27201717, 737.64532451, 773.74542196, 738.84866109, 777.35543171, 808.64218283, 796.60881701, 760.50871956, 772.54208538, 741.25533426, 707.56190997, 663.03845645, 642.58173457, 675.07182227, 691.91853441, 730.42530502, 688.30852467, 718.39193921, 744.86534400, 759.30538298, 837.52226079, 818.26887548, 812.25219257, 794.20214385, 788.18546094, 845.94561686, 824.28555839, 866.40233874, 866.40233874, 870.01234849, 916.94247517, 872.41902165, 924.16249466, 865.19900216, 877.23236798, 843.53894369, 847.14895344, 842.33560711, 861.58899242, 842.33560711, 862.79232900, 847.14895344, 842.33560711, 848.35229002, 862.79232900, 854.36897293, 844.74228027, 837.52226079, 818.26887548, 817.06553890, 783.37211461, 773.74542196, 758.10204640, 682.29184176, 604.07496396, 601.66829079, 555.94150069, 499.38468136, 565.56819335, 535.48477881, 571.58487625, 518.63806666, 535.48477881, 588.43158840, 637.76838824, 628.14169559, 655.81843697, 743.66200742, 851.96229976, 766.52540247, 813.45552915, 817.06553890, 785.77878778, 797.81215359, 764.11872931, 729.22196844, 779.76210487, 800.21882676, 770.13541222, 748.47535375, 747.27201717, 782.16877803, 782.16877803, 815.86220232, 800.21882676, 843.53894369, 891.67240696, 837.52226079, 841.13227053, 841.13227053, 848.35229002, 831.50557788, 821.87888522, 825.48889497, 842.33560711, 796.60881701, 826.69223155, 831.50557788, 789.38879752, 792.99880727, 779.76210487, 734.03531477, 734.03531477, 758.10204640, 778.55876829, 782.16877803, 788.18546094, 737.64532451, 720.79861237, 678.68183202, 693.12187099, 697.93521732, 685.90185151, 766.52540247, 794.20214385, 788.18546094, 765.32206589, 740.05199768, 673.86848569, 720.79861237, 722.00194895, 720.79861237, 708.76524656, 643.78507115, 589.63492498, 601.66829079, 587.22825181, 569.17820309, 545.11147146, 577.60155916, 654.61510038, 637.76838824, 661.83511987, 608.88831028, 570.38153967, 560.75484702, 589.63492498, 529.46809590, 476.52128631, 511.41804717, 458.47123758, 586.02491523, 601.66829079, 613.70165661, 608.88831028, 583.61824207, 578.80489574, 643.78507115, 620.92167610, 649.80175406, 642.58173457, 649.80175406, 640.17506140, 638.97172482, 664.24179304, 638.97172482, 642.58173457, 607.68497370, 636.56505166, 623.32834926, 602.87162738, 599.26161763, 794.20214385, 748.47535375, 725.61195870, 713.57859288, 713.57859288, 720.79861237, 715.98526604, 732.83197819, 758.10204640, 728.01863186, 736.44198793, 705.15523681, 701.54522707, 722.00194895, 734.03531477, 740.05199768, 723.20528553, 792.99880727, 805.03217308, 824.28555839, 878.43570456, 861.58899242, 833.91225104, 854.36897293, 866.40233874, 890.46907037, 830.30224130, 826.69223155, 860.38565584, 873.62235823, 878.43570456, 854.36897293, 823.08222181, 845.94561686, 826.69223155, 778.55876829, 752.08536349, 768.93207563, 809.84551941, 740.05199768, 723.20528553, 806.23550966, 824.28555839, 830.30224130, 791.79547068, 785.77878778, 782.16877803, 748.47535375, 734.03531477, 776.15209512, 746.06868058, 744.86534400, 737.64532451, 665.44512962, 691.91853441, 604.07496396, 608.88831028, 601.66829079, 482.53796922, 531.87476906, 518.63806666, 547.51814462, 518.63806666, 519.84140325, 531.87476906, 529.46809590, 523.45141299, 498.18134478, 521.04473983, 487.35131554, 430.79449621, 439.21785228, 472.91127656, 436.81117912, 475.31794973, 490.96132529, 511.41804717, 481.33463264, 519.84140325, 509.01137401, 496.97800820, 499.38468136, 499.38468136, 464.48792049, 453.65789126, 439.21785228, 439.21785228, 392.28772560, 376.64435004, 345.35759892, 306.85082830, 300.83414540, 324.90087703, 296.02079907, 288.80077958], 
							[1648.57111678, 1696.70458004, 1552.30419025, 1419.93716627, 1444.00389791, 1239.43667904, 1108.27299164, 1190.09987919, 1119.10302088, 1125.11970378, 1141.96641593, 1179.26984996, 1191.30321577, 1168.43982072, 1107.06965506, 1022.83609435, 985.53266032, 884.45238747, 1022.83609435, 986.73599690, 944.61921655, 856.77564609, 956.65258236, 914.53580201, 960.26259211, 1156.40645491, 1275.53677648, 1456.03726372, 1335.70360556, 1275.53677648, 1347.73697138, 1197.31989868, 1173.25316705, 1215.36994740, 1153.99978174, 1143.16975251, 1191.30321577, 1169.64315730, 1161.21980123, 1168.43982072, 1192.50655235, 1178.06651337, 1215.36994740, 1168.43982072, 1137.15306960, 1104.66298190, 1097.44296241, 1103.45964532, 1149.18643542, 1160.01646465, 1141.96641593, 1138.35640618, 1054.12284547, 1019.22608460, 1108.27299164, 1335.70360556, 1311.63687393, 1468.07062954, 1552.30419025, 1444.00389791, 1335.70360556, 1528.23745862, 1492.13736117, 1540.27082443, 1492.13736117, 1600.43765351, 1648.57111678, 1576.37092188, 1480.10399535, 1540.27082443, 1468.07062954, 1504.17072698, 1612.47101933, 1528.23745862, 1612.47101933, 1708.73794585, 1853.13833565, 1708.73794585, 1732.80467749, 1853.13833565, 1780.93814075, 1756.87140912, 1672.63784841, 1684.67121422, 1768.90477493, 1756.87140912, 1768.90477493, 1780.93814075, 1732.80467749, 1696.70458004, 1696.70458004, 1419.93716627, 1540.27082443, 1456.03726372, 1468.07062954, 1371.80370301, 1444.00389791, 1456.03726372, 1516.20409280, 1684.67121422, 1636.53775096, 1708.73794585, 1756.87140912, 1636.53775096, 1648.57111678, 1865.17170146, 1865.17170146, 1889.23843309, 1865.17170146, 1780.93814075, 1913.30516472, 1889.23843309, 1817.03823820, 1805.00487238, 1877.20506728, 1925.33853054, 1913.30516472, 1949.40526217, 2069.73892033, 2178.03921267, 2045.67218870, 1997.53872544, 1937.37189636, 1913.30516472, 1925.33853054, 1949.40526217, 1925.33853054, 1901.27179891, 1817.03823820, 1865.17170146, 1792.97150657, 1768.90477493, 1660.60448259, 1792.97150657, 1805.00487238, 1792.97150657, 1768.90477493, 1853.13833565, 1817.03823820, 1792.97150657, 1792.97150657, 1624.50438514, 1588.40428770, 1552.30419025, 1684.67121422, 1660.60448259, 1732.80467749, 1841.10496983, 1792.97150657, 1768.90477493, 1684.67121422, 1528.23745862, 1516.20409280, 1684.67121422, 1648.57111678, 1696.70458004, 1672.63784841, 1624.50438514, 1624.50438514, 1552.30419025, 1395.87043464, 1528.23745862, 1480.10399535, 1540.27082443, 1732.80467749, 1744.83804330, 1732.80467749, 1768.90477493, 1576.37092188, 1624.50438514, 1708.73794585, 1648.57111678, 1648.57111678, 1648.57111678, 1444.00389791, 1684.67121422, 1696.70458004, 1768.90477493, 1877.20506728, 1841.10496983, 1805.00487238, 1949.40526217, 1937.37189636, 1961.43862799, 1925.33853054, 1985.50535962, 1768.90477493, 1865.17170146, 1925.33853054, 1780.93814075, 1792.97150657, 1612.47101933, 1792.97150657, 1768.90477493, 1805.00487238, 1829.07160401, 1889.23843309, 2033.63882288, 2069.73892033, 2190.07257849, 2069.73892033, 2166.00584686, 2129.90574941, 2141.93911523, 2129.90574941, 1937.37189636, 2045.67218870, 2057.70555452, 1985.50535962, 1997.53872544, 1913.30516472, 1961.43862799, 1997.53872544, 2045.67218870, 2081.77228615, 2129.90574941, 2129.90574941, 2190.07257849, 2141.93911523, 2045.67218870, 2153.97248104, 2141.93911523, 2166.00584686, 2117.87238359, 2226.17267594, 2190.07257849, 2226.17267594, 2214.13931012, 2166.00584686, 2153.97248104, 2166.00584686, 2105.83901778, 2009.57209125, 1997.53872544, 2033.63882288, 1961.43862799, 1865.17170146, 1949.40526217, 2021.60545707, 2069.73892033, 2033.63882288, 2057.70555452, 2117.87238359, 2009.57209125, 2045.67218870, 2105.83901778, 2117.87238359, 2153.97248104, 2202.10594431, 2069.73892033, 2069.73892033, 1937.37189636, 1925.33853054, 1865.17170146, 1552.30419025, 1660.60448259, 1504.17072698, 1504.17072698, 1419.93716627, 1564.33755606, 1732.80467749, 1600.43765351, 1576.37092188, 1588.40428770, 1648.57111678, 1528.23745862, 1395.87043464, 1323.67023975, 1540.27082443, 1552.30419025, 1756.87140912, 1805.00487238, 1949.40526217, 1901.27179891, 2009.57209125, 1889.23843309, 2105.83901778, 2081.77228615, 2141.93911523, 2057.70555452, 1985.50535962, 2166.00584686, 2153.97248104, 1997.53872544, 2093.80565196, 2141.93911523, 2033.63882288, 2033.63882288, 2021.60545707, 2045.67218870, 2141.93911523], 
							[848.35229002, 854.36897293, 884.45238747, 854.36897293, 851.96229976, 803.82883650, 805.03217308, 776.15209512, 815.86220232, 786.98212436, 851.96229976, 818.26887548, 836.31892420, 830.30224130, 812.25219257, 795.40548043, 776.15209512, 754.49203666, 770.13541222, 737.64532451, 744.86534400, 718.39193921, 741.25533426, 713.57859288, 681.08850518, 722.00194895, 748.47535375, 784.57545120, 777.35543171, 776.15209512, 779.76210487, 731.62864161, 709.96858314, 707.56190997, 689.51186125, 667.85180278, 638.97172482, 613.70165661, 637.76838824, 643.78507115, 651.00509064, 683.49517834, 725.61195870, 701.54522707, 691.91853441, 669.05513936, 676.27515885, 684.69851492, 731.62864161, 732.83197819, 715.98526604, 723.20528553, 738.84866109, 746.06868058, 776.15209512, 770.13541222, 766.52540247, 797.81215359, 874.82569481, 890.46907037, 866.40233874, 950.63589945, 959.05925553, 941.00920680, 942.21254338, 934.99252389, 936.19586048, 902.50243619, 877.23236798, 859.18231925, 863.99566558, 882.04571430, 906.11244594, 863.99566558, 914.53580201, 890.46907037, 908.51911910, 896.48575328, 880.84237772, 891.67240696, 850.75896318, 791.79547068, 770.13541222, 773.74542196, 818.26887548, 826.69223155, 845.94561686, 803.82883650, 676.27515885, 703.95190023, 651.00509064, 626.93835901, 637.76838824, 595.65160789, 631.75170533, 582.41490549, 594.44827130, 619.71833952, 664.24179304, 683.49517834, 648.59841748, 731.62864161, 732.83197819, 718.39193921, 784.57545120, 795.40548043, 773.74542196, 771.33874880, 725.61195870, 719.59527579, 770.13541222, 784.57545120, 785.77878778, 758.10204640, 791.79547068, 789.38879752, 790.59213410, 830.30224130, 878.43570456, 885.65572405, 902.50243619, 891.67240696, 860.38565584, 866.40233874, 892.87574354, 878.43570456, 918.14581175, 886.85906063, 907.31578252, 927.77250440, 902.50243619, 900.09576303, 860.38565584, 880.84237772, 877.23236798, 886.85906063, 868.80901191, 849.55562660, 866.40233874, 859.18231925, 836.31892420, 806.23550966, 785.77878778, 794.20214385, 788.18546094, 773.74542196, 770.13541222, 774.94875854, 799.01549017, 799.01549017, 799.01549017, 722.00194895, 679.88516860, 697.93521732, 711.17191972, 717.18860263, 655.81843697, 629.34503217, 578.80489574, 559.55151044, 547.51814462, 548.72148120, 543.90813488, 560.75484702, 595.65160789, 589.63492498, 607.68497370, 583.61824207, 546.31480804, 555.94150069, 547.51814462, 517.43473008, 487.35131554, 517.43473008, 463.28458391, 547.51814462, 534.28144222, 571.58487625, 565.56819335, 593.24493472, 607.68497370, 695.52854416, 749.67869033, 732.83197819, 693.12187099, 714.78192946, 670.25847594, 709.96858314, 720.79861237, 683.49517834, 658.22511013, 636.56505166, 654.61510038, 685.90185151, 691.91853441, 706.35857339, 719.59527579, 709.96858314, 699.13855390, 708.76524656, 712.37525630, 726.81529528, 773.74542196, 776.15209512, 776.15209512, 815.86220232, 774.94875854, 776.15209512, 734.03531477, 736.44198793, 728.01863186, 720.79861237, 756.89870982, 782.16877803, 801.42216334, 797.81215359, 791.79547068, 801.42216334, 808.64218283, 762.91539273, 755.69537324, 729.22196844, 752.08536349, 755.69537324, 753.28870007, 734.03531477, 748.47535375, 738.84866109, 696.73188074, 701.54522707, 695.52854416, 697.93521732, 679.88516860, 696.73188074, 703.95190023, 659.42844671, 663.03845645, 699.13855390, 713.57859288, 737.64532451, 693.12187099, 666.64846620, 707.56190997, 653.41176380, 673.86848569, 672.66514911, 664.24179304, 649.80175406, 629.34503217, 584.82157865, 625.73502243, 587.22825181, 565.56819335, 572.78821284, 477.72462289, 530.67143248, 499.38468136, 540.29812513, 511.41804717, 523.45141299, 535.48477881, 516.23139350, 471.70793998, 457.26790100, 476.52128631, 477.72462289, 444.03119861, 439.21785228, 459.67457417, 475.31794973, 516.23139350, 523.45141299, 573.99154942, 542.70479830, 547.51814462, 546.31480804, 537.89145197, 512.62138376, 505.40136427, 478.92795947, 488.55465212, 459.67457417, 458.47123758, 421.16780356, 413.94778407, 409.13443774, 386.27104269, 387.47437927, 389.88105243, 364.61098422, 342.95092575], 
							[2178.03921267, 2166.00584686, 2081.77228615, 2153.97248104, 2033.63882288, 1708.73794585, 1612.47101933, 1756.87140912, 1672.63784841, 1648.57111678, 1768.90477493, 1684.67121422, 1756.87140912, 1817.03823820, 1756.87140912, 1696.70458004, 1588.40428770, 1540.27082443, 1672.63784841, 1624.50438514, 1636.53775096, 1516.20409280, 1720.77131167, 1684.67121422, 1708.73794585, 2033.63882288, 2166.00584686, 2322.43960246, 2045.67218870, 1997.53872544, 2129.90574941, 1985.50535962, 1961.43862799, 2033.63882288, 2033.63882288, 2081.77228615, 2069.73892033, 2045.67218870, 2069.73892033, 2129.90574941, 2141.93911523, 2190.07257849, 2298.37287083, 2286.33950502, 2178.03921267, 2129.90574941, 2093.80565196, 2081.77228615, 2226.17267594, 2262.27277339, 2250.23940757, 2250.23940757, 2250.23940757, 2238.20604175, 2310.40623665, 2442.77326062, 2358.53969991, 2406.67316318, 2442.77326062, 2394.63979736, 2370.57306573, 2442.77326062, 2382.60643154, 2394.63979736, 2442.77326062, 2695.47394276, 2779.70750347, 2791.74086928, 2695.47394276, 2731.57404020, 2731.57404020, 2863.94106418, 2863.94106418, 2791.74086928, 2900.04116163, 2912.07452744, 2888.00779581, 2839.87433255, 2851.90769836, 2924.10789326, 2767.67413765, 2743.60740602, 2599.20701623, 2671.40721112, 2743.60740602, 2827.84096673, 2912.07452744, 2912.07452744, 2863.94106418, 2803.77423510, 2839.87433255, 2478.87335807, 2454.80662644, 2394.63979736, 2502.94008970, 2322.43960246, 2346.50633410, 2370.57306573, 2406.67316318, 2695.47394276, 2514.97345552, 2599.20701623, 2611.24038205, 2551.07355297, 2587.17365041, 2587.17365041, 2587.17365041, 2647.34047949, 2563.10691878, 2454.80662644, 2719.54067439, 2683.44057694, 2611.24038205, 2587.17365041, 2659.37384531, 2731.57404020, 2671.40721112, 2695.47394276, 2623.27374786, 2647.34047949, 2599.20701623, 2563.10691878, 2563.10691878, 2466.83999225, 2442.77326062, 2346.50633410, 2430.73989481, 2418.70652899, 2406.67316318, 2394.63979736, 2406.67316318, 2370.57306573, 2214.13931012, 2226.17267594, 2226.17267594, 2262.27277339, 2262.27277339, 2214.13931012, 2202.10594431, 2153.97248104, 2166.00584686, 2069.73892033, 2081.77228615, 2021.60545707, 2190.07257849, 2141.93911523, 2202.10594431, 2262.27277339, 2286.33950502, 2238.20604175, 2166.00584686, 2105.83901778, 1937.37189636, 2226.17267594, 2190.07257849, 2141.93911523, 2153.97248104, 2153.97248104, 2166.00584686, 2081.77228615, 1961.43862799, 2226.17267594, 2069.73892033, 2069.73892033, 2202.10594431, 2190.07257849, 2214.13931012, 2214.13931012, 2166.00584686, 2166.00584686, 2153.97248104, 2033.63882288, 1997.53872544, 1925.33853054, 1768.90477493, 2081.77228615, 2117.87238359, 2117.87238359, 2202.10594431, 2129.90574941, 2129.90574941, 2238.20604175, 2166.00584686, 2166.00584686, 2226.17267594, 2298.37287083, 2286.33950502, 2310.40623665, 2346.50633410, 2238.20604175, 2262.27277339, 2153.97248104, 2238.20604175, 2153.97248104, 2238.20604175, 2274.30613920, 2430.73989481, 2563.10691878, 2575.14028460, 2755.64077184, 2635.30711368, 2743.60740602, 2671.40721112, 2551.07355297, 2466.83999225, 2442.77326062, 2442.77326062, 2478.87335807, 2358.53969991, 2418.70652899, 2346.50633410, 2514.97345552, 2563.10691878, 2599.20701623, 2599.20701623, 2611.24038205, 2587.17365041, 2719.54067439, 2611.24038205, 2527.00682133, 2599.20701623, 2430.73989481, 2563.10691878, 2514.97345552, 2514.97345552, 2563.10691878, 2551.07355297, 2514.97345552, 2430.73989481, 2442.77326062, 2370.57306573, 2382.60643154, 2370.57306573, 2358.53969991, 2514.97345552, 2394.63979736, 2346.50633410, 2370.57306573, 2430.73989481, 2454.80662644, 2394.63979736, 2334.47296828, 2442.77326062, 2394.63979736, 2430.73989481, 2430.73989481, 2406.67316318, 2418.70652899, 2382.60643154, 2262.27277339, 2105.83901778, 2093.80565196, 2009.57209125, 2009.57209125, 1648.57111678, 1756.87140912, 1612.47101933, 1576.37092188, 1492.13736117, 1600.43765351, 1913.30516472, 1829.07160401, 1768.90477493, 1829.07160401, 1973.47199380, 1865.17170146, 1720.77131167, 1468.07062954, 1756.87140912, 1684.67121422, 1792.97150657, 1817.03823820, 1853.13833565, 1949.40526217, 1937.37189636, 2021.60545707, 2202.10594431, 2141.93911523, 2250.23940757, 2190.07257849, 2117.87238359, 2262.27277339, 2274.30613920, 2214.13931012, 2310.40623665, 2298.37287083, 2238.20604175, 2141.93911523, 2069.73892033, 2069.73892033, 2105.83901778], 
							[661.83511987, 659.42844671, 631.75170533, 612.49832003, 613.70165661, 601.66829079, 565.56819335, 623.32834926, 629.34503217, 631.75170533, 643.78507115, 611.29498345, 640.17506140, 649.80175406, 640.17506140, 602.87162738, 584.82157865, 557.14483728, 605.27830054, 589.63492498, 584.82157865, 542.70479830, 559.55151044, 565.56819335, 516.23139350, 594.44827130, 619.71833952, 635.36171508, 637.76838824, 631.75170533, 659.42844671, 665.44512962, 699.13855390, 693.12187099, 689.51186125, 670.25847594, 693.12187099, 640.17506140, 690.71519783, 685.90185151, 697.93521732, 673.86848569, 678.68183202, 675.07182227, 677.47849543, 651.00509064, 647.39508089, 634.15837850, 683.49517834, 675.07182227, 638.97172482, 675.07182227, 676.27515885, 677.47849543, 682.29184176, 679.88516860, 699.13855390, 737.64532451, 770.13541222, 805.03217308, 752.08536349, 812.25219257, 803.82883650, 766.52540247, 777.35543171, 777.35543171, 774.94875854, 758.10204640, 722.00194895, 754.49203666, 761.71205615, 770.13541222, 764.11872931, 752.08536349, 758.10204640, 756.89870982, 776.15209512, 771.33874880, 778.55876829, 799.01549017, 792.99880727, 780.96544145, 758.10204640, 742.45867084, 754.49203666, 766.52540247, 791.79547068, 782.16877803, 766.52540247, 773.74542196, 728.01863186, 673.86848569, 712.37525630, 709.96858314, 728.01863186, 685.90185151, 702.74856365, 687.10518809, 722.00194895, 722.00194895, 717.18860263, 762.91539273, 755.69537324, 738.84866109, 786.98212436, 780.96544145, 813.45552915, 824.28555839, 768.93207563, 784.57545120, 800.21882676, 806.23550966, 806.23550966, 807.43884625, 806.23550966, 815.86220232, 820.67554864, 848.35229002, 836.31892420, 883.24905089, 888.06239721, 866.40233874, 872.41902165, 878.43570456, 900.09576303, 884.45238747, 888.06239721, 883.24905089, 913.33246543, 886.85906063, 877.23236798, 871.21568507, 860.38565584, 845.94561686, 857.97898267, 884.45238747, 854.36897293, 865.19900216, 866.40233874, 883.24905089, 890.46907037, 879.63904114, 866.40233874, 857.97898267, 860.38565584, 851.96229976, 841.13227053, 862.79232900, 859.18231925, 854.36897293, 851.96229976, 839.92893395, 838.72559737, 827.89556813, 797.81215359, 758.10204640, 754.49203666, 746.06868058, 740.05199768, 724.40862212, 715.98526604, 729.22196844, 700.34189048, 731.62864161, 740.05199768, 731.62864161, 740.05199768, 735.23865135, 718.39193921, 696.73188074, 697.93521732, 643.78507115, 628.14169559, 643.78507115, 649.80175406, 752.08536349, 701.54522707, 731.62864161, 728.01863186, 734.03531477, 730.42530502, 780.96544145, 771.33874880, 762.91539273, 753.28870007, 818.26887548, 765.32206589, 803.82883650, 797.81215359, 779.76210487, 777.35543171, 738.84866109, 794.20214385, 773.74542196, 791.79547068, 800.21882676, 811.04885599, 799.01549017, 797.81215359, 791.79547068, 800.21882676, 805.03217308, 805.03217308, 796.60881701, 796.60881701, 818.26887548, 800.21882676, 802.62549992, 802.62549992, 825.48889497, 827.89556813, 813.45552915, 830.30224130, 853.16563635, 894.07908012, 885.65572405, 896.48575328, 879.63904114, 861.58899242, 867.60567532, 889.26573379, 876.02903140, 868.80901191, 863.99566558, 867.60567532, 890.46907037, 870.01234849, 863.99566558, 825.48889497, 833.91225104, 833.91225104, 835.11558762, 820.67554864, 830.30224130, 806.23550966, 767.72873905, 768.93207563, 771.33874880, 772.54208538, 790.59213410, 799.01549017, 800.21882676, 809.84551941, 777.35543171, 778.55876829, 797.81215359, 790.59213410, 792.99880727, 778.55876829, 740.05199768, 708.76524656, 688.30852467, 652.20842722, 672.66514911, 620.92167610, 654.61510038, 618.51500294, 679.88516860, 651.00509064, 644.98840773, 658.22511013, 660.63178329, 632.95504192, 647.39508089, 660.63178329, 653.41176380, 626.93835901, 625.73502243, 658.22511013, 706.35857339, 712.37525630, 693.12187099, 700.34189048, 665.44512962, 643.78507115, 696.73188074, 694.32520758, 666.64846620, 676.27515885, 653.41176380, 631.75170533, 618.51500294, 628.14169559, 576.39822258, 598.05828105, 573.99154942, 589.63492498, 584.82157865, 578.80489574, 539.09478855, 535.48477881], 
							[758.10204640, 758.10204640, 718.39193921, 709.96858314, 677.47849543, 533.07810564, 486.14797896, 516.23139350, 545.11147146, 515.02805692, 536.68811539, 505.40136427, 530.67143248, 540.29812513, 527.06142274, 535.48477881, 509.01137401, 464.48792049, 517.43473008, 499.38468136, 487.35131554, 457.26790100, 460.87791075, 444.03119861, 433.20116937, 472.91127656, 523.45141299, 535.48477881, 533.07810564, 521.04473983, 564.36485676, 513.82472034, 517.43473008, 509.01137401, 521.04473983, 499.38468136, 511.41804717, 492.16466187, 521.04473983, 561.95818360, 535.48477881, 524.65474957, 545.11147146, 524.65474957, 510.21471059, 494.57133503, 492.16466187, 498.18134478, 528.26475932, 541.50146171, 516.23139350, 534.28144222, 540.29812513, 541.50146171, 553.53482753, 553.53482753, 576.39822258, 551.12815437, 548.72148120, 616.10832977, 580.00823233, 619.71833952, 601.66829079, 619.71833952, 622.12501268, 655.81843697, 664.24179304, 638.97172482, 616.10832977, 589.63492498, 592.04159814, 598.05828105, 600.46495421, 595.65160789, 652.20842722, 701.54522707, 685.90185151, 647.39508089, 678.68183202, 667.85180278, 648.59841748, 625.73502243, 630.54836875, 630.54836875, 613.70165661, 607.68497370, 604.07496396, 572.78821284, 505.40136427, 500.58801794, 466.89459366, 446.43787177, 493.36799845, 456.06456442, 487.35131554, 469.30126682, 489.75798871, 502.99469110, 529.46809590, 509.01137401, 487.35131554, 552.33149095, 531.87476906, 481.33463264, 498.18134478, 529.46809590, 533.07810564, 541.50146171, 523.45141299, 496.97800820, 515.02805692, 535.48477881, 531.87476906, 523.45141299, 518.63806666, 534.28144222, 529.46809590, 557.14483728, 557.14483728, 599.26161763, 624.53168584, 577.60155916, 564.36485676, 547.51814462, 557.14483728, 553.53482753, 595.65160789, 598.05828105, 606.48163712, 613.70165661, 666.64846620, 657.02177355, 637.76838824, 637.76838824, 655.81843697, 640.17506140, 612.49832003, 632.95504192, 631.75170533, 613.70165661, 613.70165661, 613.70165661, 612.49832003, 554.73816411, 565.56819335, 567.97486651, 583.61824207, 599.26161763, 613.70165661, 608.88831028, 605.27830054, 595.65160789, 616.10832977, 602.87162738, 601.66829079, 575.19488600, 549.92481779, 527.06142274, 487.35131554, 492.16466187, 481.33463264, 469.30126682, 470.50460340, 482.53796922, 531.87476906, 522.24807641, 534.28144222, 504.19802769, 480.13129605, 459.67457417, 481.33463264, 459.67457417, 446.43787177, 468.09793024, 446.43787177, 523.45141299, 495.77467161, 518.63806666, 512.62138376, 504.19802769, 492.16466187, 552.33149095, 542.70479830, 570.38153967, 581.21156891, 549.92481779, 500.58801794, 505.40136427, 546.31480804, 511.41804717, 527.06142274, 488.55465212, 494.57133503, 494.57133503, 500.58801794, 522.24807641, 529.46809590, 533.07810564, 541.50146171, 551.12815437, 570.38153967, 582.41490549, 583.61824207, 582.41490549, 589.63492498, 583.61824207, 581.21156891, 586.02491523, 565.56819335, 563.16152018, 572.78821284, 571.58487625, 587.22825181, 614.90499319, 635.36171508, 612.49832003, 605.27830054, 614.90499319, 595.65160789, 593.24493472, 620.92167610, 614.90499319, 617.31166635, 604.07496396, 622.12501268, 641.37839799, 643.78507115, 632.95504192, 611.29498345, 622.12501268, 589.63492498, 582.41490549, 558.34817386, 566.77152993, 604.07496396, 553.53482753, 554.73816411, 611.29498345, 625.73502243, 611.29498345, 612.49832003, 611.29498345, 622.12501268, 599.26161763, 593.24493472, 625.73502243, 620.92167610, 623.32834926, 622.12501268, 601.66829079, 586.02491523, 563.16152018, 523.45141299, 511.41804717, 469.30126682, 482.53796922, 464.48792049, 505.40136427, 481.33463264, 490.96132529, 506.60470085, 505.40136427, 515.02805692, 499.38468136, 535.48477881, 500.58801794, 464.48792049, 486.14797896, 502.99469110, 502.99469110, 517.43473008, 516.23139350, 535.48477881, 510.21471059, 527.06142274, 529.46809590, 517.43473008, 494.57133503, 493.36799845, 482.53796922, 476.52128631, 509.01137401, 480.13129605, 407.93110116, 380.25435978, 353.78095499, 354.98429157, 339.34091601, 344.15426233, 317.68085754, 316.47752096], 
							[741.25533426, 746.06868058, 785.77878778, 790.59213410, 802.62549992, 728.01863186, 753.28870007, 695.52854416, 709.96858314, 663.03845645, 687.10518809, 685.90185151, 683.49517834, 649.80175406, 613.70165661, 617.31166635, 565.56819335, 528.26475932, 577.60155916, 576.39822258, 565.56819335, 523.45141299, 536.68811539, 517.43473008, 474.11461315, 515.02805692, 638.97172482, 646.19174431, 606.48163712, 600.46495421, 629.34503217, 601.66829079, 565.56819335, 573.99154942, 559.55151044, 534.28144222, 537.89145197, 516.23139350, 546.31480804, 559.55151044, 543.90813488, 541.50146171, 561.95818360, 558.34817386, 518.63806666, 529.46809590, 517.43473008, 511.41804717, 529.46809590, 531.87476906, 506.60470085, 524.65474957, 505.40136427, 487.35131554, 527.06142274, 529.46809590, 523.45141299, 553.53482753, 589.63492498, 596.85494447, 593.24493472, 625.73502243, 619.71833952, 600.46495421, 632.95504192, 648.59841748, 640.17506140, 619.71833952, 601.66829079, 577.60155916, 586.02491523, 604.07496396, 598.05828105, 580.00823233, 604.07496396, 582.41490549, 599.26161763, 589.63492498, 586.02491523, 566.77152993, 583.61824207, 577.60155916, 569.17820309, 540.29812513, 551.12815437, 536.68811539, 531.87476906, 509.01137401, 480.13129605, 469.30126682, 417.55779381, 417.55779381, 450.04788151, 418.76113039, 421.16780356, 405.52442800, 428.38782305, 450.04788151, 474.11461315, 454.86122784, 435.60784253, 472.91127656, 475.31794973, 457.26790100, 490.96132529, 500.58801794, 506.60470085, 571.58487625, 541.50146171, 517.43473008, 535.48477881, 555.94150069, 557.14483728, 545.11147146, 535.48477881, 539.09478855, 529.46809590, 564.36485676, 565.56819335, 616.10832977, 631.75170533, 595.65160789, 584.82157865, 577.60155916, 586.02491523, 571.58487625, 594.44827130, 613.70165661, 598.05828105, 589.63492498, 588.43158840, 583.61824207, 561.95818360, 563.16152018, 564.36485676, 577.60155916, 577.60155916, 584.82157865, 580.00823233, 577.60155916, 580.00823233, 541.50146171, 572.78821284, 533.07810564, 549.92481779, 541.50146171, 561.95818360, 561.95818360, 563.16152018, 554.73816411, 570.38153967, 561.95818360, 529.46809590, 511.41804717, 548.72148120, 541.50146171, 542.70479830, 527.06142274, 501.79135452, 517.43473008, 487.35131554, 502.99469110, 487.35131554, 502.99469110, 530.67143248, 528.26475932, 534.28144222, 499.38468136, 506.60470085, 481.33463264, 501.79135452, 488.55465212, 488.55465212, 511.41804717, 482.53796922, 534.28144222, 519.84140325, 524.65474957, 519.84140325, 500.58801794, 494.57133503, 535.48477881, 543.90813488, 551.12815437, 551.12815437, 540.29812513, 496.97800820, 510.21471059, 535.48477881, 525.85808615, 527.06142274, 504.19802769, 499.38468136, 490.96132529, 499.38468136, 488.55465212, 487.35131554, 494.57133503, 494.57133503, 496.97800820, 513.82472034, 522.24807641, 524.65474957, 513.82472034, 541.50146171, 547.51814462, 565.56819335, 551.12815437, 534.28144222, 506.60470085, 528.26475932, 523.45141299, 529.46809590, 539.09478855, 553.53482753, 553.53482753, 570.38153967, 569.17820309, 543.90813488, 552.33149095, 543.90813488, 559.55151044, 566.77152993, 566.77152993, 566.77152993, 567.97486651, 561.95818360, 551.12815437, 536.68811539, 543.90813488, 543.90813488, 512.62138376, 522.24807641, 506.60470085, 546.31480804, 529.46809590, 518.63806666, 535.48477881, 540.29812513, 560.75484702, 548.72148120, 543.90813488, 552.33149095, 522.24807641, 527.06142274, 553.53482753, 533.07810564, 529.46809590, 513.82472034, 502.99469110, 505.40136427, 481.33463264, 463.28458391, 493.36799845, 419.96446697, 428.38782305, 412.74444748, 421.16780356, 403.11775483, 394.69439876, 404.32109141, 386.27104269, 395.89773534, 361.00097448, 377.84768662, 361.00097448, 329.71422336, 340.54425259, 351.37428182, 344.15426233, 362.20431106, 403.11775483, 418.76113039, 386.27104269, 401.91441825, 415.15112065, 430.79449621, 451.25121810, 463.28458391, 446.43787177, 427.18448646, 403.11775483, 403.11775483, 326.10421361, 328.51088677, 312.86751121, 298.42747223, 324.90087703, 306.85082830, 285.19076984, 264.73404795], 
							[1095.03628924, 1072.17289419, 1039.68280649, 1055.32618205, 1001.17603588, 1010.80272853, 961.46592869, 770.13541222, 697.93521732, 728.01863186, 784.57545120, 848.35229002, 894.07908012, 889.26573379, 876.02903140, 839.92893395, 786.98212436, 732.83197819, 801.42216334, 774.94875854, 741.25533426, 677.47849543, 684.69851492, 675.07182227, 622.12501268, 673.86848569, 766.52540247, 792.99880727, 783.37211461, 818.26887548, 819.47221206, 813.45552915, 811.04885599, 824.28555839, 807.43884625, 789.38879752, 800.21882676, 801.42216334, 826.69223155, 854.36897293, 844.74228027, 843.53894369, 836.31892420, 827.89556813, 808.64218283, 768.93207563, 776.15209512, 788.18546094, 823.08222181, 838.72559737, 795.40548043, 808.64218283, 854.36897293, 862.79232900, 922.95915808, 903.70577277, 851.96229976, 830.30224130, 824.28555839, 921.75582150, 889.26573379, 914.53580201, 900.09576303, 879.63904114, 872.41902165, 933.78918731, 938.60253364, 914.53580201, 854.36897293, 866.40233874, 857.97898267, 863.99566558, 854.36897293, 849.55562660, 863.99566558, 855.57230951, 865.19900216, 892.87574354, 884.45238747, 913.33246543, 872.41902165, 884.45238747, 865.19900216, 848.35229002, 845.94561686, 792.99880727, 771.33874880, 734.03531477, 634.15837850, 624.53168584, 578.80489574, 549.92481779, 581.21156891, 583.61824207, 625.73502243, 575.19488600, 578.80489574, 613.70165661, 661.83511987, 661.83511987, 628.14169559, 752.08536349, 722.00194895, 643.78507115, 667.85180278, 685.90185151, 720.79861237, 717.18860263, 722.00194895, 667.85180278, 718.39193921, 707.56190997, 683.49517834, 670.25847594, 684.69851492, 715.98526604, 724.40862212, 725.61195870, 750.88202691, 803.82883650, 827.89556813, 806.23550966, 799.01549017, 758.10204640, 779.76210487, 770.13541222, 760.50871956, 799.01549017, 771.33874880, 777.35543171, 794.20214385, 794.20214385, 728.01863186, 713.57859288, 728.01863186, 703.95190023, 731.62864161, 766.52540247, 746.06868058, 708.76524656, 722.00194895, 718.39193921, 719.59527579, 663.03845645, 655.81843697, 664.24179304, 679.88516860, 695.52854416, 700.34189048, 707.56190997, 713.57859288, 696.73188074, 613.70165661, 634.15837850, 635.36171508, 641.37839799, 625.73502243, 601.66829079, 545.11147146, 613.70165661, 590.83826156, 577.60155916, 543.90813488, 555.94150069, 636.56505166, 611.29498345, 641.37839799, 625.73502243, 595.65160789, 559.55151044, 605.27830054, 563.16152018, 510.21471059, 534.28144222, 476.52128631, 589.63492498, 641.37839799, 655.81843697, 647.39508089, 651.00509064, 649.80175406, 678.68183202, 684.69851492, 703.95190023, 720.79861237, 732.83197819, 675.07182227, 676.27515885, 705.15523681, 682.29184176, 678.68183202, 644.98840773, 695.52854416, 665.44512962, 689.51186125, 687.10518809, 688.30852467, 700.34189048, 705.15523681, 720.79861237, 729.22196844, 740.05199768, 737.64532451, 743.66200742, 741.25533426, 758.10204640, 741.25533426, 728.01863186, 715.98526604, 696.73188074, 708.76524656, 726.81529528, 744.86534400, 736.44198793, 768.93207563, 767.72873905, 777.35543171, 777.35543171, 741.25533426, 770.13541222, 770.13541222, 766.52540247, 770.13541222, 759.30538298, 765.32206589, 774.94875854, 770.13541222, 756.89870982, 725.61195870, 744.86534400, 705.15523681, 690.71519783, 641.37839799, 673.86848569, 695.52854416, 625.73502243, 638.97172482, 687.10518809, 683.49517834, 708.76524656, 655.81843697, 636.56505166, 682.29184176, 646.19174431, 643.78507115, 669.05513936, 659.42844671, 676.27515885, 659.42844671, 581.21156891, 613.70165661, 541.50146171, 524.65474957, 494.57133503, 492.16466187, 492.16466187, 471.70793998, 507.80803743, 463.28458391, 472.91127656, 482.53796922, 487.35131554, 466.89459366, 453.65789126, 477.72462289, 439.21785228, 405.52442800, 376.64435004, 418.76113039, 389.88105243, 434.40450595, 452.45455468, 484.94464238, 478.92795947, 489.75798871, 493.36799845, 476.52128631, 466.89459366, 457.26790100, 442.82786202, 431.99783279, 463.28458391, 429.59115963, 362.20431106, 373.03434029, 315.27418438, 336.93424284, 306.85082830, 312.86751121, 300.83414540, 302.03748198], 
							[784.57545120, 792.99880727, 770.13541222, 782.16877803, 764.11872931, 776.15209512, 666.64846620, 681.08850518, 684.69851492, 706.35857339, 715.98526604, 701.54522707, 697.93521732, 734.03531477, 684.69851492, 649.80175406, 613.70165661, 571.58487625, 619.71833952, 637.76838824, 601.66829079, 554.73816411, 625.73502243, 570.38153967, 553.53482753, 529.46809590, 634.15837850, 653.41176380, 678.68183202, 685.90185151, 687.10518809, 671.46181253, 617.31166635, 634.15837850, 644.98840773, 631.75170533, 651.00509064, 629.34503217, 653.41176380, 661.83511987, 637.76838824, 638.97172482, 672.66514911, 694.32520758, 661.83511987, 670.25847594, 648.59841748, 631.75170533, 672.66514911, 659.42844671, 623.32834926, 649.80175406, 631.75170533, 619.71833952, 697.93521732, 676.27515885, 700.34189048, 732.83197819, 746.06868058, 780.96544145, 734.03531477, 794.20214385, 782.16877803, 770.13541222, 791.79547068, 841.13227053, 819.47221206, 803.82883650, 777.35543171, 764.11872931, 817.06553890, 767.72873905, 770.13541222, 753.28870007, 817.06553890, 803.82883650, 855.57230951, 842.33560711, 849.55562660, 843.53894369, 832.70891446, 824.28555839, 792.99880727, 811.04885599, 777.35543171, 795.40548043, 799.01549017, 806.23550966, 695.52854416, 690.71519783, 655.81843697, 583.61824207, 646.19174431, 577.60155916, 612.49832003, 573.99154942, 592.04159814, 637.76838824, 700.34189048, 699.13855390, 666.64846620, 754.49203666, 736.44198793, 657.02177355, 736.44198793, 743.66200742, 784.57545120, 771.33874880, 753.28870007, 701.54522707, 752.08536349, 743.66200742, 803.82883650, 806.23550966, 792.99880727, 818.26887548, 817.06553890, 833.91225104, 872.41902165, 920.55248491, 926.56916782, 901.29909961, 851.96229976, 835.11558762, 854.36897293, 862.79232900, 862.79232900, 844.74228027, 890.46907037, 902.50243619, 902.50243619, 914.53580201, 844.74228027, 854.36897293, 885.65572405, 871.21568507, 890.46907037, 857.97898267, 865.19900216, 885.65572405, 882.04571430, 879.63904114, 877.23236798, 854.36897293, 861.58899242, 854.36897293, 878.43570456, 886.85906063, 892.87574354, 865.19900216, 861.58899242, 842.33560711, 795.40548043, 790.59213410, 722.00194895, 743.66200742, 711.17191972, 669.05513936, 671.46181253, 660.63178329, 590.83826156, 620.92167610, 622.12501268, 651.00509064, 726.81529528, 730.42530502, 740.05199768, 684.69851492, 649.80175406, 607.68497370, 610.09164687, 577.60155916, 563.16152018, 573.99154942, 536.68811539, 647.39508089, 641.37839799, 660.63178329, 651.00509064, 631.75170533, 631.75170533, 697.93521732, 682.29184176, 694.32520758, 709.96858314, 746.06868058, 683.49517834, 694.32520758, 749.67869033, 730.42530502, 756.89870982, 724.40862212, 758.10204640, 746.06868058, 754.49203666, 736.44198793, 746.06868058, 759.30538298, 744.86534400, 756.89870982, 747.27201717, 755.69537324, 766.52540247, 765.32206589, 799.01549017, 803.82883650, 813.45552915, 792.99880727, 770.13541222, 772.54208538, 762.91539273, 758.10204640, 794.20214385, 764.11872931, 773.74542196, 797.81215359, 800.21882676, 812.25219257, 789.38879752, 789.38879752, 832.70891446, 811.04885599, 821.87888522, 797.81215359, 819.47221206, 826.69223155, 824.28555839, 791.79547068, 774.94875854, 742.45867084, 740.05199768, 724.40862212, 705.15523681, 715.98526604, 707.56190997, 689.51186125, 670.25847594, 734.03531477, 711.17191972, 728.01863186, 712.37525630, 714.78192946, 740.05199768, 693.12187099, 691.91853441, 702.74856365, 675.07182227, 685.90185151, 640.17506140, 583.61824207, 607.68497370, 551.12815437, 541.50146171, 545.11147146, 523.45141299, 511.41804717, 504.19802769, 577.60155916, 542.70479830, 551.12815437, 567.97486651, 539.09478855, 522.24807641, 505.40136427, 539.09478855, 505.40136427, 450.04788151, 459.67457417, 470.50460340, 470.50460340, 518.63806666, 551.12815437, 601.66829079, 567.97486651, 575.19488600, 570.38153967, 569.17820309, 541.50146171, 549.92481779, 523.45141299, 502.99469110, 494.57133503, 480.13129605, 433.20116937, 445.23453519, 397.10107192, 379.05102320, 379.05102320, 397.10107192, 373.03434029, 376.64435004], 
							[1624.50438514, 1624.50438514, 1600.43765351, 1492.13736117, 1444.00389791, 1383.83706883, 1311.63687393, 1251.47004485, 1227.40331322, 1227.40331322, 1251.47004485, 1263.50341067, 1263.50341067, 1299.60350811, 1239.43667904, 1227.40331322, 1074.57956736, 1022.83609435, 1085.40959659, 1072.17289419, 1119.10302088, 1016.81941144, 1072.17289419, 1064.95287471, 998.76936272, 1070.96955761, 1152.79644516, 1185.28653286, 1227.40331322, 1179.26984996, 1215.36994740, 1161.21980123, 1157.60979149, 1215.36994740, 1203.33658159, 1203.33658159, 1203.33658159, 1179.26984996, 1275.53677648, 1347.73697138, 1323.67023975, 1323.67023975, 1299.60350811, 1383.83706883, 1347.73697138, 1275.53677648, 1275.53677648, 1263.50341067, 1335.70360556, 1371.80370301, 1287.57014230, 1287.57014230, 1323.67023975, 1407.90380046, 1516.20409280, 1588.40428770, 1540.27082443, 1600.43765351, 1612.47101933, 1600.43765351, 1576.37092188, 1660.60448259, 1636.53775096, 1576.37092188, 1624.50438514, 1600.43765351, 1564.33755606, 1528.23745862, 1480.10399535, 1576.37092188, 1552.30419025, 1564.33755606, 1600.43765351, 1528.23745862, 1588.40428770, 1576.37092188, 1624.50438514, 1636.53775096, 1696.70458004, 1732.80467749, 1756.87140912, 1708.73794585, 1720.77131167, 1732.80467749, 1732.80467749, 1696.70458004, 1696.70458004, 1648.57111678, 1456.03726372, 1419.93716627, 1311.63687393, 1239.43667904, 1347.73697138, 1311.63687393, 1359.77033719, 1335.70360556, 1323.67023975, 1347.73697138, 1468.07062954, 1407.90380046, 1335.70360556, 1564.33755606, 1504.17072698, 1359.77033719, 1383.83706883, 1395.87043464, 1431.97053209, 1480.10399535, 1407.90380046, 1371.80370301, 1407.90380046, 1419.93716627, 1395.87043464, 1347.73697138, 1347.73697138, 1383.83706883, 1383.83706883, 1431.97053209, 1395.87043464, 1480.10399535, 1456.03726372, 1383.83706883, 1419.93716627, 1419.93716627, 1480.10399535, 1456.03726372, 1456.03726372, 1468.07062954, 1480.10399535, 1431.97053209, 1444.00389791, 1456.03726372, 1383.83706883, 1335.70360556, 1347.73697138, 1323.67023975, 1299.60350811, 1299.60350811, 1287.57014230, 1263.50341067, 1263.50341067, 1239.43667904, 1202.13324501, 1062.54620154, 1113.08633797, 1186.48986945, 1215.36994740, 1275.53677648, 1263.50341067, 1215.36994740, 1251.47004485, 1263.50341067, 1179.26984996, 1299.60350811, 1239.43667904, 1239.43667904, 1179.26984996, 1083.00292343, 1022.83609435, 963.87260185, 938.60253364, 912.12912884, 879.63904114, 934.99252389, 1022.83609435, 1015.61607486, 1021.63275777, 974.70263109, 961.46592869, 938.60253364, 938.60253364, 907.31578252, 843.53894369, 895.28241670, 826.69223155, 984.32932374, 933.78918731, 967.48261160, 968.68594818, 938.60253364, 944.61921655, 1008.39605537, 1001.17603588, 1008.39605537, 1014.41273828, 1046.90282598, 1005.98938221, 975.90596767, 1033.66612358, 1005.98938221, 1030.05611384, 969.88928476, 1010.80272853, 965.07593843, 1016.81941144, 1008.39605537, 1083.00292343, 1113.08633797, 1153.99978174, 1164.82981098, 1197.31989868, 1215.36994740, 1215.36994740, 1203.33658159, 1227.40331322, 1227.40331322, 1239.43667904, 1251.47004485, 1215.36994740, 1173.25316705, 1202.13324501, 1181.67652312, 1227.40331322, 1263.50341067, 1335.70360556, 1275.53677648, 1335.70360556, 1287.57014230, 1263.50341067, 1227.40331322, 1275.53677648, 1311.63687393, 1299.60350811, 1275.53677648, 1263.50341067, 1311.63687393, 1311.63687393, 1311.63687393, 1275.53677648, 1251.47004485, 1287.57014230, 1239.43667904, 1239.43667904, 1251.47004485, 1287.57014230, 1203.33658159, 1197.31989868, 1275.53677648, 1299.60350811, 1359.77033719, 1299.60350811, 1287.57014230, 1335.70360556, 1275.53677648, 1239.43667904, 1323.67023975, 1275.53677648, 1239.43667904, 1203.33658159, 1095.03628924, 1123.91636720, 1070.96955761, 989.14267007, 998.76936272, 832.70891446, 856.77564609, 837.52226079, 889.26573379, 832.70891446, 843.53894369, 854.36897293, 836.31892420, 806.23550966, 797.81215359, 843.53894369, 794.20214385, 738.84866109, 728.01863186, 755.69537324, 741.25533426, 782.16877803, 815.86220232, 889.26573379, 884.45238747, 884.45238747, 920.55248491, 928.97584099, 920.55248491, 888.06239721, 836.31892420, 814.65886573, 839.92893395, 812.25219257, 722.00194895, 724.40862212, 669.05513936, 649.80175406, 613.70165661, 634.15837850, 631.75170533, 583.61824207], 
							[838.72559737, 826.69223155, 836.31892420, 807.43884625, 813.45552915, 768.93207563, 742.45867084, 685.90185151, 685.90185151, 675.07182227, 758.10204640, 720.79861237, 740.05199768, 740.05199768, 722.00194895, 717.18860263, 709.96858314, 697.93521732, 773.74542196, 779.76210487, 777.35543171, 737.64532451, 750.88202691, 742.45867084, 696.73188074, 709.96858314, 788.18546094, 777.35543171, 785.77878778, 782.16877803, 801.42216334, 794.20214385, 766.52540247, 806.23550966, 782.16877803, 783.37211461, 801.42216334, 773.74542196, 813.45552915, 830.30224130, 806.23550966, 842.33560711, 837.52226079, 833.91225104, 820.67554864, 817.06553890, 817.06553890, 809.84551941, 818.26887548, 819.47221206, 821.87888522, 821.87888522, 827.89556813, 790.59213410, 829.09890471, 857.97898267, 833.91225104, 853.16563635, 866.40233874, 902.50243619, 890.46907037, 926.56916782, 927.77250440, 932.58585073, 928.97584099, 934.99252389, 927.77250440, 909.72245568, 904.90910935, 896.48575328, 906.11244594, 919.34914833, 901.29909961, 896.48575328, 926.56916782, 914.53580201, 957.85591894, 962.66926527, 989.14267007, 1033.66612358, 1013.20940170, 1032.46278700, 1034.86946017, 1049.30949914, 1036.07279675, 1034.86946017, 1002.37937246, 989.14267007, 951.83923604, 920.55248491, 914.53580201, 915.73913859, 884.45238747, 855.57230951, 883.24905089, 854.36897293, 853.16563635, 890.46907037, 922.95915808, 932.58585073, 884.45238747, 880.84237772, 900.09576303, 873.62235823, 884.45238747, 866.40233874, 880.84237772, 867.60567532, 855.57230951, 836.31892420, 871.21568507, 860.38565584, 874.82569481, 863.99566558, 871.21568507, 866.40233874, 868.80901191, 903.70577277, 906.11244594, 880.84237772, 890.46907037, 854.36897293, 842.33560711, 851.96229976, 871.21568507, 855.57230951, 872.41902165, 859.18231925, 872.41902165, 859.18231925, 872.41902165, 876.02903140, 866.40233874, 877.23236798, 876.02903140, 843.53894369, 844.74228027, 849.55562660, 836.31892420, 832.70891446, 824.28555839, 808.64218283, 797.81215359, 796.60881701, 801.42216334, 807.43884625, 794.20214385, 800.21882676, 800.21882676, 807.43884625, 819.47221206, 829.09890471, 771.33874880, 782.16877803, 782.16877803, 753.28870007, 743.66200742, 705.15523681, 653.41176380, 664.24179304, 677.47849543, 684.69851492, 630.54836875, 664.24179304, 675.07182227, 671.46181253, 706.35857339, 697.93521732, 689.51186125, 655.81843697, 683.49517834, 641.37839799, 649.80175406, 661.83511987, 618.51500294, 697.93521732, 661.83511987, 695.52854416, 672.66514911, 684.69851492, 655.81843697, 702.74856365, 707.56190997, 719.59527579, 734.03531477, 746.06868058, 700.34189048, 696.73188074, 718.39193921, 687.10518809, 707.56190997, 688.30852467, 706.35857339, 711.17191972, 726.81529528, 718.39193921, 774.94875854, 764.11872931, 801.42216334, 808.64218283, 806.23550966, 838.72559737, 824.28555839, 820.67554864, 848.35229002, 843.53894369, 839.92893395, 836.31892420, 829.09890471, 831.50557788, 825.48889497, 813.45552915, 806.23550966, 845.94561686, 872.41902165, 856.77564609, 877.23236798, 901.29909961, 872.41902165, 856.77564609, 880.84237772, 902.50243619, 922.95915808, 916.94247517, 922.95915808, 930.17917757, 944.61921655, 927.77250440, 915.73913859, 924.16249466, 886.85906063, 900.09576303, 865.19900216, 868.80901191, 885.65572405, 848.35229002, 836.31892420, 863.99566558, 872.41902165, 874.82569481, 850.75896318, 832.70891446, 868.80901191, 824.28555839, 812.25219257, 830.30224130, 825.48889497, 815.86220232, 776.15209512, 711.17191972, 740.05199768, 700.34189048, 678.68183202, 677.47849543, 644.98840773, 649.80175406, 628.14169559, 659.42844671, 631.75170533, 634.15837850, 663.03845645, 628.14169559, 619.71833952, 587.22825181, 646.19174431, 586.02491523, 545.11147146, 554.73816411, 569.17820309, 569.17820309, 614.90499319, 647.39508089, 696.73188074, 715.98526604, 682.29184176, 713.57859288, 712.37525630, 702.74856365, 690.71519783, 667.85180278, 655.81843697, 646.19174431, 632.95504192, 593.24493472, 578.80489574, 523.45141299, 509.01137401, 494.57133503, 502.99469110, 474.11461315, 446.43787177], 
							[974.70263109, 949.43256287, 841.13227053, 877.23236798, 872.41902165, 839.92893395, 830.30224130, 885.65572405, 866.40233874, 848.35229002, 854.36897293, 795.40548043, 830.30224130, 836.31892420, 811.04885599, 806.23550966, 741.25533426, 699.13855390, 766.52540247, 755.69537324, 796.60881701, 779.76210487, 871.21568507, 872.41902165, 836.31892420, 944.61921655, 1069.76622103, 1114.28967455, 1024.03943093, 1001.17603588, 1044.49615282, 973.49929450, 918.14581175, 986.73599690, 1009.59939195, 992.75267981, 974.70263109, 938.60253364, 992.75267981, 985.53266032, 1003.58270904, 1050.51283573, 1086.61293317, 1085.40959659, 1061.34286496, 1016.81941144, 1016.81941144, 1022.83609435, 1076.98624052, 1057.73285522, 1040.88614307, 1005.98938221, 989.14267007, 936.19586048, 1005.98938221, 1040.88614307, 1001.17603588, 1083.00292343, 1119.10302088, 1084.20626001, 1046.90282598, 1058.93619180, 1083.00292343, 1028.85277726, 1072.17289419, 1089.01960634, 1102.25630873, 1104.66298190, 1079.39291368, 1083.00292343, 1070.96955761, 1058.93619180, 1087.81626976, 1078.18957710, 1090.22294292, 1085.40959659, 1123.91636720, 1107.06965506, 1133.54305986, 1119.10302088, 1072.17289419, 1045.69948940, 1034.86946017, 1039.68280649, 1031.25945042, 1024.03943093, 1069.76622103, 1070.96955761, 1037.27613333, 1018.02274802, 991.54934323, 932.58585073, 954.24590920, 932.58585073, 973.49929450, 910.92579226, 938.60253364, 962.66926527, 973.49929450, 987.93933348, 961.46592869, 1040.88614307, 1028.85277726, 1020.42942119, 1028.85277726, 1028.85277726, 1016.81941144, 1010.80272853, 950.63589945, 938.60253364, 1009.59939195, 1012.00606512, 1005.98938221, 991.54934323, 995.15935297, 1034.86946017, 999.97269930, 1044.49615282, 1028.85277726, 1052.91950889, 1052.91950889, 989.14267007, 975.90596767, 962.66926527, 978.31264083, 963.87260185, 990.34600665, 969.88928476, 956.65258236, 962.66926527, 948.22922629, 954.24590920, 926.56916782, 914.53580201, 925.36583124, 920.55248491, 913.33246543, 921.75582150, 900.09576303, 883.24905089, 888.06239721, 896.48575328, 866.40233874, 842.33560711, 860.38565584, 874.82569481, 884.45238747, 889.26573379, 906.11244594, 891.67240696, 860.38565584, 818.26887548, 778.55876829, 815.86220232, 811.04885599, 807.43884625, 746.06868058, 738.84866109, 725.61195870, 699.13855390, 673.86848569, 747.27201717, 703.95190023, 699.13855390, 752.08536349, 766.52540247, 785.77878778, 749.67869033, 728.01863186, 711.17191972, 719.59527579, 700.34189048, 685.90185151, 691.91853441, 669.05513936, 722.00194895, 740.05199768, 717.18860263, 728.01863186, 722.00194895, 741.25533426, 813.45552915, 784.57545120, 782.16877803, 788.18546094, 776.15209512, 711.17191972, 747.27201717, 748.47535375, 715.98526604, 737.64532451, 725.61195870, 746.06868058, 741.25533426, 768.93207563, 768.93207563, 850.75896318, 880.84237772, 857.97898267, 878.43570456, 885.65572405, 965.07593843, 937.39919706, 915.73913859, 924.16249466, 936.19586048, 920.55248491, 943.41587996, 903.70577277, 914.53580201, 890.46907037, 878.43570456, 930.17917757, 969.88928476, 992.75267981, 1010.80272853, 1042.08947966, 1043.29281624, 1043.29281624, 1034.86946017, 1034.86946017, 1040.88614307, 1004.78604563, 1040.88614307, 1038.47946991, 1046.90282598, 1048.10616256, 1050.51283573, 1012.00606512, 1015.61607486, 987.93933348, 968.68594818, 932.58585073, 932.58585073, 966.27927501, 908.51911910, 934.99252389, 950.63589945, 963.87260185, 973.49929450, 957.85591894, 928.97584099, 962.66926527, 890.46907037, 913.33246543, 907.31578252, 904.90910935, 906.11244594, 882.04571430, 803.82883650, 782.16877803, 738.84866109, 762.91539273, 749.67869033, 700.34189048, 728.01863186, 741.25533426, 783.37211461, 764.11872931, 759.30538298, 803.82883650, 770.13541222, 758.10204640, 732.83197819, 777.35543171, 786.98212436, 743.66200742, 749.67869033, 771.33874880, 847.14895344, 850.75896318, 879.63904114, 926.56916782, 914.53580201, 938.60253364, 962.66926527, 959.05925553, 975.90596767, 1009.59939195, 957.85591894, 942.21254338, 948.22922629, 962.66926527, 949.43256287, 959.05925553, 913.33246543, 874.82569481, 843.53894369, 861.58899242, 845.94561686, 844.74228027], 
							[1512.76255018, 1524.21831443, 1489.83898830, 1466.91542642, 1431.97053209, 1407.90380046, 1335.70360556, 1359.77033719, 1359.77033719, 1359.77033719, 1431.97053209, 1359.77033719, 1347.73697138, 1323.67023975, 1299.60350811, 1263.50341067, 1203.33658159, 1179.26984996, 1263.50341067, 1275.53677648, 1275.53677648, 1191.30321577, 1251.47004485, 1215.36994740, 1182.87985970, 1263.50341067, 1371.80370301, 1383.83706883, 1359.77033719, 1347.73697138, 1347.73697138, 1263.50341067, 1239.43667904, 1311.63687393, 1275.53677648, 1275.53677648, 1311.63687393, 1275.53677648, 1263.50341067, 1287.57014230, 1251.47004485, 1251.47004485, 1275.53677648, 1263.50341067, 1251.47004485, 1239.43667904, 1251.47004485, 1239.43667904, 1287.57014230, 1263.50341067, 1227.40331322, 1263.50341067, 1275.53677648, 1227.40331322, 1263.50341067, 1275.53677648, 1299.60350811, 1395.87043464, 1480.10399535, 1504.17072698, 1468.07062954, 1564.33755606, 1540.27082443, 1516.20409280, 1504.17072698, 1552.30419025, 1588.40428770, 1492.13736117, 1480.10399535, 1468.07062954, 1480.10399535, 1468.07062954, 1528.23745862, 1504.17072698, 1564.33755606, 1564.33755606, 1588.40428770, 1600.43765351, 1600.43765351, 1636.53775096, 1660.60448259, 1624.50438514, 1624.50438514, 1696.70458004, 1636.53775096, 1636.53775096, 1672.63784841, 1660.60448259, 1576.37092188, 1528.23745862, 1504.17072698, 1444.00389791, 1431.97053209, 1444.00389791, 1492.13736117, 1444.00389791, 1456.03726372, 1504.17072698, 1564.33755606, 1588.40428770, 1480.10399535, 1612.47101933, 1576.37092188, 1564.33755606, 1636.53775096, 1648.57111678, 1624.50438514, 1612.47101933, 1612.47101933, 1576.37092188, 1564.33755606, 1648.57111678, 1600.43765351, 1564.33755606, 1612.47101933, 1636.53775096, 1648.57111678, 1708.73794585, 1756.87140912, 1792.97150657, 1829.07160401, 1792.97150657, 1792.97150657, 1768.90477493, 1792.97150657, 1780.93814075, 1792.97150657, 1792.97150657, 1805.00487238, 1792.97150657, 1780.93814075, 1780.93814075, 1744.83804330, 1696.70458004, 1672.63784841, 1732.80467749, 1732.80467749, 1708.73794585, 1696.70458004, 1720.77131167, 1708.73794585, 1720.77131167, 1696.70458004, 1732.80467749, 1708.73794585, 1696.70458004, 1660.60448259, 1660.60448259, 1672.63784841, 1648.57111678, 1600.43765351, 1588.40428770, 1576.37092188, 1576.37092188, 1552.30419025, 1456.03726372, 1395.87043464, 1419.93716627, 1323.67023975, 1299.60350811, 1323.67023975, 1431.97053209, 1371.80370301, 1383.83706883, 1456.03726372, 1468.07062954, 1504.17072698, 1504.17072698, 1480.10399535, 1395.87043464, 1383.83706883, 1323.67023975, 1395.87043464, 1444.00389791, 1335.70360556, 1468.07062954, 1456.03726372, 1480.10399535, 1504.17072698, 1468.07062954, 1444.00389791, 1564.33755606, 1564.33755606, 1600.43765351, 1576.37092188, 1624.50438514, 1504.17072698, 1516.20409280, 1612.47101933, 1588.40428770, 1636.53775096, 1540.27082443, 1612.47101933, 1600.43765351, 1516.20409280, 1552.30419025, 1612.47101933, 1588.40428770, 1612.47101933, 1660.60448259, 1636.53775096, 1684.67121422, 1636.53775096, 1660.60448259, 1708.73794585, 1696.70458004, 1708.73794585, 1756.87140912, 1672.63784841, 1672.63784841, 1648.57111678, 1624.50438514, 1684.67121422, 1672.63784841, 1792.97150657, 1805.00487238, 1877.20506728, 1865.17170146, 1805.00487238, 1792.97150657, 1877.20506728, 1780.93814075, 1853.13833565, 1853.13833565, 1901.27179891, 1865.17170146, 1925.33853054, 1913.30516472, 1889.23843309, 1817.03823820, 1805.00487238, 1768.90477493, 1780.93814075, 1780.93814075, 1768.90477493, 1648.57111678, 1672.63784841, 1720.77131167, 1805.00487238, 1829.07160401, 1768.90477493, 1768.90477493, 1841.10496983, 1720.77131167, 1720.77131167, 1768.90477493, 1720.77131167, 1817.03823820, 1805.00487238, 1756.87140912, 1768.90477493, 1732.80467749, 1744.83804330, 1744.83804330, 1528.23745862, 1648.57111678, 1564.33755606, 1732.80467749, 1648.57111678, 1696.70458004, 1708.73794585, 1672.63784841, 1624.50438514, 1648.57111678, 1744.83804330, 1756.87140912, 1817.03823820, 1817.03823820, 1913.30516472, 1925.33853054, 1961.43862799, 2009.57209125, 2105.83901778, 1973.47199380, 1925.33853054, 1985.50535962, 2069.73892033, 2105.83901778, 2141.93911523, 2105.83901778, 2129.90574941, 2238.20604175, 2298.37287083, 2286.33950502, 2274.30613920, 2238.20604175, 2178.03921267, 2081.77228615, 2117.87238359, 1973.47199380, 1889.23843309], 
							[991.54934323, 998.76936272, 920.55248491, 794.20214385, 794.20214385, 643.78507115, 649.80175406, 637.76838824, 626.93835901, 608.88831028, 648.59841748, 679.88516860, 655.81843697, 629.34503217, 604.07496396, 578.80489574, 548.72148120, 511.41804717, 547.51814462, 523.45141299, 511.41804717, 475.31794973, 480.13129605, 427.18448646, 293.61412591, 469.30126682, 653.41176380, 661.83511987, 652.20842722, 619.71833952, 647.39508089, 596.85494447, 571.58487625, 577.60155916, 576.39822258, 518.63806666, 500.58801794, 478.92795947, 541.50146171, 555.94150069, 567.97486651, 565.56819335, 572.78821284, 569.17820309, 545.11147146, 535.48477881, 504.19802769, 494.57133503, 549.92481779, 554.73816411, 547.51814462, 563.16152018, 523.45141299, 523.45141299, 596.85494447, 663.03845645, 652.20842722, 648.59841748, 719.59527579, 699.13855390, 669.05513936, 722.00194895, 722.00194895, 719.59527579, 720.79861237, 700.34189048, 688.30852467, 679.88516860, 654.61510038, 619.71833952, 623.32834926, 631.75170533, 582.41490549, 542.70479830, 599.26161763, 590.83826156, 611.29498345, 602.87162738, 607.68497370, 599.26161763, 594.44827130, 583.61824207, 582.41490549, 549.92481779, 559.55151044, 559.55151044, 565.56819335, 531.87476906, 489.75798871, 480.13129605, 466.89459366, 428.38782305, 492.16466187, 445.23453519, 490.96132529, 453.65789126, 445.23453519, 460.87791075, 529.46809590, 565.56819335, 547.51814462, 643.78507115, 625.73502243, 565.56819335, 592.04159814, 611.29498345, 619.71833952, 606.48163712, 589.63492498, 582.41490549, 596.85494447, 611.29498345, 601.66829079, 581.21156891, 583.61824207, 605.27830054, 601.66829079, 600.46495421, 611.29498345, 618.51500294, 632.95504192, 604.07496396, 613.70165661, 612.49832003, 612.49832003, 601.66829079, 577.60155916, 573.99154942, 563.16152018, 569.17820309, 565.56819335, 572.78821284, 553.53482753, 542.70479830, 541.50146171, 509.01137401, 529.46809590, 530.67143248, 517.43473008, 502.99469110, 505.40136427, 487.35131554, 486.14797896, 472.91127656, 470.50460340, 475.31794973, 476.52128631, 487.35131554, 500.58801794, 492.16466187, 483.74130580, 552.33149095, 619.71833952, 628.14169559, 631.75170533, 604.07496396, 576.39822258, 569.17820309, 517.43473008, 513.82472034, 494.57133503, 478.92795947, 493.36799845, 501.79135452, 528.26475932, 522.24807641, 528.26475932, 513.82472034, 495.77467161, 471.70793998, 486.14797896, 457.26790100, 404.32109141, 436.81117912, 403.11775483, 486.14797896, 507.80803743, 495.77467161, 463.28458391, 468.09793024, 484.94464238, 539.09478855, 524.65474957, 507.80803743, 537.89145197, 555.94150069, 510.21471059, 506.60470085, 513.82472034, 487.35131554, 475.31794973, 457.26790100, 468.09793024, 462.08124733, 476.52128631, 480.13129605, 492.16466187, 499.38468136, 495.77467161, 500.58801794, 500.58801794, 553.53482753, 543.90813488, 525.85808615, 553.53482753, 540.29812513, 530.67143248, 517.43473008, 505.40136427, 501.79135452, 492.16466187, 511.41804717, 498.18134478, 519.84140325, 537.89145197, 534.28144222, 558.34817386, 554.73816411, 536.68811539, 575.19488600, 569.17820309, 587.22825181, 592.04159814, 559.55151044, 588.43158840, 577.60155916, 577.60155916, 554.73816411, 524.65474957, 531.87476906, 515.02805692, 495.77467161, 493.36799845, 500.58801794, 509.01137401, 488.55465212, 481.33463264, 527.06142274, 541.50146171, 519.84140325, 511.41804717, 506.60470085, 536.68811539, 481.33463264, 460.87791075, 468.09793024, 453.65789126, 446.43787177, 370.62766713, 364.61098422, 375.44101346, 312.86751121, 315.27418438, 312.86751121, 274.36074060, 298.42747223, 268.34405769, 264.73404795, 193.73718964, 196.14386280, 205.77055545, 220.21059443, 249.09067239, 226.22727734, 252.70068213, 280.37742351, 235.85396999, 244.27732606, 240.66731632, 228.63395050, 231.04062366, 241.87065290, 271.95406744, 276.76741377, 270.75073086, 269.54739428, 270.75073086, 265.93738453, 267.14072111, 250.29400897, 235.85396999, 228.63395050, 238.26064315, 214.19391152, 216.60058469, 191.33051647, 191.33051647, 193.73718964, 185.31383356, 181.70382382, 125.14700449], 
							[1197.31989868, 1239.43667904, 1175.65984021, 1227.40331322, 1215.36994740, 1174.45650363, 1126.32304037, 1048.10616256, 1044.49615282, 1052.91950889, 1103.45964532, 1128.72971353, 1146.77976225, 1155.20311832, 1122.71303062, 1120.30635746, 1107.06965506, 1107.06965506, 1191.30321577, 1197.31989868, 1162.42313781, 1116.69634771, 1151.59310858, 1126.32304037, 1131.13638669, 1179.26984996, 1188.89654261, 1227.40331322, 1203.33658159, 1215.36994740, 1227.40331322, 1203.33658159, 1194.91322552, 1215.36994740, 1227.40331322, 1196.11656210, 1215.36994740, 1197.31989868, 1251.47004485, 1263.50341067, 1251.47004485, 1227.40331322, 1227.40331322, 1239.43667904, 1227.40331322, 1169.64315730, 1163.62647440, 1133.54305986, 1202.13324501, 1172.04983047, 1169.64315730, 1166.03314756, 1173.25316705, 1185.28653286, 1215.36994740, 1227.40331322, 1239.43667904, 1185.28653286, 1239.43667904, 1251.47004485, 1335.70360556, 1347.73697138, 1311.63687393, 1275.53677648, 1251.47004485, 1251.47004485, 1251.47004485, 1239.43667904, 1227.40331322, 1227.40331322, 1203.33658159, 1179.26984996, 1179.26984996, 1203.33658159, 1203.33658159, 1203.33658159, 1203.33658159, 1215.36994740, 1263.50341067, 1287.57014230, 1323.67023975, 1299.60350811, 1299.60350811, 1287.57014230, 1239.43667904, 1192.50655235, 1185.28653286, 1145.57642567, 1083.00292343, 1080.59625027, 1034.86946017, 986.73599690, 1034.86946017, 1002.37937246, 1060.13952838, 967.48261160, 1007.19271879, 1010.80272853, 1066.15621129, 1019.22608460, 1010.80272853, 1073.37623078, 1040.88614307, 1013.20940170, 1010.80272853, 1025.24276751, 1072.17289419, 1040.88614307, 1022.83609435, 992.75267981, 1022.83609435, 1018.02274802, 1010.80272853, 1005.98938221, 1031.25945042, 1028.85277726, 1030.05611384, 1063.74953812, 1058.93619180, 1046.90282598, 1052.91950889, 1003.58270904, 1012.00606512, 992.75267981, 1013.20940170, 1018.02274802, 1025.24276751, 1005.98938221, 1037.27613333, 998.76936272, 999.97269930, 1014.41273828, 992.75267981, 991.54934323, 1005.98938221, 951.83923604, 975.90596767, 980.71931399, 1003.58270904, 987.93933348, 999.97269930, 1001.17603588, 991.54934323, 938.60253364, 928.97584099, 947.02588971, 986.73599690, 986.73599690, 975.90596767, 957.85591894, 985.53266032, 974.70263109, 926.56916782, 973.49929450, 968.68594818, 979.51597741, 926.56916782, 830.30224130, 782.16877803, 848.35229002, 809.84551941, 765.32206589, 758.10204640, 788.18546094, 820.67554864, 802.62549992, 841.13227053, 776.15209512, 817.06553890, 817.06553890, 826.69223155, 758.10204640, 703.95190023, 791.79547068, 707.56190997, 842.33560711, 797.81215359, 850.75896318, 826.69223155, 803.82883650, 833.91225104, 850.75896318, 878.43570456, 861.58899242, 892.87574354, 900.09576303, 843.53894369, 844.74228027, 888.06239721, 842.33560711, 867.60567532, 842.33560711, 871.21568507, 853.16563635, 892.87574354, 884.45238747, 901.29909961, 894.07908012, 918.14581175, 941.00920680, 934.99252389, 968.68594818, 966.27927501, 992.75267981, 1022.83609435, 1026.44610409, 986.73599690, 1003.58270904, 989.14267007, 999.97269930, 997.56602614, 1010.80272853, 986.73599690, 1040.88614307, 1057.73285522, 1012.00606512, 1049.30949914, 1055.32618205, 1003.58270904, 1018.02274802, 1068.56288445, 1066.15621129, 1074.57956736, 1055.32618205, 1074.57956736, 1085.40959659, 1092.62961608, 1072.17289419, 1060.13952838, 1075.78290394, 1038.47946991, 1033.66612358, 993.95601639, 1008.39605537, 1052.91950889, 968.68594818, 979.51597741, 1043.29281624, 1074.57956736, 1052.91950889, 1034.86946017, 1014.41273828, 1033.66612358, 998.76936272, 984.32932374, 1033.66612358, 1004.78604563, 1025.24276751, 996.36268955, 932.58585073, 962.66926527, 885.65572405, 891.67240696, 878.43570456, 756.89870982, 748.47535375, 738.84866109, 782.16877803, 744.86534400, 741.25533426, 740.05199768, 730.42530502, 712.37525630, 658.22511013, 714.78192946, 659.42844671, 625.73502243, 594.44827130, 635.36171508, 626.93835901, 677.47849543, 726.81529528, 820.67554864, 799.01549017, 830.30224130, 832.70891446, 823.08222181, 819.47221206, 797.81215359, 760.50871956, 746.06868058, 749.67869033, 786.98212436, 675.07182227, 652.20842722, 589.63492498, 586.02491523, 566.77152993, 572.78821284, 527.06142274, 487.35131554], 
							[998.76936272, 989.14267007, 971.09262134, 942.21254338, 980.71931399, 944.61921655, 860.38565584, 806.23550966, 827.89556813, 770.13541222, 877.23236798, 850.75896318, 842.33560711, 818.26887548, 818.26887548, 770.13541222, 774.94875854, 699.13855390, 817.06553890, 812.25219257, 792.99880727, 729.22196844, 779.76210487, 747.27201717, 722.00194895, 844.74228027, 908.51911910, 977.10930425, 972.29595792, 960.26259211, 975.90596767, 903.70577277, 873.62235823, 895.28241670, 859.18231925, 819.47221206, 866.40233874, 862.79232900, 926.56916782, 922.95915808, 902.50243619, 962.66926527, 950.63589945, 921.75582150, 866.40233874, 845.94561686, 865.19900216, 844.74228027, 890.46907037, 889.26573379, 889.26573379, 890.46907037, 890.46907037, 885.65572405, 944.61921655, 1022.83609435, 1097.44296241, 1164.82981098, 1167.23648414, 1169.64315730, 1107.06965506, 1155.20311832, 1178.06651337, 1141.96641593, 1152.79644516, 1131.13638669, 1116.69634771, 1121.50969404, 1083.00292343, 1054.12284547, 1058.93619180, 1089.01960634, 1129.93305011, 1067.35954787, 1083.00292343, 1089.01960634, 1129.93305011, 1114.28967455, 1127.52637695, 1107.06965506, 1102.25630873, 1084.20626001, 1101.05297215, 1081.79958685, 1050.51283573, 1043.29281624, 1030.05611384, 962.66926527, 870.01234849, 878.43570456, 831.50557788, 782.16877803, 844.74228027, 842.33560711, 876.02903140, 830.30224130, 853.16563635, 851.96229976, 938.60253364, 974.70263109, 949.43256287, 1090.22294292, 1064.95287471, 981.92265058, 1028.85277726, 1034.86946017, 1048.10616256, 1070.96955761, 1004.78604563, 992.75267981, 1003.58270904, 1027.64944068, 1034.86946017, 1038.47946991, 1033.66612358, 1054.12284547, 1046.90282598, 1069.76622103, 1084.20626001, 1133.54305986, 1128.72971353, 1070.96955761, 1081.79958685, 1055.32618205, 1080.59625027, 1048.10616256, 1033.66612358, 1043.29281624, 1052.91950889, 1046.90282598, 1027.64944068, 1033.66612358, 986.73599690, 978.31264083, 971.09262134, 977.10930425, 986.73599690, 1003.58270904, 972.29595792, 974.70263109, 992.75267981, 941.00920680, 951.83923604, 900.09576303, 920.55248491, 959.05925553, 967.48261160, 1025.24276751, 1020.42942119, 997.56602614, 1004.78604563, 963.87260185, 936.19586048, 1009.59939195, 933.78918731, 912.12912884, 885.65572405, 859.18231925, 825.48889497, 841.13227053, 809.84551941, 794.20214385, 761.71205615, 805.03217308, 812.25219257, 824.28555839, 827.89556813, 801.42216334, 777.35543171, 737.64532451, 732.83197819, 696.73188074, 681.08850518, 722.00194895, 673.86848569, 743.66200742, 794.20214385, 773.74542196, 788.18546094, 821.87888522, 799.01549017, 886.85906063, 872.41902165, 877.23236798, 871.21568507, 895.28241670, 849.55562660, 872.41902165, 889.26573379, 843.53894369, 850.75896318, 812.25219257, 838.72559737, 805.03217308, 813.45552915, 854.36897293, 885.65572405, 953.04257262, 926.56916782, 932.58585073, 962.66926527, 986.73599690, 996.36268955, 991.54934323, 1004.78604563, 1010.80272853, 987.93933348, 986.73599690, 938.60253364, 918.14581175, 949.43256287, 996.36268955, 980.71931399, 1044.49615282, 1057.73285522, 1038.47946991, 1051.71617231, 1074.57956736, 1045.69948940, 1043.29281624, 1055.32618205, 1046.90282598, 1044.49615282, 1037.27613333, 1045.69948940, 1061.34286496, 1083.00292343, 1070.96955761, 1051.71617231, 1049.30949914, 1026.44610409, 1010.80272853, 1015.61607486, 1028.85277726, 1039.68280649, 986.73599690, 984.32932374, 1008.39605537, 1024.03943093, 1040.88614307, 1027.64944068, 1013.20940170, 1028.85277726, 992.75267981, 978.31264083, 1005.98938221, 1004.78604563, 1013.20940170, 960.26259211, 908.51911910, 921.75582150, 830.30224130, 819.47221206, 807.43884625, 720.79861237, 759.30538298, 729.22196844, 779.76210487, 726.81529528, 753.28870007, 796.60881701, 746.06868058, 737.64532451, 711.17191972, 734.03531477, 713.57859288, 688.30852467, 678.68183202, 709.96858314, 709.96858314, 776.15209512, 782.16877803, 833.91225104, 825.48889497, 835.11558762, 824.28555839, 815.86220232, 796.60881701, 789.38879752, 779.76210487, 776.15209512, 771.33874880, 780.96544145, 715.98526604, 703.95190023, 726.81529528, 669.05513936, 669.05513936, 665.44512962, 634.15837850, 620.92167610], 
							[1335.70360556, 1335.70360556, 1335.70360556, 1275.53677648, 1251.47004485, 1275.53677648, 1251.47004485, 1227.40331322, 1215.36994740, 1156.40645491, 1192.50655235, 1155.20311832, 1157.60979149, 1150.38977200, 1089.01960634, 1018.02274802, 878.43570456, 842.33560711, 953.04257262, 945.82255313, 892.87574354, 818.26887548, 865.19900216, 831.50557788, 818.26887548, 873.62235823, 1046.90282598, 1083.00292343, 1095.03628924, 1058.93619180, 1086.61293317, 996.36268955, 945.82255313, 1014.41273828, 992.75267981, 945.82255313, 981.92265058, 962.66926527, 998.76936272, 1042.08947966, 1045.69948940, 1040.88614307, 1107.06965506, 1034.86946017, 965.07593843, 939.80587022, 955.44924578, 991.54934323, 1083.00292343, 1092.62961608, 1052.91950889, 1076.98624052, 1105.86631848, 1110.67966481, 1263.50341067, 1150.38977200, 1275.53677648, 1191.30321577, 1227.40331322, 1275.53677648, 1227.40331322, 1299.60350811, 1251.47004485, 1215.36994740, 1215.36994740, 1192.50655235, 1170.84649388, 1144.37308909, 1135.94973302, 1150.38977200, 1173.25316705, 1172.04983047, 1179.26984996, 1149.18643542, 1181.67652312, 1175.65984021, 1185.28653286, 1203.33658159, 1187.69320603, 1227.40331322, 1181.67652312, 1134.74639644, 1143.16975251, 1137.15306960, 1166.03314756, 1083.00292343, 1060.13952838, 1026.44610409, 914.53580201, 843.53894369, 746.06868058, 647.39508089, 709.96858314, 685.90185151, 734.03531477, 695.52854416, 711.17191972, 782.16877803, 897.68908986, 902.50243619, 848.35229002, 973.49929450, 966.27927501, 878.43570456, 926.56916782, 961.46592869, 985.53266032, 1052.91950889, 1008.39605537, 955.44924578, 980.71931399, 999.97269930, 992.75267981, 968.68594818, 984.32932374, 967.48261160, 974.70263109, 1031.25945042, 1016.81941144, 1105.86631848, 1101.05297215, 1044.49615282, 1020.42942119, 1010.80272853, 1034.86946017, 1013.20940170, 1010.80272853, 1016.81941144, 1048.10616256, 1042.08947966, 1010.80272853, 1102.25630873, 1101.05297215, 1104.66298190, 1114.28967455, 1109.47632822, 1101.05297215, 1179.26984996, 1163.62647440, 1161.21980123, 1161.21980123, 1139.55974276, 1121.50969404, 1095.03628924, 1096.23962583, 1092.62961608, 1101.05297215, 1090.22294292, 1120.30635746, 1095.03628924, 1143.16975251, 1097.44296241, 1028.85277726, 1076.98624052, 1022.83609435, 987.93933348, 939.80587022, 888.06239721, 830.30224130, 853.16563635, 792.99880727, 782.16877803, 772.54208538, 812.25219257, 860.38565584, 866.40233874, 895.28241670, 806.23550966, 782.16877803, 738.84866109, 790.59213410, 748.47535375, 694.32520758, 728.01863186, 649.80175406, 786.98212436, 845.94561686, 863.99566558, 874.82569481, 878.43570456, 938.60253364, 1004.78604563, 954.24590920, 973.49929450, 1010.80272853, 1051.71617231, 959.05925553, 974.70263109, 968.68594818, 926.56916782, 926.56916782, 890.46907037, 901.29909961, 896.48575328, 973.49929450, 974.70263109, 980.71931399, 1028.85277726, 1137.15306960, 1131.13638669, 1161.21980123, 1239.43667904, 1167.23648414, 1227.40331322, 1227.40331322, 1263.50341067, 1263.50341067, 1311.63687393, 1263.50341067, 1251.47004485, 1359.77033719, 1359.77033719, 1468.07062954, 1456.03726372, 1444.00389791, 1383.83706883, 1444.00389791, 1395.87043464, 1371.80370301, 1383.83706883, 1468.07062954, 1456.03726372, 1468.07062954, 1419.93716627, 1468.07062954, 1516.20409280, 1468.07062954, 1456.03726372, 1371.80370301, 1395.87043464, 1371.80370301, 1335.70360556, 1359.77033719, 1335.70360556, 1371.80370301, 1227.40331322, 1227.40331322, 1299.60350811, 1335.70360556, 1395.87043464, 1383.83706883, 1323.67023975, 1359.77033719, 1335.70360556, 1359.77033719, 1359.77033719, 1335.70360556, 1335.70360556, 1299.60350811, 1203.33658159, 1227.40331322, 1172.04983047, 1099.84963557, 1111.88300139, 961.46592869, 1019.22608460, 992.75267981, 1070.96955761, 1075.78290394, 1323.67023975, 1215.36994740, 1203.33658159, 1153.99978174, 1187.69320603, 1173.25316705, 1131.13638669, 1104.66298190, 1157.60979149, 1191.30321577, 1193.70988894, 1215.36994740, 1203.33658159, 1251.47004485, 1239.43667904, 1227.40331322, 1239.43667904, 1275.53677648, 1227.40331322, 1215.36994740, 1155.20311832, 1138.35640618, 1191.30321577, 1152.79644516, 1046.90282598, 971.09262134, 912.12912884, 914.53580201, 890.46907037, 915.73913859, 883.24905089, 805.03217308], 
							[805.03217308, 809.84551941, 818.26887548, 707.56190997, 754.49203666, 732.83197819, 655.81843697, 632.95504192, 625.73502243, 573.99154942, 593.24493472, 593.24493472, 614.90499319, 619.71833952, 595.65160789, 584.82157865, 576.39822258, 593.24493472, 632.95504192, 614.90499319, 644.98840773, 625.73502243, 632.95504192, 618.51500294, 559.55151044, 589.63492498, 647.39508089, 644.98840773, 726.81529528, 728.01863186, 777.35543171, 722.00194895, 694.32520758, 740.05199768, 736.44198793, 689.51186125, 697.93521732, 673.86848569, 709.96858314, 679.88516860, 693.12187099, 683.49517834, 699.13855390, 673.86848569, 625.73502243, 613.70165661, 625.73502243, 593.24493472, 631.75170533, 661.83511987, 625.73502243, 648.59841748, 661.83511987, 660.63178329, 666.64846620, 665.44512962, 644.98840773, 652.20842722, 647.39508089, 689.51186125, 671.46181253, 703.95190023, 703.95190023, 746.06868058, 758.10204640, 800.21882676, 786.98212436, 754.49203666, 736.44198793, 722.00194895, 722.00194895, 752.08536349, 756.89870982, 715.98526604, 743.66200742, 756.89870982, 753.28870007, 776.15209512, 780.96544145, 782.16877803, 746.06868058, 732.83197819, 737.64532451, 743.66200742, 728.01863186, 713.57859288, 685.90185151, 651.00509064, 540.29812513, 537.89145197, 535.48477881, 464.48792049, 504.19802769, 498.18134478, 563.16152018, 517.43473008, 504.19802769, 559.55151044, 595.65160789, 553.53482753, 541.50146171, 595.65160789, 610.09164687, 575.19488600, 572.78821284, 571.58487625, 589.63492498, 671.46181253, 640.17506140, 617.31166635, 625.73502243, 631.75170533, 613.70165661, 607.68497370, 610.09164687, 608.88831028, 620.92167610, 626.93835901, 636.56505166, 629.34503217, 642.58173457, 608.88831028, 590.83826156, 580.00823233, 589.63492498, 583.61824207, 581.21156891, 584.82157865, 590.83826156, 570.38153967, 563.16152018, 557.14483728, 542.70479830, 509.01137401, 505.40136427, 499.38468136, 515.02805692, 543.90813488, 583.61824207, 563.16152018, 566.77152993, 541.50146171, 547.51814462, 496.97800820, 515.02805692, 489.75798871, 487.35131554, 507.80803743, 533.07810564, 529.46809590, 523.45141299, 495.77467161, 478.92795947, 489.75798871, 494.57133503, 493.36799845, 478.92795947, 460.87791075, 446.43787177, 453.65789126, 419.96446697, 422.37114014, 406.72776458, 404.32109141, 441.62452544, 440.42118886, 456.06456442, 445.23453519, 427.18448646, 403.11775483, 415.15112065, 374.23767687, 367.01765738, 383.86436953, 330.91755994, 405.52442800, 430.79449621, 419.96446697, 435.60784253, 423.57447672, 417.55779381, 471.70793998, 506.60470085, 577.60155916, 604.07496396, 691.91853441, 626.93835901, 613.70165661, 632.95504192, 605.27830054, 643.78507115, 619.71833952, 712.37525630, 770.13541222, 764.11872931, 806.23550966, 871.21568507, 894.07908012, 998.76936272, 922.95915808, 920.55248491, 956.65258236, 981.92265058, 904.90910935, 914.53580201, 938.60253364, 908.51911910, 830.30224130, 830.30224130, 833.91225104, 848.35229002, 825.48889497, 856.77564609, 857.97898267, 902.50243619, 885.65572405, 874.82569481, 874.82569481, 848.35229002, 830.30224130, 807.43884625, 820.67554864, 812.25219257, 767.72873905, 806.23550966, 818.26887548, 847.14895344, 844.74228027, 803.82883650, 791.79547068, 772.54208538, 728.01863186, 682.29184176, 720.79861237, 747.27201717, 709.96858314, 707.56190997, 734.03531477, 743.66200742, 776.15209512, 766.52540247, 736.44198793, 768.93207563, 734.03531477, 711.17191972, 747.27201717, 734.03531477, 746.06868058, 706.35857339, 657.02177355, 655.81843697, 628.14169559, 607.68497370, 601.66829079, 577.60155916, 582.41490549, 573.99154942, 598.05828105, 567.97486651, 592.04159814, 602.87162738, 619.71833952, 629.34503217, 631.75170533, 660.63178329, 622.12501268, 644.98840773, 614.90499319, 655.81843697, 661.83511987, 707.56190997, 709.96858314, 732.83197819, 715.98526604, 717.18860263, 715.98526604, 719.59527579, 696.73188074, 689.51186125, 673.86848569, 651.00509064, 673.86848569, 672.66514911, 632.95504192, 607.68497370, 564.36485676, 537.89145197, 533.07810564, 601.66829079, 571.58487625, 576.39822258], 
							[754.49203666, 726.81529528, 700.34189048, 702.74856365, 709.96858314, 690.71519783, 684.69851492, 697.93521732, 695.52854416, 691.91853441, 708.76524656, 706.35857339, 703.95190023, 712.37525630, 685.90185151, 695.52854416, 653.41176380, 637.76838824, 673.86848569, 653.41176380, 649.80175406, 607.68497370, 632.95504192, 606.48163712, 617.31166635, 669.05513936, 713.57859288, 697.93521732, 667.85180278, 659.42844671, 677.47849543, 676.27515885, 659.42844671, 677.47849543, 659.42844671, 635.36171508, 619.71833952, 598.05828105, 636.56505166, 649.80175406, 655.81843697, 664.24179304, 663.03845645, 658.22511013, 637.76838824, 636.56505166, 618.51500294, 613.70165661, 635.36171508, 618.51500294, 602.87162738, 608.88831028, 607.68497370, 589.63492498, 652.20842722, 646.19174431, 722.00194895, 792.99880727, 791.79547068, 788.18546094, 779.76210487, 853.16563635, 883.24905089, 849.55562660, 883.24905089, 862.79232900, 862.79232900, 849.55562660, 800.21882676, 782.16877803, 773.74542196, 791.79547068, 807.43884625, 779.76210487, 794.20214385, 801.42216334, 857.97898267, 814.65886573, 817.06553890, 821.87888522, 817.06553890, 824.28555839, 817.06553890, 785.77878778, 829.09890471, 821.87888522, 824.28555839, 812.25219257, 785.77878778, 780.96544145, 774.94875854, 728.01863186, 780.96544145, 770.13541222, 770.13541222, 752.08536349, 740.05199768, 764.11872931, 808.64218283, 806.23550966, 801.42216334, 853.16563635, 843.53894369, 811.04885599, 842.33560711, 850.75896318, 853.16563635, 836.31892420, 806.23550966, 783.37211461, 818.26887548, 818.26887548, 836.31892420, 812.25219257, 823.08222181, 838.72559737, 851.96229976, 901.29909961, 921.75582150, 986.73599690, 960.26259211, 956.65258236, 962.66926527, 962.66926527, 979.51597741, 943.41587996, 934.99252389, 974.70263109, 969.88928476, 961.46592869, 938.60253364, 959.05925553, 921.75582150, 894.07908012, 920.55248491, 926.56916782, 938.60253364, 955.44924578, 934.99252389, 932.58585073, 936.19586048, 892.87574354, 902.50243619, 872.41902165, 888.06239721, 861.58899242, 876.02903140, 894.07908012, 914.53580201, 902.50243619, 867.60567532, 802.62549992, 812.25219257, 805.03217308, 784.57545120, 742.45867084, 726.81529528, 758.10204640, 730.42530502, 700.34189048, 667.85180278, 749.67869033, 743.66200742, 740.05199768, 761.71205615, 734.03531477, 731.62864161, 714.78192946, 713.57859288, 709.96858314, 714.78192946, 678.68183202, 655.81843697, 711.17191972, 693.12187099, 742.45867084, 758.10204640, 750.88202691, 752.08536349, 776.15209512, 785.77878778, 849.55562660, 836.31892420, 838.72559737, 890.46907037, 948.22922629, 867.60567532, 914.53580201, 968.68594818, 932.58585073, 954.24590920, 936.19586048, 968.68594818, 955.44924578, 959.05925553, 975.90596767, 990.34600665, 1019.22608460, 992.75267981, 992.75267981, 990.34600665, 1038.47946991, 1010.80272853, 995.15935297, 1014.41273828, 1024.03943093, 1042.08947966, 1046.90282598, 1033.66612358, 1025.24276751, 1022.83609435, 1026.44610409, 1068.56288445, 1111.88300139, 1113.08633797, 1151.59310858, 1119.10302088, 1138.35640618, 1117.89968430, 1092.62961608, 1115.49301113, 1107.06965506, 1117.89968430, 1103.45964532, 1143.16975251, 1132.33972327, 1146.77976225, 1149.18643542, 1114.28967455, 1095.03628924, 1080.59625027, 1099.84963557, 1075.78290394, 1076.98624052, 1091.42627950, 1033.66612358, 1051.71617231, 1048.10616256, 1046.90282598, 1080.59625027, 1070.96955761, 1064.95287471, 1092.62961608, 1064.95287471, 1063.74953812, 1078.18957710, 1075.78290394, 1104.66298190, 1119.10302088, 1105.86631848, 1111.88300139, 1099.84963557, 1103.45964532, 1107.06965506, 1024.03943093, 1042.08947966, 1028.85277726, 1050.51283573, 1019.22608460, 1075.78290394, 1103.45964532, 1045.69948940, 984.32932374, 920.55248491, 1016.81941144, 986.73599690, 980.71931399, 968.68594818, 974.70263109, 1022.83609435, 986.73599690, 998.76936272, 1031.25945042, 1038.47946991, 1008.39605537, 1028.85277726, 1018.02274802, 1024.03943093, 1048.10616256, 1048.10616256, 1032.46278700, 1002.37937246, 1014.41273828, 989.14267007, 997.56602614, 961.46592869, 966.27927501, 961.46592869, 925.36583124, 890.46907037, 768.93207563], 
							[576.50652287, 611.51158403, 581.98170432, 578.69659545, 560.75484702, 506.60470085, 510.21471059, 490.96132529, 468.09793024, 521.04473983, 563.16152018, 518.63806666, 535.48477881, 576.39822258, 559.55151044, 553.53482753, 513.82472034, 495.77467161, 528.26475932, 524.65474957, 493.36799845, 463.28458391, 436.81117912, 417.55779381, 381.45769636, 445.23453519, 529.46809590, 540.29812513, 522.24807641, 522.24807641, 527.06142274, 493.36799845, 480.13129605, 488.55465212, 475.31794973, 450.04788151, 458.47123758, 388.67771585, 429.59115963, 439.21785228, 421.16780356, 422.37114014, 433.20116937, 410.33777432, 393.49106218, 400.71108167, 381.45769636, 377.84768662, 403.11775483, 421.16780356, 411.54111090, 418.76113039, 407.93110116, 383.86436953, 409.13443774, 389.88105243, 409.13443774, 434.40450595, 459.67457417, 476.52128631, 466.89459366, 492.16466187, 493.36799845, 490.96132529, 493.36799845, 511.41804717, 513.82472034, 502.99469110, 480.13129605, 471.70793998, 454.86122784, 451.25121810, 463.28458391, 453.65789126, 470.50460340, 466.89459366, 465.69125707, 459.67457417, 476.52128631, 471.70793998, 465.69125707, 465.69125707, 452.45455468, 431.99783279, 451.25121810, 442.82786202, 441.62452544, 399.50774509, 373.03434029, 412.74444748, 373.03434029, 356.18762815, 371.83100371, 347.76427208, 362.20431106, 336.93424284, 358.59430131, 358.59430131, 379.05102320, 383.86436953, 377.84768662, 397.10107192, 427.18448646, 389.88105243, 415.15112065, 409.13443774, 415.15112065, 456.06456442, 472.91127656, 438.01451570, 452.45455468, 457.26790100, 466.89459366, 471.70793998, 452.45455468, 453.65789126, 446.43787177, 462.08124733, 480.13129605, 494.57133503, 489.75798871, 463.28458391, 451.25121810, 454.86122784, 457.26790100, 440.42118886, 431.99783279, 441.62452544, 468.09793024, 463.28458391, 470.50460340, 480.13129605, 450.04788151, 451.25121810, 462.08124733, 460.87791075, 478.92795947, 493.36799845, 502.99469110, 511.41804717, 506.60470085, 487.35131554, 502.99469110, 490.96132529, 475.31794973, 474.11461315, 469.30126682, 489.75798871, 483.74130580, 492.16466187, 472.91127656, 445.23453519, 441.62452544, 466.89459366, 458.47123758, 457.26790100, 429.59115963, 419.96446697, 453.65789126, 448.84454493, 439.21785228, 523.45141299, 492.16466187, 499.38468136, 523.45141299, 537.89145197, 553.53482753, 524.65474957, 480.13129605, 466.89459366, 458.47123758, 427.18448646, 400.71108167, 406.72776458, 383.86436953, 446.43787177, 445.23453519, 445.23453519, 448.84454493, 435.60784253, 429.59115963, 476.52128631, 463.28458391, 480.13129605, 458.47123758, 517.43473008, 509.01137401, 522.24807641, 557.14483728, 549.92481779, 559.55151044, 547.51814462, 547.51814462, 549.92481779, 558.34817386, 545.11147146, 555.94150069, 588.43158840, 575.19488600, 577.60155916, 598.05828105, 594.44827130, 582.41490549, 575.19488600, 577.60155916, 596.85494447, 599.26161763, 592.04159814, 586.02491523, 567.97486651, 569.17820309, 541.50146171, 548.72148120, 576.39822258, 581.21156891, 586.02491523, 594.44827130, 598.05828105, 573.99154942, 558.34817386, 571.58487625, 565.56819335, 558.34817386, 541.50146171, 577.60155916, 587.22825181, 586.02491523, 586.02491523, 573.99154942, 565.56819335, 564.36485676, 542.70479830, 530.67143248, 548.72148120, 548.72148120, 539.09478855, 523.45141299, 545.11147146, 554.73816411, 561.95818360, 535.48477881, 539.09478855, 555.94150069, 530.67143248, 530.67143248, 555.94150069, 534.28144222, 545.11147146, 516.23139350, 480.13129605, 499.38468136, 493.36799845, 474.11461315, 477.72462289, 429.59115963, 463.28458391, 416.35445723, 439.21785228, 423.57447672, 435.60784253, 472.91127656, 447.64120835, 431.99783279, 407.93110116, 406.72776458, 405.52442800, 393.49106218, 410.33777432, 460.87791075, 486.14797896, 500.58801794, 481.33463264, 472.91127656, 458.47123758, 480.13129605, 492.16466187, 487.35131554, 470.50460340, 478.92795947, 446.43787177, 424.77781330, 456.06456442, 466.89459366, 421.16780356, 387.47437927, 380.25435978, 391.08438902, 369.42433055, 379.05102320, 368.22099397, 351.37428182], 
							[2021.60545707, 1985.50535962, 1937.37189636, 1889.23843309, 1913.30516472, 1913.30516472, 1660.60448259, 1576.37092188, 1492.13736117, 1407.90380046, 1516.20409280, 1552.30419025, 1492.13736117, 1383.83706883, 1335.70360556, 1227.40331322, 1045.69948940, 1038.47946991, 1064.95287471, 1125.11970378, 1058.93619180, 998.76936272, 1131.13638669, 1095.03628924, 1046.90282598, 1119.10302088, 1299.60350811, 1480.10399535, 1504.17072698, 1528.23745862, 1600.43765351, 1480.10399535, 1383.83706883, 1456.03726372, 1383.83706883, 1311.63687393, 1431.97053209, 1311.63687393, 1456.03726372, 1504.17072698, 1480.10399535, 1552.30419025, 1468.07062954, 1407.90380046, 1335.70360556, 1323.67023975, 1371.80370301, 1323.67023975, 1347.73697138, 1456.03726372, 1395.87043464, 1444.00389791, 1468.07062954, 1576.37092188, 1792.97150657, 1961.43862799, 1853.13833565, 2021.60545707, 1901.27179891, 2069.73892033, 2009.57209125, 2117.87238359, 2081.77228615, 2021.60545707, 2033.63882288, 2117.87238359, 2153.97248104, 2021.60545707, 2009.57209125, 1937.37189636, 2009.57209125, 1985.50535962, 2033.63882288, 1985.50535962, 2069.73892033, 2129.90574941, 2153.97248104, 2141.93911523, 2214.13931012, 2298.37287083, 2190.07257849, 2190.07257849, 2214.13931012, 2166.00584686, 2081.77228615, 1997.53872544, 2021.60545707, 1925.33853054, 1744.83804330, 1696.70458004, 1744.83804330, 1624.50438514, 1672.63784841, 1696.70458004, 1841.10496983, 1672.63784841, 1744.83804330, 1853.13833565, 2117.87238359, 2166.00584686, 2057.70555452, 2478.87335807, 2214.13931012, 2057.70555452, 2274.30613920, 2286.33950502, 2238.20604175, 2298.37287083, 2250.23940757, 2153.97248104, 2274.30613920, 2286.33950502, 2166.00584686, 2117.87238359, 2166.00584686, 2226.17267594, 2274.30613920, 2141.93911523, 2310.40623665, 2358.53969991, 2406.67316318, 2226.17267594, 2214.13931012, 2178.03921267, 2262.27277339, 2202.10594431, 2202.10594431, 2226.17267594, 2238.20604175, 2190.07257849, 2045.67218870, 2093.80565196, 1841.10496983, 1865.17170146, 1877.20506728, 1780.93814075, 1744.83804330, 1889.23843309, 1901.27179891, 1913.30516472, 1901.27179891, 1805.00487238, 1805.00487238, 1708.73794585, 1792.97150657, 1756.87140912, 1865.17170146, 1961.43862799, 1985.50535962, 1961.43862799, 1997.53872544, 1973.47199380, 1768.90477493, 1925.33853054, 1853.13833565, 1792.97150657, 1708.73794585, 1588.40428770, 1564.33755606, 1552.30419025, 1492.13736117, 1456.03726372, 1468.07062954, 1407.90380046, 1576.37092188, 1588.40428770, 1636.53775096, 1528.23745862, 1456.03726372, 1311.63687393, 1359.77033719, 1215.36994740, 1172.04983047, 1215.36994740, 1143.16975251, 1323.67023975, 1395.87043464, 1444.00389791, 1444.00389791, 1468.07062954, 1407.90380046, 1588.40428770, 1504.17072698, 1564.33755606, 1612.47101933, 1841.10496983, 1624.50438514, 1720.77131167, 1720.77131167, 1672.63784841, 1684.67121422, 1516.20409280, 1612.47101933, 1564.33755606, 1672.63784841, 1624.50438514, 1792.97150657, 1792.97150657, 1792.97150657, 1829.07160401, 1780.93814075, 1841.10496983, 1780.93814075, 1841.10496983, 1901.27179891, 1817.03823820, 1853.13833565, 1841.10496983, 1768.90477493, 1756.87140912, 1829.07160401, 1780.93814075, 1901.27179891, 1973.47199380, 2081.77228615, 1997.53872544, 2033.63882288, 2033.63882288, 1985.50535962, 1925.33853054, 1997.53872544, 1961.43862799, 2009.57209125, 1985.50535962, 2081.77228615, 2069.73892033, 2129.90574941, 2045.67218870, 2009.57209125, 2033.63882288, 2021.60545707, 1949.40526217, 1961.43862799, 2021.60545707, 2045.67218870, 1853.13833565, 1817.03823820, 2033.63882288, 1973.47199380, 2021.60545707, 1973.47199380, 1937.37189636, 2021.60545707, 1913.30516472, 1877.20506728, 2009.57209125, 1925.33853054, 2009.57209125, 1961.43862799, 1865.17170146, 1889.23843309, 1805.00487238, 1624.50438514, 1624.50438514, 1359.77033719, 1419.93716627, 1311.63687393, 1359.77033719, 1299.60350811, 1323.67023975, 1419.93716627, 1359.77033719, 1275.53677648, 1239.43667904, 1347.73697138, 1251.47004485, 1158.81312807, 986.73599690, 1031.25945042, 1045.69948940, 1155.20311832, 1203.33658159, 1311.63687393, 1287.57014230, 1323.67023975, 1407.90380046, 1383.83706883, 1335.70360556, 1359.77033719, 1275.53677648, 1215.36994740, 1263.50341067, 1227.40331322, 1131.13638669, 1096.23962583, 1025.24276751, 978.31264083, 963.87260185, 980.71931399, 936.19586048, 854.36897293], 
							[1197.31989868, 1167.23648414, 1126.32304037, 999.97269930, 1036.07279675, 1022.83609435, 892.87574354, 1046.90282598, 985.53266032, 953.04257262, 986.73599690, 1022.83609435, 1028.85277726, 979.51597741, 951.83923604, 904.90910935, 890.46907037, 898.89242645, 902.50243619, 889.26573379, 843.53894369, 823.08222181, 797.81215359, 782.16877803, 777.35543171, 815.86220232, 928.97584099, 927.77250440, 945.82255313, 904.90910935, 969.88928476, 951.83923604, 902.50243619, 908.51911910, 877.23236798, 876.02903140, 880.84237772, 821.87888522, 884.45238747, 916.94247517, 892.87574354, 920.55248491, 885.65572405, 888.06239721, 854.36897293, 842.33560711, 865.19900216, 877.23236798, 878.43570456, 937.39919706, 891.67240696, 908.51911910, 888.06239721, 884.45238747, 898.89242645, 939.80587022, 1025.24276751, 975.90596767, 1004.78604563, 1095.03628924, 1022.83609435, 1119.10302088, 1119.10302088, 1068.56288445, 1109.47632822, 1141.96641593, 1188.89654261, 1196.11656210, 1151.59310858, 1129.93305011, 1147.98309883, 1149.18643542, 1173.25316705, 1122.71303062, 1167.23648414, 1153.99978174, 1160.01646465, 1116.69634771, 1139.55974276, 1107.06965506, 1117.89968430, 1102.25630873, 1131.13638669, 1143.16975251, 1155.20311832, 1113.08633797, 1058.93619180, 1074.57956736, 1064.95287471, 1010.80272853, 968.68594818, 872.41902165, 938.60253364, 914.53580201, 991.54934323, 938.60253364, 957.85591894, 1022.83609435, 1048.10616256, 1048.10616256, 1034.86946017, 1141.96641593, 1102.25630873, 1046.90282598, 1155.20311832, 1113.08633797, 1141.96641593, 1139.55974276, 1141.96641593, 1135.94973302, 1129.93305011, 1141.96641593, 1147.98309883, 1149.18643542, 1149.18643542, 1153.99978174, 1215.36994740, 1187.69320603, 1263.50341067, 1287.57014230, 1311.63687393, 1299.60350811, 1263.50341067, 1299.60350811, 1299.60350811, 1227.40331322, 1202.13324501, 1239.43667904, 1227.40331322, 1263.50341067, 1203.33658159, 1203.33658159, 1215.36994740, 1191.30321577, 1202.13324501, 1239.43667904, 1227.40331322, 1275.53677648, 1287.57014230, 1287.57014230, 1299.60350811, 1203.33658159, 1239.43667904, 1203.33658159, 1215.36994740, 1215.36994740, 1227.40331322, 1227.40331322, 1251.47004485, 1251.47004485, 1203.33658159, 1173.25316705, 1083.00292343, 1092.62961608, 1117.89968430, 1086.61293317, 1107.06965506, 1039.68280649, 971.09262134, 956.65258236, 928.97584099, 948.22922629, 947.02588971, 954.24590920, 1004.78604563, 984.32932374, 997.56602614, 904.90910935, 878.43570456, 867.60567532, 897.68908986, 786.98212436, 746.06868058, 766.52540247, 715.98526604, 906.11244594, 936.19586048, 944.61921655, 932.58585073, 962.66926527, 909.72245568, 1004.78604563, 998.76936272, 1103.45964532, 1083.00292343, 1076.98624052, 980.71931399, 1064.95287471, 1070.96955761, 1010.80272853, 992.75267981, 947.02588971, 980.71931399, 954.24590920, 967.48261160, 980.71931399, 1034.86946017, 1137.15306960, 1173.25316705, 1178.06651337, 1174.45650363, 1203.33658159, 1263.50341067, 1251.47004485, 1215.36994740, 1275.53677648, 1299.60350811, 1359.77033719, 1311.63687393, 1299.60350811, 1335.70360556, 1323.67023975, 1323.67023975, 1347.73697138, 1431.97053209, 1419.93716627, 1480.10399535, 1480.10399535, 1407.90380046, 1407.90380046, 1431.97053209, 1383.83706883, 1323.67023975, 1311.63687393, 1335.70360556, 1347.73697138, 1323.67023975, 1311.63687393, 1287.57014230, 1215.36994740, 1227.40331322, 1203.33658159, 1180.47318654, 1196.11656210, 1239.43667904, 1125.11970378, 1251.47004485, 1287.57014230, 1287.57014230, 1335.70360556, 1263.50341067, 1287.57014230, 1323.67023975, 1239.43667904, 1227.40331322, 1287.57014230, 1215.36994740, 1215.36994740, 1227.40331322, 1192.50655235, 1287.57014230, 1227.40331322, 1227.40331322, 1203.33658159, 998.76936272, 1103.45964532, 1061.34286496, 1096.23962583, 1080.59625027, 1137.15306960, 1117.89968430, 1113.08633797, 1098.64629899, 1046.90282598, 1095.03628924, 1024.03943093, 1018.02274802, 1004.78604563, 1054.12284547, 1044.49615282, 1092.62961608, 1116.69634771, 1167.23648414, 1108.27299164, 1161.21980123, 1191.30321577, 1251.47004485, 1202.13324501, 1197.31989868, 1172.04983047, 1147.98309883, 1143.16975251, 1174.45650363, 1227.40331322, 1132.33972327, 1070.96955761, 974.70263109, 955.44924578, 992.75267981, 932.58585073, 891.67240696], 
							[948.22922629, 944.61921655, 948.22922629, 934.99252389, 939.80587022, 902.50243619, 836.31892420, 752.08536349, 728.01863186, 728.01863186, 835.11558762, 824.28555839, 784.57545120, 779.76210487, 765.32206589, 723.20528553, 719.59527579, 673.86848569, 713.57859288, 709.96858314, 722.00194895, 709.96858314, 759.30538298, 752.08536349, 708.76524656, 737.64532451, 915.73913859, 919.34914833, 912.12912884, 921.75582150, 933.78918731, 910.92579226, 920.55248491, 928.97584099, 934.99252389, 921.75582150, 920.55248491, 891.67240696, 925.36583124, 960.26259211, 973.49929450, 962.66926527, 1004.78604563, 1007.19271879, 951.83923604, 950.63589945, 902.50243619, 854.36897293, 898.89242645, 916.94247517, 909.72245568, 937.39919706, 942.21254338, 920.55248491, 961.46592869, 968.68594818, 967.48261160, 1057.73285522, 1083.00292343, 1057.73285522, 1076.98624052, 1119.10302088, 1089.01960634, 1083.00292343, 1092.62961608, 1097.44296241, 1060.13952838, 1061.34286496, 1048.10616256, 1013.20940170, 1028.85277726, 1025.24276751, 1030.05611384, 960.26259211, 992.75267981, 974.70263109, 984.32932374, 975.90596767, 965.07593843, 979.51597741, 959.05925553, 936.19586048, 908.51911910, 910.92579226, 921.75582150, 908.51911910, 891.67240696, 888.06239721, 860.38565584, 838.72559737, 826.69223155, 765.32206589, 800.21882676, 776.15209512, 820.67554864, 806.23550966, 779.76210487, 805.03217308, 885.65572405, 861.58899242, 851.96229976, 926.56916782, 913.33246543, 865.19900216, 884.45238747, 859.18231925, 867.60567532, 860.38565584, 821.87888522, 809.84551941, 835.11558762, 843.53894369, 844.74228027, 841.13227053, 836.31892420, 884.45238747, 886.85906063, 908.51911910, 879.63904114, 902.50243619, 892.87574354, 861.58899242, 865.19900216, 860.38565584, 854.36897293, 862.79232900, 930.17917757, 944.61921655, 943.41587996, 930.17917757, 963.87260185, 931.38251415, 900.09576303, 908.51911910, 908.51911910, 890.46907037, 891.67240696, 916.94247517, 914.53580201, 892.87574354, 874.82569481, 815.86220232, 854.36897293, 842.33560711, 862.79232900, 872.41902165, 882.04571430, 878.43570456, 870.01234849, 863.99566558, 836.31892420, 860.38565584, 884.45238747, 891.67240696, 866.40233874, 873.62235823, 825.48889497, 788.18546094, 703.95190023, 740.05199768, 730.42530502, 832.70891446, 811.04885599, 797.81215359, 806.23550966, 800.21882676, 824.28555839, 813.45552915, 808.64218283, 776.15209512, 758.10204640, 694.32520758, 670.25847594, 685.90185151, 619.71833952, 712.37525630, 679.88516860, 703.95190023, 687.10518809, 665.44512962, 648.59841748, 734.03531477, 738.84866109, 752.08536349, 734.03531477, 790.59213410, 712.37525630, 730.42530502, 764.11872931, 756.89870982, 797.81215359, 725.61195870, 805.03217308, 786.98212436, 780.96544145, 805.03217308, 805.03217308, 806.23550966, 790.59213410, 825.48889497, 830.30224130, 825.48889497, 836.31892420, 825.48889497, 854.36897293, 889.26573379, 872.41902165, 860.38565584, 830.30224130, 818.26887548, 824.28555839, 831.50557788, 851.96229976, 907.31578252, 924.16249466, 878.43570456, 903.70577277, 896.48575328, 898.89242645, 903.70577277, 927.77250440, 937.39919706, 949.43256287, 942.21254338, 957.85591894, 962.66926527, 971.09262134, 953.04257262, 941.00920680, 953.04257262, 918.14581175, 920.55248491, 920.55248491, 909.72245568, 914.53580201, 856.77564609, 873.62235823, 895.28241670, 938.60253364, 927.77250440, 918.14581175, 902.50243619, 902.50243619, 883.24905089, 879.63904114, 900.09576303, 874.82569481, 854.36897293, 848.35229002, 776.15209512, 784.57545120, 752.08536349, 711.17191972, 711.17191972, 614.90499319, 669.05513936, 673.86848569, 729.22196844, 669.05513936, 697.93521732, 700.34189048, 691.91853441, 660.63178329, 643.78507115, 729.22196844, 642.58173457, 607.68497370, 617.31166635, 661.83511987, 623.32834926, 660.63178329, 719.59527579, 720.79861237, 728.01863186, 688.30852467, 700.34189048, 683.49517834, 658.22511013, 661.83511987, 614.90499319, 586.02491523, 565.56819335, 582.41490549, 554.73816411, 477.72462289, 457.26790100, 433.20116937, 430.79449621, 451.25121810, 427.18448646, 406.72776458], 
							[643.78507115, 634.15837850, 577.60155916, 608.88831028, 607.68497370, 530.67143248, 523.45141299, 577.60155916, 571.58487625, 543.90813488, 605.27830054, 571.58487625, 589.63492498, 583.61824207, 554.73816411, 543.90813488, 523.45141299, 494.57133503, 513.82472034, 519.84140325, 517.43473008, 486.14797896, 516.23139350, 488.55465212, 471.70793998, 507.80803743, 564.36485676, 561.95818360, 525.85808615, 511.41804717, 552.33149095, 529.46809590, 523.45141299, 542.70479830, 511.41804717, 511.41804717, 511.41804717, 496.97800820, 539.09478855, 543.90813488, 547.51814462, 535.48477881, 546.31480804, 546.31480804, 533.07810564, 527.06142274, 515.02805692, 506.60470085, 529.46809590, 523.45141299, 505.40136427, 512.62138376, 521.04473983, 504.19802769, 529.46809590, 527.06142274, 531.87476906, 673.86848569, 634.15837850, 631.75170533, 619.71833952, 661.83511987, 638.97172482, 620.92167610, 637.76838824, 628.14169559, 622.12501268, 620.92167610, 610.09164687, 623.32834926, 622.12501268, 620.92167610, 617.31166635, 631.75170533, 636.56505166, 631.75170533, 654.61510038, 623.32834926, 625.73502243, 622.12501268, 619.71833952, 612.49832003, 616.10832977, 610.09164687, 641.37839799, 631.75170533, 643.78507115, 640.17506140, 613.70165661, 619.71833952, 618.51500294, 572.78821284, 601.66829079, 581.21156891, 600.46495421, 577.60155916, 571.58487625, 583.61824207, 625.73502243, 612.49832003, 583.61824207, 628.14169559, 613.70165661, 605.27830054, 612.49832003, 614.90499319, 613.70165661, 604.07496396, 586.02491523, 589.63492498, 607.68497370, 586.02491523, 593.24493472, 605.27830054, 596.85494447, 595.65160789, 596.85494447, 596.85494447, 581.21156891, 573.99154942, 572.78821284, 554.73816411, 548.72148120, 545.11147146, 549.92481779, 535.48477881, 554.73816411, 539.09478855, 539.09478855, 535.48477881, 539.09478855, 535.48477881, 524.65474957, 519.84140325, 523.45141299, 518.63806666, 499.38468136, 494.57133503, 494.57133503, 493.36799845, 494.57133503, 489.75798871, 483.74130580, 472.91127656, 478.92795947, 458.47123758, 458.47123758, 481.33463264, 481.33463264, 470.50460340, 468.09793024, 411.54111090, 434.40450595, 445.23453519, 445.23453519, 445.23453519, 439.21785228, 431.99783279, 418.76113039, 415.15112065, 415.15112065, 412.74444748, 403.11775483, 412.74444748, 422.37114014, 415.15112065, 423.57447672, 410.33777432, 419.96446697, 399.50774509, 388.67771585, 385.06770611, 379.05102320, 393.49106218, 376.64435004, 407.93110116, 399.50774509, 401.91441825, 388.67771585, 382.66103294, 377.84768662, 395.89773534, 388.67771585, 391.08438902, 392.28772560, 409.13443774, 394.69439876, 406.72776458, 412.74444748, 407.93110116, 418.76113039, 401.91441825, 413.94778407, 401.91441825, 410.33777432, 398.30440851, 417.55779381, 417.55779381, 421.16780356, 433.20116937, 429.59115963, 444.03119861, 434.40450595, 454.86122784, 459.67457417, 468.09793024, 456.06456442, 456.06456442, 453.65789126, 456.06456442, 436.81117912, 444.03119861, 453.65789126, 465.69125707, 475.31794973, 472.91127656, 471.70793998, 498.18134478, 505.40136427, 489.75798871, 488.55465212, 505.40136427, 506.60470085, 487.35131554, 474.11461315, 489.75798871, 482.53796922, 472.91127656, 464.48792049, 462.08124733, 469.30126682, 475.31794973, 474.11461315, 469.30126682, 453.65789126, 450.04788151, 442.82786202, 444.03119861, 444.03119861, 451.25121810, 434.40450595, 429.59115963, 436.81117912, 427.18448646, 444.03119861, 439.21785228, 428.38782305, 429.59115963, 422.37114014, 397.10107192, 397.10107192, 385.06770611, 381.45769636, 381.45769636, 348.96760866, 361.00097448, 362.20431106, 367.01765738, 351.37428182, 351.37428182, 353.78095499, 357.39096473, 351.37428182, 338.13757943, 357.39096473, 353.78095499, 341.74758917, 348.96760866, 357.39096473, 363.40764764, 359.79763789, 359.79763789, 363.40764764, 371.83100371, 368.22099397, 363.40764764, 371.83100371, 371.83100371, 387.47437927, 393.49106218, 382.66103294, 363.40764764, 353.78095499, 336.93424284, 336.93424284, 327.30755019, 326.10421361, 332.12089652, 338.13757943, 350.17094524, 346.56093550], 
							[1792.97150657, 1768.90477493, 1768.90477493, 1829.07160401, 1877.20506728, 1841.10496983, 1889.23843309, 1913.30516472, 1913.30516472, 1877.20506728, 1901.27179891, 1768.90477493, 1792.97150657, 1829.07160401, 1744.83804330, 1744.83804330, 1636.53775096, 1624.50438514, 1612.47101933, 1492.13736117, 1444.00389791, 1383.83706883, 1552.30419025, 1540.27082443, 1456.03726372, 1552.30419025, 1588.40428770, 1564.33755606, 1636.53775096, 1720.77131167, 1636.53775096, 1504.17072698, 1492.13736117, 1564.33755606, 1588.40428770, 1528.23745862, 1492.13736117, 1492.13736117, 1492.13736117, 1660.60448259, 1564.33755606, 1600.43765351, 1660.60448259, 1672.63784841, 1564.33755606, 1540.27082443, 1504.17072698, 1528.23745862, 1588.40428770, 1660.60448259, 1588.40428770, 1552.30419025, 1600.43765351, 1540.27082443, 1684.67121422, 1636.53775096, 1696.70458004, 1792.97150657, 1768.90477493, 1768.90477493, 1684.67121422, 1720.77131167, 1696.70458004, 1696.70458004, 1672.63784841, 1636.53775096, 1672.63784841, 1636.53775096, 1648.57111678, 1540.27082443, 1492.13736117, 1600.43765351, 1576.37092188, 1576.37092188, 1672.63784841, 1648.57111678, 1600.43765351, 1660.60448259, 1744.83804330, 1829.07160401, 1805.00487238, 1768.90477493, 1792.97150657, 1817.03823820, 1949.40526217, 1925.33853054, 1913.30516472, 1913.30516472, 1853.13833565, 1829.07160401, 1756.87140912, 1624.50438514, 1684.67121422, 1720.77131167, 1829.07160401, 1744.83804330, 1829.07160401, 1913.30516472, 1913.30516472, 1949.40526217, 1961.43862799, 1889.23843309, 1949.40526217, 1949.40526217, 2069.73892033, 2105.83901778, 2093.80565196, 2117.87238359, 2045.67218870, 2057.70555452, 2057.70555452, 2117.87238359, 2081.77228615, 2021.60545707, 2057.70555452, 2129.90574941, 2057.70555452, 2226.17267594, 2286.33950502, 2310.40623665, 2322.43960246, 2250.23940757, 2093.80565196, 2117.87238359, 2202.10594431, 2129.90574941, 2081.77228615, 2069.73892033, 2081.77228615, 2009.57209125, 1997.53872544, 1997.53872544, 1901.27179891, 1961.43862799, 1973.47199380, 1985.50535962, 2129.90574941, 2129.90574941, 2093.80565196, 1997.53872544, 2021.60545707, 2009.57209125, 2033.63882288, 1925.33853054, 2021.60545707, 2057.70555452, 2045.67218870, 2129.90574941, 2129.90574941, 2117.87238359, 2129.90574941, 1973.47199380, 1937.37189636, 1853.13833565, 1805.00487238, 1865.17170146, 1829.07160401, 1792.97150657, 1684.67121422, 1636.53775096, 1660.60448259, 1780.93814075, 1564.33755606, 1648.57111678, 1708.73794585, 1636.53775096, 1684.67121422, 1552.30419025, 1576.37092188, 1516.20409280, 1504.17072698, 1383.83706883, 1419.93716627, 1528.23745862, 1564.33755606, 1696.70458004, 1696.70458004, 1708.73794585, 1708.73794585, 1853.13833565, 1829.07160401, 2045.67218870, 2009.57209125, 2093.80565196, 2033.63882288, 2009.57209125, 1985.50535962, 2141.93911523, 2166.00584686, 2117.87238359, 2202.10594431, 2178.03921267, 2226.17267594, 2178.03921267, 2250.23940757, 2141.93911523, 2214.13931012, 2370.57306573, 2406.67316318, 2539.04018715, 2563.10691878, 2827.84096673, 2623.27374786, 2707.50730857, 2707.50730857, 2863.94106418, 2839.87433255, 2863.94106418, 2731.57404020, 2635.30711368, 2611.24038205, 2623.27374786, 2803.77423510, 2803.77423510, 2803.77423510, 2743.60740602, 2803.77423510, 2888.00779581, 2815.80760092, 2851.90769836, 2936.14125907, 3104.60838050, 3128.67511213, 3212.90867284, 3285.10886773, 3465.60935497, 3417.47589171, 3345.27569681, 3249.00877029, 3128.67511213, 3080.54164886, 3104.60838050, 3068.50828305, 3008.34145397, 3056.47491723, 3020.37481979, 3080.54164886, 3297.14223355, 3333.24233100, 3357.30906263, 3333.24233100, 3261.04213610, 3333.24233100, 3224.94203866, 3273.07550192, 3405.44252589, 3405.44252589, 3682.20993966, 4043.21091413, 3910.84389016, 3778.47686619, 3934.91062179, 4031.17754832, 3983.04408506, 3670.17657384, 3826.61032945, 3814.57696363, 3910.84389016, 3898.81052434, 4392.17852280, 4524.54554677, 4488.44544932, 4476.41208351, 4332.01169372, 4356.07842535, 4500.47881514, 4440.31198606, 4536.57891259, 4392.17852280, 4620.81247330, 4813.34632635, 4849.44642380, 4392.17852280, 4247.77813300, 4115.41110903, 4211.67803556, 3971.01071924, 3862.71042690, 4079.31101158, 4007.11081669, 4139.47784066, 4079.31101158, 4476.41208351, 4608.77910748, 4548.61227840, 4524.54554677, 4259.81149882, 4476.41208351, 4705.04603401, 4476.41208351, 5017.91354522], 
							[974.70263109, 943.41587996, 860.38565584, 772.54208538, 789.38879752, 734.03531477, 788.18546094, 703.95190023, 707.56190997, 661.83511987, 722.00194895, 673.86848569, 709.96858314, 734.03531477, 703.95190023, 672.66514911, 660.63178329, 606.48163712, 642.58173457, 647.39508089, 614.90499319, 625.73502243, 646.19174431, 659.42844671, 614.90499319, 625.73502243, 695.52854416, 735.23865135, 746.06868058, 758.10204640, 765.32206589, 689.51186125, 667.85180278, 673.86848569, 637.76838824, 622.12501268, 666.64846620, 641.37839799, 677.47849543, 690.71519783, 683.49517834, 679.88516860, 700.34189048, 703.95190023, 691.91853441, 643.78507115, 641.37839799, 637.76838824, 673.86848569, 708.76524656, 675.07182227, 684.69851492, 696.73188074, 703.95190023, 742.45867084, 703.95190023, 722.00194895, 818.26887548, 866.40233874, 878.43570456, 836.31892420, 879.63904114, 878.43570456, 890.46907037, 874.82569481, 884.45238747, 896.48575328, 856.77564609, 829.09890471, 830.30224130, 812.25219257, 854.36897293, 896.48575328, 842.33560711, 891.67240696, 866.40233874, 861.58899242, 876.02903140, 868.80901191, 870.01234849, 877.23236798, 877.23236798, 847.14895344, 898.89242645, 904.90910935, 872.41902165, 862.79232900, 823.08222181, 746.06868058, 724.40862212, 691.91853441, 685.90185151, 703.95190023, 695.52854416, 752.08536349, 713.57859288, 703.95190023, 770.13541222, 835.11558762, 813.45552915, 790.59213410, 877.23236798, 827.89556813, 740.05199768, 842.33560711, 838.72559737, 842.33560711, 821.87888522, 780.96544145, 760.50871956, 829.09890471, 836.31892420, 808.64218283, 800.21882676, 803.82883650, 839.92893395, 861.58899242, 868.80901191, 877.23236798, 871.21568507, 867.60567532, 819.47221206, 813.45552915, 903.70577277, 974.70263109, 950.63589945, 928.97584099, 934.99252389, 915.73913859, 890.46907037, 916.94247517, 888.06239721, 867.60567532, 848.35229002, 874.82569481, 826.69223155, 820.67554864, 854.36897293, 836.31892420, 783.37211461, 797.81215359, 786.98212436, 794.20214385, 725.61195870, 747.27201717, 794.20214385, 806.23550966, 862.79232900, 842.33560711, 836.31892420, 871.21568507, 830.30224130, 747.27201717, 728.01863186, 762.91539273, 738.84866109, 709.96858314, 660.63178329, 620.92167610, 602.87162738, 625.73502243, 602.87162738, 623.32834926, 660.63178329, 706.35857339, 665.44512962, 697.93521732, 651.00509064, 649.80175406, 630.54836875, 637.76838824, 577.60155916, 580.00823233, 637.76838824, 571.58487625, 643.78507115, 622.12501268, 625.73502243, 625.73502243, 660.63178329, 658.22511013, 746.06868058, 748.47535375, 759.30538298, 746.06868058, 759.30538298, 724.40862212, 722.00194895, 764.11872931, 720.79861237, 749.67869033, 677.47849543, 755.69537324, 705.15523681, 731.62864161, 741.25533426, 746.06868058, 770.13541222, 826.69223155, 857.97898267, 821.87888522, 860.38565584, 853.16563635, 860.38565584, 821.87888522, 845.94561686, 889.26573379, 842.33560711, 835.11558762, 786.98212436, 807.43884625, 800.21882676, 847.14895344, 862.79232900, 890.46907037, 872.41902165, 922.95915808, 928.97584099, 859.18231925, 883.24905089, 904.90910935, 883.24905089, 921.75582150, 886.85906063, 891.67240696, 900.09576303, 900.09576303, 880.84237772, 847.14895344, 868.80901191, 824.28555839, 820.67554864, 827.89556813, 830.30224130, 853.16563635, 786.98212436, 758.10204640, 789.38879752, 817.06553890, 847.14895344, 806.23550966, 799.01549017, 830.30224130, 794.20214385, 784.57545120, 832.70891446, 782.16877803, 777.35543171, 736.44198793, 677.47849543, 695.52854416, 646.19174431, 643.78507115, 631.75170533, 511.41804717, 558.34817386, 533.07810564, 521.04473983, 517.43473008, 500.58801794, 506.60470085, 487.35131554, 477.72462289, 451.25121810, 463.28458391, 451.25121810, 431.99783279, 409.13443774, 451.25121810, 488.55465212, 515.02805692, 512.62138376, 583.61824207, 567.97486651, 542.70479830, 545.11147146, 552.33149095, 521.04473983, 511.41804717, 493.36799845, 469.30126682, 451.25121810, 462.08124733, 407.93110116, 380.25435978, 367.01765738, 362.20431106, 361.00097448, 374.23767687, 386.27104269, 362.20431106], 
							[1756.87140912, 1600.43765351, 1516.20409280, 1660.60448259, 1648.57111678, 1648.57111678, 1684.67121422, 1732.80467749, 1756.87140912, 1756.87140912, 1805.00487238, 1756.87140912, 1792.97150657, 1817.03823820, 1829.07160401, 1792.97150657, 1720.77131167, 1744.83804330, 1732.80467749, 1780.93814075, 1805.00487238, 1696.70458004, 1708.73794585, 1708.73794585, 1684.67121422, 1708.73794585, 1768.90477493, 1768.90477493, 1720.77131167, 1756.87140912, 1684.67121422, 1684.67121422, 1732.80467749, 1708.73794585, 1648.57111678, 1696.70458004, 1648.57111678, 1636.53775096, 1696.70458004, 1684.67121422, 1720.77131167, 1732.80467749, 1768.90477493, 1805.00487238, 1780.93814075, 1720.77131167, 1708.73794585, 1660.60448259, 1696.70458004, 1720.77131167, 1684.67121422, 1648.57111678, 1624.50438514, 1576.37092188, 1648.57111678, 1720.77131167, 1744.83804330, 1817.03823820, 2045.67218870, 2021.60545707, 1961.43862799, 2105.83901778, 2105.83901778, 2033.63882288, 2057.70555452, 1973.47199380, 2033.63882288, 2009.57209125, 1865.17170146, 1841.10496983, 1865.17170146, 1865.17170146, 1961.43862799, 1937.37189636, 1973.47199380, 1997.53872544, 1961.43862799, 1985.50535962, 1985.50535962, 2045.67218870, 2069.73892033, 2021.60545707, 2021.60545707, 2033.63882288, 2129.90574941, 2141.93911523, 2226.17267594, 2262.27277339, 2214.13931012, 2153.97248104, 2178.03921267, 1997.53872544, 2081.77228615, 2129.90574941, 2153.97248104, 2069.73892033, 2141.93911523, 2226.17267594, 2226.17267594, 2226.17267594, 2202.10594431, 2346.50633410, 2286.33950502, 2250.23940757, 2394.63979736, 2454.80662644, 2466.83999225, 2478.87335807, 2394.63979736, 2406.67316318, 2406.67316318, 2442.77326062, 2442.77326062, 2394.63979736, 2382.60643154, 2382.60643154, 2418.70652899, 2527.00682133, 2563.10691878, 2551.07355297, 2659.37384531, 2683.44057694, 2611.24038205, 2659.37384531, 2659.37384531, 2611.24038205, 2551.07355297, 2563.10691878, 2611.24038205, 2587.17365041, 2587.17365041, 2599.20701623, 2539.04018715, 2418.70652899, 2466.83999225, 2442.77326062, 2490.90672389, 2490.90672389, 2539.04018715, 2502.94008970, 2539.04018715, 2539.04018715, 2551.07355297, 2490.90672389, 2527.00682133, 2490.90672389, 2478.87335807, 2502.94008970, 2527.00682133, 2514.97345552, 2490.90672389, 2394.63979736, 2310.40623665, 2250.23940757, 2226.17267594, 2166.00584686, 2093.80565196, 2057.70555452, 1997.53872544, 1997.53872544, 1961.43862799, 2129.90574941, 2045.67218870, 2021.60545707, 2093.80565196, 2057.70555452, 2081.77228615, 1961.43862799, 2009.57209125, 1973.47199380, 2021.60545707, 1961.43862799, 1949.40526217, 2009.57209125, 2021.60545707, 2129.90574941, 2262.27277339, 2238.20604175, 2202.10594431, 2202.10594431, 2214.13931012, 2346.50633410, 2298.37287083, 2310.40623665, 2274.30613920, 2298.37287083, 2153.97248104, 2274.30613920, 2322.43960246, 2262.27277339, 2310.40623665, 2238.20604175, 2310.40623665, 2346.50633410, 2358.53969991, 2382.60643154, 2430.73989481, 2466.83999225, 2539.04018715, 2599.20701623, 2635.30711368, 2755.64077184, 2659.37384531, 2731.57404020, 2731.57404020, 2779.70750347, 2767.67413765, 2791.74086928, 2707.50730857, 2719.54067439, 2659.37384531, 2707.50730857, 2803.77423510, 2839.87433255, 2827.84096673, 2827.84096673, 2839.87433255, 2863.94106418, 2851.90769836, 2815.80760092, 2851.90769836, 2888.00779581, 2972.24135652, 3008.34145397, 3116.64174631, 3249.00877029, 3297.14223355, 3261.04213610, 3116.64174631, 3044.44155142, 3056.47491723, 3212.90867284, 3152.74184376, 3140.70847794, 3249.00877029, 3152.74184376, 3176.80857539, 3273.07550192, 3345.27569681, 3429.50925753, 3321.20896518, 3249.00877029, 3369.34242845, 3236.97540447, 3297.14223355, 3417.47589171, 3429.50925753, 3585.94301313, 3742.37676874, 3826.61032945, 3778.47686619, 3886.77715853, 3995.07745087, 4007.11081669, 3706.27667129, 3874.74379271, 3898.81052434, 3766.44350037, 3802.54359782, 4139.47784066, 3971.01071924, 3718.31003711, 3706.27667129, 3742.37676874, 3766.44350037, 3742.37676874, 3802.54359782, 4091.34437740, 4031.17754832, 4319.97832790, 4584.71237585, 4596.74574167, 4295.91159627, 4115.41110903, 4019.14418250, 4103.37774321, 4067.27764577, 4031.17754832, 4007.11081669, 3946.94398761, 3983.04408506, 4127.44447485, 4283.87823045, 4223.71140137, 3983.04408506, 4031.17754832, 3670.17657384, 3790.51023200, 4007.11081669, 3995.07745087, 4295.91159627], 
							[593.24493472, 566.77152993, 541.50146171, 540.29812513, 549.92481779, 505.40136427, 519.84140325, 512.62138376, 515.02805692, 527.06142274, 548.72148120, 547.51814462, 537.89145197, 535.48477881, 518.63806666, 504.19802769, 472.91127656, 462.08124733, 499.38468136, 483.74130580, 466.89459366, 434.40450595, 460.87791075, 451.25121810, 421.16780356, 463.28458391, 529.46809590, 522.24807641, 513.82472034, 513.82472034, 523.45141299, 496.97800820, 515.02805692, 499.38468136, 509.01137401, 489.75798871, 471.70793998, 429.59115963, 448.84454493, 447.64120835, 445.23453519, 458.47123758, 459.67457417, 456.06456442, 441.62452544, 445.23453519, 440.42118886, 421.16780356, 458.47123758, 460.87791075, 445.23453519, 452.45455468, 433.20116937, 435.60784253, 460.87791075, 463.28458391, 493.36799845, 529.46809590, 535.48477881, 528.26475932, 521.04473983, 549.92481779, 564.36485676, 546.31480804, 571.58487625, 547.51814462, 547.51814462, 545.11147146, 511.41804717, 499.38468136, 494.57133503, 498.18134478, 502.99469110, 486.14797896, 498.18134478, 492.16466187, 518.63806666, 505.40136427, 496.97800820, 493.36799845, 487.35131554, 475.31794973, 475.31794973, 452.45455468, 475.31794973, 459.67457417, 454.86122784, 439.21785228, 410.33777432, 406.72776458, 389.88105243, 367.01765738, 382.66103294, 380.25435978, 379.05102320, 368.22099397, 371.83100371, 381.45769636, 407.93110116, 404.32109141, 397.10107192, 423.57447672, 406.72776458, 395.89773534, 399.50774509, 418.76113039, 410.33777432, 410.33777432, 394.69439876, 397.10107192, 393.49106218, 409.13443774, 407.93110116, 404.32109141, 403.11775483, 400.71108167, 413.94778407, 447.64120835, 459.67457417, 474.11461315, 465.69125707, 446.43787177, 433.20116937, 433.20116937, 428.38782305, 424.77781330, 416.35445723, 422.37114014, 429.59115963, 428.38782305, 421.16780356, 416.35445723, 415.15112065, 407.93110116, 409.13443774, 410.33777432, 403.11775483, 406.72776458, 406.72776458, 413.94778407, 417.55779381, 400.71108167, 405.52442800, 373.03434029, 369.42433055, 361.00097448, 361.00097448, 374.23767687, 373.03434029, 361.00097448, 361.00097448, 332.12089652, 350.17094524, 354.98429157, 348.96760866, 345.35759892, 326.10421361, 309.25750147, 303.24081856, 282.78409667, 286.39410642, 283.98743325, 280.37742351, 291.20745274, 302.03748198, 292.41078933, 298.42747223, 276.76741377, 267.14072111, 252.70068213, 251.49734555, 241.87065290, 223.82060418, 237.05730657, 221.41393101, 257.51402846, 269.54739428, 253.90401872, 261.12403820, 264.73404795, 280.37742351, 304.44415514, 290.00411616, 285.19076984, 297.22413565, 311.66417463, 277.97075035, 286.39410642, 294.81746249, 285.19076984, 282.78409667, 269.54739428, 286.39410642, 308.05416489, 315.27418438, 318.88419412, 327.30755019, 326.10421361, 321.29086728, 328.51088677, 345.35759892, 353.78095499, 336.93424284, 340.54425259, 341.74758917, 367.01765738, 368.22099397, 357.39096473, 356.18762815, 347.76427208, 341.74758917, 335.73090626, 348.96760866, 381.45769636, 393.49106218, 422.37114014, 403.11775483, 406.72776458, 399.50774509, 393.49106218, 388.67771585, 379.05102320, 386.27104269, 382.66103294, 393.49106218, 393.49106218, 401.91441825, 387.47437927, 383.86436953, 381.45769636, 367.01765738, 361.00097448, 358.59430131, 358.59430131, 363.40764764, 351.37428182, 346.56093550, 352.57761841, 359.79763789, 367.01765738, 353.78095499, 346.56093550, 363.40764764, 340.54425259, 342.95092575, 351.37428182, 342.95092575, 344.15426233, 338.13757943, 312.86751121, 324.90087703, 282.78409667, 285.19076984, 283.98743325, 255.10735530, 279.17408693, 270.75073086, 277.97075035, 274.36074060, 285.19076984, 294.81746249, 273.15740402, 263.53071137, 250.29400897, 253.90401872, 257.51402846, 253.90401872, 241.87065290, 252.70068213, 257.51402846, 271.95406744, 274.36074060, 279.17408693, 274.36074060, 258.71736504, 262.32737479, 271.95406744, 269.54739428, 298.42747223, 291.20745274, 275.56407718, 270.75073086, 276.76741377, 246.68399923, 258.71736504, 238.26064315, 232.24396025, 238.26064315, 244.27732606, 221.41393101, 199.75387254], 
							[383.86436953, 376.64435004, 356.18762815, 342.95092575, 324.90087703, 288.80077958, 308.05416489, 321.29086728, 316.47752096, 321.29086728, 346.56093550, 342.95092575, 361.00097448, 364.61098422, 357.39096473, 348.96760866, 324.90087703, 304.44415514, 328.51088677, 356.18762815, 348.96760866, 316.47752096, 316.47752096, 309.25750147, 282.78409667, 317.68085754, 359.79763789, 464.48792049, 450.04788151, 431.99783279, 425.98114988, 371.83100371, 347.76427208, 347.76427208, 342.95092575, 320.08753070, 323.69754045, 310.46083805, 318.88419412, 335.73090626, 322.49420387, 335.73090626, 348.96760866, 362.20431106, 399.50774509, 385.06770611, 383.86436953, 418.76113039, 441.62452544, 465.69125707, 450.04788151, 438.01451570, 433.20116937, 415.15112065, 476.52128631, 482.53796922, 523.45141299, 644.98840773, 613.70165661, 605.27830054, 558.34817386, 613.70165661, 595.65160789, 629.34503217, 616.10832977, 637.76838824, 626.93835901, 606.48163712, 587.22825181, 565.56819335, 561.95818360, 561.95818360, 571.58487625, 559.55151044, 559.55151044, 558.34817386, 569.17820309, 552.33149095, 549.92481779, 547.51814462, 535.48477881, 509.01137401, 481.33463264, 481.33463264, 501.79135452, 493.36799845, 466.89459366, 428.38782305, 389.88105243, 401.91441825, 379.05102320, 321.29086728, 367.01765738, 368.22099397, 389.88105243, 392.28772560, 423.57447672, 483.74130580, 496.97800820, 500.58801794, 482.53796922, 533.07810564, 516.23139350, 476.52128631, 500.58801794, 517.43473008, 534.28144222, 548.72148120, 565.56819335, 557.14483728, 573.99154942, 604.07496396, 578.80489574, 563.16152018, 565.56819335, 584.82157865, 583.61824207, 602.87162738, 610.09164687, 638.97172482, 672.66514911, 636.56505166, 625.73502243, 626.93835901, 625.73502243, 610.09164687, 599.26161763, 607.68497370, 624.53168584, 616.10832977, 619.71833952, 614.90499319, 596.85494447, 577.60155916, 577.60155916, 587.22825181, 601.66829079, 608.88831028, 636.56505166, 620.92167610, 631.75170533, 584.82157865, 587.22825181, 559.55151044, 584.82157865, 567.97486651, 567.97486651, 595.65160789, 602.87162738, 605.27830054, 589.63492498, 559.55151044, 547.51814462, 553.53482753, 547.51814462, 521.04473983, 511.41804717, 512.62138376, 494.57133503, 480.13129605, 457.26790100, 489.75798871, 463.28458391, 460.87791075, 492.16466187, 471.70793998, 483.74130580, 444.03119861, 424.77781330, 438.01451570, 423.57447672, 391.08438902, 368.22099397, 389.88105243, 379.05102320, 442.82786202, 481.33463264, 456.06456442, 428.38782305, 442.82786202, 436.81117912, 523.45141299, 509.01137401, 528.26475932, 571.58487625, 581.21156891, 545.11147146, 546.31480804, 547.51814462, 523.45141299, 517.43473008, 499.38468136, 505.40136427, 494.57133503, 502.99469110, 505.40136427, 504.19802769, 539.09478855, 560.75484702, 583.61824207, 578.80489574, 607.68497370, 583.61824207, 601.66829079, 607.68497370, 649.80175406, 661.83511987, 649.80175406, 654.61510038, 699.13855390, 677.47849543, 708.76524656, 717.18860263, 744.86534400, 728.01863186, 742.45867084, 709.96858314, 723.20528553, 707.56190997, 718.39193921, 722.00194895, 702.74856365, 707.56190997, 708.76524656, 740.05199768, 759.30538298, 753.28870007, 776.15209512, 758.10204640, 760.50871956, 755.69537324, 740.05199768, 729.22196844, 728.01863186, 697.93521732, 659.42844671, 690.71519783, 713.57859288, 719.59527579, 731.62864161, 717.18860263, 715.98526604, 750.88202691, 734.03531477, 691.91853441, 731.62864161, 703.95190023, 696.73188074, 684.69851492, 611.29498345, 673.86848569, 619.71833952, 610.09164687, 619.71833952, 501.79135452, 547.51814462, 543.90813488, 581.21156891, 545.11147146, 575.19488600, 625.73502243, 655.81843697, 643.78507115, 561.95818360, 582.41490549, 560.75484702, 558.34817386, 548.72148120, 537.89145197, 576.39822258, 600.46495421, 599.26161763, 605.27830054, 590.83826156, 537.89145197, 509.01137401, 493.36799845, 481.33463264, 475.31794973, 460.87791075, 506.60470085, 481.33463264, 490.96132529, 487.35131554, 454.86122784, 433.20116937, 409.13443774, 409.13443774, 431.99783279, 412.74444748, 433.20116937], 
							[1720.77131167, 1672.63784841, 1672.63784841, 1660.60448259, 1636.53775096, 1612.47101933, 1444.00389791, 1275.53677648, 1239.43667904, 1197.31989868, 1323.67023975, 1299.60350811, 1359.77033719, 1299.60350811, 1287.57014230, 1239.43667904, 1107.06965506, 1095.03628924, 1139.55974276, 1191.30321577, 1227.40331322, 1164.82981098, 1197.31989868, 1166.03314756, 1167.23648414, 1263.50341067, 1395.87043464, 1576.37092188, 1540.27082443, 1492.13736117, 1480.10399535, 1395.87043464, 1359.77033719, 1395.87043464, 1371.80370301, 1359.77033719, 1431.97053209, 1419.93716627, 1504.17072698, 1600.43765351, 1576.37092188, 1600.43765351, 1672.63784841, 1648.57111678, 1540.27082443, 1468.07062954, 1468.07062954, 1431.97053209, 1564.33755606, 1576.37092188, 1504.17072698, 1480.10399535, 1564.33755606, 1552.30419025, 1612.47101933, 1684.67121422, 1780.93814075, 1841.10496983, 1720.77131167, 1768.90477493, 1636.53775096, 1780.93814075, 1780.93814075, 1732.80467749, 1708.73794585, 1780.93814075, 1829.07160401, 1841.10496983, 1841.10496983, 1841.10496983, 1865.17170146, 1925.33853054, 1925.33853054, 1853.13833565, 1913.30516472, 1973.47199380, 2069.73892033, 2214.13931012, 2310.40623665, 2250.23940757, 2310.40623665, 2286.33950502, 2190.07257849, 2286.33950502, 2214.13931012, 2202.10594431, 2129.90574941, 2093.80565196, 2009.57209125, 1961.43862799, 1805.00487238, 1660.60448259, 1780.93814075, 1708.73794585, 1889.23843309, 1708.73794585, 1780.93814075, 1889.23843309, 2021.60545707, 1985.50535962, 1913.30516472, 2117.87238359, 2045.67218870, 1853.13833565, 1913.30516472, 1973.47199380, 1985.50535962, 1961.43862799, 1973.47199380, 1889.23843309, 1925.33853054, 1937.37189636, 1901.27179891, 1913.30516472, 1949.40526217, 2009.57209125, 2069.73892033, 2093.80565196, 2057.70555452, 2093.80565196, 2093.80565196, 2045.67218870, 2009.57209125, 1985.50535962, 2045.67218870, 1973.47199380, 1997.53872544, 2045.67218870, 2093.80565196, 2021.60545707, 2009.57209125, 2081.77228615, 1997.53872544, 2021.60545707, 1961.43862799, 1853.13833565, 1889.23843309, 1925.33853054, 1913.30516472, 1877.20506728, 1877.20506728, 1829.07160401, 1829.07160401, 1792.97150657, 1829.07160401, 1841.10496983, 1913.30516472, 1997.53872544, 1997.53872544, 1949.40526217, 1961.43862799, 1985.50535962, 1865.17170146, 1985.50535962, 1889.23843309, 1877.20506728, 1720.77131167, 1624.50438514, 1528.23745862, 1612.47101933, 1576.37092188, 1672.63784841, 1576.37092188, 1588.40428770, 1768.90477493, 1744.83804330, 1732.80467749, 1672.63784841, 1600.43765351, 1612.47101933, 1648.57111678, 1540.27082443, 1431.97053209, 1540.27082443, 1395.87043464, 1636.53775096, 1612.47101933, 1672.63784841, 1660.60448259, 1600.43765351, 1612.47101933, 1780.93814075, 1732.80467749, 1768.90477493, 1805.00487238, 1901.27179891, 1744.83804330, 1756.87140912, 1829.07160401, 1756.87140912, 1768.90477493, 1684.67121422, 1744.83804330, 1792.97150657, 1841.10496983, 1913.30516472, 1997.53872544, 1937.37189636, 1997.53872544, 2057.70555452, 1937.37189636, 1997.53872544, 2009.57209125, 2105.83901778, 2117.87238359, 2021.60545707, 2021.60545707, 2117.87238359, 2045.67218870, 1997.53872544, 2057.70555452, 2057.70555452, 2129.90574941, 2153.97248104, 2178.03921267, 2105.83901778, 2190.07257849, 2153.97248104, 2129.90574941, 2081.77228615, 2178.03921267, 2190.07257849, 2274.30613920, 2226.17267594, 2298.37287083, 2358.53969991, 2358.53969991, 2298.37287083, 2202.10594431, 2238.20604175, 2238.20604175, 2226.17267594, 2166.00584686, 2202.10594431, 2178.03921267, 2069.73892033, 2033.63882288, 2153.97248104, 2202.10594431, 2202.10594431, 2093.80565196, 2190.07257849, 2226.17267594, 2178.03921267, 2153.97248104, 2190.07257849, 2190.07257849, 2238.20604175, 2166.00584686, 2105.83901778, 2069.73892033, 1985.50535962, 1817.03823820, 1805.00487238, 1480.10399535, 1624.50438514, 1552.30419025, 1648.57111678, 1612.47101933, 1636.53775096, 1708.73794585, 1684.67121422, 1600.43765351, 1576.37092188, 1732.80467749, 1624.50438514, 1600.43765351, 1612.47101933, 1696.70458004, 1696.70458004, 1780.93814075, 1901.27179891, 2033.63882288, 2069.73892033, 2009.57209125, 2093.80565196, 2105.83901778, 2069.73892033, 1985.50535962, 1925.33853054, 1913.30516472, 1973.47199380, 1961.43862799, 1805.00487238, 1805.00487238, 1732.80467749, 1624.50438514, 1684.67121422, 1744.83804330, 1648.57111678, 1708.73794585], 
							[883.24905089, 853.16563635, 764.11872931, 824.28555839, 842.33560711, 732.83197819, 726.81529528, 782.16877803, 778.55876829, 773.74542196, 818.26887548, 756.89870982, 799.01549017, 764.11872931, 750.88202691, 761.71205615, 714.78192946, 696.73188074, 767.72873905, 759.30538298, 759.30538298, 715.98526604, 786.98212436, 794.20214385, 770.13541222, 920.55248491, 1002.37937246, 1062.54620154, 974.70263109, 907.31578252, 960.26259211, 908.51911910, 871.21568507, 894.07908012, 896.48575328, 855.57230951, 885.65572405, 876.02903140, 898.89242645, 906.11244594, 913.33246543, 956.65258236, 985.53266032, 974.70263109, 972.29595792, 926.56916782, 916.94247517, 924.16249466, 951.83923604, 968.68594818, 938.60253364, 926.56916782, 912.12912884, 855.57230951, 936.19586048, 983.12598716, 947.02588971, 1019.22608460, 1045.69948940, 1022.83609435, 1010.80272853, 1064.95287471, 1064.95287471, 1008.39605537, 1045.69948940, 1062.54620154, 1069.76622103, 1070.96955761, 1026.44610409, 1025.24276751, 972.29595792, 998.76936272, 1002.37937246, 1003.58270904, 1058.93619180, 1070.96955761, 1062.54620154, 1058.93619180, 1040.88614307, 1044.49615282, 995.15935297, 979.51597741, 955.44924578, 962.66926527, 974.70263109, 974.70263109, 991.54934323, 1014.41273828, 985.53266032, 973.49929450, 928.97584099, 855.57230951, 878.43570456, 883.24905089, 924.16249466, 888.06239721, 908.51911910, 937.39919706, 938.60253364, 955.44924578, 928.97584099, 981.92265058, 968.68594818, 889.26573379, 908.51911910, 930.17917757, 960.26259211, 953.04257262, 926.56916782, 884.45238747, 974.70263109, 980.71931399, 983.12598716, 951.83923604, 984.32932374, 1004.78604563, 980.71931399, 992.75267981, 995.15935297, 1022.83609435, 1009.59939195, 969.88928476, 955.44924578, 944.61921655, 962.66926527, 933.78918731, 924.16249466, 913.33246543, 927.77250440, 966.27927501, 937.39919706, 949.43256287, 919.34914833, 904.90910935, 912.12912884, 909.72245568, 897.68908986, 891.67240696, 878.43570456, 856.77564609, 854.36897293, 836.31892420, 847.14895344, 831.50557788, 842.33560711, 854.36897293, 854.36897293, 854.36897293, 878.43570456, 883.24905089, 844.74228027, 813.45552915, 776.15209512, 783.37211461, 785.77878778, 800.21882676, 744.86534400, 720.79861237, 688.30852467, 676.27515885, 673.86848569, 673.86848569, 707.56190997, 691.91853441, 712.37525630, 719.59527579, 726.81529528, 703.95190023, 709.96858314, 691.91853441, 688.30852467, 682.29184176, 666.64846620, 670.25847594, 637.76838824, 740.05199768, 694.32520758, 701.54522707, 732.83197819, 705.15523681, 726.81529528, 754.49203666, 758.10204640, 776.15209512, 780.96544145, 774.94875854, 776.15209512, 770.13541222, 776.15209512, 750.88202691, 759.30538298, 729.22196844, 737.64532451, 758.10204640, 756.89870982, 762.91539273, 842.33560711, 888.06239721, 860.38565584, 868.80901191, 878.43570456, 931.38251415, 897.68908986, 908.51911910, 925.36583124, 944.61921655, 934.99252389, 937.39919706, 919.34914833, 909.72245568, 907.31578252, 868.80901191, 925.36583124, 959.05925553, 972.29595792, 991.54934323, 975.90596767, 1001.17603588, 974.70263109, 962.66926527, 980.71931399, 957.85591894, 967.48261160, 971.09262134, 975.90596767, 977.10930425, 985.53266032, 978.31264083, 944.61921655, 925.36583124, 925.36583124, 910.92579226, 878.43570456, 900.09576303, 909.72245568, 851.96229976, 878.43570456, 892.87574354, 920.55248491, 913.33246543, 912.12912884, 896.48575328, 904.90910935, 867.60567532, 883.24905089, 890.46907037, 876.02903140, 903.70577277, 867.60567532, 848.35229002, 788.18546094, 734.03531477, 729.22196844, 724.40862212, 659.42844671, 660.63178329, 632.95504192, 643.78507115, 642.58173457, 632.95504192, 666.64846620, 724.40862212, 709.96858314, 664.24179304, 740.05199768, 722.00194895, 708.76524656, 694.32520758, 691.91853441, 697.93521732, 709.96858314, 741.25533426, 826.69223155, 844.74228027, 863.99566558, 865.19900216, 900.09576303, 894.07908012, 924.16249466, 895.28241670, 885.65572405, 926.56916782, 937.39919706, 907.31578252, 932.58585073, 903.70577277, 874.82569481, 821.87888522, 837.52226079, 800.21882676, 801.42216334], 
							[764.11872931, 764.11872931, 740.05199768, 699.13855390, 702.74856365, 631.75170533, 567.97486651, 637.76838824, 595.65160789, 659.42844671, 706.35857339, 697.93521732, 735.23865135, 722.00194895, 703.95190023, 728.01863186, 661.83511987, 637.76838824, 663.03845645, 637.76838824, 642.58173457, 601.66829079, 631.75170533, 612.49832003, 553.53482753, 611.29498345, 746.06868058, 782.16877803, 783.37211461, 791.79547068, 836.31892420, 844.74228027, 838.72559737, 825.48889497, 832.70891446, 814.65886573, 800.21882676, 770.13541222, 829.09890471, 867.60567532, 866.40233874, 856.77564609, 865.19900216, 830.30224130, 803.82883650, 821.87888522, 836.31892420, 821.87888522, 855.57230951, 890.46907037, 902.50243619, 919.34914833, 906.11244594, 888.06239721, 914.53580201, 954.24590920, 934.99252389, 932.58585073, 1005.98938221, 971.09262134, 934.99252389, 980.71931399, 972.29595792, 998.76936272, 1008.39605537, 1058.93619180, 1045.69948940, 1021.63275777, 973.49929450, 966.27927501, 980.71931399, 999.97269930, 1005.98938221, 968.68594818, 986.73599690, 956.65258236, 1008.39605537, 1073.37623078, 1092.62961608, 1102.25630873, 1103.45964532, 1084.20626001, 1062.54620154, 1058.93619180, 1030.05611384, 999.97269930, 999.97269930, 941.00920680, 833.91225104, 878.43570456, 900.09576303, 791.79547068, 842.33560711, 862.79232900, 912.12912884, 830.30224130, 878.43570456, 877.23236798, 938.60253364, 955.44924578, 920.55248491, 998.76936272, 932.58585073, 918.14581175, 932.58585073, 965.07593843, 983.12598716, 993.95601639, 957.85591894, 926.56916782, 961.46592869, 956.65258236, 938.60253364, 926.56916782, 926.56916782, 921.75582150, 912.12912884, 918.14581175, 937.39919706, 945.82255313, 944.61921655, 912.12912884, 873.62235823, 877.23236798, 895.28241670, 879.63904114, 894.07908012, 868.80901191, 895.28241670, 871.21568507, 871.21568507, 873.62235823, 830.30224130, 797.81215359, 800.21882676, 788.18546094, 783.37211461, 803.82883650, 794.20214385, 782.16877803, 788.18546094, 759.30538298, 753.28870007, 712.37525630, 724.40862212, 731.62864161, 719.59527579, 746.06868058, 749.67869033, 748.47535375, 758.10204640, 767.72873905, 711.17191972, 746.06868058, 725.61195870, 737.64532451, 701.54522707, 661.83511987, 612.49832003, 623.32834926, 606.48163712, 589.63492498, 594.44827130, 605.27830054, 649.80175406, 624.53168584, 624.53168584, 604.07496396, 567.97486651, 545.11147146, 569.17820309, 537.89145197, 493.36799845, 539.09478855, 482.53796922, 599.26161763, 589.63492498, 601.66829079, 601.66829079, 589.63492498, 575.19488600, 630.54836875, 612.49832003, 610.09164687, 642.58173457, 637.76838824, 618.51500294, 601.66829079, 619.71833952, 589.63492498, 589.63492498, 554.73816411, 595.65160789, 592.04159814, 616.10832977, 608.88831028, 628.14169559, 643.78507115, 683.49517834, 684.69851492, 670.25847594, 681.08850518, 705.15523681, 703.95190023, 728.01863186, 718.39193921, 747.27201717, 735.23865135, 712.37525630, 693.12187099, 725.61195870, 731.62864161, 730.42530502, 744.86534400, 826.69223155, 830.30224130, 809.84551941, 792.99880727, 789.38879752, 782.16877803, 788.18546094, 806.23550966, 807.43884625, 782.16877803, 808.64218283, 814.65886573, 818.26887548, 802.62549992, 761.71205615, 759.30538298, 755.69537324, 730.42530502, 703.95190023, 722.00194895, 759.30538298, 687.10518809, 707.56190997, 735.23865135, 770.13541222, 776.15209512, 744.86534400, 725.61195870, 776.15209512, 737.64532451, 737.64532451, 768.93207563, 736.44198793, 754.49203666, 732.83197819, 658.22511013, 673.86848569, 601.66829079, 565.56819335, 584.82157865, 521.04473983, 531.87476906, 523.45141299, 536.68811539, 500.58801794, 523.45141299, 535.48477881, 521.04473983, 517.43473008, 498.18134478, 537.89145197, 493.36799845, 451.25121810, 457.26790100, 481.33463264, 465.69125707, 493.36799845, 517.43473008, 553.53482753, 531.87476906, 543.90813488, 571.58487625, 551.12815437, 552.33149095, 536.68811539, 502.99469110, 492.16466187, 487.35131554, 470.50460340, 409.13443774, 421.16780356, 388.67771585, 377.84768662, 377.84768662, 403.11775483, 381.45769636, 353.78095499], 
							[1016.81941144, 1034.86946017, 960.26259211, 860.38565584, 824.28555839, 752.08536349, 740.05199768, 758.10204640, 762.91539273, 767.72873905, 862.79232900, 807.43884625, 815.86220232, 815.86220232, 806.23550966, 776.15209512, 729.22196844, 702.74856365, 752.08536349, 791.79547068, 824.28555839, 777.35543171, 742.45867084, 679.88516860, 611.29498345, 720.79861237, 889.26573379, 1001.17603588, 959.05925553, 915.73913859, 927.77250440, 866.40233874, 842.33560711, 843.53894369, 814.65886573, 842.33560711, 824.28555839, 785.77878778, 829.09890471, 823.08222181, 842.33560711, 818.26887548, 812.25219257, 806.23550966, 783.37211461, 772.54208538, 738.84866109, 723.20528553, 758.10204640, 776.15209512, 747.27201717, 811.04885599, 814.65886573, 815.86220232, 848.35229002, 833.91225104, 883.24905089, 890.46907037, 908.51911910, 934.99252389, 872.41902165, 891.67240696, 880.84237772, 890.46907037, 879.63904114, 876.02903140, 885.65572405, 870.01234849, 829.09890471, 817.06553890, 808.64218283, 848.35229002, 836.31892420, 808.64218283, 819.47221206, 830.30224130, 854.36897293, 849.55562660, 843.53894369, 848.35229002, 837.52226079, 831.50557788, 842.33560711, 812.25219257, 797.81215359, 808.64218283, 803.82883650, 780.96544145, 724.40862212, 734.03531477, 700.34189048, 652.20842722, 690.71519783, 655.81843697, 685.90185151, 653.41176380, 683.49517834, 709.96858314, 772.54208538, 767.72873905, 746.06868058, 794.20214385, 770.13541222, 747.27201717, 758.10204640, 776.15209512, 766.52540247, 777.35543171, 750.88202691, 746.06868058, 761.71205615, 766.52540247, 754.49203666, 742.45867084, 767.72873905, 765.32206589, 780.96544145, 780.96544145, 795.40548043, 838.72559737, 830.30224130, 818.26887548, 818.26887548, 861.58899242, 860.38565584, 824.28555839, 853.16563635, 831.50557788, 818.26887548, 795.40548043, 791.79547068, 802.62549992, 802.62549992, 794.20214385, 767.72873905, 760.50871956, 750.88202691, 765.32206589, 770.13541222, 777.35543171, 794.20214385, 786.98212436, 790.59213410, 760.50871956, 755.69537324, 755.69537324, 756.89870982, 774.94875854, 772.54208538, 746.06868058, 748.47535375, 737.64532451, 708.76524656, 712.37525630, 719.59527579, 732.83197819, 720.79861237, 684.69851492, 651.00509064, 659.42844671, 643.78507115, 629.34503217, 622.12501268, 640.17506140, 685.90185151, 671.46181253, 687.10518809, 681.08850518, 681.08850518, 644.98840773, 644.98840773, 606.48163712, 595.65160789, 613.70165661, 595.65160789, 691.91853441, 690.71519783, 697.93521732, 672.66514911, 655.81843697, 646.19174431, 695.52854416, 700.34189048, 690.71519783, 728.01863186, 734.03531477, 701.54522707, 700.34189048, 724.40862212, 691.91853441, 701.54522707, 661.83511987, 679.88516860, 685.90185151, 685.90185151, 688.30852467, 714.78192946, 718.39193921, 720.79861237, 740.05199768, 748.47535375, 762.91539273, 759.30538298, 770.13541222, 780.96544145, 788.18546094, 797.81215359, 800.21882676, 786.98212436, 759.30538298, 764.11872931, 764.11872931, 782.16877803, 795.40548043, 801.42216334, 818.26887548, 831.50557788, 817.06553890, 789.38879752, 785.77878778, 821.87888522, 815.86220232, 821.87888522, 802.62549992, 814.65886573, 829.09890471, 817.06553890, 808.64218283, 808.64218283, 818.26887548, 807.43884625, 784.57545120, 779.76210487, 779.76210487, 785.77878778, 753.28870007, 746.06868058, 817.06553890, 813.45552915, 818.26887548, 806.23550966, 803.82883650, 826.69223155, 783.37211461, 788.18546094, 825.48889497, 797.81215359, 820.67554864, 814.65886573, 778.55876829, 786.98212436, 758.10204640, 722.00194895, 723.20528553, 664.24179304, 697.93521732, 696.73188074, 734.03531477, 689.51186125, 709.96858314, 717.18860263, 700.34189048, 682.29184176, 691.91853441, 683.49517834, 654.61510038, 626.93835901, 671.46181253, 676.27515885, 694.32520758, 714.78192946, 729.22196844, 782.16877803, 744.86534400, 735.23865135, 752.08536349, 791.79547068, 776.15209512, 762.91539273, 731.62864161, 728.01863186, 746.06868058, 747.27201717, 684.69851492, 666.64846620, 628.14169559, 626.93835901, 658.22511013, 667.85180278, 679.88516860, 654.61510038], 
							[885.17438942, 850.47016240, 822.69715410, 757.89747918, 779.88243853, 749.79902399, 698.88585322, 665.33682933, 664.16959284, 644.50707310, 723.18121880, 718.56040633, 781.03764165, 821.54195098, 793.76894268, 775.26162605, 717.40520321, 672.28008140, 704.67390218, 681.53373971, 673.43528452, 625.98772311, 658.39357725, 576.22975546, 538.04788573, 613.25642207, 694.26504075, 694.26504075, 718.56040633, 717.40520321, 731.29170736, 732.44691048, 735.91251984, 734.75731672, 745.16617815, 717.40520321, 708.13951153, 684.99934907, 703.51869906, 697.73065010, 697.73065010, 683.84414595, 688.47699179, 688.47699179, 665.33682933, 643.35186998, 647.97268245, 630.62056895, 651.45032517, 684.99934907, 680.37853660, 682.68894283, 686.15455219, 690.78739803, 704.67390218, 713.92756049, 698.88585322, 726.65886152, 747.48861775, 787.98089372, 757.89747918, 786.82569060, 777.57203229, 803.02260099, 797.24658540, 804.17780411, 833.10601553, 827.32999994, 775.26162605, 771.78398333, 746.33341463, 774.09438957, 771.78398333, 765.99593438, 793.76894268, 781.03764165, 812.28829267, 804.17780411, 804.17780411, 800.71219475, 807.65544683, 786.82569060, 797.24658540, 777.57203229, 782.19284476, 778.72723541, 775.26162605, 740.54536567, 700.04105634, 650.29512206, 653.76073141, 572.76414610, 613.25642207, 569.28650338, 601.69235753, 562.35528467, 599.38195129, 624.83251999, 660.70398349, 672.28008140, 642.18463350, 726.65886152, 711.61715425, 651.45032517, 688.47699179, 710.46195114, 724.34845529, 724.34845529, 702.36349594, 664.16959284, 704.67390218, 701.19625946, 698.88585322, 672.28008140, 675.74569076, 711.61715425, 715.08276361, 722.02601568, 718.56040633, 740.54536567, 771.78398333, 718.56040633, 737.06772295, 706.98430841, 719.71560945, 724.34845529, 733.60211360, 724.34845529, 734.75731672, 711.61715425, 717.40520321, 711.61715425, 691.94260114, 710.46195114, 705.82910530, 684.99934907, 686.15455219, 702.36349594, 694.26504075, 691.94260114, 696.57544698, 703.51869906, 689.63219491, 637.56382102, 643.35186998, 620.19967415, 620.19967415, 676.90089387, 676.90089387, 660.70398349, 650.29512206, 704.67390218, 715.08276361, 718.56040633, 698.88585322, 682.68894283, 642.18463350, 632.93097518, 578.55219506, 582.01780442, 578.55219506, 610.09164687, 558.34817386, 566.77152993, 613.70165661, 613.70165661, 634.15837850, 616.10832977, 618.51500294, 559.55151044, 589.63492498, 517.43473008, 482.53796922, 502.99469110, 480.13129605, 558.34817386, 555.94150069, 595.65160789, 582.41490549, 555.94150069, 561.95818360, 617.31166635, 613.70165661, 624.53168584, 636.56505166, 605.27830054, 569.17820309, 583.61824207, 618.51500294, 583.61824207, 600.46495421, 565.56819335, 577.60155916, 572.78821284, 608.88831028, 599.26161763, 624.53168584, 638.97172482, 644.98840773, 660.63178329, 661.83511987, 673.86848569, 666.64846620, 673.86848569, 671.46181253, 695.52854416, 685.90185151, 683.49517834, 670.25847594, 646.19174431, 660.63178329, 649.80175406, 681.08850518, 682.29184176, 722.00194895, 713.57859288, 725.61195870, 712.37525630, 690.71519783, 697.93521732, 718.39193921, 725.61195870, 742.45867084, 732.83197819, 749.67869033, 764.11872931, 764.11872931, 753.28870007, 725.61195870, 741.25533426, 744.86534400, 724.40862212, 725.61195870, 737.64532451, 746.06868058, 701.54522707, 676.27515885, 731.62864161, 752.08536349, 756.89870982, 718.39193921, 706.35857339, 734.03531477, 712.37525630, 706.35857339, 740.05199768, 740.05199768, 742.45867084, 712.37525630, 671.46181253, 703.95190023, 635.36171508, 607.68497370, 593.24493472, 539.09478855, 560.75484702, 531.87476906, 565.56819335, 529.46809590, 530.67143248, 545.11147146, 529.46809590, 555.94150069, 539.09478855, 589.63492498, 536.68811539, 511.41804717, 523.45141299, 543.90813488, 517.43473008, 553.53482753, 571.58487625, 616.10832977, 587.22825181, 605.27830054, 623.32834926, 594.44827130, 571.58487625, 555.94150069, 529.46809590, 506.60470085, 533.07810564, 524.65474957, 463.28458391, 466.89459366, 445.23453519, 427.18448646, 430.79449621, 439.21785228, 417.55779381, 419.96446697], 
							[922.97119144, 912.45402972, 899.60239503, 882.08181440, 852.87683557, 765.24986569, 705.66063817, 718.51227287, 706.83990802, 698.65721927, 727.86219811, 712.67609045, 741.88106928, 743.04830576, 704.49340169, 704.49340169, 662.43678816, 643.74897105, 665.93849762, 662.43678816, 660.10231520, 626.21635706, 625.04912057, 607.52853995, 590.00795932, 634.39904581, 700.99169224, 705.66063817, 746.56204858, 705.66063817, 718.51227287, 690.47453052, 684.63834809, 693.98827333, 678.79013231, 667.10573410, 655.43336926, 648.41791699, 648.41791699, 675.28842286, 670.61947692, 695.15550982, 706.83990802, 690.47453052, 670.61947692, 663.60402465, 663.60402465, 644.91620754, 660.10231520, 677.62289582, 660.10231520, 676.45565934, 686.97282106, 683.47111161, 747.72928507, 771.08604812, 848.19585626, 945.17275137, 898.43515855, 900.76963151, 858.71301799, 896.10068558, 922.97119144, 912.45402972, 940.50380544, 1017.61361359, 1010.59816131, 1000.08099959, 968.54154779, 961.52609552, 969.70878427, 969.70878427, 1015.26710725, 1002.41547256, 1017.61361359, 1010.59816131, 1022.28255952, 1017.61361359, 1003.58270904, 998.91376311, 987.22936490, 968.54154779, 952.17617028, 940.50380544, 948.67446083, 946.33998786, 956.85714958, 872.73188916, 828.34080267, 828.34080267, 776.93426390, 743.04830576, 755.89994046, 741.88106928, 755.89994046, 739.54659631, 736.04488686, 747.72928507, 782.77044632, 828.34080267, 794.45484453, 854.04407205, 861.04749096, 808.47371571, 817.82364094, 817.82364094, 815.48916798, 826.00632970, 794.45484453, 784.57545120, 800.21882676, 801.42216334, 802.62549992, 791.79547068, 801.42216334, 836.31892420, 808.64218283, 813.45552915, 818.26887548, 815.86220232, 820.67554864, 800.21882676, 794.20214385, 782.16877803, 773.74542196, 778.55876829, 779.76210487, 783.37211461, 785.77878778, 780.96544145, 779.76210487, 780.96544145, 782.16877803, 771.33874880, 771.33874880, 766.52540247, 758.10204640, 766.52540247, 746.06868058, 740.05199768, 734.03531477, 717.18860263, 711.17191972, 699.13855390, 694.32520758, 678.68183202, 672.66514911, 683.49517834, 700.34189048, 689.51186125, 688.30852467, 676.27515885, 661.83511987, 730.42530502, 703.95190023, 706.35857339, 688.30852467, 667.85180278, 667.85180278, 672.66514911, 667.85180278, 661.83511987, 634.15837850, 637.76838824, 677.47849543, 666.64846620, 675.07182227, 665.44512962, 658.22511013, 644.98840773, 631.75170533, 626.93835901, 623.32834926, 654.61510038, 636.56505166, 684.69851492, 661.83511987, 671.46181253, 678.68183202, 666.64846620, 660.63178329, 696.73188074, 695.52854416, 705.15523681, 691.91853441, 734.03531477, 709.96858314, 735.23865135, 734.03531477, 715.98526604, 724.40862212, 671.46181253, 719.59527579, 715.98526604, 712.37525630, 717.18860263, 737.64532451, 735.23865135, 750.88202691, 754.49203666, 777.35543171, 803.82883650, 794.20214385, 795.40548043, 786.98212436, 794.20214385, 790.59213410, 794.20214385, 779.76210487, 790.59213410, 762.91539273, 774.94875854, 812.25219257, 836.31892420, 854.36897293, 845.94561686, 849.55562660, 872.41902165, 872.41902165, 841.13227053, 860.38565584, 833.91225104, 843.53894369, 857.97898267, 885.65572405, 863.99566558, 865.19900216, 853.16563635, 839.92893395, 832.70891446, 829.09890471, 830.30224130, 806.23550966, 813.45552915, 819.47221206, 789.38879752, 792.99880727, 812.25219257, 819.47221206, 821.87888522, 809.84551941, 794.20214385, 795.40548043, 782.16877803, 771.33874880, 792.99880727, 778.55876829, 786.98212436, 761.71205615, 717.18860263, 715.98526604, 683.49517834, 682.29184176, 682.29184176, 617.31166635, 643.78507115, 620.92167610, 642.58173457, 614.90499319, 612.49832003, 641.37839799, 649.80175406, 644.98840773, 643.78507115, 672.66514911, 670.25847594, 629.34503217, 612.49832003, 626.93835901, 632.95504192, 648.59841748, 655.81843697, 671.46181253, 648.59841748, 628.14169559, 659.42844671, 661.83511987, 643.78507115, 636.56505166, 630.54836875, 607.68497370, 608.88831028, 610.09164687, 598.05828105, 601.66829079, 571.58487625, 577.60155916, 588.43158840, 593.24493472, 580.00823233, 584.82157865], 
							[1636.53775096, 1600.43765351, 1576.37092188, 1540.27082443, 1552.30419025, 1564.33755606, 1588.40428770, 1612.47101933, 1672.63784841, 1684.67121422, 1756.87140912, 1660.60448259, 1732.80467749, 1708.73794585, 1696.70458004, 1672.63784841, 1636.53775096, 1648.57111678, 1648.57111678, 1612.47101933, 1600.43765351, 1516.20409280, 1480.10399535, 1492.13736117, 1492.13736117, 1588.40428770, 1600.43765351, 1612.47101933, 1600.43765351, 1576.37092188, 1552.30419025, 1504.17072698, 1480.10399535, 1516.20409280, 1528.23745862, 1552.30419025, 1552.30419025, 1540.27082443, 1564.33755606, 1600.43765351, 1612.47101933, 1624.50438514, 1648.57111678, 1672.63784841, 1648.57111678, 1648.57111678, 1636.53775096, 1648.57111678, 1672.63784841, 1732.80467749, 1648.57111678, 1624.50438514, 1588.40428770, 1552.30419025, 1672.63784841, 1696.70458004, 1684.67121422, 1696.70458004, 1817.03823820, 1805.00487238, 1732.80467749, 1865.17170146, 1841.10496983, 1720.77131167, 1720.77131167, 1660.60448259, 1696.70458004, 1624.50438514, 1600.43765351, 1636.53775096, 1552.30419025, 1612.47101933, 1660.60448259, 1612.47101933, 1636.53775096, 1636.53775096, 1648.57111678, 1708.73794585, 1696.70458004, 1732.80467749, 1696.70458004, 1720.77131167, 1744.83804330, 1744.83804330, 1768.90477493, 1744.83804330, 1805.00487238, 1792.97150657, 1768.90477493, 1756.87140912, 1744.83804330, 1612.47101933, 1696.70458004, 1708.73794585, 1792.97150657, 1805.00487238, 1853.13833565, 1901.27179891, 1877.20506728, 1937.37189636, 1865.17170146, 1925.33853054, 1973.47199380, 1901.27179891, 1985.50535962, 1985.50535962, 2069.73892033, 2045.67218870, 2033.63882288, 1937.37189636, 1961.43862799, 1985.50535962, 2021.60545707, 1985.50535962, 2033.63882288, 2033.63882288, 2033.63882288, 2057.70555452, 2141.93911523, 2153.97248104, 2166.00584686, 2105.83901778, 2045.67218870, 2129.90574941, 2105.83901778, 2117.87238359, 2057.70555452, 2081.77228615, 2105.83901778, 2081.77228615, 2069.73892033, 2129.90574941, 2081.77228615, 2069.73892033, 2081.77228615, 2093.80565196, 2166.00584686, 2129.90574941, 2141.93911523, 2153.97248104, 2117.87238359, 2081.77228615, 2105.83901778, 2045.67218870, 2081.77228615, 2057.70555452, 2057.70555452, 2057.70555452, 2033.63882288, 2045.67218870, 1949.40526217, 1877.20506728, 1817.03823820, 1756.87140912, 1780.93814075, 1768.90477493, 1768.90477493, 1792.97150657, 1732.80467749, 1696.70458004, 1600.43765351, 1720.77131167, 1636.53775096, 1648.57111678, 1684.67121422, 1672.63784841, 1660.60448259, 1516.20409280, 1540.27082443, 1576.37092188, 1588.40428770, 1576.37092188, 1588.40428770, 1696.70458004, 1660.60448259, 1768.90477493, 1853.13833565, 1901.27179891, 1937.37189636, 2069.73892033, 2045.67218870, 2105.83901778, 2093.80565196, 2093.80565196, 2166.00584686, 2202.10594431, 2045.67218870, 2129.90574941, 2166.00584686, 2009.57209125, 2129.90574941, 2093.80565196, 2202.10594431, 2178.03921267, 2117.87238359, 2081.77228615, 2178.03921267, 2286.33950502, 2214.13931012, 2262.27277339, 2250.23940757, 2563.10691878, 2418.70652899, 2466.83999225, 2478.87335807, 2406.67316318, 2478.87335807, 2478.87335807, 2382.60643154, 2322.43960246, 2202.10594431, 2250.23940757, 2382.60643154, 2454.80662644, 2442.77326062, 2454.80662644, 2490.90672389, 2575.14028460, 2454.80662644, 2406.67316318, 2466.83999225, 2478.87335807, 2551.07355297, 2551.07355297, 2707.50730857, 2779.70750347, 2743.60740602, 2767.67413765, 2647.34047949, 2514.97345552, 2442.77326062, 2478.87335807, 2442.77326062, 2466.83999225, 2502.94008970, 2430.73989481, 2454.80662644, 2575.14028460, 2539.04018715, 2623.27374786, 2683.44057694, 2707.50730857, 2803.77423510, 2587.17365041, 2671.40721112, 2767.67413765, 2755.64077184, 2912.07452744, 2888.00779581, 3008.34145397, 2972.24135652, 2900.04116163, 3104.60838050, 3080.54164886, 2803.77423510, 2912.07452744, 3008.34145397, 3092.57501468, 3092.57501468, 3321.20896518, 3176.80857539, 3032.40818560, 3032.40818560, 3056.47491723, 3092.57501468, 3236.97540447, 3357.30906263, 3477.64272079, 3309.17559937, 3549.84291568, 3814.57696363, 3730.34340292, 3513.74281824, 3405.44252589, 3549.84291568, 3682.20993966, 3561.87628150, 3682.20993966, 3790.51023200, 3802.54359782, 3934.91062179, 4031.17754832, 4199.64466974, 4404.21188861, 4332.01169372, 4536.57891259, 4151.51120648, 4007.11081669, 4283.87823045, 4043.21091413, 4368.11179116], 
							[948.22922629, 937.39919706, 855.57230951, 784.57545120, 818.26887548, 765.32206589, 685.90185151, 707.56190997, 719.59527579, 691.91853441, 765.32206589, 776.15209512, 758.10204640, 778.55876829, 719.59527579, 697.93521732, 690.71519783, 671.46181253, 703.95190023, 691.91853441, 707.56190997, 682.29184176, 649.80175406, 613.70165661, 606.48163712, 655.81843697, 877.23236798, 990.34600665, 948.22922629, 885.65572405, 900.09576303, 824.28555839, 782.16877803, 806.23550966, 782.16877803, 764.11872931, 776.15209512, 750.88202691, 803.82883650, 818.26887548, 806.23550966, 860.38565584, 842.33560711, 857.97898267, 783.37211461, 794.20214385, 761.71205615, 759.30538298, 803.82883650, 833.91225104, 808.64218283, 818.26887548, 785.77878778, 772.54208538, 841.13227053, 812.25219257, 832.70891446, 913.33246543, 962.66926527, 979.51597741, 912.12912884, 973.49929450, 968.68594818, 977.10930425, 961.46592869, 998.76936272, 992.75267981, 951.83923604, 914.53580201, 900.09576303, 933.78918731, 962.66926527, 1055.32618205, 1045.69948940, 1070.96955761, 1074.57956736, 1044.49615282, 1068.56288445, 1049.30949914, 1069.76622103, 1061.34286496, 1038.47946991, 1046.90282598, 1046.90282598, 1033.66612358, 977.10930425, 969.88928476, 921.75582150, 833.91225104, 844.74228027, 817.06553890, 738.84866109, 770.13541222, 755.69537324, 788.18546094, 758.10204640, 788.18546094, 803.82883650, 918.14581175, 998.76936272, 966.27927501, 967.48261160, 938.60253364, 879.63904114, 920.55248491, 960.26259211, 959.05925553, 939.80587022, 915.73913859, 884.45238747, 926.56916782, 933.78918731, 962.66926527, 927.77250440, 938.60253364, 962.66926527, 956.65258236, 968.68594818, 962.66926527, 957.85591894, 984.32932374, 915.73913859, 926.56916782, 934.99252389, 926.56916782, 910.92579226, 947.02588971, 956.65258236, 973.49929450, 943.41587996, 960.26259211, 956.65258236, 915.73913859, 921.75582150, 936.19586048, 920.55248491, 904.90910935, 936.19586048, 888.06239721, 884.45238747, 892.87574354, 880.84237772, 889.26573379, 857.97898267, 851.96229976, 879.63904114, 896.48575328, 898.89242645, 898.89242645, 889.26573379, 882.04571430, 884.45238747, 795.40548043, 806.23550966, 826.69223155, 808.64218283, 780.96544145, 735.23865135, 699.13855390, 708.76524656, 693.12187099, 661.83511987, 687.10518809, 685.90185151, 719.59527579, 673.86848569, 706.35857339, 676.27515885, 666.64846620, 643.78507115, 626.93835901, 599.26161763, 569.17820309, 617.31166635, 557.14483728, 660.63178329, 660.63178329, 685.90185151, 679.88516860, 666.64846620, 657.02177355, 702.74856365, 729.22196844, 705.15523681, 732.83197819, 768.93207563, 722.00194895, 711.17191972, 749.67869033, 740.05199768, 748.47535375, 699.13855390, 726.81529528, 725.61195870, 728.01863186, 741.25533426, 818.26887548, 791.79547068, 791.79547068, 805.03217308, 819.47221206, 860.38565584, 856.77564609, 845.94561686, 854.36897293, 854.36897293, 857.97898267, 854.36897293, 845.94561686, 873.62235823, 884.45238747, 877.23236798, 866.40233874, 915.73913859, 922.95915808, 927.77250440, 915.73913859, 938.60253364, 914.53580201, 913.33246543, 950.63589945, 934.99252389, 938.60253364, 934.99252389, 950.63589945, 943.41587996, 967.48261160, 954.24590920, 943.41587996, 954.24590920, 951.83923604, 927.77250440, 891.67240696, 921.75582150, 959.05925553, 891.67240696, 914.53580201, 957.85591894, 968.68594818, 966.27927501, 942.21254338, 957.85591894, 966.27927501, 938.60253364, 918.14581175, 968.68594818, 939.80587022, 959.05925553, 931.38251415, 845.94561686, 912.12912884, 812.25219257, 796.60881701, 788.18546094, 695.52854416, 758.10204640, 731.62864161, 774.94875854, 753.28870007, 783.37211461, 835.11558762, 831.50557788, 796.60881701, 783.37211461, 824.28555839, 776.15209512, 742.45867084, 776.15209512, 801.42216334, 791.79547068, 826.69223155, 836.31892420, 877.23236798, 848.35229002, 849.55562660, 883.24905089, 885.65572405, 862.79232900, 843.53894369, 809.84551941, 791.79547068, 821.87888522, 813.45552915, 748.47535375, 741.25533426, 697.93521732, 703.95190023, 691.91853441, 702.74856365, 661.83511987, 696.73188074], 
							[651.00509064, 649.80175406, 625.73502243, 601.66829079, 613.70165661, 553.53482753, 518.63806666, 495.77467161, 490.96132529, 484.94464238, 523.45141299, 539.09478855, 519.84140325, 547.51814462, 534.28144222, 523.45141299, 505.40136427, 477.72462289, 542.70479830, 518.63806666, 517.43473008, 493.36799845, 486.14797896, 493.36799845, 451.25121810, 494.57133503, 540.29812513, 565.56819335, 555.94150069, 561.95818360, 571.58487625, 529.46809590, 539.09478855, 546.31480804, 541.50146171, 507.80803743, 522.24807641, 493.36799845, 529.46809590, 546.31480804, 533.07810564, 535.48477881, 506.60470085, 506.60470085, 482.53796922, 487.35131554, 483.74130580, 488.55465212, 506.60470085, 517.43473008, 493.36799845, 505.40136427, 510.21471059, 494.57133503, 509.01137401, 510.21471059, 493.36799845, 517.43473008, 527.06142274, 571.58487625, 528.26475932, 571.58487625, 565.56819335, 557.14483728, 576.39822258, 605.27830054, 612.49832003, 583.61824207, 560.75484702, 547.51814462, 539.09478855, 542.70479830, 557.14483728, 528.26475932, 553.53482753, 553.53482753, 552.33149095, 554.73816411, 570.38153967, 561.95818360, 551.12815437, 549.92481779, 546.31480804, 531.87476906, 523.45141299, 513.82472034, 501.79135452, 478.92795947, 434.40450595, 445.23453519, 428.38782305, 394.69439876, 438.01451570, 398.30440851, 419.96446697, 397.10107192, 419.96446697, 418.76113039, 469.30126682, 457.26790100, 436.81117912, 484.94464238, 469.30126682, 428.38782305, 459.67457417, 469.30126682, 465.69125707, 489.75798871, 466.89459366, 465.69125707, 468.09793024, 484.94464238, 469.30126682, 465.69125707, 478.92795947, 486.14797896, 489.75798871, 507.80803743, 522.24807641, 529.46809590, 536.68811539, 505.40136427, 502.99469110, 493.36799845, 495.77467161, 499.38468136, 488.55465212, 484.94464238, 496.97800820, 484.94464238, 488.55465212, 490.96132529, 482.53796922, 486.14797896, 478.92795947, 463.28458391, 465.69125707, 466.89459366, 486.14797896, 487.35131554, 487.35131554, 474.11461315, 475.31794973, 425.98114988, 434.40450595, 445.23453519, 431.99783279, 466.89459366, 472.91127656, 465.69125707, 474.11461315, 464.48792049, 428.38782305, 444.03119861, 450.04788151, 452.45455468, 438.01451570, 430.79449621, 415.15112065, 434.40450595, 417.55779381, 405.52442800, 388.67771585, 398.30440851, 433.20116937, 429.59115963, 434.40450595, 433.20116937, 424.77781330, 401.91441825, 401.91441825, 367.01765738, 339.34091601, 365.81432080, 339.34091601, 391.08438902, 388.67771585, 409.13443774, 413.94778407, 397.10107192, 389.88105243, 403.11775483, 433.20116937, 427.18448646, 412.74444748, 440.42118886, 405.52442800, 417.55779381, 444.03119861, 433.20116937, 429.59115963, 404.32109141, 442.82786202, 418.76113039, 409.13443774, 409.13443774, 416.35445723, 438.01451570, 431.99783279, 444.03119861, 445.23453519, 454.86122784, 459.67457417, 445.23453519, 494.57133503, 504.19802769, 555.94150069, 540.29812513, 522.24807641, 499.38468136, 509.01137401, 492.16466187, 502.99469110, 519.84140325, 551.12815437, 558.34817386, 553.53482753, 546.31480804, 533.07810564, 521.04473983, 541.50146171, 541.50146171, 557.14483728, 546.31480804, 555.94150069, 559.55151044, 561.95818360, 541.50146171, 531.87476906, 542.70479830, 519.84140325, 505.40136427, 466.89459366, 487.35131554, 511.41804717, 469.30126682, 475.31794973, 524.65474957, 531.87476906, 534.28144222, 505.40136427, 492.16466187, 527.06142274, 493.36799845, 502.99469110, 527.06142274, 505.40136427, 519.84140325, 499.38468136, 475.31794973, 493.36799845, 459.67457417, 431.99783279, 422.37114014, 385.06770611, 391.08438902, 376.64435004, 400.71108167, 382.66103294, 387.47437927, 394.69439876, 386.27104269, 385.06770611, 386.27104269, 411.54111090, 377.84768662, 351.37428182, 357.39096473, 377.84768662, 361.00097448, 383.86436953, 399.50774509, 415.15112065, 391.08438902, 392.28772560, 419.96446697, 403.11775483, 399.50774509, 394.69439876, 380.25435978, 363.40764764, 357.39096473, 383.86436953, 340.54425259, 329.71422336, 300.83414540, 288.80077958, 290.00411616, 291.20745274, 276.76741377, 253.90401872], 
							[1985.50535962, 1937.37189636, 1901.27179891, 1889.23843309, 1841.10496983, 1600.43765351, 1540.27082443, 1516.20409280, 1516.20409280, 1528.23745862, 1672.63784841, 1600.43765351, 1612.47101933, 1612.47101933, 1588.40428770, 1516.20409280, 1335.70360556, 1407.90380046, 1516.20409280, 1528.23745862, 1492.13736117, 1419.93716627, 1540.27082443, 1528.23745862, 1431.97053209, 1841.10496983, 1997.53872544, 2045.67218870, 1829.07160401, 1744.83804330, 1865.17170146, 1648.57111678, 1732.80467749, 1696.70458004, 1720.77131167, 1672.63784841, 1696.70458004, 1660.60448259, 1792.97150657, 1853.13833565, 1853.13833565, 1925.33853054, 1961.43862799, 2009.57209125, 1961.43862799, 1889.23843309, 1829.07160401, 1817.03823820, 1853.13833565, 1937.37189636, 1913.30516472, 1913.30516472, 1973.47199380, 1961.43862799, 2153.97248104, 2310.40623665, 2286.33950502, 2430.73989481, 2466.83999225, 2382.60643154, 2418.70652899, 2466.83999225, 2442.77326062, 2394.63979736, 2442.77326062, 2755.64077184, 2803.77423510, 2779.70750347, 2743.60740602, 2719.54067439, 2731.57404020, 2875.97442999, 2912.07452744, 2815.80760092, 2900.04116163, 2888.00779581, 2815.80760092, 2803.77423510, 2791.74086928, 2875.97442999, 2803.77423510, 2791.74086928, 2707.50730857, 2599.20701623, 2683.44057694, 2815.80760092, 2900.04116163, 2851.90769836, 2912.07452744, 2912.07452744, 2936.14125907, 2647.34047949, 2575.14028460, 2539.04018715, 2599.20701623, 2551.07355297, 2490.90672389, 2514.97345552, 2478.87335807, 2707.50730857, 2539.04018715, 2599.20701623, 2671.40721112, 2611.24038205, 2635.30711368, 2647.34047949, 2635.30711368, 2755.64077184, 2743.60740602, 2611.24038205, 2755.64077184, 2755.64077184, 2695.47394276, 2647.34047949, 2683.44057694, 2755.64077184, 2779.70750347, 2791.74086928, 2743.60740602, 2719.54067439, 2707.50730857, 2683.44057694, 2659.37384531, 2623.27374786, 2611.24038205, 2514.97345552, 2599.20701623, 2623.27374786, 2623.27374786, 2551.07355297, 2514.97345552, 2406.67316318, 2274.30613920, 2298.37287083, 2334.47296828, 2346.50633410, 2442.77326062, 2442.77326062, 2454.80662644, 2418.70652899, 2442.77326062, 2286.33950502, 2286.33950502, 2250.23940757, 2346.50633410, 2334.47296828, 2346.50633410, 2394.63979736, 2382.60643154, 2346.50633410, 2322.43960246, 2129.90574941, 2069.73892033, 2382.60643154, 2346.50633410, 2394.63979736, 2358.53969991, 2370.57306573, 2310.40623665, 2178.03921267, 1889.23843309, 1997.53872544, 2045.67218870, 2069.73892033, 2202.10594431, 2190.07257849, 2298.37287083, 2310.40623665, 2178.03921267, 2226.17267594, 2226.17267594, 2141.93911523, 2105.83901778, 2045.67218870, 1925.33853054, 2129.90574941, 2129.90574941, 2129.90574941, 2274.30613920, 2190.07257849, 2274.30613920, 2430.73989481, 2346.50633410, 2274.30613920, 2238.20604175, 2310.40623665, 2226.17267594, 2238.20604175, 2286.33950502, 2202.10594431, 2178.03921267, 1985.50535962, 2190.07257849, 2214.13931012, 2358.53969991, 2274.30613920, 2442.77326062, 2454.80662644, 2502.94008970, 2659.37384531, 2527.00682133, 2707.50730857, 2599.20701623, 2418.70652899, 2502.94008970, 2406.67316318, 2430.73989481, 2442.77326062, 2370.57306573, 2370.57306573, 2406.67316318, 2514.97345552, 2502.94008970, 2587.17365041, 2563.10691878, 2551.07355297, 2454.80662644, 2563.10691878, 2502.94008970, 2370.57306573, 2454.80662644, 2430.73989481, 2502.94008970, 2466.83999225, 2575.14028460, 2563.10691878, 2442.77326062, 2406.67316318, 2286.33950502, 2322.43960246, 2274.30613920, 2262.27277339, 2238.20604175, 2262.27277339, 2418.70652899, 2310.40623665, 2298.37287083, 2346.50633410, 2466.83999225, 2418.70652899, 2358.53969991, 2346.50633410, 2454.80662644, 2322.43960246, 2358.53969991, 2442.77326062, 2346.50633410, 2358.53969991, 2262.27277339, 2045.67218870, 1973.47199380, 2057.70555452, 1961.43862799, 1901.27179891, 1660.60448259, 1780.93814075, 1672.63784841, 1600.43765351, 1540.27082443, 1564.33755606, 1889.23843309, 1672.63784841, 1588.40428770, 1612.47101933, 1684.67121422, 1624.50438514, 1588.40428770, 1431.97053209, 1624.50438514, 1600.43765351, 1672.63784841, 1720.77131167, 1817.03823820, 1756.87140912, 1792.97150657, 1817.03823820, 2009.57209125, 2009.57209125, 1985.50535962, 1961.43862799, 1853.13833565, 2021.60545707, 2045.67218870, 1901.27179891, 1985.50535962, 1937.37189636, 1841.10496983, 1805.00487238, 1877.20506728, 1853.13833565, 1744.83804330], 
							[998.76936272, 971.09262134, 966.27927501, 919.34914833, 938.60253364, 922.95915808, 842.33560711, 752.08536349, 744.86534400, 697.93521732, 744.86534400, 734.03531477, 758.10204640, 784.57545120, 737.64532451, 713.57859288, 661.83511987, 619.71833952, 651.00509064, 654.61510038, 679.88516860, 629.34503217, 610.09164687, 631.75170533, 624.53168584, 671.46181253, 776.15209512, 770.13541222, 784.57545120, 839.92893395, 827.89556813, 783.37211461, 752.08536349, 740.05199768, 741.25533426, 722.00194895, 752.08536349, 748.47535375, 776.15209512, 830.30224130, 815.86220232, 800.21882676, 836.31892420, 797.81215359, 758.10204640, 734.03531477, 719.59527579, 693.12187099, 722.00194895, 764.11872931, 737.64532451, 744.86534400, 773.74542196, 786.98212436, 824.28555839, 847.14895344, 839.92893395, 837.52226079, 839.92893395, 879.63904114, 868.80901191, 1010.80272853, 980.71931399, 963.87260185, 983.12598716, 1028.85277726, 1022.83609435, 1004.78604563, 989.14267007, 979.51597741, 1012.00606512, 1028.85277726, 1044.49615282, 1034.86946017, 1101.05297215, 1081.79958685, 1119.10302088, 1150.38977200, 1143.16975251, 1172.04983047, 1203.33658159, 1156.40645491, 1141.96641593, 1149.18643542, 1140.76307935, 1188.89654261, 1188.89654261, 1145.57642567, 1062.54620154, 1064.95287471, 986.73599690, 896.48575328, 956.65258236, 892.87574354, 1009.59939195, 938.60253364, 954.24590920, 979.51597741, 1043.29281624, 1052.91950889, 981.92265058, 1125.11970378, 1067.35954787, 977.10930425, 1052.91950889, 1030.05611384, 1026.44610409, 1043.29281624, 1054.12284547, 995.15935297, 1021.63275777, 986.73599690, 995.15935297, 986.73599690, 986.73599690, 1027.64944068, 1043.29281624, 1081.79958685, 1076.98624052, 1072.17289419, 1085.40959659, 1075.78290394, 1019.22608460, 1012.00606512, 1022.83609435, 1034.86946017, 1026.44610409, 1016.81941144, 1043.29281624, 1056.52951863, 1037.27613333, 1034.86946017, 999.97269930, 995.15935297, 986.73599690, 969.88928476, 980.71931399, 1016.81941144, 991.54934323, 1016.81941144, 1020.42942119, 1037.27613333, 1003.58270904, 1004.78604563, 1024.03943093, 1001.17603588, 1008.39605537, 1043.29281624, 1027.64944068, 1022.83609435, 1028.85277726, 1020.42942119, 959.05925553, 991.54934323, 950.63589945, 906.11244594, 851.96229976, 803.82883650, 818.26887548, 836.31892420, 818.26887548, 819.47221206, 796.60881701, 829.09890471, 867.60567532, 896.48575328, 902.50243619, 831.50557788, 826.69223155, 780.96544145, 809.84551941, 764.11872931, 728.01863186, 780.96544145, 740.05199768, 854.36897293, 854.36897293, 873.62235823, 833.91225104, 811.04885599, 814.65886573, 889.26573379, 885.65572405, 886.85906063, 895.28241670, 909.72245568, 833.91225104, 830.30224130, 865.19900216, 831.50557788, 850.75896318, 806.23550966, 842.33560711, 829.09890471, 877.23236798, 872.41902165, 897.68908986, 948.22922629, 936.19586048, 991.54934323, 996.36268955, 1001.17603588, 977.10930425, 1037.27613333, 1043.29281624, 1054.12284547, 1049.30949914, 1064.95287471, 1012.00606512, 1005.98938221, 1046.90282598, 1016.81941144, 1026.44610409, 1107.06965506, 1121.50969404, 1083.00292343, 1093.83295266, 1098.64629899, 1076.98624052, 1074.57956736, 1126.32304037, 1153.99978174, 1129.93305011, 1110.67966481, 1132.33972327, 1167.23648414, 1191.30321577, 1185.28653286, 1166.03314756, 1164.82981098, 1132.33972327, 1107.06965506, 1070.96955761, 1076.98624052, 1095.03628924, 1010.80272853, 1001.17603588, 1057.73285522, 1070.96955761, 1139.55974276, 1105.86631848, 1111.88300139, 1158.81312807, 1101.05297215, 1070.96955761, 1125.11970378, 1080.59625027, 1119.10302088, 1076.98624052, 980.71931399, 1033.66612358, 985.53266032, 897.68908986, 910.92579226, 755.69537324, 831.50557788, 806.23550966, 803.82883650, 823.08222181, 797.81215359, 842.33560711, 805.03217308, 786.98212436, 766.52540247, 838.72559737, 750.88202691, 683.49517834, 697.93521732, 732.83197819, 711.17191972, 740.05199768, 778.55876829, 855.57230951, 914.53580201, 949.43256287, 1004.78604563, 992.75267981, 992.75267981, 945.82255313, 886.85906063, 841.13227053, 884.45238747, 900.09576303, 782.16877803, 730.42530502, 706.35857339, 661.83511987, 657.02177355, 667.85180278, 637.76838824, 653.41176380], 
							[1696.70458004, 1780.93814075, 1708.73794585, 1684.67121422, 1636.53775096, 1492.13736117, 1516.20409280, 1468.07062954, 1468.07062954, 1552.30419025, 1600.43765351, 1528.23745862, 1552.30419025, 1564.33755606, 1552.30419025, 1516.20409280, 1444.00389791, 1335.70360556, 1444.00389791, 1492.13736117, 1528.23745862, 1371.80370301, 1359.77033719, 1347.73697138, 1287.57014230, 1480.10399535, 1660.60448259, 1660.60448259, 1600.43765351, 1588.40428770, 1648.57111678, 1564.33755606, 1528.23745862, 1564.33755606, 1528.23745862, 1528.23745862, 1552.30419025, 1456.03726372, 1564.33755606, 1624.50438514, 1528.23745862, 1528.23745862, 1552.30419025, 1576.37092188, 1528.23745862, 1468.07062954, 1444.00389791, 1395.87043464, 1480.10399535, 1492.13736117, 1468.07062954, 1480.10399535, 1456.03726372, 1431.97053209, 1504.17072698, 1492.13736117, 1492.13736117, 1552.30419025, 1612.47101933, 1648.57111678, 1600.43765351, 1672.63784841, 1636.53775096, 1624.50438514, 1636.53775096, 1672.63784841, 1672.63784841, 1624.50438514, 1600.43765351, 1612.47101933, 1624.50438514, 1612.47101933, 1624.50438514, 1600.43765351, 1660.60448259, 1612.47101933, 1660.60448259, 1636.53775096, 1660.60448259, 1732.80467749, 1708.73794585, 1708.73794585, 1696.70458004, 1672.63784841, 1672.63784841, 1672.63784841, 1708.73794585, 1660.60448259, 1612.47101933, 1540.27082443, 1528.23745862, 1395.87043464, 1456.03726372, 1419.93716627, 1504.17072698, 1371.80370301, 1395.87043464, 1636.53775096, 1624.50438514, 1660.60448259, 1564.33755606, 1684.67121422, 1624.50438514, 1528.23745862, 1588.40428770, 1564.33755606, 1564.33755606, 1576.37092188, 1504.17072698, 1480.10399535, 1540.27082443, 1612.47101933, 1564.33755606, 1516.20409280, 1540.27082443, 1552.30419025, 1576.37092188, 1552.30419025, 1552.30419025, 1648.57111678, 1588.40428770, 1540.27082443, 1612.47101933, 1564.33755606, 1588.40428770, 1564.33755606, 1612.47101933, 1564.33755606, 1600.43765351, 1588.40428770, 1612.47101933, 1612.47101933, 1600.43765351, 1612.47101933, 1600.43765351, 1588.40428770, 1576.37092188, 1564.33755606, 1588.40428770, 1588.40428770, 1564.33755606, 1540.27082443, 1504.17072698, 1480.10399535, 1516.20409280, 1480.10399535, 1480.10399535, 1528.23745862, 1540.27082443, 1516.20409280, 1528.23745862, 1504.17072698, 1395.87043464, 1444.00389791, 1431.97053209, 1444.00389791, 1383.83706883, 1299.60350811, 1227.40331322, 1227.40331322, 1194.91322552, 1185.28653286, 1179.26984996, 1203.33658159, 1227.40331322, 1215.36994740, 1239.43667904, 1203.33658159, 1215.36994740, 1169.64315730, 1178.06651337, 1062.54620154, 1067.35954787, 1097.44296241, 1048.10616256, 1186.48986945, 1149.18643542, 1149.18643542, 1179.26984996, 1157.60979149, 1173.25316705, 1215.36994740, 1191.30321577, 1215.36994740, 1239.43667904, 1251.47004485, 1239.43667904, 1263.50341067, 1275.53677648, 1251.47004485, 1227.40331322, 1191.30321577, 1239.43667904, 1227.40331322, 1299.60350811, 1287.57014230, 1335.70360556, 1299.60350811, 1287.57014230, 1347.73697138, 1383.83706883, 1407.90380046, 1407.90380046, 1395.87043464, 1456.03726372, 1468.07062954, 1480.10399535, 1468.07062954, 1395.87043464, 1359.77033719, 1347.73697138, 1359.77033719, 1371.80370301, 1468.07062954, 1516.20409280, 1456.03726372, 1504.17072698, 1552.30419025, 1492.13736117, 1468.07062954, 1516.20409280, 1552.30419025, 1564.33755606, 1492.13736117, 1576.37092188, 1588.40428770, 1576.37092188, 1528.23745862, 1492.13736117, 1492.13736117, 1419.93716627, 1456.03726372, 1431.97053209, 1468.07062954, 1492.13736117, 1383.83706883, 1395.87043464, 1395.87043464, 1444.00389791, 1492.13736117, 1431.97053209, 1456.03726372, 1516.20409280, 1468.07062954, 1468.07062954, 1504.17072698, 1480.10399535, 1516.20409280, 1552.30419025, 1516.20409280, 1504.17072698, 1480.10399535, 1395.87043464, 1419.93716627, 1263.50341067, 1311.63687393, 1227.40331322, 1311.63687393, 1227.40331322, 1311.63687393, 1299.60350811, 1311.63687393, 1239.43667904, 1184.08319628, 1203.33658159, 1167.23648414, 1155.20311832, 1161.21980123, 1173.25316705, 1323.67023975, 1359.77033719, 1371.80370301, 1444.00389791, 1419.93716627, 1407.90380046, 1456.03726372, 1468.07062954, 1456.03726372, 1431.97053209, 1395.87043464, 1395.87043464, 1407.90380046, 1395.87043464, 1323.67023975, 1323.67023975, 1371.80370301, 1407.90380046, 1323.67023975, 1323.67023975, 1323.67023975, 1227.40331322], 
							[1783.11617996, 1859.69652001, 1804.99283902, 1804.99283902, 1903.44983812, 1881.57317907, 1848.75819049, 1772.17785044, 1739.36286186, 1739.36286186, 1815.93116854, 1837.81986096, 1859.69652001, 1870.63484954, 1848.75819049, 1804.99283902, 1761.23952091, 1673.72085133, 1761.23952091, 1772.17785044, 1728.42453233, 1695.59751038, 1717.48620280, 1750.30119138, 1695.59751038, 1673.72085133, 1739.36286186, 1750.30119138, 1706.53583991, 1695.59751038, 1706.53583991, 1695.59751038, 1717.48620280, 1728.42453233, 1728.42453233, 1706.53583991, 1783.11617996, 1673.72085133, 1761.23952091, 1815.93116854, 1848.75819049, 1804.99283902, 1881.57317907, 1859.69652001, 1772.17785044, 1739.36286186, 1783.11617996, 1772.17785044, 1804.99283902, 1815.93116854, 1837.81986096, 1826.86949807, 1859.69652001, 1881.57317907, 2001.90683723, 2122.24049539, 1969.09184865, 1892.51150859, 2056.59848486, 2023.78349628, 1990.96850770, 2133.17882491, 2144.11715444, 2045.66015533, 2078.47514391, 2078.47514391, 2078.47514391, 2012.84516675, 1969.09184865, 2056.59848486, 2034.72182581, 2034.72182581, 2111.30216586, 2001.90683723, 2100.36383633, 2078.47514391, 2111.30216586, 2122.24049539, 2133.17882491, 2176.93214302, 2187.87047255, 2220.69749449, 2286.32747165, 2428.53778886, 2351.96948218, 2373.84614123, 2395.72280028, 2395.72280028, 2319.14246023, 2341.01911928, 2176.93214302, 2056.59848486, 1980.03017817, 2001.90683723, 2133.17882491, 2001.90683723, 2001.90683723, 2023.78349628, 2144.11715444, 2144.11715444, 2089.41347344, 2155.05548397, 2089.41347344, 2001.90683723, 2067.53681439, 2089.41347344, 2133.17882491, 2089.41347344, 2034.72182581, 1990.96850770, 2045.66015533, 2034.72182581, 2012.84516675, 1980.03017817, 1980.03017817, 1980.03017817, 2001.90683723, 2023.78349628, 2012.84516675, 1969.09184865, 1969.09184865, 1903.44983812, 1892.51150859, 1859.69652001, 1903.44983812, 1914.38816765, 1914.38816765, 1881.57317907, 1881.57317907, 1870.63484954, 1848.75819049, 1859.69652001, 1815.93116854, 1826.86949807, 1804.99283902, 1772.17785044, 1772.17785044, 1728.42453233, 1695.59751038, 1619.02920370, 1662.78252180, 1629.96753322, 1750.30119138, 1662.78252180, 1695.59751038, 1750.30119138, 1750.30119138, 1750.30119138, 1783.11617996, 1750.30119138, 1804.99283902, 1783.11617996, 1794.05450949, 1826.86949807, 1804.99283902, 1750.30119138, 1706.53583991, 1597.15254464, 1498.69554554, 1542.44886365, 1498.69554554, 1476.81888649, 1454.93019407, 1454.93019407, 1520.57220459, 1531.51053412, 1542.44886365, 1487.75721601, 1465.88055696, 1433.05353501, 1443.99186454, 1454.93019407, 1378.36188738, 1433.05353501, 1323.65820638, 1422.11520549, 1465.88055696, 1487.75721601, 1487.75721601, 1454.93019407, 1487.75721601, 1531.51053412, 1542.44886365, 1553.38719317, 1498.69554554, 1575.26385222, 1476.81888649, 1487.75721601, 1542.44886365, 1498.69554554, 1542.44886365, 1487.75721601, 1509.63387507, 1465.88055696, 1454.93019407, 1454.93019407, 1509.63387507, 1487.75721601, 1509.63387507, 1542.44886365, 1575.26385222, 1629.96753322, 1619.02920370, 1629.96753322, 1629.96753322, 1586.21421512, 1662.78252180, 1662.78252180, 1619.02920370, 1608.09087417, 1564.32552270, 1553.38719317, 1564.32552270, 1673.72085133, 1684.65918086, 1673.72085133, 1662.78252180, 1684.65918086, 1695.59751038, 1695.59751038, 1706.53583991, 1706.53583991, 1739.36286186, 1717.48620280, 1750.30119138, 1794.05450949, 1804.99283902, 1728.42453233, 1684.65918086, 1695.59751038, 1662.78252180, 1640.90586275, 1608.09087417, 1629.96753322, 1629.96753322, 1575.26385222, 1575.26385222, 1695.59751038, 1706.53583991, 1728.42453233, 1728.42453233, 1673.72085133, 1717.48620280, 1695.59751038, 1651.84419228, 1651.84419228, 1662.78252180, 1629.96753322, 1629.96753322, 1542.44886365, 1531.51053412, 1520.57220459, 1443.99186454, 1443.99186454, 1258.02822922, 1334.60856927, 1290.84321780, 1356.48522833, 1345.54689880, 1367.42355785, 1367.42355785, 1345.54689880, 1301.78154733, 1301.78154733, 1411.17687596, 1299.60350811, 1263.50341067, 1251.47004485, 1359.77033719, 1456.03726372, 1612.47101933, 1492.13736117, 1552.30419025, 1528.23745862, 1576.37092188, 1612.47101933, 1696.70458004, 1708.73794585, 1720.77131167, 1696.70458004, 1756.87140912, 1792.97150657, 1780.93814075, 1684.67121422, 1624.50438514, 1612.47101933, 1516.20409280, 1528.23745862, 1612.47101933, 1564.33755606, 1540.27082443], 
							[801.09726246, 762.15729068, 778.84756907, 735.45525193, 728.77673391, 710.97938587, 673.14648374, 634.20651196, 633.08740894, 611.95681857, 686.50351980, 641.99209964, 645.32534197, 654.23003268, 625.30182126, 617.51623357, 578.57626179, 559.65981073, 578.57626179, 620.84947590, 606.38537019, 567.44539841, 595.26654018, 601.93302484, 574.12391644, 607.50447321, 672.03941409, 656.45620535, 682.05117444, 680.93207142, 669.81324141, 625.30182126, 628.64709695, 627.52799393, 624.19475160, 623.07564858, 656.45620535, 646.44444499, 655.34913570, 662.01562036, 650.89679035, 647.55151465, 670.92031106, 656.45620535, 639.76592697, 613.06388822, 599.71888553, 585.24274646, 607.50447321, 620.84947590, 618.62330323, 626.42092428, 638.65885731, 650.89679035, 708.75321319, 702.07469516, 774.56369084, 801.43419670, 787.41532553, 806.10314264, 769.88271153, 799.09972374, 792.08427146, 771.06198138, 778.06540029, 786.24808904, 794.41874443, 776.89816380, 751.19489442, 769.88271153, 762.87929263, 771.06198138, 781.56710974, 781.56710974, 795.58598092, 804.93590616, 814.28583139, 794.41874443, 789.74979850, 800.26696022, 800.26696022, 774.56369084, 774.56369084, 760.54481966, 775.73092732, 745.35871200, 768.71547505, 724.32438856, 664.74719440, 648.39385026, 621.52334439, 591.13909571, 626.19229033, 600.48902094, 629.69399978, 606.33723673, 614.50789212, 607.50447321, 651.89555971, 662.41272143, 646.04734392, 698.62111917, 665.91443088, 619.17683806, 654.23003268, 664.74719440, 668.24890385, 689.28322730, 689.28322730, 662.41272143, 676.43159261, 660.06621510, 669.41614034, 653.06279619, 651.89555971, 695.11940972, 677.59882909, 684.60224800, 686.93672097, 671.75061331, 696.28664620, 671.75061331, 671.75061331, 648.39385026, 657.73174213, 665.91443088, 678.76606558, 682.26777503, 688.10395745, 683.43501151, 696.28664620, 692.78493675, 678.76606558, 678.76606558, 678.76606558, 657.73174213, 660.06621510, 660.06621510, 654.23003268, 650.72832323, 650.72832323, 660.06621510, 664.74719440, 661.24548495, 669.41614034, 665.91443088, 668.24890385, 681.10053854, 678.76606558, 669.41614034, 665.91443088, 667.08166737, 630.86123626, 657.73174213, 642.54563447, 641.37839799, 636.70945205, 635.54221557, 611.00618267, 634.37497908, 633.19570923, 619.17683806, 600.48902094, 609.83894618, 637.87668853, 630.86123626, 633.19570923, 633.19570923, 622.69058087, 614.50789212, 627.35952681, 589.97185922, 586.47014977, 601.65625743, 579.46673086, 625.02505384, 619.17683806, 627.35952681, 625.02505384, 620.35610791, 619.17683806, 674.08508627, 668.24890385, 675.26435612, 675.26435612, 693.95217324, 672.91784979, 667.08166737, 699.78835566, 677.59882909, 688.10395745, 663.57995792, 684.60224800, 678.76606558, 685.76948448, 682.26777503, 692.78493675, 698.62111917, 702.13486199, 697.45388269, 702.13486199, 712.63999035, 717.32096965, 714.97446332, 728.99333449, 718.48820613, 719.65544262, 717.32096965, 704.46933496, 693.95217324, 705.63657144, 711.47275386, 724.32438856, 755.69537324, 784.57545120, 759.30538298, 784.57545120, 791.79547068, 759.30538298, 764.11872931, 783.37211461, 767.72873905, 777.35543171, 772.54208538, 780.96544145, 780.96544145, 782.16877803, 779.76210487, 770.13541222, 768.93207563, 764.11872931, 756.89870982, 735.23865135, 752.08536349, 768.93207563, 722.00194895, 736.44198793, 749.67869033, 755.69537324, 759.30538298, 738.84866109, 746.06868058, 767.72873905, 738.84866109, 735.23865135, 762.91539273, 749.67869033, 756.89870982, 737.64532451, 715.98526604, 725.61195870, 707.56190997, 691.91853441, 681.08850518, 629.34503217, 643.78507115, 653.41176380, 653.41176380, 626.93835901, 641.37839799, 665.44512962, 661.83511987, 679.88516860, 672.66514911, 723.20528553, 658.22511013, 646.19174431, 660.63178329, 663.03845645, 640.17506140, 660.63178329, 651.00509064, 667.85180278, 648.59841748, 660.63178329, 667.85180278, 654.61510038, 659.42844671, 655.81843697, 649.80175406, 655.81843697, 655.81843697, 658.22511013, 637.76838824, 646.19174431, 613.70165661, 611.29498345, 625.73502243, 624.53168584, 608.88831028, 617.31166635], 
							[385.06770611, 365.81432080, 353.78095499, 332.12089652, 346.56093550, 340.54425259, 327.30755019, 336.93424284, 324.90087703, 330.91755994, 345.35759892, 336.93424284, 341.74758917, 342.95092575, 334.52756968, 328.51088677, 305.64749172, 298.42747223, 320.08753070, 310.46083805, 302.03748198, 286.39410642, 294.81746249, 282.78409667, 262.32737479, 282.78409667, 342.95092575, 345.35759892, 332.12089652, 323.69754045, 330.91755994, 305.64749172, 303.24081856, 311.66417463, 312.86751121, 312.86751121, 305.64749172, 300.83414540, 297.22413565, 299.63080882, 308.05416489, 317.68085754, 320.08753070, 309.25750147, 306.85082830, 294.81746249, 292.41078933, 287.59744300, 315.27418438, 310.46083805, 297.22413565, 311.66417463, 304.44415514, 285.19076984, 321.29086728, 326.10421361, 336.93424284, 409.13443774, 397.10107192, 386.27104269, 362.20431106, 403.11775483, 400.71108167, 387.47437927, 407.93110116, 407.93110116, 409.13443774, 403.11775483, 373.03434029, 373.03434029, 362.20431106, 381.45769636, 374.23767687, 362.20431106, 383.86436953, 374.23767687, 388.67771585, 381.45769636, 373.03434029, 370.62766713, 339.34091601, 338.13757943, 336.93424284, 323.69754045, 328.51088677, 329.71422336, 338.13757943, 328.51088677, 311.66417463, 303.24081856, 298.42747223, 276.76741377, 287.59744300, 285.19076984, 294.81746249, 282.78409667, 287.59744300, 291.20745274, 298.42747223, 297.22413565, 287.59744300, 318.88419412, 308.05416489, 290.00411616, 303.24081856, 300.83414540, 303.24081856, 317.68085754, 310.46083805, 296.02079907, 302.03748198, 308.05416489, 315.27418438, 308.05416489, 311.66417463, 322.49420387, 329.71422336, 340.54425259, 339.34091601, 341.74758917, 335.73090626, 329.71422336, 314.07084779, 321.29086728, 318.88419412, 318.88419412, 312.86751121, 335.73090626, 339.34091601, 340.54425259, 336.93424284, 353.78095499, 336.93424284, 330.91755994, 347.76427208, 351.37428182, 338.13757943, 348.96760866, 348.96760866, 357.39096473, 357.39096473, 332.12089652, 338.13757943, 321.29086728, 317.68085754, 315.27418438, 321.29086728, 333.32423310, 332.12089652, 323.69754045, 327.30755019, 304.44415514, 326.10421361, 323.69754045, 318.88419412, 297.22413565, 306.85082830, 299.63080882, 294.81746249, 283.98743325, 267.14072111, 277.97075035, 273.15740402, 281.58076009, 285.19076984, 282.78409667, 282.78409667, 270.75073086, 271.95406744, 267.14072111, 261.12403820, 245.48066264, 237.05730657, 239.46397974, 239.46397974, 258.71736504, 277.97075035, 263.53071137, 264.73404795, 268.34405769, 279.17408693, 294.81746249, 276.76741377, 279.17408693, 286.39410642, 315.27418438, 303.24081856, 318.88419412, 321.29086728, 314.07084779, 317.68085754, 298.42747223, 302.03748198, 300.83414540, 316.47752096, 338.13757943, 341.74758917, 342.95092575, 326.10421361, 339.34091601, 334.52756968, 354.98429157, 344.15426233, 347.76427208, 344.15426233, 371.83100371, 363.40764764, 352.57761841, 352.57761841, 354.98429157, 346.56093550, 351.37428182, 363.40764764, 371.83100371, 377.84768662, 411.54111090, 388.67771585, 398.30440851, 394.69439876, 377.84768662, 381.45769636, 375.44101346, 380.25435978, 376.64435004, 389.88105243, 395.89773534, 399.50774509, 394.69439876, 365.81432080, 364.61098422, 359.79763789, 368.22099397, 350.17094524, 358.59430131, 361.00097448, 345.35759892, 344.15426233, 339.34091601, 340.54425259, 341.74758917, 344.15426233, 342.95092575, 348.96760866, 340.54425259, 344.15426233, 346.56093550, 341.74758917, 352.57761841, 336.93424284, 323.69754045, 322.49420387, 328.51088677, 314.07084779, 314.07084779, 271.95406744, 310.46083805, 275.56407718, 296.02079907, 285.19076984, 311.66417463, 326.10421361, 311.66417463, 304.44415514, 314.07084779, 318.88419412, 316.47752096, 318.88419412, 314.07084779, 310.46083805, 324.90087703, 330.91755994, 320.08753070, 316.47752096, 303.24081856, 275.56407718, 291.20745274, 286.39410642, 282.78409667, 290.00411616, 281.58076009, 268.34405769, 264.73404795, 268.34405769, 255.10735530, 232.24396025, 235.85396999, 229.83728708, 231.04062366, 228.63395050, 231.04062366, 214.19391152], 
							[734.03531477, 693.12187099, 683.49517834, 702.74856365, 697.93521732, 695.52854416, 717.18860263, 758.10204640, 768.93207563, 752.08536349, 796.60881701, 760.50871956, 776.15209512, 766.52540247, 772.54208538, 758.10204640, 723.20528553, 759.30538298, 791.79547068, 783.37211461, 778.55876829, 740.05199768, 738.84866109, 723.20528553, 655.81843697, 758.10204640, 785.77878778, 782.16877803, 734.03531477, 728.01863186, 734.03531477, 681.08850518, 681.08850518, 705.15523681, 676.27515885, 701.54522707, 718.39193921, 717.18860263, 732.83197819, 746.06868058, 744.86534400, 761.71205615, 776.15209512, 768.93207563, 764.11872931, 749.67869033, 738.84866109, 724.40862212, 738.84866109, 737.64532451, 707.56190997, 691.91853441, 695.52854416, 677.47849543, 743.66200742, 742.45867084, 791.79547068, 818.26887548, 847.14895344, 850.75896318, 824.28555839, 931.38251415, 897.68908986, 861.58899242, 880.84237772, 890.46907037, 892.87574354, 878.43570456, 837.52226079, 818.26887548, 824.28555839, 809.84551941, 841.13227053, 830.30224130, 859.18231925, 839.92893395, 839.92893395, 830.30224130, 842.33560711, 857.97898267, 836.31892420, 800.21882676, 844.74228027, 823.08222181, 833.91225104, 815.86220232, 841.13227053, 839.92893395, 830.30224130, 826.69223155, 801.42216334, 752.08536349, 806.23550966, 809.84551941, 841.13227053, 813.45552915, 818.26887548, 845.94561686, 838.72559737, 842.33560711, 821.87888522, 872.41902165, 897.68908986, 863.99566558, 909.72245568, 915.73913859, 938.60253364, 966.27927501, 942.21254338, 894.07908012, 938.60253364, 938.60253364, 932.58585073, 926.56916782, 933.78918731, 951.83923604, 948.22922629, 987.93933348, 1003.58270904, 1038.47946991, 1024.03943093, 992.75267981, 972.29595792, 973.49929450, 941.00920680, 918.14581175, 882.04571430, 932.58585073, 908.51911910, 919.34914833, 904.90910935, 928.97584099, 904.90910935, 895.28241670, 903.70577277, 896.48575328, 897.68908986, 931.38251415, 903.70577277, 922.95915808, 895.28241670, 855.57230951, 851.96229976, 818.26887548, 854.36897293, 845.94561686, 841.13227053, 847.14895344, 870.01234849, 851.96229976, 813.45552915, 767.72873905, 752.08536349, 758.10204640, 756.89870982, 740.05199768, 729.22196844, 741.25533426, 712.37525630, 682.29184176, 665.44512962, 708.76524656, 685.90185151, 693.12187099, 701.54522707, 672.66514911, 679.88516860, 659.42844671, 655.81843697, 648.59841748, 637.76838824, 595.65160789, 632.95504192, 648.59841748, 646.19174431, 691.91853441, 743.66200742, 731.62864161, 747.27201717, 772.54208538, 791.79547068, 853.16563635, 820.67554864, 874.82569481, 913.33246543, 920.55248491, 859.18231925, 872.41902165, 894.07908012, 847.14895344, 906.11244594, 882.04571430, 920.55248491, 890.46907037, 867.60567532, 862.79232900, 907.31578252, 968.68594818, 928.97584099, 950.63589945, 973.49929450, 1038.47946991, 990.34600665, 996.36268955, 1004.78604563, 1010.80272853, 1022.83609435, 992.75267981, 980.71931399, 978.31264083, 918.14581175, 924.16249466, 938.60253364, 977.10930425, 985.53266032, 993.95601639, 986.73599690, 984.32932374, 966.27927501, 945.82255313, 949.43256287, 908.51911910, 903.70577277, 904.90910935, 949.43256287, 912.12912884, 938.60253364, 920.55248491, 914.53580201, 877.23236798, 870.01234849, 866.40233874, 860.38565584, 878.43570456, 880.84237772, 844.74228027, 839.92893395, 854.36897293, 854.36897293, 919.34914833, 906.11244594, 891.67240696, 920.55248491, 882.04571430, 861.58899242, 873.62235823, 870.01234849, 890.46907037, 882.04571430, 861.58899242, 886.85906063, 854.36897293, 878.43570456, 876.02903140, 829.09890471, 848.35229002, 836.31892420, 848.35229002, 850.75896318, 868.80901191, 872.41902165, 818.26887548, 806.23550966, 803.82883650, 831.50557788, 830.30224130, 848.35229002, 857.97898267, 845.94561686, 833.91225104, 906.11244594, 884.45238747, 872.41902165, 857.97898267, 860.38565584, 891.67240696, 908.51911910, 901.29909961, 890.46907037, 859.18231925, 849.55562660, 904.90910935, 880.84237772, 926.56916782, 867.60567532, 878.43570456, 872.41902165, 801.42216334, 818.26887548, 782.16877803, 774.94875854], 
							[1034.86946017, 1045.69948940, 1044.49615282, 998.76936272, 983.12598716, 948.22922629, 906.11244594, 790.59213410, 771.33874880, 741.25533426, 830.30224130, 842.33560711, 873.62235823, 888.06239721, 830.30224130, 867.60567532, 790.59213410, 705.15523681, 764.11872931, 747.27201717, 741.25533426, 663.03845645, 663.03845645, 665.44512962, 613.70165661, 673.86848569, 752.08536349, 795.40548043, 854.36897293, 859.18231925, 860.38565584, 842.33560711, 830.30224130, 801.42216334, 788.18546094, 758.10204640, 765.32206589, 747.27201717, 802.62549992, 802.62549992, 784.57545120, 789.38879752, 805.03217308, 771.33874880, 734.03531477, 717.18860263, 703.95190023, 673.86848569, 731.62864161, 746.06868058, 742.45867084, 761.71205615, 748.47535375, 747.27201717, 806.23550966, 783.37211461, 809.84551941, 813.45552915, 819.47221206, 832.70891446, 844.74228027, 873.62235823, 921.75582150, 962.66926527, 986.73599690, 1022.83609435, 1022.83609435, 973.49929450, 914.53580201, 891.67240696, 874.82569481, 906.11244594, 954.24590920, 903.70577277, 927.77250440, 907.31578252, 920.55248491, 920.55248491, 944.61921655, 938.60253364, 914.53580201, 903.70577277, 913.33246543, 908.51911910, 900.09576303, 877.23236798, 854.36897293, 784.57545120, 682.29184176, 681.08850518, 669.05513936, 602.87162738, 647.39508089, 643.78507115, 695.52854416, 653.41176380, 649.80175406, 688.30852467, 755.69537324, 749.67869033, 720.79861237, 812.25219257, 794.20214385, 725.61195870, 753.28870007, 774.94875854, 792.99880727, 824.28555839, 788.18546094, 756.89870982, 789.38879752, 806.23550966, 796.60881701, 770.13541222, 783.37211461, 795.40548043, 803.82883650, 817.06553890, 817.06553890, 820.67554864, 833.91225104, 796.60881701, 789.38879752, 795.40548043, 813.45552915, 794.20214385, 773.74542196, 792.99880727, 797.81215359, 771.33874880, 782.16877803, 782.16877803, 749.67869033, 728.01863186, 717.18860263, 697.93521732, 713.57859288, 737.64532451, 729.22196844, 718.39193921, 715.98526604, 706.35857339, 695.52854416, 661.83511987, 638.97172482, 641.37839799, 648.59841748, 697.93521732, 669.05513936, 651.00509064, 687.10518809, 684.69851492, 696.73188074, 691.91853441, 682.29184176, 699.13855390, 715.98526604, 642.58173457, 589.63492498, 590.83826156, 578.80489574, 560.75484702, 542.70479830, 604.07496396, 631.75170533, 606.48163712, 630.54836875, 582.41490549, 578.80489574, 535.48477881, 535.48477881, 506.60470085, 469.30126682, 511.41804717, 469.30126682, 541.50146171, 535.48477881, 530.67143248, 541.50146171, 543.90813488, 543.90813488, 583.61824207, 581.21156891, 649.80175406, 605.27830054, 613.70165661, 594.44827130, 587.22825181, 607.68497370, 583.61824207, 594.44827130, 557.14483728, 578.80489574, 624.53168584, 638.97172482, 613.70165661, 617.31166635, 634.15837850, 614.90499319, 626.93835901, 632.95504192, 647.39508089, 671.46181253, 652.20842722, 673.86848569, 691.91853441, 681.08850518, 684.69851492, 644.98840773, 631.75170533, 661.83511987, 638.97172482, 648.59841748, 671.46181253, 726.81529528, 699.13855390, 715.98526604, 729.22196844, 714.78192946, 712.37525630, 726.81529528, 724.40862212, 718.39193921, 719.59527579, 711.17191972, 719.59527579, 717.18860263, 701.54522707, 661.83511987, 687.10518809, 655.81843697, 630.54836875, 605.27830054, 628.14169559, 629.34503217, 619.71833952, 614.90499319, 643.78507115, 658.22511013, 666.64846620, 636.56505166, 632.95504192, 660.63178329, 611.29498345, 618.51500294, 658.22511013, 636.56505166, 648.59841748, 611.29498345, 564.36485676, 570.38153967, 527.06142274, 511.41804717, 511.41804717, 457.26790100, 476.52128631, 470.50460340, 484.94464238, 457.26790100, 458.47123758, 471.70793998, 457.26790100, 447.64120835, 433.20116937, 460.87791075, 424.77781330, 376.64435004, 392.28772560, 410.33777432, 389.88105243, 425.98114988, 446.43787177, 481.33463264, 453.65789126, 465.69125707, 487.35131554, 476.52128631, 492.16466187, 471.70793998, 448.84454493, 422.37114014, 415.15112065, 421.16780356, 364.61098422, 353.78095499, 332.12089652, 314.07084779, 318.88419412, 323.69754045, 296.02079907, 290.00411616], 
							[418.76113039, 404.32109141, 373.03434029, 373.03434029, 362.20431106, 340.54425259, 341.74758917, 344.15426233, 340.54425259, 339.34091601, 361.00097448, 348.96760866, 357.39096473, 359.79763789, 341.74758917, 328.51088677, 300.83414540, 294.81746249, 312.86751121, 317.68085754, 310.46083805, 297.22413565, 300.83414540, 291.20745274, 263.53071137, 290.00411616, 342.95092575, 338.13757943, 328.51088677, 323.69754045, 323.69754045, 305.64749172, 303.24081856, 312.86751121, 327.30755019, 324.90087703, 316.47752096, 309.25750147, 311.66417463, 320.08753070, 324.90087703, 338.13757943, 333.32423310, 324.90087703, 318.88419412, 310.46083805, 315.27418438, 314.07084779, 330.91755994, 328.51088677, 330.91755994, 338.13757943, 336.93424284, 323.69754045, 347.76427208, 361.00097448, 362.20431106, 421.16780356, 407.93110116, 399.50774509, 385.06770611, 433.20116937, 425.98114988, 410.33777432, 424.77781330, 431.99783279, 441.62452544, 441.62452544, 410.33777432, 387.47437927, 391.08438902, 400.71108167, 413.94778407, 393.49106218, 407.93110116, 410.33777432, 438.01451570, 412.74444748, 417.55779381, 405.52442800, 379.05102320, 385.06770611, 373.03434029, 357.39096473, 379.05102320, 379.05102320, 385.06770611, 377.84768662, 342.95092575, 348.96760866, 342.95092575, 312.86751121, 314.07084779, 315.27418438, 318.88419412, 310.46083805, 317.68085754, 333.32423310, 350.17094524, 346.56093550, 342.95092575, 353.78095499, 342.95092575, 339.34091601, 338.13757943, 323.69754045, 336.93424284, 340.54425259, 332.12089652, 312.86751121, 342.95092575, 345.35759892, 350.17094524, 342.95092575, 347.76427208, 348.96760866, 354.98429157, 362.20431106, 357.39096473, 363.40764764, 359.79763789, 361.00097448, 346.56093550, 352.57761841, 359.79763789, 351.37428182, 357.39096473, 379.05102320, 385.06770611, 406.72776458, 407.93110116, 415.15112065, 391.08438902, 393.49106218, 416.35445723, 421.16780356, 412.74444748, 428.38782305, 421.16780356, 434.40450595, 430.79449621, 413.94778407, 401.91441825, 391.08438902, 376.64435004, 375.44101346, 377.84768662, 385.06770611, 388.67771585, 382.66103294, 376.64435004, 348.96760866, 377.84768662, 364.61098422, 352.57761841, 359.79763789, 342.95092575, 342.95092575, 338.13757943, 318.88419412, 294.81746249, 320.08753070, 314.07084779, 321.29086728, 344.15426233, 332.12089652, 340.54425259, 328.51088677, 324.90087703, 304.44415514, 309.25750147, 281.58076009, 285.19076984, 288.80077958, 265.93738453, 300.83414540, 299.63080882, 297.22413565, 303.24081856, 318.88419412, 324.90087703, 338.13757943, 316.47752096, 326.10421361, 327.30755019, 352.57761841, 333.32423310, 346.56093550, 347.76427208, 351.37428182, 345.35759892, 335.73090626, 346.56093550, 351.37428182, 356.18762815, 363.40764764, 371.83100371, 373.03434029, 365.81432080, 370.62766713, 376.64435004, 409.13443774, 400.71108167, 393.49106218, 397.10107192, 415.15112065, 400.71108167, 392.28772560, 367.01765738, 357.39096473, 362.20431106, 369.42433055, 382.66103294, 403.11775483, 395.89773534, 413.94778407, 399.50774509, 410.33777432, 392.28772560, 389.88105243, 398.30440851, 395.89773534, 400.71108167, 383.86436953, 388.67771585, 404.32109141, 404.32109141, 395.89773534, 371.83100371, 369.42433055, 374.23767687, 389.88105243, 369.42433055, 381.45769636, 392.28772560, 369.42433055, 367.01765738, 370.62766713, 371.83100371, 380.25435978, 371.83100371, 365.81432080, 385.06770611, 370.62766713, 374.23767687, 379.05102320, 369.42433055, 373.03434029, 364.61098422, 367.01765738, 365.81432080, 351.37428182, 339.34091601, 342.95092575, 296.02079907, 333.32423310, 308.05416489, 317.68085754, 315.27418438, 347.76427208, 351.37428182, 346.56093550, 338.13757943, 347.76427208, 339.34091601, 344.15426233, 357.39096473, 352.57761841, 358.59430131, 373.03434029, 387.47437927, 376.64435004, 367.01765738, 374.23767687, 357.39096473, 371.83100371, 363.40764764, 375.44101346, 380.25435978, 387.47437927, 367.01765738, 361.00097448, 363.40764764, 335.73090626, 336.93424284, 341.74758917, 339.34091601, 336.93424284, 342.95092575, 317.68085754, 309.25750147], 
							[814.65886573, 809.84551941, 824.28555839, 777.35543171, 767.72873905, 776.15209512, 691.91853441, 607.68497370, 580.00823233, 563.16152018, 598.05828105, 601.66829079, 607.68497370, 622.12501268, 587.22825181, 563.16152018, 505.40136427, 475.31794973, 523.45141299, 528.26475932, 523.45141299, 535.48477881, 560.75484702, 540.29812513, 533.07810564, 553.53482753, 598.05828105, 649.80175406, 689.51186125, 709.96858314, 740.05199768, 664.24179304, 605.27830054, 647.39508089, 637.76838824, 624.53168584, 637.76838824, 614.90499319, 673.86848569, 699.13855390, 678.68183202, 661.83511987, 679.88516860, 654.61510038, 619.71833952, 608.88831028, 604.07496396, 594.44827130, 616.10832977, 659.42844671, 625.73502243, 655.81843697, 671.46181253, 689.51186125, 740.05199768, 693.12187099, 697.93521732, 661.83511987, 707.56190997, 715.98526604, 667.85180278, 697.93521732, 677.47849543, 669.05513936, 665.44512962, 715.98526604, 722.00194895, 693.12187099, 663.03845645, 647.39508089, 647.39508089, 684.69851492, 681.08850518, 653.41176380, 675.07182227, 675.07182227, 670.25847594, 682.29184176, 694.32520758, 705.15523681, 691.91853441, 665.44512962, 685.90185151, 678.68183202, 661.83511987, 614.90499319, 635.36171508, 583.61824207, 482.53796922, 488.55465212, 470.50460340, 409.13443774, 453.65789126, 441.62452544, 471.70793998, 429.59115963, 451.25121810, 474.11461315, 506.60470085, 517.43473008, 521.04473983, 628.14169559, 618.51500294, 555.94150069, 599.26161763, 592.04159814, 592.04159814, 649.80175406, 669.05513936, 625.73502243, 654.61510038, 673.86848569, 638.97172482, 661.83511987, 651.00509064, 661.83511987, 671.46181253, 693.12187099, 667.85180278, 669.05513936, 685.90185151, 640.17506140, 632.95504192, 619.71833952, 642.58173457, 652.20842722, 647.39508089, 646.19174431, 652.20842722, 632.95504192, 649.80175406, 653.41176380, 614.90499319, 607.68497370, 601.66829079, 565.56819335, 598.05828105, 601.66829079, 618.51500294, 623.32834926, 628.14169559, 619.71833952, 636.56505166, 584.82157865, 587.22825181, 600.46495421, 612.49832003, 616.10832977, 600.46495421, 607.68497370, 612.49832003, 607.68497370, 577.60155916, 594.44827130, 587.22825181, 596.85494447, 577.60155916, 517.43473008, 480.13129605, 509.01137401, 509.01137401, 482.53796922, 463.28458391, 486.14797896, 558.34817386, 529.46809590, 553.53482753, 511.41804717, 482.53796922, 469.30126682, 484.94464238, 456.06456442, 417.55779381, 430.79449621, 386.27104269, 487.35131554, 487.35131554, 489.75798871, 515.02805692, 490.96132529, 493.36799845, 524.65474957, 517.43473008, 541.50146171, 552.33149095, 569.17820309, 518.63806666, 509.01137401, 517.43473008, 506.60470085, 527.06142274, 487.35131554, 493.36799845, 476.52128631, 516.23139350, 493.36799845, 527.06142274, 551.12815437, 561.95818360, 564.36485676, 554.73816411, 560.75484702, 575.19488600, 595.65160789, 619.71833952, 605.27830054, 601.66829079, 584.82157865, 588.43158840, 571.58487625, 589.63492498, 613.70165661, 611.29498345, 661.83511987, 652.20842722, 637.76838824, 644.98840773, 642.58173457, 629.34503217, 635.36171508, 654.61510038, 649.80175406, 693.12187099, 684.69851492, 705.15523681, 709.96858314, 699.13855390, 683.49517834, 667.85180278, 676.27515885, 649.80175406, 644.98840773, 612.49832003, 631.75170533, 667.85180278, 608.88831028, 605.27830054, 651.00509064, 690.71519783, 669.05513936, 659.42844671, 649.80175406, 689.51186125, 666.64846620, 660.63178329, 693.12187099, 675.07182227, 702.74856365, 670.25847594, 602.87162738, 661.83511987, 629.34503217, 589.63492498, 559.55151044, 457.26790100, 484.94464238, 480.13129605, 507.80803743, 472.91127656, 484.94464238, 506.60470085, 507.80803743, 484.94464238, 469.30126682, 513.82472034, 463.28458391, 421.16780356, 423.57447672, 442.82786202, 430.79449621, 463.28458391, 486.14797896, 527.06142274, 505.40136427, 511.41804717, 536.68811539, 527.06142274, 527.06142274, 522.24807641, 482.53796922, 475.31794973, 484.94464238, 470.50460340, 403.11775483, 389.88105243, 357.39096473, 340.54425259, 338.13757943, 340.54425259, 320.08753070, 308.05416489], 
							[1046.90282598, 1009.59939195, 987.93933348, 971.09262134, 904.90910935, 851.96229976, 782.16877803, 755.69537324, 730.42530502, 737.64532451, 770.13541222, 770.13541222, 826.69223155, 807.43884625, 767.72873905, 759.30538298, 696.73188074, 675.07182227, 709.96858314, 707.56190997, 703.95190023, 642.58173457, 646.19174431, 625.73502243, 616.10832977, 638.97172482, 717.18860263, 735.23865135, 786.98212436, 744.86534400, 776.15209512, 735.23865135, 735.23865135, 754.49203666, 766.52540247, 712.37525630, 742.45867084, 722.00194895, 758.10204640, 762.91539273, 764.11872931, 748.47535375, 770.13541222, 746.06868058, 708.76524656, 708.76524656, 708.76524656, 688.30852467, 728.01863186, 761.71205615, 723.20528553, 758.10204640, 748.47535375, 773.74542196, 821.87888522, 847.14895344, 839.92893395, 800.21882676, 812.25219257, 829.09890471, 883.24905089, 892.87574354, 877.23236798, 866.40233874, 855.57230951, 873.62235823, 850.75896318, 831.50557788, 821.87888522, 833.91225104, 808.64218283, 813.45552915, 812.25219257, 797.81215359, 807.43884625, 839.92893395, 832.70891446, 842.33560711, 854.36897293, 843.53894369, 859.18231925, 818.26887548, 830.30224130, 832.70891446, 792.99880727, 824.28555839, 817.06553890, 736.44198793, 649.80175406, 647.39508089, 649.80175406, 565.56819335, 629.34503217, 624.53168584, 667.85180278, 618.51500294, 626.93835901, 667.85180278, 718.39193921, 708.76524656, 657.02177355, 774.94875854, 756.89870982, 679.88516860, 696.73188074, 734.03531477, 770.13541222, 770.13541222, 736.44198793, 701.54522707, 736.44198793, 738.84866109, 711.17191972, 722.00194895, 724.40862212, 737.64532451, 724.40862212, 752.08536349, 752.08536349, 774.94875854, 766.52540247, 735.23865135, 719.59527579, 703.95190023, 725.61195870, 718.39193921, 736.44198793, 741.25533426, 750.88202691, 707.56190997, 730.42530502, 752.08536349, 691.91853441, 715.98526604, 685.90185151, 661.83511987, 695.52854416, 709.96858314, 715.98526604, 705.15523681, 705.15523681, 705.15523681, 670.25847594, 673.86848569, 661.83511987, 664.24179304, 684.69851492, 723.20528553, 715.98526604, 711.17191972, 720.79861237, 722.00194895, 660.63178329, 695.52854416, 719.59527579, 734.03531477, 715.98526604, 667.85180278, 620.92167610, 658.22511013, 635.36171508, 607.68497370, 569.17820309, 602.87162738, 669.05513936, 677.47849543, 694.32520758, 659.42844671, 623.32834926, 610.09164687, 607.68497370, 565.56819335, 527.06142274, 572.78821284, 487.35131554, 600.46495421, 651.00509064, 679.88516860, 672.66514911, 649.80175406, 601.66829079, 654.61510038, 631.75170533, 670.25847594, 663.03845645, 694.32520758, 657.02177355, 638.97172482, 681.08850518, 638.97172482, 643.78507115, 607.68497370, 631.75170533, 624.53168584, 651.00509064, 642.58173457, 683.49517834, 685.90185151, 679.88516860, 691.91853441, 693.12187099, 712.37525630, 722.00194895, 732.83197819, 794.20214385, 791.79547068, 764.11872931, 759.30538298, 735.23865135, 715.98526604, 747.27201717, 752.08536349, 776.15209512, 783.37211461, 818.26887548, 777.35543171, 782.16877803, 792.99880727, 754.49203666, 772.54208538, 817.06553890, 835.11558762, 809.84551941, 814.65886573, 830.30224130, 848.35229002, 838.72559737, 829.09890471, 815.86220232, 824.28555839, 799.01549017, 799.01549017, 785.77878778, 802.62549992, 812.25219257, 766.52540247, 761.71205615, 794.20214385, 839.92893395, 823.08222181, 821.87888522, 815.86220232, 835.11558762, 795.40548043, 789.38879752, 806.23550966, 800.21882676, 805.03217308, 799.01549017, 786.98212436, 788.18546094, 711.17191972, 659.42844671, 628.14169559, 542.70479830, 552.33149095, 560.75484702, 559.55151044, 528.26475932, 531.87476906, 543.90813488, 553.53482753, 523.45141299, 496.97800820, 530.67143248, 482.53796922, 422.37114014, 417.55779381, 456.06456442, 439.21785228, 474.11461315, 507.80803743, 558.34817386, 553.53482753, 573.99154942, 598.05828105, 595.65160789, 593.24493472, 584.82157865, 543.90813488, 524.65474957, 572.78821284, 537.89145197, 481.33463264, 462.08124733, 409.13443774, 394.69439876, 387.47437927, 400.71108167, 381.45769636, 373.03434029], 
							[892.87574354, 950.63589945, 956.65258236, 974.70263109, 932.58585073, 807.43884625, 742.45867084, 702.74856365, 682.29184176, 669.05513936, 766.52540247, 821.87888522, 833.91225104, 866.40233874, 823.08222181, 819.47221206, 746.06868058, 705.15523681, 774.94875854, 771.33874880, 747.27201717, 669.05513936, 679.88516860, 652.20842722, 554.73816411, 637.76838824, 800.21882676, 812.25219257, 830.30224130, 820.67554864, 814.65886573, 782.16877803, 789.38879752, 783.37211461, 758.10204640, 752.08536349, 752.08536349, 689.51186125, 740.05199768, 746.06868058, 752.08536349, 723.20528553, 738.84866109, 722.00194895, 701.54522707, 677.47849543, 673.86848569, 675.07182227, 689.51186125, 720.79861237, 697.93521732, 712.37525630, 675.07182227, 691.91853441, 722.00194895, 719.59527579, 709.96858314, 722.00194895, 746.06868058, 778.55876829, 791.79547068, 807.43884625, 800.21882676, 779.76210487, 776.15209512, 794.20214385, 780.96544145, 758.10204640, 758.10204640, 722.00194895, 728.01863186, 743.66200742, 731.62864161, 722.00194895, 747.27201717, 744.86534400, 755.69537324, 765.32206589, 788.18546094, 790.59213410, 782.16877803, 784.57545120, 779.76210487, 755.69537324, 723.20528553, 691.91853441, 679.88516860, 661.83511987, 601.66829079, 596.85494447, 577.60155916, 531.87476906, 577.60155916, 565.56819335, 606.48163712, 548.72148120, 565.56819335, 600.46495421, 637.76838824, 637.76838824, 619.71833952, 672.66514911, 631.75170533, 605.27830054, 638.97172482, 661.83511987, 722.00194895, 717.18860263, 711.17191972, 667.85180278, 702.74856365, 709.96858314, 685.90185151, 669.05513936, 669.05513936, 683.49517834, 685.90185151, 690.71519783, 696.73188074, 720.79861237, 701.54522707, 676.27515885, 670.25847594, 658.22511013, 691.91853441, 684.69851492, 694.32520758, 691.91853441, 685.90185151, 659.42844671, 669.05513936, 672.66514911, 630.54836875, 642.58173457, 636.56505166, 610.09164687, 626.93835901, 625.73502243, 613.70165661, 613.70165661, 616.10832977, 618.51500294, 617.31166635, 611.29498345, 613.70165661, 616.10832977, 624.53168584, 631.75170533, 625.73502243, 629.34503217, 626.93835901, 632.95504192, 576.39822258, 611.29498345, 624.53168584, 625.73502243, 605.27830054, 589.63492498, 519.84140325, 530.67143248, 546.31480804, 517.43473008, 513.82472034, 545.11147146, 576.39822258, 567.97486651, 587.22825181, 577.60155916, 571.58487625, 551.12815437, 554.73816411, 502.99469110, 475.31794973, 481.33463264, 445.23453519, 540.29812513, 536.68811539, 551.12815437, 553.53482753, 529.46809590, 539.09478855, 573.99154942, 571.58487625, 577.60155916, 583.61824207, 588.43158840, 572.78821284, 565.56819335, 581.21156891, 577.60155916, 584.82157865, 564.36485676, 581.21156891, 564.36485676, 586.02491523, 577.60155916, 590.83826156, 600.46495421, 607.68497370, 649.80175406, 637.76838824, 642.58173457, 671.46181253, 689.51186125, 754.49203666, 740.05199768, 728.01863186, 723.20528553, 702.74856365, 707.56190997, 714.78192946, 714.78192946, 758.10204640, 779.76210487, 792.99880727, 789.38879752, 792.99880727, 779.76210487, 771.33874880, 762.91539273, 774.94875854, 782.16877803, 782.16877803, 770.13541222, 771.33874880, 786.98212436, 777.35543171, 770.13541222, 748.47535375, 758.10204640, 726.81529528, 696.73188074, 691.91853441, 720.79861237, 749.67869033, 684.69851492, 687.10518809, 741.25533426, 760.50871956, 753.28870007, 725.61195870, 730.42530502, 764.11872931, 706.35857339, 707.56190997, 752.08536349, 728.01863186, 744.86534400, 713.57859288, 681.08850518, 677.47849543, 641.37839799, 580.00823233, 602.87162738, 493.36799845, 517.43473008, 513.82472034, 542.70479830, 523.45141299, 518.63806666, 519.84140325, 523.45141299, 510.21471059, 515.02805692, 552.33149095, 502.99469110, 465.69125707, 469.30126682, 521.04473983, 496.97800820, 535.48477881, 527.06142274, 565.56819335, 547.51814462, 565.56819335, 583.61824207, 601.66829079, 587.22825181, 565.56819335, 524.65474957, 523.45141299, 527.06142274, 509.01137401, 472.91127656, 465.69125707, 427.18448646, 410.33777432, 406.72776458, 413.94778407, 398.30440851, 400.71108167], 
							[417.55779381, 403.11775483, 367.01765738, 370.62766713, 374.23767687, 357.39096473, 342.95092575, 354.98429157, 358.59430131, 348.96760866, 367.01765738, 371.83100371, 359.79763789, 358.59430131, 358.59430131, 351.37428182, 312.86751121, 328.51088677, 336.93424284, 336.93424284, 332.12089652, 312.86751121, 324.90087703, 312.86751121, 280.37742351, 297.22413565, 365.81432080, 357.39096473, 339.34091601, 339.34091601, 351.37428182, 338.13757943, 347.76427208, 354.98429157, 354.98429157, 348.96760866, 340.54425259, 335.73090626, 339.34091601, 346.56093550, 346.56093550, 353.78095499, 357.39096473, 345.35759892, 334.52756968, 338.13757943, 338.13757943, 334.52756968, 353.78095499, 336.93424284, 336.93424284, 347.76427208, 346.56093550, 340.54425259, 374.23767687, 379.05102320, 385.06770611, 459.67457417, 448.84454493, 427.18448646, 424.77781330, 453.65789126, 454.86122784, 440.42118886, 463.28458391, 464.48792049, 458.47123758, 453.65789126, 448.84454493, 422.37114014, 428.38782305, 412.74444748, 418.76113039, 405.52442800, 419.96446697, 433.20116937, 454.86122784, 439.21785228, 450.04788151, 431.99783279, 395.89773534, 392.28772560, 369.42433055, 382.66103294, 405.52442800, 403.11775483, 407.93110116, 405.52442800, 377.84768662, 388.67771585, 385.06770611, 370.62766713, 376.64435004, 381.45769636, 375.44101346, 371.83100371, 376.64435004, 376.64435004, 388.67771585, 388.67771585, 409.13443774, 423.57447672, 422.37114014, 403.11775483, 430.79449621, 427.18448646, 427.18448646, 451.25121810, 431.99783279, 421.16780356, 442.82786202, 451.25121810, 439.21785228, 446.43787177, 445.23453519, 458.47123758, 477.72462289, 490.96132529, 490.96132529, 504.19802769, 488.55465212, 492.16466187, 494.57133503, 489.75798871, 486.14797896, 478.92795947, 481.33463264, 489.75798871, 498.18134478, 489.75798871, 502.99469110, 517.43473008, 515.02805692, 512.62138376, 527.06142274, 539.09478855, 522.24807641, 539.09478855, 539.09478855, 534.28144222, 542.70479830, 512.62138376, 504.19802769, 500.58801794, 488.55465212, 478.92795947, 481.33463264, 490.96132529, 501.79135452, 493.36799845, 472.91127656, 439.21785228, 459.67457417, 451.25121810, 444.03119861, 439.21785228, 412.74444748, 421.16780356, 403.11775483, 409.13443774, 368.22099397, 410.33777432, 391.08438902, 400.71108167, 424.77781330, 409.13443774, 421.16780356, 391.08438902, 395.89773534, 389.88105243, 385.06770611, 352.57761841, 338.13757943, 346.56093550, 334.52756968, 357.39096473, 379.05102320, 381.45769636, 373.03434029, 370.62766713, 386.27104269, 410.33777432, 391.08438902, 398.30440851, 406.72776458, 439.21785228, 419.96446697, 428.38782305, 431.99783279, 431.99783279, 430.79449621, 403.11775483, 418.76113039, 411.54111090, 423.57447672, 436.81117912, 434.40450595, 442.82786202, 435.60784253, 429.59115963, 433.20116937, 444.03119861, 435.60784253, 425.98114988, 430.79449621, 462.08124733, 442.82786202, 431.99783279, 417.55779381, 418.76113039, 427.18448646, 435.60784253, 440.42118886, 460.87791075, 458.47123758, 469.30126682, 464.48792049, 472.91127656, 453.65789126, 463.28458391, 470.50460340, 478.92795947, 481.33463264, 451.25121810, 465.69125707, 475.31794973, 475.31794973, 468.09793024, 458.47123758, 446.43787177, 448.84454493, 458.47123758, 439.21785228, 447.64120835, 459.67457417, 439.21785228, 435.60784253, 436.81117912, 438.01451570, 445.23453519, 446.43787177, 438.01451570, 460.87791075, 427.18448646, 411.54111090, 421.16780356, 423.57447672, 433.20116937, 429.59115963, 415.15112065, 410.33777432, 395.89773534, 381.45769636, 400.71108167, 344.15426233, 369.42433055, 342.95092575, 358.59430131, 352.57761841, 400.71108167, 409.13443774, 388.67771585, 404.32109141, 406.72776458, 418.76113039, 417.55779381, 433.20116937, 429.59115963, 446.43787177, 445.23453519, 459.67457417, 466.89459366, 444.03119861, 428.38782305, 411.54111090, 423.57447672, 416.35445723, 418.76113039, 441.62452544, 427.18448646, 425.98114988, 409.13443774, 416.35445723, 397.10107192, 369.42433055, 373.03434029, 361.00097448, 359.79763789, 334.52756968, 306.85082830, 283.98743325], 
							[974.70263109, 926.56916782, 878.43570456, 876.02903140, 782.16877803, 663.03845645, 703.95190023, 703.95190023, 708.76524656, 702.74856365, 746.06868058, 712.37525630, 715.98526604, 730.42530502, 709.96858314, 697.93521732, 672.66514911, 657.02177355, 693.12187099, 664.24179304, 666.64846620, 611.29498345, 640.17506140, 605.27830054, 553.53482753, 655.81843697, 738.84866109, 744.86534400, 829.09890471, 796.60881701, 838.72559737, 800.21882676, 764.11872931, 774.94875854, 712.37525630, 673.86848569, 708.76524656, 665.44512962, 722.00194895, 726.81529528, 720.79861237, 722.00194895, 728.01863186, 729.22196844, 691.91853441, 700.34189048, 677.47849543, 670.25847594, 720.79861237, 717.18860263, 693.12187099, 693.12187099, 695.52854416, 695.52854416, 812.25219257, 838.72559737, 992.75267981, 1202.13324501, 1141.96641593, 1121.50969404, 1064.95287471, 1113.08633797, 1176.86317679, 1121.50969404, 1119.10302088, 1155.20311832, 1137.15306960, 1114.28967455, 1058.93619180, 1032.46278700, 1036.07279675, 1024.03943093, 1101.05297215, 1076.98624052, 1095.03628924, 1089.01960634, 1131.13638669, 1101.05297215, 1109.47632822, 1119.10302088, 1109.47632822, 1064.95287471, 1025.24276751, 1027.64944068, 1061.34286496, 1048.10616256, 1045.69948940, 939.80587022, 847.14895344, 849.55562660, 821.87888522, 748.47535375, 807.43884625, 827.89556813, 830.30224130, 783.37211461, 817.06553890, 817.06553890, 847.14895344, 903.70577277, 855.57230951, 961.46592869, 962.66926527, 878.43570456, 930.17917757, 939.80587022, 939.80587022, 974.70263109, 907.31578252, 902.50243619, 914.53580201, 932.58585073, 925.36583124, 910.92579226, 910.92579226, 947.02588971, 950.63589945, 954.24590920, 945.82255313, 942.21254338, 930.17917757, 890.46907037, 902.50243619, 870.01234849, 872.41902165, 872.41902165, 877.23236798, 889.26573379, 878.43570456, 892.87574354, 890.46907037, 888.06239721, 863.99566558, 831.50557788, 825.48889497, 812.25219257, 807.43884625, 813.45552915, 802.62549992, 801.42216334, 792.99880727, 764.11872931, 785.77878778, 768.93207563, 772.54208538, 764.11872931, 758.10204640, 764.11872931, 794.20214385, 789.38879752, 778.55876829, 759.30538298, 732.83197819, 788.18546094, 779.76210487, 752.08536349, 755.69537324, 734.03531477, 711.17191972, 703.95190023, 672.66514911, 675.07182227, 664.24179304, 673.86848569, 703.95190023, 713.57859288, 723.20528553, 675.07182227, 673.86848569, 671.46181253, 667.85180278, 647.39508089, 631.75170533, 653.41176380, 654.61510038, 699.13855390, 715.98526604, 738.84866109, 732.83197819, 754.49203666, 736.44198793, 811.04885599, 824.28555839, 827.89556813, 820.67554864, 848.35229002, 797.81215359, 803.82883650, 812.25219257, 768.93207563, 795.40548043, 758.10204640, 812.25219257, 792.99880727, 817.06553890, 838.72559737, 860.38565584, 836.31892420, 842.33560711, 854.36897293, 877.23236798, 943.41587996, 863.99566558, 878.43570456, 883.24905089, 898.89242645, 916.94247517, 902.50243619, 890.46907037, 907.31578252, 909.72245568, 907.31578252, 948.22922629, 980.71931399, 996.36268955, 1002.37937246, 1015.61607486, 1026.44610409, 980.71931399, 954.24590920, 974.70263109, 986.73599690, 973.49929450, 968.68594818, 1019.22608460, 1009.59939195, 1004.78604563, 984.32932374, 944.61921655, 936.19586048, 890.46907037, 895.28241670, 865.19900216, 879.63904114, 918.14581175, 859.18231925, 882.04571430, 925.36583124, 928.97584099, 928.97584099, 944.61921655, 928.97584099, 963.87260185, 927.77250440, 927.77250440, 943.41587996, 921.75582150, 932.58585073, 914.53580201, 854.36897293, 850.75896318, 824.28555839, 786.98212436, 791.79547068, 643.78507115, 687.10518809, 671.46181253, 670.25847594, 653.41176380, 646.19174431, 707.56190997, 679.88516860, 661.83511987, 661.83511987, 700.34189048, 732.83197819, 713.57859288, 722.00194895, 754.49203666, 782.16877803, 796.60881701, 772.54208538, 838.72559737, 848.35229002, 801.42216334, 856.77564609, 854.36897293, 851.96229976, 837.52226079, 814.65886573, 777.35543171, 794.20214385, 837.52226079, 758.10204640, 770.13541222, 753.28870007, 753.28870007, 764.11872931, 747.27201717, 706.35857339, 675.07182227], 
							[934.99252389, 920.55248491, 896.48575328, 861.58899242, 866.40233874, 758.10204640, 782.16877803, 812.25219257, 750.88202691, 736.44198793, 809.84551941, 833.91225104, 832.70891446, 835.11558762, 801.42216334, 767.72873905, 723.20528553, 685.90185151, 718.39193921, 685.90185151, 689.51186125, 626.93835901, 632.95504192, 598.05828105, 539.09478855, 685.90185151, 780.96544145, 800.21882676, 758.10204640, 756.89870982, 761.71205615, 736.44198793, 717.18860263, 750.88202691, 722.00194895, 700.34189048, 693.12187099, 661.83511987, 691.91853441, 707.56190997, 676.27515885, 665.44512962, 689.51186125, 666.64846620, 651.00509064, 625.73502243, 646.19174431, 642.58173457, 638.97172482, 648.59841748, 626.93835901, 635.36171508, 624.53168584, 613.70165661, 670.25847594, 701.54522707, 709.96858314, 742.45867084, 776.15209512, 792.99880727, 742.45867084, 789.38879752, 818.26887548, 796.60881701, 823.08222181, 818.26887548, 802.62549992, 786.98212436, 756.89870982, 734.03531477, 725.61195870, 734.03531477, 766.52540247, 717.18860263, 764.11872931, 756.89870982, 770.13541222, 779.76210487, 776.15209512, 786.98212436, 786.98212436, 772.54208538, 788.18546094, 773.74542196, 753.28870007, 778.55876829, 774.94875854, 753.28870007, 722.00194895, 722.00194895, 694.32520758, 617.31166635, 666.64846620, 649.80175406, 660.63178329, 667.85180278, 653.41176380, 655.81843697, 672.66514911, 717.18860263, 702.74856365, 779.76210487, 762.91539273, 703.95190023, 740.05199768, 744.86534400, 731.62864161, 728.01863186, 746.06868058, 697.93521732, 711.17191972, 722.00194895, 730.42530502, 740.05199768, 742.45867084, 746.06868058, 747.27201717, 749.67869033, 758.10204640, 762.91539273, 788.18546094, 758.10204640, 748.47535375, 746.06868058, 768.93207563, 755.69537324, 724.40862212, 713.57859288, 728.01863186, 735.23865135, 741.25533426, 782.16877803, 728.01863186, 726.81529528, 728.01863186, 730.42530502, 717.18860263, 726.81529528, 719.59527579, 702.74856365, 685.90185151, 648.59841748, 654.61510038, 634.15837850, 617.31166635, 604.07496396, 614.90499319, 634.15837850, 636.56505166, 607.68497370, 619.71833952, 626.93835901, 681.08850518, 675.07182227, 683.49517834, 670.25847594, 649.80175406, 606.48163712, 583.61824207, 553.53482753, 539.09478855, 557.14483728, 539.09478855, 548.72148120, 571.58487625, 577.60155916, 587.22825181, 569.17820309, 560.75484702, 523.45141299, 528.26475932, 496.97800820, 509.01137401, 534.28144222, 480.13129605, 565.56819335, 554.73816411, 564.36485676, 534.28144222, 552.33149095, 573.99154942, 589.63492498, 592.04159814, 599.26161763, 586.02491523, 613.70165661, 559.55151044, 581.21156891, 588.43158840, 571.58487625, 565.56819335, 511.41804717, 547.51814462, 543.90813488, 551.12815437, 541.50146171, 559.55151044, 566.77152993, 558.34817386, 560.75484702, 560.75484702, 611.29498345, 592.04159814, 606.48163712, 611.29498345, 592.04159814, 587.22825181, 593.24493472, 572.78821284, 565.56819335, 578.80489574, 555.94150069, 567.97486651, 612.49832003, 610.09164687, 602.87162738, 634.15837850, 636.56505166, 620.92167610, 595.65160789, 624.53168584, 624.53168584, 683.49517834, 637.76838824, 628.14169559, 620.92167610, 613.70165661, 614.90499319, 575.19488600, 555.94150069, 554.73816411, 539.09478855, 517.43473008, 518.63806666, 522.24807641, 518.63806666, 527.06142274, 522.24807641, 535.48477881, 547.51814462, 528.26475932, 530.67143248, 537.89145197, 507.80803743, 506.60470085, 509.01137401, 499.38468136, 490.96132529, 433.20116937, 422.37114014, 431.99783279, 424.77781330, 395.89773534, 400.71108167, 324.90087703, 354.98429157, 264.73404795, 262.32737479, 232.24396025, 225.02394076, 275.56407718, 258.71736504, 264.73404795, 256.31069188, 259.92070162, 252.70068213, 220.21059443, 212.99057494, 228.63395050, 223.82060418, 227.43061392, 223.82060418, 234.65063341, 244.27732606, 249.09067239, 247.88733581, 246.68399923, 246.68399923, 239.46397974, 221.41393101, 204.56721887, 176.89047749, 186.51717015, 146.80706295, 126.35034107, 108.30029234, 121.53699474, 123.94366790, 123.94366790, 123.94366790, 104.69028260], 
							[620.92167610, 595.65160789, 587.22825181, 592.04159814, 566.77152993, 505.40136427, 505.40136427, 513.82472034, 511.41804717, 509.01137401, 555.94150069, 540.29812513, 527.06142274, 541.50146171, 511.41804717, 507.80803743, 492.16466187, 453.65789126, 499.38468136, 511.41804717, 496.97800820, 482.53796922, 484.94464238, 458.47123758, 429.59115963, 466.89459366, 564.36485676, 542.70479830, 533.07810564, 551.12815437, 563.16152018, 515.02805692, 546.31480804, 521.04473983, 533.07810564, 528.26475932, 530.67143248, 488.55465212, 505.40136427, 513.82472034, 527.06142274, 522.24807641, 522.24807641, 502.99469110, 493.36799845, 501.79135452, 496.97800820, 484.94464238, 507.80803743, 496.97800820, 484.94464238, 481.33463264, 472.91127656, 439.21785228, 501.79135452, 498.18134478, 521.04473983, 595.65160789, 577.60155916, 588.43158840, 567.97486651, 634.15837850, 629.34503217, 634.15837850, 619.71833952, 616.10832977, 624.53168584, 606.48163712, 577.60155916, 584.82157865, 577.60155916, 590.83826156, 593.24493472, 572.78821284, 601.66829079, 588.43158840, 605.27830054, 588.43158840, 594.44827130, 600.46495421, 576.39822258, 588.43158840, 583.61824207, 555.94150069, 564.36485676, 554.73816411, 571.58487625, 559.55151044, 548.72148120, 519.84140325, 481.33463264, 469.30126682, 480.13129605, 465.69125707, 494.57133503, 487.35131554, 477.72462289, 492.16466187, 559.55151044, 569.17820309, 545.11147146, 604.07496396, 601.66829079, 564.36485676, 602.87162738, 594.44827130, 601.66829079, 619.71833952, 623.32834926, 611.29498345, 607.68497370, 606.48163712, 600.46495421, 611.29498345, 606.48163712, 611.29498345, 637.76838824, 652.20842722, 669.05513936, 681.08850518, 688.30852467, 678.68183202, 666.64846620, 638.97172482, 654.61510038, 666.64846620, 664.24179304, 661.83511987, 664.24179304, 677.47849543, 669.05513936, 682.29184176, 659.42844671, 653.41176380, 673.86848569, 652.20842722, 643.78507115, 664.24179304, 652.20842722, 653.41176380, 653.41176380, 642.58173457, 643.78507115, 630.54836875, 640.17506140, 632.95504192, 612.49832003, 628.14169559, 636.56505166, 629.34503217, 606.48163712, 588.43158840, 572.78821284, 565.56819335, 546.31480804, 535.48477881, 519.84140325, 533.07810564, 507.80803743, 493.36799845, 462.08124733, 505.40136427, 506.60470085, 502.99469110, 511.41804717, 505.40136427, 519.84140325, 498.18134478, 507.80803743, 505.40136427, 481.33463264, 488.55465212, 452.45455468, 472.91127656, 457.26790100, 489.75798871, 541.50146171, 529.46809590, 529.46809590, 542.70479830, 530.67143248, 589.63492498, 572.78821284, 582.41490549, 559.55151044, 583.61824207, 535.48477881, 586.02491523, 572.78821284, 570.38153967, 583.61824207, 548.72148120, 576.39822258, 575.19488600, 589.63492498, 601.66829079, 632.95504192, 648.59841748, 636.56505166, 628.14169559, 643.78507115, 652.20842722, 643.78507115, 649.80175406, 700.34189048, 711.17191972, 682.29184176, 673.86848569, 663.03845645, 672.66514911, 655.81843697, 647.39508089, 685.90185151, 691.91853441, 741.25533426, 772.54208538, 738.84866109, 743.66200742, 744.86534400, 736.44198793, 753.28870007, 723.20528553, 758.10204640, 723.20528553, 746.06868058, 723.20528553, 717.18860263, 718.39193921, 672.66514911, 685.90185151, 667.85180278, 655.81843697, 628.14169559, 648.59841748, 663.03845645, 631.75170533, 601.66829079, 617.31166635, 636.56505166, 659.42844671, 614.90499319, 610.09164687, 614.90499319, 610.09164687, 602.87162738, 606.48163712, 606.48163712, 605.27830054, 613.70165661, 595.65160789, 595.65160789, 561.95818360, 542.70479830, 563.16152018, 499.38468136, 528.26475932, 511.41804717, 536.68811539, 542.70479830, 575.19488600, 577.60155916, 539.09478855, 505.40136427, 510.21471059, 516.23139350, 517.43473008, 502.99469110, 487.35131554, 517.43473008, 511.41804717, 577.60155916, 553.53482753, 571.58487625, 557.14483728, 542.70479830, 570.38153967, 549.92481779, 542.70479830, 530.67143248, 527.06142274, 499.38468136, 487.35131554, 487.35131554, 471.70793998, 466.89459366, 475.31794973, 484.94464238, 460.87791075, 468.09793024, 478.92795947, 476.52128631], 
							[1468.07062954, 1468.07062954, 1444.00389791, 1383.83706883, 1347.73697138, 1323.67023975, 1251.47004485, 1215.36994740, 1287.57014230, 1251.47004485, 1311.63687393, 1395.87043464, 1395.87043464, 1431.97053209, 1419.93716627, 1383.83706883, 1347.73697138, 1299.60350811, 1359.77033719, 1407.90380046, 1359.77033719, 1198.52323526, 1140.76307935, 1116.69634771, 1070.96955761, 1143.16975251, 1275.53677648, 1347.73697138, 1275.53677648, 1323.67023975, 1275.53677648, 1215.36994740, 1227.40331322, 1239.43667904, 1197.31989868, 1191.30321577, 1191.30321577, 1133.54305986, 1167.23648414, 1191.30321577, 1227.40331322, 1203.33658159, 1251.47004485, 1263.50341067, 1215.36994740, 1203.33658159, 1191.30321577, 1198.52323526, 1239.43667904, 1251.47004485, 1215.36994740, 1215.36994740, 1227.40331322, 1215.36994740, 1275.53677648, 1275.53677648, 1311.63687393, 1287.57014230, 1383.83706883, 1468.07062954, 1431.97053209, 1395.87043464, 1371.80370301, 1383.83706883, 1323.67023975, 1419.93716627, 1407.90380046, 1407.90380046, 1347.73697138, 1371.80370301, 1347.73697138, 1371.80370301, 1383.83706883, 1335.70360556, 1407.90380046, 1407.90380046, 1468.07062954, 1444.00389791, 1492.13736117, 1480.10399535, 1444.00389791, 1395.87043464, 1419.93716627, 1407.90380046, 1395.87043464, 1359.77033719, 1335.70360556, 1335.70360556, 1299.60350811, 1299.60350811, 1335.70360556, 1287.57014230, 1275.53677648, 1251.47004485, 1299.60350811, 1215.36994740, 1263.50341067, 1263.50341067, 1323.67023975, 1335.70360556, 1323.67023975, 1347.73697138, 1299.60350811, 1263.50341067, 1275.53677648, 1263.50341067, 1311.63687393, 1359.77033719, 1359.77033719, 1287.57014230, 1347.73697138, 1347.73697138, 1335.70360556, 1335.70360556, 1323.67023975, 1371.80370301, 1383.83706883, 1395.87043464, 1468.07062954, 1456.03726372, 1480.10399535, 1444.00389791, 1444.00389791, 1419.93716627, 1395.87043464, 1383.83706883, 1383.83706883, 1431.97053209, 1431.97053209, 1456.03726372, 1444.00389791, 1419.93716627, 1431.97053209, 1407.90380046, 1407.90380046, 1395.87043464, 1419.93716627, 1431.97053209, 1468.07062954, 1456.03726372, 1468.07062954, 1444.00389791, 1419.93716627, 1383.83706883, 1419.93716627, 1444.00389791, 1456.03726372, 1444.00389791, 1419.93716627, 1407.90380046, 1395.87043464, 1311.63687393, 1215.36994740, 1275.53677648, 1251.47004485, 1239.43667904, 1203.33658159, 1215.36994740, 1175.65984021, 1176.86317679, 1161.21980123, 1215.36994740, 1198.52323526, 1197.31989868, 1203.33658159, 1202.13324501, 1203.33658159, 1163.62647440, 1163.62647440, 1146.77976225, 1139.55974276, 1139.55974276, 1117.89968430, 1131.13638669, 1108.27299164, 1123.91636720, 1146.77976225, 1147.98309883, 1126.32304037, 1127.52637695, 1099.84963557, 1137.15306960, 1151.59310858, 1197.31989868, 1175.65984021, 1215.36994740, 1191.30321577, 1192.50655235, 1251.47004485, 1227.40331322, 1239.43667904, 1203.33658159, 1194.91322552, 1194.91322552, 1197.31989868, 1227.40331322, 1239.43667904, 1275.53677648, 1275.53677648, 1299.60350811, 1359.77033719, 1371.80370301, 1323.67023975, 1323.67023975, 1335.70360556, 1347.73697138, 1335.70360556, 1335.70360556, 1335.70360556, 1359.77033719, 1335.70360556, 1311.63687393, 1347.73697138, 1383.83706883, 1395.87043464, 1395.87043464, 1407.90380046, 1468.07062954, 1431.97053209, 1431.97053209, 1456.03726372, 1468.07062954, 1480.10399535, 1480.10399535, 1492.13736117, 1492.13736117, 1480.10399535, 1468.07062954, 1419.93716627, 1407.90380046, 1419.93716627, 1407.90380046, 1383.83706883, 1371.80370301, 1371.80370301, 1335.70360556, 1323.67023975, 1371.80370301, 1371.80370301, 1395.87043464, 1359.77033719, 1359.77033719, 1383.83706883, 1347.73697138, 1323.67023975, 1347.73697138, 1311.63687393, 1347.73697138, 1323.67023975, 1311.63687393, 1311.63687393, 1299.60350811, 1287.57014230, 1287.57014230, 1251.47004485, 1215.36994740, 1227.40331322, 1227.40331322, 1196.11656210, 1190.09987919, 1227.40331322, 1203.33658159, 1185.28653286, 1215.36994740, 1215.36994740, 1200.92990842, 1196.11656210, 1215.36994740, 1251.47004485, 1323.67023975, 1371.80370301, 1407.90380046, 1383.83706883, 1347.73697138, 1347.73697138, 1371.80370301, 1371.80370301, 1371.80370301, 1371.80370301, 1395.87043464, 1359.77033719, 1359.77033719, 1383.83706883, 1359.77033719, 1323.67023975, 1299.60350811, 1287.57014230, 1287.57014230, 1275.53677648, 1275.53677648, 1263.50341067], 
							[415.15112065, 398.30440851, 368.22099397, 373.03434029, 379.05102320, 371.83100371, 354.98429157, 357.39096473, 356.18762815, 367.01765738, 374.23767687, 371.83100371, 367.01765738, 364.61098422, 354.98429157, 348.96760866, 314.07084779, 315.27418438, 330.91755994, 334.52756968, 333.32423310, 321.29086728, 333.32423310, 317.68085754, 297.22413565, 320.08753070, 374.23767687, 364.61098422, 342.95092575, 361.00097448, 365.81432080, 350.17094524, 353.78095499, 356.18762815, 352.57761841, 344.15426233, 340.54425259, 335.73090626, 339.34091601, 340.54425259, 354.98429157, 350.17094524, 359.79763789, 352.57761841, 351.37428182, 344.15426233, 344.15426233, 336.93424284, 358.59430131, 348.96760866, 352.57761841, 358.59430131, 357.39096473, 350.17094524, 381.45769636, 405.52442800, 407.93110116, 487.35131554, 475.31794973, 469.30126682, 451.25121810, 498.18134478, 486.14797896, 481.33463264, 490.96132529, 486.14797896, 494.57133503, 489.75798871, 457.26790100, 444.03119861, 436.81117912, 434.40450595, 448.84454493, 440.42118886, 445.23453519, 446.43787177, 464.48792049, 452.45455468, 458.47123758, 438.01451570, 400.71108167, 398.30440851, 387.47437927, 382.66103294, 407.93110116, 395.89773534, 412.74444748, 401.91441825, 385.06770611, 404.32109141, 395.89773534, 373.03434029, 375.44101346, 376.64435004, 367.01765738, 364.61098422, 369.42433055, 373.03434029, 388.67771585, 388.67771585, 395.89773534, 416.35445723, 419.96446697, 407.93110116, 419.96446697, 413.94778407, 415.15112065, 425.98114988, 407.93110116, 388.67771585, 424.77781330, 421.16780356, 411.54111090, 417.55779381, 421.16780356, 418.76113039, 440.42118886, 447.64120835, 438.01451570, 444.03119861, 435.60784253, 429.59115963, 413.94778407, 411.54111090, 411.54111090, 407.93110116, 401.91441825, 436.81117912, 446.43787177, 442.82786202, 445.23453519, 463.28458391, 445.23453519, 452.45455468, 470.50460340, 469.30126682, 456.06456442, 466.89459366, 469.30126682, 476.52128631, 478.92795947, 464.48792049, 460.87791075, 448.84454493, 440.42118886, 430.79449621, 427.18448646, 442.82786202, 451.25121810, 440.42118886, 444.03119861, 411.54111090, 435.60784253, 424.77781330, 422.37114014, 407.93110116, 409.13443774, 409.13443774, 406.72776458, 393.49106218, 369.42433055, 403.11775483, 392.28772560, 398.30440851, 416.35445723, 401.91441825, 407.93110116, 397.10107192, 381.45769636, 375.44101346, 376.64435004, 358.59430131, 344.15426233, 347.76427208, 332.12089652, 358.59430131, 365.81432080, 373.03434029, 359.79763789, 364.61098422, 380.25435978, 395.89773534, 374.23767687, 394.69439876, 403.11775483, 429.59115963, 401.91441825, 416.35445723, 427.18448646, 422.37114014, 422.37114014, 401.91441825, 415.15112065, 416.35445723, 421.16780356, 425.98114988, 429.59115963, 430.79449621, 422.37114014, 423.57447672, 425.98114988, 444.03119861, 433.20116937, 425.98114988, 423.57447672, 447.64120835, 434.40450595, 429.59115963, 406.72776458, 405.52442800, 401.91441825, 407.93110116, 423.57447672, 442.82786202, 442.82786202, 459.67457417, 448.84454493, 458.47123758, 444.03119861, 436.81117912, 447.64120835, 444.03119861, 441.62452544, 427.18448646, 435.60784253, 450.04788151, 452.45455468, 440.42118886, 415.15112065, 413.94778407, 419.96446697, 433.20116937, 418.76113039, 417.55779381, 431.99783279, 413.94778407, 422.37114014, 415.15112065, 410.33777432, 416.35445723, 412.74444748, 404.32109141, 418.76113039, 401.91441825, 398.30440851, 406.72776458, 416.35445723, 423.57447672, 413.94778407, 409.13443774, 407.93110116, 415.15112065, 405.52442800, 411.54111090, 357.39096473, 367.01765738, 352.57761841, 361.00097448, 353.78095499, 399.50774509, 403.11775483, 386.27104269, 401.91441825, 395.89773534, 405.52442800, 412.74444748, 431.99783279, 433.20116937, 430.79449621, 431.99783279, 451.25121810, 456.06456442, 444.03119861, 419.96446697, 410.33777432, 418.76113039, 415.15112065, 421.16780356, 436.81117912, 424.77781330, 425.98114988, 403.11775483, 401.91441825, 385.06770611, 361.00097448, 382.66103294, 374.23767687, 368.22099397, 358.59430131, 332.12089652, 292.41078933], 
							[805.03217308, 779.76210487, 746.06868058, 766.52540247, 755.69537324, 703.95190023, 683.49517834, 709.96858314, 747.27201717, 693.12187099, 719.59527579, 728.01863186, 717.18860263, 734.03531477, 691.91853441, 676.27515885, 663.03845645, 602.87162738, 655.81843697, 700.34189048, 677.47849543, 625.73502243, 673.86848569, 649.80175406, 635.36171508, 677.47849543, 752.08536349, 765.32206589, 730.42530502, 718.39193921, 724.40862212, 682.29184176, 685.90185151, 719.59527579, 722.00194895, 690.71519783, 697.93521732, 675.07182227, 720.79861237, 720.79861237, 715.98526604, 725.61195870, 732.83197819, 728.01863186, 719.59527579, 703.95190023, 699.13855390, 681.08850518, 728.01863186, 720.79861237, 699.13855390, 693.12187099, 696.73188074, 654.61510038, 715.98526604, 786.98212436, 770.13541222, 843.53894369, 849.55562660, 870.01234849, 877.23236798, 943.41587996, 914.53580201, 883.24905089, 914.53580201, 910.92579226, 922.95915808, 926.56916782, 890.46907037, 894.07908012, 898.89242645, 898.89242645, 931.38251415, 902.50243619, 913.33246543, 918.14581175, 962.66926527, 939.80587022, 943.41587996, 977.10930425, 962.66926527, 947.02588971, 932.58585073, 945.82255313, 961.46592869, 957.85591894, 959.05925553, 956.65258236, 927.77250440, 910.92579226, 879.63904114, 818.26887548, 866.40233874, 831.50557788, 878.43570456, 833.91225104, 833.91225104, 832.70891446, 882.04571430, 865.19900216, 824.28555839, 889.26573379, 889.26573379, 890.46907037, 908.51911910, 924.16249466, 932.58585073, 938.60253364, 926.56916782, 904.90910935, 930.17917757, 960.26259211, 949.43256287, 954.24590920, 947.02588971, 962.66926527, 941.00920680, 962.66926527, 989.14267007, 1015.61607486, 990.34600665, 983.12598716, 1002.37937246, 996.36268955, 1040.88614307, 1036.07279675, 1031.25945042, 1045.69948940, 1033.66612358, 1049.30949914, 1033.66612358, 1018.02274802, 1009.59939195, 989.14267007, 1016.81941144, 1015.61607486, 1007.19271879, 1031.25945042, 1026.44610409, 1031.25945042, 1038.47946991, 1015.61607486, 1018.02274802, 992.75267981, 1028.85277726, 1008.39605537, 1002.37937246, 1008.39605537, 1022.83609435, 1022.83609435, 1010.80272853, 995.15935297, 944.61921655, 943.41587996, 880.84237772, 831.50557788, 819.47221206, 844.74228027, 818.26887548, 792.99880727, 743.66200742, 818.26887548, 782.16877803, 829.09890471, 883.24905089, 866.40233874, 866.40233874, 857.97898267, 818.26887548, 824.28555839, 811.04885599, 791.79547068, 777.35543171, 818.26887548, 796.60881701, 842.33560711, 850.75896318, 830.30224130, 849.55562660, 865.19900216, 848.35229002, 930.17917757, 902.50243619, 916.94247517, 942.21254338, 972.29595792, 884.45238747, 925.36583124, 956.65258236, 934.99252389, 969.88928476, 955.44924578, 973.49929450, 977.10930425, 995.15935297, 1008.39605537, 1037.27613333, 1043.29281624, 1034.86946017, 1081.79958685, 1090.22294292, 1149.18643542, 1111.88300139, 1117.89968430, 1125.11970378, 1131.13638669, 1131.13638669, 1133.54305986, 1129.93305011, 1085.40959659, 1083.00292343, 1085.40959659, 1107.06965506, 1166.03314756, 1194.91322552, 1179.26984996, 1178.06651337, 1203.33658159, 1176.86317679, 1202.13324501, 1190.09987919, 1161.21980123, 1197.31989868, 1196.11656210, 1184.08319628, 1191.30321577, 1194.91322552, 1188.89654261, 1178.06651337, 1174.45650363, 1160.01646465, 1161.21980123, 1138.35640618, 1152.79644516, 1145.57642567, 1103.45964532, 1116.69634771, 1125.11970378, 1132.33972327, 1161.21980123, 1161.21980123, 1158.81312807, 1178.06651337, 1139.55974276, 1127.52637695, 1153.99978174, 1144.37308909, 1168.43982072, 1164.82981098, 1150.38977200, 1134.74639644, 1122.71303062, 1127.52637695, 1131.13638669, 1022.83609435, 1050.51283573, 1052.91950889, 1073.37623078, 1089.01960634, 1096.23962583, 1085.40959659, 1075.78290394, 1050.51283573, 1052.91950889, 1050.51283573, 1080.59625027, 1116.69634771, 1101.05297215, 1109.47632822, 1123.91636720, 1179.26984996, 1167.23648414, 1191.30321577, 1227.40331322, 1239.43667904, 1251.47004485, 1311.63687393, 1299.60350811, 1287.57014230, 1287.57014230, 1323.67023975, 1299.60350811, 1323.67023975, 1323.67023975, 1323.67023975, 1287.57014230, 1239.43667904, 1184.08319628, 1203.33658159, 1131.13638669, 1113.08633797], 
							[750.88202691, 758.10204640, 770.13541222, 743.66200742, 734.03531477, 631.75170533, 655.81843697, 617.31166635, 583.61824207, 593.24493472, 648.59841748, 643.78507115, 620.92167610, 619.71833952, 587.22825181, 601.66829079, 565.56819335, 529.46809590, 547.51814462, 518.63806666, 515.02805692, 476.52128631, 477.72462289, 445.23453519, 373.03434029, 502.99469110, 622.12501268, 685.90185151, 647.39508089, 606.48163712, 683.49517834, 631.75170533, 642.58173457, 619.71833952, 636.56505166, 613.70165661, 604.07496396, 531.87476906, 577.60155916, 595.65160789, 577.60155916, 570.38153967, 595.65160789, 549.92481779, 541.50146171, 535.48477881, 521.04473983, 509.01137401, 539.09478855, 546.31480804, 521.04473983, 524.65474957, 488.55465212, 488.55465212, 537.89145197, 565.56819335, 559.55151044, 578.80489574, 624.53168584, 620.92167610, 600.46495421, 658.22511013, 723.20528553, 712.37525630, 711.17191972, 743.66200742, 728.01863186, 711.17191972, 661.83511987, 636.56505166, 628.14169559, 637.76838824, 625.73502243, 590.83826156, 623.32834926, 636.56505166, 625.73502243, 617.31166635, 616.10832977, 624.53168584, 623.32834926, 613.70165661, 617.31166635, 607.68497370, 600.46495421, 598.05828105, 589.63492498, 565.56819335, 500.58801794, 515.02805692, 502.99469110, 463.28458391, 509.01137401, 499.38468136, 496.97800820, 466.89459366, 493.36799845, 495.77467161, 545.11147146, 547.51814462, 540.29812513, 612.49832003, 601.66829079, 543.90813488, 569.17820309, 571.58487625, 598.05828105, 590.83826156, 569.17820309, 559.55151044, 575.19488600, 578.80489574, 565.56819335, 565.56819335, 554.73816411, 570.38153967, 560.75484702, 573.99154942, 601.66829079, 613.70165661, 634.15837850, 611.29498345, 608.88831028, 614.90499319, 610.09164687, 592.04159814, 561.95818360, 575.19488600, 600.46495421, 580.00823233, 589.63492498, 612.49832003, 590.83826156, 565.56819335, 564.36485676, 570.38153967, 559.55151044, 577.60155916, 573.99154942, 543.90813488, 543.90813488, 500.58801794, 509.01137401, 490.96132529, 496.97800820, 492.16466187, 490.96132529, 504.19802769, 523.45141299, 507.80803743, 507.80803743, 543.90813488, 629.34503217, 756.89870982, 743.66200742, 697.93521732, 663.03845645, 631.75170533, 624.53168584, 560.75484702, 511.41804717, 529.46809590, 553.53482753, 555.94150069, 573.99154942, 547.51814462, 563.16152018, 531.87476906, 496.97800820, 463.28458391, 475.31794973, 428.38782305, 407.93110116, 386.27104269, 370.62766713, 458.47123758, 501.79135452, 466.89459366, 433.20116937, 469.30126682, 529.46809590, 643.78507115, 587.22825181, 571.58487625, 570.38153967, 611.29498345, 554.73816411, 554.73816411, 548.72148120, 530.67143248, 496.97800820, 481.33463264, 505.40136427, 489.75798871, 484.94464238, 496.97800820, 513.82472034, 511.41804717, 509.01137401, 534.28144222, 521.04473983, 586.02491523, 559.55151044, 546.31480804, 534.28144222, 543.90813488, 528.26475932, 489.75798871, 481.33463264, 480.13129605, 484.94464238, 481.33463264, 481.33463264, 500.58801794, 500.58801794, 528.26475932, 529.46809590, 528.26475932, 504.19802769, 502.99469110, 501.79135452, 495.77467161, 512.62138376, 501.79135452, 499.38468136, 492.16466187, 489.75798871, 493.36799845, 475.31794973, 463.28458391, 451.25121810, 434.40450595, 433.20116937, 429.59115963, 436.81117912, 430.79449621, 415.15112065, 417.55779381, 428.38782305, 433.20116937, 428.38782305, 415.15112065, 433.20116937, 404.32109141, 394.69439876, 397.10107192, 389.88105243, 381.45769636, 353.78095499, 348.96760866, 339.34091601, 300.83414540, 299.63080882, 290.00411616, 246.68399923, 285.19076984, 258.71736504, 253.90401872, 208.17722861, 203.36388229, 216.60058469, 209.38056520, 204.56721887, 184.11049698, 199.75387254, 194.94052622, 174.48380433, 156.43375561, 187.72050673, 175.68714091, 176.89047749, 167.26378484, 172.07713117, 174.48380433, 168.46712142, 169.67045800, 160.04376535, 162.45043851, 158.84042877, 143.19705321, 101.08027285, 103.48694602, 107.09695576, 91.45358020, 86.64023387, 84.23356071, 77.01354122, 86.64023387, 81.82688755, 79.42021438, 68.59018515], 
							[1311.63687393, 1227.40331322, 1251.47004485, 1182.87985970, 1215.36994740, 1251.47004485, 1129.93305011, 955.44924578, 951.83923604, 908.51911910, 914.53580201, 938.60253364, 953.04257262, 954.24590920, 932.58585073, 903.70577277, 824.28555839, 759.30538298, 789.38879752, 819.47221206, 808.64218283, 736.44198793, 788.18546094, 788.18546094, 779.76210487, 784.57545120, 896.48575328, 983.12598716, 1064.95287471, 1076.98624052, 1087.81626976, 1022.83609435, 914.53580201, 956.65258236, 914.53580201, 879.63904114, 915.73913859, 903.70577277, 950.63589945, 977.10930425, 942.21254338, 938.60253364, 941.00920680, 922.95915808, 867.60567532, 850.75896318, 859.18231925, 801.42216334, 857.97898267, 914.53580201, 860.38565584, 908.51911910, 933.78918731, 967.48261160, 1057.73285522, 1051.71617231, 1080.59625027, 1028.85277726, 992.75267981, 1051.71617231, 1030.05611384, 1092.62961608, 1050.51283573, 1049.30949914, 1033.66612358, 1040.88614307, 1032.46278700, 996.36268955, 981.92265058, 941.00920680, 961.46592869, 998.76936272, 1021.63275777, 953.04257262, 1019.22608460, 1014.41273828, 1022.83609435, 1038.47946991, 1030.05611384, 1032.46278700, 1010.80272853, 992.75267981, 991.54934323, 1028.85277726, 1031.25945042, 984.32932374, 1003.58270904, 921.75582150, 777.35543171, 801.42216334, 729.22196844, 688.30852467, 746.06868058, 706.35857339, 764.11872931, 699.13855390, 725.61195870, 762.91539273, 836.31892420, 823.08222181, 783.37211461, 897.68908986, 855.57230951, 788.18546094, 842.33560711, 860.38565584, 890.46907037, 902.50243619, 857.97898267, 825.48889497, 874.82569481, 882.04571430, 879.63904114, 844.74228027, 861.58899242, 857.97898267, 896.48575328, 928.97584099, 914.53580201, 938.60253364, 937.39919706, 874.82569481, 880.84237772, 857.97898267, 902.50243619, 891.67240696, 909.72245568, 898.89242645, 922.95915808, 876.02903140, 901.29909961, 902.50243619, 866.40233874, 903.70577277, 865.19900216, 820.67554864, 851.96229976, 848.35229002, 837.52226079, 854.36897293, 860.38565584, 832.70891446, 832.70891446, 803.82883650, 784.57545120, 802.62549992, 839.92893395, 878.43570456, 860.38565584, 871.21568507, 874.82569481, 877.23236798, 797.81215359, 851.96229976, 850.75896318, 863.99566558, 842.33560711, 776.15209512, 709.96858314, 740.05199768, 756.89870982, 708.76524656, 707.56190997, 729.22196844, 778.55876829, 789.38879752, 821.87888522, 805.03217308, 783.37211461, 747.27201717, 772.54208538, 723.20528553, 673.86848569, 714.78192946, 654.61510038, 773.74542196, 762.91539273, 788.18546094, 797.81215359, 782.16877803, 758.10204640, 839.92893395, 820.67554864, 824.28555839, 860.38565584, 867.60567532, 833.91225104, 825.48889497, 861.58899242, 821.87888522, 854.36897293, 788.18546094, 829.09890471, 797.81215359, 833.91225104, 820.67554864, 831.50557788, 842.33560711, 866.40233874, 867.60567532, 910.92579226, 907.31578252, 888.06239721, 878.43570456, 998.76936272, 936.19586048, 941.00920680, 951.83923604, 914.53580201, 913.33246543, 925.36583124, 957.85591894, 933.78918731, 986.73599690, 1009.59939195, 998.76936272, 1066.15621129, 1076.98624052, 1031.25945042, 1010.80272853, 1062.54620154, 1079.39291368, 1070.96955761, 1046.90282598, 1114.28967455, 1104.66298190, 1129.93305011, 1113.08633797, 1066.15621129, 1107.06965506, 1058.93619180, 1032.46278700, 1009.59939195, 1043.29281624, 1076.98624052, 987.93933348, 989.14267007, 1083.00292343, 1092.62961608, 1113.08633797, 1116.69634771, 1093.83295266, 1131.13638669, 1087.81626976, 1091.42627950, 1137.15306960, 1123.91636720, 1147.98309883, 1203.33658159, 1152.79644516, 1115.49301113, 1107.06965506, 1039.68280649, 1025.24276751, 837.52226079, 850.75896318, 831.50557788, 874.82569481, 835.11558762, 853.16563635, 860.38565584, 832.70891446, 826.69223155, 825.48889497, 877.23236798, 803.82883650, 741.25533426, 738.84866109, 801.42216334, 753.28870007, 817.06553890, 836.31892420, 890.46907037, 862.79232900, 902.50243619, 949.43256287, 969.88928476, 1016.81941144, 969.88928476, 894.07908012, 892.87574354, 914.53580201, 920.55248491, 842.33560711, 799.01549017, 722.00194895, 712.37525630, 676.27515885, 684.69851492, 625.73502243, 646.19174431], 
							[1299.60350811, 1323.67023975, 1383.83706883, 1275.53677648, 1299.60350811, 1275.53677648, 1173.25316705, 1083.00292343, 1113.08633797, 1055.32618205, 1137.15306960, 1173.25316705, 1143.16975251, 1191.30321577, 1125.11970378, 1093.83295266, 1030.05611384, 991.54934323, 1084.20626001, 1125.11970378, 1076.98624052, 977.10930425, 1044.49615282, 1018.02274802, 979.51597741, 992.75267981, 1135.94973302, 1138.35640618, 1164.82981098, 1185.28653286, 1203.33658159, 1075.78290394, 1024.03943093, 1037.27613333, 1076.98624052, 1034.86946017, 1060.13952838, 1054.12284547, 1129.93305011, 1143.16975251, 1123.91636720, 1128.72971353, 1164.82981098, 1138.35640618, 1070.96955761, 1048.10616256, 1079.39291368, 1027.64944068, 1068.56288445, 1137.15306960, 1096.23962583, 1131.13638669, 1153.99978174, 1166.03314756, 1251.47004485, 1198.52323526, 1185.28653286, 1170.84649388, 1227.40331322, 1275.53677648, 1251.47004485, 1335.70360556, 1311.63687393, 1263.50341067, 1287.57014230, 1311.63687393, 1311.63687393, 1239.43667904, 1287.57014230, 1275.53677648, 1203.33658159, 1239.43667904, 1263.50341067, 1251.47004485, 1251.47004485, 1263.50341067, 1275.53677648, 1275.53677648, 1263.50341067, 1275.53677648, 1251.47004485, 1227.40331322, 1251.47004485, 1251.47004485, 1227.40331322, 1251.47004485, 1239.43667904, 1203.33658159, 1161.21980123, 1157.60979149, 1132.33972327, 1110.67966481, 1125.11970378, 1125.11970378, 1125.11970378, 1091.42627950, 1119.10302088, 1137.15306960, 1152.79644516, 1146.77976225, 1116.69634771, 1185.28653286, 1143.16975251, 1085.40959659, 1108.27299164, 1114.28967455, 1105.86631848, 1108.27299164, 1078.18957710, 1055.32618205, 1097.44296241, 1108.27299164, 1123.91636720, 1089.01960634, 1101.05297215, 1104.66298190, 1119.10302088, 1131.13638669, 1120.30635746, 1164.82981098, 1164.82981098, 1143.16975251, 1137.15306960, 1174.45650363, 1196.11656210, 1162.42313781, 1161.21980123, 1158.81312807, 1167.23648414, 1149.18643542, 1140.76307935, 1140.76307935, 1116.69634771, 1111.88300139, 1114.28967455, 1099.84963557, 1107.06965506, 1108.27299164, 1095.03628924, 1057.73285522, 1070.96955761, 1068.56288445, 1087.81626976, 1079.39291368, 1117.89968430, 1137.15306960, 1160.01646465, 1167.23648414, 1138.35640618, 1123.91636720, 1125.11970378, 1125.11970378, 1083.00292343, 1095.03628924, 1093.83295266, 1096.23962583, 1078.18957710, 1044.49615282, 1010.80272853, 1045.69948940, 1004.78604563, 962.66926527, 962.66926527, 960.26259211, 984.32932374, 991.54934323, 1015.61607486, 1034.86946017, 1018.02274802, 1016.81941144, 1034.86946017, 997.56602614, 969.88928476, 987.93933348, 953.04257262, 1025.24276751, 1021.63275777, 1013.20940170, 1027.64944068, 1019.22608460, 1032.46278700, 1056.52951863, 1058.93619180, 1075.78290394, 1101.05297215, 1089.01960634, 1052.91950889, 1064.95287471, 1076.98624052, 1049.30949914, 1067.35954787, 1022.83609435, 1062.54620154, 1043.29281624, 1069.76622103, 1048.10616256, 1061.34286496, 1087.81626976, 1115.49301113, 1157.60979149, 1155.20311832, 1166.03314756, 1166.03314756, 1155.20311832, 1172.04983047, 1163.62647440, 1178.06651337, 1184.08319628, 1155.20311832, 1119.10302088, 1167.23648414, 1145.57642567, 1191.30321577, 1200.92990842, 1203.33658159, 1203.33658159, 1227.40331322, 1251.47004485, 1193.70988894, 1168.43982072, 1203.33658159, 1227.40331322, 1239.43667904, 1251.47004485, 1227.40331322, 1263.50341067, 1239.43667904, 1239.43667904, 1239.43667904, 1239.43667904, 1227.40331322, 1227.40331322, 1227.40331322, 1227.40331322, 1227.40331322, 1199.72657184, 1227.40331322, 1215.36994740, 1227.40331322, 1239.43667904, 1239.43667904, 1203.33658159, 1239.43667904, 1215.36994740, 1215.36994740, 1215.36994740, 1227.40331322, 1215.36994740, 1200.92990842, 1167.23648414, 1170.84649388, 1168.43982072, 1147.98309883, 1155.20311832, 1110.67966481, 1090.22294292, 1086.61293317, 1101.05297215, 1079.39291368, 1090.22294292, 1103.45964532, 1074.57956736, 1055.32618205, 1038.47946991, 1083.00292343, 1045.69948940, 1008.39605537, 1001.17603588, 1014.41273828, 1012.00606512, 1062.54620154, 1086.61293317, 1111.88300139, 1095.03628924, 1117.89968430, 1164.82981098, 1175.65984021, 1163.62647440, 1144.37308909, 1117.89968430, 1117.89968430, 1120.30635746, 1105.86631848, 1054.12284547, 1058.93619180, 1020.42942119, 1012.00606512, 1030.05611384, 1005.98938221, 992.75267981, 986.73599690], 
							[914.53580201, 878.43570456, 902.50243619, 830.30224130, 818.26887548, 811.04885599, 770.13541222, 772.54208538, 832.70891446, 771.33874880, 812.25219257, 800.21882676, 794.20214385, 819.47221206, 762.91539273, 717.18860263, 709.96858314, 637.76838824, 703.95190023, 758.10204640, 740.05199768, 709.96858314, 715.98526604, 736.44198793, 734.03531477, 746.06868058, 818.26887548, 849.55562660, 831.50557788, 843.53894369, 832.70891446, 747.27201717, 723.20528553, 747.27201717, 764.11872931, 725.61195870, 697.93521732, 676.27515885, 672.66514911, 742.45867084, 740.05199768, 820.67554864, 860.38565584, 872.41902165, 850.75896318, 860.38565584, 850.75896318, 890.46907037, 925.36583124, 978.31264083, 926.56916782, 944.61921655, 932.58585073, 932.58585073, 1022.83609435, 1003.58270904, 1016.81941144, 998.76936272, 1083.00292343, 1227.40331322, 1190.09987919, 1287.57014230, 1263.50341067, 1193.70988894, 1197.31989868, 1215.36994740, 1185.28653286, 1155.20311832, 1087.81626976, 1155.20311832, 1152.79644516, 1137.15306960, 1129.93305011, 1101.05297215, 1126.32304037, 1117.89968430, 1107.06965506, 1107.06965506, 1113.08633797, 1155.20311832, 1143.16975251, 1107.06965506, 1089.01960634, 1070.96955761, 1158.81312807, 1119.10302088, 1107.06965506, 1074.57956736, 1010.80272853, 1046.90282598, 977.10930425, 844.74228027, 908.51911910, 977.10930425, 1076.98624052, 1036.07279675, 1022.83609435, 1061.34286496, 1101.05297215, 1099.84963557, 1064.95287471, 1143.16975251, 1151.59310858, 1058.93619180, 1179.26984996, 1215.36994740, 1251.47004485, 1239.43667904, 1215.36994740, 1203.33658159, 1287.57014230, 1263.50341067, 1227.40331322, 1197.31989868, 1200.92990842, 1227.40331322, 1192.50655235, 1227.40331322, 1263.50341067, 1359.77033719, 1311.63687393, 1287.57014230, 1287.57014230, 1287.57014230, 1263.50341067, 1263.50341067, 1227.40331322, 1203.33658159, 1227.40331322, 1215.36994740, 1199.72657184, 1227.40331322, 1156.40645491, 1123.91636720, 1120.30635746, 1084.20626001, 1151.59310858, 1185.28653286, 1197.31989868, 1215.36994740, 1202.13324501, 1127.52637695, 1152.79644516, 1097.44296241, 1125.11970378, 1116.69634771, 1119.10302088, 1132.33972327, 1166.03314756, 1129.93305011, 1101.05297215, 1022.83609435, 1005.98938221, 992.75267981, 992.75267981, 978.31264083, 954.24590920, 910.92579226, 868.80901191, 851.96229976, 850.75896318, 843.53894369, 835.11558762, 826.69223155, 876.02903140, 861.58899242, 862.79232900, 818.26887548, 867.60567532, 933.78918731, 998.76936272, 973.49929450, 962.66926527, 1064.95287471, 1003.58270904, 1095.03628924, 1149.18643542, 1167.23648414, 1227.40331322, 1299.60350811, 1359.77033719, 1419.93716627, 1468.07062954, 1612.47101933, 1648.57111678, 1648.57111678, 1516.20409280, 1540.27082443, 1684.67121422, 1588.40428770, 1672.63784841, 1660.60448259, 1913.30516472, 1829.07160401, 1708.73794585, 1696.70458004, 1768.90477493, 1732.80467749, 1624.50438514, 1636.53775096, 1684.67121422, 1780.93814075, 1636.53775096, 1660.60448259, 1612.47101933, 1564.33755606, 1708.73794585, 1708.73794585, 1732.80467749, 1780.93814075, 1504.17072698, 1516.20409280, 1480.10399535, 1576.37092188, 1576.37092188, 1660.60448259, 1696.70458004, 1744.83804330, 1660.60448259, 1600.43765351, 1624.50438514, 1468.07062954, 1588.40428770, 1564.33755606, 1576.37092188, 1564.33755606, 1516.20409280, 1528.23745862, 1516.20409280, 1528.23745862, 1492.13736117, 1492.13736117, 1444.00389791, 1504.17072698, 1480.10399535, 1419.93716627, 1383.83706883, 1480.10399535, 1540.27082443, 1576.37092188, 1528.23745862, 1516.20409280, 1600.43765351, 1504.17072698, 1492.13736117, 1540.27082443, 1528.23745862, 1660.60448259, 1672.63784841, 1636.53775096, 1684.67121422, 1672.63784841, 1744.83804330, 1732.80467749, 1756.87140912, 1877.20506728, 1973.47199380, 2021.60545707, 1865.17170146, 2238.20604175, 2105.83901778, 2129.90574941, 2117.87238359, 1985.50535962, 2009.57209125, 1973.47199380, 2057.70555452, 2178.03921267, 2129.90574941, 2141.93911523, 2202.10594431, 2250.23940757, 2226.17267594, 2178.03921267, 2370.57306573, 2406.67316318, 2310.40623665, 2238.20604175, 2286.33950502, 2166.00584686, 2286.33950502, 2370.57306573, 2406.67316318, 2478.87335807, 2900.04116163, 3080.54164886, 2731.57404020, 2394.63979736, 2539.04018715, 2262.27277339, 2262.27277339], 
							[470.50460340, 470.50460340, 431.99783279, 421.16780356, 429.59115963, 399.50774509, 403.11775483, 387.47437927, 386.27104269, 392.28772560, 418.76113039, 397.10107192, 400.71108167, 400.71108167, 391.08438902, 373.03434029, 346.56093550, 342.95092575, 357.39096473, 344.15426233, 353.78095499, 327.30755019, 334.52756968, 324.90087703, 310.46083805, 324.90087703, 375.44101346, 375.44101346, 367.01765738, 368.22099397, 371.83100371, 373.03434029, 361.00097448, 375.44101346, 380.25435978, 377.84768662, 364.61098422, 356.18762815, 353.78095499, 351.37428182, 359.79763789, 365.81432080, 363.40764764, 359.79763789, 353.78095499, 348.96760866, 345.35759892, 340.54425259, 342.95092575, 339.34091601, 332.12089652, 336.93424284, 333.32423310, 316.47752096, 346.56093550, 358.59430131, 364.61098422, 409.13443774, 415.15112065, 417.55779381, 400.71108167, 435.60784253, 425.98114988, 412.74444748, 415.15112065, 425.98114988, 415.15112065, 424.77781330, 387.47437927, 394.69439876, 383.86436953, 400.71108167, 401.91441825, 381.45769636, 400.71108167, 388.67771585, 407.93110116, 389.88105243, 389.88105243, 382.66103294, 348.96760866, 353.78095499, 351.37428182, 340.54425259, 358.59430131, 357.39096473, 364.61098422, 352.57761841, 333.32423310, 323.69754045, 317.68085754, 309.25750147, 318.88419412, 317.68085754, 314.07084779, 309.25750147, 304.44415514, 315.27418438, 324.90087703, 321.29086728, 312.86751121, 344.15426233, 341.74758917, 318.88419412, 322.49420387, 330.91755994, 324.90087703, 341.74758917, 324.90087703, 318.88419412, 332.12089652, 335.73090626, 342.95092575, 342.95092575, 344.15426233, 344.15426233, 359.79763789, 363.40764764, 359.79763789, 376.64435004, 382.66103294, 367.01765738, 365.81432080, 354.98429157, 359.79763789, 352.57761841, 353.78095499, 377.84768662, 377.84768662, 385.06770611, 382.66103294, 399.50774509, 389.88105243, 389.88105243, 400.71108167, 398.30440851, 377.84768662, 383.86436953, 383.86436953, 397.10107192, 388.67771585, 373.03434029, 382.66103294, 371.83100371, 370.62766713, 370.62766713, 364.61098422, 371.83100371, 374.23767687, 363.40764764, 359.79763789, 347.76427208, 336.93424284, 346.56093550, 321.29086728, 308.05416489, 297.22413565, 306.85082830, 297.22413565, 290.00411616, 271.95406744, 290.00411616, 280.37742351, 298.42747223, 298.42747223, 303.24081856, 304.44415514, 286.39410642, 288.80077958, 287.59744300, 286.39410642, 257.51402846, 247.88733581, 240.66731632, 243.07398948, 259.92070162, 274.36074060, 268.34405769, 258.71736504, 256.31069188, 293.61412591, 304.44415514, 296.02079907, 305.64749172, 324.90087703, 346.56093550, 327.30755019, 335.73090626, 335.73090626, 329.71422336, 330.91755994, 308.05416489, 318.88419412, 312.86751121, 329.71422336, 338.13757943, 358.59430131, 368.22099397, 348.96760866, 376.64435004, 383.86436953, 415.15112065, 394.69439876, 385.06770611, 385.06770611, 404.32109141, 399.50774509, 394.69439876, 368.22099397, 359.79763789, 358.59430131, 373.03434029, 382.66103294, 398.30440851, 394.69439876, 411.54111090, 395.89773534, 401.91441825, 400.71108167, 373.03434029, 377.84768662, 368.22099397, 358.59430131, 362.20431106, 364.61098422, 377.84768662, 379.05102320, 363.40764764, 354.98429157, 350.17094524, 344.15426233, 356.18762815, 352.57761841, 353.78095499, 356.18762815, 342.95092575, 341.74758917, 338.13757943, 338.13757943, 354.98429157, 346.56093550, 340.54425259, 347.76427208, 333.32423310, 336.93424284, 336.93424284, 327.30755019, 336.93424284, 318.88419412, 300.83414540, 309.25750147, 287.59744300, 282.78409667, 293.61412591, 257.51402846, 265.93738453, 255.10735530, 252.70068213, 245.48066264, 262.32737479, 276.76741377, 268.34405769, 263.53071137, 256.31069188, 258.71736504, 268.34405769, 268.34405769, 258.71736504, 259.92070162, 282.78409667, 282.78409667, 287.59744300, 276.76741377, 270.75073086, 269.54739428, 269.54739428, 269.54739428, 257.51402846, 265.93738453, 253.90401872, 240.66731632, 238.26064315, 243.07398948, 222.61726759, 210.58390178, 214.19391152, 204.56721887, 198.55053596, 192.53385305, 193.73718964, 192.53385305], 
							[853.16563635, 830.30224130, 814.65886573, 785.77878778, 786.98212436, 772.54208538, 743.66200742, 643.78507115, 648.59841748, 623.32834926, 666.64846620, 709.96858314, 664.24179304, 706.35857339, 655.81843697, 626.93835901, 629.34503217, 582.41490549, 599.26161763, 620.92167610, 619.71833952, 546.31480804, 570.38153967, 577.60155916, 558.34817386, 564.36485676, 666.64846620, 646.19174431, 661.83511987, 695.52854416, 714.78192946, 654.61510038, 623.32834926, 619.71833952, 629.34503217, 604.07496396, 604.07496396, 612.49832003, 652.20842722, 647.39508089, 649.80175406, 652.20842722, 631.75170533, 622.12501268, 599.26161763, 595.65160789, 586.02491523, 576.39822258, 616.10832977, 631.75170533, 618.51500294, 626.93835901, 677.47849543, 669.05513936, 675.07182227, 690.71519783, 679.88516860, 644.98840773, 653.41176380, 693.12187099, 685.90185151, 701.54522707, 748.47535375, 782.16877803, 765.32206589, 770.13541222, 776.15209512, 735.23865135, 715.98526604, 690.71519783, 679.88516860, 693.12187099, 712.37525630, 697.93521732, 729.22196844, 720.79861237, 734.03531477, 752.08536349, 747.27201717, 761.71205615, 719.59527579, 697.93521732, 696.73188074, 676.27515885, 632.95504192, 611.29498345, 593.24493472, 558.34817386, 481.33463264, 483.74130580, 466.89459366, 403.11775483, 459.67457417, 421.16780356, 448.84454493, 416.35445723, 418.76113039, 446.43787177, 502.99469110, 516.23139350, 509.01137401, 577.60155916, 540.29812513, 470.50460340, 506.60470085, 519.84140325, 530.67143248, 560.75484702, 531.87476906, 498.18134478, 533.07810564, 561.95818360, 555.94150069, 536.68811539, 541.50146171, 549.92481779, 535.48477881, 571.58487625, 543.90813488, 587.22825181, 631.75170533, 584.82157865, 561.95818360, 559.55151044, 567.97486651, 553.53482753, 565.56819335, 547.51814462, 567.97486651, 555.94150069, 564.36485676, 567.97486651, 552.33149095, 563.16152018, 534.28144222, 493.36799845, 530.67143248, 537.89145197, 569.17820309, 567.97486651, 561.95818360, 555.94150069, 537.89145197, 492.16466187, 494.57133503, 501.79135452, 513.82472034, 564.36485676, 531.87476906, 531.87476906, 552.33149095, 524.65474957, 493.36799845, 527.06142274, 513.82472034, 529.46809590, 478.92795947, 445.23453519, 446.43787177, 444.03119861, 442.82786202, 405.52442800, 389.88105243, 413.94778407, 450.04788151, 462.08124733, 469.30126682, 452.45455468, 454.86122784, 400.71108167, 409.13443774, 383.86436953, 326.10421361, 346.56093550, 314.07084779, 409.13443774, 412.74444748, 416.35445723, 418.76113039, 392.28772560, 385.06770611, 425.98114988, 416.35445723, 419.96446697, 440.42118886, 463.28458391, 439.21785228, 416.35445723, 436.81117912, 415.15112065, 415.15112065, 385.06770611, 398.30440851, 401.91441825, 417.55779381, 417.55779381, 433.20116937, 438.01451570, 445.23453519, 442.82786202, 456.06456442, 462.08124733, 489.75798871, 493.36799845, 517.43473008, 518.63806666, 528.26475932, 505.40136427, 495.77467161, 489.75798871, 511.41804717, 498.18134478, 521.04473983, 758.10204640, 758.10204640, 694.32520758, 700.34189048, 706.35857339, 663.03845645, 675.07182227, 683.49517834, 700.34189048, 685.90185151, 640.17506140, 644.98840773, 673.86848569, 653.41176380, 652.20842722, 623.32834926, 622.12501268, 607.68497370, 566.77152993, 572.78821284, 581.21156891, 613.70165661, 554.73816411, 555.94150069, 600.46495421, 608.88831028, 586.02491523, 578.80489574, 577.60155916, 622.12501268, 582.41490549, 573.99154942, 622.12501268, 599.26161763, 612.49832003, 605.27830054, 557.14483728, 549.92481779, 513.82472034, 484.94464238, 463.28458391, 367.01765738, 409.13443774, 405.52442800, 412.74444748, 392.28772560, 391.08438902, 407.93110116, 392.28772560, 388.67771585, 371.83100371, 405.52442800, 370.62766713, 334.52756968, 323.69754045, 338.13757943, 327.30755019, 354.98429157, 377.84768662, 421.16780356, 397.10107192, 427.18448646, 433.20116937, 429.59115963, 429.59115963, 421.16780356, 391.08438902, 379.05102320, 391.08438902, 393.49106218, 339.34091601, 316.47752096, 285.19076984, 271.95406744, 270.75073086, 273.15740402, 265.93738453, 250.29400897], 
							[642.97883564, 632.35337362, 615.84359572, 605.21813371, 593.42543521, 589.88762566, 555.67676665, 516.73679487, 519.09533457, 522.63314412, 524.99168382, 510.84044562, 524.99168382, 526.18298703, 509.66117577, 493.13936450, 495.49790420, 450.67361654, 477.80885645, 477.80885645, 471.91250720, 445.95653714, 495.49790420, 475.45031675, 437.68961482, 442.41872759, 524.99168382, 519.09533457, 517.91606472, 529.72079658, 550.94765388, 513.19898532, 484.88447555, 483.70520570, 477.80885645, 466.00412459, 489.60155495, 482.52593585, 508.48190592, 497.86847727, 502.58555667, 508.48190592, 524.99168382, 513.19898532, 519.09533457, 469.55396750, 469.55396750, 442.41872759, 460.10777534, 489.60155495, 469.55396750, 468.36266429, 471.91250720, 473.09177705, 476.62958660, 469.55396750, 474.27104690, 510.84044562, 515.55752502, 528.54152673, 519.09533457, 536.79641568, 519.09533457, 524.99168382, 532.07933628, 534.43787598, 542.69276493, 528.54152673, 515.55752502, 533.25860613, 502.58555667, 497.86847727, 516.73679487, 501.40628682, 524.99168382, 517.91606472, 523.81241397, 502.58555667, 508.48190592, 503.76482652, 495.49790420, 497.86847727, 503.76482652, 483.70520570, 474.27104690, 476.62958660, 476.62958660, 467.18339444, 434.15180527, 417.64202737, 428.25545602, 395.22386686, 395.22386686, 378.70205559, 404.65802566, 414.10421782, 428.25545602, 442.41872759, 436.51034497, 443.59799744, 436.51034497, 461.28704519, 458.92850549, 421.17983692, 436.51034497, 438.86888467, 436.51034497, 448.31507684, 442.41872759, 436.51034497, 461.28704519, 451.85288639, 457.74923564, 460.10777534, 450.67361654, 474.27104690, 477.80885645, 492.16466187, 471.70793998, 483.74130580, 481.33463264, 465.69125707, 463.28458391, 450.04788151, 456.06456442, 439.21785228, 453.65789126, 440.42118886, 450.04788151, 446.43787177, 448.84454493, 451.25121810, 433.20116937, 433.20116937, 431.99783279, 427.18448646, 419.96446697, 425.98114988, 440.42118886, 428.38782305, 433.20116937, 405.52442800, 407.93110116, 385.06770611, 407.93110116, 391.08438902, 392.28772560, 413.94778407, 421.16780356, 400.71108167, 406.72776458, 393.49106218, 370.62766713, 383.86436953, 382.66103294, 361.00097448, 339.34091601, 338.13757943, 312.86751121, 282.78409667, 299.63080882, 294.81746249, 281.58076009, 297.22413565, 317.68085754, 312.86751121, 317.68085754, 304.44415514, 304.44415514, 292.41078933, 288.80077958, 265.93738453, 271.95406744, 244.27732606, 225.02394076, 253.90401872, 268.34405769, 267.14072111, 270.75073086, 257.51402846, 263.53071137, 280.37742351, 290.00411616, 293.61412591, 286.39410642, 279.17408693, 267.14072111, 267.14072111, 267.14072111, 259.92070162, 267.14072111, 239.46397974, 243.07398948, 241.87065290, 251.49734555, 256.31069188, 259.92070162, 268.34405769, 281.58076009, 291.20745274, 308.05416489, 321.29086728, 335.73090626, 321.29086728, 314.07084779, 348.96760866, 416.35445723, 382.66103294, 371.83100371, 361.00097448, 341.74758917, 345.35759892, 341.74758917, 369.42433055, 373.03434029, 394.69439876, 388.67771585, 409.13443774, 394.69439876, 386.27104269, 395.89773534, 395.89773534, 401.91441825, 385.06770611, 398.30440851, 398.30440851, 393.49106218, 388.67771585, 377.84768662, 374.23767687, 363.40764764, 351.37428182, 336.93424284, 342.95092575, 345.35759892, 335.73090626, 338.13757943, 341.74758917, 346.56093550, 347.76427208, 348.96760866, 336.93424284, 342.95092575, 340.54425259, 330.91755994, 340.54425259, 335.73090626, 328.51088677, 334.52756968, 308.05416489, 312.86751121, 296.02079907, 287.59744300, 287.59744300, 249.09067239, 270.75073086, 262.32737479, 264.73404795, 244.27732606, 247.88733581, 251.49734555, 251.49734555, 246.68399923, 250.29400897, 258.71736504, 241.87065290, 229.83728708, 214.19391152, 243.07398948, 239.46397974, 245.48066264, 258.71736504, 269.54739428, 262.32737479, 259.92070162, 263.53071137, 263.53071137, 265.93738453, 252.70068213, 240.66731632, 231.04062366, 233.44729683, 234.65063341, 208.17722861, 209.38056520, 200.95720913, 200.95720913, 204.56721887, 203.36388229, 200.95720913, 190.12717989], 
							[1612.47101933, 1588.40428770, 1347.73697138, 1407.90380046, 1347.73697138, 1131.13638669, 1083.00292343, 1203.33658159, 1263.50341067, 1239.43667904, 1251.47004485, 1239.43667904, 1263.50341067, 1347.73697138, 1299.60350811, 1239.43667904, 1174.45650363, 1076.98624052, 1182.87985970, 1157.60979149, 1167.23648414, 1111.88300139, 1176.86317679, 1162.42313781, 1168.43982072, 1359.77033719, 1528.23745862, 1576.37092188, 1480.10399535, 1407.90380046, 1480.10399535, 1323.67023975, 1275.53677648, 1275.53677648, 1299.60350811, 1359.77033719, 1371.80370301, 1299.60350811, 1371.80370301, 1383.83706883, 1359.77033719, 1431.97053209, 1516.20409280, 1600.43765351, 1444.00389791, 1395.87043464, 1359.77033719, 1395.87043464, 1456.03726372, 1444.00389791, 1407.90380046, 1407.90380046, 1456.03726372, 1468.07062954, 1588.40428770, 1756.87140912, 1636.53775096, 1660.60448259, 1744.83804330, 1708.73794585, 1744.83804330, 1829.07160401, 1805.00487238, 1792.97150657, 1768.90477493, 1985.50535962, 2021.60545707, 1985.50535962, 1877.20506728, 1937.37189636, 1985.50535962, 1949.40526217, 1961.43862799, 1925.33853054, 2021.60545707, 1961.43862799, 1973.47199380, 1949.40526217, 1913.30516472, 1973.47199380, 1901.27179891, 1925.33853054, 1877.20506728, 1961.43862799, 1985.50535962, 2021.60545707, 2069.73892033, 2021.60545707, 2021.60545707, 1973.47199380, 1949.40526217, 1708.73794585, 1817.03823820, 1829.07160401, 1805.00487238, 1720.77131167, 1744.83804330, 1792.97150657, 1792.97150657, 1853.13833565, 1756.87140912, 1829.07160401, 1805.00487238, 1744.83804330, 1768.90477493, 1792.97150657, 1564.33755606, 1660.60448259, 1600.43765351, 1552.30419025, 1648.57111678, 1648.57111678, 1684.67121422, 1660.60448259, 1696.70458004, 1720.77131167, 1732.80467749, 1792.97150657, 1720.77131167, 1708.73794585, 1708.73794585, 1684.67121422, 1660.60448259, 1660.60448259, 1684.67121422, 1612.47101933, 1696.70458004, 1696.70458004, 1696.70458004, 1672.63784841, 1672.63784841, 1684.67121422, 1624.50438514, 1600.43765351, 1576.37092188, 1552.30419025, 1540.27082443, 1540.27082443, 1576.37092188, 1564.33755606, 1576.37092188, 1576.37092188, 1564.33755606, 1528.23745862, 1624.50438514, 1588.40428770, 1600.43765351, 1624.50438514, 1612.47101933, 1600.43765351, 1552.30419025, 1504.17072698, 1444.00389791, 1480.10399535, 1444.00389791, 1431.97053209, 1419.93716627, 1407.90380046, 1347.73697138, 1299.60350811, 1215.36994740, 1275.53677648, 1227.40331322, 1275.53677648, 1335.70360556, 1347.73697138, 1371.80370301, 1419.93716627, 1323.67023975, 1239.43667904, 1239.43667904, 1131.13638669, 1121.50969404, 1091.42627950, 1064.95287471, 1162.42313781, 1143.16975251, 1121.50969404, 1227.40331322, 1215.36994740, 1239.43667904, 1275.53677648, 1251.47004485, 1239.43667904, 1239.43667904, 1275.53677648, 1299.60350811, 1299.60350811, 1263.50341067, 1215.36994740, 1263.50341067, 1119.10302088, 1251.47004485, 1215.36994740, 1299.60350811, 1287.57014230, 1359.77033719, 1395.87043464, 1492.13736117, 1576.37092188, 1576.37092188, 1624.50438514, 1492.13736117, 1444.00389791, 1504.17072698, 1444.00389791, 1468.07062954, 1444.00389791, 1407.90380046, 1395.87043464, 1395.87043464, 1383.83706883, 1431.97053209, 1468.07062954, 1480.10399535, 1468.07062954, 1407.90380046, 1468.07062954, 1444.00389791, 1383.83706883, 1431.97053209, 1419.93716627, 1419.93716627, 1431.97053209, 1456.03726372, 1480.10399535, 1468.07062954, 1419.93716627, 1407.90380046, 1383.83706883, 1323.67023975, 1227.40331322, 1251.47004485, 1263.50341067, 1299.60350811, 1239.43667904, 1251.47004485, 1263.50341067, 1287.57014230, 1335.70360556, 1299.60350811, 1287.57014230, 1323.67023975, 1311.63687393, 1311.63687393, 1311.63687393, 1287.57014230, 1299.60350811, 1251.47004485, 1135.94973302, 1131.13638669, 1113.08633797, 1014.41273828, 996.36268955, 782.16877803, 860.38565584, 844.74228027, 830.30224130, 797.81215359, 821.87888522, 951.83923604, 927.77250440, 882.04571430, 876.02903140, 944.61921655, 859.18231925, 829.09890471, 734.03531477, 830.30224130, 776.15209512, 829.09890471, 823.08222181, 860.38565584, 848.35229002, 838.72559737, 892.87574354, 986.73599690, 1039.68280649, 1070.96955761, 998.76936272, 932.58585073, 1031.25945042, 932.58585073, 862.79232900, 878.43570456, 921.75582150, 902.50243619, 854.36897293, 879.63904114, 791.79547068, 752.08536349], 
							[1203.33658159, 1215.36994740, 1227.40331322, 1239.43667904, 1203.33658159, 1113.08633797, 1024.03943093, 908.51911910, 800.21882676, 843.53894369, 888.06239721, 914.53580201, 902.50243619, 890.46907037, 842.33560711, 812.25219257, 730.42530502, 774.94875854, 740.05199768, 782.16877803, 782.16877803, 711.17191972, 773.74542196, 754.49203666, 731.62864161, 779.76210487, 844.74228027, 896.48575328, 950.63589945, 974.70263109, 1042.08947966, 896.48575328, 871.21568507, 874.82569481, 884.45238747, 902.50243619, 928.97584099, 892.87574354, 947.02588971, 985.53266032, 966.27927501, 944.61921655, 951.83923604, 931.38251415, 876.02903140, 885.65572405, 884.45238747, 855.57230951, 908.51911910, 972.29595792, 954.24590920, 979.51597741, 1022.83609435, 1079.39291368, 1188.89654261, 1167.23648414, 1125.11970378, 1058.93619180, 1049.30949914, 1152.79644516, 1126.32304037, 1131.13638669, 1109.47632822, 1104.66298190, 1076.98624052, 1110.67966481, 1070.96955761, 1052.91950889, 1014.41273828, 1016.81941144, 995.15935297, 998.76936272, 1033.66612358, 987.93933348, 1040.88614307, 1031.25945042, 1038.47946991, 1048.10616256, 1091.42627950, 1076.98624052, 1075.78290394, 1066.15621129, 1058.93619180, 1040.88614307, 1003.58270904, 1015.61607486, 1015.61607486, 972.29595792, 814.65886573, 812.25219257, 824.28555839, 723.20528553, 762.91539273, 713.57859288, 842.33560711, 749.67869033, 770.13541222, 836.31892420, 950.63589945, 937.39919706, 900.09576303, 1070.96955761, 1004.78604563, 904.90910935, 960.26259211, 998.76936272, 1069.76622103, 1085.40959659, 1028.85277726, 969.88928476, 1022.83609435, 1021.63275777, 1063.74953812, 1001.17603588, 1003.58270904, 1022.83609435, 1044.49615282, 1086.61293317, 1064.95287471, 1103.45964532, 1113.08633797, 1042.08947966, 1044.49615282, 1004.78604563, 1064.95287471, 1058.93619180, 1019.22608460, 1058.93619180, 1093.83295266, 1062.54620154, 1089.01960634, 1070.96955761, 1018.02274802, 1004.78604563, 980.71931399, 950.63589945, 967.48261160, 979.51597741, 992.75267981, 1005.98938221, 968.68594818, 969.88928476, 960.26259211, 910.92579226, 922.95915808, 932.58585073, 955.44924578, 986.73599690, 989.14267007, 996.36268955, 992.75267981, 990.34600665, 934.99252389, 992.75267981, 973.49929450, 984.32932374, 950.63589945, 902.50243619, 862.79232900, 871.21568507, 844.74228027, 844.74228027, 826.69223155, 896.48575328, 972.29595792, 956.65258236, 951.83923604, 938.60253364, 904.90910935, 882.04571430, 906.11244594, 889.26573379, 849.55562660, 900.09576303, 849.55562660, 936.19586048, 948.22922629, 980.71931399, 993.95601639, 984.32932374, 932.58585073, 1004.78604563, 974.70263109, 1019.22608460, 989.14267007, 1036.07279675, 980.71931399, 979.51597741, 1034.86946017, 993.95601639, 1010.80272853, 944.61921655, 986.73599690, 968.68594818, 987.93933348, 1003.58270904, 1034.86946017, 1068.56288445, 1066.15621129, 1074.57956736, 1043.29281624, 1074.57956736, 1064.95287471, 1087.81626976, 1105.86631848, 1089.01960634, 1103.45964532, 1111.88300139, 1090.22294292, 1060.13952838, 1109.47632822, 1116.69634771, 1143.16975251, 1158.81312807, 1166.03314756, 1122.71303062, 1161.21980123, 1138.35640618, 1108.27299164, 1139.55974276, 1167.23648414, 1186.48986945, 1182.87985970, 1191.30321577, 1174.45650363, 1187.69320603, 1186.48986945, 1167.23648414, 1126.32304037, 1156.40645491, 1127.52637695, 1105.86631848, 1101.05297215, 1126.32304037, 1172.04983047, 1115.49301113, 1103.45964532, 1157.60979149, 1187.69320603, 1173.25316705, 1160.01646465, 1147.98309883, 1158.81312807, 1135.94973302, 1137.15306960, 1158.81312807, 1157.60979149, 1178.06651337, 1172.04983047, 1129.93305011, 1153.99978174, 1084.20626001, 1030.05611384, 1030.05611384, 987.93933348, 936.19586048, 927.77250440, 930.17917757, 868.80901191, 908.51911910, 926.56916782, 948.22922629, 950.63589945, 947.02588971, 1014.41273828, 933.78918731, 862.79232900, 885.65572405, 902.50243619, 867.60567532, 898.89242645, 925.36583124, 1003.58270904, 977.10930425, 962.66926527, 1010.80272853, 1002.37937246, 1024.03943093, 1028.85277726, 956.65258236, 951.83923604, 1005.98938221, 997.56602614, 926.56916782, 882.04571430, 830.30224130, 794.20214385, 782.16877803, 790.59213410, 770.13541222, 776.15209512], 
							[2539.04018715, 2575.14028460, 2442.77326062, 2406.67316318, 2322.43960246, 1805.00487238, 1612.47101933, 1720.77131167, 1624.50438514, 1672.63784841, 1732.80467749, 1600.43765351, 1792.97150657, 1805.00487238, 1732.80467749, 1660.60448259, 1516.20409280, 1504.17072698, 1660.60448259, 1684.67121422, 1648.57111678, 1528.23745862, 1648.57111678, 1636.53775096, 1648.57111678, 1949.40526217, 2141.93911523, 2358.53969991, 2202.10594431, 2141.93911523, 2262.27277339, 2105.83901778, 2105.83901778, 2153.97248104, 2166.00584686, 2141.93911523, 2190.07257849, 2129.90574941, 2141.93911523, 2153.97248104, 2153.97248104, 2190.07257849, 2214.13931012, 2262.27277339, 2166.00584686, 2105.83901778, 2045.67218870, 2081.77228615, 2141.93911523, 2117.87238359, 2166.00584686, 2129.90574941, 2117.87238359, 2117.87238359, 2382.60643154, 2502.94008970, 2430.73989481, 2418.70652899, 2478.87335807, 2394.63979736, 2454.80662644, 2502.94008970, 2502.94008970, 2454.80662644, 2478.87335807, 2719.54067439, 2791.74086928, 2779.70750347, 2683.44057694, 2803.77423510, 2863.94106418, 2948.17462489, 2996.30808815, 2960.20799071, 3020.37481979, 3020.37481979, 2984.27472234, 2924.10789326, 2863.94106418, 2984.27472234, 2948.17462489, 2827.84096673, 2635.30711368, 2755.64077184, 2767.67413765, 2803.77423510, 2875.97442999, 2875.97442999, 2863.94106418, 2743.60740602, 2707.50730857, 2418.70652899, 2358.53969991, 2322.43960246, 2370.57306573, 2298.37287083, 2382.60643154, 2358.53969991, 2406.67316318, 2575.14028460, 2442.77326062, 2695.47394276, 2707.50730857, 2599.20701623, 2743.60740602, 2755.64077184, 2707.50730857, 2743.60740602, 2743.60740602, 2599.20701623, 2779.70750347, 2791.74086928, 2743.60740602, 2695.47394276, 2767.67413765, 2827.84096673, 2839.87433255, 2936.14125907, 2803.77423510, 2803.77423510, 2803.77423510, 2707.50730857, 2719.54067439, 2719.54067439, 2755.64077184, 2695.47394276, 2719.54067439, 2767.67413765, 2779.70750347, 2743.60740602, 2719.54067439, 2683.44057694, 2490.90672389, 2490.90672389, 2539.04018715, 2527.00682133, 2647.34047949, 2599.20701623, 2599.20701623, 2551.07355297, 2563.10691878, 2430.73989481, 2382.60643154, 2298.37287083, 2466.83999225, 2442.77326062, 2575.14028460, 2623.27374786, 2647.34047949, 2611.24038205, 2587.17365041, 2454.80662644, 2298.37287083, 2575.14028460, 2575.14028460, 2514.97345552, 2502.94008970, 2478.87335807, 2370.57306573, 2214.13931012, 2033.63882288, 2226.17267594, 2129.90574941, 2166.00584686, 2346.50633410, 2430.73989481, 2514.97345552, 2551.07355297, 2370.57306573, 2358.53969991, 2406.67316318, 2334.47296828, 2226.17267594, 2238.20604175, 2057.70555452, 2298.37287083, 2370.57306573, 2442.77326062, 2539.04018715, 2454.80662644, 2490.90672389, 2599.20701623, 2502.94008970, 2466.83999225, 2490.90672389, 2527.00682133, 2502.94008970, 2539.04018715, 2527.00682133, 2262.27277339, 2382.60643154, 2202.10594431, 2346.50633410, 2322.43960246, 2430.73989481, 2430.73989481, 2551.07355297, 2659.37384531, 2755.64077184, 2875.97442999, 2743.60740602, 2875.97442999, 2815.80760092, 2791.74086928, 2851.90769836, 2851.90769836, 2851.90769836, 2791.74086928, 2683.44057694, 2551.07355297, 2527.00682133, 2755.64077184, 2719.54067439, 2803.77423510, 2779.70750347, 2731.57404020, 2599.20701623, 2767.67413765, 2731.57404020, 2671.40721112, 2779.70750347, 2755.64077184, 2827.84096673, 2827.84096673, 2851.90769836, 2888.00779581, 2839.87433255, 2779.70750347, 2635.30711368, 2599.20701623, 2539.04018715, 2490.90672389, 2418.70652899, 2382.60643154, 2527.00682133, 2406.67316318, 2454.80662644, 2502.94008970, 2539.04018715, 2623.27374786, 2539.04018715, 2466.83999225, 2575.14028460, 2466.83999225, 2466.83999225, 2478.87335807, 2466.83999225, 2430.73989481, 2382.60643154, 2190.07257849, 2117.87238359, 2105.83901778, 1997.53872544, 2033.63882288, 1696.70458004, 1768.90477493, 1564.33755606, 1528.23745862, 1456.03726372, 1516.20409280, 1792.97150657, 1696.70458004, 1636.53775096, 1684.67121422, 1877.20506728, 1732.80467749, 1672.63784841, 1480.10399535, 1780.93814075, 1648.57111678, 1780.93814075, 1756.87140912, 1841.10496983, 1829.07160401, 1805.00487238, 1877.20506728, 1997.53872544, 1985.50535962, 2093.80565196, 1937.37189636, 1853.13833565, 1985.50535962, 1997.53872544, 1780.93814075, 1877.20506728, 1889.23843309, 1841.10496983, 1768.90477493, 1780.93814075, 1684.67121422, 1636.53775096], 
							[1756.87140912, 1576.37092188, 1419.93716627, 1431.97053209, 1492.13736117, 1456.03726372, 1552.30419025, 1624.50438514, 1576.37092188, 1552.30419025, 1684.67121422, 1696.70458004, 1708.73794585, 1732.80467749, 1696.70458004, 1720.77131167, 1528.23745862, 1492.13736117, 1576.37092188, 1624.50438514, 1576.37092188, 1480.10399535, 1576.37092188, 1540.27082443, 1480.10399535, 1708.73794585, 1865.17170146, 2129.90574941, 1961.43862799, 1961.43862799, 1949.40526217, 1841.10496983, 1841.10496983, 1877.20506728, 1792.97150657, 1732.80467749, 1648.57111678, 1588.40428770, 1708.73794585, 1744.83804330, 1768.90477493, 1817.03823820, 1877.20506728, 1853.13833565, 1792.97150657, 1780.93814075, 1756.87140912, 1732.80467749, 1805.00487238, 1853.13833565, 1768.90477493, 1829.07160401, 1805.00487238, 1720.77131167, 2009.57209125, 2274.30613920, 2322.43960246, 2635.30711368, 2611.24038205, 2502.94008970, 2394.63979736, 2683.44057694, 2647.34047949, 2539.04018715, 2551.07355297, 2635.30711368, 2671.40721112, 2647.34047949, 2418.70652899, 2382.60643154, 2346.50633410, 2370.57306573, 2502.94008970, 2466.83999225, 2527.00682133, 2527.00682133, 2671.40721112, 2587.17365041, 2611.24038205, 2647.34047949, 2599.20701623, 2478.87335807, 2370.57306573, 2358.53969991, 2478.87335807, 2406.67316318, 2406.67316318, 2394.63979736, 2310.40623665, 2322.43960246, 2250.23940757, 2129.90574941, 2226.17267594, 2226.17267594, 2262.27277339, 2178.03921267, 2153.97248104, 2334.47296828, 2466.83999225, 2551.07355297, 2490.90672389, 2731.57404020, 2707.50730857, 2635.30711368, 2791.74086928, 2755.64077184, 2803.77423510, 2863.94106418, 2743.60740602, 2683.44057694, 2719.54067439, 2803.77423510, 2779.70750347, 2659.37384531, 2683.44057694, 2755.64077184, 2767.67413765, 2888.00779581, 2948.17462489, 3008.34145397, 2996.30808815, 2924.10789326, 2863.94106418, 2863.94106418, 2875.97442999, 2791.74086928, 2659.37384531, 2707.50730857, 2707.50730857, 2743.60740602, 2647.34047949, 2671.40721112, 2563.10691878, 2539.04018715, 2563.10691878, 2466.83999225, 2406.67316318, 2478.87335807, 2394.63979736, 2382.60643154, 2406.67316318, 2250.23940757, 2322.43960246, 2286.33950502, 2394.63979736, 2394.63979736, 2394.63979736, 2490.90672389, 2490.90672389, 2406.67316318, 2298.37287083, 2214.13931012, 2141.93911523, 2214.13931012, 2141.93911523, 2141.93911523, 2057.70555452, 2045.67218870, 1997.53872544, 1889.23843309, 1805.00487238, 1949.40526217, 1937.37189636, 1949.40526217, 2069.73892033, 2045.67218870, 2021.60545707, 1973.47199380, 1913.30516472, 1889.23843309, 1913.30516472, 1817.03823820, 1829.07160401, 1865.17170146, 1780.93814075, 1973.47199380, 2081.77228615, 2069.73892033, 2045.67218870, 2105.83901778, 2117.87238359, 2370.57306573, 2274.30613920, 2322.43960246, 2382.60643154, 2430.73989481, 2238.20604175, 2334.47296828, 2382.60643154, 2250.23940757, 2334.47296828, 2226.17267594, 2334.47296828, 2226.17267594, 2250.23940757, 2334.47296828, 2430.73989481, 2551.07355297, 2587.17365041, 2695.47394276, 2707.50730857, 2888.00779581, 2731.57404020, 2755.64077184, 2659.37384531, 2791.74086928, 2863.94106418, 2791.74086928, 2659.37384531, 2599.20701623, 2647.34047949, 2659.37384531, 2695.47394276, 2827.84096673, 2827.84096673, 2815.80760092, 2719.54067439, 2791.74086928, 2695.47394276, 2635.30711368, 2647.34047949, 2527.00682133, 2454.80662644, 2466.83999225, 2647.34047949, 2551.07355297, 2575.14028460, 2527.00682133, 2394.63979736, 2346.50633410, 2274.30613920, 2262.27277339, 2298.37287083, 2382.60643154, 2406.67316318, 2274.30613920, 2310.40623665, 2418.70652899, 2322.43960246, 2430.73989481, 2358.53969991, 2298.37287083, 2358.53969991, 2262.27277339, 2274.30613920, 2262.27277339, 2250.23940757, 2298.37287083, 2310.40623665, 2178.03921267, 2202.10594431, 2141.93911523, 2093.80565196, 2093.80565196, 1792.97150657, 1925.33853054, 1829.07160401, 1865.17170146, 1805.00487238, 1925.33853054, 2033.63882288, 1937.37189636, 1841.10496983, 1660.60448259, 1768.90477493, 1648.57111678, 1600.43765351, 1431.97053209, 1564.33755606, 1672.63784841, 1780.93814075, 1829.07160401, 1708.73794585, 1696.70458004, 1660.60448259, 1636.53775096, 1744.83804330, 1720.77131167, 1889.23843309, 1829.07160401, 1720.77131167, 1841.10496983, 1865.17170146, 2021.60545707, 2081.77228615, 1997.53872544, 1961.43862799, 1925.33853054, 1925.33853054, 1805.00487238, 1732.80467749], 
							[836.31892420, 792.99880727, 740.05199768, 744.86534400, 755.69537324, 703.95190023, 713.57859288, 722.00194895, 747.27201717, 738.84866109, 776.15209512, 770.13541222, 803.82883650, 817.06553890, 806.23550966, 812.25219257, 792.99880727, 752.08536349, 797.81215359, 772.54208538, 743.66200742, 707.56190997, 681.08850518, 673.86848569, 647.39508089, 685.90185151, 754.49203666, 778.55876829, 740.05199768, 729.22196844, 759.30538298, 728.01863186, 729.22196844, 756.89870982, 715.98526604, 691.91853441, 697.93521732, 685.90185151, 701.54522707, 709.96858314, 720.79861237, 718.39193921, 722.00194895, 730.42530502, 715.98526604, 712.37525630, 697.93521732, 687.10518809, 707.56190997, 690.71519783, 683.49517834, 677.47849543, 700.34189048, 660.63178329, 702.74856365, 722.00194895, 722.00194895, 746.06868058, 829.09890471, 806.23550966, 797.81215359, 859.18231925, 860.38565584, 824.28555839, 866.40233874, 848.35229002, 819.47221206, 818.26887548, 796.60881701, 830.30224130, 806.23550966, 794.20214385, 830.30224130, 823.08222181, 844.74228027, 842.33560711, 842.33560711, 824.28555839, 845.94561686, 847.14895344, 857.97898267, 843.53894369, 843.53894369, 824.28555839, 831.50557788, 826.69223155, 870.01234849, 817.06553890, 812.25219257, 776.15209512, 770.13541222, 728.01863186, 738.84866109, 728.01863186, 776.15209512, 715.98526604, 722.00194895, 779.76210487, 811.04885599, 814.65886573, 817.06553890, 845.94561686, 856.77564609, 841.13227053, 832.70891446, 842.33560711, 831.50557788, 871.21568507, 890.46907037, 823.08222181, 886.85906063, 896.48575328, 898.89242645, 908.51911910, 906.11244594, 919.34914833, 925.36583124, 950.63589945, 930.17917757, 934.99252389, 938.60253364, 898.89242645, 889.26573379, 896.48575328, 922.95915808, 922.95915808, 938.60253364, 933.78918731, 938.60253364, 963.87260185, 936.19586048, 961.46592869, 947.02588971, 944.61921655, 960.26259211, 941.00920680, 931.38251415, 943.41587996, 925.36583124, 931.38251415, 925.36583124, 907.31578252, 902.50243619, 882.04571430, 862.79232900, 855.57230951, 850.75896318, 901.29909961, 919.34914833, 906.11244594, 882.04571430, 835.11558762, 833.91225104, 843.53894369, 831.50557788, 820.67554864, 803.82883650, 759.30538298, 717.18860263, 722.00194895, 667.85180278, 755.69537324, 689.51186125, 737.64532451, 768.93207563, 784.57545120, 784.57545120, 756.89870982, 740.05199768, 734.03531477, 711.17191972, 681.08850518, 712.37525630, 709.96858314, 670.25847594, 758.10204640, 746.06868058, 743.66200742, 732.83197819, 724.40862212, 729.22196844, 778.55876829, 779.76210487, 800.21882676, 809.84551941, 830.30224130, 825.48889497, 824.28555839, 848.35229002, 836.31892420, 878.43570456, 843.53894369, 877.23236798, 874.82569481, 898.89242645, 910.92579226, 906.11244594, 918.14581175, 914.53580201, 944.61921655, 950.63589945, 938.60253364, 932.58585073, 949.43256287, 950.63589945, 943.41587996, 945.82255313, 942.21254338, 932.58585073, 902.50243619, 912.12912884, 909.72245568, 914.53580201, 953.04257262, 990.34600665, 986.73599690, 981.92265058, 962.66926527, 945.82255313, 922.95915808, 937.39919706, 938.60253364, 943.41587996, 948.22922629, 957.85591894, 950.63589945, 941.00920680, 931.38251415, 912.12912884, 889.26573379, 868.80901191, 883.24905089, 878.43570456, 889.26573379, 915.73913859, 878.43570456, 863.99566558, 865.19900216, 843.53894369, 866.40233874, 866.40233874, 872.41902165, 901.29909961, 867.60567532, 848.35229002, 882.04571430, 868.80901191, 880.84237772, 900.09576303, 848.35229002, 865.19900216, 786.98212436, 772.54208538, 789.38879752, 661.83511987, 755.69537324, 760.50871956, 755.69537324, 753.28870007, 773.74542196, 758.10204640, 785.77878778, 746.06868058, 705.15523681, 762.91539273, 774.94875854, 771.33874880, 797.81215359, 807.43884625, 860.38565584, 884.45238747, 850.75896318, 844.74228027, 841.13227053, 779.76210487, 812.25219257, 812.25219257, 802.62549992, 812.25219257, 826.69223155, 782.16877803, 728.01863186, 758.10204640, 741.25533426, 717.18860263, 728.01863186, 779.76210487, 755.69537324, 747.27201717, 722.00194895, 670.25847594], 
							[3824.51652380, 3586.23181391, 3359.84810282, 3526.65461976, 3455.16439344, 3312.19597419, 3121.56339293, 3204.96665140, 3216.87968356, 3228.79271572, 3300.28294203, 3419.42529697, 3336.02203850, 3324.10900634, 3240.70574787, 3169.22755493, 3061.99823214, 2978.59497367, 3145.40149061, 3014.33407015, 2990.50800583, 2907.10474736, 2930.94284504, 2787.96239242, 2704.55913395, 3097.73732862, 3467.07742560, 3455.16439344, 3312.19597419, 3169.22755493, 3359.84810282, 3252.61878003, 3181.14058709, 3145.40149061, 3014.33407015, 2978.59497367, 2978.59497367, 2954.76890936, 2990.50800583, 3050.08519998, 3014.33407015, 3061.99823214, 3133.47642509, 3121.56339293, 3097.73732862, 3050.08519998, 2954.76890936, 2930.94284504, 3038.16013446, 3050.08519998, 3073.91126430, 3061.99823214, 3097.73732862, 3264.54384556, 3336.02203850, 3371.77316834, 3479.00249113, 4968.29997796, 4837.23255750, 4682.35110608, 4479.80549267, 4610.86087977, 4539.38268682, 4491.71852482, 4503.63155698, 4455.97942835, 4467.89246051, 4515.54458914, 4372.57616988, 4229.60775062, 4170.03055647, 4312.99897573, 4348.75010557, 4265.34684710, 4396.40223420, 4324.92404125, 4479.80549267, 4336.83707341, 4301.08594357, 4253.43381494, 4229.60775062, 4086.62729800, 4074.71426584, 4003.23607289, 4229.60775062, 4217.69471847, 4241.52078278, 4110.46539568, 3979.39797521, 3967.48494306, 3967.48494306, 3610.05787823, 3776.85236180, 3860.25562027, 3895.99471674, 3693.46113670, 3669.62303902, 3764.93932964, 3895.99471674, 3991.31100737, 3836.42955595, 4158.11752431, 4050.88820153, 4038.97516937, 4050.88820153, 4074.71426584, 3979.39797521, 4015.14910505, 3824.51652380, 3800.69045948, 3848.34258811, 3836.42955595, 3872.16865243, 3848.34258811, 3872.16865243, 3860.25562027, 3919.83281442, 3848.34258811, 3884.08168459, 3800.69045948, 3788.76539396, 3693.46113670, 3776.85236180, 3669.62303902, 3669.62303902, 3574.30674839, 3610.05787823, 3574.30674839, 3610.05787823, 3586.23181391, 3610.05787823, 3610.05787823, 3526.65461976, 3538.56765191, 3514.74158760, 3479.00249113, 3395.59923266, 3371.77316834, 3371.77316834, 3347.94710403, 3347.94710403, 3312.19597419, 3336.02203850, 3288.36990987, 3312.19597419, 3300.28294203, 3324.10900634, 3336.02203850, 3312.19597419, 3288.36990987, 3228.79271572, 3145.40149061, 3061.99823214, 3252.61878003, 3157.31452277, 3181.14058709, 3097.73732862, 3109.65036077, 3061.99823214, 3097.73732862, 3050.08519998, 3240.70574787, 3193.05361924, 3216.87968356, 3312.19597419, 3204.96665140, 3240.70574787, 3193.05361924, 3240.70574787, 3216.87968356, 3204.96665140, 3157.31452277, 3073.91126430, 3061.99823214, 3097.73732862, 3240.70574787, 3061.99823214, 3109.65036077, 3109.65036077, 3038.16013446, 3073.91126430, 3157.31452277, 3097.73732862, 3073.91126430, 3193.05361924, 3383.68620050, 3264.54384556, 3261.04213610, 3285.10886773, 3261.04213610, 3273.07550192, 3212.90867284, 3224.94203866, 3176.80857539, 3164.77520958, 3152.74184376, 3212.90867284, 3188.84194121, 3188.84194121, 3285.10886773, 3321.20896518, 3297.14223355, 3249.00877029, 3297.14223355, 3236.97540447, 3261.04213610, 3309.17559937, 3297.14223355, 3285.10886773, 3297.14223355, 3261.04213610, 3261.04213610, 3285.10886773, 3297.14223355, 3345.27569681, 3321.20896518, 3405.44252589, 3417.47589171, 3393.40916008, 3357.30906263, 3393.40916008, 3405.44252589, 3333.24233100, 3357.30906263, 3321.20896518, 3369.34242845, 3345.27569681, 3297.14223355, 3261.04213610, 3249.00877029, 3249.00877029, 3309.17559937, 3249.00877029, 3261.04213610, 3212.90867284, 3188.84194121, 3188.84194121, 3176.80857539, 3236.97540447, 3249.00877029, 3152.74184376, 3128.67511213, 3140.70847794, 3116.64174631, 3116.64174631, 3104.60838050, 3080.54164886, 3092.57501468, 3092.57501468, 3068.50828305, 2984.27472234, 3068.50828305, 3032.40818560, 3056.47491723, 2767.67413765, 2960.20799071, 2863.94106418, 2815.80760092, 2659.37384531, 2587.17365041, 2623.27374786, 2695.47394276, 2683.44057694, 2707.50730857, 2743.60740602, 2683.44057694, 2719.54067439, 2527.00682133, 2647.34047949, 2683.44057694, 2731.57404020, 2707.50730857, 2647.34047949, 2755.64077184, 2671.40721112, 2743.60740602, 2888.00779581, 2912.07452744, 2875.97442999, 2779.70750347, 2719.54067439, 2707.50730857, 2719.54067439, 2695.47394276, 2731.57404020, 2731.57404020, 2707.50730857, 2767.67413765, 2731.57404020, 2791.74086928, 2755.64077184], 
							[1517.63606333, 1517.63606333, 1496.84240720, 1538.42971946, 1495.31416974, 1495.31416974, 1484.39990695, 1222.44556650, 1189.70277812, 1113.30293855, 1233.35982930, 1266.10261768, 1222.44556650, 1211.53130371, 1178.78851532, 1167.87425253, 1071.82392659, 987.78289973, 1024.89379990, 1091.47441296, 1091.47441296, 994.32905073, 1036.90309899, 1004.14827724, 937.57969754, 1004.14827724, 1146.04572694, 1146.04572694, 1146.04572694, 1211.53130371, 1189.70277812, 1211.53130371, 1156.95998973, 1156.95998973, 1146.04572694, 1124.21720135, 1156.95998973, 1069.64588737, 1156.95998973, 1222.44556650, 1200.61704091, 1222.44556650, 1255.18835489, 1244.27409209, 1146.04572694, 1135.13146414, 1167.87425253, 1146.04572694, 1255.18835489, 1277.01688048, 1211.53130371, 1211.53130371, 1222.44556650, 1266.10261768, 1331.60022782, 1375.25727900, 1375.25727900, 1260.65150297, 1317.95439098, 1386.71304326, 1375.25727900, 1444.01593127, 1409.63660514, 1375.25727900, 1363.78948138, 1340.87795286, 1340.87795286, 1283.57506485, 1306.48659336, 1283.57506485, 1306.48659336, 1295.03082910, 1317.95439098, 1295.03082910, 1340.87795286, 1352.33371712, 1340.87795286, 1409.63660514, 1432.56016701, 1432.56016701, 1432.56016701, 1375.25727900, 1386.71304326, 1432.56016701, 1398.16880751, 1409.63660514, 1409.63660514, 1386.71304326, 1272.10726723, 1283.57506485, 1249.18370535, 1203.34861495, 1237.72794109, 1249.18370535, 1306.48659336, 1260.65150297, 1272.10726723, 1283.57506485, 1352.33371712, 1352.33371712, 1295.03082910, 1409.63660514, 1478.39525741, 1295.03082910, 1398.16880751, 1409.63660514, 1444.01593127, 1455.47169553, 1329.41015524, 1306.48659336, 1386.71304326, 1375.25727900, 1375.25727900, 1317.95439098, 1352.33371712, 1386.71304326, 1386.71304326, 1455.47169553, 1409.63660514, 1398.16880751, 1444.01593127, 1375.25727900, 1375.25727900, 1340.87795286, 1398.16880751, 1363.78948138, 1398.16880751, 1375.25727900, 1421.09236939, 1375.25727900, 1375.25727900, 1363.78948138, 1329.41015524, 1317.95439098, 1283.57506485, 1260.65150297, 1260.65150297, 1306.48659336, 1317.95439098, 1317.95439098, 1306.48659336, 1317.95439098, 1295.03082910, 1249.18370535, 1237.72794109, 1249.18370535, 1306.48659336, 1352.33371712, 1375.25727900, 1363.78948138, 1409.63660514, 1398.16880751, 1295.03082910, 1363.78948138, 1317.95439098, 1260.65150297, 1237.72794109, 1136.87630219, 1087.59966917, 1123.12216506, 1111.66640080, 1088.74283892, 1014.24427116, 1060.09139492, 1143.74735407, 1157.50149120, 1191.88081733, 1168.96928882, 1146.04572694, 1110.51119768, 1146.04572694, 1123.12216506, 1113.95274031, 1140.31784481, 1033.72629041, 1168.96928882, 1191.88081733, 1139.16264169, 1180.42505307, 1168.96928882, 1168.96928882, 1237.72794109, 1226.27217683, 1226.27217683, 1295.03082910, 1283.57506485, 1203.34861495, 1191.88081733, 1260.65150297, 1191.88081733, 1226.27217683, 1180.42505307, 1214.80437921, 1203.34861495, 1272.10726723, 1306.48659336, 1317.95439098, 1340.87795286, 1352.33371712, 1340.87795286, 1329.41015524, 1363.78948138, 1363.78948138, 1398.16880751, 1466.93949315, 1444.01593127, 1432.56016701, 1466.93949315, 1421.09236939, 1398.16880751, 1432.56016701, 1466.93949315, 1593.00103344, 1504.17072698, 1624.50438514, 1624.50438514, 1684.67121422, 1612.47101933, 1588.40428770, 1588.40428770, 1660.60448259, 1648.57111678, 1648.57111678, 1648.57111678, 1672.63784841, 1684.67121422, 1648.57111678, 1648.57111678, 1588.40428770, 1612.47101933, 1552.30419025, 1504.17072698, 1516.20409280, 1528.23745862, 1588.40428770, 1456.03726372, 1456.03726372, 1552.30419025, 1600.43765351, 1660.60448259, 1576.37092188, 1588.40428770, 1648.57111678, 1576.37092188, 1576.37092188, 1648.57111678, 1636.53775096, 1648.57111678, 1612.47101933, 1576.37092188, 1648.57111678, 1576.37092188, 1468.07062954, 1444.00389791, 1191.30321577, 1299.60350811, 1215.36994740, 1275.53677648, 1215.36994740, 1239.43667904, 1263.50341067, 1227.40331322, 1215.36994740, 1197.31989868, 1311.63687393, 1227.40331322, 1155.20311832, 1215.36994740, 1227.40331322, 1227.40331322, 1287.57014230, 1299.60350811, 1407.90380046, 1347.73697138, 1383.83706883, 1395.87043464, 1407.90380046, 1395.87043464, 1371.80370301, 1299.60350811, 1263.50341067, 1275.53677648, 1299.60350811, 1191.30321577, 1167.23648414, 1126.32304037, 1083.00292343, 1072.17289419, 1113.08633797, 1099.84963557, 1095.03628924], 
							[1648.57111678, 1660.60448259, 1564.33755606, 1612.47101933, 1504.17072698, 1191.30321577, 1046.90282598, 1016.81941144, 1064.95287471, 968.68594818, 998.76936272, 974.70263109, 998.76936272, 1039.68280649, 942.21254338, 843.53894369, 818.26887548, 720.79861237, 770.13541222, 796.60881701, 788.18546094, 744.86534400, 776.15209512, 794.20214385, 806.23550966, 989.14267007, 1084.20626001, 1263.50341067, 1263.50341067, 1227.40331322, 1215.36994740, 1072.17289419, 1012.00606512, 1042.08947966, 950.63589945, 938.60253364, 1034.86946017, 985.53266032, 938.60253364, 974.70263109, 971.09262134, 983.12598716, 984.32932374, 997.56602614, 890.46907037, 889.26573379, 880.84237772, 855.57230951, 896.48575328, 914.53580201, 932.58585073, 972.29595792, 914.53580201, 872.41902165, 888.06239721, 1141.96641593, 1083.00292343, 1143.16975251, 1168.43982072, 1156.40645491, 1164.82981098, 1215.36994740, 1215.36994740, 1239.43667904, 1203.33658159, 1311.63687393, 1311.63687393, 1287.57014230, 1215.36994740, 1215.36994740, 1172.04983047, 1197.31989868, 1287.57014230, 1275.53677648, 1419.93716627, 1516.20409280, 1540.27082443, 1468.07062954, 1431.97053209, 1600.43765351, 1516.20409280, 1540.27082443, 1468.07062954, 1480.10399535, 1540.27082443, 1419.93716627, 1468.07062954, 1383.83706883, 1239.43667904, 1311.63687393, 1299.60350811, 1063.74953812, 1110.67966481, 1102.25630873, 1089.01960634, 1032.46278700, 1046.90282598, 1070.96955761, 1160.01646465, 1371.80370301, 1299.60350811, 1395.87043464, 1347.73697138, 1263.50341067, 1299.60350811, 1444.00389791, 1431.97053209, 1492.13736117, 1371.80370301, 1323.67023975, 1395.87043464, 1395.87043464, 1323.67023975, 1323.67023975, 1323.67023975, 1468.07062954, 1431.97053209, 1516.20409280, 1552.30419025, 1600.43765351, 1528.23745862, 1492.13736117, 1444.00389791, 1407.90380046, 1468.07062954, 1383.83706883, 1419.93716627, 1456.03726372, 1456.03726372, 1407.90380046, 1359.77033719, 1371.80370301, 1311.63687393, 1287.57014230, 1287.57014230, 1323.67023975, 1395.87043464, 1419.93716627, 1371.80370301, 1323.67023975, 1323.67023975, 1215.36994740, 1203.33658159, 1144.37308909, 1194.91322552, 1185.28653286, 1215.36994740, 1263.50341067, 1251.47004485, 1239.43667904, 1192.50655235, 1089.01960634, 1058.93619180, 1155.20311832, 1131.13638669, 1125.11970378, 1058.93619180, 1009.59939195, 997.56602614, 922.95915808, 924.16249466, 1004.78604563, 953.04257262, 993.95601639, 1140.76307935, 1133.54305986, 1149.18643542, 1140.76307935, 1038.47946991, 1020.42942119, 1040.88614307, 1040.88614307, 1040.88614307, 1022.83609435, 938.60253364, 1045.69948940, 1081.79958685, 1105.86631848, 1166.03314756, 1163.62647440, 1161.21980123, 1287.57014230, 1215.36994740, 1192.50655235, 1203.33658159, 1198.52323526, 1096.23962583, 1113.08633797, 1115.49301113, 1056.52951863, 1028.85277726, 981.92265058, 1044.49615282, 1048.10616256, 1105.86631848, 1076.98624052, 1162.42313781, 1275.53677648, 1275.53677648, 1456.03726372, 1359.77033719, 1407.90380046, 1383.83706883, 1323.67023975, 1371.80370301, 1359.77033719, 1383.83706883, 1395.87043464, 1323.67023975, 1383.83706883, 1371.80370301, 1371.80370301, 1444.00389791, 1444.00389791, 1504.17072698, 1468.07062954, 1431.97053209, 1504.17072698, 1468.07062954, 1407.90380046, 1456.03726372, 1407.90380046, 1431.97053209, 1431.97053209, 1468.07062954, 1492.13736117, 1540.27082443, 1540.27082443, 1468.07062954, 1504.17072698, 1456.03726372, 1371.80370301, 1311.63687393, 1299.60350811, 1359.77033719, 1311.63687393, 1275.53677648, 1335.70360556, 1347.73697138, 1431.97053209, 1383.83706883, 1359.77033719, 1419.93716627, 1323.67023975, 1323.67023975, 1371.80370301, 1347.73697138, 1347.73697138, 1311.63687393, 1191.30321577, 1113.08633797, 1073.37623078, 1072.17289419, 1089.01960634, 836.31892420, 974.70263109, 889.26573379, 932.58585073, 853.16563635, 856.77564609, 1022.83609435, 897.68908986, 872.41902165, 907.31578252, 965.07593843, 877.23236798, 805.03217308, 690.71519783, 796.60881701, 843.53894369, 916.94247517, 974.70263109, 1016.81941144, 991.54934323, 1054.12284547, 1039.68280649, 1058.93619180, 1044.49615282, 1048.10616256, 972.29595792, 872.41902165, 980.71931399, 918.14581175, 844.74228027, 818.26887548, 874.82569481, 833.91225104, 833.91225104, 752.08536349, 720.79861237, 718.39193921], 
							[818.26887548, 756.89870982, 728.01863186, 655.81843697, 559.55151044, 529.46809590, 522.24807641, 581.21156891, 600.46495421, 631.75170533, 619.71833952, 601.66829079, 616.10832977, 601.66829079, 589.63492498, 575.19488600, 553.53482753, 535.48477881, 578.80489574, 565.56819335, 548.72148120, 506.60470085, 515.02805692, 487.35131554, 436.81117912, 498.18134478, 566.77152993, 592.04159814, 565.56819335, 582.41490549, 613.70165661, 601.66829079, 580.00823233, 576.39822258, 581.21156891, 528.26475932, 523.45141299, 487.35131554, 510.21471059, 524.65474957, 525.85808615, 515.02805692, 529.46809590, 513.82472034, 494.57133503, 525.85808615, 530.67143248, 493.36799845, 516.23139350, 517.43473008, 512.62138376, 501.79135452, 483.74130580, 456.06456442, 496.97800820, 541.50146171, 540.29812513, 517.43473008, 605.27830054, 614.90499319, 578.80489574, 623.32834926, 649.80175406, 679.88516860, 693.12187099, 720.79861237, 748.47535375, 752.08536349, 682.29184176, 726.81529528, 646.19174431, 671.46181253, 717.18860263, 679.88516860, 708.76524656, 705.15523681, 713.57859288, 684.69851492, 685.90185151, 673.86848569, 687.10518809, 685.90185151, 664.24179304, 657.02177355, 673.86848569, 676.27515885, 669.05513936, 634.15837850, 584.82157865, 555.94150069, 547.51814462, 524.65474957, 547.51814462, 535.48477881, 613.70165661, 560.75484702, 583.61824207, 611.29498345, 625.73502243, 679.88516860, 673.86848569, 734.03531477, 722.00194895, 743.66200742, 814.65886573, 820.67554864, 848.35229002, 874.82569481, 814.65886573, 833.91225104, 823.08222181, 830.30224130, 820.67554864, 818.26887548, 819.47221206, 838.72559737, 883.24905089, 914.53580201, 950.63589945, 1111.88300139, 1081.79958685, 1042.08947966, 1028.85277726, 1004.78604563, 1031.25945042, 1028.85277726, 1028.85277726, 1052.91950889, 1016.81941144, 1007.19271879, 1052.91950889, 1020.42942119, 1010.80272853, 1007.19271879, 999.97269930, 1004.78604563, 1026.44610409, 1013.20940170, 1007.19271879, 1013.20940170, 1014.41273828, 992.75267981, 999.97269930, 910.92579226, 938.60253364, 956.65258236, 959.05925553, 992.75267981, 1010.80272853, 1002.37937246, 998.76936272, 912.12912884, 883.24905089, 806.23550966, 811.04885599, 786.98212436, 705.15523681, 722.00194895, 701.54522707, 705.15523681, 623.32834926, 661.83511987, 684.69851492, 705.15523681, 764.11872931, 749.67869033, 752.08536349, 703.95190023, 673.86848569, 663.03845645, 696.73188074, 648.59841748, 607.68497370, 660.63178329, 654.61510038, 749.67869033, 800.21882676, 802.62549992, 806.23550966, 830.30224130, 882.04571430, 912.12912884, 902.50243619, 925.36583124, 937.39919706, 926.56916782, 870.01234849, 902.50243619, 937.39919706, 903.70577277, 932.58585073, 968.68594818, 1036.07279675, 1028.85277726, 1018.02274802, 986.73599690, 1038.47946991, 1046.90282598, 1026.44610409, 1014.41273828, 1042.08947966, 1070.96955761, 1036.07279675, 1049.30949914, 1072.17289419, 1105.86631848, 1099.84963557, 1095.03628924, 1056.52951863, 1070.96955761, 1069.76622103, 1052.91950889, 1072.17289419, 1101.05297215, 1109.47632822, 1179.26984996, 1161.21980123, 1173.25316705, 1167.23648414, 1150.38977200, 1137.15306960, 1127.52637695, 1081.79958685, 1108.27299164, 1120.30635746, 1119.10302088, 1108.27299164, 1098.64629899, 1085.40959659, 1083.00292343, 1067.35954787, 1067.35954787, 1073.37623078, 1131.13638669, 1132.33972327, 1092.62961608, 1122.71303062, 1133.54305986, 1155.20311832, 1199.72657184, 1187.69320603, 1167.23648414, 1188.89654261, 1153.99978174, 1158.81312807, 1179.26984996, 1193.70988894, 1180.47318654, 1169.64315730, 1160.01646465, 1173.25316705, 1125.11970378, 1158.81312807, 1164.82981098, 1068.56288445, 1104.66298190, 1137.15306960, 1164.82981098, 1182.87985970, 1299.60350811, 1251.47004485, 1200.92990842, 1215.36994740, 1162.42313781, 1173.25316705, 1251.47004485, 1263.50341067, 1263.50341067, 1239.43667904, 1263.50341067, 1287.57014230, 1359.77033719, 1359.77033719, 1371.80370301, 1371.80370301, 1431.97053209, 1492.13736117, 1456.03726372, 1444.00389791, 1359.77033719, 1468.07062954, 1516.20409280, 1588.40428770, 1672.63784841, 1648.57111678, 1576.37092188, 1540.27082443, 1444.00389791, 1419.93716627, 1528.23745862, 1504.17072698], 
							[1719.06257372, 2177.48567785, 2085.80346370, 1948.27412579, 1744.83804330, 1636.53775096, 1552.30419025, 1504.17072698, 1516.20409280, 1456.03726372, 1648.57111678, 1528.23745862, 1540.27082443, 1564.33755606, 1480.10399535, 1456.03726372, 1419.93716627, 1263.50341067, 1383.83706883, 1444.00389791, 1444.00389791, 1335.70360556, 1347.73697138, 1263.50341067, 1014.41273828, 1173.25316705, 1407.90380046, 1456.03726372, 1468.07062954, 1528.23745862, 1516.20409280, 1383.83706883, 1359.77033719, 1383.83706883, 1371.80370301, 1347.73697138, 1371.80370301, 1239.43667904, 1395.87043464, 1419.93716627, 1395.87043464, 1419.93716627, 1419.93716627, 1395.87043464, 1347.73697138, 1311.63687393, 1359.77033719, 1323.67023975, 1395.87043464, 1431.97053209, 1407.90380046, 1431.97053209, 1419.93716627, 1383.83706883, 1444.00389791, 1456.03726372, 1504.17072698, 1480.10399535, 1528.23745862, 1684.67121422, 1576.37092188, 1672.63784841, 1624.50438514, 1588.40428770, 1624.50438514, 1588.40428770, 1588.40428770, 1708.73794585, 1576.37092188, 1528.23745862, 1516.20409280, 1576.37092188, 1612.47101933, 1564.33755606, 1624.50438514, 1588.40428770, 1624.50438514, 1612.47101933, 1612.47101933, 1636.53775096, 1624.50438514, 1588.40428770, 1600.43765351, 1588.40428770, 1528.23745862, 1456.03726372, 1456.03726372, 1395.87043464, 1239.43667904, 1251.47004485, 1251.47004485, 1182.87985970, 1251.47004485, 1215.36994740, 1263.50341067, 1203.33658159, 1227.40331322, 1275.53677648, 1347.73697138, 1335.70360556, 1335.70360556, 1480.10399535, 1456.03726372, 1347.73697138, 1371.80370301, 1444.00389791, 1468.07062954, 1456.03726372, 1395.87043464, 1359.77033719, 1431.97053209, 1395.87043464, 1407.90380046, 1347.73697138, 1335.70360556, 1383.83706883, 1419.93716627, 1444.00389791, 1444.00389791, 1444.00389791, 1468.07062954, 1383.83706883, 1335.70360556, 1347.73697138, 1395.87043464, 1383.83706883, 1395.87043464, 1383.83706883, 1419.93716627, 1347.73697138, 1407.90380046, 1347.73697138, 1299.60350811, 1239.43667904, 1287.57014230, 1287.57014230, 1287.57014230, 1287.57014230, 1287.57014230, 1263.50341067, 1263.50341067, 1263.50341067, 1275.53677648, 1203.33658159, 1263.50341067, 1239.43667904, 1275.53677648, 1323.67023975, 1275.53677648, 1263.50341067, 1299.60350811, 1275.53677648, 1227.40331322, 1299.60350811, 1287.57014230, 1299.60350811, 1251.47004485, 1178.06651337, 1101.05297215, 1104.66298190, 1104.66298190, 1080.59625027, 1040.88614307, 1086.61293317, 1185.28653286, 1169.64315730, 1227.40331322, 1149.18643542, 1121.50969404, 1079.39291368, 1098.64629899, 1003.58270904, 965.07593843, 1055.32618205, 955.44924578, 1123.91636720, 1134.74639644, 1110.67966481, 1158.81312807, 1095.03628924, 1097.44296241, 1161.21980123, 1176.86317679, 1186.48986945, 1188.89654261, 1185.28653286, 1146.77976225, 1131.13638669, 1188.89654261, 1121.50969404, 1144.37308909, 1089.01960634, 1117.89968430, 1129.93305011, 1161.21980123, 1155.20311832, 1167.23648414, 1186.48986945, 1176.86317679, 1158.81312807, 1180.47318654, 1203.33658159, 1251.47004485, 1227.40331322, 1287.57014230, 1275.53677648, 1263.50341067, 1263.50341067, 1227.40331322, 1191.30321577, 1227.40331322, 1263.50341067, 1299.60350811, 1383.83706883, 1359.77033719, 1311.63687393, 1359.77033719, 1371.80370301, 1335.70360556, 1311.63687393, 1371.80370301, 1407.90380046, 1395.87043464, 1383.83706883, 1419.93716627, 1444.00389791, 1419.93716627, 1407.90380046, 1347.73697138, 1431.97053209, 1371.80370301, 1335.70360556, 1359.77033719, 1347.73697138, 1395.87043464, 1299.60350811, 1323.67023975, 1395.87043464, 1444.00389791, 1444.00389791, 1419.93716627, 1419.93716627, 1419.93716627, 1383.83706883, 1347.73697138, 1419.93716627, 1407.90380046, 1444.00389791, 1395.87043464, 1311.63687393, 1323.67023975, 1275.53677648, 1179.26984996, 1179.26984996, 1102.25630873, 1036.07279675, 998.76936272, 975.90596767, 987.93933348, 1016.81941144, 1044.49615282, 1039.68280649, 1032.46278700, 1002.37937246, 1064.95287471, 974.70263109, 926.56916782, 922.95915808, 959.05925553, 928.97584099, 985.53266032, 1021.63275777, 1067.35954787, 1026.44610409, 1038.47946991, 1052.91950889, 1107.06965506, 1066.15621129, 1051.71617231, 999.97269930, 989.14267007, 984.32932374, 986.73599690, 916.94247517, 898.89242645, 833.91225104, 832.70891446, 830.30224130, 849.55562660, 831.50557788, 830.30224130], 
							[3002.37290452, 2918.98167942, 2776.00122680, 2895.14358174, 2859.40448527, 2740.26213032, 2728.34909817, 2787.91425895, 2799.83932448, 2811.75235664, 2835.57842095, 2787.91425895, 2847.49145311, 2907.05661390, 2823.66538879, 2835.57842095, 2692.61000169, 2680.69696953, 2787.91425895, 2764.08819464, 2716.43606601, 2633.03280754, 2740.26213032, 2573.46764675, 2513.89045260, 2811.75235664, 2954.72077589, 3026.19896884, 2942.80774374, 2895.14358174, 2871.31751742, 2823.66538879, 2776.00122680, 2787.91425895, 2680.69696953, 2728.34909817, 2704.52303385, 2656.85887185, 2776.00122680, 2776.00122680, 2752.17516248, 2716.43606601, 2752.17516248, 2799.83932448, 2740.26213032, 2716.43606601, 2680.69696953, 2644.94583970, 2740.26213032, 2823.66538879, 2776.00122680, 2835.57842095, 2859.40448527, 2918.98167942, 2978.54684021, 3097.68919515, 3073.86313084, 4050.82803470, 3741.05309850, 3526.59445293, 3598.08467924, 3645.73680787, 3562.34558277, 3467.02929214, 3431.29019567, 3538.50748509, 3478.94232430, 3526.59445293, 3478.94232430, 3431.29019567, 3455.11625998, 3514.68142077, 3562.34558277, 3490.85535645, 3621.91074356, 3574.25861492, 3657.66187339, 3574.25861492, 3502.76838861, 3490.85535645, 3490.85535645, 3419.37716351, 3395.53906583, 3395.53906583, 3586.17164708, 3550.43255061, 3633.82377571, 3562.34558277, 3467.02929214, 3526.59445293, 3514.68142077, 3312.14784072, 3419.37716351, 3514.68142077, 3514.68142077, 3347.88693720, 3335.97390504, 3359.79996935, 3383.62603367, 3371.71300151, 3240.65761441, 3324.06087288, 3335.97390504, 3371.71300151, 3395.53906583, 3478.94232430, 3371.71300151, 3335.97390504, 3204.91851794, 3133.42829163, 3181.09245362, 3169.16738810, 3121.51525947, 3145.34132378, 3216.83155010, 3181.09245362, 3216.83155010, 3169.16738810, 3204.91851794, 3133.42829163, 3097.68919515, 3133.42829163, 3157.25435594, 3121.51525947, 3121.51525947, 3085.77616299, 3097.68919515, 3097.68919515, 3133.42829163, 3109.60222731, 3085.77616299, 3085.77616299, 3061.95009868, 3061.95009868, 3073.86313084, 2978.54684021, 2930.89471158, 2942.80774374, 2918.98167942, 2883.23054958, 2895.14358174, 2811.75235664, 2823.66538879, 2847.49145311, 2835.57842095, 2835.57842095, 2835.57842095, 2859.40448527, 2859.40448527, 2823.66538879, 2799.83932448, 2716.43606601, 2537.72855028, 2656.85887185, 2704.52303385, 2704.52303385, 2621.11977538, 2692.61000169, 2609.20674322, 2597.29371106, 2537.72855028, 2668.78393738, 2644.94583970, 2692.61000169, 2835.57842095, 2740.26213032, 2776.00122680, 2680.69696953, 2692.61000169, 2668.78393738, 2644.94583970, 2644.94583970, 2644.94583970, 2704.52303385, 2716.43606601, 2799.83932448, 2704.52303385, 2692.61000169, 2716.43606601, 2656.85887185, 2692.61000169, 2752.17516248, 2728.34909817, 2740.26213032, 2811.75235664, 2966.63380805, 2978.54684021, 2960.20799071, 2972.24135652, 2948.17462489, 2984.27472234, 2863.94106418, 2888.00779581, 2912.07452744, 2924.10789326, 2888.00779581, 2912.07452744, 2924.10789326, 2948.17462489, 2996.30808815, 3008.34145397, 2996.30808815, 2996.30808815, 3020.37481979, 2984.27472234, 2996.30808815, 2984.27472234, 2972.24135652, 2924.10789326, 2912.07452744, 2912.07452744, 2875.97442999, 2948.17462489, 2984.27472234, 3008.34145397, 2996.30808815, 3044.44155142, 3056.47491723, 3044.44155142, 3020.37481979, 3044.44155142, 3008.34145397, 2996.30808815, 2972.24135652, 2960.20799071, 2984.27472234, 3020.37481979, 3008.34145397, 2888.00779581, 2888.00779581, 2888.00779581, 2948.17462489, 2912.07452744, 2936.14125907, 2912.07452744, 2912.07452744, 2888.00779581, 2900.04116163, 2972.24135652, 2936.14125907, 2888.00779581, 2863.94106418, 2900.04116163, 2851.90769836, 2900.04116163, 2900.04116163, 2888.00779581, 2863.94106418, 2863.94106418, 2803.77423510, 2803.77423510, 2851.90769836, 2851.90769836, 2888.00779581, 2623.27374786, 2791.74086928, 2719.54067439, 2707.50730857, 2539.04018715, 2502.94008970, 2514.97345552, 2635.30711368, 2575.14028460, 2743.60740602, 2767.67413765, 2563.10691878, 2575.14028460, 2502.94008970, 2647.34047949, 2611.24038205, 2647.34047949, 2599.20701623, 2611.24038205, 2755.64077184, 2635.30711368, 2695.47394276, 2719.54067439, 2743.60740602, 2647.34047949, 2599.20701623, 2611.24038205, 2623.27374786, 2671.40721112, 2671.40721112, 2659.37384531, 2611.24038205, 2635.30711368, 2707.50730857, 2719.54067439, 2671.40721112, 2623.27374786], 
							[690.71519783, 697.93521732, 685.90185151, 661.83511987, 654.61510038, 601.66829079, 611.29498345, 543.90813488, 528.26475932, 531.87476906, 601.66829079, 571.58487625, 587.22825181, 610.09164687, 577.60155916, 547.51814462, 492.16466187, 475.31794973, 506.60470085, 507.80803743, 511.41804717, 499.38468136, 509.01137401, 481.33463264, 466.89459366, 489.75798871, 649.80175406, 643.78507115, 660.63178329, 688.30852467, 709.96858314, 667.85180278, 690.71519783, 693.12187099, 664.24179304, 660.63178329, 647.39508089, 611.29498345, 661.83511987, 667.85180278, 691.91853441, 694.32520758, 703.95190023, 679.88516860, 669.05513936, 661.83511987, 651.00509064, 629.34503217, 670.25847594, 682.29184176, 648.59841748, 649.80175406, 658.22511013, 637.76838824, 677.47849543, 667.85180278, 669.05513936, 722.00194895, 746.06868058, 761.71205615, 764.11872931, 788.18546094, 782.16877803, 770.13541222, 780.96544145, 799.01549017, 796.60881701, 794.20214385, 801.42216334, 754.49203666, 734.03531477, 729.22196844, 718.39193921, 693.12187099, 711.17191972, 714.78192946, 742.45867084, 744.86534400, 756.89870982, 759.30538298, 749.67869033, 761.71205615, 726.81529528, 703.95190023, 737.64532451, 719.59527579, 746.06868058, 722.00194895, 637.76838824, 634.15837850, 635.36171508, 602.87162738, 607.68497370, 608.88831028, 640.17506140, 611.29498345, 595.65160789, 624.53168584, 644.98840773, 654.61510038, 661.83511987, 728.01863186, 732.83197819, 679.88516860, 703.95190023, 719.59527579, 709.96858314, 696.73188074, 648.59841748, 641.37839799, 700.34189048, 688.30852467, 676.27515885, 688.30852467, 681.08850518, 703.95190023, 697.93521732, 713.57859288, 707.56190997, 699.13855390, 709.96858314, 701.54522707, 766.52540247, 770.13541222, 766.52540247, 758.10204640, 783.37211461, 784.57545120, 805.03217308, 779.76210487, 780.96544145, 780.96544145, 753.28870007, 748.47535375, 756.89870982, 778.55876829, 752.08536349, 753.28870007, 755.69537324, 735.23865135, 712.37525630, 648.59841748, 673.86848569, 661.83511987, 691.91853441, 712.37525630, 697.93521732, 697.93521732, 683.49517834, 695.52854416, 672.66514911, 718.39193921, 711.17191972, 724.40862212, 719.59527579, 720.79861237, 685.90185151, 630.54836875, 608.88831028, 590.83826156, 569.17820309, 600.46495421, 537.89145197, 565.56819335, 610.09164687, 589.63492498, 565.56819335, 541.50146171, 554.73816411, 535.48477881, 533.07810564, 501.79135452, 471.70793998, 481.33463264, 464.48792049, 537.89145197, 511.41804717, 521.04473983, 541.50146171, 534.28144222, 515.02805692, 581.21156891, 571.58487625, 583.61824207, 607.68497370, 653.41176380, 595.65160789, 629.34503217, 654.61510038, 641.37839799, 648.59841748, 595.65160789, 613.70165661, 619.71833952, 631.75170533, 664.24179304, 675.07182227, 683.49517834, 672.66514911, 660.63178329, 663.03845645, 694.32520758, 690.71519783, 699.13855390, 728.01863186, 746.06868058, 742.45867084, 759.30538298, 732.83197819, 711.17191972, 688.30852467, 712.37525630, 720.79861237, 776.15209512, 771.33874880, 758.10204640, 780.96544145, 773.74542196, 741.25533426, 771.33874880, 786.98212436, 774.94875854, 780.96544145, 767.72873905, 740.05199768, 746.06868058, 756.89870982, 741.25533426, 720.79861237, 707.56190997, 684.69851492, 691.91853441, 685.90185151, 703.95190023, 687.10518809, 640.17506140, 653.41176380, 638.97172482, 652.20842722, 690.71519783, 660.63178329, 648.59841748, 681.08850518, 653.41176380, 637.76838824, 664.24179304, 661.83511987, 675.07182227, 638.97172482, 595.65160789, 641.37839799, 601.66829079, 577.60155916, 593.24493472, 493.36799845, 541.50146171, 521.04473983, 566.77152993, 531.87476906, 523.45141299, 547.51814462, 510.21471059, 488.55465212, 504.19802769, 521.04473983, 489.75798871, 460.87791075, 434.40450595, 471.70793998, 466.89459366, 493.36799845, 506.60470085, 551.12815437, 566.77152993, 560.75484702, 533.07810564, 535.48477881, 535.48477881, 543.90813488, 517.43473008, 499.38468136, 483.74130580, 499.38468136, 441.62452544, 406.72776458, 418.76113039, 404.32109141, 431.99783279, 422.37114014, 409.13443774, 397.10107192], 
							[1008.39605537, 1025.24276751, 1031.25945042, 1039.68280649, 1030.05611384, 1028.85277726, 1031.25945042, 1034.86946017, 1036.07279675, 1001.17603588, 1046.90282598, 1020.42942119, 1044.49615282, 1046.90282598, 1020.42942119, 1014.41273828, 978.31264083, 925.36583124, 965.07593843, 974.70263109, 955.44924578, 914.53580201, 939.80587022, 904.90910935, 849.55562660, 914.53580201, 1004.78604563, 978.31264083, 998.76936272, 998.76936272, 1014.41273828, 977.10930425, 985.53266032, 1004.78604563, 996.36268955, 963.87260185, 985.53266032, 972.29595792, 1016.81941144, 1013.20940170, 992.75267981, 1018.02274802, 1068.56288445, 1064.95287471, 1021.63275777, 1042.08947966, 1022.83609435, 1024.03943093, 1048.10616256, 992.75267981, 968.68594818, 984.32932374, 991.54934323, 983.12598716, 1050.51283573, 1095.03628924, 1150.38977200, 1131.13638669, 1151.59310858, 1180.47318654, 1203.33658159, 1251.47004485, 1215.36994740, 1187.69320603, 1203.33658159, 1275.53677648, 1287.57014230, 1287.57014230, 1239.43667904, 1263.50341067, 1251.47004485, 1239.43667904, 1263.50341067, 1203.33658159, 1203.33658159, 1239.43667904, 1227.40331322, 1251.47004485, 1263.50341067, 1275.53677648, 1251.47004485, 1239.43667904, 1227.40331322, 1239.43667904, 1251.47004485, 1251.47004485, 1227.40331322, 1239.43667904, 1181.67652312, 1181.67652312, 1101.05297215, 1064.95287471, 1048.10616256, 1113.08633797, 1115.49301113, 1098.64629899, 1064.95287471, 1109.47632822, 1115.49301113, 1143.16975251, 1104.66298190, 1166.03314756, 1170.84649388, 1135.94973302, 1169.64315730, 1191.30321577, 1194.91322552, 1192.50655235, 1203.33658159, 1166.03314756, 1215.36994740, 1215.36994740, 1168.43982072, 1167.23648414, 1190.09987919, 1202.13324501, 1227.40331322, 1227.40331322, 1287.57014230, 1311.63687393, 1287.57014230, 1263.50341067, 1251.47004485, 1197.31989868, 1215.36994740, 1203.33658159, 1202.13324501, 1227.40331322, 1239.43667904, 1251.47004485, 1239.43667904, 1251.47004485, 1203.33658159, 1198.52323526, 1239.43667904, 1215.36994740, 1227.40331322, 1251.47004485, 1251.47004485, 1275.53677648, 1263.50341067, 1239.43667904, 1263.50341067, 1239.43667904, 1263.50341067, 1251.47004485, 1239.43667904, 1251.47004485, 1263.50341067, 1251.47004485, 1251.47004485, 1203.33658159, 1131.13638669, 1157.60979149, 1121.50969404, 1139.55974276, 1108.27299164, 1107.06965506, 1127.52637695, 1116.69634771, 1060.13952838, 1076.98624052, 1060.13952838, 1085.40959659, 1140.76307935, 1131.13638669, 1107.06965506, 1102.25630873, 1089.01960634, 1070.96955761, 1049.30949914, 1008.39605537, 993.95601639, 993.95601639, 980.71931399, 1039.68280649, 1051.71617231, 1030.05611384, 1052.91950889, 1063.74953812, 1052.91950889, 1064.95287471, 1058.93619180, 1064.95287471, 1067.35954787, 1121.50969404, 1119.10302088, 1131.13638669, 1153.99978174, 1143.16975251, 1164.82981098, 1123.91636720, 1128.72971353, 1135.94973302, 1090.22294292, 1108.27299164, 1115.49301113, 1135.94973302, 1101.05297215, 1126.32304037, 1123.91636720, 1144.37308909, 1131.13638669, 1151.59310858, 1172.04983047, 1202.13324501, 1179.26984996, 1179.26984996, 1158.81312807, 1145.57642567, 1125.11970378, 1134.74639644, 1128.72971353, 1173.25316705, 1103.45964532, 1115.49301113, 1139.55974276, 1157.60979149, 1147.98309883, 1145.57642567, 1158.81312807, 1153.99978174, 1134.74639644, 1113.08633797, 1125.11970378, 1146.77976225, 1138.35640618, 1151.59310858, 1149.18643542, 1096.23962583, 1114.28967455, 1121.50969404, 1089.01960634, 1091.42627950, 1098.64629899, 1061.34286496, 1026.44610409, 1021.63275777, 968.68594818, 1008.39605537, 1010.80272853, 1008.39605537, 1050.51283573, 1018.02274802, 1010.80272853, 1002.37937246, 978.31264083, 984.32932374, 980.71931399, 943.41587996, 894.07908012, 902.50243619, 878.43570456, 882.04571430, 825.48889497, 795.40548043, 760.50871956, 773.74542196, 767.72873905, 794.20214385, 813.45552915, 808.64218283, 753.28870007, 713.57859288, 719.59527579, 750.88202691, 754.49203666, 723.20528553, 722.00194895, 724.40862212, 783.37211461, 842.33560711, 914.53580201, 836.31892420, 809.84551941, 854.36897293, 845.94561686, 835.11558762, 831.50557788, 841.13227053, 818.26887548, 756.89870982, 760.50871956, 722.00194895, 746.06868058, 735.23865135, 728.01863186, 734.03531477, 728.01863186, 696.73188074, 661.83511987], 
							[1004.76197889, 1004.76197889, 968.04817979, 1020.34518763, 996.56725677, 922.95915808, 876.23359861, 852.86480220, 855.19927517, 827.16153282, 892.58694276, 889.07319994, 876.23359861, 887.90596346, 838.84593102, 817.81160758, 765.23783233, 712.66405708, 759.40164991, 775.75499405, 760.56888639, 719.67950935, 759.40164991, 736.03285349, 724.34845529, 789.77386523, 872.71985580, 882.06978104, 869.21814634, 869.21814634, 911.27475987, 870.38538283, 827.16153282, 828.32876930, 835.33218821, 811.97542516, 807.29444586, 764.07059584, 810.80818867, 822.48055352, 800.29102695, 810.80818867, 841.18040399, 834.16495172, 795.61004765, 774.58775757, 788.60662874, 775.75499405, 817.81160758, 832.99771524, 814.30989813, 835.33218821, 841.18040399, 837.67869454, 911.27475987, 917.12297566, 947.49519098, 953.33137340, 906.60581393, 950.99690043, 931.14184683, 969.69675091, 979.03464278, 946.32795449, 965.01577160, 976.70016981, 966.18300809, 942.82624504, 934.64355629, 969.69675091, 967.35024457, 975.53293333, 981.38114911, 953.33137340, 972.03122387, 946.32795449, 952.16413691, 956.84511622, 956.84511622, 994.22075044, 986.05009505, 952.16413691, 963.84853512, 949.82966394, 943.99348152, 946.32795449, 976.70016981, 934.64355629, 876.23359861, 889.07319994, 850.53032923, 817.81160758, 837.67869454, 811.97542516, 859.86822111, 811.97542516, 863.38196392, 868.05090986, 885.57149049, 891.41970627, 864.54920041, 922.95915808, 890.25246979, 855.19927517, 893.75417924, 899.59036166, 880.90254455, 877.40083510, 876.23359861, 843.51487696, 886.73872697, 870.38538283, 865.71643689, 841.18040399, 869.21814634, 870.38538283, 873.88709228, 875.05432876, 871.55261931, 862.20269407, 872.71985580, 852.86480220, 841.18040399, 844.68211345, 842.34764048, 856.36651165, 863.38196392, 866.88367338, 865.71643689, 859.86822111, 869.21814634, 872.71985580, 861.03545759, 857.53374814, 836.51145806, 818.97884406, 814.30989813, 818.97884406, 823.65982337, 818.97884406, 823.65982337, 825.99429633, 807.29444586, 803.79273640, 824.82705985, 813.14266164, 829.49600579, 875.05432876, 852.86480220, 857.53374814, 855.19927517, 851.69756572, 832.99771524, 861.03545759, 849.35105938, 823.65982337, 818.97884406, 827.16153282, 832.99771524, 843.51487696, 806.12720937, 782.75841296, 775.75499405, 782.75841296, 825.99429633, 821.31331703, 837.67869454, 807.29444586, 808.46168234, 766.40506881, 762.90335936, 737.20008998, 724.34845529, 741.86903591, 719.67950935, 782.75841296, 780.42393999, 786.27215578, 773.42052108, 772.25328460, 771.08604812, 799.12379047, 788.60662874, 796.77728413, 811.97542516, 816.64437109, 806.12720937, 802.62549992, 827.16153282, 796.77728413, 823.65982337, 794.44281116, 817.81160758, 799.12379047, 807.29444586, 815.47713461, 842.34764048, 832.99771524, 821.31331703, 824.82705985, 823.65982337, 838.84593102, 813.14266164, 823.65982337, 830.66324227, 838.84593102, 855.19927517, 863.38196392, 862.20269407, 858.70098462, 828.32876930, 828.32876930, 864.54920041, 892.87574354, 907.31578252, 892.87574354, 908.51911910, 908.51911910, 873.62235823, 878.43570456, 884.45238747, 883.24905089, 868.80901191, 855.57230951, 879.63904114, 878.43570456, 888.06239721, 870.01234849, 871.21568507, 862.79232900, 856.77564609, 856.77564609, 856.77564609, 848.35229002, 849.55562660, 844.74228027, 838.72559737, 855.57230951, 856.77564609, 872.41902165, 866.40233874, 866.40233874, 882.04571430, 855.57230951, 860.38565584, 871.21568507, 859.18231925, 862.79232900, 844.74228027, 824.28555839, 842.33560711, 824.28555839, 827.89556813, 836.31892420, 762.91539273, 771.33874880, 752.08536349, 768.93207563, 725.61195870, 736.44198793, 760.50871956, 747.27201717, 756.89870982, 768.93207563, 794.20214385, 760.50871956, 738.84866109, 752.08536349, 776.15209512, 777.35543171, 794.20214385, 801.42216334, 824.28555839, 785.77878778, 776.15209512, 789.38879752, 827.89556813, 808.64218283, 803.82883650, 788.18546094, 784.57545120, 782.16877803, 803.82883650, 746.06868058, 758.10204640, 722.00194895, 722.00194895, 732.83197819, 742.45867084, 725.61195870, 728.01863186], 
							[758.10204640, 792.99880727, 758.10204640, 730.42530502, 717.18860263, 625.73502243, 614.90499319, 590.83826156, 589.63492498, 589.63492498, 636.56505166, 652.20842722, 631.75170533, 632.95504192, 605.27830054, 590.83826156, 589.63492498, 512.62138376, 559.55151044, 577.60155916, 547.51814462, 499.38468136, 535.48477881, 501.79135452, 451.25121810, 481.33463264, 541.50146171, 566.77152993, 571.58487625, 551.12815437, 569.17820309, 513.82472034, 500.58801794, 521.04473983, 496.97800820, 490.96132529, 502.99469110, 463.28458391, 495.77467161, 506.60470085, 495.77467161, 505.40136427, 510.21471059, 483.74130580, 459.67457417, 444.03119861, 456.06456442, 441.62452544, 481.33463264, 482.53796922, 460.87791075, 486.14797896, 500.58801794, 489.75798871, 522.24807641, 513.82472034, 516.23139350, 545.11147146, 565.56819335, 567.97486651, 549.92481779, 599.26161763, 601.66829079, 703.95190023, 709.96858314, 740.05199768, 734.03531477, 759.30538298, 690.71519783, 703.95190023, 703.95190023, 683.49517834, 699.13855390, 664.24179304, 691.91853441, 696.73188074, 712.37525630, 685.90185151, 693.12187099, 682.29184176, 672.66514911, 655.81843697, 653.41176380, 613.70165661, 637.76838824, 587.22825181, 569.17820309, 534.28144222, 463.28458391, 500.58801794, 477.72462289, 445.23453519, 504.19802769, 471.70793998, 493.36799845, 445.23453519, 450.04788151, 481.33463264, 528.26475932, 558.34817386, 529.46809590, 608.88831028, 557.14483728, 521.04473983, 548.72148120, 570.38153967, 589.63492498, 595.65160789, 563.16152018, 529.46809590, 581.21156891, 582.41490549, 588.43158840, 557.14483728, 571.58487625, 611.29498345, 612.49832003, 673.86848569, 685.90185151, 709.96858314, 711.17191972, 689.51186125, 663.03845645, 643.78507115, 655.81843697, 673.86848569, 670.25847594, 664.24179304, 667.85180278, 697.93521732, 691.91853441, 691.91853441, 663.03845645, 673.86848569, 675.07182227, 665.44512962, 673.86848569, 730.42530502, 728.01863186, 711.17191972, 714.78192946, 728.01863186, 726.81529528, 658.22511013, 663.03845645, 691.91853441, 684.69851492, 708.76524656, 709.96858314, 712.37525630, 701.54522707, 706.35857339, 702.74856365, 671.46181253, 647.39508089, 631.75170533, 612.49832003, 567.97486651, 519.84140325, 541.50146171, 494.57133503, 565.56819335, 534.28144222, 549.92481779, 594.44827130, 589.63492498, 601.66829079, 558.34817386, 531.87476906, 510.21471059, 510.21471059, 481.33463264, 460.87791075, 487.35131554, 435.60784253, 492.16466187, 525.85808615, 511.41804717, 533.07810564, 548.72148120, 542.70479830, 576.39822258, 580.00823233, 593.24493472, 583.61824207, 655.81843697, 599.26161763, 596.85494447, 604.07496396, 551.12815437, 554.73816411, 545.11147146, 559.55151044, 533.07810564, 529.46809590, 528.26475932, 552.33149095, 589.63492498, 572.78821284, 572.78821284, 575.19488600, 577.60155916, 610.09164687, 587.22825181, 586.02491523, 610.09164687, 607.68497370, 595.65160789, 588.43158840, 577.60155916, 588.43158840, 589.63492498, 564.36485676, 577.60155916, 625.73502243, 610.09164687, 629.34503217, 661.83511987, 626.93835901, 605.27830054, 626.93835901, 618.51500294, 624.53168584, 599.26161763, 608.88831028, 629.34503217, 624.53168584, 625.73502243, 584.82157865, 577.60155916, 566.77152993, 554.73816411, 527.06142274, 535.48477881, 554.73816411, 524.65474957, 498.18134478, 536.68811539, 542.70479830, 533.07810564, 518.63806666, 521.04473983, 540.29812513, 495.77467161, 499.38468136, 521.04473983, 496.97800820, 490.96132529, 463.28458391, 421.16780356, 430.79449621, 400.71108167, 387.47437927, 401.91441825, 329.71422336, 354.98429157, 338.13757943, 347.76427208, 336.93424284, 339.34091601, 344.15426233, 327.30755019, 323.69754045, 298.42747223, 311.66417463, 283.98743325, 269.54739428, 241.87065290, 271.95406744, 264.73404795, 270.75073086, 279.17408693, 300.83414540, 296.02079907, 312.86751121, 308.05416489, 302.03748198, 303.24081856, 298.42747223, 288.80077958, 276.76741377, 263.53071137, 286.39410642, 240.66731632, 252.70068213, 210.58390178, 210.58390178, 204.56721887, 210.58390178, 198.55053596, 175.68714091], 
							[494.57133503, 482.53796922, 469.30126682, 475.31794973, 469.30126682, 410.33777432, 427.18448646, 451.25121810, 412.74444748, 440.42118886, 474.11461315, 445.23453519, 446.43787177, 439.21785228, 434.40450595, 425.98114988, 406.72776458, 397.10107192, 422.37114014, 415.15112065, 442.82786202, 436.81117912, 457.26790100, 430.79449621, 403.11775483, 433.20116937, 481.33463264, 482.53796922, 457.26790100, 565.56819335, 624.53168584, 600.46495421, 571.58487625, 563.16152018, 552.33149095, 541.50146171, 543.90813488, 528.26475932, 578.80489574, 600.46495421, 583.61824207, 595.65160789, 584.82157865, 570.38153967, 565.56819335, 553.53482753, 549.92481779, 547.51814462, 549.92481779, 559.55151044, 543.90813488, 531.87476906, 521.04473983, 499.38468136, 537.89145197, 555.94150069, 572.78821284, 617.31166635, 600.46495421, 619.71833952, 596.85494447, 665.44512962, 669.05513936, 665.44512962, 654.61510038, 678.68183202, 679.88516860, 651.00509064, 611.29498345, 602.87162738, 618.51500294, 632.95504192, 636.56505166, 600.46495421, 655.81843697, 640.17506140, 632.95504192, 623.32834926, 619.71833952, 619.71833952, 590.83826156, 577.60155916, 553.53482753, 529.46809590, 536.68811539, 553.53482753, 537.89145197, 502.99469110, 475.31794973, 470.50460340, 472.91127656, 446.43787177, 419.96446697, 439.21785228, 462.08124733, 439.21785228, 478.92795947, 475.31794973, 505.40136427, 516.23139350, 474.11461315, 505.40136427, 487.35131554, 468.09793024, 480.13129605, 481.33463264, 494.57133503, 506.60470085, 493.36799845, 474.11461315, 471.70793998, 486.14797896, 487.35131554, 483.74130580, 486.14797896, 502.99469110, 499.38468136, 530.67143248, 534.28144222, 535.48477881, 552.33149095, 524.65474957, 505.40136427, 500.58801794, 509.01137401, 513.82472034, 523.45141299, 527.06142274, 521.04473983, 512.62138376, 522.24807641, 511.41804717, 493.36799845, 476.52128631, 483.74130580, 493.36799845, 493.36799845, 510.21471059, 533.07810564, 524.65474957, 533.07810564, 509.01137401, 536.68811539, 507.80803743, 502.99469110, 507.80803743, 512.62138376, 516.23139350, 534.28144222, 525.85808615, 523.45141299, 512.62138376, 472.91127656, 463.28458391, 458.47123758, 454.86122784, 440.42118886, 433.20116937, 407.93110116, 395.89773534, 357.39096473, 375.44101346, 385.06770611, 399.50774509, 418.76113039, 399.50774509, 418.76113039, 388.67771585, 379.05102320, 353.78095499, 359.79763789, 342.95092575, 310.46083805, 324.90087703, 303.24081856, 342.95092575, 370.62766713, 354.98429157, 358.59430131, 376.64435004, 379.05102320, 433.20116937, 415.15112065, 421.16780356, 400.71108167, 458.47123758, 427.18448646, 430.79449621, 446.43787177, 439.21785228, 428.38782305, 404.32109141, 435.60784253, 448.84454493, 482.53796922, 481.33463264, 511.41804717, 501.79135452, 470.50460340, 474.11461315, 498.18134478, 519.84140325, 527.06142274, 516.23139350, 523.45141299, 554.73816411, 543.90813488, 530.67143248, 516.23139350, 507.80803743, 509.01137401, 509.01137401, 543.90813488, 551.12815437, 563.16152018, 599.26161763, 557.14483728, 582.41490549, 570.38153967, 564.36485676, 565.56819335, 539.09478855, 549.92481779, 539.09478855, 543.90813488, 542.70479830, 539.09478855, 539.09478855, 518.63806666, 510.21471059, 504.19802769, 490.96132529, 477.72462289, 492.16466187, 502.99469110, 478.92795947, 477.72462289, 493.36799845, 501.79135452, 516.23139350, 507.80803743, 484.94464238, 515.02805692, 499.38468136, 492.16466187, 511.41804717, 512.62138376, 509.01137401, 511.41804717, 474.11461315, 481.33463264, 480.13129605, 469.30126682, 475.31794973, 444.03119861, 490.96132529, 478.92795947, 524.65474957, 540.29812513, 577.60155916, 570.38153967, 558.34817386, 560.75484702, 553.53482753, 564.36485676, 578.80489574, 587.22825181, 598.05828105, 604.07496396, 607.68497370, 626.93835901, 618.51500294, 604.07496396, 588.43158840, 599.26161763, 618.51500294, 625.73502243, 618.51500294, 605.27830054, 607.68497370, 602.87162738, 624.53168584, 659.42844671, 670.25847594, 648.59841748, 663.03845645, 647.39508089, 588.43158840, 588.43158840, 604.07496396, 588.43158840], 
							[1032.46278700, 997.56602614, 997.56602614, 1021.63275777, 987.93933348, 914.53580201, 801.42216334, 746.06868058, 764.11872931, 705.15523681, 749.67869033, 722.00194895, 707.56190997, 720.79861237, 697.93521732, 685.90185151, 636.56505166, 589.63492498, 620.92167610, 661.83511987, 631.75170533, 596.85494447, 625.73502243, 624.53168584, 589.63492498, 625.73502243, 673.86848569, 712.37525630, 746.06868058, 773.74542196, 797.81215359, 709.96858314, 711.17191972, 714.78192946, 719.59527579, 693.12187099, 722.00194895, 697.93521732, 730.42530502, 743.66200742, 737.64532451, 711.17191972, 734.03531477, 709.96858314, 664.24179304, 657.02177355, 651.00509064, 638.97172482, 671.46181253, 728.01863186, 687.10518809, 722.00194895, 737.64532451, 744.86534400, 859.18231925, 835.11558762, 817.06553890, 782.16877803, 771.33874880, 824.28555839, 809.84551941, 824.28555839, 809.84551941, 818.26887548, 825.48889497, 888.06239721, 914.53580201, 895.28241670, 863.99566558, 867.60567532, 851.96229976, 883.24905089, 892.87574354, 854.36897293, 888.06239721, 884.45238747, 877.23236798, 901.29909961, 913.33246543, 891.67240696, 920.55248491, 884.45238747, 878.43570456, 890.46907037, 872.41902165, 853.16563635, 871.21568507, 802.62549992, 687.10518809, 688.30852467, 661.83511987, 601.66829079, 661.83511987, 625.73502243, 672.66514911, 643.78507115, 670.25847594, 685.90185151, 749.67869033, 818.26887548, 796.60881701, 956.65258236, 903.70577277, 866.40233874, 986.73599690, 997.56602614, 1018.02274802, 1034.86946017, 984.32932374, 990.34600665, 1021.63275777, 1005.98938221, 998.76936272, 975.90596767, 961.46592869, 959.05925553, 998.76936272, 1002.37937246, 1008.39605537, 1010.80272853, 1012.00606512, 957.85591894, 973.49929450, 932.58585073, 918.14581175, 937.39919706, 953.04257262, 949.43256287, 962.66926527, 951.83923604, 956.65258236, 972.29595792, 909.72245568, 878.43570456, 877.23236798, 843.53894369, 871.21568507, 861.58899242, 890.46907037, 904.90910935, 957.85591894, 900.09576303, 901.29909961, 866.40233874, 862.79232900, 889.26573379, 925.36583124, 948.22922629, 967.48261160, 943.41587996, 950.63589945, 932.58585073, 871.21568507, 908.51911910, 889.26573379, 874.82569481, 836.31892420, 830.30224130, 747.27201717, 711.17191972, 722.00194895, 697.93521732, 705.15523681, 688.30852467, 772.54208538, 780.96544145, 806.23550966, 776.15209512, 743.66200742, 720.79861237, 735.23865135, 681.08850518, 643.78507115, 679.88516860, 631.75170533, 746.06868058, 755.69537324, 752.08536349, 758.10204640, 737.64532451, 750.88202691, 788.18546094, 792.99880727, 782.16877803, 805.03217308, 827.89556813, 794.20214385, 768.93207563, 795.40548043, 776.15209512, 791.79547068, 723.20528553, 728.01863186, 720.79861237, 746.06868058, 759.30538298, 776.15209512, 785.77878778, 767.72873905, 764.11872931, 758.10204640, 777.35543171, 801.42216334, 779.76210487, 848.35229002, 821.87888522, 835.11558762, 821.87888522, 785.77878778, 796.60881701, 817.06553890, 824.28555839, 813.45552915, 872.41902165, 936.19586048, 886.85906063, 913.33246543, 920.55248491, 868.80901191, 868.80901191, 930.17917757, 930.17917757, 1010.80272853, 968.68594818, 1005.98938221, 1096.23962583, 1095.03628924, 1046.90282598, 1028.85277726, 1046.90282598, 1030.05611384, 971.09262134, 956.65258236, 969.88928476, 1005.98938221, 930.17917757, 932.58585073, 1008.39605537, 1002.37937246, 1058.93619180, 1069.76622103, 1024.03943093, 1024.03943093, 1016.81941144, 1034.86946017, 1062.54620154, 1037.27613333, 1069.76622103, 1033.66612358, 990.34600665, 1032.46278700, 992.75267981, 920.55248491, 897.68908986, 752.08536349, 789.38879752, 755.69537324, 823.08222181, 771.33874880, 806.23550966, 839.92893395, 904.90910935, 890.46907037, 902.50243619, 938.60253364, 873.62235823, 809.84551941, 841.13227053, 830.30224130, 813.45552915, 888.06239721, 900.09576303, 926.56916782, 939.80587022, 974.70263109, 978.31264083, 980.71931399, 977.10930425, 944.61921655, 919.34914833, 926.56916782, 962.66926527, 933.78918731, 824.28555839, 848.35229002, 826.69223155, 782.16877803, 771.33874880, 825.48889497, 761.71205615, 778.55876829], 
							[1167.23648414, 1168.43982072, 1147.98309883, 1137.15306960, 1103.45964532, 1125.11970378, 1081.79958685, 914.53580201, 825.48889497, 806.23550966, 831.50557788, 884.45238747, 878.43570456, 841.13227053, 842.33560711, 799.01549017, 708.76524656, 655.81843697, 728.01863186, 701.54522707, 717.18860263, 696.73188074, 695.52854416, 722.00194895, 722.00194895, 771.33874880, 818.26887548, 872.41902165, 878.43570456, 890.46907037, 969.88928476, 867.60567532, 854.36897293, 844.74228027, 829.09890471, 801.42216334, 837.52226079, 835.11558762, 889.26573379, 930.17917757, 907.31578252, 890.46907037, 866.40233874, 861.58899242, 797.81215359, 808.64218283, 809.84551941, 789.38879752, 848.35229002, 894.07908012, 855.57230951, 878.43570456, 890.46907037, 886.85906063, 984.32932374, 1005.98938221, 1012.00606512, 981.92265058, 956.65258236, 987.93933348, 974.70263109, 1003.58270904, 972.29595792, 963.87260185, 978.31264083, 986.73599690, 963.87260185, 955.44924578, 939.80587022, 948.22922629, 943.41587996, 945.82255313, 973.49929450, 916.94247517, 961.46592869, 980.71931399, 993.95601639, 1001.17603588, 992.75267981, 1020.42942119, 1025.24276751, 966.27927501, 961.46592869, 1010.80272853, 957.85591894, 987.93933348, 979.51597741, 920.55248491, 806.23550966, 789.38879752, 754.49203666, 736.44198793, 776.15209512, 754.49203666, 821.87888522, 767.72873905, 799.01549017, 841.13227053, 890.46907037, 877.23236798, 842.33560711, 961.46592869, 926.56916782, 826.69223155, 867.60567532, 890.46907037, 909.72245568, 904.90910935, 878.43570456, 855.57230951, 915.73913859, 910.92579226, 909.72245568, 908.51911910, 907.31578252, 928.97584099, 936.19586048, 951.83923604, 1002.37937246, 990.34600665, 1019.22608460, 959.05925553, 948.22922629, 926.56916782, 921.75582150, 934.99252389, 947.02588971, 909.72245568, 947.02588971, 904.90910935, 904.90910935, 927.77250440, 878.43570456, 879.63904114, 851.96229976, 823.08222181, 844.74228027, 844.74228027, 853.16563635, 825.48889497, 825.48889497, 824.28555839, 812.25219257, 773.74542196, 774.94875854, 801.42216334, 818.26887548, 815.86220232, 825.48889497, 836.31892420, 844.74228027, 839.92893395, 774.94875854, 803.82883650, 818.26887548, 815.86220232, 791.79547068, 758.10204640, 714.78192946, 720.79861237, 689.51186125, 673.86848569, 671.46181253, 732.83197819, 771.33874880, 752.08536349, 764.11872931, 734.03531477, 715.98526604, 708.76524656, 711.17191972, 681.08850518, 612.49832003, 648.59841748, 611.29498345, 753.28870007, 752.08536349, 746.06868058, 760.50871956, 736.44198793, 724.40862212, 777.35543171, 758.10204640, 780.96544145, 794.20214385, 817.06553890, 772.54208538, 770.13541222, 808.64218283, 770.13541222, 792.99880727, 748.47535375, 770.13541222, 759.30538298, 777.35543171, 780.96544145, 795.40548043, 823.08222181, 827.89556813, 851.96229976, 825.48889497, 845.94561686, 845.94561686, 854.36897293, 872.41902165, 842.33560711, 842.33560711, 842.33560711, 806.23550966, 818.26887548, 833.91225104, 825.48889497, 850.75896318, 856.77564609, 895.28241670, 855.57230951, 860.38565584, 867.60567532, 845.94561686, 859.18231925, 912.12912884, 914.53580201, 902.50243619, 900.09576303, 916.94247517, 939.80587022, 920.55248491, 926.56916782, 916.94247517, 918.14581175, 895.28241670, 882.04571430, 876.02903140, 885.65572405, 897.68908986, 865.19900216, 866.40233874, 896.48575328, 883.24905089, 896.48575328, 882.04571430, 860.38565584, 862.79232900, 832.70891446, 830.30224130, 854.36897293, 844.74228027, 839.92893395, 821.87888522, 795.40548043, 777.35543171, 735.23865135, 709.96858314, 709.96858314, 640.17506140, 641.37839799, 657.02177355, 685.90185151, 690.71519783, 701.54522707, 734.03531477, 708.76524656, 675.07182227, 617.31166635, 663.03845645, 607.68497370, 558.34817386, 564.36485676, 602.87162738, 554.73816411, 632.95504192, 669.05513936, 732.83197819, 715.98526604, 732.83197819, 782.16877803, 738.84866109, 738.84866109, 712.37525630, 671.46181253, 654.61510038, 700.34189048, 685.90185151, 595.65160789, 576.39822258, 517.43473008, 502.99469110, 481.33463264, 484.94464238, 472.91127656, 471.70793998], 
							[784.57545120, 774.94875854, 790.59213410, 789.38879752, 742.45867084, 631.75170533, 575.19488600, 599.26161763, 640.17506140, 619.71833952, 661.83511987, 644.98840773, 663.03845645, 690.71519783, 688.30852467, 711.17191972, 640.17506140, 626.93835901, 638.97172482, 670.25847594, 663.03845645, 619.71833952, 667.85180278, 634.15837850, 578.80489574, 619.71833952, 685.90185151, 703.95190023, 673.86848569, 703.95190023, 711.17191972, 695.52854416, 685.90185151, 679.88516860, 688.30852467, 649.80175406, 634.15837850, 614.90499319, 648.59841748, 658.22511013, 628.14169559, 661.83511987, 667.85180278, 655.81843697, 625.73502243, 604.07496396, 600.46495421, 599.26161763, 612.49832003, 637.76838824, 613.70165661, 616.10832977, 635.36171508, 599.26161763, 613.70165661, 622.12501268, 675.07182227, 715.98526604, 691.91853441, 746.06868058, 728.01863186, 758.10204640, 764.11872931, 794.20214385, 830.30224130, 818.26887548, 812.25219257, 785.77878778, 748.47535375, 740.05199768, 749.67869033, 766.52540247, 801.42216334, 782.16877803, 783.37211461, 790.59213410, 784.57545120, 755.69537324, 786.98212436, 790.59213410, 759.30538298, 755.69537324, 746.06868058, 732.83197819, 718.39193921, 703.95190023, 700.34189048, 675.07182227, 628.14169559, 626.93835901, 598.05828105, 580.00823233, 612.49832003, 571.58487625, 618.51500294, 586.02491523, 593.24493472, 641.37839799, 661.83511987, 660.63178329, 637.76838824, 691.91853441, 703.95190023, 646.19174431, 667.85180278, 681.08850518, 702.74856365, 728.01863186, 688.30852467, 685.90185151, 703.95190023, 709.96858314, 701.54522707, 742.45867084, 728.01863186, 722.00194895, 722.00194895, 778.55876829, 782.16877803, 785.77878778, 789.38879752, 764.11872931, 755.69537324, 747.27201717, 778.55876829, 796.60881701, 783.37211461, 800.21882676, 806.23550966, 788.18546094, 788.18546094, 783.37211461, 772.54208538, 764.11872931, 753.28870007, 741.25533426, 766.52540247, 785.77878778, 780.96544145, 788.18546094, 790.59213410, 768.93207563, 768.93207563, 752.08536349, 737.64532451, 729.22196844, 752.08536349, 758.10204640, 765.32206589, 770.13541222, 754.49203666, 723.20528553, 726.81529528, 719.59527579, 703.95190023, 679.88516860, 605.27830054, 596.85494447, 570.38153967, 600.46495421, 566.77152993, 567.97486651, 554.73816411, 593.24493472, 608.88831028, 577.60155916, 595.65160789, 582.41490549, 563.16152018, 541.50146171, 536.68811539, 511.41804717, 480.13129605, 505.40136427, 457.26790100, 546.31480804, 581.21156891, 577.60155916, 551.12815437, 547.51814462, 527.06142274, 584.82157865, 578.80489574, 625.73502243, 614.90499319, 612.49832003, 557.14483728, 589.63492498, 600.46495421, 580.00823233, 583.61824207, 541.50146171, 588.43158840, 566.77152993, 596.85494447, 617.31166635, 612.49832003, 625.73502243, 626.93835901, 623.32834926, 614.90499319, 648.59841748, 654.61510038, 655.81843697, 652.20842722, 655.81843697, 658.22511013, 658.22511013, 611.29498345, 619.71833952, 616.10832977, 613.70165661, 644.98840773, 673.86848569, 705.15523681, 693.12187099, 678.68183202, 697.93521732, 661.83511987, 658.22511013, 689.51186125, 690.71519783, 700.34189048, 665.44512962, 690.71519783, 671.46181253, 691.91853441, 661.83511987, 660.63178329, 673.86848569, 673.86848569, 647.39508089, 642.58173457, 649.80175406, 630.54836875, 589.63492498, 589.63492498, 611.29498345, 616.10832977, 637.76838824, 612.49832003, 611.29498345, 611.29498345, 599.26161763, 592.04159814, 618.51500294, 594.44827130, 589.63492498, 534.28144222, 494.57133503, 501.79135452, 475.31794973, 470.50460340, 469.30126682, 457.26790100, 448.84454493, 459.67457417, 476.52128631, 488.55465212, 488.55465212, 494.57133503, 481.33463264, 469.30126682, 462.08124733, 470.50460340, 472.91127656, 463.28458391, 474.11461315, 487.35131554, 483.74130580, 504.19802769, 504.19802769, 506.60470085, 501.79135452, 513.82472034, 529.46809590, 519.84140325, 518.63806666, 504.19802769, 481.33463264, 475.31794973, 469.30126682, 469.30126682, 453.65789126, 424.77781330, 410.33777432, 382.66103294, 368.22099397, 381.45769636, 365.81432080, 345.35759892], 
							[877.23236798, 850.75896318, 794.20214385, 753.28870007, 734.03531477, 719.59527579, 719.59527579, 679.88516860, 699.13855390, 673.86848569, 726.81529528, 738.84866109, 725.61195870, 718.39193921, 673.86848569, 628.14169559, 625.73502243, 583.61824207, 600.46495421, 608.88831028, 610.09164687, 593.24493472, 601.66829079, 593.24493472, 587.22825181, 619.71833952, 644.98840773, 701.54522707, 682.29184176, 690.71519783, 666.64846620, 637.76838824, 631.75170533, 652.20842722, 646.19174431, 636.56505166, 634.15837850, 596.85494447, 613.70165661, 634.15837850, 618.51500294, 628.14169559, 626.93835901, 613.70165661, 586.02491523, 583.61824207, 584.82157865, 572.78821284, 595.65160789, 612.49832003, 610.09164687, 610.09164687, 647.39508089, 715.98526604, 752.08536349, 754.49203666, 939.80587022, 1153.99978174, 1203.33658159, 1191.30321577, 1111.88300139, 1188.89654261, 1161.21980123, 1186.48986945, 1175.65984021, 1172.04983047, 1167.23648414, 1156.40645491, 1125.11970378, 1058.93619180, 1052.91950889, 1076.98624052, 1160.01646465, 1102.25630873, 1107.06965506, 1108.27299164, 1172.04983047, 1138.35640618, 1153.99978174, 1164.82981098, 1126.32304037, 1073.37623078, 1058.93619180, 1044.49615282, 1090.22294292, 1080.59625027, 1068.56288445, 992.75267981, 916.94247517, 943.41587996, 902.50243619, 779.76210487, 860.38565584, 842.33560711, 938.60253364, 890.46907037, 878.43570456, 1019.22608460, 1048.10616256, 1095.03628924, 1069.76622103, 1215.36994740, 1168.43982072, 1110.67966481, 1173.25316705, 1152.79644516, 1146.77976225, 1149.18643542, 1125.11970378, 1104.66298190, 1086.61293317, 1089.01960634, 1045.69948940, 1051.71617231, 1052.91950889, 1062.54620154, 1057.73285522, 1074.57956736, 1113.08633797, 1120.30635746, 1108.27299164, 1058.93619180, 1057.73285522, 1042.08947966, 1052.91950889, 1010.80272853, 997.56602614, 987.93933348, 997.56602614, 977.10930425, 974.70263109, 993.95601639, 967.48261160, 932.58585073, 931.38251415, 901.29909961, 920.55248491, 950.63589945, 936.19586048, 928.97584099, 962.66926527, 910.92579226, 913.33246543, 884.45238747, 872.41902165, 873.62235823, 894.07908012, 919.34914833, 950.63589945, 944.61921655, 918.14581175, 908.51911910, 916.94247517, 877.23236798, 878.43570456, 851.96229976, 830.30224130, 789.38879752, 719.59527579, 714.78192946, 693.12187099, 718.39193921, 703.95190023, 715.98526604, 749.67869033, 725.61195870, 744.86534400, 689.51186125, 678.68183202, 643.78507115, 632.95504192, 584.82157865, 581.21156891, 600.46495421, 598.05828105, 661.83511987, 725.61195870, 709.96858314, 687.10518809, 700.34189048, 712.37525630, 770.13541222, 821.87888522, 943.41587996, 925.36583124, 924.16249466, 826.69223155, 831.50557788, 851.96229976, 812.25219257, 825.48889497, 764.11872931, 779.76210487, 743.66200742, 756.89870982, 737.64532451, 759.30538298, 794.20214385, 788.18546094, 788.18546094, 812.25219257, 835.11558762, 819.47221206, 824.28555839, 821.87888522, 854.36897293, 910.92579226, 883.24905089, 872.41902165, 902.50243619, 879.63904114, 872.41902165, 931.38251415, 984.32932374, 971.09262134, 965.07593843, 978.31264083, 992.75267981, 983.12598716, 968.68594818, 1002.37937246, 1009.59939195, 1046.90282598, 1079.39291368, 1089.01960634, 1081.79958685, 1092.62961608, 1057.73285522, 1050.51283573, 1043.29281624, 1019.22608460, 1018.02274802, 997.56602614, 1052.91950889, 1034.86946017, 1012.00606512, 1027.64944068, 1045.69948940, 1061.34286496, 1103.45964532, 1090.22294292, 1083.00292343, 1102.25630873, 1055.32618205, 1052.91950889, 1108.27299164, 1108.27299164, 1139.55974276, 1150.38977200, 1146.77976225, 1147.98309883, 1139.55974276, 1116.69634771, 1116.69634771, 1076.98624052, 1083.00292343, 1073.37623078, 1101.05297215, 1089.01960634, 1178.06651337, 1167.23648414, 1138.35640618, 1108.27299164, 1084.20626001, 1090.22294292, 1087.81626976, 1083.00292343, 1098.64629899, 1115.49301113, 1097.44296241, 1156.40645491, 1133.54305986, 1167.23648414, 1155.20311832, 1227.40331322, 1251.47004485, 1263.50341067, 1287.57014230, 1287.57014230, 1263.50341067, 1275.53677648, 1335.70360556, 1323.67023975, 1419.93716627, 1335.70360556, 1359.77033719, 1287.57014230, 1215.36994740, 1227.40331322, 1076.98624052, 1083.00292343], 
							[1744.83804330, 1648.57111678, 1636.53775096, 1624.50438514, 1588.40428770, 1636.53775096, 1347.73697138, 1203.33658159, 1202.13324501, 1134.74639644, 1275.53677648, 1275.53677648, 1155.20311832, 1170.84649388, 1097.44296241, 1022.83609435, 974.70263109, 950.63589945, 998.76936272, 1058.93619180, 1008.39605537, 966.27927501, 1021.63275777, 998.76936272, 989.14267007, 1032.46278700, 1095.03628924, 1251.47004485, 1178.06651337, 1215.36994740, 1275.53677648, 1190.09987919, 1086.61293317, 1092.62961608, 1083.00292343, 1076.98624052, 1119.10302088, 1034.86946017, 1167.23648414, 1191.30321577, 1215.36994740, 1131.13638669, 1149.18643542, 1179.26984996, 1083.00292343, 1089.01960634, 1107.06965506, 1072.17289419, 1115.49301113, 1186.48986945, 1173.25316705, 1131.13638669, 1185.28653286, 1197.31989868, 1371.80370301, 1552.30419025, 1335.70360556, 1239.43667904, 1251.47004485, 1419.93716627, 1299.60350811, 1444.00389791, 1371.80370301, 1371.80370301, 1395.87043464, 1431.97053209, 1419.93716627, 1359.77033719, 1323.67023975, 1347.73697138, 1371.80370301, 1419.93716627, 1395.87043464, 1359.77033719, 1407.90380046, 1419.93716627, 1419.93716627, 1456.03726372, 1444.00389791, 1492.13736117, 1468.07062954, 1407.90380046, 1407.90380046, 1407.90380046, 1335.70360556, 1383.83706883, 1347.73697138, 1311.63687393, 1203.33658159, 1203.33658159, 1089.01960634, 1119.10302088, 1117.89968430, 1119.10302088, 1107.06965506, 1105.86631848, 1140.76307935, 1173.25316705, 1263.50341067, 1227.40331322, 1175.65984021, 1239.43667904, 1335.70360556, 1203.33658159, 1287.57014230, 1347.73697138, 1431.97053209, 1528.23745862, 1395.87043464, 1359.77033719, 1444.00389791, 1480.10399535, 1431.97053209, 1468.07062954, 1444.00389791, 1564.33755606, 1684.67121422, 1696.70458004, 1612.47101933, 1636.53775096, 1684.67121422, 1588.40428770, 1528.23745862, 1480.10399535, 1540.27082443, 1492.13736117, 1431.97053209, 1444.00389791, 1444.00389791, 1407.90380046, 1359.77033719, 1395.87043464, 1359.77033719, 1359.77033719, 1359.77033719, 1347.73697138, 1323.67023975, 1407.90380046, 1407.90380046, 1564.33755606, 1516.20409280, 1456.03726372, 1528.23745862, 1468.07062954, 1468.07062954, 1444.00389791, 1528.23745862, 1540.27082443, 1528.23745862, 1540.27082443, 1528.23745862, 1564.33755606, 1395.87043464, 1444.00389791, 1444.00389791, 1383.83706883, 1383.83706883, 1263.50341067, 1299.60350811, 1239.43667904, 1263.50341067, 1227.40331322, 1155.20311832, 1215.36994740, 1299.60350811, 1359.77033719, 1395.87043464, 1323.67023975, 1335.70360556, 1263.50341067, 1251.47004485, 1173.25316705, 1138.35640618, 1202.13324501, 1113.08633797, 1311.63687393, 1311.63687393, 1359.77033719, 1431.97053209, 1431.97053209, 1431.97053209, 1552.30419025, 1516.20409280, 1780.93814075, 1780.93814075, 1744.83804330, 1732.80467749, 1780.93814075, 1937.37189636, 1792.97150657, 1877.20506728, 1877.20506728, 1949.40526217, 1937.37189636, 1961.43862799, 1925.33853054, 2045.67218870, 2262.27277339, 2214.13931012, 2178.03921267, 2238.20604175, 2250.23940757, 2153.97248104, 2214.13931012, 2262.27277339, 2286.33950502, 2358.53969991, 2334.47296828, 2274.30613920, 2298.37287083, 2250.23940757, 2238.20604175, 2286.33950502, 2346.50633410, 2346.50633410, 2286.33950502, 2322.43960246, 2322.43960246, 2202.10594431, 2226.17267594, 2274.30613920, 2129.90574941, 2093.80565196, 2129.90574941, 2238.20604175, 2250.23940757, 2178.03921267, 2166.00584686, 2129.90574941, 2093.80565196, 2117.87238359, 2093.80565196, 2093.80565196, 2105.83901778, 2178.03921267, 2069.73892033, 2069.73892033, 2117.87238359, 2190.07257849, 2214.13931012, 2214.13931012, 2238.20604175, 2262.27277339, 2153.97248104, 2117.87238359, 2214.13931012, 2178.03921267, 2190.07257849, 2370.57306573, 2358.53969991, 2442.77326062, 2346.50633410, 2394.63979736, 2370.57306573, 2370.57306573, 2406.67316318, 2454.80662644, 2478.87335807, 2635.30711368, 2888.00779581, 2827.84096673, 2767.67413765, 2767.67413765, 2659.37384531, 2767.67413765, 2743.60740602, 2912.07452744, 2924.10789326, 2948.17462489, 2960.20799071, 3068.50828305, 2984.27472234, 2972.24135652, 2623.27374786, 2659.37384531, 2683.44057694, 2707.50730857, 2514.97345552, 2478.87335807, 2514.97345552, 2659.37384531, 2647.34047949, 2551.07355297, 2791.74086928, 2659.37384531, 2635.30711368, 2418.70652899, 2238.20604175, 2286.33950502, 2238.20604175, 2346.50633410], 
							[939.74570339, 914.53580201, 934.01782126, 861.81762637, 859.53128686, 811.39782360, 751.79656271, 702.51992970, 676.15482519, 689.90896232, 719.71560945, 717.41723658, 722.00194895, 712.83252420, 693.35050495, 701.37675994, 630.31973480, 607.39617292, 650.94492381, 650.94492381, 654.38646643, 617.70876743, 599.36991792, 610.83771555, 581.04310179, 636.04761693, 732.31454346, 796.48848335, 779.30483697, 764.40753009, 802.22839885, 756.38127509, 742.62713796, 762.10915722, 736.89925583, 710.54618470, 722.00194895, 689.90896232, 736.89925583, 755.23810534, 740.34079846, 717.41723658, 716.27406682, 704.80626920, 676.15482519, 676.15482519, 668.14060356, 649.80175406, 665.84223069, 743.77030771, 727.72983108, 744.92551083, 767.83703935, 762.10915722, 812.54099335, 819.41204523, 805.65790810, 786.17588885, 793.05897410, 843.47877686, 804.51473835, 846.92031949, 842.33560711, 865.25916899, 833.16618236, 903.08003775, 948.91512814, 931.73148176, 859.53128686, 884.74118825, 876.71493325, 853.79137137, 859.53128686, 814.82733286, 854.94657449, 866.40233874, 860.67445661, 870.98705112, 881.29964562, 869.84388137, 874.42859374, 840.03723424, 850.36186211, 836.60772498, 803.37156860, 812.54099335, 779.30483697, 756.38127509, 660.11434856, 709.39098158, 677.31002831, 624.59185267, 679.59636782, 668.14060356, 719.71560945, 687.62262282, 692.20733519, 709.39098158, 771.27858197, 791.90377098, 762.10915722, 835.46455523, 796.48848335, 752.93973247, 813.68416310, 807.95628098, 823.99675761, 891.61224013, 861.81762637, 845.77714973, 869.84388137, 879.01330612, 862.96079612, 849.20665899, 848.06348924, 864.10396587, 865.25916899, 875.57176349, 840.03723424, 868.68867825, 888.17069750, 833.16618236, 841.19243736, 850.36186211, 859.53128686, 848.06348924, 813.68416310, 811.39782360, 810.24262048, 805.65790810, 811.39782360, 818.26887548, 815.98253597, 770.13541222, 768.99224246, 749.51022321, 756.38127509, 795.34531360, 760.96598746, 765.55069984, 777.00646410, 766.69386959, 754.09493558, 699.07838707, 709.39098158, 733.45771321, 742.62713796, 786.17588885, 771.27858197, 749.51022321, 749.51022321, 728.87300083, 704.80626920, 756.38127509, 733.45771321, 749.51022321, 707.10464207, 675.01165544, 624.59185267, 642.93070218, 636.04761693, 624.59185267, 609.69454579, 613.12405505, 668.14060356, 658.97117881, 687.62262282, 645.21704168, 625.73502243, 630.31973480, 626.87819218, 574.16001654, 564.99059179, 594.79723891, 553.53482753, 632.60607431, 653.24329668, 669.28377331, 658.97117881, 639.48915955, 632.60607431, 683.03791044, 676.15482519, 701.37675994, 722.00194895, 731.17137370, 683.03791044, 681.89474069, 719.71560945, 755.23810534, 750.65339296, 694.49367470, 741.48396821, 723.14511871, 740.34079846, 733.45771321, 749.51022321, 825.13992736, 813.68416310, 802.22839885, 803.37156860, 818.26887548, 818.26887548, 813.68416310, 833.16618236, 832.02301261, 827.43830023, 828.58146998, 823.99675761, 773.57695484, 818.26887548, 798.78685622, 813.68416310, 824.28555839, 877.23236798, 902.50243619, 934.99252389, 939.80587022, 898.89242645, 878.43570456, 925.36583124, 900.09576303, 896.48575328, 896.48575328, 904.90910935, 926.56916782, 903.70577277, 900.09576303, 879.63904114, 914.53580201, 871.21568507, 865.19900216, 853.16563635, 849.55562660, 876.02903140, 827.89556813, 807.43884625, 842.33560711, 860.38565584, 890.46907037, 861.58899242, 862.79232900, 870.01234849, 842.33560711, 842.33560711, 872.41902165, 850.75896318, 872.41902165, 877.23236798, 812.25219257, 806.23550966, 777.35543171, 779.76210487, 742.45867084, 688.30852467, 701.54522707, 746.06868058, 820.67554864, 758.10204640, 808.64218283, 813.45552915, 776.15209512, 765.32206589, 747.27201717, 782.16877803, 746.06868058, 728.01863186, 768.93207563, 779.76210487, 777.35543171, 817.06553890, 859.18231925, 886.85906063, 843.53894369, 872.41902165, 889.26573379, 860.38565584, 862.79232900, 863.99566558, 820.67554864, 802.62549992, 812.25219257, 823.08222181, 737.64532451, 701.54522707, 690.71519783, 664.24179304, 664.24179304, 708.76524656, 677.47849543, 678.68183202], 
							[593.24493472, 607.68497370, 555.94150069, 552.33149095, 548.72148120, 493.36799845, 500.58801794, 492.16466187, 496.97800820, 493.36799845, 528.26475932, 513.82472034, 523.45141299, 545.11147146, 523.45141299, 502.99469110, 493.36799845, 468.09793024, 511.41804717, 499.38468136, 487.35131554, 462.08124733, 469.30126682, 442.82786202, 439.21785228, 470.50460340, 516.23139350, 523.45141299, 534.28144222, 545.11147146, 542.70479830, 516.23139350, 483.74130580, 493.36799845, 510.21471059, 483.74130580, 496.97800820, 487.35131554, 490.96132529, 489.75798871, 490.96132529, 486.14797896, 494.57133503, 483.74130580, 469.30126682, 463.28458391, 460.87791075, 460.87791075, 478.92795947, 483.74130580, 469.30126682, 478.92795947, 478.92795947, 454.86122784, 478.92795947, 507.80803743, 555.94150069, 563.16152018, 608.88831028, 608.88831028, 601.66829079, 641.37839799, 630.54836875, 614.90499319, 612.49832003, 629.34503217, 613.70165661, 613.70165661, 617.31166635, 608.88831028, 602.87162738, 608.88831028, 623.32834926, 613.70165661, 610.09164687, 616.10832977, 637.76838824, 651.00509064, 643.78507115, 660.63178329, 637.76838824, 619.71833952, 605.27830054, 613.70165661, 610.09164687, 619.71833952, 629.34503217, 622.12501268, 596.85494447, 586.02491523, 554.73816411, 523.45141299, 535.48477881, 535.48477881, 541.50146171, 521.04473983, 530.67143248, 535.48477881, 524.65474957, 555.94150069, 537.89145197, 582.41490549, 575.19488600, 553.53482753, 548.72148120, 535.48477881, 530.67143248, 589.63492498, 590.83826156, 575.19488600, 607.68497370, 619.71833952, 601.66829079, 613.70165661, 616.10832977, 635.36171508, 626.93835901, 630.54836875, 623.32834926, 634.15837850, 646.19174431, 625.73502243, 631.75170533, 619.71833952, 607.68497370, 616.10832977, 625.73502243, 640.17506140, 661.83511987, 663.03845645, 659.42844671, 661.83511987, 631.75170533, 631.75170533, 643.78507115, 661.83511987, 666.64846620, 701.54522707, 679.88516860, 669.05513936, 679.88516860, 665.44512962, 673.86848569, 660.63178329, 655.81843697, 655.81843697, 648.59841748, 651.00509064, 659.42844671, 660.63178329, 634.15837850, 599.26161763, 589.63492498, 580.00823233, 577.60155916, 551.12815437, 563.16152018, 566.77152993, 557.14483728, 557.14483728, 527.06142274, 576.39822258, 563.16152018, 565.56819335, 590.83826156, 593.24493472, 596.85494447, 561.95818360, 535.48477881, 521.04473983, 499.38468136, 451.25121810, 445.23453519, 456.06456442, 436.81117912, 481.33463264, 487.35131554, 501.79135452, 492.16466187, 480.13129605, 525.85808615, 553.53482753, 552.33149095, 531.87476906, 549.92481779, 589.63492498, 553.53482753, 577.60155916, 581.21156891, 563.16152018, 553.53482753, 536.68811539, 577.60155916, 582.41490549, 575.19488600, 601.66829079, 590.83826156, 598.05828105, 589.63492498, 600.46495421, 604.07496396, 612.49832003, 622.12501268, 632.95504192, 631.75170533, 685.90185151, 667.85180278, 658.22511013, 655.81843697, 651.00509064, 649.80175406, 659.42844671, 691.91853441, 679.88516860, 685.90185151, 709.96858314, 701.54522707, 694.32520758, 657.02177355, 655.81843697, 669.05513936, 637.76838824, 642.58173457, 607.68497370, 618.51500294, 629.34503217, 608.88831028, 631.75170533, 602.87162738, 594.44827130, 600.46495421, 602.87162738, 587.22825181, 584.82157865, 604.07496396, 548.72148120, 561.95818360, 582.41490549, 571.58487625, 580.00823233, 569.17820309, 554.73816411, 577.60155916, 549.92481779, 563.16152018, 559.55151044, 558.34817386, 534.28144222, 518.63806666, 481.33463264, 480.13129605, 451.25121810, 441.62452544, 451.25121810, 386.27104269, 415.15112065, 391.08438902, 405.52442800, 413.94778407, 400.71108167, 412.74444748, 403.11775483, 394.69439876, 388.67771585, 438.01451570, 430.79449621, 457.26790100, 423.57447672, 493.36799845, 489.75798871, 499.38468136, 487.35131554, 484.94464238, 453.65789126, 425.98114988, 450.04788151, 441.62452544, 430.79449621, 442.82786202, 438.01451570, 422.37114014, 409.13443774, 391.08438902, 362.20431106, 354.98429157, 361.00097448, 367.01765738, 379.05102320, 405.52442800, 361.00097448, 347.76427208], 
							[661.83511987, 607.68497370, 605.27830054, 611.29498345, 619.71833952, 563.16152018, 539.09478855, 536.68811539, 523.45141299, 529.46809590, 559.55151044, 580.00823233, 577.60155916, 565.56819335, 529.46809590, 535.48477881, 539.09478855, 481.33463264, 528.26475932, 505.40136427, 495.77467161, 477.72462289, 507.80803743, 484.94464238, 456.06456442, 522.24807641, 566.77152993, 618.51500294, 601.66829079, 590.83826156, 602.87162738, 587.22825181, 592.04159814, 588.43158840, 570.38153967, 543.90813488, 537.89145197, 524.65474957, 551.12815437, 548.72148120, 552.33149095, 543.90813488, 555.94150069, 537.89145197, 541.50146171, 533.07810564, 533.07810564, 519.84140325, 553.53482753, 529.46809590, 511.41804717, 511.41804717, 494.57133503, 459.67457417, 501.79135452, 527.06142274, 572.78821284, 629.34503217, 617.31166635, 607.68497370, 589.63492498, 614.90499319, 636.56505166, 620.92167610, 642.58173457, 623.32834926, 619.71833952, 619.71833952, 583.61824207, 564.36485676, 553.53482753, 554.73816411, 581.21156891, 578.80489574, 593.24493472, 599.26161763, 599.26161763, 595.65160789, 593.24493472, 592.04159814, 571.58487625, 559.55151044, 546.31480804, 529.46809590, 553.53482753, 559.55151044, 563.16152018, 541.50146171, 499.38468136, 509.01137401, 494.57133503, 464.48792049, 501.79135452, 506.60470085, 509.01137401, 481.33463264, 481.33463264, 487.35131554, 509.01137401, 516.23139350, 496.97800820, 531.87476906, 517.43473008, 504.19802769, 531.87476906, 549.92481779, 545.11147146, 563.16152018, 547.51814462, 527.06142274, 521.04473983, 531.87476906, 534.28144222, 525.85808615, 523.45141299, 552.33149095, 555.94150069, 593.24493472, 582.41490549, 601.66829079, 619.71833952, 595.65160789, 569.17820309, 588.43158840, 589.63492498, 563.16152018, 582.41490549, 559.55151044, 575.19488600, 584.82157865, 598.05828105, 583.61824207, 557.14483728, 563.16152018, 561.95818360, 578.80489574, 555.94150069, 559.55151044, 553.53482753, 553.53482753, 565.56819335, 541.50146171, 547.51814462, 529.46809590, 531.87476906, 529.46809590, 537.89145197, 560.75484702, 555.94150069, 537.89145197, 536.68811539, 511.41804717, 498.18134478, 511.41804717, 493.36799845, 486.14797896, 469.30126682, 469.30126682, 464.48792049, 451.25121810, 445.23453519, 477.72462289, 444.03119861, 452.45455468, 470.50460340, 463.28458391, 474.11461315, 456.06456442, 435.60784253, 419.96446697, 418.76113039, 386.27104269, 394.69439876, 418.76113039, 400.71108167, 445.23453519, 440.42118886, 439.21785228, 456.06456442, 448.84454493, 466.89459366, 487.35131554, 474.11461315, 478.92795947, 490.96132529, 528.26475932, 494.57133503, 493.36799845, 522.24807641, 494.57133503, 496.97800820, 483.74130580, 504.19802769, 496.97800820, 509.01137401, 505.40136427, 525.85808615, 534.28144222, 513.82472034, 530.67143248, 523.45141299, 541.50146171, 527.06142274, 533.07810564, 539.09478855, 584.82157865, 578.80489574, 577.60155916, 560.75484702, 547.51814462, 537.89145197, 561.95818360, 565.56819335, 601.66829079, 612.49832003, 628.14169559, 600.46495421, 630.54836875, 596.85494447, 607.68497370, 602.87162738, 607.68497370, 596.85494447, 592.04159814, 605.27830054, 628.14169559, 643.78507115, 640.17506140, 625.73502243, 625.73502243, 626.93835901, 629.34503217, 617.31166635, 623.32834926, 629.34503217, 601.66829079, 613.70165661, 625.73502243, 643.78507115, 652.20842722, 635.36171508, 622.12501268, 652.20842722, 630.54836875, 637.76838824, 654.61510038, 649.80175406, 646.19174431, 637.76838824, 617.31166635, 628.14169559, 601.66829079, 601.66829079, 619.71833952, 521.04473983, 571.58487625, 566.77152993, 573.99154942, 528.26475932, 517.43473008, 565.56819335, 519.84140325, 512.62138376, 489.75798871, 519.84140325, 501.79135452, 511.41804717, 560.75484702, 567.97486651, 561.95818360, 582.41490549, 587.22825181, 601.66829079, 584.82157865, 546.31480804, 534.28144222, 535.48477881, 531.87476906, 541.50146171, 501.79135452, 469.30126682, 521.04473983, 504.19802769, 470.50460340, 456.06456442, 460.87791075, 433.20116937, 442.82786202, 451.25121810, 399.50774509, 393.49106218], 
							[2382.60643154, 2382.60643154, 2370.57306573, 2250.23940757, 2214.13931012, 1732.80467749, 1516.20409280, 1624.50438514, 1624.50438514, 1660.60448259, 1744.83804330, 1504.17072698, 1684.67121422, 1708.73794585, 1648.57111678, 1552.30419025, 1395.87043464, 1359.77033719, 1528.23745862, 1540.27082443, 1540.27082443, 1431.97053209, 1588.40428770, 1612.47101933, 1612.47101933, 1901.27179891, 2141.93911523, 2322.43960246, 2009.57209125, 1949.40526217, 2069.73892033, 1889.23843309, 1817.03823820, 1829.07160401, 1792.97150657, 1792.97150657, 1865.17170146, 1805.00487238, 1889.23843309, 2021.60545707, 2045.67218870, 2105.83901778, 2141.93911523, 2153.97248104, 2009.57209125, 2057.70555452, 2021.60545707, 1997.53872544, 2045.67218870, 2093.80565196, 2057.70555452, 2057.70555452, 2021.60545707, 2009.57209125, 2141.93911523, 2406.67316318, 2298.37287083, 2322.43960246, 2310.40623665, 2214.13931012, 2214.13931012, 2250.23940757, 2262.27277339, 2178.03921267, 2286.33950502, 2454.80662644, 2466.83999225, 2514.97345552, 2466.83999225, 2563.10691878, 2659.37384531, 2707.50730857, 2719.54067439, 2635.30711368, 2815.80760092, 2875.97442999, 2779.70750347, 2695.47394276, 2695.47394276, 2803.77423510, 2755.64077184, 2695.47394276, 2635.30711368, 2707.50730857, 2755.64077184, 2803.77423510, 2875.97442999, 2888.00779581, 2791.74086928, 2731.57404020, 2731.57404020, 2430.73989481, 2370.57306573, 2322.43960246, 2226.17267594, 2153.97248104, 2129.90574941, 2190.07257849, 2178.03921267, 2382.60643154, 2250.23940757, 2370.57306573, 2346.50633410, 2286.33950502, 2394.63979736, 2370.57306573, 2310.40623665, 2382.60643154, 2334.47296828, 2153.97248104, 2310.40623665, 2310.40623665, 2334.47296828, 2286.33950502, 2346.50633410, 2382.60643154, 2418.70652899, 2442.77326062, 2370.57306573, 2358.53969991, 2382.60643154, 2358.53969991, 2298.37287083, 2238.20604175, 2250.23940757, 2093.80565196, 2286.33950502, 2262.27277339, 2274.30613920, 2202.10594431, 2178.03921267, 2153.97248104, 2081.77228615, 2069.73892033, 2069.73892033, 2141.93911523, 2166.00584686, 2129.90574941, 2141.93911523, 2081.77228615, 2105.83901778, 2033.63882288, 2045.67218870, 1973.47199380, 2178.03921267, 2166.00584686, 2202.10594431, 2298.37287083, 2262.27277339, 2250.23940757, 2250.23940757, 2069.73892033, 1937.37189636, 2129.90574941, 2045.67218870, 2105.83901778, 2117.87238359, 2069.73892033, 2021.60545707, 1937.37189636, 1865.17170146, 1973.47199380, 1925.33853054, 1901.27179891, 2009.57209125, 2069.73892033, 2190.07257849, 2202.10594431, 2093.80565196, 2009.57209125, 1997.53872544, 1913.30516472, 1889.23843309, 1865.17170146, 1841.10496983, 2009.57209125, 1997.53872544, 1985.50535962, 2009.57209125, 1985.50535962, 1997.53872544, 2105.83901778, 2033.63882288, 2021.60545707, 2033.63882288, 2190.07257849, 2117.87238359, 2166.00584686, 2274.30613920, 2166.00584686, 2238.20604175, 2045.67218870, 2105.83901778, 2129.90574941, 2190.07257849, 2178.03921267, 2298.37287083, 2358.53969991, 2358.53969991, 2406.67316318, 2442.77326062, 2514.97345552, 2394.63979736, 2358.53969991, 2382.60643154, 2430.73989481, 2490.90672389, 2514.97345552, 2346.50633410, 2382.60643154, 2358.53969991, 2454.80662644, 2454.80662644, 2514.97345552, 2514.97345552, 2539.04018715, 2454.80662644, 2563.10691878, 2527.00682133, 2430.73989481, 2382.60643154, 2346.50633410, 2346.50633410, 2394.63979736, 2418.70652899, 2418.70652899, 2454.80662644, 2382.60643154, 2238.20604175, 2190.07257849, 2178.03921267, 2190.07257849, 2178.03921267, 2202.10594431, 2298.37287083, 2166.00584686, 2190.07257849, 2250.23940757, 2274.30613920, 2262.27277339, 2250.23940757, 2238.20604175, 2322.43960246, 2262.27277339, 2238.20604175, 2274.30613920, 2238.20604175, 2214.13931012, 2226.17267594, 2129.90574941, 2057.70555452, 2057.70555452, 2033.63882288, 2009.57209125, 1624.50438514, 1636.53775096, 1528.23745862, 1552.30419025, 1492.13736117, 1504.17072698, 1756.87140912, 1588.40428770, 1552.30419025, 1624.50438514, 1756.87140912, 1600.43765351, 1540.27082443, 1407.90380046, 1624.50438514, 1600.43765351, 1696.70458004, 1732.80467749, 1829.07160401, 1660.60448259, 1708.73794585, 1504.17072698, 1660.60448259, 1708.73794585, 1877.20506728, 1805.00487238, 1696.70458004, 1841.10496983, 1756.87140912, 1744.83804330, 1829.07160401, 1756.87140912, 1732.80467749, 1684.67121422, 1672.63784841, 1612.47101933, 1564.33755606], 
							[986.73599690, 974.70263109, 926.56916782, 825.48889497, 809.84551941, 809.84551941, 765.32206589, 722.00194895, 738.84866109, 770.13541222, 815.86220232, 759.30538298, 782.16877803, 802.62549992, 758.10204640, 756.89870982, 709.96858314, 666.64846620, 735.23865135, 740.05199768, 712.37525630, 676.27515885, 696.73188074, 651.00509064, 596.85494447, 661.83511987, 755.69537324, 801.42216334, 797.81215359, 805.03217308, 837.52226079, 844.74228027, 800.21882676, 795.40548043, 765.32206589, 788.18546094, 766.52540247, 728.01863186, 764.11872931, 815.86220232, 809.84551941, 758.10204640, 788.18546094, 779.76210487, 746.06868058, 755.69537324, 749.67869033, 717.18860263, 753.28870007, 765.32206589, 752.08536349, 764.11872931, 774.94875854, 772.54208538, 767.72873905, 783.37211461, 758.10204640, 790.59213410, 806.23550966, 841.13227053, 824.28555839, 848.35229002, 842.33560711, 832.70891446, 849.55562660, 906.11244594, 888.06239721, 857.97898267, 835.11558762, 815.86220232, 815.86220232, 823.08222181, 817.06553890, 785.77878778, 838.72559737, 830.30224130, 837.52226079, 830.30224130, 849.55562660, 871.21568507, 884.45238747, 898.89242645, 918.14581175, 921.75582150, 914.53580201, 888.06239721, 902.50243619, 849.55562660, 788.18546094, 789.38879752, 779.76210487, 777.35543171, 826.69223155, 806.23550966, 878.43570456, 813.45552915, 830.30224130, 888.06239721, 891.67240696, 953.04257262, 938.60253364, 944.61921655, 932.58585073, 896.48575328, 920.55248491, 926.56916782, 932.58585073, 963.87260185, 951.83923604, 972.29595792, 1007.19271879, 1030.05611384, 1010.80272853, 1009.59939195, 1003.58270904, 998.76936272, 1001.17603588, 999.97269930, 1010.80272853, 1021.63275777, 996.36268955, 962.66926527, 966.27927501, 956.65258236, 966.27927501, 954.24590920, 950.63589945, 961.46592869, 974.70263109, 956.65258236, 950.63589945, 954.24590920, 915.73913859, 926.56916782, 944.61921655, 928.97584099, 962.66926527, 937.39919706, 948.22922629, 943.41587996, 943.41587996, 919.34914833, 926.56916782, 898.89242645, 910.92579226, 924.16249466, 953.04257262, 961.46592869, 951.83923604, 956.65258236, 932.58585073, 913.33246543, 850.75896318, 902.50243619, 904.90910935, 930.17917757, 866.40233874, 842.33560711, 796.60881701, 812.25219257, 824.28555839, 779.76210487, 765.32206589, 813.45552915, 865.19900216, 826.69223155, 859.18231925, 833.91225104, 813.45552915, 782.16877803, 803.82883650, 821.87888522, 795.40548043, 827.89556813, 771.33874880, 843.53894369, 838.72559737, 853.16563635, 859.18231925, 837.52226079, 854.36897293, 980.71931399, 977.10930425, 943.41587996, 941.00920680, 938.60253364, 920.55248491, 914.53580201, 956.65258236, 951.83923604, 956.65258236, 1002.37937246, 1151.59310858, 1203.33658159, 1239.43667904, 1215.36994740, 1263.50341067, 1215.36994740, 1167.23648414, 1227.40331322, 1187.69320603, 1215.36994740, 1359.77033719, 1371.80370301, 1540.27082443, 1648.57111678, 1588.40428770, 1444.00389791, 1444.00389791, 1407.90380046, 1359.77033719, 1347.73697138, 1347.73697138, 1407.90380046, 1323.67023975, 1383.83706883, 1371.80370301, 1431.97053209, 1419.93716627, 1383.83706883, 1371.80370301, 1359.77033719, 1359.77033719, 1287.57014230, 1311.63687393, 1395.87043464, 1347.73697138, 1347.73697138, 1323.67023975, 1287.57014230, 1203.33658159, 1167.23648414, 1178.06651337, 1227.40331322, 1371.80370301, 1311.63687393, 1287.57014230, 1347.73697138, 1335.70360556, 1371.80370301, 1287.57014230, 1263.50341067, 1299.60350811, 1215.36994740, 1215.36994740, 1251.47004485, 1275.53677648, 1239.43667904, 1196.11656210, 1143.16975251, 1108.27299164, 1013.20940170, 963.87260185, 980.71931399, 860.38565584, 938.60253364, 892.87574354, 896.48575328, 844.74228027, 850.75896318, 901.29909961, 876.02903140, 888.06239721, 878.43570456, 913.33246543, 856.77564609, 782.16877803, 792.99880727, 837.52226079, 836.31892420, 866.40233874, 873.62235823, 1004.78604563, 944.61921655, 968.68594818, 1009.59939195, 1060.13952838, 1026.44610409, 993.95601639, 944.61921655, 950.63589945, 974.70263109, 983.12598716, 910.92579226, 910.92579226, 850.75896318, 867.60567532, 841.13227053, 837.52226079, 830.30224130, 832.70891446], 
							[635.36171508, 614.90499319, 586.02491523, 598.05828105, 565.56819335, 542.70479830, 525.85808615, 546.31480804, 545.11147146, 523.45141299, 553.53482753, 527.06142274, 530.67143248, 525.85808615, 496.97800820, 489.75798871, 470.50460340, 441.62452544, 498.18134478, 499.38468136, 464.48792049, 436.81117912, 456.06456442, 456.06456442, 428.38782305, 486.14797896, 553.53482753, 566.77152993, 543.90813488, 531.87476906, 533.07810564, 501.79135452, 498.18134478, 498.18134478, 500.58801794, 521.04473983, 505.40136427, 487.35131554, 521.04473983, 537.89145197, 511.41804717, 515.02805692, 522.24807641, 513.82472034, 499.38468136, 493.36799845, 483.74130580, 472.91127656, 510.21471059, 490.96132529, 477.72462289, 480.13129605, 481.33463264, 447.64120835, 509.01137401, 554.73816411, 566.77152993, 637.76838824, 637.76838824, 630.54836875, 637.76838824, 723.20528553, 713.57859288, 693.12187099, 694.32520758, 705.15523681, 702.74856365, 691.91853441, 671.46181253, 669.05513936, 640.17506140, 652.20842722, 685.90185151, 651.00509064, 681.08850518, 682.29184176, 750.88202691, 741.25533426, 759.30538298, 783.37211461, 764.11872931, 738.84866109, 756.89870982, 740.05199768, 762.91539273, 740.05199768, 744.86534400, 737.64532451, 707.56190997, 695.52854416, 658.22511013, 614.90499319, 629.34503217, 628.14169559, 702.74856365, 635.36171508, 646.19174431, 675.07182227, 728.01863186, 766.52540247, 730.42530502, 817.06553890, 800.21882676, 806.23550966, 839.92893395, 849.55562660, 842.33560711, 842.33560711, 772.54208538, 836.31892420, 842.33560711, 841.13227053, 820.67554864, 815.86220232, 842.33560711, 853.16563635, 837.52226079, 866.40233874, 871.21568507, 908.51911910, 890.46907037, 876.02903140, 895.28241670, 876.02903140, 884.45238747, 872.41902165, 878.43570456, 866.40233874, 862.79232900, 855.57230951, 849.55562660, 859.18231925, 836.31892420, 835.11558762, 842.33560711, 872.41902165, 868.80901191, 904.90910935, 892.87574354, 891.67240696, 895.28241670, 884.45238747, 896.48575328, 872.41902165, 833.91225104, 837.52226079, 844.74228027, 867.60567532, 854.36897293, 844.74228027, 837.52226079, 788.18546094, 756.89870982, 770.13541222, 712.37525630, 663.03845645, 673.86848569, 691.91853441, 667.85180278, 646.19174431, 602.87162738, 637.76838824, 622.12501268, 665.44512962, 705.15523681, 685.90185151, 682.29184176, 670.25847594, 663.03845645, 623.32834926, 620.92167610, 618.51500294, 589.63492498, 625.73502243, 595.65160789, 660.63178329, 632.95504192, 631.75170533, 625.73502243, 646.19174431, 671.46181253, 700.34189048, 700.34189048, 728.01863186, 715.98526604, 761.71205615, 691.91853441, 735.23865135, 752.08536349, 723.20528553, 742.45867084, 723.20528553, 728.01863186, 744.86534400, 771.33874880, 778.55876829, 837.52226079, 841.13227053, 835.11558762, 845.94561686, 836.31892420, 908.51911910, 872.41902165, 854.36897293, 880.84237772, 878.43570456, 851.96229976, 856.77564609, 842.33560711, 850.75896318, 838.72559737, 830.30224130, 870.01234849, 902.50243619, 919.34914833, 953.04257262, 941.00920680, 960.26259211, 932.58585073, 931.38251415, 938.60253364, 914.53580201, 933.78918731, 904.90910935, 919.34914833, 916.94247517, 921.75582150, 920.55248491, 896.48575328, 848.35229002, 839.92893395, 837.52226079, 827.89556813, 814.65886573, 832.70891446, 792.99880727, 777.35543171, 805.03217308, 803.82883650, 827.89556813, 833.91225104, 815.86220232, 860.38565584, 836.31892420, 823.08222181, 851.96229976, 836.31892420, 837.52226079, 830.30224130, 791.79547068, 797.81215359, 776.15209512, 734.03531477, 748.47535375, 629.34503217, 651.00509064, 642.58173457, 684.69851492, 679.88516860, 703.95190023, 714.78192946, 714.78192946, 684.69851492, 702.74856365, 752.08536349, 725.61195870, 706.35857339, 702.74856365, 660.63178329, 703.95190023, 717.18860263, 728.01863186, 764.11872931, 758.10204640, 713.57859288, 740.05199768, 749.67869033, 729.22196844, 712.37525630, 679.88516860, 671.46181253, 678.68183202, 685.90185151, 635.36171508, 625.73502243, 614.90499319, 608.88831028, 598.05828105, 610.09164687, 559.55151044, 547.51814462], 
							[524.65474957, 524.65474957, 502.99469110, 478.92795947, 472.91127656, 447.64120835, 416.35445723, 415.15112065, 417.55779381, 457.26790100, 475.31794973, 459.67457417, 458.47123758, 471.70793998, 452.45455468, 439.21785228, 417.55779381, 397.10107192, 421.16780356, 450.04788151, 433.20116937, 398.30440851, 421.16780356, 406.72776458, 371.83100371, 391.08438902, 535.48477881, 511.41804717, 499.38468136, 492.16466187, 490.96132529, 476.52128631, 447.64120835, 446.43787177, 434.40450595, 430.79449621, 425.98114988, 393.49106218, 410.33777432, 416.35445723, 403.11775483, 421.16780356, 418.76113039, 416.35445723, 393.49106218, 392.28772560, 394.69439876, 377.84768662, 417.55779381, 409.13443774, 398.30440851, 397.10107192, 383.86436953, 382.66103294, 442.82786202, 453.65789126, 450.04788151, 492.16466187, 489.75798871, 492.16466187, 476.52128631, 504.19802769, 505.40136427, 502.99469110, 513.82472034, 559.55151044, 565.56819335, 547.51814462, 505.40136427, 504.19802769, 493.36799845, 511.41804717, 513.82472034, 494.57133503, 501.79135452, 493.36799845, 500.58801794, 510.21471059, 510.21471059, 521.04473983, 511.41804717, 502.99469110, 507.80803743, 483.74130580, 475.31794973, 465.69125707, 454.86122784, 450.04788151, 409.13443774, 404.32109141, 382.66103294, 354.98429157, 373.03434029, 357.39096473, 354.98429157, 341.74758917, 353.78095499, 353.78095499, 394.69439876, 403.11775483, 377.84768662, 397.10107192, 401.91441825, 368.22099397, 419.96446697, 438.01451570, 435.60784253, 453.65789126, 451.25121810, 433.20116937, 451.25121810, 457.26790100, 444.03119861, 445.23453519, 442.82786202, 463.28458391, 463.28458391, 474.11461315, 474.11461315, 487.35131554, 499.38468136, 488.55465212, 474.11461315, 469.30126682, 478.92795947, 457.26790100, 454.86122784, 480.13129605, 489.75798871, 505.40136427, 493.36799845, 490.96132529, 482.53796922, 490.96132529, 495.77467161, 486.14797896, 502.99469110, 510.21471059, 510.21471059, 487.35131554, 500.58801794, 483.74130580, 484.94464238, 464.48792049, 468.09793024, 459.67457417, 448.84454493, 458.47123758, 463.28458391, 466.89459366, 457.26790100, 463.28458391, 465.69125707, 471.70793998, 435.60784253, 421.16780356, 421.16780356, 411.54111090, 404.32109141, 400.71108167, 388.67771585, 433.20116937, 409.13443774, 411.54111090, 428.38782305, 416.35445723, 423.57447672, 394.69439876, 398.30440851, 375.44101346, 393.49106218, 354.98429157, 346.56093550, 365.81432080, 356.18762815, 394.69439876, 404.32109141, 409.13443774, 394.69439876, 398.30440851, 392.28772560, 409.13443774, 421.16780356, 411.54111090, 412.74444748, 438.01451570, 399.50774509, 403.11775483, 415.15112065, 394.69439876, 421.16780356, 415.15112065, 417.55779381, 418.76113039, 415.15112065, 434.40450595, 476.52128631, 464.48792049, 464.48792049, 459.67457417, 469.30126682, 492.16466187, 471.70793998, 483.74130580, 487.35131554, 510.21471059, 511.41804717, 496.97800820, 489.75798871, 493.36799845, 493.36799845, 450.04788151, 471.70793998, 498.18134478, 502.99469110, 529.46809590, 521.04473983, 513.82472034, 511.41804717, 510.21471059, 513.82472034, 494.57133503, 490.96132529, 494.57133503, 513.82472034, 507.80803743, 501.79135452, 492.16466187, 484.94464238, 482.53796922, 471.70793998, 469.30126682, 463.28458391, 476.52128631, 463.28458391, 457.26790100, 458.47123758, 463.28458391, 464.48792049, 469.30126682, 454.86122784, 439.21785228, 452.45455468, 445.23453519, 444.03119861, 452.45455468, 452.45455468, 441.62452544, 428.38782305, 404.32109141, 407.93110116, 399.50774509, 393.49106218, 394.69439876, 375.44101346, 391.08438902, 370.62766713, 382.66103294, 393.49106218, 423.57447672, 430.79449621, 405.52442800, 401.91441825, 391.08438902, 395.89773534, 380.25435978, 386.27104269, 371.83100371, 399.50774509, 391.08438902, 398.30440851, 405.52442800, 404.32109141, 412.74444748, 397.10107192, 405.52442800, 409.13443774, 407.93110116, 397.10107192, 399.50774509, 386.27104269, 383.86436953, 383.86436953, 344.15426233, 338.13757943, 327.30755019, 326.10421361, 324.90087703, 341.74758917, 323.69754045, 306.85082830], 
							[3453.57598916, 3429.50925753, 3285.10886773, 3357.30906263, 3297.14223355, 3285.10886773, 3381.37579426, 3429.50925753, 3453.57598916, 3477.64272079, 3585.94301313, 3465.60935497, 3477.64272079, 3441.54262334, 3381.37579426, 3405.44252589, 3321.20896518, 3357.30906263, 3345.27569681, 3429.50925753, 3285.10886773, 3056.47491723, 3104.60838050, 3020.37481979, 2984.27472234, 3297.14223355, 3381.37579426, 3477.64272079, 3273.07550192, 3152.74184376, 3176.80857539, 2984.27472234, 2972.24135652, 3140.70847794, 3092.57501468, 3080.54164886, 3032.40818560, 2936.14125907, 3008.34145397, 3140.70847794, 3140.70847794, 3164.77520958, 3212.90867284, 3309.17559937, 3176.80857539, 3140.70847794, 3068.50828305, 3068.50828305, 3092.57501468, 3080.54164886, 2996.30808815, 3032.40818560, 3032.40818560, 2972.24135652, 3092.57501468, 3080.54164886, 3224.94203866, 3056.47491723, 3345.27569681, 3357.30906263, 3273.07550192, 3489.67608660, 3405.44252589, 3297.14223355, 3261.04213610, 3236.97540447, 3261.04213610, 3104.60838050, 3152.74184376, 3008.34145397, 3104.60838050, 3104.60838050, 3128.67511213, 3080.54164886, 3116.64174631, 3080.54164886, 3152.74184376, 3116.64174631, 3116.64174631, 3128.67511213, 3092.57501468, 3188.84194121, 3224.94203866, 3176.80857539, 3116.64174631, 3020.37481979, 3032.40818560, 3044.44155142, 2984.27472234, 2984.27472234, 2996.30808815, 2839.87433255, 2996.30808815, 2924.10789326, 3032.40818560, 2900.04116163, 2972.24135652, 3116.64174631, 3104.60838050, 3020.37481979, 3080.54164886, 2972.24135652, 2996.30808815, 3020.37481979, 3176.80857539, 2924.10789326, 2924.10789326, 2900.04116163, 2791.74086928, 2647.34047949, 2454.80662644, 2683.44057694, 2659.37384531, 2635.30711368, 2623.27374786, 2599.20701623, 2671.40721112, 2647.34047949, 2659.37384531, 2623.27374786, 2803.77423510, 2659.37384531, 2611.24038205, 2731.57404020, 2755.64077184, 2695.47394276, 2659.37384531, 2671.40721112, 2671.40721112, 2623.27374786, 2659.37384531, 2707.50730857, 2683.44057694, 2779.70750347, 2647.34047949, 2683.44057694, 2635.30711368, 2671.40721112, 2755.64077184, 2743.60740602, 2707.50730857, 2707.50730857, 2707.50730857, 2671.40721112, 2647.34047949, 2731.57404020, 2743.60740602, 2755.64077184, 2779.70750347, 2791.74086928, 2803.77423510, 2683.44057694, 2683.44057694, 2659.37384531, 2671.40721112, 2623.27374786, 2418.70652899, 2418.70652899, 2358.53969991, 2322.43960246, 2334.47296828, 2478.87335807, 2346.50633410, 2382.60643154, 2514.97345552, 2478.87335807, 2466.83999225, 2466.83999225, 2406.67316318, 2430.73989481, 2478.87335807, 2358.53969991, 2370.57306573, 2454.80662644, 2418.70652899, 2466.83999225, 2587.17365041, 2635.30711368, 2779.70750347, 2755.64077184, 2863.94106418, 3020.37481979, 2863.94106418, 2912.07452744, 3032.40818560, 3056.47491723, 2839.87433255, 2972.24135652, 3152.74184376, 3152.74184376, 3044.44155142, 2972.24135652, 3116.64174631, 3164.77520958, 3128.67511213, 3080.54164886, 3056.47491723, 3236.97540447, 3417.47589171, 3622.04311058, 3585.94301313, 3838.64369527, 3694.24330547, 3730.34340292, 3646.10984221, 3730.34340292, 3766.44350037, 3658.14320803, 3597.97637895, 3597.97637895, 3561.87628150, 3513.74281824, 3585.94301313, 3682.20993966, 3826.61032945, 3814.57696363, 3778.47686619, 3922.87725598, 3742.37676874, 3706.27667129, 3934.91062179, 4055.24427995, 4043.21091413, 3946.94398761, 4043.21091413, 4151.51120648, 4163.54457229, 4175.57793811, 3946.94398761, 3910.84389016, 3802.54359782, 4019.14418250, 3838.64369527, 3934.91062179, 4007.11081669, 3934.91062179, 3838.64369527, 3922.87725598, 4007.11081669, 4091.34437740, 4079.31101158, 4043.21091413, 4187.61130393, 3934.91062179, 3946.94398761, 4079.31101158, 4043.21091413, 4115.41110903, 4295.91159627, 4584.71237585, 4560.64564422, 4295.91159627, 4596.74574167, 4596.74574167, 4079.31101158, 4344.04505953, 4319.97832790, 4428.27862024, 4440.31198606, 5090.11374012, 4825.37969217, 4861.47978961, 4861.47978961, 4837.41305798, 4993.84681359, 4921.64661869, 5210.44739828, 5066.04700848, 5186.38066664, 5703.81539673, 5980.58281049, 5896.34924978, 5535.34827530, 5559.41500694, 5427.04798296, 5715.84876254, 5427.04798296, 5246.54749572, 5595.51510438, 5475.18144622, 5643.64856765, 5812.11568907, 5691.78203091, 5896.34924978, 5655.68193346, 5956.51607886, 5535.34827530, 5583.48173857, 5751.94885999, 5571.44837275, 5836.18242070], 
							[419.96446697, 398.30440851, 391.08438902, 391.08438902, 375.44101346, 369.42433055, 356.18762815, 332.12089652, 346.56093550, 347.76427208, 353.78095499, 352.57761841, 373.03434029, 382.66103294, 364.61098422, 353.78095499, 324.90087703, 327.30755019, 342.95092575, 334.52756968, 344.15426233, 326.10421361, 332.12089652, 300.83414540, 252.70068213, 299.63080882, 377.84768662, 362.20431106, 367.01765738, 376.64435004, 380.25435978, 352.57761841, 358.59430131, 354.98429157, 364.61098422, 338.13757943, 336.93424284, 318.88419412, 322.49420387, 334.52756968, 330.91755994, 352.57761841, 344.15426233, 340.54425259, 333.32423310, 329.71422336, 320.08753070, 329.71422336, 346.56093550, 346.56093550, 346.56093550, 336.93424284, 334.52756968, 332.12089652, 345.35759892, 348.96760866, 371.83100371, 391.08438902, 419.96446697, 424.77781330, 434.40450595, 494.57133503, 487.35131554, 480.13129605, 466.89459366, 476.52128631, 466.89459366, 454.86122784, 433.20116937, 416.35445723, 421.16780356, 416.35445723, 441.62452544, 421.16780356, 429.59115963, 416.35445723, 421.16780356, 417.55779381, 421.16780356, 423.57447672, 419.96446697, 397.10107192, 382.66103294, 382.66103294, 381.45769636, 401.91441825, 412.74444748, 388.67771585, 347.76427208, 354.98429157, 321.29086728, 306.85082830, 320.08753070, 306.85082830, 317.68085754, 305.64749172, 318.88419412, 317.68085754, 342.95092575, 347.76427208, 332.12089652, 379.05102320, 363.40764764, 371.83100371, 375.44101346, 376.64435004, 409.13443774, 425.98114988, 421.16780356, 392.28772560, 418.76113039, 428.38782305, 411.54111090, 403.11775483, 401.91441825, 427.18448646, 422.37114014, 435.60784253, 454.86122784, 453.65789126, 450.04788151, 429.59115963, 419.96446697, 410.33777432, 406.72776458, 394.69439876, 389.88105243, 428.38782305, 450.04788151, 445.23453519, 436.81117912, 445.23453519, 430.79449621, 415.15112065, 424.77781330, 429.59115963, 418.76113039, 439.21785228, 430.79449621, 424.77781330, 430.79449621, 415.15112065, 417.55779381, 389.88105243, 403.11775483, 400.71108167, 397.10107192, 423.57447672, 419.96446697, 412.74444748, 415.15112065, 405.52442800, 391.08438902, 409.13443774, 373.03434029, 393.49106218, 365.81432080, 362.20431106, 354.98429157, 348.96760866, 344.15426233, 350.17094524, 348.96760866, 356.18762815, 369.42433055, 369.42433055, 377.84768662, 361.00097448, 348.96760866, 330.91755994, 327.30755019, 312.86751121, 305.64749172, 293.61412591, 300.83414540, 336.93424284, 341.74758917, 336.93424284, 333.32423310, 340.54425259, 335.73090626, 351.37428182, 352.57761841, 351.37428182, 359.79763789, 363.40764764, 329.71422336, 344.15426233, 351.37428182, 350.17094524, 350.17094524, 342.95092575, 356.18762815, 345.35759892, 351.37428182, 358.59430131, 374.23767687, 373.03434029, 371.83100371, 376.64435004, 389.88105243, 406.72776458, 397.10107192, 383.86436953, 403.11775483, 422.37114014, 419.96446697, 403.11775483, 394.69439876, 385.06770611, 370.62766713, 385.06770611, 380.25435978, 407.93110116, 409.13443774, 428.38782305, 435.60784253, 411.54111090, 409.13443774, 395.89773534, 410.33777432, 389.88105243, 418.76113039, 400.71108167, 403.11775483, 407.93110116, 406.72776458, 403.11775483, 395.89773534, 380.25435978, 373.03434029, 370.62766713, 363.40764764, 380.25435978, 370.62766713, 375.44101346, 364.61098422, 367.01765738, 367.01765738, 380.25435978, 371.83100371, 371.83100371, 382.66103294, 364.61098422, 364.61098422, 374.23767687, 365.81432080, 368.22099397, 342.95092575, 324.90087703, 335.73090626, 322.49420387, 314.07084779, 320.08753070, 285.19076984, 304.44415514, 268.34405769, 285.19076984, 267.14072111, 271.95406744, 324.90087703, 318.88419412, 309.25750147, 316.47752096, 324.90087703, 327.30755019, 336.93424284, 328.51088677, 353.78095499, 356.18762815, 373.03434029, 374.23767687, 376.64435004, 362.20431106, 354.98429157, 361.00097448, 354.98429157, 340.54425259, 346.56093550, 356.18762815, 342.95092575, 326.10421361, 316.47752096, 297.22413565, 282.78409667, 294.81746249, 283.98743325, 281.58076009, 285.19076984, 273.15740402, 246.68399923], 
							[722.00194895, 706.35857339, 693.12187099, 709.96858314, 702.74856365, 657.02177355, 623.32834926, 653.41176380, 655.81843697, 652.20842722, 709.96858314, 701.54522707, 717.18860263, 685.90185151, 696.73188074, 691.91853441, 643.78507115, 588.43158840, 647.39508089, 620.92167610, 623.32834926, 589.63492498, 584.82157865, 555.94150069, 515.02805692, 577.60155916, 673.86848569, 649.80175406, 634.15837850, 648.59841748, 653.41176380, 641.37839799, 653.41176380, 655.81843697, 637.76838824, 622.12501268, 617.31166635, 593.24493472, 623.32834926, 625.73502243, 622.12501268, 611.29498345, 606.48163712, 604.07496396, 592.04159814, 578.80489574, 581.21156891, 572.78821284, 634.15837850, 628.14169559, 610.09164687, 612.49832003, 592.04159814, 589.63492498, 601.66829079, 612.49832003, 632.95504192, 676.27515885, 643.78507115, 647.39508089, 654.61510038, 689.51186125, 684.69851492, 669.05513936, 725.61195870, 741.25533426, 758.10204640, 755.69537324, 719.59527579, 697.93521732, 676.27515885, 690.71519783, 702.74856365, 681.08850518, 693.12187099, 696.73188074, 731.62864161, 711.17191972, 724.40862212, 714.78192946, 703.95190023, 694.32520758, 675.07182227, 673.86848569, 708.76524656, 681.08850518, 703.95190023, 682.29184176, 665.44512962, 655.81843697, 625.73502243, 580.00823233, 605.27830054, 587.22825181, 604.07496396, 573.99154942, 605.27830054, 613.70165661, 654.61510038, 660.63178329, 637.76838824, 694.32520758, 679.88516860, 648.59841748, 654.61510038, 673.86848569, 655.81843697, 678.68183202, 637.76838824, 619.71833952, 637.76838824, 665.44512962, 659.42844671, 644.98840773, 643.78507115, 646.19174431, 657.02177355, 682.29184176, 673.86848569, 703.95190023, 694.32520758, 666.64846620, 653.41176380, 653.41176380, 655.81843697, 652.20842722, 637.76838824, 649.80175406, 667.85180278, 665.44512962, 665.44512962, 642.58173457, 637.76838824, 626.93835901, 635.36171508, 641.37839799, 649.80175406, 658.22511013, 659.42844671, 666.64846620, 673.86848569, 655.81843697, 653.41176380, 632.95504192, 634.15837850, 643.78507115, 636.56505166, 657.02177355, 641.37839799, 641.37839799, 641.37839799, 602.87162738, 575.19488600, 583.61824207, 577.60155916, 555.94150069, 524.65474957, 541.50146171, 511.41804717, 502.99469110, 475.31794973, 487.35131554, 475.31794973, 481.33463264, 492.16466187, 484.94464238, 489.75798871, 475.31794973, 429.59115963, 440.42118886, 450.04788151, 429.59115963, 409.13443774, 425.98114988, 411.54111090, 463.28458391, 482.53796922, 464.48792049, 468.09793024, 465.69125707, 506.60470085, 588.43158840, 531.87476906, 547.51814462, 561.95818360, 610.09164687, 561.95818360, 582.41490549, 619.71833952, 601.66829079, 604.07496396, 595.65160789, 605.27830054, 601.66829079, 618.51500294, 622.12501268, 628.14169559, 636.56505166, 623.32834926, 630.54836875, 643.78507115, 665.44512962, 658.22511013, 646.19174431, 643.78507115, 669.05513936, 673.86848569, 671.46181253, 659.42844671, 641.37839799, 636.56505166, 637.76838824, 673.86848569, 693.12187099, 685.90185151, 715.98526604, 705.15523681, 700.34189048, 677.47849543, 689.51186125, 696.73188074, 688.30852467, 720.79861237, 708.76524656, 712.37525630, 724.40862212, 747.27201717, 738.84866109, 728.01863186, 712.37525630, 709.96858314, 705.15523681, 673.86848569, 665.44512962, 678.68183202, 637.76838824, 659.42844671, 666.64846620, 667.85180278, 685.90185151, 675.07182227, 660.63178329, 695.52854416, 681.08850518, 661.83511987, 667.85180278, 654.61510038, 642.58173457, 634.15837850, 589.63492498, 577.60155916, 564.36485676, 539.09478855, 541.50146171, 487.35131554, 545.11147146, 529.46809590, 553.53482753, 567.97486651, 598.05828105, 625.73502243, 581.21156891, 541.50146171, 529.46809590, 553.53482753, 558.34817386, 542.70479830, 525.85808615, 540.29812513, 530.67143248, 582.41490549, 581.21156891, 605.27830054, 589.63492498, 535.48477881, 540.29812513, 540.29812513, 523.45141299, 542.70479830, 530.67143248, 517.43473008, 502.99469110, 511.41804717, 481.33463264, 499.38468136, 477.72462289, 476.52128631, 464.48792049, 460.87791075, 419.96446697, 357.39096473], 
							[744.86534400, 700.34189048, 673.86848569, 691.91853441, 740.05199768, 715.98526604, 699.13855390, 691.91853441, 732.83197819, 734.03531477, 762.91539273, 765.32206589, 758.10204640, 764.11872931, 770.13541222, 758.10204640, 767.72873905, 756.89870982, 776.15209512, 770.13541222, 746.06868058, 717.18860263, 709.96858314, 709.96858314, 681.08850518, 719.59527579, 818.26887548, 823.08222181, 790.59213410, 792.99880727, 799.01549017, 753.28870007, 776.15209512, 788.18546094, 758.10204640, 744.86534400, 737.64532451, 726.81529528, 770.13541222, 784.57545120, 777.35543171, 788.18546094, 794.20214385, 799.01549017, 785.77878778, 783.37211461, 740.05199768, 726.81529528, 767.72873905, 734.03531477, 729.22196844, 740.05199768, 734.03531477, 718.39193921, 726.81529528, 744.86534400, 755.69537324, 805.03217308, 830.30224130, 833.91225104, 817.06553890, 854.36897293, 831.50557788, 830.30224130, 837.52226079, 820.67554864, 812.25219257, 817.06553890, 796.60881701, 773.74542196, 766.52540247, 776.15209512, 774.94875854, 766.52540247, 786.98212436, 767.72873905, 789.38879752, 783.37211461, 772.54208538, 773.74542196, 761.71205615, 752.08536349, 742.45867084, 748.47535375, 773.74542196, 778.55876829, 788.18546094, 770.13541222, 760.50871956, 729.22196844, 722.00194895, 719.59527579, 732.83197819, 722.00194895, 744.86534400, 712.37525630, 688.30852467, 734.03531477, 743.66200742, 776.15209512, 758.10204640, 794.20214385, 794.20214385, 780.96544145, 771.33874880, 805.03217308, 791.79547068, 823.08222181, 832.70891446, 782.16877803, 833.91225104, 841.13227053, 832.70891446, 851.96229976, 859.18231925, 845.94561686, 833.91225104, 859.18231925, 878.43570456, 919.34914833, 926.56916782, 891.67240696, 900.09576303, 914.53580201, 922.95915808, 914.53580201, 897.68908986, 932.58585073, 928.97584099, 926.56916782, 924.16249466, 937.39919706, 914.53580201, 902.50243619, 909.72245568, 926.56916782, 921.75582150, 938.60253364, 918.14581175, 915.73913859, 912.12912884, 916.94247517, 902.50243619, 890.46907037, 860.38565584, 837.52226079, 839.92893395, 878.43570456, 872.41902165, 865.19900216, 835.11558762, 791.79547068, 765.32206589, 748.47535375, 752.08536349, 750.88202691, 743.66200742, 746.06868058, 695.52854416, 702.74856365, 655.81843697, 708.76524656, 681.08850518, 697.93521732, 719.59527579, 706.35857339, 722.00194895, 687.10518809, 671.46181253, 659.42844671, 658.22511013, 642.58173457, 622.12501268, 634.15837850, 634.15837850, 685.90185151, 694.32520758, 667.85180278, 669.05513936, 667.85180278, 691.91853441, 735.23865135, 746.06868058, 758.10204640, 758.10204640, 758.10204640, 700.34189048, 724.40862212, 765.32206589, 764.11872931, 770.13541222, 729.22196844, 766.52540247, 737.64532451, 768.93207563, 796.60881701, 808.64218283, 808.64218283, 805.03217308, 806.23550966, 818.26887548, 836.31892420, 821.87888522, 818.26887548, 827.89556813, 859.18231925, 837.52226079, 841.13227053, 837.52226079, 823.08222181, 808.64218283, 794.20214385, 808.64218283, 832.70891446, 866.40233874, 895.28241670, 862.79232900, 857.97898267, 848.35229002, 859.18231925, 883.24905089, 868.80901191, 878.43570456, 896.48575328, 926.56916782, 916.94247517, 909.72245568, 916.94247517, 879.63904114, 878.43570456, 879.63904114, 860.38565584, 842.33560711, 863.99566558, 870.01234849, 814.65886573, 819.47221206, 841.13227053, 832.70891446, 831.50557788, 838.72559737, 836.31892420, 842.33560711, 842.33560711, 836.31892420, 853.16563635, 854.36897293, 863.99566558, 883.24905089, 838.72559737, 873.62235823, 836.31892420, 854.36897293, 860.38565584, 800.21882676, 845.94561686, 818.26887548, 830.30224130, 811.04885599, 889.26573379, 872.41902165, 839.92893395, 813.45552915, 805.03217308, 862.79232900, 862.79232900, 860.38565584, 895.28241670, 922.95915808, 948.22922629, 955.44924578, 932.58585073, 953.04257262, 962.66926527, 953.04257262, 953.04257262, 974.70263109, 963.87260185, 997.56602614, 992.75267981, 977.10930425, 959.05925553, 967.48261160, 927.77250440, 953.04257262, 932.58585073, 998.76936272, 962.66926527, 949.43256287, 908.51911910, 837.52226079], 
							[922.95915808, 878.43570456, 854.36897293, 926.56916782, 836.31892420, 860.38565584, 842.33560711, 685.90185151, 734.03531477, 687.10518809, 760.50871956, 752.08536349, 746.06868058, 736.44198793, 706.35857339, 667.85180278, 613.70165661, 571.58487625, 625.73502243, 688.30852467, 722.00194895, 673.86848569, 722.00194895, 758.10204640, 756.89870982, 741.25533426, 829.09890471, 833.91225104, 819.47221206, 901.29909961, 909.72245568, 747.27201717, 735.23865135, 738.84866109, 701.54522707, 671.46181253, 712.37525630, 659.42844671, 746.06868058, 762.91539273, 746.06868058, 776.15209512, 782.16877803, 768.93207563, 709.96858314, 697.93521732, 717.18860263, 709.96858314, 758.10204640, 776.15209512, 752.08536349, 782.16877803, 812.25219257, 823.08222181, 962.66926527, 991.54934323, 981.92265058, 980.71931399, 908.51911910, 992.75267981, 950.63589945, 1037.27613333, 974.70263109, 962.66926527, 961.46592869, 915.73913859, 918.14581175, 910.92579226, 896.48575328, 853.16563635, 836.31892420, 829.09890471, 854.36897293, 849.55562660, 896.48575328, 871.21568507, 841.13227053, 890.46907037, 897.68908986, 878.43570456, 854.36897293, 837.52226079, 847.14895344, 817.06553890, 821.87888522, 827.89556813, 853.16563635, 848.35229002, 823.08222181, 811.04885599, 782.16877803, 740.05199768, 818.26887548, 831.50557788, 872.41902165, 786.98212436, 766.52540247, 854.36897293, 866.40233874, 890.46907037, 885.65572405, 960.26259211, 907.31578252, 925.36583124, 995.15935297, 1083.00292343, 1083.00292343, 1076.98624052, 1028.85277726, 985.53266032, 1034.86946017, 1127.52637695, 1046.90282598, 1044.49615282, 1024.03943093, 1101.05297215, 1097.44296241, 1122.71303062, 1117.89968430, 1202.13324501, 1239.43667904, 1167.23648414, 1152.79644516, 1156.40645491, 1161.21980123, 1191.30321577, 1161.21980123, 1162.42313781, 1146.77976225, 1132.33972327, 1121.50969404, 1162.42313781, 1126.32304037, 1156.40645491, 1185.28653286, 1147.98309883, 1190.09987919, 1191.30321577, 1182.87985970, 1149.18643542, 1143.16975251, 1113.08633797, 1127.52637695, 1110.67966481, 1081.79958685, 1096.23962583, 1139.55974276, 1178.06651337, 1162.42313781, 1155.20311832, 1168.43982072, 1134.74639644, 1101.05297215, 1093.83295266, 1064.95287471, 989.14267007, 984.32932374, 914.53580201, 884.45238747, 897.68908986, 878.43570456, 835.11558762, 806.23550966, 888.06239721, 965.07593843, 904.90910935, 916.94247517, 833.91225104, 842.33560711, 806.23550966, 841.13227053, 782.16877803, 753.28870007, 841.13227053, 788.18546094, 912.12912884, 980.71931399, 962.66926527, 939.80587022, 918.14581175, 896.48575328, 1042.08947966, 1037.27613333, 1046.90282598, 1037.27613333, 1054.12284547, 953.04257262, 938.60253364, 984.32932374, 962.66926527, 979.51597741, 854.36897293, 950.63589945, 901.29909961, 961.46592869, 926.56916782, 956.65258236, 965.07593843, 986.73599690, 1049.30949914, 1045.69948940, 1056.52951863, 1105.86631848, 1138.35640618, 1150.38977200, 1139.55974276, 1108.27299164, 1149.18643542, 1074.57956736, 1060.13952838, 1093.83295266, 1119.10302088, 1140.76307935, 1143.16975251, 1166.03314756, 1109.47632822, 1155.20311832, 1133.54305986, 1092.62961608, 1096.23962583, 1152.79644516, 1184.08319628, 1170.84649388, 1173.25316705, 1215.36994740, 1383.83706883, 1371.80370301, 1347.73697138, 1251.47004485, 1251.47004485, 1263.50341067, 1251.47004485, 1191.30321577, 1185.28653286, 1203.33658159, 1131.13638669, 1128.72971353, 1251.47004485, 1287.57014230, 1335.70360556, 1311.63687393, 1251.47004485, 1311.63687393, 1239.43667904, 1168.43982072, 1191.30321577, 1162.42313781, 1155.20311832, 1067.35954787, 980.71931399, 1039.68280649, 990.34600665, 975.90596767, 989.14267007, 765.32206589, 819.47221206, 838.72559737, 863.99566558, 827.89556813, 841.13227053, 906.11244594, 874.82569481, 848.35229002, 868.80901191, 949.43256287, 877.23236798, 850.75896318, 872.41902165, 888.06239721, 878.43570456, 891.67240696, 888.06239721, 950.63589945, 920.55248491, 902.50243619, 936.19586048, 908.51911910, 890.46907037, 884.45238747, 832.70891446, 777.35543171, 819.47221206, 777.35543171, 734.03531477, 691.91853441, 654.61510038, 634.15837850, 629.34503217, 669.05513936, 616.10832977, 626.93835901], 
							[1010.80272853, 934.99252389, 920.55248491, 913.33246543, 938.60253364, 931.38251415, 974.70263109, 1001.17603588, 1016.81941144, 1009.59939195, 1046.90282598, 968.68594818, 984.32932374, 968.68594818, 956.65258236, 942.21254338, 942.21254338, 891.67240696, 938.60253364, 969.88928476, 944.61921655, 896.48575328, 870.01234849, 890.46907037, 896.48575328, 967.48261160, 980.71931399, 985.53266032, 945.82255313, 942.21254338, 916.94247517, 874.82569481, 901.29909961, 900.09576303, 857.97898267, 862.79232900, 863.99566558, 849.55562660, 854.36897293, 870.01234849, 866.40233874, 896.48575328, 907.31578252, 902.50243619, 904.90910935, 910.92579226, 885.65572405, 848.35229002, 872.41902165, 860.38565584, 836.31892420, 813.45552915, 831.50557788, 841.13227053, 910.92579226, 936.19586048, 996.36268955, 1018.02274802, 1056.52951863, 1045.69948940, 980.71931399, 1127.52637695, 1113.08633797, 1068.56288445, 1064.95287471, 1022.83609435, 1063.74953812, 1045.69948940, 995.15935297, 963.87260185, 972.29595792, 1012.00606512, 1063.74953812, 996.36268955, 1045.69948940, 1019.22608460, 1034.86946017, 1016.81941144, 1021.63275777, 1062.54620154, 1045.69948940, 1002.37937246, 1010.80272853, 986.73599690, 1012.00606512, 1013.20940170, 1038.47946991, 1036.07279675, 992.75267981, 975.90596767, 965.07593843, 915.73913859, 949.43256287, 974.70263109, 1004.78604563, 973.49929450, 989.14267007, 1009.59939195, 1016.81941144, 1032.46278700, 1002.37937246, 1052.91950889, 1083.00292343, 1069.76622103, 1127.52637695, 1113.08633797, 1119.10302088, 1169.64315730, 1152.79644516, 1116.69634771, 1147.98309883, 1176.86317679, 1161.21980123, 1155.20311832, 1156.40645491, 1197.31989868, 1203.33658159, 1263.50341067, 1311.63687393, 1323.67023975, 1311.63687393, 1251.47004485, 1251.47004485, 1263.50341067, 1239.43667904, 1203.33658159, 1161.21980123, 1203.33658159, 1197.31989868, 1185.28653286, 1164.82981098, 1227.40331322, 1170.84649388, 1150.38977200, 1163.62647440, 1150.38977200, 1191.30321577, 1198.52323526, 1191.30321577, 1196.11656210, 1202.13324501, 1168.43982072, 1173.25316705, 1133.54305986, 1176.86317679, 1168.43982072, 1157.60979149, 1166.03314756, 1190.09987919, 1197.31989868, 1151.59310858, 1101.05297215, 1058.93619180, 1032.46278700, 1033.66612358, 1020.42942119, 1005.98938221, 1068.56288445, 1005.98938221, 1014.41273828, 989.14267007, 1083.00292343, 1022.83609435, 1016.81941144, 1079.39291368, 1028.85277726, 1032.46278700, 993.95601639, 1013.20940170, 1015.61607486, 998.76936272, 956.65258236, 968.68594818, 1009.59939195, 1016.81941144, 1052.91950889, 1144.37308909, 1103.45964532, 1127.52637695, 1202.13324501, 1215.36994740, 1275.53677648, 1227.40331322, 1299.60350811, 1311.63687393, 1335.70360556, 1251.47004485, 1299.60350811, 1311.63687393, 1263.50341067, 1275.53677648, 1263.50341067, 1287.57014230, 1239.43667904, 1263.50341067, 1215.36994740, 1251.47004485, 1263.50341067, 1197.31989868, 1178.06651337, 1251.47004485, 1323.67023975, 1215.36994740, 1227.40331322, 1251.47004485, 1275.53677648, 1299.60350811, 1275.53677648, 1263.50341067, 1287.57014230, 1239.43667904, 1202.13324501, 1202.13324501, 1251.47004485, 1287.57014230, 1287.57014230, 1323.67023975, 1347.73697138, 1347.73697138, 1251.47004485, 1287.57014230, 1215.36994740, 1203.33658159, 1215.36994740, 1287.57014230, 1239.43667904, 1227.40331322, 1263.50341067, 1239.43667904, 1172.04983047, 1173.25316705, 1187.69320603, 1227.40331322, 1239.43667904, 1215.36994740, 1200.92990842, 1191.30321577, 1227.40331322, 1239.43667904, 1299.60350811, 1287.57014230, 1263.50341067, 1287.57014230, 1239.43667904, 1239.43667904, 1251.47004485, 1215.36994740, 1263.50341067, 1275.53677648, 1287.57014230, 1263.50341067, 1275.53677648, 1275.53677648, 1299.60350811, 1287.57014230, 1263.50341067, 1239.43667904, 1311.63687393, 1299.60350811, 1383.83706883, 1359.77033719, 1251.47004485, 1263.50341067, 1275.53677648, 1263.50341067, 1323.67023975, 1359.77033719, 1383.83706883, 1359.77033719, 1359.77033719, 1480.10399535, 1492.13736117, 1480.10399535, 1480.10399535, 1492.13736117, 1516.20409280, 1516.20409280, 1600.43765351, 1540.27082443, 1516.20409280, 1516.20409280, 1576.37092188, 1564.33755606, 1636.53775096, 1588.40428770, 1492.13736117, 1444.00389791, 1335.70360556, 1311.63687393, 1239.43667904, 1287.57014230], 
							[1287.57014230, 1275.53677648, 1215.36994740, 1185.28653286, 1119.10302088, 944.61921655, 902.50243619, 902.50243619, 878.43570456, 872.41902165, 960.26259211, 985.53266032, 956.65258236, 943.41587996, 922.95915808, 885.65572405, 835.11558762, 758.10204640, 827.89556813, 799.01549017, 770.13541222, 717.18860263, 708.76524656, 652.20842722, 594.44827130, 746.06868058, 866.40233874, 902.50243619, 998.76936272, 951.83923604, 987.93933348, 927.77250440, 904.90910935, 903.70577277, 863.99566558, 843.53894369, 861.58899242, 832.70891446, 884.45238747, 960.26259211, 932.58585073, 995.15935297, 986.73599690, 948.22922629, 938.60253364, 902.50243619, 870.01234849, 874.82569481, 908.51911910, 949.43256287, 927.77250440, 979.51597741, 934.99252389, 966.27927501, 1140.76307935, 1143.16975251, 1371.80370301, 1612.47101933, 1528.23745862, 1516.20409280, 1431.97053209, 1528.23745862, 1636.53775096, 1564.33755606, 1588.40428770, 1624.50438514, 1588.40428770, 1588.40428770, 1456.03726372, 1480.10399535, 1492.13736117, 1528.23745862, 1552.30419025, 1480.10399535, 1504.17072698, 1504.17072698, 1576.37092188, 1504.17072698, 1528.23745862, 1516.20409280, 1492.13736117, 1431.97053209, 1395.87043464, 1395.87043464, 1444.00389791, 1407.90380046, 1347.73697138, 1263.50341067, 1044.49615282, 1117.89968430, 1050.51283573, 948.22922629, 1010.80272853, 998.76936272, 1027.64944068, 986.73599690, 998.76936272, 1016.81941144, 1116.69634771, 1149.18643542, 1132.33972327, 1251.47004485, 1198.52323526, 1131.13638669, 1185.28653286, 1161.21980123, 1162.42313781, 1215.36994740, 1164.82981098, 1107.06965506, 1140.76307935, 1141.96641593, 1134.74639644, 1113.08633797, 1110.67966481, 1172.04983047, 1122.71303062, 1155.20311832, 1179.26984996, 1179.26984996, 1161.21980123, 1131.13638669, 1116.69634771, 1111.88300139, 1110.67966481, 1099.84963557, 1075.78290394, 1078.18957710, 1076.98624052, 1066.15621129, 1050.51283573, 1046.90282598, 1010.80272853, 957.85591894, 971.09262134, 984.32932374, 965.07593843, 981.92265058, 966.27927501, 934.99252389, 931.38251415, 900.09576303, 920.55248491, 879.63904114, 878.43570456, 885.65572405, 878.43570456, 902.50243619, 974.70263109, 942.21254338, 937.39919706, 900.09576303, 889.26573379, 897.68908986, 898.89242645, 909.72245568, 854.36897293, 847.14895344, 835.11558762, 795.40548043, 786.98212436, 796.60881701, 786.98212436, 786.98212436, 826.69223155, 807.43884625, 823.08222181, 796.60881701, 792.99880727, 748.47535375, 779.76210487, 742.45867084, 693.12187099, 706.35857339, 673.86848569, 758.10204640, 782.16877803, 770.13541222, 755.69537324, 770.13541222, 764.11872931, 896.48575328, 876.02903140, 962.66926527, 947.02588971, 1021.63275777, 974.70263109, 980.71931399, 986.73599690, 957.85591894, 944.61921655, 901.29909961, 903.70577277, 960.26259211, 950.63589945, 960.26259211, 979.51597741, 990.34600665, 1013.20940170, 996.36268955, 1046.90282598, 1119.10302088, 1167.23648414, 1116.69634771, 1126.32304037, 1135.94973302, 1133.54305986, 1113.08633797, 1173.25316705, 1151.59310858, 1263.50341067, 1335.70360556, 1323.67023975, 1431.97053209, 1468.07062954, 1419.93716627, 1383.83706883, 1347.73697138, 1347.73697138, 1395.87043464, 1395.87043464, 1383.83706883, 1419.93716627, 1407.90380046, 1407.90380046, 1456.03726372, 1492.13736117, 1468.07062954, 1419.93716627, 1395.87043464, 1359.77033719, 1311.63687393, 1227.40331322, 1239.43667904, 1323.67023975, 1215.36994740, 1251.47004485, 1287.57014230, 1263.50341067, 1299.60350811, 1335.70360556, 1299.60350811, 1407.90380046, 1347.73697138, 1371.80370301, 1371.80370301, 1323.67023975, 1311.63687393, 1156.40645491, 1095.03628924, 980.71931399, 908.51911910, 922.95915808, 968.68594818, 892.87574354, 904.90910935, 902.50243619, 876.02903140, 811.04885599, 794.20214385, 874.82569481, 843.53894369, 847.14895344, 815.86220232, 829.09890471, 783.37211461, 705.15523681, 644.98840773, 720.79861237, 703.95190023, 697.93521732, 699.13855390, 805.03217308, 812.25219257, 818.26887548, 873.62235823, 862.79232900, 832.70891446, 835.11558762, 805.03217308, 800.21882676, 755.69537324, 791.79547068, 730.42530502, 731.62864161, 706.35857339, 690.71519783, 657.02177355, 654.61510038, 658.22511013, 614.90499319], 
							[1552.30419025, 1600.43765351, 1588.40428770, 1588.40428770, 1540.27082443, 1323.67023975, 1263.50341067, 1275.53677648, 1239.43667904, 1227.40331322, 1311.63687393, 1335.70360556, 1263.50341067, 1275.53677648, 1215.36994740, 1203.33658159, 1101.05297215, 978.31264083, 954.24590920, 866.40233874, 853.16563635, 820.67554864, 758.10204640, 697.93521732, 673.86848569, 889.26573379, 1109.47632822, 1119.10302088, 1073.37623078, 1064.95287471, 1083.00292343, 1010.80272853, 972.29595792, 981.92265058, 956.65258236, 932.58585073, 912.12912884, 825.48889497, 921.75582150, 939.80587022, 913.33246543, 933.78918731, 914.53580201, 914.53580201, 873.62235823, 838.72559737, 841.13227053, 819.47221206, 885.65572405, 884.45238747, 849.55562660, 871.21568507, 826.69223155, 836.31892420, 918.14581175, 986.73599690, 974.70263109, 1068.56288445, 1203.33658159, 1167.23648414, 1095.03628924, 1188.89654261, 1227.40331322, 1162.42313781, 1191.30321577, 1167.23648414, 1145.57642567, 1133.54305986, 1084.20626001, 1025.24276751, 1032.46278700, 1022.83609435, 1022.83609435, 974.70263109, 1032.46278700, 1020.42942119, 1058.93619180, 1058.93619180, 1087.81626976, 1081.79958685, 1058.93619180, 1058.93619180, 1028.85277726, 977.10930425, 979.51597741, 991.54934323, 968.68594818, 938.60253364, 860.38565584, 861.58899242, 855.57230951, 800.21882676, 890.46907037, 867.60567532, 865.19900216, 830.30224130, 842.33560711, 854.36897293, 938.60253364, 998.76936272, 974.70263109, 1197.31989868, 1140.76307935, 1063.74953812, 1109.47632822, 1119.10302088, 1116.69634771, 1101.05297215, 1068.56288445, 1046.90282598, 1072.17289419, 1113.08633797, 1089.01960634, 1070.96955761, 1062.54620154, 1075.78290394, 1070.96955761, 1081.79958685, 1086.61293317, 1086.61293317, 1085.40959659, 1067.35954787, 1034.86946017, 1031.25945042, 1028.85277726, 978.31264083, 955.44924578, 950.63589945, 951.83923604, 968.68594818, 953.04257262, 996.36268955, 932.58585073, 924.16249466, 925.36583124, 890.46907037, 882.04571430, 885.65572405, 862.79232900, 857.97898267, 854.36897293, 808.64218283, 805.03217308, 771.33874880, 786.98212436, 813.45552915, 815.86220232, 901.29909961, 914.53580201, 889.26573379, 896.48575328, 902.50243619, 922.95915808, 914.53580201, 908.51911910, 903.70577277, 885.65572405, 817.06553890, 818.26887548, 794.20214385, 767.72873905, 770.13541222, 762.91539273, 753.28870007, 814.65886573, 806.23550966, 817.06553890, 797.81215359, 756.89870982, 740.05199768, 734.03531477, 707.56190997, 675.07182227, 684.69851492, 640.17506140, 715.98526604, 736.44198793, 722.00194895, 685.90185151, 722.00194895, 722.00194895, 780.96544145, 753.28870007, 758.10204640, 744.86534400, 806.23550966, 736.44198793, 728.01863186, 770.13541222, 728.01863186, 723.20528553, 685.90185151, 740.05199768, 707.56190997, 723.20528553, 715.98526604, 754.49203666, 768.93207563, 747.27201717, 750.88202691, 764.11872931, 829.09890471, 785.77878778, 779.76210487, 794.20214385, 780.96544145, 780.96544145, 773.74542196, 758.10204640, 735.23865135, 738.84866109, 744.86534400, 735.23865135, 782.16877803, 803.82883650, 807.43884625, 815.86220232, 808.64218283, 782.16877803, 800.21882676, 801.42216334, 805.03217308, 876.02903140, 826.69223155, 813.45552915, 818.26887548, 808.64218283, 770.13541222, 756.89870982, 752.08536349, 732.83197819, 706.35857339, 691.91853441, 695.52854416, 714.78192946, 700.34189048, 687.10518809, 725.61195870, 718.39193921, 722.00194895, 723.20528553, 715.98526604, 713.57859288, 682.29184176, 665.44512962, 676.27515885, 648.59841748, 634.15837850, 578.80489574, 545.11147146, 541.50146171, 504.19802769, 488.55465212, 486.14797896, 389.88105243, 430.79449621, 389.88105243, 364.61098422, 330.91755994, 300.83414540, 348.96760866, 324.90087703, 317.68085754, 303.24081856, 321.29086728, 302.03748198, 276.76741377, 262.32737479, 270.75073086, 261.12403820, 268.34405769, 276.76741377, 286.39410642, 277.97075035, 285.19076984, 281.58076009, 279.17408693, 283.98743325, 279.17408693, 270.75073086, 222.61726759, 197.34719938, 214.19391152, 174.48380433, 149.21373612, 141.99371663, 139.58704346, 162.45043851, 145.60372637, 168.46712142, 132.36702397], 
							[898.89242645, 908.51911910, 878.43570456, 859.18231925, 824.28555839, 766.52540247, 725.61195870, 673.86848569, 638.97172482, 613.70165661, 646.19174431, 685.90185151, 691.91853441, 715.98526604, 655.81843697, 654.61510038, 601.66829079, 553.53482753, 590.83826156, 595.65160789, 565.56819335, 505.40136427, 527.06142274, 517.43473008, 539.09478855, 537.89145197, 651.00509064, 682.29184176, 661.83511987, 696.73188074, 705.15523681, 638.97172482, 618.51500294, 655.81843697, 634.15837850, 634.15837850, 637.76838824, 581.21156891, 637.76838824, 634.15837850, 649.80175406, 636.56505166, 617.31166635, 625.73502243, 584.82157865, 555.94150069, 581.21156891, 578.80489574, 634.15837850, 658.22511013, 620.92167610, 646.19174431, 620.92167610, 649.80175406, 714.78192946, 696.73188074, 694.32520758, 695.52854416, 678.68183202, 705.15523681, 703.95190023, 723.20528553, 744.86534400, 709.96858314, 713.57859288, 702.74856365, 738.84866109, 696.73188074, 697.93521732, 664.24179304, 693.12187099, 675.07182227, 690.71519783, 677.47849543, 697.93521732, 707.56190997, 706.35857339, 709.96858314, 722.00194895, 720.79861237, 694.32520758, 677.47849543, 690.71519783, 664.24179304, 643.78507115, 648.59841748, 607.68497370, 575.19488600, 496.97800820, 488.55465212, 493.36799845, 418.76113039, 433.20116937, 423.57447672, 423.57447672, 386.27104269, 419.96446697, 417.55779381, 457.26790100, 462.08124733, 445.23453519, 510.21471059, 499.38468136, 442.82786202, 474.11461315, 470.50460340, 492.16466187, 509.01137401, 475.31794973, 468.09793024, 481.33463264, 500.58801794, 504.19802769, 493.36799845, 504.19802769, 504.19802769, 493.36799845, 523.45141299, 533.07810564, 577.60155916, 583.61824207, 559.55151044, 540.29812513, 535.48477881, 545.11147146, 529.46809590, 522.24807641, 519.84140325, 559.55151044, 601.66829079, 625.73502243, 601.66829079, 583.61824207, 608.88831028, 599.26161763, 599.26161763, 614.90499319, 631.75170533, 629.34503217, 635.36171508, 703.95190023, 673.86848569, 702.74856365, 685.90185151, 705.15523681, 718.39193921, 718.39193921, 746.06868058, 813.45552915, 788.18546094, 770.13541222, 736.44198793, 709.96858314, 740.05199768, 703.95190023, 679.88516860, 664.24179304, 625.73502243, 611.29498345, 625.73502243, 577.60155916, 571.58487625, 570.38153967, 587.22825181, 661.83511987, 628.14169559, 660.63178329, 626.93835901, 599.26161763, 571.58487625, 589.63492498, 530.67143248, 472.91127656, 493.36799845, 442.82786202, 553.53482753, 565.56819335, 595.65160789, 605.27830054, 571.58487625, 546.31480804, 607.68497370, 654.61510038, 644.98840773, 616.10832977, 648.59841748, 602.87162738, 598.05828105, 641.37839799, 612.49832003, 630.54836875, 587.22825181, 624.53168584, 596.85494447, 631.75170533, 625.73502243, 673.86848569, 737.64532451, 741.25533426, 741.25533426, 758.10204640, 830.30224130, 808.64218283, 800.21882676, 835.11558762, 821.87888522, 903.70577277, 853.16563635, 832.70891446, 850.75896318, 866.40233874, 914.53580201, 926.56916782, 903.70577277, 880.84237772, 898.89242645, 937.39919706, 1001.17603588, 974.70263109, 977.10930425, 986.73599690, 1010.80272853, 1037.27613333, 1080.59625027, 1089.01960634, 1188.89654261, 1152.79644516, 1096.23962583, 1078.18957710, 1090.22294292, 1043.29281624, 1016.81941144, 1021.63275777, 1062.54620154, 1095.03628924, 1089.01960634, 1113.08633797, 1113.08633797, 1128.72971353, 1141.96641593, 1141.96641593, 1133.54305986, 1192.50655235, 1166.03314756, 1155.20311832, 1172.04983047, 1147.98309883, 1166.03314756, 1103.45964532, 1043.29281624, 1086.61293317, 1045.69948940, 1022.83609435, 1010.80272853, 930.17917757, 977.10930425, 968.68594818, 1056.52951863, 1049.30949914, 1054.12284547, 1038.47946991, 1058.93619180, 1022.83609435, 1052.91950889, 1049.30949914, 1076.98624052, 1131.13638669, 1121.50969404, 1144.37308909, 1200.92990842, 1275.53677648, 1263.50341067, 1287.57014230, 1203.33658159, 1263.50341067, 1227.40331322, 1215.36994740, 1263.50341067, 1335.70360556, 1335.70360556, 1359.77033719, 1468.07062954, 1444.00389791, 1419.93716627, 1456.03726372, 1395.87043464, 1311.63687393, 1239.43667904, 1287.57014230, 1263.50341067, 1239.43667904], 
							[599.26161763, 566.77152993, 555.94150069, 540.29812513, 535.48477881, 507.80803743, 493.36799845, 498.18134478, 513.82472034, 511.41804717, 523.45141299, 509.01137401, 512.62138376, 516.23139350, 505.40136427, 493.36799845, 472.91127656, 469.30126682, 487.35131554, 469.30126682, 471.70793998, 451.25121810, 442.82786202, 444.03119861, 412.74444748, 451.25121810, 521.04473983, 512.62138376, 494.57133503, 499.38468136, 502.99469110, 487.35131554, 489.75798871, 494.57133503, 488.55465212, 474.11461315, 464.48792049, 435.60784253, 434.40450595, 447.64120835, 448.84454493, 435.60784253, 454.86122784, 445.23453519, 431.99783279, 427.18448646, 421.16780356, 415.15112065, 425.98114988, 424.77781330, 413.94778407, 413.94778407, 413.94778407, 397.10107192, 425.98114988, 445.23453519, 475.31794973, 517.43473008, 549.92481779, 541.50146171, 511.41804717, 561.95818360, 561.95818360, 540.29812513, 554.73816411, 547.51814462, 551.12815437, 570.38153967, 546.31480804, 507.80803743, 507.80803743, 496.97800820, 507.80803743, 493.36799845, 505.40136427, 493.36799845, 519.84140325, 517.43473008, 504.19802769, 505.40136427, 501.79135452, 493.36799845, 499.38468136, 478.92795947, 492.16466187, 484.94464238, 495.77467161, 475.31794973, 450.04788151, 456.06456442, 424.77781330, 391.08438902, 418.76113039, 403.11775483, 413.94778407, 413.94778407, 419.96446697, 448.84454493, 450.04788151, 478.92795947, 469.30126682, 481.33463264, 477.72462289, 444.03119861, 476.52128631, 481.33463264, 469.30126682, 504.19802769, 488.55465212, 483.74130580, 481.33463264, 482.53796922, 489.75798871, 488.55465212, 490.96132529, 490.96132529, 506.60470085, 533.07810564, 567.97486651, 580.00823233, 577.60155916, 569.17820309, 563.16152018, 565.56819335, 572.78821284, 552.33149095, 531.87476906, 558.34817386, 561.95818360, 571.58487625, 564.36485676, 565.56819335, 563.16152018, 547.51814462, 558.34817386, 555.94150069, 564.36485676, 582.41490549, 565.56819335, 563.16152018, 564.36485676, 535.48477881, 535.48477881, 529.46809590, 535.48477881, 534.28144222, 536.68811539, 546.31480804, 541.50146171, 533.07810564, 541.50146171, 483.74130580, 495.77467161, 489.75798871, 487.35131554, 466.89459366, 458.47123758, 445.23453519, 423.57447672, 412.74444748, 400.71108167, 430.79449621, 413.94778407, 415.15112065, 433.20116937, 413.94778407, 428.38782305, 406.72776458, 393.49106218, 393.49106218, 380.25435978, 375.44101346, 379.05102320, 399.50774509, 376.64435004, 421.16780356, 411.54111090, 422.37114014, 421.16780356, 436.81117912, 436.81117912, 466.89459366, 451.25121810, 477.72462289, 469.30126682, 528.26475932, 482.53796922, 490.96132529, 529.46809590, 517.43473008, 515.02805692, 509.01137401, 528.26475932, 542.70479830, 535.48477881, 552.33149095, 559.55151044, 563.16152018, 551.12815437, 558.34817386, 571.58487625, 583.61824207, 564.36485676, 563.16152018, 571.58487625, 598.05828105, 616.10832977, 607.68497370, 601.66829079, 614.90499319, 608.88831028, 611.29498345, 643.78507115, 657.02177355, 655.81843697, 667.85180278, 651.00509064, 663.03845645, 654.61510038, 632.95504192, 647.39508089, 653.41176380, 679.88516860, 654.61510038, 643.78507115, 672.66514911, 667.85180278, 661.83511987, 644.98840773, 644.98840773, 623.32834926, 636.56505166, 618.51500294, 623.32834926, 638.97172482, 601.66829079, 614.90499319, 629.34503217, 631.75170533, 643.78507115, 628.14169559, 625.73502243, 642.58173457, 619.71833952, 630.54836875, 634.15837850, 632.95504192, 637.76838824, 628.14169559, 613.70165661, 572.78821284, 586.02491523, 569.17820309, 576.39822258, 521.04473983, 564.36485676, 548.72148120, 569.17820309, 567.97486651, 593.24493472, 641.37839799, 601.66829079, 582.41490549, 569.17820309, 587.22825181, 583.61824207, 594.44827130, 604.07496396, 602.87162738, 596.85494447, 628.14169559, 629.34503217, 652.20842722, 638.97172482, 606.48163712, 623.32834926, 637.76838824, 618.51500294, 641.37839799, 626.93835901, 638.97172482, 620.92167610, 631.75170533, 571.58487625, 580.00823233, 560.75484702, 565.56819335, 566.77152993, 540.29812513, 546.31480804, 488.55465212], 
							[617.70876743, 614.26722480, 636.04761693, 601.66829079, 610.83771555, 561.54904916, 542.06702991, 505.40136427, 527.16972303, 488.20568452, 555.82116704, 561.54904916, 608.53934267, 675.01165544, 641.77549906, 608.53934267, 570.71847392, 495.08876976, 553.53482753, 529.46809590, 515.71395877, 464.13895288, 472.16520788, 448.09847625, 418.30386249, 473.30837764, 539.78069040, 536.33914778, 550.09328491, 545.50857253, 544.36540278, 504.24616115, 476.74992026, 520.29867115, 528.31289278, 531.75443540, 528.31289278, 499.67348214, 530.61126565, 503.10299140, 505.40136427, 530.61126565, 527.16972303, 507.68770377, 471.02203813, 471.02203813, 490.50405739, 480.19146288, 501.95982164, 520.29867115, 492.79039689, 509.98607664, 536.33914778, 534.05280827, 561.54904916, 595.94040867, 583.32944129, 693.35050495, 666.98540044, 691.05213207, 675.01165544, 723.14511871, 760.96598746, 725.43145821, 734.60088296, 750.65339296, 802.21636548, 779.30483697, 750.65339296, 744.91347747, 746.06868058, 739.18559534, 738.04242559, 701.36472658, 700.22155683, 710.53415133, 736.89925583, 719.70357608, 722.00194895, 716.26203346, 707.10464207, 684.18108019, 660.11434856, 641.77549906, 678.45319807, 668.12857019, 631.46290455, 602.81146055, 530.61126565, 535.19597803, 530.61126565, 515.71395877, 521.44184090, 504.24616115, 536.33914778, 498.51827902, 518.00029828, 536.33914778, 613.12405505, 630.31973480, 585.61578080, 642.91866881, 608.53934267, 576.44635604, 600.51308768, 641.77549906, 645.21704168, 693.35050495, 662.40068807, 625.73502243, 658.97117881, 684.18108019, 702.51992970, 693.35050495, 702.51992970, 722.00194895, 735.75608608, 744.91347747, 759.81078435, 774.72012459, 788.46222835, 733.45771321, 738.04242559, 720.84674583, 742.62713796, 744.91347747, 739.18559534, 756.38127509, 757.52444484, 763.25232697, 756.38127509, 739.18559534, 712.83252420, 712.83252420, 713.97569395, 694.49367470, 700.22155683, 752.93973247, 742.62713796, 738.04242559, 731.17137370, 710.53415133, 695.63684445, 688.76579257, 701.36472658, 695.63684445, 692.19530183, 709.39098158, 728.01863186, 720.79861237, 724.40862212, 688.30852467, 671.46181253, 673.86848569, 678.68183202, 647.39508089, 613.70165661, 586.02491523, 551.12815437, 551.12815437, 537.89145197, 555.94150069, 547.51814462, 569.17820309, 608.88831028, 598.05828105, 620.92167610, 595.65160789, 546.31480804, 523.45141299, 554.73816411, 524.65474957, 502.99469110, 511.41804717, 463.28458391, 560.75484702, 548.72148120, 563.16152018, 549.92481779, 530.67143248, 552.33149095, 599.26161763, 578.80489574, 600.46495421, 600.46495421, 608.88831028, 583.61824207, 596.85494447, 613.70165661, 613.70165661, 616.10832977, 604.07496396, 625.73502243, 630.54836875, 643.78507115, 649.80175406, 675.07182227, 675.07182227, 653.41176380, 641.37839799, 659.42844671, 670.25847594, 669.05513936, 679.88516860, 701.54522707, 713.57859288, 701.54522707, 706.35857339, 703.95190023, 679.88516860, 684.69851492, 687.10518809, 689.51186125, 697.93521732, 714.78192946, 723.20528553, 715.98526604, 714.78192946, 683.49517834, 682.29184176, 694.32520758, 659.42844671, 665.44512962, 671.46181253, 689.51186125, 689.51186125, 694.32520758, 693.12187099, 665.44512962, 648.59841748, 632.95504192, 631.75170533, 606.48163712, 631.75170533, 632.95504192, 596.85494447, 608.88831028, 631.75170533, 631.75170533, 631.75170533, 625.73502243, 606.48163712, 625.73502243, 602.87162738, 586.02491523, 601.66829079, 589.63492498, 598.05828105, 587.22825181, 535.48477881, 541.50146171, 517.43473008, 495.77467161, 493.36799845, 421.16780356, 456.06456442, 458.47123758, 515.02805692, 478.92795947, 510.21471059, 521.04473983, 512.62138376, 490.96132529, 475.31794973, 504.19802769, 488.55465212, 464.48792049, 486.14797896, 522.24807641, 547.51814462, 570.38153967, 570.38153967, 582.41490549, 554.73816411, 555.94150069, 567.97486651, 573.99154942, 552.33149095, 541.50146171, 523.45141299, 499.38468136, 493.36799845, 471.70793998, 445.23453519, 422.37114014, 394.69439876, 386.27104269, 381.45769636, 392.28772560, 368.22099397, 376.64435004], 
							[776.15209512, 722.00194895, 770.13541222, 782.16877803, 755.69537324, 758.10204640, 873.62235823, 821.87888522, 847.14895344, 842.33560711, 891.67240696, 885.65572405, 903.70577277, 888.06239721, 898.89242645, 907.31578252, 861.58899242, 854.36897293, 926.56916782, 968.68594818, 914.53580201, 902.50243619, 922.95915808, 920.55248491, 885.65572405, 933.78918731, 1030.05611384, 1028.85277726, 992.75267981, 960.26259211, 957.85591894, 938.60253364, 961.46592869, 989.14267007, 932.58585073, 950.63589945, 927.77250440, 896.48575328, 962.66926527, 980.71931399, 967.48261160, 995.15935297, 1007.19271879, 1036.07279675, 1016.81941144, 1002.37937246, 984.32932374, 998.76936272, 1033.66612358, 1038.47946991, 998.76936272, 1020.42942119, 977.10930425, 1004.78604563, 1072.17289419, 1081.79958685, 1125.11970378, 1107.06965506, 1134.74639644, 1203.33658159, 1191.30321577, 1227.40331322, 1203.33658159, 1198.52323526, 1161.21980123, 1151.59310858, 1187.69320603, 1160.01646465, 1107.06965506, 1109.47632822, 1072.17289419, 1083.00292343, 1104.66298190, 1121.50969404, 1113.08633797, 1134.74639644, 1146.77976225, 1122.71303062, 1135.94973302, 1099.84963557, 1076.98624052, 1075.78290394, 1095.03628924, 1139.55974276, 1178.06651337, 1127.52637695, 1161.21980123, 1129.93305011, 1119.10302088, 1099.84963557, 1129.93305011, 1117.89968430, 1160.01646465, 1227.40331322, 1275.53677648, 1263.50341067, 1263.50341067, 1335.70360556, 1335.70360556, 1444.00389791, 1407.90380046, 1600.43765351, 1660.60448259, 1576.37092188, 1672.63784841, 1552.30419025, 1540.27082443, 1552.30419025, 1612.47101933, 1540.27082443, 1516.20409280, 1588.40428770, 1648.57111678, 1624.50438514, 1600.43765351, 1648.57111678, 1612.47101933, 1612.47101933, 1684.67121422, 1756.87140912, 1720.77131167, 1660.60448259, 1564.33755606, 1528.23745862, 1576.37092188, 1564.33755606, 1576.37092188, 1504.17072698, 1540.27082443, 1516.20409280, 1492.13736117, 1516.20409280, 1480.10399535, 1480.10399535, 1456.03726372, 1444.00389791, 1504.17072698, 1552.30419025, 1540.27082443, 1480.10399535, 1456.03726372, 1383.83706883, 1383.83706883, 1359.77033719, 1371.80370301, 1395.87043464, 1371.80370301, 1395.87043464, 1407.90380046, 1371.80370301, 1275.53677648, 1187.69320603, 1129.93305011, 1123.91636720, 1121.50969404, 1169.64315730, 1149.18643542, 1105.86631848, 1093.83295266, 1081.79958685, 1073.37623078, 1117.89968430, 1070.96955761, 1064.95287471, 1117.89968430, 1083.00292343, 1107.06965506, 1013.20940170, 1049.30949914, 1061.34286496, 1067.35954787, 1010.80272853, 1037.27613333, 1161.21980123, 1137.15306960, 1215.36994740, 1203.33658159, 1299.60350811, 1299.60350811, 1263.50341067, 1263.50341067, 1468.07062954, 1371.80370301, 1323.67023975, 1299.60350811, 1323.67023975, 1251.47004485, 1287.57014230, 1395.87043464, 1347.73697138, 1311.63687393, 1227.40331322, 1299.60350811, 1299.60350811, 1323.67023975, 1311.63687393, 1347.73697138, 1444.00389791, 1383.83706883, 1456.03726372, 1383.83706883, 1504.17072698, 1444.00389791, 1564.33755606, 1564.33755606, 1540.27082443, 1552.30419025, 1504.17072698, 1456.03726372, 1492.13736117, 1552.30419025, 1552.30419025, 1636.53775096, 1588.40428770, 1624.50438514, 1612.47101933, 1636.53775096, 1636.53775096, 1564.33755606, 1552.30419025, 1648.57111678, 1660.60448259, 1648.57111678, 1612.47101933, 1672.63784841, 1792.97150657, 1732.80467749, 1732.80467749, 1636.53775096, 1540.27082443, 1588.40428770, 1612.47101933, 1600.43765351, 1612.47101933, 1552.30419025, 1492.13736117, 1480.10399535, 1564.33755606, 1552.30419025, 1612.47101933, 1528.23745862, 1516.20409280, 1576.37092188, 1480.10399535, 1444.00389791, 1444.00389791, 1383.83706883, 1395.87043464, 1431.97053209, 1335.70360556, 1359.77033719, 1275.53677648, 1299.60350811, 1275.53677648, 1275.53677648, 1359.77033719, 1323.67023975, 1383.83706883, 1335.70360556, 1371.80370301, 1383.83706883, 1359.77033719, 1383.83706883, 1431.97053209, 1444.00389791, 1492.13736117, 1444.00389791, 1672.63784841, 1588.40428770, 1648.57111678, 1672.63784841, 1696.70458004, 1732.80467749, 1660.60448259, 1756.87140912, 1744.83804330, 1792.97150657, 1708.73794585, 1756.87140912, 1756.87140912, 1720.77131167, 1756.87140912, 1708.73794585, 1684.67121422, 1805.00487238, 1684.67121422, 1612.47101933, 1371.80370301, 1480.10399535, 1480.10399535, 1516.20409280], 
							[529.46809590, 510.21471059, 493.36799845, 487.35131554, 476.52128631, 423.57447672, 407.93110116, 410.33777432, 387.47437927, 395.89773534, 433.20116937, 424.77781330, 458.47123758, 474.11461315, 464.48792049, 468.09793024, 430.79449621, 427.18448646, 457.26790100, 439.21785228, 428.38782305, 398.30440851, 415.15112065, 380.25435978, 370.62766713, 411.54111090, 481.33463264, 540.29812513, 530.67143248, 516.23139350, 534.28144222, 487.35131554, 483.74130580, 493.36799845, 478.92795947, 464.48792049, 453.65789126, 410.33777432, 427.18448646, 439.21785228, 421.16780356, 421.16780356, 442.82786202, 421.16780356, 421.16780356, 403.11775483, 399.50774509, 403.11775483, 421.16780356, 433.20116937, 409.13443774, 387.47437927, 373.03434029, 370.62766713, 411.54111090, 438.01451570, 421.16780356, 498.18134478, 481.33463264, 476.52128631, 444.03119861, 511.41804717, 511.41804717, 511.41804717, 507.80803743, 517.43473008, 499.38468136, 499.38468136, 468.09793024, 448.84454493, 444.03119861, 446.43787177, 463.28458391, 462.08124733, 494.57133503, 477.72462289, 499.38468136, 490.96132529, 496.97800820, 510.21471059, 492.16466187, 486.14797896, 486.14797896, 464.48792049, 481.33463264, 456.06456442, 444.03119861, 400.71108167, 367.01765738, 379.05102320, 374.23767687, 356.18762815, 353.78095499, 339.34091601, 341.74758917, 327.30755019, 336.93424284, 389.88105243, 405.52442800, 406.72776458, 388.67771585, 419.96446697, 398.30440851, 382.66103294, 409.13443774, 417.55779381, 397.10107192, 418.76113039, 407.93110116, 393.49106218, 413.94778407, 439.21785228, 435.60784253, 422.37114014, 427.18448646, 440.42118886, 445.23453519, 468.09793024, 475.31794973, 492.16466187, 487.35131554, 459.67457417, 451.25121810, 452.45455468, 459.67457417, 457.26790100, 441.62452544, 460.87791075, 474.11461315, 465.69125707, 457.26790100, 457.26790100, 451.25121810, 457.26790100, 440.42118886, 450.04788151, 448.84454493, 493.36799845, 483.74130580, 494.57133503, 496.97800820, 469.30126682, 477.72462289, 460.87791075, 451.25121810, 463.28458391, 474.11461315, 475.31794973, 481.33463264, 475.31794973, 474.11461315, 441.62452544, 412.74444748, 433.20116937, 425.98114988, 403.11775483, 379.05102320, 387.47437927, 373.03434029, 345.35759892, 329.71422336, 341.74758917, 330.91755994, 344.15426233, 369.42433055, 362.20431106, 358.59430131, 338.13757943, 327.30755019, 300.83414540, 322.49420387, 308.05416489, 270.75073086, 281.58076009, 264.73404795, 298.42747223, 334.52756968, 320.08753070, 303.24081856, 321.29086728, 342.95092575, 359.79763789, 358.59430131, 375.44101346, 381.45769636, 392.28772560, 358.59430131, 361.00097448, 383.86436953, 363.40764764, 338.13757943, 318.88419412, 328.51088677, 327.30755019, 342.95092575, 354.98429157, 362.20431106, 376.64435004, 367.01765738, 379.05102320, 394.69439876, 417.55779381, 403.11775483, 421.16780356, 425.98114988, 444.03119861, 430.79449621, 438.01451570, 422.37114014, 413.94778407, 415.15112065, 411.54111090, 438.01451570, 435.60784253, 454.86122784, 476.52128631, 447.64120835, 471.70793998, 452.45455468, 431.99783279, 433.20116937, 407.93110116, 431.99783279, 415.15112065, 427.18448646, 438.01451570, 445.23453519, 424.77781330, 416.35445723, 398.30440851, 397.10107192, 387.47437927, 375.44101346, 376.64435004, 392.28772560, 375.44101346, 367.01765738, 370.62766713, 381.45769636, 387.47437927, 381.45769636, 369.42433055, 383.86436953, 364.61098422, 358.59430131, 380.25435978, 362.20431106, 356.18762815, 352.57761841, 320.08753070, 326.10421361, 329.71422336, 316.47752096, 317.68085754, 270.75073086, 302.03748198, 282.78409667, 300.83414540, 290.00411616, 287.59744300, 314.07084779, 296.02079907, 280.37742351, 275.56407718, 293.61412591, 269.54739428, 256.31069188, 249.09067239, 250.29400897, 247.88733581, 262.32737479, 282.78409667, 312.86751121, 308.05416489, 298.42747223, 294.81746249, 298.42747223, 288.80077958, 280.37742351, 276.76741377, 261.12403820, 251.49734555, 247.88733581, 229.83728708, 229.83728708, 222.61726759, 227.43061392, 220.21059443, 212.99057494, 196.14386280, 181.70382382], 
							[551.41695515, 529.22742858, 484.82430872, 495.34147044, 483.74130580, 451.25121810, 430.79449621, 453.65789126, 447.64120835, 468.09793024, 500.58801794, 488.55465212, 502.99469110, 493.36799845, 476.52128631, 468.09793024, 445.23453519, 424.77781330, 463.28458391, 452.45455468, 465.69125707, 431.99783279, 428.38782305, 416.35445723, 411.54111090, 440.42118886, 488.55465212, 513.82472034, 478.92795947, 477.72462289, 482.53796922, 459.67457417, 463.28458391, 457.26790100, 451.25121810, 447.64120835, 444.03119861, 412.74444748, 433.20116937, 442.82786202, 427.18448646, 428.38782305, 442.82786202, 424.77781330, 409.13443774, 398.30440851, 413.94778407, 411.54111090, 421.16780356, 428.38782305, 412.74444748, 415.15112065, 407.93110116, 405.52442800, 436.81117912, 463.28458391, 468.09793024, 519.84140325, 511.41804717, 521.04473983, 499.38468136, 540.29812513, 559.55151044, 539.09478855, 560.75484702, 571.58487625, 559.55151044, 536.68811539, 517.43473008, 518.63806666, 527.06142274, 525.85808615, 541.50146171, 527.06142274, 535.48477881, 515.02805692, 531.87476906, 530.67143248, 541.50146171, 543.90813488, 525.85808615, 509.01137401, 500.58801794, 496.97800820, 504.19802769, 499.38468136, 510.21471059, 492.16466187, 440.42118886, 458.47123758, 439.21785228, 429.59115963, 425.98114988, 427.18448646, 439.21785228, 419.96446697, 413.94778407, 435.60784253, 464.48792049, 462.08124733, 434.40450595, 471.70793998, 457.26790100, 457.26790100, 463.28458391, 459.67457417, 457.26790100, 480.13129605, 434.40450595, 434.40450595, 429.59115963, 464.48792049, 448.84454493, 434.40450595, 444.03119861, 447.64120835, 447.64120835, 507.80803743, 539.09478855, 522.24807641, 518.63806666, 509.01137401, 481.33463264, 488.55465212, 516.23139350, 489.75798871, 475.31794973, 493.36799845, 498.18134478, 513.82472034, 492.16466187, 492.16466187, 469.30126682, 465.69125707, 463.28458391, 470.50460340, 470.50460340, 495.77467161, 495.77467161, 502.99469110, 504.19802769, 493.36799845, 482.53796922, 457.26790100, 460.87791075, 450.04788151, 433.20116937, 454.86122784, 454.86122784, 445.23453519, 448.84454493, 419.96446697, 409.13443774, 416.35445723, 407.93110116, 407.93110116, 377.84768662, 376.64435004, 356.18762815, 371.83100371, 344.15426233, 373.03434029, 346.56093550, 353.78095499, 367.01765738, 365.81432080, 365.81432080, 353.78095499, 336.93424284, 302.03748198, 327.30755019, 292.41078933, 277.97075035, 281.58076009, 282.78409667, 310.46083805, 324.90087703, 315.27418438, 294.81746249, 327.30755019, 361.00097448, 386.27104269, 377.84768662, 368.22099397, 348.96760866, 370.62766713, 340.54425259, 344.15426233, 362.20431106, 326.10421361, 327.30755019, 317.68085754, 322.49420387, 321.29086728, 336.93424284, 346.56093550, 362.20431106, 375.44101346, 351.37428182, 368.22099397, 398.30440851, 416.35445723, 397.10107192, 411.54111090, 419.96446697, 440.42118886, 417.55779381, 409.13443774, 405.52442800, 397.10107192, 410.33777432, 401.91441825, 427.18448646, 435.60784253, 447.64120835, 457.26790100, 440.42118886, 458.47123758, 454.86122784, 429.59115963, 445.23453519, 418.76113039, 446.43787177, 430.79449621, 456.06456442, 465.69125707, 454.86122784, 451.25121810, 433.20116937, 413.94778407, 428.38782305, 418.76113039, 407.93110116, 425.98114988, 416.35445723, 392.28772560, 401.91441825, 409.13443774, 412.74444748, 406.72776458, 399.50774509, 394.69439876, 415.15112065, 381.45769636, 385.06770611, 394.69439876, 383.86436953, 383.86436953, 359.79763789, 333.32423310, 334.52756968, 318.88419412, 329.71422336, 333.32423310, 285.19076984, 320.08753070, 298.42747223, 306.85082830, 299.63080882, 293.61412591, 318.88419412, 288.80077958, 282.78409667, 256.31069188, 276.76741377, 274.36074060, 258.71736504, 241.87065290, 263.53071137, 259.92070162, 275.56407718, 280.37742351, 302.03748198, 322.49420387, 312.86751121, 302.03748198, 310.46083805, 304.44415514, 283.98743325, 288.80077958, 263.53071137, 267.14072111, 247.88733581, 245.48066264, 244.27732606, 244.27732606, 256.31069188, 255.10735530, 240.66731632, 228.63395050, 216.60058469], 
							[1480.10399535, 1468.07062954, 1468.07062954, 1444.00389791, 1311.63687393, 1155.20311832, 1125.11970378, 1107.06965506, 1155.20311832, 1132.33972327, 1251.47004485, 1263.50341067, 1275.53677648, 1239.43667904, 1197.31989868, 1141.96641593, 1090.22294292, 1024.03943093, 1073.37623078, 1099.84963557, 1058.93619180, 971.09262134, 985.53266032, 953.04257262, 950.63589945, 957.85591894, 1152.79644516, 1263.50341067, 1179.26984996, 1251.47004485, 1198.52323526, 1179.26984996, 1114.28967455, 1161.21980123, 1141.96641593, 1119.10302088, 1131.13638669, 1097.44296241, 1157.60979149, 1185.28653286, 1193.70988894, 1191.30321577, 1227.40331322, 1190.09987919, 1167.23648414, 1144.37308909, 1122.71303062, 1098.64629899, 1143.16975251, 1191.30321577, 1161.21980123, 1179.26984996, 1193.70988894, 1167.23648414, 1202.13324501, 1299.60350811, 1299.60350811, 1263.50341067, 1311.63687393, 1371.80370301, 1323.67023975, 1371.80370301, 1347.73697138, 1359.77033719, 1383.83706883, 1383.83706883, 1371.80370301, 1311.63687393, 1287.57014230, 1347.73697138, 1299.60350811, 1347.73697138, 1383.83706883, 1323.67023975, 1395.87043464, 1383.83706883, 1383.83706883, 1371.80370301, 1383.83706883, 1407.90380046, 1371.80370301, 1335.70360556, 1347.73697138, 1395.87043464, 1347.73697138, 1359.77033719, 1383.83706883, 1311.63687393, 1263.50341067, 1191.30321577, 1170.84649388, 1134.74639644, 1143.16975251, 1110.67966481, 1215.36994740, 1164.82981098, 1143.16975251, 1215.36994740, 1192.50655235, 1203.33658159, 1175.65984021, 1251.47004485, 1191.30321577, 1166.03314756, 1167.23648414, 1198.52323526, 1182.87985970, 1179.26984996, 1144.37308909, 1131.13638669, 1160.01646465, 1176.86317679, 1185.28653286, 1155.20311832, 1174.45650363, 1203.33658159, 1192.50655235, 1191.30321577, 1180.47318654, 1169.64315730, 1186.48986945, 1161.21980123, 1114.28967455, 1155.20311832, 1191.30321577, 1155.20311832, 1162.42313781, 1145.57642567, 1158.81312807, 1161.21980123, 1188.89654261, 1202.13324501, 1163.62647440, 1155.20311832, 1185.28653286, 1143.16975251, 1152.79644516, 1180.47318654, 1152.79644516, 1173.25316705, 1173.25316705, 1147.98309883, 1157.60979149, 1123.91636720, 1145.57642567, 1149.18643542, 1175.65984021, 1182.87985970, 1185.28653286, 1187.69320603, 1167.23648414, 1164.82981098, 1097.44296241, 1158.81312807, 1155.20311832, 1163.62647440, 1108.27299164, 1105.86631848, 1018.02274802, 1033.66612358, 1068.56288445, 1033.66612358, 993.95601639, 1046.90282598, 1103.45964532, 1083.00292343, 1128.72971353, 1086.61293317, 1080.59625027, 1039.68280649, 1055.32618205, 962.66926527, 1002.37937246, 1040.88614307, 986.73599690, 1052.91950889, 1052.91950889, 1049.30949914, 1028.85277726, 1018.02274802, 1046.90282598, 1131.13638669, 1126.32304037, 1115.49301113, 1121.50969404, 1150.38977200, 1072.17289419, 1149.18643542, 1160.01646465, 1131.13638669, 1143.16975251, 1109.47632822, 1155.20311832, 1145.57642567, 1160.01646465, 1153.99978174, 1181.67652312, 1193.70988894, 1196.11656210, 1197.31989868, 1172.04983047, 1191.30321577, 1149.18643542, 1186.48986945, 1169.64315730, 1199.72657184, 1215.36994740, 1215.36994740, 1202.13324501, 1158.81312807, 1186.48986945, 1123.91636720, 1143.16975251, 1227.40331322, 1227.40331322, 1193.70988894, 1239.43667904, 1227.40331322, 1227.40331322, 1187.69320603, 1239.43667904, 1275.53677648, 1275.53677648, 1227.40331322, 1239.43667904, 1299.60350811, 1263.50341067, 1251.47004485, 1227.40331322, 1239.43667904, 1215.36994740, 1181.67652312, 1166.03314756, 1173.25316705, 1163.62647440, 1129.93305011, 1128.72971353, 1191.30321577, 1202.13324501, 1202.13324501, 1198.52323526, 1193.70988894, 1198.52323526, 1134.74639644, 1121.50969404, 1164.82981098, 1138.35640618, 1143.16975251, 1155.20311832, 1054.12284547, 1095.03628924, 1038.47946991, 1056.52951863, 1063.74953812, 915.73913859, 972.29595792, 937.39919706, 1021.63275777, 950.63589945, 975.90596767, 963.87260185, 963.87260185, 914.53580201, 897.68908986, 951.83923604, 886.85906063, 860.38565584, 890.46907037, 926.56916782, 934.99252389, 960.26259211, 1005.98938221, 1055.32618205, 1010.80272853, 996.36268955, 1040.88614307, 1026.44610409, 1027.64944068, 1012.00606512, 973.49929450, 926.56916782, 922.95915808, 950.63589945, 878.43570456, 878.43570456, 853.16563635, 848.35229002, 848.35229002, 850.75896318, 808.64218283, 817.06553890], 
							[1516.20409280, 1528.23745862, 1492.13736117, 1504.17072698, 1504.17072698, 1419.93716627, 1444.00389791, 1444.00389791, 1444.00389791, 1431.97053209, 1444.00389791, 1468.07062954, 1504.17072698, 1468.07062954, 1419.93716627, 1383.83706883, 1371.80370301, 1371.80370301, 1431.97053209, 1456.03726372, 1419.93716627, 1347.73697138, 1359.77033719, 1335.70360556, 1251.47004485, 1335.70360556, 1468.07062954, 1480.10399535, 1407.90380046, 1383.83706883, 1359.77033719, 1323.67023975, 1323.67023975, 1347.73697138, 1323.67023975, 1323.67023975, 1347.73697138, 1335.70360556, 1383.83706883, 1407.90380046, 1383.83706883, 1371.80370301, 1383.83706883, 1407.90380046, 1359.77033719, 1371.80370301, 1395.87043464, 1371.80370301, 1407.90380046, 1407.90380046, 1407.90380046, 1383.83706883, 1407.90380046, 1395.87043464, 1407.90380046, 1383.83706883, 1407.90380046, 1419.93716627, 1612.47101933, 1648.57111678, 1684.67121422, 1780.93814075, 1768.90477493, 1744.83804330, 1696.70458004, 1660.60448259, 1636.53775096, 1636.53775096, 1564.33755606, 1564.33755606, 1528.23745862, 1540.27082443, 1540.27082443, 1468.07062954, 1552.30419025, 1516.20409280, 1528.23745862, 1504.17072698, 1516.20409280, 1516.20409280, 1516.20409280, 1492.13736117, 1540.27082443, 1540.27082443, 1528.23745862, 1540.27082443, 1564.33755606, 1516.20409280, 1480.10399535, 1516.20409280, 1468.07062954, 1407.90380046, 1431.97053209, 1407.90380046, 1468.07062954, 1371.80370301, 1359.77033719, 1456.03726372, 1468.07062954, 1456.03726372, 1444.00389791, 1468.07062954, 1504.17072698, 1383.83706883, 1456.03726372, 1468.07062954, 1504.17072698, 1492.13736117, 1431.97053209, 1407.90380046, 1480.10399535, 1504.17072698, 1480.10399535, 1468.07062954, 1516.20409280, 1504.17072698, 1492.13736117, 1516.20409280, 1492.13736117, 1468.07062954, 1431.97053209, 1395.87043464, 1407.90380046, 1468.07062954, 1492.13736117, 1444.00389791, 1492.13736117, 1468.07062954, 1468.07062954, 1444.00389791, 1431.97053209, 1444.00389791, 1407.90380046, 1431.97053209, 1383.83706883, 1383.83706883, 1359.77033719, 1371.80370301, 1383.83706883, 1383.83706883, 1371.80370301, 1263.50341067, 1263.50341067, 1251.47004485, 1263.50341067, 1275.53677648, 1311.63687393, 1323.67023975, 1335.70360556, 1335.70360556, 1347.73697138, 1287.57014230, 1311.63687393, 1311.63687393, 1299.60350811, 1287.57014230, 1227.40331322, 1263.50341067, 1203.33658159, 1227.40331322, 1178.06651337, 1131.13638669, 1122.71303062, 1170.84649388, 1203.33658159, 1200.92990842, 1227.40331322, 1194.91322552, 1185.28653286, 1178.06651337, 1150.38977200, 1115.49301113, 1162.42313781, 1152.79644516, 1083.00292343, 1176.86317679, 1202.13324501, 1182.87985970, 1187.69320603, 1185.28653286, 1182.87985970, 1203.33658159, 1215.36994740, 1227.40331322, 1299.60350811, 1275.53677648, 1227.40331322, 1251.47004485, 1251.47004485, 1239.43667904, 1275.53677648, 1203.33658159, 1275.53677648, 1275.53677648, 1299.60350811, 1299.60350811, 1287.57014230, 1311.63687393, 1419.93716627, 1444.00389791, 1468.07062954, 1528.23745862, 1419.93716627, 1468.07062954, 1504.17072698, 1552.30419025, 1540.27082443, 1528.23745862, 1492.13736117, 1468.07062954, 1431.97053209, 1492.13736117, 1480.10399535, 1552.30419025, 1600.43765351, 1612.47101933, 1672.63784841, 1660.60448259, 1600.43765351, 1564.33755606, 1624.50438514, 1648.57111678, 1612.47101933, 1564.33755606, 1564.33755606, 1648.57111678, 1612.47101933, 1600.43765351, 1564.33755606, 1552.30419025, 1564.33755606, 1492.13736117, 1468.07062954, 1468.07062954, 1444.00389791, 1347.73697138, 1431.97053209, 1419.93716627, 1407.90380046, 1468.07062954, 1431.97053209, 1419.93716627, 1468.07062954, 1419.93716627, 1431.97053209, 1480.10399535, 1419.93716627, 1456.03726372, 1383.83706883, 1359.77033719, 1383.83706883, 1395.87043464, 1383.83706883, 1371.80370301, 1215.36994740, 1311.63687393, 1287.57014230, 1275.53677648, 1227.40331322, 1239.43667904, 1239.43667904, 1239.43667904, 1215.36994740, 1215.36994740, 1227.40331322, 1239.43667904, 1180.47318654, 1173.25316705, 1178.06651337, 1203.33658159, 1359.77033719, 1419.93716627, 1444.00389791, 1419.93716627, 1383.83706883, 1431.97053209, 1395.87043464, 1395.87043464, 1407.90380046, 1395.87043464, 1395.87043464, 1371.80370301, 1335.70360556, 1287.57014230, 1287.57014230, 1155.20311832, 1185.28653286, 1175.65984021, 1299.60350811, 1215.36994740, 1227.40331322], 
							[1624.50438514, 1540.27082443, 1456.03726372, 1419.93716627, 1444.00389791, 1468.07062954, 1564.33755606, 1624.50438514, 1624.50438514, 1600.43765351, 1624.50438514, 1540.27082443, 1564.33755606, 1492.13736117, 1444.00389791, 1395.87043464, 1287.57014230, 1263.50341067, 1311.63687393, 1239.43667904, 1193.70988894, 1025.24276751, 1072.17289419, 1048.10616256, 1073.37623078, 1173.25316705, 1311.63687393, 1275.53677648, 1287.57014230, 1203.33658159, 1227.40331322, 1166.03314756, 1119.10302088, 1197.31989868, 1132.33972327, 1143.16975251, 1125.11970378, 1072.17289419, 1108.27299164, 1167.23648414, 1155.20311832, 1191.30321577, 1215.36994740, 1215.36994740, 1202.13324501, 1215.36994740, 1170.84649388, 1166.03314756, 1190.09987919, 1227.40331322, 1169.64315730, 1151.59310858, 1143.16975251, 1138.35640618, 1215.36994740, 1200.92990842, 1251.47004485, 1239.43667904, 1431.97053209, 1456.03726372, 1431.97053209, 1552.30419025, 1552.30419025, 1528.23745862, 1576.37092188, 1588.40428770, 1612.47101933, 1564.33755606, 1492.13736117, 1540.27082443, 1552.30419025, 1588.40428770, 1612.47101933, 1636.53775096, 1660.60448259, 1672.63784841, 1744.83804330, 1744.83804330, 1768.90477493, 1805.00487238, 1780.93814075, 1744.83804330, 1720.77131167, 1780.93814075, 1865.17170146, 1865.17170146, 1853.13833565, 1865.17170146, 1865.17170146, 1829.07160401, 1805.00487238, 1684.67121422, 1768.90477493, 1792.97150657, 1829.07160401, 1792.97150657, 1829.07160401, 1901.27179891, 1901.27179891, 1925.33853054, 1865.17170146, 1949.40526217, 1973.47199380, 1889.23843309, 2069.73892033, 2045.67218870, 2033.63882288, 2093.80565196, 2033.63882288, 1985.50535962, 2021.60545707, 2033.63882288, 2021.60545707, 1997.53872544, 1997.53872544, 2045.67218870, 2021.60545707, 2105.83901778, 2178.03921267, 2226.17267594, 2226.17267594, 2153.97248104, 2178.03921267, 2214.13931012, 2226.17267594, 2166.00584686, 2117.87238359, 2178.03921267, 2141.93911523, 2153.97248104, 2153.97248104, 2202.10594431, 2117.87238359, 2093.80565196, 2105.83901778, 2117.87238359, 2190.07257849, 2190.07257849, 2166.00584686, 2153.97248104, 2166.00584686, 2129.90574941, 2141.93911523, 2057.70555452, 2021.60545707, 2021.60545707, 2045.67218870, 2057.70555452, 2166.00584686, 2105.83901778, 2033.63882288, 1865.17170146, 1841.10496983, 1805.00487238, 1792.97150657, 1696.70458004, 1648.57111678, 1744.83804330, 1696.70458004, 1660.60448259, 1636.53775096, 1696.70458004, 1624.50438514, 1600.43765351, 1696.70458004, 1660.60448259, 1636.53775096, 1480.10399535, 1480.10399535, 1456.03726372, 1456.03726372, 1299.60350811, 1299.60350811, 1359.77033719, 1347.73697138, 1516.20409280, 1588.40428770, 1516.20409280, 1516.20409280, 1612.47101933, 1612.47101933, 1732.80467749, 1612.47101933, 1648.57111678, 1612.47101933, 1696.70458004, 1660.60448259, 1672.63784841, 1756.87140912, 1756.87140912, 1792.97150657, 1696.70458004, 1744.83804330, 1696.70458004, 1648.57111678, 1636.53775096, 1732.80467749, 1901.27179891, 1865.17170146, 1961.43862799, 1985.50535962, 2069.73892033, 1937.37189636, 2021.60545707, 2009.57209125, 1985.50535962, 2045.67218870, 1997.53872544, 1985.50535962, 1973.47199380, 1961.43862799, 1913.30516472, 1949.40526217, 2057.70555452, 2009.57209125, 2081.77228615, 2141.93911523, 2202.10594431, 2153.97248104, 2117.87238359, 2153.97248104, 2141.93911523, 2117.87238359, 2153.97248104, 2226.17267594, 2238.20604175, 2310.40623665, 2262.27277339, 2166.00584686, 2141.93911523, 2081.77228615, 2141.93911523, 2093.80565196, 2153.97248104, 2153.97248104, 2069.73892033, 2093.80565196, 2117.87238359, 2117.87238359, 2214.13931012, 2141.93911523, 2129.90574941, 2166.00584686, 2069.73892033, 2081.77228615, 2069.73892033, 2069.73892033, 2093.80565196, 2141.93911523, 2093.80565196, 2069.73892033, 1973.47199380, 1985.50535962, 1985.50535962, 1913.30516472, 1805.00487238, 1768.90477493, 1997.53872544, 1901.27179891, 1973.47199380, 1841.10496983, 1817.03823820, 1829.07160401, 1805.00487238, 1756.87140912, 1768.90477493, 1768.90477493, 1805.00487238, 1829.07160401, 1961.43862799, 1985.50535962, 1997.53872544, 1973.47199380, 1877.20506728, 1805.00487238, 1865.17170146, 1841.10496983, 1841.10496983, 1841.10496983, 1853.13833565, 1817.03823820, 1768.90477493, 1792.97150657, 1817.03823820, 1612.47101933, 1600.43765351, 1576.37092188, 1456.03726372, 1383.83706883, 1371.80370301, 1383.83706883], 
							[1684.67121422, 1540.27082443, 1564.33755606, 1576.37092188, 1203.33658159, 1111.88300139, 920.55248491, 926.56916782, 818.26887548, 788.18546094, 892.87574354, 939.80587022, 993.95601639, 956.65258236, 902.50243619, 902.50243619, 802.62549992, 737.64532451, 772.54208538, 752.08536349, 776.15209512, 681.08850518, 719.59527579, 700.34189048, 635.36171508, 697.93521732, 800.21882676, 950.63589945, 945.82255313, 960.26259211, 962.66926527, 973.49929450, 890.46907037, 924.16249466, 906.11244594, 872.41902165, 914.53580201, 842.33560711, 913.33246543, 950.63589945, 913.33246543, 909.72245568, 924.16249466, 922.95915808, 853.16563635, 797.81215359, 791.79547068, 755.69537324, 796.60881701, 882.04571430, 847.14895344, 883.24905089, 912.12912884, 953.04257262, 1125.11970378, 1419.93716627, 1359.77033719, 1251.47004485, 1227.40331322, 1311.63687393, 1239.43667904, 1299.60350811, 1263.50341067, 1227.40331322, 1239.43667904, 1263.50341067, 1263.50341067, 1251.47004485, 1203.33658159, 1174.45650363, 1168.43982072, 1173.25316705, 1215.36994740, 1174.45650363, 1215.36994740, 1227.40331322, 1227.40331322, 1263.50341067, 1287.57014230, 1311.63687393, 1311.63687393, 1323.67023975, 1299.60350811, 1347.73697138, 1299.60350811, 1185.28653286, 1141.96641593, 1010.80272853, 854.36897293, 884.45238747, 812.25219257, 761.71205615, 837.52226079, 812.25219257, 890.46907037, 802.62549992, 801.42216334, 861.58899242, 924.16249466, 926.56916782, 882.04571430, 1107.06965506, 1030.05611384, 926.56916782, 975.90596767, 995.15935297, 1042.08947966, 1089.01960634, 1046.90282598, 990.34600665, 1042.08947966, 1031.25945042, 1022.83609435, 1014.41273828, 1031.25945042, 1046.90282598, 1068.56288445, 1105.86631848, 1083.00292343, 1131.13638669, 1117.89968430, 1060.13952838, 1043.29281624, 1020.42942119, 1044.49615282, 1057.73285522, 1044.49615282, 1046.90282598, 1062.54620154, 1028.85277726, 1054.12284547, 1044.49615282, 989.14267007, 1004.78604563, 989.14267007, 927.77250440, 969.88928476, 974.70263109, 969.88928476, 990.34600665, 984.32932374, 950.63589945, 954.24590920, 902.50243619, 894.07908012, 910.92579226, 937.39919706, 1010.80272853, 973.49929450, 983.12598716, 992.75267981, 1022.83609435, 985.53266032, 1016.81941144, 1012.00606512, 1040.88614307, 967.48261160, 918.14581175, 854.36897293, 882.04571430, 873.62235823, 818.26887548, 818.26887548, 830.30224130, 913.33246543, 942.21254338, 949.43256287, 902.50243619, 863.99566558, 847.14895344, 848.35229002, 806.23550966, 784.57545120, 815.86220232, 746.06868058, 880.84237772, 950.63589945, 973.49929450, 975.90596767, 926.56916782, 908.51911910, 986.73599690, 962.66926527, 968.68594818, 1004.78604563, 1009.59939195, 963.87260185, 943.41587996, 998.76936272, 949.43256287, 968.68594818, 926.56916782, 962.66926527, 932.58585073, 966.27927501, 980.71931399, 998.76936272, 1010.80272853, 1085.40959659, 1097.44296241, 1107.06965506, 1122.71303062, 1173.25316705, 1191.30321577, 1323.67023975, 1323.67023975, 1287.57014230, 1311.63687393, 1227.40331322, 1187.69320603, 1287.57014230, 1335.70360556, 1371.80370301, 1419.93716627, 1456.03726372, 1371.80370301, 1444.00389791, 1383.83706883, 1335.70360556, 1311.63687393, 1395.87043464, 1456.03726372, 1431.97053209, 1468.07062954, 1492.13736117, 1564.33755606, 1564.33755606, 1564.33755606, 1540.27082443, 1528.23745862, 1552.30419025, 1456.03726372, 1480.10399535, 1540.27082443, 1612.47101933, 1456.03726372, 1444.00389791, 1576.37092188, 1672.63784841, 1817.03823820, 1732.80467749, 1684.67121422, 1780.93814075, 1708.73794585, 1660.60448259, 1829.07160401, 1817.03823820, 1973.47199380, 2214.13931012, 2141.93911523, 2178.03921267, 2093.80565196, 1829.07160401, 1817.03823820, 1612.47101933, 1552.30419025, 1419.93716627, 1468.07062954, 1407.90380046, 1431.97053209, 1419.93716627, 1407.90380046, 1335.70360556, 1203.33658159, 1299.60350811, 1155.20311832, 859.18231925, 790.59213410, 962.66926527, 878.43570456, 955.44924578, 977.10930425, 1160.01646465, 1227.40331322, 1263.50341067, 1311.63687393, 1287.57014230, 1251.47004485, 1239.43667904, 1105.86631848, 1068.56288445, 1095.03628924, 1101.05297215, 974.70263109, 938.60253364, 874.82569481, 866.40233874, 855.57230951, 872.41902165, 842.33560711, 827.89556813], 
							[1311.63687393, 1263.50341067, 1227.40331322, 1215.36994740, 1239.43667904, 1203.33658159, 1196.11656210, 1162.42313781, 1167.23648414, 1179.26984996, 1251.47004485, 1198.52323526, 1203.33658159, 1227.40331322, 1239.43667904, 1185.28653286, 1192.50655235, 1120.30635746, 1144.37308909, 1162.42313781, 1138.35640618, 1141.96641593, 1113.08633797, 1062.54620154, 986.73599690, 1057.73285522, 1157.60979149, 1157.60979149, 1191.30321577, 1184.08319628, 1182.87985970, 1167.23648414, 1137.15306960, 1161.21980123, 1135.94973302, 1143.16975251, 1131.13638669, 1083.00292343, 1080.59625027, 1132.33972327, 1140.76307935, 1098.64629899, 1141.96641593, 1134.74639644, 1079.39291368, 1083.00292343, 1113.08633797, 1083.00292343, 1161.21980123, 1173.25316705, 1134.74639644, 1172.04983047, 1191.30321577, 1188.89654261, 1215.36994740, 1239.43667904, 1239.43667904, 1335.70360556, 1371.80370301, 1371.80370301, 1335.70360556, 1456.03726372, 1444.00389791, 1383.83706883, 1431.97053209, 1407.90380046, 1419.93716627, 1371.80370301, 1371.80370301, 1335.70360556, 1407.90380046, 1335.70360556, 1371.80370301, 1335.70360556, 1395.87043464, 1383.83706883, 1407.90380046, 1407.90380046, 1431.97053209, 1456.03726372, 1444.00389791, 1407.90380046, 1383.83706883, 1407.90380046, 1395.87043464, 1371.80370301, 1383.83706883, 1359.77033719, 1299.60350811, 1287.57014230, 1203.33658159, 1167.23648414, 1166.03314756, 1137.15306960, 1215.36994740, 1197.31989868, 1172.04983047, 1275.53677648, 1299.60350811, 1287.57014230, 1275.53677648, 1371.80370301, 1275.53677648, 1239.43667904, 1275.53677648, 1251.47004485, 1287.57014230, 1323.67023975, 1227.40331322, 1251.47004485, 1311.63687393, 1323.67023975, 1335.70360556, 1299.60350811, 1311.63687393, 1323.67023975, 1311.63687393, 1287.57014230, 1287.57014230, 1287.57014230, 1335.70360556, 1299.60350811, 1263.50341067, 1251.47004485, 1251.47004485, 1239.43667904, 1287.57014230, 1287.57014230, 1287.57014230, 1275.53677648, 1311.63687393, 1287.57014230, 1239.43667904, 1227.40331322, 1239.43667904, 1227.40331322, 1227.40331322, 1251.47004485, 1239.43667904, 1215.36994740, 1215.36994740, 1239.43667904, 1239.43667904, 1179.26984996, 1193.70988894, 1188.89654261, 1203.33658159, 1203.33658159, 1227.40331322, 1202.13324501, 1215.36994740, 1203.33658159, 1156.40645491, 1203.33658159, 1190.09987919, 1178.06651337, 1125.11970378, 1042.08947966, 1015.61607486, 1031.25945042, 986.73599690, 956.65258236, 943.41587996, 1010.80272853, 1016.81941144, 1009.59939195, 1016.81941144, 974.70263109, 1030.05611384, 986.73599690, 971.09262134, 975.90596767, 965.07593843, 971.09262134, 925.36583124, 1010.80272853, 1021.63275777, 995.15935297, 1010.80272853, 991.54934323, 1003.58270904, 1066.15621129, 1051.71617231, 1125.11970378, 1167.23648414, 1263.50341067, 1156.40645491, 1145.57642567, 1197.31989868, 1166.03314756, 1149.18643542, 1137.15306960, 1134.74639644, 1114.28967455, 1178.06651337, 1179.26984996, 1203.33658159, 1203.33658159, 1215.36994740, 1188.89654261, 1239.43667904, 1287.57014230, 1323.67023975, 1323.67023975, 1323.67023975, 1335.70360556, 1359.77033719, 1335.70360556, 1335.70360556, 1287.57014230, 1263.50341067, 1275.53677648, 1275.53677648, 1383.83706883, 1395.87043464, 1407.90380046, 1419.93716627, 1419.93716627, 1419.93716627, 1395.87043464, 1407.90380046, 1407.90380046, 1444.00389791, 1431.97053209, 1480.10399535, 1492.13736117, 1480.10399535, 1492.13736117, 1468.07062954, 1456.03726372, 1431.97053209, 1419.93716627, 1407.90380046, 1431.97053209, 1431.97053209, 1371.80370301, 1383.83706883, 1395.87043464, 1395.87043464, 1419.93716627, 1407.90380046, 1407.90380046, 1407.90380046, 1383.83706883, 1383.83706883, 1419.93716627, 1407.90380046, 1407.90380046, 1395.87043464, 1371.80370301, 1371.80370301, 1347.73697138, 1323.67023975, 1323.67023975, 1311.63687393, 1287.57014230, 1275.53677648, 1275.53677648, 1287.57014230, 1335.70360556, 1323.67023975, 1323.67023975, 1311.63687393, 1275.53677648, 1299.60350811, 1275.53677648, 1251.47004485, 1287.57014230, 1335.70360556, 1335.70360556, 1383.83706883, 1383.83706883, 1419.93716627, 1383.83706883, 1371.80370301, 1395.87043464, 1395.87043464, 1419.93716627, 1504.17072698, 1504.17072698, 1576.37092188, 1528.23745862, 1552.30419025, 1516.20409280, 1528.23745862, 1516.20409280, 1431.97053209, 1383.83706883, 1419.93716627, 1444.00389791, 1335.70360556], 
							[1175.65984021, 1173.25316705, 1178.06651337, 1173.25316705, 1191.30321577, 1113.08633797, 947.02588971, 878.43570456, 807.43884625, 783.37211461, 948.22922629, 919.34914833, 878.43570456, 902.50243619, 848.35229002, 819.47221206, 764.11872931, 697.93521732, 725.61195870, 752.08536349, 740.05199768, 683.49517834, 732.83197819, 735.23865135, 665.44512962, 687.10518809, 813.45552915, 836.31892420, 879.63904114, 914.53580201, 937.39919706, 818.26887548, 776.15209512, 813.45552915, 790.59213410, 774.94875854, 801.42216334, 754.49203666, 770.13541222, 812.25219257, 806.23550966, 778.55876829, 814.65886573, 770.13541222, 735.23865135, 702.74856365, 697.93521732, 655.81843697, 706.35857339, 746.06868058, 713.57859288, 724.40862212, 755.69537324, 778.55876829, 842.33560711, 841.13227053, 831.50557788, 836.31892420, 831.50557788, 861.58899242, 819.47221206, 878.43570456, 902.50243619, 908.51911910, 888.06239721, 928.97584099, 938.60253364, 891.67240696, 879.63904114, 826.69223155, 842.33560711, 865.19900216, 890.46907037, 854.36897293, 878.43570456, 884.45238747, 877.23236798, 896.48575328, 912.12912884, 908.51911910, 870.01234849, 847.14895344, 859.18231925, 845.94561686, 831.50557788, 783.37211461, 795.40548043, 714.78192946, 605.27830054, 618.51500294, 548.72148120, 475.31794973, 523.45141299, 501.79135452, 507.80803743, 457.26790100, 484.94464238, 523.45141299, 590.83826156, 589.63492498, 558.34817386, 661.83511987, 661.83511987, 601.66829079, 636.56505166, 625.73502243, 624.53168584, 657.02177355, 620.92167610, 602.87162738, 636.56505166, 649.80175406, 652.20842722, 652.20842722, 752.08536349, 742.45867084, 732.83197819, 750.88202691, 725.61195870, 765.32206589, 801.42216334, 736.44198793, 719.59527579, 709.96858314, 713.57859288, 687.10518809, 709.96858314, 699.13855390, 728.01863186, 709.96858314, 722.00194895, 715.98526604, 701.54522707, 708.76524656, 681.08850518, 648.59841748, 655.81843697, 671.46181253, 678.68183202, 669.05513936, 666.64846620, 600.46495421, 601.66829079, 555.94150069, 564.36485676, 569.17820309, 567.97486651, 642.58173457, 608.88831028, 612.49832003, 617.31166635, 623.32834926, 755.69537324, 732.83197819, 723.20528553, 705.15523681, 640.17506140, 600.46495421, 565.56819335, 564.36485676, 542.70479830, 548.72148120, 554.73816411, 593.24493472, 619.71833952, 623.32834926, 628.14169559, 590.83826156, 547.51814462, 553.53482753, 576.39822258, 536.68811539, 495.77467161, 528.26475932, 496.97800820, 588.43158840, 606.48163712, 611.29498345, 567.97486651, 571.58487625, 577.60155916, 637.76838824, 628.14169559, 660.63178329, 663.03845645, 654.61510038, 626.93835901, 637.76838824, 657.02177355, 649.80175406, 655.81843697, 601.66829079, 637.76838824, 605.27830054, 636.56505166, 649.80175406, 631.75170533, 661.83511987, 683.49517834, 658.22511013, 649.80175406, 666.64846620, 705.15523681, 690.71519783, 728.01863186, 729.22196844, 705.15523681, 691.91853441, 688.30852467, 690.71519783, 687.10518809, 706.35857339, 709.96858314, 773.74542196, 782.16877803, 766.52540247, 767.72873905, 772.54208538, 734.03531477, 719.59527579, 767.72873905, 748.47535375, 711.17191972, 717.18860263, 753.28870007, 766.52540247, 772.54208538, 741.25533426, 718.39193921, 728.01863186, 729.22196844, 695.52854416, 659.42844671, 683.49517834, 730.42530502, 659.42844671, 667.85180278, 726.81529528, 743.66200742, 753.28870007, 741.25533426, 743.66200742, 756.89870982, 722.00194895, 709.96858314, 718.39193921, 724.40862212, 754.49203666, 724.40862212, 676.27515885, 715.98526604, 655.81843697, 625.73502243, 602.87162738, 488.55465212, 513.82472034, 499.38468136, 517.43473008, 481.33463264, 490.96132529, 499.38468136, 499.38468136, 481.33463264, 456.06456442, 481.33463264, 444.03119861, 397.10107192, 375.44101346, 399.50774509, 386.27104269, 412.74444748, 440.42118886, 486.14797896, 463.28458391, 499.38468136, 510.21471059, 490.96132529, 498.18134478, 488.55465212, 442.82786202, 429.59115963, 417.55779381, 427.18448646, 354.98429157, 346.56093550, 323.69754045, 315.27418438, 290.00411616, 292.41078933, 276.76741377, 281.58076009], 
							[909.72245568, 890.46907037, 794.20214385, 762.91539273, 758.10204640, 709.96858314, 625.73502243, 679.88516860, 683.49517834, 730.42530502, 743.66200742, 705.15523681, 709.96858314, 715.98526604, 703.95190023, 694.32520758, 673.86848569, 642.58173457, 726.81529528, 746.06868058, 746.06868058, 722.00194895, 759.30538298, 718.39193921, 709.96858314, 772.54208538, 914.53580201, 896.48575328, 860.38565584, 872.41902165, 855.57230951, 824.28555839, 806.23550966, 838.72559737, 821.87888522, 805.03217308, 803.82883650, 767.72873905, 806.23550966, 833.91225104, 818.26887548, 806.23550966, 801.42216334, 795.40548043, 758.10204640, 759.30538298, 742.45867084, 740.05199768, 770.13541222, 805.03217308, 789.38879752, 788.18546094, 742.45867084, 719.59527579, 770.13541222, 806.23550966, 783.37211461, 788.18546094, 836.31892420, 896.48575328, 866.40233874, 884.45238747, 908.51911910, 877.23236798, 871.21568507, 872.41902165, 878.43570456, 870.01234849, 836.31892420, 830.30224130, 833.91225104, 844.74228027, 888.06239721, 859.18231925, 908.51911910, 890.46907037, 892.87574354, 895.28241670, 892.87574354, 907.31578252, 872.41902165, 839.92893395, 853.16563635, 860.38565584, 853.16563635, 853.16563635, 851.96229976, 865.19900216, 818.26887548, 792.99880727, 747.27201717, 688.30852467, 756.89870982, 722.00194895, 782.16877803, 736.44198793, 756.89870982, 770.13541222, 755.69537324, 824.28555839, 780.96544145, 829.09890471, 802.62549992, 750.88202691, 797.81215359, 794.20214385, 820.67554864, 823.08222181, 782.16877803, 758.10204640, 805.03217308, 815.86220232, 806.23550966, 806.23550966, 803.82883650, 814.65886573, 818.26887548, 848.35229002, 865.19900216, 866.40233874, 888.06239721, 859.18231925, 842.33560711, 854.36897293, 890.46907037, 843.53894369, 845.94561686, 850.75896318, 859.18231925, 872.41902165, 876.02903140, 859.18231925, 838.72559737, 849.55562660, 871.21568507, 860.38565584, 859.18231925, 885.65572405, 920.55248491, 896.48575328, 902.50243619, 895.28241670, 926.56916782, 898.89242645, 913.33246543, 926.56916782, 962.66926527, 956.65258236, 1010.80272853, 992.75267981, 956.65258236, 931.38251415, 885.65572405, 884.45238747, 848.35229002, 807.43884625, 811.04885599, 815.86220232, 797.81215359, 786.98212436, 758.10204640, 827.89556813, 777.35543171, 801.42216334, 835.11558762, 839.92893395, 848.35229002, 836.31892420, 823.08222181, 817.06553890, 770.13541222, 688.30852467, 630.54836875, 653.41176380, 624.53168584, 723.20528553, 766.52540247, 750.88202691, 734.03531477, 723.20528553, 740.05199768, 805.03217308, 794.20214385, 838.72559737, 812.25219257, 803.82883650, 735.23865135, 759.30538298, 807.43884625, 808.64218283, 808.64218283, 823.08222181, 818.26887548, 802.62549992, 821.87888522, 783.37211461, 800.21882676, 845.94561686, 894.07908012, 878.43570456, 900.09576303, 925.36583124, 916.94247517, 885.65572405, 918.14581175, 934.99252389, 921.75582150, 908.51911910, 883.24905089, 894.07908012, 877.23236798, 874.82569481, 906.11244594, 913.33246543, 925.36583124, 920.55248491, 956.65258236, 990.34600665, 956.65258236, 931.38251415, 950.63589945, 921.75582150, 934.99252389, 945.82255313, 973.49929450, 1016.81941144, 993.95601639, 1009.59939195, 966.27927501, 954.24590920, 922.95915808, 915.73913859, 900.09576303, 925.36583124, 963.87260185, 902.50243619, 908.51911910, 930.17917757, 918.14581175, 949.43256287, 960.26259211, 920.55248491, 983.12598716, 945.82255313, 926.56916782, 968.68594818, 959.05925553, 967.48261160, 974.70263109, 928.97584099, 950.63589945, 902.50243619, 920.55248491, 924.16249466, 764.11872931, 843.53894369, 812.25219257, 833.91225104, 812.25219257, 897.68908986, 896.48575328, 873.62235823, 824.28555839, 815.86220232, 832.70891446, 817.06553890, 797.81215359, 830.30224130, 850.75896318, 862.79232900, 914.53580201, 925.36583124, 931.38251415, 866.40233874, 856.77564609, 889.26573379, 901.29909961, 931.38251415, 903.70577277, 862.79232900, 865.19900216, 859.18231925, 902.50243619, 886.85906063, 896.48575328, 877.23236798, 836.31892420, 835.11558762, 871.21568507, 806.23550966, 813.45552915], 
							[856.77564609, 830.30224130, 791.79547068, 783.37211461, 799.01549017, 789.38879752, 805.03217308, 838.72559737, 851.96229976, 830.30224130, 854.36897293, 830.30224130, 812.25219257, 794.20214385, 771.33874880, 761.71205615, 758.10204640, 703.95190023, 761.71205615, 746.06868058, 746.06868058, 689.51186125, 706.35857339, 670.25847594, 651.00509064, 670.25847594, 734.03531477, 766.52540247, 752.08536349, 722.00194895, 722.00194895, 667.85180278, 681.08850518, 703.95190023, 670.25847594, 673.86848569, 643.78507115, 613.70165661, 638.97172482, 672.66514911, 648.59841748, 697.93521732, 689.51186125, 679.88516860, 661.83511987, 661.83511987, 637.76838824, 628.14169559, 651.00509064, 673.86848569, 631.75170533, 634.15837850, 638.97172482, 619.71833952, 740.05199768, 723.20528553, 794.20214385, 770.13541222, 849.55562660, 878.43570456, 838.72559737, 962.66926527, 955.44924578, 968.68594818, 938.60253364, 926.56916782, 957.85591894, 938.60253364, 914.53580201, 876.02903140, 851.96229976, 890.46907037, 925.36583124, 904.90910935, 907.31578252, 885.65572405, 956.65258236, 938.60253364, 1003.58270904, 1007.19271879, 1004.78604563, 962.66926527, 955.44924578, 986.73599690, 1036.07279675, 1042.08947966, 1044.49615282, 1031.25945042, 987.93933348, 980.71931399, 957.85591894, 902.50243619, 938.60253364, 980.71931399, 1022.83609435, 987.93933348, 1015.61607486, 1057.73285522, 1056.52951863, 1117.89968430, 1101.05297215, 1275.53677648, 1263.50341067, 1184.08319628, 1251.47004485, 1227.40331322, 1227.40331322, 1299.60350811, 1239.43667904, 1203.33658159, 1227.40331322, 1251.47004485, 1251.47004485, 1227.40331322, 1227.40331322, 1263.50341067, 1251.47004485, 1311.63687393, 1335.70360556, 1395.87043464, 1395.87043464, 1335.70360556, 1359.77033719, 1335.70360556, 1311.63687393, 1263.50341067, 1263.50341067, 1263.50341067, 1287.57014230, 1299.60350811, 1299.60350811, 1347.73697138, 1275.53677648, 1251.47004485, 1275.53677648, 1251.47004485, 1299.60350811, 1323.67023975, 1287.57014230, 1311.63687393, 1287.57014230, 1251.47004485, 1251.47004485, 1203.33658159, 1227.40331322, 1184.08319628, 1169.64315730, 1179.26984996, 1215.36994740, 1188.89654261, 1157.60979149, 1076.98624052, 1054.12284547, 1080.59625027, 1096.23962583, 1098.64629899, 1070.96955761, 1083.00292343, 1043.29281624, 989.14267007, 960.26259211, 1040.88614307, 1005.98938221, 1003.58270904, 1046.90282598, 1033.66612358, 1040.88614307, 969.88928476, 996.36268955, 981.92265058, 980.71931399, 926.56916782, 933.78918731, 1003.58270904, 1016.81941144, 1086.61293317, 1151.59310858, 1113.08633797, 1132.33972327, 1186.48986945, 1186.48986945, 1263.50341067, 1239.43667904, 1299.60350811, 1407.90380046, 1480.10399535, 1395.87043464, 1504.17072698, 1492.13736117, 1456.03726372, 1492.13736117, 1431.97053209, 1492.13736117, 1431.97053209, 1407.90380046, 1359.77033719, 1444.00389791, 1492.13736117, 1419.93716627, 1395.87043464, 1383.83706883, 1444.00389791, 1395.87043464, 1383.83706883, 1359.77033719, 1347.73697138, 1359.77033719, 1323.67023975, 1287.57014230, 1311.63687393, 1198.52323526, 1172.04983047, 1169.64315730, 1187.69320603, 1227.40331322, 1251.47004485, 1287.57014230, 1299.60350811, 1299.60350811, 1239.43667904, 1227.40331322, 1185.28653286, 1179.26984996, 1169.64315730, 1227.40331322, 1203.33658159, 1203.33658159, 1203.33658159, 1192.50655235, 1169.64315730, 1152.79644516, 1139.55974276, 1157.60979149, 1203.33658159, 1193.70988894, 1178.06651337, 1170.84649388, 1186.48986945, 1182.87985970, 1275.53677648, 1263.50341067, 1263.50341067, 1263.50341067, 1203.33658159, 1215.36994740, 1227.40331322, 1227.40331322, 1287.57014230, 1311.63687393, 1287.57014230, 1311.63687393, 1275.53677648, 1311.63687393, 1299.60350811, 1323.67023975, 1347.73697138, 1431.97053209, 1468.07062954, 1431.97053209, 1492.13736117, 1468.07062954, 1431.97053209, 1419.93716627, 1444.00389791, 1504.17072698, 1492.13736117, 1528.23745862, 1552.30419025, 1540.27082443, 1576.37092188, 1648.57111678, 1624.50438514, 1636.53775096, 1636.53775096, 1708.73794585, 1768.90477493, 1780.93814075, 1889.23843309, 1937.37189636, 1841.10496983, 1961.43862799, 2033.63882288, 2021.60545707, 2117.87238359, 1985.50535962, 2045.67218870, 1913.30516472, 1732.80467749, 1768.90477493, 1720.77131167, 1792.97150657], 
							[1153.99978174, 1144.37308909, 1120.30635746, 1115.49301113, 1115.49301113, 1095.03628924, 1036.07279675, 1039.68280649, 1044.49615282, 1024.03943093, 1074.57956736, 1070.96955761, 1048.10616256, 1034.86946017, 1056.52951863, 1044.49615282, 1015.61607486, 944.61921655, 957.85591894, 937.39919706, 974.70263109, 914.53580201, 913.33246543, 878.43570456, 819.47221206, 866.40233874, 974.70263109, 986.73599690, 938.60253364, 956.65258236, 963.87260185, 950.63589945, 962.66926527, 986.73599690, 968.68594818, 949.43256287, 955.44924578, 955.44924578, 943.41587996, 951.83923604, 934.99252389, 954.24590920, 979.51597741, 974.70263109, 915.73913859, 896.48575328, 882.04571430, 867.60567532, 890.46907037, 890.46907037, 908.51911910, 900.09576303, 878.43570456, 843.53894369, 904.90910935, 931.38251415, 962.66926527, 1034.86946017, 1014.41273828, 1021.63275777, 1024.03943093, 1081.79958685, 1092.62961608, 1057.73285522, 1066.15621129, 1070.96955761, 1060.13952838, 1043.29281624, 1019.22608460, 1005.98938221, 974.70263109, 1010.80272853, 1046.90282598, 1034.86946017, 1083.00292343, 1063.74953812, 1119.10302088, 1085.40959659, 1084.20626001, 1117.89968430, 1114.28967455, 1105.86631848, 1099.84963557, 1086.61293317, 1099.84963557, 1086.61293317, 1144.37308909, 1137.15306960, 1092.62961608, 1069.76622103, 1036.07279675, 998.76936272, 997.56602614, 986.73599690, 1070.96955761, 1004.78604563, 1044.49615282, 1045.69948940, 1105.86631848, 1078.18957710, 1089.01960634, 1167.23648414, 1144.37308909, 1157.60979149, 1215.36994740, 1239.43667904, 1227.40331322, 1251.47004485, 1192.50655235, 1215.36994740, 1251.47004485, 1239.43667904, 1196.11656210, 1187.69320603, 1199.72657184, 1203.33658159, 1203.33658159, 1215.36994740, 1275.53677648, 1275.53677648, 1287.57014230, 1239.43667904, 1227.40331322, 1227.40331322, 1239.43667904, 1179.26984996, 1215.36994740, 1215.36994740, 1203.33658159, 1215.36994740, 1215.36994740, 1239.43667904, 1190.09987919, 1161.21980123, 1172.04983047, 1167.23648414, 1197.31989868, 1203.33658159, 1203.33658159, 1203.33658159, 1203.33658159, 1184.08319628, 1192.50655235, 1185.28653286, 1188.89654261, 1174.45650363, 1215.36994740, 1227.40331322, 1227.40331322, 1239.43667904, 1227.40331322, 1184.08319628, 1131.13638669, 1145.57642567, 1075.78290394, 1028.85277726, 977.10930425, 991.54934323, 987.93933348, 967.48261160, 949.43256287, 950.63589945, 912.12912884, 963.87260185, 1033.66612358, 1004.78604563, 1028.85277726, 986.73599690, 998.76936272, 919.34914833, 959.05925553, 918.14581175, 951.83923604, 978.31264083, 928.97584099, 1034.86946017, 1009.59939195, 1022.83609435, 1037.27613333, 1084.20626001, 1055.32618205, 1135.94973302, 1131.13638669, 1116.69634771, 1138.35640618, 1194.91322552, 1095.03628924, 1139.55974276, 1152.79644516, 1115.49301113, 1146.77976225, 1092.62961608, 1139.55974276, 1128.72971353, 1158.81312807, 1173.25316705, 1215.36994740, 1227.40331322, 1239.43667904, 1227.40331322, 1263.50341067, 1347.73697138, 1239.43667904, 1299.60350811, 1371.80370301, 1371.80370301, 1335.70360556, 1323.67023975, 1299.60350811, 1263.50341067, 1275.53677648, 1251.47004485, 1299.60350811, 1359.77033719, 1395.87043464, 1431.97053209, 1468.07062954, 1492.13736117, 1431.97053209, 1456.03726372, 1444.00389791, 1468.07062954, 1444.00389791, 1444.00389791, 1456.03726372, 1227.40331322, 1172.04983047, 1191.30321577, 1179.26984996, 1178.06651337, 1141.96641593, 1162.42313781, 1145.57642567, 1149.18643542, 1134.74639644, 1103.45964532, 1078.18957710, 1107.06965506, 1113.08633797, 1170.84649388, 1099.84963557, 1127.52637695, 1153.99978174, 1091.42627950, 1090.22294292, 1147.98309883, 1140.76307935, 1166.03314756, 1161.21980123, 1126.32304037, 1163.62647440, 1147.98309883, 1108.27299164, 1098.64629899, 1075.78290394, 1113.08633797, 1070.96955761, 1076.98624052, 1101.05297215, 1063.74953812, 1111.88300139, 1091.42627950, 1030.05611384, 957.85591894, 1083.00292343, 1056.52951863, 1038.47946991, 1014.41273828, 1010.80272853, 977.10930425, 1046.90282598, 1072.17289419, 1139.55974276, 1227.40331322, 1239.43667904, 1215.36994740, 1263.50341067, 1215.36994740, 1287.57014230, 1263.50341067, 1323.67023975, 1359.77033719, 1287.57014230, 1227.40331322, 1275.53677648, 1174.45650363, 1182.87985970, 1203.33658159, 1179.26984996, 1064.95287471, 1052.91950889], 
							[830.30224130, 817.06553890, 800.21882676, 772.54208538, 770.13541222, 667.85180278, 610.09164687, 620.92167610, 589.63492498, 584.82157865, 624.53168584, 602.87162738, 631.75170533, 657.02177355, 691.91853441, 718.39193921, 652.20842722, 622.12501268, 657.02177355, 649.80175406, 628.14169559, 587.22825181, 611.29498345, 584.82157865, 551.12815437, 581.21156891, 649.80175406, 671.46181253, 667.85180278, 685.90185151, 683.49517834, 673.86848569, 654.61510038, 654.61510038, 651.00509064, 623.32834926, 631.75170533, 589.63492498, 631.75170533, 654.61510038, 630.54836875, 620.92167610, 620.92167610, 613.70165661, 584.82157865, 576.39822258, 571.58487625, 577.60155916, 607.68497370, 635.36171508, 608.88831028, 618.51500294, 611.29498345, 614.90499319, 669.05513936, 664.24179304, 644.98840773, 620.92167610, 648.59841748, 688.30852467, 661.83511987, 701.54522707, 709.96858314, 723.20528553, 724.40862212, 728.01863186, 747.27201717, 719.59527579, 685.90185151, 673.86848569, 673.86848569, 672.66514911, 681.08850518, 670.25847594, 687.10518809, 702.74856365, 702.74856365, 711.17191972, 728.01863186, 724.40862212, 732.83197819, 709.96858314, 695.52854416, 719.59527579, 684.69851492, 661.83511987, 655.81843697, 619.71833952, 553.53482753, 551.12815437, 523.45141299, 512.62138376, 546.31480804, 509.01137401, 539.09478855, 495.77467161, 530.67143248, 541.50146171, 590.83826156, 600.46495421, 571.58487625, 649.80175406, 619.71833952, 551.12815437, 577.60155916, 595.65160789, 611.29498345, 643.78507115, 595.65160789, 577.60155916, 589.63492498, 607.68497370, 613.70165661, 610.09164687, 632.95504192, 629.34503217, 631.75170533, 635.36171508, 646.19174431, 648.59841748, 657.02177355, 623.32834926, 630.54836875, 620.92167610, 634.15837850, 628.14169559, 637.76838824, 618.51500294, 635.36171508, 602.87162738, 623.32834926, 626.93835901, 589.63492498, 605.27830054, 584.82157865, 554.73816411, 561.95818360, 566.77152993, 577.60155916, 578.80489574, 590.83826156, 564.36485676, 557.14483728, 521.04473983, 575.19488600, 537.89145197, 553.53482753, 599.26161763, 565.56819335, 553.53482753, 540.29812513, 535.48477881, 512.62138376, 516.23139350, 522.24807641, 537.89145197, 502.99469110, 478.92795947, 458.47123758, 470.50460340, 487.35131554, 445.23453519, 451.25121810, 454.86122784, 492.16466187, 494.57133503, 501.79135452, 475.31794973, 478.92795947, 446.43787177, 462.08124733, 435.60784253, 400.71108167, 422.37114014, 399.50774509, 481.33463264, 469.30126682, 481.33463264, 501.79135452, 471.70793998, 459.67457417, 504.19802769, 513.82472034, 541.50146171, 534.28144222, 552.33149095, 494.57133503, 496.97800820, 527.06142274, 494.57133503, 495.77467161, 482.53796922, 516.23139350, 488.55465212, 523.45141299, 509.01137401, 536.68811539, 548.72148120, 558.34817386, 553.53482753, 551.12815437, 583.61824207, 586.02491523, 595.65160789, 644.98840773, 664.24179304, 625.73502243, 612.49832003, 601.66829079, 605.27830054, 594.44827130, 612.49832003, 619.71833952, 770.13541222, 824.28555839, 794.20214385, 748.47535375, 734.03531477, 711.17191972, 685.90185151, 722.00194895, 723.20528553, 706.35857339, 694.32520758, 689.51186125, 722.00194895, 720.79861237, 714.78192946, 685.90185151, 673.86848569, 669.05513936, 623.32834926, 611.29498345, 616.10832977, 635.36171508, 584.82157865, 580.00823233, 637.76838824, 646.19174431, 653.41176380, 601.66829079, 598.05828105, 619.71833952, 583.61824207, 617.31166635, 637.76838824, 617.31166635, 613.70165661, 602.87162738, 566.77152993, 592.04159814, 565.56819335, 519.84140325, 517.43473008, 435.60784253, 458.47123758, 463.28458391, 493.36799845, 470.50460340, 478.92795947, 498.18134478, 489.75798871, 477.72462289, 463.28458391, 487.35131554, 446.43787177, 409.13443774, 419.96446697, 436.81117912, 422.37114014, 451.25121810, 468.09793024, 515.02805692, 495.77467161, 511.41804717, 528.26475932, 506.60470085, 498.18134478, 493.36799845, 470.50460340, 450.04788151, 442.82786202, 438.01451570, 394.69439876, 393.49106218, 365.81432080, 344.15426233, 341.74758917, 354.98429157, 340.54425259, 324.90087703], 
							[642.58173457, 625.73502243, 594.44827130, 581.21156891, 571.58487625, 559.55151044, 561.95818360, 552.33149095, 552.33149095, 529.46809590, 589.63492498, 583.61824207, 557.14483728, 551.12815437, 533.07810564, 500.58801794, 494.57133503, 452.45455468, 482.53796922, 490.96132529, 499.38468136, 464.48792049, 481.33463264, 469.30126682, 445.23453519, 475.31794973, 529.46809590, 540.29812513, 524.65474957, 535.48477881, 534.28144222, 500.58801794, 505.40136427, 512.62138376, 522.24807641, 493.36799845, 493.36799845, 468.09793024, 484.94464238, 505.40136427, 496.97800820, 477.72462289, 490.96132529, 475.31794973, 457.26790100, 452.45455468, 452.45455468, 458.47123758, 459.67457417, 494.57133503, 464.48792049, 478.92795947, 513.82472034, 517.43473008, 586.02491523, 587.22825181, 725.61195870, 843.53894369, 853.16563635, 818.26887548, 791.79547068, 842.33560711, 848.35229002, 829.09890471, 825.48889497, 855.57230951, 850.75896318, 824.28555839, 797.81215359, 767.72873905, 752.08536349, 754.49203666, 812.25219257, 759.30538298, 789.38879752, 773.74542196, 838.72559737, 808.64218283, 826.69223155, 825.48889497, 800.21882676, 773.74542196, 754.49203666, 752.08536349, 782.16877803, 794.20214385, 794.20214385, 755.69537324, 695.52854416, 707.56190997, 655.81843697, 565.56819335, 614.90499319, 613.70165661, 661.83511987, 632.95504192, 625.73502243, 742.45867084, 736.44198793, 772.54208538, 728.01863186, 830.30224130, 853.16563635, 813.45552915, 866.40233874, 862.79232900, 866.40233874, 884.45238747, 860.38565584, 825.48889497, 842.33560711, 842.33560711, 821.87888522, 827.89556813, 826.69223155, 831.50557788, 818.26887548, 842.33560711, 892.87574354, 932.58585073, 909.72245568, 882.04571430, 862.79232900, 870.01234849, 879.63904114, 845.94561686, 848.35229002, 851.96229976, 842.33560711, 845.94561686, 830.30224130, 838.72559737, 818.26887548, 772.54208538, 782.16877803, 770.13541222, 791.79547068, 820.67554864, 806.23550966, 800.21882676, 801.42216334, 743.66200742, 720.79861237, 677.47849543, 695.52854416, 697.93521732, 706.35857339, 760.50871956, 777.35543171, 749.67869033, 750.88202691, 711.17191972, 735.23865135, 730.42530502, 719.59527579, 715.98526604, 678.68183202, 637.76838824, 635.36171508, 607.68497370, 577.60155916, 595.65160789, 567.97486651, 589.63492498, 623.32834926, 619.71833952, 622.12501268, 589.63492498, 586.02491523, 551.12815437, 545.11147146, 505.40136427, 499.38468136, 518.63806666, 481.33463264, 565.56819335, 592.04159814, 571.58487625, 555.94150069, 554.73816411, 598.05828105, 624.53168584, 626.93835901, 672.66514911, 643.78507115, 659.42844671, 601.66829079, 625.73502243, 626.93835901, 605.27830054, 618.51500294, 541.50146171, 559.55151044, 559.55151044, 558.34817386, 545.11147146, 583.61824207, 586.02491523, 590.83826156, 601.66829079, 607.68497370, 628.14169559, 623.32834926, 623.32834926, 628.14169559, 659.42844671, 720.79861237, 661.83511987, 664.24179304, 670.25847594, 640.17506140, 652.20842722, 706.35857339, 725.61195870, 725.61195870, 746.06868058, 756.89870982, 753.28870007, 755.69537324, 709.96858314, 750.88202691, 742.45867084, 765.32206589, 794.20214385, 785.77878778, 802.62549992, 788.18546094, 777.35543171, 758.10204640, 762.91539273, 765.32206589, 754.49203666, 750.88202691, 767.72873905, 764.11872931, 713.57859288, 711.17191972, 746.06868058, 747.27201717, 774.94875854, 762.91539273, 765.32206589, 771.33874880, 719.59527579, 738.84866109, 746.06868058, 740.05199768, 750.88202691, 718.39193921, 691.91853441, 700.34189048, 673.86848569, 632.95504192, 660.63178329, 581.21156891, 630.54836875, 576.39822258, 620.92167610, 604.07496396, 677.47849543, 689.51186125, 675.07182227, 677.47849543, 649.80175406, 684.69851492, 673.86848569, 625.73502243, 644.98840773, 651.00509064, 678.68183202, 753.28870007, 726.81529528, 768.93207563, 713.57859288, 830.30224130, 824.28555839, 831.50557788, 862.79232900, 856.77564609, 876.02903140, 871.21568507, 939.80587022, 932.58585073, 948.22922629, 924.16249466, 934.99252389, 895.28241670, 818.26887548, 848.35229002, 747.27201717, 755.69537324], 
							[1335.70360556, 1239.43667904, 1188.89654261, 1239.43667904, 1197.31989868, 1227.40331322, 1101.05297215, 1087.81626976, 1072.17289419, 1049.30949914, 1129.93305011, 1107.06965506, 1111.88300139, 1125.11970378, 1036.07279675, 986.73599690, 945.82255313, 904.90910935, 939.80587022, 962.66926527, 967.48261160, 901.29909961, 974.70263109, 963.87260185, 903.70577277, 1099.84963557, 1239.43667904, 1215.36994740, 1161.21980123, 1133.54305986, 1155.20311832, 1102.25630873, 1042.08947966, 1064.95287471, 1031.25945042, 992.75267981, 1018.02274802, 990.34600665, 1062.54620154, 1057.73285522, 1010.80272853, 1086.61293317, 1138.35640618, 1143.16975251, 1086.61293317, 1095.03628924, 1090.22294292, 1076.98624052, 1092.62961608, 1091.42627950, 1049.30949914, 1073.37623078, 1081.79958685, 1067.35954787, 1215.36994740, 1263.50341067, 1263.50341067, 1516.20409280, 1383.83706883, 1359.77033719, 1371.80370301, 1371.80370301, 1431.97053209, 1347.73697138, 1323.67023975, 1311.63687393, 1359.77033719, 1323.67023975, 1239.43667904, 1203.33658159, 1227.40331322, 1239.43667904, 1311.63687393, 1239.43667904, 1275.53677648, 1287.57014230, 1311.63687393, 1275.53677648, 1287.57014230, 1323.67023975, 1299.60350811, 1287.57014230, 1275.53677648, 1275.53677648, 1299.60350811, 1263.50341067, 1299.60350811, 1263.50341067, 1215.36994740, 1239.43667904, 1203.33658159, 1107.06965506, 1179.26984996, 1193.70988894, 1197.31989868, 1138.35640618, 1108.27299164, 1139.55974276, 1203.33658159, 1263.50341067, 1239.43667904, 1311.63687393, 1311.63687393, 1239.43667904, 1239.43667904, 1263.50341067, 1239.43667904, 1263.50341067, 1263.50341067, 1263.50341067, 1347.73697138, 1371.80370301, 1371.80370301, 1419.93716627, 1419.93716627, 1516.20409280, 1492.13736117, 1504.17072698, 1516.20409280, 1540.27082443, 1528.23745862, 1456.03726372, 1456.03726372, 1468.07062954, 1456.03726372, 1456.03726372, 1468.07062954, 1456.03726372, 1492.13736117, 1480.10399535, 1456.03726372, 1468.07062954, 1395.87043464, 1419.93716627, 1444.00389791, 1431.97053209, 1431.97053209, 1395.87043464, 1359.77033719, 1359.77033719, 1371.80370301, 1287.57014230, 1263.50341067, 1263.50341067, 1275.53677648, 1287.57014230, 1311.63687393, 1335.70360556, 1287.57014230, 1287.57014230, 1275.53677648, 1251.47004485, 1158.81312807, 1275.53677648, 1227.40331322, 1215.36994740, 1203.33658159, 1179.26984996, 1120.30635746, 1120.30635746, 1090.22294292, 1175.65984021, 1104.66298190, 1150.38977200, 1215.36994740, 1215.36994740, 1227.40331322, 1163.62647440, 1176.86317679, 1155.20311832, 1134.74639644, 1122.71303062, 1132.33972327, 1194.91322552, 1149.18643542, 1263.50341067, 1275.53677648, 1251.47004485, 1299.60350811, 1347.73697138, 1323.67023975, 1419.93716627, 1407.90380046, 1371.80370301, 1395.87043464, 1335.70360556, 1323.67023975, 1335.70360556, 1383.83706883, 1275.53677648, 1227.40331322, 1239.43667904, 1287.57014230, 1311.63687393, 1299.60350811, 1359.77033719, 1444.00389791, 1468.07062954, 1528.23745862, 1540.27082443, 1552.30419025, 1636.53775096, 1588.40428770, 1576.37092188, 1576.37092188, 1576.37092188, 1600.43765351, 1600.43765351, 1552.30419025, 1552.30419025, 1564.33755606, 1612.47101933, 1672.63784841, 1768.90477493, 1756.87140912, 1720.77131167, 1720.77131167, 1768.90477493, 1744.83804330, 1744.83804330, 1780.93814075, 1756.87140912, 1817.03823820, 1756.87140912, 1805.00487238, 1780.93814075, 1817.03823820, 1829.07160401, 1756.87140912, 1720.77131167, 1684.67121422, 1708.73794585, 1660.60448259, 1720.77131167, 1732.80467749, 1600.43765351, 1612.47101933, 1684.67121422, 1744.83804330, 1817.03823820, 1780.93814075, 1732.80467749, 1841.10496983, 1708.73794585, 1732.80467749, 1732.80467749, 1744.83804330, 1744.83804330, 1732.80467749, 1576.37092188, 1624.50438514, 1564.33755606, 1456.03726372, 1431.97053209, 1323.67023975, 1444.00389791, 1468.07062954, 1624.50438514, 1516.20409280, 1648.57111678, 1744.83804330, 1708.73794585, 1720.77131167, 1684.67121422, 1732.80467749, 1684.67121422, 1636.53775096, 1720.77131167, 1805.00487238, 1925.33853054, 1901.27179891, 1901.27179891, 1985.50535962, 1961.43862799, 1913.30516472, 2045.67218870, 1985.50535962, 1913.30516472, 1985.50535962, 1949.40526217, 1937.37189636, 1949.40526217, 2117.87238359, 1949.40526217, 1949.40526217, 2057.70555452, 1961.43862799, 2033.63882288, 1961.43862799, 1961.43862799, 2009.57209125], 
							[1072.17289419, 1108.27299164, 1064.95287471, 966.27927501, 927.77250440, 879.63904114, 878.43570456, 914.53580201, 925.36583124, 968.68594818, 1033.66612358, 992.75267981, 974.70263109, 926.56916782, 927.77250440, 912.12912884, 854.36897293, 872.41902165, 882.04571430, 920.55248491, 914.53580201, 890.46907037, 866.40233874, 783.37211461, 677.47849543, 824.28555839, 938.60253364, 950.63589945, 981.92265058, 932.58585073, 986.73599690, 918.14581175, 931.38251415, 928.97584099, 991.54934323, 960.26259211, 937.39919706, 891.67240696, 930.17917757, 944.61921655, 914.53580201, 922.95915808, 942.21254338, 932.58585073, 919.34914833, 910.92579226, 896.48575328, 856.77564609, 877.23236798, 891.67240696, 908.51911910, 908.51911910, 876.02903140, 872.41902165, 927.77250440, 941.00920680, 930.17917757, 990.34600665, 1028.85277726, 1026.44610409, 1028.85277726, 1058.93619180, 1067.35954787, 1323.67023975, 1323.67023975, 1287.57014230, 1275.53677648, 1227.40331322, 1172.04983047, 1161.21980123, 1263.50341067, 1299.60350811, 1299.60350811, 1239.43667904, 1263.50341067, 1263.50341067, 1299.60350811, 1287.57014230, 1311.63687393, 1347.73697138, 1323.67023975, 1299.60350811, 1335.70360556, 1323.67023975, 1299.60350811, 1299.60350811, 1263.50341067, 1251.47004485, 1161.21980123, 1140.76307935, 1070.96955761, 985.53266032, 1058.93619180, 1061.34286496, 1083.00292343, 1028.85277726, 1034.86946017, 1044.49615282, 1128.72971353, 1122.71303062, 1086.61293317, 1215.36994740, 1202.13324501, 1121.50969404, 1190.09987919, 1199.72657184, 1179.26984996, 1182.87985970, 1109.47632822, 1097.44296241, 1125.11970378, 1143.16975251, 1137.15306960, 1116.69634771, 1129.93305011, 1143.16975251, 1153.99978174, 1163.62647440, 1126.32304037, 1152.79644516, 1121.50969404, 1091.42627950, 1092.62961608, 1196.11656210, 1178.06651337, 1117.89968430, 1117.89968430, 1101.05297215, 1117.89968430, 1102.25630873, 1131.13638669, 1119.10302088, 1119.10302088, 1089.01960634, 1066.15621129, 1037.27613333, 1058.93619180, 1068.56288445, 1052.91950889, 1058.93619180, 1070.96955761, 1067.35954787, 1022.83609435, 985.53266032, 1044.49615282, 1051.71617231, 1070.96955761, 1110.67966481, 1073.37623078, 1049.30949914, 1076.98624052, 1057.73285522, 1002.37937246, 1046.90282598, 1055.32618205, 1049.30949914, 999.97269930, 951.83923604, 866.40233874, 872.41902165, 896.48575328, 995.15935297, 973.49929450, 954.24590920, 1002.37937246, 1003.58270904, 1009.59939195, 969.88928476, 1002.37937246, 948.22922629, 972.29595792, 903.70577277, 856.77564609, 949.43256287, 901.29909961, 1019.22608460, 980.71931399, 986.73599690, 986.73599690, 978.31264083, 969.88928476, 1031.25945042, 1064.95287471, 1070.96955761, 1108.27299164, 1105.86631848, 1028.85277726, 1048.10616256, 1117.89968430, 1075.78290394, 1076.98624052, 1022.83609435, 1046.90282598, 1021.63275777, 1069.76622103, 1046.90282598, 1034.86946017, 1081.79958685, 1042.08947966, 1048.10616256, 1045.69948940, 1080.59625027, 1068.56288445, 1099.84963557, 1107.06965506, 1072.17289419, 1091.42627950, 1084.20626001, 1070.96955761, 1026.44610409, 1075.78290394, 1073.37623078, 1092.62961608, 1108.27299164, 1120.30635746, 1114.28967455, 1153.99978174, 1173.25316705, 1129.93305011, 1117.89968430, 1153.99978174, 1133.54305986, 1129.93305011, 1107.06965506, 1145.57642567, 1155.20311832, 1133.54305986, 1121.50969404, 1104.66298190, 1122.71303062, 1102.25630873, 1113.08633797, 1080.59625027, 1095.03628924, 1105.86631848, 1021.63275777, 1018.02274802, 1104.66298190, 1091.42627950, 1109.47632822, 1099.84963557, 1097.44296241, 1132.33972327, 1081.79958685, 1055.32618205, 1119.10302088, 1113.08633797, 1127.52637695, 1073.37623078, 1044.49615282, 1105.86631848, 1039.68280649, 991.54934323, 995.15935297, 919.34914833, 974.70263109, 908.51911910, 965.07593843, 926.56916782, 943.41587996, 953.04257262, 903.70577277, 865.19900216, 829.09890471, 885.65572405, 873.62235823, 886.85906063, 936.19586048, 953.04257262, 1022.83609435, 1046.90282598, 1060.13952838, 1110.67966481, 1083.00292343, 1040.88614307, 1105.86631848, 1109.47632822, 1074.57956736, 1079.39291368, 1040.88614307, 1034.86946017, 1002.37937246, 1007.19271879, 966.27927501, 989.14267007, 922.95915808, 957.85591894, 975.90596767, 945.82255313, 924.16249466, 973.49929450], 
							[1468.07062954, 1383.83706883, 1335.70360556, 1335.70360556, 1347.73697138, 1167.23648414, 1125.11970378, 1227.40331322, 1251.47004485, 1215.36994740, 1239.43667904, 1227.40331322, 1227.40331322, 1251.47004485, 1203.33658159, 1179.26984996, 1107.06965506, 1095.03628924, 1144.37308909, 1196.11656210, 1145.57642567, 1113.08633797, 1169.64315730, 1102.25630873, 1037.27613333, 1143.16975251, 1227.40331322, 1251.47004485, 1203.33658159, 1239.43667904, 1227.40331322, 1192.50655235, 1144.37308909, 1182.87985970, 1156.40645491, 1126.32304037, 1156.40645491, 1137.15306960, 1203.33658159, 1227.40331322, 1227.40331322, 1203.33658159, 1227.40331322, 1227.40331322, 1155.20311832, 1164.82981098, 1164.82981098, 1156.40645491, 1215.36994740, 1215.36994740, 1179.26984996, 1190.09987919, 1190.09987919, 1198.52323526, 1227.40331322, 1215.36994740, 1191.30321577, 1227.40331322, 1383.83706883, 1395.87043464, 1335.70360556, 1431.97053209, 1419.93716627, 1335.70360556, 1371.80370301, 1347.73697138, 1299.60350811, 1299.60350811, 1299.60350811, 1347.73697138, 1335.70360556, 1371.80370301, 1347.73697138, 1311.63687393, 1395.87043464, 1419.93716627, 1444.00389791, 1444.00389791, 1444.00389791, 1480.10399535, 1456.03726372, 1419.93716627, 1407.90380046, 1419.93716627, 1395.87043464, 1407.90380046, 1480.10399535, 1407.90380046, 1359.77033719, 1383.83706883, 1335.70360556, 1299.60350811, 1335.70360556, 1311.63687393, 1383.83706883, 1275.53677648, 1275.53677648, 1299.60350811, 1323.67023975, 1407.90380046, 1347.73697138, 1407.90380046, 1371.80370301, 1287.57014230, 1359.77033719, 1395.87043464, 1335.70360556, 1299.60350811, 1227.40331322, 1227.40331322, 1299.60350811, 1335.70360556, 1335.70360556, 1275.53677648, 1287.57014230, 1335.70360556, 1311.63687393, 1335.70360556, 1287.57014230, 1287.57014230, 1299.60350811, 1299.60350811, 1287.57014230, 1287.57014230, 1335.70360556, 1323.67023975, 1359.77033719, 1323.67023975, 1371.80370301, 1383.83706883, 1419.93716627, 1407.90380046, 1371.80370301, 1335.70360556, 1347.73697138, 1311.63687393, 1287.57014230, 1323.67023975, 1335.70360556, 1371.80370301, 1347.73697138, 1335.70360556, 1311.63687393, 1311.63687393, 1323.67023975, 1287.57014230, 1311.63687393, 1335.70360556, 1323.67023975, 1311.63687393, 1275.53677648, 1335.70360556, 1275.53677648, 1263.50341067, 1263.50341067, 1251.47004485, 1215.36994740, 1179.26984996, 1179.26984996, 1186.48986945, 1170.84649388, 1132.33972327, 1095.03628924, 1119.10302088, 1187.69320603, 1198.52323526, 1215.36994740, 1215.36994740, 1202.13324501, 1150.38977200, 1157.60979149, 1105.86631848, 1101.05297215, 1132.33972327, 1101.05297215, 1227.40331322, 1179.26984996, 1146.77976225, 1137.15306960, 1123.91636720, 1137.15306960, 1191.30321577, 1196.11656210, 1202.13324501, 1200.92990842, 1203.33658159, 1176.86317679, 1143.16975251, 1202.13324501, 1155.20311832, 1180.47318654, 1155.20311832, 1203.33658159, 1179.26984996, 1186.48986945, 1192.50655235, 1227.40331322, 1251.47004485, 1263.50341067, 1263.50341067, 1239.43667904, 1299.60350811, 1275.53677648, 1323.67023975, 1299.60350811, 1263.50341067, 1251.47004485, 1239.43667904, 1299.60350811, 1287.57014230, 1287.57014230, 1275.53677648, 1347.73697138, 1335.70360556, 1419.93716627, 1407.90380046, 1395.87043464, 1444.00389791, 1371.80370301, 1371.80370301, 1383.83706883, 1371.80370301, 1395.87043464, 1347.73697138, 1383.83706883, 1407.90380046, 1419.93716627, 1395.87043464, 1359.77033719, 1347.73697138, 1371.80370301, 1359.77033719, 1359.77033719, 1359.77033719, 1359.77033719, 1311.63687393, 1299.60350811, 1311.63687393, 1287.57014230, 1311.63687393, 1263.50341067, 1215.36994740, 1227.40331322, 1169.64315730, 1146.77976225, 1181.67652312, 1158.81312807, 1203.33658159, 1173.25316705, 1087.81626976, 1089.01960634, 1052.91950889, 1052.91950889, 1051.71617231, 902.50243619, 924.16249466, 897.68908986, 910.92579226, 909.72245568, 966.27927501, 1103.45964532, 1050.51283573, 965.07593843, 960.26259211, 985.53266032, 931.38251415, 906.11244594, 844.74228027, 854.36897293, 861.58899242, 914.53580201, 926.56916782, 937.39919706, 922.95915808, 920.55248491, 902.50243619, 872.41902165, 862.79232900, 867.60567532, 904.90910935, 897.68908986, 886.85906063, 910.92579226, 854.36897293, 854.36897293, 844.74228027, 842.33560711, 851.96229976, 854.36897293, 839.92893395, 853.16563635], 
							[595.65160789, 577.60155916, 576.39822258, 535.48477881, 563.16152018, 522.24807641, 493.36799845, 518.63806666, 524.65474957, 523.45141299, 541.50146171, 549.92481779, 529.46809590, 547.51814462, 519.84140325, 499.38468136, 487.35131554, 469.30126682, 504.19802769, 510.21471059, 505.40136427, 469.30126682, 477.72462289, 489.75798871, 447.64120835, 457.26790100, 545.11147146, 529.46809590, 507.80803743, 516.23139350, 539.09478855, 483.74130580, 495.77467161, 498.18134478, 489.75798871, 490.96132529, 480.13129605, 463.28458391, 481.33463264, 487.35131554, 487.35131554, 496.97800820, 505.40136427, 495.77467161, 481.33463264, 469.30126682, 481.33463264, 464.48792049, 476.52128631, 493.36799845, 470.50460340, 484.94464238, 477.72462289, 450.04788151, 493.36799845, 482.53796922, 499.38468136, 552.33149095, 576.39822258, 586.02491523, 557.14483728, 595.65160789, 580.00823233, 563.16152018, 582.41490549, 589.63492498, 594.44827130, 583.61824207, 561.95818360, 575.19488600, 554.73816411, 580.00823233, 589.63492498, 570.38153967, 595.65160789, 589.63492498, 587.22825181, 573.99154942, 586.02491523, 586.02491523, 569.17820309, 577.60155916, 559.55151044, 559.55151044, 576.39822258, 583.61824207, 592.04159814, 577.60155916, 558.34817386, 553.53482753, 536.68811539, 504.19802769, 505.40136427, 494.57133503, 504.19802769, 482.53796922, 487.35131554, 516.23139350, 529.46809590, 530.67143248, 527.06142274, 563.16152018, 551.12815437, 540.29812513, 555.94150069, 567.97486651, 571.58487625, 563.16152018, 580.00823233, 560.75484702, 640.17506140, 614.90499319, 629.34503217, 613.70165661, 635.36171508, 635.36171508, 613.70165661, 648.59841748, 661.83511987, 672.66514911, 673.86848569, 665.44512962, 640.17506140, 670.25847594, 690.71519783, 689.51186125, 695.52854416, 712.37525630, 705.15523681, 700.34189048, 684.69851492, 718.39193921, 666.64846620, 661.83511987, 669.05513936, 670.25847594, 657.02177355, 667.85180278, 688.30852467, 675.07182227, 673.86848569, 683.49517834, 685.90185151, 657.02177355, 631.75170533, 634.15837850, 634.15837850, 643.78507115, 631.75170533, 637.76838824, 617.31166635, 601.66829079, 565.56819335, 578.80489574, 571.58487625, 596.85494447, 576.39822258, 551.12815437, 502.99469110, 552.33149095, 501.79135452, 553.53482753, 554.73816411, 559.55151044, 561.95818360, 570.38153967, 565.56819335, 547.51814462, 542.70479830, 519.84140325, 527.06142274, 495.77467161, 528.26475932, 502.99469110, 488.55465212, 541.50146171, 522.24807641, 529.46809590, 522.24807641, 529.46809590, 529.46809590, 594.44827130, 573.99154942, 571.58487625, 581.21156891, 625.73502243, 566.77152993, 571.58487625, 599.26161763, 595.65160789, 598.05828105, 558.34817386, 576.39822258, 552.33149095, 594.44827130, 596.85494447, 631.75170533, 635.36171508, 630.54836875, 625.73502243, 635.36171508, 658.22511013, 649.80175406, 670.25847594, 671.46181253, 678.68183202, 702.74856365, 687.10518809, 682.29184176, 663.03845645, 683.49517834, 676.27515885, 693.12187099, 709.96858314, 711.17191972, 714.78192946, 713.57859288, 701.54522707, 695.52854416, 681.08850518, 681.08850518, 673.86848569, 672.66514911, 694.32520758, 703.95190023, 703.95190023, 715.98526604, 699.13855390, 678.68183202, 673.86848569, 676.27515885, 675.07182227, 666.64846620, 654.61510038, 673.86848569, 637.76838824, 629.34503217, 647.39508089, 642.58173457, 673.86848569, 657.02177355, 637.76838824, 661.83511987, 654.61510038, 640.17506140, 644.98840773, 649.80175406, 649.80175406, 653.41176380, 613.70165661, 600.46495421, 622.12501268, 595.65160789, 608.88831028, 547.51814462, 582.41490549, 535.48477881, 539.09478855, 533.07810564, 583.61824207, 581.21156891, 577.60155916, 555.94150069, 528.26475932, 583.61824207, 564.36485676, 558.34817386, 541.50146171, 583.61824207, 589.63492498, 644.98840773, 619.71833952, 634.15837850, 619.71833952, 590.83826156, 623.32834926, 630.54836875, 631.75170533, 631.75170533, 631.75170533, 634.15837850, 631.75170533, 634.15837850, 578.80489574, 564.36485676, 535.48477881, 613.70165661, 588.43158840, 542.70479830, 510.21471059, 486.14797896], 
							[1275.53677648, 1263.50341067, 1239.43667904, 1199.72657184, 1179.26984996, 1080.59625027, 968.68594818, 986.73599690, 949.43256287, 932.58585073, 1004.78604563, 950.63589945, 939.80587022, 953.04257262, 914.53580201, 902.50243619, 824.28555839, 758.10204640, 824.28555839, 800.21882676, 807.43884625, 740.05199768, 740.05199768, 722.00194895, 653.41176380, 782.16877803, 950.63589945, 933.78918731, 896.48575328, 915.73913859, 920.55248491, 849.55562660, 839.92893395, 830.30224130, 806.23550966, 794.20214385, 783.37211461, 756.89870982, 842.33560711, 860.38565584, 856.77564609, 836.31892420, 830.30224130, 826.69223155, 800.21882676, 758.10204640, 768.93207563, 756.89870982, 783.37211461, 812.25219257, 783.37211461, 829.09890471, 811.04885599, 890.46907037, 1004.78604563, 1063.74953812, 1078.18957710, 1383.83706883, 1311.63687393, 1371.80370301, 1299.60350811, 1371.80370301, 1395.87043464, 1323.67023975, 1311.63687393, 1371.80370301, 1323.67023975, 1323.67023975, 1239.43667904, 1203.33658159, 1202.13324501, 1227.40331322, 1263.50341067, 1192.50655235, 1239.43667904, 1239.43667904, 1299.60350811, 1275.53677648, 1275.53677648, 1287.57014230, 1263.50341067, 1192.50655235, 1191.30321577, 1150.38977200, 1178.06651337, 1182.87985970, 1161.21980123, 1079.39291368, 973.49929450, 957.85591894, 915.73913859, 807.43884625, 902.50243619, 878.43570456, 900.09576303, 850.75896318, 888.06239721, 889.26573379, 986.73599690, 1026.44610409, 1064.95287471, 1287.57014230, 1191.30321577, 1125.11970378, 1167.23648414, 1185.28653286, 1143.16975251, 1132.33972327, 1113.08633797, 1073.37623078, 1089.01960634, 1143.16975251, 1131.13638669, 1084.20626001, 1099.84963557, 1113.08633797, 1119.10302088, 1120.30635746, 1153.99978174, 1161.21980123, 1215.36994740, 1150.38977200, 1140.76307935, 1115.49301113, 1122.71303062, 1091.42627950, 1090.22294292, 1098.64629899, 1141.96641593, 1115.49301113, 1113.08633797, 1140.76307935, 1073.37623078, 1054.12284547, 1042.08947966, 1016.81941144, 998.76936272, 1005.98938221, 998.76936272, 983.12598716, 974.70263109, 921.75582150, 926.56916782, 861.58899242, 868.80901191, 892.87574354, 876.02903140, 919.34914833, 950.63589945, 932.58585073, 939.80587022, 1095.03628924, 1081.79958685, 1078.18957710, 1064.95287471, 1034.86946017, 1004.78604563, 957.85591894, 927.77250440, 895.28241670, 842.33560711, 854.36897293, 879.63904114, 906.11244594, 950.63589945, 939.80587022, 979.51597741, 891.67240696, 865.19900216, 818.26887548, 855.57230951, 824.28555839, 819.47221206, 836.31892420, 797.81215359, 900.09576303, 934.99252389, 902.50243619, 842.33560711, 863.99566558, 902.50243619, 922.95915808, 951.83923604, 956.65258236, 938.60253364, 986.73599690, 908.51911910, 866.40233874, 945.82255313, 907.31578252, 909.72245568, 842.33560711, 866.40233874, 847.14895344, 890.46907037, 901.29909961, 894.07908012, 908.51911910, 909.72245568, 920.55248491, 931.38251415, 984.32932374, 955.44924578, 932.58585073, 962.66926527, 956.65258236, 934.99252389, 902.50243619, 894.07908012, 912.12912884, 921.75582150, 885.65572405, 926.56916782, 937.39919706, 962.66926527, 957.85591894, 983.12598716, 986.73599690, 931.38251415, 927.77250440, 984.32932374, 998.76936272, 989.14267007, 1001.17603588, 1008.39605537, 1001.17603588, 1010.80272853, 983.12598716, 962.66926527, 937.39919706, 936.19586048, 891.67240696, 884.45238747, 889.26573379, 915.73913859, 865.19900216, 870.01234849, 922.95915808, 939.80587022, 931.38251415, 949.43256287, 926.56916782, 944.61921655, 908.51911910, 884.45238747, 919.34914833, 879.63904114, 892.87574354, 849.55562660, 782.16877803, 790.59213410, 738.84866109, 709.96858314, 715.98526604, 588.43158840, 664.24179304, 605.27830054, 605.27830054, 577.60155916, 565.56819335, 596.85494447, 573.99154942, 551.12815437, 512.62138376, 570.38153967, 517.43473008, 480.13129605, 450.04788151, 481.33463264, 464.48792049, 489.75798871, 523.45141299, 577.60155916, 555.94150069, 553.53482753, 577.60155916, 565.56819335, 563.16152018, 545.11147146, 505.40136427, 481.33463264, 446.43787177, 480.13129605, 421.16780356, 381.45769636, 391.08438902, 361.00097448, 364.61098422, 373.03434029, 365.81432080, 324.90087703], 
							[1516.20409280, 1480.10399535, 1492.13736117, 1528.23745862, 1504.17072698, 1492.13736117, 1456.03726372, 1371.80370301, 1347.73697138, 1287.57014230, 1335.70360556, 1371.80370301, 1359.77033719, 1395.87043464, 1359.77033719, 1335.70360556, 1227.40331322, 1203.33658159, 1190.09987919, 1239.43667904, 1299.60350811, 1170.84649388, 1203.33658159, 1169.64315730, 1104.66298190, 1141.96641593, 1347.73697138, 1299.60350811, 1275.53677648, 1311.63687393, 1323.67023975, 1251.47004485, 1215.36994740, 1251.47004485, 1251.47004485, 1200.92990842, 1263.50341067, 1239.43667904, 1287.57014230, 1335.70360556, 1335.70360556, 1311.63687393, 1371.80370301, 1323.67023975, 1251.47004485, 1215.36994740, 1203.33658159, 1188.89654261, 1251.47004485, 1287.57014230, 1287.57014230, 1299.60350811, 1335.70360556, 1335.70360556, 1456.03726372, 1395.87043464, 1407.90380046, 1347.73697138, 1444.00389791, 1468.07062954, 1480.10399535, 1552.30419025, 1528.23745862, 1468.07062954, 1480.10399535, 1480.10399535, 1468.07062954, 1444.00389791, 1431.97053209, 1492.13736117, 1444.00389791, 1492.13736117, 1516.20409280, 1480.10399535, 1516.20409280, 1528.23745862, 1540.27082443, 1540.27082443, 1516.20409280, 1528.23745862, 1492.13736117, 1480.10399535, 1504.17072698, 1516.20409280, 1504.17072698, 1528.23745862, 1540.27082443, 1540.27082443, 1444.00389791, 1456.03726372, 1407.90380046, 1395.87043464, 1419.93716627, 1431.97053209, 1444.00389791, 1383.83706883, 1419.93716627, 1468.07062954, 1480.10399535, 1480.10399535, 1431.97053209, 1492.13736117, 1492.13736117, 1383.83706883, 1407.90380046, 1431.97053209, 1419.93716627, 1468.07062954, 1383.83706883, 1347.73697138, 1395.87043464, 1407.90380046, 1383.83706883, 1359.77033719, 1371.80370301, 1407.90380046, 1419.93716627, 1419.93716627, 1371.80370301, 1395.87043464, 1383.83706883, 1359.77033719, 1323.67023975, 1347.73697138, 1371.80370301, 1359.77033719, 1359.77033719, 1335.70360556, 1371.80370301, 1359.77033719, 1347.73697138, 1359.77033719, 1311.63687393, 1263.50341067, 1275.53677648, 1215.36994740, 1275.53677648, 1299.60350811, 1275.53677648, 1275.53677648, 1299.60350811, 1287.57014230, 1299.60350811, 1299.60350811, 1323.67023975, 1311.63687393, 1323.67023975, 1347.73697138, 1323.67023975, 1311.63687393, 1311.63687393, 1299.60350811, 1239.43667904, 1299.60350811, 1275.53677648, 1263.50341067, 1239.43667904, 1180.47318654, 1133.54305986, 1176.86317679, 1191.30321577, 1203.33658159, 1203.33658159, 1203.33658159, 1263.50341067, 1251.47004485, 1263.50341067, 1239.43667904, 1263.50341067, 1215.36994740, 1251.47004485, 1196.11656210, 1215.36994740, 1227.40331322, 1175.65984021, 1251.47004485, 1239.43667904, 1263.50341067, 1299.60350811, 1311.63687393, 1311.63687393, 1371.80370301, 1359.77033719, 1371.80370301, 1359.77033719, 1395.87043464, 1347.73697138, 1335.70360556, 1395.87043464, 1359.77033719, 1359.77033719, 1323.67023975, 1335.70360556, 1347.73697138, 1383.83706883, 1371.80370301, 1407.90380046, 1444.00389791, 1456.03726372, 1468.07062954, 1468.07062954, 1480.10399535, 1456.03726372, 1456.03726372, 1480.10399535, 1480.10399535, 1480.10399535, 1480.10399535, 1456.03726372, 1444.00389791, 1444.00389791, 1431.97053209, 1492.13736117, 1480.10399535, 1504.17072698, 1516.20409280, 1504.17072698, 1516.20409280, 1492.13736117, 1504.17072698, 1528.23745862, 1540.27082443, 1540.27082443, 1528.23745862, 1540.27082443, 1552.30419025, 1540.27082443, 1504.17072698, 1492.13736117, 1468.07062954, 1456.03726372, 1456.03726372, 1444.00389791, 1444.00389791, 1444.00389791, 1419.93716627, 1407.90380046, 1407.90380046, 1431.97053209, 1468.07062954, 1419.93716627, 1395.87043464, 1468.07062954, 1444.00389791, 1431.97053209, 1468.07062954, 1444.00389791, 1468.07062954, 1468.07062954, 1468.07062954, 1444.00389791, 1419.93716627, 1419.93716627, 1444.00389791, 1347.73697138, 1359.77033719, 1335.70360556, 1371.80370301, 1335.70360556, 1335.70360556, 1407.90380046, 1407.90380046, 1407.90380046, 1395.87043464, 1431.97053209, 1468.07062954, 1504.17072698, 1540.27082443, 1564.33755606, 1624.50438514, 1696.70458004, 1744.83804330, 1973.47199380, 1961.43862799, 1913.30516472, 2105.83901778, 2129.90574941, 2129.90574941, 2105.83901778, 2045.67218870, 2117.87238359, 2190.07257849, 2153.97248104, 2105.83901778, 2178.03921267, 2166.00584686, 2129.90574941, 2153.97248104, 2166.00584686, 2178.03921267, 2226.17267594], 
							[909.72245568, 871.21568507, 885.65572405, 860.38565584, 837.52226079, 736.44198793, 746.06868058, 711.17191972, 703.95190023, 707.56190997, 770.13541222, 796.60881701, 831.50557788, 885.65572405, 866.40233874, 843.53894369, 786.98212436, 713.57859288, 780.96544145, 785.77878778, 759.30538298, 700.34189048, 712.37525630, 691.91853441, 682.29184176, 728.01863186, 830.30224130, 823.08222181, 820.67554864, 837.52226079, 848.35229002, 830.30224130, 813.45552915, 818.26887548, 811.04885599, 791.79547068, 795.40548043, 767.72873905, 795.40548043, 802.62549992, 791.79547068, 809.84551941, 818.26887548, 820.67554864, 802.62549992, 774.94875854, 794.20214385, 777.35543171, 805.03217308, 812.25219257, 788.18546094, 794.20214385, 794.20214385, 782.16877803, 803.82883650, 803.82883650, 842.33560711, 903.70577277, 900.09576303, 931.38251415, 934.99252389, 1119.10302088, 1101.05297215, 1083.00292343, 1084.20626001, 1098.64629899, 1097.44296241, 1068.56288445, 1036.07279675, 1008.39605537, 1037.27613333, 1005.98938221, 1072.17289419, 997.56602614, 1010.80272853, 986.73599690, 1045.69948940, 1027.64944068, 1022.83609435, 1037.27613333, 1016.81941144, 978.31264083, 974.70263109, 955.44924578, 990.34600665, 995.15935297, 975.90596767, 921.75582150, 842.33560711, 842.33560711, 819.47221206, 724.40862212, 779.76210487, 746.06868058, 782.16877803, 736.44198793, 755.69537324, 788.18546094, 842.33560711, 843.53894369, 823.08222181, 902.50243619, 871.21568507, 820.67554864, 847.14895344, 845.94561686, 830.30224130, 837.52226079, 829.09890471, 777.35543171, 854.36897293, 901.29909961, 873.62235823, 857.97898267, 867.60567532, 903.70577277, 914.53580201, 925.36583124, 928.97584099, 927.77250440, 971.09262134, 945.82255313, 931.38251415, 924.16249466, 927.77250440, 918.14581175, 933.78918731, 932.58585073, 932.58585073, 916.94247517, 937.39919706, 934.99252389, 904.90910935, 915.73913859, 890.46907037, 884.45238747, 891.67240696, 879.63904114, 920.55248491, 931.38251415, 920.55248491, 936.19586048, 895.28241670, 853.16563635, 831.50557788, 833.91225104, 818.26887548, 815.86220232, 841.13227053, 843.53894369, 838.72559737, 818.26887548, 764.11872931, 799.01549017, 806.23550966, 792.99880727, 764.11872931, 728.01863186, 693.12187099, 684.69851492, 684.69851492, 687.10518809, 697.93521732, 715.98526604, 743.66200742, 758.10204640, 794.20214385, 735.23865135, 741.25533426, 691.91853441, 700.34189048, 651.00509064, 670.25847594, 671.46181253, 601.66829079, 678.68183202, 673.86848569, 676.27515885, 648.59841748, 653.41176380, 667.85180278, 708.76524656, 702.74856365, 722.00194895, 717.18860263, 749.67869033, 695.52854416, 697.93521732, 729.22196844, 703.95190023, 703.95190023, 691.91853441, 714.78192946, 681.08850518, 676.27515885, 697.93521732, 697.93521732, 732.83197819, 746.06868058, 786.98212436, 837.52226079, 856.77564609, 813.45552915, 818.26887548, 812.25219257, 844.74228027, 843.53894369, 839.92893395, 830.30224130, 809.84551941, 820.67554864, 806.23550966, 848.35229002, 876.02903140, 891.67240696, 963.87260185, 931.38251415, 944.61921655, 922.95915808, 903.70577277, 895.28241670, 904.90910935, 909.72245568, 892.87574354, 950.63589945, 925.36583124, 904.90910935, 903.70577277, 882.04571430, 898.89242645, 878.43570456, 854.36897293, 813.45552915, 812.25219257, 837.52226079, 805.03217308, 836.31892420, 854.36897293, 861.58899242, 877.23236798, 848.35229002, 847.14895344, 868.80901191, 854.36897293, 843.53894369, 866.40233874, 842.33560711, 851.96229976, 831.50557788, 795.40548043, 808.64218283, 797.81215359, 776.15209512, 756.89870982, 689.51186125, 699.13855390, 675.07182227, 697.93521732, 663.03845645, 660.63178329, 702.74856365, 683.49517834, 655.81843697, 660.63178329, 697.93521732, 673.86848569, 632.95504192, 613.70165661, 649.80175406, 631.75170533, 657.02177355, 685.90185151, 702.74856365, 697.93521732, 671.46181253, 660.63178329, 652.20842722, 646.19174431, 624.53168584, 613.70165661, 596.85494447, 584.82157865, 590.83826156, 529.46809590, 531.87476906, 513.82472034, 547.51814462, 554.73816411, 561.95818360, 560.75484702, 522.24807641], 
							[1612.47101933, 1624.50438514, 1600.43765351, 1744.83804330, 1564.33755606, 1672.63784841, 1684.67121422, 1696.70458004, 1756.87140912, 1684.67121422, 1756.87140912, 1648.57111678, 1756.87140912, 1780.93814075, 1780.93814075, 1672.63784841, 1708.73794585, 1648.57111678, 1720.77131167, 1744.83804330, 1720.77131167, 1564.33755606, 1672.63784841, 1564.33755606, 1552.30419025, 1576.37092188, 1708.73794585, 1732.80467749, 1672.63784841, 1708.73794585, 1672.63784841, 1588.40428770, 1600.43765351, 1636.53775096, 1624.50438514, 1624.50438514, 1636.53775096, 1636.53775096, 1708.73794585, 1708.73794585, 1672.63784841, 1708.73794585, 1780.93814075, 1792.97150657, 1732.80467749, 1708.73794585, 1708.73794585, 1744.83804330, 1817.03823820, 1841.10496983, 1805.00487238, 1817.03823820, 1768.90477493, 1780.93814075, 1841.10496983, 1913.30516472, 1985.50535962, 1973.47199380, 2045.67218870, 2081.77228615, 2057.70555452, 2153.97248104, 2166.00584686, 2081.77228615, 2105.83901778, 2190.07257849, 2093.80565196, 2093.80565196, 1961.43862799, 1901.27179891, 1889.23843309, 1913.30516472, 1985.50535962, 1937.37189636, 1997.53872544, 1961.43862799, 1973.47199380, 1985.50535962, 2021.60545707, 2057.70555452, 1985.50535962, 1997.53872544, 2021.60545707, 2045.67218870, 2105.83901778, 2069.73892033, 2105.83901778, 2129.90574941, 2081.77228615, 2069.73892033, 1997.53872544, 1961.43862799, 1937.37189636, 1913.30516472, 2045.67218870, 1949.40526217, 1973.47199380, 2069.73892033, 2081.77228615, 2069.73892033, 2069.73892033, 2226.17267594, 2178.03921267, 2141.93911523, 2202.10594431, 2262.27277339, 2322.43960246, 2382.60643154, 2298.37287083, 2250.23940757, 2250.23940757, 2334.47296828, 2322.43960246, 2322.43960246, 2322.43960246, 2334.47296828, 2358.53969991, 2370.57306573, 2454.80662644, 2442.77326062, 2478.87335807, 2527.00682133, 2490.90672389, 2502.94008970, 2502.94008970, 2490.90672389, 2454.80662644, 2454.80662644, 2442.77326062, 2418.70652899, 2430.73989481, 2430.73989481, 2394.63979736, 2370.57306573, 2406.67316318, 2406.67316318, 2418.70652899, 2406.67316318, 2454.80662644, 2466.83999225, 2466.83999225, 2418.70652899, 2478.87335807, 2430.73989481, 2430.73989481, 2454.80662644, 2478.87335807, 2502.94008970, 2527.00682133, 2514.97345552, 2442.77326062, 2358.53969991, 2214.13931012, 2178.03921267, 2105.83901778, 2105.83901778, 2081.77228615, 2081.77228615, 2033.63882288, 2033.63882288, 2033.63882288, 2033.63882288, 1997.53872544, 1997.53872544, 2009.57209125, 2033.63882288, 2045.67218870, 1961.43862799, 1937.37189636, 1901.27179891, 1853.13833565, 1829.07160401, 1829.07160401, 1925.33853054, 1853.13833565, 2021.60545707, 2081.77228615, 2081.77228615, 2045.67218870, 2069.73892033, 2093.80565196, 2153.97248104, 2166.00584686, 2117.87238359, 2117.87238359, 2286.33950502, 2202.10594431, 2202.10594431, 2262.27277339, 2238.20604175, 2238.20604175, 2214.13931012, 2262.27277339, 2238.20604175, 2250.23940757, 2214.13931012, 2262.27277339, 2286.33950502, 2286.33950502, 2286.33950502, 2322.43960246, 2382.60643154, 2430.73989481, 2478.87335807, 2478.87335807, 2466.83999225, 2442.77326062, 2406.67316318, 2454.80662644, 2502.94008970, 2502.94008970, 2502.94008970, 2563.10691878, 2599.20701623, 2647.34047949, 2599.20701623, 2623.27374786, 2755.64077184, 2695.47394276, 2611.24038205, 2707.50730857, 2743.60740602, 2803.77423510, 2791.74086928, 2839.87433255, 2912.07452744, 2863.94106418, 2815.80760092, 2743.60740602, 2635.30711368, 2647.34047949, 2707.50730857, 2659.37384531, 2671.40721112, 2743.60740602, 2695.47394276, 2743.60740602, 2791.74086928, 2888.00779581, 2996.30808815, 2912.07452744, 2851.90769836, 2936.14125907, 2875.97442999, 2900.04116163, 2972.24135652, 3032.40818560, 3092.57501468, 3188.84194121, 3249.00877029, 3224.94203866, 3176.80857539, 3333.24233100, 3357.30906263, 3032.40818560, 3008.34145397, 2960.20799071, 2984.27472234, 2912.07452744, 3152.74184376, 2960.20799071, 2791.74086928, 2587.17365041, 2839.87433255, 2791.74086928, 2912.07452744, 2875.97442999, 3080.54164886, 3116.64174631, 3261.04213610, 3465.60935497, 3537.80954987, 3489.67608660, 3429.50925753, 3501.70945242, 3513.74281824, 3345.27569681, 3080.54164886, 3309.17559937, 3236.97540447, 3381.37579426, 3453.57598916, 3417.47589171, 3489.67608660, 3345.27569681, 3465.60935497, 3273.07550192, 3128.67511213, 3405.44252589, 3441.54262334, 3682.20993966], 
							[765.32206589, 709.96858314, 691.91853441, 756.89870982, 734.03531477, 741.25533426, 742.45867084, 702.74856365, 714.78192946, 719.59527579, 755.69537324, 703.95190023, 741.25533426, 760.50871956, 725.61195870, 700.34189048, 684.69851492, 665.44512962, 699.13855390, 703.95190023, 695.52854416, 691.91853441, 697.93521732, 673.86848569, 659.42844671, 712.37525630, 744.86534400, 720.79861237, 725.61195870, 734.03531477, 750.88202691, 712.37525630, 690.71519783, 691.91853441, 693.12187099, 677.47849543, 681.08850518, 652.20842722, 649.80175406, 669.05513936, 688.30852467, 684.69851492, 715.98526604, 697.93521732, 673.86848569, 689.51186125, 673.86848569, 683.49517834, 738.84866109, 768.93207563, 766.52540247, 756.89870982, 750.88202691, 738.84866109, 830.30224130, 884.45238747, 922.95915808, 914.53580201, 1007.19271879, 974.70263109, 902.50243619, 1012.00606512, 990.34600665, 939.80587022, 939.80587022, 920.55248491, 913.33246543, 886.85906063, 820.67554864, 842.33560711, 823.08222181, 836.31892420, 862.79232900, 855.57230951, 876.02903140, 818.26887548, 842.33560711, 876.02903140, 884.45238747, 890.46907037, 896.48575328, 892.87574354, 854.36897293, 866.40233874, 918.14581175, 926.56916782, 931.38251415, 908.51911910, 901.29909961, 912.12912884, 895.28241670, 824.28555839, 832.70891446, 870.01234849, 890.46907037, 890.46907037, 909.72245568, 914.53580201, 891.67240696, 924.16249466, 908.51911910, 1010.80272853, 1008.39605537, 993.95601639, 1010.80272853, 1050.51283573, 1083.00292343, 1069.76622103, 1033.66612358, 997.56602614, 1028.85277726, 1078.18957710, 1054.12284547, 1034.86946017, 1033.66612358, 1052.91950889, 1031.25945042, 1080.59625027, 1058.93619180, 1026.44610409, 1057.73285522, 1051.71617231, 1039.68280649, 999.97269930, 972.29595792, 971.09262134, 926.56916782, 941.00920680, 950.63589945, 947.02588971, 947.02588971, 930.17917757, 909.72245568, 950.63589945, 960.26259211, 973.49929450, 983.12598716, 981.92265058, 1025.24276751, 991.54934323, 991.54934323, 949.43256287, 967.48261160, 961.46592869, 987.93933348, 969.88928476, 981.92265058, 977.10930425, 990.34600665, 973.49929450, 968.68594818, 945.82255313, 884.45238747, 886.85906063, 882.04571430, 838.72559737, 830.30224130, 823.08222181, 814.65886573, 806.23550966, 762.91539273, 796.60881701, 730.42530502, 756.89870982, 754.49203666, 738.84866109, 738.84866109, 670.25847594, 693.12187099, 669.05513936, 657.02177355, 587.22825181, 619.71833952, 653.41176380, 652.20842722, 719.59527579, 761.71205615, 774.94875854, 756.89870982, 779.76210487, 780.96544145, 930.17917757, 901.29909961, 894.07908012, 919.34914833, 913.33246543, 836.31892420, 859.18231925, 894.07908012, 867.60567532, 904.90910935, 818.26887548, 885.65572405, 882.04571430, 857.97898267, 848.35229002, 898.89242645, 926.56916782, 888.06239721, 926.56916782, 954.24590920, 986.73599690, 984.32932374, 1009.59939195, 1021.63275777, 1013.20940170, 1024.03943093, 1020.42942119, 969.88928476, 978.31264083, 971.09262134, 954.24590920, 962.66926527, 996.36268955, 1027.64944068, 1034.86946017, 1044.49615282, 1048.10616256, 1026.44610409, 1034.86946017, 1066.15621129, 1064.95287471, 1069.76622103, 1080.59625027, 1101.05297215, 1164.82981098, 1170.84649388, 1153.99978174, 1114.28967455, 1050.51283573, 1050.51283573, 1079.39291368, 1009.59939195, 1019.22608460, 1036.07279675, 978.31264083, 1007.19271879, 1045.69948940, 1072.17289419, 1090.22294292, 1078.18957710, 1033.66612358, 1064.95287471, 1038.47946991, 1025.24276751, 1003.58270904, 968.68594818, 979.51597741, 973.49929450, 919.34914833, 959.05925553, 850.75896318, 807.43884625, 808.64218283, 749.67869033, 843.53894369, 847.14895344, 902.50243619, 892.87574354, 954.24590920, 920.55248491, 856.77564609, 856.77564609, 845.94561686, 850.75896318, 897.68908986, 896.48575328, 879.63904114, 891.67240696, 928.97584099, 944.61921655, 932.58585073, 897.68908986, 888.06239721, 928.97584099, 981.92265058, 962.66926527, 965.07593843, 1034.86946017, 1007.19271879, 1043.29281624, 980.71931399, 1004.78604563, 944.61921655, 943.41587996, 944.61921655, 885.65572405, 936.19586048, 912.12912884, 823.08222181, 841.13227053], 
							[1517.26302899, 1328.91678724, 1192.89162006, 1234.74366637, 1224.27463811, 1151.02754039, 1171.95356354, 1192.89162006, 1171.95356354, 1182.42259180, 1339.38581550, 1370.76883355, 1318.45979235, 1339.38581550, 1318.45979235, 1328.91678724, 1234.74366637, 1098.70646582, 1234.74366637, 1234.74366637, 1151.02754039, 1088.24947092, 1098.70646582, 1077.78044266, 1046.39742462, 1161.49656865, 1224.27463811, 1318.45979235, 1266.13871778, 1328.91678724, 1423.08990812, 1349.84281040, 1391.70689007, 1433.55893638, 1423.08990812, 1402.16388496, 1412.63291322, 1402.16388496, 1464.95398779, 1485.88001094, 1464.95398779, 1475.41098268, 1548.65808041, 1538.20108551, 1538.20108551, 1506.80603410, 1464.95398779, 1485.88001094, 1538.20108551, 1559.12710867, 1517.26302899, 1569.58410356, 838.16002917, 876.87136700, 894.66871504, 936.52076135, 958.49368733, 1067.32344777, 1020.23688733, 1040.11600766, 1046.39742462, 1213.81764321, 1287.06474093, 1255.66968952, 1203.34861495, 1234.74366637, 1224.27463811, 1203.34861495, 1098.70646582, 1098.70646582, 1151.02754039, 1130.10151723, 1151.02754039, 1077.78044266, 1192.89162006, 1203.34861495, 1287.06474093, 1266.13871778, 1297.52173583, 1318.45979235, 1297.52173583, 1266.13871778, 1318.45979235, 1307.99076409, 1318.45979235, 1297.52173583, 1297.52173583, 1276.59571267, 1171.95356354, 1151.02754039, 1109.17549408, 1077.78044266, 1151.02754039, 1130.10151723, 1203.34861495, 1140.57054549, 1203.34861495, 1203.34861495, 1255.66968952, 1297.52173583, 1287.06474093, 1402.16388496, 1297.52173583, 1339.38581550, 1391.70689007, 1402.16388496, 1475.41098268, 1559.12710867, 1475.41098268, 1423.08990812, 1485.88001094, 1496.33700584, 1496.33700584, 1444.01593127, 1444.01593127, 1475.41098268, 1506.80603410, 1548.65808041, 1559.12710867, 1590.51012671, 1590.51012671, 1485.88001094, 1485.88001094, 1444.01593127, 1485.88001094, 1475.41098268, 1475.41098268, 1538.20108551, 1527.73205725, 1527.73205725, 1485.88001094, 1475.41098268, 1433.55893638, 1454.48495953, 1433.55893638, 1381.23786181, 1454.48495953, 1423.08990812, 1391.70689007, 1360.31183866, 1370.76883355, 1349.84281040, 1360.31183866, 1307.99076409, 1276.59571267, 1276.59571267, 1297.52173583, 1328.91678724, 1318.45979235, 1307.99076409, 1307.99076409, 1276.59571267, 1171.95356354, 1266.13871778, 1297.52173583, 1266.13871778, 1213.81764321, 1140.57054549, 1098.70646582, 1119.64452234, 1088.24947092, 1015.00237320, 1024.41246527, 1088.24947092, 1140.57054549, 1119.64452234, 1140.57054549, 1119.64452234, 1056.85441951, 1056.85441951, 1056.85441951, 1013.95547038, 1004.53334494, 1077.78044266, 977.10930425, 1101.05297215, 1092.62961608, 1036.07279675, 1049.30949914, 1004.78604563, 1004.78604563, 1022.83609435, 1013.20940170, 1021.63275777, 977.10930425, 1016.81941144, 950.63589945, 936.19586048, 962.66926527, 965.07593843, 984.32932374, 963.87260185, 1022.83609435, 969.88928476, 991.54934323, 1026.44610409, 1022.83609435, 1046.90282598, 1058.93619180, 1024.03943093, 1040.88614307, 1092.62961608, 1080.59625027, 1102.25630873, 1141.96641593, 1197.31989868, 1203.33658159, 1202.13324501, 1191.30321577, 1156.40645491, 1104.66298190, 1125.11970378, 1157.60979149, 1251.47004485, 1239.43667904, 1323.67023975, 1299.60350811, 1371.80370301, 1287.57014230, 1299.60350811, 1323.67023975, 1347.73697138, 1371.80370301, 1323.67023975, 1383.83706883, 1359.77033719, 1371.80370301, 1335.70360556, 1287.57014230, 1287.57014230, 1251.47004485, 1275.53677648, 1275.53677648, 1287.57014230, 1299.60350811, 1263.50341067, 1263.50341067, 1383.83706883, 1359.77033719, 1395.87043464, 1335.70360556, 1347.73697138, 1371.80370301, 1311.63687393, 1299.60350811, 1347.73697138, 1359.77033719, 1359.77033719, 1311.63687393, 1239.43667904, 1215.36994740, 1251.47004485, 1227.40331322, 1161.21980123, 1101.05297215, 1150.38977200, 1046.90282598, 1028.85277726, 962.66926527, 998.76936272, 1004.78604563, 1084.20626001, 1024.03943093, 1021.63275777, 1107.06965506, 1057.73285522, 993.95601639, 983.12598716, 1074.57956736, 1048.10616256, 1096.23962583, 1156.40645491, 1251.47004485, 1275.53677648, 1251.47004485, 1287.57014230, 1311.63687393, 1275.53677648, 1275.53677648, 1287.57014230, 1227.40331322, 1215.36994740, 1203.33658159, 1072.17289419, 1026.44610409, 1002.37937246, 999.97269930, 1004.78604563, 1001.17603588, 1001.17603588, 1054.12284547], 
							[725.61195870, 746.06868058, 717.18860263, 632.95504192, 583.61824207, 583.61824207, 529.46809590, 551.12815437, 529.46809590, 565.56819335, 607.68497370, 593.24493472, 655.81843697, 659.42844671, 628.14169559, 630.54836875, 583.61824207, 571.58487625, 599.26161763, 614.90499319, 628.14169559, 596.85494447, 584.82157865, 555.94150069, 530.67143248, 577.60155916, 625.73502243, 649.80175406, 625.73502243, 649.80175406, 669.05513936, 629.34503217, 625.73502243, 614.90499319, 611.29498345, 583.61824207, 598.05828105, 563.16152018, 560.75484702, 590.83826156, 599.26161763, 592.04159814, 604.07496396, 583.61824207, 565.56819335, 573.99154942, 575.19488600, 586.02491523, 606.48163712, 622.12501268, 614.90499319, 618.51500294, 606.48163712, 602.87162738, 647.39508089, 666.64846620, 673.86848569, 672.66514911, 770.13541222, 753.28870007, 746.06868058, 765.32206589, 752.08536349, 742.45867084, 730.42530502, 790.59213410, 815.86220232, 784.57545120, 743.66200742, 734.03531477, 724.40862212, 741.25533426, 736.44198793, 725.61195870, 734.03531477, 744.86534400, 774.94875854, 782.16877803, 782.16877803, 790.59213410, 768.93207563, 766.52540247, 731.62864161, 726.81529528, 703.95190023, 681.08850518, 642.58173457, 607.68497370, 553.53482753, 581.21156891, 566.77152993, 524.65474957, 531.87476906, 517.43473008, 565.56819335, 522.24807641, 521.04473983, 536.68811539, 583.61824207, 589.63492498, 557.14483728, 607.68497370, 590.83826156, 557.14483728, 583.61824207, 596.85494447, 616.10832977, 625.73502243, 599.26161763, 570.38153967, 577.60155916, 593.24493472, 577.60155916, 575.19488600, 578.80489574, 586.02491523, 587.22825181, 593.24493472, 616.10832977, 610.09164687, 620.92167610, 594.44827130, 599.26161763, 598.05828105, 598.05828105, 571.58487625, 567.97486651, 577.60155916, 578.80489574, 583.61824207, 595.65160789, 573.99154942, 566.77152993, 559.55151044, 555.94150069, 584.82157865, 573.99154942, 583.61824207, 598.05828105, 595.65160789, 606.48163712, 554.73816411, 557.14483728, 542.70479830, 542.70479830, 536.68811539, 534.28144222, 577.60155916, 564.36485676, 551.12815437, 558.34817386, 567.97486651, 559.55151044, 564.36485676, 549.92481779, 533.07810564, 499.38468136, 499.38468136, 487.35131554, 480.13129605, 472.91127656, 454.86122784, 457.26790100, 474.11461315, 504.19802769, 482.53796922, 502.99469110, 466.89459366, 456.06456442, 434.40450595, 434.40450595, 433.20116937, 380.25435978, 407.93110116, 381.45769636, 457.26790100, 469.30126682, 453.65789126, 454.86122784, 457.26790100, 466.89459366, 478.92795947, 484.94464238, 477.72462289, 487.35131554, 472.91127656, 486.14797896, 470.50460340, 480.13129605, 457.26790100, 469.30126682, 445.23453519, 460.87791075, 452.45455468, 471.70793998, 462.08124733, 472.91127656, 481.33463264, 482.53796922, 495.77467161, 513.82472034, 525.85808615, 536.68811539, 515.02805692, 559.55151044, 563.16152018, 608.88831028, 583.61824207, 547.51814462, 567.97486651, 571.58487625, 549.92481779, 577.60155916, 595.65160789, 614.90499319, 619.71833952, 623.32834926, 625.73502243, 601.66829079, 586.02491523, 605.27830054, 601.66829079, 583.61824207, 575.19488600, 593.24493472, 623.32834926, 601.66829079, 604.07496396, 593.24493472, 605.27830054, 595.65160789, 586.02491523, 573.99154942, 571.58487625, 599.26161763, 553.53482753, 554.73816411, 559.55151044, 584.82157865, 586.02491523, 567.97486651, 576.39822258, 598.05828105, 565.56819335, 573.99154942, 587.22825181, 573.99154942, 586.02491523, 559.55151044, 539.09478855, 557.14483728, 499.38468136, 484.94464238, 475.31794973, 379.05102320, 440.42118886, 421.16780356, 431.99783279, 417.55779381, 423.57447672, 439.21785228, 436.81117912, 440.42118886, 430.79449621, 447.64120835, 416.35445723, 394.69439876, 385.06770611, 391.08438902, 387.47437927, 421.16780356, 439.21785228, 445.23453519, 429.59115963, 430.79449621, 441.62452544, 460.87791075, 440.42118886, 429.59115963, 410.33777432, 394.69439876, 389.88105243, 391.08438902, 342.95092575, 330.91755994, 329.71422336, 300.83414540, 290.00411616, 304.44415514, 279.17408693, 275.56407718], 
							[1624.50438514, 1564.33755606, 1480.10399535, 1588.40428770, 1540.27082443, 1504.17072698, 1516.20409280, 1492.13736117, 1492.13736117, 1431.97053209, 1564.33755606, 1480.10399535, 1516.20409280, 1492.13736117, 1444.00389791, 1419.93716627, 1395.87043464, 1299.60350811, 1335.70360556, 1395.87043464, 1311.63687393, 1239.43667904, 1347.73697138, 1323.67023975, 1275.53677648, 1335.70360556, 1431.97053209, 1660.60448259, 1612.47101933, 1576.37092188, 1636.53775096, 1528.23745862, 1456.03726372, 1504.17072698, 1480.10399535, 1504.17072698, 1492.13736117, 1480.10399535, 1576.37092188, 1624.50438514, 1528.23745862, 1576.37092188, 1600.43765351, 1576.37092188, 1504.17072698, 1456.03726372, 1468.07062954, 1456.03726372, 1528.23745862, 1576.37092188, 1528.23745862, 1552.30419025, 1552.30419025, 1600.43765351, 1648.57111678, 1696.70458004, 1672.63784841, 1636.53775096, 1696.70458004, 1756.87140912, 1708.73794585, 1805.00487238, 1756.87140912, 1708.73794585, 1768.90477493, 1756.87140912, 1708.73794585, 1696.70458004, 1684.67121422, 1744.83804330, 1720.77131167, 1756.87140912, 1756.87140912, 1744.83804330, 1792.97150657, 1829.07160401, 1817.03823820, 1877.20506728, 1913.30516472, 1937.37189636, 1913.30516472, 1841.10496983, 1925.33853054, 1925.33853054, 1877.20506728, 1889.23843309, 1973.47199380, 1913.30516472, 1829.07160401, 1889.23843309, 1901.27179891, 1829.07160401, 1901.27179891, 1961.43862799, 1985.50535962, 1937.37189636, 1865.17170146, 1889.23843309, 1925.33853054, 1913.30516472, 1889.23843309, 1949.40526217, 1961.43862799, 1829.07160401, 1937.37189636, 1925.33853054, 1889.23843309, 1949.40526217, 1817.03823820, 1829.07160401, 1913.30516472, 1925.33853054, 1877.20506728, 1841.10496983, 1853.13833565, 1889.23843309, 1901.27179891, 1913.30516472, 1877.20506728, 1913.30516472, 1889.23843309, 1853.13833565, 1817.03823820, 1805.00487238, 1853.13833565, 1805.00487238, 1853.13833565, 1841.10496983, 1865.17170146, 1841.10496983, 1889.23843309, 1865.17170146, 1829.07160401, 1865.17170146, 1841.10496983, 1780.93814075, 1853.13833565, 1853.13833565, 1853.13833565, 1853.13833565, 1877.20506728, 1853.13833565, 1841.10496983, 1841.10496983, 1829.07160401, 1829.07160401, 1841.10496983, 1901.27179891, 1877.20506728, 1853.13833565, 1889.23843309, 1877.20506728, 1780.93814075, 1853.13833565, 1865.17170146, 1865.17170146, 1829.07160401, 1708.73794585, 1576.37092188, 1612.47101933, 1636.53775096, 1588.40428770, 1564.33755606, 1660.60448259, 1865.17170146, 1865.17170146, 1949.40526217, 1901.27179891, 1805.00487238, 1720.77131167, 1732.80467749, 1684.67121422, 1672.63784841, 1720.77131167, 1588.40428770, 1732.80467749, 1732.80467749, 1696.70458004, 1648.57111678, 1636.53775096, 1636.53775096, 1720.77131167, 1720.77131167, 1768.90477493, 1780.93814075, 1829.07160401, 1744.83804330, 1720.77131167, 1841.10496983, 1732.80467749, 1756.87140912, 1708.73794585, 1780.93814075, 1732.80467749, 1792.97150657, 1805.00487238, 1841.10496983, 1901.27179891, 1961.43862799, 1949.40526217, 1817.03823820, 1877.20506728, 1853.13833565, 1841.10496983, 1925.33853054, 1877.20506728, 1877.20506728, 1925.33853054, 1877.20506728, 1865.17170146, 1817.03823820, 1817.03823820, 1913.30516472, 1889.23843309, 1925.33853054, 1889.23843309, 1925.33853054, 1913.30516472, 1817.03823820, 1768.90477493, 1889.23843309, 1913.30516472, 1925.33853054, 1817.03823820, 1925.33853054, 1949.40526217, 1937.37189636, 1877.20506728, 1829.07160401, 1865.17170146, 1829.07160401, 1792.97150657, 1780.93814075, 1817.03823820, 1829.07160401, 1744.83804330, 1756.87140912, 1865.17170146, 1901.27179891, 1925.33853054, 1853.13833565, 1877.20506728, 1913.30516472, 1829.07160401, 1829.07160401, 1889.23843309, 1817.03823820, 1877.20506728, 1841.10496983, 1780.93814075, 1756.87140912, 1708.73794585, 1660.60448259, 1624.50438514, 1492.13736117, 1480.10399535, 1444.00389791, 1468.07062954, 1431.97053209, 1456.03726372, 1516.20409280, 1468.07062954, 1492.13736117, 1407.90380046, 1492.13736117, 1419.93716627, 1371.80370301, 1383.83706883, 1444.00389791, 1407.90380046, 1468.07062954, 1588.40428770, 1708.73794585, 1696.70458004, 1732.80467749, 1696.70458004, 1732.80467749, 1768.90477493, 1672.63784841, 1528.23745862, 1468.07062954, 1552.30419025, 1564.33755606, 1492.13736117, 1395.87043464, 1359.77033719, 1335.70360556, 1335.70360556, 1395.87043464, 1335.70360556, 1299.60350811], 
							[493.36799845, 513.82472034, 474.11461315, 483.74130580, 480.13129605, 423.57447672, 401.91441825, 380.25435978, 392.28772560, 400.71108167, 431.99783279, 445.23453519, 439.21785228, 451.25121810, 424.77781330, 405.52442800, 385.06770611, 361.00097448, 392.28772560, 436.81117912, 434.40450595, 394.69439876, 392.28772560, 379.05102320, 348.96760866, 382.66103294, 535.48477881, 505.40136427, 487.35131554, 490.96132529, 482.53796922, 451.25121810, 445.23453519, 451.25121810, 438.01451570, 422.37114014, 434.40450595, 403.11775483, 422.37114014, 424.77781330, 410.33777432, 405.52442800, 409.13443774, 394.69439876, 380.25435978, 393.49106218, 373.03434029, 375.44101346, 389.88105243, 406.72776458, 391.08438902, 394.69439876, 376.64435004, 373.03434029, 412.74444748, 403.11775483, 409.13443774, 424.77781330, 453.65789126, 470.50460340, 450.04788151, 486.14797896, 481.33463264, 482.53796922, 509.01137401, 522.24807641, 517.43473008, 518.63806666, 475.31794973, 464.48792049, 466.89459366, 475.31794973, 475.31794973, 458.47123758, 474.11461315, 464.48792049, 487.35131554, 483.74130580, 481.33463264, 476.52128631, 478.92795947, 470.50460340, 475.31794973, 447.64120835, 436.81117912, 427.18448646, 423.57447672, 394.69439876, 359.79763789, 361.00097448, 346.56093550, 332.12089652, 335.73090626, 334.52756968, 339.34091601, 327.30755019, 317.68085754, 323.69754045, 362.20431106, 361.00097448, 353.78095499, 377.84768662, 380.25435978, 352.57761841, 367.01765738, 382.66103294, 393.49106218, 445.23453519, 427.18448646, 447.64120835, 468.09793024, 498.18134478, 474.11461315, 476.52128631, 463.28458391, 465.69125707, 466.89459366, 476.52128631, 488.55465212, 519.84140325, 511.41804717, 483.74130580, 474.11461315, 465.69125707, 462.08124733, 470.50460340, 466.89459366, 502.99469110, 507.80803743, 504.19802769, 493.36799845, 494.57133503, 482.53796922, 478.92795947, 484.94464238, 484.94464238, 499.38468136, 529.46809590, 516.23139350, 512.62138376, 516.23139350, 511.41804717, 510.21471059, 481.33463264, 504.19802769, 481.33463264, 493.36799845, 507.80803743, 495.77467161, 495.77467161, 495.77467161, 504.19802769, 469.30126682, 484.94464238, 445.23453519, 442.82786202, 404.32109141, 395.89773534, 382.66103294, 387.47437927, 385.06770611, 400.71108167, 411.54111090, 415.15112065, 444.03119861, 468.09793024, 465.69125707, 410.33777432, 394.69439876, 385.06770611, 405.52442800, 387.47437927, 356.18762815, 362.20431106, 339.34091601, 371.83100371, 385.06770611, 381.45769636, 364.61098422, 373.03434029, 369.42433055, 403.11775483, 401.91441825, 397.10107192, 388.67771585, 407.93110116, 391.08438902, 409.13443774, 407.93110116, 404.32109141, 405.52442800, 403.11775483, 418.76113039, 411.54111090, 431.99783279, 444.03119861, 452.45455468, 457.26790100, 450.04788151, 450.04788151, 451.25121810, 468.09793024, 487.35131554, 468.09793024, 481.33463264, 515.02805692, 496.97800820, 482.53796922, 490.96132529, 499.38468136, 490.96132529, 487.35131554, 484.94464238, 501.79135452, 499.38468136, 510.21471059, 492.16466187, 489.75798871, 483.74130580, 481.33463264, 483.74130580, 477.72462289, 477.72462289, 468.09793024, 486.14797896, 487.35131554, 476.52128631, 464.48792049, 451.25121810, 456.06456442, 472.91127656, 456.06456442, 446.43787177, 470.50460340, 469.30126682, 440.42118886, 425.98114988, 439.21785228, 428.38782305, 438.01451570, 429.59115963, 425.98114988, 433.20116937, 409.13443774, 409.13443774, 419.96446697, 415.15112065, 413.94778407, 386.27104269, 357.39096473, 367.01765738, 353.78095499, 341.74758917, 342.95092575, 300.83414540, 334.52756968, 306.85082830, 333.32423310, 316.47752096, 323.69754045, 332.12089652, 323.69754045, 309.25750147, 311.66417463, 322.49420387, 302.03748198, 310.46083805, 294.81746249, 326.10421361, 330.91755994, 339.34091601, 332.12089652, 336.93424284, 362.20431106, 356.18762815, 377.84768662, 391.08438902, 385.06770611, 375.44101346, 367.01765738, 351.37428182, 354.98429157, 351.37428182, 316.47752096, 294.81746249, 304.44415514, 288.80077958, 282.78409667, 294.81746249, 285.19076984, 287.59744300], 
							[1128.72971353, 1083.00292343, 1076.98624052, 992.75267981, 987.93933348, 926.56916782, 906.11244594, 908.51911910, 872.41902165, 903.70577277, 974.70263109, 998.76936272, 1024.03943093, 1036.07279675, 986.73599690, 944.61921655, 897.68908986, 885.65572405, 1005.98938221, 981.92265058, 991.54934323, 914.53580201, 891.67240696, 872.41902165, 829.09890471, 969.88928476, 1020.42942119, 1056.52951863, 1056.52951863, 1070.96955761, 1110.67966481, 1076.98624052, 1066.15621129, 1064.95287471, 1076.98624052, 1040.88614307, 1033.66612358, 1062.54620154, 1038.47946991, 1080.59625027, 1058.93619180, 1073.37623078, 1062.54620154, 1070.96955761, 1055.32618205, 1014.41273828, 1028.85277726, 1022.83609435, 1048.10616256, 1070.96955761, 1066.15621129, 1075.78290394, 1040.88614307, 1046.90282598, 1050.51283573, 1070.96955761, 1073.37623078, 1227.40331322, 1203.33658159, 1227.40331322, 1203.33658159, 1287.57014230, 1299.60350811, 1239.43667904, 1227.40331322, 1275.53677648, 1227.40331322, 1227.40331322, 1203.33658159, 1202.13324501, 1196.11656210, 1169.64315730, 1215.36994740, 1181.67652312, 1203.33658159, 1227.40331322, 1287.57014230, 1275.53677648, 1287.57014230, 1275.53677648, 1263.50341067, 1227.40331322, 1251.47004485, 1200.92990842, 1215.36994740, 1193.70988894, 1187.69320603, 1125.11970378, 1083.00292343, 1042.08947966, 1045.69948940, 1024.03943093, 995.15935297, 971.09262134, 1028.85277726, 1010.80272853, 1033.66612358, 1034.86946017, 1090.22294292, 1102.25630873, 1055.32618205, 1157.60979149, 1141.96641593, 1107.06965506, 1131.13638669, 1139.55974276, 1185.28653286, 1199.72657184, 1163.62647440, 1128.72971353, 1147.98309883, 1170.84649388, 1141.96641593, 1134.74639644, 1141.96641593, 1155.20311832, 1169.64315730, 1167.23648414, 1160.01646465, 1155.20311832, 1173.25316705, 1239.43667904, 1299.60350811, 1335.70360556, 1431.97053209, 1383.83706883, 1395.87043464, 1407.90380046, 1407.90380046, 1395.87043464, 1371.80370301, 1371.80370301, 1323.67023975, 1275.53677648, 1287.57014230, 1287.57014230, 1299.60350811, 1299.60350811, 1311.63687393, 1227.40331322, 1185.28653286, 1147.98309883, 1126.32304037, 1073.37623078, 1127.52637695, 1116.69634771, 1109.47632822, 1153.99978174, 1153.99978174, 1109.47632822, 1134.74639644, 1095.03628924, 1083.00292343, 1133.54305986, 1075.78290394, 1036.07279675, 1013.20940170, 1022.83609435, 1038.47946991, 1015.61607486, 928.97584099, 972.29595792, 950.63589945, 990.34600665, 1051.71617231, 1090.22294292, 1120.30635746, 1038.47946991, 1010.80272853, 975.90596767, 1003.58270904, 957.85591894, 1010.80272853, 1020.42942119, 979.51597741, 1022.83609435, 1073.37623078, 1037.27613333, 1058.93619180, 1022.83609435, 1043.29281624, 1010.80272853, 1021.63275777, 1013.20940170, 996.36268955, 987.93933348, 934.99252389, 1020.42942119, 1010.80272853, 997.56602614, 995.15935297, 957.85591894, 980.71931399, 1020.42942119, 1016.81941144, 1034.86946017, 1044.49615282, 1014.41273828, 1067.35954787, 1070.96955761, 1103.45964532, 1113.08633797, 1107.06965506, 1075.78290394, 1086.61293317, 1097.44296241, 1110.67966481, 1098.64629899, 1093.83295266, 1083.00292343, 1067.35954787, 1040.88614307, 1049.30949914, 1084.20626001, 1105.86631848, 1169.64315730, 1139.55974276, 1141.96641593, 1128.72971353, 1119.10302088, 1132.33972327, 1117.89968430, 1132.33972327, 1125.11970378, 1123.91636720, 1132.33972327, 1125.11970378, 1111.88300139, 1083.00292343, 1073.37623078, 1049.30949914, 1040.88614307, 1030.05611384, 1027.64944068, 1025.24276751, 1004.78604563, 1014.41273828, 1032.46278700, 1008.39605537, 1010.80272853, 1027.64944068, 1020.42942119, 1044.49615282, 1028.85277726, 1012.00606512, 998.76936272, 971.09262134, 963.87260185, 908.51911910, 845.94561686, 860.38565584, 806.23550966, 823.08222181, 833.91225104, 708.76524656, 773.74542196, 685.90185151, 722.00194895, 657.02177355, 629.34503217, 611.29498345, 604.07496396, 594.44827130, 589.63492498, 628.14169559, 642.58173457, 599.26161763, 586.02491523, 606.48163712, 595.65160789, 623.32834926, 623.32834926, 695.52854416, 742.45867084, 697.93521732, 690.71519783, 663.03845645, 638.97172482, 643.78507115, 612.49832003, 594.44827130, 608.88831028, 617.31166635, 563.16152018, 534.28144222, 499.38468136, 484.94464238, 505.40136427, 470.50460340, 458.47123758, 413.94778407], 
							[1251.47004485, 1215.36994740, 1191.30321577, 1163.62647440, 1187.69320603, 1143.16975251, 1090.22294292, 969.88928476, 956.65258236, 890.46907037, 987.93933348, 1022.83609435, 962.66926527, 957.85591894, 901.29909961, 857.97898267, 728.01863186, 724.40862212, 794.20214385, 789.38879752, 783.37211461, 706.35857339, 764.11872931, 782.16877803, 740.05199768, 770.13541222, 821.87888522, 878.43570456, 896.48575328, 919.34914833, 992.75267981, 854.36897293, 800.21882676, 835.11558762, 835.11558762, 794.20214385, 805.03217308, 783.37211461, 831.50557788, 897.68908986, 880.84237772, 913.33246543, 920.55248491, 884.45238747, 806.23550966, 800.21882676, 823.08222181, 806.23550966, 876.02903140, 890.46907037, 849.55562660, 859.18231925, 900.09576303, 918.14581175, 1022.83609435, 1051.71617231, 1076.98624052, 1070.96955761, 1045.69948940, 1024.03943093, 1027.64944068, 1093.83295266, 1092.62961608, 1043.29281624, 1044.49615282, 1034.86946017, 1058.93619180, 1058.93619180, 1022.83609435, 1043.29281624, 1013.20940170, 1052.91950889, 1063.74953812, 1034.86946017, 1064.95287471, 1052.91950889, 1105.86631848, 1093.83295266, 1125.11970378, 1074.57956736, 1073.37623078, 1058.93619180, 1058.93619180, 1055.32618205, 1095.03628924, 1042.08947966, 974.70263109, 902.50243619, 799.01549017, 838.72559737, 820.67554864, 746.06868058, 800.21882676, 754.49203666, 835.11558762, 748.47535375, 734.03531477, 817.06553890, 865.19900216, 866.40233874, 836.31892420, 962.66926527, 932.58585073, 849.55562660, 908.51911910, 983.12598716, 968.68594818, 963.87260185, 992.75267981, 914.53580201, 963.87260185, 968.68594818, 944.61921655, 922.95915808, 932.58585073, 957.85591894, 950.63589945, 965.07593843, 950.63589945, 996.36268955, 1001.17603588, 944.61921655, 953.04257262, 928.97584099, 951.83923604, 926.56916782, 908.51911910, 939.80587022, 962.66926527, 955.44924578, 945.82255313, 927.77250440, 920.55248491, 928.97584099, 941.00920680, 902.50243619, 922.95915808, 937.39919706, 930.17917757, 947.02588971, 951.83923604, 914.53580201, 926.56916782, 857.97898267, 855.57230951, 872.41902165, 896.48575328, 948.22922629, 922.95915808, 950.63589945, 904.90910935, 926.56916782, 902.50243619, 906.11244594, 897.68908986, 892.87574354, 880.84237772, 802.62549992, 724.40862212, 778.55876829, 770.13541222, 708.76524656, 723.20528553, 722.00194895, 782.16877803, 770.13541222, 792.99880727, 752.08536349, 738.84866109, 724.40862212, 722.00194895, 643.78507115, 636.56505166, 671.46181253, 611.29498345, 709.96858314, 722.00194895, 734.03531477, 728.01863186, 697.93521732, 703.95190023, 778.55876829, 792.99880727, 785.77878778, 838.72559737, 800.21882676, 777.35543171, 737.64532451, 788.18546094, 752.08536349, 761.71205615, 722.00194895, 734.03531477, 705.15523681, 723.20528553, 755.69537324, 729.22196844, 771.33874880, 783.37211461, 786.98212436, 768.93207563, 778.55876829, 776.15209512, 780.96544145, 830.30224130, 855.57230951, 830.30224130, 823.08222181, 835.11558762, 807.43884625, 799.01549017, 830.30224130, 826.69223155, 860.38565584, 902.50243619, 866.40233874, 919.34914833, 962.66926527, 928.97584099, 973.49929450, 1007.19271879, 1002.37937246, 999.97269930, 998.76936272, 1016.81941144, 1046.90282598, 1062.54620154, 1046.90282598, 1010.80272853, 1024.03943093, 1016.81941144, 1004.78604563, 980.71931399, 1015.61607486, 1039.68280649, 959.05925553, 949.43256287, 1010.80272853, 1048.10616256, 1063.74953812, 1064.95287471, 1079.39291368, 1087.81626976, 1054.12284547, 1028.85277726, 1042.08947966, 1022.83609435, 1042.08947966, 985.53266032, 926.56916782, 926.56916782, 889.26573379, 872.41902165, 857.97898267, 709.96858314, 770.13541222, 765.32206589, 796.60881701, 741.25533426, 759.30538298, 780.96544145, 764.11872931, 756.89870982, 772.54208538, 808.64218283, 728.01863186, 676.27515885, 690.71519783, 722.00194895, 703.95190023, 762.91539273, 776.15209512, 819.47221206, 800.21882676, 802.62549992, 803.82883650, 788.18546094, 795.40548043, 806.23550966, 765.32206589, 741.25533426, 777.35543171, 766.52540247, 697.93521732, 673.86848569, 623.32834926, 602.87162738, 602.87162738, 604.07496396, 577.60155916, 587.22825181], 
							[2286.33950502, 2310.40623665, 2226.17267594, 2057.70555452, 1949.40526217, 1636.53775096, 1588.40428770, 1648.57111678, 1648.57111678, 1889.23843309, 1913.30516472, 1937.37189636, 1949.40526217, 2069.73892033, 2141.93911523, 2286.33950502, 2190.07257849, 2057.70555452, 2153.97248104, 2105.83901778, 2069.73892033, 1937.37189636, 1841.10496983, 1684.67121422, 1660.60448259, 2021.60545707, 2105.83901778, 2238.20604175, 2153.97248104, 2190.07257849, 2190.07257849, 2214.13931012, 2310.40623665, 2262.27277339, 2226.17267594, 2141.93911523, 2117.87238359, 2021.60545707, 2153.97248104, 2178.03921267, 2117.87238359, 2093.80565196, 2081.77228615, 2021.60545707, 2021.60545707, 1961.43862799, 1937.37189636, 1949.40526217, 2033.63882288, 2045.67218870, 2009.57209125, 2021.60545707, 1973.47199380, 1877.20506728, 2021.60545707, 2045.67218870, 2021.60545707, 2141.93911523, 2166.00584686, 2166.00584686, 2081.77228615, 2166.00584686, 2178.03921267, 2202.10594431, 2129.90574941, 1997.53872544, 2033.63882288, 1961.43862799, 1949.40526217, 1889.23843309, 1889.23843309, 1961.43862799, 1961.43862799, 1901.27179891, 1925.33853054, 1937.37189636, 1913.30516472, 1913.30516472, 1925.33853054, 1949.40526217, 1901.27179891, 1889.23843309, 1817.03823820, 1732.80467749, 1805.00487238, 1756.87140912, 1768.90477493, 1696.70458004, 1588.40428770, 1516.20409280, 1395.87043464, 1263.50341067, 1371.80370301, 1287.57014230, 1383.83706883, 1275.53677648, 1251.47004485, 1383.83706883, 1504.17072698, 1516.20409280, 1480.10399535, 1636.53775096, 1600.43765351, 1528.23745862, 1564.33755606, 1696.70458004, 1708.73794585, 1672.63784841, 1552.30419025, 1528.23745862, 1540.27082443, 1552.30419025, 1540.27082443, 1516.20409280, 1492.13736117, 1492.13736117, 1492.13736117, 1612.47101933, 1588.40428770, 1564.33755606, 1600.43765351, 1528.23745862, 1516.20409280, 1468.07062954, 1504.17072698, 1480.10399535, 1504.17072698, 1564.33755606, 1588.40428770, 1684.67121422, 1672.63784841, 1672.63784841, 1576.37092188, 1588.40428770, 1576.37092188, 1504.17072698, 1516.20409280, 1600.43765351, 1588.40428770, 1576.37092188, 1624.50438514, 1588.40428770, 1600.43765351, 1564.33755606, 1564.33755606, 1564.33755606, 1552.30419025, 1576.37092188, 1624.50438514, 1612.47101933, 1612.47101933, 1588.40428770, 1444.00389791, 1516.20409280, 1431.97053209, 1407.90380046, 1395.87043464, 1359.77033719, 1275.53677648, 1251.47004485, 1215.36994740, 1227.40331322, 1193.70988894, 1215.36994740, 1299.60350811, 1287.57014230, 1287.57014230, 1227.40331322, 1110.67966481, 993.95601639, 960.26259211, 915.73913859, 898.89242645, 933.78918731, 866.40233874, 1009.59939195, 1022.83609435, 1052.91950889, 1022.83609435, 1024.03943093, 1057.73285522, 1215.36994740, 1215.36994740, 1227.40331322, 1203.33658159, 1191.30321577, 1156.40645491, 1108.27299164, 1167.23648414, 1120.30635746, 1139.55974276, 1034.86946017, 1143.16975251, 1084.20626001, 1131.13638669, 1107.06965506, 1141.96641593, 1175.65984021, 1179.26984996, 1193.70988894, 1227.40331322, 1251.47004485, 1395.87043464, 1347.73697138, 1431.97053209, 1371.80370301, 1347.73697138, 1347.73697138, 1323.67023975, 1323.67023975, 1347.73697138, 1323.67023975, 1359.77033719, 1419.93716627, 1444.00389791, 1480.10399535, 1468.07062954, 1564.33755606, 1504.17072698, 1431.97053209, 1419.93716627, 1431.97053209, 1456.03726372, 1468.07062954, 1552.30419025, 1564.33755606, 1564.33755606, 1516.20409280, 1528.23745862, 1540.27082443, 1504.17072698, 1456.03726372, 1456.03726372, 1480.10399535, 1528.23745862, 1407.90380046, 1407.90380046, 1347.73697138, 1431.97053209, 1456.03726372, 1371.80370301, 1335.70360556, 1419.93716627, 1359.77033719, 1251.47004485, 1323.67023975, 1263.50341067, 1239.43667904, 1167.23648414, 987.93933348, 1015.61607486, 910.92579226, 903.70577277, 903.70577277, 783.37211461, 791.79547068, 759.30538298, 782.16877803, 734.03531477, 792.99880727, 807.43884625, 770.13541222, 744.86534400, 622.12501268, 685.90185151, 488.55465212, 476.52128631, 433.20116937, 509.01137401, 537.89145197, 566.77152993, 607.68497370, 695.52854416, 652.20842722, 659.42844671, 689.51186125, 666.64846620, 666.64846620, 660.63178329, 577.60155916, 516.23139350, 505.40136427, 504.19802769, 459.67457417, 428.38782305, 398.30440851, 411.54111090, 401.91441825, 397.10107192, 381.45769636, 373.03434029], 
							[1431.97053209, 1431.97053209, 1335.70360556, 1287.57014230, 1263.50341067, 1251.47004485, 1287.57014230, 1287.57014230, 1251.47004485, 1227.40331322, 1311.63687393, 1323.67023975, 1347.73697138, 1347.73697138, 1311.63687393, 1263.50341067, 1203.33658159, 1263.50341067, 1323.67023975, 1335.70360556, 1311.63687393, 1227.40331322, 1323.67023975, 1335.70360556, 1227.40331322, 1299.60350811, 1431.97053209, 1444.00389791, 1480.10399535, 1444.00389791, 1492.13736117, 1323.67023975, 1431.97053209, 1504.17072698, 1480.10399535, 1492.13736117, 1456.03726372, 1456.03726372, 1480.10399535, 1540.27082443, 1480.10399535, 1504.17072698, 1564.33755606, 1528.23745862, 1516.20409280, 1492.13736117, 1444.00389791, 1504.17072698, 1552.30419025, 1564.33755606, 1540.27082443, 1528.23745862, 1552.30419025, 1564.33755606, 1600.43765351, 1612.47101933, 1636.53775096, 1660.60448259, 1636.53775096, 1636.53775096, 1660.60448259, 1660.60448259, 1612.47101933, 1648.57111678, 1624.50438514, 1624.50438514, 1612.47101933, 1588.40428770, 1540.27082443, 1504.17072698, 1516.20409280, 1600.43765351, 1588.40428770, 1540.27082443, 1588.40428770, 1528.23745862, 1552.30419025, 1540.27082443, 1564.33755606, 1588.40428770, 1552.30419025, 1564.33755606, 1624.50438514, 1648.57111678, 1600.43765351, 1576.37092188, 1600.43765351, 1588.40428770, 1564.33755606, 1576.37092188, 1528.23745862, 1431.97053209, 1419.93716627, 1419.93716627, 1444.00389791, 1347.73697138, 1323.67023975, 1383.83706883, 1492.13736117, 1540.27082443, 1504.17072698, 1576.37092188, 1540.27082443, 1504.17072698, 1504.17072698, 1540.27082443, 1564.33755606, 1564.33755606, 1480.10399535, 1444.00389791, 1516.20409280, 1516.20409280, 1492.13736117, 1444.00389791, 1456.03726372, 1504.17072698, 1516.20409280, 1468.07062954, 1468.07062954, 1444.00389791, 1444.00389791, 1456.03726372, 1395.87043464, 1444.00389791, 1468.07062954, 1468.07062954, 1492.13736117, 1456.03726372, 1456.03726372, 1504.17072698, 1516.20409280, 1516.20409280, 1492.13736117, 1444.00389791, 1456.03726372, 1395.87043464, 1456.03726372, 1456.03726372, 1468.07062954, 1431.97053209, 1395.87043464, 1419.93716627, 1444.00389791, 1444.00389791, 1480.10399535, 1444.00389791, 1431.97053209, 1444.00389791, 1456.03726372, 1444.00389791, 1444.00389791, 1468.07062954, 1468.07062954, 1456.03726372, 1431.97053209, 1359.77033719, 1335.70360556, 1335.70360556, 1299.60350811, 1239.43667904, 1227.40331322, 1371.80370301, 1239.43667904, 1263.50341067, 1383.83706883, 1347.73697138, 1347.73697138, 1347.73697138, 1311.63687393, 1359.77033719, 1359.77033719, 1263.50341067, 1347.73697138, 1383.83706883, 1347.73697138, 1407.90380046, 1383.83706883, 1444.00389791, 1431.97053209, 1468.07062954, 1456.03726372, 1528.23745862, 1600.43765351, 1588.40428770, 1648.57111678, 1636.53775096, 1636.53775096, 1660.60448259, 1756.87140912, 1660.60448259, 1720.77131167, 1720.77131167, 1792.97150657, 1780.93814075, 1732.80467749, 1708.73794585, 1841.10496983, 1949.40526217, 1913.30516472, 1973.47199380, 2045.67218870, 2117.87238359, 2009.57209125, 1997.53872544, 1913.30516472, 2033.63882288, 2117.87238359, 2093.80565196, 2069.73892033, 1901.27179891, 1817.03823820, 1913.30516472, 1985.50535962, 2009.57209125, 1985.50535962, 1985.50535962, 2069.73892033, 2141.93911523, 2129.90574941, 2105.83901778, 2178.03921267, 2129.90574941, 2153.97248104, 2117.87238359, 2334.47296828, 2322.43960246, 2334.47296828, 2346.50633410, 2346.50633410, 2298.37287083, 2262.27277339, 2226.17267594, 2105.83901778, 2178.03921267, 2238.20604175, 2262.27277339, 2262.27277339, 2370.57306573, 2406.67316318, 2454.80662644, 2394.63979736, 2382.60643154, 2478.87335807, 2382.60643154, 2406.67316318, 2478.87335807, 2502.94008970, 2599.20701623, 2683.44057694, 2839.87433255, 2815.80760092, 2743.60740602, 2924.10789326, 2924.10789326, 2707.50730857, 2707.50730857, 2851.90769836, 2875.97442999, 2900.04116163, 3212.90867284, 3020.37481979, 2912.07452744, 2972.24135652, 3008.34145397, 3092.57501468, 3116.64174631, 3273.07550192, 3381.37579426, 3357.30906263, 3381.37579426, 3549.84291568, 3561.87628150, 3489.67608660, 3597.97637895, 3549.84291568, 3670.17657384, 3718.31003711, 3886.77715853, 3838.64369527, 3971.01071924, 3958.97735342, 4175.57793811, 4259.81149882, 4283.87823045, 4319.97832790, 4271.84486464, 4055.24427995, 3862.71042690, 4115.41110903, 4007.11081669, 4223.71140137], 
							[552.33149095, 559.55151044, 561.95818360, 527.06142274, 511.41804717, 481.33463264, 448.84454493, 433.20116937, 457.26790100, 468.09793024, 505.40136427, 471.70793998, 487.35131554, 505.40136427, 481.33463264, 463.28458391, 427.18448646, 434.40450595, 446.43787177, 453.65789126, 442.82786202, 415.15112065, 421.16780356, 403.11775483, 361.00097448, 400.71108167, 499.38468136, 481.33463264, 463.28458391, 489.75798871, 487.35131554, 458.47123758, 433.20116937, 450.04788151, 442.82786202, 419.96446697, 445.23453519, 412.74444748, 419.96446697, 421.16780356, 405.52442800, 424.77781330, 430.79449621, 427.18448646, 418.76113039, 398.30440851, 413.94778407, 397.10107192, 403.11775483, 421.16780356, 409.13443774, 413.94778407, 399.50774509, 387.47437927, 413.94778407, 421.16780356, 418.76113039, 451.25121810, 451.25121810, 469.30126682, 447.64120835, 487.35131554, 481.33463264, 478.92795947, 476.52128631, 513.82472034, 512.62138376, 505.40136427, 475.31794973, 474.11461315, 464.48792049, 481.33463264, 489.75798871, 475.31794973, 484.94464238, 481.33463264, 495.77467161, 482.53796922, 489.75798871, 492.16466187, 486.14797896, 481.33463264, 472.91127656, 464.48792049, 475.31794973, 474.11461315, 475.31794973, 459.67457417, 442.82786202, 425.98114988, 412.74444748, 387.47437927, 389.88105243, 379.05102320, 406.72776458, 380.25435978, 407.93110116, 422.37114014, 452.45455468, 440.42118886, 439.21785228, 480.13129605, 505.40136427, 481.33463264, 507.80803743, 496.97800820, 515.02805692, 569.17820309, 572.78821284, 548.72148120, 587.22825181, 577.60155916, 563.16152018, 542.70479830, 529.46809590, 543.90813488, 545.11147146, 559.55151044, 549.92481779, 567.97486651, 569.17820309, 559.55151044, 534.28144222, 551.12815437, 566.77152993, 539.09478855, 557.14483728, 546.31480804, 547.51814462, 555.94150069, 540.29812513, 543.90813488, 506.60470085, 510.21471059, 523.45141299, 522.24807641, 507.80803743, 505.40136427, 505.40136427, 504.19802769, 500.58801794, 487.35131554, 489.75798871, 475.31794973, 466.89459366, 452.45455468, 465.69125707, 472.91127656, 477.72462289, 483.74130580, 478.92795947, 484.94464238, 444.03119861, 445.23453519, 463.28458391, 451.25121810, 419.96446697, 392.28772560, 403.11775483, 400.71108167, 393.49106218, 411.54111090, 387.47437927, 398.30440851, 406.72776458, 405.52442800, 409.13443774, 409.13443774, 400.71108167, 379.05102320, 377.84768662, 369.42433055, 356.18762815, 346.56093550, 336.93424284, 361.00097448, 381.45769636, 379.05102320, 369.42433055, 388.67771585, 383.86436953, 401.91441825, 404.32109141, 405.52442800, 415.15112065, 404.32109141, 391.08438902, 387.47437927, 416.35445723, 403.11775483, 407.93110116, 379.05102320, 400.71108167, 369.42433055, 391.08438902, 401.91441825, 407.93110116, 415.15112065, 411.54111090, 430.79449621, 446.43787177, 452.45455468, 452.45455468, 459.67457417, 469.30126682, 477.72462289, 470.50460340, 468.09793024, 464.48792049, 457.26790100, 466.89459366, 462.08124733, 464.48792049, 456.06456442, 489.75798871, 500.58801794, 513.82472034, 516.23139350, 495.77467161, 470.50460340, 488.55465212, 489.75798871, 476.52128631, 477.72462289, 489.75798871, 486.14797896, 494.57133503, 484.94464238, 464.48792049, 464.48792049, 460.87791075, 445.23453519, 442.82786202, 451.25121810, 465.69125707, 436.81117912, 428.38782305, 450.04788151, 446.43787177, 457.26790100, 441.62452544, 435.60784253, 442.82786202, 421.16780356, 438.01451570, 445.23453519, 428.38782305, 435.60784253, 431.99783279, 403.11775483, 417.55779381, 397.10107192, 387.47437927, 418.76113039, 345.35759892, 371.83100371, 359.79763789, 380.25435978, 362.20431106, 362.20431106, 371.83100371, 364.61098422, 352.57761841, 356.18762815, 379.05102320, 352.57761841, 341.74758917, 333.32423310, 346.56093550, 340.54425259, 364.61098422, 375.44101346, 387.47437927, 373.03434029, 356.18762815, 375.44101346, 383.86436953, 367.01765738, 359.79763789, 350.17094524, 353.78095499, 348.96760866, 344.15426233, 308.05416489, 306.85082830, 285.19076984, 276.76741377, 286.39410642, 288.80077958, 271.95406744, 252.70068213], 
							[758.10204640, 734.03531477, 738.84866109, 748.47535375, 735.23865135, 732.83197819, 679.88516860, 629.34503217, 634.15837850, 602.87162738, 661.83511987, 632.95504192, 673.86848569, 640.17506140, 608.88831028, 596.85494447, 551.12815437, 518.63806666, 553.53482753, 573.99154942, 555.94150069, 541.50146171, 506.60470085, 513.82472034, 501.79135452, 535.48477881, 602.87162738, 637.76838824, 628.14169559, 631.75170533, 687.10518809, 700.34189048, 709.96858314, 697.93521732, 697.93521732, 643.78507115, 649.80175406, 577.60155916, 594.44827130, 620.92167610, 617.31166635, 595.65160789, 604.07496396, 586.02491523, 560.75484702, 565.56819335, 561.95818360, 569.17820309, 586.02491523, 613.70165661, 584.82157865, 606.48163712, 624.53168584, 641.37839799, 678.68183202, 667.85180278, 752.08536349, 848.35229002, 800.21882676, 795.40548043, 758.10204640, 857.97898267, 883.24905089, 877.23236798, 854.36897293, 863.99566558, 850.75896318, 848.35229002, 817.06553890, 770.13541222, 758.10204640, 779.76210487, 813.45552915, 783.37211461, 800.21882676, 795.40548043, 866.40233874, 855.57230951, 856.77564609, 862.79232900, 830.30224130, 799.01549017, 815.86220232, 780.96544145, 805.03217308, 799.01549017, 764.11872931, 695.52854416, 616.10832977, 655.81843697, 607.68497370, 577.60155916, 625.73502243, 599.26161763, 666.64846620, 606.48163712, 625.73502243, 718.39193921, 705.15523681, 746.06868058, 728.01863186, 814.65886573, 792.99880727, 729.22196844, 771.33874880, 770.13541222, 758.10204640, 766.52540247, 756.89870982, 706.35857339, 732.83197819, 734.03531477, 731.62864161, 712.37525630, 722.00194895, 735.23865135, 729.22196844, 746.06868058, 791.79547068, 788.18546094, 788.18546094, 753.28870007, 744.86534400, 743.66200742, 749.67869033, 729.22196844, 718.39193921, 728.01863186, 741.25533426, 724.40862212, 732.83197819, 731.62864161, 707.56190997, 706.35857339, 701.54522707, 678.68183202, 671.46181253, 679.88516860, 685.90185151, 679.88516860, 691.91853441, 655.81843697, 649.80175406, 614.90499319, 610.09164687, 640.17506140, 660.63178329, 684.69851492, 663.03845645, 661.83511987, 649.80175406, 636.56505166, 613.70165661, 697.93521732, 679.88516860, 670.25847594, 637.76838824, 616.10832977, 590.83826156, 583.61824207, 537.89145197, 535.48477881, 529.46809590, 542.70479830, 582.41490549, 561.95818360, 580.00823233, 546.31480804, 535.48477881, 502.99469110, 511.41804717, 490.96132529, 450.04788151, 454.86122784, 435.60784253, 537.89145197, 541.50146171, 540.29812513, 528.26475932, 519.84140325, 515.02805692, 569.17820309, 546.31480804, 613.70165661, 607.68497370, 611.29498345, 565.56819335, 553.53482753, 589.63492498, 572.78821284, 576.39822258, 525.85808615, 531.87476906, 540.29812513, 565.56819335, 553.53482753, 560.75484702, 570.38153967, 583.61824207, 584.82157865, 589.63492498, 637.76838824, 607.68497370, 590.83826156, 618.51500294, 675.07182227, 687.10518809, 661.83511987, 638.97172482, 637.76838824, 646.19174431, 636.56505166, 664.24179304, 688.30852467, 694.32520758, 688.30852467, 701.54522707, 728.01863186, 696.73188074, 678.68183202, 709.96858314, 726.81529528, 749.67869033, 759.30538298, 753.28870007, 762.91539273, 748.47535375, 728.01863186, 715.98526604, 712.37525630, 691.91853441, 673.86848569, 667.85180278, 703.95190023, 699.13855390, 653.41176380, 649.80175406, 661.83511987, 688.30852467, 690.71519783, 681.08850518, 673.86848569, 703.95190023, 673.86848569, 659.42844671, 702.74856365, 669.05513936, 679.88516860, 630.54836875, 584.82157865, 600.46495421, 546.31480804, 522.24807641, 515.02805692, 434.40450595, 492.16466187, 472.91127656, 494.57133503, 481.33463264, 487.35131554, 509.01137401, 488.55465212, 472.91127656, 464.48792049, 496.97800820, 460.87791075, 421.16780356, 419.96446697, 451.25121810, 428.38782305, 476.52128631, 474.11461315, 511.41804717, 494.57133503, 522.24807641, 517.43473008, 506.60470085, 490.96132529, 490.96132529, 464.48792049, 458.47123758, 463.28458391, 428.38782305, 397.10107192, 409.13443774, 374.23767687, 364.61098422, 364.61098422, 365.81432080, 345.35759892, 326.10421361], 
							[1648.57111678, 1576.37092188, 1540.27082443, 1576.37092188, 1528.23745862, 1552.30419025, 1588.40428770, 1588.40428770, 1696.70458004, 1696.70458004, 1720.77131167, 1636.53775096, 1660.60448259, 1660.60448259, 1612.47101933, 1588.40428770, 1540.27082443, 1588.40428770, 1624.50438514, 1564.33755606, 1564.33755606, 1444.00389791, 1492.13736117, 1468.07062954, 1347.73697138, 1492.13736117, 1516.20409280, 1540.27082443, 1444.00389791, 1444.00389791, 1371.80370301, 1311.63687393, 1335.70360556, 1359.77033719, 1383.83706883, 1335.70360556, 1335.70360556, 1287.57014230, 1275.53677648, 1347.73697138, 1371.80370301, 1407.90380046, 1431.97053209, 1407.90380046, 1395.87043464, 1347.73697138, 1323.67023975, 1335.70360556, 1383.83706883, 1323.67023975, 1287.57014230, 1287.57014230, 1263.50341067, 1239.43667904, 1359.77033719, 1383.83706883, 1431.97053209, 1480.10399535, 1600.43765351, 1660.60448259, 1600.43765351, 1684.67121422, 1672.63784841, 1624.50438514, 1576.37092188, 1552.30419025, 1624.50438514, 1552.30419025, 1552.30419025, 1528.23745862, 1516.20409280, 1528.23745862, 1564.33755606, 1588.40428770, 1600.43765351, 1624.50438514, 1648.57111678, 1684.67121422, 1684.67121422, 1805.00487238, 1756.87140912, 1708.73794585, 1696.70458004, 1720.77131167, 1756.87140912, 1792.97150657, 1792.97150657, 1768.90477493, 1708.73794585, 1744.83804330, 1708.73794585, 1588.40428770, 1756.87140912, 1780.93814075, 1841.10496983, 1768.90477493, 1792.97150657, 1865.17170146, 1901.27179891, 1937.37189636, 1937.37189636, 2009.57209125, 2069.73892033, 2045.67218870, 2105.83901778, 2069.73892033, 2166.00584686, 2166.00584686, 2069.73892033, 2057.70555452, 2057.70555452, 2069.73892033, 2057.70555452, 2009.57209125, 1997.53872544, 2057.70555452, 2069.73892033, 2202.10594431, 2238.20604175, 2274.30613920, 2286.33950502, 2178.03921267, 2141.93911523, 2178.03921267, 2117.87238359, 2057.70555452, 1973.47199380, 2117.87238359, 2081.77228615, 2069.73892033, 2081.77228615, 2141.93911523, 1985.50535962, 1925.33853054, 1997.53872544, 1901.27179891, 1985.50535962, 2009.57209125, 1901.27179891, 1913.30516472, 1925.33853054, 1853.13833565, 1853.13833565, 1805.00487238, 1853.13833565, 1889.23843309, 1889.23843309, 1913.30516472, 1973.47199380, 1937.37189636, 1889.23843309, 1792.97150657, 1720.77131167, 1720.77131167, 1696.70458004, 1588.40428770, 1660.60448259, 1756.87140912, 1648.57111678, 1576.37092188, 1516.20409280, 1684.67121422, 1660.60448259, 1672.63784841, 1732.80467749, 1696.70458004, 1660.60448259, 1576.37092188, 1636.53775096, 1600.43765351, 1600.43765351, 1516.20409280, 1516.20409280, 1600.43765351, 1588.40428770, 1672.63784841, 1792.97150657, 1708.73794585, 1756.87140912, 1853.13833565, 1877.20506728, 1925.33853054, 1853.13833565, 1841.10496983, 1829.07160401, 1853.13833565, 1768.90477493, 1829.07160401, 1877.20506728, 1792.97150657, 1829.07160401, 1708.73794585, 1744.83804330, 1732.80467749, 1744.83804330, 1744.83804330, 1829.07160401, 1913.30516472, 1937.37189636, 2009.57209125, 2021.60545707, 2129.90574941, 2069.73892033, 2069.73892033, 2093.80565196, 2117.87238359, 2093.80565196, 2105.83901778, 2057.70555452, 2033.63882288, 2021.60545707, 1997.53872544, 2045.67218870, 2093.80565196, 2105.83901778, 2117.87238359, 2166.00584686, 2226.17267594, 2202.10594431, 2153.97248104, 2166.00584686, 2190.07257849, 2238.20604175, 2238.20604175, 2346.50633410, 2382.60643154, 2454.80662644, 2406.67316318, 2394.63979736, 2322.43960246, 2238.20604175, 2274.30613920, 2238.20604175, 2286.33950502, 2322.43960246, 2202.10594431, 2190.07257849, 2190.07257849, 2202.10594431, 2250.23940757, 2238.20604175, 2226.17267594, 2310.40623665, 2190.07257849, 2238.20604175, 2262.27277339, 2250.23940757, 2298.37287083, 2370.57306573, 2322.43960246, 2298.37287083, 2226.17267594, 2262.27277339, 2274.30613920, 2141.93911523, 2069.73892033, 2081.77228615, 2202.10594431, 2226.17267594, 2382.60643154, 2262.27277339, 2238.20604175, 2166.00584686, 2238.20604175, 2250.23940757, 2310.40623665, 2334.47296828, 2370.57306573, 2370.57306573, 2418.70652899, 2587.17365041, 2671.40721112, 2635.30711368, 2551.07355297, 2635.30711368, 2671.40721112, 2659.37384531, 2707.50730857, 2755.64077184, 2683.44057694, 2767.67413765, 2875.97442999, 2875.97442999, 2972.24135652, 2936.14125907, 2888.00779581, 2683.44057694, 2671.40721112, 2659.37384531, 2514.97345552, 2539.04018715], 
							[495.77467161, 481.33463264, 470.50460340, 442.82786202, 442.82786202, 409.13443774, 397.10107192, 404.32109141, 409.13443774, 403.11775483, 423.57447672, 433.20116937, 440.42118886, 446.43787177, 421.16780356, 410.33777432, 418.76113039, 380.25435978, 392.28772560, 415.15112065, 416.35445723, 393.49106218, 412.74444748, 409.13443774, 379.05102320, 421.16780356, 450.04788151, 441.62452544, 429.59115963, 418.76113039, 434.40450595, 404.32109141, 409.13443774, 407.93110116, 391.08438902, 386.27104269, 385.06770611, 380.25435978, 395.89773534, 398.30440851, 399.50774509, 410.33777432, 411.54111090, 403.11775483, 388.67771585, 379.05102320, 375.44101346, 373.03434029, 389.88105243, 387.47437927, 379.05102320, 385.06770611, 385.06770611, 399.50774509, 412.74444748, 429.59115963, 493.36799845, 512.62138376, 515.02805692, 522.24807641, 487.35131554, 527.06142274, 521.04473983, 524.65474957, 524.65474957, 528.26475932, 529.46809590, 512.62138376, 506.60470085, 504.19802769, 498.18134478, 498.18134478, 511.41804717, 502.99469110, 511.41804717, 506.60470085, 527.06142274, 516.23139350, 507.80803743, 505.40136427, 498.18134478, 489.75798871, 476.52128631, 452.45455468, 463.28458391, 472.91127656, 493.36799845, 462.08124733, 421.16780356, 404.32109141, 389.88105243, 373.03434029, 391.08438902, 369.42433055, 405.52442800, 371.83100371, 391.08438902, 404.32109141, 440.42118886, 441.62452544, 421.16780356, 463.28458391, 452.45455468, 436.81117912, 446.43787177, 454.86122784, 440.42118886, 454.86122784, 436.81117912, 439.21785228, 489.75798871, 493.36799845, 482.53796922, 470.50460340, 477.72462289, 496.97800820, 489.75798871, 512.62138376, 509.01137401, 515.02805692, 525.85808615, 509.01137401, 486.14797896, 498.18134478, 519.84140325, 498.18134478, 499.38468136, 513.82472034, 511.41804717, 513.82472034, 519.84140325, 553.53482753, 537.89145197, 534.28144222, 537.89145197, 525.85808615, 530.67143248, 545.11147146, 540.29812513, 540.29812513, 547.51814462, 534.28144222, 535.48477881, 495.77467161, 495.77467161, 501.79135452, 504.19802769, 517.43473008, 513.82472034, 505.40136427, 499.38468136, 462.08124733, 451.25121810, 450.04788151, 460.87791075, 439.21785228, 409.13443774, 399.50774509, 391.08438902, 395.89773534, 380.25435978, 398.30440851, 379.05102320, 395.89773534, 418.76113039, 422.37114014, 484.94464238, 460.87791075, 446.43787177, 448.84454493, 446.43787177, 424.77781330, 413.94778407, 411.54111090, 394.69439876, 430.79449621, 436.81117912, 425.98114988, 430.79449621, 444.03119861, 444.03119861, 462.08124733, 460.87791075, 493.36799845, 476.52128631, 490.96132529, 472.91127656, 457.26790100, 486.14797896, 472.91127656, 475.31794973, 445.23453519, 478.92795947, 486.14797896, 512.62138376, 519.84140325, 513.82472034, 527.06142274, 500.58801794, 502.99469110, 509.01137401, 512.62138376, 510.21471059, 505.40136427, 511.41804717, 523.45141299, 529.46809590, 535.48477881, 535.48477881, 534.28144222, 529.46809590, 507.80803743, 531.87476906, 542.70479830, 542.70479830, 558.34817386, 552.33149095, 536.68811539, 536.68811539, 530.67143248, 541.50146171, 527.06142274, 575.19488600, 560.75484702, 572.78821284, 589.63492498, 586.02491523, 569.17820309, 547.51814462, 545.11147146, 529.46809590, 528.26475932, 517.43473008, 527.06142274, 543.90813488, 523.45141299, 516.23139350, 521.04473983, 547.51814462, 552.33149095, 541.50146171, 537.89145197, 549.92481779, 530.67143248, 536.68811539, 540.29812513, 551.12815437, 549.92481779, 537.89145197, 519.84140325, 537.89145197, 521.04473983, 516.23139350, 516.23139350, 440.42118886, 505.40136427, 481.33463264, 496.97800820, 476.52128631, 500.58801794, 522.24807641, 527.06142274, 501.79135452, 500.58801794, 522.24807641, 505.40136427, 517.43473008, 527.06142274, 542.70479830, 559.55151044, 576.39822258, 581.21156891, 589.63492498, 570.38153967, 584.82157865, 593.24493472, 592.04159814, 580.00823233, 569.17820309, 559.55151044, 578.80489574, 588.43158840, 584.82157865, 543.90813488, 548.72148120, 535.48477881, 569.17820309, 551.12815437, 541.50146171, 511.41804717, 484.94464238], 
							[638.97172482, 631.75170533, 602.87162738, 566.77152993, 577.60155916, 533.07810564, 517.43473008, 516.23139350, 487.35131554, 471.70793998, 545.11147146, 541.50146171, 553.53482753, 553.53482753, 527.06142274, 499.38468136, 481.33463264, 457.26790100, 494.57133503, 495.77467161, 487.35131554, 469.30126682, 499.38468136, 495.77467161, 441.62452544, 498.18134478, 637.76838824, 637.76838824, 672.66514911, 676.27515885, 695.52854416, 671.46181253, 679.88516860, 676.27515885, 653.41176380, 647.39508089, 629.34503217, 631.75170533, 655.81843697, 663.03845645, 649.80175406, 651.00509064, 670.25847594, 651.00509064, 628.14169559, 629.34503217, 619.71833952, 611.29498345, 644.98840773, 642.58173457, 625.73502243, 641.37839799, 640.17506140, 622.12501268, 648.59841748, 604.07496396, 652.20842722, 707.56190997, 722.00194895, 726.81529528, 701.54522707, 749.67869033, 732.83197819, 726.81529528, 731.62864161, 794.20214385, 771.33874880, 762.91539273, 752.08536349, 734.03531477, 713.57859288, 722.00194895, 715.98526604, 677.47849543, 696.73188074, 697.93521732, 734.03531477, 746.06868058, 759.30538298, 760.50871956, 754.49203666, 732.83197819, 720.79861237, 708.76524656, 695.52854416, 677.47849543, 673.86848569, 631.75170533, 601.66829079, 604.07496396, 587.22825181, 507.80803743, 552.33149095, 535.48477881, 545.11147146, 530.67143248, 529.46809590, 541.50146171, 602.87162738, 602.87162738, 587.22825181, 637.76838824, 602.87162738, 622.12501268, 681.08850518, 665.44512962, 637.76838824, 607.68497370, 607.68497370, 581.21156891, 607.68497370, 605.27830054, 604.07496396, 595.65160789, 600.46495421, 619.71833952, 607.68497370, 625.73502243, 654.61510038, 647.39508089, 652.20842722, 642.58173457, 649.80175406, 646.19174431, 651.00509064, 642.58173457, 657.02177355, 666.64846620, 673.86848569, 657.02177355, 655.81843697, 654.61510038, 629.34503217, 618.51500294, 605.27830054, 630.54836875, 630.54836875, 634.15837850, 623.32834926, 577.60155916, 571.58487625, 543.90813488, 565.56819335, 555.94150069, 555.94150069, 545.11147146, 558.34817386, 576.39822258, 551.12815437, 543.90813488, 518.63806666, 588.43158840, 643.78507115, 626.93835901, 619.71833952, 588.43158840, 578.80489574, 541.50146171, 522.24807641, 493.36799845, 456.06456442, 498.18134478, 465.69125707, 477.72462289, 489.75798871, 472.91127656, 483.74130580, 456.06456442, 459.67457417, 416.35445723, 442.82786202, 428.38782305, 389.88105243, 403.11775483, 374.23767687, 421.16780356, 425.98114988, 428.38782305, 404.32109141, 403.11775483, 421.16780356, 453.65789126, 454.86122784, 466.89459366, 442.82786202, 475.31794973, 445.23453519, 451.25121810, 478.92795947, 463.28458391, 523.45141299, 507.80803743, 493.36799845, 495.77467161, 523.45141299, 572.78821284, 551.12815437, 567.97486651, 559.55151044, 573.99154942, 577.60155916, 582.41490549, 561.95818360, 575.19488600, 587.22825181, 613.70165661, 587.22825181, 595.65160789, 599.26161763, 593.24493472, 582.41490549, 587.22825181, 594.44827130, 634.15837850, 624.53168584, 641.37839799, 637.76838824, 654.61510038, 636.56505166, 625.73502243, 642.58173457, 643.78507115, 644.98840773, 619.71833952, 624.53168584, 651.00509064, 642.58173457, 635.36171508, 626.93835901, 619.71833952, 613.70165661, 589.63492498, 573.99154942, 588.43158840, 571.58487625, 530.67143248, 541.50146171, 564.36485676, 576.39822258, 576.39822258, 575.19488600, 566.77152993, 593.24493472, 580.00823233, 560.75484702, 581.21156891, 572.78821284, 573.99154942, 517.43473008, 511.41804717, 512.62138376, 475.31794973, 454.86122784, 460.87791075, 395.89773534, 427.18448646, 406.72776458, 444.03119861, 409.13443774, 421.16780356, 416.35445723, 397.10107192, 386.27104269, 375.44101346, 424.77781330, 405.52442800, 380.25435978, 379.05102320, 388.67771585, 394.69439876, 393.49106218, 407.93110116, 433.20116937, 442.82786202, 422.37114014, 421.16780356, 427.18448646, 415.15112065, 421.16780356, 409.13443774, 387.47437927, 381.45769636, 380.25435978, 330.91755994, 323.69754045, 300.83414540, 286.39410642, 281.58076009, 279.17408693, 277.97075035, 274.36074060], 
							[866.40233874, 876.02903140, 812.25219257, 754.49203666, 752.08536349, 687.10518809, 613.70165661, 607.68497370, 625.73502243, 601.66829079, 637.76838824, 602.87162738, 646.19174431, 661.83511987, 625.73502243, 617.31166635, 601.66829079, 577.60155916, 681.08850518, 663.03845645, 625.73502243, 589.63492498, 601.66829079, 593.24493472, 541.50146171, 624.53168584, 638.97172482, 683.49517834, 649.80175406, 673.86848569, 670.25847594, 629.34503217, 640.17506140, 647.39508089, 638.97172482, 608.88831028, 637.76838824, 593.24493472, 625.73502243, 631.75170533, 611.29498345, 625.73502243, 629.34503217, 604.07496396, 559.55151044, 560.75484702, 561.95818360, 546.31480804, 582.41490549, 601.66829079, 590.83826156, 595.65160789, 573.99154942, 580.00823233, 607.68497370, 598.05828105, 580.00823233, 589.63492498, 632.95504192, 654.61510038, 637.76838824, 660.63178329, 661.83511987, 640.17506140, 652.20842722, 685.90185151, 708.76524656, 687.10518809, 670.25847594, 644.98840773, 648.59841748, 682.29184176, 667.85180278, 658.22511013, 681.08850518, 684.69851492, 693.12187099, 696.73188074, 707.56190997, 703.95190023, 685.90185151, 673.86848569, 666.64846620, 661.83511987, 624.53168584, 631.75170533, 598.05828105, 559.55151044, 500.58801794, 523.45141299, 486.14797896, 435.60784253, 463.28458391, 434.40450595, 463.28458391, 433.20116937, 445.23453519, 466.89459366, 516.23139350, 511.41804717, 483.74130580, 541.50146171, 523.45141299, 478.92795947, 487.35131554, 517.43473008, 528.26475932, 553.53482753, 529.46809590, 512.62138376, 529.46809590, 548.72148120, 536.68811539, 524.65474957, 523.45141299, 529.46809590, 539.09478855, 546.31480804, 552.33149095, 577.60155916, 600.46495421, 554.73816411, 539.09478855, 540.29812513, 551.12815437, 535.48477881, 551.12815437, 539.09478855, 551.12815437, 529.46809590, 535.48477881, 531.87476906, 531.87476906, 519.84140325, 501.79135452, 482.53796922, 489.75798871, 507.80803743, 511.41804717, 507.80803743, 501.79135452, 484.94464238, 487.35131554, 445.23453519, 464.48792049, 466.89459366, 462.08124733, 484.94464238, 483.74130580, 484.94464238, 490.96132529, 492.16466187, 440.42118886, 469.30126682, 478.92795947, 478.92795947, 436.81117912, 424.77781330, 403.11775483, 421.16780356, 421.16780356, 409.13443774, 400.71108167, 404.32109141, 445.23453519, 427.18448646, 442.82786202, 434.40450595, 412.74444748, 404.32109141, 424.77781330, 382.66103294, 364.61098422, 391.08438902, 363.40764764, 417.55779381, 424.77781330, 431.99783279, 427.18448646, 427.18448646, 413.94778407, 445.23453519, 553.53482753, 626.93835901, 661.83511987, 702.74856365, 701.54522707, 688.30852467, 706.35857339, 702.74856365, 697.93521732, 673.86848569, 658.22511013, 699.13855390, 673.86848569, 663.03845645, 652.20842722, 672.66514911, 648.59841748, 649.80175406, 651.00509064, 664.24179304, 675.07182227, 651.00509064, 926.56916782, 1227.40331322, 1109.47632822, 986.73599690, 957.85591894, 951.83923604, 928.97584099, 906.11244594, 919.34914833, 973.49929450, 934.99252389, 955.44924578, 931.38251415, 918.14581175, 910.92579226, 884.45238747, 873.62235823, 896.48575328, 874.82569481, 797.81215359, 837.52226079, 824.28555839, 841.13227053, 829.09890471, 814.65886573, 795.40548043, 724.40862212, 632.95504192, 573.99154942, 637.76838824, 687.10518809, 644.98840773, 643.78507115, 678.68183202, 695.52854416, 679.88516860, 667.85180278, 644.98840773, 697.93521732, 654.61510038, 634.15837850, 673.86848569, 649.80175406, 672.66514911, 630.54836875, 582.41490549, 582.41490549, 537.89145197, 509.01137401, 500.58801794, 421.16780356, 469.30126682, 458.47123758, 475.31794973, 458.47123758, 457.26790100, 446.43787177, 452.45455468, 457.26790100, 439.21785228, 457.26790100, 431.99783279, 386.27104269, 382.66103294, 409.13443774, 386.27104269, 435.60784253, 439.21785228, 480.13129605, 464.48792049, 475.31794973, 489.75798871, 477.72462289, 469.30126682, 459.67457417, 439.21785228, 419.96446697, 415.15112065, 419.96446697, 381.45769636, 369.42433055, 357.39096473, 338.13757943, 336.93424284, 333.32423310, 304.44415514, 290.00411616], 
							[637.76838824, 634.15837850, 631.75170533, 581.21156891, 588.43158840, 517.43473008, 490.96132529, 527.06142274, 527.06142274, 524.65474957, 560.75484702, 530.67143248, 619.71833952, 611.29498345, 582.41490549, 582.41490549, 528.26475932, 487.35131554, 553.53482753, 553.53482753, 536.68811539, 493.36799845, 511.41804717, 486.14797896, 419.96446697, 487.35131554, 552.33149095, 559.55151044, 578.80489574, 567.97486651, 573.99154942, 546.31480804, 529.46809590, 543.90813488, 540.29812513, 524.65474957, 525.85808615, 490.96132529, 495.77467161, 512.62138376, 505.40136427, 525.85808615, 541.50146171, 518.63806666, 510.21471059, 494.57133503, 487.35131554, 470.50460340, 489.75798871, 492.16466187, 483.74130580, 482.53796922, 466.89459366, 448.84454493, 486.14797896, 471.70793998, 509.01137401, 594.44827130, 580.00823233, 576.39822258, 541.50146171, 588.43158840, 592.04159814, 613.70165661, 613.70165661, 614.90499319, 612.49832003, 601.66829079, 561.95818360, 537.89145197, 537.89145197, 547.51814462, 553.53482753, 530.67143248, 545.11147146, 533.07810564, 559.55151044, 539.09478855, 554.73816411, 537.89145197, 521.04473983, 516.23139350, 492.16466187, 490.96132529, 492.16466187, 476.52128631, 468.09793024, 433.20116937, 386.27104269, 406.72776458, 376.64435004, 361.00097448, 371.83100371, 363.40764764, 363.40764764, 341.74758917, 358.59430131, 362.20431106, 413.94778407, 398.30440851, 381.45769636, 427.18448646, 410.33777432, 377.84768662, 397.10107192, 407.93110116, 400.71108167, 428.38782305, 421.16780356, 398.30440851, 406.72776458, 418.76113039, 424.77781330, 423.57447672, 428.38782305, 442.82786202, 440.42118886, 459.67457417, 459.67457417, 475.31794973, 469.30126682, 457.26790100, 431.99783279, 438.01451570, 436.81117912, 425.98114988, 417.55779381, 427.18448646, 435.60784253, 435.60784253, 431.99783279, 441.62452544, 428.38782305, 422.37114014, 421.16780356, 425.98114988, 409.13443774, 439.21785228, 417.55779381, 417.55779381, 416.35445723, 406.72776458, 395.89773534, 371.83100371, 375.44101346, 374.23767687, 371.83100371, 376.64435004, 380.25435978, 376.64435004, 375.44101346, 369.42433055, 354.98429157, 354.98429157, 353.78095499, 356.18762815, 340.54425259, 352.57761841, 322.49420387, 312.86751121, 304.44415514, 305.64749172, 296.02079907, 304.44415514, 322.49420387, 294.81746249, 309.25750147, 293.61412591, 294.81746249, 269.54739428, 259.92070162, 246.68399923, 233.44729683, 247.88733581, 227.43061392, 258.71736504, 250.29400897, 265.93738453, 252.70068213, 262.32737479, 255.10735530, 287.59744300, 348.96760866, 353.78095499, 342.95092575, 341.74758917, 315.27418438, 321.29086728, 324.90087703, 318.88419412, 306.85082830, 293.61412591, 302.03748198, 310.46083805, 324.90087703, 314.07084779, 328.51088677, 338.13757943, 334.52756968, 341.74758917, 354.98429157, 379.05102320, 369.42433055, 358.59430131, 373.03434029, 444.03119861, 442.82786202, 397.10107192, 393.49106218, 397.10107192, 393.49106218, 381.45769636, 383.86436953, 428.38782305, 430.79449621, 448.84454493, 427.18448646, 441.62452544, 427.18448646, 431.99783279, 428.38782305, 431.99783279, 428.38782305, 413.94778407, 424.77781330, 435.60784253, 433.20116937, 422.37114014, 409.13443774, 406.72776458, 395.89773534, 380.25435978, 379.05102320, 377.84768662, 387.47437927, 376.64435004, 373.03434029, 380.25435978, 381.45769636, 385.06770611, 380.25435978, 370.62766713, 383.86436953, 364.61098422, 363.40764764, 385.06770611, 370.62766713, 373.03434029, 361.00097448, 338.13757943, 347.76427208, 324.90087703, 310.46083805, 317.68085754, 262.32737479, 287.59744300, 271.95406744, 277.97075035, 268.34405769, 288.80077958, 296.02079907, 305.64749172, 287.59744300, 270.75073086, 285.19076984, 279.17408693, 268.34405769, 251.49734555, 269.54739428, 270.75073086, 283.98743325, 280.37742351, 299.63080882, 286.39410642, 287.59744300, 285.19076984, 281.58076009, 274.36074060, 270.75073086, 264.73404795, 246.68399923, 246.68399923, 244.27732606, 223.82060418, 215.39724810, 209.38056520, 200.95720913, 200.95720913, 199.75387254, 187.72050673, 167.26378484], 
							[825.48889497, 829.09890471, 807.43884625, 776.15209512, 764.11872931, 631.75170533, 669.05513936, 661.83511987, 589.63492498, 636.56505166, 661.83511987, 642.58173457, 626.93835901, 619.71833952, 598.05828105, 577.60155916, 551.12815437, 505.40136427, 542.70479830, 523.45141299, 529.46809590, 476.52128631, 472.91127656, 403.11775483, 368.22099397, 511.41804717, 583.61824207, 659.42844671, 613.70165661, 599.26161763, 654.61510038, 669.05513936, 651.00509064, 654.61510038, 660.63178329, 631.75170533, 625.73502243, 571.58487625, 617.31166635, 635.36171508, 625.73502243, 614.90499319, 632.95504192, 604.07496396, 583.61824207, 571.58487625, 548.72148120, 552.33149095, 586.02491523, 589.63492498, 554.73816411, 569.17820309, 564.36485676, 541.50146171, 583.61824207, 619.71833952, 606.48163712, 625.73502243, 691.91853441, 677.47849543, 655.81843697, 691.91853441, 703.95190023, 684.69851492, 689.51186125, 707.56190997, 723.20528553, 701.54522707, 689.51186125, 646.19174431, 612.49832003, 589.63492498, 577.60155916, 523.45141299, 559.55151044, 548.72148120, 587.22825181, 566.77152993, 563.16152018, 567.97486651, 566.77152993, 564.36485676, 566.77152993, 553.53482753, 558.34817386, 565.56819335, 566.77152993, 545.11147146, 500.58801794, 498.18134478, 488.55465212, 469.30126682, 490.96132529, 482.53796922, 511.41804717, 463.28458391, 471.70793998, 502.99469110, 551.12815437, 559.55151044, 549.92481779, 637.76838824, 613.70165661, 557.14483728, 583.61824207, 600.46495421, 593.24493472, 590.83826156, 571.58487625, 583.61824207, 586.02491523, 589.63492498, 572.78821284, 561.95818360, 583.61824207, 590.83826156, 589.63492498, 598.05828105, 593.24493472, 601.66829079, 623.32834926, 604.07496396, 602.87162738, 583.61824207, 593.24493472, 571.58487625, 563.16152018, 560.75484702, 583.61824207, 561.95818360, 566.77152993, 580.00823233, 546.31480804, 543.90813488, 541.50146171, 523.45141299, 527.06142274, 527.06142274, 517.43473008, 529.46809590, 517.43473008, 477.72462289, 498.18134478, 483.74130580, 484.94464238, 486.14797896, 490.96132529, 499.38468136, 516.23139350, 501.79135452, 489.75798871, 539.09478855, 608.88831028, 672.66514911, 655.81843697, 610.09164687, 577.60155916, 548.72148120, 529.46809590, 505.40136427, 488.55465212, 519.84140325, 507.80803743, 512.62138376, 512.62138376, 523.45141299, 534.28144222, 507.80803743, 510.21471059, 463.28458391, 484.94464238, 457.26790100, 423.57447672, 460.87791075, 424.77781330, 502.99469110, 546.31480804, 511.41804717, 482.53796922, 493.36799845, 507.80803743, 546.31480804, 534.28144222, 528.26475932, 551.12815437, 557.14483728, 517.43473008, 516.23139350, 540.29812513, 510.21471059, 499.38468136, 481.33463264, 490.96132529, 490.96132529, 499.38468136, 489.75798871, 496.97800820, 506.60470085, 510.21471059, 505.40136427, 529.46809590, 564.36485676, 530.67143248, 535.48477881, 540.29812513, 547.51814462, 531.87476906, 537.89145197, 511.41804717, 506.60470085, 500.58801794, 511.41804717, 505.40136427, 549.92481779, 549.92481779, 555.94150069, 576.39822258, 576.39822258, 570.38153967, 570.38153967, 588.43158840, 589.63492498, 595.65160789, 580.00823233, 589.63492498, 583.61824207, 576.39822258, 561.95818360, 534.28144222, 542.70479830, 525.85808615, 517.43473008, 506.60470085, 507.80803743, 521.04473983, 495.77467161, 488.55465212, 518.63806666, 525.85808615, 531.87476906, 530.67143248, 519.84140325, 539.09478855, 494.57133503, 492.16466187, 502.99469110, 492.16466187, 490.96132529, 454.86122784, 421.16780356, 436.81117912, 404.32109141, 386.27104269, 386.27104269, 317.68085754, 359.79763789, 354.98429157, 376.64435004, 342.95092575, 318.88419412, 334.52756968, 324.90087703, 308.05416489, 299.63080882, 291.20745274, 274.36074060, 246.68399923, 227.43061392, 261.12403820, 246.68399923, 256.31069188, 262.32737479, 270.75073086, 267.14072111, 264.73404795, 286.39410642, 290.00411616, 296.02079907, 283.98743325, 271.95406744, 264.73404795, 233.44729683, 258.71736504, 222.61726759, 216.60058469, 198.55053596, 186.51717015, 187.72050673, 190.12717989, 174.48380433, 170.87379459], 
							[894.07908012, 848.35229002, 836.31892420, 843.53894369, 824.28555839, 782.16877803, 753.28870007, 748.47535375, 766.52540247, 752.08536349, 764.11872931, 790.59213410, 752.08536349, 734.03531477, 709.96858314, 697.93521732, 673.86848569, 637.76838824, 655.81843697, 658.22511013, 616.10832977, 610.09164687, 616.10832977, 596.85494447, 535.48477881, 635.36171508, 790.59213410, 765.32206589, 731.62864161, 730.42530502, 734.03531477, 693.12187099, 695.52854416, 670.25847594, 652.20842722, 647.39508089, 614.90499319, 611.29498345, 619.71833952, 637.76838824, 637.76838824, 634.15837850, 661.83511987, 631.75170533, 613.70165661, 608.88831028, 610.09164687, 593.24493472, 636.56505166, 613.70165661, 598.05828105, 606.48163712, 573.99154942, 577.60155916, 612.49832003, 685.90185151, 673.86848569, 766.52540247, 830.30224130, 818.26887548, 782.16877803, 824.28555839, 863.99566558, 819.47221206, 877.23236798, 855.57230951, 851.96229976, 854.36897293, 802.62549992, 802.62549992, 776.15209512, 794.20214385, 790.59213410, 735.23865135, 794.20214385, 800.21882676, 796.60881701, 802.62549992, 797.81215359, 803.82883650, 800.21882676, 786.98212436, 794.20214385, 778.55876829, 755.69537324, 812.25219257, 788.18546094, 779.76210487, 749.67869033, 760.50871956, 747.27201717, 678.68183202, 740.05199768, 740.05199768, 759.30538298, 707.56190997, 702.74856365, 726.81529528, 786.98212436, 782.16877803, 794.20214385, 866.40233874, 837.52226079, 790.59213410, 815.86220232, 830.30224130, 821.87888522, 788.18546094, 784.57545120, 797.81215359, 776.15209512, 820.67554864, 832.70891446, 830.30224130, 819.47221206, 842.33560711, 847.14895344, 851.96229976, 870.01234849, 877.23236798, 872.41902165, 835.11558762, 835.11558762, 824.28555839, 842.33560711, 836.31892420, 812.25219257, 826.69223155, 830.30224130, 849.55562660, 868.80901191, 912.12912884, 874.82569481, 874.82569481, 856.77564609, 844.74228027, 848.35229002, 843.53894369, 844.74228027, 829.09890471, 824.28555839, 801.42216334, 813.45552915, 752.08536349, 770.13541222, 791.79547068, 770.13541222, 779.76210487, 768.93207563, 774.94875854, 767.72873905, 866.40233874, 897.68908986, 904.90910935, 902.50243619, 873.62235823, 866.40233874, 870.01234849, 866.40233874, 818.26887548, 788.18546094, 774.94875854, 791.79547068, 821.87888522, 860.38565584, 847.14895344, 844.74228027, 837.52226079, 785.77878778, 767.72873905, 812.25219257, 750.88202691, 744.86534400, 770.13541222, 784.57545120, 827.89556813, 848.35229002, 829.09890471, 836.31892420, 824.28555839, 877.23236798, 926.56916782, 914.53580201, 920.55248491, 963.87260185, 1026.44610409, 939.80587022, 938.60253364, 999.97269930, 936.19586048, 954.24590920, 896.48575328, 922.95915808, 907.31578252, 938.60253364, 938.60253364, 974.70263109, 980.71931399, 971.09262134, 992.75267981, 986.73599690, 1058.93619180, 1010.80272853, 1033.66612358, 1022.83609435, 1045.69948940, 1054.12284547, 1038.47946991, 1021.63275777, 1022.83609435, 1005.98938221, 1036.07279675, 1046.90282598, 1104.66298190, 1127.52637695, 1144.37308909, 1141.96641593, 1185.28653286, 1137.15306960, 1141.96641593, 1191.30321577, 1164.82981098, 1196.11656210, 1175.65984021, 1227.40331322, 1251.47004485, 1192.50655235, 1188.89654261, 1149.18643542, 1135.94973302, 1109.47632822, 1119.10302088, 1095.03628924, 1101.05297215, 1116.69634771, 1085.40959659, 1083.00292343, 1064.95287471, 1076.98624052, 1101.05297215, 1107.06965506, 1111.88300139, 1135.94973302, 1083.00292343, 1070.96955761, 1086.61293317, 1060.13952838, 1086.61293317, 1028.85277726, 944.61921655, 896.48575328, 933.78918731, 944.61921655, 941.00920680, 807.43884625, 856.77564609, 813.45552915, 824.28555839, 836.31892420, 825.48889497, 956.65258236, 896.48575328, 856.77564609, 839.92893395, 903.70577277, 879.63904114, 837.52226079, 760.50871956, 746.06868058, 753.28870007, 797.81215359, 842.33560711, 859.18231925, 882.04571430, 867.60567532, 826.69223155, 818.26887548, 870.01234849, 892.87574354, 860.38565584, 860.38565584, 854.36897293, 902.50243619, 809.84551941, 800.21882676, 867.60567532, 934.99252389, 876.02903140, 888.06239721, 890.46907037, 888.06239721], 
							[1107.06965506, 1191.30321577, 1179.26984996, 1111.88300139, 1012.00606512, 938.60253364, 914.53580201, 885.65572405, 950.63589945, 981.92265058, 1070.96955761, 1016.81941144, 1002.37937246, 1097.44296241, 1116.69634771, 1194.91322552, 1239.43667904, 1095.03628924, 1146.77976225, 1167.23648414, 1137.15306960, 1109.47632822, 1052.91950889, 920.55248491, 926.56916782, 1093.83295266, 1147.98309883, 1359.77033719, 1299.60350811, 1287.57014230, 1227.40331322, 1188.89654261, 1197.31989868, 1167.23648414, 1166.03314756, 1114.28967455, 1096.23962583, 1070.96955761, 1108.27299164, 1127.52637695, 1132.33972327, 1116.69634771, 1113.08633797, 1080.59625027, 1070.96955761, 1043.29281624, 1016.81941144, 993.95601639, 1014.41273828, 1032.46278700, 1031.25945042, 1012.00606512, 971.09262134, 1015.61607486, 1058.93619180, 1058.93619180, 1158.81312807, 1311.63687393, 1335.70360556, 1275.53677648, 1198.52323526, 1275.53677648, 1275.53677648, 1311.63687393, 1323.67023975, 1287.57014230, 1287.57014230, 1239.43667904, 1178.06651337, 1138.35640618, 1139.55974276, 1239.43667904, 1227.40331322, 1169.64315730, 1179.26984996, 1185.28653286, 1203.33658159, 1185.28653286, 1190.09987919, 1196.11656210, 1182.87985970, 1163.62647440, 1123.91636720, 1127.52637695, 1119.10302088, 1096.23962583, 1085.40959659, 1046.90282598, 949.43256287, 986.73599690, 908.51911910, 831.50557788, 880.84237772, 842.33560711, 872.41902165, 830.30224130, 837.52226079, 900.09576303, 931.38251415, 959.05925553, 920.55248491, 1003.58270904, 1008.39605537, 944.61921655, 962.66926527, 974.70263109, 977.10930425, 1013.20940170, 973.49929450, 942.21254338, 957.85591894, 956.65258236, 951.83923604, 962.66926527, 962.66926527, 955.44924578, 956.65258236, 974.70263109, 968.68594818, 998.76936272, 990.34600665, 950.63589945, 943.41587996, 957.85591894, 981.92265058, 951.83923604, 942.21254338, 954.24590920, 954.24590920, 944.61921655, 931.38251415, 920.55248491, 913.33246543, 914.53580201, 900.09576303, 860.38565584, 857.97898267, 859.18231925, 863.99566558, 855.57230951, 866.40233874, 853.16563635, 857.97898267, 827.89556813, 850.75896318, 860.38565584, 842.33560711, 877.23236798, 883.24905089, 870.01234849, 861.58899242, 861.58899242, 784.57545120, 800.21882676, 803.82883650, 825.48889497, 806.23550966, 770.13541222, 723.20528553, 741.25533426, 734.03531477, 726.81529528, 685.90185151, 697.93521732, 738.84866109, 732.83197819, 744.86534400, 738.84866109, 707.56190997, 673.86848569, 696.73188074, 629.34503217, 608.88831028, 637.76838824, 573.99154942, 667.85180278, 672.66514911, 685.90185151, 695.52854416, 682.29184176, 673.86848569, 744.86534400, 728.01863186, 722.00194895, 736.44198793, 730.42530502, 693.12187099, 675.07182227, 720.79861237, 688.30852467, 696.73188074, 651.00509064, 676.27515885, 677.47849543, 701.54522707, 697.93521732, 709.96858314, 715.98526604, 741.25533426, 728.01863186, 743.66200742, 755.69537324, 740.05199768, 744.86534400, 772.54208538, 783.37211461, 807.43884625, 806.23550966, 792.99880727, 797.81215359, 805.03217308, 788.18546094, 805.03217308, 809.84551941, 821.87888522, 830.30224130, 841.13227053, 832.70891446, 794.20214385, 782.16877803, 830.30224130, 820.67554864, 811.04885599, 814.65886573, 837.52226079, 838.72559737, 845.94561686, 827.89556813, 806.23550966, 835.11558762, 806.23550966, 806.23550966, 788.18546094, 800.21882676, 813.45552915, 765.32206589, 770.13541222, 817.06553890, 815.86220232, 836.31892420, 818.26887548, 806.23550966, 821.87888522, 808.64218283, 785.77878778, 829.09890471, 824.28555839, 819.47221206, 789.38879752, 746.06868058, 762.91539273, 724.40862212, 715.98526604, 709.96858314, 590.83826156, 654.61510038, 622.12501268, 642.58173457, 632.95504192, 638.97172482, 669.05513936, 647.39508089, 632.95504192, 625.73502243, 682.29184176, 618.51500294, 604.07496396, 616.10832977, 618.51500294, 605.27830054, 637.76838824, 651.00509064, 685.90185151, 681.08850518, 669.05513936, 675.07182227, 665.44512962, 653.41176380, 631.75170533, 625.73502243, 629.34503217, 648.59841748, 654.61510038, 590.83826156, 581.21156891, 553.53482753, 535.48477881, 547.51814462, 563.16152018, 541.50146171, 512.62138376], 
							[819.47221206, 741.25533426, 699.13855390, 765.32206589, 813.45552915, 765.32206589, 770.13541222, 788.18546094, 779.76210487, 741.25533426, 743.66200742, 758.10204640, 770.13541222, 767.72873905, 750.88202691, 711.17191972, 649.80175406, 679.88516860, 735.23865135, 731.62864161, 723.20528553, 685.90185151, 688.30852467, 715.98526604, 697.93521732, 746.06868058, 796.60881701, 803.82883650, 803.82883650, 803.82883650, 807.43884625, 784.57545120, 803.82883650, 794.20214385, 786.98212436, 774.94875854, 752.08536349, 758.10204640, 786.98212436, 788.18546094, 821.87888522, 824.28555839, 838.72559737, 861.58899242, 849.55562660, 830.30224130, 812.25219257, 808.64218283, 843.53894369, 865.19900216, 805.03217308, 797.81215359, 764.11872931, 764.11872931, 826.69223155, 806.23550966, 844.74228027, 908.51911910, 949.43256287, 1044.49615282, 1008.39605537, 1081.79958685, 1068.56288445, 1034.86946017, 1073.37623078, 1076.98624052, 1046.90282598, 1057.73285522, 1015.61607486, 971.09262134, 981.92265058, 987.93933348, 1007.19271879, 957.85591894, 986.73599690, 965.07593843, 993.95601639, 974.70263109, 953.04257262, 973.49929450, 955.44924578, 956.65258236, 925.36583124, 895.28241670, 978.31264083, 950.63589945, 980.71931399, 971.09262134, 896.48575328, 934.99252389, 902.50243619, 863.99566558, 902.50243619, 903.70577277, 954.24590920, 891.67240696, 901.29909961, 950.63589945, 968.68594818, 986.73599690, 985.53266032, 1069.76622103, 1028.85277726, 985.53266032, 1046.90282598, 1049.30949914, 1092.62961608, 1126.32304037, 1070.96955761, 1062.54620154, 1119.10302088, 1126.32304037, 1103.45964532, 1092.62961608, 1093.83295266, 1109.47632822, 1146.77976225, 1166.03314756, 1155.20311832, 1194.91322552, 1167.23648414, 1161.21980123, 1133.54305986, 1111.88300139, 1126.32304037, 1129.93305011, 1152.79644516, 1143.16975251, 1128.72971353, 1141.96641593, 1107.06965506, 1095.03628924, 1087.81626976, 1085.40959659, 1085.40959659, 1085.40959659, 1095.03628924, 1090.22294292, 1081.79958685, 1099.84963557, 1097.44296241, 1107.06965506, 1104.66298190, 1087.81626976, 1104.66298190, 1083.00292343, 1089.01960634, 1089.01960634, 1083.00292343, 1057.73285522, 1052.91950889, 1046.90282598, 1022.83609435, 1040.88614307, 955.44924578, 922.95915808, 886.85906063, 842.33560711, 800.21882676, 755.69537324, 696.73188074, 752.08536349, 743.66200742, 768.93207563, 805.03217308, 803.82883650, 808.64218283, 776.15209512, 778.55876829, 792.99880727, 806.23550966, 752.08536349, 746.06868058, 791.79547068, 778.55876829, 830.30224130, 886.85906063, 883.24905089, 843.53894369, 890.46907037, 874.82569481, 986.73599690, 953.04257262, 971.09262134, 1043.29281624, 1016.81941144, 902.50243619, 957.85591894, 977.10930425, 932.58585073, 950.63589945, 928.97584099, 950.63589945, 937.39919706, 925.36583124, 965.07593843, 998.76936272, 979.51597741, 974.70263109, 990.34600665, 1022.83609435, 1069.76622103, 1030.05611384, 1038.47946991, 1083.00292343, 1098.64629899, 1095.03628924, 1111.88300139, 1083.00292343, 1089.01960634, 1099.84963557, 1086.61293317, 1128.72971353, 1146.77976225, 1182.87985970, 1203.33658159, 1275.53677648, 1251.47004485, 1203.33658159, 1251.47004485, 1263.50341067, 1239.43667904, 1239.43667904, 1275.53677648, 1275.53677648, 1287.57014230, 1299.60350811, 1287.57014230, 1227.40331322, 1196.11656210, 1185.28653286, 1203.33658159, 1120.30635746, 1140.76307935, 1158.81312807, 1117.89968430, 1076.98624052, 1089.01960634, 1143.16975251, 1149.18643542, 1156.40645491, 1155.20311832, 1194.91322552, 1126.32304037, 1139.55974276, 1184.08319628, 1161.21980123, 1167.23648414, 1153.99978174, 1135.94973302, 1160.01646465, 1104.66298190, 1144.37308909, 1143.16975251, 950.63589945, 993.95601639, 1022.83609435, 1028.85277726, 995.15935297, 1042.08947966, 1084.20626001, 1083.00292343, 1015.61607486, 1010.80272853, 1068.56288445, 1089.01960634, 1046.90282598, 1052.91950889, 1060.13952838, 1102.25630873, 1143.16975251, 1107.06965506, 1064.95287471, 1055.32618205, 1058.93619180, 1117.89968430, 1091.42627950, 1080.59625027, 1095.03628924, 1076.98624052, 1049.30949914, 1007.19271879, 1008.39605537, 938.60253364, 975.90596767, 972.29595792, 985.53266032, 898.89242645, 896.48575328, 836.31892420, 833.91225104], 
							[671.46181253, 616.10832977, 602.87162738, 613.70165661, 587.22825181, 516.23139350, 528.26475932, 557.14483728, 546.31480804, 553.53482753, 595.65160789, 580.00823233, 604.07496396, 606.48163712, 590.83826156, 577.60155916, 545.11147146, 552.33149095, 598.05828105, 589.63492498, 587.22825181, 569.17820309, 577.60155916, 558.34817386, 535.48477881, 593.24493472, 655.81843697, 643.78507115, 635.36171508, 624.53168584, 631.75170533, 583.61824207, 592.04159814, 604.07496396, 589.63492498, 584.82157865, 593.24493472, 584.82157865, 593.24493472, 611.29498345, 602.87162738, 602.87162738, 610.09164687, 604.07496396, 592.04159814, 584.82157865, 573.99154942, 565.56819335, 601.66829079, 587.22825181, 566.77152993, 570.38153967, 563.16152018, 540.29812513, 567.97486651, 624.53168584, 655.81843697, 770.13541222, 752.08536349, 755.69537324, 712.37525630, 778.55876829, 788.18546094, 758.10204640, 758.10204640, 754.49203666, 746.06868058, 744.86534400, 708.76524656, 700.34189048, 722.00194895, 725.61195870, 750.88202691, 728.01863186, 742.45867084, 734.03531477, 770.13541222, 770.13541222, 771.33874880, 782.16877803, 755.69537324, 741.25533426, 736.44198793, 729.22196844, 753.28870007, 750.88202691, 750.88202691, 734.03531477, 706.35857339, 690.71519783, 690.71519783, 676.27515885, 691.91853441, 661.83511987, 683.49517834, 665.44512962, 643.78507115, 682.29184176, 729.22196844, 734.03531477, 725.61195870, 748.47535375, 740.05199768, 722.00194895, 718.39193921, 690.71519783, 709.96858314, 699.13855390, 699.13855390, 681.08850518, 709.96858314, 737.64532451, 722.00194895, 722.00194895, 723.20528553, 741.25533426, 740.05199768, 770.13541222, 765.32206589, 792.99880727, 796.60881701, 776.15209512, 768.93207563, 759.30538298, 768.93207563, 776.15209512, 766.52540247, 778.55876829, 794.20214385, 785.77878778, 790.59213410, 796.60881701, 768.93207563, 753.28870007, 761.71205615, 762.91539273, 768.93207563, 788.18546094, 779.76210487, 778.55876829, 792.99880727, 776.15209512, 786.98212436, 768.93207563, 778.55876829, 767.72873905, 772.54208538, 776.15209512, 786.98212436, 780.96544145, 780.96544145, 747.27201717, 689.51186125, 682.29184176, 703.95190023, 703.95190023, 675.07182227, 653.41176380, 635.36171508, 655.81843697, 613.70165661, 640.17506140, 616.10832977, 629.34503217, 660.63178329, 652.20842722, 649.80175406, 613.70165661, 618.51500294, 600.46495421, 618.51500294, 583.61824207, 583.61824207, 588.43158840, 571.58487625, 619.71833952, 649.80175406, 631.75170533, 640.17506140, 671.46181253, 683.49517834, 729.22196844, 712.37525630, 715.98526604, 676.27515885, 712.37525630, 685.90185151, 706.35857339, 713.57859288, 679.88516860, 667.85180278, 632.95504192, 667.85180278, 661.83511987, 667.85180278, 699.13855390, 711.17191972, 713.57859288, 685.90185151, 711.17191972, 720.79861237, 741.25533426, 731.62864161, 731.62864161, 715.98526604, 744.86534400, 742.45867084, 737.64532451, 726.81529528, 717.18860263, 713.57859288, 693.12187099, 713.57859288, 732.83197819, 771.33874880, 772.54208538, 762.91539273, 779.76210487, 767.72873905, 741.25533426, 752.08536349, 743.66200742, 762.91539273, 752.08536349, 773.74542196, 770.13541222, 762.91539273, 749.67869033, 738.84866109, 720.79861237, 723.20528553, 722.00194895, 707.56190997, 712.37525630, 715.98526604, 690.71519783, 688.30852467, 689.51186125, 688.30852467, 694.32520758, 707.56190997, 688.30852467, 708.76524656, 688.30852467, 675.07182227, 690.71519783, 683.49517834, 684.69851492, 655.81843697, 631.75170533, 630.54836875, 628.14169559, 623.32834926, 630.54836875, 565.56819335, 605.27830054, 588.43158840, 571.58487625, 558.34817386, 560.75484702, 593.24493472, 565.56819335, 525.85808615, 484.94464238, 534.28144222, 528.26475932, 521.04473983, 494.57133503, 529.46809590, 541.50146171, 567.97486651, 564.36485676, 607.68497370, 611.29498345, 577.60155916, 578.80489574, 565.56819335, 566.77152993, 541.50146171, 511.41804717, 494.57133503, 502.99469110, 511.41804717, 481.33463264, 493.36799845, 490.96132529, 502.99469110, 523.45141299, 494.57133503, 501.79135452, 472.91127656], 
							[3477.64272079, 3465.60935497, 3345.27569681, 3333.24233100, 3285.10886773, 3188.84194121, 3056.47491723, 3032.40818560, 3104.60838050, 3068.50828305, 3200.87530702, 3044.44155142, 3128.67511213, 3140.70847794, 3080.54164886, 3008.34145397, 3080.54164886, 2912.07452744, 2984.27472234, 2960.20799071, 2936.14125907, 2815.80760092, 2767.67413765, 2827.84096673, 2743.60740602, 3020.37481979, 3068.50828305, 3224.94203866, 3224.94203866, 2984.27472234, 3092.57501468, 2996.30808815, 2984.27472234, 3080.54164886, 3068.50828305, 2936.14125907, 2936.14125907, 2936.14125907, 3080.54164886, 3080.54164886, 3128.67511213, 3056.47491723, 3116.64174631, 3164.77520958, 3068.50828305, 3008.34145397, 2936.14125907, 2948.17462489, 3008.34145397, 2960.20799071, 2912.07452744, 2839.87433255, 2863.94106418, 2779.70750347, 2791.74086928, 2731.57404020, 2719.54067439, 2803.77423510, 3044.44155142, 3032.40818560, 2948.17462489, 2984.27472234, 2960.20799071, 2900.04116163, 3056.47491723, 2900.04116163, 2839.87433255, 2803.77423510, 2779.70750347, 2731.57404020, 2719.54067439, 2743.60740602, 2815.80760092, 2707.50730857, 2755.64077184, 2839.87433255, 2803.77423510, 2851.90769836, 2912.07452744, 2936.14125907, 2972.24135652, 2936.14125907, 2996.30808815, 2948.17462489, 2900.04116163, 2851.90769836, 2827.84096673, 2815.80760092, 2707.50730857, 2731.57404020, 2683.44057694, 2563.10691878, 2551.07355297, 2502.94008970, 2623.27374786, 2466.83999225, 2418.70652899, 2478.87335807, 2599.20701623, 2599.20701623, 2599.20701623, 2623.27374786, 2575.14028460, 2527.00682133, 2418.70652899, 2502.94008970, 2454.80662644, 2454.80662644, 2442.77326062, 2406.67316318, 2382.60643154, 2334.47296828, 2322.43960246, 2322.43960246, 2322.43960246, 2346.50633410, 2334.47296828, 2382.60643154, 2358.53969991, 2322.43960246, 2310.40623665, 2274.30613920, 2262.27277339, 2310.40623665, 2334.47296828, 2310.40623665, 2334.47296828, 2310.40623665, 2334.47296828, 2322.43960246, 2286.33950502, 2310.40623665, 2298.37287083, 2310.40623665, 2346.50633410, 2274.30613920, 2286.33950502, 2250.23940757, 2274.30613920, 2286.33950502, 2286.33950502, 2226.17267594, 2286.33950502, 2274.30613920, 2418.70652899, 2406.67316318, 2442.77326062, 2466.83999225, 2466.83999225, 2478.87335807, 2478.87335807, 2418.70652899, 2430.73989481, 2382.60643154, 2394.63979736, 2406.67316318, 2334.47296828, 2298.37287083, 2298.37287083, 2406.67316318, 2298.37287083, 2286.33950502, 2250.23940757, 2262.27277339, 2322.43960246, 2274.30613920, 2310.40623665, 2238.20604175, 2214.13931012, 2274.30613920, 2298.37287083, 2274.30613920, 2286.33950502, 2310.40623665, 2298.37287083, 2382.60643154, 2322.43960246, 2382.60643154, 2358.53969991, 2406.67316318, 2406.67316318, 2406.67316318, 2430.73989481, 2478.87335807, 2527.00682133, 2527.00682133, 2502.94008970, 2575.14028460, 2647.34047949, 2671.40721112, 2647.34047949, 2671.40721112, 2719.54067439, 2755.64077184, 2671.40721112, 2575.14028460, 2647.34047949, 2695.47394276, 2683.44057694, 2779.70750347, 2671.40721112, 2803.77423510, 2791.74086928, 2767.67413765, 2743.60740602, 2888.00779581, 2863.94106418, 2839.87433255, 2839.87433255, 2839.87433255, 2827.84096673, 2767.67413765, 2803.77423510, 2863.94106418, 2948.17462489, 2924.10789326, 2936.14125907, 2984.27472234, 2936.14125907, 2960.20799071, 2984.27472234, 2960.20799071, 2803.77423510, 2683.44057694, 2875.97442999, 2875.97442999, 2863.94106418, 2755.64077184, 2767.67413765, 2779.70750347, 2815.80760092, 2791.74086928, 2707.50730857, 2683.44057694, 2695.47394276, 2695.47394276, 2695.47394276, 2791.74086928, 2827.84096673, 2839.87433255, 2779.70750347, 2791.74086928, 2827.84096673, 2803.77423510, 2803.77423510, 2851.90769836, 2900.04116163, 2851.90769836, 2791.74086928, 2815.80760092, 2839.87433255, 2875.97442999, 2827.84096673, 2863.94106418, 2683.44057694, 2707.50730857, 2671.40721112, 2731.57404020, 2683.44057694, 2996.30808815, 3008.34145397, 3032.40818560, 2936.14125907, 2839.87433255, 3044.44155142, 3080.54164886, 3044.44155142, 3092.57501468, 3176.80857539, 3200.87530702, 3345.27569681, 3489.67608660, 3429.50925753, 3393.40916008, 3453.57598916, 3513.74281824, 3646.10984221, 3754.41013455, 3754.41013455, 3706.27667129, 3610.00974476, 3706.27667129, 3850.67706108, 3754.41013455, 3597.97637895, 3694.24330547, 3477.64272079, 3465.60935497, 3622.04311058, 3585.94301313, 3525.77618405], 
							[489.75798871, 487.35131554, 492.16466187, 490.96132529, 465.69125707, 471.70793998, 451.25121810, 433.20116937, 433.20116937, 436.81117912, 445.23453519, 451.25121810, 464.48792049, 464.48792049, 436.81117912, 434.40450595, 403.11775483, 387.47437927, 430.79449621, 429.59115963, 411.54111090, 393.49106218, 403.11775483, 388.67771585, 352.57761841, 397.10107192, 457.26790100, 456.06456442, 438.01451570, 424.77781330, 446.43787177, 409.13443774, 413.94778407, 421.16780356, 421.16780356, 419.96446697, 412.74444748, 398.30440851, 440.42118886, 435.60784253, 421.16780356, 436.81117912, 435.60784253, 419.96446697, 405.52442800, 418.76113039, 403.11775483, 388.67771585, 411.54111090, 411.54111090, 409.13443774, 415.15112065, 403.11775483, 398.30440851, 447.64120835, 453.65789126, 469.30126682, 476.52128631, 522.24807641, 507.80803743, 481.33463264, 521.04473983, 524.65474957, 506.60470085, 524.65474957, 521.04473983, 517.43473008, 506.60470085, 498.18134478, 501.79135452, 489.75798871, 509.01137401, 521.04473983, 500.58801794, 518.63806666, 498.18134478, 513.82472034, 507.80803743, 502.99469110, 500.58801794, 487.35131554, 487.35131554, 486.14797896, 469.30126682, 464.48792049, 477.72462289, 498.18134478, 489.75798871, 427.18448646, 430.79449621, 410.33777432, 376.64435004, 407.93110116, 397.10107192, 394.69439876, 371.83100371, 380.25435978, 383.86436953, 417.55779381, 416.35445723, 424.77781330, 445.23453519, 436.81117912, 422.37114014, 409.13443774, 418.76113039, 434.40450595, 450.04788151, 440.42118886, 438.01451570, 482.53796922, 471.70793998, 470.50460340, 463.28458391, 463.28458391, 469.30126682, 459.67457417, 492.16466187, 507.80803743, 540.29812513, 533.07810564, 506.60470085, 490.96132529, 488.55465212, 507.80803743, 490.96132529, 493.36799845, 509.01137401, 524.65474957, 516.23139350, 511.41804717, 512.62138376, 493.36799845, 495.77467161, 501.79135452, 505.40136427, 487.35131554, 496.97800820, 510.21471059, 512.62138376, 511.41804717, 504.19802769, 490.96132529, 463.28458391, 477.72462289, 469.30126682, 476.52128631, 496.97800820, 500.58801794, 498.18134478, 483.74130580, 469.30126682, 431.99783279, 439.21785228, 444.03119861, 429.59115963, 421.16780356, 418.76113039, 392.28772560, 406.72776458, 379.05102320, 397.10107192, 403.11775483, 411.54111090, 440.42118886, 431.99783279, 440.42118886, 412.74444748, 404.32109141, 397.10107192, 403.11775483, 397.10107192, 388.67771585, 381.45769636, 375.44101346, 403.11775483, 433.20116937, 415.15112065, 425.98114988, 424.77781330, 424.77781330, 436.81117912, 441.62452544, 444.03119861, 441.62452544, 468.09793024, 435.60784253, 448.84454493, 465.69125707, 452.45455468, 444.03119861, 446.43787177, 464.48792049, 451.25121810, 451.25121810, 465.69125707, 462.08124733, 477.72462289, 470.50460340, 480.13129605, 499.38468136, 515.02805692, 493.36799845, 504.19802769, 506.60470085, 541.50146171, 540.29812513, 528.26475932, 516.23139350, 506.60470085, 509.01137401, 510.21471059, 541.50146171, 542.70479830, 547.51814462, 571.58487625, 548.72148120, 547.51814462, 548.72148120, 535.48477881, 536.68811539, 534.28144222, 540.29812513, 518.63806666, 529.46809590, 531.87476906, 521.04473983, 521.04473983, 517.43473008, 513.82472034, 501.79135452, 501.79135452, 494.57133503, 505.40136427, 510.21471059, 472.91127656, 472.91127656, 478.92795947, 483.74130580, 500.58801794, 486.14797896, 475.31794973, 478.92795947, 470.50460340, 483.74130580, 498.18134478, 488.55465212, 489.75798871, 465.69125707, 433.20116937, 433.20116937, 434.40450595, 418.76113039, 424.77781330, 370.62766713, 386.27104269, 391.08438902, 395.89773534, 376.64435004, 385.06770611, 403.11775483, 389.88105243, 376.64435004, 385.06770611, 403.11775483, 405.52442800, 389.88105243, 364.61098422, 413.94778407, 421.16780356, 427.18448646, 403.11775483, 398.30440851, 374.23767687, 382.66103294, 389.88105243, 377.84768662, 367.01765738, 374.23767687, 369.42433055, 364.61098422, 357.39096473, 358.59430131, 315.27418438, 310.46083805, 309.25750147, 275.56407718, 277.97075035, 269.54739428, 247.88733581, 206.97389203], 
							[969.88928476, 938.60253364, 918.14581175, 914.53580201, 913.33246543, 898.89242645, 974.70263109, 1013.20940170, 1060.13952838, 1044.49615282, 1080.59625027, 998.76936272, 1010.80272853, 1016.81941144, 1007.19271879, 1022.83609435, 1012.00606512, 1004.78604563, 1072.17289419, 1061.34286496, 1064.95287471, 983.12598716, 996.36268955, 968.68594818, 944.61921655, 989.14267007, 1040.88614307, 1076.98624052, 1107.06965506, 1095.03628924, 1107.06965506, 1040.88614307, 1034.86946017, 1066.15621129, 1028.85277726, 962.66926527, 995.15935297, 992.75267981, 1055.32618205, 1089.01960634, 1091.42627950, 1104.66298190, 1110.67966481, 1117.89968430, 1114.28967455, 1117.89968430, 1109.47632822, 1127.52637695, 1149.18643542, 1145.57642567, 1147.98309883, 1131.13638669, 1113.08633797, 1122.71303062, 1173.25316705, 1175.65984021, 1239.43667904, 1193.70988894, 1287.57014230, 1347.73697138, 1299.60350811, 1407.90380046, 1371.80370301, 1311.63687393, 1323.67023975, 1311.63687393, 1323.67023975, 1287.57014230, 1263.50341067, 1263.50341067, 1164.82981098, 1128.72971353, 1166.03314756, 1174.45650363, 1179.26984996, 1162.42313781, 1203.33658159, 1194.91322552, 1239.43667904, 1239.43667904, 1239.43667904, 1173.25316705, 1167.23648414, 1147.98309883, 1143.16975251, 1158.81312807, 1153.99978174, 1149.18643542, 1091.42627950, 1034.86946017, 990.34600665, 956.65258236, 968.68594818, 945.82255313, 1021.63275777, 987.93933348, 986.73599690, 1039.68280649, 1072.17289419, 1092.62961608, 1076.98624052, 1107.06965506, 1089.01960634, 1044.49615282, 1099.84963557, 1098.64629899, 1074.57956736, 1116.69634771, 1092.62961608, 1060.13952838, 1073.37623078, 1093.83295266, 1072.17289419, 1101.05297215, 1116.69634771, 1125.11970378, 1097.44296241, 1091.42627950, 1140.76307935, 1143.16975251, 1147.98309883, 1108.27299164, 1111.88300139, 1101.05297215, 1089.01960634, 1085.40959659, 1076.98624052, 1080.59625027, 1095.03628924, 1073.37623078, 1081.79958685, 1032.46278700, 1002.37937246, 995.15935297, 965.07593843, 941.00920680, 962.66926527, 965.07593843, 956.65258236, 939.80587022, 939.80587022, 896.48575328, 901.29909961, 890.46907037, 874.82569481, 884.45238747, 894.07908012, 934.99252389, 925.36583124, 934.99252389, 901.29909961, 896.48575328, 818.26887548, 811.04885599, 829.09890471, 860.38565584, 825.48889497, 789.38879752, 748.47535375, 732.83197819, 689.51186125, 703.95190023, 709.96858314, 675.07182227, 725.61195870, 725.61195870, 750.88202691, 695.52854416, 697.93521732, 691.91853441, 687.10518809, 630.54836875, 616.10832977, 625.73502243, 631.75170533, 709.96858314, 789.38879752, 776.15209512, 782.16877803, 853.16563635, 878.43570456, 926.56916782, 879.63904114, 926.56916782, 909.72245568, 943.41587996, 867.60567532, 860.38565584, 914.53580201, 897.68908986, 870.01234849, 830.30224130, 895.28241670, 894.07908012, 872.41902165, 859.18231925, 906.11244594, 949.43256287, 944.61921655, 915.73913859, 950.63589945, 999.97269930, 993.95601639, 1032.46278700, 1040.88614307, 1018.02274802, 1013.20940170, 1032.46278700, 987.93933348, 1019.22608460, 1005.98938221, 974.70263109, 1062.54620154, 1055.32618205, 1049.30949914, 1045.69948940, 1063.74953812, 1113.08633797, 1075.78290394, 1033.66612358, 1076.98624052, 1045.69948940, 1057.73285522, 1075.78290394, 1111.88300139, 1115.49301113, 1098.64629899, 1084.20626001, 1062.54620154, 1055.32618205, 1057.73285522, 1067.35954787, 1043.29281624, 1052.91950889, 1056.52951863, 1050.51283573, 1036.07279675, 1057.73285522, 1072.17289419, 1079.39291368, 1080.59625027, 1076.98624052, 1080.59625027, 1058.93619180, 1063.74953812, 1080.59625027, 1078.18957710, 1085.40959659, 1068.56288445, 1042.08947966, 1052.91950889, 1025.24276751, 992.75267981, 998.76936272, 938.60253364, 1021.63275777, 1009.59939195, 1080.59625027, 1021.63275777, 1074.57956736, 1072.17289419, 1034.86946017, 991.54934323, 1010.80272853, 1045.69948940, 1040.88614307, 1008.39605537, 1037.27613333, 1026.44610409, 1092.62961608, 1149.18643542, 1151.59310858, 1109.47632822, 1079.39291368, 1062.54620154, 1101.05297215, 1113.08633797, 1098.64629899, 1069.76622103, 1091.42627950, 1080.59625027, 1114.28967455, 1143.16975251, 1139.55974276, 1164.82981098, 1128.72971353, 1156.40645491, 1117.89968430, 1138.35640618, 1125.11970378, 1119.10302088], 
							[728.01863186, 697.93521732, 730.42530502, 714.78192946, 700.34189048, 691.91853441, 667.85180278, 682.29184176, 675.07182227, 637.76838824, 654.61510038, 640.17506140, 667.85180278, 661.83511987, 651.00509064, 623.32834926, 605.27830054, 595.65160789, 623.32834926, 610.09164687, 601.66829079, 536.68811539, 529.46809590, 490.96132529, 436.81117912, 487.35131554, 594.44827130, 598.05828105, 589.63492498, 613.70165661, 619.71833952, 583.61824207, 571.58487625, 595.65160789, 587.22825181, 581.21156891, 589.63492498, 535.48477881, 561.95818360, 563.16152018, 570.38153967, 586.02491523, 598.05828105, 580.00823233, 559.55151044, 553.53482753, 555.94150069, 553.53482753, 583.61824207, 602.87162738, 567.97486651, 595.65160789, 595.65160789, 571.58487625, 611.29498345, 631.75170533, 631.75170533, 667.85180278, 676.27515885, 732.83197819, 726.81529528, 813.45552915, 783.37211461, 746.06868058, 753.28870007, 770.13541222, 789.38879752, 765.32206589, 758.10204640, 758.10204640, 753.28870007, 753.28870007, 741.25533426, 744.86534400, 770.13541222, 740.05199768, 786.98212436, 764.11872931, 776.15209512, 762.91539273, 774.94875854, 785.77878778, 770.13541222, 776.15209512, 767.72873905, 772.54208538, 786.98212436, 749.67869033, 652.20842722, 657.02177355, 667.85180278, 610.09164687, 660.63178329, 623.32834926, 613.70165661, 595.65160789, 600.46495421, 641.37839799, 693.12187099, 707.56190997, 659.42844671, 701.54522707, 682.29184176, 666.64846620, 705.15523681, 709.96858314, 731.62864161, 743.66200742, 713.57859288, 700.34189048, 731.62864161, 746.06868058, 782.16877803, 784.57545120, 785.77878778, 796.60881701, 827.89556813, 833.91225104, 854.36897293, 913.33246543, 906.11244594, 885.65572405, 883.24905089, 876.02903140, 895.28241670, 867.60567532, 866.40233874, 889.26573379, 908.51911910, 909.72245568, 904.90910935, 918.14581175, 851.96229976, 876.02903140, 902.50243619, 890.46907037, 902.50243619, 921.75582150, 897.68908986, 903.70577277, 920.55248491, 895.28241670, 901.29909961, 877.23236798, 884.45238747, 857.97898267, 863.99566558, 880.84237772, 880.84237772, 884.45238747, 865.19900216, 824.28555839, 756.89870982, 768.93207563, 774.94875854, 754.49203666, 741.25533426, 729.22196844, 711.17191972, 700.34189048, 673.86848569, 701.54522707, 728.01863186, 748.47535375, 807.43884625, 785.77878778, 794.20214385, 746.06868058, 744.86534400, 761.71205615, 744.86534400, 675.07182227, 679.88516860, 694.32520758, 661.83511987, 697.93521732, 742.45867084, 741.25533426, 748.47535375, 737.64532451, 718.39193921, 712.37525630, 761.71205615, 749.67869033, 770.13541222, 811.04885599, 779.76210487, 782.16877803, 836.31892420, 848.35229002, 871.21568507, 832.70891446, 848.35229002, 817.06553890, 796.60881701, 814.65886573, 827.89556813, 859.18231925, 818.26887548, 849.55562660, 862.79232900, 874.82569481, 860.38565584, 861.58899242, 912.12912884, 896.48575328, 908.51911910, 920.55248491, 878.43570456, 910.92579226, 868.80901191, 896.48575328, 889.26573379, 918.14581175, 877.23236798, 894.07908012, 883.24905089, 915.73913859, 890.46907037, 874.82569481, 856.77564609, 860.38565584, 872.41902165, 849.55562660, 823.08222181, 826.69223155, 824.28555839, 842.33560711, 830.30224130, 794.20214385, 819.47221206, 829.09890471, 815.86220232, 818.26887548, 830.30224130, 782.16877803, 760.50871956, 780.96544145, 779.76210487, 796.60881701, 783.37211461, 786.98212436, 829.09890471, 764.11872931, 768.93207563, 772.54208538, 760.50871956, 754.49203666, 770.13541222, 755.69537324, 726.81529528, 685.90185151, 655.81843697, 649.80175406, 618.51500294, 671.46181253, 622.12501268, 653.41176380, 644.98840773, 682.29184176, 723.20528553, 681.08850518, 667.85180278, 703.95190023, 682.29184176, 694.32520758, 684.69851492, 654.61510038, 673.86848569, 658.22511013, 717.18860263, 746.06868058, 815.86220232, 782.16877803, 747.27201717, 783.37211461, 788.18546094, 782.16877803, 788.18546094, 815.86220232, 782.16877803, 762.91539273, 761.71205615, 749.67869033, 817.06553890, 806.23550966, 770.13541222, 839.92893395, 824.28555839, 762.91539273, 746.06868058], 
							[1323.67023975, 1335.70360556, 1323.67023975, 1323.67023975, 1335.70360556, 1323.67023975, 1359.77033719, 1359.77033719, 1395.87043464, 1407.90380046, 1431.97053209, 1419.93716627, 1395.87043464, 1419.93716627, 1419.93716627, 1419.93716627, 1383.83706883, 1335.70360556, 1383.83706883, 1371.80370301, 1347.73697138, 1311.63687393, 1275.53677648, 1347.73697138, 1323.67023975, 1359.77033719, 1383.83706883, 1371.80370301, 1347.73697138, 1323.67023975, 1371.80370301, 1311.63687393, 1323.67023975, 1335.70360556, 1335.70360556, 1359.77033719, 1323.67023975, 1323.67023975, 1371.80370301, 1383.83706883, 1383.83706883, 1371.80370301, 1383.83706883, 1407.90380046, 1383.83706883, 1383.83706883, 1371.80370301, 1383.83706883, 1383.83706883, 1419.93716627, 1359.77033719, 1383.83706883, 1407.90380046, 1395.87043464, 1492.13736117, 1552.30419025, 1564.33755606, 1612.47101933, 1600.43765351, 1600.43765351, 1588.40428770, 1720.77131167, 1684.67121422, 1684.67121422, 1660.60448259, 1672.63784841, 1660.60448259, 1612.47101933, 1588.40428770, 1552.30419025, 1444.00389791, 1444.00389791, 1504.17072698, 1492.13736117, 1516.20409280, 1492.13736117, 1576.37092188, 1564.33755606, 1552.30419025, 1552.30419025, 1552.30419025, 1576.37092188, 1576.37092188, 1576.37092188, 1588.40428770, 1588.40428770, 1648.57111678, 1636.53775096, 1564.33755606, 1588.40428770, 1552.30419025, 1492.13736117, 1492.13736117, 1480.10399535, 1576.37092188, 1504.17072698, 1540.27082443, 1624.50438514, 1660.60448259, 1708.73794585, 1672.63784841, 1732.80467749, 1768.90477493, 1732.80467749, 1756.87140912, 1877.20506728, 1889.23843309, 1913.30516472, 1877.20506728, 1805.00487238, 1817.03823820, 1865.17170146, 1817.03823820, 1829.07160401, 1841.10496983, 1829.07160401, 1841.10496983, 1889.23843309, 1949.40526217, 2033.63882288, 1997.53872544, 1973.47199380, 1937.37189636, 1985.50535962, 1949.40526217, 1913.30516472, 1973.47199380, 1973.47199380, 1937.37189636, 1901.27179891, 1901.27179891, 1925.33853054, 1853.13833565, 1829.07160401, 1853.13833565, 1865.17170146, 1805.00487238, 1805.00487238, 1913.30516472, 1901.27179891, 1877.20506728, 1841.10496983, 1841.10496983, 1805.00487238, 1841.10496983, 1865.17170146, 1901.27179891, 1889.23843309, 1877.20506728, 1865.17170146, 1853.13833565, 1708.73794585, 1684.67121422, 1672.63784841, 1660.60448259, 1576.37092188, 1588.40428770, 1588.40428770, 1540.27082443, 1564.33755606, 1528.23745862, 1552.30419025, 1528.23745862, 1564.33755606, 1588.40428770, 1636.53775096, 1600.43765351, 1516.20409280, 1516.20409280, 1504.17072698, 1516.20409280, 1492.13736117, 1504.17072698, 1528.23745862, 1504.17072698, 1600.43765351, 1612.47101933, 1588.40428770, 1588.40428770, 1624.50438514, 1672.63784841, 1756.87140912, 1708.73794585, 1732.80467749, 1744.83804330, 1817.03823820, 1780.93814075, 1768.90477493, 1780.93814075, 1780.93814075, 1768.90477493, 1684.67121422, 1732.80467749, 1744.83804330, 1768.90477493, 1792.97150657, 1805.00487238, 1805.00487238, 1817.03823820, 1889.23843309, 1973.47199380, 2033.63882288, 1985.50535962, 1937.37189636, 1985.50535962, 1961.43862799, 2009.57209125, 2021.60545707, 2033.63882288, 2009.57209125, 1985.50535962, 1985.50535962, 2105.83901778, 2153.97248104, 2190.07257849, 2202.10594431, 2153.97248104, 2310.40623665, 2310.40623665, 2298.37287083, 2298.37287083, 2286.33950502, 2262.27277339, 2250.23940757, 2370.57306573, 2442.77326062, 2514.97345552, 2514.97345552, 2478.87335807, 2370.57306573, 2334.47296828, 2370.57306573, 2286.33950502, 2286.33950502, 2286.33950502, 2141.93911523, 2214.13931012, 2286.33950502, 2322.43960246, 2358.53969991, 2358.53969991, 2334.47296828, 2394.63979736, 2310.40623665, 2322.43960246, 2322.43960246, 2274.30613920, 2418.70652899, 2527.00682133, 2647.34047949, 2575.14028460, 2514.97345552, 2635.30711368, 2647.34047949, 2454.80662644, 2563.10691878, 2514.97345552, 2502.94008970, 2527.00682133, 2767.67413765, 2539.04018715, 2599.20701623, 2647.34047949, 2635.30711368, 2647.34047949, 2755.64077184, 2924.10789326, 3092.57501468, 2960.20799071, 3140.70847794, 3417.47589171, 3261.04213610, 3212.90867284, 3188.84194121, 3164.77520958, 3224.94203866, 3224.94203866, 3020.37481979, 3140.70847794, 3116.64174631, 3261.04213610, 3381.37579426, 3381.37579426, 3273.07550192, 3357.30906263, 3477.64272079, 3285.10886773, 3212.90867284, 3333.24233100, 3321.20896518, 3453.57598916], 
							[517.43473008, 541.50146171, 535.48477881, 533.07810564, 517.43473008, 488.55465212, 468.09793024, 477.72462289, 460.87791075, 481.33463264, 505.40136427, 517.43473008, 513.82472034, 560.75484702, 539.09478855, 529.46809590, 522.24807641, 475.31794973, 547.51814462, 527.06142274, 524.65474957, 493.36799845, 499.38468136, 472.91127656, 452.45455468, 501.79135452, 549.92481779, 543.90813488, 522.24807641, 557.14483728, 578.80489574, 539.09478855, 566.77152993, 563.16152018, 589.63492498, 564.36485676, 577.60155916, 557.14483728, 575.19488600, 578.80489574, 571.58487625, 563.16152018, 566.77152993, 559.55151044, 543.90813488, 537.89145197, 502.99469110, 521.04473983, 547.51814462, 531.87476906, 517.43473008, 517.43473008, 523.45141299, 517.43473008, 515.02805692, 537.89145197, 535.48477881, 565.56819335, 593.24493472, 592.04159814, 575.19488600, 618.51500294, 605.27830054, 583.61824207, 589.63492498, 604.07496396, 598.05828105, 587.22825181, 548.72148120, 566.77152993, 551.12815437, 545.11147146, 531.87476906, 535.48477881, 557.14483728, 540.29812513, 561.95818360, 546.31480804, 561.95818360, 565.56819335, 549.92481779, 519.84140325, 527.06142274, 530.67143248, 531.87476906, 540.29812513, 559.55151044, 551.12815437, 522.24807641, 525.85808615, 510.21471059, 489.75798871, 506.60470085, 499.38468136, 511.41804717, 499.38468136, 511.41804717, 498.18134478, 516.23139350, 528.26475932, 494.57133503, 522.24807641, 531.87476906, 499.38468136, 516.23139350, 537.89145197, 533.07810564, 575.19488600, 592.04159814, 555.94150069, 592.04159814, 607.68497370, 580.00823233, 592.04159814, 599.26161763, 600.46495421, 589.63492498, 592.04159814, 593.24493472, 642.58173457, 659.42844671, 619.71833952, 636.56505166, 619.71833952, 623.32834926, 638.97172482, 634.15837850, 646.19174431, 661.83511987, 697.93521732, 672.66514911, 695.52854416, 679.88516860, 666.64846620, 676.27515885, 675.07182227, 672.66514911, 697.93521732, 666.64846620, 683.49517834, 696.73188074, 683.49517834, 689.51186125, 666.64846620, 675.07182227, 675.07182227, 665.44512962, 673.86848569, 685.90185151, 687.10518809, 661.83511987, 620.92167610, 571.58487625, 589.63492498, 573.99154942, 522.24807641, 549.92481779, 559.55151044, 531.87476906, 530.67143248, 509.01137401, 553.53482753, 522.24807641, 543.90813488, 560.75484702, 572.78821284, 575.19488600, 517.43473008, 535.48477881, 517.43473008, 510.21471059, 474.11461315, 433.20116937, 462.08124733, 399.50774509, 463.28458391, 487.35131554, 480.13129605, 460.87791075, 469.30126682, 489.75798871, 523.45141299, 534.28144222, 541.50146171, 553.53482753, 589.63492498, 516.23139350, 521.04473983, 563.16152018, 566.77152993, 571.58487625, 546.31480804, 576.39822258, 575.19488600, 587.22825181, 601.66829079, 606.48163712, 610.09164687, 606.48163712, 614.90499319, 619.71833952, 631.75170533, 620.92167610, 631.75170533, 629.34503217, 651.00509064, 666.64846620, 649.80175406, 644.98840773, 651.00509064, 640.17506140, 642.58173457, 654.61510038, 661.83511987, 671.46181253, 684.69851492, 672.66514911, 672.66514911, 657.02177355, 649.80175406, 665.44512962, 653.41176380, 649.80175406, 629.34503217, 649.80175406, 647.39508089, 629.34503217, 625.73502243, 612.49832003, 623.32834926, 614.90499319, 608.88831028, 605.27830054, 607.68497370, 623.32834926, 567.97486651, 587.22825181, 613.70165661, 608.88831028, 622.12501268, 602.87162738, 598.05828105, 608.88831028, 592.04159814, 592.04159814, 601.66829079, 596.85494447, 608.88831028, 596.85494447, 583.61824207, 564.36485676, 549.92481779, 540.29812513, 552.33149095, 488.55465212, 528.26475932, 478.92795947, 517.43473008, 513.82472034, 557.14483728, 584.82157865, 553.53482753, 552.33149095, 581.21156891, 588.43158840, 581.21156891, 590.83826156, 594.44827130, 612.49832003, 624.53168584, 632.95504192, 643.78507115, 577.60155916, 589.63492498, 580.00823233, 589.63492498, 605.27830054, 613.70165661, 611.29498345, 614.90499319, 605.27830054, 575.19488600, 592.04159814, 547.51814462, 537.89145197, 546.31480804, 567.97486651, 567.97486651, 552.33149095, 522.24807641, 513.82472034], 
							[539.09478855, 505.40136427, 483.74130580, 454.86122784, 456.06456442, 445.23453519, 419.96446697, 401.91441825, 412.74444748, 409.13443774, 434.40450595, 417.55779381, 440.42118886, 446.43787177, 463.28458391, 495.77467161, 478.92795947, 431.99783279, 484.94464238, 483.74130580, 468.09793024, 423.57447672, 438.01451570, 419.96446697, 379.05102320, 421.16780356, 469.30126682, 478.92795947, 445.23453519, 445.23453519, 446.43787177, 418.76113039, 419.96446697, 422.37114014, 409.13443774, 398.30440851, 409.13443774, 407.93110116, 412.74444748, 413.94778407, 422.37114014, 405.52442800, 415.15112065, 404.32109141, 392.28772560, 379.05102320, 377.84768662, 380.25435978, 392.28772560, 400.71108167, 387.47437927, 385.06770611, 376.64435004, 371.83100371, 400.71108167, 446.43787177, 446.43787177, 463.28458391, 464.48792049, 477.72462289, 466.89459366, 516.23139350, 507.80803743, 501.79135452, 496.97800820, 505.40136427, 516.23139350, 500.58801794, 478.92795947, 486.14797896, 477.72462289, 486.14797896, 498.18134478, 493.36799845, 501.79135452, 496.97800820, 525.85808615, 528.26475932, 518.63806666, 527.06142274, 516.23139350, 504.19802769, 521.04473983, 512.62138376, 511.41804717, 504.19802769, 505.40136427, 470.50460340, 450.04788151, 465.69125707, 456.06456442, 422.37114014, 427.18448646, 424.77781330, 468.09793024, 422.37114014, 409.13443774, 431.99783279, 452.45455468, 453.65789126, 454.86122784, 505.40136427, 472.91127656, 460.87791075, 469.30126682, 506.60470085, 502.99469110, 493.36799845, 481.33463264, 469.30126682, 487.35131554, 500.58801794, 502.99469110, 496.97800820, 501.79135452, 498.18134478, 489.75798871, 511.41804717, 527.06142274, 517.43473008, 517.43473008, 504.19802769, 504.19802769, 490.96132529, 505.40136427, 486.14797896, 487.35131554, 486.14797896, 499.38468136, 494.57133503, 489.75798871, 484.94464238, 472.91127656, 482.53796922, 481.33463264, 475.31794973, 470.50460340, 481.33463264, 472.91127656, 470.50460340, 475.31794973, 445.23453519, 456.06456442, 435.60784253, 435.60784253, 435.60784253, 448.84454493, 468.09793024, 475.31794973, 466.89459366, 475.31794973, 451.25121810, 425.98114988, 431.99783279, 419.96446697, 411.54111090, 399.50774509, 394.69439876, 374.23767687, 389.88105243, 369.42433055, 397.10107192, 362.20431106, 385.06770611, 427.18448646, 400.71108167, 405.52442800, 387.47437927, 392.28772560, 377.84768662, 381.45769636, 356.18762815, 350.17094524, 359.79763789, 344.15426233, 377.84768662, 391.08438902, 394.69439876, 398.30440851, 399.50774509, 403.11775483, 431.99783279, 422.37114014, 411.54111090, 416.35445723, 427.18448646, 415.15112065, 403.11775483, 433.20116937, 436.81117912, 430.79449621, 398.30440851, 417.55779381, 415.15112065, 419.96446697, 415.15112065, 435.60784253, 459.67457417, 447.64120835, 457.26790100, 458.47123758, 470.50460340, 474.11461315, 462.08124733, 475.31794973, 477.72462289, 469.30126682, 458.47123758, 445.23453519, 445.23453519, 451.25121810, 456.06456442, 463.28458391, 493.36799845, 510.21471059, 505.40136427, 512.62138376, 519.84140325, 513.82472034, 490.96132529, 493.36799845, 496.97800820, 499.38468136, 486.14797896, 498.18134478, 529.46809590, 506.60470085, 499.38468136, 477.72462289, 484.94464238, 476.52128631, 469.30126682, 463.28458391, 474.11461315, 474.11461315, 445.23453519, 452.45455468, 464.48792049, 460.87791075, 472.91127656, 463.28458391, 450.04788151, 465.69125707, 453.65789126, 446.43787177, 463.28458391, 445.23453519, 448.84454493, 435.60784253, 403.11775483, 412.74444748, 403.11775483, 380.25435978, 380.25435978, 340.54425259, 361.00097448, 356.18762815, 400.71108167, 368.22099397, 367.01765738, 361.00097448, 367.01765738, 357.39096473, 333.32423310, 361.00097448, 352.57761841, 329.71422336, 297.22413565, 330.91755994, 326.10421361, 348.96760866, 367.01765738, 361.00097448, 357.39096473, 363.40764764, 363.40764764, 364.61098422, 341.74758917, 357.39096473, 346.56093550, 345.35759892, 340.54425259, 339.34091601, 297.22413565, 297.22413565, 292.41078933, 288.80077958, 282.78409667, 287.59744300, 269.54739428, 263.53071137], 
							[861.58899242, 883.24905089, 931.38251415, 928.97584099, 902.50243619, 842.33560711, 754.49203666, 729.22196844, 734.03531477, 752.08536349, 770.13541222, 819.47221206, 866.40233874, 880.84237772, 839.92893395, 820.67554864, 752.08536349, 709.96858314, 749.67869033, 750.88202691, 722.00194895, 651.00509064, 638.97172482, 625.73502243, 536.68811539, 625.73502243, 746.06868058, 729.22196844, 734.03531477, 744.86534400, 823.08222181, 752.08536349, 741.25533426, 743.66200742, 722.00194895, 685.90185151, 706.35857339, 672.66514911, 676.27515885, 691.91853441, 684.69851492, 679.88516860, 673.86848569, 660.63178329, 632.95504192, 602.87162738, 601.66829079, 590.83826156, 619.71833952, 647.39508089, 607.68497370, 616.10832977, 607.68497370, 617.31166635, 675.07182227, 683.49517834, 677.47849543, 782.16877803, 748.47535375, 794.20214385, 803.82883650, 863.99566558, 890.46907037, 903.70577277, 896.48575328, 916.94247517, 949.43256287, 950.63589945, 870.01234849, 865.19900216, 861.58899242, 862.79232900, 890.46907037, 850.75896318, 883.24905089, 859.18231925, 874.82569481, 872.41902165, 879.63904114, 880.84237772, 902.50243619, 879.63904114, 865.19900216, 831.50557788, 851.96229976, 782.16877803, 761.71205615, 679.88516860, 611.29498345, 634.15837850, 592.04159814, 517.43473008, 554.73816411, 541.50146171, 575.19488600, 537.89145197, 547.51814462, 565.56819335, 631.75170533, 625.73502243, 601.66829079, 676.27515885, 663.03845645, 622.12501268, 659.42844671, 677.47849543, 700.34189048, 711.17191972, 695.52854416, 685.90185151, 711.17191972, 699.13855390, 691.91853441, 673.86848569, 703.95190023, 706.35857339, 705.15523681, 703.95190023, 708.76524656, 750.88202691, 748.47535375, 695.52854416, 679.88516860, 675.07182227, 679.88516860, 673.86848569, 667.85180278, 669.05513936, 673.86848569, 664.24179304, 673.86848569, 654.61510038, 631.75170533, 612.49832003, 606.48163712, 578.80489574, 581.21156891, 620.92167610, 617.31166635, 614.90499319, 620.92167610, 587.22825181, 584.82157865, 555.94150069, 577.60155916, 553.53482753, 570.38153967, 573.99154942, 600.46495421, 578.80489574, 605.27830054, 596.85494447, 543.90813488, 589.63492498, 583.61824207, 588.43158840, 548.72148120, 521.04473983, 507.80803743, 506.60470085, 493.36799845, 470.50460340, 471.70793998, 499.38468136, 542.70479830, 519.84140325, 541.50146171, 507.80803743, 492.16466187, 470.50460340, 484.94464238, 463.28458391, 416.35445723, 436.81117912, 391.08438902, 474.11461315, 484.94464238, 474.11461315, 481.33463264, 462.08124733, 458.47123758, 502.99469110, 496.97800820, 528.26475932, 529.46809590, 519.84140325, 490.96132529, 499.38468136, 507.80803743, 481.33463264, 489.75798871, 459.67457417, 504.19802769, 480.13129605, 493.36799845, 500.58801794, 506.60470085, 506.60470085, 516.23139350, 534.28144222, 541.50146171, 580.00823233, 565.56819335, 571.58487625, 583.61824207, 580.00823233, 595.65160789, 565.56819335, 554.73816411, 560.75484702, 573.99154942, 569.17820309, 582.41490549, 728.01863186, 736.44198793, 728.01863186, 702.74856365, 700.34189048, 711.17191972, 694.32520758, 691.91853441, 684.69851492, 707.56190997, 755.69537324, 772.54208538, 746.06868058, 738.84866109, 728.01863186, 712.37525630, 697.93521732, 661.83511987, 646.19174431, 616.10832977, 643.78507115, 672.66514911, 605.27830054, 610.09164687, 625.73502243, 651.00509064, 637.76838824, 620.92167610, 631.75170533, 638.97172482, 593.24493472, 586.02491523, 619.71833952, 610.09164687, 608.88831028, 543.90813488, 489.75798871, 494.57133503, 493.36799845, 480.13129605, 486.14797896, 395.89773534, 406.72776458, 397.10107192, 405.52442800, 386.27104269, 388.67771585, 394.69439876, 387.47437927, 376.64435004, 356.18762815, 385.06770611, 365.81432080, 320.08753070, 304.44415514, 335.73090626, 308.05416489, 341.74758917, 361.00097448, 375.44101346, 370.62766713, 369.42433055, 376.64435004, 374.23767687, 373.03434029, 362.20431106, 341.74758917, 321.29086728, 320.08753070, 304.44415514, 271.95406744, 243.07398948, 246.68399923, 222.61726759, 219.00725785, 225.02394076, 219.00725785, 215.39724810], 
							[831.50557788, 825.48889497, 792.99880727, 776.15209512, 685.90185151, 690.71519783, 659.42844671, 602.87162738, 565.56819335, 537.89145197, 576.39822258, 619.71833952, 625.73502243, 618.51500294, 586.02491523, 565.56819335, 525.85808615, 489.75798871, 505.40136427, 505.40136427, 505.40136427, 460.87791075, 468.09793024, 434.40450595, 380.25435978, 407.93110116, 504.19802769, 493.36799845, 482.53796922, 512.62138376, 529.46809590, 541.50146171, 553.53482753, 549.92481779, 540.29812513, 505.40136427, 506.60470085, 496.97800820, 537.89145197, 548.72148120, 543.90813488, 543.90813488, 537.89145197, 539.09478855, 518.63806666, 494.57133503, 499.38468136, 506.60470085, 523.45141299, 545.11147146, 528.26475932, 555.94150069, 563.16152018, 558.34817386, 613.70165661, 667.85180278, 613.70165661, 637.76838824, 598.05828105, 646.19174431, 625.73502243, 649.80175406, 679.88516860, 673.86848569, 669.05513936, 734.03531477, 711.17191972, 682.29184176, 663.03845645, 643.78507115, 617.31166635, 649.80175406, 684.69851492, 635.36171508, 654.61510038, 640.17506140, 631.75170533, 655.81843697, 659.42844671, 661.83511987, 683.49517834, 657.02177355, 671.46181253, 631.75170533, 622.12501268, 630.54836875, 601.66829079, 578.80489574, 511.41804717, 499.38468136, 468.09793024, 394.69439876, 436.81117912, 442.82786202, 475.31794973, 445.23453519, 446.43787177, 494.57133503, 535.48477881, 547.51814462, 511.41804717, 601.66829079, 567.97486651, 513.82472034, 531.87476906, 547.51814462, 553.53482753, 577.60155916, 554.73816411, 536.68811539, 554.73816411, 554.73816411, 565.56819335, 536.68811539, 552.33149095, 564.36485676, 565.56819335, 624.53168584, 622.12501268, 675.07182227, 689.51186125, 647.39508089, 663.03845645, 670.25847594, 667.85180278, 651.00509064, 657.02177355, 659.42844671, 663.03845645, 642.58173457, 643.78507115, 667.85180278, 636.56505166, 641.37839799, 641.37839799, 626.93835901, 642.58173457, 635.36171508, 640.17506140, 647.39508089, 647.39508089, 637.76838824, 658.22511013, 637.76838824, 637.76838824, 632.95504192, 642.58173457, 655.81843697, 638.97172482, 648.59841748, 660.63178329, 622.12501268, 586.02491523, 584.82157865, 583.61824207, 583.61824207, 588.43158840, 553.53482753, 507.80803743, 528.26475932, 513.82472034, 505.40136427, 492.16466187, 518.63806666, 564.36485676, 548.72148120, 573.99154942, 554.73816411, 541.50146171, 512.62138376, 512.62138376, 496.97800820, 505.40136427, 511.41804717, 452.45455468, 505.40136427, 518.63806666, 531.87476906, 542.70479830, 525.85808615, 519.84140325, 583.61824207, 582.41490549, 598.05828105, 595.65160789, 613.70165661, 577.60155916, 558.34817386, 580.00823233, 554.73816411, 547.51814462, 506.60470085, 528.26475932, 518.63806666, 551.12815437, 546.31480804, 576.39822258, 587.22825181, 589.63492498, 625.73502243, 657.02177355, 652.20842722, 634.15837850, 622.12501268, 636.56505166, 649.80175406, 643.78507115, 630.54836875, 625.73502243, 641.37839799, 661.83511987, 641.37839799, 655.81843697, 697.93521732, 709.96858314, 681.08850518, 672.66514911, 671.46181253, 658.22511013, 665.44512962, 691.91853441, 678.68183202, 667.85180278, 669.05513936, 672.66514911, 682.29184176, 677.47849543, 669.05513936, 661.83511987, 655.81843697, 619.71833952, 613.70165661, 593.24493472, 580.00823233, 590.83826156, 555.94150069, 553.53482753, 557.14483728, 567.97486651, 551.12815437, 560.75484702, 547.51814462, 564.36485676, 545.11147146, 541.50146171, 565.56819335, 542.70479830, 557.14483728, 548.72148120, 507.80803743, 511.41804717, 482.53796922, 462.08124733, 468.09793024, 429.59115963, 428.38782305, 422.37114014, 441.62452544, 412.74444748, 441.62452544, 458.47123758, 438.01451570, 451.25121810, 406.72776458, 427.18448646, 397.10107192, 379.05102320, 381.45769636, 393.49106218, 409.13443774, 421.16780356, 430.79449621, 451.25121810, 441.62452544, 436.81117912, 462.08124733, 442.82786202, 435.60784253, 430.79449621, 409.13443774, 405.52442800, 412.74444748, 401.91441825, 357.39096473, 352.57761841, 310.46083805, 310.46083805, 309.25750147, 303.24081856, 290.00411616, 290.00411616], 
							[993.04148059, 949.80559721, 957.98828597, 980.18984590, 981.35708238, 875.04229540, 799.09972374, 736.02082013, 736.02082013, 698.63315254, 747.69318497, 712.65202371, 724.33642192, 730.17260434, 695.13144309, 665.91443088, 647.22661377, 606.33723673, 667.08166737, 696.29867957, 665.91443088, 616.85439845, 665.91443088, 670.59541019, 639.04392502, 692.78493675, 725.50365841, 738.35529310, 760.54481966, 772.22921787, 767.56027193, 698.63315254, 684.61428136, 709.15031426, 693.95217324, 700.96762551, 702.13486199, 688.11599082, 713.81926020, 717.32096965, 719.65544262, 743.02423903, 765.22579896, 765.22579896, 712.65202371, 714.98649668, 706.80380793, 712.65202371, 753.54140076, 755.87587372, 716.15373317, 718.48820613, 731.33984083, 746.52594849, 803.78070304, 855.17520844, 922.93509135, 911.26272650, 851.67349898, 864.52513368, 834.15291836, 905.41451072, 891.39563954, 882.04571430, 882.04571430, 899.57832830, 903.08003775, 864.52513368, 852.84073547, 838.82186429, 872.70782243, 868.02684313, 870.37334946, 842.32357375, 883.21295079, 882.04571430, 905.41451072, 884.39222064, 904.24727423, 934.61948955, 933.45225307, 911.26272650, 918.26614541, 862.19066071, 904.24727423, 871.54058595, 879.71124134, 836.48739132, 747.69318497, 720.82267910, 697.46591605, 705.63657144, 724.33642192, 691.61770027, 759.37758318, 717.32096965, 741.85700255, 740.68976606, 813.11859491, 807.28241249, 760.54481966, 834.15291836, 810.78412194, 753.54140076, 789.76183186, 771.06198138, 782.74637959, 792.09630483, 748.87245482, 718.48820613, 769.89474490, 771.06198138, 759.37758318, 781.57914311, 793.26354131, 776.91019717, 783.91361608, 794.43077780, 781.57914311, 766.39303545, 786.24808904, 743.02423903, 764.05856248, 736.02082013, 761.71205615, 760.54481966, 759.37758318, 757.04311021, 755.87587372, 745.35871200, 772.22921787, 775.74296069, 748.87245482, 748.87245482, 744.19147552, 731.33984083, 733.67431379, 731.33984083, 736.02082013, 733.67431379, 741.85700255, 736.02082013, 693.95217324, 660.07824846, 669.42817370, 655.40930253, 660.07824846, 676.43159261, 688.11599082, 688.11599082, 675.26435612, 668.26093722, 650.72832323, 679.93330206, 674.09711964, 663.57995792, 654.24206604, 649.56108674, 661.24548495, 682.27980839, 663.57995792, 658.91101198, 650.72832323, 671.76264667, 713.81926020, 668.26093722, 700.96762551, 668.26093722, 671.76264667, 657.74377550, 672.92988316, 661.24548495, 643.72490432, 664.74719440, 654.24206604, 676.43159261, 677.59882909, 692.78493675, 688.11599082, 684.61428136, 671.76264667, 727.83813137, 718.48820613, 711.48478723, 685.78151785, 699.80038902, 695.13144309, 697.46591605, 711.48478723, 702.13486199, 695.13144309, 668.26093722, 695.13144309, 676.43159261, 676.43159261, 688.11599082, 704.46933496, 711.48478723, 727.83813137, 716.15373317, 723.16918544, 729.00536786, 739.52252958, 746.52594849, 751.20692779, 755.87587372, 743.02423903, 727.83813137, 726.67089489, 712.65202371, 717.32096965, 707.97104441, 744.19147552, 767.72873905, 773.74542196, 758.10204640, 767.72873905, 778.55876829, 764.11872931, 760.50871956, 779.76210487, 760.50871956, 782.16877803, 777.35543171, 784.57545120, 788.18546094, 780.96544145, 782.16877803, 768.93207563, 762.91539273, 753.28870007, 741.25533426, 730.42530502, 742.45867084, 742.45867084, 734.03531477, 728.01863186, 743.66200742, 762.91539273, 752.08536349, 746.06868058, 747.27201717, 749.67869033, 708.76524656, 711.17191972, 713.57859288, 688.30852467, 683.49517834, 675.07182227, 663.03845645, 663.03845645, 643.78507115, 631.75170533, 638.97172482, 613.70165661, 622.12501268, 613.70165661, 637.76838824, 617.31166635, 618.51500294, 637.76838824, 626.93835901, 624.53168584, 625.73502243, 658.22511013, 618.51500294, 611.29498345, 614.90499319, 625.73502243, 620.92167610, 631.75170533, 619.71833952, 641.37839799, 636.56505166, 638.97172482, 651.00509064, 663.03845645, 658.22511013, 642.58173457, 636.56505166, 626.93835901, 631.75170533, 636.56505166, 616.10832977, 613.70165661, 598.05828105, 595.65160789, 599.26161763, 608.88831028, 606.48163712, 595.65160789], 
							[619.71833952, 590.83826156, 557.14483728, 529.46809590, 535.48477881, 517.43473008, 507.80803743, 523.45141299, 504.19802769, 482.53796922, 522.24807641, 499.38468136, 510.21471059, 517.43473008, 489.75798871, 465.69125707, 469.30126682, 451.25121810, 476.52128631, 525.85808615, 516.23139350, 500.58801794, 523.45141299, 511.41804717, 451.25121810, 413.94778407, 505.40136427, 493.36799845, 469.30126682, 453.65789126, 460.87791075, 427.18448646, 413.94778407, 416.35445723, 403.11775483, 392.28772560, 418.76113039, 367.01765738, 379.05102320, 419.96446697, 398.30440851, 407.93110116, 416.35445723, 428.38782305, 406.72776458, 393.49106218, 380.25435978, 425.98114988, 453.65789126, 494.57133503, 465.69125707, 452.45455468, 445.23453519, 434.40450595, 496.97800820, 480.13129605, 559.55151044, 600.46495421, 570.38153967, 614.90499319, 558.34817386, 637.76838824, 623.32834926, 612.49832003, 613.70165661, 607.68497370, 613.70165661, 611.29498345, 587.22825181, 565.56819335, 589.63492498, 584.82157865, 599.26161763, 577.60155916, 612.49832003, 607.68497370, 623.32834926, 623.32834926, 653.41176380, 661.83511987, 667.85180278, 641.37839799, 631.75170533, 636.56505166, 669.05513936, 667.85180278, 657.02177355, 635.36171508, 607.68497370, 601.66829079, 565.56819335, 527.06142274, 554.73816411, 533.07810564, 586.02491523, 613.70165661, 649.80175406, 676.27515885, 697.93521732, 743.66200742, 695.52854416, 780.96544145, 768.93207563, 697.93521732, 760.50871956, 720.79861237, 722.00194895, 750.88202691, 713.57859288, 709.96858314, 712.37525630, 753.28870007, 752.08536349, 734.03531477, 740.05199768, 744.86534400, 741.25533426, 785.77878778, 830.30224130, 942.21254338, 980.71931399, 941.00920680, 955.44924578, 962.66926527, 949.43256287, 903.70577277, 908.51911910, 906.11244594, 914.53580201, 903.70577277, 920.55248491, 978.31264083, 943.41587996, 920.55248491, 938.60253364, 916.94247517, 950.63589945, 953.04257262, 919.34914833, 916.94247517, 909.72245568, 878.43570456, 886.85906063, 817.06553890, 856.77564609, 830.30224130, 831.50557788, 851.96229976, 847.14895344, 820.67554864, 770.13541222, 722.00194895, 703.95190023, 725.61195870, 725.61195870, 722.00194895, 702.74856365, 747.27201717, 722.00194895, 683.49517834, 634.15837850, 697.93521732, 675.07182227, 679.88516860, 730.42530502, 718.39193921, 728.01863186, 673.86848569, 661.83511987, 675.07182227, 651.00509064, 581.21156891, 536.68811539, 661.83511987, 631.75170533, 747.27201717, 827.89556813, 817.06553890, 842.33560711, 866.40233874, 866.40233874, 974.70263109, 956.65258236, 1039.68280649, 1141.96641593, 1215.36994740, 1081.79958685, 1104.66298190, 1152.79644516, 1097.44296241, 1164.82981098, 1127.52637695, 1144.37308909, 1105.86631848, 1095.03628924, 1079.39291368, 1117.89968430, 1158.81312807, 1085.40959659, 1095.03628924, 1116.69634771, 1125.11970378, 1090.22294292, 1074.57956736, 1064.95287471, 1055.32618205, 1070.96955761, 1046.90282598, 1061.34286496, 1120.30635746, 1045.69948940, 1063.74953812, 979.51597741, 989.14267007, 1016.81941144, 998.76936272, 999.97269930, 1005.98938221, 993.95601639, 950.63589945, 942.21254338, 920.55248491, 901.29909961, 854.36897293, 926.56916782, 906.11244594, 902.50243619, 902.50243619, 884.45238747, 849.55562660, 843.53894369, 844.74228027, 819.47221206, 871.21568507, 889.26573379, 835.11558762, 861.58899242, 841.13227053, 830.30224130, 901.29909961, 863.99566558, 866.40233874, 872.41902165, 807.43884625, 774.94875854, 767.72873905, 770.13541222, 779.76210487, 813.45552915, 771.33874880, 786.98212436, 770.13541222, 773.74542196, 767.72873905, 676.27515885, 659.42844671, 632.95504192, 683.49517834, 653.41176380, 749.67869033, 799.01549017, 765.32206589, 782.16877803, 754.49203666, 744.86534400, 735.23865135, 722.00194895, 770.13541222, 734.03531477, 714.78192946, 748.47535375, 750.88202691, 732.83197819, 744.86534400, 723.20528553, 736.44198793, 729.22196844, 713.57859288, 703.95190023, 664.24179304, 673.86848569, 669.05513936, 677.47849543, 637.76838824, 561.95818360, 559.55151044, 524.65474957, 517.43473008, 515.02805692, 486.14797896, 490.96132529], 
							[950.63589945, 950.63589945, 948.22922629, 863.99566558, 861.58899242, 830.30224130, 789.38879752, 722.00194895, 699.13855390, 646.19174431, 673.86848569, 709.96858314, 734.03531477, 735.23865135, 717.18860263, 678.68183202, 620.92167610, 584.82157865, 638.97172482, 646.19174431, 641.37839799, 596.85494447, 632.95504192, 630.54836875, 628.14169559, 604.07496396, 722.00194895, 737.64532451, 712.37525630, 774.94875854, 770.13541222, 702.74856365, 673.86848569, 667.85180278, 670.25847594, 649.80175406, 670.25847594, 653.41176380, 677.47849543, 730.42530502, 693.12187099, 676.27515885, 647.39508089, 634.15837850, 607.68497370, 593.24493472, 594.44827130, 596.85494447, 630.54836875, 641.37839799, 605.27830054, 625.73502243, 659.42844671, 665.44512962, 746.06868058, 738.84866109, 729.22196844, 687.10518809, 689.51186125, 729.22196844, 722.00194895, 742.45867084, 756.89870982, 752.08536349, 736.44198793, 807.43884625, 805.03217308, 783.37211461, 744.86534400, 730.42530502, 738.84866109, 742.45867084, 754.49203666, 722.00194895, 765.32206589, 753.28870007, 755.69537324, 788.18546094, 794.20214385, 795.40548043, 767.72873905, 782.16877803, 773.74542196, 789.38879752, 714.78192946, 713.57859288, 700.34189048, 637.76838824, 567.97486651, 573.99154942, 575.19488600, 524.65474957, 570.38153967, 518.63806666, 536.68811539, 506.60470085, 504.19802769, 563.16152018, 607.68497370, 625.73502243, 625.73502243, 740.05199768, 731.62864161, 715.98526604, 707.56190997, 732.83197819, 872.41902165, 987.93933348, 1044.49615282, 1022.83609435, 1162.42313781, 1170.84649388, 1144.37308909, 1167.23648414, 1151.59310858, 1139.55974276, 1179.26984996, 1063.74953812, 1014.41273828, 1034.86946017, 1058.93619180, 1028.85277726, 1043.29281624, 1033.66612358, 975.90596767, 1016.81941144, 962.66926527, 1010.80272853, 986.73599690, 998.76936272, 983.12598716, 895.28241670, 842.33560711, 845.94561686, 848.35229002, 848.35229002, 849.55562660, 872.41902165, 870.01234849, 842.33560711, 842.33560711, 812.25219257, 808.64218283, 723.20528553, 768.93207563, 759.30538298, 735.23865135, 762.91539273, 806.23550966, 770.13541222, 767.72873905, 746.06868058, 713.57859288, 678.68183202, 709.96858314, 697.93521732, 695.52854416, 617.31166635, 582.41490549, 548.72148120, 540.29812513, 541.50146171, 540.29812513, 552.33149095, 605.27830054, 595.65160789, 619.71833952, 589.63492498, 583.61824207, 551.12815437, 559.55151044, 506.60470085, 464.48792049, 492.16466187, 445.23453519, 513.82472034, 565.56819335, 523.45141299, 553.53482753, 529.46809590, 521.04473983, 559.55151044, 569.17820309, 592.04159814, 593.24493472, 601.66829079, 571.58487625, 565.56819335, 584.82157865, 542.70479830, 557.14483728, 518.63806666, 542.70479830, 518.63806666, 535.48477881, 540.29812513, 547.51814462, 569.17820309, 565.56819335, 588.43158840, 616.10832977, 620.92167610, 637.76838824, 630.54836875, 649.80175406, 673.86848569, 664.24179304, 644.98840773, 616.10832977, 601.66829079, 623.32834926, 625.73502243, 640.17506140, 685.90185151, 697.93521732, 734.03531477, 732.83197819, 719.59527579, 697.93521732, 685.90185151, 720.79861237, 734.03531477, 730.42530502, 712.37525630, 740.05199768, 738.84866109, 740.05199768, 725.61195870, 700.34189048, 711.17191972, 703.95190023, 678.68183202, 676.27515885, 700.34189048, 717.18860263, 657.02177355, 630.54836875, 646.19174431, 661.83511987, 689.51186125, 679.88516860, 661.83511987, 678.68183202, 638.97172482, 632.95504192, 657.02177355, 652.20842722, 660.63178329, 624.53168584, 581.21156891, 567.97486651, 535.48477881, 509.01137401, 518.63806666, 462.08124733, 464.48792049, 474.11461315, 482.53796922, 475.31794973, 498.18134478, 502.99469110, 495.77467161, 480.13129605, 444.03119861, 472.91127656, 425.98114988, 389.88105243, 404.32109141, 421.16780356, 404.32109141, 436.81117912, 451.25121810, 469.30126682, 463.28458391, 466.89459366, 489.75798871, 472.91127656, 476.52128631, 457.26790100, 433.20116937, 411.54111090, 434.40450595, 454.86122784, 374.23767687, 364.61098422, 345.35759892, 327.30755019, 314.07084779, 320.08753070, 296.02079907, 298.42747223], 
							[611.29498345, 594.44827130, 557.14483728, 577.60155916, 569.17820309, 499.38468136, 487.35131554, 529.46809590, 515.02805692, 518.63806666, 536.68811539, 534.28144222, 548.72148120, 551.12815437, 527.06142274, 523.45141299, 481.33463264, 492.16466187, 519.84140325, 498.18134478, 504.19802769, 481.33463264, 492.16466187, 481.33463264, 447.64120835, 490.96132529, 559.55151044, 542.70479830, 528.26475932, 507.80803743, 557.14483728, 511.41804717, 507.80803743, 516.23139350, 500.58801794, 490.96132529, 501.79135452, 482.53796922, 509.01137401, 505.40136427, 500.58801794, 500.58801794, 500.58801794, 501.79135452, 486.14797896, 488.55465212, 487.35131554, 481.33463264, 500.58801794, 493.36799845, 480.13129605, 488.55465212, 504.19802769, 494.57133503, 524.65474957, 530.67143248, 542.70479830, 713.57859288, 679.88516860, 678.68183202, 643.78507115, 703.95190023, 679.88516860, 649.80175406, 683.49517834, 655.81843697, 663.03845645, 672.66514911, 630.54836875, 620.92167610, 629.34503217, 637.76838824, 649.80175406, 632.95504192, 661.83511987, 637.76838824, 670.25847594, 640.17506140, 641.37839799, 638.97172482, 626.93835901, 625.73502243, 619.71833952, 606.48163712, 646.19174431, 631.75170533, 641.37839799, 629.34503217, 612.49832003, 608.88831028, 594.44827130, 571.58487625, 600.46495421, 599.26161763, 611.29498345, 593.24493472, 595.65160789, 606.48163712, 661.83511987, 647.39508089, 637.76838824, 651.00509064, 655.81843697, 643.78507115, 667.85180278, 663.03845645, 671.46181253, 658.22511013, 631.75170533, 619.71833952, 648.59841748, 635.36171508, 625.73502243, 652.20842722, 641.37839799, 657.02177355, 663.03845645, 646.19174431, 631.75170533, 629.34503217, 619.71833952, 605.27830054, 601.66829079, 596.85494447, 596.85494447, 567.97486651, 598.05828105, 567.97486651, 580.00823233, 577.60155916, 582.41490549, 577.60155916, 569.17820309, 573.99154942, 572.78821284, 547.51814462, 529.46809590, 521.04473983, 523.45141299, 539.09478855, 539.09478855, 527.06142274, 534.28144222, 506.60470085, 498.18134478, 493.36799845, 505.40136427, 522.24807641, 519.84140325, 513.82472034, 505.40136427, 484.94464238, 481.33463264, 486.14797896, 484.94464238, 465.69125707, 458.47123758, 446.43787177, 440.42118886, 439.21785228, 445.23453519, 456.06456442, 445.23453519, 457.26790100, 470.50460340, 462.08124733, 460.87791075, 453.65789126, 448.84454493, 427.18448646, 422.37114014, 429.59115963, 417.55779381, 430.79449621, 401.91441825, 457.26790100, 427.18448646, 439.21785228, 422.37114014, 424.77781330, 412.74444748, 423.57447672, 425.98114988, 428.38782305, 421.16780356, 450.04788151, 428.38782305, 436.81117912, 436.81117912, 436.81117912, 433.20116937, 425.98114988, 424.77781330, 421.16780356, 418.76113039, 417.55779381, 428.38782305, 424.77781330, 425.98114988, 435.60784253, 438.01451570, 444.03119861, 441.62452544, 445.23453519, 453.65789126, 458.47123758, 466.89459366, 462.08124733, 444.03119861, 442.82786202, 435.60784253, 446.43787177, 439.21785228, 462.08124733, 459.67457417, 465.69125707, 459.67457417, 489.75798871, 486.14797896, 476.52128631, 476.52128631, 489.75798871, 492.16466187, 475.31794973, 477.72462289, 478.92795947, 481.33463264, 470.50460340, 458.47123758, 460.87791075, 454.86122784, 459.67457417, 459.67457417, 451.25121810, 446.43787177, 435.60784253, 439.21785228, 439.21785228, 457.26790100, 458.47123758, 434.40450595, 438.01451570, 447.64120835, 430.79449621, 425.98114988, 430.79449621, 425.98114988, 419.96446697, 404.32109141, 389.88105243, 369.42433055, 373.03434029, 379.05102320, 377.84768662, 351.37428182, 361.00097448, 361.00097448, 368.22099397, 363.40764764, 364.61098422, 369.42433055, 369.42433055, 362.20431106, 348.96760866, 367.01765738, 368.22099397, 357.39096473, 348.96760866, 357.39096473, 357.39096473, 353.78095499, 348.96760866, 365.81432080, 379.05102320, 357.39096473, 373.03434029, 376.64435004, 375.44101346, 373.03434029, 382.66103294, 376.64435004, 354.98429157, 353.78095499, 353.78095499, 356.18762815, 344.15426233, 336.93424284, 348.96760866, 352.57761841, 353.78095499, 339.34091601], 
							[661.83511987, 624.53168584, 587.22825181, 583.61824207, 572.78821284, 535.48477881, 535.48477881, 536.68811539, 545.11147146, 531.87476906, 547.51814462, 565.56819335, 589.63492498, 636.56505166, 631.75170533, 649.80175406, 611.29498345, 602.87162738, 677.47849543, 689.51186125, 663.03845645, 637.76838824, 647.39508089, 632.95504192, 613.70165661, 670.25847594, 829.09890471, 824.28555839, 783.37211461, 776.15209512, 774.94875854, 697.93521732, 714.78192946, 720.79861237, 707.56190997, 693.12187099, 694.32520758, 626.93835901, 693.12187099, 714.78192946, 712.37525630, 714.78192946, 740.05199768, 741.25533426, 730.42530502, 700.34189048, 703.95190023, 691.91853441, 706.35857339, 738.84866109, 699.13855390, 711.17191972, 670.25847594, 648.59841748, 695.52854416, 728.01863186, 809.84551941, 800.21882676, 807.43884625, 856.77564609, 818.26887548, 863.99566558, 853.16563635, 842.33560711, 890.46907037, 924.16249466, 987.93933348, 996.36268955, 937.39919706, 925.36583124, 909.72245568, 920.55248491, 956.65258236, 914.53580201, 954.24590920, 943.41587996, 950.63589945, 980.71931399, 965.07593843, 959.05925553, 938.60253364, 876.02903140, 854.36897293, 849.55562660, 901.29909961, 880.84237772, 892.87574354, 861.58899242, 842.33560711, 837.52226079, 803.82883650, 742.45867084, 785.77878778, 800.21882676, 854.36897293, 833.91225104, 865.19900216, 902.50243619, 891.67240696, 927.77250440, 892.87574354, 959.05925553, 934.99252389, 884.45238747, 949.43256287, 915.73913859, 924.16249466, 986.73599690, 968.68594818, 987.93933348, 1061.34286496, 1056.52951863, 1045.69948940, 1039.68280649, 1034.86946017, 1040.88614307, 1058.93619180, 1131.13638669, 1167.23648414, 1203.33658159, 1197.31989868, 1170.84649388, 1143.16975251, 1147.98309883, 1158.81312807, 1146.77976225, 1114.28967455, 1155.20311832, 1161.21980123, 1152.79644516, 1138.35640618, 1168.43982072, 1133.54305986, 1108.27299164, 1133.54305986, 1093.83295266, 1161.21980123, 1179.26984996, 1131.13638669, 1178.06651337, 1127.52637695, 1107.06965506, 1131.13638669, 1105.86631848, 1119.10302088, 1115.49301113, 1133.54305986, 1167.23648414, 1191.30321577, 1157.60979149, 1140.76307935, 1057.73285522, 993.95601639, 1004.78604563, 1001.17603588, 995.15935297, 974.70263109, 995.15935297, 915.73913859, 914.53580201, 884.45238747, 979.51597741, 906.11244594, 930.17917757, 961.46592869, 944.61921655, 947.02588971, 919.34914833, 908.51911910, 866.40233874, 833.91225104, 808.64218283, 861.58899242, 882.04571430, 876.02903140, 986.73599690, 1002.37937246, 1021.63275777, 1033.66612358, 1086.61293317, 1087.81626976, 1185.28653286, 1147.98309883, 1166.03314756, 1200.92990842, 1251.47004485, 1179.26984996, 1193.70988894, 1239.43667904, 1181.67652312, 1251.47004485, 1287.57014230, 1347.73697138, 1263.50341067, 1263.50341067, 1227.40331322, 1275.53677648, 1323.67023975, 1323.67023975, 1299.60350811, 1359.77033719, 1407.90380046, 1395.87043464, 1383.83706883, 1383.83706883, 1395.87043464, 1456.03726372, 1419.93716627, 1383.83706883, 1359.77033719, 1395.87043464, 1347.73697138, 1359.77033719, 1383.83706883, 1371.80370301, 1395.87043464, 1419.93716627, 1480.10399535, 1419.93716627, 1371.80370301, 1407.90380046, 1323.67023975, 1311.63687393, 1383.83706883, 1395.87043464, 1419.93716627, 1395.87043464, 1383.83706883, 1359.77033719, 1335.70360556, 1323.67023975, 1287.57014230, 1323.67023975, 1347.73697138, 1311.63687393, 1275.53677648, 1275.53677648, 1335.70360556, 1347.73697138, 1371.80370301, 1383.83706883, 1359.77033719, 1407.90380046, 1347.73697138, 1347.73697138, 1383.83706883, 1419.93716627, 1419.93716627, 1468.07062954, 1552.30419025, 1552.30419025, 1540.27082443, 1612.47101933, 1600.43765351, 1480.10399535, 1564.33755606, 1612.47101933, 1612.47101933, 1648.57111678, 1805.00487238, 1696.70458004, 1708.73794585, 1672.63784841, 1660.60448259, 1672.63784841, 1684.67121422, 1780.93814075, 1768.90477493, 1744.83804330, 1792.97150657, 1853.13833565, 1817.03823820, 1792.97150657, 1756.87140912, 1841.10496983, 1829.07160401, 1829.07160401, 1817.03823820, 1817.03823820, 1817.03823820, 1853.13833565, 1937.37189636, 1985.50535962, 2166.00584686, 2178.03921267, 2238.20604175, 2129.90574941, 1961.43862799, 2057.70555452, 1973.47199380, 2093.80565196], 
							[515.02805692, 496.97800820, 481.33463264, 470.50460340, 440.42118886, 403.11775483, 386.27104269, 391.08438902, 374.23767687, 371.83100371, 410.33777432, 398.30440851, 413.94778407, 397.10107192, 385.06770611, 386.27104269, 373.03434029, 354.98429157, 381.45769636, 367.01765738, 363.40764764, 353.78095499, 342.95092575, 339.34091601, 296.02079907, 336.93424284, 381.45769636, 391.08438902, 381.45769636, 374.23767687, 375.44101346, 356.18762815, 346.56093550, 358.59430131, 352.57761841, 335.73090626, 324.90087703, 309.25750147, 324.90087703, 328.51088677, 339.34091601, 335.73090626, 335.73090626, 329.71422336, 314.07084779, 302.03748198, 292.41078933, 287.59744300, 321.29086728, 326.10421361, 312.86751121, 312.86751121, 303.24081856, 294.81746249, 324.90087703, 332.12089652, 332.12089652, 370.62766713, 373.03434029, 377.84768662, 379.05102320, 429.59115963, 444.03119861, 421.16780356, 421.16780356, 441.62452544, 439.21785228, 435.60784253, 403.11775483, 400.71108167, 393.49106218, 397.10107192, 412.74444748, 393.49106218, 403.11775483, 403.11775483, 404.32109141, 418.76113039, 416.35445723, 421.16780356, 415.15112065, 383.86436953, 379.05102320, 379.05102320, 382.66103294, 354.98429157, 351.37428182, 315.27418438, 311.66417463, 309.25750147, 304.44415514, 283.98743325, 296.02079907, 296.02079907, 306.85082830, 303.24081856, 298.42747223, 333.32423310, 336.93424284, 336.93424284, 336.93424284, 398.30440851, 418.76113039, 407.93110116, 452.45455468, 436.81117912, 427.18448646, 447.64120835, 436.81117912, 419.96446697, 428.38782305, 439.21785228, 436.81117912, 419.96446697, 427.18448646, 440.42118886, 440.42118886, 456.06456442, 453.65789126, 487.35131554, 504.19802769, 482.53796922, 452.45455468, 446.43787177, 457.26790100, 460.87791075, 433.20116937, 440.42118886, 446.43787177, 458.47123758, 457.26790100, 445.23453519, 433.20116937, 416.35445723, 418.76113039, 418.76113039, 423.57447672, 454.86122784, 460.87791075, 460.87791075, 448.84454493, 422.37114014, 421.16780356, 401.91441825, 419.96446697, 423.57447672, 407.93110116, 433.20116937, 425.98114988, 434.40450595, 431.99783279, 407.93110116, 389.88105243, 379.05102320, 377.84768662, 371.83100371, 342.95092575, 342.95092575, 323.69754045, 310.46083805, 294.81746249, 306.85082830, 294.81746249, 311.66417463, 333.32423310, 328.51088677, 330.91755994, 312.86751121, 296.02079907, 276.76741377, 271.95406744, 263.53071137, 226.22727734, 239.46397974, 217.80392127, 251.49734555, 276.76741377, 267.14072111, 271.95406744, 285.19076984, 292.41078933, 335.73090626, 342.95092575, 346.56093550, 336.93424284, 344.15426233, 309.25750147, 316.47752096, 330.91755994, 309.25750147, 294.81746249, 283.98743325, 290.00411616, 288.80077958, 305.64749172, 328.51088677, 335.73090626, 347.76427208, 334.52756968, 350.17094524, 365.81432080, 379.05102320, 371.83100371, 379.05102320, 376.64435004, 400.71108167, 386.27104269, 401.91441825, 385.06770611, 373.03434029, 374.23767687, 359.79763789, 391.08438902, 397.10107192, 423.57447672, 430.79449621, 405.52442800, 416.35445723, 411.54111090, 394.69439876, 397.10107192, 385.06770611, 401.91441825, 385.06770611, 401.91441825, 403.11775483, 401.91441825, 392.28772560, 371.83100371, 369.42433055, 353.78095499, 345.35759892, 342.95092575, 346.56093550, 362.20431106, 332.12089652, 335.73090626, 336.93424284, 347.76427208, 336.93424284, 333.32423310, 327.30755019, 342.95092575, 322.49420387, 327.30755019, 324.90087703, 322.49420387, 320.08753070, 303.24081856, 290.00411616, 283.98743325, 264.73404795, 268.34405769, 263.53071137, 228.63395050, 249.09067239, 246.68399923, 251.49734555, 247.88733581, 246.68399923, 258.71736504, 247.88733581, 249.09067239, 249.09067239, 251.49734555, 243.07398948, 226.22727734, 220.21059443, 237.05730657, 238.26064315, 249.09067239, 238.26064315, 273.15740402, 263.53071137, 262.32737479, 257.51402846, 250.29400897, 247.88733581, 246.68399923, 241.87065290, 226.22727734, 227.43061392, 221.41393101, 209.38056520, 190.12717989, 210.58390178, 209.38056520, 220.21059443, 205.77055545, 197.34719938, 184.11049698], 
							[470.50460340, 457.26790100, 469.30126682, 433.20116937, 410.33777432, 376.64435004, 376.64435004, 433.20116937, 403.11775483, 419.96446697, 464.48792049, 468.09793024, 522.24807641, 499.38468136, 487.35131554, 484.94464238, 427.18448646, 403.11775483, 434.40450595, 421.16780356, 409.13443774, 383.86436953, 374.23767687, 361.00097448, 320.08753070, 397.10107192, 427.18448646, 452.45455468, 451.25121810, 457.26790100, 451.25121810, 397.10107192, 387.47437927, 386.27104269, 367.01765738, 352.57761841, 351.37428182, 324.90087703, 329.71422336, 340.54425259, 351.37428182, 362.20431106, 348.96760866, 342.95092575, 342.95092575, 338.13757943, 322.49420387, 318.88419412, 334.52756968, 344.15426233, 342.95092575, 339.34091601, 326.10421361, 318.88419412, 335.73090626, 351.37428182, 392.28772560, 487.35131554, 498.18134478, 484.94464238, 470.50460340, 498.18134478, 493.36799845, 496.97800820, 493.36799845, 533.07810564, 547.51814462, 528.26475932, 490.96132529, 486.14797896, 474.11461315, 468.09793024, 466.89459366, 458.47123758, 474.11461315, 471.70793998, 471.70793998, 463.28458391, 476.52128631, 471.70793998, 459.67457417, 453.65789126, 446.43787177, 409.13443774, 447.64120835, 447.64120835, 427.18448646, 398.30440851, 348.96760866, 354.98429157, 342.95092575, 300.83414540, 324.90087703, 309.25750147, 330.91755994, 306.85082830, 323.69754045, 316.47752096, 373.03434029, 363.40764764, 419.96446697, 451.25121810, 442.82786202, 422.37114014, 433.20116937, 445.23453519, 481.33463264, 492.16466187, 463.28458391, 459.67457417, 462.08124733, 471.70793998, 471.70793998, 471.70793998, 471.70793998, 469.30126682, 476.52128631, 500.58801794, 493.36799845, 577.60155916, 604.07496396, 590.83826156, 587.22825181, 580.00823233, 589.63492498, 577.60155916, 570.38153967, 553.53482753, 559.55151044, 558.34817386, 543.90813488, 535.48477881, 511.41804717, 499.38468136, 493.36799845, 484.94464238, 512.62138376, 510.21471059, 505.40136427, 511.41804717, 504.19802769, 471.70793998, 468.09793024, 427.18448646, 434.40450595, 442.82786202, 444.03119861, 483.74130580, 511.41804717, 486.14797896, 484.94464238, 457.26790100, 448.84454493, 460.87791075, 457.26790100, 454.86122784, 434.40450595, 417.55779381, 397.10107192, 397.10107192, 385.06770611, 381.45769636, 387.47437927, 407.93110116, 431.99783279, 436.81117912, 436.81117912, 415.15112065, 391.08438902, 379.05102320, 392.28772560, 362.20431106, 336.93424284, 348.96760866, 326.10421361, 407.93110116, 409.13443774, 409.13443774, 382.66103294, 389.88105243, 398.30440851, 430.79449621, 416.35445723, 427.18448646, 429.59115963, 433.20116937, 416.35445723, 410.33777432, 430.79449621, 453.65789126, 451.25121810, 421.16780356, 439.21785228, 427.18448646, 439.21785228, 453.65789126, 453.65789126, 478.92795947, 494.57133503, 542.70479830, 560.75484702, 567.97486651, 569.17820309, 545.11147146, 561.95818360, 629.34503217, 600.46495421, 557.14483728, 547.51814462, 548.72148120, 553.53482753, 564.36485676, 553.53482753, 573.99154942, 598.05828105, 588.43158840, 599.26161763, 622.12501268, 631.75170533, 614.90499319, 631.75170533, 614.90499319, 620.92167610, 610.09164687, 616.10832977, 618.51500294, 608.88831028, 606.48163712, 594.44827130, 588.43158840, 564.36485676, 525.85808615, 511.41804717, 517.43473008, 543.90813488, 511.41804717, 516.23139350, 559.55151044, 555.94150069, 567.97486651, 552.33149095, 535.48477881, 546.31480804, 505.40136427, 499.38468136, 519.84140325, 512.62138376, 523.45141299, 515.02805692, 493.36799845, 511.41804717, 472.91127656, 445.23453519, 440.42118886, 383.86436953, 405.52442800, 386.27104269, 403.11775483, 393.49106218, 403.11775483, 403.11775483, 403.11775483, 413.94778407, 375.44101346, 418.76113039, 374.23767687, 370.62766713, 365.81432080, 391.08438902, 376.64435004, 391.08438902, 406.72776458, 433.20116937, 417.55779381, 411.54111090, 429.59115963, 436.81117912, 416.35445723, 413.94778407, 389.88105243, 380.25435978, 387.47437927, 373.03434029, 336.93424284, 324.90087703, 302.03748198, 308.05416489, 292.41078933, 309.25750147, 285.19076984, 281.58076009], 
							[480.13129605, 459.67457417, 428.38782305, 401.91441825, 400.71108167, 365.81432080, 354.98429157, 346.56093550, 330.91755994, 336.93424284, 374.23767687, 383.86436953, 386.27104269, 380.25435978, 369.42433055, 373.03434029, 339.34091601, 330.91755994, 358.59430131, 348.96760866, 348.96760866, 316.47752096, 321.29086728, 286.39410642, 270.75073086, 312.86751121, 348.96760866, 364.61098422, 357.39096473, 342.95092575, 346.56093550, 328.51088677, 316.47752096, 320.08753070, 308.05416489, 294.81746249, 291.20745274, 275.56407718, 287.59744300, 286.39410642, 291.20745274, 311.66417463, 309.25750147, 299.63080882, 291.20745274, 285.19076984, 277.97075035, 270.75073086, 286.39410642, 318.88419412, 304.44415514, 296.02079907, 290.00411616, 286.39410642, 305.64749172, 320.08753070, 322.49420387, 364.61098422, 371.83100371, 371.83100371, 351.37428182, 403.11775483, 427.18448646, 416.35445723, 417.55779381, 446.43787177, 441.62452544, 439.21785228, 403.11775483, 391.08438902, 381.45769636, 387.47437927, 385.06770611, 373.03434029, 375.44101346, 373.03434029, 391.08438902, 383.86436953, 382.66103294, 381.45769636, 371.83100371, 356.18762815, 357.39096473, 340.54425259, 344.15426233, 332.12089652, 321.29086728, 286.39410642, 264.73404795, 280.37742351, 263.53071137, 238.26064315, 259.92070162, 249.09067239, 259.92070162, 246.68399923, 252.70068213, 269.54739428, 271.95406744, 276.76741377, 269.54739428, 300.83414540, 287.59744300, 268.34405769, 286.39410642, 287.59744300, 282.78409667, 294.81746249, 292.41078933, 283.98743325, 292.41078933, 323.69754045, 316.47752096, 306.85082830, 306.85082830, 320.08753070, 311.66417463, 336.93424284, 334.52756968, 338.13757943, 347.76427208, 328.51088677, 323.69754045, 320.08753070, 321.29086728, 314.07084779, 306.85082830, 306.85082830, 311.66417463, 303.24081856, 305.64749172, 297.22413565, 277.97075035, 276.76741377, 271.95406744, 273.15740402, 268.34405769, 282.78409667, 286.39410642, 281.58076009, 279.17408693, 269.54739428, 263.53071137, 243.07398948, 251.49734555, 252.70068213, 253.90401872, 263.53071137, 279.17408693, 264.73404795, 258.71736504, 250.29400897, 244.27732606, 246.68399923, 244.27732606, 238.26064315, 232.24396025, 225.02394076, 215.39724810, 214.19391152, 198.55053596, 210.58390178, 208.17722861, 221.41393101, 228.63395050, 216.60058469, 219.00725785, 220.21059443, 204.56721887, 196.14386280, 185.31383356, 97.47026311, 111.91030209, 116.72364841, 111.91030209, 139.58704346, 186.51717015, 179.29715066, 172.07713117, 169.67045800, 185.31383356, 263.53071137, 247.88733581, 222.61726759, 229.83728708, 240.66731632, 215.39724810, 208.17722861, 215.39724810, 211.78723836, 210.58390178, 199.75387254, 198.55053596, 210.58390178, 214.19391152, 226.22727734, 223.82060418, 227.43061392, 219.00725785, 246.68399923, 252.70068213, 259.92070162, 250.29400897, 249.09067239, 250.29400897, 273.15740402, 273.15740402, 259.92070162, 251.49734555, 252.70068213, 250.29400897, 249.09067239, 246.68399923, 261.12403820, 270.75073086, 286.39410642, 274.36074060, 275.56407718, 275.56407718, 271.95406744, 270.75073086, 267.14072111, 273.15740402, 264.73404795, 268.34405769, 269.54739428, 276.76741377, 267.14072111, 264.73404795, 257.51402846, 249.09067239, 238.26064315, 226.22727734, 232.24396025, 239.46397974, 227.43061392, 227.43061392, 234.65063341, 232.24396025, 229.83728708, 227.43061392, 210.58390178, 234.65063341, 221.41393101, 216.60058469, 223.82060418, 220.21059443, 219.00725785, 204.56721887, 192.53385305, 194.94052622, 168.46712142, 169.67045800, 166.06044826, 149.21373612, 167.26378484, 156.43375561, 161.24710193, 146.80706295, 145.60372637, 150.41707270, 146.80706295, 141.99371663, 132.36702397, 141.99371663, 138.38370688, 129.96035081, 119.13032158, 138.38370688, 135.97703372, 144.40038979, 145.60372637, 156.43375561, 156.43375561, 155.23041902, 149.21373612, 146.80706295, 143.19705321, 143.19705321, 137.18037030, 126.35034107, 122.74033132, 120.33365816, 116.72364841, 109.50362892, 109.50362892, 107.09695576, 105.89361918, 110.70696551, 101.08027285, 101.08027285], 
							[1016.81941144, 1032.46278700, 1040.88614307, 1040.88614307, 1052.91950889, 1020.42942119, 1004.78604563, 963.87260185, 926.56916782, 904.90910935, 920.55248491, 920.55248491, 968.68594818, 957.85591894, 947.02588971, 944.61921655, 904.90910935, 831.50557788, 827.89556813, 833.91225104, 809.84551941, 713.57859288, 740.05199768, 693.12187099, 649.80175406, 743.66200742, 863.99566558, 854.36897293, 838.72559737, 842.33560711, 876.02903140, 791.79547068, 776.15209512, 827.89556813, 762.91539273, 731.62864161, 734.03531477, 667.85180278, 722.00194895, 768.93207563, 765.32206589, 749.67869033, 759.30538298, 753.28870007, 730.42530502, 711.17191972, 718.39193921, 720.79861237, 723.20528553, 754.49203666, 735.23865135, 802.62549992, 790.59213410, 788.18546094, 847.14895344, 868.80901191, 879.63904114, 890.46907037, 890.46907037, 907.31578252, 888.06239721, 902.50243619, 907.31578252, 909.72245568, 924.16249466, 907.31578252, 888.06239721, 854.36897293, 814.65886573, 820.67554864, 833.91225104, 821.87888522, 851.96229976, 842.33560711, 862.79232900, 867.60567532, 890.46907037, 897.68908986, 909.72245568, 924.16249466, 920.55248491, 903.70577277, 878.43570456, 854.36897293, 825.48889497, 808.64218283, 789.38879752, 765.32206589, 696.73188074, 700.34189048, 679.88516860, 612.49832003, 730.42530502, 690.71519783, 722.00194895, 703.95190023, 717.18860263, 743.66200742, 820.67554864, 845.94561686, 830.30224130, 936.19586048, 918.14581175, 805.03217308, 866.40233874, 844.74228027, 843.53894369, 831.50557788, 832.70891446, 794.20214385, 847.14895344, 886.85906063, 880.84237772, 848.35229002, 836.31892420, 854.36897293, 866.40233874, 888.06239721, 889.26573379, 960.26259211, 980.71931399, 992.75267981, 974.70263109, 984.32932374, 1003.58270904, 990.34600665, 959.05925553, 962.66926527, 962.66926527, 956.65258236, 942.21254338, 966.27927501, 918.14581175, 903.70577277, 894.07908012, 892.87574354, 892.87574354, 912.12912884, 898.89242645, 878.43570456, 877.23236798, 848.35229002, 836.31892420, 788.18546094, 812.25219257, 794.20214385, 800.21882676, 809.84551941, 815.86220232, 812.25219257, 803.82883650, 807.43884625, 918.14581175, 971.09262134, 960.26259211, 922.95915808, 906.11244594, 901.29909961, 915.73913859, 862.79232900, 835.11558762, 812.25219257, 808.64218283, 806.23550966, 835.11558762, 831.50557788, 838.72559737, 825.48889497, 778.55876829, 722.00194895, 738.84866109, 706.35857339, 691.91853441, 720.79861237, 687.10518809, 771.33874880, 842.33560711, 823.08222181, 774.94875854, 778.55876829, 814.65886573, 878.43570456, 860.38565584, 853.16563635, 860.38565584, 927.77250440, 907.31578252, 890.46907037, 901.29909961, 865.19900216, 836.31892420, 818.26887548, 839.92893395, 844.74228027, 851.96229976, 859.18231925, 861.58899242, 862.79232900, 868.80901191, 913.33246543, 914.53580201, 944.61921655, 934.99252389, 920.55248491, 919.34914833, 920.55248491, 901.29909961, 884.45238747, 889.26573379, 867.60567532, 877.23236798, 873.62235823, 889.26573379, 897.68908986, 903.70577277, 900.09576303, 898.89242645, 903.70577277, 876.02903140, 879.63904114, 897.68908986, 895.28241670, 894.07908012, 889.26573379, 885.65572405, 916.94247517, 906.11244594, 880.84237772, 859.18231925, 849.55562660, 844.74228027, 832.70891446, 817.06553890, 818.26887548, 824.28555839, 806.23550966, 803.82883650, 819.47221206, 820.67554864, 829.09890471, 817.06553890, 815.86220232, 823.08222181, 808.64218283, 800.21882676, 815.86220232, 809.84551941, 818.26887548, 789.38879752, 766.52540247, 777.35543171, 717.18860263, 725.61195870, 729.22196844, 691.91853441, 699.13855390, 681.08850518, 695.52854416, 675.07182227, 677.47849543, 697.93521732, 676.27515885, 628.14169559, 624.53168584, 671.46181253, 590.83826156, 542.70479830, 435.60784253, 466.89459366, 439.21785228, 483.74130580, 506.60470085, 577.60155916, 589.63492498, 598.05828105, 600.46495421, 590.83826156, 618.51500294, 616.10832977, 553.53482753, 536.68811539, 566.77152993, 528.26475932, 411.54111090, 391.08438902, 385.06770611, 362.20431106, 368.22099397, 387.47437927, 348.96760866, 347.76427208], 
							[1371.80370301, 1335.70360556, 1299.60350811, 1323.67023975, 1335.70360556, 1251.47004485, 1263.50341067, 1251.47004485, 1287.57014230, 1263.50341067, 1299.60350811, 1311.63687393, 1299.60350811, 1287.57014230, 1227.40331322, 1186.48986945, 1143.16975251, 1061.34286496, 1132.33972327, 1087.81626976, 1089.01960634, 1012.00606512, 1018.02274802, 992.75267981, 961.46592869, 1092.62961608, 1263.50341067, 1287.57014230, 1227.40331322, 1251.47004485, 1227.40331322, 1143.16975251, 1083.00292343, 1101.05297215, 1078.18957710, 1025.24276751, 1040.88614307, 1043.29281624, 1083.00292343, 1119.10302088, 1095.03628924, 1101.05297215, 1092.62961608, 1105.86631848, 1058.93619180, 993.95601639, 991.54934323, 963.87260185, 980.71931399, 997.56602614, 974.70263109, 981.92265058, 986.73599690, 963.87260185, 1016.81941144, 1078.18957710, 1058.93619180, 1090.22294292, 1149.18643542, 1144.37308909, 1117.89968430, 1161.21980123, 1190.09987919, 1135.94973302, 1173.25316705, 1131.13638669, 1120.30635746, 1093.83295266, 1081.79958685, 1032.46278700, 1010.80272853, 993.95601639, 1027.64944068, 977.10930425, 998.76936272, 1024.03943093, 1024.03943093, 1019.22608460, 1022.83609435, 998.76936272, 1034.86946017, 1012.00606512, 1007.19271879, 1004.78604563, 1003.58270904, 1064.95287471, 1052.91950889, 1049.30949914, 993.95601639, 1021.63275777, 989.14267007, 933.78918731, 919.34914833, 932.58585073, 972.29595792, 924.16249466, 933.78918731, 983.12598716, 1022.83609435, 1080.59625027, 1027.64944068, 1081.79958685, 1070.96955761, 1040.88614307, 1091.42627950, 1101.05297215, 1068.56288445, 1046.90282598, 1010.80272853, 995.15935297, 1040.88614307, 1076.98624052, 1097.44296241, 1103.45964532, 1109.47632822, 1150.38977200, 1151.59310858, 1114.28967455, 1172.04983047, 1160.01646465, 1190.09987919, 1176.86317679, 1179.26984996, 1169.64315730, 1194.91322552, 1161.21980123, 1169.64315730, 1153.99978174, 1173.25316705, 1193.70988894, 1215.36994740, 1323.67023975, 1239.43667904, 1215.36994740, 1215.36994740, 1198.52323526, 1187.69320603, 1215.36994740, 1215.36994740, 1227.40331322, 1200.92990842, 1168.43982072, 1167.23648414, 1146.77976225, 1172.04983047, 1167.23648414, 1156.40645491, 1170.84649388, 1181.67652312, 1152.79644516, 1146.77976225, 1185.28653286, 1190.09987919, 1168.43982072, 1186.48986945, 1182.87985970, 1192.50655235, 1107.06965506, 1070.96955761, 1052.91950889, 969.88928476, 1004.78604563, 1034.86946017, 1012.00606512, 1083.00292343, 1085.40959659, 1110.67966481, 1081.79958685, 1073.37623078, 1033.66612358, 1051.71617231, 1022.83609435, 996.36268955, 1016.81941144, 986.73599690, 1105.86631848, 1084.20626001, 1098.64629899, 1083.00292343, 1109.47632822, 1109.47632822, 1203.33658159, 1180.47318654, 1215.36994740, 1239.43667904, 1323.67023975, 1167.23648414, 1196.11656210, 1203.33658159, 1182.87985970, 1191.30321577, 1155.20311832, 1203.33658159, 1151.59310858, 1179.26984996, 1215.36994740, 1263.50341067, 1215.36994740, 1251.47004485, 1287.57014230, 1263.50341067, 1371.80370301, 1275.53677648, 1275.53677648, 1311.63687393, 1383.83706883, 1383.83706883, 1359.77033719, 1299.60350811, 1311.63687393, 1287.57014230, 1263.50341067, 1395.87043464, 1407.90380046, 1444.00389791, 1419.93716627, 1419.93716627, 1456.03726372, 1419.93716627, 1383.83706883, 1431.97053209, 1431.97053209, 1492.13736117, 1431.97053209, 1492.13736117, 1468.07062954, 1456.03726372, 1407.90380046, 1359.77033719, 1383.83706883, 1347.73697138, 1287.57014230, 1311.63687393, 1335.70360556, 1347.73697138, 1215.36994740, 1251.47004485, 1251.47004485, 1287.57014230, 1311.63687393, 1287.57014230, 1263.50341067, 1323.67023975, 1275.53677648, 1251.47004485, 1299.60350811, 1227.40331322, 1215.36994740, 1143.16975251, 1057.73285522, 1076.98624052, 1058.93619180, 1046.90282598, 1040.88614307, 857.97898267, 918.14581175, 868.80901191, 927.77250440, 949.43256287, 918.14581175, 937.39919706, 876.02903140, 836.31892420, 746.06868058, 842.33560711, 836.31892420, 738.84866109, 661.83511987, 661.83511987, 691.91853441, 741.25533426, 806.23550966, 856.77564609, 854.36897293, 833.91225104, 829.09890471, 788.18546094, 827.89556813, 815.86220232, 805.03217308, 748.47535375, 761.71205615, 794.20214385, 708.76524656, 720.79861237, 702.74856365, 753.28870007, 761.71205615, 736.44198793, 684.69851492, 677.47849543], 
							[1347.73697138, 1347.73697138, 1275.53677648, 1287.57014230, 1287.57014230, 1239.43667904, 1203.33658159, 1110.67966481, 1009.59939195, 968.68594818, 1026.44610409, 1004.78604563, 1040.88614307, 1037.27613333, 950.63589945, 919.34914833, 848.35229002, 806.23550966, 845.94561686, 877.23236798, 845.94561686, 808.64218283, 830.30224130, 829.09890471, 821.87888522, 874.82569481, 1016.81941144, 1066.15621129, 986.73599690, 1013.20940170, 1057.73285522, 967.48261160, 894.07908012, 882.04571430, 863.99566558, 860.38565584, 860.38565584, 861.58899242, 894.07908012, 926.56916782, 894.07908012, 903.70577277, 931.38251415, 909.72245568, 876.02903140, 878.43570456, 874.82569481, 839.92893395, 906.11244594, 896.48575328, 878.43570456, 902.50243619, 906.11244594, 920.55248491, 1020.42942119, 990.34600665, 971.09262134, 1034.86946017, 1107.06965506, 1095.03628924, 1010.80272853, 1075.78290394, 1086.61293317, 1084.20626001, 1085.40959659, 1086.61293317, 1098.64629899, 1061.34286496, 1043.29281624, 1025.24276751, 1022.83609435, 1063.74953812, 1092.62961608, 1062.54620154, 1122.71303062, 1083.00292343, 1078.18957710, 1089.01960634, 1098.64629899, 1111.88300139, 1092.62961608, 1093.83295266, 1119.10302088, 1166.03314756, 1143.16975251, 1123.91636720, 1134.74639644, 1117.89968430, 1064.95287471, 1057.73285522, 1015.61607486, 939.80587022, 998.76936272, 950.63589945, 996.36268955, 939.80587022, 986.73599690, 985.53266032, 1004.78604563, 1022.83609435, 992.75267981, 1104.66298190, 1044.49615282, 984.32932374, 1097.44296241, 1137.15306960, 1141.96641593, 1145.57642567, 1096.23962583, 1101.05297215, 1188.89654261, 1203.33658159, 1166.03314756, 1173.25316705, 1176.86317679, 1170.84649388, 1150.38977200, 1184.08319628, 1203.33658159, 1311.63687393, 1299.60350811, 1251.47004485, 1215.36994740, 1287.57014230, 1323.67023975, 1227.40331322, 1251.47004485, 1227.40331322, 1275.53677648, 1239.43667904, 1227.40331322, 1263.50341067, 1181.67652312, 1174.45650363, 1167.23648414, 1108.27299164, 1121.50969404, 1131.13638669, 1160.01646465, 1191.30321577, 1185.28653286, 1145.57642567, 1145.57642567, 1128.72971353, 1145.57642567, 1192.50655235, 1200.92990842, 1215.36994740, 1263.50341067, 1239.43667904, 1239.43667904, 1215.36994740, 1111.88300139, 1083.00292343, 1123.91636720, 1084.20626001, 1051.71617231, 1004.78604563, 933.78918731, 897.68908986, 859.18231925, 906.11244594, 867.60567532, 879.63904114, 942.21254338, 930.17917757, 936.19586048, 902.50243619, 878.43570456, 862.79232900, 851.96229976, 814.65886573, 754.49203666, 747.27201717, 729.22196844, 818.26887548, 829.09890471, 812.25219257, 861.58899242, 896.48575328, 876.02903140, 1049.30949914, 1009.59939195, 1002.37937246, 1031.25945042, 1058.93619180, 965.07593843, 992.75267981, 1015.61607486, 969.88928476, 986.73599690, 965.07593843, 980.71931399, 986.73599690, 1005.98938221, 1019.22608460, 1068.56288445, 1090.22294292, 1143.16975251, 1185.28653286, 1167.23648414, 1203.33658159, 1190.09987919, 1178.06651337, 1147.98309883, 1176.86317679, 1178.06651337, 1158.81312807, 1114.28967455, 1123.91636720, 1185.28653286, 1196.11656210, 1239.43667904, 1263.50341067, 1287.57014230, 1239.43667904, 1323.67023975, 1419.93716627, 1468.07062954, 1431.97053209, 1431.97053209, 1419.93716627, 1431.97053209, 1395.87043464, 1468.07062954, 1456.03726372, 1407.90380046, 1419.93716627, 1371.80370301, 1431.97053209, 1359.77033719, 1359.77033719, 1311.63687393, 1275.53677648, 1299.60350811, 1179.26984996, 1155.20311832, 1155.20311832, 1169.64315730, 1239.43667904, 1227.40331322, 1227.40331322, 1311.63687393, 1239.43667904, 1203.33658159, 1227.40331322, 1227.40331322, 1251.47004485, 1197.31989868, 1143.16975251, 1123.91636720, 1058.93619180, 972.29595792, 989.14267007, 818.26887548, 902.50243619, 843.53894369, 856.77564609, 855.57230951, 847.14895344, 888.06239721, 888.06239721, 848.35229002, 833.91225104, 845.94561686, 842.33560711, 785.77878778, 741.25533426, 808.64218283, 812.25219257, 847.14895344, 886.85906063, 1020.42942119, 992.75267981, 985.53266032, 1025.24276751, 1040.88614307, 1034.86946017, 997.56602614, 928.97584099, 974.70263109, 914.53580201, 912.12912884, 807.43884625, 855.57230951, 827.89556813, 823.08222181, 823.08222181, 851.96229976, 785.77878778, 794.20214385], 
							[960.26259211, 949.43256287, 904.90910935, 926.56916782, 908.51911910, 842.33560711, 854.36897293, 868.80901191, 853.16563635, 841.13227053, 873.62235823, 902.50243619, 878.43570456, 873.62235823, 835.11558762, 818.26887548, 788.18546094, 770.13541222, 820.67554864, 786.98212436, 799.01549017, 752.08536349, 744.86534400, 730.42530502, 679.88516860, 786.98212436, 906.11244594, 880.84237772, 848.35229002, 854.36897293, 866.40233874, 814.65886573, 801.42216334, 808.64218283, 776.15209512, 773.74542196, 762.91539273, 741.25533426, 779.76210487, 803.82883650, 794.20214385, 796.60881701, 789.38879752, 776.15209512, 764.11872931, 742.45867084, 737.64532451, 725.61195870, 764.11872931, 767.72873905, 737.64532451, 737.64532451, 741.25533426, 730.42530502, 770.13541222, 782.16877803, 802.62549992, 863.99566558, 890.46907037, 885.65572405, 862.79232900, 895.28241670, 933.78918731, 883.24905089, 894.07908012, 867.60567532, 857.97898267, 843.53894369, 800.21882676, 795.40548043, 770.13541222, 786.98212436, 815.86220232, 770.13541222, 812.25219257, 789.38879752, 802.62549992, 791.79547068, 803.82883650, 806.23550966, 818.26887548, 790.59213410, 803.82883650, 779.76210487, 773.74542196, 817.06553890, 812.25219257, 800.21882676, 767.72873905, 770.13541222, 754.49203666, 728.01863186, 785.77878778, 770.13541222, 773.74542196, 753.28870007, 762.91539273, 764.11872931, 794.20214385, 841.13227053, 806.23550966, 856.77564609, 845.94561686, 799.01549017, 829.09890471, 836.31892420, 815.86220232, 797.81215359, 794.20214385, 786.98212436, 803.82883650, 838.72559737, 826.69223155, 836.31892420, 830.30224130, 851.96229976, 842.33560711, 838.72559737, 836.31892420, 832.70891446, 837.52226079, 836.31892420, 818.26887548, 807.43884625, 815.86220232, 784.57545120, 788.18546094, 802.62549992, 800.21882676, 806.23550966, 818.26887548, 848.35229002, 812.25219257, 799.01549017, 791.79547068, 789.38879752, 760.50871956, 768.93207563, 789.38879752, 771.33874880, 770.13541222, 738.84866109, 734.03531477, 706.35857339, 725.61195870, 713.57859288, 708.76524656, 719.59527579, 743.66200742, 734.03531477, 718.39193921, 773.74542196, 783.37211461, 772.54208538, 786.98212436, 764.11872931, 764.11872931, 728.01863186, 722.00194895, 702.74856365, 678.68183202, 679.88516860, 677.47849543, 673.86848569, 696.73188074, 699.13855390, 711.17191972, 697.93521732, 677.47849543, 640.17506140, 649.80175406, 617.31166635, 616.10832977, 641.37839799, 602.87162738, 677.47849543, 685.90185151, 678.68183202, 703.95190023, 684.69851492, 734.03531477, 780.96544145, 779.76210487, 784.57545120, 794.20214385, 844.74228027, 794.20214385, 791.79547068, 806.23550966, 758.10204640, 752.08536349, 728.01863186, 768.93207563, 730.42530502, 740.05199768, 754.49203666, 799.01549017, 796.60881701, 790.59213410, 818.26887548, 829.09890471, 895.28241670, 857.97898267, 876.02903140, 865.19900216, 880.84237772, 880.84237772, 855.57230951, 829.09890471, 843.53894369, 839.92893395, 836.31892420, 876.02903140, 914.53580201, 914.53580201, 942.21254338, 948.22922629, 979.51597741, 947.02588971, 962.66926527, 949.43256287, 953.04257262, 980.71931399, 962.66926527, 975.90596767, 963.87260185, 936.19586048, 920.55248491, 886.85906063, 903.70577277, 874.82569481, 863.99566558, 861.58899242, 837.52226079, 843.53894369, 829.09890471, 817.06553890, 831.50557788, 814.65886573, 851.96229976, 832.70891446, 832.70891446, 872.41902165, 849.55562660, 844.74228027, 859.18231925, 835.11558762, 829.09890471, 761.71205615, 723.20528553, 711.17191972, 709.96858314, 715.98526604, 722.00194895, 619.71833952, 673.86848569, 620.92167610, 659.42844671, 651.00509064, 641.37839799, 689.51186125, 625.73502243, 601.66829079, 539.09478855, 590.83826156, 580.00823233, 536.68811539, 487.35131554, 529.46809590, 518.63806666, 576.39822258, 581.21156891, 594.44827130, 607.68497370, 600.46495421, 582.41490549, 589.63492498, 602.87162738, 622.12501268, 607.68497370, 581.21156891, 584.82157865, 614.90499319, 560.75484702, 552.33149095, 573.99154942, 600.46495421, 570.38153967, 577.60155916, 547.51814462, 513.82472034], 
							[1299.60350811, 1275.53677648, 1215.36994740, 1143.16975251, 1190.09987919, 1161.21980123, 1167.23648414, 1215.36994740, 1239.43667904, 1239.43667904, 1263.50341067, 1239.43667904, 1239.43667904, 1227.40331322, 1203.33658159, 1173.25316705, 1129.93305011, 1090.22294292, 1155.20311832, 1173.25316705, 1108.27299164, 1084.20626001, 1034.86946017, 1046.90282598, 1010.80272853, 1031.25945042, 1185.28653286, 1181.67652312, 1125.11970378, 1133.54305986, 1167.23648414, 1131.13638669, 1091.42627950, 1087.81626976, 1034.86946017, 1020.42942119, 1010.80272853, 1013.20940170, 997.56602614, 1046.90282598, 1009.59939195, 1008.39605537, 1016.81941144, 1010.80272853, 995.15935297, 984.32932374, 968.68594818, 962.66926527, 986.73599690, 1040.88614307, 1038.47946991, 1079.39291368, 1115.49301113, 1111.88300139, 1166.03314756, 1215.36994740, 1359.77033719, 1395.87043464, 1395.87043464, 1383.83706883, 1347.73697138, 1468.07062954, 1504.17072698, 1395.87043464, 1335.70360556, 1468.07062954, 1419.93716627, 1419.93716627, 1347.73697138, 1347.73697138, 1359.77033719, 1383.83706883, 1431.97053209, 1431.97053209, 1456.03726372, 1444.00389791, 1564.33755606, 1564.33755606, 1600.43765351, 1660.60448259, 1612.47101933, 1600.43765351, 1540.27082443, 1552.30419025, 1624.50438514, 1624.50438514, 1624.50438514, 1576.37092188, 1564.33755606, 1564.33755606, 1528.23745862, 1431.97053209, 1480.10399535, 1528.23745862, 1600.43765351, 1540.27082443, 1564.33755606, 1672.63784841, 1660.60448259, 1696.70458004, 1708.73794585, 1925.33853054, 1925.33853054, 1841.10496983, 1937.37189636, 1889.23843309, 1913.30516472, 1925.33853054, 1805.00487238, 1780.93814075, 1792.97150657, 1805.00487238, 1744.83804330, 1732.80467749, 1744.83804330, 1805.00487238, 1805.00487238, 1865.17170146, 1901.27179891, 1889.23843309, 1841.10496983, 1817.03823820, 1829.07160401, 1817.03823820, 1817.03823820, 1768.90477493, 1756.87140912, 1792.97150657, 1756.87140912, 1817.03823820, 1792.97150657, 1780.93814075, 1768.90477493, 1732.80467749, 1768.90477493, 1756.87140912, 1768.90477493, 1708.73794585, 1720.77131167, 1708.73794585, 1696.70458004, 1624.50438514, 1660.60448259, 1636.53775096, 1684.67121422, 1672.63784841, 1636.53775096, 1684.67121422, 1708.73794585, 1660.60448259, 1648.57111678, 1612.47101933, 1588.40428770, 1540.27082443, 1528.23745862, 1383.83706883, 1359.77033719, 1383.83706883, 1335.70360556, 1335.70360556, 1215.36994740, 1347.73697138, 1287.57014230, 1311.63687393, 1371.80370301, 1371.80370301, 1359.77033719, 1227.40331322, 1287.57014230, 1227.40331322, 1215.36994740, 1173.25316705, 1196.11656210, 1227.40331322, 1215.36994740, 1323.67023975, 1371.80370301, 1359.77033719, 1335.70360556, 1371.80370301, 1335.70360556, 1492.13736117, 1456.03726372, 1540.27082443, 1528.23745862, 1516.20409280, 1419.93716627, 1456.03726372, 1528.23745862, 1504.17072698, 1492.13736117, 1383.83706883, 1431.97053209, 1371.80370301, 1431.97053209, 1383.83706883, 1456.03726372, 1468.07062954, 1468.07062954, 1468.07062954, 1492.13736117, 1612.47101933, 1540.27082443, 1576.37092188, 1612.47101933, 1660.60448259, 1696.70458004, 1684.67121422, 1624.50438514, 1660.60448259, 1600.43765351, 1600.43765351, 1696.70458004, 1732.80467749, 1744.83804330, 1744.83804330, 1768.90477493, 1805.00487238, 1805.00487238, 1720.77131167, 1768.90477493, 1768.90477493, 1768.90477493, 1805.00487238, 1853.13833565, 1853.13833565, 1889.23843309, 1865.17170146, 1829.07160401, 1792.97150657, 1792.97150657, 1768.90477493, 1768.90477493, 1829.07160401, 1792.97150657, 1732.80467749, 1744.83804330, 1792.97150657, 1817.03823820, 1841.10496983, 1805.00487238, 1805.00487238, 1877.20506728, 1829.07160401, 1805.00487238, 1901.27179891, 1901.27179891, 1937.37189636, 1925.33853054, 1937.37189636, 1949.40526217, 1937.37189636, 1949.40526217, 1949.40526217, 1901.27179891, 1901.27179891, 1913.30516472, 1961.43862799, 1985.50535962, 2045.67218870, 2033.63882288, 2009.57209125, 1985.50535962, 1997.53872544, 2009.57209125, 2021.60545707, 2045.67218870, 2069.73892033, 2057.70555452, 2033.63882288, 2093.80565196, 2141.93911523, 2214.13931012, 2226.17267594, 2202.10594431, 2238.20604175, 2238.20604175, 2226.17267594, 2286.33950502, 2274.30613920, 2346.50633410, 2370.57306573, 2346.50633410, 2370.57306573, 2370.57306573, 2334.47296828, 2286.33950502, 2190.07257849, 2178.03921267, 2021.60545707, 2081.77228615], 
							[753.28870007, 740.05199768, 728.01863186, 642.58173457, 655.81843697, 613.70165661, 542.70479830, 529.46809590, 529.46809590, 519.84140325, 561.95818360, 554.73816411, 554.73816411, 564.36485676, 541.50146171, 553.53482753, 505.40136427, 451.25121810, 494.57133503, 492.16466187, 484.94464238, 457.26790100, 451.25121810, 450.04788151, 406.72776458, 460.87791075, 517.43473008, 540.29812513, 531.87476906, 547.51814462, 552.33149095, 495.77467161, 487.35131554, 487.35131554, 465.69125707, 457.26790100, 459.67457417, 464.48792049, 483.74130580, 482.53796922, 468.09793024, 462.08124733, 470.50460340, 458.47123758, 435.60784253, 428.38782305, 429.59115963, 445.23453519, 457.26790100, 470.50460340, 453.65789126, 463.28458391, 492.16466187, 477.72462289, 496.97800820, 551.12815437, 577.60155916, 564.36485676, 555.94150069, 626.93835901, 611.29498345, 647.39508089, 653.41176380, 634.15837850, 629.34503217, 644.98840773, 643.78507115, 643.78507115, 625.73502243, 592.04159814, 595.65160789, 595.65160789, 601.66829079, 566.77152993, 578.80489574, 586.02491523, 596.85494447, 584.82157865, 589.63492498, 590.83826156, 581.21156891, 570.38153967, 552.33149095, 531.87476906, 540.29812513, 515.02805692, 499.38468136, 464.48792049, 415.15112065, 433.20116937, 399.50774509, 365.81432080, 410.33777432, 391.08438902, 419.96446697, 386.27104269, 409.13443774, 412.74444748, 456.06456442, 447.64120835, 421.16780356, 493.36799845, 487.35131554, 434.40450595, 459.67457417, 463.28458391, 476.52128631, 511.41804717, 516.23139350, 529.46809590, 541.50146171, 563.16152018, 589.63492498, 576.39822258, 569.17820309, 566.77152993, 547.51814462, 571.58487625, 592.04159814, 602.87162738, 630.54836875, 598.05828105, 589.63492498, 567.97486651, 607.68497370, 601.66829079, 589.63492498, 577.60155916, 580.00823233, 573.99154942, 586.02491523, 573.99154942, 559.55151044, 529.46809590, 536.68811539, 513.82472034, 545.11147146, 545.11147146, 541.50146171, 541.50146171, 549.92481779, 522.24807641, 537.89145197, 499.38468136, 511.41804717, 505.40136427, 494.57133503, 540.29812513, 531.87476906, 533.07810564, 527.06142274, 507.80803743, 511.41804717, 502.99469110, 516.23139350, 512.62138376, 488.55465212, 462.08124733, 425.98114988, 447.64120835, 428.38782305, 440.42118886, 433.20116937, 446.43787177, 480.13129605, 466.89459366, 480.13129605, 463.28458391, 452.45455468, 427.18448646, 433.20116937, 427.18448646, 401.91441825, 419.96446697, 391.08438902, 447.64120835, 460.87791075, 513.82472034, 489.75798871, 475.31794973, 474.11461315, 516.23139350, 510.21471059, 537.89145197, 571.58487625, 594.44827130, 530.67143248, 530.67143248, 529.46809590, 565.56819335, 577.60155916, 565.56819335, 581.21156891, 583.61824207, 588.43158840, 583.61824207, 620.92167610, 629.34503217, 643.78507115, 623.32834926, 649.80175406, 654.61510038, 637.76838824, 643.78507115, 634.15837850, 667.85180278, 649.80175406, 678.68183202, 652.20842722, 630.54836875, 673.86848569, 649.80175406, 673.86848569, 696.73188074, 703.95190023, 706.35857339, 708.76524656, 737.64532451, 749.67869033, 729.22196844, 736.44198793, 720.79861237, 735.23865135, 699.13855390, 724.40862212, 740.05199768, 728.01863186, 694.32520758, 679.88516860, 685.90185151, 696.73188074, 663.03845645, 641.37839799, 654.61510038, 687.10518809, 631.75170533, 637.76838824, 695.52854416, 684.69851492, 703.95190023, 697.93521732, 679.88516860, 706.35857339, 658.22511013, 671.46181253, 685.90185151, 672.66514911, 672.66514911, 653.41176380, 604.07496396, 623.32834926, 587.22825181, 559.55151044, 542.70479830, 487.35131554, 507.80803743, 499.38468136, 535.48477881, 502.99469110, 524.65474957, 518.63806666, 513.82472034, 517.43473008, 493.36799845, 511.41804717, 495.77467161, 466.89459366, 512.62138376, 541.50146171, 512.62138376, 592.04159814, 599.26161763, 607.68497370, 595.65160789, 595.65160789, 619.71833952, 592.04159814, 605.27830054, 601.66829079, 555.94150069, 557.14483728, 543.90813488, 570.38153967, 601.66829079, 558.34817386, 523.45141299, 512.62138376, 487.35131554, 475.31794973, 472.91127656, 419.96446697], 
							[835.11558762, 813.45552915, 792.99880727, 770.13541222, 743.66200742, 683.49517834, 708.76524656, 679.88516860, 652.20842722, 647.39508089, 684.69851492, 673.86848569, 684.69851492, 688.30852467, 661.83511987, 631.75170533, 598.05828105, 554.73816411, 580.00823233, 593.24493472, 583.61824207, 559.55151044, 559.55151044, 519.84140325, 465.69125707, 561.95818360, 648.59841748, 642.58173457, 622.12501268, 607.68497370, 614.90499319, 586.02491523, 595.65160789, 600.46495421, 578.80489574, 569.17820309, 565.56819335, 555.94150069, 572.78821284, 572.78821284, 582.41490549, 572.78821284, 582.41490549, 566.77152993, 548.72148120, 540.29812513, 530.67143248, 524.65474957, 551.12815437, 564.36485676, 535.48477881, 536.68811539, 525.85808615, 519.84140325, 565.56819335, 586.02491523, 590.83826156, 607.68497370, 663.03845645, 655.81843697, 623.32834926, 659.42844671, 720.79861237, 715.98526604, 709.96858314, 682.29184176, 703.95190023, 682.29184176, 625.73502243, 613.70165661, 598.05828105, 613.70165661, 601.66829079, 553.53482753, 588.43158840, 593.24493472, 606.48163712, 578.80489574, 601.66829079, 617.31166635, 612.49832003, 595.65160789, 590.83826156, 565.56819335, 577.60155916, 589.63492498, 589.63492498, 558.34817386, 524.65474957, 529.46809590, 533.07810564, 481.33463264, 525.85808615, 498.18134478, 535.48477881, 505.40136427, 517.43473008, 540.29812513, 563.16152018, 575.19488600, 559.55151044, 616.10832977, 601.66829079, 566.77152993, 612.49832003, 610.09164687, 611.29498345, 606.48163712, 586.02491523, 560.75484702, 593.24493472, 588.43158840, 581.21156891, 580.00823233, 571.58487625, 613.70165661, 607.68497370, 604.07496396, 619.71833952, 616.10832977, 637.76838824, 608.88831028, 601.66829079, 595.65160789, 629.34503217, 613.70165661, 632.95504192, 638.97172482, 651.00509064, 666.64846620, 677.47849543, 683.49517834, 637.76838824, 622.12501268, 619.71833952, 635.36171508, 600.46495421, 619.71833952, 599.26161763, 575.19488600, 583.61824207, 553.53482753, 545.11147146, 535.48477881, 547.51814462, 553.53482753, 565.56819335, 583.61824207, 623.32834926, 589.63492498, 599.26161763, 626.93835901, 629.34503217, 637.76838824, 624.53168584, 619.71833952, 586.02491523, 560.75484702, 549.92481779, 522.24807641, 493.36799845, 521.04473983, 519.84140325, 509.01137401, 548.72148120, 563.16152018, 557.14483728, 511.41804717, 513.82472034, 487.35131554, 505.40136427, 469.30126682, 454.86122784, 468.09793024, 427.18448646, 472.91127656, 521.04473983, 517.43473008, 523.45141299, 506.60470085, 530.67143248, 566.77152993, 557.14483728, 555.94150069, 541.50146171, 580.00823233, 517.43473008, 531.87476906, 517.43473008, 500.58801794, 483.74130580, 457.26790100, 492.16466187, 464.48792049, 481.33463264, 464.48792049, 481.33463264, 484.94464238, 488.55465212, 481.33463264, 499.38468136, 541.50146171, 539.09478855, 546.31480804, 523.45141299, 523.45141299, 515.02805692, 506.60470085, 488.55465212, 501.79135452, 499.38468136, 493.36799845, 492.16466187, 531.87476906, 525.85808615, 536.68811539, 539.09478855, 533.07810564, 528.26475932, 522.24807641, 531.87476906, 522.24807641, 547.51814462, 523.45141299, 525.85808615, 533.07810564, 529.46809590, 518.63806666, 502.99469110, 499.38468136, 481.33463264, 482.53796922, 471.70793998, 474.11461315, 477.72462289, 464.48792049, 459.67457417, 466.89459366, 463.28458391, 477.72462289, 469.30126682, 451.25121810, 480.13129605, 448.84454493, 445.23453519, 454.86122784, 444.03119861, 431.99783279, 387.47437927, 373.03434029, 373.03434029, 358.59430131, 353.78095499, 345.35759892, 290.00411616, 312.86751121, 283.98743325, 287.59744300, 249.09067239, 222.61726759, 244.27732606, 228.63395050, 220.21059443, 205.77055545, 210.58390178, 228.63395050, 204.56721887, 193.73718964, 211.78723836, 200.95720913, 214.19391152, 205.77055545, 216.60058469, 226.22727734, 228.63395050, 228.63395050, 234.65063341, 231.04062366, 223.82060418, 210.58390178, 192.53385305, 180.50048724, 186.51717015, 168.46712142, 152.82374586, 158.84042877, 150.41707270, 160.04376535, 139.58704346, 150.41707270, 140.79038005], 
							[1188.89654261, 1125.11970378, 1076.98624052, 985.53266032, 983.12598716, 914.53580201, 866.40233874, 938.60253364, 956.65258236, 961.46592869, 993.95601639, 972.29595792, 926.56916782, 937.39919706, 910.92579226, 888.06239721, 866.40233874, 795.40548043, 851.96229976, 843.53894369, 901.29909961, 890.46907037, 866.40233874, 818.26887548, 768.93207563, 823.08222181, 896.48575328, 884.45238747, 862.79232900, 900.09576303, 880.84237772, 845.94561686, 825.48889497, 841.13227053, 813.45552915, 817.06553890, 818.26887548, 765.32206589, 833.91225104, 867.60567532, 826.69223155, 814.65886573, 773.74542196, 807.43884625, 794.20214385, 780.96544145, 786.98212436, 782.16877803, 782.16877803, 799.01549017, 772.54208538, 782.16877803, 772.54208538, 760.50871956, 794.20214385, 776.15209512, 872.41902165, 1012.00606512, 1009.59939195, 1143.16975251, 1115.49301113, 1311.63687393, 1275.53677648, 1359.77033719, 1347.73697138, 1311.63687393, 1311.63687393, 1227.40331322, 1180.47318654, 1203.33658159, 1275.53677648, 1275.53677648, 1287.57014230, 1239.43667904, 1287.57014230, 1215.36994740, 1215.36994740, 1192.50655235, 1175.65984021, 1188.89654261, 1143.16975251, 1104.66298190, 1080.59625027, 1068.56288445, 1091.42627950, 1113.08633797, 1092.62961608, 1042.08947966, 972.29595792, 1008.39605537, 986.73599690, 904.90910935, 930.17917757, 939.80587022, 1032.46278700, 977.10930425, 1046.90282598, 1091.42627950, 1116.69634771, 1122.71303062, 1057.73285522, 1089.01960634, 1057.73285522, 1056.52951863, 1050.51283573, 1055.32618205, 1107.06965506, 1089.01960634, 1107.06965506, 1098.64629899, 1081.79958685, 1086.61293317, 1072.17289419, 1064.95287471, 1078.18957710, 1096.23962583, 1117.89968430, 1149.18643542, 1153.99978174, 1157.60979149, 1169.64315730, 1170.84649388, 1163.62647440, 1137.15306960, 1131.13638669, 1107.06965506, 1119.10302088, 1095.03628924, 1107.06965506, 1137.15306960, 1105.86631848, 1126.32304037, 1125.11970378, 1110.67966481, 1149.18643542, 1161.21980123, 1147.98309883, 1178.06651337, 1167.23648414, 1155.20311832, 1132.33972327, 1110.67966481, 1127.52637695, 1095.03628924, 1095.03628924, 1096.23962583, 1066.15621129, 1096.23962583, 1096.23962583, 1081.79958685, 1108.27299164, 1067.35954787, 1007.19271879, 980.71931399, 980.71931399, 948.22922629, 914.53580201, 878.43570456, 876.02903140, 861.58899242, 836.31892420, 861.58899242, 910.92579226, 872.41902165, 908.51911910, 904.90910935, 934.99252389, 878.43570456, 872.41902165, 845.94561686, 820.67554864, 759.30538298, 736.44198793, 794.20214385, 753.28870007, 803.82883650, 872.41902165, 850.75896318, 860.38565584, 932.58585073, 916.94247517, 1002.37937246, 978.31264083, 977.10930425, 957.85591894, 1043.29281624, 974.70263109, 969.88928476, 992.75267981, 983.12598716, 968.68594818, 945.82255313, 996.36268955, 995.15935297, 993.95601639, 1007.19271879, 1058.93619180, 1050.51283573, 1031.25945042, 1076.98624052, 1116.69634771, 1197.31989868, 1203.33658159, 1178.06651337, 1215.36994740, 1263.50341067, 1239.43667904, 1215.36994740, 1162.42313781, 1168.43982072, 1116.69634771, 1122.71303062, 1199.72657184, 1239.43667904, 1227.40331322, 1323.67023975, 1263.50341067, 1263.50341067, 1263.50341067, 1239.43667904, 1239.43667904, 1202.13324501, 1215.36994740, 1192.50655235, 1192.50655235, 1161.21980123, 1141.96641593, 1156.40645491, 1145.57642567, 1109.47632822, 1089.01960634, 1096.23962583, 1080.59625027, 1107.06965506, 1145.57642567, 1074.57956736, 1103.45964532, 1131.13638669, 1108.27299164, 1121.50969404, 1087.81626976, 1075.78290394, 1116.69634771, 1098.64629899, 1081.79958685, 1093.83295266, 1109.47632822, 1073.37623078, 1004.78604563, 1015.61607486, 1018.02274802, 979.51597741, 934.99252389, 939.80587022, 791.79547068, 853.16563635, 848.35229002, 903.70577277, 943.41587996, 962.66926527, 1001.17603588, 1004.78604563, 942.21254338, 919.34914833, 902.50243619, 888.06239721, 924.16249466, 956.65258236, 971.09262134, 986.73599690, 1063.74953812, 1046.90282598, 1057.73285522, 1016.81941144, 975.90596767, 1022.83609435, 974.70263109, 980.71931399, 963.87260185, 906.11244594, 902.50243619, 879.63904114, 874.82569481, 809.84551941, 811.04885599, 824.28555839, 839.92893395, 863.99566558, 868.80901191, 836.31892420, 811.04885599], 
							[697.93521732, 702.74856365, 687.10518809, 667.85180278, 683.49517834, 658.22511013, 637.76838824, 563.16152018, 533.07810564, 529.46809590, 586.02491523, 559.55151044, 573.99154942, 577.60155916, 541.50146171, 522.24807641, 493.36799845, 428.38782305, 469.30126682, 468.09793024, 471.70793998, 466.89459366, 482.53796922, 475.31794973, 453.65789126, 477.72462289, 518.63806666, 524.65474957, 510.21471059, 504.19802769, 512.62138376, 477.72462289, 454.86122784, 482.53796922, 459.67457417, 445.23453519, 459.67457417, 436.81117912, 470.50460340, 475.31794973, 457.26790100, 453.65789126, 454.86122784, 454.86122784, 434.40450595, 424.77781330, 423.57447672, 416.35445723, 436.81117912, 463.28458391, 436.81117912, 463.28458391, 478.92795947, 474.11461315, 523.45141299, 506.60470085, 507.80803743, 523.45141299, 505.40136427, 537.89145197, 506.60470085, 546.31480804, 539.09478855, 614.90499319, 628.14169559, 646.19174431, 643.78507115, 625.73502243, 604.07496396, 599.26161763, 590.83826156, 612.49832003, 601.66829079, 592.04159814, 616.10832977, 618.51500294, 611.29498345, 616.10832977, 625.73502243, 625.73502243, 595.65160789, 566.77152993, 565.56819335, 529.46809590, 519.84140325, 525.85808615, 523.45141299, 463.28458391, 401.91441825, 410.33777432, 391.08438902, 351.37428182, 421.16780356, 387.47437927, 412.74444748, 374.23767687, 391.08438902, 409.13443774, 413.94778407, 433.20116937, 404.32109141, 469.30126682, 454.86122784, 403.11775483, 441.62452544, 447.64120835, 451.25121810, 460.87791075, 434.40450595, 415.15112065, 447.64120835, 463.28458391, 487.35131554, 469.30126682, 470.50460340, 475.31794973, 483.74130580, 509.01137401, 516.23139350, 559.55151044, 567.97486651, 535.48477881, 522.24807641, 505.40136427, 525.85808615, 516.23139350, 522.24807641, 513.82472034, 527.06142274, 558.34817386, 541.50146171, 534.28144222, 530.67143248, 522.24807641, 517.43473008, 518.63806666, 517.43473008, 547.51814462, 542.70479830, 546.31480804, 541.50146171, 535.48477881, 551.12815437, 518.63806666, 511.41804717, 502.99469110, 483.74130580, 540.29812513, 537.89145197, 546.31480804, 537.89145197, 531.87476906, 522.24807641, 516.23139350, 477.72462289, 488.55465212, 457.26790100, 433.20116937, 445.23453519, 459.67457417, 428.38782305, 421.16780356, 391.08438902, 418.76113039, 438.01451570, 442.82786202, 463.28458391, 427.18448646, 410.33777432, 404.32109141, 397.10107192, 393.49106218, 361.00097448, 394.69439876, 361.00097448, 421.16780356, 430.79449621, 433.20116937, 421.16780356, 421.16780356, 430.79449621, 453.65789126, 466.89459366, 498.18134478, 504.19802769, 528.26475932, 484.94464238, 489.75798871, 527.06142274, 483.74130580, 481.33463264, 452.45455468, 454.86122784, 438.01451570, 470.50460340, 465.69125707, 481.33463264, 533.07810564, 515.02805692, 519.84140325, 506.60470085, 530.67143248, 547.51814462, 534.28144222, 542.70479830, 581.21156891, 553.53482753, 577.60155916, 637.76838824, 620.92167610, 711.17191972, 719.59527579, 667.85180278, 717.18860263, 743.66200742, 728.01863186, 722.00194895, 701.54522707, 684.69851492, 665.44512962, 669.05513936, 661.83511987, 667.85180278, 625.73502243, 623.32834926, 654.61510038, 653.41176380, 638.97172482, 613.70165661, 616.10832977, 588.43158840, 570.38153967, 537.89145197, 563.16152018, 560.75484702, 513.82472034, 492.16466187, 559.55151044, 565.56819335, 543.90813488, 533.07810564, 519.84140325, 545.11147146, 509.01137401, 511.41804717, 547.51814462, 523.45141299, 527.06142274, 511.41804717, 487.35131554, 498.18134478, 458.47123758, 421.16780356, 421.16780356, 364.61098422, 399.50774509, 381.45769636, 405.52442800, 386.27104269, 401.91441825, 409.13443774, 403.11775483, 389.88105243, 356.18762815, 373.03434029, 348.96760866, 320.08753070, 306.85082830, 321.29086728, 326.10421361, 346.56093550, 367.01765738, 397.10107192, 377.84768662, 395.89773534, 407.93110116, 403.11775483, 393.49106218, 404.32109141, 379.05102320, 365.81432080, 371.83100371, 383.86436953, 305.64749172, 322.49420387, 298.42747223, 273.15740402, 270.75073086, 297.22413565, 270.75073086, 271.95406744], 
							[648.59841748, 602.87162738, 590.83826156, 559.55151044, 589.63492498, 577.60155916, 601.66829079, 595.65160789, 602.87162738, 584.82157865, 614.90499319, 587.22825181, 596.85494447, 592.04159814, 582.41490549, 567.97486651, 548.72148120, 542.70479830, 577.60155916, 565.56819335, 563.16152018, 535.48477881, 547.51814462, 529.46809590, 493.36799845, 565.56819335, 613.70165661, 596.85494447, 587.22825181, 572.78821284, 589.63492498, 537.89145197, 524.65474957, 536.68811539, 533.07810564, 528.26475932, 534.28144222, 535.48477881, 541.50146171, 554.73816411, 552.33149095, 572.78821284, 589.63492498, 588.43158840, 583.61824207, 576.39822258, 567.97486651, 571.58487625, 589.63492498, 578.80489574, 553.53482753, 558.34817386, 564.36485676, 546.31480804, 601.66829079, 602.87162738, 634.15837850, 649.80175406, 691.91853441, 693.12187099, 663.03845645, 756.89870982, 741.25533426, 746.06868058, 740.05199768, 728.01863186, 718.39193921, 711.17191972, 682.29184176, 664.24179304, 661.83511987, 677.47849543, 712.37525630, 697.93521732, 705.15523681, 697.93521732, 683.49517834, 672.66514911, 685.90185151, 701.54522707, 700.34189048, 673.86848569, 695.52854416, 681.08850518, 697.93521732, 678.68183202, 689.51186125, 699.13855390, 691.91853441, 647.39508089, 622.12501268, 592.04159814, 612.49832003, 632.95504192, 655.81843697, 638.97172482, 649.80175406, 655.81843697, 677.47849543, 694.32520758, 659.42844671, 691.91853441, 695.52854416, 649.80175406, 697.93521732, 696.73188074, 711.17191972, 737.64532451, 722.00194895, 697.93521732, 715.98526604, 744.86534400, 740.05199768, 729.22196844, 740.05199768, 761.71205615, 754.49203666, 790.59213410, 823.08222181, 845.94561686, 843.53894369, 825.48889497, 805.03217308, 835.11558762, 853.16563635, 841.13227053, 826.69223155, 818.26887548, 827.89556813, 825.48889497, 812.25219257, 832.70891446, 819.47221206, 803.82883650, 835.11558762, 844.74228027, 868.80901191, 877.23236798, 857.97898267, 866.40233874, 850.75896318, 839.92893395, 845.94561686, 824.28555839, 842.33560711, 833.91225104, 837.52226079, 848.35229002, 850.75896318, 850.75896318, 824.28555839, 776.15209512, 776.15209512, 748.47535375, 754.49203666, 764.11872931, 765.32206589, 752.08536349, 749.67869033, 725.61195870, 723.20528553, 767.72873905, 725.61195870, 728.01863186, 741.25533426, 717.18860263, 709.96858314, 690.71519783, 709.96858314, 703.95190023, 715.98526604, 690.71519783, 684.69851492, 709.96858314, 717.18860263, 761.71205615, 817.06553890, 802.62549992, 789.38879752, 806.23550966, 799.01549017, 866.40233874, 886.85906063, 900.09576303, 913.33246543, 927.77250440, 872.41902165, 932.58585073, 922.95915808, 903.70577277, 932.58585073, 908.51911910, 936.19586048, 890.46907037, 866.40233874, 877.23236798, 913.33246543, 932.58585073, 896.48575328, 884.45238747, 894.07908012, 954.24590920, 924.16249466, 912.12912884, 918.14581175, 926.56916782, 965.07593843, 941.00920680, 922.95915808, 938.60253364, 913.33246543, 902.50243619, 936.19586048, 957.85591894, 962.66926527, 973.49929450, 987.93933348, 980.71931399, 966.27927501, 939.80587022, 960.26259211, 914.53580201, 890.46907037, 889.26573379, 947.02588971, 908.51911910, 919.34914833, 913.33246543, 908.51911910, 859.18231925, 850.75896318, 842.33560711, 836.31892420, 854.36897293, 876.02903140, 838.72559737, 823.08222181, 860.38565584, 855.57230951, 890.46907037, 877.23236798, 884.45238747, 901.29909961, 849.55562660, 803.82883650, 790.59213410, 788.18546094, 807.43884625, 788.18546094, 800.21882676, 808.64218283, 808.64218283, 820.67554864, 830.30224130, 790.59213410, 800.21882676, 792.99880727, 831.50557788, 819.47221206, 860.38565584, 836.31892420, 815.86220232, 818.26887548, 829.09890471, 784.57545120, 835.11558762, 835.11558762, 854.36897293, 836.31892420, 839.92893395, 860.38565584, 861.58899242, 843.53894369, 820.67554864, 794.20214385, 825.48889497, 801.42216334, 780.96544145, 786.98212436, 765.32206589, 776.15209512, 792.99880727, 782.16877803, 754.49203666, 719.59527579, 719.59527579, 682.29184176, 665.44512962, 644.98840773, 599.26161763, 614.90499319], 
							[718.39193921, 712.37525630, 637.76838824, 679.88516860, 648.59841748, 612.49832003, 661.83511987, 643.78507115, 658.22511013, 655.81843697, 688.30852467, 676.27515885, 685.90185151, 693.12187099, 709.96858314, 699.13855390, 634.15837850, 625.73502243, 644.98840773, 637.76838824, 643.78507115, 619.71833952, 636.56505166, 625.73502243, 613.70165661, 607.68497370, 713.57859288, 684.69851492, 632.95504192, 661.83511987, 647.39508089, 611.29498345, 614.90499319, 618.51500294, 613.70165661, 595.65160789, 613.70165661, 577.60155916, 592.04159814, 637.76838824, 584.82157865, 602.87162738, 619.71833952, 592.04159814, 588.43158840, 563.16152018, 570.38153967, 571.58487625, 607.68497370, 599.26161763, 577.60155916, 590.83826156, 612.49832003, 586.02491523, 681.08850518, 653.41176380, 661.83511987, 667.85180278, 709.96858314, 749.67869033, 705.15523681, 768.93207563, 759.30538298, 786.98212436, 782.16877803, 771.33874880, 779.76210487, 737.64532451, 758.10204640, 759.30538298, 777.35543171, 790.59213410, 786.98212436, 761.71205615, 803.82883650, 785.77878778, 792.99880727, 776.15209512, 748.47535375, 742.45867084, 738.84866109, 688.30852467, 714.78192946, 720.79861237, 736.44198793, 715.98526604, 743.66200742, 730.42530502, 712.37525630, 707.56190997, 640.17506140, 601.66829079, 671.46181253, 643.78507115, 671.46181253, 630.54836875, 652.20842722, 684.69851492, 744.86534400, 765.32206589, 767.72873905, 800.21882676, 799.01549017, 735.23865135, 844.74228027, 854.36897293, 854.36897293, 878.43570456, 866.40233874, 808.64218283, 827.89556813, 826.69223155, 818.26887548, 819.47221206, 821.87888522, 835.11558762, 863.99566558, 880.84237772, 866.40233874, 924.16249466, 936.19586048, 930.17917757, 890.46907037, 919.34914833, 919.34914833, 925.36583124, 874.82569481, 931.38251415, 927.77250440, 904.90910935, 926.56916782, 947.02588971, 908.51911910, 901.29909961, 912.12912884, 897.68908986, 918.14581175, 944.61921655, 945.82255313, 947.02588971, 930.17917757, 926.56916782, 921.75582150, 894.07908012, 883.24905089, 865.19900216, 879.63904114, 912.12912884, 950.63589945, 928.97584099, 877.23236798, 805.03217308, 832.70891446, 801.42216334, 764.11872931, 803.82883650, 765.32206589, 730.42530502, 729.22196844, 752.08536349, 726.81529528, 743.66200742, 671.46181253, 697.93521732, 761.71205615, 737.64532451, 740.05199768, 689.51186125, 687.10518809, 619.71833952, 628.14169559, 592.04159814, 567.97486651, 613.70165661, 588.43158840, 690.71519783, 665.44512962, 675.07182227, 708.76524656, 697.93521732, 713.57859288, 777.35543171, 779.76210487, 791.79547068, 776.15209512, 801.42216334, 743.66200742, 731.62864161, 746.06868058, 734.03531477, 734.03531477, 738.84866109, 790.59213410, 803.82883650, 802.62549992, 842.33560711, 855.57230951, 908.51911910, 904.90910935, 896.48575328, 902.50243619, 954.24590920, 894.07908012, 938.60253364, 938.60253364, 937.39919706, 919.34914833, 924.16249466, 883.24905089, 927.77250440, 916.94247517, 907.31578252, 959.05925553, 979.51597741, 971.09262134, 983.12598716, 998.76936272, 1027.64944068, 1010.80272853, 974.70263109, 999.97269930, 959.05925553, 984.32932374, 967.48261160, 980.71931399, 991.54934323, 997.56602614, 983.12598716, 945.82255313, 949.43256287, 916.94247517, 889.26573379, 862.79232900, 896.48575328, 909.72245568, 860.38565584, 874.82569481, 907.31578252, 916.94247517, 912.12912884, 920.55248491, 912.12912884, 959.05925553, 909.72245568, 912.12912884, 926.56916782, 889.26573379, 880.84237772, 874.82569481, 862.79232900, 867.60567532, 859.18231925, 832.70891446, 844.74228027, 675.07182227, 790.59213410, 750.88202691, 817.06553890, 799.01549017, 857.97898267, 868.80901191, 832.70891446, 807.43884625, 782.16877803, 855.57230951, 807.43884625, 805.03217308, 820.67554864, 836.31892420, 889.26573379, 950.63589945, 920.55248491, 938.60253364, 962.66926527, 916.94247517, 902.50243619, 871.21568507, 839.92893395, 879.63904114, 819.47221206, 808.64218283, 831.50557788, 863.99566558, 812.25219257, 826.69223155, 776.15209512, 737.64532451, 749.67869033, 741.25533426, 684.69851492, 714.78192946], 
							[649.80175406, 631.75170533, 611.29498345, 578.80489574, 593.24493472, 543.90813488, 552.33149095, 571.58487625, 524.65474957, 589.63492498, 607.68497370, 589.63492498, 589.63492498, 613.70165661, 613.70165661, 584.82157865, 553.53482753, 511.41804717, 608.88831028, 583.61824207, 553.53482753, 511.41804717, 552.33149095, 547.51814462, 529.46809590, 553.53482753, 605.27830054, 604.07496396, 592.04159814, 601.66829079, 604.07496396, 543.90813488, 537.89145197, 560.75484702, 551.12815437, 542.70479830, 569.17820309, 547.51814462, 571.58487625, 571.58487625, 566.77152993, 565.56819335, 577.60155916, 573.99154942, 559.55151044, 549.92481779, 561.95818360, 539.09478855, 582.41490549, 584.82157865, 565.56819335, 565.56819335, 582.41490549, 553.53482753, 613.70165661, 592.04159814, 595.65160789, 619.71833952, 660.63178329, 693.12187099, 672.66514911, 722.00194895, 697.93521732, 697.93521732, 709.96858314, 709.96858314, 717.18860263, 695.52854416, 664.24179304, 678.68183202, 663.03845645, 695.52854416, 688.30852467, 669.05513936, 734.03531477, 734.03531477, 758.10204640, 718.39193921, 742.45867084, 732.83197819, 697.93521732, 693.12187099, 689.51186125, 682.29184176, 690.71519783, 697.93521732, 691.91853441, 673.86848569, 655.81843697, 628.14169559, 601.66829079, 569.17820309, 643.78507115, 625.73502243, 635.36171508, 605.27830054, 622.12501268, 643.78507115, 690.71519783, 683.49517834, 685.90185151, 728.01863186, 712.37525630, 655.81843697, 734.03531477, 776.15209512, 782.16877803, 783.37211461, 743.66200742, 770.13541222, 796.60881701, 792.99880727, 794.20214385, 760.50871956, 768.93207563, 770.13541222, 782.16877803, 806.23550966, 835.11558762, 914.53580201, 896.48575328, 860.38565584, 879.63904114, 868.80901191, 888.06239721, 880.84237772, 830.30224130, 855.57230951, 886.85906063, 884.45238747, 883.24905089, 898.89242645, 878.43570456, 848.35229002, 865.19900216, 873.62235823, 890.46907037, 924.16249466, 941.00920680, 874.82569481, 872.41902165, 848.35229002, 859.18231925, 811.04885599, 807.43884625, 838.72559737, 830.30224130, 890.46907037, 898.89242645, 892.87574354, 854.36897293, 782.16877803, 740.05199768, 747.27201717, 687.10518809, 714.78192946, 728.01863186, 703.95190023, 667.85180278, 689.51186125, 655.81843697, 661.83511987, 630.54836875, 667.85180278, 719.59527579, 669.05513936, 689.51186125, 630.54836875, 613.70165661, 617.31166635, 629.34503217, 617.31166635, 588.43158840, 637.76838824, 596.85494447, 673.86848569, 679.88516860, 707.56190997, 702.74856365, 722.00194895, 712.37525630, 770.13541222, 753.28870007, 765.32206589, 770.13541222, 766.52540247, 734.03531477, 741.25533426, 750.88202691, 707.56190997, 732.83197819, 699.13855390, 770.13541222, 782.16877803, 791.79547068, 764.11872931, 818.26887548, 842.33560711, 839.92893395, 820.67554864, 830.30224130, 876.02903140, 859.18231925, 874.82569481, 902.50243619, 868.80901191, 845.94561686, 853.16563635, 859.18231925, 865.19900216, 865.19900216, 847.14895344, 886.85906063, 956.65258236, 950.63589945, 932.58585073, 921.75582150, 938.60253364, 913.33246543, 884.45238747, 900.09576303, 915.73913859, 924.16249466, 900.09576303, 922.95915808, 927.77250440, 932.58585073, 914.53580201, 916.94247517, 913.33246543, 908.51911910, 894.07908012, 845.94561686, 848.35229002, 872.41902165, 836.31892420, 860.38565584, 878.43570456, 867.60567532, 914.53580201, 866.40233874, 868.80901191, 915.73913859, 851.96229976, 826.69223155, 820.67554864, 801.42216334, 783.37211461, 788.18546094, 734.03531477, 754.49203666, 744.86534400, 752.08536349, 758.10204640, 644.98840773, 689.51186125, 693.12187099, 738.84866109, 714.78192946, 738.84866109, 734.03531477, 712.37525630, 691.91853441, 688.30852467, 742.45867084, 705.15523681, 687.10518809, 722.00194895, 740.05199768, 734.03531477, 778.55876829, 782.16877803, 788.18546094, 758.10204640, 764.11872931, 755.69537324, 780.96544145, 759.30538298, 755.69537324, 740.05199768, 702.74856365, 744.86534400, 756.89870982, 734.03531477, 693.12187099, 687.10518809, 673.86848569, 673.86848569, 670.25847594, 611.29498345, 601.66829079], 
							[2658.79624375, 2670.25200801, 2670.25200801, 2555.64623198, 2372.28180367, 1971.17362093, 1719.05054036, 2074.32363271, 2108.70295884, 2005.55294707, 2143.08228498, 2039.93227321, 2062.85583508, 2028.47650895, 2005.55294707, 1948.25005905, 1730.50630462, 1604.44476433, 1845.11208065, 1936.79429480, 1994.09718281, 1845.11208065, 2028.47650895, 2074.32363271, 2028.47650895, 2521.26690584, 2567.11402960, 2865.08423393, 2612.94911999, 2544.19046772, 2601.49335573, 2326.44671328, 2441.05248931, 2441.05248931, 2429.58469169, 2418.12892743, 2441.05248931, 2406.66112981, 2498.35537733, 2578.56979386, 2578.56979386, 2704.63133414, 2761.93422216, 2761.93422216, 2658.79624375, 2578.56979386, 2544.19046772, 2567.11402960, 2761.93422216, 2716.09913177, 2670.25200801, 2727.55489602, 2750.47845790, 2727.55489602, 2899.46356007, 2956.76644808, 2876.53999819, 2968.22221234, 2830.70490780, 2796.31354829, 2842.16067205, 2887.99576244, 2922.37508858, 2830.70490780, 2922.37508858, 3174.51020252, 3185.96596678, 3163.05443826, 3036.98086461, 3071.36019075, 3128.66307876, 3243.26885479, 3300.57174281, 3197.43376440, 3323.49530469, 3415.17751884, 3403.72175458, 3289.11597855, 3312.02750706, 3403.72175458, 3380.79819270, 3334.95106894, 3277.64818093, 3300.57174281, 3346.41886657, 3415.17751884, 3541.23905913, 3529.78329487, 3495.39193537, 3392.25395696, 3472.48040685, 3163.05443826, 3140.13087638, 3140.13087638, 3266.19241667, 3128.66307876, 3048.44866223, 3082.82798837, 3197.43376440, 3438.10108072, 3277.64818093, 3346.41886657, 3300.57174281, 3266.19241667, 3266.19241667, 3243.26885479, 3243.26885479, 3289.11597855, 3231.81309054, 3140.13087638, 3312.02750706, 3243.26885479, 3151.58664064, 3105.75155025, 3140.13087638, 3197.43376440, 3151.58664064, 3220.34529291, 3082.82798837, 3151.58664064, 3105.75155025, 3048.44866223, 3014.06933610, 2979.67797660, 3002.60153848, 2956.76644808, 2991.14577422, 2991.14577422, 3002.60153848, 2991.14577422, 3002.60153848, 2922.37508858, 2761.93422216, 2842.16067205, 2807.78134592, 2819.23711017, 2887.99576244, 2853.61643631, 2819.23711017, 2739.01066028, 2681.71980563, 2452.50825357, 2544.19046772, 2509.81114158, 2601.49335573, 2670.25200801, 2727.55489602, 2761.93422216, 2807.78134592, 2704.63133414, 2704.63133414, 2555.64623198, 2395.20536555, 2739.01066028, 2635.87268187, 2601.49335573, 2532.73470346, 2441.05248931, 2360.82603942, 2314.97891566, 2131.61448736, 2303.52315140, 2120.15872310, 2143.08228498, 2303.52315140, 2360.82603942, 2486.88757970, 2544.19046772, 2383.74960130, 2337.90247754, 2418.12892743, 2292.06738715, 2246.22026339, 2211.84093725, 2097.23516122, 2303.52315140, 2280.59958952, 2292.06738715, 2372.28180367, 2383.74960130, 2383.74960130, 2429.58469169, 2280.59958952, 2326.44671328, 2269.14382527, 2360.82603942, 2257.68806101, 2269.14382527, 2395.20536555, 2349.37027516, 2383.74960130, 2257.68806101, 2292.06738715, 2326.44671328, 2418.12892743, 2475.43181545, 2601.49335573, 2739.01066028, 2716.09913177, 2853.61643631, 2784.85778404, 2887.99576244, 2807.78134592, 2670.25200801, 2739.01066028, 2670.25200801, 2647.34047949, 2624.41691761, 2521.26690584, 2590.03759148, 2475.43181545, 2590.03759148, 2475.43181545, 2719.54067439, 2671.40721112, 2743.60740602, 2827.84096673, 2948.17462489, 2875.97442999, 2827.84096673, 2996.30808815, 2972.24135652, 3080.54164886, 3068.50828305, 3056.47491723, 3068.50828305, 3056.47491723, 2972.24135652, 2803.77423510, 2803.77423510, 2743.60740602, 2731.57404020, 2695.47394276, 2671.40721112, 2743.60740602, 2659.37384531, 2599.20701623, 2611.24038205, 2743.60740602, 2863.94106418, 2827.84096673, 2815.80760092, 2888.00779581, 2779.70750347, 2779.70750347, 2803.77423510, 2803.77423510, 2755.64077184, 2803.77423510, 2623.27374786, 2454.80662644, 2563.10691878, 2551.07355297, 2587.17365041, 2394.63979736, 2310.40623665, 2214.13931012, 2141.93911523, 2153.97248104, 2286.33950502, 2454.80662644, 2382.60643154, 2298.37287083, 2358.53969991, 2527.00682133, 2358.53969991, 2274.30613920, 2153.97248104, 2370.57306573, 2394.63979736, 2502.94008970, 2466.83999225, 2539.04018715, 2551.07355297, 2430.73989481, 2442.77326062, 2587.17365041, 2623.27374786, 2731.57404020, 2635.30711368, 2514.97345552, 2707.50730857, 2755.64077184, 2563.10691878, 2755.64077184, 2827.84096673, 2683.44057694, 2635.30711368, 2611.24038205, 2599.20701623, 2731.57404020], 
							[806.81311122, 739.19762870, 727.72983108, 693.35050495, 693.35050495, 687.62262282, 624.59185267, 624.59185267, 640.63232931, 630.31973480, 663.55589118, 676.16685856, 658.97117881, 672.72531594, 630.31973480, 608.55137604, 566.14579491, 509.98607664, 574.16001654, 610.83771555, 601.66829079, 551.24848803, 573.01684679, 583.32944129, 563.84742203, 587.91415367, 647.51541455, 673.86848569, 664.69906094, 660.11434856, 660.11434856, 633.76127742, 608.55137604, 641.77549906, 632.61810767, 622.29347980, 658.97117881, 634.90444718, 617.70876743, 631.46290455, 634.90444718, 655.52963619, 653.24329668, 656.68483930, 623.44868292, 624.59185267, 618.86397054, 599.38195129, 621.15031005, 655.52963619, 620.00714030, 650.94492381, 648.65858430, 639.48915955, 664.69906094, 670.42694307, 699.07838707, 692.20733519, 710.54618470, 767.84907271, 730.02820395, 768.99224246, 772.43378509, 750.65339296, 774.72012459, 774.72012459, 781.60320984, 774.72012459, 744.92551083, 722.00194895, 718.56040633, 732.31454346, 734.61291633, 713.98772732, 746.06868058, 725.44349158, 741.48396821, 731.17137370, 741.48396821, 755.23810534, 754.09493558, 739.19762870, 738.05445895, 749.51022321, 750.65339296, 727.72983108, 732.31454346, 722.00194895, 694.49367470, 673.86848569, 639.48915955, 609.69454579, 641.77549906, 622.29347980, 647.51541455, 618.86397054, 625.73502243, 640.63232931, 665.84223069, 647.51541455, 632.61810767, 687.62262282, 686.47945306, 632.61810767, 641.77549906, 676.16685856, 704.81830257, 703.66309945, 710.54618470, 691.06416544, 700.23359019, 709.40301494, 734.61291633, 705.96147232, 713.98772732, 733.46974658, 740.34079846, 767.84907271, 788.47426172, 825.15196073, 849.21869236, 810.25465385, 780.44800672, 781.60320984, 801.08522909, 794.20214385, 780.44800672, 790.76060122, 789.61743147, 779.30483697, 778.16166722, 777.01849746, 787.33109197, 764.40753009, 775.86329434, 775.86329434, 775.86329434, 780.44800672, 794.20214385, 793.05897410, 786.18792221, 773.57695484, 788.47426172, 774.72012459, 787.33109197, 780.44800672, 770.13541222, 795.34531360, 812.25219257, 819.47221206, 805.03217308, 782.16877803, 764.11872931, 779.76210487, 747.27201717, 712.37525630, 695.52854416, 641.37839799, 623.32834926, 631.75170533, 588.43158840, 588.43158840, 583.61824207, 622.12501268, 644.98840773, 616.10832977, 640.17506140, 594.44827130, 595.65160789, 599.26161763, 578.80489574, 576.39822258, 566.77152993, 589.63492498, 543.90813488, 600.46495421, 598.05828105, 624.53168584, 632.95504192, 625.73502243, 637.76838824, 691.91853441, 694.32520758, 726.81529528, 723.20528553, 731.62864161, 681.08850518, 701.54522707, 709.96858314, 673.86848569, 683.49517834, 651.00509064, 678.68183202, 690.71519783, 694.32520758, 701.54522707, 717.18860263, 753.28870007, 743.66200742, 753.28870007, 752.08536349, 761.71205615, 760.50871956, 764.11872931, 786.98212436, 768.93207563, 770.13541222, 776.15209512, 767.72873905, 748.47535375, 749.67869033, 770.13541222, 774.94875854, 800.21882676, 806.23550966, 824.28555839, 799.01549017, 838.72559737, 811.04885599, 796.60881701, 801.42216334, 817.06553890, 817.06553890, 801.42216334, 825.48889497, 835.11558762, 826.69223155, 820.67554864, 799.01549017, 788.18546094, 765.32206589, 746.06868058, 731.62864161, 740.05199768, 753.28870007, 718.39193921, 708.76524656, 720.79861237, 742.45867084, 764.11872931, 735.23865135, 741.25533426, 764.11872931, 723.20528553, 762.91539273, 762.91539273, 764.11872931, 776.15209512, 764.11872931, 722.00194895, 746.06868058, 679.88516860, 649.80175406, 648.59841748, 596.85494447, 613.70165661, 590.83826156, 607.68497370, 588.43158840, 598.05828105, 599.26161763, 595.65160789, 592.04159814, 598.05828105, 642.58173457, 606.48163712, 581.21156891, 601.66829079, 624.53168584, 599.26161763, 625.73502243, 623.32834926, 619.71833952, 625.73502243, 616.10832977, 611.29498345, 618.51500294, 599.26161763, 587.22825181, 584.82157865, 581.21156891, 555.94150069, 561.95818360, 506.60470085, 471.70793998, 444.03119861, 444.03119861, 415.15112065, 431.99783279, 418.76113039, 427.18448646], 
							[866.40233874, 829.09890471, 764.11872931, 766.52540247, 777.35543171, 734.03531477, 672.66514911, 688.30852467, 713.57859288, 699.13855390, 738.84866109, 758.10204640, 771.33874880, 829.09890471, 841.13227053, 885.65572405, 820.67554864, 837.52226079, 888.06239721, 848.35229002, 842.33560711, 815.86220232, 826.69223155, 786.98212436, 715.98526604, 825.48889497, 944.61921655, 912.12912884, 886.85906063, 872.41902165, 902.50243619, 867.60567532, 851.96229976, 891.67240696, 872.41902165, 845.94561686, 826.69223155, 830.30224130, 856.77564609, 866.40233874, 888.06239721, 919.34914833, 916.94247517, 897.68908986, 892.87574354, 878.43570456, 880.84237772, 870.01234849, 891.67240696, 889.26573379, 856.77564609, 867.60567532, 868.80901191, 870.01234849, 985.53266032, 974.70263109, 1060.13952838, 1251.47004485, 1179.26984996, 1172.04983047, 1143.16975251, 1198.52323526, 1215.36994740, 1179.26984996, 1215.36994740, 1239.43667904, 1239.43667904, 1227.40331322, 1167.23648414, 1196.11656210, 1179.26984996, 1202.13324501, 1227.40331322, 1227.40331322, 1275.53677648, 1287.57014230, 1335.70360556, 1371.80370301, 1347.73697138, 1395.87043464, 1371.80370301, 1347.73697138, 1359.77033719, 1299.60350811, 1347.73697138, 1347.73697138, 1299.60350811, 1251.47004485, 1203.33658159, 1239.43667904, 1203.33658159, 1140.76307935, 1178.06651337, 1188.89654261, 1215.36994740, 1125.11970378, 1155.20311832, 1179.26984996, 1215.36994740, 1203.33658159, 1227.40331322, 1251.47004485, 1251.47004485, 1215.36994740, 1239.43667904, 1239.43667904, 1263.50341067, 1299.60350811, 1251.47004485, 1227.40331322, 1227.40331322, 1263.50341067, 1215.36994740, 1239.43667904, 1227.40331322, 1287.57014230, 1263.50341067, 1275.53677648, 1287.57014230, 1311.63687393, 1299.60350811, 1263.50341067, 1251.47004485, 1239.43667904, 1227.40331322, 1215.36994740, 1239.43667904, 1251.47004485, 1251.47004485, 1275.53677648, 1263.50341067, 1275.53677648, 1251.47004485, 1251.47004485, 1263.50341067, 1263.50341067, 1251.47004485, 1275.53677648, 1251.47004485, 1227.40331322, 1227.40331322, 1227.40331322, 1227.40331322, 1227.40331322, 1200.92990842, 1188.89654261, 1198.52323526, 1187.69320603, 1203.33658159, 1194.91322552, 1190.09987919, 1155.20311832, 1092.62961608, 1132.33972327, 1110.67966481, 1095.03628924, 1044.49615282, 1016.81941144, 992.75267981, 980.71931399, 972.29595792, 1033.66612358, 955.44924578, 969.88928476, 1034.86946017, 999.97269930, 986.73599690, 938.60253364, 914.53580201, 874.82569481, 890.46907037, 889.26573379, 904.90910935, 961.46592869, 938.60253364, 990.34600665, 974.70263109, 961.46592869, 957.85591894, 967.48261160, 984.32932374, 1057.73285522, 1022.83609435, 1016.81941144, 1025.24276751, 1093.83295266, 1008.39605537, 1068.56288445, 1046.90282598, 1033.66612358, 1040.88614307, 981.92265058, 1034.86946017, 1034.86946017, 1038.47946991, 1056.52951863, 1076.98624052, 1105.86631848, 1137.15306960, 1174.45650363, 1196.11656210, 1263.50341067, 1167.23648414, 1215.36994740, 1197.31989868, 1239.43667904, 1227.40331322, 1227.40331322, 1179.26984996, 1161.21980123, 1167.23648414, 1191.30321577, 1203.33658159, 1263.50341067, 1263.50341067, 1263.50341067, 1299.60350811, 1311.63687393, 1251.47004485, 1263.50341067, 1275.53677648, 1287.57014230, 1299.60350811, 1287.57014230, 1287.57014230, 1311.63687393, 1287.57014230, 1251.47004485, 1203.33658159, 1203.33658159, 1196.11656210, 1197.31989868, 1186.48986945, 1174.45650363, 1200.92990842, 1151.59310858, 1125.11970378, 1161.21980123, 1181.67652312, 1203.33658159, 1172.04983047, 1164.82981098, 1215.36994740, 1151.59310858, 1113.08633797, 1119.10302088, 1097.44296241, 1001.17603588, 1044.49615282, 980.71931399, 986.73599690, 934.99252389, 937.39919706, 955.44924578, 789.38879752, 872.41902165, 842.33560711, 868.80901191, 871.21568507, 878.43570456, 937.39919706, 934.99252389, 933.78918731, 941.00920680, 1022.83609435, 995.15935297, 1034.86946017, 1001.17603588, 1002.37937246, 1028.85277726, 1048.10616256, 1058.93619180, 1141.96641593, 1128.72971353, 1089.01960634, 1103.45964532, 1113.08633797, 1108.27299164, 1086.61293317, 1058.93619180, 1019.22608460, 986.73599690, 1045.69948940, 892.87574354, 943.41587996, 891.67240696, 947.02588971, 955.44924578, 979.51597741, 950.63589945, 936.19586048], 
							[1275.53677648, 1239.43667904, 1202.13324501, 1215.36994740, 1251.47004485, 1143.16975251, 1103.45964532, 1043.29281624, 1068.56288445, 1093.83295266, 1135.94973302, 1120.30635746, 1116.69634771, 1131.13638669, 1089.01960634, 1046.90282598, 1004.78604563, 972.29595792, 1052.91950889, 1107.06965506, 1119.10302088, 1026.44610409, 1134.74639644, 1036.07279675, 1076.98624052, 1105.86631848, 1149.18643542, 1203.33658159, 1174.45650363, 1215.36994740, 1227.40331322, 1172.04983047, 1155.20311832, 1135.94973302, 1114.28967455, 1084.20626001, 1076.98624052, 1034.86946017, 1091.42627950, 1128.72971353, 1054.12284547, 1119.10302088, 1119.10302088, 1107.06965506, 1061.34286496, 1044.49615282, 1087.81626976, 1051.71617231, 1109.47632822, 1113.08633797, 1092.62961608, 1101.05297215, 1113.08633797, 1076.98624052, 1095.03628924, 1092.62961608, 1078.18957710, 1107.06965506, 1275.53677648, 1251.47004485, 1191.30321577, 1251.47004485, 1239.43667904, 1227.40331322, 1175.65984021, 1239.43667904, 1200.92990842, 1190.09987919, 1199.72657184, 1185.28653286, 1143.16975251, 1137.15306960, 1147.98309883, 1107.06965506, 1109.47632822, 1181.67652312, 1185.28653286, 1200.92990842, 1215.36994740, 1168.43982072, 1197.31989868, 1133.54305986, 1149.18643542, 1140.76307935, 1162.42313781, 1122.71303062, 1113.08633797, 1098.64629899, 1064.95287471, 978.31264083, 999.97269930, 939.80587022, 998.76936272, 962.66926527, 969.88928476, 969.88928476, 986.73599690, 1004.78604563, 1095.03628924, 1143.16975251, 1147.98309883, 1215.36994740, 1215.36994740, 1186.48986945, 1251.47004485, 1203.33658159, 1287.57014230, 1323.67023975, 1323.67023975, 1275.53677648, 1287.57014230, 1323.67023975, 1263.50341067, 1287.57014230, 1263.50341067, 1347.73697138, 1371.80370301, 1371.80370301, 1480.10399535, 1456.03726372, 1504.17072698, 1395.87043464, 1371.80370301, 1407.90380046, 1419.93716627, 1383.83706883, 1347.73697138, 1419.93716627, 1456.03726372, 1419.93716627, 1419.93716627, 1444.00389791, 1359.77033719, 1275.53677648, 1335.70360556, 1323.67023975, 1323.67023975, 1323.67023975, 1299.60350811, 1347.73697138, 1359.77033719, 1371.80370301, 1359.77033719, 1299.60350811, 1323.67023975, 1311.63687393, 1323.67023975, 1395.87043464, 1383.83706883, 1359.77033719, 1323.67023975, 1287.57014230, 1227.40331322, 1263.50341067, 1203.33658159, 1185.28653286, 1145.57642567, 1081.79958685, 995.15935297, 1014.41273828, 967.48261160, 960.26259211, 951.83923604, 943.41587996, 983.12598716, 1008.39605537, 1027.64944068, 981.92265058, 943.41587996, 912.12912884, 957.85591894, 938.60253364, 868.80901191, 885.65572405, 805.03217308, 921.75582150, 916.94247517, 934.99252389, 944.61921655, 939.80587022, 948.22922629, 1026.44610409, 1022.83609435, 1058.93619180, 1040.88614307, 1070.96955761, 1027.64944068, 1034.86946017, 1081.79958685, 1095.03628924, 1131.13638669, 1025.24276751, 1128.72971353, 1113.08633797, 1137.15306960, 1131.13638669, 1109.47632822, 1174.45650363, 1173.25316705, 1227.40331322, 1203.33658159, 1203.33658159, 1163.62647440, 1227.40331322, 1263.50341067, 1191.30321577, 1185.28653286, 1215.36994740, 1199.72657184, 1196.11656210, 1263.50341067, 1197.31989868, 1215.36994740, 1215.36994740, 1311.63687393, 1359.77033719, 1419.93716627, 1431.97053209, 1323.67023975, 1347.73697138, 1371.80370301, 1287.57014230, 1287.57014230, 1275.53677648, 1299.60350811, 1287.57014230, 1299.60350811, 1275.53677648, 1239.43667904, 1215.36994740, 1227.40331322, 1227.40331322, 1215.36994740, 1215.36994740, 1251.47004485, 1199.72657184, 1152.79644516, 1179.26984996, 1179.26984996, 1239.43667904, 1227.40331322, 1215.36994740, 1263.50341067, 1215.36994740, 1203.33658159, 1263.50341067, 1227.40331322, 1239.43667904, 1203.33658159, 1119.10302088, 1149.18643542, 1150.38977200, 1125.11970378, 1085.40959659, 890.46907037, 980.71931399, 993.95601639, 1046.90282598, 1020.42942119, 1027.64944068, 1089.01960634, 1046.90282598, 1010.80272853, 1015.61607486, 1078.18957710, 992.75267981, 928.97584099, 962.66926527, 992.75267981, 972.29595792, 1034.86946017, 1092.62961608, 1227.40331322, 1162.42313781, 1192.50655235, 1227.40331322, 1251.47004485, 1275.53677648, 1239.43667904, 1153.99978174, 1103.45964532, 1138.35640618, 1133.54305986, 1078.18957710, 1045.69948940, 948.22922629, 922.95915808, 938.60253364, 974.70263109, 1010.80272853, 1013.20940170], 
							[576.39822258, 589.63492498, 613.70165661, 582.41490549, 577.60155916, 541.50146171, 559.55151044, 529.46809590, 517.43473008, 498.18134478, 553.53482753, 494.57133503, 529.46809590, 535.48477881, 583.61824207, 547.51814462, 505.40136427, 499.38468136, 507.80803743, 557.14483728, 545.11147146, 513.82472034, 513.82472034, 529.46809590, 495.77467161, 507.80803743, 575.19488600, 541.50146171, 531.87476906, 547.51814462, 545.11147146, 515.02805692, 512.62138376, 528.26475932, 509.01137401, 484.94464238, 502.99469110, 487.35131554, 502.99469110, 512.62138376, 483.74130580, 527.06142274, 519.84140325, 495.77467161, 488.55465212, 488.55465212, 504.19802769, 476.52128631, 529.46809590, 513.82472034, 494.57133503, 504.19802769, 500.58801794, 496.97800820, 510.21471059, 507.80803743, 561.95818360, 632.95504192, 610.09164687, 723.20528553, 654.61510038, 689.51186125, 669.05513936, 641.37839799, 649.80175406, 640.17506140, 631.75170533, 614.90499319, 612.49832003, 611.29498345, 598.05828105, 588.43158840, 589.63492498, 593.24493472, 608.88831028, 594.44827130, 604.07496396, 590.83826156, 607.68497370, 599.26161763, 600.46495421, 602.87162738, 616.10832977, 592.04159814, 578.80489574, 548.72148120, 546.31480804, 528.26475932, 536.68811539, 542.70479830, 483.74130580, 403.11775483, 465.69125707, 468.09793024, 470.50460340, 462.08124733, 492.16466187, 541.50146171, 614.90499319, 595.65160789, 543.90813488, 595.65160789, 571.58487625, 577.60155916, 595.65160789, 604.07496396, 587.22825181, 635.36171508, 601.66829079, 565.56819335, 620.92167610, 626.93835901, 648.59841748, 635.36171508, 649.80175406, 703.95190023, 720.79861237, 690.71519783, 717.18860263, 741.25533426, 761.71205615, 730.42530502, 671.46181253, 681.08850518, 699.13855390, 709.96858314, 691.91853441, 683.49517834, 681.08850518, 681.08850518, 689.51186125, 691.91853441, 675.07182227, 661.83511987, 669.05513936, 661.83511987, 659.42844671, 673.86848569, 697.93521732, 685.90185151, 696.73188074, 661.83511987, 700.34189048, 673.86848569, 670.25847594, 671.46181253, 685.90185151, 690.71519783, 671.46181253, 669.05513936, 671.46181253, 649.80175406, 599.26161763, 608.88831028, 610.09164687, 578.80489574, 567.97486651, 540.29812513, 530.67143248, 506.60470085, 489.75798871, 493.36799845, 458.47123758, 465.69125707, 464.48792049, 446.43787177, 463.28458391, 405.52442800, 393.49106218, 369.42433055, 379.05102320, 361.00097448, 367.01765738, 387.47437927, 358.59430131, 421.16780356, 429.59115963, 407.93110116, 401.91441825, 451.25121810, 442.82786202, 498.18134478, 498.18134478, 476.52128631, 456.06456442, 476.52128631, 454.86122784, 433.20116937, 471.70793998, 464.48792049, 419.96446697, 380.25435978, 421.16780356, 393.49106218, 411.54111090, 416.35445723, 445.23453519, 441.62452544, 493.36799845, 501.79135452, 528.26475932, 559.55151044, 523.45141299, 540.29812513, 540.29812513, 564.36485676, 558.34817386, 517.43473008, 501.79135452, 488.55465212, 471.70793998, 469.30126682, 502.99469110, 517.43473008, 531.87476906, 559.55151044, 611.29498345, 610.09164687, 595.65160789, 590.83826156, 616.10832977, 611.29498345, 619.71833952, 622.12501268, 637.76838824, 642.58173457, 655.81843697, 638.97172482, 631.75170533, 607.68497370, 613.70165661, 606.48163712, 584.82157865, 611.29498345, 651.00509064, 602.87162738, 617.31166635, 647.39508089, 643.78507115, 644.98840773, 631.75170533, 623.32834926, 641.37839799, 617.31166635, 614.90499319, 611.29498345, 607.68497370, 592.04159814, 584.82157865, 541.50146171, 537.89145197, 470.50460340, 477.72462289, 498.18134478, 421.16780356, 439.21785228, 460.87791075, 487.35131554, 457.26790100, 448.84454493, 434.40450595, 423.57447672, 397.10107192, 371.83100371, 379.05102320, 417.55779381, 407.93110116, 435.60784253, 444.03119861, 427.18448646, 440.42118886, 447.64120835, 422.37114014, 429.59115963, 419.96446697, 429.59115963, 457.26790100, 459.67457417, 470.50460340, 484.94464238, 481.33463264, 452.45455468, 506.60470085, 557.14483728, 553.53482753, 534.28144222, 505.40136427, 493.36799845, 535.48477881, 515.02805692, 534.28144222], 
							[533.07810564, 535.48477881, 506.60470085, 500.58801794, 475.31794973, 430.79449621, 425.98114988, 410.33777432, 430.79449621, 434.40450595, 475.31794973, 530.67143248, 509.01137401, 525.85808615, 490.96132529, 470.50460340, 458.47123758, 418.76113039, 445.23453519, 466.89459366, 435.60784253, 403.11775483, 415.15112065, 392.28772560, 358.59430131, 405.52442800, 494.57133503, 488.55465212, 475.31794973, 507.80803743, 488.55465212, 452.45455468, 445.23453519, 457.26790100, 442.82786202, 429.59115963, 421.16780356, 373.03434029, 416.35445723, 421.16780356, 406.72776458, 409.13443774, 427.18448646, 404.32109141, 389.88105243, 385.06770611, 381.45769636, 373.03434029, 386.27104269, 394.69439876, 382.66103294, 395.89773534, 383.86436953, 370.62766713, 391.08438902, 406.72776458, 411.54111090, 434.40450595, 447.64120835, 447.64120835, 434.40450595, 457.26790100, 459.67457417, 460.87791075, 458.47123758, 477.72462289, 500.58801794, 510.21471059, 478.92795947, 459.67457417, 466.89459366, 463.28458391, 466.89459366, 464.48792049, 468.09793024, 469.30126682, 475.31794973, 474.11461315, 478.92795947, 476.52128631, 474.11461315, 454.86122784, 452.45455468, 431.99783279, 440.42118886, 436.81117912, 439.21785228, 417.55779381, 376.64435004, 367.01765738, 354.98429157, 318.88419412, 335.73090626, 324.90087703, 346.56093550, 324.90087703, 341.74758917, 340.54425259, 365.81432080, 379.05102320, 364.61098422, 401.91441825, 403.11775483, 367.01765738, 376.64435004, 387.47437927, 387.47437927, 409.13443774, 382.66103294, 369.42433055, 394.69439876, 395.89773534, 393.49106218, 387.47437927, 392.28772560, 403.11775483, 409.13443774, 428.38782305, 424.77781330, 433.20116937, 487.35131554, 487.35131554, 466.89459366, 462.08124733, 469.30126682, 448.84454493, 447.64120835, 456.06456442, 451.25121810, 460.87791075, 446.43787177, 453.65789126, 450.04788151, 445.23453519, 441.62452544, 433.20116937, 425.98114988, 436.81117912, 442.82786202, 451.25121810, 446.43787177, 418.76113039, 425.98114988, 397.10107192, 393.49106218, 385.06770611, 367.01765738, 381.45769636, 397.10107192, 383.86436953, 382.66103294, 379.05102320, 405.52442800, 417.55779381, 413.94778407, 403.11775483, 393.49106218, 375.44101346, 367.01765738, 361.00097448, 328.51088677, 340.54425259, 341.74758917, 348.96760866, 361.00097448, 359.79763789, 365.81432080, 359.79763789, 354.98429157, 336.93424284, 330.91755994, 308.05416489, 265.93738453, 279.17408693, 259.92070162, 305.64749172, 320.08753070, 322.49420387, 309.25750147, 315.27418438, 311.66417463, 334.52756968, 351.37428182, 335.73090626, 327.30755019, 327.30755019, 312.86751121, 317.68085754, 323.69754045, 317.68085754, 315.27418438, 303.24081856, 310.46083805, 329.71422336, 322.49420387, 324.90087703, 336.93424284, 336.93424284, 342.95092575, 359.79763789, 377.84768662, 374.23767687, 369.42433055, 352.57761841, 359.79763789, 383.86436953, 385.06770611, 375.44101346, 364.61098422, 364.61098422, 367.01765738, 367.01765738, 367.01765738, 383.86436953, 386.27104269, 397.10107192, 391.08438902, 398.30440851, 403.11775483, 391.08438902, 409.13443774, 391.08438902, 392.28772560, 394.69439876, 405.52442800, 407.93110116, 398.30440851, 400.71108167, 379.05102320, 369.42433055, 371.83100371, 356.18762815, 342.95092575, 359.79763789, 364.61098422, 347.76427208, 346.56093550, 352.57761841, 361.00097448, 358.59430131, 354.98429157, 348.96760866, 358.59430131, 351.37428182, 339.34091601, 354.98429157, 344.15426233, 350.17094524, 342.95092575, 320.08753070, 330.91755994, 298.42747223, 294.81746249, 300.83414540, 257.51402846, 279.17408693, 273.15740402, 267.14072111, 253.90401872, 247.88733581, 258.71736504, 257.51402846, 258.71736504, 261.12403820, 258.71736504, 251.49734555, 246.68399923, 239.46397974, 253.90401872, 243.07398948, 252.70068213, 258.71736504, 262.32737479, 267.14072111, 274.36074060, 288.80077958, 292.41078933, 286.39410642, 281.58076009, 263.53071137, 251.49734555, 269.54739428, 265.93738453, 263.53071137, 239.46397974, 222.61726759, 221.41393101, 222.61726759, 217.80392127, 198.55053596, 181.70382382], 
							[830.30224130, 835.11558762, 829.09890471, 823.08222181, 801.42216334, 640.17506140, 619.71833952, 655.81843697, 628.14169559, 631.75170533, 711.17191972, 703.95190023, 693.12187099, 673.86848569, 671.46181253, 661.83511987, 617.31166635, 589.63492498, 631.75170533, 644.98840773, 640.17506140, 584.82157865, 606.48163712, 589.63492498, 553.53482753, 577.60155916, 646.19174431, 740.05199768, 685.90185151, 691.91853441, 689.51186125, 682.29184176, 655.81843697, 646.19174431, 648.59841748, 625.73502243, 632.95504192, 619.71833952, 643.78507115, 657.02177355, 653.41176380, 638.97172482, 667.85180278, 632.95504192, 647.39508089, 610.09164687, 610.09164687, 608.88831028, 623.32834926, 653.41176380, 699.13855390, 678.68183202, 665.44512962, 646.19174431, 665.44512962, 685.90185151, 671.46181253, 664.24179304, 770.13541222, 808.64218283, 744.86534400, 782.16877803, 770.13541222, 772.54208538, 782.16877803, 854.36897293, 844.74228027, 837.52226079, 802.62549992, 747.27201717, 758.10204640, 752.08536349, 748.47535375, 723.20528553, 762.91539273, 823.08222181, 831.50557788, 809.84551941, 808.64218283, 812.25219257, 790.59213410, 776.15209512, 788.18546094, 771.33874880, 762.91539273, 737.64532451, 753.28870007, 746.06868058, 676.27515885, 677.47849543, 646.19174431, 586.02491523, 605.27830054, 589.63492498, 618.51500294, 598.05828105, 613.70165661, 613.70165661, 661.83511987, 638.97172482, 623.32834926, 694.32520758, 659.42844671, 632.95504192, 655.81843697, 652.20842722, 647.39508089, 661.83511987, 634.15837850, 616.10832977, 646.19174431, 649.80175406, 649.80175406, 648.59841748, 644.98840773, 654.61510038, 659.42844671, 660.63178329, 658.22511013, 664.24179304, 669.05513936, 640.17506140, 630.54836875, 646.19174431, 649.80175406, 642.58173457, 638.97172482, 626.93835901, 628.14169559, 623.32834926, 619.71833952, 616.10832977, 598.05828105, 601.66829079, 596.85494447, 589.63492498, 582.41490549, 589.63492498, 577.60155916, 577.60155916, 580.00823233, 565.56819335, 576.39822258, 552.33149095, 541.50146171, 549.92481779, 565.56819335, 572.78821284, 566.77152993, 545.11147146, 564.36485676, 559.55151044, 516.23139350, 511.41804717, 535.48477881, 571.58487625, 523.45141299, 499.38468136, 463.28458391, 481.33463264, 471.70793998, 446.43787177, 446.43787177, 469.30126682, 495.77467161, 483.74130580, 486.14797896, 476.52128631, 493.36799845, 462.08124733, 447.64120835, 434.40450595, 417.55779381, 423.57447672, 409.13443774, 498.18134478, 490.96132529, 481.33463264, 492.16466187, 476.52128631, 452.45455468, 498.18134478, 501.79135452, 505.40136427, 500.58801794, 504.19802769, 475.31794973, 469.30126682, 481.33463264, 469.30126682, 469.30126682, 446.43787177, 475.31794973, 463.28458391, 474.11461315, 481.33463264, 482.53796922, 490.96132529, 482.53796922, 488.55465212, 496.97800820, 502.99469110, 519.84140325, 531.87476906, 553.53482753, 580.00823233, 572.78821284, 581.21156891, 555.94150069, 537.89145197, 547.51814462, 541.50146171, 559.55151044, 590.83826156, 613.70165661, 593.24493472, 611.29498345, 596.85494447, 576.39822258, 576.39822258, 590.83826156, 598.05828105, 607.68497370, 589.63492498, 610.09164687, 618.51500294, 616.10832977, 601.66829079, 584.82157865, 598.05828105, 577.60155916, 565.56819335, 551.12815437, 570.38153967, 578.80489574, 534.28144222, 530.67143248, 586.02491523, 572.78821284, 587.22825181, 558.34817386, 559.55151044, 588.43158840, 552.33149095, 547.51814462, 572.78821284, 563.16152018, 560.75484702, 564.36485676, 499.38468136, 509.01137401, 482.53796922, 464.48792049, 460.87791075, 369.42433055, 411.54111090, 393.49106218, 421.16780356, 392.28772560, 405.52442800, 409.13443774, 403.11775483, 392.28772560, 381.45769636, 413.94778407, 380.25435978, 362.20431106, 362.20431106, 388.67771585, 374.23767687, 400.71108167, 409.13443774, 445.23453519, 418.76113039, 444.03119861, 456.06456442, 448.84454493, 438.01451570, 429.59115963, 405.52442800, 389.88105243, 391.08438902, 370.62766713, 348.96760866, 320.08753070, 300.83414540, 297.22413565, 283.98743325, 282.78409667, 270.75073086, 257.51402846], 
							[4091.34437740, 4031.17754832, 4079.31101158, 4392.17852280, 4452.34535187, 4596.74574167, 4705.04603401, 4596.74574167, 4644.87920493, 4584.71237585, 4885.54652125, 4536.57891259, 4608.77910748, 4440.31198606, 4307.94496208, 4211.67803556, 4103.37774321, 4091.34437740, 4151.51120648, 4271.84486464, 4163.54457229, 3718.31003711, 3802.54359782, 3730.34340292, 3549.84291568, 3790.51023200, 3850.67706108, 3910.84389016, 3874.74379271, 3874.74379271, 3814.57696363, 3561.87628150, 3489.67608660, 3453.57598916, 3429.50925753, 3249.00877029, 3092.57501468, 3297.14223355, 3309.17559937, 3309.17559937, 3345.27569681, 3453.57598916, 3369.34242845, 3453.57598916, 3212.90867284, 3104.60838050, 3116.64174631, 2851.90769836, 2815.80760092, 3044.44155142, 2851.90769836, 2815.80760092, 2647.34047949, 2382.60643154, 2563.10691878, 2635.30711368, 2719.54067439, 2888.00779581, 3008.34145397, 3176.80857539, 2851.90769836, 3116.64174631, 3128.67511213, 3092.57501468, 3032.40818560, 2875.97442999, 2924.10789326, 2839.87433255, 2743.60740602, 2743.60740602, 2683.44057694, 2924.10789326, 3116.64174631, 3116.64174631, 3236.97540447, 3164.77520958, 3212.90867284, 3080.54164886, 3261.04213610, 3453.57598916, 3441.54262334, 3104.60838050, 3273.07550192, 3273.07550192, 3405.44252589, 3393.40916008, 3369.34242845, 3333.24233100, 3333.24233100, 3200.87530702, 3020.37481979, 2815.80760092, 2996.30808815, 3212.90867284, 3393.40916008, 3164.77520958, 3381.37579426, 3537.80954987, 3501.70945242, 3513.74281824, 3345.27569681, 3790.51023200, 3766.44350037, 3104.60838050, 3261.04213610, 3273.07550192, 3200.87530702, 3200.87530702, 3104.60838050, 2948.17462489, 3008.34145397, 2984.27472234, 3092.57501468, 3140.70847794, 3092.57501468, 3285.10886773, 3405.44252589, 3429.50925753, 3489.67608660, 3622.04311058, 3658.14320803, 3381.37579426, 3309.17559937, 3585.94301313, 3525.77618405, 3501.70945242, 3465.60935497, 3369.34242845, 3309.17559937, 3273.07550192, 3249.00877029, 3429.50925753, 3164.77520958, 3297.14223355, 3188.84194121, 3068.50828305, 3020.37481979, 2996.30808815, 3032.40818560, 3020.37481979, 3032.40818560, 2984.27472234, 2731.57404020, 2635.30711368, 2731.57404020, 2779.70750347, 2924.10789326, 2948.17462489, 2888.00779581, 2875.97442999, 2815.80760092, 2575.14028460, 2478.87335807, 2623.27374786, 2623.27374786, 2575.14028460, 2502.94008970, 2514.97345552, 2478.87335807, 2286.33950502, 2190.07257849, 2105.83901778, 2129.90574941, 2166.00584686, 2226.17267594, 2129.90574941, 2129.90574941, 2093.80565196, 2093.80565196, 1768.90477493, 1925.33853054, 1792.97150657, 1792.97150657, 1877.20506728, 1732.80467749, 1889.23843309, 1973.47199380, 1913.30516472, 1985.50535962, 2117.87238359, 2178.03921267, 2502.94008970, 2214.13931012, 2226.17267594, 2274.30613920, 2238.20604175, 2262.27277339, 2153.97248104, 2117.87238359, 2021.60545707, 2033.63882288, 1841.10496983, 1901.27179891, 1937.37189636, 1949.40526217, 2045.67218870, 2129.90574941, 2166.00584686, 2202.10594431, 2214.13931012, 2274.30613920, 2358.53969991, 2394.63979736, 2394.63979736, 2575.14028460, 2611.24038205, 2587.17365041, 2599.20701623, 2539.04018715, 2502.94008970, 2611.24038205, 2514.97345552, 2527.00682133, 2647.34047949, 2623.27374786, 2695.47394276, 2755.64077184, 2912.07452744, 2695.47394276, 2719.54067439, 2803.77423510, 2707.50730857, 2888.00779581, 2936.14125907, 3020.37481979, 3116.64174631, 3140.70847794, 2996.30808815, 2912.07452744, 2755.64077184, 2731.57404020, 2900.04116163, 2731.57404020, 2888.00779581, 2960.20799071, 2647.34047949, 2599.20701623, 2815.80760092, 2695.47394276, 2755.64077184, 2815.80760092, 2755.64077184, 2924.10789326, 2731.57404020, 2659.37384531, 2815.80760092, 2803.77423510, 2863.94106418, 2948.17462489, 2791.74086928, 2960.20799071, 2791.74086928, 2599.20701623, 2659.37384531, 2406.67316318, 2514.97345552, 2262.27277339, 2454.80662644, 2406.67316318, 2659.37384531, 2707.50730857, 2803.77423510, 2683.44057694, 2514.97345552, 2599.20701623, 2527.00682133, 2575.14028460, 2743.60740602, 2575.14028460, 2635.30711368, 2803.77423510, 2827.84096673, 2996.30808815, 3164.77520958, 3465.60935497, 3405.44252589, 3273.07550192, 3333.24233100, 3345.27569681, 3369.34242845, 3597.97637895, 3585.94301313, 3549.84291568, 3561.87628150, 3297.14223355, 3333.24233100, 3212.90867284, 2888.00779581, 3200.87530702, 3044.44155142, 3032.40818560], 
							[884.45238747, 823.08222181, 734.03531477, 768.93207563, 783.37211461, 758.10204640, 723.20528553, 764.11872931, 762.91539273, 734.03531477, 782.16877803, 758.10204640, 758.10204640, 782.16877803, 764.11872931, 737.64532451, 649.80175406, 644.98840773, 711.17191972, 691.91853441, 651.00509064, 617.31166635, 617.31166635, 617.31166635, 584.82157865, 685.90185151, 806.23550966, 932.58585073, 891.67240696, 872.41902165, 886.85906063, 833.91225104, 806.23550966, 806.23550966, 791.79547068, 758.10204640, 711.17191972, 720.79861237, 748.47535375, 770.13541222, 776.15209512, 772.54208538, 799.01549017, 807.43884625, 765.32206589, 737.64532451, 740.05199768, 759.30538298, 800.21882676, 805.03217308, 788.18546094, 791.79547068, 764.11872931, 726.81529528, 854.36897293, 1046.90282598, 1050.51283573, 1215.36994740, 1263.50341067, 1239.43667904, 1191.30321577, 1335.70360556, 1323.67023975, 1251.47004485, 1263.50341067, 1335.70360556, 1371.80370301, 1371.80370301, 1251.47004485, 1251.47004485, 1263.50341067, 1275.53677648, 1359.77033719, 1335.70360556, 1359.77033719, 1419.93716627, 1624.50438514, 1588.40428770, 1576.37092188, 1648.57111678, 1588.40428770, 1468.07062954, 1359.77033719, 1383.83706883, 1444.00389791, 1456.03726372, 1468.07062954, 1431.97053209, 1383.83706883, 1359.77033719, 1323.67023975, 1251.47004485, 1323.67023975, 1359.77033719, 1323.67023975, 1227.40331322, 1311.63687393, 1347.73697138, 1444.00389791, 1492.13736117, 1468.07062954, 1576.37092188, 1564.33755606, 1492.13736117, 1588.40428770, 1612.47101933, 1600.43765351, 1612.47101933, 1564.33755606, 1480.10399535, 1492.13736117, 1516.20409280, 1492.13736117, 1468.07062954, 1480.10399535, 1528.23745862, 1516.20409280, 1624.50438514, 1648.57111678, 1696.70458004, 1672.63784841, 1624.50438514, 1612.47101933, 1600.43765351, 1600.43765351, 1528.23745862, 1492.13736117, 1444.00389791, 1480.10399535, 1444.00389791, 1395.87043464, 1431.97053209, 1359.77033719, 1371.80370301, 1311.63687393, 1347.73697138, 1359.77033719, 1395.87043464, 1323.67023975, 1323.67023975, 1359.77033719, 1227.40331322, 1227.40331322, 1203.33658159, 1263.50341067, 1251.47004485, 1275.53677648, 1299.60350811, 1359.77033719, 1275.53677648, 1227.40331322, 1196.11656210, 1153.99978174, 1215.36994740, 1203.33658159, 1175.65984021, 1086.61293317, 1095.03628924, 1025.24276751, 973.49929450, 901.29909961, 914.53580201, 949.43256287, 950.63589945, 1018.02274802, 991.54934323, 1021.63275777, 956.65258236, 944.61921655, 920.55248491, 934.99252389, 878.43570456, 860.38565584, 882.04571430, 827.89556813, 956.65258236, 1019.22608460, 1004.78604563, 983.12598716, 1026.44610409, 1012.00606512, 1172.04983047, 1127.52637695, 1149.18643542, 1191.30321577, 1287.57014230, 1173.25316705, 1180.47318654, 1227.40331322, 1193.70988894, 1251.47004485, 1139.55974276, 1191.30321577, 1199.72657184, 1251.47004485, 1287.57014230, 1359.77033719, 1419.93716627, 1395.87043464, 1576.37092188, 1600.43765351, 1708.73794585, 1588.40428770, 1552.30419025, 1552.30419025, 1576.37092188, 1576.37092188, 1552.30419025, 1504.17072698, 1480.10399535, 1492.13736117, 1492.13736117, 1600.43765351, 1636.53775096, 1636.53775096, 1600.43765351, 1552.30419025, 1588.40428770, 1552.30419025, 1492.13736117, 1504.17072698, 1431.97053209, 1431.97053209, 1371.80370301, 1468.07062954, 1407.90380046, 1480.10399535, 1419.93716627, 1347.73697138, 1371.80370301, 1347.73697138, 1347.73697138, 1335.70360556, 1395.87043464, 1419.93716627, 1299.60350811, 1335.70360556, 1407.90380046, 1395.87043464, 1444.00389791, 1371.80370301, 1371.80370301, 1407.90380046, 1335.70360556, 1299.60350811, 1323.67023975, 1287.57014230, 1311.63687393, 1275.53677648, 1172.04983047, 1168.43982072, 1084.20626001, 1058.93619180, 1039.68280649, 891.67240696, 961.46592869, 915.73913859, 909.72245568, 811.04885599, 844.74228027, 918.14581175, 842.33560711, 872.41902165, 855.57230951, 895.28241670, 842.33560711, 788.18546094, 712.37525630, 772.54208538, 838.72559737, 863.99566558, 880.84237772, 883.24905089, 850.75896318, 832.70891446, 830.30224130, 833.91225104, 839.92893395, 907.31578252, 842.33560711, 812.25219257, 873.62235823, 829.09890471, 778.55876829, 720.79861237, 736.44198793, 722.00194895, 684.69851492, 675.07182227, 602.87162738, 573.99154942], 
							[961.46592869, 986.73599690, 968.68594818, 950.63589945, 903.70577277, 749.67869033, 754.49203666, 788.18546094, 749.67869033, 838.72559737, 853.16563635, 830.30224130, 836.31892420, 859.18231925, 880.84237772, 901.29909961, 861.58899242, 800.21882676, 850.75896318, 815.86220232, 815.86220232, 749.67869033, 753.28870007, 720.79861237, 706.35857339, 792.99880727, 845.94561686, 890.46907037, 873.62235823, 866.40233874, 909.72245568, 855.57230951, 855.57230951, 876.02903140, 857.97898267, 836.31892420, 842.33560711, 835.11558762, 861.58899242, 866.40233874, 845.94561686, 842.33560711, 850.75896318, 847.14895344, 812.25219257, 826.69223155, 842.33560711, 836.31892420, 867.60567532, 878.43570456, 902.50243619, 901.29909961, 880.84237772, 873.62235823, 901.29909961, 930.17917757, 926.56916782, 914.53580201, 948.22922629, 978.31264083, 962.66926527, 995.15935297, 974.70263109, 972.29595792, 986.73599690, 967.48261160, 951.83923604, 932.58585073, 907.31578252, 879.63904114, 879.63904114, 890.46907037, 890.46907037, 867.60567532, 903.70577277, 907.31578252, 925.36583124, 919.34914833, 922.95915808, 919.34914833, 906.11244594, 890.46907037, 886.85906063, 885.65572405, 860.38565584, 830.30224130, 813.45552915, 747.27201717, 719.59527579, 703.95190023, 661.83511987, 601.66829079, 637.76838824, 618.51500294, 666.64846620, 619.71833952, 634.15837850, 661.83511987, 746.06868058, 724.40862212, 709.96858314, 811.04885599, 803.82883650, 734.03531477, 760.50871956, 797.81215359, 794.20214385, 824.28555839, 801.42216334, 776.15209512, 809.84551941, 820.67554864, 797.81215359, 779.76210487, 782.16877803, 824.28555839, 851.96229976, 883.24905089, 877.23236798, 912.12912884, 918.14581175, 896.48575328, 892.87574354, 908.51911910, 948.22922629, 932.58585073, 910.92579226, 926.56916782, 914.53580201, 922.95915808, 914.53580201, 920.55248491, 890.46907037, 889.26573379, 898.89242645, 891.67240696, 889.26573379, 927.77250440, 919.34914833, 915.73913859, 922.95915808, 848.35229002, 856.77564609, 824.28555839, 837.52226079, 850.75896318, 837.52226079, 896.48575328, 872.41902165, 865.19900216, 842.33560711, 848.35229002, 785.77878778, 797.81215359, 806.23550966, 791.79547068, 758.10204640, 687.10518809, 644.98840773, 683.49517834, 653.41176380, 641.37839799, 671.46181253, 685.90185151, 712.37525630, 681.08850518, 709.96858314, 673.86848569, 655.81843697, 635.36171508, 617.31166635, 608.88831028, 588.43158840, 625.73502243, 576.39822258, 648.59841748, 640.17506140, 661.83511987, 644.98840773, 631.75170533, 634.15837850, 703.95190023, 709.96858314, 719.59527579, 719.59527579, 758.10204640, 697.93521732, 705.15523681, 742.45867084, 752.08536349, 782.16877803, 730.42530502, 780.96544145, 773.74542196, 790.59213410, 785.77878778, 803.82883650, 806.23550966, 839.92893395, 827.89556813, 824.28555839, 838.72559737, 866.40233874, 860.38565584, 872.41902165, 895.28241670, 894.07908012, 901.29909961, 890.46907037, 878.43570456, 876.02903140, 874.82569481, 867.60567532, 880.84237772, 913.33246543, 924.16249466, 957.85591894, 956.65258236, 954.24590920, 944.61921655, 956.65258236, 928.97584099, 931.38251415, 936.19586048, 938.60253364, 938.60253364, 959.05925553, 932.58585073, 897.68908986, 892.87574354, 889.26573379, 843.53894369, 835.11558762, 854.36897293, 870.01234849, 805.03217308, 823.08222181, 832.70891446, 861.58899242, 868.80901191, 854.36897293, 835.11558762, 856.77564609, 836.31892420, 817.06553890, 841.13227053, 825.48889497, 826.69223155, 777.35543171, 728.01863186, 697.93521732, 676.27515885, 663.03845645, 661.83511987, 554.73816411, 611.29498345, 619.71833952, 659.42844671, 651.00509064, 711.17191972, 719.59527579, 713.57859288, 697.93521732, 696.73188074, 722.00194895, 691.91853441, 696.73188074, 666.64846620, 715.98526604, 715.98526604, 729.22196844, 737.64532451, 760.50871956, 729.22196844, 715.98526604, 725.61195870, 697.93521732, 700.34189048, 691.91853441, 676.27515885, 643.78507115, 659.42844671, 659.42844671, 577.60155916, 601.66829079, 577.60155916, 588.43158840, 571.58487625, 577.60155916, 529.46809590, 548.72148120], 
							[2948.17462489, 2875.97442999, 2972.24135652, 2839.87433255, 2647.34047949, 2406.67316318, 2166.00584686, 1997.53872544, 2153.97248104, 2069.73892033, 2250.23940757, 2105.83901778, 2129.90574941, 2190.07257849, 2081.77228615, 2021.60545707, 1913.30516472, 1768.90477493, 1805.00487238, 1913.30516472, 1865.17170146, 1624.50438514, 1744.83804330, 1696.70458004, 1540.27082443, 1889.23843309, 2166.00584686, 2250.23940757, 2105.83901778, 2214.13931012, 2274.30613920, 2238.20604175, 2129.90574941, 2166.00584686, 2069.73892033, 2033.63882288, 2129.90574941, 2069.73892033, 2190.07257849, 2274.30613920, 2250.23940757, 2238.20604175, 2250.23940757, 2214.13931012, 2045.67218870, 1985.50535962, 2021.60545707, 2021.60545707, 2166.00584686, 2250.23940757, 2190.07257849, 2178.03921267, 2274.30613920, 2322.43960246, 2827.84096673, 2960.20799071, 2888.00779581, 2827.84096673, 2623.27374786, 2815.80760092, 2743.60740602, 2863.94106418, 2803.77423510, 2719.54067439, 2755.64077184, 2695.47394276, 2683.44057694, 2647.34047949, 2635.30711368, 2527.00682133, 2635.30711368, 2671.40721112, 2731.57404020, 2647.34047949, 2695.47394276, 2707.50730857, 2767.67413765, 2743.60740602, 2743.60740602, 2863.94106418, 2827.84096673, 2743.60740602, 2671.40721112, 2647.34047949, 2707.50730857, 2623.27374786, 2659.37384531, 2611.24038205, 2129.90574941, 2178.03921267, 2021.60545707, 1829.07160401, 1985.50535962, 1937.37189636, 2045.67218870, 1925.33853054, 1913.30516472, 2021.60545707, 2346.50633410, 2382.60643154, 2190.07257849, 2563.10691878, 2527.00682133, 2238.20604175, 2334.47296828, 2430.73989481, 2502.94008970, 2442.77326062, 2346.50633410, 2322.43960246, 2406.67316318, 2382.60643154, 2346.50633410, 2298.37287083, 2322.43960246, 2358.53969991, 2430.73989481, 2442.77326062, 2418.70652899, 2527.00682133, 2563.10691878, 2406.67316318, 2394.63979736, 2382.60643154, 2454.80662644, 2454.80662644, 2490.90672389, 2382.60643154, 2394.63979736, 2358.53969991, 2346.50633410, 2394.63979736, 2190.07257849, 2166.00584686, 2117.87238359, 2009.57209125, 2105.83901778, 2153.97248104, 2141.93911523, 2141.93911523, 2141.93911523, 2093.80565196, 2093.80565196, 1985.50535962, 2021.60545707, 2057.70555452, 2105.83901778, 2178.03921267, 2190.07257849, 2178.03921267, 2178.03921267, 2190.07257849, 2069.73892033, 2214.13931012, 2141.93911523, 2153.97248104, 2129.90574941, 1985.50535962, 1865.17170146, 1889.23843309, 1829.07160401, 1684.67121422, 1732.80467749, 1720.77131167, 1913.30516472, 1877.20506728, 1925.33853054, 1865.17170146, 1792.97150657, 1768.90477493, 1792.97150657, 1684.67121422, 1660.60448259, 1696.70458004, 1564.33755606, 1792.97150657, 1780.93814075, 1817.03823820, 1817.03823820, 1792.97150657, 1817.03823820, 1925.33853054, 1937.37189636, 1973.47199380, 1997.53872544, 2069.73892033, 1949.40526217, 1949.40526217, 2057.70555452, 1949.40526217, 2009.57209125, 1817.03823820, 1865.17170146, 1841.10496983, 1973.47199380, 1973.47199380, 2057.70555452, 2093.80565196, 2105.83901778, 2117.87238359, 2129.90574941, 2226.17267594, 2214.13931012, 2226.17267594, 2370.57306573, 2322.43960246, 2190.07257849, 2262.27277339, 2190.07257849, 2105.83901778, 2202.10594431, 2214.13931012, 2298.37287083, 2406.67316318, 2442.77326062, 2262.27277339, 2382.60643154, 2346.50633410, 2262.27277339, 2262.27277339, 2406.67316318, 2442.77326062, 2478.87335807, 2466.83999225, 2551.07355297, 2647.34047949, 2659.37384531, 2659.37384531, 2575.14028460, 2611.24038205, 2587.17365041, 2478.87335807, 2418.70652899, 2539.04018715, 2635.30711368, 2430.73989481, 2394.63979736, 2527.00682133, 2599.20701623, 2779.70750347, 2695.47394276, 2695.47394276, 2839.87433255, 2647.34047949, 2647.34047949, 2791.74086928, 2731.57404020, 2815.80760092, 2851.90769836, 2791.74086928, 2839.87433255, 2707.50730857, 2478.87335807, 2430.73989481, 1961.43862799, 1997.53872544, 1925.33853054, 2033.63882288, 1949.40526217, 2033.63882288, 2081.77228615, 1997.53872544, 1997.53872544, 1913.30516472, 2045.67218870, 1925.33853054, 1817.03823820, 1853.13833565, 1961.43862799, 1877.20506728, 1949.40526217, 2033.63882288, 2166.00584686, 2117.87238359, 2105.83901778, 2214.13931012, 2226.17267594, 2069.73892033, 1901.27179891, 1877.20506728, 1817.03823820, 1841.10496983, 1792.97150657, 1660.60448259, 1612.47101933, 1516.20409280, 1528.23745862, 1480.10399535, 1564.33755606, 1504.17072698, 1468.07062954], 
							[1431.97053209, 1371.80370301, 1371.80370301, 1407.90380046, 1407.90380046, 1407.90380046, 1311.63687393, 1119.10302088, 986.73599690, 1043.29281624, 1107.06965506, 1045.69948940, 1064.95287471, 1051.71617231, 1024.03943093, 980.71931399, 913.33246543, 850.75896318, 926.56916782, 967.48261160, 906.11244594, 867.60567532, 938.60253364, 926.56916782, 912.12912884, 956.65258236, 1022.83609435, 1113.08633797, 1155.20311832, 1107.06965506, 1191.30321577, 1093.83295266, 997.56602614, 1002.37937246, 978.31264083, 986.73599690, 1005.98938221, 1002.37937246, 1107.06965506, 1143.16975251, 1108.27299164, 1096.23962583, 1096.23962583, 1036.07279675, 986.73599690, 1001.17603588, 1027.64944068, 974.70263109, 1057.73285522, 1087.81626976, 1052.91950889, 1042.08947966, 1093.83295266, 1084.20626001, 1155.20311832, 1101.05297215, 1111.88300139, 1052.91950889, 1067.35954787, 1157.60979149, 1084.20626001, 1113.08633797, 1102.25630873, 1102.25630873, 1167.23648414, 1239.43667904, 1197.31989868, 1151.59310858, 1146.77976225, 1122.71303062, 1140.76307935, 1163.62647440, 1170.84649388, 1135.94973302, 1149.18643542, 1155.20311832, 1179.26984996, 1156.40645491, 1196.11656210, 1200.92990842, 1174.45650363, 1166.03314756, 1155.20311832, 1161.21980123, 1179.26984996, 1120.30635746, 1055.32618205, 971.09262134, 901.29909961, 879.63904114, 819.47221206, 794.20214385, 851.96229976, 814.65886573, 865.19900216, 805.03217308, 813.45552915, 872.41902165, 928.97584099, 909.72245568, 888.06239721, 992.75267981, 979.51597741, 873.62235823, 928.97584099, 914.53580201, 950.63589945, 975.90596767, 915.73913859, 890.46907037, 934.99252389, 926.56916782, 930.17917757, 914.53580201, 920.55248491, 962.66926527, 979.51597741, 1002.37937246, 975.90596767, 1002.37937246, 1010.80272853, 950.63589945, 950.63589945, 959.05925553, 973.49929450, 990.34600665, 986.73599690, 977.10930425, 986.73599690, 950.63589945, 986.73599690, 986.73599690, 948.22922629, 962.66926527, 919.34914833, 876.02903140, 889.26573379, 932.58585073, 888.06239721, 904.90910935, 904.90910935, 894.07908012, 890.46907037, 854.36897293, 860.38565584, 874.82569481, 913.33246543, 962.66926527, 960.26259211, 960.26259211, 974.70263109, 990.34600665, 938.60253364, 977.10930425, 941.00920680, 968.68594818, 903.70577277, 842.33560711, 788.18546094, 818.26887548, 838.72559737, 753.28870007, 779.76210487, 791.79547068, 839.92893395, 851.96229976, 865.19900216, 836.31892420, 774.94875854, 765.32206589, 774.94875854, 732.83197819, 663.03845645, 735.23865135, 682.29184176, 773.74542196, 783.37211461, 796.60881701, 851.96229976, 812.25219257, 808.64218283, 856.77564609, 883.24905089, 889.26573379, 913.33246543, 938.60253364, 897.68908986, 931.38251415, 953.04257262, 900.09576303, 896.48575328, 841.13227053, 914.53580201, 906.11244594, 924.16249466, 900.09576303, 908.51911910, 939.80587022, 1067.35954787, 1024.03943093, 1057.73285522, 1056.52951863, 1084.20626001, 1099.84963557, 1131.13638669, 1123.91636720, 1098.64629899, 1089.01960634, 1061.34286496, 1068.56288445, 1119.10302088, 1134.74639644, 1191.30321577, 1263.50341067, 1299.60350811, 1251.47004485, 1335.70360556, 1323.67023975, 1251.47004485, 1251.47004485, 1287.57014230, 1335.70360556, 1335.70360556, 1311.63687393, 1335.70360556, 1371.80370301, 1371.80370301, 1347.73697138, 1335.70360556, 1299.60350811, 1275.53677648, 1251.47004485, 1215.36994740, 1251.47004485, 1299.60350811, 1179.26984996, 1162.42313781, 1227.40331322, 1275.53677648, 1347.73697138, 1287.57014230, 1287.57014230, 1323.67023975, 1263.50341067, 1263.50341067, 1347.73697138, 1335.70360556, 1383.83706883, 1371.80370301, 1335.70360556, 1371.80370301, 1299.60350811, 1166.03314756, 1162.42313781, 937.39919706, 987.93933348, 924.16249466, 991.54934323, 924.16249466, 962.66926527, 960.26259211, 945.82255313, 914.53580201, 890.46907037, 961.46592869, 871.21568507, 784.57545120, 801.42216334, 862.79232900, 902.50243619, 974.70263109, 1068.56288445, 1178.06651337, 1169.64315730, 1203.33658159, 1239.43667904, 1263.50341067, 1227.40331322, 1227.40331322, 1116.69634771, 1105.86631848, 1084.20626001, 1091.42627950, 975.90596767, 942.21254338, 987.93933348, 915.73913859, 918.14581175, 961.46592869, 930.17917757, 959.05925553], 
							[800.21882676, 792.99880727, 776.15209512, 746.06868058, 740.05199768, 744.86534400, 732.83197819, 697.93521732, 697.93521732, 667.85180278, 723.20528553, 706.35857339, 703.95190023, 709.96858314, 697.93521732, 665.44512962, 660.63178329, 649.80175406, 683.49517834, 644.98840773, 635.36171508, 622.12501268, 601.66829079, 602.87162738, 553.53482753, 601.66829079, 651.00509064, 651.00509064, 637.76838824, 649.80175406, 670.25847594, 596.85494447, 570.38153967, 601.66829079, 558.34817386, 561.95818360, 559.55151044, 540.29812513, 553.53482753, 559.55151044, 547.51814462, 583.61824207, 564.36485676, 560.75484702, 542.70479830, 535.48477881, 534.28144222, 518.63806666, 546.31480804, 539.09478855, 518.63806666, 519.84140325, 525.85808615, 506.60470085, 554.73816411, 566.77152993, 599.26161763, 631.75170533, 625.73502243, 642.58173457, 610.09164687, 702.74856365, 695.52854416, 653.41176380, 649.80175406, 647.39508089, 661.83511987, 644.98840773, 631.75170533, 624.53168584, 619.71833952, 624.53168584, 643.78507115, 634.15837850, 648.59841748, 648.59841748, 657.02177355, 631.75170533, 659.42844671, 653.41176380, 641.37839799, 625.73502243, 624.53168584, 608.88831028, 625.73502243, 636.56505166, 652.20842722, 629.34503217, 611.29498345, 571.58487625, 552.33149095, 515.02805692, 518.63806666, 539.09478855, 565.56819335, 563.16152018, 553.53482753, 565.56819335, 592.04159814, 581.21156891, 560.75484702, 565.56819335, 575.19488600, 558.34817386, 590.83826156, 577.60155916, 589.63492498, 631.75170533, 607.68497370, 586.02491523, 614.90499319, 613.70165661, 618.51500294, 599.26161763, 599.26161763, 613.70165661, 614.90499319, 635.36171508, 649.80175406, 696.73188074, 684.69851492, 653.41176380, 679.88516860, 697.93521732, 705.15523681, 702.74856365, 694.32520758, 703.95190023, 702.74856365, 696.73188074, 673.86848569, 688.30852467, 687.10518809, 643.78507115, 663.03845645, 673.86848569, 669.05513936, 671.46181253, 653.41176380, 665.44512962, 681.08850518, 658.22511013, 660.63178329, 634.15837850, 634.15837850, 628.14169559, 620.92167610, 626.93835901, 658.22511013, 657.02177355, 648.59841748, 635.36171508, 566.77152993, 593.24493472, 586.02491523, 601.66829079, 548.72148120, 560.75484702, 517.43473008, 529.46809590, 522.24807641, 541.50146171, 495.77467161, 496.97800820, 527.06142274, 524.65474957, 529.46809590, 515.02805692, 510.21471059, 505.40136427, 511.41804717, 515.02805692, 495.77467161, 513.82472034, 513.82472034, 553.53482753, 561.95818360, 606.48163712, 583.61824207, 572.78821284, 572.78821284, 624.53168584, 613.70165661, 602.87162738, 612.49832003, 610.09164687, 608.88831028, 596.85494447, 625.73502243, 606.48163712, 629.34503217, 593.24493472, 601.66829079, 587.22825181, 598.05828105, 589.63492498, 607.68497370, 630.54836875, 613.70165661, 626.93835901, 631.75170533, 661.83511987, 660.63178329, 647.39508089, 653.41176380, 638.97172482, 665.44512962, 665.44512962, 646.19174431, 653.41176380, 653.41176380, 657.02177355, 666.64846620, 700.34189048, 709.96858314, 701.54522707, 718.39193921, 730.42530502, 718.39193921, 701.54522707, 714.78192946, 707.56190997, 715.98526604, 695.52854416, 714.78192946, 722.00194895, 713.57859288, 685.90185151, 673.86848569, 659.42844671, 643.78507115, 638.97172482, 636.56505166, 661.83511987, 659.42844671, 610.09164687, 616.10832977, 648.59841748, 663.03845645, 661.83511987, 635.36171508, 631.75170533, 646.19174431, 630.54836875, 634.15837850, 640.17506140, 640.17506140, 624.53168584, 612.49832003, 594.44827130, 610.09164687, 594.44827130, 590.83826156, 584.82157865, 511.41804717, 555.94150069, 569.17820309, 600.46495421, 567.97486651, 606.48163712, 608.88831028, 599.26161763, 575.19488600, 548.72148120, 589.63492498, 592.04159814, 589.63492498, 576.39822258, 601.66829079, 598.05828105, 618.51500294, 626.93835901, 624.53168584, 604.07496396, 619.71833952, 635.36171508, 637.76838824, 635.36171508, 628.14169559, 602.87162738, 598.05828105, 572.78821284, 563.16152018, 533.07810564, 529.46809590, 511.41804717, 494.57133503, 496.97800820, 515.02805692, 484.94464238, 493.36799845], 
							[595.65160789, 573.99154942, 560.75484702, 541.50146171, 535.48477881, 499.38468136, 537.89145197, 540.29812513, 535.48477881, 518.63806666, 600.46495421, 553.53482753, 576.39822258, 558.34817386, 553.53482753, 543.90813488, 500.58801794, 500.58801794, 529.46809590, 523.45141299, 501.79135452, 480.13129605, 475.31794973, 457.26790100, 430.79449621, 462.08124733, 525.85808615, 495.77467161, 505.40136427, 487.35131554, 492.16466187, 470.50460340, 464.48792049, 481.33463264, 462.08124733, 439.21785228, 415.15112065, 410.33777432, 427.18448646, 451.25121810, 450.04788151, 458.47123758, 456.06456442, 463.28458391, 441.62452544, 450.04788151, 433.20116937, 442.82786202, 470.50460340, 459.67457417, 456.06456442, 453.65789126, 451.25121810, 416.35445723, 457.26790100, 445.23453519, 483.74130580, 492.16466187, 522.24807641, 555.94150069, 525.85808615, 575.19488600, 572.78821284, 553.53482753, 542.70479830, 541.50146171, 577.60155916, 557.14483728, 535.48477881, 530.67143248, 527.06142274, 531.87476906, 553.53482753, 531.87476906, 552.33149095, 523.45141299, 521.04473983, 517.43473008, 540.29812513, 535.48477881, 523.45141299, 521.04473983, 524.65474957, 549.92481779, 542.70479830, 537.89145197, 535.48477881, 529.46809590, 511.41804717, 501.79135452, 494.57133503, 450.04788151, 493.36799845, 480.13129605, 524.65474957, 499.38468136, 511.41804717, 535.48477881, 539.09478855, 536.68811539, 510.21471059, 537.89145197, 554.73816411, 528.26475932, 584.82157865, 577.60155916, 598.05828105, 635.36171508, 599.26161763, 587.22825181, 604.07496396, 608.88831028, 611.29498345, 602.87162738, 600.46495421, 617.31166635, 622.12501268, 655.81843697, 667.85180278, 707.56190997, 719.59527579, 691.91853441, 671.46181253, 683.49517834, 715.98526604, 690.71519783, 672.66514911, 673.86848569, 688.30852467, 689.51186125, 663.03845645, 711.17191972, 683.49517834, 661.83511987, 688.30852467, 676.27515885, 693.12187099, 703.95190023, 695.52854416, 697.93521732, 684.69851492, 661.83511987, 685.90185151, 669.05513936, 693.12187099, 672.66514911, 682.29184176, 695.52854416, 689.51186125, 683.49517834, 675.07182227, 631.75170533, 614.90499319, 631.75170533, 619.71833952, 630.54836875, 619.71833952, 610.09164687, 564.36485676, 543.90813488, 531.87476906, 601.66829079, 583.61824207, 577.60155916, 594.44827130, 570.38153967, 582.41490549, 548.72148120, 560.75484702, 537.89145197, 534.28144222, 484.94464238, 489.75798871, 535.48477881, 501.79135452, 565.56819335, 608.88831028, 624.53168584, 601.66829079, 634.15837850, 623.32834926, 651.00509064, 635.36171508, 648.59841748, 644.98840773, 684.69851492, 660.63178329, 681.08850518, 676.27515885, 632.95504192, 632.95504192, 614.90499319, 651.00509064, 636.56505166, 618.51500294, 625.73502243, 649.80175406, 657.02177355, 677.47849543, 687.10518809, 715.98526604, 802.62549992, 748.47535375, 756.89870982, 768.93207563, 750.88202691, 770.13541222, 758.10204640, 743.66200742, 758.10204640, 732.83197819, 715.98526604, 723.20528553, 770.13541222, 764.11872931, 770.13541222, 789.38879752, 807.43884625, 780.96544145, 782.16877803, 774.94875854, 788.18546094, 773.74542196, 752.08536349, 791.79547068, 794.20214385, 805.03217308, 785.77878778, 759.30538298, 743.66200742, 729.22196844, 719.59527579, 703.95190023, 712.37525630, 725.61195870, 673.86848569, 666.64846620, 713.57859288, 709.96858314, 712.37525630, 694.32520758, 669.05513936, 685.90185151, 663.03845645, 661.83511987, 665.44512962, 663.03845645, 637.76838824, 641.37839799, 601.66829079, 619.71833952, 590.83826156, 566.77152993, 577.60155916, 498.18134478, 529.46809590, 577.60155916, 613.70165661, 578.80489574, 605.27830054, 608.88831028, 605.27830054, 570.38153967, 553.53482753, 565.56819335, 559.55151044, 549.92481779, 548.72148120, 543.90813488, 558.34817386, 601.66829079, 602.87162738, 599.26161763, 600.46495421, 590.83826156, 631.75170533, 631.75170533, 619.71833952, 616.10832977, 600.46495421, 589.63492498, 617.31166635, 625.73502243, 584.82157865, 525.85808615, 541.50146171, 546.31480804, 517.43473008, 507.80803743, 501.79135452, 470.50460340], 
							[896.48575328, 889.26573379, 854.36897293, 842.33560711, 867.60567532, 773.74542196, 708.76524656, 689.51186125, 706.35857339, 679.88516860, 732.83197819, 749.67869033, 709.96858314, 726.81529528, 720.79861237, 690.71519783, 643.78507115, 649.80175406, 695.52854416, 685.90185151, 657.02177355, 646.19174431, 691.91853441, 679.88516860, 661.83511987, 637.76838824, 725.61195870, 738.84866109, 734.03531477, 753.28870007, 801.42216334, 709.96858314, 720.79861237, 703.95190023, 708.76524656, 687.10518809, 723.20528553, 734.03531477, 776.15209512, 806.23550966, 768.93207563, 752.08536349, 771.33874880, 741.25533426, 723.20528553, 705.15523681, 734.03531477, 722.00194895, 750.88202691, 789.38879752, 736.44198793, 790.59213410, 808.64218283, 817.06553890, 889.26573379, 981.92265058, 947.02588971, 894.07908012, 908.51911910, 921.75582150, 890.46907037, 933.78918731, 894.07908012, 892.87574354, 913.33246543, 938.60253364, 903.70577277, 897.68908986, 884.45238747, 882.04571430, 856.77564609, 862.79232900, 876.02903140, 859.18231925, 933.78918731, 926.56916782, 902.50243619, 918.14581175, 912.12912884, 932.58585073, 890.46907037, 873.62235823, 908.51911910, 925.36583124, 906.11244594, 912.12912884, 912.12912884, 880.84237772, 812.25219257, 768.93207563, 773.74542196, 740.05199768, 746.06868058, 701.54522707, 777.35543171, 728.01863186, 780.96544145, 866.40233874, 892.87574354, 873.62235823, 854.36897293, 955.44924578, 902.50243619, 821.87888522, 872.41902165, 919.34914833, 961.46592869, 927.77250440, 916.94247517, 916.94247517, 961.46592869, 956.65258236, 950.63589945, 938.60253364, 934.99252389, 967.48261160, 998.76936272, 1024.03943093, 1027.64944068, 1091.42627950, 1119.10302088, 1093.83295266, 1083.00292343, 1070.96955761, 1092.62961608, 1099.84963557, 1115.49301113, 1123.91636720, 1121.50969404, 1114.28967455, 1125.11970378, 1128.72971353, 1083.00292343, 1064.95287471, 1056.52951863, 1052.91950889, 1113.08633797, 1105.86631848, 1113.08633797, 1119.10302088, 1107.06965506, 1089.01960634, 1129.93305011, 1116.69634771, 1168.43982072, 1141.96641593, 1152.79644516, 1175.65984021, 1200.92990842, 1194.91322552, 1186.48986945, 1164.82981098, 1083.00292343, 1104.66298190, 1091.42627950, 1040.88614307, 1022.83609435, 983.12598716, 938.60253364, 973.49929450, 948.22922629, 901.29909961, 907.31578252, 938.60253364, 1056.52951863, 1063.74953812, 1037.27613333, 969.88928476, 951.83923604, 926.56916782, 965.07593843, 890.46907037, 865.19900216, 866.40233874, 758.10204640, 950.63589945, 886.85906063, 950.63589945, 960.26259211, 919.34914833, 918.14581175, 1034.86946017, 1001.17603588, 1057.73285522, 1052.91950889, 1024.03943093, 965.07593843, 1001.17603588, 998.76936272, 951.83923604, 986.73599690, 947.02588971, 1010.80272853, 1005.98938221, 1034.86946017, 1046.90282598, 1086.61293317, 1076.98624052, 1108.27299164, 1079.39291368, 1131.13638669, 1202.13324501, 1202.13324501, 1184.08319628, 1215.36994740, 1197.31989868, 1192.50655235, 1181.67652312, 1143.16975251, 1133.54305986, 1155.20311832, 1152.79644516, 1215.36994740, 1227.40331322, 1251.47004485, 1181.67652312, 1263.50341067, 1215.36994740, 1190.09987919, 1227.40331322, 1275.53677648, 1275.53677648, 1275.53677648, 1263.50341067, 1287.57014230, 1311.63687393, 1287.57014230, 1287.57014230, 1251.47004485, 1287.57014230, 1263.50341067, 1215.36994740, 1191.30321577, 1215.36994740, 1263.50341067, 1181.67652312, 1169.64315730, 1215.36994740, 1263.50341067, 1287.57014230, 1275.53677648, 1227.40331322, 1239.43667904, 1215.36994740, 1227.40331322, 1287.57014230, 1251.47004485, 1275.53677648, 1263.50341067, 1215.36994740, 1239.43667904, 1186.48986945, 1156.40645491, 1155.20311832, 1090.22294292, 1093.83295266, 1048.10616256, 1141.96641593, 1069.76622103, 1146.77976225, 1146.77976225, 1079.39291368, 1067.35954787, 1046.90282598, 1111.88300139, 1028.85277726, 999.97269930, 999.97269930, 1027.64944068, 997.56602614, 1062.54620154, 1107.06965506, 1141.96641593, 1143.16975251, 1188.89654261, 1202.13324501, 1180.47318654, 1157.60979149, 1127.52637695, 1075.78290394, 1049.30949914, 1073.37623078, 1076.98624052, 1024.03943093, 1037.27613333, 953.04257262, 944.61921655, 995.15935297, 962.66926527, 912.12912884, 916.94247517], 
							[663.03845645, 667.85180278, 659.42844671, 651.00509064, 634.15837850, 587.22825181, 547.51814462, 554.73816411, 529.46809590, 567.97486651, 608.88831028, 595.65160789, 601.66829079, 581.21156891, 572.78821284, 548.72148120, 500.58801794, 466.89459366, 493.36799845, 504.19802769, 488.55465212, 457.26790100, 459.67457417, 454.86122784, 397.10107192, 458.47123758, 517.43473008, 536.68811539, 535.48477881, 529.46809590, 545.11147146, 507.80803743, 498.18134478, 502.99469110, 517.43473008, 522.24807641, 501.79135452, 446.43787177, 483.74130580, 494.57133503, 480.13129605, 471.70793998, 487.35131554, 471.70793998, 447.64120835, 445.23453519, 445.23453519, 438.01451570, 471.70793998, 469.30126682, 452.45455468, 458.47123758, 448.84454493, 441.62452544, 469.30126682, 469.30126682, 496.97800820, 540.29812513, 565.56819335, 559.55151044, 519.84140325, 576.39822258, 565.56819335, 584.82157865, 601.66829079, 599.26161763, 601.66829079, 601.66829079, 566.77152993, 558.34817386, 554.73816411, 569.17820309, 577.60155916, 560.75484702, 578.80489574, 583.61824207, 575.19488600, 582.41490549, 586.02491523, 593.24493472, 589.63492498, 584.82157865, 551.12815437, 560.75484702, 548.72148120, 539.09478855, 555.94150069, 536.68811539, 511.41804717, 472.91127656, 442.82786202, 411.54111090, 456.06456442, 435.60784253, 444.03119861, 424.77781330, 442.82786202, 444.03119861, 480.13129605, 489.75798871, 468.09793024, 517.43473008, 493.36799845, 440.42118886, 476.52128631, 483.74130580, 481.33463264, 505.40136427, 487.35131554, 454.86122784, 482.53796922, 496.97800820, 487.35131554, 488.55465212, 481.33463264, 476.52128631, 482.53796922, 523.45141299, 525.85808615, 543.90813488, 555.94150069, 519.84140325, 513.82472034, 494.57133503, 505.40136427, 487.35131554, 488.55465212, 489.75798871, 490.96132529, 499.38468136, 502.99469110, 513.82472034, 502.99469110, 505.40136427, 504.19802769, 499.38468136, 489.75798871, 488.55465212, 480.13129605, 469.30126682, 487.35131554, 470.50460340, 478.92795947, 446.43787177, 442.82786202, 442.82786202, 448.84454493, 450.04788151, 452.45455468, 452.45455468, 441.62452544, 445.23453519, 440.42118886, 483.74130580, 462.08124733, 451.25121810, 451.25121810, 424.77781330, 405.52442800, 376.64435004, 361.00097448, 385.06770611, 359.79763789, 367.01765738, 383.86436953, 385.06770611, 383.86436953, 363.40764764, 359.79763789, 347.76427208, 352.57761841, 336.93424284, 291.20745274, 316.47752096, 286.39410642, 339.34091601, 338.13757943, 328.51088677, 340.54425259, 330.91755994, 336.93424284, 368.22099397, 367.01765738, 358.59430131, 364.61098422, 389.88105243, 356.18762815, 357.39096473, 367.01765738, 364.61098422, 353.78095499, 338.13757943, 352.57761841, 341.74758917, 357.39096473, 369.42433055, 389.88105243, 409.13443774, 391.08438902, 399.50774509, 446.43787177, 511.41804717, 475.31794973, 454.86122784, 456.06456442, 494.57133503, 500.58801794, 478.92795947, 464.48792049, 460.87791075, 451.25121810, 442.82786202, 471.70793998, 505.40136427, 534.28144222, 551.12815437, 528.26475932, 548.72148120, 533.07810564, 540.29812513, 535.48477881, 534.28144222, 528.26475932, 527.06142274, 547.51814462, 560.75484702, 576.39822258, 566.77152993, 558.34817386, 543.90813488, 534.28144222, 548.72148120, 517.43473008, 512.62138376, 515.02805692, 466.89459366, 475.31794973, 487.35131554, 482.53796922, 512.62138376, 505.40136427, 488.55465212, 501.79135452, 486.14797896, 476.52128631, 488.55465212, 484.94464238, 480.13129605, 454.86122784, 415.15112065, 450.04788151, 430.79449621, 409.13443774, 423.57447672, 365.81432080, 404.32109141, 385.06770611, 410.33777432, 392.28772560, 421.16780356, 446.43787177, 438.01451570, 393.49106218, 382.66103294, 417.55779381, 421.16780356, 415.15112065, 442.82786202, 468.09793024, 477.72462289, 494.57133503, 513.82472034, 533.07810564, 527.06142274, 506.60470085, 533.07810564, 540.29812513, 531.87476906, 541.50146171, 518.63806666, 504.19802769, 507.80803743, 516.23139350, 458.47123758, 490.96132529, 488.55465212, 482.53796922, 457.26790100, 457.26790100, 445.23453519, 438.01451570], 
							[1877.20506728, 1925.33853054, 1865.17170146, 1829.07160401, 1817.03823820, 1744.83804330, 1708.73794585, 1780.93814075, 1780.93814075, 1780.93814075, 1805.00487238, 1780.93814075, 1841.10496983, 1853.13833565, 1768.90477493, 1732.80467749, 1684.67121422, 1612.47101933, 1672.63784841, 1672.63784841, 1648.57111678, 1395.87043464, 1287.57014230, 1143.16975251, 1156.40645491, 1275.53677648, 1552.30419025, 1516.20409280, 1528.23745862, 1528.23745862, 1660.60448259, 1528.23745862, 1564.33755606, 1612.47101933, 1576.37092188, 1564.33755606, 1540.27082443, 1468.07062954, 1528.23745862, 1552.30419025, 1492.13736117, 1504.17072698, 1540.27082443, 1504.17072698, 1468.07062954, 1431.97053209, 1383.83706883, 1359.77033719, 1468.07062954, 1492.13736117, 1407.90380046, 1407.90380046, 1407.90380046, 1407.90380046, 1480.10399535, 1528.23745862, 1516.20409280, 1600.43765351, 1612.47101933, 1588.40428770, 1516.20409280, 1624.50438514, 1624.50438514, 1696.70458004, 1744.83804330, 1744.83804330, 1684.67121422, 1660.60448259, 1600.43765351, 1552.30419025, 1552.30419025, 1564.33755606, 1624.50438514, 1564.33755606, 1588.40428770, 1588.40428770, 1648.57111678, 1636.53775096, 1648.57111678, 1684.67121422, 1660.60448259, 1588.40428770, 1588.40428770, 1588.40428770, 1684.67121422, 1696.70458004, 1708.73794585, 1672.63784841, 1624.50438514, 1684.67121422, 1648.57111678, 1564.33755606, 1552.30419025, 1576.37092188, 1612.47101933, 1552.30419025, 1528.23745862, 1588.40428770, 1708.73794585, 1708.73794585, 1696.70458004, 1780.93814075, 1877.20506728, 1805.00487238, 1853.13833565, 1949.40526217, 1949.40526217, 1925.33853054, 1877.20506728, 1829.07160401, 1877.20506728, 1853.13833565, 1841.10496983, 1829.07160401, 1841.10496983, 1853.13833565, 1841.10496983, 1901.27179891, 1949.40526217, 2009.57209125, 2105.83901778, 2045.67218870, 2069.73892033, 2093.80565196, 2141.93911523, 2117.87238359, 2117.87238359, 2105.83901778, 2093.80565196, 2069.73892033, 2069.73892033, 2105.83901778, 2033.63882288, 2033.63882288, 2081.77228615, 2057.70555452, 2105.83901778, 2117.87238359, 2105.83901778, 2093.80565196, 2081.77228615, 2033.63882288, 2021.60545707, 2021.60545707, 2021.60545707, 1985.50535962, 1997.53872544, 2033.63882288, 2021.60545707, 2009.57209125, 1961.43862799, 1817.03823820, 1877.20506728, 1853.13833565, 1841.10496983, 1805.00487238, 1792.97150657, 1792.97150657, 1732.80467749, 1648.57111678, 1480.10399535, 1528.23745862, 1492.13736117, 1516.20409280, 1624.50438514, 1600.43765351, 1624.50438514, 1480.10399535, 1444.00389791, 1419.93716627, 1456.03726372, 1299.60350811, 1287.57014230, 1359.77033719, 1239.43667904, 1419.93716627, 1444.00389791, 1528.23745862, 1552.30419025, 1552.30419025, 1684.67121422, 1708.73794585, 1684.67121422, 1756.87140912, 1756.87140912, 1829.07160401, 1720.77131167, 1708.73794585, 1708.73794585, 1648.57111678, 1720.77131167, 1708.73794585, 1732.80467749, 1696.70458004, 1708.73794585, 1732.80467749, 1708.73794585, 1768.90477493, 1756.87140912, 1768.90477493, 1817.03823820, 1841.10496983, 1792.97150657, 1805.00487238, 1841.10496983, 1877.20506728, 1877.20506728, 1865.17170146, 1853.13833565, 1841.10496983, 1853.13833565, 1817.03823820, 1877.20506728, 1913.30516472, 1961.43862799, 1925.33853054, 1985.50535962, 1973.47199380, 1937.37189636, 1925.33853054, 1949.40526217, 1949.40526217, 1973.47199380, 2009.57209125, 2117.87238359, 2141.93911523, 2105.83901778, 2045.67218870, 2069.73892033, 2081.77228615, 2081.77228615, 2021.60545707, 1997.53872544, 2069.73892033, 2057.70555452, 1973.47199380, 1973.47199380, 1997.53872544, 1997.53872544, 2033.63882288, 1997.53872544, 1997.53872544, 2033.63882288, 1973.47199380, 1985.50535962, 2009.57209125, 1985.50535962, 1937.37189636, 1925.33853054, 1817.03823820, 1805.00487238, 1817.03823820, 1829.07160401, 1817.03823820, 1684.67121422, 1720.77131167, 1744.83804330, 1780.93814075, 1768.90477493, 1817.03823820, 1841.10496983, 1829.07160401, 1780.93814075, 1732.80467749, 1768.90477493, 1805.00487238, 1805.00487238, 1853.13833565, 1913.30516472, 1877.20506728, 2009.57209125, 2033.63882288, 2045.67218870, 2021.60545707, 2033.63882288, 2117.87238359, 2105.83901778, 2081.77228615, 2069.73892033, 2021.60545707, 2009.57209125, 2057.70555452, 2033.63882288, 1961.43862799, 1925.33853054, 1961.43862799, 1865.17170146, 1865.17170146, 1937.37189636, 1925.33853054, 1901.27179891], 
							[571.58487625, 547.51814462, 530.67143248, 543.90813488, 547.51814462, 528.26475932, 476.52128631, 460.87791075, 492.16466187, 465.69125707, 490.96132529, 505.40136427, 519.84140325, 489.75798871, 470.50460340, 439.21785228, 409.13443774, 403.11775483, 429.59115963, 430.79449621, 398.30440851, 421.16780356, 433.20116937, 422.37114014, 383.86436953, 421.16780356, 471.70793998, 516.23139350, 492.16466187, 517.43473008, 515.02805692, 484.94464238, 462.08124733, 498.18134478, 484.94464238, 475.31794973, 481.33463264, 448.84454493, 487.35131554, 498.18134478, 478.92795947, 495.77467161, 476.52128631, 460.87791075, 433.20116937, 438.01451570, 441.62452544, 428.38782305, 427.18448646, 451.25121810, 441.62452544, 472.91127656, 472.91127656, 459.67457417, 484.94464238, 472.91127656, 463.28458391, 465.69125707, 507.80803743, 511.41804717, 505.40136427, 527.06142274, 545.11147146, 527.06142274, 536.68811539, 547.51814462, 547.51814462, 523.45141299, 528.26475932, 511.41804717, 487.35131554, 502.99469110, 490.96132529, 469.30126682, 488.55465212, 493.36799845, 509.01137401, 506.60470085, 522.24807641, 524.65474957, 507.80803743, 511.41804717, 516.23139350, 501.79135452, 462.08124733, 422.37114014, 421.16780356, 397.10107192, 365.81432080, 361.00097448, 348.96760866, 318.88419412, 336.93424284, 326.10421361, 340.54425259, 318.88419412, 332.12089652, 354.98429157, 376.64435004, 376.64435004, 367.01765738, 394.69439876, 388.67771585, 354.98429157, 362.20431106, 367.01765738, 368.22099397, 380.25435978, 374.23767687, 365.81432080, 395.89773534, 405.52442800, 400.71108167, 389.88105243, 394.69439876, 392.28772560, 397.10107192, 416.35445723, 405.52442800, 412.74444748, 441.62452544, 421.16780356, 406.72776458, 421.16780356, 412.74444748, 412.74444748, 413.94778407, 398.30440851, 388.67771585, 386.27104269, 391.08438902, 391.08438902, 374.23767687, 377.84768662, 373.03434029, 361.00097448, 361.00097448, 367.01765738, 371.83100371, 367.01765738, 371.83100371, 353.78095499, 347.76427208, 338.13757943, 341.74758917, 336.93424284, 334.52756968, 358.59430131, 358.59430131, 359.79763789, 358.59430131, 350.17094524, 356.18762815, 357.39096473, 370.62766713, 358.59430131, 326.10421361, 322.49420387, 308.05416489, 306.85082830, 305.64749172, 292.41078933, 294.81746249, 300.83414540, 318.88419412, 312.86751121, 321.29086728, 304.44415514, 290.00411616, 282.78409667, 280.37742351, 270.75073086, 233.44729683, 252.70068213, 231.04062366, 275.56407718, 276.76741377, 268.34405769, 293.61412591, 296.02079907, 318.88419412, 323.69754045, 314.07084779, 305.64749172, 310.46083805, 303.24081856, 293.61412591, 285.19076984, 310.46083805, 303.24081856, 303.24081856, 282.78409667, 294.81746249, 312.86751121, 358.59430131, 350.17094524, 344.15426233, 382.66103294, 383.86436953, 411.54111090, 413.94778407, 412.74444748, 409.13443774, 404.32109141, 403.11775483, 416.35445723, 431.99783279, 406.72776458, 403.11775483, 403.11775483, 419.96446697, 409.13443774, 416.35445723, 439.21785228, 445.23453519, 446.43787177, 456.06456442, 459.67457417, 445.23453519, 439.21785228, 450.04788151, 451.25121810, 439.21785228, 431.99783279, 439.21785228, 445.23453519, 439.21785228, 428.38782305, 413.94778407, 415.15112065, 400.71108167, 394.69439876, 376.64435004, 389.88105243, 407.93110116, 389.88105243, 393.49106218, 406.72776458, 407.93110116, 401.91441825, 393.49106218, 392.28772560, 407.93110116, 379.05102320, 388.67771585, 409.13443774, 399.50774509, 413.94778407, 416.35445723, 403.11775483, 415.15112065, 391.08438902, 380.25435978, 379.05102320, 365.81432080, 376.64435004, 379.05102320, 415.15112065, 412.74444748, 463.28458391, 459.67457417, 446.43787177, 434.40450595, 419.96446697, 439.21785228, 442.82786202, 440.42118886, 448.84454493, 463.28458391, 453.65789126, 459.67457417, 448.84454493, 471.70793998, 459.67457417, 463.28458391, 489.75798871, 477.72462289, 470.50460340, 476.52128631, 469.30126682, 456.06456442, 441.62452544, 456.06456442, 425.98114988, 407.93110116, 397.10107192, 385.06770611, 427.18448646, 415.15112065, 441.62452544, 434.40450595], 
							[671.46181253, 629.34503217, 607.68497370, 589.63492498, 583.61824207, 572.78821284, 523.45141299, 480.13129605, 493.36799845, 476.52128631, 500.58801794, 530.67143248, 527.06142274, 534.28144222, 515.02805692, 495.77467161, 459.67457417, 433.20116937, 492.16466187, 470.50460340, 504.19802769, 458.47123758, 484.94464238, 439.21785228, 404.32109141, 453.65789126, 493.36799845, 513.82472034, 493.36799845, 529.46809590, 595.65160789, 540.29812513, 510.21471059, 516.23139350, 478.92795947, 478.92795947, 502.99469110, 481.33463264, 507.80803743, 506.60470085, 505.40136427, 478.92795947, 492.16466187, 475.31794973, 445.23453519, 436.81117912, 453.65789126, 430.79449621, 457.26790100, 489.75798871, 464.48792049, 472.91127656, 492.16466187, 471.70793998, 515.02805692, 507.80803743, 496.97800820, 522.24807641, 518.63806666, 561.95818360, 566.77152993, 587.22825181, 592.04159814, 599.26161763, 612.49832003, 649.80175406, 651.00509064, 631.75170533, 614.90499319, 567.97486651, 572.78821284, 598.05828105, 586.02491523, 565.56819335, 577.60155916, 582.41490549, 600.46495421, 576.39822258, 598.05828105, 601.66829079, 595.65160789, 610.09164687, 590.83826156, 606.48163712, 571.58487625, 530.67143248, 500.58801794, 481.33463264, 421.16780356, 423.57447672, 393.49106218, 368.22099397, 388.67771585, 381.45769636, 385.06770611, 387.47437927, 392.28772560, 418.76113039, 500.58801794, 487.35131554, 448.84454493, 509.01137401, 486.14797896, 438.01451570, 458.47123758, 457.26790100, 468.09793024, 501.79135452, 465.69125707, 451.25121810, 477.72462289, 534.28144222, 535.48477881, 558.34817386, 553.53482753, 552.33149095, 539.09478855, 540.29812513, 546.31480804, 563.16152018, 547.51814462, 527.06142274, 507.80803743, 517.43473008, 505.40136427, 499.38468136, 512.62138376, 499.38468136, 516.23139350, 502.99469110, 502.99469110, 516.23139350, 492.16466187, 465.69125707, 458.47123758, 475.31794973, 472.91127656, 500.58801794, 512.62138376, 516.23139350, 531.87476906, 488.55465212, 478.92795947, 465.69125707, 470.50460340, 475.31794973, 458.47123758, 483.74130580, 472.91127656, 487.35131554, 466.89459366, 476.52128631, 447.64120835, 445.23453519, 446.43787177, 444.03119861, 428.38782305, 411.54111090, 387.47437927, 397.10107192, 381.45769636, 369.42433055, 380.25435978, 404.32109141, 425.98114988, 422.37114014, 422.37114014, 406.72776458, 385.06770611, 362.20431106, 403.11775483, 367.01765738, 321.29086728, 348.96760866, 312.86751121, 370.62766713, 385.06770611, 409.13443774, 395.89773534, 409.13443774, 397.10107192, 488.55465212, 448.84454493, 451.25121810, 458.47123758, 474.11461315, 464.48792049, 435.60784253, 448.84454493, 434.40450595, 440.42118886, 398.30440851, 433.20116937, 434.40450595, 442.82786202, 431.99783279, 454.86122784, 459.67457417, 458.47123758, 464.48792049, 470.50460340, 478.92795947, 490.96132529, 484.94464238, 492.16466187, 536.68811539, 516.23139350, 493.36799845, 483.74130580, 500.58801794, 563.16152018, 598.05828105, 559.55151044, 583.61824207, 607.68497370, 608.88831028, 612.49832003, 598.05828105, 607.68497370, 588.43158840, 578.80489574, 583.61824207, 557.14483728, 564.36485676, 572.78821284, 671.46181253, 706.35857339, 658.22511013, 631.75170533, 629.34503217, 601.66829079, 569.17820309, 541.50146171, 545.11147146, 572.78821284, 524.65474957, 515.02805692, 549.92481779, 563.16152018, 560.75484702, 552.33149095, 547.51814462, 576.39822258, 529.46809590, 527.06142274, 555.94150069, 540.29812513, 536.68811539, 515.02805692, 459.67457417, 463.28458391, 421.16780356, 382.66103294, 386.27104269, 334.52756968, 369.42433055, 369.42433055, 367.01765738, 367.01765738, 368.22099397, 374.23767687, 373.03434029, 364.61098422, 352.57761841, 367.01765738, 344.15426233, 333.32423310, 321.29086728, 354.98429157, 339.34091601, 354.98429157, 379.05102320, 397.10107192, 382.66103294, 393.49106218, 412.74444748, 400.71108167, 382.66103294, 380.25435978, 362.20431106, 353.78095499, 354.98429157, 347.76427208, 300.83414540, 306.85082830, 271.95406744, 259.92070162, 259.92070162, 267.14072111, 253.90401872, 247.88733581], 
							[961.46592869, 951.83923604, 969.88928476, 955.44924578, 930.17917757, 877.23236798, 758.10204640, 715.98526604, 703.95190023, 700.34189048, 790.59213410, 819.47221206, 794.20214385, 823.08222181, 772.54208538, 750.88202691, 719.59527579, 685.90185151, 734.03531477, 720.79861237, 717.18860263, 653.41176380, 682.29184176, 624.53168584, 582.41490549, 619.71833952, 746.06868058, 770.13541222, 759.30538298, 779.76210487, 788.18546094, 723.20528553, 707.56190997, 724.40862212, 709.96858314, 684.69851492, 718.39193921, 648.59841748, 664.24179304, 709.96858314, 703.95190023, 679.88516860, 683.49517834, 690.71519783, 667.85180278, 634.15837850, 643.78507115, 625.73502243, 657.02177355, 689.51186125, 664.24179304, 685.90185151, 682.29184176, 683.49517834, 722.00194895, 746.06868058, 750.88202691, 697.93521732, 738.84866109, 784.57545120, 746.06868058, 794.20214385, 830.30224130, 872.41902165, 879.63904114, 902.50243619, 924.16249466, 930.17917757, 865.19900216, 841.13227053, 866.40233874, 868.80901191, 871.21568507, 868.80901191, 883.24905089, 878.43570456, 873.62235823, 849.55562660, 854.36897293, 874.82569481, 876.02903140, 837.52226079, 844.74228027, 818.26887548, 803.82883650, 759.30538298, 753.28870007, 684.69851492, 622.12501268, 638.97172482, 625.73502243, 554.73816411, 607.68497370, 578.80489574, 648.59841748, 601.66829079, 595.65160789, 625.73502243, 669.05513936, 673.86848569, 653.41176380, 722.00194895, 709.96858314, 653.41176380, 685.90185151, 724.40862212, 764.11872931, 768.93207563, 788.18546094, 747.27201717, 743.66200742, 752.08536349, 748.47535375, 750.88202691, 779.76210487, 817.06553890, 825.48889497, 844.74228027, 842.33560711, 856.77564609, 874.82569481, 827.89556813, 815.86220232, 818.26887548, 827.89556813, 808.64218283, 830.30224130, 827.89556813, 814.65886573, 807.43884625, 806.23550966, 813.45552915, 785.77878778, 767.72873905, 773.74542196, 758.10204640, 782.16877803, 783.37211461, 766.52540247, 790.59213410, 796.60881701, 791.79547068, 782.16877803, 750.88202691, 742.45867084, 749.67869033, 759.30538298, 770.13541222, 770.13541222, 782.16877803, 801.42216334, 788.18546094, 685.90185151, 741.25533426, 718.39193921, 762.91539273, 731.62864161, 681.08850518, 625.73502243, 644.98840773, 637.76838824, 657.02177355, 631.75170533, 676.27515885, 741.25533426, 718.39193921, 728.01863186, 703.95190023, 659.42844671, 625.73502243, 602.87162738, 589.63492498, 551.12815437, 583.61824207, 548.72148120, 631.75170533, 613.70165661, 640.17506140, 648.59841748, 631.75170533, 634.15837850, 658.22511013, 673.86848569, 671.46181253, 637.76838824, 661.83511987, 608.88831028, 601.66829079, 649.80175406, 638.97172482, 661.83511987, 613.70165661, 641.37839799, 622.12501268, 643.78507115, 651.00509064, 658.22511013, 653.41176380, 699.13855390, 671.46181253, 675.07182227, 691.91853441, 746.06868058, 766.52540247, 738.84866109, 750.88202691, 777.35543171, 794.20214385, 753.28870007, 758.10204640, 765.32206589, 761.71205615, 758.10204640, 785.77878778, 788.18546094, 770.13541222, 780.96544145, 782.16877803, 767.72873905, 791.79547068, 794.20214385, 786.98212436, 800.21882676, 814.65886573, 813.45552915, 854.36897293, 821.87888522, 824.28555839, 782.16877803, 803.82883650, 776.15209512, 766.52540247, 741.25533426, 747.27201717, 762.91539273, 726.81529528, 699.13855390, 752.08536349, 740.05199768, 759.30538298, 741.25533426, 743.66200742, 784.57545120, 749.67869033, 737.64532451, 782.16877803, 756.89870982, 762.91539273, 709.96858314, 644.98840773, 655.81843697, 612.49832003, 604.07496396, 604.07496396, 527.06142274, 541.50146171, 561.95818360, 624.53168584, 599.26161763, 589.63492498, 624.53168584, 598.05828105, 564.36485676, 536.68811539, 582.41490549, 524.65474957, 486.14797896, 494.57133503, 529.46809590, 498.18134478, 541.50146171, 581.21156891, 607.68497370, 589.63492498, 587.22825181, 606.48163712, 602.87162738, 575.19488600, 589.63492498, 548.72148120, 523.45141299, 511.41804717, 482.53796922, 457.26790100, 469.30126682, 427.18448646, 425.98114988, 434.40450595, 429.59115963, 388.67771585, 351.37428182], 
							[2166.00584686, 2178.03921267, 2178.03921267, 2214.13931012, 2178.03921267, 2153.97248104, 2117.87238359, 2081.77228615, 2033.63882288, 2129.90574941, 2202.10594431, 2153.97248104, 2166.00584686, 2190.07257849, 2190.07257849, 2129.90574941, 2057.70555452, 2009.57209125, 1985.50535962, 1877.20506728, 1853.13833565, 1756.87140912, 1696.70458004, 1648.57111678, 1612.47101933, 1817.03823820, 2093.80565196, 1985.50535962, 2009.57209125, 2033.63882288, 2057.70555452, 2033.63882288, 1985.50535962, 2033.63882288, 1973.47199380, 1949.40526217, 1841.10496983, 1768.90477493, 1636.53775096, 1708.73794585, 1684.67121422, 1720.77131167, 1732.80467749, 1708.73794585, 1684.67121422, 1648.57111678, 1624.50438514, 1588.40428770, 1600.43765351, 1588.40428770, 1564.33755606, 1648.57111678, 1612.47101933, 1600.43765351, 1768.90477493, 1805.00487238, 1829.07160401, 1829.07160401, 1937.37189636, 1997.53872544, 1949.40526217, 2045.67218870, 2045.67218870, 1937.37189636, 1901.27179891, 1865.17170146, 1889.23843309, 1877.20506728, 1792.97150657, 1805.00487238, 1780.93814075, 1756.87140912, 1853.13833565, 1853.13833565, 1865.17170146, 1877.20506728, 2021.60545707, 2009.57209125, 2009.57209125, 2069.73892033, 2129.90574941, 2093.80565196, 2045.67218870, 1973.47199380, 1997.53872544, 1985.50535962, 1937.37189636, 1925.33853054, 1889.23843309, 1865.17170146, 1853.13833565, 1805.00487238, 1817.03823820, 1817.03823820, 1913.30516472, 1805.00487238, 1805.00487238, 1853.13833565, 1973.47199380, 1985.50535962, 1925.33853054, 1997.53872544, 1997.53872544, 1925.33853054, 2009.57209125, 1985.50535962, 1973.47199380, 1949.40526217, 1901.27179891, 1841.10496983, 1913.30516472, 1913.30516472, 1865.17170146, 1841.10496983, 1853.13833565, 1841.10496983, 1817.03823820, 1829.07160401, 1901.27179891, 1877.20506728, 1853.13833565, 1829.07160401, 1829.07160401, 1805.00487238, 1792.97150657, 1780.93814075, 1768.90477493, 1780.93814075, 1780.93814075, 1780.93814075, 1841.10496983, 1829.07160401, 1780.93814075, 1780.93814075, 1732.80467749, 1672.63784841, 1672.63784841, 1648.57111678, 1648.57111678, 1624.50438514, 1600.43765351, 1600.43765351, 1612.47101933, 1612.47101933, 1636.53775096, 1672.63784841, 1660.60448259, 1696.70458004, 1696.70458004, 1708.73794585, 1684.67121422, 1780.93814075, 1756.87140912, 1792.97150657, 1780.93814075, 1780.93814075, 1732.80467749, 1696.70458004, 1672.63784841, 1672.63784841, 1732.80467749, 1708.73794585, 1660.60448259, 1672.63784841, 1696.70458004, 1684.67121422, 1708.73794585, 1636.53775096, 1612.47101933, 1600.43765351, 1600.43765351, 1564.33755606, 1564.33755606, 1564.33755606, 1564.33755606, 1636.53775096, 1684.67121422, 1708.73794585, 1696.70458004, 1720.77131167, 1696.70458004, 1805.00487238, 1792.97150657, 1829.07160401, 1853.13833565, 1949.40526217, 1889.23843309, 1877.20506728, 1925.33853054, 1877.20506728, 1853.13833565, 1792.97150657, 1853.13833565, 1817.03823820, 1817.03823820, 1805.00487238, 1841.10496983, 1865.17170146, 1949.40526217, 2057.70555452, 2045.67218870, 2081.77228615, 2057.70555452, 2045.67218870, 2057.70555452, 2021.60545707, 2021.60545707, 2033.63882288, 1985.50535962, 1925.33853054, 1913.30516472, 1937.37189636, 1973.47199380, 2021.60545707, 2021.60545707, 1961.43862799, 1985.50535962, 1985.50535962, 1973.47199380, 1997.53872544, 1997.53872544, 1997.53872544, 1985.50535962, 1985.50535962, 2009.57209125, 2033.63882288, 2045.67218870, 2045.67218870, 2009.57209125, 1997.53872544, 1937.37189636, 1937.37189636, 1925.33853054, 1901.27179891, 1889.23843309, 1865.17170146, 1841.10496983, 1853.13833565, 1877.20506728, 1913.30516472, 1925.33853054, 1901.27179891, 1961.43862799, 1889.23843309, 1889.23843309, 1949.40526217, 1937.37189636, 1913.30516472, 1901.27179891, 1817.03823820, 1768.90477493, 1792.97150657, 1780.93814075, 1792.97150657, 1696.70458004, 1696.70458004, 1672.63784841, 1708.73794585, 1660.60448259, 1636.53775096, 1660.60448259, 1612.47101933, 1588.40428770, 1648.57111678, 1684.67121422, 1708.73794585, 1732.80467749, 1660.60448259, 1684.67121422, 1720.77131167, 1576.37092188, 1636.53775096, 1672.63784841, 1684.67121422, 1648.57111678, 1648.57111678, 1648.57111678, 1564.33755606, 1684.67121422, 1624.50438514, 1636.53775096, 1600.43765351, 1696.70458004, 1576.37092188, 1612.47101933, 1684.67121422, 1684.67121422, 1672.63784841, 1672.63784841, 1672.63784841, 1624.50438514], 
							[660.63178329, 649.80175406, 625.73502243, 687.10518809, 673.86848569, 596.85494447, 665.44512962, 655.81843697, 691.91853441, 675.07182227, 715.98526604, 682.29184176, 705.15523681, 701.54522707, 697.93521732, 728.01863186, 700.34189048, 702.74856365, 728.01863186, 718.39193921, 740.05199768, 709.96858314, 729.22196844, 691.91853441, 625.73502243, 684.69851492, 746.06868058, 724.40862212, 724.40862212, 722.00194895, 744.86534400, 705.15523681, 673.86848569, 673.86848569, 685.90185151, 657.02177355, 638.97172482, 608.88831028, 649.80175406, 649.80175406, 630.54836875, 637.76838824, 641.37839799, 632.95504192, 629.34503217, 663.03845645, 657.02177355, 652.20842722, 665.44512962, 677.47849543, 661.83511987, 651.00509064, 636.56505166, 658.22511013, 676.27515885, 673.86848569, 731.62864161, 770.13541222, 764.11872931, 774.94875854, 748.47535375, 797.81215359, 779.76210487, 803.82883650, 794.20214385, 789.38879752, 772.54208538, 762.91539273, 711.17191972, 713.57859288, 732.83197819, 737.64532451, 741.25533426, 697.93521732, 728.01863186, 718.39193921, 707.56190997, 702.74856365, 715.98526604, 695.52854416, 691.91853441, 672.66514911, 643.78507115, 637.76838824, 670.25847594, 634.15837850, 631.75170533, 612.49832003, 572.78821284, 542.70479830, 537.89145197, 476.52128631, 531.87476906, 515.02805692, 569.17820309, 543.90813488, 563.16152018, 612.49832003, 629.34503217, 642.58173457, 614.90499319, 643.78507115, 623.32834926, 605.27830054, 636.56505166, 634.15837850, 626.93835901, 649.80175406, 685.90185151, 654.61510038, 664.24179304, 676.27515885, 664.24179304, 652.20842722, 665.44512962, 679.88516860, 705.15523681, 709.96858314, 728.01863186, 749.67869033, 754.49203666, 732.83197819, 690.71519783, 693.12187099, 696.73188074, 694.32520758, 693.12187099, 697.93521732, 697.93521732, 701.54522707, 683.49517834, 691.91853441, 678.68183202, 671.46181253, 685.90185151, 677.47849543, 678.68183202, 690.71519783, 683.49517834, 691.91853441, 690.71519783, 676.27515885, 685.90185151, 657.02177355, 665.44512962, 659.42844671, 659.42844671, 665.44512962, 691.91853441, 671.46181253, 661.83511987, 643.78507115, 611.29498345, 594.44827130, 588.43158840, 578.80489574, 576.39822258, 569.17820309, 567.97486651, 523.45141299, 489.75798871, 517.43473008, 528.26475932, 524.65474957, 547.51814462, 518.63806666, 531.87476906, 512.62138376, 493.36799845, 483.74130580, 498.18134478, 476.52128631, 431.99783279, 452.45455468, 400.71108167, 462.08124733, 496.97800820, 506.60470085, 515.02805692, 559.55151044, 523.45141299, 570.38153967, 581.21156891, 578.80489574, 567.97486651, 593.24493472, 545.11147146, 537.89145197, 554.73816411, 548.72148120, 565.56819335, 553.53482753, 573.99154942, 552.33149095, 576.39822258, 581.21156891, 583.61824207, 583.61824207, 589.63492498, 611.29498345, 599.26161763, 619.71833952, 612.49832003, 595.65160789, 626.93835901, 642.58173457, 661.83511987, 637.76838824, 636.56505166, 631.75170533, 625.73502243, 623.32834926, 643.78507115, 644.98840773, 660.63178329, 678.68183202, 665.44512962, 675.07182227, 660.63178329, 648.59841748, 661.83511987, 629.34503217, 657.02177355, 638.97172482, 648.59841748, 634.15837850, 636.56505166, 626.93835901, 629.34503217, 626.93835901, 613.70165661, 588.43158840, 571.58487625, 581.21156891, 589.63492498, 576.39822258, 571.58487625, 572.78821284, 605.27830054, 616.10832977, 595.65160789, 581.21156891, 598.05828105, 589.63492498, 592.04159814, 604.07496396, 596.85494447, 598.05828105, 587.22825181, 559.55151044, 565.56819335, 543.90813488, 529.46809590, 529.46809590, 499.38468136, 502.99469110, 484.94464238, 502.99469110, 493.36799845, 525.85808615, 535.48477881, 509.01137401, 498.18134478, 478.92795947, 496.97800820, 460.87791075, 468.09793024, 465.69125707, 499.38468136, 499.38468136, 533.07810564, 513.82472034, 519.84140325, 517.43473008, 528.26475932, 552.33149095, 548.72148120, 541.50146171, 537.89145197, 529.46809590, 515.02805692, 518.63806666, 528.26475932, 499.38468136, 492.16466187, 494.57133503, 469.30126682, 466.89459366, 475.31794973, 466.89459366, 472.91127656], 
							[920.55248491, 861.58899242, 866.40233874, 889.26573379, 789.38879752, 741.25533426, 696.73188074, 607.68497370, 554.73816411, 558.34817386, 628.14169559, 598.05828105, 637.76838824, 610.09164687, 586.02491523, 569.17820309, 546.31480804, 528.26475932, 553.53482753, 557.14483728, 554.73816411, 482.53796922, 529.46809590, 524.65474957, 517.43473008, 552.33149095, 602.87162738, 617.31166635, 632.95504192, 643.78507115, 654.61510038, 602.87162738, 569.17820309, 587.22825181, 600.46495421, 573.99154942, 607.68497370, 577.60155916, 616.10832977, 637.76838824, 644.98840773, 613.70165661, 607.68497370, 586.02491523, 573.99154942, 545.11147146, 555.94150069, 539.09478855, 592.04159814, 619.71833952, 612.49832003, 630.54836875, 659.42844671, 638.97172482, 697.93521732, 731.62864161, 722.00194895, 644.98840773, 649.80175406, 734.03531477, 722.00194895, 746.06868058, 734.03531477, 708.76524656, 701.54522707, 731.62864161, 761.71205615, 713.57859288, 687.10518809, 673.86848569, 679.88516860, 694.32520758, 706.35857339, 679.88516860, 708.76524656, 737.64532451, 731.62864161, 746.06868058, 750.88202691, 722.00194895, 737.64532451, 711.17191972, 697.93521732, 703.95190023, 664.24179304, 658.22511013, 649.80175406, 583.61824207, 533.07810564, 513.82472034, 507.80803743, 440.42118886, 478.92795947, 463.28458391, 505.40136427, 477.72462289, 480.13129605, 511.41804717, 569.17820309, 555.94150069, 535.48477881, 661.83511987, 619.71833952, 551.12815437, 606.48163712, 601.66829079, 628.14169559, 649.80175406, 628.14169559, 606.48163712, 590.83826156, 631.75170533, 625.73502243, 610.09164687, 623.32834926, 631.75170533, 637.76838824, 649.80175406, 636.56505166, 675.07182227, 679.88516860, 626.93835901, 629.34503217, 613.70165661, 636.56505166, 632.95504192, 635.36171508, 617.31166635, 640.17506140, 620.92167610, 641.37839799, 643.78507115, 612.49832003, 611.29498345, 601.66829079, 577.60155916, 584.82157865, 604.07496396, 593.24493472, 610.09164687, 602.87162738, 600.46495421, 601.66829079, 564.36485676, 541.50146171, 586.02491523, 600.46495421, 636.56505166, 604.07496396, 612.49832003, 634.15837850, 631.75170533, 561.95818360, 625.73502243, 616.10832977, 618.51500294, 590.83826156, 553.53482753, 505.40136427, 549.92481779, 548.72148120, 511.41804717, 505.40136427, 502.99469110, 548.72148120, 561.95818360, 576.39822258, 546.31480804, 523.45141299, 506.60470085, 517.43473008, 487.35131554, 415.15112065, 463.28458391, 410.33777432, 516.23139350, 607.68497370, 583.61824207, 595.65160789, 559.55151044, 546.31480804, 601.66829079, 578.80489574, 595.65160789, 623.32834926, 628.14169559, 602.87162738, 590.83826156, 618.51500294, 592.04159814, 664.24179304, 631.75170533, 665.44512962, 677.47849543, 722.00194895, 685.90185151, 697.93521732, 724.40862212, 718.39193921, 714.78192946, 715.98526604, 737.64532451, 772.54208538, 760.50871956, 776.15209512, 770.13541222, 776.15209512, 758.10204640, 743.66200742, 722.00194895, 770.13541222, 730.42530502, 776.15209512, 776.15209512, 803.82883650, 771.33874880, 796.60881701, 812.25219257, 773.74542196, 760.50871956, 807.43884625, 824.28555839, 814.65886573, 809.84551941, 833.91225104, 848.35229002, 871.21568507, 845.94561686, 838.72559737, 851.96229976, 832.70891446, 772.54208538, 806.23550966, 821.87888522, 863.99566558, 801.42216334, 802.62549992, 867.60567532, 919.34914833, 954.24590920, 942.21254338, 953.04257262, 997.56602614, 953.04257262, 926.56916782, 1020.42942119, 1005.98938221, 1044.49615282, 1089.01960634, 980.71931399, 999.97269930, 980.71931399, 883.24905089, 885.65572405, 658.22511013, 712.37525630, 697.93521732, 746.06868058, 690.71519783, 715.98526604, 730.42530502, 730.42530502, 697.93521732, 657.02177355, 701.54522707, 652.20842722, 619.71833952, 616.10832977, 643.78507115, 619.71833952, 671.46181253, 693.12187099, 755.69537324, 714.78192946, 758.10204640, 778.55876829, 786.98212436, 768.93207563, 767.72873905, 718.39193921, 722.00194895, 748.47535375, 737.64532451, 654.61510038, 647.39508089, 601.66829079, 571.58487625, 575.19488600, 566.77152993, 558.34817386, 535.48477881], 
							[928.97584099, 938.60253364, 920.55248491, 914.53580201, 914.53580201, 896.48575328, 788.18546094, 729.22196844, 754.49203666, 699.13855390, 736.44198793, 756.89870982, 782.16877803, 836.31892420, 801.42216334, 749.67869033, 690.71519783, 685.90185151, 722.00194895, 740.05199768, 728.01863186, 661.83511987, 713.57859288, 713.57859288, 673.86848569, 722.00194895, 818.26887548, 850.75896318, 937.39919706, 974.70263109, 989.14267007, 879.63904114, 830.30224130, 880.84237772, 868.80901191, 831.50557788, 844.74228027, 836.31892420, 900.09576303, 932.58585073, 884.45238747, 908.51911910, 986.73599690, 992.75267981, 941.00920680, 926.56916782, 927.77250440, 897.68908986, 950.63589945, 955.44924578, 991.54934323, 974.70263109, 1009.59939195, 1085.40959659, 1143.16975251, 1200.92990842, 1191.30321577, 1155.20311832, 1123.91636720, 1141.96641593, 1167.23648414, 1239.43667904, 1198.52323526, 1198.52323526, 1173.25316705, 1179.26984996, 1178.06651337, 1140.76307935, 1105.86631848, 1089.01960634, 1101.05297215, 1116.69634771, 1126.32304037, 1099.84963557, 1137.15306960, 1122.71303062, 1153.99978174, 1179.26984996, 1172.04983047, 1197.31989868, 1187.69320603, 1185.28653286, 1146.77976225, 1150.38977200, 1114.28967455, 1089.01960634, 1063.74953812, 1034.86946017, 938.60253364, 944.61921655, 926.56916782, 806.23550966, 865.19900216, 806.23550966, 865.19900216, 800.21882676, 824.28555839, 871.21568507, 932.58585073, 949.43256287, 956.65258236, 1022.83609435, 1022.83609435, 938.60253364, 1009.59939195, 1016.81941144, 1049.30949914, 1084.20626001, 1061.34286496, 968.68594818, 1052.91950889, 1052.91950889, 1022.83609435, 1008.39605537, 1004.78604563, 1052.91950889, 1070.96955761, 1079.39291368, 1057.73285522, 1061.34286496, 1102.25630873, 1018.02274802, 1007.19271879, 1004.78604563, 1039.68280649, 1064.95287471, 1055.32618205, 1075.78290394, 1084.20626001, 1083.00292343, 1075.78290394, 1089.01960634, 1083.00292343, 1083.00292343, 1084.20626001, 1085.40959659, 1058.93619180, 1036.07279675, 1048.10616256, 1046.90282598, 1042.08947966, 1036.07279675, 1038.47946991, 1009.59939195, 1025.24276751, 1031.25945042, 1052.91950889, 1061.34286496, 1063.74953812, 1076.98624052, 1087.81626976, 1078.18957710, 993.95601639, 975.90596767, 943.41587996, 949.43256287, 943.41587996, 897.68908986, 849.55562660, 892.87574354, 872.41902165, 855.57230951, 849.55562660, 863.99566558, 938.60253364, 967.48261160, 983.12598716, 953.04257262, 948.22922629, 922.95915808, 932.58585073, 886.85906063, 902.50243619, 936.19586048, 849.55562660, 948.22922629, 949.43256287, 937.39919706, 943.41587996, 930.17917757, 926.56916782, 993.95601639, 1002.37937246, 995.15935297, 1014.41273828, 1040.88614307, 1009.59939195, 991.54934323, 1033.66612358, 1013.20940170, 1046.90282598, 1074.57956736, 1091.42627950, 1095.03628924, 1070.96955761, 1066.15621129, 1119.10302088, 1125.11970378, 1101.05297215, 1114.28967455, 1095.03628924, 1128.72971353, 1129.93305011, 1147.98309883, 1173.25316705, 1197.31989868, 1203.33658159, 1215.36994740, 1191.30321577, 1193.70988894, 1227.40331322, 1227.40331322, 1299.60350811, 1251.47004485, 1203.33658159, 1215.36994740, 1239.43667904, 1251.47004485, 1197.31989868, 1192.50655235, 1227.40331322, 1163.62647440, 1125.11970378, 1117.89968430, 1096.23962583, 1117.89968430, 1113.08633797, 1093.83295266, 1092.62961608, 1081.79958685, 1040.88614307, 1031.25945042, 1021.63275777, 1025.24276751, 1066.15621129, 998.76936272, 1003.58270904, 1037.27613333, 1046.90282598, 1033.66612358, 1016.81941144, 1015.61607486, 1022.83609435, 995.15935297, 967.48261160, 1007.19271879, 959.05925553, 961.46592869, 902.50243619, 890.46907037, 835.11558762, 800.21882676, 791.79547068, 807.43884625, 640.17506140, 706.35857339, 709.96858314, 740.05199768, 742.45867084, 760.50871956, 776.15209512, 722.00194895, 689.51186125, 660.63178329, 696.73188074, 658.22511013, 600.46495421, 595.65160789, 658.22511013, 661.83511987, 702.74856365, 738.84866109, 784.57545120, 771.33874880, 768.93207563, 770.13541222, 823.08222181, 843.53894369, 856.77564609, 823.08222181, 826.69223155, 827.89556813, 833.91225104, 797.81215359, 791.79547068, 767.72873905, 701.54522707, 743.66200742, 749.67869033, 747.27201717, 767.72873905], 
							];
		var assetsReturns = new Array(assetsPrices.length);
		for (var k = 0; k < assetsPrices.length; ++k) {
			assetsReturns[k] = new Array(assetsPrices[k].length - 1);
			for (var i = 1; i < assetsPrices[k].length; ++i) {
				assetsReturns[k][i-1] = (assetsPrices[k][i] - assetsPrices[k][i-1])/assetsPrices[k][i-1];
			}
		}
		
		// Construct minimum weights constraints
		var lowerBounds = new Array(assetsPrices.length);
		for (var k = 0; k < assetsPrices.length; ++k) {
			lowerBounds[k] = 0.01;
		}
		
		// Compute the weights of the minimum tracking error portfolio
		var weights = PortfolioAllocation.minimumTrackingErrorWeights(assetsReturns, indexReturns, {constraints: { minWeights: lowerBounds, maxNbAssets: 10 }});
		
		// Check that the number of non-zero weights is lower than 10, and that the minimum weights matches with the lower bounds
		var nbNonZeroAssets = 0;
		var lowerBoundsOk = true;
		for (var k = 0; k < weights.length; ++k) {
			if (weights[k] != 0) {
				nbNonZeroAssets = nbNonZeroAssets + 1;
				
				if (weights[k] < lowerBounds[k]) {
					lowerBoundsOk = false;
				}
			}
		}
		assert.equal(nbNonZeroAssets <= 10, true, "Minimum tracking error portfolio, indtrack2 OR data set ((Nikkei)), number of assets");
		assert.equal(lowerBoundsOk, true, "Minimum tracking error portfolio, indtrack2 OR data set ((Nikkei)), minimum bounds");

	}
});